
Require Import Tactics.
Require Import Syntax.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Breduce.
Require Import Determinism.
Require Import Standard.  (* shadows Determinism.halt *)
Require Preservation.
Require Progress.
Require Wf_nat.


(* Embedding the language into the CBV untyped lambda calculus. *)

Definition lm_i := lm_lam (var 0).


Definition lm_let (* lm *) m1 (* lm => lm *) m2 :=
  lm_app (lm_lam m2) m1.


Lemma val_i :
  val lm_i.
Proof.
apply val_lam.
Qed.


Lemma reduce_i :
  reduce lm_i lm_i.
Proof.
unfold lm_i.
eauto with lambda.
Qed.


Definition step_let {m1 m1' m2}
  (H : step m1 m1')
  : step (lm_let m1 m2) (lm_let m1' m2)
  :=
  step_app2 val_lam H.


Definition step_let_beta {m1 m2}
  (H : val m1)
  : step (lm_let m1 m2) (subst1 m1 m2)
  :=
  step_beta H.


Definition reduce_let {m1 m1' m2 m2'}
  (H1 : reduce m1 m1') (H2 : reduce m2 m2')
  : reduce (lm_let m1 m2) (lm_let m1' m2')
  :=
  reduce_app (reduce_lam H2) H1.


Definition reduce_let_beta {m1 m1' m2 m2'}
  (Hv : val m1) (H1 : reduce m1 m1') (H2 : reduce m2 m2')
  : reduce (lm_let m1 m2) (subst1 m1' m2')
  :=
  reduce_beta Hv H2 H1.


Hint Resolve val_i step_let step_let_beta reduce_let reduce_let_beta : lambda.


Definition lm_theta :=
  lm_app
  (lm_lam (lm_lam (lm_app (var 0) (lm_lam (lm_app (lm_app (var 2) (var 2)) (var 1))))))
  (lm_lam (lm_lam (lm_app (var 0) (lm_lam (lm_app (lm_app (var 2) (var 2)) (var 1)))))).


Lemma step_theta :
  forall m,
    val m
    -> plus step (lm_app lm_theta m) (lm_app m (lm_lam (lm_app lm_theta (shift1 m)))).
Proof.
intro m.
unfold lm_theta.
eexists; split.
  eauto with lambda; done.

  simp_sub.
  #2 eapply star_step.
    eauto with lambda; done.
    
    simp_sub.
    apply star_refl.
Qed.


Lemma subst_theta :
  forall s, subst s lm_theta = lm_theta.
Proof.
intro s.
unfold lm_theta.
simp_sub.
reflexivity.
Qed.


Ltac simp_sub_lm :=
  unfold lm_let, lm_i;
  simp_sub;
  repeat (match goal with |- appcontext [lm_app (lm_lam ?m1) ?m2] => fold (lm_let m2 m1) end);
  repeat (fold lm_i).


Inductive tembed : context -> (* tm *) term -> (* lm *) term -> Prop :=
| tembed_var {G i t}
    : index i G (cl_tm t)
      -> tembed G (var i) (var i)

| tembed_varu {G i}
    : index i G cl_utm
      -> tembed G (var i) (var i)

| tembed_star {G}
    : tembed G tm_star lm_i

| tembed_lam {G t e m}
    : tembed (G; cl_utm) e m
      -> tembed G (tm_lam t e) (lm_lam m)

| tembed_app {G e1 e2 m1 m2}
    : tembed G e1 m1
      -> tembed G e2 m2
      -> tembed G (tm_app e1 e2) (lm_app m1 m2)

| tembed_pair {G e1 e2 m1 m2}
    : tembed G e1 m1
      -> tembed G e2 m2
      -> tembed G (tm_pair e1 e2)
         (lm_let m1 (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1)))))

| tembed_pair_val1 {G e1 e2 m1 m2}
    : tvalue G e1
      -> tembed G e1 m1
      -> tembed G e2 m2
      -> tembed G (tm_pair e1 e2)
         (lm_let m2 (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1)) (var 1))))

| tembed_pair_val2 {G e1 e2 m1 m2}
    : tvalue G e1
      -> tvalue G e2
      -> tembed G e1 m1
      -> tembed G e2 m2
      -> tembed G (tm_pair e1 e2)
         (lm_lam (lm_app (lm_app (var 0) (shift1 m1)) (shift1 m2)))

| tembed_pi1 {G e m}
    : tembed G e m
      -> tembed G (tm_pi1 e) (lm_app m (lm_lam (lm_lam (var 1))))

| tembed_pi2 {G e m}
    : tembed G e m
      -> tembed G (tm_pi2 e) (lm_app m (lm_lam (lm_lam (var 0))))

| tembed_plam {G k e m}
    : tembed (G; cl_ucn) e m
      -> tembed G (tm_plam k e) (lm_lam m)

| tembed_papp {G e c m}
    : tembed G e m
      -> tembed G (tm_papp e c) (lm_app m lm_i)

| tembed_pack {G c e k t m}
    : tembed G e m
      -> tembed G (tm_pack c e k t) (lm_let m (lm_lam (var 1)))

| tembed_pack_val {G c e k t m}
    : tvalue G e
      -> tembed G e m
      -> tembed G (tm_pack c e k t) (lm_lam (shift1 m))

| tembed_unpack {G e1 e2 m1 m2}
    : tembed G e1 m1
      -> tembed (G; cl_ucn; cl_utm) e2 m2
      -> tembed G (tm_unpack e1 e2)
         (lm_let (lm_app m1 lm_i) (subst (dot (var 0) (dot lm_i sh1)) m2))

| tembed_fix {G t e m}
    : tembed G e m
      -> tembed G (tm_fix t e) (lm_app lm_theta (lm_lam (lm_app (shift1 m) (var 0))))

| tembed_lett {G e1 e2 m1 m2}
    : tembed G e1 m1
      -> tembed (G; cl_utm) e2 m2
      -> tembed G (tm_lett e1 e2) (lm_let m1 m2)

| tembed_letm {G M e m1 m2}
    : membed G M m1
      -> tembed (G; cl_ucn; cl_umd (var 0)) e m2
      -> tembed G (tm_letm M e) (lm_let m1 (subst (dot (var 0) (dot lm_i sh1)) m2))

| tembed_ext {G M m}
    : membed G M m
      -> tembed G (tm_ext M) (lm_let m (var 0))
    

with membed : context -> (* md *) term -> (* lm *) term -> Prop :=
| membed_var {G i T c}
    : index i G (cl_md T c)
      -> membed G (var i) (var i)

| membed_varu {G i c}
    : index i G (cl_umd c)
      -> membed G (var i) (var i)

| membed_star {G}
    : membed G md_star lm_i

| membed_stat {G c}
    : membed G (md_stat c) lm_i

| membed_dyn {G e m}
    : tembed G e m
      -> membed G (md_dyn e) m

| membed_lamgn {G T M m}
    : membed (G; cl_ucn; cl_umd (var 0)) M m
      -> membed G (md_lamgn T M) (lm_lam (subst (dot (var 0) (dot lm_i sh1)) m))

| membed_appgn {G M1 M2 m1 m2}
    : membed G M1 m1
      -> membed G M2 m2
      -> membed G (md_appgn M1 M2) (lm_app m1 m2)

| membed_lamap {G T M m}
    : membed (G; cl_ucn; cl_umd (var 0)) M m
      -> membed G (md_lamap T M) (lm_lam (subst (dot (var 0) (dot lm_i sh1)) m))

| membed_appap {G M1 M2 m1 m2}
    : membed G M1 m1
      -> membed G M2 m2
      -> membed G (md_appap M1 M2) (lm_app m1 m2)

| membed_pair {G M1 M2 m1 m2}
    : membed G M1 m1
      -> membed G M2 m2
      -> membed G (md_pair M1 M2)
         (lm_let m1 (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1)))))

| membed_pair_val1 {G M1 M2 m1 m2}
    : mvalue G M1
      -> membed G M1 m1
      -> membed G M2 m2
      -> membed G (md_pair M1 M2)
         (lm_let m2 (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1)) (var 1))))

| membed_pair_val2 {G M1 M2 m1 m2}
    : mvalue G M1
      -> mvalue G M2
      -> membed G M1 m1
      -> membed G M2 m2
      -> membed G (md_pair M1 M2)
         (lm_lam (lm_app (lm_app (var 0) (shift1 m1)) (shift1 m2)))

| membed_pi1 {G M m}
    : membed G M m
      -> membed G (md_pi1 M) (lm_app m (lm_lam (lm_lam (var 1))))

| membed_pi2 {G M m}
    : membed G M m
      -> membed G (md_pi2 M) (lm_app m (lm_lam (lm_lam (var 0))))

| membed_unpack {G e M T m1 m2}
    : tembed G e m1
      -> membed (G; cl_ucn; cl_utm) M m2
      -> membed G (md_unpack e M T)
         (lm_let (lm_app m1 lm_i) (subst (dot (var 0) (dot lm_i sh1)) m2))

| membed_lett {G e M m1 m2}
    : tembed G e m1
      -> membed (G; cl_utm) M m2
      -> membed G (md_lett e M) (lm_let m1 m2)

| membed_letm {G M1 M2 T m1 m2}
    : membed G M1 m1
      -> membed (G; cl_ucn; cl_umd (var 0)) M2 m2
      -> membed G (md_letm M1 M2 T) (lm_let m1 (subst (dot (var 0) (dot lm_i sh1)) m2))

| membed_seal {G M T m}
    : membed G M m
      -> membed G (md_seal M T) (lm_let lm_i (shift1 m))
.


Inductive embed : context -> term -> term -> term -> Prop :=
| embed_cn {G c1 c2 k}
    : embed G c1 c2 (cl_cn k)

| embed_tm {G e m t}
    : tvalue G e
      -> tembed G e m
      -> embed G e m (cl_tm t)

| embed_md {G M m T c}
    : mvalue G M
      -> membed G M m
      -> fstm G M c
      -> embed G M m (cl_md T c)

| embed_ucn {G c1 c2}
    : embed G c1 c2 cl_ucn

| embed_utm {G e m}
    : tvalue G e
      -> tembed G e m
      -> embed G e m cl_utm

| embed_umd {G M m c}
    : mvalue G M
      -> membed G M m
      -> fstm G M c
      -> embed G M m (cl_umd c)
.


Scheme tembed_mut_ind := Minimality for tembed Sort Prop
with   membed_mut_ind := Minimality for membed Sort Prop.
Combined Scheme embeds_ind from tembed_mut_ind, membed_mut_ind.


Hint Constructors tembed membed embed : embed.


Lemma embeds_hlm :
  (forall G e m,
     tembed G e m
     -> hlm m)
  /\
  (forall G M m,
     membed G M m
     -> hlm m).
Proof.
#6 exploit 
  (embeds_ind
     (fun G e m => hlm m)
     (fun G M m => hlm m));
try (intros; eauto 18 using hlm, shift_hlm; done).

(* unpack *)
intros G e1 e2 m1 m2 _ Hm1 _ Hm2.
sor (refine (subst_under_hlm 1 lm_i _ _ Hm2)) as H.
  eauto using hlm; done.
simp_sub in H.
eauto using hlm; done.

(* letm *)
intros G M e m1 m2 _ Hm1 _ Hm2.
sor (refine (subst_under_hlm 1 lm_i _ _ Hm2)) as H.
  eauto using hlm; done.
simp_sub in H.
eauto using hlm; done.

(* lamgn *)
intros G T M m _ Hm.
sor (refine (subst_under_hlm 1 lm_i _ _ Hm)) as H.
  eauto using hlm; done.
simp_sub in H.
eauto using hlm; done.

(* lamap *)
intros G T M m _ Hm.
sor (refine (subst_under_hlm 1 lm_i _ _ Hm)) as H.
  eauto using hlm; done.
simp_sub in H.
eauto using hlm; done.

(* unpack *)
intros G e M T m1 m2 _ Hm1 _ Hm2.
sor (refine (subst_under_hlm 1 lm_i _ _ Hm2)) as H.
  eauto using hlm; done.
simp_sub in H.
eauto using hlm; done.

(* letm *)
intros G M1 M2 T m1 m2 _ Hm1 _ Hm2.
sor (refine (subst_under_hlm 1 lm_i _ _ Hm2)) as H.
  eauto using hlm; done.
simp_sub in H.
eauto using hlm; done.
Qed.
 

Lemma tembed_hlm :
  forall G e m,
    tembed G e m
    -> hlm m.
Proof.
exact (embeds_hlm andel).
Qed.


Lemma membed_hlm :
  forall G M m,
    membed G M m
    -> hlm m.
Proof.
exact (embeds_hlm ander).
Qed.


Hint Resolve tembed_hlm membed_hlm : embed.


Lemma embed_behaved :
  behavedrel embed.
Proof.
intros G i C HC Hindex.
destruct HC; simp_sub; eauto with embed static dynamic.
Qed.


Lemma embed_oof :
  forall G e m C,
    embed G e m C -> oof G e C.
Proof.
intros G e m C H.
invert H; intros; subst; auto using oof.
Qed.


Lemma subst_embeds_wc :
  (forall G1 G2 s s' e m,
     subrel (wcrel embed) G1 s s' G2
     -> tembed G2 e m
     -> tembed G1 (subst s e) (subst s' m))
  /\
  (forall G1 G2 s s' M m,
     subrel (wcrel embed) G1 s s' G2
     -> membed G2 M m
     -> membed G1 (subst s M) (subst s' m)).
Proof.
assert (forall G1 G2 s s' e, subrel (wcrel embed) G1 s s' G2 -> tvalue G2 e -> tvalue G1 (subst s e)) as Hsubstv.
  intros G1 G2 s s' e Hs Hval.
  eapply subst_tvalue; eauto; [].
  refine (subrel_subof_mono _#7 Hs); [].
  intros; eapply embed_oof; eauto; [].
  eapply wcrel_elim; eauto; done.
assert (forall G1 G2 s s' M, subrel (wcrel embed) G1 s s' G2 -> mvalue G2 M -> mvalue G1 (subst s M)) as Hsubstvm.
  intros G1 G2 s s' M Hs Hval.
  eapply subst_mvalue; eauto; [].
  refine (subrel_subof_mono _#7 Hs); [].
  intros; eapply embed_oof; eauto; [].
  eapply wcrel_elim; eauto; done.
#17 exploit
  (embeds_ind
     (fun G2 e m =>
        forall G1 s s',
          subrel (wcrel embed) G1 s s' G2
          -> tembed G1 (subst s e) (subst s' m))
     (fun G2 M m =>
        forall G1 s s',
          subrel (wcrel embed) G1 s s' G2
          -> membed G1 (subst s M) (subst s' m))) as Hind;
try (intros; simp_sub_lm; rewrite -> ?subst_compose; eauto with embed; done).

(* var *)
intros G2 i t Hindex G1 s s' Hs.
simp_sub.
so (subrel_index _ (wcrel_behaved _ embed_behaved) _#6 iscl_tm Hs Hindex) as H.
invert (wcrel_elim _#5 H); auto; done.

(* varu *)
intros G2 i Hindex G1 s s' Hs.
simp_sub.
so (subrel_index _ (wcrel_behaved _ embed_behaved) _#6 iscl_utm Hs Hindex) as H.
invert (wcrel_elim _#5 H); auto; done.

(* lam *)
intros G t e m _ IH G1 s s' Hs.
simp_sub.
apply tembed_lam; [].
apply IH; [].
so (subrel_wc_bind _ embed_behaved _#5 iscl_utm Hs) as H.
simp_sub in H.
assumption.

(* pair_val2 *)
intros.
simp_sub_lm.
rewrite -> ?subst_compose.
apply tembed_pair_val2; eauto; done.

(* plam *)
intros G2 k e m _ IH G1 s s' Hs.
simp_sub.
apply tembed_plam; [].
apply IH; [].
so (subrel_wc_bind _ embed_behaved _#5 iscl_ucn Hs) as H.
simp_sub in H.
assumption.

(* unpack *)
intros G2 e1 e2 m1 m2 _ IH1 _ IH2 G1 s s' Hs.
simp_sub_lm.
repl (dot (var 0) (dot lm_i (compose s' sh1))) with (compose (under 2 s') (dot (var 0) (dot lm_i sh1))) by (simp_sub; reflexivity).
rewrite -> subst_compose.
apply tembed_unpack; auto; [].
apply IH2; [].
simp_sub.
repl (dot (var 1) (compose s (sh 2))) with (compose (dot (var 0) (compose s sh1)) sh1) by (simp_sub; reflexivity).
repl (dot (var 1) (compose s' (sh 2))) with (compose (dot (var 0) (compose s' sh1)) sh1) by (simp_sub; reflexivity).
repl (G1; cl_ucn; cl_utm) with (G1; subst s cl_ucn; subst (dot (var 0) (compose s sh1)) cl_utm) by (simp_sub; reflexivity).
apply subrel_wc_bind; auto using embed_behaved with static; [].
apply subrel_wc_bind; auto using embed_behaved with static; done.

(* lett *)
intros G2 e1 e2 m1 m2 _ IH1 _ IH2 G1 s s' Hs.
simp_sub_lm.
apply tembed_lett; auto; [].
apply IH2; [].
repl (G1; cl_utm) with (G1; subst s cl_utm) by (simp_sub; reflexivity).
apply subrel_wc_bind; auto using embed_behaved with static; done.

(* letm *)
intros G2 M e m1 m2 _ IH1 _ IH2 G1 s s' H3.
simp_sub_lm.
repl (dot (var 0) (dot lm_i (compose s' sh1))) with (compose (under 2 s') (dot (var 0) (dot lm_i sh1))) by (simp_sub; reflexivity).
rewrite -> subst_compose.
apply tembed_letm; auto; [].
apply IH2; [].
simp_sub.
repl (dot (var 1) (compose s (sh 2))) with (compose (dot (var 0) (compose s sh1)) sh1) by (simp_sub; reflexivity).
repl (dot (var 1) (compose s' (sh 2))) with (compose (dot (var 0) (compose s' sh1)) sh1) by (simp_sub; reflexivity).
repl (G1; cl_ucn; cl_umd (var 0)) with (G1; subst s cl_ucn; subst (dot (var 0) (compose s sh1)) (cl_umd (var 0))) by (simp_sub; reflexivity).
apply subrel_wc_bind; auto using embed_behaved with static; [].
apply subrel_wc_bind; auto using embed_behaved with static; done.

(* var *)
intros G2 i T c Hindex G1 s s' Hs.
simp_sub.
so (subrel_index _ (wcrel_behaved _ embed_behaved) _#6 iscl_md Hs Hindex) as H.
invert (wcrel_elim _#5 H); auto; done.

(* varu *)
intros G2 i c Hindex G1 s s' Hs.
simp_sub.
so (subrel_index _ (wcrel_behaved _ embed_behaved) _#6 iscl_umd Hs Hindex) as H.
invert (wcrel_elim _#5 H); auto; done.

(* lamgn *)
intros G T M m _ IH G1 s s' H1.
simp_sub_lm.
repl (dot (var 0) (dot lm_i (compose s' sh1))) with (compose (under 2 s') (dot (var 0) (dot lm_i sh1))) by (simp_sub; reflexivity).
rewrite -> subst_compose.
apply membed_lamgn; [].
apply IH; [].
repl (dot (var 1) (compose s (sh 2))) with (compose (dot (var 0) (compose s sh1)) sh1) by (simp_sub; reflexivity).
repl (under 2 s') with (dot (var 0) (compose (dot (var 0) (compose s' sh1)) sh1)) by (simp_sub; reflexivity).
repl (G1; cl_ucn; cl_umd (var 0)) with (G1; subst s cl_ucn; subst (dot (var 0) (compose s sh1)) (cl_umd (var 0))) by (simp_sub; reflexivity).
apply subrel_wc_bind; eauto using embed_behaved with static; [].
apply subrel_wc_bind; eauto using embed_behaved with static; done.

(* lamap *)
intros G T M m _ IH G1 s s' H1.
simp_sub_lm.
repl (dot (var 0) (dot lm_i (compose s' sh1))) with (compose (under 2 s') (dot (var 0) (dot lm_i sh1))) by (simp_sub; reflexivity).
rewrite -> subst_compose.
apply membed_lamap; [].
apply IH; [].
repl (dot (var 1) (compose s (sh 2))) with (compose (dot (var 0) (compose s sh1)) sh1) by (simp_sub; reflexivity).
repl (under 2 s') with (dot (var 0) (compose (dot (var 0) (compose s' sh1)) sh1)) by (simp_sub; reflexivity).
repl (G1; cl_ucn; cl_umd (var 0)) with (G1; subst s cl_ucn; subst (dot (var 0) (compose s sh1)) (cl_umd (var 0))) by (simp_sub; reflexivity).
apply subrel_wc_bind; eauto using embed_behaved with static; [].
apply subrel_wc_bind; eauto using embed_behaved with static; done.

(* pair *)
intros.
simp_sub_lm.
rewrite -> ?subst_compose.
apply membed_pair_val2; eauto; done.

(* unpack *)
intros G e M T m1 m2 _ IH1 _ IH2 G1 s s' H3.
simp_sub_lm.
repl (dot (var 0) (dot lm_i (compose s' sh1))) with (compose (under 2 s') (dot (var 0) (dot lm_i sh1))) by (simp_sub; reflexivity).
rewrite -> subst_compose.
apply membed_unpack; auto; [].
apply IH2; [].
simp_sub.
repl (dot (var 1) (compose s (sh 2))) with (compose (dot (var 0) (compose s sh1)) sh1) by (simp_sub; reflexivity).
repl (dot (var 1) (compose s' (sh 2))) with (compose (dot (var 0) (compose s' sh1)) sh1) by (simp_sub; reflexivity).
repl (G1; cl_ucn; cl_utm) with (G1; subst s cl_ucn; subst (dot (var 0) (compose s sh1)) cl_utm) by (simp_sub; reflexivity).
apply subrel_wc_bind; auto using embed_behaved with static; [].
apply subrel_wc_bind; auto using embed_behaved with static; done.

(* lett *)
intros G e M m1 m2 _ IH1 _ IH2 G1 s s' H3.
simp_sub_lm.
apply membed_lett; auto; [].
apply IH2; [].
repl (G1; cl_utm) with (G1; subst s cl_utm) by (simp_sub; reflexivity).
apply subrel_wc_bind; auto using embed_behaved with static; done.

(* letm *)
intros G M1 M2 T m1 m2 _ IH1 _ IH2 G1 s s' H3.
simp_sub_lm.
repl (dot (var 0) (dot lm_i (compose s' sh1))) with (compose (under 2 s') (dot (var 0) (dot lm_i sh1))) by (simp_sub; reflexivity).
rewrite -> subst_compose.
apply membed_letm; auto; [].
apply IH2; [].
simp_sub.
repl (dot (var 1) (compose s (sh 2))) with (compose (dot (var 0) (compose s sh1)) sh1) by (simp_sub; reflexivity).
repl (dot (var 1) (compose s' (sh 2))) with (compose (dot (var 0) (compose s' sh1)) sh1) by (simp_sub; reflexivity).
repl (G1; cl_ucn; cl_umd (var 0)) with (G1; subst s cl_ucn; subst (dot (var 0) (compose s sh1)) (cl_umd (var 0))) by (simp_sub; reflexivity).
apply subrel_wc_bind; auto using embed_behaved with static; [].
apply subrel_wc_bind; auto using embed_behaved with static; done.

(* epilogue *)
destruct_all Hind; split; eauto; done.
Qed.


Lemma subst_tembed_wc :
  forall G1 G2 s s' e m,
    subrel (wcrel embed) G1 s s' G2
    -> tembed G2 e m
    -> tembed G1 (subst s e) (subst s' m).
Proof.
exact (subst_embeds_wc andel).
Qed.


Lemma subst_membed_wc :
  forall G1 G2 s s' M m,
    subrel (wcrel embed) G1 s s' G2
    -> membed G2 M m
    -> membed G1 (subst s M) (subst s' m).
Proof.
exact (subst_embeds_wc ander).
Qed.


Lemma subst_embed_wc :
  forall G1 G2 s s' t t' C,
    subrel (wcrel embed) G1 s s' G2
    -> embed G2 t t' C
    -> embed G1 (subst s t) (subst s' t') (subst s C).
Proof.
assert (forall G1 G2 s s', subrel (wcrel embed) G1 s s' G2 -> subof oof G1 s G2) as Hoof.
  intros G1 G2 s s' H.
  refine (subrel_subof_mono _#7 H); [].
  intros; eapply embed_oof; eauto; [].
  eapply wcrel_elim; eauto; done.
assert (forall G1 G2 s s', subrel (wcrel embed) G1 s s' G2 -> subof tof G1 s G2) as Htof.
  intros G1 G2 s s' H.
  refine (subof_mono _#6 (Hoof _#4 H)); [].
  intros; apply oof_tof; auto; done.
intros G1 G2 s s' t t' C Hs H.
destruct H; simp_sub; eauto 7 using subst_tembed_wc, subst_membed_wc, subst_tvalue, subst_mvalue, subst_fstm with embed; done.
Qed.


Lemma embed_wc :
  forall G e m C,
    embed G e m C -> wcrel embed G e m C.
Proof.
exact (wcrel_intro _ subst_embed_wc).
Qed.


Lemma subst_tembed :
  forall G1 G2 s s' e m,
    subrel embed G1 s s' G2
    -> tembed G2 e m
    -> tembed G1 (subst s e) (subst s' m).
Proof.
intros G1 G2 s s' e m Hs H.
eapply subst_tembed_wc; eauto using subrel_mono, embed_wc; done.
Qed.


Lemma subst_membed :
  forall G1 G2 s s' M m,
    subrel embed G1 s s' G2
    -> membed G2 M m
    -> membed G1 (subst s M) (subst s' m).
Proof.
intros G1 G2 s s' M m Hs H.
eapply subst_membed_wc; eauto using subrel_mono, embed_wc; done.
Qed.


Lemma can_embeds :
  (forall G e t,
     oft G e t
     -> exists m, tembed G e m)
  /\
  (forall G p M T,
     ofm G p M T
     -> exists m, membed G M m).
Proof.
#12 exploit
  (dynamic_ind
     (fun G e t => exists m, tembed G e m)
     (fun G p M T => exists m, membed G M m)) as Hind;
try (intros;
     repeat (match goal with
             | H : exists _, tembed _ _ _ |- _ => destruct H
             | H : exists _, membed _ _ _ |- _ => destruct H
             end);
     eexists;
     eauto with embed;
     done).

(* lam *)
intros G t1 t2 e _ _ (m & IH).
exists (lm_lam m).
apply tembed_lam; [].
sor (refine (@subst_tembed (G; cl_utm) _ (under 1 id) (under 1 id) _ _ _ IH)) as H.
  simp_sub.
  apply subrel_dot; eauto using subrel_sh with sequence; [].
  simp_sub.
  apply embed_tm; eauto using tvalue with embed sequence; done.
simp_sub in H.
assumption.

(* plam *)
intros G k t e _ _ (m & IH).
exists (lm_lam m).
apply tembed_plam; [].
sor (refine (@subst_tembed (G; cl_ucn) _ (under 1 id) (under 1 id) _ _ _ IH)) as H.
  simp_sub.
  apply subrel_dot; eauto using subrel_sh with sequence; [].
  simp_sub.
  apply embed_cn; eauto; done.
simp_sub in H.
assumption.

(* unpack *)
intros G k t t' e1 e2 _ (m1 & IH1) _ (m2 & IH2) _.
exists (lm_let (lm_app m1 lm_i) (subst (dot (var 0) (dot lm_i sh1)) m2)).
apply tembed_unpack; auto; [].
sor (refine (@subst_tembed (G; cl_ucn; cl_utm) _ (under 2 id) (under 2 id) _ _ _ IH2)) as H.
  simp_sub.
  #2 apply subrel_dot.
    apply subrel_dot; eauto using subrel_sh with sequence; [].
    simp_sub.
    apply embed_cn; eauto; done.

    simp_sub.
    apply embed_tm; eauto using tvalue with embed sequence; done.
simp_sub in H.
assumption.

(* lett *)
intros G t1 t2 e1 e2 _ (m1 & IH1) _ (m2 & IH2).
exists (lm_let m1 m2).
apply tembed_lett; auto; [].
sor (refine (@subst_tembed (G; cl_utm) _ (under 1 id) (under 1 id) _ _ _ IH2)) as H.
  simp_sub.
  apply subrel_dot; eauto using subrel_sh with sequence; [].
  simp_sub.
  apply embed_tm; eauto using tvalue with embed sequence; done.
simp_sub in H.
assumption.

(* letm *)
intros G T k t M e _ (m1 & IH1) _ _ (m2 & IH2) _.
exists (lm_let m1 (subst (dot (var 0) (dot lm_i sh1)) m2)).
apply tembed_letm; auto; [].
sor (refine (@subst_tembed (G; cl_ucn; cl_umd (var 0)) _ (under 2 id) (under 2 id) _ _ _ IH2)) as H.
  simp_sub.
  #2 apply subrel_dot.
    apply subrel_dot; eauto using subrel_sh with sequence; [].
    simp_sub.
    apply embed_cn; eauto; done.

    simp_sub.
    apply embed_md; eauto using mvalue with embed sequence static; [].
    repl (var 1) with (shift1 (var 0)) by (simp_sub; reflexivity).
    apply fstm_varu; auto using index_0; done.
simp_sub in H.
assumption.

(* lamgn *)
intros G T1 T2 k1 M _ _ _ (m & IH).
exists (lm_lam (subst (dot (var 0) (dot lm_i sh1)) m)).
apply membed_lamgn; [].
sor (refine (@subst_membed (G; cl_ucn; cl_umd (var 0)) _ (under 2 id) (under 2 id) _ _ _ IH)) as H.
  simp_sub.
  #2 apply subrel_dot.
    apply subrel_dot; eauto using subrel_sh with sequence; [].
    simp_sub.
    apply embed_cn; eauto; done.

    simp_sub.
    apply embed_md; eauto using mvalue with embed sequence static; [].
    repl (var 1) with (shift1 (var 0)) by (simp_sub; reflexivity).
    apply fstm_varu; auto using index_0; done.
simp_sub in H.
assumption.

(* lamap *)
intros G T1 T2 k1 M _ _ _ (m & IH).
exists (lm_lam (subst (dot (var 0) (dot lm_i sh1)) m)).
apply membed_lamap; [].
sor (refine (@subst_membed (G; cl_ucn; cl_umd (var 0)) _ (under 2 id) (under 2 id) _ _ _ IH)) as H.
  simp_sub.
  #2 apply subrel_dot.
    apply subrel_dot; eauto using subrel_sh with sequence; [].
    simp_sub.
    apply embed_cn; eauto; done.

    simp_sub.
    apply embed_md; eauto using mvalue with embed sequence static; [].
    repl (var 1) with (shift1 (var 0)) by (simp_sub; reflexivity).
    apply fstm_varu; auto using index_0; done.
simp_sub in H.
assumption.

(* unpack *)
intros G k t T e M _ (m1 & IH1) _ (m2 & IH2) _.
exists (lm_let (lm_app m1 lm_i) (subst (dot (var 0) (dot lm_i sh1)) m2)).
apply membed_unpack; auto; [].
sor (refine (@subst_membed (G; cl_ucn; cl_utm) _ (under 2 id) (under 2 id) _ _ _ IH2)) as H.
  simp_sub.
  #2 apply subrel_dot.
    apply subrel_dot; eauto using subrel_sh with sequence; [].
    simp_sub.
    apply embed_cn; eauto; done.

    simp_sub.
    apply embed_tm; eauto using tvalue with embed sequence; done.
simp_sub in H.
assumption.

(* lett *)
intros G p t T e M _ (m1 & IH1) _ (m2 & IH2).
exists (lm_let m1 m2).
apply membed_lett; auto; [].
sor (refine (@subst_membed (G; cl_utm) _ (under 1 id) (under 1 id) _ _ _ IH2)) as H.
  simp_sub.
  apply subrel_dot; eauto using subrel_sh with sequence; [].
  simp_sub.
  apply embed_tm; eauto using tvalue with embed sequence; done.
simp_sub in H.
assumption.

(* letm *)
intros G T1 T2 k1 M1 M2 _ (m1 & IH1) _ _ (m2 & IH2) _.
exists (lm_let m1 (subst (dot (var 0) (dot lm_i sh1)) m2)).
apply membed_letm; auto; [].
sor (refine (@subst_membed (G; cl_ucn; cl_umd (var 0)) _ (under 2 id) (under 2 id) _ _ _ IH2)) as H.
  simp_sub.
  #2 apply subrel_dot.
    apply subrel_dot; eauto using subrel_sh with sequence; [].
    simp_sub.
    apply embed_cn; eauto; done.

    simp_sub.
    apply embed_md; eauto using mvalue with embed sequence static; [].
    repl (var 1) with (shift1 (var 0)) by (simp_sub; reflexivity).
    apply fstm_varu; auto using index_0; done.
simp_sub in H.
assumption.

(* extsigma *)
intros G T1 T2 M _ (m & IH) _ _.
invert IH; [].
intros m' Hembed _.
eauto; done.

(* epilogue *)
destruct_all Hind.
split; eauto; done.
Qed.


Lemma can_tembed :
  forall G e t,
    oft G e t
    -> exists m, tembed G e m.
Proof.
exact (can_embeds andel).
Qed.


Lemma can_membed :
  forall G p M T,
    ofm G p M T
    -> exists m, membed G M m.
Proof.
exact (can_embeds ander).
Qed.


Lemma tembed_val :
  forall G v m,
    tembed G v m
    -> tvalue G v
    -> exists m',
         tembed G v m'
         /\ val m'
         /\ star step m m'.
Proof.
intros G v m Hembed Hval.
revert m Hembed.
induct Hval.

(* var *)
intros G i t Hindex m Hembed.
assert (m = var i).
  invert Hembed; intros; subst; reflexivity.
subst m.
exists (var i).
eauto using star_refl with embed lambda; done.

(* varu *)
intros G i Hindex m Hembed.
assert (m = var i).
  invert Hembed; intros; subst; reflexivity.
subst m.
exists (var i).
eauto using star_refl with embed lambda; done.

(* star *)
intros G m Hembed.
invert Hembed; [].
intros <-.
exists lm_i.
eauto using star_refl with embed lambda; done.

(* lam *)
intros G t e m Hembed.
invert Hembed; [].
intros m1 _ <-.
exists (lm_lam m1).
eauto using star_refl with embed lambda; done.

(* pair *)
intros G v1 v2 Hval1 IH1 Hval2 IH2 m Hembed.
#3 invert Hembed.
  intros m1 m2 Hembed1 Hembed2 <-.
  so (IH1 _ Hembed1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (IH2 _ Hembed2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists (lm_lam (lm_app (lm_app (var 0) (shift1 m1')) (shift1 m2'))).
  do2 2 split.
    apply tembed_pair_val2; auto; done.
    
    apply val_lam; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun m => lm_let m (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1)))))); eauto with lambda; done.

      #2 eapply star_step.
        apply step_let_beta; eauto; done.

        simp_sub_lm.
        #2 eapply star_trans.
          apply (star_map' _ _ (fun m => lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1')) (var 1))))); eauto with lambda; done.

          #2 eapply star_step.
            eapply step_let_beta; eauto; done.

            simp_sub_lm.
            apply star_refl; done.

  intros m1 m2 Hval1' Hembed1 Hembed2 <-.
  so (IH2 _ Hembed2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists (lm_lam (lm_app (lm_app (var 0) (shift1 m1)) (shift1 m2'))).
  do2 2 split.
    apply tembed_pair_val2; auto; done.

    apply val_lam; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun m => lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1)) (var 1))))); eauto with lambda; done.

      #2 eapply star_step.
        eapply step_let_beta; eauto; done.

        simp_sub_lm.
        apply star_refl; done.

  intros m1 m2 Hval1' Hembed1 Hval2' Hembed2 <-.
  exists (lm_lam (lm_app (lm_app (var 0) (shift1 m1)) (shift1 m2))).
  do2 2 split.
    apply tembed_pair_val2; auto; done.

    apply val_lam; done.

    apply star_refl; done.

(* plam *)
intros G k e m Hembed.
invert Hembed; [].
intros m1 _ <-.
exists (lm_lam m1).
eauto using star_refl with embed lambda; done.

(* pack *)
intros G c v k t Hval IH m Hembed.
#2 invert Hembed.
  intros m1 Hembed1 <-.
  so (IH _ Hembed1) as (m1' & Hembed1' & Hval' & Hstep).
  exists (lm_lam (shift1 m1')).
  do2 2 split.
    apply tembed_pack_val; auto; done.

    apply val_lam; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun m => lm_let m (lm_lam (var 1)))); eauto with lambda; done.

      #2 eapply star_step.
        apply step_let_beta; eauto; done.

        simp_sub_lm.
        apply star_refl; done.

  intros m1 Hval' Hembed1 <-.
  exists (lm_lam (shift1 m1)).
  do2 2 split.
    apply tembed_pack_val; auto; done.

    apply val_lam; done.

    apply star_refl; done.
Qed.


Lemma membed_val :
  forall G V m,
    membed G V m
    -> mvalue G V
    -> exists m',
         membed G V m'
         /\ val m'
         /\ star step m m'.
Proof.
intros G V m Hembed Hval.
revert m Hembed.
induct Hval.

(* var *)
intros G i T c Hindex m Hembed.
assert (m = var i).
  invert Hembed; intros; subst; reflexivity.
subst m.
exists (var i).
eauto using star_refl with embed lambda; done.

(* varu *)
intros G i c Hindex m Hembed.
assert (m = var i).
  invert Hembed; intros; subst; reflexivity.
subst m.
exists (var i).
eauto using star_refl with embed lambda; done.

(* star *)
intros G m Hembed.
invert Hembed; [].
intros <-.
eauto using star_refl with embed lambda; done.

(* stat *)
intros G c m Hembed.
invert Hembed; [].
intros <-.
eauto using star_refl with embed lambda; done.

(* dyn *)
intros G v Hval m Hembed.
invert Hembed; [].
intros Hembed'.
so (tembed_val _#3 Hembed' Hval) as (m' & Hembed'' & Hval' & Hstep).
eauto with embed lambda; done.

(* lamgn *)
intros G T M m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
eauto using star_refl with embed lambda; done.

(* lamap *)
intros G T M m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
eauto using star_refl with embed lambda; done.

(* pair *)
intros G V1 V2 Hval1 IH1 Hval2 IH2 m Hembed.
#3 invert Hembed.
  intros m1 m2 Hembed1 Hembed2 <-.
  so (IH1 _ Hembed1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (IH2 _ Hembed2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists (lm_lam (lm_app (lm_app (var 0) (shift1 m1')) (shift1 m2'))).
  do2 2 split.
    apply membed_pair_val2; auto; done.
    
    apply val_lam; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun m => lm_let m (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1)))))); eauto with lambda; done.

      #2 eapply star_step.
        apply step_let_beta; eauto; done.

        simp_sub_lm.
        #2 eapply star_trans.
          apply (star_map' _ _ (fun m => lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1')) (var 1))))); eauto with lambda; done.

          #2 eapply star_step.
            eapply step_let_beta; eauto; done.

            simp_sub_lm.
            apply star_refl; done.

  intros m1 m2 Hval1' Hembed1 Hembed2 <-.
  so (IH2 _ Hembed2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists (lm_lam (lm_app (lm_app (var 0) (shift1 m1)) (shift1 m2'))).
  do2 2 split.
    apply membed_pair_val2; auto; done.

    apply val_lam; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun m => lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1)) (var 1))))); eauto with lambda; done.

      #2 eapply star_step.
        eapply step_let_beta; eauto; done.

        simp_sub_lm.
        apply star_refl; done.

  intros m1 m2 Hval1' Hembed1 Hval2' Hembed2 <-.
  exists (lm_lam (lm_app (lm_app (var 0) (shift1 m1)) (shift1 m2))).
  do2 2 split.
    apply membed_pair_val2; auto; done.

    apply val_lam; done.

    apply star_refl; done.
Qed.


Lemma tembed_val' :
  forall G v m,
    tembed G v m
    -> tvalue G v
    -> exists m',
         tembed G v m'
         /\ val m'
         /\ star reduce m m'.
Proof.
intros G v m Hembed Hval.
so (tembed_val _#3 Hembed Hval) as (m' & Hembed' & Hval' & Hsteps).
so (steps_reduces _ _ (tembed_hlm _#3 Hembed) Hsteps) as Hreds.
eauto; done.
Qed.  


Lemma membed_val' :
  forall G V m,
    membed G V m
    -> mvalue G V
    -> exists m',
         membed G V m'
         /\ val m'
         /\ star reduce m m'.
Proof.
intros G V m Hembed Hval.
so (membed_val _#3 Hembed Hval) as (m' & Hembed' & Hval' & Hsteps).
so (steps_reduces _ _ (membed_hlm _#3 Hembed) Hsteps) as Hreds.
eauto; done.
Qed.  


Lemma embeds_step :
  (forall G e e' m,
     tembed G e m
     -> tstep G e e'
     -> exists m',
          tembed G e' m'
          /\ plus step m m')
  /\
  (forall G M M' m,
     membed G M m
     -> mstep G M M'
     -> exists m',
          membed G M' m'
          /\ plus step m m').
Proof.
exploit
  (Dynamic.step_ind
     (fun G e e' =>
        forall m,
          tembed G e m
          -> exists m',
               tembed G e' m'
               /\ plus step m m')
     (fun G M M' =>
        forall m,
          membed G M m
          -> exists m',
               membed G M' m'
               /\ plus step m m')) as Hind.

(* app1 *)
intros G e1 e1' e2 _ IH m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hstep).
exists (lm_app m1' m2).
split; eauto with embed; [].
apply (plus_map' _ _ (fun m => lm_app m m2)); auto with lambda; done.

(* app2 *)
intros G v1 e2 e2' Hval _ IH m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (tembed_val _#3 Hembed1 Hval) as (m1' & Hembed1' & Hval' & Hstep1).
so (IH _ Hembed2) as (m2' & Hembed2' & Hstep2).
exists (lm_app m1' m2').
split; auto with embed; [].
#2 apply (star_plus_trans _#3 (lm_app m1' m2)).
  apply (star_map' _ _ (fun m => lm_app m m2)); auto with lambda; done.

  apply (plus_map' _ _ (fun m => lm_app m1' m)); auto with lambda; done.

(* app3 *)
intros G t e1 v2 Hval m Hembed.
invert Hembed; [].
intros n m2 Hembed' Hembed2 <-.
invert Hembed'; [].
intros m1 Hembed1 <-.
so (tembed_val _#3 Hembed2 Hval) as (m2' & Hembed2' & Hval' & Hstep).
exists (subst1 m2' m1).
split.
  eapply subst_tembed; eauto; [].
  apply subrel_dot; eauto using subrel_sh with sequence; [].
  simp_sub.
  apply embed_utm; auto; done.

  apply plusr_plus.
  exists (lm_app (lm_lam m1) m2').
  split.
    apply (star_map' _ _ (fun m => lm_app (lm_lam m1) m)); auto with lambda; done.

    apply step_beta; auto; done.

(* pair1 *)
intros G e1 e1' e2 Hstep IH m Hembed.
#3 invert Hembed.
  intros m1 m2 Hembed1 Hembed2 <-.
  so (IH _ Hembed1) as (m1' & Hembed1' & Hstep').
  exists (lm_let m1' (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1))))).
  split.
    apply tembed_pair; auto; done.

    apply (plus_map' _ _ (fun m => lm_let m (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1)))))); auto with lambda; done.

  intros _ _ Hval1 _ _ _.
  destruct (tstep_value_contra _#3 Hstep Hval1); done.

  intros _ _ Hval1 _ _ _ _.
  destruct (tstep_value_contra _#3 Hstep Hval1); done.

(* pair2 *)
intros G v1 e2 e2' Hval1 Hstep IH m Hembed.
#3 invert Hembed.
  intros m1 m2 Hembed1 Hembed2 <-.
  so (tembed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (IH _ Hembed2) as (m2' & Hembed2' & Hstep2).
  exists (lm_let m2' (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1')) (var 1)))).
  split.
    apply tembed_pair_val1; auto; done.

    #2 eapply star_plus_trans.
      apply (star_map' _ _ (fun m => lm_let m (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1)))))); eauto with lambda; done.

      #2 eapply plus_step.
        apply step_let_beta; auto; done.

        simp_sub_lm.
        apply (plus_map' _ _ (fun m => lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1')) (var 1))))); auto with lambda; done.

  intros m1 m2 _ Hembed1 Hembed2 <-.
  so (IH _ Hembed2) as (m2' & Hembed2' & Hstep').
  exists (lm_let m2' (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1)) (var 1)))).
  split.
    apply tembed_pair_val1; auto; done.

    apply (plus_map' _ _ (fun m => lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1)) (var 1))))); auto with lambda; done.

  intros _ _ _ Hval2 _ _ _.
  destruct (tstep_value_contra _#3 Hstep Hval2); done.

(* pi11 *)
intros G e e' _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hstep1).
exists (lm_app m1' (lm_lam (lm_lam (var 1)))).
split.
  apply tembed_pi1; auto; done.

  apply (plus_map' _ _ (fun m => lm_app m (lm_lam (lm_lam (var 1))))); auto with lambda; done.

(* pi12 *)
intros G v1 v2 Hval1 Hval2 m Hembed.
invert Hembed; [].
intros n Hembed' <-.
#3 invert Hembed'.
  intros m1 m2 Hembed1 Hembed2 <-.
  so (tembed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (tembed_val _#3 Hembed2 Hval2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists m1'.
  split; auto; [].
  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1))))) (lm_lam (lm_lam (var 1))))); eauto with lambda; done.

    #2 eapply plus_step.
      exact (step_app1 (step_let_beta Hval1')).

      simp_sub_lm.
      #2 eapply star_plus_trans.
        apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1')) (var 1)))) (lm_lam (lm_lam (var 1))))); eauto with lambda; done.

        #2 eapply plus_step.
          exact (step_app1 (step_let_beta Hval2')).

          simp_sub_lm.
          #2 eapply plus_step.
            exact (step_let_beta val_lam).

            simp_sub_lm.
            #2 eapply plus_step.
              exact (step_app1 (step_let_beta Hval1')).

              simp_sub_lm.
              eexists; split.
                exact (step_let_beta Hval2').

                simp_sub_lm.
                apply star_refl; done.

  intros m1 m2 _ Hembed1 Hembed2 <-.
  so (tembed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (tembed_val _#3 Hembed2 Hval2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists m1'.
  split; auto; [].
  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1)) (var 1)))) (lm_lam (lm_lam (var 1))))); eauto with lambda; done.

    #2 eapply plus_step.
      exact (step_app1 (step_let_beta Hval2')).

      simp_sub_lm.
      #2 eapply plus_step.
        exact (step_let_beta val_lam).

        simp_sub_lm.
        #2 eapply star_plus_trans.
          apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (var 1))) m2')); eauto with lambda; done.

          #2 eapply plus_step.
            exact (step_app1 (step_let_beta Hval1')).
  
            simp_sub_lm.
            eexists; split.
              exact (step_let_beta Hval2').
  
              simp_sub_lm.
              apply star_refl; done.

  intros m1 m2 _ _ Hembed1 Hembed2 <-.
  so (tembed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (tembed_val _#3 Hembed2 Hval2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists m1'.
  split; auto; [].
  #2 eapply plus_step.
    exact (step_beta val_lam).

    simp_sub_lm.
    #2 eapply star_plus_trans.
      apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (var 1))) m2)); eauto with lambda; done.

      #2 eapply plus_step.
        exact (step_app1 (step_let_beta Hval1')).

        simp_sub_lm.
        #2 eapply star_plus_trans.
          apply (star_map' _ _ (fun m => lm_let m (shift1 m1'))); eauto with lambda; done.

          eexists; split.
            exact (step_let_beta Hval2').

            simp_sub_lm.
            apply star_refl; done.

(* pi21 *)
intros G e e' _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hstep1).
exists (lm_app m1' (lm_lam (lm_lam (var 0)))).
split.
  apply tembed_pi2; auto; done.

  apply (plus_map' _ _ (fun m => lm_app m (lm_lam (lm_lam (var 0))))); auto with lambda; done.

(* pi22 *)
intros G v1 v2 Hval1 Hval2 m Hembed.
invert Hembed; [].
intros n Hembed' <-.
#3 invert Hembed'.
  intros m1 m2 Hembed1 Hembed2 <-.
  so (tembed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (tembed_val _#3 Hembed2 Hval2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists m2'.
  split; auto; [].
  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1))))) (lm_lam (lm_lam (var 0))))); eauto with lambda; done.

    #2 eapply plus_step.
      exact (step_app1 (step_let_beta Hval1')).

      simp_sub_lm.
      #2 eapply star_plus_trans.
        apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1')) (var 1)))) (lm_lam (lm_lam (var 0))))); eauto with lambda; done.

        #2 eapply plus_step.
          exact (step_app1 (step_let_beta Hval2')).

          simp_sub_lm.
          #2 eapply plus_step.
            exact (step_let_beta val_lam).

            simp_sub_lm.
            #2 eapply plus_step.
              exact (step_app1 (step_let_beta Hval1')).

              simp_sub_lm.
              eexists; split.
                exact (step_let_beta Hval2').

                simp_sub_lm.
                apply star_refl; done.

  intros m1 m2 _ Hembed1 Hembed2 <-.
  so (tembed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (tembed_val _#3 Hembed2 Hval2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists m2'.
  split; auto; [].
  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1)) (var 1)))) (lm_lam (lm_lam (var 0))))); eauto with lambda; done.

    #2 eapply plus_step.
      exact (step_app1 (step_let_beta Hval2')).

      simp_sub_lm.
      #2 eapply plus_step.
        exact (step_let_beta val_lam).

        simp_sub_lm.
        #2 eapply star_plus_trans.
          apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (var 0))) m2')); eauto with lambda; done.

          #2 eapply plus_step.
            exact (step_app1 (step_let_beta Hval1')).
  
            simp_sub_lm.
            eexists; split.
              exact (step_let_beta Hval2').
  
              simp_sub_lm.
              apply star_refl; done.

  intros m1 m2 _ _ Hembed1 Hembed2 <-.
  so (tembed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (tembed_val _#3 Hembed2 Hval2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists m2'.
  split; auto; [].
  #2 eapply plus_step.
    exact (step_beta val_lam).

    simp_sub_lm.
    #2 eapply star_plus_trans.
      apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (var 0))) m2)); eauto with lambda; done.

      #2 eapply plus_step.
        exact (step_app1 (step_let_beta Hval1')).

        simp_sub_lm.
        #2 eapply star_plus_trans.
          apply (star_map' _ _ (fun m => lm_let m (var 0))); eauto with lambda; done.

          eexists; split.
            exact (step_let_beta Hval2').

            simp_sub_lm.
            apply star_refl; done.

(* papp1 *)
intros G e e' c _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hstep1).
exists (lm_app m1' lm_i).
split; auto with embed; [].
apply (plus_map' _ _ (fun m => lm_app m lm_i)); auto with lambda; done.

(* papp2 *)
intros G k e c m Hembed.
invert Hembed; [].
intros n Hembed' <-.
invert Hembed'; [].
intros m1 Hembed1 <-.
exists (subst1 lm_i m1).
split.
  eapply subst_tembed; eauto; [].
  apply subrel_dot; eauto using subrel_sh with sequence; [].
  simp_sub.
  apply embed_ucn; auto; done.

  eexists; split.
    exact (step_beta val_lam).

    simp_sub_lm.
    apply star_refl; done.

(* pack *)
intros G c e e' k t Hstep IH m Hembed.
#2 invert Hembed.
  intros m1 Hembed1 <-.
  so (IH _ Hembed1) as (m1' & Hembed1' & Hstep1).
  exists (lm_let m1' (lm_lam (var 1))).
  split.
    apply tembed_pack; auto; done.

    apply (plus_map' _ _ (fun m => lm_let m (lm_lam (var 1)))); auto with lambda; done.

  intros _ Hval _ _.
  destruct (tstep_value_contra _#3 Hstep Hval); done.

(* unpack1 *)
intros G e1 e1' e2 _ IH m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hstep1).
exists (lm_let (lm_app m1' lm_i) (subst (dot (var 0) (dot lm_i sh1)) m2)).
split.
  apply tembed_unpack; auto; done.

  apply (plus_map' _ _ (fun m => lm_let (lm_app m lm_i) (subst (dot (var 0) (dot lm_i sh1)) m2))); auto with lambda; done.

(* unpack2 *)
intros G c v k t e Hval m Hembed.
invert Hembed; [].
intros n m2 Hembed' Hembed2 <-.
#2 invert Hembed'.
  intros m1 Hembed1 <-.
  so (tembed_val _#3 Hembed1 Hval) as (m1' & Hembed1' & Hval' & Hstep1).
  exists (subst (dot m1' (dot lm_i id)) m2).
  split.
    eapply subst_tembed; eauto; [].
    #2 apply subrel_dot.
      apply subrel_dot; eauto using subrel_sh with sequence; [].
      simp_sub.
      apply embed_ucn; done.

      simp_sub.
      apply embed_utm; auto; done.

    #2 eapply star_plus_trans.
      apply (star_map' _ _ (fun m => lm_let (lm_app (lm_let m (lm_lam (var 1))) lm_i) (subst (dot (var 0) (dot lm_i sh1)) m2))); eauto with lambda; done.

      #2 eapply plus_step.
        exact (step_let (step_app1 (step_let_beta Hval'))).

        simp_sub_lm.
        #2 eapply plus_step.
          exact (step_let (step_let_beta val_i)).

          simp_sub_lm.
          eexists; split.
            exact (step_let_beta Hval').

            simp_sub_lm.
            apply star_refl; done.

  intros m1 _ Hembed1 <-.
  so (tembed_val _#3 Hembed1 Hval) as (m1' & Hembed1' & Hval' & Hstep1).
  exists (subst (dot m1' (dot lm_i id)) m2).
  split.
    eapply subst_tembed; eauto; [].
    #2 apply subrel_dot.
      apply subrel_dot; eauto using subrel_sh with sequence; [].
      simp_sub.
      apply embed_ucn; done.

      simp_sub.
      apply embed_utm; auto; done.

    #2 eapply plus_step.
      exact (step_let (step_beta val_lam)).

      simp_sub_lm.
      #2 eapply star_plus_trans.
        apply (star_map' _ _ (fun m => lm_let m (subst (dot (var 0) (dot lm_i sh1)) m2))); eauto with lambda; done.

        eexists; split.
          exact (step_let_beta Hval').

          simp_sub_lm.
          apply star_refl; done.

(* fix *)
intros G t e m Hembed.      
invert Hembed; [].
intros n Hembed' <-.
exists (lm_app n (lm_lam (lm_app lm_theta (lm_lam (lm_app (subst (sh 2) n) (var 0)))))).
split.
  apply tembed_app; auto; [].
  apply tembed_lam; [].
  simp_sub.
  #2 sor (refine (subst_tembed (G; cl_utm) _ sh1 sh1 _ _ _ Hembed)) as H.
    apply subrel_sh; auto with sequence; done.

    simp_sub in H.
    unfold shift1 in H; rewrite -> subst_theta in H.
    assumption.

  #2 eapply plus_star_trans.
    apply step_theta; auto with lambda; done.

    #2 eapply star_step.
      eauto with lambda; done.

      simp_sub.
      apply star_refl; done.

(* lett1 *)
intros G e1 e1' e2 _ IH m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hstep1).
exists (lm_let m1' m2).
split.
  auto with embed; done.

  apply (plus_map' _ _ (fun m => lm_let m m2)); auto with lambda; done.

(* lett2 *)
intros G v1 e2 Hval m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (tembed_val _#3 Hembed1 Hval) as (m1' & Hembed1' & Hval' & Hstep1).
exists (subst1 m1' m2).
split.
  eapply subst_tembed; eauto; [].
  apply subrel_dot; eauto using subrel_sh with sequence; [].
  simp_sub.
  apply embed_utm; auto; done.

  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_let m m2)); eauto with lambda; done.

    eexists; split.
      exact (step_let_beta Hval').

      apply star_refl; done.

(* letm1 *)
intros G M M' e _ IH m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hsteps).
eexists; split.
  eauto with embed; done.

  apply (plus_map' _ _ (fun m => lm_let m _)); eauto with lambda; done.

(* letm2 *)
intros G V e c Hval Hfst m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (membed_val _#3 Hembed1 Hval) as (m1' & Hembed1' & Hval' & Hsteps1).
exists (subst (dot m1' (dot lm_i id)) m2).
split.
  refine (subst_tembed _#6 _ Hembed2); [].
  #2 apply subrel_dot.
    apply subrel_dot; eauto using subrel_sh with sequence; [].
    simp_sub.
    apply embed_ucn; eauto; done.

    simp_sub.
    apply embed_umd; auto; done.

  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_let m _)); eauto with lambda; done.

    eexists; split.
      apply step_let_beta; auto; done.

      simp_sub_lm.
      apply star_refl; done.

(* ext1 *)
intros G M M' _ IH m Hembed.    
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hsteps).
eexists.
split.
  eauto with embed; done.

  apply (plus_map' _ _ (fun m => lm_let m _)); eauto with lambda; done.

(* ext2 *)
intros G v Hval m Hembed.
invert Hembed; [].
intros m1 Hembed' <-.
invert Hembed'; [].
intros Hembed1.
so (tembed_val _#3 Hembed1 Hval) as (m1' & Hembed1' & Hval' & Hsteps).
exists m1'.
split.
  assumption.

  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_let m (var 0))); eauto with lambda; done.

    eexists; split.
      apply step_let_beta; auto; done.

      simp_sub.
      apply star_refl; done.

(* dyn *)
intros G e e' _ IH m Hembed.
invert Hembed; [].
intros Hembed'.
so (IH _ Hembed') as (m' & Hembed'' & Hsteps).
eauto with embed; done.

(* appgn1 *)
intros G M1 M1' M2 _ IH m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hstep).
exists (lm_app m1' m2).
split; eauto with embed.
apply (plus_map' _ _ (fun m => lm_app m m2)); auto with lambda; done.

(* appgn2 *)
intros G V1 M2 M2' Hval _ IH m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (membed_val _#3 Hembed1 Hval) as (m1' & Hembed1' & Hval' & Hstep1).
so (IH _ Hembed2) as (m2' & Hembed2' & Hstep2).
exists (lm_app m1' m2').
split; auto with embed; [].
#2 apply (star_plus_trans _#3 (lm_app m1' m2)).
  apply (star_map' _ _ (fun m => lm_app m m2)); auto with lambda; done.

  apply (plus_map' _ _ (fun m => lm_app m1' m)); auto with lambda; done.

(* appgn3 *)
intros G T M1 V2 c2 Hval Hfst m Hembed.
invert Hembed; [].
intros n m2 Hembed' Hembed2 <-.
invert Hembed'; [].
intros m1 Hembed1 <-.
so (membed_val _#3 Hembed2 Hval) as (m2' & Hembed2' & Hval' & Hstep).
exists (subst (dot m2' (dot lm_i id)) m1).
split.
  refine (subst_membed _#6 _ Hembed1); [].
  #2 apply subrel_dot.
    apply subrel_dot; eauto using subrel_sh with sequence; [].
    simp_sub.
    apply embed_ucn; eauto; done.

    simp_sub.
    apply embed_umd; auto; done.

  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_app _ m)); eauto with lambda; done.
    
    eexists; split.
      apply step_beta; auto; done.

      simp_sub.
      apply star_refl; done.

(* appap1 *)
intros G M1 M1' M2 _ IH m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hstep).
exists (lm_app m1' m2).
split; eauto with embed.
apply (plus_map' _ _ (fun m => lm_app m m2)); auto with lambda; done.

(* appap2 *)
intros G V1 M2 M2' Hval _ IH m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (membed_val _#3 Hembed1 Hval) as (m1' & Hembed1' & Hval' & Hstep1).
so (IH _ Hembed2) as (m2' & Hembed2' & Hstep2).
exists (lm_app m1' m2').
split; auto with embed; [].
#2 apply (star_plus_trans _#3 (lm_app m1' m2)).
  apply (star_map' _ _ (fun m => lm_app m m2)); auto with lambda; done.

  apply (plus_map' _ _ (fun m => lm_app m1' m)); auto with lambda; done.

(* appap3 *)
intros G T M1 V2 c2 Hval Hfst m Hembed.
invert Hembed; [].
intros n m2 Hembed' Hembed2 <-.
invert Hembed'; [].
intros m1 Hembed1 <-.
so (membed_val _#3 Hembed2 Hval) as (m2' & Hembed2' & Hval' & Hstep).
exists (subst (dot m2' (dot lm_i id)) m1).
split.
  refine (subst_membed _#6 _ Hembed1); [].
  #2 apply subrel_dot.
    apply subrel_dot; eauto using subrel_sh with sequence; [].
    simp_sub.
    apply embed_ucn; eauto; done.

    simp_sub.
    apply embed_umd; auto; done.

  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_app _ m)); eauto with lambda; done.
    
    eexists; split.
      apply step_beta; auto; done.

      simp_sub.
      apply star_refl; done.

(* pair1 *)
intros G M1 M1' M2 Hstep IH m Hembed.
#3 invert Hembed.
  intros m1 m2 Hembed1 Hembed2 <-.
  so (IH _ Hembed1) as (m1' & Hembed1' & Hstep').
  exists (lm_let m1' (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1))))).
  split.
    apply membed_pair; auto; done.

    apply (plus_map' _ _ (fun m => lm_let m (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1)))))); auto with lambda; done.

  intros _ _ Hval1 _ _ _.
  destruct (mstep_value_contra _#3 Hstep Hval1); done.

  intros _ _ Hval1 _ _ _ _.
  destruct (mstep_value_contra _#3 Hstep Hval1); done.

(* pair2 *)
intros G V1 M2 M2' Hval1 Hstep IH m Hembed.
#3 invert Hembed.
  intros m1 m2 Hembed1 Hembed2 <-.
  so (membed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (IH _ Hembed2) as (m2' & Hembed2' & Hstep2).
  exists (lm_let m2' (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1')) (var 1)))).
  split.
    apply membed_pair_val1; auto; done.

    #2 eapply star_plus_trans.
      apply (star_map' _ _ (fun m => lm_let m (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1)))))); eauto with lambda; done.

      #2 eapply plus_step.
        apply step_let_beta; auto; done.

        simp_sub_lm.
        apply (plus_map' _ _ (fun m => lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1')) (var 1))))); auto with lambda; done.

  intros m1 m2 _ Hembed1 Hembed2 <-.
  so (IH _ Hembed2) as (m2' & Hembed2' & Hstep').
  exists (lm_let m2' (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1)) (var 1)))).
  split.
    apply membed_pair_val1; auto; done.

    apply (plus_map' _ _ (fun m => lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1)) (var 1))))); auto with lambda; done.

  intros _ _ _ Hval2 _ _ _.
  destruct (mstep_value_contra _#3 Hstep Hval2); done.

(* pi11 *)
intros G M M' _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hstep1).
exists (lm_app m1' (lm_lam (lm_lam (var 1)))).
split.
  apply membed_pi1; auto; done.

  apply (plus_map' _ _ (fun m => lm_app m (lm_lam (lm_lam (var 1))))); auto with lambda; done.

(* pi12 *)
intros G V1 V2 Hval1 Hval2 m Hembed.
invert Hembed; [].
intros n Hembed' <-.
#3 invert Hembed'.
  intros m1 m2 Hembed1 Hembed2 <-.
  so (membed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (membed_val _#3 Hembed2 Hval2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists m1'.
  split; auto; [].
  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1))))) (lm_lam (lm_lam (var 1))))); eauto with lambda; done.

    #2 eapply plus_step.
      exact (step_app1 (step_let_beta Hval1')).

      simp_sub_lm.
      #2 eapply star_plus_trans.
        apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1')) (var 1)))) (lm_lam (lm_lam (var 1))))); eauto with lambda; done.

        #2 eapply plus_step.
          exact (step_app1 (step_let_beta Hval2')).

          simp_sub_lm.
          #2 eapply plus_step.
            exact (step_let_beta val_lam).

            simp_sub_lm.
            #2 eapply plus_step.
              exact (step_app1 (step_let_beta Hval1')).

              simp_sub_lm.
              eexists; split.
                exact (step_let_beta Hval2').

                simp_sub_lm.
                apply star_refl; done.

  intros m1 m2 _ Hembed1 Hembed2 <-.
  so (membed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (membed_val _#3 Hembed2 Hval2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists m1'.
  split; auto; [].
  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1)) (var 1)))) (lm_lam (lm_lam (var 1))))); eauto with lambda; done.

    #2 eapply plus_step.
      exact (step_app1 (step_let_beta Hval2')).

      simp_sub_lm.
      #2 eapply plus_step.
        exact (step_let_beta val_lam).

        simp_sub_lm.
        #2 eapply star_plus_trans.
          apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (var 1))) m2')); eauto with lambda; done.

          #2 eapply plus_step.
            exact (step_app1 (step_let_beta Hval1')).
  
            simp_sub_lm.
            eexists; split.
              exact (step_let_beta Hval2').
  
              simp_sub_lm.
              apply star_refl; done.

  intros m1 m2 _ _ Hembed1 Hembed2 <-.
  so (membed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (membed_val _#3 Hembed2 Hval2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists m1'.
  split; auto; [].
  #2 eapply plus_step.
    exact (step_beta val_lam).

    simp_sub_lm.
    #2 eapply star_plus_trans.
      apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (var 1))) m2)); eauto with lambda; done.

      #2 eapply plus_step.
        exact (step_app1 (step_let_beta Hval1')).

        simp_sub_lm.
        #2 eapply star_plus_trans.
          apply (star_map' _ _ (fun m => lm_let m (shift1 m1'))); eauto with lambda; done.

          eexists; split.
            exact (step_let_beta Hval2').

            simp_sub_lm.
            apply star_refl; done.

(* pi21 *)
intros G M M' _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hstep1).
exists (lm_app m1' (lm_lam (lm_lam (var 0)))).
split.
  apply membed_pi2; auto; done.

  apply (plus_map' _ _ (fun m => lm_app m (lm_lam (lm_lam (var 0))))); auto with lambda; done.

(* pi22 *)
intros G V1 V2 Hval1 Hval2 m Hembed.
invert Hembed; [].
intros n Hembed' <-.
#3 invert Hembed'.
  intros m1 m2 Hembed1 Hembed2 <-.
  so (membed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (membed_val _#3 Hembed2 Hval2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists m2'.
  split; auto; [].
  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_let (shift1 m2) (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1))))) (lm_lam (lm_lam (var 0))))); eauto with lambda; done.

    #2 eapply plus_step.
      exact (step_app1 (step_let_beta Hval1')).

      simp_sub_lm.
      #2 eapply star_plus_trans.
        apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1')) (var 1)))) (lm_lam (lm_lam (var 0))))); eauto with lambda; done.

        #2 eapply plus_step.
          exact (step_app1 (step_let_beta Hval2')).

          simp_sub_lm.
          #2 eapply plus_step.
            exact (step_let_beta val_lam).

            simp_sub_lm.
            #2 eapply plus_step.
              exact (step_app1 (step_let_beta Hval1')).

              simp_sub_lm.
              eexists; split.
                exact (step_let_beta Hval2').

                simp_sub_lm.
                apply star_refl; done.

  intros m1 m2 _ Hembed1 Hembed2 <-.
  so (membed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (membed_val _#3 Hembed2 Hval2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists m2'.
  split; auto; [].
  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1)) (var 1)))) (lm_lam (lm_lam (var 0))))); eauto with lambda; done.

    #2 eapply plus_step.
      exact (step_app1 (step_let_beta Hval2')).

      simp_sub_lm.
      #2 eapply plus_step.
        exact (step_let_beta val_lam).

        simp_sub_lm.
        #2 eapply star_plus_trans.
          apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (var 0))) m2')); eauto with lambda; done.

          #2 eapply plus_step.
            exact (step_app1 (step_let_beta Hval1')).
  
            simp_sub_lm.
            eexists; split.
              exact (step_let_beta Hval2').
  
              simp_sub_lm.
              apply star_refl; done.

  intros m1 m2 _ _ Hembed1 Hembed2 <-.
  so (membed_val _#3 Hembed1 Hval1) as (m1' & Hembed1' & Hval1' & Hstep1).
  so (membed_val _#3 Hembed2 Hval2) as (m2' & Hembed2' & Hval2' & Hstep2).
  exists m2'.
  split; auto; [].
  #2 eapply plus_step.
    exact (step_beta val_lam).

    simp_sub_lm.
    #2 eapply star_plus_trans.
      apply (star_map' _ _ (fun m => lm_app (lm_let m (lm_lam (var 0))) m2)); eauto with lambda; done.

      #2 eapply plus_step.
        exact (step_app1 (step_let_beta Hval1')).

        simp_sub_lm.
        #2 eapply star_plus_trans.
          apply (star_map' _ _ (fun m => lm_let m (var 0))); eauto with lambda; done.

          eexists; split.
            exact (step_let_beta Hval2').

            simp_sub_lm.
            apply star_refl; done.

(* unpack1 *)
intros G e e' M T _ IH m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hstep1).
exists (lm_let (lm_app m1' lm_i) (subst (dot (var 0) (dot lm_i sh1)) m2)).
split.
  apply membed_unpack; auto; done.

  apply (plus_map' _ _ (fun m => lm_let (lm_app m lm_i) (subst (dot (var 0) (dot lm_i sh1)) m2))); auto with lambda; done.

(* unpack2 *)
intros G c v k t M T Hval m Hembed.
invert Hembed; [].
intros n m2 Hembed' Hembed2 <-.
#2 invert Hembed'.
  intros m1 Hembed1 <-.
  so (tembed_val _#3 Hembed1 Hval) as (m1' & Hembed1' & Hval' & Hstep1).
  exists (subst (dot m1' (dot lm_i id)) m2).
  split.
    eapply subst_membed; eauto; [].
    #2 apply subrel_dot.
      apply subrel_dot; eauto using subrel_sh with sequence; [].
      simp_sub.
      apply embed_ucn; done.

      simp_sub.
      apply embed_utm; auto; done.

    #2 eapply star_plus_trans.
      apply (star_map' _ _ (fun m => lm_let (lm_app (lm_let m (lm_lam (var 1))) lm_i) (subst (dot (var 0) (dot lm_i sh1)) m2))); eauto with lambda; done.

      #2 eapply plus_step.
        exact (step_let (step_app1 (step_let_beta Hval'))).

        simp_sub_lm.
        #2 eapply plus_step.
          exact (step_let (step_let_beta val_i)).

          simp_sub_lm.
          eexists; split.
            exact (step_let_beta Hval').

            simp_sub_lm.
            apply star_refl; done.

  intros m1 _ Hembed1 <-.
  so (tembed_val _#3 Hembed1 Hval) as (m1' & Hembed1' & Hval' & Hstep1).
  exists (subst (dot m1' (dot lm_i id)) m2).
  split.
    eapply subst_membed; eauto; [].
    #2 apply subrel_dot.
      apply subrel_dot; eauto using subrel_sh with sequence; [].
      simp_sub.
      apply embed_ucn; done.

      simp_sub.
      apply embed_utm; auto; done.

    #2 eapply plus_step.
      exact (step_let (step_beta val_lam)).

      simp_sub_lm.
      #2 eapply star_plus_trans.
        apply (star_map' _ _ (fun m => lm_let m (subst (dot (var 0) (dot lm_i sh1)) m2))); eauto with lambda; done.

        eexists; split.
          exact (step_let_beta Hval').

          simp_sub_lm.
          apply star_refl; done.

(* lett1 *)
intros G e1 e1' M2 _ IH m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hstep1).
exists (lm_let m1' m2).
split.
  auto with embed; done.

  apply (plus_map' _ _ (fun m => lm_let m m2)); auto with lambda; done.

(* lett2 *)
intros G v1 M2 Hval m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (tembed_val _#3 Hembed1 Hval) as (m1' & Hembed1' & Hval' & Hstep1).
exists (subst1 m1' m2).
split.
  eapply subst_membed; eauto; [].
  apply subrel_dot; eauto using subrel_sh with sequence; [].
  simp_sub.
  apply embed_utm; auto; done.

  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_let m m2)); eauto with lambda; done.

    eexists; split.
      exact (step_let_beta Hval').

      apply star_refl; done.

(* letm1 *)
intros G M1 M1' M2 T _ IH m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hsteps).
eexists; split.
  eauto with embed; done.

  apply (plus_map' _ _ (fun m => lm_let m _)); eauto with lambda; done.

(* letm2 *)
intros G V M T c Hval Hfst m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (membed_val _#3 Hembed1 Hval) as (m1' & Hembed1' & Hval' & Hsteps1).
exists (subst (dot m1' (dot lm_i id)) m2).
split.
  refine (subst_membed _#6 _ Hembed2); [].
  #2 apply subrel_dot.
    apply subrel_dot; eauto using subrel_sh with sequence; [].
    simp_sub.
    apply embed_ucn; eauto; done.

    simp_sub.
    apply embed_umd; auto; done.

  #2 eapply star_plus_trans.
    apply (star_map' _ _ (fun m => lm_let m _)); eauto with lambda; done.

    eexists; split.
      apply step_let_beta; auto; done.

      simp_sub_lm.
      apply star_refl; done.

(* seal *)
intros G M T m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
exists m1.
split.
  assumption.

  eexists; split.
    apply step_let_beta; eauto with lambda; done.

    simp_sub_lm.
    apply star_refl; done.

(* epilogue *)
destruct_all Hind.
eauto 8; done.
Qed.


Lemma tembed_step :
  forall G e e' m,
    tembed G e m
    -> tstep G e e'
    -> exists m',
         tembed G e' m'
         /\ plus step m m'.
Proof.
exact (embeds_step andel).
Qed.


Lemma membed_step :
  forall G M M' m,
    membed G M m
    -> mstep G M M'
    -> exists m',
         membed G M' m'
         /\ plus step m m'.
Proof.
exact (embeds_step ander).
Qed.


Lemma tembed_halt :
  forall G e m,
    tembed G e m
    -> thalt G e
    -> halt m.
Proof.
intros G e m Hembed Hhalt.
destruct Hhalt as (v & Hstep & Hval).
unfold halt.
revert m Hembed Hval.
induct Hstep.

(* refl *)
intros e m Hembed Hval.
so (tembed_val _#3 Hembed Hval) as (m' & Hembed' & Hval' & Hstep).
eauto; done.

(* step *)
intros e1 e2 v Hstep _ IH m1 Hembed Hval.
so (tembed_step _#4 Hembed Hstep) as (m2 & Hembed2 & Hstep').
so (IH _ Hembed2 Hval) as (m3 & Hstep'' & Hval').
exists m3.
split; auto; [].
apply plus_star; [].
eapply plus_star_trans; eauto; done.
Qed.


Lemma membed_halt :
  forall G M m,
    membed G M m
    -> mhalt G M
    -> halt m.
Proof.
intros G M m Hembed Hhalt.
destruct Hhalt as (v & Hstep & Hval).
unfold halt.
revert m Hembed Hval.
induct Hstep.

(* refl *)
intros M m Hembed Hval.
so (membed_val _#3 Hembed Hval) as (m' & Hembed' & Hval' & Hstep).
eauto; done.

(* step *)
intros M1 M2 v Hstep _ IH m1 Hembed Hval.
so (membed_step _#4 Hembed Hstep) as (m2 & Hembed2 & Hstep').
so (IH _ Hembed2 Hval) as (m3 & Hstep'' & Hval').
exists m3.
split; auto; [].
apply plus_star; [].
eapply plus_star_trans; eauto; done.
Qed.


Lemma step_shorter :
  forall i j m m' n,
    expon step i m n
    -> val n
    -> expon step j m m'
    -> j <= i /\ expon step (i-j) m' n.
Proof.
intros i j m p n Hmn Hval Hmp.
revert i Hmn.
induct Hmp.

(* 0 *)
intros m i Hmn.
split.
  omega.

  repl (i-0) with i by omega.
  assumption.

(* S *)
intros j m q p Hmq Hqp IH i Hmn.
#2 invert Hmn.
  intros <- <-.
  exfalso.
  eapply step_val_contra; eauto; done.

  intros i' q' Hmq' Hqn <-.
  so (step_determ _ _ _ Hmq Hmq').
  subst q'.
  so (IH _ Hqn) as (Hlt & Hpn).
  split.
    omega.

    repl (S i' - S j) with (i' - j) by omega.
    assumption.
Qed.


Lemma tembed_halt_conv :
  forall e t m,
    oft nil e t
    -> tembed nil e m
    -> halt m
    -> thalt nil e.
Proof.
intros e t m Hof Hembed Hhalt.
destruct Hhalt as (n & Hstepm & Hval).
unfold thalt.
so (star_expon _#4 Hstepm) as (i & Hstepmexp).
renameover Hstepmexp into Hstepm.
revert e m n Hof Hembed Hval Hstepm.
wfinduct i using Wf_nat.lt_wf.
intros i IH e m n Hof Hembed Hval Hstepm.
so (Progress.progress_tm _ _ Hof) as [Hval' | (e' & Hstepe)].
  exists e.
  auto using star_refl; done.

  so (Preservation.preservation_tm _#4 wf_nil Hof Hstepe) as Hof'.
  so (tembed_step _#4 Hembed Hstepe) as (m' & Hembed' & Hstepm').
  so (plus_expon _#4 Hstepm') as (j & Hstepm'j).
  so (step_shorter _#5 Hstepm Hval Hstepm'j) as (Hlt & Hstepm'').
  assert (i - S j < i) as Hlt' by omega.
  so (IH _ Hlt' _#3 Hof' Hembed' Hval Hstepm'') as (v & Hstepe' & Hvalv).
  eauto using star_step; done.
Qed.


Lemma membed_halt_conv :
  forall M p t m,
    ofm nil p M t
    -> membed nil M m
    -> halt m
    -> mhalt nil M.
Proof.
intros M p t m Hof Hembed Hhalt.
destruct Hhalt as (n & Hstepm & Hval).
unfold thalt.
so (star_expon _#4 Hstepm) as (i & Hstepmexp).
renameover Hstepmexp into Hstepm.
revert M m n Hof Hembed Hval Hstepm.
wfinduct i using Wf_nat.lt_wf.
intros i IH M m n Hof Hembed Hval Hstepm.
so (Progress.progress_md _ _ _ Hof) as [Hval' | (M' & HstepM)].
  exists M.
  auto using star_refl; done.

  so (Preservation.preservation_md _#5 wf_nil Hof HstepM) as Hof'.
  so (membed_step _#4 Hembed HstepM) as (m' & Hembed' & Hstepm').
  so (plus_expon _#4 Hstepm') as (j & Hstepm'j).
  so (step_shorter _#5 Hstepm Hval Hstepm'j) as (Hlt & Hstepm'').
  assert (i - S j < i) as Hlt' by omega.
  so (IH _ Hlt' _#3 Hof' Hembed' Hval Hstepm'') as (V & Hstepe' & Hvalv).
  unfold mhalt.
  eauto using star_step; done.
Qed.


Lemma treduce_value :
  forall G e e',
    treduce G e e'
    -> tvalue G e
    -> tvalue G e'.
Proof.
intros G e e' Hred Hval.
revert Hval.
induct Hred; try (intros; invert Hval; intros; subst; eauto with dynamic; done).
Qed.


Lemma mreduce_value :
  forall G M M',
    mreduce G M M'
    -> mvalue G M
    -> mvalue G M'.
Proof.
intros G M M' Hred Hval.
revert Hval.
induct Hred; try (intros; invert Hval; intros; subst; eauto using treduce_value with dynamic; done).
Qed.


Lemma treduces_value :
  forall G e e',
    star (treduce G) e e'
    -> tvalue G e
    -> tvalue G e'.
Proof.
intros G e e' Hred Hval.
revert Hval.
induct Hred; intros; eauto using treduce_value.
Qed.


Lemma mreduces_value :
  forall G M M',
    star (mreduce G) M M'
    -> mvalue G M
    -> mvalue G M'.
Proof.
intros G e e' Hred Hval.
revert Hval.
induct Hred; intros; eauto using mreduce_value.
Qed.


Lemma embeds_reduce :
  (forall G e e' m,
     tembed G e m
     -> treduce G e e'
     -> exists m',
          tembed G e' m'
          /\ star reduce m m')
  /\
  (forall G M M' m,
     membed G M m
     -> mreduce G M M'
     -> exists m',
          membed G M' m'
          /\ star reduce m m').
Proof.
exploit
  (breduce_ind
     (fun G e e' =>
        forall m,
          tembed G e m
          -> exists m',
               tembed G e' m'
               /\ star reduce m m')
     (fun G M M' =>
        forall m,
          membed G M m
          -> exists m',
               membed G M' m'
               /\ star reduce m m')) as Hind.

(* refl *)
intros G e m Hembed.
eauto using star_refl; done.

(* trans *)
intros G e1 e2 e3 _ IH1 _ IH2 m Hembed1.
so (IH1 _ Hembed1) as (m2 & Hembed2 & Hred1).
so (IH2 _ Hembed2) as (m3 & Hembed3 & Hred2).
eauto using star_trans; done.

(* lam *)
intros G t t' e e' _ IH m Hembed.
invert Hembed; [].
intros n Hembed' <-.
so (IH _ Hembed') as (n' & Hembed'' & Hred).
exists (lm_lam n').
split.
  auto with embed; done.

  apply (star_map' _ _ lm_lam); auto with lambda; done.

(* app1 *)
intros G e1 e1' e2 e2' _ IH1 _ IH2 m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1).
so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2).
exists (lm_app m1' m2').
split.
  eauto with embed; done.

  #2 eapply star_trans.
    apply (star_map' _ _ (fun m => lm_app m m2)); eauto with lambda embed; done.

    apply (star_map' _ _ (fun m => lm_app m1' m)); eauto with lambda embed; done.

(* app2 *)
intros G t e1 v2 Hval m Hembed.
so (tembed_step _#4 Hembed (tstep_app3 Hval)) as (m' & Hembed' & Hstep).
eauto 6 using tembed_hlm, steps_reduces, plus_star; done.

(* pair *)    
intros G e1 e1' e2 e2' Hrede1 IH1 Hrede2 IH2 m Hembed.
#3 invert Hembed.
  intros m1 m2 Hembed1 Hembed2 <-.
  so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1).
  so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2).
  exists (lm_let m1' (lm_let (shift1 m2') (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1))))).
  split.
    eauto with embed; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun m => lm_let m _)); eauto 9 using shift_hlm with embed lambda; done.

      apply (star_map' _ _ (fun m => lm_let _ (lm_let (shift1 m) _))); eauto 9 using shift_reduce with embed lambda; done.

  intros m1 m2 Hval1 Hembed1 Hembed2 <-.
  so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1').
  so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2').
  exists (lm_let m2' (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1')) (var 1)))).
  split.
    eauto using treduce_value with embed; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun m => lm_let m _)); eauto 8 using shift_hlm with embed lambda; done.

      apply (star_map' _ _ (fun m => lm_let _ (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m)) (var 1))))); eauto 8 using shift_reduce with embed lambda; done.

  intros m1 m2 Hval1 Hval2 Hembed1 Hembed2 <-.
  so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1').
  so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2').
  exists (lm_lam (lm_app (lm_app (var 0) (shift1 m1')) (shift1 m2'))).
  split.
    eauto using treduce_value with embed; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun m => lm_lam (lm_app (lm_app (var 0) (shift1 m)) (shift1 m2)))); eauto 8 using shift_reduce, shift_hlm with embed lambda; done.

      apply (star_map' _ _ (fun m => lm_lam (lm_app (lm_app (var 0) (shift1 m1')) (shift1 m)))); eauto 8 using shift_reduce, shift_hlm with embed lambda; done.

(* pi11 *)
intros G e e' _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hred).
exists (lm_app m1' (lm_lam (lm_lam (var 1)))).
split.
  eauto with embed; done.

  apply (star_map' _ _ (fun m => lm_app m _)); eauto with embed lambda; done.

(* pi12 *)
intros G v1 v2 Hval1 Hval2 m Hembed.
so (tembed_step _#4 Hembed (tstep_pi12 Hval1 Hval2)) as (m' & Hembed' & Hstep).
eauto 6 using tembed_hlm, steps_reduces, plus_star; done.

(* pi21 *)
intros G e e' _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hred).
exists (lm_app m1' (lm_lam (lm_lam (var 0)))).
split.
  eauto with embed; done.

  apply (star_map' _ _ (fun m => lm_app m (lm_lam (lm_lam (var 0))))); eauto with embed lambda; done.

(* pi22 *)
intros G v1 v2 Hval1 Hval2 m Hembed.
so (tembed_step _#4 Hembed (tstep_pi22 Hval1 Hval2)) as (m' & Hembed' & Hstep).
eauto 6 using tembed_hlm, steps_reduces, plus_star; done.

(* plam *)
intros G k k' e1 e1' _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hred).
exists (lm_lam m1').
split.
  eauto with embed; done.

  apply (star_map' _ _ lm_lam); eauto with embed lambda; done.

(* papp1 *)
intros G e1 e1' c c' _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hred).
exists (lm_app m1' lm_i).
split.
  eauto with embed; done.

  apply (star_map' _ _ (fun m => lm_app m lm_i)); eauto with embed lambda; done.

(* papp2 *)
intros G k e c m Hembed.
so (tembed_step _#4 Hembed tstep_papp2) as (m' & Hembed' & Hstep).
eauto 6 using tembed_hlm, steps_reduces, plus_star; done.

(* pack *)
intros G c c' e1 e1' k k' t t' Hred IH m Hembed.
#2 invert Hembed.
  intros m1 Hembed1 <-.
  so (IH _ Hembed1) as (m1' & Hembed1' & Hred1).
  exists (lm_let m1' (lm_lam (var 1))).
  split.
    eauto with embed; done.

    apply (star_map' _ _ (fun m => lm_let m _)); eauto with embed lambda; done.

  intros m1 Hval Hembed1 <-.
  so (IH _ Hembed1) as (m1' & Hembed1' & Hred1).
  exists (lm_lam (shift1 m1')).
  split.
    eauto using treduce_value with embed; done.

    apply (star_map' _ _ (fun m => lm_lam (shift1 m))); eauto using shift_reduce with embed lambda; done.

(* unpack1 *)
intros G e1 e1' e2 e2' _ IH1 _ IH2 m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1).
so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2).
exists (lm_let (lm_app m1' lm_i) (subst (dot (var 0) (dot lm_i sh1)) m2')).
split.
  eauto with embed; done.

  #2 eapply star_trans.
    exploit (subst_under_hlm 1 lm_i m2) as H; eauto with embed lambda; [].
    simp_sub in H.
    apply (star_map' _ _ (fun m => lm_let (lm_app m _) _)); eauto 8 using subst_under_hlm with embed lambda; done.

    repl (dot (var 0) (dot lm_i sh1)) with (under 1 (dot lm_i id)) by (simp_sub; reflexivity).
    apply (star_map' _ _ (fun m => lm_let _ (subst (under 1 (dot lm_i id)) m))); eauto 8 using subst_under_reduce with embed lambda; done.

(* unpack2 *)
intros G c v k t e Hval m Hembed.
so (tembed_step _#4 Hembed (tstep_unpack2 Hval)) as (m' & Hembed' & Hstep).
eauto 6 using tembed_hlm, steps_reduces, plus_star; done.

(* fix1 *)
intros G t t' e1 e1' _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hred1).
exists (lm_app lm_theta (lm_lam (lm_app (shift1 m1') (var 0)))).
split.
  eauto with embed; done.

  assert (hlm lm_theta).
    unfold lm_theta.
    eauto 15 with lambda; done.
  apply (star_map' _ _ (fun m => lm_app _ (lm_lam (lm_app (shift1 m) (var 0))))); eauto 6 using shift_reduce with embed lambda; done.

(* fix2 *)
intros G t e m Hembed.
so (tembed_step _#4 Hembed tstep_fix) as (m' & Hembed' & Hstep).
eauto 6 using tembed_hlm, steps_reduces, plus_star; done.

(* lett1 *)
intros G e1 e1' e2 e2' _ IH1 _ IH2 m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1).
so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2).
exists (lm_let m1' m2').
split.
  eauto with embed; done.

  #2 eapply star_trans.
    apply (star_map' _ _ (fun m => lm_let m m2)); eauto with embed lambda; done.

    apply (star_map' _ _ (fun m => lm_let m1' m)); eauto with embed lambda; done.

(* lett2 *)
intros G v1 e2 Hval m Hembed.
so (tembed_step _#4 Hembed (tstep_lett2 Hval)) as (m' & Hembed' & Hstep).
eauto 6 using tembed_hlm, steps_reduces, plus_star; done.

(* letm1 *)
intros G M M' e e' _ IH1 _ IH2 m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1).
so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2).
exists (lm_let m1' (subst (dot (var 0) (dot lm_i sh1)) m2')).
split.
  eauto with embed; done.

  #2 eapply star_trans.
    exploit (subst_under_hlm 1 lm_i m2) as H; eauto with embed lambda; [].
    apply (star_map' _ _ (fun m => lm_let m _)); eauto with embed lambda; done.

    repl (dot (var 0) (dot lm_i sh1)) with (under 1 (dot lm_i id)) by (simp_sub; reflexivity).
    apply (star_map' _ _ (fun m => lm_let m1' (subst (under 1 (dot lm_i id)) m))); eauto 8 using subst_under_reduce with embed lambda; done.

(* letm2 *)
intros G V e c Hval Hfst m Hembed.
so (tembed_step _#4 Hembed (tstep_letm2 Hval Hfst)) as (m' & Hembed' & Hstep).
eauto 6 using tembed_hlm, steps_reduces, plus_star; done.

(* ext1 *)
intros G M M' _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed' & Hred).
exists (lm_let m1' (var 0)).
split.
  eauto with embed; done.

  apply (star_map' _ _ (fun m => lm_let m (var 0))); eauto with embed lambda; done.

(* ext2 *)
intros G v Hval m Hembed.
so (tembed_step _#4 Hembed (tstep_ext2 Hval)) as (m' & Hembed' & Hstep).
eauto 6 using tembed_hlm, steps_reduces, plus_star; done.

(* refl *)
intros G M m Hembed.
eauto using star_refl; done.

(* trans *)
intros G M1 M2 M3 _ IH1 _ IH2 m Hembed1.
so (IH1 _ Hembed1) as (m2 & Hembed2 & Hred1).
so (IH2 _ Hembed2) as (m3 & Hembed3 & Hred2).
eauto using star_trans; done.

(* stat *)
intros G c c' m Hembed.
invert Hembed; [].
intros <-.
exists lm_i.
split.
  apply membed_stat; done.

  apply star_refl; done.

(* dyn *)
intros G e e' _ IH m Hembed.
invert Hembed; [].
intros Hembed'.
so (IH _ Hembed') as (m' & Hembed'' & Hred).
eauto with embed; done.

(* lamgn *)
intros G T T' M M' _ IH m Hembed.
invert Hembed; [].
intros n Hembed' <-.
so (IH _ Hembed') as (n' & Hembed'' & Hred).
exists (lm_lam (subst (dot (var 0) (dot lm_i sh1)) n')).
split.
  eauto with embed; done.

  repl (dot (var 0) (dot lm_i sh1)) with (under 1 (dot lm_i id)) by (simp_sub; reflexivity).
  apply (star_map' _ _ (fun m => lm_lam (subst (under 1 (dot lm_i id)) m))); eauto using subst_under_reduce with embed lambda; done.

(* appgn1 *)
intros G M1 M1' M2 M2' _ IH1 _ IH2 m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1).
so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2).
exists (lm_app m1' m2').
split.
  eauto with embed; done.

  #2 eapply star_trans.
    apply (star_map' _ _ (fun m => lm_app m m2)); eauto with lambda embed; done.

    apply (star_map' _ _ (fun m => lm_app m1' m)); eauto with lambda embed; done.

(* appgn2 *)
intros G T M1 V2 c2 Hval Hfst m Hembed.
so (membed_step _#4 Hembed (mstep_appgn3 Hval Hfst)) as (m' & Hembed' & Hstep).
eauto 6 using membed_hlm, steps_reduces, plus_star; done.

(* lamap *)
intros G T T' M M' _ IH m Hembed.
invert Hembed; [].
intros n Hembed' <-.
so (IH _ Hembed') as (n' & Hembed'' & Hred).
exists (lm_lam (subst (dot (var 0) (dot lm_i sh1)) n')).
split.
  eauto with embed; done.

  repl (dot (var 0) (dot lm_i sh1)) with (under 1 (dot lm_i id)) by (simp_sub; reflexivity).
  apply (star_map' _ _ (fun m => lm_lam (subst (under 1 (dot lm_i id)) m))); eauto using subst_under_reduce with embed lambda; done.

(* appap1 *)
intros G M1 M1' M2 M2' _ IH1 _ IH2 m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1).
so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2).
exists (lm_app m1' m2').
split.
  eauto with embed; done.

  #2 eapply star_trans.
    apply (star_map' _ _ (fun m => lm_app m m2)); eauto with lambda embed; done.

    apply (star_map' _ _ (fun m => lm_app m1' m)); eauto with lambda embed; done.

(* appap2 *)
intros G T M1 V2 c2 Hval Hfst m Hembed.
so (membed_step _#4 Hembed (mstep_appap3 Hval Hfst)) as (m' & Hembed' & Hstep).
eauto 6 using membed_hlm, steps_reduces, plus_star; done.

(* pair *)    
intros G M1 M1' M2 M2' HredM1 IH1 HredM2 IH2 m Hembed.
#3 invert Hembed.
  intros m1 m2 Hembed1 Hembed2 <-.
  so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1).
  so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2).
  exists (lm_let m1' (lm_let (shift1 m2') (lm_lam (lm_app (lm_app (var 0) (var 2)) (var 1))))).
  split.
    eauto with embed; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun m => lm_let m _)); eauto 9 using shift_hlm with embed lambda; done.

      apply (star_map' _ _ (fun m => lm_let _ (lm_let (shift1 m) _))); eauto 9 using shift_reduce with embed lambda; done.

  intros m1 m2 Hval1 Hembed1 Hembed2 <-.
  so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1').
  so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2').
  exists (lm_let m2' (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m1')) (var 1)))).
  split.
    eauto using mreduce_value with embed; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun m => lm_let m _)); eauto 8 using shift_hlm with embed lambda; done.

      apply (star_map' _ _ (fun m => lm_let _ (lm_lam (lm_app (lm_app (var 0) (subst (sh 2) m)) (var 1))))); eauto 8 using shift_reduce with embed lambda; done.

  intros m1 m2 Hval1 Hval2 Hembed1 Hembed2 <-.
  so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1').
  so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2').
  exists (lm_lam (lm_app (lm_app (var 0) (shift1 m1')) (shift1 m2'))).
  split.
    eauto using mreduce_value with embed; done.

    #2 eapply star_trans.
      apply (star_map' _ _ (fun m => lm_lam (lm_app (lm_app (var 0) (shift1 m)) (shift1 m2)))); eauto 8 using shift_reduce, shift_hlm with embed lambda; done.

      apply (star_map' _ _ (fun m => lm_lam (lm_app (lm_app (var 0) (shift1 m1')) (shift1 m)))); eauto 8 using shift_reduce, shift_hlm with embed lambda; done.

(* pi11 *)
intros G M M' _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hred).
exists (lm_app m1' (lm_lam (lm_lam (var 1)))).
split.
  eauto with embed; done.

  apply (star_map' _ _ (fun m => lm_app m _)); eauto with embed lambda; done.

(* pi12 *)
intros G V1 V2 Hval1 Hval2 m Hembed.
so (membed_step _#4 Hembed (mstep_pi12 Hval1 Hval2)) as (m' & Hembed' & Hstep).
eauto 6 using membed_hlm, steps_reduces, plus_star; done.

(* pi21 *)
intros G M M' _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hred).
exists (lm_app m1' (lm_lam (lm_lam (var 0)))).
split.
  eauto with embed; done.

  apply (star_map' _ _ (fun m => lm_app m (lm_lam (lm_lam (var 0))))); eauto with embed lambda; done.

(* pi22 *)
intros G V1 V2 Hval1 Hval2 m Hembed.
so (membed_step _#4 Hembed (mstep_pi22 Hval1 Hval2)) as (m' & Hembed' & Hstep).
eauto 6 using membed_hlm, steps_reduces, plus_star; done.

(* unpack1 *)
intros G e e' M M' T T' _ IH1 _ IH2 m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1).
so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2).
exists (lm_let (lm_app m1' lm_i) (subst (dot (var 0) (dot lm_i sh1)) m2')).
split.
  eauto with embed; done.

  #2 eapply star_trans.
    exploit (subst_under_hlm 1 lm_i m2) as H; eauto with embed lambda; [].
    simp_sub in H.
    apply (star_map' _ _ (fun m => lm_let (lm_app m _) _)); eauto 8 using subst_under_hlm with embed lambda; done.

    repl (dot (var 0) (dot lm_i sh1)) with (under 1 (dot lm_i id)) by (simp_sub; reflexivity).
    apply (star_map' _ _ (fun m => lm_let _ (subst (under 1 (dot lm_i id)) m))); eauto 8 using subst_under_reduce with embed lambda; done.

(* unpack2 *)
intros G c v k t M T Hval m Hembed.
so (membed_step _#4 Hembed (mstep_unpack2 Hval)) as (m' & Hembed' & Hstep).
eauto 6 using membed_hlm, steps_reduces, plus_star; done.

(* lett1 *)
intros G e1 e1' M2 M2' _ IH1 _ IH2 m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1).
so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2).
exists (lm_let m1' m2').
split.
  eauto with embed; done.

  #2 eapply star_trans.
    apply (star_map' _ _ (fun m => lm_let m m2)); eauto with embed lambda; done.

    apply (star_map' _ _ (fun m => lm_let m1' m)); eauto with embed lambda; done.

(* lett2 *)
intros G v1 M2 Hval m Hembed.
so (membed_step _#4 Hembed (mstep_lett2 Hval)) as (m' & Hembed' & Hstep).
eauto 6 using membed_hlm, steps_reduces, plus_star; done.

(* letm1 *)
intros G M1 M1' M2 M2' T T' _ IH1 _ IH2 m Hembed.
invert Hembed; [].
intros m1 m2 Hembed1 Hembed2 <-.
so (IH1 _ Hembed1) as (m1' & Hembed1' & Hred1).
so (IH2 _ Hembed2) as (m2' & Hembed2' & Hred2).
exists (lm_let m1' (subst (dot (var 0) (dot lm_i sh1)) m2')).
split.
  eauto with embed; done.

  #2 eapply star_trans.
    exploit (subst_under_hlm 1 lm_i m2) as H; eauto with embed lambda; [].
    apply (star_map' _ _ (fun m => lm_let m _)); eauto with embed lambda; done.

    repl (dot (var 0) (dot lm_i sh1)) with (under 1 (dot lm_i id)) by (simp_sub; reflexivity).
    apply (star_map' _ _ (fun m => lm_let m1' (subst (under 1 (dot lm_i id)) m))); eauto 8 using subst_under_reduce with embed lambda; done.

(* letm2 *)
intros G V M T c Hval Hfst m Hembed.
so (membed_step _#4 Hembed (mstep_letm2 Hval Hfst)) as (m' & Hembed' & Hstep).
eauto 6 using membed_hlm, steps_reduces, plus_star; done.

(* seal1 *)
intros G M M' T T' _ IH m Hembed.
invert Hembed; [].
intros m1 Hembed1 <-.
so (IH _ Hembed1) as (m1' & Hembed1' & Hred).
exists (lm_let lm_i (shift1 m1')).
split.
  eauto with embed; done.

  apply (star_map' _ _ (fun m => lm_let lm_i (shift1 m))); eauto using shift_reduce with embed lambda; done.

(* seal2 *)
intros G M T m Hembed.
so (membed_step _#4 Hembed mstep_seal) as (m' & Hembed' & Hstep).
eauto 6 using membed_hlm, steps_reduces, plus_star; done.

(* epilogue *)
destruct_all Hind.
split; eauto 8; done.

(* tedium *)
Local Ltac exvar := solve [exact nonsense | exact nil].
Grab Existential Variables.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
exvar.
Qed.


Lemma tembed_reduce :
  forall G e e' m,
    tembed G e m
    -> treduce G e e'
    -> exists m',
         tembed G e' m'
         /\ star reduce m m'.
Proof.
exact (embeds_reduce andel).
Qed.


Lemma membed_reduce :
  forall G M M' m,
    membed G M m
    -> mreduce G M M'
    -> exists m',
         membed G M' m'
         /\ star reduce m m'.
Proof.
exact (embeds_reduce ander).
Qed.


Lemma adequacy_treduce :
  forall e e' t t',
    oft nil e t
    -> oft nil e' t'
    -> treduce nil e e'
    -> (thalt nil e <-> thalt nil e').
Proof.
intros e e' t t' Hof Hof' Hred.
so (can_tembed _#3 Hof) as (m & Hembed).
so (tembed_reduce _#4 Hembed Hred) as (m' & Hembed' & Hred').
split.
  intro Hhalt.
  eapply tembed_halt_conv; eauto; [].
  apply (adequacy_reduce m); auto; [].
  eapply tembed_halt; eauto; done.

  intro Hhalt.
  eapply tembed_halt_conv; eauto; [].
  apply (adequacy_reduce m m'); auto; [].
  eapply tembed_halt; eauto; done.
Qed.


Lemma adequacy_mreduce :
  forall M M' p p' t t',
    ofm nil p M t
    -> ofm nil p' M' t'
    -> mreduce nil M M'
    -> (mhalt nil M <-> mhalt nil M').
Proof.
intros M M' p p' t t' Hof Hof' Hred.
so (can_membed _#4 Hof) as (m & Hembed).
so (membed_reduce _#4 Hembed Hred) as (m' & Hembed' & Hred').
split.
  intro Hhalt.
  eapply membed_halt_conv; eauto; [].
  apply (adequacy_reduce m); auto; [].
  eapply membed_halt; eauto; done.

  intro Hhalt.
  eapply membed_halt_conv; eauto; [].
  apply (adequacy_reduce m m'); auto; [].
  eapply membed_halt; eauto; done.
Qed.
