
Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import Subst.
Require Import TSubst.

Require Compare_dec.


Inductive natk : context -> (* cn *) term -> (* kd *) term -> Prop :=
| natk_var {G i k}
    : index i G (cl_cn k)
      -> natk G (var i) (subst (sh (S i)) k)

| natk_unit {G}
    : natk G cn_unit kd_type

| natk_arrow {G t1 t2}
    : natk G (cn_arrow t1 t2) kd_type

| natk_prod {G t1 t2}
    : natk G (cn_prod t1 t2) kd_type

| natk_all {G k t}
    : natk G (cn_all k t) kd_type

| natk_exists {G k t}
    : natk G (cn_exists k t) kd_type

| natk_app {G p c k1 k2}
    : natk G p (kd_pi k1 k2)
      -> natk G (cn_app p c) (subst1 c k2)

| natk_pi1 {G p k1 k2}
    : natk G p (kd_sigma k1 k2)
      -> natk G (cn_pi1 p) k1

| natk_pi2 {G p k1 k2}
    : natk G p (kd_sigma k1 k2)
      -> natk G (cn_pi2 p) (subst1 (cn_pi1 p) k2)

| natk_star {G}
    : natk G cn_star kd_one
.


Inductive headred : context -> (* cn *) term -> (* cn *) term -> Prop :=
| headred_app {G c1 c1' c2}
    : headred G c1 c1'
      -> headred G (cn_app c1 c2) (cn_app c1' c2)

| headred_beta {G k1 c1 c2}
    : headred G (cn_app (cn_lam k1 c1) c2) (subst1 c2 c1)

| headred_pi1 {G c c'}
    : headred G c c'
      -> headred G (cn_pi1 c) (cn_pi1 c')

| headred_beta1 {G c1 c2}
    : headred G (cn_pi1 (cn_pair c1 c2)) c1

| headred_pi2 {G c c'}
    : headred G c c'
      -> headred G (cn_pi2 c) (cn_pi2 c')

| headred_beta2 {G c1 c2}
    : headred G (cn_pi2 (cn_pair c1 c2)) c2

| headred_sing {G c c'}
    : natk G c (kd_sing c')
      -> headred G c c'
.


Inductive headnorm : context -> (* cn *) term -> (* cn *) term -> Prop :=
| headnorm_more {G c c' c''}
    : headred G c c'
      -> headnorm G c' c''
      -> headnorm G c c''

| headnorm_done {G c}
    : natk G c kd_type
      -> headnorm G c c
.


Inductive aeq : context -> context -> (* cn *) term -> term -> (* kd *) term -> term -> Prop :=
| aeq_type {G G' c c' p p'}
    : headnorm G c p
      -> headnorm G' c' p'
      -> peq G G' p p' kd_type kd_type
      -> aeq G G' c c' kd_type kd_type

| aeq_sing {G G' c c' b b'}
    : aeq G G' c c' (kd_sing b) (kd_sing b')

| aeq_pi {G G' c c' k1 k1' k2 k2'}
    : aeq (G; cl_cn k1) (G'; cl_cn k1') (cn_app (shift1 c) (var 0)) (cn_app (shift1 c') (var 0)) k2 k2'
      -> aeq G G' c c' (kd_pi k1 k2) (kd_pi k1' k2')

| aeq_sigma {G G' c c' k1 k1' k2 k2'}
    : aeq G G' (cn_pi1 c) (cn_pi1 c') k1 k1'
      -> aeq G G' (cn_pi2 c) (cn_pi2 c') (subst1 (cn_pi1 c) k2) (subst1 (cn_pi1 c') k2')
      -> aeq G G' c c' (kd_sigma k1 k2) (kd_sigma k1' k2')

| aeq_one {G G' c c'}
    : aeq G G' c c' kd_one kd_one


with peq : context -> context -> (* cn *) term -> term -> (* kd *) term -> term -> Prop :=
| peq_var {G G' i k k'}
    : index i G (cl_cn k)
      -> index i G' (cl_cn k')
      -> peq G G' (var i) (var i) (subst (sh (S i)) k) (subst (sh (S i)) k')

| peq_unit {G G'}
    : peq G G' cn_unit cn_unit kd_type kd_type

| peq_arrow {G G' t1 t1' t2 t2'}
    : aeq G G' t1 t1' kd_type kd_type
      -> aeq G G' t2 t2' kd_type kd_type
      -> peq G G' (cn_arrow t1 t2) (cn_arrow t1' t2') kd_type kd_type

| peq_prod {G G' t1 t1' t2 t2'}
    : aeq G G' t1 t1' kd_type kd_type
      -> aeq G G' t2 t2' kd_type kd_type
      -> peq G G' (cn_prod t1 t2) (cn_prod t1' t2') kd_type kd_type

| peq_all {G G' k k' t t'}
    : aeqk G G' k k'
      -> aeq (G; cl_cn k) (G'; cl_cn k') t t' kd_type kd_type
      -> peq G G' (cn_all k t) (cn_all k' t') kd_type kd_type

| peq_exists {G G' k k' t t'}
    : aeqk G G' k k'
      -> aeq (G; cl_cn k) (G'; cl_cn k') t t' kd_type kd_type
      -> peq G G' (cn_exists k t) (cn_exists k' t') kd_type kd_type

| peq_app {G G' p p' c c' k1 k1' k2 k2'}
    : peq G G' p p' (kd_pi k1 k2) (kd_pi k1' k2')
      -> aeq G G' c c' k1 k1'
      -> peq G G' (cn_app p c) (cn_app p' c') (subst1 c k2) (subst1 c' k2')

| peq_pi1 {G G' p p' k1 k1' k2 k2'}
    : peq G G' p p' (kd_sigma k1 k2) (kd_sigma k1' k2')
      -> peq G G' (cn_pi1 p) (cn_pi1 p') k1 k1'

| peq_pi2 {G G' p p' k1 k1' k2 k2'}
    : peq G G' p p' (kd_sigma k1 k2) (kd_sigma k1' k2')
      -> peq G G' (cn_pi2 p) (cn_pi2 p') (subst1 (cn_pi1 p) k2) (subst1 (cn_pi1 p') k2')

| peq_star {G G'}
    : peq G G' cn_star cn_star kd_one kd_one


with aeqk : context -> context -> (* kd *) term -> term -> Prop :=
| aeqk_type {G G'}
    : aeqk G G' kd_type kd_type

| aeqk_sing {G G' c c'}
    : aeq G G' c c' kd_type kd_type
      -> aeqk G G' (kd_sing c) (kd_sing c')

| aeqk_pi {G G' k1 k1' k2 k2'}
    : aeqk G G' k1 k1'
      -> aeqk (G; cl_cn k1) (G'; cl_cn k1') k2 k2'
      -> aeqk G G' (kd_pi k1 k2) (kd_pi k1' k2')

| aeqk_sigma {G G' k1 k1' k2 k2'}
    : aeqk G G' k1 k1'
      -> aeqk (G; cl_cn k1) (G'; cl_cn k1') k2 k2'
      -> aeqk G G' (kd_sigma k1 k2) (kd_sigma k1' k2')

| aeqk_one {G G'}
    : aeqk G G' kd_one kd_one
.


Scheme aeq_mut_ind := Minimality for aeq Sort Prop
with   peq_mut_ind := Minimality for peq Sort Prop
with   aeqk_mut_ind := Minimality for aeqk Sort Prop.

Combined Scheme alg_ind from aeq_mut_ind, peq_mut_ind, aeqk_mut_ind.


Inductive path : (* cn *) term -> Prop :=
| path_var {i}
    : path (var i)

| path_unit
    : path cn_unit

| path_arrow {t1 t2}
    : path (cn_arrow t1 t2)
 
| path_prod {t1 t2}
    : path (cn_prod t1 t2)

| path_all {k t}
    : path (cn_all k t)
 
| path_exists {k t}
    : path (cn_exists k t)

| path_app {p c}
    : path p
      -> path (cn_app p c)

| path_pi1 {p}
    : path p
      -> path (cn_pi1 p)

| path_pi2 {p}
    : path p
      -> path (cn_pi2 p)

| path_star
    : path cn_star
.


Definition extend i (G G' : context) :=
  exists G1,
    G' = G;; G1 /\ length G1 = i.


Hint Constructors natk headred headnorm aeq peq aeqk path : alg.


Lemma extend_0 :
  forall G, extend 0 G G.
Proof.
intros G.
exists nil.
auto.
Qed.


Lemma extend_1 :
  forall G cl, extend 1 G (G; cl).
Proof.
intros G cl.
exists (nil; cl).
auto.
Qed.


Lemma extend_trans :
  forall i j D1 D2 D3,
    extend i D1 D2
    -> extend j D2 D3
    -> extend (i+j) D1 D3.
Proof.
intros i j D1 D2 D3 Hext1 Hext2.
destruct Hext1 as (G1 & -> & Hlen1).
destruct Hext2 as (G2 & -> & Hlen2).
rewrite -> app_assoc.
exists (G1;; G2).
split; auto; [].
rewrite -> app_length.  
omega.
Qed.


Hint Resolve extend_0 extend_1 : alg.


Lemma shift_path :
  forall i p, path p -> path (subst (sh i) p).
Proof.
intros i p Hpath.
induct Hpath; intros; simp_sub; auto with alg.
Qed.


Lemma weaken_natk_gen :
  forall i j G1 G2 G3 p k,
    length G2 = i
    -> length G3 = j
    -> natk (G1;; G3) p k
    -> natk (G1;; G2;; subst_ctx (sh i) G3) (subst (under j (sh i)) p) (subst (under j (sh i)) k).
Proof.
intros i j G1 G2 G3 p k Hlen2 Hlen3 Hnatk.
remember (G1;; G3) as G eqn:HeqG.
revert HeqG.
#3 induct Hnatk; try (simp_sub; intros; subst; eauto with alg; done).

(* var *)
intros G n k Hindex ->.
simp_sub.
so (Compare_dec.lt_dec n j) as [Hlt | Hnlt].
  rewrite -> project_under_lt; auto; [].
  so (index_app_lt _#5 (eqconv (eqsymm Hlen3) Hlt) Hindex) as Hindex'.
  so (index_subst_ctx _#3 (sh i) Hindex') as Hindex''.
  rewrite -> Hlen3 in Hindex''.
  so (index_app_left _ _ (G1;; G2) _ _ Hindex'') as Hindex'''.
  simp_sub in Hindex'''.
  so (natk_var Hindex''') as Hnatk.
  simp_sub in Hnatk.
  rewrite -> compose_sh_under_leq; [| omega].
  repl (j - (1 + n)) with (j - n - 1) by omega.
  assumption.

  assert (n >= j) as Hgeq by omega; clear Hnlt.
  rewrite -> project_under_ge; auto; [].
  simp_sub.
  repl (j + i + (n - j)) with (n + i) by omega.
  so (index_app_ge _#5 (eqconv (eqsymm Hlen3) Hgeq) Hindex) as Hindex'.
  so (index_app_right _ (G2;; subst_ctx (sh i) G3) _#3 Hindex') as Hindex''.
  rewrite -> app_length in Hindex''; [].
  rewrite -> subst_ctx_length in Hindex''; [].
  rewrite -> Hlen2 in Hindex''; [].
  rewrite -> Hlen3 in Hindex''; [].
  rewrite <- app_assoc in Hindex''; [].
  repl (n - j + (j + i)) with (n + i) in Hindex'' by omega.
  so (natk_var Hindex'') as Hnatk.
  rewrite -> compose_sh_under_geq; auto; [].
  simp_sub.
  repl (1 + n - j + i + j) with (S (n + i)) by omega.
  assumption.

(* app *)
intros G p c k1 k2 _ IH ->.
simp_sub.
so (IH (eq_refl _)) as Hnatk.
simp_sub in Hnatk.
so (@natk_app _ _ (subst (under j (sh i)) c) _ _ Hnatk) as Hnatk'.
simp_sub in Hnatk'.
assumption.

(* pi2 *)
intros G p k1 k2 _ IH ->.
simp_sub.
so (IH (eq_refl _)) as Hnatk.
simp_sub in Hnatk.
so (@natk_pi2 _#4 Hnatk) as Hnatk'.
simp_sub in Hnatk'.
assumption.
Qed.


Lemma weaken_natk :
  forall i G1 G2 p k,
    extend i G1 G2
    -> natk G1 p k
    -> natk G2 (subst (sh i) p) (subst (sh i) k).
Proof.
intros i G1 G2 p k Hext Hnatk.
destruct Hext as (G & -> & Hlen).
exploit (weaken_natk_gen i 0 G1 G nil p k); auto; done.
Qed.


Lemma weaken_headred_gen :
  forall i j G1 G2 G3 c c',
    length G2 = i
    -> length G3 = j
    -> headred (G1;; G3) c c'
    -> headred (G1;; G2;; subst_ctx (sh i) G3) (subst (under j (sh i)) c) (subst (under j (sh i)) c').
Proof.
intros i j G1 G2 G3 c c' Hlen2 Hlen3 Hred.
remember (G1;; G3) as G eqn:HeqG.
revert HeqG.
#2 induct Hred; try (simp_sub; intros; subst; eauto with alg; done).

(* beta *)
intros G k1 c1 c2 ->.
simp_sub.
so (@headred_beta
      (G1;; G2;; subst_ctx (sh i) G3)
      (subst (under j (sh i)) k1)
      (subst (dot (var 0) (compose (under j (sh i)) sh1)) c1)
      (subst (under j (sh i)) c2)) as Hred.
simp_sub in Hred.
assumption.

(* sing *)
intros G c c' Hnatk ->.
apply headred_sing; [].
so (weaken_natk_gen _#7 Hlen2 Hlen3 Hnatk) as Hnatk'.
simp_sub in Hnatk'.
assumption.
Qed.


Lemma weaken_headred :
  forall i G1 G2 c c',
    extend i G1 G2
    -> headred G1 c c'
    -> headred G2 (subst (sh i) c) (subst (sh i) c').
Proof.
intros i G1 G2 c c' Hext Hred.
destruct Hext as (G & -> & Hlen).
exploit (weaken_headred_gen i 0 G1 G nil c c'); auto; done.
Qed.


Lemma weaken_headnorm_gen :
  forall i j G1 G2 G3 c c',
    length G2 = i
    -> length G3 = j
    -> headnorm (G1;; G3) c c'
    -> headnorm (G1;; G2;; subst_ctx (sh i) G3) (subst (under j (sh i)) c) (subst (under j (sh i)) c').
Proof.
intros i j G1 G2 G3 c c' Hlen2 Hlen3 Hnorm.
remember (G1;; G3) as G eqn:HeqG.
revert HeqG.
induct Hnorm.

intros G c c' c'' Hred _ IH ->.
eapply headnorm_more; eauto using weaken_headred_gen; done.

intros G c Hnatk ->.
so (weaken_natk_gen _#7 Hlen2 Hlen3 Hnatk) as Hnatk'.
simp_sub in Hnatk'.
apply headnorm_done; auto; done.
Qed.


Lemma weaken_alg :
  (forall i j G1 G1' G2 G2' G3 G3' c c' k k',
     length G2 = i
     -> length G2' = i
     -> length G3 = j
     -> length G3' = j
     -> aeq (G1;; G3) (G1';; G3') c c' k k'
     -> aeq (G1;; G2;; subst_ctx (sh i) G3) (G1';; G2';; subst_ctx (sh i) G3') (subst (under j (sh i)) c) (subst (under j (sh i)) c') (subst (under j (sh i)) k) (subst (under j (sh i)) k'))
  /\
  (forall i j G1 G1' G2 G2' G3 G3' p p' k k',
     length G2 = i
     -> length G2' = i
     -> length G3 = j
     -> length G3' = j
     -> peq (G1;; G3) (G1';; G3') p p' k k'
     -> peq (G1;; G2;; subst_ctx (sh i) G3) (G1';; G2';; subst_ctx (sh i) G3') (subst (under j (sh i)) p) (subst (under j (sh i)) p') (subst (under j (sh i)) k) (subst (under j (sh i)) k'))
  /\
  (forall i j G1 G1' G2 G2' G3 G3' k k',
     length G2 = i
     -> length G2' = i
     -> length G3 = j
     -> length G3' = j
     -> aeqk (G1;; G3) (G1';; G3') k k'
     -> aeqk (G1;; G2;; subst_ctx (sh i) G3) (G1';; G2';; subst_ctx (sh i) G3') (subst (under j (sh i)) k) (subst (under j (sh i)) k')).
Proof.
exploit
  (alg_ind
     (fun G G' c c' k k' =>
        forall i j G1 G1' G2 G2' G3 G3',
          G = G1;; G3
          -> G' = G1';; G3'
          -> length G2 = i
          -> length G2' = i
          -> length G3 = j
          -> length G3' = j
          -> aeq (G1;; G2;; subst_ctx (sh i) G3) (G1';; G2';; subst_ctx (sh i) G3') (subst (under j (sh i)) c) (subst (under j (sh i)) c') (subst (under j (sh i)) k) (subst (under j (sh i)) k'))
     (fun G G' p p' k k' =>
        forall i j G1 G1' G2 G2' G3 G3',
          G = G1;; G3
          -> G' = G1';; G3'
          -> length G2 = i
          -> length G2' = i
          -> length G3 = j
          -> length G3' = j
          -> peq (G1;; G2;; subst_ctx (sh i) G3) (G1';; G2';; subst_ctx (sh i) G3') (subst (under j (sh i)) p) (subst (under j (sh i)) p') (subst (under j (sh i)) k) (subst (under j (sh i)) k'))
     (fun G G' k k' =>
        forall i j G1 G1' G2 G2' G3 G3',
          G = G1;; G3
          -> G' = G1';; G3'
          -> length G2 = i
          -> length G2' = i
          -> length G3 = j
          -> length G3' = j
          -> aeqk (G1;; G2;; subst_ctx (sh i) G3) (G1';; G2';; subst_ctx (sh i) G3') (subst (under j (sh i)) k) (subst (under j (sh i)) k'))) as Hind;
  try (intros; subst; simp_sub; eauto using weaken_headnorm_gen with alg; done).

(* pi *)
intros G G' c c' k1 k1' k2 k2' _ IH i j G1 G1' G2 G2' G3 G3' -> -> Hlen2 Hlen2' Hlen3 Hlen3'.
simp_sub.
apply aeq_pi; [].
simp_sub.
#3 exploit (IH i (S j) G1 G1' G2 G2' (G3; cl_cn k1) (G3'; cl_cn k1')) as Haeq; auto.
  simpl; auto; done.

  simpl; auto; done.

  simpl subst_ctx in Haeq.
  simp_sub in Haeq.
  autorewrite with canonlist in Haeq.
  rewrite -> Hlen3, Hlen3' in Haeq.
  assumption.

(* sigma *)
intros G G' c c' k1 k1' k2 k2' _ IH1 _ IH2 i j G1 G1' G2 G2' G3 G3' -> -> Hlen2 Hlen2' Hlen3 Hlen3'.
simp_sub.
apply aeq_sigma; eauto; [].
clear IH1.
simp_sub.
so (IH2 _#8 (eq_refl _) (eq_refl _) Hlen2 Hlen2' Hlen3 Hlen3') as Haeq.
simp_sub in Haeq.
assumption.

(* var *)
intros G G' n k k' Hindex Hindex' i j G1 G1' G2 G2' G3 G3' -> -> Hlen2 Hlen2' Hlen3 Hlen3'.
simp_sub.
so (Compare_dec.lt_dec n j) as [Hlt | Hnlt].
  rewrite -> !project_under_lt; auto; [].
  so (index_app_lt _#5 (eqconv (eqsymm Hlen3) Hlt) Hindex) as Hindex2.
  so (index_subst_ctx _#3 (sh i) Hindex2) as Hindex3.
  rewrite -> Hlen3 in Hindex3.
  so (index_app_left _ _ (G1;; G2) _ _ Hindex3) as Hindex4.
  simp_sub in Hindex4.
  so (index_app_lt _#5 (eqconv (eqsymm Hlen3') Hlt) Hindex') as Hindex2'.
  so (index_subst_ctx _#3 (sh i) Hindex2') as Hindex3'.
  rewrite -> Hlen3' in Hindex3'.
  so (index_app_left _ _ (G1';; G2') _ _ Hindex3') as Hindex4'.
  simp_sub in Hindex4'.
  so (peq_var Hindex4 Hindex4') as Hpeq.
  simp_sub in Hpeq.
  rewrite -> !compose_sh_under_leq; [| omega].
  repl (j - (1 + n)) with (j - n - 1) by omega.
  assumption.

  assert (n >= j) as Hgeq by omega; clear Hnlt.
  rewrite -> !project_under_ge; auto; [].
  simp_sub.
  repl (j + i + (n - j)) with (n + i) by omega.
  so (index_app_ge _#5 (eqconv (eqsymm Hlen3) Hgeq) Hindex) as Hindex2.
  so (index_app_right _ (G2;; subst_ctx (sh i) G3) _#3 Hindex2) as Hindex3.
  rewrite -> app_length in Hindex3; [].
  rewrite -> subst_ctx_length in Hindex3; [].
  rewrite -> Hlen2 in Hindex3; [].
  rewrite -> Hlen3 in Hindex3; [].
  rewrite <- app_assoc in Hindex3; [].
  repl (n - j + (j + i)) with (n + i) in Hindex3 by omega.
  so (index_app_ge _#5 (eqconv (eqsymm Hlen3') Hgeq) Hindex') as Hindex2'.
  so (index_app_right _ (G2';; subst_ctx (sh i) G3') _#3 Hindex2') as Hindex3'.
  rewrite -> app_length in Hindex3'; [].
  rewrite -> subst_ctx_length in Hindex3'; [].
  rewrite -> Hlen2' in Hindex3'; [].
  rewrite -> Hlen3' in Hindex3'; [].
  rewrite <- app_assoc in Hindex3'; [].
  repl (n - j + (j + i)) with (n + i) in Hindex3' by omega.
  so (peq_var Hindex3 Hindex3') as Hpeq.
  simp_sub in Hpeq.
  rewrite -> compose_sh_under_geq; auto; [].
  simp_sub.
  repl (1 + n - j + i + j) with (S (n + i)) by omega.
  assumption.

(* all *)
intros G G' k k' t t' _ IH1 _ IH2 i j G1 G1' G2 G2' G3 G3' -> -> Hlen2 Hlen2' Hlen3 Hlen3'.
simp_sub.
apply peq_all; auto; [].
#3 exploit (IH2 i (S j) G1 G1' G2 G2' (G3; cl_cn k) (G3'; cl_cn k')) as Haeq; auto.
  simpl; auto; done.

  simpl; auto; done.

  simpl subst_ctx in Haeq.
  simp_sub in Haeq.
  autorewrite with canonlist in Haeq.
  rewrite -> Hlen3, Hlen3' in Haeq.
  assumption.

(* exists *)
intros G G' k k' t t' _ IH1 _ IH2 i j G1 G1' G2 G2' G3 G3' -> -> Hlen2 Hlen2' Hlen3 Hlen3'.
simp_sub.
apply peq_exists; auto; [].
#3 exploit (IH2 i (S j) G1 G1' G2 G2' (G3; cl_cn k) (G3'; cl_cn k')) as Haeq; auto.
  simpl; auto; done.

  simpl; auto; done.

  simpl subst_ctx in Haeq.
  simp_sub in Haeq.
  autorewrite with canonlist in Haeq.
  rewrite -> Hlen3, Hlen3' in Haeq.
  assumption.

(* app *)
intros G G' p p' c c' k1 k1' k2 k2' _ IH1 _ IH2 i j G1 G1' G2 G2' G3 G3' -> -> Hlen2 Hlen2' Hlen3 Hlen3'.
simp_sub.
so (IH1 _#8 (eq_refl _) (eq_refl _) Hlen2 Hlen2' Hlen3 Hlen3') as Hpeq.
simp_sub in Hpeq.
rewrite -> !(split_dot _ (under j (sh i))).
rewrite -> !subst_compose.
eapply peq_app; eauto; done.

(* pi2 *)
intros G G' p p' k1 k1' k2 k2' _ IH i j G1 G1' G2 G2' G3 G3' -> -> Hlen2 Hlen2' Hlen3 Hlen3'.
simp_sub.
so (IH _#8 (eq_refl _) (eq_refl _) Hlen2 Hlen2' Hlen3 Hlen3') as Hpeq.
simp_sub in Hpeq.
rewrite -> !(split_dot _ (under j (sh i))).
rewrite -> !subst_compose.
eapply peq_pi2; eauto; done.

(* pi *)
intros G G' k1 k1' k2 k2' _ IH1 _ IH2 i j G1 G1' G2 G2' G3 G3' -> -> Hlen2 Hlen2' Hlen3 Hlen3'.
simp_sub.
apply aeqk_pi; auto; [].
#3 exploit (IH2 i (S j) G1 G1' G2 G2' (G3; cl_cn k1) (G3'; cl_cn k1')) as Haeq; auto.
  simpl; auto; done.

  simpl; auto; done.

  simpl subst_ctx in Haeq.
  simp_sub in Haeq.
  autorewrite with canonlist in Haeq.
  rewrite -> Hlen3, Hlen3' in Haeq.
  assumption.

(* sigma *)
intros G G' k1 k1' k2 k2' _ IH1 _ IH2 i j G1 G1' G2 G2' G3 G3' -> -> Hlen2 Hlen2' Hlen3 Hlen3'.
simp_sub.
apply aeqk_sigma; auto; [].
#3 exploit (IH2 i (S j) G1 G1' G2 G2' (G3; cl_cn k1) (G3'; cl_cn k1')) as Haeq; auto.
  simpl; auto; done.

  simpl; auto; done.

  simpl subst_ctx in Haeq.
  simp_sub in Haeq.
  autorewrite with canonlist in Haeq.
  rewrite -> Hlen3, Hlen3' in Haeq.
  assumption.

(* epilogue *)
destruct Hind as (Hind & Hindp & Hindk).
do2 2 split; intros; eauto; done.
Qed.


Lemma weaken_aeq :
  forall i G1 G1' G2 G2' c c' k k',
    extend i G1 G2
    -> extend i G1' G2'
    -> aeq G1 G1' c c' k k'
    -> aeq G2 G2' (subst (sh i) c) (subst (sh i) c') (subst (sh i) k) (subst (sh i) k').
Proof.
intros i G1 G1' G2 G2' c c' k k' Hext Hext' Haeq.
destruct Hext as (G & -> & Hlen).
destruct Hext' as (G' & -> & Hlen').
exploit (weaken_alg andel i 0 G1 G1' G G' nil nil c c' k k'); auto; done.
Qed.


Lemma weaken_peq :
  forall i G1 G1' G2 G2' p p' k k',
    extend i G1 G2
    -> extend i G1' G2'
    -> peq G1 G1' p p' k k'
    -> peq G2 G2' (subst (sh i) p) (subst (sh i) p') (subst (sh i) k) (subst (sh i) k').
Proof.
intros i G1 G1' G2 G2' p p' k k' Hext Hext' Haeq.
destruct Hext as (G & -> & Hlen).
destruct Hext' as (G' & -> & Hlen').
exploit (weaken_alg ander andel i 0 G1 G1' G G' nil nil p p' k k'); auto; done.
Qed.


Lemma natk_path :
  forall G p k,
    natk G p k
    -> path p.
Proof.
intros G p k Hnatk.
induct Hnatk; auto with alg.
Qed.


Lemma peq_natk :
  forall G G' p p' k k',
    peq G G' p p' k k'
    -> natk G p k /\ natk G' p' k'.
Proof.
intros G G' p p' k k' Hpeq.
#3 induct Hpeq; try (intros; auto with alg; done).

(* app *)
intros G G' p p' c c' k1 k1' k2 k2' _ (IH1 & IH2) _.
eauto using natk_app; done.

(* pi1 *)
intros G G' p p' k1 k1' k2 k2' _ (IH1 & IH2).
eauto using natk_pi1; done.

(* pi2 *)
intros G G' p p' k1 k1' k2 k2' _ (IH1 & IH2).
eauto using natk_pi2; done.
Qed.


Lemma natk_fun :
  forall G p k k',
    natk G p k
    -> natk G p k'
    -> k = k'.
Proof.
intros G p k k' Hnatk Hnatk'.
revert k' Hnatk'.
induct Hnatk.

(* var *)
intros G i k Hindex k' Hnatk'.
invert Hnatk'.
intros k'' Hindex' <-.
f_equal; [].
so (index_fun _#5 Hindex Hindex') as Heq.
injection Heq; auto; done.

(* unit *)
intros G k' Hnatk'.
invert Hnatk'; auto; done.

(* arrow *)
intros G t1 t2 k' Hnatk'.
invert Hnatk'; auto; done.

(* prod *)
intros G t1 t2 k' Hnatk'.
invert Hnatk'; auto; done.

(* all *)
intros G t1 t2 k' Hnatk'.
invert Hnatk'; auto; done.

(* exists *)
intros G t1 t2 k' Hnatk'.
invert Hnatk'; auto; done.

(* app *)
intros G p c k1 k2 _ IH k' Hnatk'.
invert Hnatk'.
intros k1' k2' Hnatk'' <-.
f_equal; [].
so (IH _ Hnatk'') as Heq.
injection Heq; auto; done.

(* pi1 *)
intros G p k1 k2 _ IH k' Hnatk'.
invert Hnatk'.
intros k'' Hnatk''.
so (IH _ Hnatk'') as Heq.
injection Heq; auto; done.

(* pi2 *)
intros G p k1 k2 _ IH k' Hnatk'.
invert Hnatk'.
intros ? ? Hnatk'' <-.
so (IH _ Hnatk'') as Heq.
f_equal; [].
injection Heq; auto; done.

(* star *)
intros G k' Hnatk'.
invert Hnatk'; auto; done.
Qed.


Lemma headred_path :
  forall G p c k,
    natk G p k
    -> headred G p c
    -> k = kd_sing c.
Proof.
intros G p c k Hnatk Hred.
revert k Hnatk.
induct Hred.

(* app *)
intros G c1 c1' c2 _ IH k Hnatk.
exfalso.
invert Hnatk.
intros k1 k2 Hnatk' <-.
so (IH _ Hnatk') as Heq.
discriminate Heq.

(* beta *)
intros G k1 c1 c2 k Hnatk.
exfalso.
invert Hnatk; [].
intros ? ? Hnatk'.
invert Hnatk'; done.

(* pi1 *)
intros G c c' _ IH k Hnatk.
exfalso.
invert Hnatk.
intros ? Hnatk'.
so (IH _ Hnatk') as Heq.
discriminate Heq.

(* beta1 *)
intros G c1 c2 k Hnatk.
exfalso.
invert Hnatk; [].
intros ? Hnatk'.
invert Hnatk'; done.

(* pi2 *)
intros G c c' _ IH k Hnatk.
exfalso.
invert Hnatk.
intros ? ? Hnatk' <-.
so (IH _ Hnatk') as Heq.
discriminate Heq.

(* beta2 *)
intros G c1 c2 k Hnatk.
exfalso.
invert Hnatk; [].
intros ? ? Hnatk' <-.
invert Hnatk'; done.

(* sing *)
intros G c b Hnatk k Hnatk'.
eapply natk_fun; eauto; done.
Qed.


Lemma headred_fun :
  forall G c c' c'',
    headred G c c'
    -> headred G c c''
    -> c' = c''.
Proof.
intros G c c' c'' Hred Hred'.
revert c'' Hred'.
induct Hred.

(* app *)
intros G c1 c1' c2 Hred IH c'' Hredapp.
#3 invert Hredapp.
  intros c1'' Hred' <-.
  f_equal; [].
  apply IH; auto; done.

  intros k c <- <-.
  exfalso.
  invert Hred; [].
  intro Hnatk.
  invert Hnatk; done.

  intro Hnatk.
  exfalso.
  invert Hnatk.
  intros k1 k2 Hnatk' _.
  so (headred_path _#4 Hnatk' Hred) as Heq.
  discriminate Heq.

(* beta *)
intros G k1 c1 c2 c'' Hredapp.
#3 invert Hredapp.
  intros b Hred' <-.
  exfalso.
  invert Hred'; [].
  intro Hnatk.
  invert Hnatk; done.

  intro; assumption.

  intro Hnatk.
  exfalso.
  invert Hnatk; [].
  intros ? ? Hnatk' _.
  invert Hnatk'; done.

(* pi1 *)
intros G c1 c1' Hred IH c'' Hredpi.
#3 invert Hredpi.
  intros c1'' Hred' <-.
  f_equal; [].
  apply IH; auto; done.

  intros d <-.
  exfalso.
  invert Hred; [].
  intros Hnatk.
  invert Hnatk; done.

  intro Hnatk.
  exfalso.
  invert Hnatk; [].
  intros ? Hnatk'.
  so (headred_path _#4 Hnatk' Hred) as Heq.
  discriminate Heq.

(* beta1 *)
intros G c1 c2 c'' Hredpi.
#3 invert Hredpi.
  intros c' Hred' _.
  exfalso.
  invert Hred'; [].
  intro Hnatk.
  invert Hnatk; done.

  auto; done.

  intro Hnatk.
  exfalso.
  invert Hnatk; [].
  intros ? Hnatk'.
  invert Hnatk'; done.

(* pi2 *)
intros G c1 c1' Hred IH c'' Hredpi.
#3 invert Hredpi.
  intros c1'' Hred' <-.
  f_equal; [].
  apply IH; auto; done.

  intros d <-.
  exfalso.
  invert Hred; [].
  intros Hnatk.
  invert Hnatk; done.

  intro Hnatk.
  exfalso.
  invert Hnatk; [].
  intros ? ? Hnatk' _.
  so (headred_path _#4 Hnatk' Hred) as Heq.
  discriminate Heq.

(* beta1 *)
intros G c1 c2 c'' Hredpi.
#3 invert Hredpi.
  intros c' Hred' _.
  exfalso.
  invert Hred'; [].
  intro Hnatk.
  invert Hnatk; done.

  auto; done.

  intro Hnatk.
  exfalso.
  invert Hnatk; [].
  intros ? ? Hnatk' _.
  invert Hnatk'; done.

(* sing *)
intros G c c' Hnatk c'' Hred'.
so (headred_path _#4 Hnatk Hred') as Heq.
injection Heq; auto; done.
Qed.


Lemma headnorm_fun :
  forall G c c' c'',
    headnorm G c c'
    -> headnorm G c c''
    -> c' = c''.
Proof.
intros G c c1 c2 Hnorm Hnorm'.
revert Hnorm'.
induct Hnorm.

(* more *)
intros G c c1' c1 Hred1 Hnorm1 IH Hnorm.
#2 invert Hnorm.
  intros c2' Hred2 Hnorm2.
  so (headred_fun _#4 Hred1 Hred2); subst c2'.
  apply IH; auto; done.

  intros Hnatk _.
  exfalso.
  so (headred_path _#4 Hnatk Hred1) as Heq.
  discriminate Heq.

(* done *)
intros G c Hnatk Hnorm.
#2 invert Hnorm.
  intros ? Hred _.
  exfalso.
  so (headred_path _#4 Hnatk Hred) as Heq.
  discriminate Heq.

  intros Hnatk' ->.
  reflexivity.
Qed.


Lemma alg_symm :
  (forall G G' c c' k k',
     aeq G G' c c' k k'
     -> aeq G' G c' c k' k)
  /\
  (forall G G' p p' k k',
     peq G G' p p' k k'
     -> peq G' G p' p k' k)
  /\
  (forall G G' k k',
     aeqk G G' k k'
     -> aeqk G' G k' k).
Proof.
exploit
  (alg_ind
     (fun G G' c c' k k' =>
        aeq G' G c' c k' k)
     (fun G G' p p' k k' =>
        peq G' G p' p k' k)
     (fun G G' k k' =>
        aeqk G' G k' k)) ;
  intros; eauto with alg.
Qed.


Lemma aeq_symm :
  forall G G' c c' k k',
    aeq G G' c c' k k'
    -> aeq G' G c' c k' k.
Proof.
exact (alg_symm andel).
Qed.


Lemma peq_symm :
  forall G G' p p' k k',
    peq G G' p p' k k'
    -> peq G' G p' p k' k.
Proof.
exact (alg_symm ander andel).
Qed.


Lemma alg_trans :
  (forall G1 G2 G3 c1 c2 c3 k1 k2 k3,
     aeq G1 G2 c1 c2 k1 k2
     -> aeq G2 G3 c2 c3 k2 k3
     -> aeq G1 G3 c1 c3 k1 k3)
  /\
  (forall G1 G2 G3 p1 p2 p3 k1 k2 k3,
     peq G1 G2 p1 p2 k1 k2
     -> peq G2 G3 p2 p3 k2 k3
     -> peq G1 G3 p1 p3 k1 k3)
  /\
  (forall G1 G2 G3 k1 k2 k3,
     aeqk G1 G2 k1 k2
     -> aeqk G2 G3 k2 k3
     -> aeqk G1 G3 k1 k3).
Proof.
exploit
  (alg_ind
     (fun G1 G2 c1 c2 k1 k2 =>
        forall G3 c3 k3,
          aeq G2 G3 c2 c3 k2 k3
          -> aeq G1 G3 c1 c3 k1 k3)
     (fun G1 G2 p1 p2 k1 k2 =>
        forall G3 p3 k3,
          peq G2 G3 p2 p3 k2 k3
          -> peq G1 G3 p1 p3 k1 k3)
     (fun G1 G2 k1 k2 =>
        forall G3 k3,
          aeqk G2 G3 k2 k3
          -> aeqk G1 G3 k1 k3)) as Hind.

(* type *)
intros G1 G2 c1 c2 p1 p2 Hnorm1 Hnorm2 _ IH G3 c3 k3 Haeq.
invert Haeq; [].
intros p2' p3 Hnorm2' Hnorm3 Hpeq <-.
so (headnorm_fun _#4 Hnorm2 Hnorm2').
subst p2'.
eapply aeq_type; eauto; done.

(* sing *)
intros G1 G2 c1 c2 b1 b2 G3 c3 k3 Haeq.
invert Haeq; [].
intros b3 <-.
apply aeq_sing; done.

(* pi *)
intros G1 G2 c1 c2 k1 k2 k1' k2' _ IH G3 c3 l Haeq.
invert Haeq; [].
intros k3 k3' Haeq' <-.
apply aeq_pi; [].
apply IH; auto; done.

(* sigma *)
intros G1 G2 c1 c2 k1 k2 k1' k2' _ IH1 _ IH2 G3 c3 l Haeq.
invert Haeq; [].
intros k3 k3' Haeq1 Haeq2 <-.
apply aeq_sigma; auto; done.

(* one *)
intros G1 G2 c1 c2 G3 c3 k3 Haeq.
invert Haeq; [].
intros <-.
apply aeq_one; done.

(* var *)
intros G1 G2 i k1 k2 Hindex1 Hindex2 G3 p3 l Hpeq.
invert Hpeq.
intros k2' k3 Hindex2' Hindex3 <- Heq <-.
apply peq_var; auto; done.

(* unit *)
intros G1 G2 G3 p3 k3 Hpeq.
invert Hpeq; intros; subst; auto with alg.

(* arrow *)
intros G1 G2 s1 t1 s2 t2 _ IH1 _ IH2 G3 p3 k3 Hpeq.
invert Hpeq.
intros; subst; auto with alg; done.

(* prod *)
intros G1 G2 s1 t1 s2 t2 _ IH1 _ IH2 G3 p3 k3 Hpeq.
invert Hpeq.
intros; subst; auto with alg; done.

(* all *)
intros G G' k k' t t' H H0 H1 H2 G3 p3 k3 Hpeq.
invert Hpeq.
intros; subst; auto with alg; done.

(* exists *)
intros G G' k k' t t' H H0 H1 H2 G3 p3 k3 Hpeq.
invert Hpeq.
intros; subst; auto with alg; done.

(* app *)
intros G1 G2 p1 p2 c1 c2 k1 k2 k1' k2' Hpeq12 IH1 _ IH2 G3 d l Hpeq.
invert Hpeq.
intros p3 c3 e k3 e' k3' Hpeq23 Haeq <- _ <-.
so (peq_natk _#6 Hpeq12) as (_ & Hnatk).
so (peq_natk _#6 Hpeq23) as (Hnatk' & _).
so (natk_fun _#4 Hnatk Hnatk') as Heq.
injection Heq.
intros <- <-.
eapply peq_app; eauto; done.

(* pi1 *)
intros G1 G2 p1 p2 k1 k2 k1' k2' Hpeq12 IH G3 d k3 Hpeq.
invert Hpeq.
intros p3 e k3' Hpeq23 <-.
so (peq_natk _#6 Hpeq12) as (_ & Hnatk).
so (peq_natk _#6 Hpeq23) as (Hnatk' & _).
so (natk_fun _#4 Hnatk Hnatk') as Heq.
injection Heq.
intros <-.
eapply peq_pi1; eauto; done.

(* pi2 *)
intros G1 G2 p1 p2 k1 k2 k1' k2' Hpeq12 IH G3 d l Hpeq.
invert Hpeq.
intros p3 e k3 e' k3' Hpeq23 <- _ <-.
so (peq_natk _#6 Hpeq12) as (_ & Hnatk).
so (peq_natk _#6 Hpeq23) as (Hnatk' & _).
so (natk_fun _#4 Hnatk Hnatk') as Heq.
injection Heq.
intros <- <-.
eapply peq_pi2; eauto; done.

(* star *)
intros G1 G2 G3 p3 k3 Hpeq.
invert Hpeq; intros; subst; auto with alg.

(* type *)
intros G1 G2 G3 k3 Haeq.
invert Haeq; intros; subst; auto with alg.

(* sing *)
intros G1 G2 c1 c2 _ IH G3 k3 Haeq.
invert Haeq.
intros c3 Haeq' <-.
apply aeqk_sing; auto; done.

(* pi *)
intros G1 G2 k1 k2 k1' k2' _ IH1 _ IH2 G3 l Haeq.
invert Haeq.
intros k3 k3' Haeq1 Haeq2 <-.
apply aeqk_pi; auto; done.

(* sigma *)
intros G1 G2 k1 k2 k1' k2' _ IH1 _ IH2 G3 l Haeq.
invert Haeq.
intros k3 k3' Haeq1 Haeq2 <-.
apply aeqk_sigma; auto; done.

(* one *)
intros G1 G2 G3 k3 Haeq.
invert Haeq; intros; subst; auto with alg.

(* epilogue *)
destruct_all Hind.
repeat2 split; intros; eauto; done.
Qed.


Lemma aeq_trans :
  forall G1 G2 G3 c1 c2 c3 k1 k2 k3,
    aeq G1 G2 c1 c2 k1 k2
    -> aeq G2 G3 c2 c3 k2 k3
    -> aeq G1 G3 c1 c3 k1 k3.
Proof.
exact (alg_trans andel).
Qed.


Lemma peq_trans :
  forall G1 G2 G3 p1 p2 p3 k1 k2 k3,
    peq G1 G2 p1 p2 k1 k2
    -> peq G2 G3 p2 p3 k2 k3
    -> peq G1 G3 p1 p3 k1 k3.
Proof.
exact (alg_trans ander andel).
Qed.


Lemma aeq_self_left :
  forall D D' c c' k k',
    aeq D D' c c' k k'
    -> aeq D D c c k k.
Proof.
intros.
eauto using aeq_symm, aeq_trans.
Qed.


Lemma peq_self_left :
  forall G G' p p' k k',
    peq G G' p p' k k'
    -> peq G G p p k k.
Proof.
intros.
eauto using peq_symm, peq_trans.
Qed.


Lemma peq_self_right :
  forall G G' p p' k k',
    peq G G' p p' k k'
    -> peq G' G' p' p' k' k'.
Proof.
eauto using peq_symm, peq_self_left.
Qed.
