
Require Import Tactics.
Require Import Extensionality.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Determinism.
Require Import Regularity.
Require Import Closure.
Require Import Preservation.
Require Declarative.
Require Inversion.
Require Consistent.
Require Progress.


Definition ev (value : term -> Prop) (step : term -> term -> Prop) (R : term -> term -> Prop) (m m' : term) :=
  (halt value step m <-> halt value step m')
  /\ (forall v v',
        star step m v
        -> star step m' v'
        -> value v
        -> value v'
        -> R v v').


Lemma ev_transpose :
  forall value step R,
    ev value step (transpose R) = transpose (ev value step R).
Proof.
intros value step R.
fextensionality 2.
intros e e'.
pextensionality.
  intro H.
  unfold transpose.
  destruct H as (Halfiff & Hrel).
  split.
    symmetry; assumption.

    intros v' v; intros.
    fold (transpose R v v').
    auto; done.

  intro H.
  unfold transpose in H.
  destruct H as (Halfiff & Hrel).
  split.
    symmetry; assumption.

    intros v v'; intros.
    unfold transpose.
    auto; done.
Qed.


Lemma steps_ev :
  forall value step R e e' d d',
    deterministic value step
    -> star step e d
    -> star step e' d'
    -> ev value step R e e'
    -> ev value step R d d'.
Proof.
intros value step R e e' d d' Hdeterm Hsteps Hsteps' Hev.
destruct Hev as (Hhaltiff & Hrel).
split.
  rewrite <- (steps_halt_iff _ _ Hdeterm _ _ Hsteps); [].
  rewrite <- (steps_halt_iff _ _ Hdeterm _ _ Hsteps'); [].
  assumption.

  intros v v' Hstepsv Hstepsv' Hval Hval'.
  apply Hrel; eauto using star_trans; done.
Qed.


Lemma anti_steps_ev :
  forall value step R e e' d d',
    deterministic value step
    -> star step e d
    -> star step e' d'
    -> ev value step R d d'
    -> ev value step R e e'.
Proof.
intros value step R e e' d d' Hdeterm Hsteps Hsteps' Hd.
destruct Hd as (Hhaltiff & Hrel).
split.
  rewrite -> (steps_halt_iff _ _ Hdeterm _ _ Hsteps); [].
  rewrite -> (steps_halt_iff _ _ Hdeterm _ _ Hsteps'); [].
  assumption.

  intros v v' Hstepsv Hstepsv' Hvalv Hvalv'.
  #2 apply Hrel; eauto.
    exact (steps_confluent_val _ _ Hdeterm _ _ _ Hsteps Hstepsv Hvalv).

    exact (steps_confluent_val _ _ Hdeterm _ _ _ Hsteps' Hstepsv' Hvalv').
Qed.


Lemma ev_funct :
  forall (value : term -> Prop) (step : term -> term -> Prop) (R R' : term -> term -> Prop) m m',
    (forall v v',
       star step m v 
       -> star step m' v'
       -> value v 
       -> value v'
       -> (R v v' -> R' v v'))
    -> ev value step R m m'
    -> ev value step R' m m'.
Proof.
intros value step R R' m m' Hincl Hev.
destruct Hev as (Hhaltiff & Hrel).
split; auto; done.
Qed.


Lemma tsteps_tev :
  forall t t' R e e' d d',
    star (tstep nil) e d
    -> star (tstep nil) e' d'
    -> tev t t' R e e'
    -> tev t t' R d d'.
Proof.
intros t t' R e e' d d' Hsteps Hsteps' Hev.
destruct Hev as (Hhaltiff & Hrel & Hofe & Hofe').
destruct (steps_ev _#7 (deterministic_tm nil) Hsteps Hsteps' (conj Hhaltiff Hrel)).
do2 3 split; eauto using preservations_tm, wf_nil.
Qed.


Lemma msteps_mev :
  forall t t' R e e' d d',
    star (mstep nil) e d
    -> star (mstep nil) e' d'
    -> mev t t' R e e'
    -> mev t t' R d d'.
Proof.
intros t t' R e e' d d' Hsteps Hsteps' Hev.
destruct Hev as (Hhaltiff & Hrel & Hofe & Hofe').
destruct (steps_ev _#7 (deterministic_md nil) Hsteps Hsteps' (conj Hhaltiff Hrel)).
do2 3 split; eauto using preservations_md, wf_nil.
Qed.


Lemma msteps_mevp :
  forall t t' R e e' d d',
    star (mstep nil) e d
    -> star (mstep nil) e' d'
    -> mevp t t' R e e'
    -> mevp t t' R d d'.
Proof.
intros t t' R e e' d d' Hsteps Hsteps' Hev.
destruct Hev as (Hhaltiff & Hrel & Hofe & Hofe').
destruct (steps_ev _#7 (deterministic_md nil) Hsteps Hsteps' (conj Hhaltiff Hrel)).
do2 3 split; eauto using preservations_md, wf_nil.
Qed.


Lemma anti_tsteps_tev :
  forall t t' R e e' d d',
    oft nil e t
    -> oft nil e' t'
    -> star (tstep nil) e d
    -> star (tstep nil) e' d'
    -> tev t t' R d d'
    -> tev t t' R e e'.
Proof.
intros t t' R e e' d d' Hofe Hofe' Hsteps Hsteps' Hev.
destruct Hev as (Hhaltiff & Hrel & _ & _).
destruct (anti_steps_ev _#7 (deterministic_tm nil) Hsteps Hsteps' (conj Hhaltiff Hrel)).
do2 3 split; auto.
Qed.


Lemma anti_msteps_mev :
  forall t t' R e e' d d',
    ofm nil impure e t
    -> ofm nil impure e' t'
    -> star (mstep nil) e d
    -> star (mstep nil) e' d'
    -> mev t t' R d d'
    -> mev t t' R e e'.
Proof.
intros t t' R e e' d d' Hofe Hofe' Hsteps Hsteps' Hev.
destruct Hev as (Hhaltiff & Hrel & _ & _).
destruct (anti_steps_ev _#7 (deterministic_md nil) Hsteps Hsteps' (conj Hhaltiff Hrel)).
do2 3 split; auto.
Qed.


Lemma anti_msteps_mevp :
  forall t t' R e e' d d',
    ofm nil pure e t
    -> ofm nil pure e' t'
    -> star (mstep nil) e d
    -> star (mstep nil) e' d'
    -> mevp t t' R d d'
    -> mevp t t' R e e'.
Proof.
intros t t' R e e' d d' Hofe Hofe' Hsteps Hsteps' Hev.
destruct Hev as (Hhaltiff & Hrel & _ & _).
destruct (anti_steps_ev _#7 (deterministic_md nil) Hsteps Hsteps' (conj Hhaltiff Hrel)).
do2 3 split; auto.
Qed.


Lemma mevp_funct :
  forall T T' (R R' : term -> term -> Prop) M M',
    (forall V V',
       star (mstep nil) M V
       -> star (mstep nil) M' V'
       -> mvalue nil V
       -> mvalue nil V'
       -> R V V' 
       -> R' V V')
    -> mevp T T' R M M'
    -> mevp T T' R' M M'.
Proof.
intros T T' R R' M M' Hincl Hev.
destruct Hev as (Hhaltiff & Hrel & HofM & HofM').
do2 3 split; auto; done.
Qed.
  

Definition md_appgnlt (F M : (* md *) term) (T : (* sg *) term) :=
  md_letm M (md_appgn (subst (sh 2) F) (var 0)) T.


Lemma subst_appgnlt :
  forall s M1 M2 T,
    subst s (md_appgnlt M1 M2 T)
    = md_appgnlt (subst s M1) (subst s M2) (subst s T).
Proof.
intros s M1 M2 T.
unfold md_appgnlt.
simp_sub.
reflexivity.
Qed.


Lemma ofm_appgnlt {G T1 T2 M1 M2} :
  wf G
  -> ofm G impure M1 (sg_pigen T1 (shift1 T2))
  -> ofm G impure M2 T1
  -> ofm G impure (md_appgnlt M1 M2 T2) T2.
Proof.
intros Hwf HofM1 HofM2.
unfold md_appgnlt.
so (Regularity.ofm_reg _#4 Hwf HofM1) as Hofpi.
invert Hofpi; clear Hofpi; [].
intros k1 HofT1 Hfst1 HofT2w.
so (Regularity.ofs_fst _#3 HofT1 Hfst1) as Hofk1.
so (Declarative.inhabitation _ _ Hofk1) as (c & Hofc).
so (subst1_ofs _#4 Hofc HofT2w) as HofT2.
simp_sub in HofT2.
clear c Hofc HofT2w.
so (Regularity.can_fsts _ _ HofT2) as (k2 & Hfst2).
eapply ofm_letm; eauto with static; [].
#2 eapply rel_quest.
  #3 eapply ofm_appgn.
    sor (refine (weaken2_ofm _#6 HofM1)) as H.
    simp_sub in H.
    exact H.

    #2 eapply rel_quest.
      eapply ofm_var; eauto with sequence; done.

      simp_sub.
      reflexivity.

    eapply fstm_var; eauto with sequence; done.

  simp_sub.
  reflexivity.
Qed.


Lemma mstep_appgnlt1 {G M1 M2 M2' T}
  : mstep G M2 M2'
    -> mstep G (md_appgnlt M1 M2 T) (md_appgnlt M1 M2' T).
Proof.
intro Hstep.
unfold md_appgnlt.
apply mstep_letm1; auto; done.
Qed.


Lemma mstep_appgnlt2 {G M1 V2 c2 T}
  : mvalue G V2
    -> fstm G V2 c2
    -> mstep G (md_appgnlt M1 V2 T) (md_appgn M1 V2).
Proof.
intros Hval Hfst.
unfold md_appgnlt.
#2 eapply rel_quest.
  eapply mstep_letm2; eauto; done.

  simp_sub.
  reflexivity.
Qed.


Lemma mstep_appgnlt2' {G T1 M1 V2 c2 T}
  : mvalue G V2
    -> fstm G V2 c2
    -> plus (mstep G) (md_appgnlt (md_lamgn T1 M1) V2 T) (subst (dot V2 (dot c2 id)) M1).
Proof.
intros HvalV2 HfstV2.
eexists.
split.
  eapply mstep_appgnlt2; eauto; done.

  #2 eapply star_step.
    eapply mstep_appgn3; eauto; done.

    simp_sub.
    apply star_refl; done.
Qed.


Lemma tps_ignore :
  forall t t' R,
    ofc nil t kd_type
    -> ofc nil t' kd_type
    -> tvrel t t' R
    -> tps t t' R (tm_lam t tm_star) (tm_lam t' tm_star).
Proof.
intros t t' R Hoft Hoft' HofR.
do2 4 split; auto with static dynamic; [].
intros v v' Hv.
so (HofR _ _ Hv) as (_ & _ & Hval & Hval').
split.
  intros _.
  exists tm_star; split; auto with dynamic; [].
  eapply star_step; eauto with dynamic; [].
  simp_sub.
  apply star_refl; done.

  intros _.
  exists tm_star; split; auto with dynamic; [].
  eapply star_step; eauto with dynamic; [].
  simp_sub.
  apply star_refl; done.
Qed.


Lemma mps_ignore :
  forall T T' R,
    ofs nil T
    -> ofs nil T'
    -> mvrel T T' R
    -> mps T T' R (md_lamgn T md_star) (md_lamgn T' md_star).
Proof.
intros T T' R HofT HofT' HofR.
so (Regularity.can_fsts _ _ HofT) as (k & HfstT).
so (Regularity.can_fsts _ _ HofT') as (k' & HfstT').
#3 do2 4 split; auto with dynamic.
  intros v v' Hv.
  so (HofR _ _ Hv) as (Hofv & Hofv' & Hval & Hval').
  so (Preservation.can_fstm_value _#4 wf_nil Hofv Hval) as (c & Hfstv).
  so (Preservation.can_fstm_value _#4 wf_nil Hofv' Hval') as (c' & Hfstv').
  split.
    intros _.
    exists md_star; split; auto with dynamic; [].
    eapply star_step; eauto with dynamic; [].
    simp_sub.
    apply star_refl; done.
  
    intros _.
    exists md_star; split; auto with dynamic; [].
    eapply star_step; eauto with dynamic; [].
    simp_sub.
    apply star_refl; done.

  apply ofm_forget; [].
  eapply ofm_lamgn; eauto; [].
  simp_sub.
  apply ofm_forget; [].
  apply ofm_star; done.

  apply ofm_forget; [].
  eapply ofm_lamgn; eauto; [].
  simp_sub.
  apply ofm_forget; [].
  apply ofm_star; done.
Qed.


Lemma ignore_halt_iff :
  forall t e,
    thalt nil (tm_app (tm_lam t tm_star) e) <-> thalt nil e.
Proof.
intros t e.
split.
  intro H.
  exact (Inversion.app_halt_inversion' _ _ H ander).

  intro H.
  destruct H as (v & Hsteps & Hval).
  exists tm_star; split; auto with dynamic; [].
  #2 apply (star_trans _ _ _ (tm_app (tm_lam t tm_star) v)).
    apply star_map'; auto with dynamic; done.

    eapply star_step; eauto with dynamic; [].
    simp_sub.
    apply star_refl; done.
Qed.


Lemma ignore_halt_iff_md :
  forall T M,
    ofm nil impure M T
    -> (mhalt nil (md_appgnlt (md_lamgn T md_star) M sg_one) <-> mhalt nil M).
Proof.
intros T M Hof.
split.
  intro H.
  unfold md_appgnlt in H.
  exact (Inversion.mletm_halt_inversion' _ _ _ H).

  intro H.
  destruct H as (V & Hsteps & Hval).
  exists md_star; split; auto with dynamic; [].
  #2 apply (star_trans _ _ _ (md_appgnlt (md_lamgn T md_star) V sg_one)).
    apply (star_map' _ _ (fun z => md_appgnlt _ z sg_one)); auto using mstep_appgnlt1 with dynamic; done.

    so (Preservation.can_fstm_value _#4 wf_nil (Preservation.preservations_md _#5 wf_nil Hof Hsteps) Hval) as (c & Hfst).
    #2 eapply star_trans.
      apply plus_star; [].
      eapply mstep_appgnlt2'; eauto; done.

      simp_sub.
      apply star_refl; done.
Qed.


Lemma tps_elim_tev :
  forall t t' R f f' e e',
    tps t t' R f f'
    -> tev t t' R e e'
    -> (thalt nil (tm_app f e) <-> thalt nil (tm_app f' e')).
Proof.
intros t t' R f f' e e' Hf He.
so Hf as (_ & _ & _ & Hvalf & Hvalf').
destruct He as (Hhaltiff & Hrel & _).
assert (thalt nil e -> (thalt nil (tm_app f e) <-> thalt nil (tm_app f' e'))) as Hprop.
  intros Hhalte.
  so (Hhaltiff andel Hhalte) as Hhalte'.
  destruct Hhalte as (v & Hsteps & Hvalv).
  destruct Hhalte' as (v' & Hsteps' & Hvalv').
  #2 rewrite -> (tsteps_thalt_iff nil (tm_app f e) (tm_app f v)).
    #2 rewrite -> (tsteps_thalt_iff nil (tm_app f' e') (tm_app f' v')).
      refine (Hf andel _ _ _); [].
      apply Hrel; auto; done.

      apply star_map'; auto with dynamic; done.

    apply star_map'; auto with dynamic; done.
split.
  intro Hhalt.
  so (Inversion.app_halt_inversion' _ _ Hhalt) as (_ & Hhalte).
  exact (Hprop Hhalte andel Hhalt).

  intro Hhalt.
  so (Inversion.app_halt_inversion' _ _ Hhalt) as (_ & Hhalte').
  exact (Hprop (Hhaltiff ander Hhalte') ander Hhalt).
Qed.


Lemma mps_elim_mev :
  forall t t' R f f' e e',
    mps t t' R f f'
    -> mev t t' R e e'
    -> (mhalt nil (md_appgnlt f e sg_one) <-> mhalt nil (md_appgnlt f' e' sg_one)).
Proof.
intros t t' R f f' e e' Hf He.
so Hf as (_ & _ & _ & Hvalf & Hvalf').
destruct He as (Hhaltiff & Hrel & Hofe & Hofe').
assert (mhalt nil e -> (mhalt nil (md_appgnlt f e sg_one) <-> mhalt nil (md_appgnlt f' e' sg_one))) as Hprop.
  intros Hhalte.
  so (Hhaltiff andel Hhalte) as Hhalte'.
  destruct Hhalte as (v & Hsteps & Hvalv).
  destruct Hhalte' as (v' & Hsteps' & Hvalv').
  so (preservations_md _#5 wf_nil Hofe Hsteps) as Hofv.
  so (preservations_md _#5 wf_nil Hofe' Hsteps') as Hofv'.
  so (can_fstm_value _#4 wf_nil Hofv Hvalv) as (c & Hfstv).
  so (can_fstm_value _#4 wf_nil Hofv' Hvalv') as (c' & Hfstv').
  #2 rewrite -> (msteps_mhalt_iff nil (md_appgnlt f e sg_one) (md_appgn f v)).
    #2 rewrite -> (msteps_mhalt_iff nil (md_appgnlt f' e' sg_one) (md_appgn f' v')).
      refine (Hf andel _ _ _); [].
      apply Hrel; auto; done.

      #2 apply (star_trans _#3 (md_appgnlt f' v' sg_one)).
        apply (star_map' _ _ (fun z => md_appgnlt _ z _)); auto using mstep_appgnlt1; done.

        apply star_one; [].
        eapply mstep_appgnlt2; eauto; done.

    #2 apply (star_trans _#3 (md_appgnlt f v sg_one)).
      apply (star_map' _ _ (fun z => md_appgnlt _ z _)); auto using mstep_appgnlt1; done.

      apply star_one; [].
      eapply mstep_appgnlt2; eauto; done.
split.
  intro Hhalt.
  unfold md_appgnlt.
  so (Inversion.mletm_halt_inversion' _ _ _ Hhalt) as Hhalte.
  exact (Hprop Hhalte andel Hhalt).

  intro Hhalt.
  so (Inversion.mletm_halt_inversion' _ _ _ Hhalt) as Hhalte'.
  exact (Hprop (Hhaltiff ander Hhalte') ander Hhalt).
Qed.


Lemma tev_tpst_intro :
  forall t t' R e e',
    tvrel t t' R
    -> oft nil e t
    -> oft nil e' t'
    -> (forall f f',
          tps t t' R f f'
          -> (thalt nil (tm_app f e) <-> thalt nil (tm_app f' e')))
    -> tev t t' (tpst t t' R) e e'.
Proof.
intros t t' R e e' HofR Hofe Hofe' Hrel.
so (Regularity.oft_reg _#3 wf_nil Hofe) as Hoft.
so (Regularity.oft_reg _#3 wf_nil Hofe') as Hoft'.
#2 do2 3 split; auto.
  so (Hrel _ _ (tps_ignore _#3 Hoft Hoft' HofR)) as Hiff.
  rewrite <- (ignore_halt_iff t e); [].
  rewrite <- (ignore_halt_iff t' e'); [].
  rewrite -> Hiff; [].
  reflexivity.

  intros v v' Hsteps Hsteps' Hvalv Hvalv'.
  #3 do2 4 split; auto.
    intros f f' Hf.
    so Hf as (_ & _ & _ & Hvalf & Hvalf').
    #2 rewrite <- (tsteps_thalt_iff nil (tm_app f e) (tm_app f v)).
      #2 rewrite <- (tsteps_thalt_iff nil (tm_app f' e') (tm_app f' v')).
        apply Hrel; auto; done.

        apply star_map'; auto; [].
        intros; apply tstep_app2; auto; done.

      apply star_map'; auto; [].
      intros; apply tstep_app2; auto; done.

  eapply Preservation.preservations_tm; eauto using wf_nil; done.

  eapply Preservation.preservations_tm; eauto using wf_nil; done.
Qed.
   

Lemma mev_mpst_intro :
  forall T T' R M M',
    mvrel T T' R
    -> ofm nil impure M T
    -> ofm nil impure M' T'
    -> (forall F F',
          mps T T' R F F'
          -> (mhalt nil (md_appgnlt F M sg_one) <-> mhalt nil (md_appgnlt F' M' sg_one)))
    -> mev T T' (mpst T T' R) M M'.
Proof.
intros t t' R e e' HofR Hofe Hofe' Hrel.
so (Regularity.ofm_reg _#4 wf_nil Hofe) as Hoft.
so (Regularity.ofm_reg _#4 wf_nil Hofe') as Hoft'.
#2 do2 3 split; auto.
  so (Hrel _ _ (mps_ignore _#3 Hoft Hoft' HofR)) as Hiff.
  rewrite <- (ignore_halt_iff_md t e Hofe); [].
  rewrite <- (ignore_halt_iff_md t' e' Hofe'); [].
  rewrite -> Hiff; [].
  reflexivity.

  intros v v' Hsteps Hsteps' Hvalv Hvalv'.
  so (Preservation.preservations_md _#5 wf_nil Hofe Hsteps) as Hofv.
  so (Preservation.preservations_md _#5 wf_nil Hofe' Hsteps') as Hofv'.
  so (Preservation.can_fstm_value _#4 wf_nil Hofv Hvalv) as (c & Hfstv).
  so (Preservation.can_fstm_value _#4 wf_nil Hofv' Hvalv') as (c' & Hfstv').
  do2 4 split; auto; [].
  intros f f' Hf.
  so Hf as (_ & _ & _ & Hvalf & Hvalf').
  #2 rewrite <- (msteps_mhalt_iff nil (md_appgnlt f e sg_one) (md_appgn f v)).
    #2 rewrite <- (msteps_mhalt_iff nil (md_appgnlt f' e' sg_one) (md_appgn f' v')).
      apply Hrel; auto; done.

      #2 apply (@star_trans _#3 (md_appgnlt f' v' sg_one)).
        apply (star_map' _ _ (fun z => md_appgnlt f' z sg_one)) ; auto; [].
        intros; apply mstep_appgnlt1; auto; done.

        unfold md_appgnlt.
        #2 eapply star_step.
          eapply mstep_letm2; eauto; done.

          simp_sub.
          apply star_refl; done.
          
    #2 apply (@star_trans _#3 (md_appgnlt f v sg_one)).
      apply (star_map' _ _ (fun z => md_appgnlt f z sg_one)) ; auto; [].
      intros; apply mstep_appgnlt1; auto; done.

      unfold md_appgnlt.
      #2 eapply star_step.
        eapply mstep_letm2; eauto; done.

        simp_sub.
        apply star_refl; done.
Qed.
   

Inductive unit_rel : term -> term -> Prop :=
| unit_rel_
    : unit_rel tm_star tm_star
.


Lemma unit_rel_tpclosed :
  tpclosed cn_unit cn_unit unit_rel.
Proof.
unfold tpclosed.
intros v v' Hst.
unfold tpst in Hst.
unfold tpt in Hst.
destruct Hst as (_ & Hofv & Hofv' & Hval & Hval').
so (Progress.canon_tm _ _ Hofv Consistent.similar_unit Hval) as H.
invert H; clear H; intros <-.
so (Progress.canon_tm _ _ Hofv' Consistent.similar_unit Hval') as H.
invert H; clear H; intros <-.
apply unit_rel_; done.
Qed.


Inductive arrow_rel
  (t1 t1' t2 t2' : (* cn *) term) (R1 R2 : term -> term -> Prop) : term -> term -> Prop :=
| arrow_rel_ {v v'}
    : oft nil v (cn_arrow t1 t2)
      -> oft nil v' (cn_arrow t1' t2')
      -> tvalue nil v
      -> tvalue nil v'
      -> (forall w w',
            R1 w w'
            -> tev t2 t2' R2 (tm_app v w) (tm_app v' w'))
      -> arrow_rel t1 t1' t2 t2' R1 R2 v v'
.


Lemma arrow_rel_tvrel :
  forall t1 t1' t2 t2' R1 R2,
    tvrel (cn_arrow t1 t2) (cn_arrow t1' t2') (arrow_rel t1 t1' t2 t2' R1 R2).
Proof.
intros t1 t1' t2 t2' R1 R2.
intros v v' Hv.
invert Hv; [].
intros Hofv Hofv' Hvalv Hvalv' _.
auto; done.
Qed.


Lemma arrow_rel_transpose :
  forall t1 t1' t2 t2' R1 R2,
    transpose (arrow_rel t1 t1' t2 t2' R1 R2) = arrow_rel t1' t1 t2' t2 (transpose R1) (transpose R2).
Proof.
intros t1 t1' t2 t2' R1 R2.
fextensionality 2.
intros v' v.
pextensionality.
  intro H.
  unfold transpose in H.
  invert H; [].
  intros Hof Hof' Hval Hval' Hrel.
  apply arrow_rel_; auto; [].
  intros w' w Hw.
  unfold transpose in Hw.
  rewrite -> tev_transpose; [].
  unfold transpose.
  apply Hrel; auto; done.

  intro H.
  unfold transpose.
  invert H; [].
  intros Hof Hof' Hval Hval' Hrel.
  apply arrow_rel_; auto; [].
  intros w w' Hw.
  unfold transpose in Hrel at 1.
  so (Hrel _ _ Hw) as H'.
  rewrite -> tev_transpose in H'; [].
  unfold transpose in H'.
  assumption.
Qed.


Lemma biorthogonality_lemma :
  forall t t' u u' R S f f' w w' g g' e e',
    tvrel u u' S
    -> tps t t' R f f'
    -> tpst u u' S w w'
    -> arrow_rel u u' t t' S R g g'
    -> star (tstep nil) (tm_app g w) e
    -> star (tstep nil) (tm_app g' w') e'
    -> (thalt nil (tm_app f e) <-> thalt nil (tm_app f' e')).
Proof.
intros t t' u u' R S f f' w w' g g' e e' HofS Hf Hw Hg Hstepse Hstepse'.
so Hf as (_ & Hoff & Hoff' & Hvalf & Hvalf').
so Hw as (_ & Hofw & Hofw' & Hvalw & Hvalw').
so (Regularity.oft_reg _#3 wf_nil Hofw) as Hofu.
so (Regularity.oft_reg _#3 wf_nil Hofw') as Hofu'.
invert Hg; [].
intros Hofg Hofg' _ _ _.
#2 rewrite <- (tsteps_thalt_iff nil (tm_app (tm_lam u (tm_app (shift1 f) (tm_app (shift1 g) (var 0)))) w) (tm_app f e)).
  #2 rewrite <- (tsteps_thalt_iff nil (tm_app (tm_lam u' (tm_app (shift1 f') (tm_app (shift1 g') (var 0)))) w') (tm_app f' e')).
    refine (Hw andel _ _ _); [].
    #3 do2 4 split; auto with dynamic.
      intros x x' Hx.
      so (HofS _ _ Hx) as (_ & _ & Hvalx & Hvalx').
      #2 rewrite -> (tsteps_thalt_iff nil (tm_app (tm_lam u (tm_app (shift1 f) (tm_app (shift1 g) (var 0)))) x) (tm_app f (tm_app g x))).
        #2 rewrite -> (tsteps_thalt_iff nil (tm_app (tm_lam u' (tm_app (shift1 f') (tm_app (shift1 g') (var 0)))) x') (tm_app f' (tm_app g' x'))).
          eapply tps_elim_tev; eauto; [].
          invert Hg; [].
          intros _ _ _ _ Hrel.
          apply Hrel; auto; done.

          #2 eapply star_step.
            apply tstep_app3; auto; done.

            simp_sub.
            apply star_refl; done.

        #2 eapply star_step.
          apply tstep_app3; auto; done.

          simp_sub.
          apply star_refl; done.

      apply oft_lam; auto; [].
      sor (refine (weaken1_oft _#4 Hoff)) as H1.
      simp_sub in H1.
      eapply oft_app; eauto; [].
      sor (refine (weaken1_oft _#4 Hofg)) as H2.
      simp_sub in H2.
      eapply oft_app; eauto; [].
      apply oft_var; auto with sequence; done.

      apply oft_lam; auto; [].
      sor (refine (weaken1_oft _#4 Hoff')) as H1.
      simp_sub in H1.
      eapply oft_app; eauto; [].
      sor (refine (weaken1_oft _#4 Hofg')) as H2.
      simp_sub in H2.
      eapply oft_app; eauto; [].
      apply oft_var; auto with sequence; done.

    #2 eapply star_step.
      apply tstep_app3; auto; done.
  
      simp_sub.
      apply star_map'; auto; [].
      intros; auto using tstep_app2; done.

  #2 eapply star_step.
    apply tstep_app3; auto; done.

    simp_sub.
    apply star_map'; auto; [].
    intros; auto using tstep_app2; done.
Qed.


Lemma app_fn_arrow :
  forall t1 t1' t2 t2' R1 R2 v v',
    tvrel t1 t1' R1
    -> ofc nil t2 kd_type
    -> ofc nil t2' kd_type
    -> R1 v v'
    -> arrow_rel (cn_arrow t1 t2) (cn_arrow t1' t2') t2 t2' (arrow_rel t1 t1' t2 t2' R1 R2) R2 (tm_lam (cn_arrow t1 t2) (tm_app (var 0) (shift1 v))) (tm_lam (cn_arrow t1' t2') (tm_app (var 0) (shift1 v'))).
Proof.
intros t1 t1' t2 t2' R1 R2 v v' HofR1 Hoft2 Hoft2' Hv.
so (HofR1 _ _ Hv) as (Hofv & Hofv' & Hvalv & Hvalv').
so (Regularity.oft_reg _#3 wf_nil Hofv) as Hoft1.
so (Regularity.oft_reg _#3 wf_nil Hofv') as Hoft1'.
assert (oft nil (tm_lam (cn_arrow t1 t2) (tm_app (var 0) (shift1 v))) (cn_arrow (cn_arrow t1 t2) t2)) as Hoflam.
  apply oft_lam; auto with static; [].
  #2 apply (@oft_app _ (shift1 t1)).
    so (@oft_var (nil; cl_tm (cn_arrow t1 t2)) _ _ index_0) as H.
    simp_sub in H.
    assumption.

    apply weaken1_oft; auto; done.
assert (oft nil (tm_lam (cn_arrow t1' t2') (tm_app (var 0) (shift1 v'))) (cn_arrow (cn_arrow t1' t2') t2')) as Hoflam'.
  apply oft_lam; auto with static; [].
  #2 apply (@oft_app _ (shift1 t1')).
    so (@oft_var (nil; cl_tm (cn_arrow t1' t2')) _ _ index_0) as H.
    simp_sub in H.
    assumption.

    apply weaken1_oft; auto; done.
apply arrow_rel_; auto with dynamic; [].
intros w w' Hw.
invert Hw; [].
intros Hofw Hofw' Hvalw Hvalw' Hrel.
so (Hrel _ _ Hv) as Hwv.
#2 eapply anti_tsteps_tev; eauto with static.
  #2 eapply star_step.
    apply tstep_app3; auto; done.
    
    simp_sub.
    apply star_refl; done.

  #2 eapply star_step.
    apply tstep_app3; auto; done.
    
    simp_sub.
    apply star_refl; done.
Qed.


Lemma arrow_rel_tpclosed :
  forall t1 t1' t2 t2' R1 R2,
    ofc nil t1 kd_type
    -> ofc nil t1' kd_type
    -> ofc nil t2 kd_type
    -> ofc nil t2' kd_type
    -> tvrel t1 t1' R1
    -> tvrel t2 t2' R2
    -> tpclosed t2 t2' R2
    -> tpclosed (cn_arrow t1 t2) (cn_arrow t1' t2') (arrow_rel t1 t1' t2 t2' R1 R2).
Proof.
intros t1 t1' t2 t2' R1 R2 Hoft1 Hoft1' Hoft2 Hoft2' HofR1 HofR2 Hcl.
intros v v' Hv.
so Hv as (_ & Hofv & Hofv' & Hvalv & Hvalv').
apply arrow_rel_; auto; [].
intros w w' Hw.
so (HofR1 _ _ Hw) as (Hofw & Hofw' & Hvalw & Hvalw').
rewrite <- (tpclosed_eq_tpst _#3 HofR2 Hcl); [].
apply tev_tpst_intro; eauto with static; [].
intros f f' Hf.
#4 exploit (biorthogonality_lemma t2 t2' (cn_arrow t1 t2) (cn_arrow t1' t2') R2 (arrow_rel t1 t1' t2 t2' R1 R2) f f' v v' (tm_lam (cn_arrow t1 t2) (tm_app (var 0) (shift1 w))) (tm_lam (cn_arrow t1' t2') (tm_app (var 0) (shift1 w'))) (tm_app v w) (tm_app v' w')) as H; auto.
  apply arrow_rel_tvrel; done.

  apply app_fn_arrow; auto; done.

  #2 eapply star_step.
    apply tstep_app3; auto; done.

    simp_sub.
    apply star_refl; done.

  #2 eapply star_step.
    apply tstep_app3; auto; done.

    simp_sub.
    apply star_refl; done.
Qed.


Inductive prod_rel
  (t1 t1' t2 t2' : (* cn *) term) (R1 R2 : term -> term -> Prop) : term -> term -> Prop :=
| prod_rel_ {v1 v1' v2 v2'}
    : oft nil (tm_pair v1 v2) (cn_prod t1 t2)
      -> oft nil (tm_pair v1' v2') (cn_prod t1' t2')
      -> tvalue nil (tm_pair v1 v2)
      -> tvalue nil (tm_pair v1' v2')
      -> R1 v1 v1'
      -> R2 v2 v2'
      -> prod_rel t1 t1' t2 t2' R1 R2 (tm_pair v1 v2) (tm_pair v1' v2')
.


Lemma prod_rel_tvrel :
  forall t1 t1' t2 t2' R1 R2,
    tvrel (cn_prod t1 t2) (cn_prod t1' t2') (prod_rel t1 t1' t2 t2' R1 R2).
Proof.
intros t1 t1' t2 t2' R1 R2.
intros v v' Hv.
invert Hv; [].
intros v1 v1' v2 v2' Hofv Hofv' Hvalv Hvalv' Hv1 Hv2 <- <-.
auto; done.
Qed.


Lemma pi1_fn_arrow :
  forall t1 t1' t2 t2' R1 R2,
    ofc nil t1 kd_type
    -> ofc nil t1' kd_type
    -> ofc nil t2 kd_type
    -> ofc nil t2' kd_type
    -> arrow_rel (cn_prod t1 t2) (cn_prod t1' t2') t1 t1' (prod_rel t1 t1' t2 t2' R1 R2) R1 (tm_lam (cn_prod t1 t2) (tm_pi1 (var 0))) (tm_lam (cn_prod t1' t2') (tm_pi1 (var 0))).
Proof.
intros t1 t1' t2 t2' R1 R2 Hoft1 Hoft1' Hoft2 Hoft2'.
assert (oft nil (tm_lam (cn_prod t1 t2) (tm_pi1 (var 0))) (cn_arrow (cn_prod t1 t2) t1)) as Hoflam.
  apply oft_lam; auto with static; [].
  apply (@oft_pi1 _ _ (shift1 t2)); [].
  so (@oft_var (nil; cl_tm (cn_prod t1 t2)) _ _ index_0) as H.
  simp_sub in H.
  assumption.
assert (oft nil (tm_lam (cn_prod t1' t2') (tm_pi1 (var 0))) (cn_arrow (cn_prod t1' t2') t1')) as Hoflam'.
  apply oft_lam; auto with static; [].
  apply (@oft_pi1 _ _ (shift1 t2')); [].
  so (@oft_var (nil; cl_tm (cn_prod t1' t2')) _ _ index_0) as H.
  simp_sub in H.
  assumption.
apply arrow_rel_; auto with dynamic; [].
intros w w' Hw.
invert Hw; [].
intros v1 v1' v2 v2' Hofv Hofv' Hval Hval' Hv1 _ <- <-.
invert Hval; []; intros Hval1 Hval2.
invert Hval'; []; intros Hval1' Hval2'.
assert (star (tstep nil) (tm_app (tm_lam (cn_prod t1 t2) (tm_pi1 (var 0))) (tm_pair v1 v2)) v1) as Hsteps.
  #2 eapply star_step.
    apply tstep_app3; auto; done.

    simp_sub.
    apply star_one; [].
    apply tstep_pi12; auto; done.
assert (star (tstep nil) (tm_app (tm_lam (cn_prod t1' t2') (tm_pi1 (var 0))) (tm_pair v1' v2')) v1') as Hsteps'.
  #2 eapply star_step.
    apply tstep_app3; auto; done.

    simp_sub.
    apply star_one; [].
    apply tstep_pi12; auto; done.
do2 3 split.
  split.
    intros _.
    exists v1'; split; auto; done.
  
    intros _.
    exists v1; split; auto; done.

  intros v v' Hstepsa Hstepsa' Hvalv Hvalv'.
  repl v with v1 by (eapply tvalue_unique; eauto).
  repl v' with v1' by (eapply tvalue_unique; eauto).
  assumption.

  eapply oft_app; eauto with static; done.

  eapply oft_app; eauto with static; done.
Qed.


Lemma pi2_fn_arrow :
  forall t1 t1' t2 t2' R1 R2,
    ofc nil t1 kd_type
    -> ofc nil t1' kd_type
    -> ofc nil t2 kd_type
    -> ofc nil t2' kd_type
    -> arrow_rel (cn_prod t1 t2) (cn_prod t1' t2') t2 t2' (prod_rel t1 t1' t2 t2' R1 R2) R2 (tm_lam (cn_prod t1 t2) (tm_pi2 (var 0))) (tm_lam (cn_prod t1' t2') (tm_pi2 (var 0))).
Proof.
intros t1 t1' t2 t2' R1 R2 Hoft1 Hoft1' Hoft2 Hoft2'.
assert (oft nil (tm_lam (cn_prod t1 t2) (tm_pi2 (var 0))) (cn_arrow (cn_prod t1 t2) t2)) as Hoflam.
  apply oft_lam; auto with static; [].
  apply (@oft_pi2 _ (shift1 t1)); [].
  so (@oft_var (nil; cl_tm (cn_prod t1 t2)) _ _ index_0) as H.
  simp_sub in H.
  assumption.
assert (oft nil (tm_lam (cn_prod t1' t2') (tm_pi2 (var 0))) (cn_arrow (cn_prod t1' t2') t2')) as Hoflam'.
  apply oft_lam; auto with static; [].
  apply (@oft_pi2 _ (shift1 t1')); [].
  so (@oft_var (nil; cl_tm (cn_prod t1' t2')) _ _ index_0) as H.
  simp_sub in H.
  assumption.
apply arrow_rel_; auto with dynamic; [].
intros w w' Hw.
invert Hw; [].
intros v1 v1' v2 v2' Hofv Hofv' Hval Hval' _ Hv2 <- <-.
invert Hval; []; intros Hval1 Hval2.
invert Hval'; []; intros Hval1' Hval2'.
assert (star (tstep nil) (tm_app (tm_lam (cn_prod t1 t2) (tm_pi2 (var 0))) (tm_pair v1 v2)) v2) as Hsteps.
  #2 eapply star_step.
    apply tstep_app3; auto; done.

    simp_sub.
    apply star_one; [].
    apply tstep_pi22; auto; done.
assert (star (tstep nil) (tm_app (tm_lam (cn_prod t1' t2') (tm_pi2 (var 0))) (tm_pair v1' v2')) v2') as Hsteps'.
  #2 eapply star_step.
    apply tstep_app3; auto; done.

    simp_sub.
    apply star_one; [].
    apply tstep_pi22; auto; done.
do2 3 split.
  split.
    intros _.
    exists v2'; split; auto; done.
  
    intros _.
    exists v2; split; auto; done.

  intros v v' Hstepsa Hstepsa' Hvalv Hvalv'.
  repl v with v2 by (eapply tvalue_unique; eauto).
  repl v' with v2' by (eapply tvalue_unique; eauto).
  assumption.

  eapply oft_app; eauto with static; done.

  eapply oft_app; eauto with static; done.
Qed.


Lemma prod_rel_tpclosed :
  forall t1 t1' t2 t2' R1 R2,
    ofc nil t1 kd_type
    -> ofc nil t1' kd_type
    -> ofc nil t2 kd_type
    -> ofc nil t2' kd_type
    -> tvrel t1 t1' R1
    -> tvrel t2 t2' R2
    -> tpclosed t1 t1' R1
    -> tpclosed t2 t2' R2
    -> tpclosed (cn_prod t1 t2) (cn_prod t1' t2') (prod_rel t1 t1' t2 t2' R1 R2).
Proof.
intros t1 t1' t2 t2' R1 R2 Hoft1 Hoft1' Hoft2 Hoft2' HofR1 HofR2 Hcl1 Hcl2.
intros v v' Hv.
so Hv as (_ & Hofv & Hofv' & Hvalv & Hvalv').
invert (Progress.canon_tm _ _ Hofv Consistent.similar_prod Hvalv); [].
intros v1 v2 <-.
invert (Progress.canon_tm _ _ Hofv' Consistent.similar_prod Hvalv'); [].
intros v1' v2' <-.
so (Inversion.pair_oft_inversion_prod _#5 wf_nil Hofv) as (Hofv1 & Hofv2).
so (Inversion.pair_oft_inversion_prod _#5 wf_nil Hofv') as (Hofv1' & Hofv2').
invert Hvalv; [].
intros Hvalv1 Hvalv2.
invert Hvalv'; [].
intros Hvalv1' Hvalv2'.
#2 apply prod_rel_; auto; clear Hofv Hofv' Hvalv Hvalv'.
  apply Hcl1; [].
  do2 4 split; auto; [].
  intros f f' Hf.
  #4 exploit (biorthogonality_lemma t1 t1' (cn_prod t1 t2) (cn_prod t1' t2') R1 (prod_rel t1 t1' t2 t2' R1 R2) f f' (tm_pair v1 v2) (tm_pair v1' v2') (tm_lam (cn_prod t1 t2) (tm_pi1 (var 0))) (tm_lam (cn_prod t1' t2') (tm_pi1 (var 0))) v1 v1') as H; auto.
    apply prod_rel_tvrel; auto; done.
    
    apply pi1_fn_arrow; auto; done.

    #2 eapply star_step.
      apply tstep_app3; auto with dynamic; done.

      simp_sub.
      apply star_one; [].
      apply tstep_pi12; auto; done.

    #2 eapply star_step.
      apply tstep_app3; auto with dynamic; done.

      simp_sub.
      apply star_one; [].
      apply tstep_pi12; auto; done.

  apply Hcl2; [].
  do2 4 split; auto; [].
  intros f f' Hf.
  #4 exploit (biorthogonality_lemma t2 t2' (cn_prod t1 t2) (cn_prod t1' t2') R2 (prod_rel t1 t1' t2 t2' R1 R2) f f' (tm_pair v1 v2) (tm_pair v1' v2') (tm_lam (cn_prod t1 t2) (tm_pi2 (var 0))) (tm_lam (cn_prod t1' t2') (tm_pi2 (var 0))) v2 v2') as H; auto.
    apply prod_rel_tvrel; auto; done.
    
    apply pi2_fn_arrow; auto; done.

    #2 eapply star_step.
      apply tstep_app3; auto with dynamic; done.

      simp_sub.
      apply star_one; [].
      apply tstep_pi22; auto; done.

    #2 eapply star_step.
      apply tstep_app3; auto with dynamic; done.

      simp_sub.
      apply star_one; [].
      apply tstep_pi22; auto; done.
Qed.


Inductive all_rel
  (C : Type) (k k' : (* kd *) term) (t t' : (* cn *)  term) 
  (R : (* cn *) term -> (* cn *) term -> C -> Prop)
  (S : (* cn *) term -> (* cn *) term -> C -> (* tm *) term -> (* tm *) term -> Prop)
  : term -> term -> Prop :=
| all_rel_ {v v'}
    : oft nil v (cn_all k t)
      -> oft nil v' (cn_all k' t')
      -> tvalue nil v
      -> tvalue nil v'
      -> (forall c c' Q,
            R c c' Q
            -> tev (subst1 c t) (subst1 c' t') (S c c' Q) (tm_papp v c) (tm_papp v' c'))
      -> all_rel C k k' t t' R S v v'.


Lemma all_rel_tvrel :
  forall C k k' t t' R S,
    tvrel (cn_all k t) (cn_all k' t') (all_rel C k k' t t' R S).
Proof.
intros C k k' t t' R S.
intros v v' Hv.
invert Hv; [].
intros; auto; done.
Qed.


Lemma papp_fn_arrow :
  forall C k k' t t' (R : term -> term -> C -> Prop) S c c' Q,
    ofc nil c k
    -> ofc nil c' k'
    -> ofc (nil; cl_cn k) t kd_type
    -> ofc (nil; cl_cn k') t' kd_type
    -> R c c' Q
    -> arrow_rel (cn_all k t) (cn_all k' t') (subst1 c t) (subst1 c' t') (all_rel C k k' t t' R S) (S c c' Q) (tm_lam (cn_all k t) (tm_papp (var 0) (shift1 c))) (tm_lam (cn_all k' t') (tm_papp (var 0) (shift1 c'))).
Proof.
intros C k k' t t' R S c c' Q Hofc Hofc' Hoft Hoft' Hc.
so (Regularity.ofc_reg _#3 wf_nil Hofc) as Hofk.
so (Regularity.ofc_reg _#3 wf_nil Hofc') as Hofk'.
assert (oft nil (tm_lam (cn_all k t) (tm_papp (var 0) (shift1 c))) (cn_arrow (cn_all k t) (subst1 c t))) as Hoflam.
  apply oft_lam; auto with static; [].
  simp_sub.
  replsub (subst (dot (shift1 c) sh1) t) with (subst1 (shift1 c) (subst (dot (var 0) (sh 2)) t)).
  #2 apply (@oft_papp _ (shift1 k)).
    so (@oft_var (nil; cl_tm (cn_all k t)) _ _ index_0) as H.
    simp_sub in H.
    eassumption.

    apply weaken1_ofc; auto; done.
assert (oft nil (tm_lam (cn_all k' t') (tm_papp (var 0) (shift1 c'))) (cn_arrow (cn_all k' t') (subst1 c' t'))) as Hoflam'.
  apply oft_lam; auto with static; [].
  simp_sub.
  replsub (subst (dot (shift1 c') sh1) t') with (subst1 (shift1 c') (subst (dot (var 0) (sh 2)) t')).
  #2 apply (@oft_papp _ (shift1 k')).
    so (@oft_var (nil; cl_tm (cn_all k' t')) _ _ index_0) as H.
    simp_sub in H.
    eassumption.

    apply weaken1_ofc; auto; done.
apply arrow_rel_; auto with dynamic; [].
intros v v' Hv.
invert Hv; [].
intros Hofv Hofv' Hvalv Hvalv' Hrel.
so (Hrel _ _ _ Hc) as Hvc; clear Hrel.
#2 eapply anti_tsteps_tev; eauto with static.
  #2 eapply star_step.
    apply tstep_app3; auto; done.

    simp_sub.
    apply star_refl; done.

  #2 eapply star_step.
    apply tstep_app3; auto; done.

    simp_sub.
    apply star_refl; done.
Qed.


Lemma all_rel_tpclosed :
  forall C k k' t t' (R : term -> term -> C -> Prop) S,
    ofk nil k
    -> ofk nil k'
    -> ofc (nil; cl_cn k) t kd_type
    -> ofc (nil; cl_cn k') t' kd_type
    -> (forall c c' Q, R c c' Q ->
          ofc nil c k 
          /\ ofc nil c' k'
          /\ tvrel (subst1 c t) (subst1 c' t') (S c c' Q)
          /\ tpclosed (subst1 c t) (subst1 c' t') (S c c' Q))
    -> tpclosed (cn_all k t) (cn_all k' t') (all_rel C k k' t t' R S).
Proof.
intros C k k' t t' R S Hofk Hofk' Hoft Hoft' Hprops.
intros v v' Hv.
so Hv as (_ & Hofv & Hofv' & Hvalv & Hvalv').
apply all_rel_; auto; [].
intros c c' Q Hc.
so (Hprops _ _ _ Hc) as (Hofc & Hofc' & HofS & Hcl); clear Hprops.
rewrite <- (tpclosed_eq_tpst _#3 HofS Hcl); [].
apply tev_tpst_intro; eauto with static; [].
intros f f' Hf.
#4 exploit (biorthogonality_lemma (subst1 c t) (subst1 c' t') (cn_all k t) (cn_all k' t') (S c c' Q) (all_rel C k k' t t' R S) f f' v v' (tm_lam (cn_all k t) (tm_papp (var 0) (shift1 c))) (tm_lam (cn_all k' t') (tm_papp (var 0) (shift1 c'))) (tm_papp v c) (tm_papp v' c')) as H; auto.
  apply all_rel_tvrel; done.

  apply papp_fn_arrow; auto; done.

  #2 eapply star_step.
    apply tstep_app3; auto; done.

    simp_sub.
    apply star_refl; done.

  #2 eapply star_step.
    apply tstep_app3; auto; done.

    simp_sub.
    apply star_refl; done.
Qed.


Inductive dyn_rel
  (t t' : (* cn *) term) (R : term -> term -> Prop) : term -> term -> Prop :=
| dyn_rel_ {v v'}
    : oft nil v t
      -> oft nil v' t'
      -> tvalue nil v
      -> tvalue nil v'
      -> R v v'
      -> dyn_rel t t' R (md_dyn v) (md_dyn v').


Lemma dyn_rel_mvrel :
  forall t t' R,
    mvrel (sg_dyn t) (sg_dyn t') (dyn_rel t t' R).
Proof.
intros t t' R.
intros V V' HV.
invert HV; [].
intros v v' Hofv Hofv' Hval Hval' Hv <- <-.
do2 3 split; auto with static dynamic.
Qed.


Lemma thalt_into_mod :
  forall e,
    thalt nil e <-> mhalt nil (md_lett e md_star).
Proof.
intro e.
split.
  intro Hhalt.
  destruct Hhalt as (v & Hsteps & Hval).
  exists md_star.
  split; auto with dynamic; [].
  #2 apply (star_trans _#3 (md_lett v md_star)).
    eapply (star_map _#4 (fun z => md_lett z _)); eauto with dynamic; done.

    #2 eapply star_step.
      apply mstep_lett2; auto; done.

      simp_sub.
      apply star_refl; done.

  intro Hhalt.
  eapply Inversion.mlett_halt_inversion'; eauto; done.
Qed.


Lemma app_thalt_into_mod :
  forall t f v,
    tvalue nil f
    -> tvalue nil v
    -> (thalt nil (tm_app f v)
        <->
        mhalt nil (md_appgn (md_lamgn (sg_dyn t) (md_lett (tm_app (subst (sh 2) f) (tm_ext (var 0))) md_star)) (md_dyn v))).
Proof.
intros t f v Hvalf Hvalv.
#2 rewrite -> (msteps_mhalt_iff _ _ (md_lett (tm_app f v) md_star)).
  apply thalt_into_mod; done.

  #2 eapply star_step.
    eapply mstep_appgn3; eauto with static dynamic; done.

    simp_sub.
    apply star_one; auto with dynamic; done.
Qed.


Lemma dyn_rel_mpclosed :
  forall t t' R,
    ofc nil t kd_type
    -> ofc nil t' kd_type
    -> tvrel t t' R
    -> tpclosed t t' R
    -> mpclosed (sg_dyn t) (sg_dyn t') (dyn_rel t t' R).
Proof.
intros t t' R Hoft Hoft' HofR Hcl.
intros V V' HV.
so HV as (_ & HofV & HofV' & HvalV & HvalV').
invert (Progress.canon_md _#3 HofV HvalV); [].
intros v <-.
invert (Progress.canon_md _#3 HofV' HvalV'); [].
intros v' <-.
so (Inversion.dyn_ofm_inversion_dyn _#4 wf_nil HofV) as Hofv.
so (Inversion.dyn_ofm_inversion_dyn _#4 wf_nil HofV') as Hofv'.
clear HofV HofV'.
invert HvalV; clear HvalV; [].
intros Hvalv.
invert HvalV'; clear HvalV'; [].
intros Hvalv'.
apply dyn_rel_; auto; [].
apply Hcl; [].
do2 4 split; auto; [].
intros f f' Hf.
so Hf as (_ & Hoff & Hoff' & Hvalf & Hvalf').
rewrite -> (app_thalt_into_mod t f); auto; [].
rewrite -> (app_thalt_into_mod t' f'); auto; [].
destruct HV as (Hincl & _).
apply Hincl; [].
clear v v' Hofv Hofv' Hvalv Hvalv' Hincl.
#3 do2 4 split; auto with dynamic.
  intros V V' HV.
  invert HV; clear HV; [].
  intros w w' Hofw Hofw' Hvalw Hvalw' Hw <- <-.
  rewrite <- (app_thalt_into_mod t f); auto; [].
  rewrite <- (app_thalt_into_mod t' f'); auto; [].
  exact (Hf andel _ _ Hw).

  apply ofm_forget; [].
  eapply ofm_lamgn; eauto with static; [].
  #2 apply (@ofm_lett _ _ cn_unit).
    #2 eapply oft_app.
      #2 eapply rel_quest.
        eapply weaken2_oft; eauto; done.

        simp_sub.
        reflexivity.
        
      apply oft_ext; [].
      apply ofm_forget; [].
      #2 eapply rel_quest.
        eapply ofm_var; eauto with sequence; done.

        simp_sub.
        reflexivity.
        
    simp_sub.
    apply ofm_forget; [].
    apply ofm_star; done.

  apply ofm_forget; [].
  eapply ofm_lamgn; eauto with static; [].
  #2 apply (@ofm_lett _ _ cn_unit).
    #2 eapply oft_app.
      #2 eapply rel_quest.
        eapply weaken2_oft; eauto; done.

        simp_sub.
        reflexivity.
        
      apply oft_ext; [].
      apply ofm_forget; [].
      #2 eapply rel_quest.
        eapply ofm_var; eauto with sequence; done.

        simp_sub.
        reflexivity.
        
    simp_sub.
    apply ofm_forget; [].
    apply ofm_star; done.
Qed.


Inductive pigen_rel
  (T1 T1' T2 T2' : (* sg *) term) (R1 R2 : term -> term -> Prop) : term -> term -> Prop :=
| pigen_rel_ {V V'}
    : ofm nil impure V (sg_pigen T1 (shift1 T2))
      -> ofm nil impure V' (sg_pigen T1' (shift1 T2'))
      -> mvalue nil V
      -> mvalue nil V'
      -> (forall W W',
            R1 W W'
            -> mev T2 T2' R2 (md_appgn V W) (md_appgn V' W'))
      -> pigen_rel T1 T1' T2 T2' R1 R2 V V'
.


Lemma biorthogonality_lemma_md :
  forall t t' u u' R S f f' w w' g g' e e',
    mvrel u u' S
    -> mps t t' R f f'
    -> mpst u u' S w w'
    -> pigen_rel u u' t t' S R g g'
    -> star (mstep nil) (md_appgn g w) e
    -> star (mstep nil) (md_appgn g' w') e'
    -> (mhalt nil (md_appgnlt f e sg_one) <-> mhalt nil (md_appgnlt f' e' sg_one)).
Proof.
intros t t' u u' R S f f' w w' g g' e e' HofS Hf Hw Hg Hstepse Hstepse'.
so Hf as (_ & Hoff & Hoff' & Hvalf & Hvalf').
so Hw as (_ & Hofw & Hofw' & Hvalw & Hvalw').
so (can_fstm_value _#4 wf_nil Hofw Hvalw) as (c & Hfstw).
so (can_fstm_value _#4 wf_nil Hofw' Hvalw') as (c' & Hfstw').
so (Regularity.ofm_reg _#4 wf_nil Hofw) as Hofu.
so (Regularity.ofm_reg _#4 wf_nil Hofw') as Hofu'.
so (can_fsts _ _ Hofu) as (k & Hfstu).
so (can_fsts _ _ Hofu') as (k' & Hfstu').
invert Hg; [].
intros Hofg Hofg' _ _ _.
#2 rewrite <- (msteps_mhalt_iff nil (md_appgn (md_lamgn u (md_appgnlt (subst (sh 2) f) (md_appgn (subst (sh 2) g) (var 0)) sg_one)) w) (md_appgnlt f e sg_one)).
  #2 rewrite <- (msteps_mhalt_iff nil (md_appgn (md_lamgn u' (md_appgnlt (subst (sh 2) f') (md_appgn (subst (sh 2) g') (var 0)) sg_one)) w') (md_appgnlt f' e' sg_one)).
    refine (Hw andel _ _ _); [].
    #3 do2 4 split; auto with dynamic.
      intros x x' Hx.
      so (HofS _ _ Hx) as (Hofx & Hofx' & Hvalx & Hvalx').
      so (can_fstm_value _#4 wf_nil Hofx Hvalx) as (d & Hfstx).
      so (can_fstm_value _#4 wf_nil Hofx' Hvalx') as (d' & Hfstx').
      #2 rewrite -> (msteps_mhalt_iff nil (md_appgn (md_lamgn u (md_appgnlt (subst (sh 2) f) (md_appgn (subst (sh 2) g) (var 0)) sg_one)) x) (md_appgnlt f (md_appgn g x) sg_one)).
        #2 rewrite -> (msteps_mhalt_iff nil (md_appgn (md_lamgn u' (md_appgnlt (subst (sh 2) f') (md_appgn (subst (sh 2) g') (var 0)) sg_one)) x') (md_appgnlt f' (md_appgn g' x') sg_one)).
          eapply mps_elim_mev; eauto; [].
          invert Hg; [].
          intros _ _ _ _ Hrel.
          apply Hrel; auto; done.

          #2 eapply star_step.
            eapply mstep_appgn3; eauto; done.

            rewrite -> subst_appgnlt; [].
            simp_sub.
            apply star_refl; done.

        #2 eapply star_step.
          eapply mstep_appgn3; eauto; done.

          rewrite -> subst_appgnlt; [].
          simp_sub.
          apply star_refl; done.

      apply ofm_forget; [].
      eapply ofm_lamgn; eauto; [].
      sor (refine (weaken2_ofm _#6 Hoff)) as H1.
      simp_sub in H1.
      replsub (shift1 sg_one) with (subst1 (var 1) (subst (sh 2) sg_one)).
      eapply ofm_appgnlt; eauto with static sequence; [].
      sor (refine (weaken2_ofm _#6 Hofg)) as H2.
      simp_sub in H2.
      #2 eapply rel_quest.
        eapply ofm_appgn; eauto with static sequence; [].
        #2 eapply rel_quest.
          eapply ofm_var; eauto with sequence; done.

          simp_sub.
          reflexivity.

        simp_sub.
        reflexivity.

      apply ofm_forget; [].
      eapply ofm_lamgn; eauto; [].
      sor (refine (weaken2_ofm _#6 Hoff')) as H1.
      simp_sub in H1.
      replsub (shift1 sg_one) with (subst1 (var 1) (subst (sh 2) sg_one)).
      eapply ofm_appgnlt; eauto with static sequence; [].
      sor (refine (weaken2_ofm _#6 Hofg')) as H2.
      simp_sub in H2.
      #2 eapply rel_quest.
        eapply ofm_appgn; eauto with static sequence; [].
        #2 eapply rel_quest.
          eapply ofm_var; eauto with sequence; done.

          simp_sub.
          reflexivity.

        simp_sub.
        reflexivity.

    #2 eapply star_step.
      eapply mstep_appgn3; eauto; done.

      rewrite -> subst_appgnlt; [].
      simp_sub.
      apply (star_map' _ _ (fun z => md_appgnlt _ z _)); auto using mstep_appgnlt1; done.

  #2 eapply star_step.
    eapply mstep_appgn3; eauto; done.

    rewrite -> subst_appgnlt; [].
    simp_sub.
    apply (star_map' _ _ (fun z => md_appgnlt _ z _)); auto using mstep_appgnlt1; done.
Qed.
