
Require Import Syntax.
Require Import Sequence.
Require Import Subst.
Require Import Static.
Require Import Dynamic.


Inductive treduce : context -> (* tm *) term -> (* tm *) term -> Prop :=
| treduce_refl {G e}
    : treduce G e e

| treduce_trans {G e1 e2 e3}
    : treduce G e1 e2
      -> treduce G e2 e3
      -> treduce G e1 e3

| treduce_lam {G t t' e e'}
    : treduce (G; cl_utm) e e'
      -> treduce G (tm_lam t e) (tm_lam t' e')

| treduce_app1 {G e1 e1' e2 e2'}
    : treduce G e1 e1'
      -> treduce G e2 e2'
      -> treduce G (tm_app e1 e2) (tm_app e1' e2')

| treduce_app2 {G t e1 v2}
    : tvalue G v2
      -> treduce G (tm_app (tm_lam t e1) v2) (subst1 v2 e1)

| treduce_pair {G e1 e1' e2 e2'}
    : treduce G e1 e1'
      -> treduce G e2 e2'
      -> treduce G (tm_pair e1 e2) (tm_pair e1' e2')

| treduce_pi11 {G e e'}
    : treduce G e e'
      -> treduce G (tm_pi1 e) (tm_pi1 e')

| treduce_pi12 {G v1 v2}
    : tvalue G v1
      -> tvalue G v2
      -> treduce G (tm_pi1 (tm_pair v1 v2)) v1

| treduce_pi21 {G e e'}
    : treduce G e e'
      -> treduce G (tm_pi2 e) (tm_pi2 e')

| treduce_pi22 {G v1 v2}
    : tvalue G v1
      -> tvalue G v2
      -> treduce G (tm_pi2 (tm_pair v1 v2)) v2

| treduce_plam {G k k' e e'}
    : treduce (G; cl_ucn) e e'
      -> treduce G (tm_plam k e) (tm_plam k' e')

| treduce_papp1 {G e e' c c'}
    : treduce G e e'
      -> treduce G (tm_papp e c) (tm_papp e' c')

| treduce_papp2 {G k e c}
    : treduce G (tm_papp (tm_plam k e) c) (subst1 c e)

| treduce_pack {G c c' e e' k k' t t'}
    : treduce G e e'
      -> treduce G (tm_pack c e k t) (tm_pack c' e' k' t')

| treduce_unpack1 {G e1 e1' e2 e2'}
    : treduce G e1 e1'
      -> treduce (G; cl_ucn; cl_utm) e2 e2'
      -> treduce G (tm_unpack e1 e2) (tm_unpack e1' e2')

| treduce_unpack2 {G c v k t e}
    : tvalue G v
      -> treduce G (tm_unpack (tm_pack c v k t) e) (subst (dot v (dot c id)) e)

| treduce_fix1 {G t t' e e'}
    : treduce G e e'
      -> treduce G (tm_fix t e) (tm_fix t' e')

| treduce_fix2 {G t e}
    : treduce G (tm_fix t e) (tm_app e (tm_lam cn_unit (shift1 (tm_fix t e))))

| treduce_lett1 {G e1 e1' e2 e2'}
    : treduce G e1 e1'
      -> treduce (G; cl_utm) e2 e2'
      -> treduce G (tm_lett e1 e2) (tm_lett e1' e2')

| treduce_lett2 {G v1 e2}
    : tvalue G v1
      -> treduce G (tm_lett v1 e2) (subst1 v1 e2)

| treduce_letm1 {G M M' e e'}
    : mreduce G M M'
      -> treduce (G; cl_ucn; cl_umd (var 0)) e e'
      -> treduce G (tm_letm M e) (tm_letm M' e')

| treduce_letm2 {G V e c}
    : mvalue G V
      -> fstm G V c
      -> treduce G (tm_letm V e) (subst (dot V (dot c id)) e)

| treduce_ext1 {G M M'}
    : mreduce G M M'
      -> treduce G (tm_ext M) (tm_ext M')

| treduce_ext2 {G v}
    : tvalue G v
      -> treduce G (tm_ext (md_dyn v)) v


with mreduce : context -> (* md *) term -> (* md *) term -> Prop :=
| mreduce_refl {G M}
    : mreduce G M M

| mreduce_trans {G M1 M2 M3}
    : mreduce G M1 M2
      -> mreduce G M2 M3
      -> mreduce G M1 M3

| mreduce_stat {G c c'}
    : mreduce G (md_stat c) (md_stat c')

| mreduce_dyn {G e e'}
    : treduce G e e'
      -> mreduce G (md_dyn e) (md_dyn e')

| mreduce_lamgn {G T T' M M'}
    : mreduce (G; cl_ucn; cl_umd (var 0)) M M'
      -> mreduce G (md_lamgn T M) (md_lamgn T' M')

| mreduce_appgn1 {G M1 M1' M2 M2'}
    : mreduce G M1 M1'
      -> mreduce G M2 M2'
      -> mreduce G (md_appgn M1 M2) (md_appgn M1' M2')

| mreduce_appgn2 {G T M1 V2 c2}
    : mvalue G V2
      -> fstm G V2 c2
      -> mreduce G (md_appgn (md_lamgn T M1) V2) (subst (dot V2 (dot c2 id)) M1)

| mreduce_lamap {G T T' M M'}
    : mreduce (G; cl_ucn; cl_umd (var 0)) M M'
      -> mreduce G (md_lamap T M) (md_lamap T' M')

| mreduce_appap1 {G M1 M1' M2 M2'}
    : mreduce G M1 M1'
      -> mreduce G M2 M2'
      -> mreduce G (md_appap M1 M2) (md_appap M1' M2')

| mreduce_appap2 {G T M1 V2 c2}
    : mvalue G V2
      -> fstm G V2 c2
      -> mreduce G (md_appap (md_lamap T M1) V2) (subst (dot V2 (dot c2 id)) M1)

| mreduce_pair {G M1 M1' M2 M2'}
    : mreduce G M1 M1'
      -> mreduce G M2 M2'
      -> mreduce G (md_pair M1 M2) (md_pair M1' M2')

| mreduce_pi11 {G M M'}
    : mreduce G M M'
      -> mreduce G (md_pi1 M) (md_pi1 M')

| mreduce_pi12 {G V1 V2}
    : mvalue G V1
      -> mvalue G V2
      -> mreduce G (md_pi1 (md_pair V1 V2)) V1

| mreduce_pi21 {G M M'}
    : mreduce G M M'
      -> mreduce G (md_pi2 M) (md_pi2 M')

| mreduce_pi22 {G V1 V2}
    : mvalue G V1
      -> mvalue G V2
      -> mreduce G (md_pi2 (md_pair V1 V2)) V2

| mreduce_unpack1 {G e e' M M' T T'}
    : treduce G e e'
      -> mreduce (G; cl_ucn; cl_utm) M M'
      -> mreduce G (md_unpack e M T) (md_unpack e' M' T')

| mreduce_unpack2 {G c v k t M T}
    : tvalue G v
      -> mreduce G (md_unpack (tm_pack c v k t) M T) (subst (dot v (dot c id)) M)

| mreduce_lett1 {G e1 e1' M2 M2'}
    : treduce G e1 e1'
      -> mreduce (G; cl_utm) M2 M2'
      -> mreduce G (md_lett e1 M2) (md_lett e1' M2')

| mreduce_lett2 {G v1 M2}
    : tvalue G v1
      -> mreduce G (md_lett v1 M2) (subst1 v1 M2)

| mreduce_letm1 {G M1 M1' M2 M2' T T'}
    : mreduce G M1 M1'
      -> mreduce (G; cl_ucn; cl_umd (var 0)) M2 M2'
      -> mreduce G (md_letm M1 M2 T) (md_letm M1' M2' T')

| mreduce_letm2 {G V M T c}
    : mvalue G V
      -> fstm G V c
      -> mreduce G (md_letm V M T) (subst (dot V (dot c id)) M)

| mreduce_seal1 {G M M' T T'}
    : mreduce G M M'
      -> mreduce G (md_seal M T) (md_seal M' T')

| mreduce_seal2 {G M T}
    : mreduce G (md_seal M T) M
.


Scheme treduce_mut_ind := Minimality for treduce Sort Prop
with   mreduce_mut_ind := Minimality for mreduce Sort Prop.
Combined Scheme breduce_ind from treduce_mut_ind, mreduce_mut_ind.

Hint Constructors treduce mreduce : reduce.
