
Require Import Tactics.
Require Import Extensionality.
Require Import Sequence.
Require Import Relation.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Regularity.
Require Import Preservation. 
Require Import Determinism.
Require Import Inversion.
Require Import Congruence.
Require Import Contextual.
Require Import Compactness.
Require Declarative.
Require Progress.


Definition trel (t t' : (* cn *) term) (R : (* tm *) term -> (* tm *) term -> Prop) :=
  forall e e',
    R e e'
    -> oft nil e t /\ oft nil e' t'.


Definition mrel (T T' : (* sg *) term) (R : (* md *) term -> (* md *) term -> Prop) :=
  forall M M',
    R M M'
    -> ofm nil impure M T /\ ofm nil impure M' T'.


Definition tvrel (t t' : (* cn *) term) (R : (* tm *) term -> (* tm *) term -> Prop) :=
  forall e e',
    R e e'
    -> oft nil e t /\ oft nil e' t' /\ tvalue nil e /\ tvalue nil e'.


Definition mvrel (T T' : (* sg *) term) (R : (* md *) term -> (* md *) term -> Prop) :=
  forall M M',
    R M M'
    -> ofm nil impure M T /\ ofm nil impure M' T' /\ mvalue nil M /\ mvalue nil M'.


(* Pitts closure *)

Definition tps (t t' : (* cn *) term) (R : (* tm *) term -> (* tm *) term -> Prop) (f f' : (* tm *) term) :=
  (forall v v', R v v' -> (thalt nil (tm_app f v) <-> thalt nil (tm_app f' v')))
  /\ oft nil f (cn_arrow t cn_unit)
  /\ oft nil f' (cn_arrow t' cn_unit)
  /\ tvalue nil f
  /\ tvalue nil f'.


Definition mps (T T' : (* sg *) term) (R : (* md *) term -> (* md *) term -> Prop) (F F' : (* md *) term) :=
  (forall V V', R V V' -> (mhalt nil (md_appgn F V) <-> mhalt nil (md_appgn F' V')))
  /\ ofm nil impure F (sg_pigen T sg_one)
  /\ ofm nil impure F' (sg_pigen T' sg_one)
  /\ mvalue nil F
  /\ mvalue nil F'.


Definition tpt (t t' : (* cn *) term) (S : (* tm *) term -> (* tm *) term -> Prop) (v v' : (* tm *) term) :=
  (forall f f', S f f' -> (thalt nil (tm_app f v) <-> thalt nil (tm_app f' v')))
  /\ oft nil v t
  /\ oft nil v' t'
  /\ tvalue nil v
  /\ tvalue nil v'.


Definition mpt (T T' : (* sg *) term) (S : (* md *) term -> (* md *) term -> Prop) (V V' : (* tm *) term) :=
  (forall F F', S F F' -> (mhalt nil (md_appgn F V) <-> mhalt nil (md_appgn F' V')))
  /\ ofm nil impure V T
  /\ ofm nil impure V' T'
  /\ mvalue nil V
  /\ mvalue nil V'.


Definition tpst (t t' : (* cn *) term) (R : (* tm *) term -> (* tm *) term -> Prop) := 
  tpt t t' (tps t t' R).


Definition mpst (T T' : (* sg *) term) (R : (* md *) term -> (* md *) term -> Prop) := 
  mpt T T' (mps T T' R).


Definition tpclosed (t t' : (* cn *) term) (R : (* tm *) term -> (* tm *) term -> Prop) :=
  forall v v',
    tpst t t' R v v'
    -> R v v'.


Definition mpclosed (T T' : (* sg *) term) (R : (* md *) term -> (* md *) term -> Prop) :=
  forall V V',
    mpst T T' R V V'
    -> R V V'.


Lemma tps_tvrel :
  forall t t' R,
    tvrel (cn_arrow t cn_unit) (cn_arrow t' cn_unit) (tps t t' R).
Proof.
intros t t' R.
intros f f' HRs.
destruct HRs.
assumption.
Qed.


Lemma mps_mvrel :
  forall T T' R,
    mvrel (sg_pigen T sg_one) (sg_pigen T' sg_one) (mps T T' R).
Proof.
intros t t' R.
intros f f' HRs.
destruct HRs.
assumption.
Qed.


Lemma tpt_tvrel :
  forall t t' R,
    tvrel t t' (tpt t t' R).
Proof.
intros t t' R.
intros v v' HRt.
destruct HRt.
assumption.
Qed.


Lemma mpt_mvrel :
  forall T T' R,
    mvrel T T' (mpt T T' R).
Proof.
intros t t' R.
intros v v' HRt.
destruct HRt.
assumption.
Qed.


Lemma tpst_tvrel :
  forall t t' R,
    tvrel t t' (tpst t t' R).
Proof.
intros t t' R.
unfold tpst.
apply tpt_tvrel; auto.
Qed.


Lemma mpst_mvrel :
  forall T T' R,
    mvrel T T' (mpst T T' R).
Proof.
intros t t' R.
unfold mpst.
apply mpt_mvrel; auto.
Qed.


Lemma tpst_expand :
  forall t t' R,
    tvrel t t' R
    -> forall v v',
         R v v'
         -> tpst t t' R v v'.
Proof.
intros t t' R HrelR v v' HR.
so (HrelR _ _ HR) as (Hof & Hof' & Hval & Hval').
do2 4 split; auto; [].
intros f f' HS.
exact (HS andel v v' HR).
Qed.


Lemma mpst_expand :
  forall T T' R,
    mvrel T T' R
    -> forall V V',
         R V V'
         -> mpst T T' R V V'.
Proof.
intros t t' R HrelR v v' HR.
so (HrelR _ _ HR) as (Hof & Hof' & Hval & Hval').
do2 4 split; auto; [].
intros f f' HS.
exact (HS andel v v' HR).
Qed.


Lemma tpclosed_eq_tpst :
  forall t t' R,
    tvrel t t' R
    -> tpclosed t t' R
    -> tpst t t' R = R.
Proof.
intros t t' R HofR Hcl.
fextensionality 2.
intros v v'.
pextensionality; auto; [].
intro Hv.
apply tpst_expand; auto; done.
Qed.


Lemma mpclosed_eq_mpst :
  forall t t' R,
    mvrel t t' R
    -> mpclosed t t' R
    -> mpst t t' R = R.
Proof.
intros t t' R HofR Hcl.
fextensionality 2.
intros v v'.
pextensionality; auto; [].
intro Hv.
apply mpst_expand; auto; done.
Qed.


Lemma tpst_mono :
  forall t t' (R1 R2 : term -> term -> Prop),
    (forall v v', R1 v v' -> R2 v v')
    -> forall v v', tpst t t' R1 v v' -> tpst t t' R2 v v'.
Proof.
intros t t' R1 R2 Hincl v v' HR1st.
destruct HR1st as (Hall1st & Hwfs).
split; auto; [].
intros f f' HR2s.
destruct HR2s as (Hall2s & Hwfsf).
split.
  intro Hhalt.
  refine (Hall1st _ _ _ andel Hhalt); [].
  split; auto; done.

  intro Hhalt.
  refine (Hall1st _ _ _ ander Hhalt); [].
  split; auto; done.
Qed.


Lemma mpst_mono :
  forall T T' (R1 R2 : term -> term -> Prop),
    (forall V V', R1 V V' -> R2 V V')
    -> forall V V', mpst T T' R1 V V' -> mpst T T' R2 V V'.
Proof.
intros t t' R1 R2 Hincl v v' HR1st.
destruct HR1st as (Hall1st & Hwfs).
split; auto; [].
intros f f' HR2s.
destruct HR2s as (Hall2s & Hwfsf).
split.
  intro Hhalt.
  refine (Hall1st _ _ _ andel Hhalt); [].
  split; auto; done.

  intro Hhalt.
  refine (Hall1st _ _ _ ander Hhalt); [].
  split; auto; done.
Qed.


Lemma mpst_mono_sg :
  forall T T' U U' R,
    subs nil T U
    -> subs nil T' U'
    -> forall V V', mpst T T' R V V' -> mpst U U' R V V'.
Proof.
intros T T' U U' R Hsubs Hsubs' V V' HV.
unfold mpst in HV.
unfold mpst.
destruct HV as (Hrel & Hof & Hof' & Hval & Hval').
do2 3 split; eauto with static; [].
intros F F' HF.
apply Hrel; [].
clear Hrel Hof Hof' Hval Hval'.
destruct HF as (Hrel & Hof & Hof' & Hval & Hval').
so (subs_fst_o _#3 wf_nil Hsubs) as (k & l & HfstU & HfstT & _).
so (subs_fst_o _#3 wf_nil Hsubs') as (k' & l' & HfstU' & HfstT' & _).
do2 3 split; eauto with static; done.
Qed.


Lemma tpst_idem :
  forall t t' R,
    tpclosed t t' (tpst t t' R).
Proof.
intros t t' R.
intros v v' HRstst.
destruct HRstst as (Hall & Hrel).
split; auto; [].
intros f f' HRs.
apply Hall; [].
clear v v' Hall Hrel.
so (tps_tvrel _#5 HRs) as Hrel.
split; auto; [].
intros w w' HRst.
exact (HRst andel _ _ HRs).
Qed.


Lemma mpst_idem :
  forall T T' R,
    mpclosed T T' (mpst T T' R).
Proof.
intros t t' R.
intros v v' HRstst.
destruct HRstst as (Hall & Hrel).
split; auto; [].
intros f f' HRs.
apply Hall; [].
clear v v' Hall Hrel.
so (mps_mvrel _#5 HRs) as Hrel.
split; auto; [].
intros w w' HRst.
exact (HRst andel _ _ HRs).
Qed.


Lemma tpst_ceqt :
  forall t1 t2 R v1 v2 v1' v2',
    tvrel t1 t2 R
    -> tpclosed t1 t2 R
    -> ceqt nil v1 v1' t1
    -> ceqt nil v2 v2' t2
    -> tvalue nil v1'
    -> tvalue nil v2'
    -> R v1 v2
    -> R v1' v2'.
Proof.
intros t1 t2 R v1 v2 v1' v2' _ Hclosed Hceq1 Hceq2 Hval1 Hval2 HR.
so (ceqt_reg _#4 wf_nil Hceq1 ander) as Hof1.
so (ceqt_reg _#4 wf_nil Hceq2 ander) as Hof2.
apply Hclosed; [].
split; auto; [].
intros f1 f2 HRs.
so (HRs andel _ _ HR) as Hiff.
so (HRs ander) as (Hoff1 & Hoff2 & _ & _).
split.
  intro Hhalt.
  #2 apply (ceqt_halt (tm_app f2 v2) _ cn_unit).
    apply ceqt_comp; eapply tcc_app; eauto using ceqt_refl; done.

    apply Hiff; [].
    apply (ceqt_halt (tm_app f1 v1') _ cn_unit); auto; [].
    apply ceqt_comp; eapply tcc_app; eauto using ceqt_refl, ceqt_symm; done.

  intro Hhalt.
  #2 apply (ceqt_halt (tm_app f1 v1) _ cn_unit).
    apply ceqt_comp; eapply tcc_app; eauto using ceqt_refl; done.

    apply Hiff; [].
    apply (ceqt_halt (tm_app f2 v2') _ cn_unit); auto; [].
    apply ceqt_comp; eapply tcc_app; eauto using ceqt_refl, ceqt_symm; done.
Qed.


Lemma appgn_halt :
  forall T1 T2 F F' V V',
    ceqm nil impure F F' (sg_pigen T1 (shift1 T2))
    -> ceqm nil impure V V' T1
    -> mvalue nil V
    -> mvalue nil V'
    -> mhalt nil (md_appgn F V)
    -> mhalt nil (md_appgn F' V').
Proof.
intros T1 T2 F F' V V' Hceq1 Hceq2 Hval Hval' Hhalt.
so (can_fstm _#3 wf_nil (mvalue_pure _#4 Hval (ceqm_reg _#5 wf_nil Hceq2 andel))) as (c & Hfstm).
so (can_fstm _#3 wf_nil (mvalue_pure _#4 Hval' (ceqm_reg _#5 wf_nil Hceq2 ander))) as (c' & Hfstm').
so (ofm_reg _#4 wf_nil (ceqm_reg _#5 wf_nil Hceq1 andel)) as Hofpi.
invert Hofpi; clear Hofpi; intros k1 HofT1 Hfsts HofT2w.
so (subst1_ofs _#4 (Declarative.inh_ofc _ _ (ofs_fst _#3 HofT1 Hfsts)) HofT2w) as HofT2.
simp_sub in HofT2.
#2 apply (msteps_mhalt _ (md_letm V' (md_appgn (subst (sh 2) F') (var 0)) T2)).
  #2 eapply star_step.
    eapply mstep_letm2; eauto; done.

    simp_sub.
    apply star_refl; done.

  #2 apply (ceqm_halt (md_letm V (md_appgn (subst (sh 2) F) (var 0)) T2) _ T2).
    apply ceqm_comp; eapply mcc_letm; eauto using eqs_refl; [].
    replsub (subst (sh 2) T2) with (subst1 (var 1) (subst (sh 3) T2)).
    #3 apply ceqm_comp; apply (mcc_appgn _#6 (subst (sh 2) T1)).
      replsub (sg_pigen (subst (sh 2) T1) (subst (sh 3) T2)) with (subst (sh 2) (sg_pigen T1 (shift1 T2))).
      exact (subst_ceqm _#7 (weaken2 _#4) wf_nil Hceq1).

      apply ceqm_refl; [].
      replsub (subst (sh 2) T1) with (shift1 (shift1 T1)).
      eapply ofm_var; eauto with sequence; done.

      replsub (var 1) with (shift1 (var 0)).
      eapply fstm_var; eauto with sequence; done.

    eapply anti_msteps_mhalt; eauto; [].
    #2 eapply star_step.
      eapply mstep_letm2; eauto; done.

      simp_sub.
      apply star_refl; done.
Qed.


Lemma mpclosed_ceqm :
  forall T1 T2 R V1 V2 V1' V2',
    mvrel T1 T2 R
    -> mpclosed T1 T2 R
    -> ceqm nil impure V1 V1' T1
    -> ceqm nil impure V2 V2' T2
    -> mvalue nil V1'
    -> mvalue nil V2'
    -> R V1 V2
    -> R V1' V2'.
Proof.
intros T1 T2 R V1 V2 V1' V2' HofR Hclosed Hceq1 Hceq2 Hval1' Hval2' HR.
so (HofR _ _ HR) as (_ & _ & Hval1 & Hval2).
so (ceqm_reg _#5 wf_nil Hceq1 ander) as Hof1'.
so (ceqm_reg _#5 wf_nil Hceq2 ander) as Hof2'.
apply Hclosed; [].
split; auto; [].
intros F1 F2 HRs.
so (HRs andel _ _ HR) as Hiff.
so (HRs ander) as (Hoff1 & Hoff2 & _ & _).
split.
  intro Hhalt.
  apply (appgn_halt T2 sg_one F2 _ V2); simp_sub; auto using ceqm_refl; [].
  apply Hiff; [].
  apply (appgn_halt T1 sg_one F1 _ V1'); simp_sub; auto using ceqm_refl; [].
  apply ceqm_symm; assumption.

  intro Hhalt.
  apply (appgn_halt T1 sg_one F1 _ V1); simp_sub; auto using ceqm_refl; [].
  apply Hiff; [].
  apply (appgn_halt T2 sg_one F2 _ V2'); simp_sub; auto using ceqm_refl; [].
  apply ceqm_symm; assumption.
Qed.


Lemma tpclosed_admissible :
  forall t1 t2 u1 u2 R f1 f2 v1 v2,
    tpclosed t1 t2 R
    -> oft nil f1 (cn_arrow (cn_arrow cn_unit u1) u1)
    -> oft nil f2 (cn_arrow (cn_arrow cn_unit u2) u2)
    -> oft (nil; cl_tm (cn_arrow cn_unit u1)) v1 (shift1 t1)
    -> oft (nil; cl_tm (cn_arrow cn_unit u2)) v2 (shift1 t2)
    -> tvalue (nil; cl_tm (cn_arrow cn_unit u1)) v1
    -> tvalue (nil; cl_tm (cn_arrow cn_unit u2)) v2
    -> (forall j, R (subst1 (fixnv j u1 f1) v1) (subst1 (fixnv j u2 f2) v2))
    -> R (subst1 (fixv u1 f1) v1) (subst1 (fixv u2 f2) v2).
Proof.
intros t1 t2 u1 u2 R f1 f2 v1 v2 Hclosed Hoff1 Hoff2 Hofv1 Hofv2 Hvalv1 Hvalv2 HallR.
so (ofc_ofcr _#3 (oft_reg _#3 wf_nil Hoff1)) as (kk & H).
invert H; intros _ Hofu1 _; clear kk H.
so (ofc_ofcr _#3 (oft_reg _#3 wf_nil Hoff2)) as (kk & H).
invert H; intros _ Hofu2 _; clear kk H.
apply Hclosed; [].
split.
  intros g1 g2 HRs.
  so (tps_tvrel _#5 HRs) as (Hofg1 & Hofg2 & _).
  split.
    intro Hhalt.
    #3 exploit (compactness u1 f1 (tm_app (shift1 g1) v1) cn_unit) as H; auto.
      eapply oft_app; eauto; [].
      replsub (cn_arrow (shift1 t1) cn_unit) with (shift1 (cn_arrow t1 cn_unit)).
      eapply subst_oft; eauto using weaken1; done.

      simp_sub.
      assumption.

      destruct H as (j & H).
      so (H j (Le.le_refl _)) as Hhalt'; clear H.
      simp_sub in Hhalt'.
      so (HRs andel _ _ (HallR j) andel Hhalt') as Hhalt''.
      refine (capproxt_halt _ _ cn_unit _ Hhalt''); [].
      apply capproxt_comp; eapply tcc_app; eauto using capproxt_refl; [].
      replsub t2 with (subst1 (fixnv j u2 f2) (shift1 t2)).
      #3 eapply funct_capprox_tm; eauto.
        apply wf_nil; done.

        eauto with static; done.

        apply subrel_dot; auto using subrel_id; [].
        simp_sub; apply deq_tm; [].
        apply capproxt_comp; apply tcc_lam; auto with static; [].
        refine (weaken1_capproxt _#6); [].
        apply fixn_approx; auto; done.

    intro Hhalt.
    #3 exploit (compactness u2 f2 (tm_app (shift1 g2) v2) cn_unit) as H; auto.
      eapply oft_app; eauto; [].
      replsub (cn_arrow (shift1 t2) cn_unit) with (shift1 (cn_arrow t2 cn_unit)).
      eapply subst_oft; eauto using weaken1; done.

      simp_sub.
      assumption.

      destruct H as (j & H).
      so (H j (Le.le_refl _)) as Hhalt'; clear H.
      simp_sub in Hhalt'.
      so (HRs andel _ _ (HallR j) ander Hhalt') as Hhalt''.
      refine (capproxt_halt _ _ cn_unit _ Hhalt''); [].
      apply capproxt_comp; eapply tcc_app; eauto using capproxt_refl; [].
      replsub t1 with (subst1 (fixnv j u1 f1) (shift1 t1)).
      #3 eapply funct_capprox_tm; eauto.
        exact wf_nil.

        eauto with static; done.

        apply subrel_dot; auto using subrel_id; [].
        simp_sub; apply deq_tm; [].
        apply capproxt_comp; apply tcc_lam; auto with static; [].
        refine (weaken1_capproxt _#6); [].
        apply fixn_approx; auto; done.

  do2 3 split.
    so (subst1_oft_tm _#5 (oft_fixv _#3 Hofu1 Hoff1) Hofv1) as H.
    simp_sub in H.
    assumption.

    so (subst1_oft_tm _#5 (oft_fixv _#3 Hofu2 Hoff2) Hofv2) as H.
    simp_sub in H.
    assumption.

    eapply subst_tvalue; eauto; [].
    apply subof_dot; auto using subof_id; [].
    simp_sub; apply oof_tm; [].
    exact tvalue_lam.

    eapply subst_tvalue; eauto; [].
    apply subof_dot; auto using subof_id; [].
    simp_sub; apply oof_tm; [].
    exact tvalue_lam.
Qed.


(* Evaluation closure *)

Definition tev (t t' : (* cn *) term) (R : (* tm *) term -> (* tm *) term -> Prop) (e e' : (* tm *) term) :=
  (thalt nil e <-> thalt nil e')
  /\ (forall v v',
        star (tstep nil) e v
        -> star (tstep nil) e' v'
        -> tvalue nil v
        -> tvalue nil v'
        -> R v v')
  /\ oft nil e t
  /\ oft nil e' t'.


Definition mev (T T' : (* sg *) term) (R : (* md *) term -> (* md *) term -> Prop) (M M' : (* md *) term) :=
  (mhalt nil M <-> mhalt nil M')
  /\ (forall V V',
        star (mstep nil) M V
        -> star (mstep nil) M' V'
        -> mvalue nil V
        -> mvalue nil V'
        -> R V V')
  /\ ofm nil impure M T
  /\ ofm nil impure M' T'.


Definition mevp (T T' : (* sg *) term) (R : (* md *) term -> (* md *) term -> Prop) (M M' : (* md *) term) :=
  (mhalt nil M <-> mhalt nil M')
  /\ (forall V V',
        star (mstep nil) M V
        -> star (mstep nil) M' V'
        -> mvalue nil V
        -> mvalue nil V'
        -> R V V')
  /\ ofm nil pure M T
  /\ ofm nil pure M' T'.


Lemma mevp_forget :
  forall T T' R M M',
    mevp T T' R M M'
    -> mev T T' R M M'.
Proof.
intros T T' R M M' HM.
destruct HM as (Hhaltiff & Hrel & Hof & Hof').
do2 3 split; eauto using ofm_forget_impure.
Qed.


Lemma tev_trel :
  forall t t' R,
    trel t t' (tev t t' R).
Proof.
intros t t' R.
intros e e' HRt.
destruct HRt as (_ & _ & ?).
assumption.
Qed.


Lemma mev_mrel :
  forall T T' R,
    mrel T T' (mev T T' R).
Proof.
intros t t' R.
intros e e' HRt.
destruct HRt as (_ & _ & ? & ?).
auto.
Qed.


Lemma mevp_mrel :
  forall T T' R,
    mrel T T' (mevp T T' R).
Proof.
intros T T' R.
intros M M' HM.
exact (mev_mrel _#5 (mevp_forget _#5 HM)).
Qed.


Lemma tev_mono :
  forall t t' (R1 R2 : term -> term -> Prop),
    (forall e e', R1 e e' -> R2 e e')
    -> forall e e', tev t t' R1 e e' -> tev t t' R2 e e'.
Proof.
intros t t' R1 R2 Hincl e e' HR1ev.
destruct HR1ev as (Hhaltiff & Hall1 & Hofs).
do2 2 split; auto.
Qed.


Lemma mev_mono :
  forall T T' (R1 R2 : term -> term -> Prop),
    (forall M M', R1 M M' -> R2 M M')
    -> forall M M', mev T T' R1 M M' -> mev T T' R2 M M'.
Proof.
intros t t' R1 R2 Hincl e e' HR1ev.
destruct HR1ev as (Hhaltiff & Hall1 & Hofs).
do2 2 split; auto.
Qed.


Lemma mevp_mono :
  forall T T' (R1 R2 : term -> term -> Prop),
    (forall M M', R1 M M' -> R2 M M')
    -> forall M M', mevp T T' R1 M M' -> mevp T T' R2 M M'.
Proof.
intros t t' R1 R2 Hincl e e' HR1ev.
destruct HR1ev as (Hhaltiff & Hall1 & Hofs).
do2 2 split; auto.
Qed.


Lemma mev_mono_sg :
  forall T T' U U' R,
    subs nil T U
    -> subs nil T' U'
    -> forall M M', mev T T' R M M' -> mev U U' R M M'.
Proof.
intros T1 T1' T2 T2' R Hsubs1 Hsubs2 M1 M2 HM.
destruct HM as (Hhalt & Hrel & Hof & Hof').
do2 3 split; eauto with static.
Qed.


Lemma mevp_mono_sg :
  forall T T' U U' R,
    subs nil T U
    -> subs nil T' U'
    -> forall M M', mevp T T' R M M' -> mevp U U' R M M'.
Proof.
intros T1 T1' T2 T2' R Hsubs1 Hsubs2 M1 M2 HM.
destruct HM as (Hhalt & Hrel & Hof & Hof').
do2 3 split; eauto with static.
Qed.


Lemma tpclosed_tev_ceqt :
  forall t1 t2 R e1 e2 e1' e2',
    tvrel t1 t2 R
    -> tpclosed t1 t2 R
    -> ceqt nil e1 e1' t1
    -> ceqt nil e2 e2' t2
    -> tev t1 t2 R e1 e2
    -> tev t1 t2 R e1' e2'.
Proof.
intros t1 t2 R e1 e2 e1' e2' Hrel Hclosed Hceq1 Hceq2 HR.
so (ceqt_reg _#4 wf_nil Hceq1 ander) as Hofe1'.
so (ceqt_reg _#4 wf_nil Hceq2 ander) as Hofe2'.
destruct HR as (Hhaltiff & Hall & Hofs).
#2 do2 2 split; auto.
  split.
    intros Hhalt.
    eapply ceqt_halt; eauto; [].
    apply Hhaltiff; [].
    eapply ceqt_halt; eauto using ceqt_symm; done.

    intros Hhalt.
      eapply ceqt_halt; eauto; [].
    apply Hhaltiff; [].
    eapply ceqt_halt; eauto using ceqt_symm; done.

  intros v1' v2' Hsteps1' Hsteps2' Hval1' Hval2'.
  so (ceqt_halt _#3 (ceqt_symm _#4 Hceq1) (eintro v1' (conj Hsteps1' Hval1'))) as (v1 & Hsteps1 & Hval1).
  so (ceqt_halt _#3 (ceqt_symm _#4 Hceq2) (eintro v2' (conj Hsteps2' Hval2'))) as (v2 & Hsteps2 & Hval2).
  so (Hall _ _ Hsteps1 Hsteps2 Hval1 Hval2) as HR.
  #2 eapply tpst_ceqt; eauto using tpst_idem, tpst_tvrel.
    eapply anti_tsteps_ceqt_trans; eauto; [].
    eapply ceqt_tsteps_trans; eauto; done.
    
    eapply anti_tsteps_ceqt_trans; eauto; [].
    eapply ceqt_tsteps_trans; eauto; done.
Qed.


Lemma mpclosed_mev_ceqm :
  forall T1 T2 R M1 M2 M1' M2',
    mvrel T1 T2 R
    -> mpclosed T1 T2 R
    -> ceqm nil impure M1 M1' T1
    -> ceqm nil impure M2 M2' T2
    -> mev T1 T2 R M1 M2
    -> mev T1 T2 R M1' M2'.
Proof.
intros t1 t2 R e1 e2 e1' e2' HofR Hclosed Hceq1 Hceq2 HR.
so (ceqm_reg _#5 wf_nil Hceq1 ander) as Hofe1'.
so (ceqm_reg _#5 wf_nil Hceq2 ander) as Hofe2'.
destruct HR as (Hhaltiff & Hall & Hofs).
#2 do2 3 split; auto.
  split.
    intros Hhalt.
    eapply ceqm_halt; eauto; [].
    apply Hhaltiff; [].
    eapply ceqm_halt; eauto using ceqm_symm; done.

    intros Hhalt.
    eapply ceqm_halt; eauto; [].
    apply Hhaltiff; [].
    eapply ceqm_halt; eauto using ceqm_symm; done.

  intros v1' v2' Hsteps1' Hsteps2' Hval1' Hval2'.
  so (ceqm_halt _#3 (ceqm_symm _#5 Hceq1) (eintro v1' (conj Hsteps1' Hval1'))) as (v1 & Hsteps1 & Hval1).
  so (ceqm_halt _#3 (ceqm_symm _#5 Hceq2) (eintro v2' (conj Hsteps2' Hval2'))) as (v2 & Hsteps2 & Hval2).
  so (Hall _ _ Hsteps1 Hsteps2 Hval1 Hval2) as HR.
  #2 eapply mpclosed_ceqm; eauto using mpst_idem, mpst_mvrel.
    eapply anti_msteps_ceqm_trans; eauto; [].
    eapply ceqm_msteps_trans; eauto; done.
    
    eapply anti_msteps_ceqm_trans; eauto; [].
    eapply ceqm_msteps_trans; eauto; done.
Qed.


Lemma mevp_ceqm :
  forall T1 T2 R M1 M2 M1' M2',
    mvrel T1 T2 R
    -> (forall V1 V2 V1' V2',
          ceqm nil pure V1 V1' T1
          -> ceqm nil pure V2 V2' T2
          -> mvalue nil V1'
          -> mvalue nil V2'
          -> R V1 V2
          -> R V1' V2')
    -> ceqm nil pure M1 M1' T1
    -> ceqm nil pure M2 M2' T2
    -> mevp T1 T2 R M1 M2
    -> mevp T1 T2 R M1' M2'.
Proof.
intros T1 T2 R M1 M2 M1' M2' HofR Hresp Hceq1 Hceq2 HM.
destruct HM as (Hhaltiff & Hrel & Hof1 & Hof2).
so (ceqm_reg _#5 wf_nil Hceq1) as (_ & HofM1').
so (ceqm_reg _#5 wf_nil Hceq2) as (_ & HofM2').
#2 do2 3 split; auto.
  transitivity (mhalt nil M1).
    symmetry.
    eapply ceqm_halt_iff; eauto; [].
    apply ceqm_comp; apply mcc_forget; eauto; done.

    transitivity (mhalt nil M2).
      assumption.

      eapply ceqm_halt_iff; eauto; [].
      apply ceqm_comp; apply mcc_forget; eauto; done.

  intros V V' HstepsV HstepsV' HvalV HvalV'.
  so (ceqm_halt _#3 (ceqm_comp _#5 (mcc_forget _#9 (ceqm_symm _#5 Hceq1))) (eintro V (conj HstepsV HvalV))) as (W & HstepsW & HvalW).
  so (ceqm_halt _#3 (ceqm_comp _#5 (mcc_forget _#9 (ceqm_symm _#5 Hceq2))) (eintro V' (conj HstepsV' HvalV'))) as (W' & HstepsW' & HvalW').
  so (Hrel _ _ HstepsW HstepsW' HvalW HvalW') as Hint.
  #2 eapply Hresp; eauto.
    eapply anti_msteps_ceqm_trans; eauto; [].
    eapply ceqm_msteps_trans; eauto; done.

    eapply anti_msteps_ceqm_trans; eauto; [].
    eapply ceqm_msteps_trans; eauto; done.
Qed.


(* Invariance under definitional equality *)

Lemma tvrel_eqc :
  forall t1 t1' t2 t2' R,
    eqc nil t1 t1' kd_type
    -> eqc nil t2 t2' kd_type
    -> tvrel t1 t2 R
    -> tvrel t1' t2' R.
Proof.
intros t1 t1' t2 t2' R Heq1 Heq2 Hrel.
intros v v' Hv.
so (Hrel _ _ Hv) as (Hof & Hof' & Hval & Hval').
#2 do2 3 split; auto.
  eapply oft_equiv; eauto; done.

  eapply oft_equiv; eauto; done.
Qed.


Lemma mvrel_eqs :
  forall T1 T1' T2 T2' R,
    eqs nil T1 T1'
    -> eqs nil T2 T2'
    -> mvrel T1 T2 R
    -> mvrel T1' T2' R.
Proof.
intros T1 T1' T2 T2' R Heq1 Heq2 Hrel.
intros v v' Hv.
so (Hrel _ _ Hv) as (Hof & Hof' & Hval & Hval').
#2 do2 3 split; auto.
  eapply ofm_subsume; eauto with static; done.

  eapply ofm_subsume; eauto with static; done.
Qed.


Lemma tps_eqc :
  forall t1 t1' t2 t2' R,
    eqc nil t1 t1' kd_type
    -> eqc nil t2 t2' kd_type
    -> tps t1 t2 R = tps t1' t2' R.
Proof.
intros t1 t1' t2 t2' R Heq1 Heq2.
fextensionality 2.
intros f f'.
pextensionality.
  intro H; destruct H as (Hrel & Hof & Hof' & Hval & Hval').
  #2 do2 4 split; auto.
    eapply oft_equiv; eauto with static; done.

    eapply oft_equiv; eauto with static; done.

  intro H; destruct H as (Hrel & Hof & Hof' & Hval & Hval').
  #2 do2 4 split; auto.
    eapply oft_equiv; eauto with static; done.

    eapply oft_equiv; eauto with static; done.
Qed.


Lemma mps_eqs :
  forall T1 T1' T2 T2' R,
    eqs nil T1 T1'
    -> eqs nil T2 T2'
    -> mps T1 T2 R = mps T1' T2' R.
Proof.
intros T1 T1' T2 T2' R Heq1 Heq2.
so (can_fsts _ _ (eqs_reg _#3 wf_nil Heq1 andel)) as (k1 & Hfst1).
so (can_fsts _ _ (eqs_reg _#3 wf_nil Heq2 andel)) as (k2 & Hfst2).
fextensionality 2.
intros f f'.
pextensionality.
  intro H; destruct H as (Hrel & Hof & Hof' & Hval & Hval').
  #2 do2 4 split; auto.
    eapply ofm_subsume; eauto with static; done.

    eapply ofm_subsume; eauto with static; done.

  intro H; destruct H as (Hrel & Hof & Hof' & Hval & Hval').
  #2 do2 4 split; auto.
    eapply ofm_subsume; eauto with static; done.

    eapply ofm_subsume; eauto with static; done.
Qed.


Lemma tpt_eqc :
  forall t1 t1' t2 t2' R,
    eqc nil t1 t1' kd_type
    -> eqc nil t2 t2' kd_type
    -> tpt t1 t2 R = tpt t1' t2' R.
Proof.
intros t1 t1' t2 t2' R Heq1 Heq2.
fextensionality 2.
intros v v'.
pextensionality.
  intro H; destruct H as (Hrel & Hof & Hof' & Hval & Hval').
  #2 do2 4 split; auto.
    eapply oft_equiv; eauto with static; done.

    eapply oft_equiv; eauto with static; done.

  intro H; destruct H as (Hrel & Hof & Hof' & Hval & Hval').
  #2 do2 4 split; auto.
    eapply oft_equiv; eauto with static; done.

    eapply oft_equiv; eauto with static; done.
Qed.


Lemma mpt_eqs :
  forall T1 T1' T2 T2' R,
    eqs nil T1 T1'
    -> eqs nil T2 T2'
    -> mpt T1 T2 R = mpt T1' T2' R.
Proof.
intros T1 T1' T2 T2' R Heq1 Heq2.
so (can_fsts _ _ (eqs_reg _#3 wf_nil Heq1 andel)) as (k1 & Hfst1).
so (can_fsts _ _ (eqs_reg _#3 wf_nil Heq2 andel)) as (k2 & Hfst2).
fextensionality 2.
intros f f'.
pextensionality.
  intro H; destruct H as (Hrel & Hof & Hof' & Hval & Hval').
  #2 do2 4 split; auto.
    eapply ofm_subsume; eauto with static; done.

    eapply ofm_subsume; eauto with static; done.

  intro H; destruct H as (Hrel & Hof & Hof' & Hval & Hval').
  #2 do2 4 split; auto.
    eapply ofm_subsume; eauto with static; done.

    eapply ofm_subsume; eauto with static; done.
Qed.


Lemma tpst_eqc :
  forall t1 t1' t2 t2' R,
    eqc nil t1 t1' kd_type
    -> eqc nil t2 t2' kd_type
    -> tpst t1 t2 R = tpst t1' t2' R.
Proof.
intros t1 t1' t2 t2' R Heq1 Heq2.
unfold tpst.
rewrite <- (tps_eqc _#5 Heq1 Heq2).
rewrite <- (tpt_eqc _#5 Heq1 Heq2).
reflexivity.
Qed.


Lemma mpst_eqs :
  forall T1 T1' T2 T2' R,
    eqs nil T1 T1'
    -> eqs nil T2 T2'
    -> mpst T1 T2 R = mpst T1' T2' R.
Proof.
intros T1 T1' T2 T2' R Heq1 Heq2.
unfold mpst.
rewrite <- (mps_eqs _#5 Heq1 Heq2).
rewrite <- (mpt_eqs _#5 Heq1 Heq2).
reflexivity.
Qed.


Lemma tpclosed_eqc :
  forall t1 t1' t2 t2' R,
    eqc nil t1 t1' kd_type
    -> eqc nil t2 t2' kd_type
    -> tpclosed t1 t2 R
    -> tpclosed t1' t2' R.
Proof.
intros t1 t1' t2 t2' R Heqc1 Heqc2 Hcl.
intros v v' Hrel.
apply Hcl; [].
rewrite -> (tpst_eqc _#5 Heqc1 Heqc2).
assumption.
Qed.


Lemma mpclosed_eqs :
  forall T1 T1' T2 T2' R,
    eqs nil T1 T1'
    -> eqs nil T2 T2'
    -> mpclosed T1 T2 R
    -> mpclosed T1' T2' R.
Proof.
intros T1 T1' T2 T2' R Heqc1 Heqc2 Hcl.
intros v v' Hrel.
apply Hcl; [].
rewrite -> (mpst_eqs _#5 Heqc1 Heqc2).
assumption.
Qed.


Lemma tev_eqc :
  forall t1 t1' t2 t2' R,
    eqc nil t1 t1' kd_type
    -> eqc nil t2 t2' kd_type
    -> tev t1 t2 R = tev t1' t2' R.
Proof.
cut (forall t1 t1' t2 t2' R v v',
       eqc nil t1 t1' kd_type
       -> eqc nil t2 t2' kd_type
       -> tev t1 t2 R v v'
       -> tev t1' t2' R v v').
  intros H t1 t1' t2 t2' R Heqc1 Heqc2.
  fextensionality 2.
  intros v v'.
  pextensionality.
    intro; eapply H; eauto; done.

    intro; eapply H; eauto using eqc_symm; done.
intros t1 t1' t2 t2' R v v' Heqc1 Heqc2 Hrel.
destruct Hrel as (Hhaltiff & Hvalrel & Hof & Hof').
do2 3 split; eauto with static; done.
Qed.


Lemma mev_eqs :
  forall T1 T1' T2 T2' R,
    eqs nil T1 T1'
    -> eqs nil T2 T2'
    -> mev T1 T2 R = mev T1' T2' R.
Proof.
cut (forall T1 T1' T2 T2' R v v',
       eqs nil T1 T1'
       -> eqs nil T2 T2'
       -> mev T1 T2 R v v'
       -> mev T1' T2' R v v').
  intros H T1 T1' T2 t2' R Heqc1 Heqc2.
  fextensionality 2.
  intros v v'.
  pextensionality.
    intro; eapply H; eauto; done.

    intro; eapply H; eauto using eqs_symm; done.
intros T1 T1' T2 T2' R v v' Heqc1 Heqc2 Hrel.
destruct Hrel as (Hhaltiff & Hvalrel & Hof & Hof').
do2 3 split; eauto with static; done.
Qed.


Lemma mevp_eqs :
  forall T1 T1' T2 T2' R,
    eqs nil T1 T1'
    -> eqs nil T2 T2'
    -> mevp T1 T2 R = mevp T1' T2' R.
Proof.
cut (forall T1 T1' T2 T2' R v v',
       eqs nil T1 T1'
       -> eqs nil T2 T2'
       -> mevp T1 T2 R v v'
       -> mevp T1' T2' R v v').
  intros H T1 T1' T2 t2' R Heqc1 Heqc2.
  fextensionality 2.
  intros v v'.
  pextensionality.
    intro; eapply H; eauto; done.

    intro; eapply H; eauto using eqs_symm; done.
intros T1 T1' T2 T2' R v v' Heqc1 Heqc2 Hrel.
destruct Hrel as (Hhaltiff & Hvalrel & Hof & Hof').
do2 3 split; eauto with static; done.
Qed.


Lemma tps_elim_tev :
  forall t t' R f f' e e',
    tps t t' R f f'
    -> tev t t' R e e'
    -> (thalt nil (tm_app f e) <-> thalt nil (tm_app f' e')).
Proof.
intros t t' R f f' e e' Hf He.
so (Hf ander ander ander) as (Hvalf & Hvalf').
split.
  intro Hhaltfe.
  so (Inversion.app_halt_inversion' _ _ Hhaltfe) as (_ & Hhalte).
  so (He andel andel Hhalte) as Hhalte'.
  destruct Hhalte as (v & Hstepsv & Hvalv).
  destruct Hhalte' as (v' & Hstepsv' & Hvalv').
  cut (thalt nil (tm_app f' v')).
    intro H.
    eapply anti_tsteps_thalt; eauto; [].
    apply star_map'; auto; [].
    intros; apply tstep_app2; auto; done.
  so (He ander andel v v' Hstepsv Hstepsv' Hvalv Hvalv') as Hv.
  refine (Hf andel _ _ Hv andel _); [].
  eapply tsteps_thalt; eauto; [].
  apply star_map'; auto; [].
  intros; apply tstep_app2; auto; done.

  intro Hhaltfe'.
  so (Inversion.app_halt_inversion' _ _ Hhaltfe') as (_ & Hhalte').
  so (He andel ander Hhalte') as Hhalte.
  destruct Hhalte as (v & Hstepsv & Hvalv).
  destruct Hhalte' as (v' & Hstepsv' & Hvalv').
  cut (thalt nil (tm_app f v)).
    intro H.
    eapply anti_tsteps_thalt; eauto; [].
    apply star_map'; auto; [].
    intros; apply tstep_app2; auto; done.
  so (He ander andel v v' Hstepsv Hstepsv' Hvalv Hvalv') as Hv.
  refine (Hf andel _ _ Hv ander _); [].
  eapply tsteps_thalt; eauto; [].
  apply star_map'; auto; [].
  intros; apply tstep_app2; auto; done.
Qed.


Lemma mps_elim_mev :
  forall T T' R f f' e e',
    mps T T' R f f'
    -> mev T T' R e e'
    -> (mhalt nil (md_appgn f e) <-> mhalt nil (md_appgn f' e')).
Proof.
intros t t' R f f' e e' Hf He.
so (Hf ander ander ander) as (Hvalf & Hvalf').
split.
  intro Hhaltfe.
  so (Inversion.appgn_halt_inversion' _ _ Hhaltfe) as (_ & Hhalte).
  so (He andel andel Hhalte) as Hhalte'.
  destruct Hhalte as (v & Hstepsv & Hvalv).
  destruct Hhalte' as (v' & Hstepsv' & Hvalv').
  cut (mhalt nil (md_appgn f' v')).
    intro H.
    eapply anti_msteps_mhalt; eauto; [].
    apply star_map'; auto; [].
    intros; apply mstep_appgn2; auto; done.
  so (He ander andel v v' Hstepsv Hstepsv' Hvalv Hvalv') as Hv.
  refine (Hf andel _ _ Hv andel _); [].
  eapply msteps_mhalt; eauto; [].
  apply star_map'; auto; [].
  intros; apply mstep_appgn2; auto; done.

  intro Hhaltfe'.
  so (Inversion.appgn_halt_inversion' _ _ Hhaltfe') as (_ & Hhalte').
  so (He andel ander Hhalte') as Hhalte.
  destruct Hhalte as (v & Hstepsv & Hvalv).
  destruct Hhalte' as (v' & Hstepsv' & Hvalv').
  cut (mhalt nil (md_appgn f v)).
    intro H.
    eapply anti_msteps_mhalt; eauto; [].
    apply star_map'; auto; [].
    intros; apply mstep_appgn2; auto; done.
  so (He ander andel v v' Hstepsv Hstepsv' Hvalv Hvalv') as Hv.
  refine (Hf andel _ _ Hv ander _); [].
  eapply msteps_mhalt; eauto; [].
  apply star_map'; auto; [].
  intros; apply mstep_appgn2; auto; done.
Qed.



(* Transposes *)

(* Like transpose_inv, but uses extensionality. *)
Lemma transpose_involution :
  forall (T : Set) (R : T -> T -> Prop),
    transpose (transpose R) = R.
Proof.
intros T R.
fextensionality 2.
intros x y.
pextensionality; eauto using transpose_inv.
Qed.


Lemma tvrel_transpose :
  forall t t' R,
    tvrel t' t (transpose R) <-> tvrel t t' R.
Proof.
intros t t' R.
split.
  intro H.
  intros v v' Hv.
  destruct_all (H _ _ Hv).
  repeat2 split; auto; done.

  intro H.
  intros v v' Hv.
  destruct_all (H _ _ Hv).
  repeat2 split; auto; done.
Qed.


Lemma tps_transpose :
  forall t t' R,
    tps t t' (transpose R) = transpose (tps t' t R).
Proof.
intros t t' R.
fextensionality 2.
intros v v'.
pextensionality.
  intro H.
  unfold transpose.
  destruct H as (Hrel & Hof & Hof' & Hval & Hval').
  unfold transpose in Hrel.
  do2 4 split; auto; [].
  intros w w' Hw.
  so (Hrel _ _ Hw) as Hiff.
  symmetry.
  assumption.

  intro H.
  unfold transpose in H.
  destruct H as (Hrel & Hof & Hof' & Hval & Hval').
  do2 4 split; auto; [].
  intros w w' Hw.
  unfold transpose in Hw.
  so (Hrel _ _ Hw) as Hiff.
  symmetry.
  assumption.
Qed.


Lemma tpt_transpose :
  forall t t' R,
    tpt t t' (transpose R) = transpose (tpt t' t R).
Proof.
intros t t' R.
fextensionality 2.
intros v v'.
pextensionality.
  intro H.
  unfold transpose.
  destruct H as (Hrel & Hof & Hof' & Hval & Hval').
  unfold transpose in Hrel.
  do2 4 split; auto; [].
  intros f f' Hf.
  so (Hrel _ _ Hf) as Hiff.
  symmetry.
  assumption.

  intro H.
  unfold transpose in H.
  destruct H as (Hrel & Hof & Hof' & Hval & Hval').
  do2 4 split; auto; [].
  intros f f' Hf.
  so (Hrel _ _ Hf) as Hiff.
  symmetry.
  assumption.
Qed.


Lemma tpst_transpose :
  forall t t' R,
    tpst t t' (transpose R) = transpose (tpst t' t R).
Proof.
intros t t' R.
unfold tpst.
rewrite -> tps_transpose; [].
rewrite -> tpt_transpose; [].
reflexivity.
Qed.


Lemma tev_transpose :
  forall t t' R,
    tev t t' (transpose R) = transpose (tev t' t R).
Proof.
intros t t' R.
fextensionality 2.
intros e e'.
pextensionality.
  intro H.
  unfold transpose.
  destruct H as (Halfiff & Hrel & Hof & Hof').
  #2 do2 3 split; auto.
    symmetry; assumption.

    intros v' v; intros.
    fold (transpose R v v').
    auto; done.

  intro H.
  unfold transpose in H.
  destruct H as (Halfiff & Hrel & Hof & Hof').
  #2 do2 3 split; auto.
    symmetry; assumption.

    intros v v'; intros.
    unfold transpose.
    auto; done.
Qed.
