
Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Formation.
Require Import Regularity.
Require Import Determinism.
Require Import Congruence.
Require Import Contextual.
Require Import Divergence.
Require Import Preservation.
Require Import Inversion.
Require Consistent.
Require Principal.
Require Wf_nat.
Require Compare_dec.
Require Compare.


Definition bottom (t : (* tp *) term) :=
  tm_fix t (tm_lam (cn_arrow cn_unit t) (tm_app (var 0) tm_star)).


Lemma oft_bottom :
  forall G t, ofc G t kd_type -> oft G (bottom t) t.
Proof.
intros G t Hof.
apply oft_fix; [].
apply oft_lam; eauto with static; [].
eapply oft_app; eauto with static; [].
replsub (cn_arrow cn_unit (shift1 t)) with (shift1 (cn_arrow cn_unit t)).
apply oft_var; auto with sequence.
Qed.


Lemma bottom_diverge :
  forall G t, diverge G (bottom t).
Proof.
intros G t.
intros e Hsteps.
remember (bottom t) as b eqn:Heqb.
assert (b = bottom t
        \/ b = tm_app (tm_lam (cn_arrow cn_unit t) (tm_app (var 0) tm_star)) (tm_lam cn_unit (shift1 (bottom t)))
        \/ b = tm_app (tm_lam cn_unit (shift1 (bottom t))) tm_star) as Hcases.
  left.
  assumption.
clear Heqb.
revert Hcases.
induct Hsteps.

(* refl *)
intros e Hcases.
destruct Hcases as [-> | [-> | ->]];
unfold bottom;
eauto with dynamic; done.

(* step *)
intros b e e' Hstep _ IH Hcases.
#3 destruct Hcases as [-> | [-> | ->]].
  invert Hstep; [].
  intros <-.
  apply IH; [].
  right; left.
  reflexivity.

  #3 invert Hstep.
    intros e'' Hstep' _.
    invert Hstep'; done.

    intros e'' _ Hstep' _.
    invert Hstep'; done.

    intros _ <-.
    apply IH; [].
    right; right.
    reflexivity.

  #3 invert Hstep.
    intros e'' Hstep' _.
    invert Hstep'; done.

    intros e'' _ Hstep' _.
    invert Hstep'; done.

    intros _ <-.
    apply IH; [].
    left.
    simp_sub.
    reflexivity.
Qed.


Fixpoint fixn (n : nat) (t : (* tp *) term) (f : (* tm *) term) {struct n} :=
  (match n with
   | 0 => bottom t

   | S i => tm_app f (tm_lam cn_unit (shift1 (fixn i t f)))
   end).


Lemma oft_fixn :
  forall G n t f,
    ofc G t kd_type
    -> oft G f (cn_arrow (cn_arrow cn_unit t) t)
    -> oft G (fixn n t f) t.
Proof.
intros G n t f Hoft Hoff.
induct n.

(* 0 *)
simpl.
apply oft_bottom; auto; done.

(* S *)
intros N IH.
simpl.
eapply oft_app; eauto; [].
apply oft_lam; eauto with static; [].
refine (subst_oft _#5 (weaken1 _#3) _); auto; done.
Qed.


Lemma fixn_approx :
  forall G n t f,
    ofc G t kd_type
    -> oft G f (cn_arrow (cn_arrow cn_unit t) t)
    -> capproxt G (fixn n t f) (tm_fix t f) t.
Proof.
intros G n t f Hoft Hoff.
induct n.

(* 0 *)
simpl.
#3 apply diverge_capproxt.
  apply oft_bottom; auto; done.

  apply oft_fix; auto; done.

  apply bottom_diverge; done.

(* S *)
intros n IH.
simpl.
#2 apply (capproxt_trans _ _ (tm_app f (tm_lam cn_unit (shift1 (tm_fix t f))))).
  apply capproxt_comp; eapply tcc_app; eauto using capproxt_refl; [].
  apply capproxt_comp; eapply tcc_lam; eauto with static; [].
  apply capproxt_with_wf; [].
  intro Hwf.
  apply weaken1_capproxt; auto; done.

  apply anti_tstep_capproxt; eauto with static dynamic; done.
Qed.


Lemma fixn_mono :
  forall G m n t f,
    m <= n
    -> ofc G t kd_type
    -> oft G f (cn_arrow (cn_arrow cn_unit t) t)
    -> capproxt G (fixn m t f) (fixn n t f) t.
Proof.
intros G m n t f Hleq Hoft Hoff.
repl n with (m + (n - m)) by omega.
remember (n - m) as i eqn:Heqi.
clear Hleq Heqi.
revert i.
induct m.

(* 0 *)
intros k.
simpl.
#3 apply diverge_capproxt.
  apply oft_bottom; auto; done.

  apply oft_fixn; auto; done.

  apply bottom_diverge; done.

(* S *)
intros m IH i.
simpl.
apply capproxt_comp; eapply tcc_app; eauto using capproxt_refl; [].
apply capproxt_comp; eapply tcc_lam; eauto with static; [].
apply weaken1_capproxt; auto; done.
Qed.


Definition fixv (t : (* tp *) term) (f : (* tm *) term) := tm_lam cn_unit (shift1 (tm_fix t f)).
Definition fixnv (n : nat) (t : (* tp *) term) (f : (* tm *) term) := tm_lam cn_unit (shift1 (fixn n t f)).


Lemma oft_fixv :
  forall G t f,
    ofc G t kd_type
    -> oft G f (cn_arrow (cn_arrow cn_unit t) t)
    -> oft G (fixv t f) (cn_arrow cn_unit t).
Proof.
intros G t f Hoft Hoff.
apply oft_lam; auto with static; [].
eapply subst_oft; eauto using weaken1, oft_fix.
Qed.


Lemma oft_fixnv :
  forall G n t f,
    ofc G t kd_type
    -> oft G f (cn_arrow (cn_arrow cn_unit t) t)
    -> oft G (fixnv n t f) (cn_arrow cn_unit t).
Proof.
intros G n t f Hoft Hoff.
apply oft_lam; auto with static; [].
eapply subst_oft; eauto using weaken1, oft_fixn.
Qed.


Lemma minus_plus_dist :
  forall m n p, m - (n + p) = m - n - p.
Proof.
intros; omega.
Qed.


Lemma id_thru :
  forall (R : context -> term -> term -> Prop) s m C,
    R s m C -> R s m (subst id C).
Proof.
intros R s m C H.
simp_sub.
assumption.
Qed.


Lemma subst_ctx_allcl :
  forall G s,
    Forall iscl G
    -> Forall iscl (subst_ctx s G).
Proof.
intros G s H.
induct H.

(* nil *)
apply Forall_nil; done.

(* cons *)
intros C G Hiscl _ IH.
simpl.
apply Forall_cons; auto; [].
invert Hiscl; intros; subst; simp_sub; auto using iscl; done.
Qed.


Lemma lower_approxt_gen :
  forall u f G e t i i',
    oft nil f (cn_arrow (cn_arrow cn_unit u) u)
    -> wf G
    -> oft (nil; cl_tm (cn_arrow cn_unit u);; subst_ctx sh1 G) e (subst (under (length G) sh1) t)
    -> i <= i'
    -> capproxt G (subst (under (length G) (dot (fixnv i u f) id)) e) (subst (under (length G) (dot (fixnv i' u f) id)) e) t.
Proof.
intros u f G e t i i' Hoff Hwf Hofe Hleq.
so (ofc_ofcr _#3 (oft_reg _#3 wf_nil Hoff)) as (l & Hofcr).
invert Hofcr; [].
intros _ Hofu _.
replsub t with (subst (under (length G) (dot (fixnv i u f) id)) (subst (under (length G) sh1) t)).
#3 refine (funct_capprox_tm _#9 Hofe).
  assumption.

  apply wf_wfh_append; auto with static; [].
  so (wf_wfh _ Hwf) as Hwfh.
  eapply subst_wfh; eauto; [].
  apply weaken1; done.

  #2 repl G with (nil;; subst_ctx (dot (fixnv i u f) id) (subst_ctx sh1 G)) at 1.
    rewrite <- subst_ctx_compose; [].
    simp_sub.
    rewrite -> subst_ctx_id; [].
    rewrite -> app_nil_r; [].
    reflexivity.

    repl (length G) with (length (subst_ctx sh1 G)).
      apply subst_ctx_length; done.
    #4 apply subrel_under.
      apply deq_behaved; exact capprox_creflexive.
      
      apply wclorel_wcrel; [].
      exact deq_capprox_wclo.
  
      apply subst_ctx_allcl; [].
      apply wf_allcl; auto; done.
  
      apply subrel_dot; auto using subrel_id; [].
      simp_sub.
      apply deq_tm; [].
      unfold fixnv.
      apply capproxt_comp; apply tcc_lam; auto with static; [].
      apply weaken1_capproxt; [].
      eapply fixn_mono; eauto; done.
Qed.


Lemma lower_approxm_gen :
  forall u f G p M T i i',
    oft nil f (cn_arrow (cn_arrow cn_unit u) u)
    -> wf G
    -> ofm (nil; cl_tm (cn_arrow cn_unit u);; subst_ctx sh1 G) p M (subst (under (length G) sh1) T)
    -> i <= i'
    -> capproxm G p (subst (under (length G) (dot (fixnv i u f) id)) M) (subst (under (length G) (dot (fixnv i' u f) id)) M) T.
Proof.
intros u f G p M T i i' Hoff Hwf HofM Hleq.
so (ofc_ofcr _#3 (oft_reg _#3 wf_nil Hoff)) as (l & Hofcr).
invert Hofcr; [].
intros _ Hofu _.
replsub T with (subst (under (length G) (dot (fixnv i u f) id)) (subst (under (length G) sh1) T)).
#3 refine (funct_capprox_md _#10 HofM).
  assumption.

  apply wf_wfh_append; auto with static; [].
  so (wf_wfh _ Hwf) as Hwfh.
  eapply subst_wfh; eauto; [].
  apply weaken1; done.

  #2 repl G with (nil;; subst_ctx (dot (fixnv i u f) id) (subst_ctx sh1 G)) at 1.
    rewrite <- subst_ctx_compose; [].
    simp_sub.
    rewrite -> subst_ctx_id; [].
    rewrite -> app_nil_r; [].
    reflexivity.

    repl (length G) with (length (subst_ctx sh1 G)).
      apply subst_ctx_length; done.
    #4 apply subrel_under.
      apply deq_behaved; exact capprox_creflexive.
  
      apply wclorel_wcrel; exact deq_capprox_wclo.
  
      apply subst_ctx_allcl; [].
      apply wf_allcl; auto; done.
  
      apply subrel_dot; auto using subrel_id; [].
      simp_sub.
      apply deq_tm; [].
      unfold fixnv.
      apply capproxt_comp; apply tcc_lam; auto with static; [].
      apply weaken1_capproxt; [].
      eapply fixn_mono; eauto; done.
Qed.


Lemma lower_approxt :
  forall u f e t i i',
    oft nil f (cn_arrow (cn_arrow cn_unit u) u)
    -> oft (nil; cl_tm (cn_arrow cn_unit u)) e (shift1 t)
    -> i <= i'
    -> capproxt nil (subst1 (fixnv i u f) e) (subst1 (fixnv i' u f) e) t.

Proof.
intros u f e t i i' Hoff Hofe Hleq.
exploit (lower_approxt_gen u f nil e t i i'); auto using wf_nil.
Qed.


Lemma lower_approxm :
  forall u f p M T i i',
    oft nil f (cn_arrow (cn_arrow cn_unit u) u)
    -> ofm (nil; cl_tm (cn_arrow cn_unit u)) p M (shift1 T)
    -> i <= i'
    -> capproxm nil p (subst1 (fixnv i u f) M) (subst1 (fixnv i' u f) M) T.
Proof.
intros u f p M T i i' Hoff HofM Hleq.
exploit (lower_approxm_gen u f nil p M T i i'); auto using wf_nil.
Qed.


Lemma tstepn_value_contra :
  forall G i e e',
    0 < i
    -> expon (tstep G) i e e'
    -> tvalue G e
    -> False.
Proof.
intros G i e e' Hlt Hsteps Hval.
repl i with (S (i-1)) in Hsteps by omega.
invert Hsteps; [].
intros e'' Hstep _.
eapply tstep_value_contra; eauto.
Qed.


Lemma mstepn_value_contra :
  forall G i M M',
    0 < i
    -> expon (mstep G) i M M'
    -> mvalue G M
    -> False.
Proof.
intros G i M M' Hlt Hsteps Hval.
repl i with (S (i-1)) in Hsteps by omega.
invert Hsteps; [].
intros e'' Hstep _.
eapply mstep_value_contra; eauto.
Qed.


Lemma unsubst_tvalue :
  forall G m e,
    tvalue G (subst1 m e)
    -> tvalue (G; cl_utm) e.
Proof.
intros G m e Hval.
revert Hval.
induct e;
try (intros; simp_sub in Hval; invert Hval; eauto using tvalue; done); [].
intros n Hval.
destruct n as [| n'].
  apply tvalue_varu; auto with sequence; done.

  simp_sub in Hval.
  invert Hval; eauto with dynamic sequence; done.
Qed.


Lemma tvalue_mvalue_contra :
  forall G v,
    tvalue G v -> mvalue G v -> False.
Proof.
intros G v Htval Hmval.
#4 invert Htval; invert Hmval; try (intros; subst; discriminate).

intros i T c Hindex Heq i' t Hindex' Heq'.
rewrite <- Heq' in Heq.
injection Heq.
intros; subst i'.
so (index_fun _#5 Hindex Hindex') as HeqC.
discriminate HeqC.

intros i c Hindex Heq i' t Hindex' Heq'.
rewrite <- Heq' in Heq.
injection Heq.
intros; subst i'.
so (index_fun _#5 Hindex Hindex') as HeqC.
discriminate HeqC.

intros i T c Hindex Heq i' Hindex' Heq'.
rewrite <- Heq' in Heq.
injection Heq.
intros; subst i'.
so (index_fun _#5 Hindex Hindex') as HeqC.
discriminate HeqC.

intros i c Hindex Heq i' Hindex' Heq'.
rewrite <- Heq' in Heq.
injection Heq.
intros; subst i'.
so (index_fun _#5 Hindex Hindex') as HeqC.
discriminate HeqC.
Qed.


Lemma unsubst_mvalue :
  forall G v M,
    tvalue G v
    -> mvalue G (subst1 v M)
    -> mvalue (G; cl_utm) M.
Proof.
intros G v M Hvalv Hval.
revert Hval.
#2 induct M;
try (intros; simp_sub in Hval; invert Hval; eauto using mvalue; done).
  intros n Hval.
  destruct n as [| n'].
    exfalso.
    simp_sub in Hval.
    eapply tvalue_mvalue_contra; eauto; done.
  
    simp_sub in Hval.
    invert Hval; eauto with dynamic sequence; done.

  intros e _ Hval.
  simp_sub in Hval.
  invert Hval; [].
  eauto using mvalue, unsubst_tvalue; done.
Qed.


Lemma unsubst_fstm :
  forall G e t M T c,
    wf G
    -> oft G e t
    -> ofm (G; cl_tm t) pure M T
    -> fstm G (subst1 e M) c
    -> fstm (G; cl_tm t) M (shift1 c).
Proof.
intros G e t M T c Hwf Hofe HofM Hfstm.
so (wf_tm Hwf (oft_reg _#3 Hwf Hofe)) as Hwf'.
so (can_fstm _#3 Hwf' HofM) as (c' & HfstM).
#2 exploit (subst_fstm G (G; cl_tm t) (dot e id) M c') as HfstMx; auto.
  apply subof_dot; auto using subof_id; [].
  simp_sub.
  apply tof_tm; done.
so (ofm_fst_o _#5 Hwf' HofM HfstM) as (k & _ & Hofc).
so (subordf_tm_cn' _#3 (ofc_fcn _#3 Hofc)) as Heqc'.
rewrite -> Heqc' in HfstMx; simp_sub in HfstMx.
so (fstm_fun _#4 Hfstm HfstMx); subst c.
rewrite <- Heqc'.
assumption.
Qed.


Lemma merge_iht :
  forall u f i e,
    (forall i' : nat,
      i' < i ->
      forall e t v,
      oft (nil; cl_tm (cn_arrow cn_unit u)) e (shift1 t) ->
      expon (tstep nil) i' (subst1 (fixv u f) e) v ->
      tvalue nil v ->
      exists (j : nat) (v' : term),
        oft (nil; cl_tm (cn_arrow cn_unit u)) v' (shift1 t) /\
        tvalue (nil; cl_utm) v' /\
        v = subst1 (fixv u f) v' /\
        (forall l : nat,
         l >= j ->
         capproxt nil (subst1 (fixnv (l - j) u f) v')
           (subst1 (fixnv l u f) e) t))
    -> (forall t v,
          nil; cl_utm = nil; cl_utm ->
          oft (nil; cl_tm (cn_arrow cn_unit u)) e (shift1 t) ->
          expon (tstep nil) i (subst1 (fixv u f) e) v ->
          tvalue nil v ->
          exists (j : nat) (v' : term),
            oft (nil; cl_tm (cn_arrow cn_unit u)) v' (shift1 t) /\
            tvalue (nil; cl_utm) v' /\
            v = subst1 (fixv u f) v' /\
            (forall l : nat,
             l >= j ->
             capproxt nil (subst1 (fixnv (l - j) u f) v')
               (subst1 (fixnv l u f) e) t))
    -> forall i' t v,
         i' <= i ->
         oft (nil; cl_tm (cn_arrow cn_unit u)) e (shift1 t) ->
         expon (tstep nil) i' (subst1 (fixv u f) e) v ->
         tvalue nil v ->
         exists (j : nat) (v' : term),
           oft (nil; cl_tm (cn_arrow cn_unit u)) v' (shift1 t) /\
           tvalue (nil; cl_utm) v' /\
           v = subst1 (fixv u f) v' /\
           (forall l : nat,
            l >= j ->
            capproxt nil (subst1 (fixnv (l - j) u f) v')
              (subst1 (fixnv l u f) e) t).
Proof.
intros u f i e IHo IHi i' t v Hleq.
so (Compare.le_decide _ _ Hleq) as [Hlt | ->].
  apply IHo; auto; done.

  apply IHi; auto; done.
Qed.       


Lemma merge_ihm :
  forall u f i M,
    (forall i' : nat,
      i' < i ->
      forall M T V p,
      ofm (nil; cl_tm (cn_arrow cn_unit u)) p M (shift1 T) ->
      expon (mstep nil) i' (subst1 (fixv u f) M) V ->
      mvalue nil V ->
      exists (j : nat) (V' : term),
        ofm (nil; cl_tm (cn_arrow cn_unit u)) p V' (shift1 T) /\
        mvalue (nil; cl_utm) V' /\
        V = subst1 (fixv u f) V' /\
        (forall l : nat,
         l >= j ->
         capproxm nil p (subst1 (fixnv (l - j) u f) V')
           (subst1 (fixnv l u f) M) T))
    -> (forall T V p,
          nil; cl_utm = nil; cl_utm ->
          ofm (nil; cl_tm (cn_arrow cn_unit u)) p M (shift1 T) ->
          expon (mstep nil) i (subst1 (fixv u f) M) V ->
          mvalue nil V ->
          exists (j : nat) (V' : term),
            ofm (nil; cl_tm (cn_arrow cn_unit u)) p V' (shift1 T) /\
            mvalue (nil; cl_utm) V' /\
            V = subst1 (fixv u f) V' /\
            (forall l : nat,
             l >= j ->
             capproxm nil p (subst1 (fixnv (l - j) u f) V')
               (subst1 (fixnv l u f) M) T))
    -> forall i' T V p,
         i' <= i ->
         ofm (nil; cl_tm (cn_arrow cn_unit u)) p M (shift1 T) ->
         expon (mstep nil) i' (subst1 (fixv u f) M) V ->
         mvalue nil V ->
         exists (j : nat) (V' : term),
           ofm (nil; cl_tm (cn_arrow cn_unit u)) p V' (shift1 T) /\
           mvalue (nil; cl_utm) V' /\
           V = subst1 (fixv u f) V' /\
           (forall l : nat,
            l >= j ->
            capproxm nil p (subst1 (fixnv (l - j) u f) V')
              (subst1 (fixnv l u f) M) T).
Proof.
intros u f i M IHo IHi i' T V p Hleq.
so (Compare.le_decide _ _ Hleq) as [Hlt | ->].
  apply IHo; auto; done.

  apply IHi; auto; done.
Qed.       


Lemma simulation :
  forall u f i,
     oft nil f (cn_arrow (cn_arrow cn_unit u) u)
     -> (forall e t v,
           oft (nil; cl_tm (cn_arrow cn_unit u)) e (shift1 t)
           -> expon (tstep nil) i (subst1 (fixv u f) e) v
           -> tvalue nil v
           -> exists j v',
                oft (nil; cl_tm (cn_arrow cn_unit u)) v' (shift1 t)
                /\ tvalue (nil; cl_utm) v' 
                /\ v = subst1 (fixv u f) v'
                /\ forall l,
                     l >= j
                     -> capproxt nil (subst1 (fixnv (l-j) u f) v') (subst1 (fixnv l u f) e) t)
        /\
        (forall M T V p,
           ofm (nil; cl_tm (cn_arrow cn_unit u)) p M (shift1 T)
           -> expon (mstep nil) i (subst1 (fixv u f) M) V
           -> mvalue nil V
           -> exists j V',
                ofm (nil; cl_tm (cn_arrow cn_unit u)) p V' (shift1 T)
                /\ mvalue (nil; cl_utm) V'
                /\ V = subst1 (fixv u f) V'
                /\ forall l,
                     l >= j
                     -> capproxm nil p (subst1 (fixnv (l-j) u f) V') (subst1 (fixnv l u f) M) T).
Proof.
intros u f i Hoff.
so (ofc_ofcr _#3 (oft_reg _#3 wf_nil Hoff)) as (l & Hofcr).
invert Hofcr; [].
intros _ Hofu _.
clear l Hofcr.
so (wf_tm wf_nil (ofc_arrow ofc_unit Hofu)) as Hwf.
wfinduct i using Wf_nat.lt_wf.
intros i IH.
so (fun i' Hlt => IH i' Hlt andel) as IHt.
so (fun i' Hlt => IH i' Hlt ander) as IHm.
clear IH.
so (Compare_dec.zerop i) as [-> | Hlti].
  assert (forall l, subof dof nil (dot (fixnv l u f) id) (nil; cl_tm (cn_arrow cn_unit u))) as Hs.
    intros l.
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply dof_tm; [].
    apply oft_fixnv; auto; done.
  split.
    intros e t v Hof Hsteps Hval.
    invert Hsteps; [].
    intros <-.
    exists 0, e.
    #2 do2 3 split; auto.
      exact (unsubst_tvalue _#3 Hval).

      intros l _.
      repl (l - 0) with l by omega.
      apply capproxt_refl; [].
      replsub t with (subst1 (fixnv l u f) (shift1 t)).
      eapply subst_oft; eauto; done.

    intros M T V p Hof Hsteps Hval.
    invert Hsteps; [].
    intros <-.
    exists 0, M.
    #2 do2 3 split; auto.
      exact (unsubst_mvalue _#3 tvalue_lam Hval).

      intros l _.
      repl (l - 0) with l by omega.
      apply capproxm_refl; [].
      replsub T with (subst1 (fixnv l u f) (shift1 T)).
      eapply subst_ofm; eauto; done.
exploit
  (dynamic_form_ind
     (fun G e =>
        forall t v,
          G = nil; cl_utm
          -> oft (nil; cl_tm (cn_arrow cn_unit u)) e (shift1 t)
          -> expon (tstep nil) i (subst1 (fixv u f) e) v
          -> tvalue nil v
          -> exists j v',
               oft (nil; cl_tm (cn_arrow cn_unit u)) v' (shift1 t)
               /\ tvalue (nil; cl_utm) v'
               /\ v = subst1 (fixv u f) v'
               /\ forall l,
                    l >= j
                    -> capproxt nil (subst1 (fixnv (l-j) u f) v') (subst1 (fixnv l u f) e) t)
     (fun G M =>
        forall T V p,
          G = nil; cl_utm
          -> ofm (nil; cl_tm (cn_arrow cn_unit u)) p M (shift1 T)
          -> expon (mstep nil) i (subst1 (fixv u f) M) V
          -> mvalue nil V
          -> exists j V',
               ofm (nil; cl_tm (cn_arrow cn_unit u)) p V' (shift1 T)
               /\ mvalue (nil; cl_utm) V'
               /\ V = subst1 (fixv u f) V'
               /\ forall l,
                    l >= j
                    -> capproxm nil p (subst1 (fixnv (l-j) u f) V') (subst1 (fixnv l u f) M) T)) as Hind.

(* var *)
intros G j t' Hindex t v -> _ _ _.
exfalso.
#2 invert Hindex.
  intros _ Heq.
  discriminate Heq.

  intros j' Hindex' _.
  invert Hindex'; done.

(* varu *)
intros G j Hindex t v -> _ Hsteps _.
assert (j = 0).
  invert Hindex; auto; [].
  intros j' Hindex' _.
  invert Hindex'; done.
subst j.
simp_sub in Hsteps.
exfalso; unfold fixv in Hsteps; eapply tstepn_value_contra; eauto with dynamic; done.

(* star *)
intros; exfalso; eapply tstepn_value_contra; eauto with dynamic; done.

(* lam *)
intros; exfalso; eapply tstepn_value_contra; eauto; simp_sub; auto with dynamic; done.

(* app *)
intros G e1 e2 _ _ _ _ t v -> Hof Hsteps Hval.
clear IHm.
so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 Hwf Hof)) as Hoft; simp_sub in Hoft.
simp_sub in Hsteps.
so (app_halt_inversion _#4 Hsteps Hval) as (i1 & i2 & i3 & t1 & e3 & v2 & Hlt1 & Hlt2 & Hlt3 & Hval2 & Hsteps1 & Hsteps2 & Hsteps3).
so (app_oft_inversion _#4 Hwf Hof) as (t' & Hof1 & Hof2).
so (ofc_fcn _#3 (oft_reg _#3 Hwf Hof1)) as Harrow.
invert Harrow; clear Harrow; [].
intros Ht' _.
so (subordf_tm_cn _#3 Ht') as Heqt'.
rewrite -> Heqt' in Hof1.
replsub (cn_arrow (subst (dot nonsense sh1) t') (shift1 t)) with (shift1 (cn_arrow (subst1 nonsense t') t)) in Hof1.
so (IHt _ Hlt1 _#3 Hof1 Hsteps1 tvalue_lam) as (j1 & v1' & Hofv1' & Hval1' & Heq & Hap1).
rewrite -> Heqt' in Hof2.
replsub (subst (dot nonsense sh1) t') with (shift1 (subst1 nonsense t')) in Hof2.
so (IHt _ Hlt2 _#3 Hof2 Hsteps2 Hval2) as (j2 & v2' & Hofv2' & Hval2' & -> & Hap2).
assert (v1' = var 0 \/ (exists t1' e3', v1' = tm_lam t1' e3')) as [-> | (t1' & e3' & ->)].
  case_eq v1'; try (intros; subst v1'; simp_sub in Heq; discriminate Heq); eauto; [].
  intros n ->.
  left.
  destruct n; auto; [].
  simp_sub in Heq; discriminate Heq.

  (* principal case *)
  simp_sub in Heq.
  injection Heq; clear Heq.
  intros -> ->.
  simp_sub in Hofv1'.
  so (subst_eqc _#6 (unshift_tm _ _) (Principal.oft_unique _#4 Hwf Hofv1' (oft_var index_0))) as Heqc.
  simp_sub in Heqc.
  so (Consistent.arrow_injective _#5 wf_nil Heqc) as (Heqct' & Heqctu).
  clear Heqc.
  simp_sub in Hofv2'.
  rewrite <- Heqt' in Hofv2'.
  simp_sub in Hsteps3.
  #2 invert Hsteps3.
    intros; subst v; invert Hval; done.
  intros i3' e Hstep Hsteps3' <-.
  so (tstep_fun _#4 Hstep tstep_fix); subst e.
  clear Hstep Hsteps3.
  assert (i3' < i) as Hlt3' by omega.
  #4 exploit (IHt i3' Hlt3' (tm_app (shift1 f) (var 0)) u v) as (j3 & v' & Hofv' & Hval' & -> & Hap3); clear IHt Hlt3'.
    #2 apply (@oft_app _ (shift1 (cn_arrow cn_unit u))).
      so (subst_oft _#5 (weaken1 _ _ (cl_tm (cn_arrow cn_unit u))) Hoff) as H.
      simp_sub in H.
      assumption.

      apply oft_var; eauto with sequence; done.

    simp_sub.
    exact Hsteps3'.

    assumption.
  exists (j1 + j2 + j3 + 1), v'.
  #2 do2 3 split; auto.
    eapply oft_equiv; eauto; [].
    eapply weaken1_eqc_type; eauto using eqc_symm; done.

    intros l Hleq.
    simp_sub.

    #2 apply (capproxt_trans _ _ (fixn (l-j1-j2) u f)).
      apply capproxt_comp; apply (@tcc_equiv _#6 u); eauto using eqc_symm; [].
      rewrite -> !minus_plus_dist; [].  (* Omega can't do this equality, for some reason. *)
      repl (l-j1-j2) with (S (l-1-j1-j2)) by omega.  (* Why so slow?! *)
      exploit (Hap3 (l-1-j1-j2)) as H; [omega |].
      simp_sub in H.
      exact H.

      #2 apply (anti_tsteps_capproxt_trans _ _ (tm_app (tm_lam cn_unit (shift1 (fixn (l-j1-j2) u f))) (subst1 (fixnv (l-j1-j2) u f) v2'))).
        #2 eapply star_step.
          apply tstep_app3; [].
          eapply subst_tvalue; eauto; [].
          apply subof_dot; auto using subof_id; [].
          simp_sub; apply oof_utm; [].
          apply tvalue_lam; done.

          simp_sub.
          apply star_refl; done.

        #2 apply capproxt_comp; apply (tcc_app _#6 cn_unit).
          #2 apply (capproxt_trans _ _ (fixnv (l-j1) u f)).
            apply capproxt_comp; apply tcc_lam; auto with static; [].
            apply weaken1_capproxt; [].
            eapply capproxt_comp; eapply tcc_equiv; eauto using eqc_symm; [].
            eapply fixn_mono; eauto; [].
            omega.
  
            apply capproxt_comp; apply (@tcc_equiv _#6 (cn_arrow (subst1 nonsense t') t)); eauto with static; [].
            exploit (Hap1 l) as H; [omega |].
            simp_sub in H.
            assumption.
  
          apply capproxt_comp; apply (tcc_equiv _#6 (subst1 nonsense t')); eauto; [].
          #2 apply (capproxt_trans _ _ (subst1 (fixnv (l-j2) u f) v2')).
            #2 eapply lower_approxt; eauto.
              simp_sub.
              rewrite <- Heqt'; [].
              assumption.
  
              omega.
              
            exploit (Hap2 l) as H; [omega |].
            assumption.

  (* regular app case *)
  simp_sub in Heq.
  injection Heq; clear Heq.
  intros -> ->.
  simp_sub in Hofv1'.
  rewrite <- Heqt' in Hofv1'.
  so (lam_oft_inversion_arrow _#5 Hwf Hofv1') as (_ & Hofe3').
  simp_sub in Hofv2'.
  rewrite <- Heqt' in Hofv2'.
  so (subst_oft _#5 (subof_dot dof (subof_id _ _) (id_thru _#4 (dof_tm Hofv2'))) Hofe3') as Hofe3'x.
  simp_sub in Hofe3'x.
  simp_sub in Hsteps3.
  replsub (subst (dot (subst1 (fixv u f) v2') (dot (fixv u f) id)) e3') with (subst1 (fixv u f) (subst1 v2' e3')) in Hsteps3.
  so (IHt _ Hlt3 _#3 Hofe3'x Hsteps3 Hval) as (j3 & v' & Hofv' & Hvalv' & Heq & Hap3).
  clear IHt Hsteps Hsteps1 Hsteps2.
  exists (j1 + j2 + j3), v'.
  do2 3 split; auto; [].
  intros l Hleq.
  simp_sub.
  rewrite -> !minus_plus_dist; [].
  exploit (Hap3 (l-j1-j2)) as Hap3'; [omega |].
  eapply capproxt_trans; eauto; [].
  #2 apply (anti_tsteps_capproxt_trans _ _ (subst1 (fixnv (l-j1-j2) u f) (tm_app (tm_lam t1' e3') v2'))).
    simp_sub.
    #2 eapply star_step.
      apply tstep_app3; eauto; [].
      eapply subst_tvalue; eauto; [].
      apply subof_dot; auto using subof_id; [].
      simp_sub.
      apply oof_utm; unfold fixnv; auto with dynamic; done.

      simp_sub.
      apply star_refl; done.

    replsub (subst1 (fixnv (l-j1-j2) u f) (tm_app (tm_lam t1' e3') v2')) with (tm_app (subst1 (fixnv (l-j1-j2) u f) (tm_lam t1' e3')) (subst1 (fixnv (l-j1-j2) u f) v2')).
    #2 apply capproxt_comp; apply (tcc_app _#6 (subst1 nonsense t')).
      exploit (Hap1 l); [omega |].
      eapply capproxt_trans; eauto; [].
      apply lower_approxt; auto; try omega; [].
      simp_sub.
      rewrite <- Heqt'; [].
      assumption.

      exploit (Hap2 l); [omega |].
      eapply capproxt_trans; eauto; [].
      apply lower_approxt; auto; try omega; [].
      simp_sub.
      rewrite <- Heqt'; [].
      assumption.

(* pair *)
intros G e1 e2 _ IH1 _ IH2 t v -> Hof Hsteps Hval.
so (merge_iht _#4 IHt IH1) as IH1'.
so (merge_iht _#4 IHt IH2) as IH2'.
clear IHt IHm IH1 IH2.
simp_sub in Hsteps.
so (pair_oft_inversion _#4 Hwf Hof) as (t1 & t2 & Heqc & Hof1 & Hof2).
so (subordf_tm_cn' _#3 (ofc_fcn _#3 (oft_reg _#3 Hwf Hof1))).
remember (subst1 nonsense t1) as t1' eqn:Heq; clear Heq; subst t1.
so (subordf_tm_cn' _#3 (ofc_fcn _#3 (oft_reg _#3 Hwf Hof2))).
remember (subst1 nonsense t2) as t2' eqn:Heq; clear Heq; subst t2.
rename t1' into t1; rename t2' into t2.
so (subst_eqc _#6 (unshift_tm _ _) Heqc) as Heqc'; simp_sub in Heqc'.
so (pair_halt_inversion _#4 Hsteps Hval) as (i1 & i2 & v1 & v2 & Hleq1 & Hleq2 & Hval1 & Hval2 & Hsteps1 & Hsteps2 & ->).
so (IH1' _#3 Hleq1 Hof1 Hsteps1 Hval1) as (j1 & v1' & Hofv1 & Hval1' & -> & Hap1).
so (IH2' _#3 Hleq2 Hof2 Hsteps2 Hval2) as (j2 & v2' & Hofv2 & Hval2' & -> & Hap2).
clear Hof Hsteps Hval Hval1 Hval2 Hsteps1 Hsteps2 IH1' IH2'.
exists (j1 + j2), (tm_pair v1' v2').
#3 do2 3 split; auto.
  eapply oft_equiv; eauto with static; done.

  eauto with dynamic; done.

  intros l Hleq.
  simp_sub.
  apply capproxt_comp; eapply tcc_equiv; eauto using eqc_symm; [].
  #2 apply capproxt_comp; apply tcc_pair.
    exploit (Hap1 l) as H; [omega |].
    eapply capproxt_trans; eauto; [].
    eapply lower_approxt; eauto; [].
    omega.

    exploit (Hap2 l) as H; [omega |].

    eapply capproxt_trans; eauto; [].
    eapply lower_approxt; eauto; [].
    omega.

(* pi1 *)
intros G e _ _ t v -> Hof Hsteps Hval.
clear IHm.
rename t into t1, v into v1, Hval into Hval1.
simp_sub in Hsteps.
so (pi1_oft_inversion _#3 Hwf Hof) as (t2s & Hof').
so (ofc_fcn _#3 (oft_reg _#3 Hwf Hof')) as H; invert H; []; clear H.
intros _ Ht2s.
so (subordf_tm_cn _#3 Ht2s) as Heqt2s.
rewrite -> Heqt2s in Hof'.
replsub (cn_prod (shift1 t1) (subst (dot nonsense sh1) t2s)) with (shift1 (cn_prod t1 (subst1 nonsense t2s))) in Hof'.
so (pi1_halt_inversion _#3 Hsteps Hval1) as (i' & v2 & Hlt & Hval2 & Hsteps').
so (IHt _ Hlt _#3 Hof' Hsteps' (tvalue_pair Hval1 Hval2)) as (j & v' & Hof'' & Hval' & Heq & Hap).
assert (exists v1' v2', v' = tm_pair v1' v2') as (v1' & v2' & ->).
  case_eq v'; try (intros; subst v'; simp_sub in Heq; discriminate Heq); eauto; [].
  intros n ->.
  destruct n; simp_sub in Heq; discriminate Heq.
simp_sub in Heq.
injection Heq; clear Heq.
intros -> ->.
simp_sub in Hof''.
so (pair_oft_inversion_prod _#5 Hwf Hof'') as (Hof1 & _).
invert Hval'; [].
intros Hval1' Hval2'.
exists j, v1'.
do2 3 split; auto; [].
intros l Hleq.
simp_sub.
#2 apply (anti_tsteps_capproxt_trans _ _ (tm_pi1 (subst1 (fixnv (l-j) u f) (tm_pair v1' v2')))).
  simp_sub.
  apply star_one; [].
  #2 apply tstep_pi12.
    eapply subst_tvalue; eauto; [].
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply oof_utm; apply tvalue_lam; done.

    eapply subst_tvalue; eauto; [].
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply oof_utm; apply tvalue_lam; done.

  apply capproxt_comp; apply (tcc_pi1 _#7 (subst1 nonsense t2s)); [].
  apply Hap; auto; done.

(* pi2 *)
intros G e _ _ t v -> Hof Hsteps Hval.
clear IHm.
rename t into t2, v into v2, Hval into Hval2.
simp_sub in Hsteps.
so (pi2_oft_inversion _#3 Hwf Hof) as (t1s & Hof').
so (ofc_fcn _#3 (oft_reg _#3 Hwf Hof')) as H; invert H; []; clear H.
intros Ht1s _.
so (subordf_tm_cn _#3 Ht1s) as Heqt1s.
rewrite -> Heqt1s in Hof'.
replsub (cn_prod (subst (dot nonsense sh1) t1s) (shift1 t2)) with (shift1 (cn_prod (subst1 nonsense t1s) t2)) in Hof'.
so (pi2_halt_inversion _#3 Hsteps Hval2) as (i' & v1 & Hlt & Hval1 & Hsteps').
so (IHt _ Hlt _#3 Hof' Hsteps' (tvalue_pair Hval1 Hval2)) as (j & v' & Hof'' & Hval' & Heq & Hap).
assert (exists v1' v2', v' = tm_pair v1' v2') as (v1' & v2' & ->).
  case_eq v'; try (intros; subst v'; simp_sub in Heq; discriminate Heq); eauto; [].
  intros n ->.
  destruct n; simp_sub in Heq; discriminate Heq.
simp_sub in Heq.
injection Heq; clear Heq.
intros -> ->.
simp_sub in Hof''.
so (pair_oft_inversion_prod _#5 Hwf Hof'') as (_ & Hof2).
invert Hval'; [].
intros Hval1' Hval2'.
exists j, v2'.
do2 3 split; auto; [].
intros l Hleq.
simp_sub.
#2 apply (anti_tsteps_capproxt_trans _ _ (tm_pi2 (subst1 (fixnv (l-j) u f) (tm_pair v1' v2')))).
  simp_sub.
  apply star_one; [].
  #2 apply tstep_pi22.
    eapply subst_tvalue; eauto; [].
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply oof_utm; apply tvalue_lam; done.

    eapply subst_tvalue; eauto; [].
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply oof_utm; apply tvalue_lam; done.

  apply capproxt_comp; apply (tcc_pi2 _#6 (subst1 nonsense t1s)); [].
  apply Hap; auto; done.

(* plam *)
intros; exfalso; eapply tstepn_value_contra; eauto; simp_sub; auto with dynamic; done.

(* papp *)
intros G e0 c _ _ Hc t v -> Hof Hsteps Hval.
clear IHm.
so (subordf_utm_cn' _ _ Hc) as Hisc.
remember (subst1 nonsense c) as x eqn:H; clear H; subst c; rename x into c.
so (papp_oft_inversion _#4 Hwf Hof) as (k & t' & Hof0 & Hofcs & Heqct).
so (subordf_tm_cn _#3 (ofc_fcn _#3 (oft_reg _#3 Hwf Hof0))) as Hisall.
simp_sub in Hisall; injection Hisall; clear Hisall.
intros Hist' Hisk.
replsub (subst (dot nonsense sh1) k) with (shift1 (subst1 nonsense k)) in Hisk.
remember (subst1 nonsense k) as x eqn:H; clear H; subst k; rename x into k.
replsub (subst (dot (var 0) (dot nonsense (sh 2))) t') with (subst (dot (var 0) (sh 2)) (subst (dot (var 0) (dot nonsense sh1)) t')) in Hist'.
remember (subst (dot (var 0) (dot nonsense sh1)) t') as x eqn:H; clear H; subst t'; rename x into t'.
replsub (cn_all (shift1 k) (subst (dot (var 0) (sh 2)) t')) with (shift1 (cn_all k t')) in Hof0.
so (subst_ofc _#5 (unshift_tm _ _) Hofcs) as Hofc; clear Hofcs; simp_sub in Hofc.
simp_sub in Hsteps.
so (papp_halt_inversion _#4 Hsteps Hval) as (i1 & i2 & k' & e1 & Hlt1 & Hlt2 & Hsteps1 & Hsteps2).
so (IHt _ Hlt1 _#3 Hof0 Hsteps1 tvalue_plam) as (j1 & v0' & Hofv0' & Hval0' & Heq & Hap1).
assert (exists k'' e2', v0' = tm_plam k'' e2') as (k'' & e1' & ->).
  case_eq v0'; try (intros; subst v0'; simp_sub in Heq; discriminate Heq); eauto; [].
  intros n ->.
  destruct n; simp_sub in Heq; discriminate Heq.
simp_sub in Heq.
injection Heq; clear Heq.
intros -> ->.
simp_sub in Hofv0'.
so (plam_oft_inversion_all _#5 Hwf Hofv0') as (Heqk & Hofe1').
simp_sub in Hsteps2.
replsub (subst (dot c (dot (fixv u f) id)) e1') with (subst1 (fixv u f) (subst1 (shift1 c) e1')) in Hsteps2.
#2 exploit (IHt _ Hlt2 (subst1 (shift1 c) e1') (subst1 c t') v) as (j2 & v' & Hofv' & Hval' & -> & Hap2); auto.
  so (subst1_oft_cn _#5 (weaken1_ofc _#4 Hofc) Hofe1') as H.
  simp_sub in H.
  simp_sub.
  assumption.
clear IHt.
exists (j1 + j2), v'.
#2 do2 3 split; auto.
  simp_sub in Heqct.
  simp_sub in Hofv'.
  eauto with static; done.

  intros l Hleq.
  so (subst_eqc _#6 (unshift_tm _ _) Heqct) as Heqctx; simp_sub in Heqctx.
  apply capproxt_comp; eapply tcc_equiv; eauto using eqc_symm; [].
  #2 apply (capproxt_trans _ _ (subst1 (fixnv (l-j1) u f) (subst1 (shift1 c) e1'))).
    rewrite -> minus_plus_dist; [].
    apply Hap2; omega.

    #2 apply (anti_tsteps_capproxt_trans _ _ (subst1 (fixnv (l-j1) u f) (tm_papp (tm_plam k'' e1') (shift1 c)))).
      #2 eapply star_step.
        simp_sub.
        eapply tstep_papp2; done.

        simp_sub.
        apply star_refl; done.

      simp_sub.
      apply capproxt_comp; apply (tcc_papp _#6 k t'); auto with static; [].
      exploit (Hap1 l) as H; [omega |].
      simp_sub in H.
      assumption.

(* pack *)
intros G k t c e Hc _ IH _ _ t' v -> Hof Hsteps Hval.
clear IHt IHm.
so (subordf_utm_cn' _ _ Hc) as Hisc.
remember (subst1 nonsense c) as x eqn:H; clear H; subst c; rename x into c.
so (pack_oft_inversion _#6 Hwf Hof) as (Heqct & Hofc & Hofe).
so (subordf_tm_cn _#3 (ofc_fcn _#3 (eqc_reg _#4 Hwf Heqct ander andel))) as Hisex.
simp_sub in Hisex; injection Hisex; clear Hisex.
intros Hist Hisk.
replsub (subst (dot nonsense sh1) k) with (shift1 (subst1 nonsense k)) in Hisk.
remember (subst1 nonsense k) as x eqn:H; clear H; subst k; rename x into k.
replsub (subst (dot (var 0) (dot nonsense (sh 2))) t) with (subst (dot (var 0) (sh 2)) (subst (dot (var 0) (dot nonsense sh1)) t)) in Hist.
remember (subst (dot (var 0) (dot nonsense sh1)) t) as x eqn:H; clear H; subst t; rename x into t.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqct ander andel)) as (z & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _; clear z Hofcr.
simp_sub in Hsteps.
so (pack_halt_inversion _#6 Hsteps Hval) as (v0 & Hval0 & Hsteps' & ->).
simp_sub in Hofe.
replsub (subst (dot (shift1 c) sh1) t) with (shift1 (subst1 c t)) in Hofe.
so (IH _ _ (eq_refl _) Hofe Hsteps' Hval0) as (j & v' & Hofv' & Hval' & -> & Hap).
exists j, (tm_pack (shift1 c) v' (shift1 k) (subst (dot (var 0) (sh 2)) t)).
do2 3 split.
  eapply oft_equiv; eauto using eqc_symm; [].
  apply oft_pack; auto; [].
  simp_sub.
  simp_sub in Hofv'.
  assumption.

  apply tvalue_pack; auto; done.

  simp_sub.
  reflexivity.

  intros l Hleq.
  simp_sub.
  so (subst_eqc _#6 (unshift_tm _ _) Heqct) as Heqctx; simp_sub in Heqctx.
  apply capproxt_comp; eapply tcc_equiv; eauto using eqc_symm; [].
  #3 apply capproxt_comp; eapply tcc_pack; eauto.
    apply eqc_refl; so (subst_ofc _#5 (unshift_tm _ _) Hofc) as H; simp_sub in H; assumption.

    apply eqk_refl; so (subst_ofk _#4 (unshift_tm _ _) Hofk) as H; simp_sub in H; assumption.
    
    #2 sor (refine (subst_ofc (nil; cl_cn k) _ (dot (var 0) (dot nonsense sh1)) _#3 Hoft)) as H.
      #2 apply subof_dot.
        apply subof_dot; eauto using subof_sh with sequence; [].
        simp_sub; apply sof_tm; done.

        simp_sub.
        apply sof_cn; eauto using ofc_var with sequence; done.
    simp_sub in H.
    apply eqc_refl; assumption.

(* unpack *)
intros G e1 e2 _ _ _ _ t v -> Hof Hsteps Hval.
clear IHm.
so (unpack_oft_inversion _#4 Hwf Hof) as (k & t1 & Hof1 & Hof2).
simp_sub in Hsteps.
so (unpack_halt_inversion _#4 Hsteps Hval) as (i1 & i2 & c & v3 & k' & t1' & Hlt1 & Hlt2 & Hval3 & Hsteps1 & Hsteps2).
so (subordf_tm_cn _#3 (ofc_fcn _#3 (oft_reg _#3 Hwf Hof1))) as Hisex.
simp_sub in Hisex; injection Hisex; clear Hisex.
intros Hist1 Hisk.
replsub (subst (dot nonsense sh1) k) with (shift1 (subst1 nonsense k)) in Hisk.
remember (subst1 nonsense k) as x eqn:H; clear H; subst k; rename x into k.
replsub (subst (dot (var 0) (dot nonsense (sh 2))) t1) with (subst (dot (var 0) (sh 2)) (subst (dot (var 0) (dot nonsense sh1)) t1)) in Hist1.
remember (subst (dot (var 0) (dot nonsense sh1)) t1) as x eqn:H; clear H; subst t1; rename x into t1.
replsub (cn_exists (shift1 k) (subst (dot (var 0) (sh 2)) t1)) with (shift1 (cn_exists k t1)) in Hof1.
so (IHt _ Hlt1 _#3 Hof1 Hsteps1 (tvalue_pack Hval3)) as (j1 & v1' & Hof1' & Hval1' & Heq & Hap1).
assert (exists c' v3' k' t1', v1' = tm_pack c' v3' k' t1') as (c' & v3' & k'' & t1'' & ->).
  case_eq v1'; try (intros; subst v1'; simp_sub in Heq; discriminate Heq); eauto; [].
  intros n ->.
  destruct n; simp_sub in Heq; discriminate Heq.
simp_sub in Heq.
injection Heq; clear Heq.
intros -> -> -> ->.
simp_sub in Hof1'.
so (pack_oft_inversion_exists _#7 Hwf Hof1') as (Heqk & Heqt1 & Hofc' & Hofv3').
so (subord_tm_cn _#4 Hofc') as Hisc'.
replsub (subst (dot nonsense sh1) c') with (shift1 (subst1 nonsense c')) in Hisc'.
remember (subst1 nonsense c') as x eqn:H; clear H; subst c'; rename x into c.
so (subord_tm_cn _#4 (eqc_reg _#4 Hwf (eqc_exists (eqk_symm Heqk) (eqc_symm Heqt1)) ander andel)) as Hisex.
simp_sub in Hisex.
injection Hisex; clear Hisex.
intros Hist1'' Hisk''.
replsub (subst (dot nonsense sh1) k'') with (shift1 (subst1 nonsense k'')) in Hisk''.
remember (subst1 nonsense k'') as x eqn:H; clear H; subst k''; rename x into k'.
replsub (subst (dot (var 0) (dot nonsense (sh 2))) t1'') with (subst (dot (var 0) (sh 2)) (subst (dot (var 0) (dot nonsense sh1)) t1'')) in Hist1''.
remember (subst (dot (var 0) (dot nonsense sh1)) t1'') as x eqn:H; clear H; subst t1''; rename x into t1'.
simp_sub in Hsteps2.
#3 exploit (IHt _ Hlt2 (subst (dot v3' (dot (shift1 c) id)) e2) t v) as (j2 & v' & Hofv' & Hval' & -> & Hap2); auto.
  simp_sub in Hof2.
  replsub (shift1 t) with (subst (dot v3' (dot (shift1 c) id)) (subst (sh 3) t)).
  eapply subst_oft; eauto; [].
  #2 apply subof_dot.
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply dof_cn; auto; done.

    simp_sub.
    simp_sub in Hofv3'.
    apply dof_tm; auto; done.

  simp_sub.
  assumption.
clear IHt Hsteps1 Hsteps2.
exists (j1 + j2), v'.
do2 3 split; auto; [].
intros l Hleq.
rewrite -> minus_plus_dist; [].
exploit (Hap2 (l-j1)) as Hap2'; [omega |]; clear Hap2.
eapply capproxt_trans; eauto; [].
#2 apply (anti_tsteps_capproxt_trans _ _ (subst1 (fixnv (l-j1) u f) (tm_unpack (tm_pack (shift1 c) v3' (shift1 k') (subst (dot (var 0) (sh 2)) t1')) e2))).
  #2 eapply star_step.
    simp_sub.
    eapply tstep_unpack2; eauto; [].
    so (unsubst_tvalue _#3 Hval3) as Hval3'.
    eapply subst_tvalue; eauto; [].
    apply subof_dot; auto using subof_id; [].
    simp_sub; apply oof_utm; [].
    exact tvalue_lam.

    simp_sub.
    apply star_refl; done.

  simp_sub.
  #3 apply capproxt_comp; eapply tcc_unpack.
    exploit (Hap1 l) as Hap1'; [omega |].
    simp_sub in Hap1'.
    eexact Hap1'.

    replsub (subst (dot (var 0) (dot (var 1) (dot (subst (sh 2) (fixnv (l-j1) u f)) (sh 2)))) e2) with (subst (under 2 (dot (fixnv (l-j1) u f) id)) e2).
    replsub (subst (dot (var 0) (dot (var 1) (dot (subst (sh 2) (fixnv l u f)) (sh 2)))) e2) with (subst (under 2 (dot (fixnv l u f) id)) e2).
    #3 apply lower_approxt_gen; auto.
      so (ofc_ofcr _#3 (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 Hwf Hof1))) as (kk & Hofcr); simp_sub in Hofcr.
      invert Hofcr; [].
      intros Hofk Hoft1 _; clear kk Hofcr.
      auto using wf; done.

      simpl app.
      simpl length.
      rewrite <- subst_compose; [].
      replace (under 2 sh1) with (under (2+0) sh1) by (f_equal; omega).
      rewrite -> compose_sh_under; [].
      simp_sub.
      simp_sub in Hof2.
      assumption.

      omega.

    so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 Hwf Hof)) as Hoft; simp_sub in Hoft.
    assumption.

(* fix *)
intros G t' e0 Ht' _ _ t v -> Hof Hsteps Hval.
clear IHm.
so (subordf_utm_cn' _ _ Ht') as Hist'.
remember (subst1 nonsense t') as x eqn:H; clear H; subst t'; rename x into t'.
so (fix_oft_inversion _#4 Hwf Hof) as (Heqct & Hofe0).
so (subst_ofc _#5 (unshift_tm _ _) (eqc_reg _#4 Hwf Heqct ander andel)) as Hoft; simp_sub in Hoft.
simp_sub in Hsteps.
repl i with (S (i-1)) in Hsteps by omega.
invert Hsteps; clear Hsteps; [].
intros e1 Hstep Hsteps.
so (tstep_fun _#4 Hstep tstep_fix); subst e1; clear Hstep.
replsub (tm_app (subst1 (fixv u f) e0) (tm_lam cn_unit (shift1 (tm_fix t' (subst1 (fixv u f) e0))))) with (subst1 (fixv u f) (tm_app e0 (tm_lam cn_unit (shift1 (tm_fix (shift1 t') e0))))) in Hsteps.
so (oft_app Hofe0 (oft_lam ofc_unit (subst_oft _#5 (weaken1 _#3) (oft_fix Hofe0)))) as Hoffix.
sor (refine (IHt _#5 Hoffix Hsteps Hval)) as (j & v' & Hofv' & Hval' & -> & Hap); [omega |].
exists j, v'.
#2 do2 3 split; auto.
  eauto with static; done.

  intros l Hleq.
  #2 apply capproxt_comp; apply (@tcc_equiv _#6 t').
    #2 apply (capproxt_trans _ _ (subst1 (fixnv l u f) (tm_app e0 (tm_lam cn_unit (shift1 (tm_fix (shift1 t') e0)))))).
      eapply Hap; eauto; done.

      #2 apply anti_tsteps_capproxt.
        so (subst1_oft_tm _#5 (oft_fixnv _ l _ _ Hofu Hoff) (oft_fix Hofe0)) as H; simp_sub in H.
        simp_sub.
        assumption.

        simp_sub.
        #2 eapply star_step.
          apply tstep_fix; done.
    
          simp_sub.
          apply star_refl; done.

    so (subst_eqc _#6 (unshift_tm _ _) Heqct) as Heqctx; simp_sub in Heqctx.
    apply eqc_symm; auto; done.

(* lett *)
intros G e1 e2 _ _ _ _ t v -> Hof Hsteps Hval.
clear IHm.
so (lett_oft_inversion _#4 Hwf Hof) as (t' & Hofe1 & Hofe2).
so (subordf_tm_cn' _#3 (ofc_fcn _#3 (oft_reg _#3 Hwf Hofe1))) as His.
remember (subst1 nonsense t') as x eqn:H; clear H; subst t'; rename x into t'.
simp_sub in Hsteps.
so (lett_halt_inversion _#4 Hsteps Hval) as (i1 & i2 & v1 & Hlt1 & Hlt2 & Hval1 & Hsteps1 & Hsteps2).
simp_sub in Hsteps2.
so (IHt _ Hlt1 _#3 Hofe1 Hsteps1 Hval1) as (j1 & v1' & Hofv1' & Hval1' & -> & Hap1).
replsub (subst (dot (subst1 (fixv u f) v1') (dot (fixv u f) id)) e2) with (subst1 (fixv u f) (subst1 v1' e2)) in Hsteps2.
so (subst1_oft_tm _#5 Hofv1' Hofe2) as Hofe2x; simp_sub in Hofe2x.
so (IHt _ Hlt2 _#3 Hofe2x Hsteps2 Hval) as (j2 & v' & Hofv' & Hval' & -> & Hap2).
exists (j1 + j2), v'.
do2 3 split; auto; [].
intros l Hleq.
rewrite -> minus_plus_dist; [].
exploit (Hap2 (l-j1)) as Hap2'; [omega |].
eapply capproxt_trans; eauto; [].
#2 apply (anti_tsteps_capproxt_trans _ _ (subst1 (fixnv (l-j1) u f) (tm_lett v1' e2))).
  simp_sub.
  #2 eapply star_step.
    apply tstep_lett2; [].
    eapply subst_tvalue; eauto; [].
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply oof_utm; apply tvalue_lam; done.

    simp_sub.
    apply star_refl; done.

  simp_sub.
  #2 apply capproxt_comp; apply (tcc_lett _#6 t').
    eapply Hap1; eauto; omega.

    replsub (subst (dot (var 0) (dot (shift1 (fixnv (l-j1) u f)) sh1)) e2) with (subst (under 1 (dot (fixnv (l-j1) u f) id)) e2).
    replsub (subst (dot (var 0) (dot (shift1 (fixnv l u f)) sh1)) e2) with (subst (under 1 (dot (fixnv l u f) id)) e2).
    #3 apply lower_approxt_gen; auto.
      apply wf_tm; auto using wf_nil; [].
      so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 Hwf Hofe1)) as H; simp_sub in H.
      assumption.

      simpl.
      simp_sub.
      simp_sub in Hofe2.
      assumption.

      omega.

(* letm *)
intros G M e _ _ _ _ t v -> Hof Hsteps Hval.
so (letm_oft_inversion _#4 Hwf Hof) as (T & k & HofM & Hfsts & Hofe).
so (subordf_tm_sg' _#3 (ofs_fsg _ _ (ofm_reg _#4 Hwf HofM))).
forget (subst1 nonsense T) as x; subst T; rename x into T.
so (subordf_tm_kd' _#3 (ofk_fkd _ _ (ofs_fst _#3 (ofm_reg _#4 Hwf HofM) Hfsts))).
forget (subst1 nonsense k) as x; subst k; rename x into k.
so (subst_fsts _ _ (dot nonsense id) Hfsts) as Hfsts'; simp_sub in Hfsts'; renameover Hfsts' into Hfsts.
simp_sub in Hsteps.
so (letm_halt_inversion _#4 Hsteps Hval) as (i1 & i2 & V & c & Hlt1 & Hlt2 & HvalV & HfstV & Hsteps1 & Hsteps2).
clear Hof Hsteps.
so (IHm _ Hlt1 _#4 HofM Hsteps1 HvalV) as (j1 & V' & HofV' & HvalV' & -> & Hap1).
clear IHm Hlt1 Hsteps1 HvalV.
simp_sub in Hsteps2.
so (subst_mvalue _#4 (subof_tm_utm_oof _ (cn_arrow cn_unit u)) HvalV') as HvalV''; simp_sub in HvalV''.
so (mvalue_pure _#4 HvalV'' HofV') as H; renameover H into HofV'.
so (unsubst_fstm _#6 wf_nil (oft_fixv _#3 Hofu Hoff) HofV' HfstV) as HfstV'.
replsub (subst (dot (subst1 (fixv u f) V') (dot c (dot (fixv u f) id))) e) with (subst1 (fixv u f) (subst (dot V' (dot (shift1 c) id)) e)) in Hsteps2.
#2 exploit (IHt _ Hlt2 (subst (dot V' (dot (shift1 c) id)) e) t v) as H; auto; clear IHt.
  simp_sub in Hofe.
  replsub (shift1 t) with (subst (dot V' (dot (shift1 c) id)) (subst (sh 3) t)).
  eapply subst_oft; eauto; [].
  #2 apply subof_dot.
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply dof_cn; [].
    so (ofm_fst _#6 Hwf HofV' HfstV' (subst_fsts _ _ sh1 Hfsts)).
    assumption.

    simp_sub.
    apply dof_md; auto; done.

  clear Hsteps2.
  destruct H as (j2 & v' & Hofv' & Hvalv' & -> & Hap2).
  exists (j1 + j2), v'.
  do2 3 split; auto; [].
  intros l Hleq.
  rewrite -> minus_plus_dist; [].
  exploit (Hap2 (l-j1)) as Hap2'; [omega |]; clear Hap2.
  eapply capproxt_trans; eauto; [].
  #2 apply (anti_tsteps_capproxt_trans _ _ (tm_letm (subst1 (fixnv (l-j1) u f) V') (subst (under 2 (dot (fixnv (l-j1) u f) id)) e))).
    simp_sub.
    #2 eapply star_step.
      #2 eapply (@tstep_letm2 _#3 c); eauto.
        eapply subst_mvalue; eauto; [].
        apply subof_dot; auto using subof_id; [].
        simp_sub.
        apply oof_tm; exact tvalue_lam.

        replsub c with (subst1 (fixnv (l-j1) u f) (shift1 c)).
        eapply subst_fstm; eauto; [].
        apply subof_dot; auto using subof_id; [].
        simp_sub.
        apply tof_tm; done.

      simp_sub.
      apply star_refl; done.

    simp_sub.
    #3 apply capproxt_comp; apply (tcc_letm _#6 T k); auto.
      eapply Hap1; eauto; omega.

      simp_sub.
      replsub (dot (var 0) (dot (var 1) (dot (subst (sh 2) (fixnv (l-j1) u f)) (sh 2)))) with (under 2 (dot (fixnv (l-j1) u f) id)).
      replsub (dot (var 0) (dot (var 1) (dot (subst (sh 2) (fixnv l u f)) (sh 2)))) with (under 2 (dot (fixnv l u f) id)).
      #3 apply lower_approxt_gen; auto.
        apply wf_md; auto using wf; [].
        so (subst_ofs _#4 (unshift_tm _ _) (ofm_reg _#4 Hwf HofM)) as H.
        simp_sub in H.
        assumption.

        simpl.
        simp_sub.
        simp_sub in Hofe.
        assumption.

        omega.

      so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 Hwf Hofv')) as H.
      simp_sub in H.
      assumption.

(* ext *)
intros G M _ _ t v -> Hof Hsteps Hval.
clear IHt.
so (ext_oft_inversion _#3 Hwf Hof) as HofM.
simp_sub in Hsteps.
so (ext_halt_inversion _#3 Hsteps Hval) as (i' & Hlt & HstepsM).
clear Hof Hsteps.
replsub (sg_dyn (shift1 t)) with (shift1 (sg_dyn t)) in HofM.
so (IHm _ Hlt _#4 HofM HstepsM (mvalue_dyn Hval)) as (j & V' & HofV' & HvalV' & Heq & Hap).
assert (exists v', V' = md_dyn v') as (v' & ->).
  case_eq V'; try (intros; subst V'; simp_sub in Heq; discriminate Heq); eauto; [].
  intros n ->.
  destruct n; simp_sub in Heq; discriminate Heq.
simp_sub in Heq.
injection Heq.
intros ->.
invert HvalV'; [].
intros Hvalv'.
simp_sub in HofV'.
so (dyn_ofm_inversion_dyn _#4 Hwf HofV') as Hofv'.
exists j, v'.
do2 3 split; auto; [].
intros l Hleq.
#2 apply (anti_tsteps_capproxt_trans _ _ (subst1 (fixnv (l-j) u f) (tm_ext (md_dyn v')))).
  simp_sub.
  apply star_one; [].
  apply tstep_ext2; [].
  eapply subst_tvalue; eauto; [].
  apply subof_dot; auto using subof_id; [].
  simp_sub.
  apply oof_utm; [].
  exact tvalue_lam.

  simp_sub.
  apply capproxt_comp; apply tcc_ext; [].
  so (Hap _ Hleq) as H.
  simp_sub in H.
  assumption.

(* var *)
intros G j T' c Hindex T V p -> _ _ _.
exfalso.
#2 invert Hindex.
  intros _ Heq.
  discriminate Heq.

  intros j' Hindex' _.
  invert Hindex'; done.

(* varu *)
intros G j c Hindex T V p -> _ _ _.
exfalso.
#2 invert Hindex.
  intros _ Heq.
  discriminate Heq.

  intros j' Hindex' _.
  invert Hindex'; done.

(* star *)
intros; exfalso; eapply mstepn_value_contra; eauto; simp_sub; auto with dynamic; done.

(* stat *)
intros; exfalso; eapply mstepn_value_contra; eauto; simp_sub; auto with dynamic; done.

(* dyn *)
intros G e _ IH T V p -> Hof Hsteps Hval.
clear IHt IHm.
so (dyn_ofm_inversion _#4 Hwf Hof) as (t & Heqs & Hofe).
so (subordf_tm_cn' _ _ _ (ofc_fcn _#3 (oft_reg _#3 Hwf Hofe))).
forget (subst1 nonsense t) as x; subst t; rename x into t.
simp_sub in Hsteps.
so (dyn_halt_inversion _#3 Hsteps Hval) as (v & Hsteps' & ->).
invert Hval; [].
intros Hvalv.
so (IH _ _ (eq_refl _) Hofe Hsteps' Hvalv) as (j & v' & Hofv' & Hvalv' & -> & Hap).
exists j, (md_dyn v').
do2 3 split; auto.
  rewrite <- Heqs; [].
  apply ofm_forget_pure; [].
  apply ofm_dyn; auto with dynamic; done.

  apply mvalue_dyn; auto; done.

  intros l Hleq.
  simp_sub.
  f_apply (subst1 nonsense) in Heqs as Heqs'.
  simp_sub in Heqs'.
  rewrite <- Heqs'; [].
  apply capproxm_forget_pure; apply capproxm_comp; apply mcc_dyn; auto; done.

(* lamgn *)
intros; exfalso; eapply mstepn_value_contra; eauto; simp_sub; auto with dynamic; done.

(* appgn *)
intros G M1 M2 _ _ _ _ T V p -> Hof Hsteps Hval.
clear IHt.
so (appgn_ofm_inversion _#5 Hwf Hof) as (T1 & T2 & c & HofM1 & HofM2 & HfstM2 & Hsubs2 & ->).
clear Hof.
so (subordf_tm_sg _#3 (ofs_fsg
 _ _ (ofm_reg _#4 Hwf HofM1))) as Hispi.
simp_sub in Hispi.
injection Hispi; clear Hispi; intros HisT2 HisT1.
replsub (subst (dot nonsense sh1) T1) with (shift1 (subst1 nonsense T1)) in HisT1.
forget (subst1 nonsense T1) as x; subst T1; rename x into T1.
replsub (subst (dot (var 0) (dot nonsense (sh 2))) T2) with (subst (dot (var 0) (sh 2)) (subst (dot (var 0) (dot nonsense sh1)) T2)) in HisT2.
forget (subst (dot (var 0) (dot nonsense sh1)) T2) as x; subst T2; rename x into T2.
so (subordf_tm_cn' _#3 (fstm_fcn _#3 (wf_fcontext _ Hwf) (ofm_fmd _#4 (wf_fcontext _ Hwf) HofM2) HfstM2)) as H.
forget (subst1 nonsense c) as x; subst c; rename x into c.
simp_sub in Hsteps.
so (appgn_halt_inversion _#4 Hsteps Hval) as (i1 & i2 & i3 & T1' & M3 & V2 & d & Hlt1 & Hlt2 & Hlt3 & Hval2 & Hsteps1 & Hsteps2 & HfstV2 & Hsteps3).
clear Hsteps.
replsub (sg_pigen (shift1 T1) (subst (dot (var 0) (sh 2)) T2)) with (shift1 (sg_pigen T1 T2)) in HofM1.
so (IHm _ Hlt1 _#4 HofM1 Hsteps1 mvalue_lamgn) as (j1 & V1' & HofV1' & HvalV1' & Heq & Hap1).
clear HofM1 Hsteps1 Hlt1.
assert (exists T1'' M3', V1' = md_lamgn T1'' M3') as (T1'' & M3' & ->).
  case_eq V1'; try (intros; subst V1'; simp_sub in Heq; discriminate Heq); eauto; [].
  intros n ->.
  destruct n; simp_sub in Heq; discriminate Heq.
simp_sub in Heq; injection Heq; clear Heq; intros -> ->.
so (IHm _ Hlt2 _#4 HofM2 Hsteps2 Hval2) as (j2 & V2' & HofV2' & HvalV2' & -> & Hap2).
clear Hval2 Hlt2.
so (subst_mvalue _#4 (subof_tm_utm_oof _ (cn_arrow cn_unit u)) HvalV2') as HvalV2''; simp_sub in HvalV2''.
so (unsubst_fstm _#6 wf_nil (oft_fixv _#3 Hofu Hoff) HofV2' HfstV2) as HfstV2'.
clear HfstV2.
simp_sub in Hsteps3.
replsub (subst (dot (subst1 (fixv u f) V2') (dot d (dot (fixv u f) id))) M3') with (subst1 (fixv u f) (subst (dot V2' (dot (shift1 d) id)) M3')) in Hsteps3.
simp_sub in HofV1'.
so (lamgn_ofm_inversion_pigen _#6 Hwf HofV1') as (k1w & Hsubs1 & HfstT1w & HofM3').
so (ofm_reg _#4 Hwf HofV1') as Hofpi.
invert Hofpi; clear Hofpi; [].
intros k1w' HofT1w HfstT1w' HofT2w.
so (fsts_fun _#3 HfstT1w HfstT1w'); subst k1w'; clear HfstT1w'.
so (subst_ofs _#4 (unshift_tm _ _) HofT1w) as HofT1; simp_sub in HofT1; clear HofT1w.
so (can_fsts _ _ HofT1) as (k1 & HfstT1).
so (subst_fsts _ _ sh1 HfstT1) as H; simp_sub in H; so (fsts_fun _#3 H HfstT1w); subst k1w; clear H HfstT1w.
#2 sor (refine (subst_ofs (nil; cl_cn k1) _ (dot (var 0) (dot nonsense sh1)) _ _ HofT2w)) as HofT2.
  #2 apply subof_dot.
    apply subof_dot; eauto using subof_sh with sequence; [].
    simp_sub.
    apply sof_tm; done.

    simp_sub.
    apply sof_cn; apply ofc_var; auto with sequence; done.
simp_sub in HofT2; clear HofT2w.
so (ofm_fst _#6 Hwf HofV2' HfstV2' (subst_fsts _ _ sh1 HfstT1)) as Hofdw.
#2 sor (refine (subst_ofm (nil; cl_tm (cn_arrow cn_unit u)) _ (dot V2' (dot (shift1 d) id)) _#4 HofM3')) as HofM3'x.
  #2 apply subof_dot.
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply dof_cn; auto; done.

    simp_sub.
    apply dof_md; auto; done.
clear HofM3'; simp_sub in HofM3'x.
replsub (subst (dot (shift1 d) sh1) T2) with (shift1 (subst1 d T2)) in HofM3'x.
so (IHm _ Hlt3 _#4 HofM3'x Hsteps3 Hval) as (j3 & V' & HofV' & HvalV' & -> & Hap3).
clear Hsteps3 HofM3'x Hlt3 IHm.
assert (eqc nil c d k1) as Heqcd.
  #6 apply (preservations_fst' nil pure (subst1 (fixv u f) M2) (subst1 (fixv u f) V2') T1 c d k1).
    exact wf_nil.

    replsub T1 with (subst1 (fixv u f) (shift1 T1)).
    eapply subst1_ofm_tm; eauto using oft_fixv; done.

    eapply expon_star; eauto; done.

    assumption.

    replsub c with (subst1 (fixv u f) (shift1 c)).
    eapply subst1_fstm_tm; eauto; done.

    replsub d with (subst1 (fixv u f) (shift1 d)).
    eapply subst1_fstm_tm; eauto; done.
assert (eqs nil (subst1 c T2) (subst1 d T2)) as HeqT2x.
  apply (funct_sg _ k1); auto; [].
  exact wf_nil.
exists (j1 + j2 + j3), V'.
#2 do2 3 split; auto.
  eapply ofm_subsume; eauto; [].
  simp_sub in Hsubs2.
  eapply subs_trans; eauto; [].
  apply subs_refl; [].
  simp_sub.
  apply eqs_symm; [].
  so (weaken1_eqs _ (cl_tm (cn_arrow cn_unit u)) _ _ HeqT2x) as H.
  simp_sub in H.
  assumption.

  intros l Hleq.
  simp_sub.
  #2 apply capproxm_comp; apply (@mcc_subsume _#7 (subst1 d T2)).
    rewrite -> !minus_plus_dist; [].
    #2 apply (capproxm_trans _#3 (subst1 (fixnv (l-j1-j2) u f) (subst (dot V2' (dot (shift1 d) id)) M3'))).
      apply Hap3; auto; omega.

      #2 apply (anti_msteps_capproxm_trans _#3 (subst1 (fixnv (l-j1-j2) u f) (md_appgn (md_lamgn T1'' M3') V2'))).
        simp_sub.
        #2 eapply star_step.
          #2 apply (@mstep_appgn3 _#4 d).
            eapply subst_mvalue; eauto; [].
            apply subof_dot; auto using subof_id; [].
            simp_sub.
            apply oof_tm; [].
            apply tvalue_lam; done.

            so (subst1_fstm_tm _ (fixnv (l-j1-j2) u f) _#3 HfstV2') as H.
            simp_sub in H.
            assumption.

          simp_sub.
          apply star_refl; done.

        simp_sub.
        #3 apply capproxm_comp; apply (@mcc_appgn _#6 T1 T2 _#4 d).
          simp_sub.
          #2 apply (capproxm_trans _#3 (subst1 (fixnv (l-j1) u f) (md_lamgn T1'' M3'))).
            replsub (md_lamgn (subst1 (fixnv (l-j1-j2) u f) T1'') (subst (dot (var 0) (dot (var 1) (dot (subst (sh 2) (fixnv (l - j1 - j2) u f)) (sh 2)))) M3')) with (subst1 (fixnv (l-j1-j2) u f) (md_lamgn T1'' M3')).
            apply lower_approxm; auto; [].
            omega.

            apply Hap1; omega.

          #2 apply (capproxm_trans _#3 (subst1 (fixnv (l-j2) u f) V2')).
            apply lower_approxm; auto; omega.

            apply Hap2; omega.

          so (subst1_fstm_tm _ (fixnv (l-j1-j2) u f) _#3 HfstV2') as H.
          simp_sub in H.
          assumption.

    #2 apply (@subs_trans _ _ (subst1 c T2)).
      apply subs_refl; [].
      apply eqs_symm; assumption.

      so (subst_subs _#5 (unshift_tm _ _) Hsubs2) as Hsubs2'.
      simp_sub in Hsubs2'.
      assumption.

(* lamap *)
intros; exfalso; eapply mstepn_value_contra; eauto; simp_sub; auto with dynamic; done.

(* appap *)
intros G M1 M2 _ _ _ _ T V p -> Hof Hsteps Hval.
clear IHt.
so (appap_ofm_inversion _#5 Hwf Hof) as (T1 & T2 & c & HofM1 & HofM2 & HfstM2 & Hsubs2).
clear Hof.
so (subordf_tm_sg _#3 (ofs_fsg
 _ _ (ofm_reg _#4 Hwf HofM1))) as Hispi.
simp_sub in Hispi.
injection Hispi; clear Hispi; intros HisT2 HisT1.
replsub (subst (dot nonsense sh1) T1) with (shift1 (subst1 nonsense T1)) in HisT1.
forget (subst1 nonsense T1) as x; subst T1; rename x into T1.
replsub (subst (dot (var 0) (dot nonsense (sh 2))) T2) with (subst (dot (var 0) (sh 2)) (subst (dot (var 0) (dot nonsense sh1)) T2)) in HisT2.
forget (subst (dot (var 0) (dot nonsense sh1)) T2) as x; subst T2; rename x into T2.
so (subordf_tm_cn' _#3 (fstm_fcn _#3 (wf_fcontext _ Hwf) (ofm_fmd _#4 (wf_fcontext _ Hwf) HofM2) HfstM2)) as H.
forget (subst1 nonsense c) as x; subst c; rename x into c.
simp_sub in Hsteps.
so (appap_halt_inversion _#4 Hsteps Hval) as (i1 & i2 & i3 & T1' & M3 & V2 & d & Hlt1 & Hlt2 & Hlt3 & Hval2 & Hsteps1 & Hsteps2 & HfstV2 & Hsteps3).
clear Hsteps.
replsub (sg_piapp (shift1 T1) (subst (dot (var 0) (sh 2)) T2)) with (shift1 (sg_piapp T1 T2)) in HofM1.
so (IHm _ Hlt1 _#4 HofM1 Hsteps1 mvalue_lamap) as (j1 & V1' & HofV1' & HvalV1' & Heq & Hap1).
clear HofM1 Hsteps1 Hlt1.
assert (exists T1'' M3', V1' = md_lamap T1'' M3') as (T1'' & M3' & ->).
  case_eq V1'; try (intros; subst V1'; simp_sub in Heq; discriminate Heq); eauto; [].
  intros n ->.
  destruct n; simp_sub in Heq; discriminate Heq.
simp_sub in Heq; injection Heq; clear Heq; intros -> ->.
so (IHm _ Hlt2 _#4 HofM2 Hsteps2 Hval2) as (j2 & V2' & HofV2' & HvalV2' & -> & Hap2).
clear Hval2 Hlt2.
so (subst_mvalue _#4 (subof_tm_utm_oof _ (cn_arrow cn_unit u)) HvalV2') as HvalV2''; simp_sub in HvalV2''.
so (unsubst_fstm _#6 wf_nil (oft_fixv _#3 Hofu Hoff) HofV2' HfstV2) as HfstV2'.
clear HfstV2.
simp_sub in Hsteps3.
replsub (subst (dot (subst1 (fixv u f) V2') (dot d (dot (fixv u f) id))) M3') with (subst1 (fixv u f) (subst (dot V2' (dot (shift1 d) id)) M3')) in Hsteps3.
simp_sub in HofV1'.
so (lamap_ofm_inversion_piapp _#6 Hwf HofV1') as (k1w & Hsubs1 & HfstT1w & HofM3').
so (ofm_reg _#4 Hwf HofV1') as Hofpi.
invert Hofpi; clear Hofpi; [].
intros k1w' HofT1w HfstT1w' HofT2w.
so (fsts_fun _#3 HfstT1w HfstT1w'); subst k1w'; clear HfstT1w'.
so (subst_ofs _#4 (unshift_tm _ _) HofT1w) as HofT1; simp_sub in HofT1; clear HofT1w.
so (can_fsts _ _ HofT1) as (k1 & HfstT1).
so (subst_fsts _ _ sh1 HfstT1) as H; simp_sub in H; so (fsts_fun _#3 H HfstT1w); subst k1w; clear H HfstT1w.
#2 sor (refine (subst_ofs (nil; cl_cn k1) _ (dot (var 0) (dot nonsense sh1)) _ _ HofT2w)) as HofT2.
  #2 apply subof_dot.
    apply subof_dot; eauto using subof_sh with sequence; [].
    simp_sub.
    apply sof_tm; done.

    simp_sub.
    apply sof_cn; apply ofc_var; auto with sequence; done.
simp_sub in HofT2; clear HofT2w.
so (ofm_fst _#6 Hwf HofV2' HfstV2' (subst_fsts _ _ sh1 HfstT1)) as Hofdw.
#2 sor (refine (subst_ofm (nil; cl_tm (cn_arrow cn_unit u)) _ (dot V2' (dot (shift1 d) id)) _#4 HofM3')) as HofM3'x.
  #2 apply subof_dot.
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply dof_cn; auto; done.

    simp_sub.
    apply dof_md; auto; done.
clear HofM3'; simp_sub in HofM3'x.
replsub (subst (dot (shift1 d) sh1) T2) with (shift1 (subst1 d T2)) in HofM3'x.
so (IHm _ Hlt3 _#4 HofM3'x Hsteps3 Hval) as (j3 & V' & HofV' & HvalV' & -> & Hap3).
clear Hsteps3 HofM3'x Hlt3 IHm.
assert (eqc nil c d k1) as Heqcd.
  #6 apply (preservations_fst' nil pure (subst1 (fixv u f) M2) (subst1 (fixv u f) V2') T1 c d k1).
    exact wf_nil.

    replsub T1 with (subst1 (fixv u f) (shift1 T1)).
    eapply subst1_ofm_tm; eauto using oft_fixv; done.

    eapply expon_star; eauto; done.

    assumption.

    replsub c with (subst1 (fixv u f) (shift1 c)).
    eapply subst1_fstm_tm; eauto; done.

    replsub d with (subst1 (fixv u f) (shift1 d)).
    eapply subst1_fstm_tm; eauto; done.
assert (eqs nil (subst1 c T2) (subst1 d T2)) as HeqT2x.
  apply (funct_sg _ k1); auto; [].
  exact wf_nil.
exists (j1 + j2 + j3), V'.
#2 do2 3 split; auto.
  apply ofm_forget_pure; [].
  eapply ofm_subsume; eauto; [].
  simp_sub in Hsubs2.
  eapply subs_trans; eauto; [].
  apply subs_refl; [].
  simp_sub.
  apply eqs_symm; [].
  so (weaken1_eqs _ (cl_tm (cn_arrow cn_unit u)) _ _ HeqT2x) as H.
  simp_sub in H.
  assumption.

  intros l Hleq.
  simp_sub.
  #2 apply capproxm_comp; apply (@mcc_subsume _#7 (subst1 d T2)).
    rewrite -> !minus_plus_dist; [].
    #2 apply (capproxm_trans _#3 (subst1 (fixnv (l-j1-j2) u f) (subst (dot V2' (dot (shift1 d) id)) M3'))).
      apply capproxm_forget_pure; [].
      apply Hap3; auto; omega.

      #2 apply (anti_msteps_capproxm_trans _#3 (subst1 (fixnv (l-j1-j2) u f) (md_appap (md_lamap T1'' M3') V2'))).
        simp_sub.
        #2 eapply star_step.
          #2 apply (@mstep_appap3 _#4 d).
            eapply subst_mvalue; eauto; [].
            apply subof_dot; auto using subof_id; [].
            simp_sub.
            apply oof_tm; [].
            apply tvalue_lam; done.

            so (subst1_fstm_tm _ (fixnv (l-j1-j2) u f) _#3 HfstV2') as H.
            simp_sub in H.
            assumption.

          simp_sub.
          apply star_refl; done.

        simp_sub.
        #3 apply capproxm_comp; apply (@mcc_appap _#7 T1 T2 _#4 d).
          simp_sub.
          #2 apply (capproxm_trans _#3 (subst1 (fixnv (l-j1) u f) (md_lamap T1'' M3'))).
            replsub (md_lamap (subst1 (fixnv (l-j1-j2) u f) T1'') (subst (dot (var 0) (dot (var 1) (dot (subst (sh 2) (fixnv (l - j1 - j2) u f)) (sh 2)))) M3')) with (subst1 (fixnv (l-j1-j2) u f) (md_lamap T1'' M3')).
            apply lower_approxm; auto; [].
            omega.

            apply Hap1; omega.

          #2 apply (capproxm_trans _#3 (subst1 (fixnv (l-j2) u f) V2')).
            apply lower_approxm; auto; omega.

            apply Hap2; omega.

          so (subst1_fstm_tm _ (fixnv (l-j1-j2) u f) _#3 HfstV2') as H.
          simp_sub in H.
          assumption.

    #2 apply (@subs_trans _ _ (subst1 c T2)).
      apply subs_refl; [].
      apply eqs_symm; assumption.

      so (subst_subs _#5 (unshift_tm _ _) Hsubs2) as Hsubs2'.
      simp_sub in Hsubs2'.
      assumption.

(* pair *)
intros G M1 M2 _ IH1 _ IH2 T V p -> Hof Hsteps Hval.
so (merge_ihm _#4 IHm IH1) as IH1'.
so (merge_ihm _#4 IHm IH2) as IH2'.
clear IHt IHm IH1 IH2.
so (pair_ofm_inversion _#5 Hwf Hof) as (T1 & T2 & Hsubs & Hof1 & Hof2).
clear Hof.
so (subordf_tm_sg' _#3 (ofs_fsg _ _ (ofm_reg _#4 Hwf Hof1))) as H.
forget (subst1 nonsense T1) as x; subst T1; rename x into T1.
so (subordf_tm_sg' _#3 (ofs_fsg _ _ (ofm_reg _#4 Hwf Hof2))) as H.
forget (subst1 nonsense T2) as x; subst T2; rename x into T2.
simp_sub in Hsteps.
so (mpair_halt_inversion _#4 Hsteps Hval) as (i1 & i2 & V1 & V2 & Hleq1 & Hleq2 & Hval1 & Hval2 & Hsteps1 & Hsteps2 & ->).
clear Hsteps Hval.
so (IH1' _#4 Hleq1 Hof1 Hsteps1 Hval1) as (j1 & V1' & HofV1' & Hval1' & -> & Hap1).
so (IH2' _#4 Hleq2 Hof2 Hsteps2 Hval2) as (j2 & V2' & HofV2' & Hval2' & -> & Hap2).
clear IH1' IH2' Hleq1 Hleq2 Hof1 Hof2 Hsteps1 Hsteps2 Hval1 Hval2.
exists (j1+j2), (md_pair V1' V2').
#3 do2 3 split; auto.
  eapply ofm_subsume; eauto; [].
  apply ofm_pair; auto; done.

  apply mvalue_pair; auto; done.

  intros l Hleq.
  rewrite -> !minus_plus_dist; [].
  so (subst_subs _#5 (unshift_tm _ _) Hsubs) as Hsubsx; simp_sub in Hsubsx.
  apply capproxm_comp; eapply mcc_subsume; eauto; [].
  simp_sub.
  #2 apply capproxm_comp; apply mcc_pair.
    #2 apply (capproxm_trans _#3 (subst1 (fixnv (l-j1) u f) V1')).
      apply lower_approxm; auto; omega.

      apply Hap1; auto; omega.

    #2 apply (capproxm_trans _#3 (subst1 (fixnv (l-j2) u f) V2')).
      apply lower_approxm; auto; omega.

      apply Hap2; auto; omega.

(* pi1 *)
intros G M _ _ T1 V1 p -> Hof Hsteps Hval1.
clear IHt.
so (pi1_ofm_inversion _#4 Hwf Hof) as (T2 & Hof').
clear Hof.
so (subordf_tm_sg _#3 (ofs_fsg _ _ (ofm_reg _#4 Hwf Hof'))) as Hissigma.
simp_sub in Hissigma.
injection Hissigma; clear Hissigma; intros HisT2.
replsub (subst (dot (var 0) (dot nonsense (sh 2))) T2) with (subst (under 1 sh1) (subst (under 1 (dot nonsense id)) T2)) in HisT2.
forget (subst (under 1 (dot nonsense id)) T2) as x; subst T2; rename x into T2. 
simp_sub in Hsteps.
so (mpi1_halt_inversion _#3 Hsteps Hval1) as (i' & V2 & Hlt & Hval2 & Hsteps').
clear Hsteps.
simp_sub in Hsteps'.
replsub (sg_sigma (shift1 T1) (subst (under 1 sh1) T2)) with (shift1 (sg_sigma T1 T2)) in Hof'.
so (IHm _ Hlt _#4 Hof' Hsteps' (mvalue_pair Hval1 Hval2)) as (j & V' & HofV' & HvalV' & Heq & Hap).
clear Hof' Hsteps'.
assert (exists V1' V2', V' = md_pair V1' V2') as (V1' & V2' & ->).
  case_eq V'; try (intros; subst V'; simp_sub in Heq; discriminate Heq); eauto; [].
  intros n ->.
  destruct n; simp_sub in Heq; discriminate Heq.
simp_sub in Heq.
injection Heq; clear Heq; intros -> ->.
so (unsubst_mvalue _#3 tvalue_lam Hval1) as Hval1'.
so (unsubst_mvalue _#3 tvalue_lam Hval2) as Hval2'.
clear Hval1 Hval2.
exists j, V1'.
#2 do2 3 split; auto.
  simp_sub in HofV'.
  so (pair_ofm_inversion_sigma_pure _#5 Hwf HofV') as (_ & HofV1' & _ & _).
  apply ofm_forget_pure; assumption.

  intros l Hleq.
  #2 apply (anti_msteps_capproxm_trans _#3 (subst1 (fixnv (l-j) u f) (md_pi1 (md_pair V1' V2')))).
    #2 eapply star_step.
      simp_sub.
      #2 apply mstep_pi12.
        eapply subst_mvalue; eauto; [].
        apply subof_dot; auto using subof_id; [].
        simp_sub; apply oof_utm; [].
        apply tvalue_lam; done.

        eapply subst_mvalue; eauto; [].
        apply subof_dot; auto using subof_id; [].
        simp_sub; apply oof_utm; [].
        apply tvalue_lam; done.

      apply star_refl; done.

    simp_sub.
    apply capproxm_forget_pure; [].
    apply capproxm_comp; apply (mcc_pi1 _#7 T2); [].
    so (Hap _ Hleq) as H.
    simp_sub in H.
    assumption.

(* pi2 *)
intros G M _ _ T V2 p -> Hof Hsteps Hval2.
clear IHt Hlti.
so (pi2_ofm_inversion _#4 Hwf Hof) as (T1 & T2 & c & Hof' & HfstM & Hsubs).
clear Hof.
so (subordf_tm_sg _#3 (ofs_fsg _ _ (ofm_reg _#4 Hwf Hof'))) as Hissigma.
simp_sub in Hissigma.
injection Hissigma; clear Hissigma; intros HisT2 HisT1.
replsub (subst (dot (var 0) (dot nonsense (sh 2))) T2) with (subst (under 1 sh1) (subst (under 1 (dot nonsense id)) T2)) in HisT2.
forget (subst (under 1 (dot nonsense id)) T2) as x; subst T2; rename x into T2. 
replsub (subst (dot nonsense sh1) T1) with (shift1 (subst1 nonsense T1)) in HisT1.
forget (subst1 nonsense T1) as x; subst T1; rename x into T1.
so (subordf_tm_cn' _#3 (fstm_fcn _#3 (wf_fcontext _ Hwf) (ofm_fmd _#4 (wf_fcontext _ Hwf) Hof') HfstM)).
forget (subst1 nonsense c) as x; subst c; rename x into c.
simp_sub in Hsteps.
so (mpi2_halt_inversion _#3 Hsteps Hval2) as (i' & V1 & Hlt & Hval1 & Hsteps').
clear Hsteps.
replsub (sg_sigma (shift1 T1) (subst (under 1 sh1) T2)) with (shift1 (sg_sigma T1 T2)) in Hof'.
so (IHm _ Hlt _#4 Hof' Hsteps' (mvalue_pair Hval1 Hval2)) as (j & V' & HofV' & HvalV' & Heq & Hap).
clear IHm.
assert (exists V1' V2', V' = md_pair V1' V2') as (V1' & V2' & ->).
  case_eq V'; try (intros; subst V'; simp_sub in Heq; discriminate Heq); eauto; [].
  intros n ->.
  destruct n; simp_sub in Heq; discriminate Heq.
simp_sub in Heq.
injection Heq; clear Heq; intros -> ->.
so (unsubst_mvalue _#3 tvalue_lam Hval1) as Hval1'.
so (unsubst_mvalue _#3 tvalue_lam Hval2) as Hval2'.
clear Hval1 Hval2.
so (subst_ofs _#4 (unshift_tm _ _) (ofm_reg _#4 Hwf Hof')) as Hofsigma; simp_sub in Hofsigma.
invert Hofsigma; clear Hofsigma; intros k1 _ HfstT1 HofT2; [].
simp_sub in HofV'.
so (pair_ofm_inversion_sigma_pure _#5 Hwf HofV') as (c1 & HofV1' & HfstV1' & HofV2').
so (subordf_tm_cn' _#3 (fstm_fcn _#3 (wf_fcontext _ Hwf) (ofm_fmd _#4 (wf_fcontext _ Hwf) HofV1') HfstV1')).
forget (subst1 nonsense c1) as x; subst c1; rename x into c1.
so (can_fstm _#3 Hwf HofV2') as (c2 & HfstV2').
so (subordf_tm_cn' _#3 (fstm_fcn _#3 (wf_fcontext _ Hwf) (ofm_fmd _#4 (wf_fcontext _ Hwf) HofV2') HfstV2')).
forget (subst1 nonsense c2) as x; subst c2; rename x into c2.
so (can_fsts _ _ HofT2) as (k2 & HfstT2).
assert (eqc nil c (cn_pair c1 c2) (kd_sigma k1 k2)) as Heqc.
  #6 apply (preservations_fst' _ pure (subst1 (fixv u f) M) (subst1 (fixv u f) (md_pair V1' V2')) (sg_sigma T1 T2)).
    exact wf_nil.

    so (subst1_ofm_tm _#6 (oft_fixv _#3 Hofu Hoff) Hof') as H; simp_sub in H.
    assumption.

    simp_sub.
    eapply expon_star; eauto; done.

    apply fsts_sigma; auto; done.

    so (subst1_fstm_tm _ (fixv u f) _#3 HfstM) as H; simp_sub in H.
    assumption.

    simp_sub.
    #2 apply fstm_pair.
      so (subst1_fstm_tm _ (fixv u f) _#3 HfstV1') as H; simp_sub in H.
      assumption.

      so (subst1_fstm_tm _ (fixv u f) _#3 HfstV2') as H; simp_sub in H.
      assumption.
exists j, V2'.
#2 do2 3 split; auto.
  simp_sub in HofV2'.
  apply ofm_forget_pure; [].
  eapply ofm_subsume; eauto; [].
  eapply subs_trans; eauto; [].
  simp_sub.
  apply subs_refl; [].
  replsub (subst (dot (shift1 c1) sh1) T2) with (shift1 (subst1 c1 T2)).
  replsub (subst (dot (cn_pi1 (shift1 c)) sh1) T2) with (shift1 (subst1 (cn_pi1 c) T2)).
  apply weaken1_eqs; [].
  eapply funct_sg; eauto using wf_nil; [].
  #2 apply (@eqc_trans _ _ (cn_pi1 (cn_pair c1 c2))).
    apply eqc_symm; [].
    #2 apply (@eqc_beta1 _ _ (subst1 c1 k2)).
      so (subst_ofc _#5 (unshift_tm _ _) (ofm_fst _#6 Hwf HofV1' HfstV1' (subst_fsts _ _ sh1 HfstT1))) as H; simp_sub in H.
      assumption.

      so (subst_ofc _#5 (unshift_tm _ _) (ofm_fst _#6 Hwf HofV2' HfstV2' (subst_fsts _ _ (dot (shift1 c1) sh1) HfstT2))) as H; simp_sub in H.
      assumption.

    apply (@eqc_pi1 _ _ k2); [].
    apply eqc_symm; auto; done.

  intros l Hleq.
  #2 apply (anti_msteps_capproxm_trans _#3 (subst1 (fixnv (l-j) u f) (md_pi2 (md_pair V1' V2')))).
    #2 eapply star_step.
      simp_sub.
      #2 apply mstep_pi22.
        eapply subst_mvalue; eauto; [].
        apply subof_dot; auto using subof_id; [].
        simp_sub; apply oof_utm; [].
        apply tvalue_lam; done.

        eapply subst_mvalue; eauto; [].
        apply subof_dot; auto using subof_id; [].
        simp_sub; apply oof_utm; [].
        apply tvalue_lam; done.

      apply star_refl; done.

    simp_sub.
    apply capproxm_forget_pure; [].
    #2 apply capproxm_comp; apply (@mcc_subsume _#7 (subst1 (cn_pi1 (cn_pair c1 c2)) T2)).
      #2 apply capproxm_comp; apply (@mcc_pi2 _#6 T1).
        so (Hap _ Hleq) as H.
        simp_sub in H.
        assumption.

        #2 apply fstm_pair.
          so (subst1_fstm_tm _ (fixnv (l-j) u f) _#3 HfstV1') as H; simp_sub in H.
          assumption.

          so (subst1_fstm_tm _ (fixnv (l-j) u f) _#3 HfstV2') as H; simp_sub in H.
          assumption.

      apply (@subs_trans _ _ (subst1 (cn_pi1 c) T2)).
        apply subs_refl; [].
        eapply funct_sg; eauto using wf_nil; [].
        apply (@eqc_pi1 _ _ k2); apply eqc_symm; assumption.

        so (subst_subs _#5 (unshift_tm _ _) Hsubs) as H; simp_sub in H.
        assumption.

(* unpack *)
intros G e M T' _ _ _ _ _ T V p -> Hof Hsteps Hval.
so (unpack_ofm_inversion _#6 Hwf Hof) as (k & t & Hofe & HofM & Hsubs & ->).
clear Hof.
so (subordf_tm_sg' _#3 (ofs_fsg _ _ (subs_reg _#3 Hwf Hsubs andel))) as H.
forget (subst1 nonsense T') as x; subst T'; rename x into T'.
so (subordf_tm_cn _#3 (ofc_fcn _#3 (oft_reg _#3 Hwf Hofe))) as Hisex; simp_sub in Hisex.
injection Hisex; clear Hisex; intros Hist Hisk.
replsub (subst (dot nonsense sh1) k) with (shift1 (subst1 nonsense k)) in Hisk.
forget (subst1 nonsense k) as x; subst k; rename x into k.
replsub (subst (dot (var 0) (dot nonsense (sh 2))) t) with (subst (under 1 sh1) (subst (under 1 (dot nonsense id)) t)) in Hist.
forget (subst (under 1 (dot nonsense id)) t) as x; subst t; rename x into t.
simp_sub in Hsteps.
so (munpack_halt_inversion _#5 Hsteps Hval) as (i1 & i2 & c & v & k' & t' & Hlt1 & Hlt2 & Hvalv & Hsteps1 & Hsteps2).
clear Hsteps.
replsub (cn_exists (shift1 k) (subst (under 1 sh1) t)) with (shift1 (cn_exists k t)) in Hofe.
so (IHt _ Hlt1 _#3 Hofe Hsteps1 (tvalue_pack Hvalv)) as (j1 & v1 & Hofv1 & Hvalv1 & Heq & Hap1).
clear Hsteps1 Hvalv IHt.
assert (exists c' v' k'' t'', v1 = tm_pack c' v' k'' t'') as (c' & v' & k'' & t'' & ->).
  case_eq v1; try (intros; subst v1; simp_sub in Heq; discriminate Heq); eauto; [].
  intros n ->.
  destruct n; simp_sub in Heq; discriminate Heq.
simp_sub in Heq; injection Heq; clear Heq; intros -> -> -> ->.
simp_sub in Hofv1.
so (pack_oft_inversion_exists _#7 Hwf Hofv1) as (Heqk & Heqt & Hofc' & Hofv').
so (subordf_tm_cn' _#3 (ofc_fcn _#3 Hofc')) as H.
forget (subst1 nonsense c') as x; subst c'; rename x into c'.
simp_sub in Hofv'.
#3 exploit (IHm _ Hlt2 (subst (dot v' (dot (shift1 c') id)) M) T' V impure) as H; auto; clear IHm.
  replsub (shift1 T') with (subst (dot v' (dot (shift1 c') id)) (subst (sh 3) T')).
  simp_sub in HofM.
  eapply subst_ofm; eauto; [].
  #2 apply subof_dot.
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply dof_cn; auto; done.

    simp_sub.
    apply dof_tm; auto; done.

    simp_sub.
    simp_sub in Hsteps2.
    assumption.
destruct H as (j2 & V' & HofV' & HvalV' & -> & Hap2).
clear Hsteps2.
exists (j1+j2), V'.
#2 do2 3 split; auto.
  eapply ofm_subsume; eauto; done.

  intros l Hleq.
  rewrite -> !minus_plus_dist; [].
  so (subst_subs _#5 (unshift_tm _ _) Hsubs) as Hsubsx; simp_sub in Hsubsx.
  apply capproxm_comp; eapply mcc_subsume; eauto; [].
  #2 apply (capproxm_trans _#3 (subst1 (fixnv (l-j1) u f) (subst (dot v' (dot (shift1 c') id)) M))).
    eapply Hap2; eauto; omega.

    #2 apply (anti_msteps_capproxm_trans _#3 (subst1 (fixnv (l-j1) u f) (md_unpack (tm_pack (shift1 c') v' k'' t'') M (shift1 T')))).
      simp_sub.
      #2 eapply star_step.
        apply mstep_unpack2; [].
        invert Hvalv1; [].
        intros Hvalv'.
        eapply subst_tvalue; eauto; [].
        apply subof_dot; auto using subof_id; [].
        simp_sub; apply oof_utm; apply tvalue_lam; done.

        simp_sub.
        apply star_refl; done.

      simp_sub.
      #3 apply capproxm_comp; apply (@mcc_unpack _#6 k t).
        simp_sub.
        exploit (Hap1 l) as H; [omega |].
        simp_sub in H.
        assumption.

        replsub (subst (dot (var 0) (dot (var 1) (dot (subst (sh 2) (fixnv (l-j1) u f)) (sh 2)))) M) with (subst (under 2 (dot (fixnv (l-j1) u f) id)) M).
        replsub (subst (dot (var 0) (dot (var 1) (dot (subst (sh 2) (fixnv l u f)) (sh 2)))) M) with (subst (under 2 (dot (fixnv l u f) id)) M).
        #3 apply lower_approxm_gen; auto.
          so (ofc_ofcr _#3 (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 Hwf Hofe))) as (kk & Hofcr); simp_sub in Hofcr.
          invert Hofcr; [].
          intros Hofk Hoft _.
          auto using wf; done.

          simpl app.
          simpl length.
          simp_sub.
          simp_sub in HofM.
          assumption.

          omega.

        exact (eqs_refl (subs_reg _#3 wf_nil Hsubsx andel)).

(* lett *)
intros G e M _ _ _ _ T V p -> Hof Hsteps Hval.
so (lett_ofm_inversion _#5 Hwf Hof) as (t & Hofe & HofM).
clear Hof.
so (subordf_tm_cn' _#3 (ofc_fcn _#3 (oft_reg _#3 Hwf Hofe))) as His.
forget (subst1 nonsense t) as x; subst t; rename x into t.
simp_sub in Hsteps.
so (mlett_halt_inversion _#4 Hsteps Hval) as (i1 & i2 & v & Hlt1 & Hlt2 & Hvalv & Hsteps1 & Hsteps2).
so (IHt _ Hlt1 _#3 Hofe Hsteps1 Hvalv) as (j1 & v' & Hofv' & Hvalv' & -> & Hap1).
clear Hsteps1 IHt.
simp_sub in Hsteps2.
replsub (subst (dot (subst1 (fixv u f) v') (dot (fixv u f) id)) M) with (subst1 (fixv u f) (subst1 v' M)) in Hsteps2.
so (subst1_ofm_tm _#6 Hofv' HofM) as HofMx; simp_sub in HofMx.
so (IHm _ Hlt2 _#4 HofMx Hsteps2 Hval) as (j2 & V' & HofV' & HvalV' & -> & Hap2).
clear HofMx Hsteps2 IHm.
exists (j1+j2), V'.
do2 3 split; auto; [].
intros l Hleq.
rewrite -> minus_plus_dist; [].
exploit (Hap2 (l-j1)) as H; [omega |].
eapply capproxm_trans; eauto; clear H; [].
#2 apply (anti_msteps_capproxm_trans _#3 (subst1 (fixnv (l-j1) u f) (md_lett v' M))).
  simp_sub.
  #2 eapply star_step.
    apply mstep_lett2; [].
    eapply subst_tvalue; eauto; [].
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply oof_utm; apply tvalue_lam; done.

    simp_sub.
    apply star_refl; done.

  simp_sub.
  #2 apply capproxm_comp; apply (mcc_lett _#7 t).
    eapply Hap1; eauto; omega.

    replsub (subst (dot (var 0) (dot (shift1 (fixnv (l-j1) u f)) sh1)) M) with (subst (under 1 (dot (fixnv (l-j1) u f) id)) M).
    replsub (subst (dot (var 0) (dot (shift1 (fixnv l u f)) sh1)) M) with (subst (under 1 (dot (fixnv l u f) id)) M).
    #3 apply lower_approxm_gen; auto.
      apply wf_tm; auto using wf_nil; [].
      so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 Hwf Hofe)) as H; simp_sub in H.
      assumption.

      simpl.
      simp_sub.
      simp_sub in HofM.
      assumption.

      omega.

(* letm *)
intros G M1 M2 T2 _ _ _ _ _ T V p -> Hof Hsteps Hval.
clear IHt Hlti.
so (letm_ofm_inversion _#6 Hwf Hof) as (T1 & k1 & HofM1 & HfstT1w & HofM2 & Hsubs & ->).
clear Hof.
so (subordf_tm_sg' _#3 (ofs_fsg _ _ (ofm_reg _#4 Hwf HofM1))) as H.
forget (subst1 nonsense T1) as x; subst T1; rename x into T1.
so (subordf_tm_sg' _#3 (ofs_fsg _ _ (subs_reg _#3 Hwf Hsubs andel))) as H.
forget (subst1 nonsense T2) as x; subst T2; rename x into T2.
so (subordf_tm_kd' _#3 (fsts_fkd _#3 (ofs_fsg _ _ (ofm_reg _#4 Hwf HofM1)) HfstT1w)) as H.
forget (subst1 nonsense k1) as x; subst k1; rename x into k1.
so (subst_fsts _ _ (dot nonsense id) HfstT1w) as HfstT1; simp_sub in HfstT1.
simp_sub in Hsteps.
so (mletm_halt_inversion _#5 Hsteps Hval) as (i1 & i2 & V1 & c1 & Hlt1 & Hlt2 & Hval1 & HfstV1 & Hsteps1 & Hsteps2).
clear Hsteps.
so (IHm _ Hlt1 _#4 HofM1 Hsteps1 Hval1) as (j1 & V1' & HofV1' & Hval1' & -> & Hap1).
clear Hlt1 Hsteps1 Hval1.
so (subst_mvalue _#4 (subof_tm_utm_oof _ (cn_arrow cn_unit u)) Hval1') as Hval1''; simp_sub in Hval1''.
so (mvalue_pure _#4 Hval1'' HofV1') as H; renameover H into HofV1'.
so (unsubst_fstm _#6 wf_nil (oft_fixv _#3 Hofu Hoff) HofV1' HfstV1) as HfstV1'.
simp_sub in Hsteps2.
replsub (subst (dot (subst1 (fixv u f) V1') (dot c1 (dot (fixv u f) id))) M2) with (subst1 (fixv u f) (subst (dot V1' (dot (shift1 c1) id)) M2)) in Hsteps2.
#2 exploit (IHm _ Hlt2 (subst (dot V1' (dot (shift1 c1) id)) M2) T2 V impure) as H; auto; clear IHm.
  simp_sub in HofM2.
  replsub (shift1 T2) with (subst (dot V1' (dot (shift1 c1) id)) (subst (sh 3) T2)).
  eapply subst_ofm; eauto; [].
  #2 apply subof_dot.
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply dof_cn; [].
    eapply ofm_fst; eauto; done.

    simp_sub.
    apply dof_md; auto; done.
clear Hsteps2.
destruct H as (j2 & V' & HofV' & HvalV' & -> & Hap2).
exists (j1 + j2), V'.
#2 do2 3 split; auto.
  eapply ofm_subsume; eauto; done.

  intros l Hleq.
  rewrite -> minus_plus_dist; [].
  so (subst_subs _#5 (unshift_tm _ _) Hsubs) as Hsubsx; simp_sub in Hsubsx.
  apply capproxm_comp; eapply mcc_subsume; eauto; [].
  exploit (Hap2 (l-j1)) as H; [omega |]; clear Hap2.
  eapply capproxm_trans; eauto; clear H; [].
  #2 apply (anti_msteps_capproxm_trans _#3 (md_letm (subst1 (fixnv (l-j1) u f) V1') (subst (under 2 (dot (fixnv (l-j1) u f) id)) M2) T2)).
    simp_sub.
    #2 eapply star_step.
      #2 eapply (@mstep_letm2 _#4 c1); eauto.
        eapply subst_mvalue; eauto; [].
        apply subof_dot; auto using subof_id; [].
        simp_sub.
        apply oof_tm; exact tvalue_lam.
  
        replsub c1 with (subst1 (fixnv (l-j1) u f) (shift1 c1)).
        eapply subst_fstm; eauto; [].
        apply subof_dot; auto using subof_id; [].
        simp_sub.
        apply tof_tm; done.
  
      simp_sub.
      apply star_refl; done.
  
    simp_sub.
    #3 apply capproxm_comp; apply (@mcc_letm _#6 T1 _ _ k1); auto.
      eapply Hap1; eauto; omega.
  
      simp_sub.
      replsub (dot (var 0) (dot (var 1) (dot (subst (sh 2) (fixnv (l-j1) u f)) (sh 2)))) with (under 2 (dot (fixnv (l-j1) u f) id)).
      replsub (dot (var 0) (dot (var 1) (dot (subst (sh 2) (fixnv l u f)) (sh 2)))) with (under 2 (dot (fixnv l u f) id)).
      #3 apply lower_approxm_gen; auto.
        apply wf_md; auto using wf; [].
        so (subst_ofs _#4 (unshift_tm _ _) (ofm_reg _#4 Hwf HofM1)) as H.
        simp_sub in H.
        assumption.
  
        simpl.
        simp_sub.
        simp_sub in HofM2.
        assumption.
  
        omega.
  
      exact (eqs_refl (subs_reg _#3 wf_nil Hsubsx andel)).

(* seal *)
intros G T' M _ _ HT T V p -> Hof Hsteps Hval.
clear IHt.
so (subordf_utm_sg' _ _ HT) as H; clear HT.
forget (subst1 nonsense T') as x; subst T'; rename x into T'.
so (seal_ofm_inversion _#5 Hwf Hof) as (HofM & Hsubs & ->).
simp_sub in Hsteps.
repl i with (S (i-1)) in Hsteps by omega.
invert Hsteps; clear Hsteps; [].
intros M1 Hstep Hsteps.
invert Hstep; [].
intros <-.
assert (i-1 < i) as Hlt by omega.
so (IHm _ Hlt _#4 HofM Hsteps Hval) as (j & V' & HofV' & HvalV' & -> & Hap).  
exists j, V'.
#2 do2 3 split; auto.
  eapply ofm_subsume; eauto; done.

  intros l Hleq.
  simp_sub.
  so (subst_subs _#5 (unshift_tm _ _) Hsubs) as Hsubsx; simp_sub in Hsubsx.
  apply capproxm_comp; eapply mcc_subsume; eauto; [].
  so (Hap _ Hleq) as H.
  eapply capproxm_trans; eauto; clear H; [].
  #2 apply anti_mstep_capproxm.
    apply ofm_seal; [].
    replsub T' with (subst1 (fixnv l u f) (shift1 T')).
    eapply subst1_ofm_tm; eauto; [].
    apply oft_fixnv; auto; done.

    apply mstep_seal; done.
    
(* epilogue *)
destruct Hind as (Hindt & Hindm).
clear IHt IHm.
split.
  intros e t v Hof Hsteps Hval.
  so (subst_ftm _#4 (subof_fof_utm_tm _#4 (subof_id _ _)) (oft_ftm _#3 (fcontext_tm fcontext_nil (fcn_arrow fcn_unit (ofc_fcn _#3 Hofu))) Hof)) as He.
  simp_sub in He.
  eapply Hindt; eauto; done.


  intros M T V p Hof Hsteps Hval.
  so (subst_fmd _#4 (subof_fof_utm_tm _#4 (subof_id _ _)) (ofm_fmd _#4 (fcontext_tm fcontext_nil (fcn_arrow fcn_unit (ofc_fcn _#3 Hofu))) Hof)) as HM.
  simp_sub in HM.
  eapply Hindm; eauto; done.
Qed.


Theorem compactness :
  forall u f e t,
    oft nil f (cn_arrow (cn_arrow cn_unit u) u)
    -> oft (nil; cl_tm (cn_arrow cn_unit u)) e t
    -> thalt nil (subst1 (fixv u f) e)
    -> exists i, forall j, j >= i -> thalt nil (subst1 (fixnv j u f) e).
Proof.
intros u f e t Hoff Hof (v & Hsteps & Hval).
so (ofc_ofcr _#3 (oft_reg _#3 wf_nil Hoff)) as (kk & Hofcr).
invert Hofcr; clear Hofcr; [].
intros _ Hofu _; clear kk.
so (subordf_tm_cn' _#3 (ofc_fcn _#3 (oft_reg _#3 (wf_tm wf_nil (ofc_arrow ofc_unit Hofu)) Hof))) as H.
forget (subst1 nonsense t) as x; subst t; rename x into t.
so (star_expon _#4 Hsteps) as (i & Hsteps').
so (simulation _#3 Hoff andel _#3 Hof Hsteps' Hval) as (j & v' & _ & Hvalv' & _ & Hap).
exists j.
intros l Hleq.
so (Hap _ Hleq) as Hap'.
eapply capproxt_halt; eauto; [].
eexists; split.
  apply star_refl; done.

  eapply subst_tvalue; eauto; [].
  apply subof_dot; auto using subof_id; [].
  simp_sub.
  apply oof_utm; [].
  apply tvalue_lam; done.
Qed.
