
Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Formation.
Require Import Functionality.
Require Import Regularity.
Require Import Singleton.
Require Import Principal.
Require Import Declarative.
Require Preservation.
Require Progress.



Definition kreln := context -> (* kd *) term -> (* kd *) term -> Prop.
Definition creln := context -> (* cn *) term -> (* cn *) term -> (* kd *) term -> Prop.
Definition sreln := context -> (* sg *) term -> (* sg *) term -> Prop.
Definition treln := context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop.
Definition mreln := context -> purity -> (* md *) term -> (* md *) term -> (* md *) term -> Prop.


(* Compatibility *)

Inductive tcc (Rk : kreln) (Rc : creln) (Rs : sreln) (Rt : treln) (Rm : mreln) : treln :=
| tcc_lam {G t1 t1' t2 e e'}
    : Rc G t1 t1' kd_type
      -> Rt (G; cl_tm t1) e e' (shift1 t2)
      -> tcc Rk Rc Rs Rt Rm G (tm_lam t1 e) (tm_lam t1' e') (cn_arrow t1 t2)

| tcc_app {G t1 t2 e1 e1' e2 e2'}
    : Rt G e1 e1' (cn_arrow t1 t2)
       -> Rt G e2 e2' t1
       -> tcc Rk Rc Rs Rt Rm G (tm_app e1 e2) (tm_app e1' e2') t2

| tcc_pair {G t1 t2 e1 e1' e2 e2'}
    : Rt G e1 e1' t1
      -> Rt G e2 e2' t2
      -> tcc Rk Rc Rs Rt Rm G (tm_pair e1 e2) (tm_pair e1' e2') (cn_prod t1 t2)

| tcc_pi1 {G t1 t2 e e'}
    : Rt G e e' (cn_prod t1 t2)
      -> tcc Rk Rc Rs Rt Rm G (tm_pi1 e) (tm_pi1 e') t1

| tcc_pi2 {G t1 t2 e e'}
    : Rt G e e' (cn_prod t1 t2)
      -> tcc Rk Rc Rs Rt Rm G (tm_pi2 e) (tm_pi2 e') t2

| tcc_plam {G k k' t e e'}
    : Rk G k k'
      -> Rt (G; cl_cn k) e e' t
      -> tcc Rk Rc Rs Rt Rm G (tm_plam k e) (tm_plam k' e') (cn_all k t)

| tcc_papp {G k t e e' c c'}
    : Rt G e e' (cn_all k t)
      -> Rc G c c' k
      -> tcc Rk Rc Rs Rt Rm G (tm_papp e c) (tm_papp e' c') (subst1 c t)

| tcc_pack {G k k' t t' c c' e e'}
    : Rc G c c' k
      -> Rk G k k'
      -> Rt G e e' (subst1 c t)
      -> Rc (G; cl_cn k) t t' kd_type
      -> tcc Rk Rc Rs Rt Rm G (tm_pack c e k t) (tm_pack c' e' k' t') (cn_exists k t)

| tcc_unpack {G k t t' e1 e1' e2 e2'}
    : Rt G e1 e1' (cn_exists k t)
      -> Rt (G; cl_cn k; cl_tm t) e2 e2' (subst (sh 2) t')
      -> ofc G t' kd_type
      -> tcc Rk Rc Rs Rt Rm G (tm_unpack e1 e2) (tm_unpack e1' e2') t'

| tcc_fix {G t t' e e'}
    : Rc G t t' kd_type
      -> Rt G e e' (cn_arrow (cn_arrow cn_unit t) t)
      -> tcc Rk Rc Rs Rt Rm G (tm_fix t e) (tm_fix t' e') t

| tcc_lett {G t1 t2 e1 e1' e2 e2'}
    : Rt G e1 e1' t1
      -> Rt (G; cl_tm t1) e2 e2' (shift1 t2)
      -> tcc Rk Rc Rs Rt Rm G (tm_lett e1 e2) (tm_lett e1' e2') t2

| tcc_equiv {G t1 t2 e e'}
    : Rt G e e' t1
      -> eqc G t1 t2 kd_type
      -> tcc Rk Rc Rs Rt Rm G e e' t2

| tcc_letm {G T k t M M' e e'}
    : Rm G impure M M' T
      -> fsts T k
      -> Rt (G; cl_cn k; cl_md (shift1 T) (var 0)) e e' (subst (sh 2) t)
      -> ofc G t kd_type
      -> tcc Rk Rc Rs Rt Rm G (tm_letm M e) (tm_letm M' e') t

| tcc_ext {G t M M'}
    : Rm G impure M M' (sg_dyn t)
      -> tcc Rk Rc Rs Rt Rm G (tm_ext M) (tm_ext M') t
.


Inductive mcc (Rk : kreln) (Rc : creln) (Rs : sreln) (Rt : treln) (Rm : mreln) : mreln :=
| mcc_stat {G c c' k}
    : Rc G c c' k
       -> mcc Rk Rc Rs Rt Rm G pure (md_stat c) (md_stat c') (sg_stat k)

| mcc_dyn {G e e' t}
    : Rt G e e' t
      -> mcc Rk Rc Rs Rt Rm G pure (md_dyn e) (md_dyn e') (sg_dyn t)

| mcc_lamgn {G T1 T1' T2 k1 M M'}
    : Rs G T1 T1'
      -> fsts T1 k1
      -> Rm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M M' (shift1 T2)
      -> mcc Rk Rc Rs Rt Rm G pure (md_lamgn T1 M) (md_lamgn T1' M') (sg_pigen T1 T2)

| mcc_appgn {G T1 T2 M1 M1' M2 M2' c2}
    : Rm G impure M1 M1' (sg_pigen T1 T2)
      -> Rm G pure M2 M2' T1
      -> fstm G M2 c2
      -> mcc Rk Rc Rs Rt Rm G impure (md_appgn M1 M2) (md_appgn M1' M2') (subst1 c2 T2)

| mcc_lamap {G T1 T1' T2 k1 M M'}
    : Rs G T1 T1'
      -> fsts T1 k1
      -> Rm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) pure M M' (shift1 T2)
      -> mcc Rk Rc Rs Rt Rm G pure (md_lamap T1 M) (md_lamap T1' M') (sg_piapp T1 T2)

| mcc_appap {G p T1 T2 M1 M1' M2 M2' c2}
    : Rm G p M1 M1' (sg_piapp T1 T2)
      -> Rm G pure M2 M2' T1
      -> fstm G M2 c2
      -> mcc Rk Rc Rs Rt Rm G p (md_appap M1 M2) (md_appap M1' M2') (subst1 c2 T2)

| mcc_pair {G p T1 T2 M1 M1' M2 M2'}
    : Rm G p M1 M1' T1
      -> Rm G p M2 M2' T2
      -> mcc Rk Rc Rs Rt Rm G p (md_pair M1 M2) (md_pair M1' M2') (sg_sigma T1 (shift1 T2))

| mcc_pi1 {G T1 T2 M M'}
    : Rm G pure M M' (sg_sigma T1 T2)
      -> mcc Rk Rc Rs Rt Rm G pure (md_pi1 M) (md_pi1 M') T1

| mcc_pi2 {G T1 T2 M M' c}
    : Rm G pure M M' (sg_sigma T1 T2)
      -> fstm G M c
      -> mcc Rk Rc Rs Rt Rm G pure (md_pi2 M) (md_pi2 M') (subst1 (cn_pi1 c) T2)

| mcc_unpack {G k t T T' e e' M M'}
    : Rt G e e' (cn_exists k t)
      -> Rm (G; cl_cn k; cl_tm t) impure M M' (subst (sh 2) T)
      -> Rs G T T'
      -> mcc Rk Rc Rs Rt Rm G impure (md_unpack e M T) (md_unpack e' M' T') T

| mcc_lett {G p t T e e' M M'}
    : Rt G e e' t
      -> Rm (G; cl_tm t) p M M' (shift1 T)
      -> mcc Rk Rc Rs Rt Rm G p (md_lett e M) (md_lett e' M') T

| mcc_letm {G T1 T2 T2' k1 M1 M1' M2 M2'}
    : Rm G impure M1 M1' T1
      -> fsts T1 k1
      -> Rm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M2 M2' (subst (sh 2) T2)
      -> Rs G T2 T2'
      -> mcc Rk Rc Rs Rt Rm G impure (md_letm M1 M2 T2) (md_letm M1' M2' T2') T2

| mcc_seal {G T T' M M'}
    : Rm G impure M M' T
      -> Rs G T T'
      -> mcc Rk Rc Rs Rt Rm G impure (md_seal M T) (md_seal M' T') T

| mcc_forget {G T M M'}
    : Rm G pure M M' T
      -> mcc Rk Rc Rs Rt Rm G impure M M' T

| mcc_subsume {G p T1 T2 M M'}
    : Rm G p M M' T1
      -> subs G T1 T2
      -> mcc Rk Rc Rs Rt Rm G p M M' T2
.


Definition compatible (Rk : kreln) (Rc : creln) (Rs : sreln) (Rt : treln) (Rm : mreln)
  :=
  (forall G e e' t, wf G -> tcc Rk Rc Rs Rt Rm G e e' t -> Rt G e e' t)
  /\
  (forall G p M M' T, wf G -> mcc Rk Rc Rs Rt Rm G p M M' T -> Rm G p M M' T)
.


(* Congruences *)

Definition sregular
  (Rk : kreln)
  (Rc : creln)
  (Rs : sreln)
  :=
  (forall G k k', wf G -> Rk G k k' -> eqk G k k')
  /\
  (forall G c c' k, wf G -> Rc G c c' k -> eqc G c c' k)
  /\
  (forall G T T', wf G -> Rs G T T' -> eqs G T T').


Definition cregular
  (Rt : treln)
  (Rm : mreln) 
  :=
  (forall G e e' t, wf G -> Rt G e e' t -> oft G e t /\ oft G e' t)
  /\
  (forall G p M M' T, wf G -> Rm G p M M' T -> ofm G p M T /\ ofm G p M' T)
  /\
  (forall G M M' T,
     wf G
     -> Rm G pure M M' T
     -> exists k c c',
          fsts T k
          /\ fstm G M c
          /\ fstm G M' c'
          /\ eqc G c c' k).


Definition creflexive
  (Rt : treln)
  (Rm : mreln) 
  :=
  (forall G e t, oft G e t -> Rt G e e t)
  /\
  (forall G p M T, ofm G p M T -> Rm G p M M T).

  
Definition ssymmetric
  (Rk : kreln)
  (Rc : creln)
  (Rs : sreln)
  :=
  (forall G k k', Rk G k k' -> Rk G k' k)
  /\
  (forall G c c' k, Rc G c c' k -> Rc G c' c k)
  /\
  (forall G T T', Rs G T T' -> Rs G T' T).


Definition csymmetric
  (Rt : treln)
  (Rm : mreln) 
  :=
  (forall G t e e', Rt G e e' t -> Rt G e' e t)
  /\
  (forall G p T M M', Rm G p M M' T -> Rm G p M' M T).


Definition ctransitive
  (Rt : treln)
  (Rm : mreln) 
  :=
  (forall G t e1 e2 e3, Rt G e1 e2 t -> Rt G e2 e3 t -> Rt G e1 e3 t)
  /\
  (forall G p T M1 M2 M3, Rm G p M1 M2 T -> Rm G p M2 M3 T -> Rm G p M1 M3 T).


Definition ssubstitutive
  (Rk : kreln)
  (Rc : creln)
  (Rs : sreln)
  :=
  (forall G1 s G2 k k',
     subof sof G1 s G2 
     -> wf G2
     -> Rk G2 k k'
     -> Rk G1 (subst s k) (subst s k'))
  /\
  (forall G1 s G2 c c' k,
     subof sof G1 s G2 
     -> wf G2
     -> Rc G2 c c' k
     -> Rc G1 (subst s c) (subst s c') (subst s k))
  /\
  (forall G1 s G2 T T',
     subof sof G1 s G2 
     -> wf G2
     -> Rs G2 T T'
     -> Rs G1 (subst s T) (subst s T')).


Definition csubstitutive (Rt : treln) (Rm : mreln) 
  :=
  (forall G1 s G2 e e' t,
     subof aof G1 s G2 
     -> wf G2
     -> Rt G2 e e' t
     -> Rt G1 (subst s e) (subst s e') (subst s t))
  /\
  (forall G1 s G2 p M M' T,
     subof aof G1 s G2 
     -> wf G2
     -> Rm G2 p M M' T
     -> Rm G1 p (subst s M) (subst s M') (subst s T)).


Inductive samek : kreln :=
| samekrefl {G k}
    : ofk G k
      -> samek G k k.


Inductive samec : creln :=
| samecrefl {G c k}
    : ofc G c k
      -> samec G c c k.


Inductive sames : sreln :=
| samesrefl {G T}
    : ofs G T
      -> sames G T T.


Hint Constructors samek samec sames : same.


(* Congruence minus symmetry. *)
Definition semicongruence Rt Rm :=
  cregular Rt Rm
  /\ creflexive Rt Rm
  /\ ctransitive Rt Rm
  /\ csubstitutive Rt Rm
  /\ compatible samek samec sames Rt Rm.


Definition congruence Rt Rm :=
  semicongruence Rt Rm
  /\ csymmetric Rt Rm.


(* Congruence minus symmetry and transitivity. *)
Definition quasicongruence Rt Rm :=
  cregular Rt Rm
  /\ creflexive Rt Rm
  /\ csubstitutive Rt Rm
  /\ compatible samek samec sames Rt Rm.


Definition consistent
  (Rm : context -> purity -> (* md *) term -> (* md *) term -> (* cn *) term -> Prop)
  :=
  forall M M' T,
    Rm nil impure M M' T
    -> mhalt nil M
    -> mhalt nil M'.


Definition self_respecting
  (Rm : mreln) 
  :=
  forall G M M' T c,
    wf G
    -> ofs G T
    -> Rm G pure M M' T
    -> fstm G M c
    -> Rm G pure M M' (singles c T).


(* Properties *)

Lemma subst_samek :
  forall G1 G2 s k k',
    subof sof G1 s G2
    -> samek G2 k k'
    -> samek G1 (subst s k) (subst s k').
Proof.
intros G1 G2 s k k' Hs Hsame.
invert Hsame; [].
intros Hof <-.
apply samekrefl; [].
eapply subst_ofk; eauto.
Qed.


Lemma subst_samec :
  forall G1 G2 s c c' k,
    subof sof G1 s G2
    -> samec G2 c c' k
    -> samec G1 (subst s c) (subst s c') (subst s k).
Proof.
intros G1 G2 s c c' k Hs Hsame.
invert Hsame; [].
intros Hof <-.
apply samecrefl; [].
eapply subst_ofc; eauto.
Qed.


Lemma subst_sames :
  forall G1 G2 s T T',
    subof sof G1 s G2
    -> sames G2 T T'
    -> sames G1 (subst s T) (subst s T').
Proof.
intros G1 G2 s T T' Hs Hsame.
invert Hsame; [].
intros Hof <-.
apply samesrefl; [].
eapply subst_ofs; eauto.
Qed.


Lemma same_sregular :
  sregular samek samec sames.
Proof.
do2 2 split.
  intros G k k' _ H.
  invert H; [].
  intros; subst; eauto using eqk_refl; done.

  intros G c c' k _ H.
  invert H; [].
  intros; subst; eauto using eqc_refl; done.

  intros G T T' _ H.
  invert H.
  intros; subst; eauto using eqs_refl; done.
Qed.


Lemma eq_sregular :
  sregular eqk eqc eqs.
Proof.
unfold sregular; auto.
Qed.


Lemma same_ssubstitutive :
  ssubstitutive samek samec sames.
Proof.
do2 2 split.
  eauto using subst_samek; done.

  eauto using subst_samec; done.

  eauto using subst_sames; done.
Qed.


Lemma eq_ssubstitutive :
  ssubstitutive eqk eqc eqs.
Proof.
do2 2 split.
  eauto using subst_eqk; done.

  eauto using subst_eqc; done.

  eauto using subst_eqs; done.
Qed.


Lemma same_ssymetric :
  ssymmetric samek samec sames.
Proof.
do2 2 split.
  intros G k k' H.
  invert H; intros; subst; eauto with same; done.

  intros G c c' k H.
  invert H; intros; subst; eauto with same; done.

  intros G T T' H.
  invert H; intros; subst; eauto with same; done.
Qed.  


Lemma eq_ssymetric :
  ssymmetric eqk eqc eqs.
Proof.
do2 2 split; eauto with static.
Qed.  


Lemma tcc_mono :
  forall (Rk Rk' : kreln) (Rc Rc' : creln) (Rs Rs' : sreln) (Rt Rt' : treln) (Rm Rm' : mreln),
    (forall G k k', Rk G k k' -> Rk' G k k')
    -> (forall G c c' k, Rc G c c' k -> Rc' G c c' k)
    -> (forall G T T', Rs G T T' -> Rs' G T T')
    -> (forall G e e' t, Rt G e e' t -> Rt' G e e' t)
    -> (forall G p M M' T, Rm G p M M' T -> Rm' G p M M' T)
    -> forall G e e' t,
         tcc Rk Rc Rs Rt Rm G e e' t
         -> tcc Rk' Rc' Rs' Rt' Rm' G e e' t.
Proof.
intros Rk Rk' Rc Rc' Rs Rs' Rt Rt' Rm Rm' Himpk Himpc Himps Himpt Himpm G e e' t H.
induct H; intros; eauto 6 using tcc.
Qed.


Lemma mcc_mono :
  forall (Rk Rk' : kreln) (Rc Rc' : creln) (Rs Rs' : sreln) (Rt Rt' : treln) (Rm Rm' : mreln),
    (forall G k k', Rk G k k' -> Rk' G k k')
    -> (forall G c c' k, Rc G c c' k -> Rc' G c c' k)
    -> (forall G T T', Rs G T T' -> Rs' G T T')
    -> (forall G e e' t, Rt G e e' t -> Rt' G e e' t)
    -> (forall G p M M' T, Rm G p M M' T -> Rm' G p M M' T)
    -> forall G p M M' T,
         mcc Rk Rc Rs Rt Rm G p M M' T
         -> mcc Rk' Rc' Rs' Rt' Rm' G p M M' T.
Proof.
intros Rk Rk' Rc Rc' Rs Rs' Rt Rt' Rm Rm' Himpk Himpc Himps Himpt Himpm G p M M' T H.
induct H; intros; eauto using mcc.
Qed.


Lemma tcc_mono_wf :
  forall Rk Rc Rs (Rt Rt' : treln) (Rm Rm' : mreln),
    sregular Rk Rc Rs
    -> cregular Rt Rm
    -> (forall G e e' t, wf G -> Rt G e e' t -> Rt' G e e' t)
    -> (forall G p M M' T, wf G -> Rm G p M M' T -> Rm' G p M M' T)
    -> forall G e e' t,
         wf G
         -> tcc Rk Rc Rs Rt Rm G e e' t
         -> tcc Rk Rc Rs Rt' Rm' G e e' t.
Proof.
intros Rk Rc Rs Rt Rt' Rm Rm' Hsreg Hreg Himpt Himpm G e e' t Hwf Hcc.
revert Hwf.
#3 induct Hcc;
  try (intros;
       repeat
         (match goal with
          | H : Rk _ _ _ |- _ =>
              let Hof := fresh in
              let Hof' := fresh
              in
                so (eqk_reg _#3 Hwf (Hsreg andel _#3 Hwf H)) as (Hof & Hof');
                revert H
     
          | H : Rc _ _ _ _ |- _ =>
              let Hof := fresh in
              let Hof' := fresh
              in
                so (eqc_reg _#4 Hwf (Hsreg anderl _#4 Hwf H)) as (Hof & Hof' & _);
                revert H
          end);
       intros;
       eauto 6 using tcc, wf;
       done).

(* unpack *)
intros G k t t' e1 e1' e2 e2' H1 H2 Hoft' Hwf.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf (Hreg andel _#4 Hwf H1 andel))) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
eauto 6 using tcc, wf; done.

(* lett *)
intros G t1 t2 e1 e1' e2 e2' H1 H2 Hwf.
so (oft_reg _#3 Hwf (Hreg andel _#4 Hwf H1 andel)).
eauto 6 using tcc, wf; done.

(* letm *)
intros G T k t M M' e e' H1 Hfst H2 H3 Hwf.
so (ofm_reg _#4 Hwf (Hreg anderl _#5 Hwf H1 andel)).
eauto 6 using tcc, wf; done.
Qed.


Lemma mcc_mono_wf :
  forall Rk Rc Rs (Rt Rt' : treln) (Rm Rm' : mreln),
    sregular Rk Rc Rs
    -> cregular Rt Rm
    -> (forall G e e' t, wf G -> Rt G e e' t -> Rt' G e e' t)
    -> (forall G p M M' T, wf G -> Rm G p M M' T -> Rm' G p M M' T)
    -> forall G p M M' T,
         wf G
         -> mcc Rk Rc Rs Rt Rm G p M M' T
         -> mcc Rk Rc Rs Rt' Rm' G p M M' T.
Proof.
intros Rk Rc Rs Rt Rt' Rm Rm' Hsreg Hreg Himpt Himpm G p M M' T Hwf Hcc.
revert Hwf.
#3 induct Hcc;
  try (intros;
       repeat
         (match goal with
          | H : Rs _ _ _ |- _ =>
              let Hof := fresh in
              let Hof' := fresh
              in
                so (eqs_reg _#3 Hwf (Hsreg anderr _#3 Hwf H)) as (Hof & Hof');
                revert H
          end);
       intros;
       eauto 6 using mcc, wf;
       done).

(* unpack *)
intros G k t T T' e e' M M' H1 H2 H3 Hwf.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf (Hreg andel _#4 Hwf H1 andel))) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
eauto 6 using mcc, wf; done.

(* lett *)
intros G p t1 t2 e1 e1' e2 e2' H1 H2 Hwf.
so (oft_reg _#3 Hwf (Hreg andel _#4 Hwf H1 andel)).
eauto 6 using mcc, wf; done.

(* letm *)
intros G T1 T2 T2' k1 M1 M1' M2 M2' H1 Hfst H2 H3 Hwf.
so (ofm_reg _#4 Hwf (Hreg anderl _#5 Hwf H1 andel)).
eauto 6 using mcc, wf; done.
Qed.


Lemma compatible_mono :
  forall (Rk Rk' : _ -> _ -> _ -> Prop) (Rc Rc' : _ -> _ -> _ -> _ -> Prop) (Rs Rs' : _ -> _ -> _ -> Prop) Rt Rm,
    (forall G k k', Rk G k k' -> Rk' G k k')
    -> (forall G c c' k, Rc G c c' k -> Rc' G c c' k)
    -> (forall G T T', Rs G T T' -> Rs' G T T')
    -> compatible Rk' Rc' Rs' Rt Rm
    -> compatible Rk Rc Rs Rt Rm.
Proof.
intros Rk Rk' Rc Rc' Rs Rs' Rt Rm Himplk Himplc Himpls (Htcomp & Hmcomp).
split; eauto using tcc_mono, mcc_mono; done.
Qed.


Lemma compatible_ext :
  forall Rk Rc Rs Rt Rt' Rm Rm',
    (forall G e e' t, Rt G e e' t <-> Rt' G e e' t)
    -> (forall G p M M' T, Rm G p M M' T <-> Rm' G p M M' T)
    -> compatible Rk Rc Rs Rt Rm
    -> compatible Rk Rc Rs Rt' Rm'.
Proof.
intros Rk Rc Rs Rt Rt' Rm Rm' Hifft Hiffm Hcomp.
so (fun G e e' t => Hifft G e e' t ander) as Himplt.
so (fun G p M M' T => Hiffm G p M M' T ander) as Himplm.
split.
  intros G e e' t Hwf Htcc.
  apply Hifft; apply Hcomp; auto; [].
  invert Htcc; intros; subst; eauto using tcc; done.

  intros G p M M' T Hwf Hmcc.
  apply Hiffm; apply Hcomp; auto; [].
  invert Hmcc; intros; subst; eauto using mcc; done.
Qed.


Lemma semicongruence_ext :
  forall Rt Rt' Rm Rm',
    (forall G e e' t, Rt G e e' t <-> Rt' G e e' t)
    -> (forall G p M M' T, Rm G p M M' T <-> Rm' G p M M' T)
    -> semicongruence Rt Rm
    -> semicongruence Rt' Rm'.
Proof.
intros Rt Rt' Rm Rm' Hifft Hiffm ((Htreg & Hmreg & Hregfst) & (Htrefl & Hmrefl) & (Httrans & Hmtrans) & (Htsubst & Hmsubst) & Hcompat).
so (fun G e e' t => Hifft G e e' t ander) as Himplt.
so (fun G p M M' T => Hiffm G p M M' T ander) as Himplm.
so (fun G e e' t => Hifft G e e' t andel) as Himplt'.
so (fun G p M M' T => Hiffm G p M M' T andel) as Himplm'.
do2 4 split.
  do2 2 split; auto; done.

  split; auto; done.

  split; eauto; done.

  split; eauto; done.

  eapply compatible_ext; eauto; done.
Qed.


Lemma subst_tcc :
  forall Rk Rc Rs Rt Rm,
    sregular Rk Rc Rs
    -> cregular Rt Rm
    -> ssubstitutive Rk Rc Rs
    -> csubstitutive Rt Rm
    -> forall G1 G2 s e e' t,
         subof aof G1 s G2
         -> wf G2
         -> tcc Rk Rc Rs Rt Rm G2 e e' t
         -> tcc Rk Rc Rs Rt Rm G1 (subst s e) (subst s e') (subst s t).
Proof.
intros Rk Rc Rs Rt Rm (Hkreg & Hcreg & Hsreg) (Htreg & Hmreg & _) (Hksubst & Hcsubst & _) (Htsubst & Hmsubst).
assert (forall G1 s G2 c c', 
          subof sof G1 s G2
          -> wf G2
          -> Rc G2 c c' kd_type -> Rc G1 (subst s c) (subst s c') kd_type) as Hcsubstt.
  intros G1 s G2 c c' Hs Hwf HR.
  replsub kd_type with (subst s kd_type).
  eapply Hcsubst; eauto; done.
intros G1 G2 s e e' t Hs Hwf Htcc.
revert G1 s Hs Hwf.
#9 induct Htcc;
try (intro G; intros;
     repeat
       (match goal with
        | H : Rt G _ _ _ |- _ =>
          let H' := fresh in so (Htsubst _#6 Hs Hwf H) as H'; clear H; simp_sub in H'

        | H : eqc G _ _ _ |- _ =>
          let H' := fresh in so (subst_eqc _#6 (subof_aof_sof _#3 Hs) H) as H'; clear H; simp_sub in H'
        end);
     simp_sub;
     eauto using tcc; done).

(* lam *)
intros G t1 t1' t2 e e' H1 H2 G1 s Hs Hwf.
simp_sub.
apply tcc_lam; simp_sub; eauto using subof_aof_sof; [].
repl (subst (compose s sh1) t2) with (subst (dot (var 0) (compose s sh1)) (shift1 t2)) by (simp_sub; reflexivity).
so (eqc_reg _#4 Hwf (Hcreg _#4 Hwf H1) andel).
eapply Htsubst; eauto using subof_bind_aof_tm, wf; done.

(* plam *)
intros G k k' t e e' H1 H2 G1 s Hs Hwf.
simp_sub.
so (eqk_reg _#3 Hwf (Hkreg _#3 Hwf H1) andel).
apply tcc_plam; simp_sub; eauto using subof_aof_sof, subof_bind_aof_cn, wf; done.

(* papp *)
intros G k t e e' c c' HR H0 G1 s Hs Hwf.
so (Htsubst _#6 Hs Hwf HR) as HR'; simp_sub in HR'.
simp_sub.
rewrite -> split_dot_compose; [].
eapply tcc_papp; eauto using subof_aof_sof; done.

(* pack *)
intros G k k' t t' c c' e e' H1 H2 HR H3 G1 s Hs Hwf.
so (Htsubst _#6 Hs Hwf HR) as HR'; simp_sub in HR'.
simp_sub.
so (eqk_reg _#3 Hwf (Hkreg _#3 Hwf H2) andel).
eapply tcc_pack; simp_sub; eauto using subof_aof_sof, subof_bind_aof_cn, wf; done.

(* unpack *)
intros G k t t' e1 e1' e2 e2' HR1 HR2 H1 G1 s Hs Hwf.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf (Htreg _#4 Hwf HR1 andel))) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
so (Htsubst _#6 Hs Hwf HR1) as HR1'; simp_sub in HR1'.
so (Htsubst _#6 (subof_bind_aof_cn_tm _#5 Hs) (wf_tm (wf_cn Hwf Hofk) Hoft) HR2) as HR2'; simp_sub in HR2'.
simp_sub.
eapply tcc_unpack; simp_sub; eauto using subst_ofc_type, subof_aof_sof; done.

(* fix *)
intros G t t' e e' H HR G1 s Hs Hwf.
simp_sub.
apply tcc_fix; eauto using subof_aof_sof; [].
so (Htsubst _#6 Hs Hwf HR) as HR'; simp_sub in HR'.
assumption.

(* lett *)
intros G t1 t2 e1 e1' e2 e2' HR1 HR2 G1 s Hs Hwf.
so (oft_reg _#3 Hwf (Htreg _#4 Hwf HR1 andel)) as Hoft.
so (Htsubst _#6 Hs Hwf HR1) as HR1'; simp_sub in HR1'.
so (Htsubst _#6 (subof_bind_aof_tm _#4 Hs) (wf_tm Hwf Hoft) HR2) as HR2'; simp_sub in HR2'.
simp_sub.
eapply tcc_lett; simp_sub; eauto; done.

(* letm *)
intros G T k t M M' e e' HR1 Hfst HR2 H2 G1 s Hs Hwf.
so (ofm_reg _#4 Hwf (Hmreg _#5 Hwf HR1 andel)) as HofT.
so (Hmsubst _#7 Hs Hwf HR1) as HR1'; simp_sub in HR1'.
so (Htsubst _#6 (subof_bind_aof_cn_md _#5 Hs) (wf_md Hwf HofT Hfst) HR2) as HR2'; simp_sub in HR2'.
simp_sub.
eapply tcc_letm; simp_sub; try eassumption; simp_sub; eauto using subst_fsts, subst_ofc_type, subof_aof_sof; done.

(* ext  *)
intros G t M M' HR G1 s Hs Hwf.
so (Hmsubst _#7 Hs Hwf HR) as HR'; simp_sub in HR'.
simp_sub.
apply tcc_ext; auto; done.
Qed.


Lemma subst_mcc :
  forall Rk Rc Rs Rt Rm,
    sregular Rk Rc Rs
    -> cregular Rt Rm
    -> ssubstitutive Rk Rc Rs
    -> csubstitutive Rt Rm
    -> forall G1 G2 s p M M' T,
         subof aof G1 s G2
         -> wf G2
         -> mcc Rk Rc Rs Rt Rm G2 p M M' T
         -> mcc Rk Rc Rs Rt Rm G1 p (subst s M) (subst s M') (subst s T).
Proof.
intros Rk Rc Rs Rt Rm (Hkreg & Hcreg & Hsreg) (Htreg & Hmreg & _) (_ & Hcsubst & Hssubst) (Htsubst & Hmsubst) G1 G2 s p M M' T Hs Hwf Hmcc.
revert G1 s Hs Hwf.
#12 induct Hmcc;
try (intro G; intros;
     repeat
       (match goal with
        | H : Rm G _ _ _ _ |- _ =>
          let H' := fresh in so (Hmsubst _#7 Hs Hwf H) as H'; clear H; simp_sub in H'

        | H : eqs G _ _ |- _ =>
          let H' := fresh in so (Hssubst _#5 (subof_aof_sof _#3 Hs) H) as H'; clear H; simp_sub in H'
        end);
     simp_sub;
     eauto using mcc; done).


(* stat *)
intros G c c' k Heqc G1 s Hs Hwf.
simp_sub.
apply mcc_stat; [].
eapply Hcsubst; eauto using subof_aof_sof; done.

(* lamgn *)
intros G T1 T1' T2 k1 M M' H1 Hfst HR G1 s Hs Hwf.
simp_sub.
so (eqs_reg _#3 Hwf (Hsreg _#3 Hwf H1) andel) as HofT1.
so (Hmsubst _#7 (subof_bind_aof_cn_md' _#5 Hs) (wf_md Hwf HofT1 Hfst) HR) as HR'; simp_sub in HR'.
eapply mcc_lamgn; simp_sub; eauto using subof_aof_sof, subst_fsts; done.

(* appgn *)
intros G T1 T2 M1 M1' M2 M2' c2 HR1 HR2 H1 G1 s Hs Hwf.
so (Hmsubst _#7 Hs Hwf HR1) as HR1'; simp_sub in HR1'.
simp_sub.
rewrite -> split_dot_compose; [].
eapply mcc_appgn; eauto using subst_fstm, subof_aof_tof; done.

(* lamap *)
intros G T1 T1' T2 k1 M M' H1 Hfst HR G1 s Hs Hwf.
simp_sub.
so (eqs_reg _#3 Hwf (Hsreg _#3 Hwf H1) andel) as HofT1.
so (Hmsubst _#7 (subof_bind_aof_cn_md' _#5 Hs) (wf_md Hwf HofT1 Hfst) HR) as HR'; simp_sub in HR'.
eapply mcc_lamap; simp_sub; eauto using subof_aof_sof, subst_fsts; done.

(* appap *)
intros G p T1 T2 M1 M1' M2 M2' c2 HR1 H0 H1 G1 s Hs Hwf.
so (Hmsubst _#7 Hs Hwf HR1) as HR1'; simp_sub in HR1'.
simp_sub.
rewrite -> split_dot_compose; [].
eapply mcc_appap; eauto using subst_fstm, subof_aof_tof; done.

(* pair *)
intros G p T1 T2 M1 M1' M2 M2' H H0 G1 s Hs Hwf.
simp_sub.
rewrite -> subst_compose; [].
apply mcc_pair; eauto; done.

(* pi2 *)
intros G T1 T2 M M' c HR H0 G1 s Hs Hwf.
so (Hmsubst _#7 Hs Hwf HR) as HR'; simp_sub in HR'.
simp_sub.
rewrite -> split_dot_compose; [].
eapply mcc_pi2; eauto using subst_fstm, subof_aof_tof; done.

(* unpack *)
intros G k t T T' e e' M M' HR1 HR2 H1 G1 s Hs Hwf.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf (Htreg _#4 Hwf HR1 andel))) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
so (Htsubst _#6 Hs Hwf HR1) as HR1'; simp_sub in HR1'.
so (Hmsubst _#7 (subof_bind_aof_cn_tm _#5 Hs) (wf_tm (wf_cn Hwf Hofk) Hoft) HR2) as HR2'; simp_sub in HR2'.
simp_sub.
eapply mcc_unpack; simp_sub; eauto using subof_aof_sof; done.

(* lett *)
intros G p t T e e' M M' HR1 HR2 G1 s Hs Hwf.
so (oft_reg _#3 Hwf (Htreg _#4 Hwf HR1 andel)) as Hoft.
so (Hmsubst _#7 (subof_bind_aof_tm _#4 Hs) (wf_tm Hwf Hoft) HR2) as HR2'; simp_sub in HR2'.
simp_sub.
eapply mcc_lett; simp_sub; eauto; done.

(* letm *)
intros G T1 T2 T2' k1 M1 M1' M2 M2' HR1 Hfst HR2 H2 G1 s Hs Hwf.
so (ofm_reg _#4 Hwf (Hmreg _#5 Hwf HR1 andel)) as HofT.
so (Hmsubst _#7 (subof_bind_aof_cn_md' _#5 Hs) (wf_md Hwf HofT Hfst) HR2) as HR2'; simp_sub in HR2'.
simp_sub.
eapply mcc_letm; eauto; simp_sub; eauto using subst_fsts, subof_aof_sof; done.

(* seal *)
intros G T T' M M' H H0 G1 s Hs Hwf.
simp_sub.
eapply mcc_seal; eauto using subof_aof_sof; done.

(* subsume *)
intros G p T1 T2 M M' H H0 G1 s Hs Hwf.
eapply mcc_subsume; eauto using subst_subs, subof_aof_sof; done.
Qed.


Lemma tconsistency :
  forall Rt Rm,
    compatible samek samec sames Rt Rm
    -> consistent Rm
    -> forall e e' t,
         Rt nil e e' t
         -> thalt nil e
         -> thalt nil e'.
Proof.
intros Rt Rm Hcomp Hcons e e' t HR (v & Hsteps & Hval).
exploit (Hcons (md_dyn e) (md_dyn e') (sg_dyn t)) as H.
  apply Hcomp; auto using wf_nil; [].
  apply mcc_forget; [].
  apply Hcomp; auto using wf_nil; [].
  apply mcc_dyn; auto; done.

  exists (md_dyn v).
  split.
    apply (star_map _ _ (tstep nil) _  md_dyn); auto; [].
    intros; apply mstep_dyn; auto; done.

    apply mvalue_dyn; auto; done.
destruct H as (V & Hsteps' & Hval').
assert (exists v', V = md_dyn v' /\ star (tstep nil) e' v') as (v' & -> & Hsteps'').
  clear HR.
  remember (md_dyn e') as M eqn:Heq.
  revert e' Heq.
  #2 induct Hsteps'.
    intros M d ->.
    exists d.
    split; auto using star_refl; done.

    intros M1 M2 M3 Hstep _ IH d1 ->.
    invert Hstep; [].
    intros d2 Hstep' <-.
    so (IH _ (eq_refl _)) as (v' & Heq & Hstepsi).
    exists v'; split; auto; [].
    eapply star_step; eauto; done.

  invert Hval'; [].
  intros Hval''.
  exists v'.
  split; auto; done.
Qed.


(* Weak compatible closure *)

Inductive tccw
  (Rk : context -> (* kd *) term -> (* kd *) term -> Prop)
  (Rc : context -> (* cn *) term -> (* cn *) term -> (* kd *) term -> Prop)
  (Rs : context -> (* sg *) term -> (* sg *) term -> Prop)
  (Rt : context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop)
  (Rm : context -> purity -> (* md *) term -> (* md *) term -> (* sg *) term -> Prop) 
  : context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop
  :=
| tccw_lam {G t1 t1' t2 e e'}
    : Rc G t1 t1' kd_type
      -> Rt (G; cl_tm t1) e e' (shift1 t2)
      -> tccw Rk Rc Rs Rt Rm G (tm_lam t1 e) (tm_lam t1' e') (cn_arrow t1 t2)

| tccw_app1 {G t1 t2 e1 e1' e2}
    : Rt G e1 e1' (cn_arrow t1 t2)
       -> oft G e2 t1
       -> tccw Rk Rc Rs Rt Rm G (tm_app e1 e2) (tm_app e1' e2) t2

| tccw_app2 {G t1 t2 e1 e2 e2'}
    : oft G e1 (cn_arrow t1 t2)
       -> Rt G e2 e2' t1
       -> tccw Rk Rc Rs Rt Rm G (tm_app e1 e2) (tm_app e1 e2') t2

| tccw_pair1 {G t1 t2 e1 e1' e2}
    : Rt G e1 e1' t1
      -> oft G e2 t2
      -> tccw Rk Rc Rs Rt Rm G (tm_pair e1 e2) (tm_pair e1' e2) (cn_prod t1 t2)

| tccw_pair2 {G t1 t2 e1 e2 e2'}
    : oft G e1 t1
      -> Rt G e2 e2' t2
      -> tccw Rk Rc Rs Rt Rm G (tm_pair e1 e2) (tm_pair e1 e2') (cn_prod t1 t2)

| tccw_pi1 {G t1 t2 e e'}
    : Rt G e e' (cn_prod t1 t2)
      -> tccw Rk Rc Rs Rt Rm G (tm_pi1 e) (tm_pi1 e') t1

| tccw_pi2 {G t1 t2 e e'}
    : Rt G e e' (cn_prod t1 t2)
      -> tccw Rk Rc Rs Rt Rm G (tm_pi2 e) (tm_pi2 e') t2

| tccw_plam {G k k' t e e'}
    : Rk G k k'
      -> Rt (G; cl_cn k) e e' t
      -> tccw Rk Rc Rs Rt Rm G (tm_plam k e) (tm_plam k' e') (cn_all k t)

| tccw_papp {G k t e e' c c'}
    : Rt G e e' (cn_all k t)
      -> Rc G c c' k
      -> tccw Rk Rc Rs Rt Rm G (tm_papp e c) (tm_papp e' c') (subst1 c t)

| tccw_pack {G k k' t t' c c' e e'}
    : Rc G c c' k
      -> Rk G k k'
      -> Rt G e e' (subst1 c t)
      -> Rc (G; cl_cn k) t t' kd_type
      -> tccw Rk Rc Rs Rt Rm G (tm_pack c e k t) (tm_pack c' e' k' t') (cn_exists k t)

| tccw_unpack1 {G k t t' e1 e1' e2}
    : Rt G e1 e1' (cn_exists k t)
      -> oft (G; cl_cn k; cl_tm t) e2 (subst (sh 2) t')
      -> ofc G t' kd_type
      -> tccw Rk Rc Rs Rt Rm G (tm_unpack e1 e2) (tm_unpack e1' e2) t'

| tccw_unpack2 {G k t t' e1 e2 e2'}
    : oft G e1 (cn_exists k t)
      -> Rt (G; cl_cn k; cl_tm t) e2 e2' (subst (sh 2) t')
      -> ofc G t' kd_type
      -> tccw Rk Rc Rs Rt Rm G (tm_unpack e1 e2) (tm_unpack e1 e2') t'

| tccw_fix {G t t' e e'}
    : Rc G t t' kd_type
      -> Rt G e e' (cn_arrow (cn_arrow cn_unit t) t)
      -> tccw Rk Rc Rs Rt Rm G (tm_fix t e) (tm_fix t' e') t

| tccw_lett1 {G t1 t2 e1 e1' e2}
    : Rt G e1 e1' t1
      -> oft (G; cl_tm t1) e2 (shift1 t2)
      -> tccw Rk Rc Rs Rt Rm G (tm_lett e1 e2) (tm_lett e1' e2) t2

| tccw_lett2 {G t1 t2 e1 e2 e2'}
    : oft G e1 t1
      -> Rt (G; cl_tm t1) e2 e2' (shift1 t2)
      -> tccw Rk Rc Rs Rt Rm G (tm_lett e1 e2) (tm_lett e1 e2') t2

| tccw_equiv {G t1 t2 e e'}
    : Rt G e e' t1
      -> eqc G t1 t2 kd_type
      -> tccw Rk Rc Rs Rt Rm G e e' t2

| tccw_letm1 {G T k t M M' e}
    : Rm G impure M M' T
      -> fsts T k
      -> oft (G; cl_cn k; cl_md (shift1 T) (var 0)) e (subst (sh 2) t)
      -> ofc G t kd_type
      -> tccw Rk Rc Rs Rt Rm G (tm_letm M e) (tm_letm M' e) t

| tccw_letm2 {G T k t M e e'}
    : ofm G impure M T
      -> fsts T k
      -> Rt (G; cl_cn k; cl_md (shift1 T) (var 0)) e e' (subst (sh 2) t)
      -> ofc G t kd_type
      -> tccw Rk Rc Rs Rt Rm G (tm_letm M e) (tm_letm M e') t

| tccw_ext {G t M M'}
    : Rm G impure M M' (sg_dyn t)
      -> tccw Rk Rc Rs Rt Rm G (tm_ext M) (tm_ext M') t
.


Inductive mccw
  (Rk : context -> (* kd *) term -> (* kd *) term -> Prop)
  (Rc : context -> (* cn *) term -> (* cn *) term -> (* kd *) term -> Prop)
  (Rs : context -> (* sg *) term -> (* sg *) term -> Prop)
  (Rt : context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop)
  (Rm : context -> purity -> (* md *) term -> (* md *) term -> (* sg *) term -> Prop) 
  : context -> purity -> (* md *) term -> (* md *) term -> (* sg *) term -> Prop
  :=
| mccw_stat {G c c' k}
    : Rc G c c' k
       -> mccw Rk Rc Rs Rt Rm G pure (md_stat c) (md_stat c') (sg_stat k)

| mccw_dyn {G e e' t}
    : Rt G e e' t
      -> mccw Rk Rc Rs Rt Rm G pure (md_dyn e) (md_dyn e') (sg_dyn t)

| mccw_lamgn {G T1 T1' T2 k1 M M'}
    : Rs G T1 T1'
      -> fsts T1 k1
      -> Rm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M M' (shift1 T2)
      -> mccw Rk Rc Rs Rt Rm G pure (md_lamgn T1 M) (md_lamgn T1' M') (sg_pigen T1 T2)

| mccw_appgn1 {G T1 T2 M1 M1' M2 c2}
    : Rm G impure M1 M1' (sg_pigen T1 T2)
      -> ofm G pure M2 T1
      -> fstm G M2 c2
      -> mccw Rk Rc Rs Rt Rm G impure (md_appgn M1 M2) (md_appgn M1' M2) (subst1 c2 T2)

| mccw_appgn2 {G T1 T2 M1 M2 M2' c2}
    : ofm G impure M1 (sg_pigen T1 T2)
      -> Rm G pure M2 M2' T1
      -> fstm G M2 c2
      -> mccw Rk Rc Rs Rt Rm G impure (md_appgn M1 M2) (md_appgn M1 M2') (subst1 c2 T2)

| mccw_lamap {G T1 T1' T2 k1 M M'}
    : Rs G T1 T1'
      -> fsts T1 k1
      -> Rm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) pure M M' (shift1 T2)
      -> mccw Rk Rc Rs Rt Rm G pure (md_lamap T1 M) (md_lamap T1' M') (sg_piapp T1 T2)

| mccw_appap1 {G p T1 T2 M1 M1' M2 c2}
    : Rm G p M1 M1' (sg_piapp T1 T2)
      -> ofm G pure M2 T1
      -> fstm G M2 c2
      -> mccw Rk Rc Rs Rt Rm G p (md_appap M1 M2) (md_appap M1' M2) (subst1 c2 T2)

| mccw_appap2 {G p T1 T2 M1 M2 M2' c2}
    : ofm G p M1 (sg_piapp T1 T2)
      -> Rm G pure M2 M2' T1
      -> fstm G M2 c2
      -> mccw Rk Rc Rs Rt Rm G p (md_appap M1 M2) (md_appap M1 M2') (subst1 c2 T2)

| mccw_pair1 {G p T1 T2 M1 M1' M2}
    : Rm G p M1 M1' T1
      -> ofm G p M2 T2
      -> mccw Rk Rc Rs Rt Rm G p (md_pair M1 M2) (md_pair M1' M2) (sg_sigma T1 (shift1 T2))

| mccw_pair2 {G p T1 T2 M1 M2 M2'}
    : ofm G p M1 T1
      -> Rm G p M2 M2' T2
      -> mccw Rk Rc Rs Rt Rm G p (md_pair M1 M2) (md_pair M1 M2') (sg_sigma T1 (shift1 T2))

| mccw_pi1 {G T1 T2 M M'}
    : Rm G pure M M' (sg_sigma T1 T2)
      -> mccw Rk Rc Rs Rt Rm G pure (md_pi1 M) (md_pi1 M') T1

| mccw_pi2 {G T1 T2 M M' c}
    : Rm G pure M M' (sg_sigma T1 T2)
      -> fstm G M c
      -> mccw Rk Rc Rs Rt Rm G pure (md_pi2 M) (md_pi2 M') (subst1 (cn_pi1 c) T2)

| mccw_unpack1 {G k t T T' e e' M}
    : Rt G e e' (cn_exists k t)
      -> ofm (G; cl_cn k; cl_tm t) impure M (subst (sh 2) T)
      -> Rs G T T'
      -> mccw Rk Rc Rs Rt Rm G impure (md_unpack e M T) (md_unpack e' M T') T

| mccw_unpack2 {G k t T T' e M M'}
    : oft G e (cn_exists k t)
      -> Rm (G; cl_cn k; cl_tm t) impure M M' (subst (sh 2) T)
      -> Rs G T T'
      -> mccw Rk Rc Rs Rt Rm G impure (md_unpack e M T) (md_unpack e M' T') T

| mccw_lett1 {G p t T e e' M}
    : Rt G e e' t
      -> ofm (G; cl_tm t) p M (shift1 T)
      -> mccw Rk Rc Rs Rt Rm G p (md_lett e M) (md_lett e' M) T

| mccw_lett2 {G p t T e M M'}
    : oft G e t
      -> Rm (G; cl_tm t) p M M' (shift1 T)
      -> mccw Rk Rc Rs Rt Rm G p (md_lett e M) (md_lett e M') T

| mccw_letm1 {G T1 T2 T2' k1 M1 M1' M2}
    : Rm G impure M1 M1' T1
      -> fsts T1 k1
      -> ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M2 (subst (sh 2) T2)
      -> Rs G T2 T2'
      -> mccw Rk Rc Rs Rt Rm G impure (md_letm M1 M2 T2) (md_letm M1' M2 T2') T2

| mccw_letm2 {G T1 T2 T2' k1 M1 M2 M2'}
    : ofm G impure M1 T1
      -> fsts T1 k1
      -> Rm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M2 M2' (subst (sh 2) T2)
      -> Rs G T2 T2'
      -> mccw Rk Rc Rs Rt Rm G impure (md_letm M1 M2 T2) (md_letm M1 M2' T2') T2

| mccw_seal {G T T' M M'}
    : Rm G impure M M' T
      -> Rs G T T'
      -> mccw Rk Rc Rs Rt Rm G impure (md_seal M T) (md_seal M' T') T

| mccw_forget {G T M M'}
    : Rm G pure M M' T
      -> mccw Rk Rc Rs Rt Rm G impure M M' T

| mccw_subsume {G p T1 T2 M M'}
    : Rm G p M M' T1
      -> subs G T1 T2
      -> mccw Rk Rc Rs Rt Rm G p M M' T2
.


Definition weakly_compatible
  (Rk : context -> (* kd *) term -> (* kd *) term -> Prop)
  (Rc : context -> (* cn *) term -> (* cn *) term -> (* kd *) term -> Prop)
  (Rs : context -> (* sg *) term -> (* sg *) term -> Prop)
  (Rt : context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop)
  (Rm : context -> purity -> (* md *) term -> (* md *) term -> (* cn *) term -> Prop) 
  :=
  (forall G e e' t, wf G -> tccw Rk Rc Rs Rt Rm G e e' t -> Rt G e e' t)
  /\
  (forall G p M M' T, wf G -> mccw Rk Rc Rs Rt Rm G p M M' T -> Rm G p M M' T)
.


Lemma weaken_compatible :
  forall Rk Rc Rs Rt Rm,
    creflexive Rt Rm
    -> compatible Rk Rc Rs Rt Rm
    -> weakly_compatible Rk Rc Rs Rt Rm.
Proof.
intros Rk Rc Rs Rt Rm (Hrefl & Hmrefl) (Htcomp & Hmcomp).
split.
  intros G e e' t Hwf Htcc.
  revert Hwf.
  induct Htcc; eauto using tcc; done.

  intros G p M M' T Hwf Hmcc.
  revert Hwf.
  induct Hmcc; eauto using mcc; done.
Qed.


(* Regularization *)

Definition treg (Rt : treln) (G : context) (e e' : (* tm *) term) (t : (* cn *) term)
  :=
  oft G e t
  /\ oft G e' t
  /\ Rt G e e' t.


Definition mreg (Rm : mreln) (G : context) (p : purity) (M M' : (* md *) term) (T : (* sg *) term)
  :=
  ofm G p M T
  /\ ofm G p M' T
  /\ Rm G p M M' T
  /\ (p = pure
        -> wf G
        -> exists k c c',
             fsts T k
             /\ fstm G M c
             /\ fstm G M' c'
             /\ eqc G c c' k).


Lemma treg_intro :
  forall (Rt : treln) G e e' t,
    oft G e t
    -> oft G e' t
    -> Rt G e e' t
    -> treg Rt G e e' t.
Proof.
intros Rt G e1 e2 t Hof1 Hof2 HR.
do2 2 split; auto.
Qed.


Lemma mreg_intro_impure :
  forall (Rm : mreln) G M M' T,
    ofm G impure M T
    -> ofm G impure M' T
    -> Rm G impure M M' T
    -> mreg Rm G impure M M' T.
Proof.
intros Rt G M1 M2 T Hof1 Hof2 HR.
do2 3 split; eauto; [].
intro Heq; discriminate Heq.
Qed.


Lemma mreg_intro_pure :
  forall (Rm : mreln) G M M' T c c' k,
    ofm G pure M T
    -> ofm G pure M' T
    -> fsts T k
    -> fstm G M c
    -> fstm G M' c'
    -> eqc G c c' k
    -> Rm G pure M M' T
    -> mreg Rm G pure M M' T.
Proof.
intros Rt G M1 M2 T c c' k Hof1 Hof2 Hfst Hfst1 Hfst2 Heqc HR.
do2 3 split; eauto; [].
intros _ Hwf.
do 3 eexists; do2 3 split; eauto using subst_fsts, subst_fstm, subst_eqc, subof_aof_tof, subof_aof_sof.
Qed.


Lemma treg_elim :
  forall Rt G e e' t,
    wf G
    -> treg Rt G e e' t
    -> Rt G e e' t.
Proof.
intros Rt G e e' t Hwf HR.
destruct HR as (_ & _ & HR').
assumption.
Qed.


Lemma mreg_elim :
  forall Rm G p M M' T,
    wf G
    -> mreg Rm G p M M' T
    -> Rm G p M M' T.
Proof.
intros Rm G p M M' T Hwf HR.
so HR as (_ & _ & HR' & _).
assumption.
Qed.


Lemma treg_reg :
  forall Rt G e e' t,
    wf G
    -> treg Rt G e e' t
    -> oft G e t /\ oft G e' t.
Proof.
intros Rt G e e' t Hwf Hreg.
so Hreg as (H & H' & _).
auto.
Qed.


Lemma mreg_reg :
  forall Rm G p M M' T,
    wf G
    -> mreg Rm G p M M' T
    -> ofm G p M T /\ ofm G p M' T.
Proof.
intros Rt G p M M' T Hwf Hreg.
so Hreg as (H & H' & _).
auto.
Qed.


Lemma mreg_fst :
  forall Rm G M M' T,
    wf G
    -> mreg Rm G pure M M' T
    -> exists k c c',
         fsts T k
         /\ fstm G M c
         /\ fstm G M' c'
         /\ eqc G c c' k.
Proof.
intros Rm G M M' T Hwf HR.
so HR as (_ & _ & _ & H).
exact (H (eq_refl _) Hwf).
Qed.


Lemma reg_cregular :
  forall Rt Rm, cregular (treg Rt) (mreg Rm).
Proof.
intros Rt Rm.
do2 2 split.
  intros G e e' t Hwf Hreg.
  eapply treg_reg; eauto; done.

  intros G p M M' T Hwf Hreg.
  eapply mreg_reg; eauto; done.

  intros G M M' T Hwf Hreg.
  eapply mreg_fst; eauto; done.
Qed.


Lemma reg_creflexive :
  forall Rt Rm,
    creflexive Rt Rm
    -> creflexive (treg Rt) (mreg Rm).
Proof.
intros Rt Rm (Htrefl & Hmrefl).
split.
  intros G e t Hof.
  do2 2 split; eauto; done.

  intros G p M T Hof.
  do2 3 split; auto; [].
  intros -> Hwf.
  so (ofm_reg _#4 Hwf Hof) as HofT.
  so (can_fsts _ _ HofT) as (k & Hfst).
  so (Preservation.can_fstm _#3 Hwf Hof) as (c & Hfstm).
  exists k, c, c.
  do2 3 split; auto; [].
  apply eqc_refl; [].
  eapply ofm_fst; eauto; done.
Qed.


Lemma reg_ctransitive :
  forall Rt Rm,
    ctransitive Rt Rm
    -> ctransitive (treg Rt) (mreg Rm).
Proof.
intros Rt Rm (Httrans & Hmtrans).
split.
  intros G t e1 e2 e3 H12 H23.
  destruct (H12) as (Hof & _ & Hrel12).
  destruct (H23) as (_ & Hof' & Hrel23).
  do2 2 split; eauto; done.

  intros G p T M1 M2 M3 H12 H23.
  destruct (H12) as (Hof & _ & Heqfst12 & Hrel12).
  destruct (H23) as (_ & Hof' & Heqfst23 & Hrel23).
  do2 3 split; eauto; [].
  intros -> Hwf.
  so (Hrel12 (eq_refl _) Hwf) as (k & c1 & c2 & Hfst & Hfst1 & Hfst2 & Heq12).
  so (Hrel23 (eq_refl _) Hwf) as (k' & c2' & c3 & Hfst' & Hfst2' & Hfst3 & Heq23).
  so (fsts_fun _#3 Hfst Hfst'); subst k'.
  so (fstm_fun _#4 Hfst2 Hfst2'); subst c2'.
  do 3 eexists; eauto with static.
Qed.


Lemma reg_csubstitutive :
  forall Rt Rm,
    csubstitutive Rt Rm
    -> csubstitutive (treg Rt) (mreg Rm).
Proof.
intros Rt Rm (Htsubst & Hmsubst).
split.
  intros G1 s G2 e e' t Hs Hwf HR.
  destruct (HR) as (Hof & Hof' & Hrel).
  do2 2 split; eauto using subst_oft, subof_aof_dof; done.
    
  intros G1 s G2 p M M' T Hs Hwf HR.
  destruct HR as (Hof & Hof' & Hrel & Hrelfst).
  do2 3 split; eauto using subst_ofm, subof_aof_dof; [].
  intros -> Hwf'.
  so (Hrelfst (eq_refl _) Hwf) as (k & c & c' & Hfst & Hfstm & Hfstm' & Heqc).
  exists (subst s k), (subst s c), (subst s c').
  do2 3 split; eauto using subst_fsts, subst_fstm, subst_eqc, subof_aof_tof, subof_aof_sof; done.
Qed.


Lemma reg_compatible :
  forall Rk Rc Rs Rt Rm,
    sregular Rk Rc Rs
    -> ssubstitutive Rk Rc Rs
    -> csubstitutive Rt Rm
    -> compatible Rk Rc Rs Rt Rm
    -> compatible Rk Rc Rs (treg Rt) (mreg Rm).
Proof.
intros Rk Rc Rs Rt Rm (Hregk & Hregc & Hregs) Hssubst Hsubst (Htcomp & Hmcomp).
split.

Local Ltac destreg :=
  repeat
    (match goal with
     | H : treg _ ?G _ _ _ |- _ =>
       let Hwf' := fresh
       in
         assert (wf G) as Hwf' by (eauto using wf);
         so (treg_reg _#5 Hwf' H) as (? & ?);
         so (treg_elim _#5 Hwf' H);
         clear H

     | H : mreg _ ?G _ _ _ _ |- _ =>
       let Hwf' := fresh
       in
         assert (wf G) as Hwf' by (eauto using wf);
         so (mreg_reg _#6 Hwf' H) as (? & ?);
         so (mreg_elim _#6 Hwf' H);
         clear H
     end).

Local Ltac dotreg Htcomp :=
  destreg; eapply treg_intro; eauto with static; apply Htcomp; eauto using tcc.

intros G e e' t Hwf Htcc.
revert Hwf.
#8 induct Htcc;
try (intros; dotreg Htcomp; done).

(* lam *)
intros G t1 t1' t2 e e' HR1 HR2 Hwf.
so (Hregc _#4 Hwf HR1) as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (Hof1 & Hof1' & _).
so (treg_reg _#5 (wf_tm Hwf Hof1) HR2) as (Hofe & Hofe').
#2 eapply treg_intro; eauto with static.
  #2 apply (@oft_equiv _ (cn_arrow t1' t2)).
    apply oft_lam; auto; [].
    eapply oft_conv_0t; eauto; done.

    apply eqc_symm; [].
    apply eqc_arrow; auto; [].
    apply eqc_refl; [].
    so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 (wf_tm Hwf Hof1) Hofe)) as H; simp_sub in H.
    assumption.

  eapply Htcomp; eauto; apply tcc_lam; auto; [].
  eapply treg_elim; eauto using wf; done.

(* plam *)
intros G k k' t e e' HR1 HR2 Hwf.
so (Hregk _#3 Hwf HR1) as Heqk.
so (eqk_reg _#3 Hwf Heqk) as (Hof1 & Hof1').
so (treg_reg _#5 (wf_cn Hwf Hof1) HR2) as (Hofe & Hofe').
#2 eapply treg_intro; eauto with static.
  #2 apply (@oft_equiv _ (cn_all k' t)).
    apply oft_plam; auto; [].
    eapply oft_conv_0c; eauto; done.

    apply eqc_symm; [].
    apply eqc_all; auto; [].
    apply eqc_refl; auto.
    exact (oft_reg _#3 (wf_cn Hwf Hof1) Hofe).

  eapply Htcomp; eauto; apply tcc_plam; auto; [].
  eapply treg_elim; eauto using wf; done.

(* papp *)
intros G k t e e' c c' HR1 HR2 Hwf.
so (Hregc _#4 Hwf HR2) as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & _).
so (treg_reg _#5 Hwf HR1) as (Hofe & Hofe').
#2 eapply treg_intro; eauto with static.
  #2 apply (@oft_equiv _ (subst1 c' t)).
    eauto with static; done.

    apply eqc_symm; [].
    eapply funct_cn_type; eauto; [].
    so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hofe)) as (kk & Hofcr).
    invert Hofcr; [].
    intros _ Hoft _; clear kk Hofcr.
    assumption.

  apply Htcomp; eauto; eapply tcc_papp; eauto; [].
  eapply treg_elim; eauto; done.

(* pack *)
intros G k k' t t' c c' e e' HRc HRk HRe HRt Hwf.
so (Hregc _#4 Hwf HRc) as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & Hofk).
so (Hregc _#4 (wf_cn Hwf Hofk) HRt) as Heqt.
so (eqc_reg _#4 (wf_cn Hwf Hofk) Heqt) as (Hoft & Hoft' & _).
so (treg_reg _#5 Hwf HRe) as (Hofe & Hofe').
so (Hregk _#3 Hwf HRk) as Heqk.
#2 eapply treg_intro; eauto with static.
  apply (@oft_equiv _ (cn_exists k' t')); eauto with static; [].
  #2 eapply oft_pack; eauto with static.
    eapply oft_equiv; eauto; [].
    replsub kd_type with (subst1 c kd_type).
    eapply funct_eqc; eauto; done.

    so (subst_ofc _#5 (subof_conv_0 _#3 (eqk_symm Heqk)) Hoft') as H; simp_sub in H.
    assumption.

  apply Htcomp; eauto; apply tcc_pack; auto; [].
  eapply treg_elim; eauto; done.

(* unpack *)
intros G k t t' e1 e1' e2 e2' HR1 HR2 Hoft' Hwf.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf (treg_reg _#5 Hwf HR1 andel))) as (l & Hofcr).
invert Hofcr; intros; [].
dotreg Htcomp; done.

(* fix *)
intros G t t' e e' HRt HRe Hwf.
so (Hregc _#4 Hwf HRt) as Heqt.
so (eqc_reg _#4 Hwf Heqt) as (Hoft & Hoft' & _).
so (treg_reg _#5 Hwf HRe) as (Hofe & Hofe').
#2 eapply treg_intro; eauto with static.
  apply (@oft_equiv _ t'); eauto with static; [].
  apply oft_fix; auto; [].
  eapply oft_equiv; eauto with static; done.

  apply Htcomp; eauto; apply tcc_fix; auto; [].
  eapply treg_elim; eauto; done.

(* lett *)
intros G t1 t2 e1 e1' e2 e2' HR1 HR2 Hwf.
so (oft_reg _#3 Hwf (treg_reg _#5 Hwf HR1 andel)).
dotreg Htcomp; done.

(* letm *)
intros G T k t M M' e e' HR1 Hfst HR2 Hoft Hwf.
so (mreg_reg _#6 Hwf HR1) as (HofM & HofM').
so (wf_md Hwf (ofm_reg _#4 Hwf HofM) Hfst) as Hwf'.
so (treg_reg _#5 Hwf' HR2) as (? & ?).
so (mreg_elim _#6 Hwf HR1).
so (treg_elim _#5 Hwf' HR2).
eapply treg_intro; eauto with static; apply Htcomp; eauto using tcc; done.



intros G p e e' t Hwf Hmcc.
revert Hwf.
induct Hmcc.

(* stat *)
intros G c c' k HRc Hwf.
so (Hregc _#4 Hwf HRc) as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & _).
eapply mreg_intro_pure; eauto with static; [].
apply Hmcomp; eauto. apply mcc_stat; auto; done.

(* dyn *)
intros G e e' t H Hwf.
destreg.
eapply mreg_intro_pure; eauto with static; apply Hmcomp; eauto using mcc; done.

(* lamgn *)
intros G T1 T1' T2 k1 M M' HRs Hfst HRm Hwf.
so (Hregs _#3 Hwf HRs) as Heqs.
so (eqs_reg _#3 Hwf Heqs) as (Hofs & Hofs').
so (mreg_reg _#6 (wf_md Hwf Hofs Hfst) HRm) as (Hofm & Hofm').
so (can_fsts _ _ Hofs') as (k1' & Hfst').
so (eqs_fst _#5 Hwf Heqs Hfst Hfst') as Heqk.
#2 eapply mreg_intro_pure; eauto with static.
  #2 apply (@ofm_subsume _ _ (sg_pigen T1' T2)).
    eapply ofm_lamgn; eauto; [].
    eapply ofm_conv_01m; eauto; [].
    apply weaken1_eqs; auto; done.

    apply subs_refl; [].
    apply eqs_symm; [].
    eapply eqs_pigen; eauto; [].
    so (subst_ofs _#4 (unshift_md _#3) (ofm_reg _#4 (wf_md Hwf Hofs Hfst) Hofm)) as H; simp_sub in H.
    apply eqs_refl; assumption.

  apply Hmcomp; eauto; eapply mcc_lamgn; eauto; [].
  eapply mreg_elim; eauto using wf; done.

(* appgn *)
intros G T1 T2 M1 M1' M2 M2' c2 HR1 HR2 HfstM2 Hwf.
so (mreg_reg _#6 Hwf HR1) as (HofM1 & HofM1').
so (mreg_reg _#6 Hwf HR2) as (HofM2 & HofM2').
so (mreg_fst _#5 Hwf HR2) as (k1 & c2a & c2' & HfstT1 & HfstM2a & HfstM2' & Heqc).
so (mreg_elim _#6 Hwf HR1) as HR1'.
so (mreg_elim _#6 Hwf HR2) as HR2'.
so (fstm_fun _#4 HfstM2 HfstM2a); subst c2a.
#2 eapply mreg_intro_impure; eauto with static.
  so (ofm_reg _#4 Hwf HofM1) as Hofpi.
  invert Hofpi; [].
  intros k1a HofT1 HfstT1a HofT2.
  so (fsts_fun _#3 HfstT1 HfstT1a); subst k1a.
  #2 apply (@ofm_subsume _ _ (subst1 c2' T2)).
    eapply ofm_appgn; eauto; done.

    apply subs_refl; apply eqs_symm; [].
    eapply funct_sg; eauto; done.

  apply Hmcomp; eauto using mcc; done.

(* lamap *)
intros G T1 T1' T2 k1 M M' HRs Hfst HRm Hwf.
so (Hregs _#3 Hwf HRs) as Heqs.
so (eqs_reg _#3 Hwf Heqs) as (Hofs & Hofs').
so (mreg_reg _#6 (wf_md Hwf Hofs Hfst) HRm) as (Hofm & Hofm').
so (can_fsts _ _ Hofs') as (k1' & Hfst').
so (eqs_fst _#5 Hwf Heqs Hfst Hfst') as Heqk.
so (mreg_fst _#5 (wf_md Hwf Hofs Hfst) HRm) as (k2w & cw & cw' & HfstT2w & Hfstm & Hfstm' & Heqcw).
so (subordf_md_kd' _#4 (ofk_fkd _ _ (eqc_reg _#4 (wf_md Hwf Hofs Hfst) Heqcw ander ander))).
forget (subst1 nonsense k2w) as k2; subst k2w.
so (subst_fsts _ _ (dot nonsense id) HfstT2w) as HfstT2; simp_sub in HfstT2.
so (subordf_md_cn' _#4 (ofc_fcn _#3 (eqc_reg _#4 (wf_md Hwf Hofs Hfst) Heqcw andel))).
forget (subst1 nonsense cw) as c; subst cw.
so (subordf_md_cn' _#4 (ofc_fcn _#3 (eqc_reg _#4 (wf_md Hwf Hofs Hfst) Heqcw ander andel))).
forget (subst1 nonsense cw') as c'; subst cw'.
so (subst_eqc _#6 (unshift_md _#3) Heqcw) as Heqc; simp_sub in Heqc.
clear HfstT2w Heqcw.
#3 apply (mreg_intro_pure _#5 (cn_lam k1 c) (cn_lam k1' c') (kd_pi k1 k2)); eauto with static.
  #2 apply (@ofm_subsume _ _ (sg_piapp T1' T2)).
    eapply ofm_lamap; eauto; [].
    eapply ofm_conv_01m; eauto; [].
    apply weaken1_eqs; auto; done.

    apply subs_refl; [].
    apply eqs_symm; [].
    eapply eqs_piapp; eauto; [].
    so (subst_ofs _#4 (unshift_md _#3) (ofm_reg _#4 (wf_md Hwf Hofs Hfst) Hofm)) as H; simp_sub in H.
    apply eqs_refl; assumption.

  apply fstm_lamap; auto; [].
  so (subst_fstm _#5 (subof_aof_tof _#3 (subof_conv_01m _#5 (eqk_symm Heqk) (weaken1_eqs _#4 (eqs_symm Heqs)))) Hfstm') as H.
  simp_sub in H.
  assumption.

  apply Hmcomp; eauto; eapply mcc_lamap; eauto; [].
  eapply mreg_elim; eauto using wf; done.

(* appap *)
intros G p T1 T2 M1 M1' M2 M2' c2 HR1 HR2 HfstM2 Hwf.
so (mreg_reg _#6 Hwf HR1) as (HofM1 & HofM1').
so (mreg_reg _#6 Hwf HR2) as (HofM2 & HofM2').
so (mreg_fst _#5 Hwf HR2) as (k1 & c2a & c2' & HfstT1 & HfstM2a & HfstM2' & Heqc).
so (mreg_elim _#6 Hwf HR1) as HR1'.
so (mreg_elim _#6 Hwf HR2) as HR2'.
so (fstm_fun _#4 HfstM2 HfstM2a); subst c2a.
assert (ofm G p (md_appap M1' M2') (subst1 c2 T2)) as Hofapp.
  so (ofm_reg _#4 Hwf HofM1) as Hofpi.
  invert Hofpi; [].
  intros k1a HofT1 HfstT1a HofT2.
  so (fsts_fun _#3 HfstT1 HfstT1a); subst k1a.
  #2 apply (@ofm_subsume _ _ (subst1 c2' T2)).
    eapply ofm_appap; eauto; done.

    apply subs_refl; apply eqs_symm; [].
    eapply funct_sg; eauto; done.
#2 destruct p.
  so (mreg_fst _#5 Hwf HR1) as (kapp & c1 & c1' & Hfstapp & HfstM1 & HfstM1' & Heqc1).
  invert Hfstapp; [].
  intros k1a k2 HfstT1a HfstT2 <-.
  so (fsts_fun _#3 HfstT1 HfstT1a); subst k1a.
  eapply (mreg_intro_pure _#5 (cn_app c1 c2) (cn_app c1' c2') (subst1 c2 k2)); eauto using subst_fsts with static; [].
  apply Hmcomp; eauto using mcc; done.

  eapply mreg_intro_impure; eauto with static; [].
  apply Hmcomp; eauto using mcc; done.

(* pair *)
intros G p T1 T2 M1 M1' M2 M2' HR1 HR2 Hwf.
so (mreg_reg _#6 Hwf HR1) as (HofM1 & HofM1').
so (mreg_reg _#6 Hwf HR2) as (HofM2 & HofM2').
so (mreg_elim _#6 Hwf HR1) as HR1'.
so (mreg_elim _#6 Hwf HR2) as HR2'.
#2 destruct p.
  so (mreg_fst _#5 Hwf HR1) as (k1 & c1 & c1' & HfstT1 & HfstM1 & HfstM1' & Heqc1).
  so (mreg_fst _#5 Hwf HR2) as (k2 & c2 & c2' & HfstT2 & HfstM2 & HfstM2' & Heqc2).
  #7 eapply (mreg_intro_pure _#5 (cn_pair c1 c2) (cn_pair c1' c2') (kd_sigma k1 (shift1 k2))).
    eauto with static; done.

    eauto with static; done.

    eauto using subst_fsts with static; done.

    apply fstm_pair; auto; done.

    apply fstm_pair; auto; done.

    apply eqc_pair; simp_sub; auto; [].
    apply weaken1_ofk; [].
    eapply eqc_reg; eauto; done.

    apply Hmcomp; eauto using mcc; done.

  eapply mreg_intro_impure; eauto with static; [].
  apply Hmcomp; eauto using mcc; done.

(* pi1 *)
intros G T1 T2 M M' HR Hwf.
so (mreg_reg _#6 Hwf HR) as (HofM & HofM').
so (mreg_elim _#6 Hwf HR) as HR'.
so (mreg_fst _#5 Hwf HR) as (ksigma & c & c' & Hfstsigma & HfstM & HfstM' & Heqc).
invert Hfstsigma; [].
intros k1 k2 HfstT1 HfstT2 <-.
eapply (mreg_intro_pure _#5 (cn_pi1 c) (cn_pi1 c') k1); eauto with static; [].
apply Hmcomp; eauto using mcc; done.

(* pi2 *)
intros G T1 T2 M M' c HR HfstM Hwf.
so (mreg_reg _#6 Hwf HR) as (HofM & HofM').
so (mreg_elim _#6 Hwf HR) as HR'.
so (mreg_fst _#5 Hwf HR) as (ksigma & ca & c' & Hfstsigma & HfstMa & HfstM' & Heqc).
invert Hfstsigma; [].
intros k1 k2 HfstT1 HfstT2 <-.
so (fstm_fun _#4 HfstM HfstMa); subst ca.
#2 eapply (mreg_intro_pure _#5 (cn_pi2 c) (cn_pi2 c') (subst1 (cn_pi1 c) k2)); eauto using subst_fsts with static.
  #2 apply (@ofm_subsume _ _ (subst1 (cn_pi1 c') T2)).
    eauto with static; done.

    apply subs_refl; apply eqs_symm; [].
    so (ofm_reg _#4 Hwf HofM) as Hofsigma.
    invert Hofsigma; [].
    intros k1a _ HfstT1a HofT2.
    so (fsts_fun _#3 HfstT1 HfstT1a); subst k1a.
    eapply funct_sg; eauto with static; done.

  apply Hmcomp; eauto using mcc; done.

(* unpack *)
intros G k t T T' e e' M M' HR1 HR2 HRs Hwf.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf (treg_reg _#5 Hwf HR1 andel))) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
clear Hofcr l.
so (Hregs _#3 Hwf HRs) as Heqs.
so (eqs_reg _#3 Hwf Heqs) as (HofT & HofT').
so (treg_reg _#5 Hwf HR1) as (Hofe & Hofe').
so (mreg_reg _#6 (wf_tm (wf_cn Hwf Hofk) Hoft) HR2) as (HofM & HofM').
#2 eapply mreg_intro_impure; eauto with static.
  #2 apply (@ofm_subsume _ _ T').
    eapply ofm_unpack; eauto; [].
    eapply ofm_subsume; eauto; [].
    refine (subst_subs _#5 (weaken2 _#4) _); eauto with static; done.

    apply subs_refl; [].
    apply eqs_symm; auto; done.

  so (treg_elim _#5 Hwf HR1).
  so (mreg_elim _#6 (wf_tm (wf_cn Hwf Hofk) Hoft) HR2).
  apply Hmcomp; eauto; eapply mcc_unpack; eauto; done.

(* lett *)
intros G p t T e e' M M' HR1 HR2 Hwf.
so (treg_reg _#5 Hwf HR1) as (Hofe & Hofe').
so (treg_elim _#5 Hwf HR1) as HR1'.
so (oft_reg _#3 Hwf Hofe) as Hoft.
so (wf_tm Hwf Hoft) as Hwf'.
so (mreg_reg _#6 Hwf' HR2) as (HofM & HofM').
so (mreg_elim _#6 Hwf' HR2) as HR2'.
#2 destruct p.
  so (mreg_fst _#5 Hwf' HR2) as (k & c & c' & HfstT1w & HfstM & HfstM' & Heqc).
  so (subst_fsts _ _ (dot nonsense id) HfstT1w) as HfstT1.
  simp_sub in HfstT1.
  #6 eapply (mreg_intro_pure _#5 (subst1 nonsense c) (subst1 nonsense c') (subst1 nonsense k)); eauto.
    eauto with static; done.

    eauto with static; done.

    apply fstm_lett; [].
    simp_sub.
    so (ofm_fst _#6 Hwf' HofM HfstM HfstT1w).
    erewrite <- subord_tm_cn; eauto; [].
    rewrite <- (subst_dot0_sh c); rewrite <- (subst_dot0_sh M); [].
    eapply subst_fstm; eauto; [].
    apply subof_dot; auto using subof_sh with sequence; [].
    simp_sub.
    apply tof_tm; done.

    apply fstm_lett; [].
    simp_sub.
    so (ofm_fst _#6 Hwf' HofM' HfstM' HfstT1w).
    erewrite <- subord_tm_cn; eauto; [].
    rewrite <- (subst_dot0_sh c'); rewrite <- (subst_dot0_sh M'); [].
    eapply subst_fstm; eauto; [].
    apply subof_dot; auto using subof_sh with sequence; [].
    simp_sub.
    apply tof_tm; done.

    eapply subst_eqc; eauto using unshift_tm; done.

    apply Hmcomp; eauto using mcc; done.

  eapply mreg_intro_impure; eauto with static; apply Hmcomp; eauto using mcc; done.

(* letm *)
intros G T1 T2 T2' k1 M1 M1' M2 M2' HR1 Hfst HR2 HRs Hwf.
so (mreg_reg _#6 Hwf HR1) as (HofM & HofM').
so (wf_md Hwf (ofm_reg _#4 Hwf HofM) Hfst) as Hwf'.
so (mreg_reg _#6 Hwf' HR2) as (? & ?).
so (mreg_elim _#6 Hwf HR1).
so (mreg_elim _#6 Hwf' HR2).
so (Hregs _#3 Hwf HRs) as Heqs.
so (eqs_reg _#3 Hwf Heqs) as (Hofs & Hofs').
#2 eapply mreg_intro_impure; eauto with static.
  #2 apply (@ofm_subsume _ _ T2').
    eapply ofm_letm; eauto; [].
    eapply ofm_subsume; eauto; [].
    refine (subst_subs _#5 (weaken2 _#4) _); eauto with static; done.

    apply subs_refl; apply eqs_symm; assumption.

  apply Hmcomp; eauto; eapply mcc_letm; eauto; done.

(* seal *)
intros G T T' M M' HRm HRs Hwf.
so (Hregs _#3 Hwf HRs) as Heqs.
so (eqs_reg _#3 Hwf Heqs) as (Hofs & Hofs').
so (mreg_reg _#6 Hwf HRm) as (Hofm & Hofm').
#2 eapply mreg_intro_impure; eauto with static.
  #2 apply (@ofm_subsume _ _ T').
    eapply ofm_seal; eauto; [].
    eapply ofm_subsume; eauto; [].
    apply subs_refl; assumption.

    apply subs_refl; apply eqs_symm; assumption.

  apply Hmcomp; eauto; apply mcc_seal; auto; [].
  eapply mreg_elim; eauto; done.

(* forget *)
intros; destreg.
eapply mreg_intro_impure; eauto with static; apply Hmcomp; eauto using mcc; done.

(* subsume *)
intros G p T1 T2 M M' HR Hsubs Hwf.
so (mreg_reg _#6 Hwf HR) as (HofM & HofM').
so (mreg_elim _#6 Hwf HR) as HR'.
#2 destruct p.
  so (mreg_fst _#5 Hwf HR) as (k1 & c & c' & Hfst1 & HfstM & HfstM' & Heqc).
  so (can_fsts _ _ (subs_reg _#3 Hwf Hsubs ander)) as (k2 & Hfst2).
  so (subs_fst _#5 Hwf Hsubs Hfst1 Hfst2).
  eapply mreg_intro_pure; eauto with static; apply Hmcomp; eauto using mcc; done.

  eapply mreg_intro_impure; eauto with static; apply Hmcomp; eauto using mcc; done.
Qed.


Lemma reg_quasicongruence :
  forall Rt Rm,
    creflexive Rt Rm
    -> csubstitutive Rt Rm
    -> compatible samek samec sames Rt Rm
    -> quasicongruence (treg Rt) (mreg Rm).
Proof.
intros Rt Rm Hrefl Hsub Hcomp.
do2 3 split;
auto using reg_cregular, reg_creflexive, reg_csubstitutive, reg_compatible, same_sregular, same_ssubstitutive.
Qed.


Lemma reg_semicongruence :
  forall Rt Rm,
    creflexive Rt Rm
    -> ctransitive Rt Rm
    -> csubstitutive Rt Rm
    -> compatible samek samec sames Rt Rm
    -> semicongruence (treg Rt) (mreg Rm).
Proof.
intros Rt Rm Hrefl Htrans Hsub Hcomp.
do2 4 split;
auto using reg_cregular, reg_creflexive, reg_ctransitive, reg_csubstitutive, reg_compatible, same_sregular, same_ssubstitutive.
Qed.


Lemma reg_consistent :
  forall (Rm : _ -> _ -> _ -> _ -> _ -> Prop),
    (forall M M' T,
       ofm nil impure M T
       -> ofm nil impure M' T
       -> Rm nil impure M M' T
       -> mhalt nil M
       -> mhalt nil M')
    -> consistent (mreg Rm).
Proof.
intros Rm Hcons.
intros M M' T HR Hhalt.
destruct HR as (Hof & Hof' & Hrel & _).
eauto.
Qed.


Lemma tcc_formation :
  forall Rt Rm G e e' t,
    cregular Rt Rm
    -> wf G
    -> tcc eqk eqc eqs Rt Rm G e e' t
    -> oft G e t /\ oft G e' t.
Proof.
intros Rt Rm G e e' t Hcreg Hwf Htcc.
so (reg_cregular (fun _ _ _ _ => True) (fun _ _ _ _ _ => True) andel) as Hreg.
eapply Hreg; eauto; [].
#6 apply (reg_compatible eqk eqc eqs (fun _ _ _ _ => True) (fun _ _ _ _ _ => True)).
  exact eq_sregular.

  exact eq_ssubstitutive.

  split; intros; trivial; done.

  split; intros; trivial; done.

  assumption.

  #3 apply (tcc_mono_wf eqk eqc eqs Rt _ Rm); auto.
    apply eq_sregular; done.

    clear G e e' t Hwf Htcc Hreg.
    intros G e e' t Hwf HR.
    so (Hcreg andel _#4 Hwf HR) as (? & ?).
    do2 2 split; auto; done.

    clear G e e' t Hwf Htcc Hreg.
    intros G p M M' T Hwf HR.
    so (Hcreg ander andel _#5 Hwf HR) as (? & ?).
    do2 3 split; auto; [].
    intros -> _.
    apply (Hcreg ander ander); auto; done.
Qed.


Lemma mcc_formation :
  forall Rt Rm G p M M' T,
    cregular Rt Rm
    -> wf G
    -> mcc eqk eqc eqs Rt Rm G p M M' T
    -> ofm G p M T
       /\ ofm G p M' T
       /\ (p = pure ->
             exists k c c',
               fsts T k
               /\ fstm G M c
               /\ fstm G M' c'
               /\ eqc G c c' k).
Proof.
intros Rt Rm G p M M' T Hcreg Hwf Hmcc.
cut (mreg (fun _ _ _ _ _ => True) G p M M' T).
  so (reg_cregular (fun _ _ _ _ => True) (fun _ _ _ _ _ => True)) as (_ & Hreg & Hregfst).
  intro Hmreg.
  so (Hreg _#5 Hwf Hmreg) as (Hof & Hof').
  do2 2 split; trivial; [].
  intros ->.
  apply Hregfst; eauto; done.
#6 apply (reg_compatible eqk eqc eqs (fun _ _ _ _ => True) (fun _ _ _ _ _ => True)).
  exact eq_sregular.

  exact eq_ssubstitutive.

  split; intros; trivial; done.

  split; intros; trivial; done.

  assumption.

  #3 apply (mcc_mono_wf eqk eqc eqs Rt _ Rm); auto.
    apply eq_sregular; done.

    clear G p M M' T Hwf Hmcc.
    intros G M M' T Hwf HR.
    so (Hcreg andel _#4 Hwf HR) as (? & ?).
    do2 2 split; auto; done.

    clear G p M M' T Hwf Hmcc.
    intros G p M M' T Hwf HR.
    so (Hcreg ander andel _#5 Hwf HR) as (? & ?).
    do2 3 split; auto; [].
    intros -> _.
    apply (Hcreg ander ander); auto; done.
Qed.


(* Symmetrization *)

Lemma transpose_sregular :
  forall Rk Rc Rs,
    sregular Rk Rc Rs
    -> sregular (fun G k k' => Rk G k' k) (fun G c c' k => Rc G c' c k) (fun G T T' => Rs G T' T).
Proof.
intros Rk Rc Rs Hreg.
do2 2 split.
  intros.
  apply eqk_symm; [].
  eapply (Hreg andel); eauto; done.

  intros.
  apply eqc_symm; [].
  eapply (Hreg anderl); eauto; done.

  intros.
  apply eqs_symm; [].
  eapply (Hreg anderr); eauto; done.
Qed.


Lemma transpose_ssubstitutive :
  forall Rk Rc Rs,
    ssubstitutive Rk Rc Rs
    -> ssubstitutive (fun G k k' => Rk G k' k) (fun G c c' k => Rc G c' c k) (fun G T T' => Rs G T' T).
Proof.
intros Rk Rc Rs (Hk & Hc & Hs).
do2 2 split; eauto.
Qed.


Lemma transpose_csubstitutive :
  forall Rt Rm,
    csubstitutive Rt Rm
    -> csubstitutive (fun G e e' t => Rt G e' e t) (fun G p M M' T => Rm G p M' M T).
Proof.
intros Rt Rm (Htsubst & Hmsubst).
split; eauto; done.
Qed.
    

Lemma transpose_compatible :
  forall Rk Rc Rs Rt Rm,
    sregular Rk Rc Rs
    -> cregular Rt Rm
    -> ssubstitutive Rk Rc Rs
    -> csubstitutive Rt Rm
    -> (forall G c c' k k', eqk G k k' -> Rc G c c' k -> Rc G c c' k')
    -> ssymmetric Rk Rc Rs
    -> compatible Rk Rc Rs Rt Rm
    -> compatible 
       (fun G k k' => Rk G k' k)
       (fun G c c' k => Rc G c' c k)
       (fun G T T' => Rs G T' T)
       (fun G e e' t => Rt G e' e t)
       (fun G p M M' T => Rm G p M' M T).

Proof.
intros Rk Rc Rs Rt Rm Hsregular (Htreg & Hmreg & Hregfst) Hssubst Hsubst Hconvert (Hksymm & Hcsymm & Hssymm) (Htcomp & Hmcomp).
split.


cut (forall G e e' t,
       wf G
       -> tcc
          (fun G k k' => Rk G k' k)
          (fun G c c' k => Rc G c' c k)
          (fun G T T' => Rs G T' T)
          (fun G e e' t => Rt G e' e t)
          (fun G p M M' T => Rm G p M' M T)
          G e e' t
       -> Rt G e' e t).
  intros H G e e' t Hcc.
  apply H; eauto using subst_tcc, transpose_sregular, transpose_ssubstitutive, transpose_csubstitutive; done.
destruct Hsregular as (Hkreg & Hcreg & Hsreg).
destruct Hsubst as (Htsubst & Hmsubst).
intros G e e' t Hwf Htcc.
revert Hwf.
#5 induct Htcc;
try (intros; eauto using tcc; done).

(* lam *)
intros G t1 t1' t2 e e' HRc HR Hwf.
so (Hcreg _#4 Hwf HRc) as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (_ & Hoft1 & _).
#2 apply Htcomp; auto; apply (@tcc_equiv _#6 (cn_arrow t1' t2)).
  apply Htcomp; auto; [].
  apply tcc_lam; auto with static; [].
  so (Htsubst _#6 (subof_conv_0t _#3 Heqc) (wf_tm Hwf Hoft1) HR) as H; simp_sub in H.
  assumption.

  so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 (wf_tm Hwf Hoft1) (Htreg _#4 (wf_tm Hwf Hoft1) HR andel))) as Hoft2; simp_sub in Hoft2.
  eauto with static; done.

(* plam *)
intros G k k' t e e' HRk HR Hwf.
so (Hkreg _#3 Hwf HRk) as Heqk.
so (eqk_reg _#3 Hwf Heqk) as (_ & Hofk).
#2 apply Htcomp; auto; apply (@tcc_equiv _#6 (cn_all k' t)).
  apply Htcomp; auto; apply tcc_plam; auto with static; [].
  so (Htsubst _#6 (subof_conv_0_aof _#3 Heqk) (wf_cn Hwf Hofk) HR) as H; simp_sub in H.
  assumption.

  so (oft_reg _#3 (wf_cn Hwf Hofk) (Htreg _#4 (wf_cn Hwf Hofk) HR andel)) as Hoft.
  eauto with static; done.

(* papp *)
intros G k t e e' c c' HR HRc Hwf.
#2 apply Htcomp; auto; apply (@tcc_equiv _#6 (subst1 c' t)).
  apply Htcomp; auto; eapply tcc_papp; eauto with static; done.

  so (ofc_ofcr _#3 (oft_reg _#3 Hwf (Htreg _#4 Hwf HR andel))) as (l & H).
  invert H; clear H; [].
  intros _ Hoft _.
  repl kd_type with (subst1 c' kd_type) by (simp_sub; reflexivity).
  eapply funct_cn; eauto with static; done.

(* pack *)
intros G k k' t t' c c' e e' HRc HRk HR HRt Hwf.
so (Hcreg _#4 Hwf HRc) as Heqc.
so (Hkreg _#3 Hwf HRk) as Heqk.
so (eqk_reg _#3 Hwf Heqk) as (_ & Hofk).
#2 apply Htcomp; auto; apply (@tcc_equiv _#6 (cn_exists k' t')).
  #2 apply Htcomp; auto; eapply tcc_pack; eauto with static.
    apply Htcomp; auto; apply (@tcc_equiv _#6 (subst1 c t)); auto; [].
    repl kd_type with (subst1 c kd_type) by (simp_sub; reflexivity).
    eapply funct_eqc; eauto with static; done.

    so (Hssubst ander andel _#6 (subof_conv_0 _#3 Heqk) (wf_cn Hwf Hofk) HRt) as HRt'; simp_sub in HRt'.
    auto; done.

  so (Hcreg _#4 (wf_cn Hwf Hofk) HRt) as Heqt.
  so (subst_eqc _#6 (subof_conv_0 _#3 Heqk) Heqt) as Heqt'; simp_sub in Heqt'.
  eauto with static; done.

(* fix *)
intros G t t' e e' HRc HR Hwf.
so (Hcreg _#4 Hwf HRc) as Heqc.
apply Htcomp; auto; apply (@tcc_equiv _#6 t'); auto with static; [].
apply Htcomp; auto; apply tcc_fix; auto with static; [].
apply Htcomp; auto; eapply tcc_equiv; eauto with static; done.






cut (forall G p M M' T,
       wf G
       -> mcc
          (fun G k k' => Rk G k' k)
          (fun G c c' k => Rc G c' c k)
          (fun G T T' => Rs G T' T)
          (fun G e e' t => Rt G e' e t)
          (fun G p M M' T => Rm G p M' M T)
          G p M M' T
       -> Rm G p M' M T).
  intros H G p M M' T Hcc.
  apply H; eauto using subst_mcc, transpose_ssubstitutive, transpose_csubstitutive; done.
destruct Hsregular as (Hkreg & Hcreg & Hsreg).
destruct Hsubst as (Htsubst & Hmsubst).
intros G p M M' T Hwf Hmcc.
revert Hwf.
#8 induct Hmcc;
try (intros; eauto using mcc; done).

(* lamgn *)
intros G T1 T1' T2 k1 M M' HRT Hfst HR Hwf.
so (Hsreg _#3 Hwf HRT) as HeqT1.
so (eqs_reg _#3 Hwf HeqT1 ander) as HofT1.
so (can_fsts _ _ (eqs_reg _#3 Hwf HeqT1 andel)) as (k1' & Hfst').
so (eqs_fst _#5 Hwf HeqT1 Hfst' Hfst) as Heqk.
#2 apply Hmcomp; auto; apply (@mcc_subsume _#7 (sg_pigen T1' T2)).
  apply Hmcomp; auto; eapply mcc_lamgn; eauto; [].
  so (Hmsubst _#7 (subof_conv_01m _#5 Heqk (weaken1_eqs _#4 HeqT1)) (wf_md Hwf HofT1 Hfst) HR) as HR'; simp_sub in HR'.
  assumption.

  apply subs_refl; [].
  eapply eqs_pigen; eauto; [].
  so (wf_md Hwf (eqs_reg _#3 Hwf HeqT1 ander) Hfst) as Hwf'.
  so (subst_ofs _#4 (subof_conv_0 _#3 Heqk) (subst_ofs _#4 (unshift_md _#3) (ofm_reg _#4 Hwf' (Hmreg _#5 Hwf' HR andel)))) as HofT2; simp_sub in HofT2.
  apply eqs_refl; assumption.

(* appgn *)
intros G T1 T2 M1 M1' M2 M2' c2 HR1 HR2 HfstM2 Hwf.
so (ofm_reg _#4 Hwf (Hmreg _#5 Hwf HR1 andel)) as Hofpi.
invert Hofpi; [].
intros k1 HofT1 Hfst HofT2.
so (Hregfst _#4 Hwf HR2) as (k' & c2' & d & Hfst' & HfstM2' & Hfstd & Heqc).
so (fsts_fun _#3 Hfst Hfst'); subst k'; clear Hfst'.
so (fstm_fun _#4 HfstM2 Hfstd); subst d; clear Hfstd.
#2 apply Hmcomp; auto; apply (mcc_subsume _#7 (subst1 c2' T2)).
  apply Hmcomp; auto; eapply mcc_appgn; eauto; done.

  apply subs_refl; [].
  eapply funct_sg; eauto with static; done.

(* lamap *)
intros G T1 T1' T2 k1 M M' HRT Hfst HR Hwf.
so (Hsreg _#3 Hwf HRT) as HeqT1.
so (eqs_reg _#3 Hwf HeqT1 ander) as HofT1.
so (can_fsts _ _ (eqs_reg _#3 Hwf HeqT1 andel)) as (k1' & Hfst').
so (eqs_fst _#5 Hwf HeqT1 Hfst' Hfst) as Heqk.
#2 apply Hmcomp; auto; apply (@mcc_subsume _#7 (sg_piapp T1' T2)).
  apply Hmcomp; auto; eapply mcc_lamap; eauto; [].
  so (Hmsubst _#7 (subof_conv_01m _#5 Heqk (weaken1_eqs _#4 HeqT1)) (wf_md Hwf HofT1 Hfst) HR) as HR'; simp_sub in HR'.
  assumption.

  apply subs_refl; [].
  eapply eqs_piapp; eauto; [].
  so (wf_md Hwf (eqs_reg _#3 Hwf HeqT1 ander) Hfst) as Hwf'.
  so (subst_ofs _#4 (subof_conv_0 _#3 Heqk) (subst_ofs _#4 (unshift_md _#3) (ofm_reg _#4 Hwf' (Hmreg _#5 Hwf' HR andel)))) as HofT2; simp_sub in HofT2.
  apply eqs_refl; assumption.

(* appap *)
intros G p T1 T2 M1 M1' M2 M2' c2 HR1 HR2 HfstM2 Hwf.
so (ofm_reg _#4 Hwf (Hmreg _#5 Hwf HR1 andel)) as Hofpi.
invert Hofpi; [].
intros k1 HofT1 Hfst HofT2.
so (Hregfst _#4 Hwf HR2) as (k' & c2' & d & Hfst' & HfstM2' & Hfstd & Heqc).
so (fsts_fun _#3 Hfst Hfst'); subst k'; clear Hfst'.
so (fstm_fun _#4 HfstM2 Hfstd); subst d; clear Hfstd.
#2 apply Hmcomp; auto; apply (mcc_subsume _#7 (subst1 c2' T2)).
  apply Hmcomp; auto; eapply mcc_appap; eauto; done.

  apply subs_refl; [].
  eapply funct_sg; eauto with static; done.

(* pi2 *)
intros G T1 T2 M M' c HR HfstM Hwf.  
so (ofm_reg _#4 Hwf (Hmreg _#5 Hwf HR andel)) as Hofsigma.
invert Hofsigma; [].
intros k1 HofT1 Hfst1 HofT2.
so (Hregfst _#4 Hwf HR) as (l & c' & d & Hfstsigma & HfstM' & Hfstd & Heqc).
invert Hfstsigma; [].
intros k1' k2 Hfst1' Hfst2 <-.
so (fsts_fun _#3 Hfst1 Hfst1'); subst k1'; clear Hfst1'.
so (fstm_fun _#4 HfstM Hfstd); subst d; clear Hfstd.
#2 apply Hmcomp; auto; apply (mcc_subsume _#7 (subst1 (cn_pi1 c') T2)).
  apply Hmcomp; auto; eapply mcc_pi2; eauto; done.

  apply subs_refl; [].
  eapply funct_sg; eauto with static; done.

(* unpack *)
intros G k t T T' e e' M M' HRe HRM HRT Hwf.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf (Htreg _#4 Hwf HRe andel))) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
#2 apply Hmcomp; auto; apply (@mcc_subsume _#7 T').
  apply Hmcomp; auto; eapply mcc_unpack; eauto; [].
  apply Hmcomp; auto using wf; eapply mcc_subsume; eauto; [].
  apply subs_refl; [].
  apply (subst_eqs _#5 (weaken2 _#4)); auto; done.

  apply subs_refl; auto; done.

(* letm *)
intros G T1 T2 T2' k1 M1 M1' M2 M2' HR1 Hfst HR2 HRT Hwf.
so (ofm_reg _#4 Hwf (Hmreg _#5 Hwf HR1 andel)) as HofT.
#2 apply Hmcomp; auto; apply (@mcc_subsume _#7 T2').
  apply Hmcomp; auto; eapply mcc_letm; eauto; [].
  apply Hmcomp; auto using wf; eapply mcc_subsume; eauto; [].
  apply subs_refl; [].
  apply (subst_eqs _#5 (weaken2 _#4)); auto; done.

  apply subs_refl; auto; done.

(* seal *)
intros G T T' M M' HR HRT Hwf.
#2 apply Hmcomp; auto; apply (@mcc_subsume _#7 T').
  apply Hmcomp; auto; eapply mcc_seal; eauto; [].
  apply Hmcomp; auto; eapply mcc_subsume; eauto; [].
  apply subs_refl; apply eqs_symm; auto; done.

  apply subs_refl; auto; done.
Qed.


Lemma transpose_ssymmetric :
  forall Rk Rc Rs,
    ssymmetric Rk Rc Rs
    -> ssymmetric (fun G k k' => Rk G k' k) (fun G c c' k => Rc G c' c k) (fun G T T' => Rs G T' T).
Proof.
intros Rk Rc Rs (Hk & Hc & Hs).
do2 2 split; eauto.
Qed.


Lemma transpose_compatible' :
  forall Rk Rc Rs Rt Rm,
    sregular Rk Rc Rs
    -> cregular Rt Rm
    -> ssubstitutive Rk Rc Rs
    -> csubstitutive Rt Rm
    -> (forall G c c' k k', eqk G k k' -> Rc G c c' k -> Rc G c c' k')
    -> ssymmetric Rk Rc Rs
    -> compatible Rk Rc Rs Rt Rm
    -> compatible Rk Rc Rs
       (fun G e e' t => Rt G e' e t)
       (fun G p M M' T => Rm G p M' M T).
Proof.
intros Rk Rc Rs Rt Rm Hsreg Hreg Hssubst Hsubst Hequiv Hsymm Hcomp.
apply transpose_compatible; eauto using transpose_sregular, transpose_ssubstitutive, transpose_ssymmetric; [].
#3 eapply compatible_mono; eauto.
  intros.
  apply (Hsymm andel); auto; done.

  intros.
  apply (Hsymm anderl); auto; done.

  intros.
  apply (Hsymm anderr); auto; done.
Qed.


Lemma transpose_semicongruence :
  forall Rt Rm,
    semicongruence Rt Rm
    -> semicongruence (fun G e e' t => Rt G e' e t) (fun G p M M' T => Rm G p M' M T).
Proof.
intros Rt Rm (Hreg & (Htrefl & Hmrefl) & (Htrans & Hmtrans) & Hsubst & Hcomp).
do2 4 split.
  destruct Hreg as (Htreg & Hmreg & Hregfst).
  do2 2 split.
    intros G e e' t Hwf HR.
    so (Htreg _#4 Hwf HR) as (H & H').
    auto; done.

    intros G p M M' T Hwf HR.
    so (Hmreg _#5 Hwf HR) as (H & H').
    auto; done.

    intros G M M' T Hwf HR.
    so (Hregfst _#4 Hwf HR) as (k & c & c' & Hfst & Hfstm & Hfstm' & Heqc).
    do 3 eexists; eauto using eqc_symm; done.

  split; auto; done.

  split; eauto; done.

  destruct Hsubst.
  split; eauto; done.

  apply transpose_compatible'; auto using same_sregular, same_ssubstitutive, same_ssymetric; [].
  intros G c c' k k' Heqk Hsame.
  invert Hsame; [].
  intros Hofc <-.
  apply samecrefl; eauto with static; done.
Qed.


Lemma treg_transpose :
  forall R G e e' t,
    treg (fun G e e' t => R G e e' t) G e e' t
    -> treg (fun G e e' t => R G e' e t) G e' e t.
Proof.
intros R G e e' t H.
so H as (Hof & Hof' & HR).
do2 2 split; auto.
Qed.


Lemma mreg_transpose :
  forall R G p M M' T,
    mreg (fun G p M M' T => R G p M M' T) G p M M' T
    -> mreg (fun G p M M' T => R G p M' M T) G p M' M T.
Proof.
intros R G p M M' T H.
so H as (Hof & Hof' & HR & Hfst).
do2 3 split; auto; [].
intros Heq Hwf.
so (Hfst Heq Hwf) as (k & c & c' & Hfsts & Hfstm & Hfstm' & Heqc).
do 3 eexists; do2 3 split; eauto using eqc_symm.
Qed.


(* Conjunction *)

Definition tintersect (Rt Rt' : treln) G e e' t :=
  Rt G e e' t /\ Rt' G e e' t.


Definition mintersect (Rm Rm' : mreln) G p M M' T :=
  Rm G p M M' T /\ Rm' G p M M' T.


Lemma conjunction_compatible :
  forall Rk Rc Rs Rt Rt' Rm Rm',
    compatible Rk Rc Rs Rt Rm
    -> compatible Rk Rc Rs Rt' Rm'
    -> compatible Rk Rc Rs (tintersect Rt Rt') (mintersect Rm Rm').
Proof.
intros Rk Rc Rs Rt Rt' Rm Rm' (Htcomp & Hmcomp) (Htcomp' & Hmcomp').
unfold tintersect, mintersect.
split.
  intros G e e' t Hwf H.
  revert Hwf.
  induct H;
  intros;
  repeat (match goal with
          | H : _ /\ _ |- _ => destruct H
          end);
  split; eauto using Htcomp, tcc;
  done.

  intros G p M M' T Hwf H.
  revert Hwf.
  induct H;
  intros;
  repeat (match goal with
          | H : _ /\ _ |- _ => destruct H
          end);
  split; eauto using Hmcomp, mcc;
  done.
Qed.


Lemma conjunction_semicongruence :
  forall Rt Rt' Rm Rm',
    semicongruence Rt Rm
    -> semicongruence Rt' Rm'
    -> semicongruence (tintersect Rt Rt') (mintersect Rm Rm').
Proof.
intros Rt Rt' Rm Rm'.
unfold tintersect, mintersect.
intros ((Htreg & Hmreg & Hregfst) & (Htrefl & Hmrefl) & (Htrans & Hmtrans) & (Htsubst & Hmsubst) & Hcomp).
intros ((Htreg' & Hmreg' & Hregfst') & (Htrefl' & Hmrefl') & (Htrans' & Hmtrans') & (Htsubst' & Hmsubst') & Hcomp').
do2 4 split; [do2 2 split | split .. |];
try (intros;
     repeat (match goal with
             | H : _ /\ _ |- _ => destruct H
             end);
     eauto;
     done); [].
apply conjunction_compatible; auto; done.
Qed.


(* Union *)

Definition tunion (Rt Rt' : context -> term -> term -> term -> Prop) G e e' t :=
  Rt G e e' t \/ Rt' G e e' t.


Definition munion (Rm Rm' : context -> purity -> term -> term -> term -> Prop) G p M M' T :=
  Rm G p M M' T \/ Rm' G p M M' T.


Lemma union_cregular :
  forall Rt Rt' Rm Rm',
    cregular Rt Rm
    -> cregular Rt' Rm'
    -> cregular (tunion Rt Rt') (munion Rm Rm').
Proof.
intros Rt Rt' Rm Rm' Hreg Hreg'.
do2 2 split.
  intros G e e' t Hwf H.
  #2 destruct H.
    apply (Hreg andel); auto; done.

    apply (Hreg' andel); auto; done.

  intros G p M M' T Hwf H.
  #2 destruct H.
    apply (Hreg anderl); auto; done.

    apply (Hreg' anderl); auto; done.

  intros G M M' T Hwf H.
  #2 destruct H as [H | H].
    apply (Hreg anderr); auto; done.

    apply (Hreg' anderr); auto; done.
Qed.


Lemma union_creflexive_1 :
  forall Rt Rt' Rm Rm',
    creflexive Rt Rm
    -> creflexive (tunion Rt Rt') (munion Rm Rm').
Proof.
intros Rt Rt' Rm Rm' Hrefl.
split.
  intros G e t H.
  left.
  apply (Hrefl andel); auto; done.

  intros G p M T H.
  left.
  apply (Hrefl ander); auto; done.
Qed.


Lemma union_csubstitutive :
  forall Rt Rt' Rm Rm',
    csubstitutive Rt Rm
    -> csubstitutive Rt' Rm'
    -> csubstitutive (tunion Rt Rt') (munion Rm Rm').
Proof.
intros Rt Rt' Rm Rm' Hsubst Hsubst'.
split.
  intros G1 s G2 e e' t Hsubof Hwf H.
  destruct H as [H | H].
    left.
    eapply (Hsubst andel); eauto; done.

    right.
    eapply (Hsubst' andel); eauto; done.

  intros G1 s G2 p M M' T Hsubof Hwf H.
  destruct H as [H | H].
    left.
    eapply (Hsubst ander); eauto; done.

    right.
    eapply (Hsubst' ander); eauto; done.
Qed.


Lemma union_compatible :
  forall Rk Rc Rs Rt Rt' Rm Rm',
    weakly_compatible Rk Rc Rs Rt Rm
    -> weakly_compatible Rk Rc Rs Rt' Rm'
    -> weakly_compatible Rk Rc Rs (tunion Rt Rt') (munion Rm Rm').
Proof.
unfold tunion, munion.
intros Rk Rc Rs Rt Rt' Rm Rm' (Htcomp & Hmcomp) (Htcomp' & Hmcomp').
split.
  intros G e e' t Hwf H.
  revert Hwf.
  induct H;
  intros;
  (match goal with
   | H : _ \/ _ |- _ => destruct H; [left | right]
   end);
  eauto using tccw;
  done.

  intros G p M M' T Hwf H.
  revert Hwf.
  induct H;
  try (intros;
       (match goal with
        | H : _ \/ _ |- _ => destruct H; [left | right]
        end);
       eauto using mccw;
       done); [].
  (* stat *)
  intros G c c' k H.
  left.
  eauto using mccw; done.
Qed.


(* Transitive closure *)

Definition ttc Rt (G : context) (e e' : (* tm *) term) (t : (* cn *) term) :=
  plus (fun e1 e2 => Rt G e1 e2 t) e e'.


Definition mtc Rm (G : context) (p : purity) (M M' : (* md *) term) (T : (* sg *) term) :=
  plus (fun M1 M2 => Rm G p M1 M2 T) M M'.


Lemma tc_creflexive :
  forall Rt Rm,
    creflexive Rt Rm
    -> creflexive (ttc Rt) (mtc Rm).
Proof.
intros Rt Rm Hrefl.
split.
  intros G e t Hofe.
  apply plus_one; [].
  apply (Hrefl andel); auto; done.

  intros G p e t Hofe.
  apply plus_one; [].
  apply (Hrefl ander); auto; done.
Qed.


Lemma tc_ctransitive :
  forall Rt Rm,
    ctransitive (ttc Rt) (mtc Rm).
Proof.
intros Rt Rm.
split.
  intros G t e1 e2 e3 H12 H23.
  eapply plus_trans; eauto; done.

  intros G p t e1 e2 e3 H12 H23.
  eapply plus_trans; eauto; done.
Qed.


Lemma tc_csubstitutive :
  forall Rt Rm,
    csubstitutive Rt Rm
    -> csubstitutive (ttc Rt) (mtc Rm).
Proof.
intros Rt Rm Hsubst.
split.
  intros G1 s G2 e e' t Hs Hwf Httc.
  unfold ttc in Httc; unfold ttc.
  eapply plus_mono_map; eauto; [].
  intros e1 e2 He.
  eapply (Hsubst andel); eauto; done.

  intros G1 s G2 p M M' T Hs Hwf Hmtc.
  unfold mtc in Hmtc; unfold mtc.
  eapply plus_mono_map; eauto; [].
  intros e1 e2 He.
  eapply (Hsubst ander); eauto; done.
Qed.


Lemma ttc_intro :
  forall (Rt : treln) G e e' t,
    Rt G e e' t
    -> ttc Rt G e e' t.
Proof.
intros Rt G e e' t HR.
apply plus_one; auto; done.
Qed.


Lemma mtc_intro :
  forall (Rm : mreln) G p e e' t,
    Rm G p e e' t
    -> mtc Rm G p e e' t.
Proof.
intros Rm G p e e' t HR.
apply plus_one; auto; done.
Qed.


Lemma tc_cregular :
  forall Rt Rm,
    cregular Rt Rm
    -> cregular (ttc Rt) (mtc Rm).
Proof.
intros Rt Rm (Htreg & Hmreg & Hfstreg).
do2 2 split.
  intros G e1 e2 t Hwf Hplus.
  so (plus_plusr _#4 Hplus) as (e2' & _ & H2).
  destruct Hplus as (e1' & H1 & _).
  split.
    exact (Htreg _#4 Hwf H1 andel).

    exact (Htreg _#4 Hwf H2 ander).

  intros G p M1 M2 T Hwf Hplus.
  so (plus_plusr _#4 Hplus) as (e2' & _ & H2).
  destruct Hplus as (e1' & H1 & _).
  split.
    exact (Hmreg _#5 Hwf H1 andel).

    exact (Hmreg _#5 Hwf H2 ander).

  intros G M1 M2 T Hwf Hplus.
  so (plus_plusi _#4 Hplus) as Hplus'.
  clear Hplus.
  revert Hwf.
  induct Hplus'.

  (* one *)
  intros M M' H Hwf.
  apply Hfstreg; auto; done.

  (* step *)
  intros M1 M2 M3 H12 _ IH Hwf.
  so (Hfstreg _#4 Hwf H12) as (k & c1 & c2 & Hfst & Hfst1 & Hfst2 & Heqc).
  so (IH Hwf) as (k' & c2' & c3 & Hfst' & Hfst2' & Hfst3 & Heqc').
  so (fsts_fun _#3 Hfst Hfst'); subst k'.
  so (fstm_fun _#4 Hfst2 Hfst2'); subst c2'.
  do 3 eexists; eauto using eqc_trans; done.
Qed.


Lemma ttc_reg :
  forall Rt Rm G e e' t,
    cregular Rt Rm
    -> wf G
    -> ttc Rt G e e' t
    -> oft G e t /\ oft G e' t.
Proof.
intros Rt Rm G e e' t Hreg.
revert G e e' t.
exact (tc_cregular Rt Rm Hreg andel).
Qed.


Lemma mtc_reg :
  forall Rt Rm G p M M' T,
    cregular Rt Rm
    -> wf G
    -> mtc Rm G p M M' T
    -> ofm G p M T /\ ofm G p M' T.
Proof.
intros Rt Rm G p M M' T Hreg.
revert G p M M' T.
exact (tc_cregular Rt Rm Hreg anderl).
Qed.


Lemma mtc_fst :
  forall Rt Rm G M M' T,
    cregular Rt Rm
    -> wf G
    -> mtc Rm G pure M M' T
    -> exists k c c',
         fsts T k
         /\ fstm G M c
         /\ fstm G M' c'
         /\ eqc G c c' k.
Proof.
intros Rt Rm G M M' T Hreg.
revert G M M' T.
exact (tc_cregular Rt Rm Hreg anderr).
Qed.


Lemma helpmapfst :
  forall f Rt Rm G p T T' M M' k c,
    cregular Rt Rm
    -> weakly_compatible samek samec sames Rt Rm
    -> wf G
    -> fsts T k
    -> fstm G M c
    -> (forall M1 M2 c1,
          Rm G pure M1 M2 T
          -> fstm G M1 c1
          -> eqc G c c1 k
          -> exists T'',
               eqs G T' T''
               /\ mccw samek samec sames Rt Rm G p (f M1) (f M2) T'')
    -> plus (fun M1 M2 => Rm G pure M1 M2 T) M M'
    -> plus (fun M1 M2 => Rm G p M1 M2 T') (f M) (f M').
Proof.
intros f Rt Rm G p T T' M M' k c Hreg Hcompat Hwf Hfst Hfstm Hincl Hplus.
assert (eqc G c c k) as Heqc.
  destruct Hplus as (M'' & HM & _).
  so (Hreg anderl _#5 Hwf HM andel) as HofM.
  so (ofm_fst _#6 Hwf HofM Hfstm Hfst) as Hofc.
  apply eqc_refl; auto; done.
revert Heqc.
so Hfstm as H; revert H.
generalize c at 1 3.
so (plus_plusi _#4 Hplus) as Hplusi; clear Hplus.
induct Hplusi.

(* one *)
intros M1 M2 HM c' Hfstm' Heqc.
apply plus_one; [].
so (Hincl _ _ _ HM Hfstm' Heqc) as (T'' & Heqs & Hmcc).
apply (Hcompat ander); auto; [].
eapply mccw_subsume; eauto with static; [].
apply (Hcompat ander); auto; done.

(* step *)
intros M1 M2 M3 HM _ IH c' Hfstm' Heqc.
exists (f M2).
split.
  so (Hincl _ _ _ HM Hfstm' Heqc) as (T'' & Heqs & Hmcc).
  apply (Hcompat ander); auto; [].
  eapply mccw_subsume; eauto with static; [].
  apply (Hcompat ander); auto; done.

  apply plus_star; [].
  so (Hreg anderr _#4 Hwf HM) as (ka & c'a & c'' & Hfsta & Hfstm'a & Hfstm'' & Heqc').
  so (fsts_fun _#3 Hfst Hfsta); subst ka.
  so (fstm_fun _#4 Hfstm' Hfstm'a); subst c'a.
  eapply IH; eauto with static; done.
Qed.


Lemma tc_comp :
  forall Rt Rm,
    cregular Rt Rm
    -> csubstitutive Rt Rm
    -> weakly_compatible samek samec sames Rt Rm
    -> compatible samek samec sames (ttc Rt) (mtc Rm).
Proof.
intros Rt Rm Hreg Hsubst Hwcomp.
split.

(* terms *)
intros G e e' t Hwf Htcc.
unfold ttc.
revert Hwf.
destruct Hwcomp as (Htcomp & _).
induct Htcc.

(* lam *)
intros G t1 t1' t2 e e' Hsame HR Hwf.
invert Hsame; clear Hsame; [].
intros Hoft1 <-.
eapply plus_mono_map; eauto using wf; [].
intros e1 e2 H12.
eauto using tccw, samec; done.

(* app *)
intros G t1 t2 e1 e1' e2 e2' HR1 HR2 Hwf.
#2 apply (plus_trans _#3 (tm_app e1' e2)).
  eapply (plus_mono_map _#3 (fun z => tm_app z _)); eauto; [].
  intros e e' He.
  simpl in He.
  so (ttc_reg _#6 Hreg Hwf HR2 andel).
  eauto using tccw; done.

  eapply plus_mono_map; eauto; [].
  intros e e' He.
  simpl in He.
  so (ttc_reg _#6 Hreg Hwf HR1 ander).
  eauto using tccw; done.

(* pair *)
intros G t1 t2 e1 e1' e2 e2' HR1 HR2 Hwf.
#2 apply (plus_trans _#3 (tm_pair e1' e2)).
  eapply (plus_mono_map _#3 (fun z => tm_pair z _)); eauto; [].
  intros e e' He.
  simpl in He.
  so (ttc_reg _#6 Hreg Hwf HR2 andel).
  eauto using tccw; done.

  eapply plus_mono_map; eauto; [].
  intros e e' He.
  simpl in He.
  so (ttc_reg _#6 Hreg Hwf HR1 ander).
  eauto using tccw; done.

(* pi1 *)
intros G t1 t2 e e' HR Hwf.
eapply plus_mono_map; eauto; [].
intros e1 e2 H12.
simpl in H12.
eauto using tccw; done.

(* pi2 *)
intros G t1 t2 e e' HR Hwf.
eapply plus_mono_map; eauto; [].
intros e1 e2 H12.
simpl in H12.
eauto using tccw; done.

(* plam *)
intros G k k' t e e' Hsame HR Hwf.
invert Hsame; clear Hsame; [].
intros Hofk <-.
eapply plus_mono_map; eauto using wf; [].
intros e1 e2 H12.
eauto using tccw, samek; done.

(* papp *)
intros G k t e e' c c' HR Hsame Hwf.
invert Hsame; clear Hsame; [].
intros Hofc <-.
eapply (plus_mono_map _#3 (fun z => tm_papp z c)); eauto; [].
intros e1 e2 H.
simpl in H.
eauto using tccw, samec; done.

(* pack *)
intros G k k' t t' c c' e e' Hsamec Hsamek HR1 Hsamet Hwf.
invert Hsamec; invert Hsamek; invert Hsamet; [].
intros Hoft <- Hofk <- Hofc <-.
eapply (plus_mono_map _#3 (fun z => tm_pack _ z _ _)); eauto; [].
intros e1 e2 H.
simpl in H.
eauto using tccw, samec; done.

(* unpack *)
intros G k t t' e1 e1' e2 e2' HR1 HR2 Hoft' Hwf.
so (ttc_reg _#6 Hreg Hwf HR1 ander) as Hofe1'.
so (ofc_ofcr _#3 (oft_reg _ _ _ Hwf Hofe1')) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
so (ttc_reg _#6 Hreg (wf_tm (wf_cn Hwf Hofk) Hoft) HR2 andel).
#2 apply (plus_trans _#3 (tm_unpack e1' e2)).
  eapply (plus_mono_map _#3 (fun z => tm_unpack z _)); eauto; [].
  intros e e' He.
  simpl in He.
  eauto using tccw; done.

  eapply plus_mono_map; eauto using wf; [].
  intros e e' He.
  simpl in He.
  eauto using tccw; done.

(* fix *)
intros G t t' e e' Hsame HR Hwf.
invert Hsame; clear Hsame; [].
intros Hof <-.
eapply plus_mono_map; eauto using wf; [].
intros e1 e2 H12.
eauto using tccw, samec; done.

(* lett *)
intros G t t' e1 e1' e2 e2' HR1 HR2 Hwf.
so (ttc_reg _#6 Hreg Hwf HR1 ander) as Hofe1'.
so (oft_reg _#3 Hwf Hofe1') as Hoft.
so (ttc_reg _#6 Hreg (wf_tm Hwf Hoft) HR2 andel).
#2 apply (plus_trans _#3 (tm_lett e1' e2)).
  eapply (plus_mono_map _#3 (fun z => tm_lett z _)); eauto; [].
  intros e e' He.
  simpl in He.
  eauto using tccw; done.

  eapply plus_mono_map; eauto using wf; [].
  intros e e' He.
  simpl in He.
  eauto using tccw; done.

(* equiv *)
intros G t1 t2 e e' HR1 Heqc Hwf.
eapply plus_mono; eauto; [].
intros e1 e2 H.
simpl in H.
eauto using tccw; done.

(* letm *)
intros G T k t M M' e e' HR1 Hfst HR2 Hoft Hwf.
so (mtc_reg _#7 Hreg Hwf HR1 ander) as HofM'.
so (ofm_reg _#4 Hwf HofM') as HofT.
so (ttc_reg _#6 Hreg (wf_md Hwf HofT Hfst) HR2 andel).
#2 apply (plus_trans _#3 (tm_letm M' e)).
  eapply (plus_mono_map _#3 (fun z => tm_letm z _)); eauto; [].
  intros ? ? He.
  simpl in He.
  eauto using tccw; done.

  eapply plus_mono_map; eauto using wf; [].
  intros ? ? He.
  simpl in He.
  eauto using tccw; done.

(* ext *)
intros G t M M' HR Hwf.
eapply plus_mono_map; eauto; [].
intros ? ? He.
simpl in He.
eauto using tccw; done.



(* modules *)
intros G p M M' T Hwf Hmcc.
unfold mtc.
revert Hwf.
so Hwcomp as (_ & Hmcomp).
induct Hmcc.

(* stat *)
intros G c c' k Hsame Hwf.
apply plus_one; eauto using mccw; done.

(* dyn *)
intros G e e' t HR Hwf.
eapply plus_mono_map; eauto using wf; [].
intros ? ? H.
simpl in H.
eauto using mccw; done.

(* lamgn *)
intros G T1 T1' T2 k1 M M' Hsame Hfst HR Hwf.
invert Hsame; clear Hsame; [].
intros HofT1 <-.
eapply plus_mono_map; eauto using wf; [].
intros ? ? H.
simpl in H.
eauto using mccw, sames; done.

(* appgn *)
intros G T1 T2 M1 M1' M2 M2' c2 HR1 HR2 HfstM2 Hwf.
so (mtc_reg _#7 Hreg Hwf HR1 ander) as HofM1'.
so (mtc_reg _#7 Hreg Hwf HR2 andel).
invert (ofm_reg _#4 Hwf HofM1'); [].
intros k1 _ Hfst1 HofT2.
#2 apply (plus_trans _#3 (md_appgn M1' M2)).
  eapply (plus_mono_map _#3 (fun z => md_appgn z _)); eauto using wf; [].
  intros ? ? He.
  simpl in He.
  eauto using mccw; done.

  eapply helpmapfst; eauto; [].
  intros M M' c2' HM HfstM Heqc.
  exists (subst1 c2' T2).
  eauto using mccw, funct_sg; done.

(* lamap *)
intros G T1 T1' T2 k1 M M' Hsame Hfst HR Hwf.
invert Hsame; clear Hsame; [].
intros HofT1 <-.
eapply plus_mono_map; eauto using wf; [].
intros ? ? H.
simpl in H.
eauto using mccw, sames; done.

(* appap *)
intros G p T1 T2 M1 M1' M2 M2' c2 HR1 HR2 HfstM2 Hwf.
so (mtc_reg _#7 Hreg Hwf HR1 ander) as HofM1'.
so (mtc_reg _#7 Hreg Hwf HR2 andel).
invert (ofm_reg _#4 Hwf HofM1'); [].
intros k1 _ Hfst1 HofT2.
#2 apply (plus_trans _#3 (md_appap M1' M2)).
  eapply (plus_mono_map _#3 (fun z => md_appap z _)); eauto using wf; [].
  intros ? ? He.
  simpl in He.
  eauto using mccw; done.

  eapply helpmapfst; eauto; [].
  intros M M' c2' HM HfstM Heqc.
  exists (subst1 c2' T2).
  eauto using mccw, funct_sg; done.

(* pair *)
intros G p T1 T2 M1 M1' M2 M2' HR1 HR2 Hwf.
so (mtc_reg _#7 Hreg Hwf HR1 ander).
so (mtc_reg _#7 Hreg Hwf HR2 andel).
#2 apply (plus_trans _#3 (md_pair M1' M2)).
  eapply (plus_mono_map _#3 (fun z => md_pair z _)); eauto using wf; [].
  intros ? ? He.
  simpl in He.
  eauto using mccw; done.

  eapply plus_mono_map; eauto using wf; [].
  intros ? ? He.
  simpl in He.
  eauto using mccw; done.

(* pi1 *)
intros G T1 T2 M M' HR Hwf.
eapply plus_mono_map; eauto; [].
intros e1 e2 H12.
simpl in H12.
eauto using mccw; done.

(* pi2 *)
intros G T1 T2 M M' c HR Hfstm Hwf.
so (mtc_reg _#7 Hreg Hwf HR andel) as HofM.
invert (ofm_reg _#4 Hwf HofM); [].
intros k1 HofT1 Hfst1 HofT2.
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
eapply (helpmapfst _#5 (sg_sigma T1 T2)); eauto using fsts_sigma; [].
intros M1 M2 c' H12 Hfstm1 Heqc.
exists (subst1 (cn_pi1 c') T2).
split; eauto using funct_sg, mccw with static; done.

(* unpack *)
intros G k t T T' e e' M M' HR1 HR2 Hsames Hwf.
invert Hsames; clear Hsames; [].
intros HofT <-.
so (ttc_reg _#6 Hreg Hwf HR1 ander) as Hofe'.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hofe')) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _; clear l Hofcr.
so (mtc_reg _#7 Hreg (wf_tm (wf_cn Hwf Hofk) Hoft) HR2 andel).
#2 apply (plus_trans _#3 (md_unpack e' M T)).
  eapply (plus_mono_map _#3 (fun z => md_unpack z _ _)); eauto using wf; [].
  intros ? ? He.
  simpl in He.
  eauto using sames, mccw; done.

  eapply (plus_mono_map _#3 (fun z => md_unpack _ z _)); eauto using wf; [].
  intros ? ? He.
  simpl in He.
  eauto using sames, mccw; done.

(* lete *)
intros G p t T e e' M M' HR1 HR2 Hwf.
so (ttc_reg _#6 Hreg Hwf HR1 ander) as Hofe'.
so (oft_reg _#3 Hwf Hofe') as Hoft.
so (mtc_reg _#7 Hreg (wf_tm Hwf Hoft) HR2 andel).
#2 apply (plus_trans _#3 (md_lett e' M)).
  eapply (plus_mono_map _#3 (fun z => md_lett z _)); eauto using wf; [].
  intros ? ? He.
  simpl in He.
  eauto using mccw; done.

  eapply plus_mono_map; eauto using wf; [].
  intros ? ? He.
  simpl in He.
  eauto using mccw; done.

(* letm *)
intros G T1 T2 T2' k1 M1 M1' M2 M2' HR1 Hfst1 HR2 Hsames Hwf.
invert Hsames; clear Hsames; [].
intros HofT2 <-.
so (mtc_reg _#7 Hreg Hwf HR1 ander) as HofM1'.
so (ofm_reg _#4 Hwf HofM1') as HofT1.
so (mtc_reg _#7 Hreg (wf_md Hwf HofT1 Hfst1) HR2 andel).
#2 apply (plus_trans _#3 (md_letm M1' M2 T2)).
  eapply (plus_mono_map _#3 (fun z => md_letm z _ _)); eauto using wf; [].
  intros ? ? He.
  simpl in He.
  eauto using sames, mccw; done.

  eapply (plus_mono_map _#3 (fun z => md_letm _ z _)); eauto using wf; [].
  intros ? ? He.
  simpl in He.
  eauto using sames, mccw; done.

(* seal *)
intros G T T' M M' HR Hsames Hwf.
invert Hsames; clear Hsames; [].
intros HofT <-.
eapply (plus_mono_map _#3 (fun z => md_seal z _)); eauto; [].
intros e1 e2 H12.
simpl in H12.
eauto using mccw, sames; done.

(* forget *)
intros G T M M' HR Hwf.
eapply plus_mono; eauto; [].
eauto using mccw; done.

(* subsume *)
intros G p T1 T2 M M' HR Hsubs Hwf.
eapply plus_mono; eauto; [].
intros M1 M2 M12.
simpl in M12.
eauto using mccw; done.
Qed.


Lemma tc_semicongruence_strong :
  forall Rt Rm,
    cregular Rt Rm
    -> creflexive Rt Rm
    -> csubstitutive Rt Rm
    -> weakly_compatible samek samec sames Rt Rm
    -> semicongruence (ttc Rt) (mtc Rm).
Proof.
intros Rt Rm Hreg Hrefl Hsubst Hcomp.
do2 4 split; auto using tc_cregular, tc_creflexive, tc_ctransitive, tc_csubstitutive; [].
apply tc_comp; auto; done.
Qed.    


Lemma tc_semicongruence :
  forall Rt Rm,
    quasicongruence Rt Rm
    -> semicongruence (ttc Rt) (mtc Rm).
Proof.
unfold quasicongruence.
intros Rt Rm H.
destruct_all H.
apply tc_semicongruence_strong; auto using weaken_compatible; done.
Qed.


Lemma mtc_consistent :
  forall Rm,
    consistent Rm
    -> consistent (mtc Rm).
Proof.
intros Rm Hcons.
intros M M' T HR Hhalt.
so (plus_star _#4 HR) as Hstar.
clear HR.
revert Hhalt.
induct Hstar; auto; [].
(* step *)
intros M1 M2 M3 HR _ IH Hhalt.
unfold consistent in Hcons.
eauto; done.
Qed.


(* Transitive and symmetric closure *)

Definition tstc Rt :=
  ttc (fun G e e' t => Rt G e e' t \/ Rt G e' e t).


Definition mstc Rm :=
  mtc (fun G p M M' T => Rm G p M M' T \/ Rm G p M' M T).


Lemma stc_congruence :
  forall Rt Rm,
    quasicongruence Rt Rm
    -> congruence (tstc Rt) (mstc Rm).
Proof.
intros Rt Rm (Hreg & Hrefl & Hsubst & Hcomp).
set (Rt' G e e' t := Rt G e' e t).
set (Rm' G p M M' T := Rm G p M' M T).
set (RRt := tunion Rt Rt').
set (RRm := munion Rm Rm').
assert (cregular Rt' Rm') as Hreg'.
  do2 2 split.
    intros.
    refine (match _ with conj x y => conj y x end); [].
    apply (Hreg andel); auto; done.

    intros.
    refine (match _ with conj x y => conj y x end); [].
    apply (Hreg anderl); auto; done.

    intros G M M' T Hwf H.
    so (Hreg anderr _#4 Hwf H) as (k & c' & c & Hfst & Hfstm' & Hfstm & Heqc).
    exists k, c, c'.
    auto using eqc_symm; done.
assert (csubstitutive Rt' Rm') as Hsubst'.
  split.
    intros; eapply (Hsubst andel); eauto; done.

    intros; eapply (Hsubst ander); eauto; done.
split.
  do2 4 split.
    apply tc_cregular; apply union_cregular; auto; done.

    apply tc_creflexive; apply union_creflexive_1; auto; done.

    apply tc_ctransitive; done.

    apply tc_csubstitutive; apply union_csubstitutive; auto; done.

    #3 apply tc_comp.
      apply union_cregular; auto; done.

      apply union_csubstitutive; auto; done.

      #2 apply union_compatible.
        apply weaken_compatible; auto; done.

        apply weaken_compatible; auto; [].
        apply transpose_compatible'; auto using same_sregular, same_ssubstitutive, same_ssymetric; [].
        intros G c c' k k' Heqk Hsame.
        invert Hsame; [].
        intros Hof ->.
        apply samecrefl; eauto with static; done.

  (* symmetry *)
  split.
    intros G t e e' Hstc.
    apply plus_transpose; [].
    apply (plus_mono _ (fun e1 e2 => Rt G e1 e2 t \/ Rt G e2 e1 t)); auto; [].
    intros e1 e2 H.
    unfold transpose.
    destruct H; [right | left]; assumption.

    intros G p T M M' Hstc.
    apply plus_transpose; [].
    apply (plus_mono _ (fun M1 M2 => Rm G p M1 M2 T \/ Rm G p M2 M1 T)); auto; [].
    intros e1 e2 H.
    unfold transpose.
    destruct H; [right | left]; assumption.
Qed.    


Lemma mstc_consistent :
  forall Rm,
    consistent Rm
    -> consistent (fun G p M M' T => Rm G p M' M T)
    -> consistent (mstc Rm).
Proof.
intros Rm Hcons Hcons'.
intros M M' T HR Hhalt.
so (plus_star _#4 HR) as HR'.
revert Hhalt.
induct HR'; auto; [].
intros M1 M2 M3 Hstep _ IH H1.
apply IH; [].
destruct Hstep as [Hstep | Hstep].
  eapply Hcons; eauto; done.

  eapply Hcons'; eauto; done.
Qed.    


Lemma tstc_intro :
  forall (Rt : treln) G e e' t,
    Rt G e e' t
    -> tstc Rt G e e' t.
Proof.
intros Rt G e e' t HR.
apply plus_one; [].
left; assumption.
Qed.


Lemma mstc_intro :
  forall (Rm : mreln) G p e e' t,
    Rm G p e e' t
    -> mstc Rm G p e e' t.
Proof.
intros Rm G p e e' t HR.
apply plus_one; [].
left; assumption.
Qed.


(* Functionality *)

Inductive deq
  (Rt : treln)
  (Rm : context -> purity -> (* md *) term -> (* md *) term -> (* cn *) term -> Prop)
  : context -> term -> term -> (* cl *) term -> Prop :=
| deq_cn {G c c' k}
    : ofc G c k
      -> eqc G c c' k
      -> deq Rt Rm G c c' (cl_cn k)

| deq_tm {G e e' t}
    : Rt G e e' t
      -> deq Rt Rm G e e' (cl_tm t)

| deq_md {G M M' T c}
    : Rm G pure M M' T
      -> fstm G M c
      -> deq Rt Rm G M M' (cl_md T c)

| deq_ucn {G c c'}
    : deq Rt Rm G c c' cl_ucn

| deq_utm {G t t'}
    : deq Rt Rm G t t' cl_utm

| deq_umd {G M M' c}
    : fstm G M c
      -> deq Rt Rm G M M' (cl_umd c)
.



Lemma deq_wclo_ifwf :
  forall Rt Rm,
    csubstitutive Rt Rm
    -> forall G m m' A A',
         wf G
         -> deq Rt Rm G m m' A
         -> deq Rt Rm (G; A') (shift1 m) (shift1 m') (shift1 A).
Proof.
intros Rt Rm (Htsubst & Hmsubst).
intros G m m' C C' Hwf H.
revert Hwf.
#4 invert H; try (intros; subst; simp_sub; eauto using deq; done).

(* cn *)
intros k H0 H1 <- Hwf.
simp_sub.
eapply deq_cn; eauto using weaken1_ofc, weaken1_eqc; done.

(* tm *)
intros t Hm <- Hwf.
simp_sub.
apply deq_tm; [].
eapply Htsubst; eauto; [].
apply subof_sh; eauto with sequence; done.

(* md *)
intros T c Hm Hfst <- Hwf.
simp_sub.
#2 apply deq_md.
  eapply Hmsubst; eauto; [].
  apply subof_sh; eauto with sequence; done.

  eapply subst_fstm; eauto; [].
  apply subof_sh; eauto with sequence; done.

(* umd *)
intros c Hm <- Hwf.
simp_sub.
apply deq_umd; auto; [].
eapply subst_fstm; eauto; [].
apply subof_sh; eauto with sequence; done.
Qed.


Lemma deq_behaved :
  forall Rt Rm,
    creflexive Rt Rm
    -> behavedrel (deq Rt Rm).
Proof.
intros Rt Rm (Htrefl & Hmrefl) G i C HC Hindex.
destruct HC; simp_sub; eauto 6 using deq with static.
Qed.


Lemma deq_seqs :
  forall Rt Rm G m m' C,
    deq Rt Rm G m m' C
    -> seqs G m m' C.
Proof.
intros Rt Rm G m m' C H.
invert H; intros; subst; eauto using seqs.
Qed.


Lemma deq_sof :
  forall Rt Rm G m m' C,
    deq Rt Rm G m m' C
    -> sof G m C.
Proof.
intros Rt Rm G m m' C H.
invert H; intros; subst; eauto using sof.
Qed.


Lemma deq_tof :
  forall Rt Rm G m m' C,
    deq Rt Rm G m m' C
    -> tof G m C.
Proof.
intros Rt Rm G m m' C H.
invert H; intros; subst; eauto using tof.
Qed.


Lemma subrel_deq_sof :
  forall Rt Rm G1 s s' G2,
    subrel (deq Rt Rm) G1 s s' G2
    -> subof sof G1 s G2.
Proof.
eauto using subrel_subof_mono, deq_sof.
Qed.


(* A weaker form of subrel_wclo_weaken1 that requires a well-formed context *)
Lemma subrel_wclo_weaken1_ifwf :
  forall (R : _ -> _ -> _ -> _ -> Prop) G1 G2 s s' A,
    (forall G m m' A A',
       wf G
       -> R G m m' A
       -> R (G; A') (shift1 m) (shift1 m') (shift1 A))
    -> wf G1
    -> subrel R G1 s s' G2
    -> subrel R (G1; A) (compose s sh1) (compose s' sh1) G2.
Proof.
intros R G1 G2 s s' A Hclo Hwf Hs.
revert Hwf.
induct Hs.

(* dot *)
intros G1 t1 t2 s1 s2 A' G2 H H0 H1 Hwf.
simp_sub.
apply subrel_dot; eauto; [].
rewrite -> subst_compose; [].
apply Hclo; auto; done.

(* sh *)
intros G1 n G2 Htrunc Hwf.
simp_sub.
apply subrel_sh; eauto with sequence; done.
Qed.


(* A weaker form of subrel_wclo_bind that requires a well-formed context *)
Lemma subrel_wclo_bind_ifwf :
  forall R,
    behavedrel R
    -> (forall G m m' A A',
          wf G
          -> R G m m' A
          -> R (G; A') (shift1 m) (shift1 m') (shift1 A))
    -> forall G G' s s' A,
         wf G
         -> iscl A
         -> subrel R G s s' G'
         -> subrel R (G; subst s A) (dot (var 0) (compose s sh1)) (dot (var 0) (compose s' sh1)) (G'; A).
Proof.
intros R Hbehaved Hclo G G' s s' A Hwf Hiscl Hss.
#2 apply subrel_dot.
  apply subrel_wclo_weaken1_ifwf; auto; done.

  replsub (subst (compose s sh1) A) with (subst (sh 1) (subst s A)).
  apply Hbehaved; auto using subst_cl with sequence; done.
Qed.


Lemma dynamic_functionality :
  forall Rt Rm,
    creflexive Rt Rm
    -> csubstitutive Rt Rm
    -> compatible eqk eqc eqs Rt Rm
    -> self_respecting Rm
    -> (forall G1 G2 s s' e t,
          wf G1
          -> wf G2
          -> subrel (deq Rt Rm) G1 s s' G2
          -> oft G2 e t
          -> Rt G1 (subst s e) (subst s' e) (subst s t))
       /\
       (forall G1 G2 p s s' M T,
          wf G1
          -> wf G2
          -> subrel (deq Rt Rm) G1 s s' G2
          -> ofm G2 p M T
          -> Rm G1 p (subst s M) (subst s' M) (subst s T)).
Proof.
intros Rt Rm Hrefl Hcsubst Hcompat Hselfresp.
destruct Hcompat as (Htcomp & Hmcomp).
#34 exploit
  (Principal.dynamic_principal_ind
     (fun G2 e t =>
        forall G1 s s',
          wf G1
          -> wf G2
          -> subrel (deq Rt Rm) G1 s s' G2
          -> Rt G1 (subst s e) (subst s' e) (subst s t))
     (fun G2 p M T =>
        forall G1 s s',
          wf G1
          -> wf G2
          -> subrel (deq Rt Rm) G1 s s' G2
          -> Rm G1 p (subst s M) (subst s' M) (subst s T))) as Hind.

(* var *)
intros G2 i t Hindex G1 s s' Hwf1 Hwf Hs.
simp_sub.
so (subrel_index _ (deq_behaved _ _ Hrefl) _#6 iscl_tm Hs Hindex) as H.
simp_sub in H.
invert H; auto; done.

(* star *)
intros; eapply (Hrefl andel); eauto with static; done.

(* lam *)
intros G2 t1 t2 e Hoft1 _ IH G1 s s' Hwf1 Hwf Hs.
simp_sub.
#2 apply Htcomp; auto; apply tcc_lam.
  repl kd_type with (subst s kd_type) by (simp_sub; reflexivity).
  eapply seqs_ofc; eauto; [].
  eapply subrel_mono; try eassumption; eauto using deq_seqs; done.

  repl (shift1 (subst s t2)) with (subst (dot (var 0) (compose s sh1)) (shift1 t2)) by (simp_sub; reflexivity).
  apply IH; eauto using subst_ofc_type, subrel_subof_mono, deq_sof with static; [].
  repl (cl_tm (subst s t1)) with (subst s (cl_tm t1)) by (simp_sub; reflexivity).
  apply subrel_wclo_bind_ifwf; auto using deq_behaved, deq_wclo_ifwf, iscl; done.

(* app *)
intros; simp_sub; apply Htcomp; eauto using tcc; done.

(* pair *)
intros; simp_sub; apply Htcomp; eauto using tcc; done.

(* pi1 *)
intros; simp_sub; apply Htcomp; eauto using tcc; done.

(* pi2 *)
intros; simp_sub; apply Htcomp; eauto using tcc; done.

(* plam *)
intros G2 k t e Hofk _ IH G1 s s' Hwf1 Hwf Hs.
simp_sub.
#2 apply Htcomp; auto; apply tcc_plam.
  eapply seqs_ofk; eauto; [].
  eapply subrel_mono; try eassumption; eauto using deq_seqs; done.

  apply IH; eauto using subst_ofk, subrel_subof_mono, deq_sof with static; [].
  repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
  apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf with static; done.
  
(* papp *)
intros G k t e c _ IH Hofc G1 s s' Hwf1 Hwf Hs.
simp_sub.
so (IH _#3 Hwf1 Hwf Hs) as HR.
simp_sub in HR.
rewrite -> split_dot; rewrite -> subst_compose.
apply Htcomp; auto; eapply tcc_papp; eauto; [].
eapply seqs_ofc; eauto; [].
eapply subrel_mono; try eassumption; eauto using deq_seqs; done.

(* pack *)
intros G k t c e Hofc _ IH Hoft G1 s s' Hwf1 Hwf Hs.
simp_sub.
#4 apply Htcomp; auto; eapply tcc_pack.
  eapply seqs_ofc; eauto; [].
  eapply subrel_mono; try eassumption; eauto using deq_seqs; done.

  #2 eapply seqs_ofk; eauto.
    eapply subrel_mono; try eassumption; eauto using deq_seqs; done.

    eapply ofc_reg; eauto; done.

  simp_sub.
  so (IH _ _ _ Hwf1 Hwf Hs) as HR.
  simp_sub in HR.
  assumption.

  repl kd_type with (subst (dot (var 0) (compose s sh1)) kd_type) by (simp_sub; reflexivity).
  eapply seqs_ofc; eauto; [].
  repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
  apply subrel_wclo_bind_ifwf; eauto using seqs_behaved, seqs_wclo with static; [].
  eapply subrel_mono; try eassumption; eauto using deq_seqs; done.

(* unpack *)
intros G k t t' e1 e2 Hofe1 IH1 _ IH2 Hoft' G1 s s' Hwf1 Hwf Hs.
simp_sub.
so (IH1 _ _ _ Hwf1 Hwf Hs) as HR1.
simp_sub in HR1.
#2 apply Htcomp; auto; eapply tcc_unpack; eauto.
  repl (subst (sh 2) (subst s t')) with (subst (dot (var 0) (dot (var 1) (compose s (sh 2)))) (subst (sh 2) t')) by (simp_sub; reflexivity).
  so (ofc_ofcr _#3 (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe1))) as (l & Hofex).
  invert Hofex; [].
  intros Hofk Hoft _.
  #2 apply IH2; eauto with static.
    #2 apply wf_tm.
      apply wf_cn; eauto using subst_ofk, subrel_deq_sof; done.

      eapply subst_ofc_type; eauto using subof_bind_sof_cn, subrel_deq_sof; done.

    repl (cl_tm (subst (dot (var 0) (compose s sh1)) t)) with (subst (dot (var 0) (compose s sh1)) (cl_tm t)) by (simp_sub; reflexivity).
    repl (dot (var 0) (dot (var 1) (compose s (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s sh1)) sh1)) by (simp_sub; reflexivity).
    repl (dot (var 0) (dot (var 1) (compose s' (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s' sh1)) sh1)) by (simp_sub; reflexivity).
    #2 apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf with static.
      apply wf_cn; eauto using subst_ofk, subrel_deq_sof; done.

      repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
      apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf with static; done.

  repl kd_type with (subst s kd_type) by (simp_sub; reflexivity).
  eapply subst_ofc; eauto; [].
  eapply subrel_subof_mono; eauto using deq_sof; done.

(* fix *)
intros G t e Hofe IH G1 s s' Hwf1 Hwf Hs.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe))) as (kk & Hofcr).
invert Hofcr; [].
intros _ Hoft _; clear kk Hofcr.
simp_sub.
#2 apply Htcomp; auto; eapply tcc_fix.
  repl kd_type with (subst s kd_type) by (simp_sub; reflexivity).
  eapply seqs_ofc; eauto; [].
  eapply subrel_mono; try eassumption; eauto using deq_seqs; done.

  so (IH _ _ _ Hwf1 Hwf Hs) as HR.
  simp_sub in HR.
  assumption.

(* lett *)
intros G t1 t2 e1 e2 Hofe1 IH1 _ IH2 G1 s s' Hwf1 Hwf Hs.
simp_sub.
apply Htcomp; auto; eapply tcc_lett; eauto; [].
repl (shift1 (subst s t2)) with (subst (dot (var 0) (compose s sh1)) (shift1 t2)) by (simp_sub; reflexivity).
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe1)) as Hoft1.
apply IH2; eauto using wf, subst_ofc_type, subrel_deq_sof; [].
repl (cl_tm (subst s t1)) with (subst s (cl_tm t1)) by (simp_sub; reflexivity).
apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf with static; done.

(* equiv *)
intros G t t' e _ IH Heqc G1 s s' Hwf1 Hwf Hs.
apply Htcomp; auto; eapply tcc_equiv; eauto; [].
repl kd_type with (subst s kd_type) by (simp_sub; reflexivity).
eapply subst_eqc; eauto; [].
eapply subrel_subof_mono; eauto using deq_sof; done.

(* letm *)
intros G T k t M e HofM IH1 HfstT _ IH2 Hoft G1 s s' Hwf1 Hwf Hs.
simp_sub.
#2 apply Htcomp; auto; eapply tcc_letm; eauto using subst_fsts.
  repl (subst (sh 2) (subst s t)) with (subst (dot (var 0) (dot (var 1) (compose s (sh 2)))) (subst (sh 2) t)) by (simp_sub; reflexivity).
  so (ofm_reg _#4 Hwf (ofmp_ofm _#4 Hwf HofM)) as HofT.
  so (ofs_fst _#3 HofT HfstT) as Hofk.
  apply IH2; eauto using wf, subst_ofs, subst_fsts, subrel_deq_sof; [].
  repl (dot (var 0) (dot (var 1) (compose s (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s sh1)) sh1)) by (simp_sub; reflexivity).
  repl (dot (var 0) (dot (var 1) (compose s' (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s' sh1)) sh1)) by (simp_sub; reflexivity).
  repl (cl_md (shift1 (subst s T)) (var 0)) with (subst (dot (var 0) (compose s sh1)) (cl_md (shift1 T) (var 0))) by (simp_sub; reflexivity).
  apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf, subst_ofk, subst_ofc_type, subrel_deq_sof with static; [].
  repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
  apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf with static; done.

  repl kd_type with (subst s kd_type) by (simp_sub; reflexivity).
  eapply subst_ofc; eauto; [].
  eapply subrel_subof_mono; eauto using deq_sof; done.

(* ext *)
intros; simp_sub; apply Htcomp; eauto using tcc; done.


(* var *)
intros G i T c Hindex G1 s s' Hwf1 Hwf Hs.
simp_sub.
so (subrel_index _ (deq_behaved _ _ Hrefl) _#6 iscl_md Hs Hindex) as H.
simp_sub in H.
assert (Rm G1 pure (project s i) (project s' i) (subst (compose (sh (1 + i)) s) T)
        /\ fstm G1 (project s i) (subst (compose (sh (1 + i)) s) c)) as (HR & Hfstm).
  invert H; auto; done.
so (wf_index_md _#4 Hwf Hindex) as (HofT & _).
rewrite -> subst_singles; eauto using ofs_hsg, subst_hsg; [].
#3 apply Hselfresp; auto.
  eapply subst_ofs; eauto using subrel_deq_sof; done.

  simp_sub.
  assumption.

  simp_sub.
  assumption.

(* star *)
intros; eapply (Hrefl ander); eauto with static; done.

(* stat *)
intros G c k Hofc G1 s s' Hwf1 Hwf Hs.
simp_sub.
apply Hmcomp; auto; eapply mcc_stat; eauto; [].
so (single_ofc _#3 Hwf Hofc) as Hofcs.
eapply seqs_ofc; eauto; [].
eapply subrel_mono; try eassumption; eauto using deq_seqs; done.

(* dyn *)
intros; simp_sub; apply Hmcomp; eauto using mcc; done.

(* lamgn *)
intros G T1 T2 k1 M HofT1 HfstT1 HofM IH G1 s s' Hwf1 Hwf Hs.
simp_sub.
#2 apply Hmcomp; auto; eapply mcc_lamgn; eauto using subst_fsts.
  eapply seqs_ofs; eauto; [].
  eapply subrel_mono; try eassumption; eauto using deq_seqs; done.

  repl (shift1 (subst (dot (var 0) (compose s sh1)) T2)) with (subst (dot (var 0) (dot (var 1) (compose s (sh 2)))) (shift1 T2)) by (simp_sub; reflexivity).
  apply IH; eauto using wf, subst_ofs, subst_fsts, subrel_deq_sof; [].
  repl (dot (var 0) (dot (var 1) (compose s (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s sh1)) sh1)) by (simp_sub; reflexivity).
  repl (dot (var 0) (dot (var 1) (compose s' (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s' sh1)) sh1)) by (simp_sub; reflexivity).
  repl (cl_md (shift1 (subst s T1)) (var 0)) with (subst (dot (var 0) (compose s sh1)) (cl_md (shift1 T1) (var 0))) by (simp_sub; reflexivity).
  so (ofs_fst _#3 HofT1 HfstT1) as Hofk1.
  apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf, subst_ofk, subrel_deq_sof with static; [].
  repl (cl_cn (subst s k1)) with (subst s (cl_cn k1)) by (simp_sub; reflexivity).
  apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf with static; done.

(* appgn *)
intros G T1 T2 M1 M2 c2 HofM1 IH1 _ IH2 HfstM2 G1 s s' Hwf1 Hwf Hs.
simp_sub.
rewrite -> split_dot; rewrite -> subst_compose.
#2 apply Hmcomp; auto; eapply mcc_appgn; eauto.
  so (IH1 _ _ _ Hwf1 Hwf Hs) as H.
  simp_sub in H.
  assumption.

  eapply subst_fstm; eauto; [].
  eapply subrel_subof_mono; eauto using deq_tof; done.

(* lamap *)
intros G T1 T2 k1 M HofT1 HfstT1 HofM IH G1 s s' Hwf1 Hwf Hs.
simp_sub.
#2 apply Hmcomp; auto; eapply mcc_lamap; eauto using subst_fsts.
  eapply seqs_ofs; eauto; [].
  eapply subrel_mono; try eassumption; eauto using deq_seqs; done.

  repl (shift1 (subst (dot (var 0) (compose s sh1)) T2)) with (subst (dot (var 0) (dot (var 1) (compose s (sh 2)))) (shift1 T2)) by (simp_sub; reflexivity).
  apply IH; eauto using wf, subst_ofs, subst_fsts, subrel_deq_sof; [].
  repl (dot (var 0) (dot (var 1) (compose s (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s sh1)) sh1)) by (simp_sub; reflexivity).
  repl (dot (var 0) (dot (var 1) (compose s' (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s' sh1)) sh1)) by (simp_sub; reflexivity).
  repl (cl_md (shift1 (subst s T1)) (var 0)) with (subst (dot (var 0) (compose s sh1)) (cl_md (shift1 T1) (var 0))) by (simp_sub; reflexivity).
  so (ofs_fst _#3 HofT1 HfstT1) as Hofk1.
  apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf, subst_ofk, subrel_deq_sof with static; [].
  repl (cl_cn (subst s k1)) with (subst s (cl_cn k1)) by (simp_sub; reflexivity).
  apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf with static; done.

(* appap *)
intros G p T1 T2 M1 M2 c2 HofM1 IH1 _ IH2 HfstM2 G1 s s' Hwf1 Hwf Hs.
simp_sub.
rewrite -> split_dot; rewrite -> subst_compose.
#2 apply Hmcomp; auto; eapply mcc_appap; eauto.
  so (IH1 _ _ _ Hwf1 Hwf Hs) as H.
  simp_sub in H.
  assumption.

  eapply subst_fstm; eauto; [].
  eapply subrel_subof_mono; eauto using deq_tof; done.

(* pair *)
intros G p T1 T2 M1 M2 _ IH1 _ IH2 G1 s s' Hwf1 Hwf Hs.
simp_sub.
rewrite -> subst_compose.
apply Hmcomp; auto; eapply mcc_pair; eauto; done.

(* pi1 *)
intros; simp_sub; apply Hmcomp; eauto using mcc; done.

(* pi2 *)
intros G T1 T2 M c _ IH Hfst G1 s s' Hwf1 Hwf Hs.
simp_sub.
rewrite -> split_dot; rewrite -> subst_compose.
apply Hmcomp; auto; [].
#2 eapply (@mcc_pi2 _#6 (subst s T1)).
  so (IH _ _ _ Hwf1 Hwf Hs) as H.
  simp_sub in H.
  assumption.

  eapply subst_fstm; eauto; [].
  eapply subrel_subof_mono; eauto using deq_tof; done.

(* unpack *)
intros G k t T e M Hofe IH1 _ IH2 HofT G1 s s' Hwf1 Hwf Hs.
simp_sub.
so (IH1 _ _ _ Hwf1 Hwf Hs) as HR1.
simp_sub in HR1.
#2 apply Hmcomp; auto; eapply mcc_unpack; eauto.
  repl (subst (sh 2) (subst s T)) with (subst (dot (var 0) (dot (var 1) (compose s (sh 2)))) (subst (sh 2) T)) by (simp_sub; reflexivity).
  so (ofc_ofcr _#3 (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe))) as (l & Hofex).
  invert Hofex; [].
  intros Hofk Hoft _.
  #2 apply IH2; eauto with static.
    #2 apply wf_tm.
      apply wf_cn; eauto using subst_ofk, subrel_deq_sof; done.

      eapply subst_ofc_type; eauto using subof_bind_sof_cn, subrel_deq_sof; done.
    
    repl (cl_tm (subst (dot (var 0) (compose s sh1)) t)) with (subst (dot (var 0) (compose s sh1)) (cl_tm t)) by (simp_sub; reflexivity).
    repl (dot (var 0) (dot (var 1) (compose s (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s sh1)) sh1)) by (simp_sub; reflexivity).
    repl (dot (var 0) (dot (var 1) (compose s' (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s' sh1)) sh1)) by (simp_sub; reflexivity).
    #2 apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf with static.
      apply wf_cn; eauto using subst_ofk, subrel_deq_sof; done.

      repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
      apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf with static; done.

  eapply seqs_ofs; eauto; [].
  eapply subrel_mono; try eassumption; eauto using deq_seqs; done.

(* lett *)
intros G p t T e M Hofe IH1 _ IH2 G1 s s' Hwf1 Hwf Hs.
simp_sub.
apply Hmcomp; auto; eapply mcc_lett; eauto; [].
repl (shift1 (subst s T)) with (subst (dot (var 0) (compose s sh1)) (shift1 T)) by (simp_sub; reflexivity).
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe)) as Hoft.
apply IH2; eauto using wf, subst_ofc_type, subrel_deq_sof; [].
repl (cl_tm (subst s t)) with (subst s (cl_tm t)) by (simp_sub; reflexivity).
apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf with static; done.

(* letm *)
intros G T1 T2 k1 M1 M2 HofM IH1 Hfst _ IH2 HofT2 G1 s s' Hwf1 Hwf Hs.
simp_sub.
#2 apply Hmcomp; auto; eapply mcc_letm; eauto using subst_fsts.
  repl (subst (sh 2) (subst s T2)) with (subst (dot (var 0) (dot (var 1) (compose s (sh 2)))) (subst (sh 2) T2)) by (simp_sub; reflexivity).
  so (ofm_reg _#4 Hwf (ofmp_ofm _#4 Hwf HofM)) as HofT.
  so (ofs_fst _#3 HofT Hfst) as Hofk.
  apply IH2; eauto using wf, subst_ofs, subst_fsts, subrel_deq_sof; [].
  repl (dot (var 0) (dot (var 1) (compose s (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s sh1)) sh1)) by (simp_sub; reflexivity).
  repl (dot (var 0) (dot (var 1) (compose s' (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s' sh1)) sh1)) by (simp_sub; reflexivity).
  repl (cl_md (shift1 (subst s T1)) (var 0)) with (subst (dot (var 0) (compose s sh1)) (cl_md (shift1 T1) (var 0))) by (simp_sub; reflexivity).
  apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf, subst_ofk, subrel_deq_sof with static; [].
  repl (cl_cn (subst s k1)) with (subst s (cl_cn k1)) by (simp_sub; reflexivity).
  apply subrel_wclo_bind_ifwf; eauto using deq_behaved, deq_wclo_ifwf with static; done.

  eapply seqs_ofs; eauto; [].
  eapply subrel_mono; try eassumption; eauto using deq_seqs; done.

(* seal *)
intros G T M HofM IH G1 s s' Hwf1 Hwf Hs.
so (ofm_reg _#4 Hwf (ofmp_ofm _#4 Hwf HofM)) as HofT.
simp_sub.
apply Hmcomp; auto; eapply mcc_seal; eauto; [].
eapply seqs_ofs; eauto; [].
eapply subrel_mono; try eassumption; eauto using deq_seqs; done.

(* forget *)
intros; simp_sub; apply Hmcomp; eauto using mcc; done.

(* subsume *)
intros G p T T' M _ IH Hsubs G1 s s' Hwf1 Hwf Hs.
apply Hmcomp; auto; eapply mcc_subsume; eauto; [].
eapply subst_subs; eauto; [].
eapply subrel_subof_mono; eauto using deq_sof; done.

(* epilogue *)
destruct Hind.
split; intros; eauto using oft_oftp, ofm_ofmp; done.
Qed.


Lemma subst_deq_tm :
  forall Rt Rm,
    creflexive Rt Rm
    -> csubstitutive Rt Rm
    -> compatible eqk eqc eqs Rt Rm
    -> self_respecting Rm
    -> forall G1 G2 s s' e t,
         wf G1
         -> wf G2
         -> subrel (deq Rt Rm) G1 s s' G2
         -> oft G2 e t
         -> Rt G1 (subst s e) (subst s' e) (subst s t).
Proof.
intros.
refine (dynamic_functionality _#6 andel _#10); eauto.
Qed.


Lemma subst_deq_md :
  forall Rt Rm,
    creflexive Rt Rm
    -> csubstitutive Rt Rm
    -> compatible eqk eqc eqs Rt Rm
    -> self_respecting Rm
    -> forall G1 G2 p s s' M T,
         wf G1
         -> wf G2
         -> subrel (deq Rt Rm) G1 s s' G2
         -> ofm G2 p M T
         -> Rm G1 p (subst s M) (subst s' M) (subst s T).
Proof.
intros.
refine (dynamic_functionality _#6 ander _#11); eauto.
Qed.
