
Require Import Tactics.
Require Import Sequence.
Require Import Subst.
Require Import Static.
Require Import Regularity.
Require Import Algorithm.
Require Import Declarative.
Require Import Sound.
Require Import Complete.


Lemma arrow_injective :
  forall G t1 t1' t2 t2',
    wf G
    -> eqc G (cn_arrow t1 t2) (cn_arrow t1' t2') kd_type
    -> eqc G t1 t1' kd_type /\ eqc G t2 t2' kd_type.
Proof.
intros G t1 t1' t2 t2' Hwf Heqc.
so (aeq_complete _#4 Hwf Heqc) as Haeq.
invert Haeq; [].
intros p p' Hnorm Hnorm' Hpeq.
assert (headnorm G (cn_arrow t1 t2) (cn_arrow t1 t2)) as Hnormr by (auto with alg).
assert (headnorm G (cn_arrow t1' t2') (cn_arrow t1' t2')) as Hnormr' by (auto with alg).
so (headnorm_fun _#4 Hnorm Hnormr) as Heq.
so (headnorm_fun _#4 Hnorm' Hnormr') as Heq'.
subst p p'.
invert Hpeq; [].
intros Haeq1 Haeq2.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc andel)) as (l & Hofcr).
invert Hofcr.
intros Hoft1 Hoft2.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc ander andel)) as (l' & Hofcr').
invert Hofcr'.
intros Hoft1' Hoft2'.
split; eapply aeq_sound; eauto using ctxeq_refl with static; done.
Qed.


Lemma prod_injective :
  forall G t1 t1' t2 t2',
    wf G
    -> eqc G (cn_prod t1 t2) (cn_prod t1' t2') kd_type
    -> eqc G t1 t1' kd_type /\ eqc G t2 t2' kd_type.
Proof.
intros G t1 t1' t2 t2' Hwf Heqc.
so (aeq_complete _#4 Hwf Heqc) as Haeq.
invert Haeq; [].
intros p p' Hnorm Hnorm' Hpeq.
assert (headnorm G (cn_prod t1 t2) (cn_prod t1 t2)) as Hnormr by (auto with alg).
assert (headnorm G (cn_prod t1' t2') (cn_prod t1' t2')) as Hnormr' by (auto with alg).
so (headnorm_fun _#4 Hnorm Hnormr) as Heq.
so (headnorm_fun _#4 Hnorm' Hnormr') as Heq'.
subst p p'.
invert Hpeq; [].
intros Haeq1 Haeq2.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc andel)) as (l & Hofcr).
invert Hofcr.
intros Hoft1 Hoft2.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc ander andel)) as (l' & Hofcr').
invert Hofcr'.
intros Hoft1' Hoft2'.
split; eapply aeq_sound; eauto using ctxeq_refl with static; done.
Qed.


Lemma all_injective :
  forall G k k' t t',
    wf G
    -> eqc G (cn_all k t) (cn_all k' t') kd_type
    -> eqk G k k' /\ eqc (G; cl_cn k) t t' kd_type.
Proof.
intros G k k' t t' Hwf Heqc.
so (aeq_complete _#4 Hwf Heqc) as Haeq.
invert Haeq; [].
intros p p' Hnorm Hnorm' Hpeq.
assert (headnorm G (cn_all k t) (cn_all k t)) as Hnormr by (auto with alg).
assert (headnorm G (cn_all k' t') (cn_all k' t')) as Hnormr' by (auto with alg).
so (headnorm_fun _#4 Hnorm Hnormr) as Heq.
so (headnorm_fun _#4 Hnorm' Hnormr') as Heq'.
subst p p'.
invert Hpeq; [].
intros Haeq1 Haeq2.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc andel)) as (l & Hofcr).
invert Hofcr.
intros Hofk Hofc.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc ander andel)) as (l' & Hofcr').
invert Hofcr'.
intros Hofk' Hofc'.
assert (eqk G k k') as Heqk.
  eapply aeqk_sound; eauto using ctxeq_refl with static; done.
split; auto; [].
eapply (aeq_sound _ (G; cl_cn k')); eauto with static; [].
apply ctxeq_cn; auto using ctxeq_refl; done.
Qed.


Lemma exists_injective :
  forall G k k' t t',
    wf G
    -> eqc G (cn_exists k t) (cn_exists k' t') kd_type
    -> eqk G k k' /\ eqc (G; cl_cn k) t t' kd_type.
Proof.
intros G k k' t t' Hwf Heqc.
so (aeq_complete _#4 Hwf Heqc) as Haeq.
invert Haeq; [].
intros p p' Hnorm Hnorm' Hpeq.
assert (headnorm G (cn_exists k t) (cn_exists k t)) as Hnormr by (auto with alg).
assert (headnorm G (cn_exists k' t') (cn_exists k' t')) as Hnormr' by (auto with alg).
so (headnorm_fun _#4 Hnorm Hnormr) as Heq.
so (headnorm_fun _#4 Hnorm' Hnormr') as Heq'.
subst p p'.
invert Hpeq; [].
intros Haeq1 Haeq2.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc andel)) as (l & Hofcr).
invert Hofcr.
intros Hofk Hofc.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc ander andel)) as (l' & Hofcr').
invert Hofcr'.
intros Hofk' Hofc'.
assert (eqk G k k') as Heqk.
  eapply aeqk_sound; eauto using ctxeq_refl with static; done.
split; auto; [].
eapply (aeq_sound _ (G; cl_cn k')); eauto with static; [].
apply ctxeq_cn; auto using ctxeq_refl; done.
Qed.


Inductive similar : (* cn *) term -> (* cn *) term -> Prop :=
| similar_unit
    : similar cn_unit cn_unit

| similar_arrow {t1 t1' t2 t2'}
    : similar (cn_arrow t1 t2) (cn_arrow t1' t2')

| similar_prod {t1 t1' t2 t2'}
    : similar (cn_prod t1 t2) (cn_prod t1' t2')

| similar_all {k k' t t'}
    : similar (cn_all k t) (cn_all k' t')

| similar_exists {k k' t t'}
    : similar (cn_exists k t) (cn_exists k' t')
.


(* Distinct base types are not equal. *)
Lemma consistent :
  forall G c c',
    wf G
    -> eqc G c c' kd_type
    -> similar c c
    -> similar c' c'
    -> similar c c'.
Proof.
intros G c c' Hwf Heqc Hbase Hbase'.
so (aeq_complete _#4 Hwf Heqc) as Haeq.
invert Haeq; [].
intros p p' Hnorm Hnorm' Hpeq.
assert (headnorm G c c) as Hnormr.
  apply headnorm_done; [].
  invert Hbase; intros; subst; eauto using natk; done.
assert (headnorm G c' c') as Hnormr'.
  apply headnorm_done; [].
  invert Hbase'; intros; subst; eauto using natk; done.
so (headnorm_fun _#4 Hnorm Hnormr) as Heq.
so (headnorm_fun _#4 Hnorm' Hnormr') as Heq'.
subst p p'.
invert Hpeq; invert Hbase; invert Hbase'; intros; subst; try discriminate; auto using similar.
Qed.
