
Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import TSubst.
Require Import Formation.
Require Import Dynamic.
Require Import Regularity.
Require Import Singleton.
Require Import Declarative.
Require Import Breduce.
Require Import Congruence.
Require Import BreduceCong.
Require Import Preservation.


Definition capproxt (G : context) (e e' : (* tm *) term) (t : (* cn *) term) :=
  exists Rt Rm,
    semicongruence Rt Rm
    /\ consistent Rm
    /\ Rt G e e' t.


Definition capproxm (G : context) (p : purity) (M M' : (* md *) term) (T : (* sg *) term) :=
  exists Rt Rm,
    semicongruence Rt Rm
    /\ consistent Rm
    /\ Rm G p M M' T.


Definition ceqt (G : context) (e e' : (* tm *) term) (t : (* cn *) term) :=
  capproxt G e e' t /\ capproxt G e' e t.
      

Definition ceqm (G : context) (p : purity) (M M' : (* md *) term) (T : (* sg *) term) :=
  capproxm G p M M' T /\ capproxm G p M' M T.


Lemma capproxt_univ :
  forall Rt Rm G e e' t,
    semicongruence Rt Rm
    -> consistent Rm
    -> Rt G e e' t
    -> capproxt G e e' t.
Proof.
intros Rt Rm G e e' t Hsc Hcons HR.
exists Rt, Rm.
do2 2 split; trivial; done.
Qed.


Lemma capproxm_univ :
  forall Rt Rm G p M M' T,
    semicongruence Rt Rm
    -> consistent Rm
    -> Rm G p M M' T
    -> capproxm G p M M' T.
Proof.
intros Rt Rm G p M M' T Hsc Hcons HR.
exists Rt, Rm.
do2 2 split; trivial; done.
Qed.


Lemma capproxt_univ_quasi :
  forall Rt Rm G e e' t,
    quasicongruence Rt Rm
    -> consistent Rm
    -> Rt G e e' t
    -> capproxt G e e' t.
Proof.
intros Rt Rm G e e' t Hquasi Hcon HR.
#3 apply (capproxt_univ (ttc Rt) (mtc Rm)).
  apply tc_semicongruence; auto; done.

  apply mtc_consistent; auto; done.

  eapply ttc_intro; eauto; done.
Qed.


Lemma capproxm_univ_quasi :
  forall Rt Rm G p M M' T,
    quasicongruence Rt Rm
    -> consistent Rm
    -> Rm G p M M' T
    -> capproxm G p M M' T.
Proof.
intros Rt Rm G p M M' T Hquasi Hcon HR.
#3 apply (capproxm_univ (ttc Rt) (mtc Rm)).
  apply tc_semicongruence; auto; done.

  apply mtc_consistent; auto; done.

  eapply mtc_intro; eauto; done.
Qed.


Lemma capprox_cregular :
  cregular capproxt capproxm.
Proof.
do2 2 split.
  intros G e e' t Hwf Hap.
  destruct Hap as (Rt & Rm & Hsc & _ & HR).
  apply (Hsc andel andel); auto; done.

  intros G p M M' T Hwf Hap.
  destruct Hap as (Rt & Rm & Hsc & _ & HR).
  apply (Hsc andel anderl); auto; done.

  intros G M M' T Hwf Hap.
  destruct Hap as (Rt & Rm & Hsc & _ & HR).
  apply (Hsc andel anderr); auto; done.
Qed.


Lemma capproxt_reg :
  forall G e e' t,
    wf G
    -> capproxt G e e' t
    -> oft G e t /\ oft G e' t.
Proof.
exact (capprox_cregular andel).
Qed.


Lemma capproxm_reg :
  forall G p M M' T,
    wf G
    -> capproxm G p M M' T
    -> ofm G p M T /\ ofm G p M' T.
Proof.
exact (capprox_cregular anderl).
Qed.


Lemma capproxm_fst :
  forall G M M' T,
    wf G
    -> capproxm G pure M M' T
    -> exists k c c',
         fsts T k
         /\ fstm G M c
         /\ fstm G M' c'
         /\ eqc G c c' k.
Proof.
exact (capprox_cregular anderr).
Qed.


Lemma treduce_capproxt :
  forall G e e' t,
    oft G e t
    -> oft G e' t
    -> treduce G e e'
    -> capproxt G e e' t.
Proof.
intros G e e' t Hof Hof' Hred.
apply (capproxt_univ treducer mreducer); auto using reducer_semicongruence, reducer_consistent; [].
eapply treg_intro; eauto using reduce_csubstitutive.
Qed.


Lemma anti_treduce_capproxt :
  forall G e e' t,
    oft G e t
    -> oft G e' t
    -> treduce G e e'
    -> capproxt G e' e t.
Proof.
intros G e e' t Hof Hof' Hred.
apply (capproxt_univ (treg (fun G e e' t => treduce G e' e)) (mreg (fun G p M M' T => mreduce G M' M))); auto using anti_reducer_semicongruence, anti_reducer_consistent; [].
eapply treg_intro; eauto using transpose_csubstitutive, reduce_csubstitutive.
Qed.


Lemma mreduce_capproxm_impure :
  forall G M M' T,
    ofm G impure M T
    -> ofm G impure M' T
    -> mreduce G M M'
    -> capproxm G impure M M' T.
Proof.
intros G M M' T Hof Hof' Hred.
apply (capproxm_univ treducer mreducer); auto using reducer_semicongruence, reducer_consistent; [].
eapply mreg_intro_impure; eauto using reduce_csubstitutive.
Qed.


Lemma anti_mreduce_capproxm_impure :
  forall G M M' T,
    ofm G impure M T
    -> ofm G impure M' T
    -> mreduce G M M'
    -> capproxm G impure M' M T.
Proof.
intros G M M' T Hof Hof' Hred.
apply (capproxm_univ (treg (fun G e e' t => treduce G e' e)) (mreg (fun G p M M' T => mreduce G M' M))); auto using anti_reducer_semicongruence, anti_reducer_consistent; [].
eapply mreg_intro_impure; eauto using transpose_csubstitutive, reduce_csubstitutive.
Qed.


Lemma mreduce_capproxm_pure :
  forall G M M' T c c' k,
    ofm G pure M T
    -> ofm G pure M' T
    -> fsts T k
    -> fstm G M c
    -> fstm G M' c'
    -> eqc G c c' k
    -> mreduce G M M'
    -> capproxm G pure M M' T.
Proof.
intros G M M' T c c' k Hof Hof' Hfst Hfstm Hfstm' Heqc Hred.
apply (capproxm_univ treducer mreducer); auto using reducer_semicongruence, reducer_consistent; [].
eapply mreg_intro_pure; eauto using reduce_csubstitutive.
Qed.


Lemma anti_mreduce_capproxm_pure :
  forall G M M' T c c' k,
    ofm G pure M T
    -> ofm G pure M' T
    -> fsts T k
    -> fstm G M c
    -> fstm G M' c'
    -> eqc G c c' k
    -> mreduce G M M'
    -> capproxm G pure M' M T.
Proof.
intros G M M' T c c' k Hof Hof' Hfst Hfstm Hfstm' Heqc Hred.
apply (capproxm_univ (treg (fun G e e' t => treduce G e' e)) (mreg (fun G p M M' T => mreduce G M' M))); auto using anti_reducer_semicongruence, anti_reducer_consistent; [].
eapply mreg_intro_pure; eauto using transpose_csubstitutive, reduce_csubstitutive, eqc_symm.
Qed.


Theorem capprox_consistent :
  consistent capproxm.
Proof.
intros M M' T Hap Hhalt.
destruct Hap as (Rt & Rm & _ & Hcons & HR).
exact (Hcons _ _ _ HR Hhalt).
Qed.


Lemma capproxt_refl :
  forall G e t, oft G e t -> capproxt G e e t.
Proof.
intros G e t Hof.
eapply treduce_capproxt; auto with reduce.
Qed.


Lemma capproxm_refl :
  forall G p M T, ofm G p M T -> capproxm G p M M T.
Proof.
intros G p M T Hof.
apply (capproxm_univ treducer mreducer); auto using reducer_semicongruence, reducer_consistent; [].
do2 3 split; auto.
  apply mreduce_refl; done.

  intros -> Hwf.
  so (can_fsts _ _ (ofm_reg _#4 Hwf Hof)) as (k & Hfst).
  so (can_fstm _#3 Hwf Hof) as (c & Hfstm).
  so (ofm_fst _#6 Hwf Hof Hfstm Hfst) as Hofc.
  exists k, c, c.
  auto with static; done.
Qed.


Lemma capprox_creflexive :
  creflexive capproxt capproxm.
Proof.
split.
  exact capproxt_refl.

  exact capproxm_refl.
Qed.


Lemma capprox_csubstitutive :
  csubstitutive capproxt capproxm.
Proof.
split.
  intros G1 s G2 e e' t Hs Hwf (Rt & Rm & Hsc & Hcons & HR).
  exists Rt, Rm.
  do2 2 split; auto; [].
  eapply (Hsc anderrrl andel); eauto; done.

  intros G1 s G2 p M M' T Hs Hwf (Rt & Rm & Hsc & Hcons & HR).
  exists Rt, Rm.
  do2 2 split; auto; [].
  eapply (Hsc anderrrl ander); eauto; done.
Qed.


Lemma subst_capproxt :
  forall G1 s G2 e e' t,
    subof aof G1 s G2 
    -> wf G2
    -> capproxt G2 e e' t
    -> capproxt G1 (subst s e) (subst s e') (subst s t).
Proof.
exact (capprox_csubstitutive andel).
Qed.


Lemma subst_capproxm :
  forall G1 s G2 p M M' T,
    subof aof G1 s G2 
    -> wf G2
    -> capproxm G2 p M M' T
    -> capproxm G1 p (subst s M) (subst s M') (subst s T).
Proof.
exact (capprox_csubstitutive ander).
Qed.


Lemma capprox_weakly_compatible :
  weakly_compatible samek samec sames capproxt capproxm.
Proof.
unfold capproxt, capproxm.
split.
  intros G e e' t Hwf Htcc.
  revert Hwf.
  induct Htcc;
  try (intros;
       (match goal with
        | H : exists Rt Rm, _ |- _ =>
          destruct H as (? & ? & Hcong & ? & ?);
          so Hcong as (_ & (? & ?) & _ & (Ht & Hm))
        end);
       do 2 eexists;
       do2 2 split; eauto;
       first [apply Ht | apply Hm];
       eauto using tcc;
       done); done.

  intros G p M M' T Hwf Hmcc.
  revert Hwf.
  induct Hmcc;
  try (intros;
       (match goal with
        | H : exists Rt Rm, _ |- _ =>
          destruct H as (? & ? & Hcong & ? & ?);
          so Hcong as (_ & (? & ?) & _ & _ & (Ht & Hm))
        end);
       do 2 eexists;
       do2 2 split; eauto;
       first [apply Ht | apply Hm];
       eauto using mcc;
       done);
  [].
  (* stat *)
  intros G c c' k Hsame Hwf.
  invert Hsame; intros Hofc <-; [].
  exists treducer, mreducer.
  do2 2 split; auto using reducer_semicongruence, reducer_consistent; [].
  apply (reducer_semicongruence ander andel ander); auto with static; done.
Qed.


Lemma tc_capprox_semicongruence :
  semicongruence (ttc capproxt) (mtc capproxm).
Proof.
  apply tc_semicongruence_strong; auto using capprox_cregular, capprox_creflexive, capprox_csubstitutive, capprox_weakly_compatible; done.
Qed.


Lemma capprox_ctransitive :
  ctransitive capproxt capproxm.
Proof.
split.
  intros G t e1 e2 e3 H12 H23.
  eapply capproxt_univ; eauto using tc_capprox_semicongruence, mtc_consistent, capprox_consistent; [].
  eapply plus_trans; eauto using plus_one; done.

  intros G p T M1 M2 M3 H12 H23.
  eapply capproxm_univ; eauto using tc_capprox_semicongruence, mtc_consistent, capprox_consistent; [].
  eapply plus_trans; eauto using plus_one; done.
Qed.


Lemma capproxt_trans :
  forall G e1 e2 e3 t,
    capproxt G e1 e2 t
    -> capproxt G e2 e3 t
    -> capproxt G e1 e3 t.
Proof.
intros.
eapply (capprox_ctransitive andel); eauto.
Qed.


Lemma capproxm_trans :
  forall G p M1 M2 M3 T,
    capproxm G p M1 M2 T
    -> capproxm G p M2 M3 T
    -> capproxm G p M1 M3 T.
Proof.
intros.
eapply (capprox_ctransitive ander); eauto.
Qed.


Lemma capprox_compatible_weak :
  compatible samek samec sames capproxt capproxm.
Proof.
split.
  intros G e e' t Hwf Hcc.
  eapply capproxt_univ; eauto using tc_capprox_semicongruence, mtc_consistent, capprox_consistent; [].
  apply (tc_capprox_semicongruence anderr anderr andel); auto; [].
  #2 apply (tcc_mono samek _ samec _ sames _ capproxt _ capproxm); auto.
    intros; apply plus_one; auto; done.

    intros; apply plus_one; auto; done.

  intros G p M M' T Hwf Hcc.
  eapply capproxm_univ; eauto using tc_capprox_semicongruence, mtc_consistent, capprox_consistent; [].
  apply (tc_capprox_semicongruence anderr anderr ander); auto; [].
  #2 apply (mcc_mono samek _ samec _ sames _ capproxt _ capproxm); auto.
    intros; apply plus_one; auto; done.

    intros; apply plus_one; auto; done.
Qed.


Lemma capproxt_comp_weak_ifwf :
  forall G e e' t,
    wf G
    -> tcc samek samec sames capproxt capproxm G e e' t 
    -> capproxt G e e' t.
Proof.
exact (capprox_compatible_weak andel).
Qed.


Lemma capproxm_comp_weak_ifwf :
  forall G p M M' T,
    wf G
    -> mcc samek samec sames capproxt capproxm G p M M' T
    -> capproxm G p M M' T.
Proof.
exact (capprox_compatible_weak ander).
Qed.


Theorem capprox_semicongruence :
  semicongruence capproxt capproxm.
Proof.
do2 4 split.
  exact capprox_cregular.

  exact capprox_creflexive.

  exact capprox_ctransitive.

  exact capprox_csubstitutive.

  exact capprox_compatible_weak.
Qed.


Lemma capproxt_halt :
  forall e e' t,
    capproxt nil e e' t
    -> thalt nil e
    -> thalt nil e'.
Proof.
exact (tconsistency _ _ capprox_compatible_weak capprox_consistent).
Qed.


Definition capproxm_halt := capprox_consistent.


Lemma capprox_with_wf :
  (forall G e e' t,
     (wf G -> capproxt G e e' t)
     -> capproxt G e e' t)
  /\
  (forall G p M M' T,
     (wf G -> capproxm G p M M' T)
     -> capproxm G p M M' T).
Proof.
assert (cregular
         (fun G e e' t => wf G -> capproxt G e e' t)
         (fun G p M M' T => wf G -> capproxm G p M M' T)) as Hreg.
  do2 2 split.
    intros G e e' t Hwf Hap.
    exact (capproxt_reg _#4 Hwf (Hap Hwf)).

    intros G p M M' T Hwf Hap.
    exact (capproxm_reg _#5 Hwf (Hap Hwf)).

    intros G M M' T Hwf Hap.
    exact (capproxm_fst _#4 Hwf (Hap Hwf)).
assert (semicongruence
         (fun G e e' t => wf G -> capproxt G e e' t)
         (fun G p M M' T => wf G -> capproxm G p M M' T)) as Hsc.
  do2 4 split.
    exact Hreg.

    split.
      intros G e t Hof Hwf.
      apply capproxt_refl; auto; done.

      intros G p M T Hof Hwf.
      apply capproxm_refl; auto; done.

    split.
      intros G t e1 e2 e3 H12 H23 Hwf.
      eapply capproxt_trans; eauto; done.

      intros G p T M1 M2 M3 H12 H23 Hwf.
      eapply capproxm_trans; eauto; done.

    split.
      intros G1 s G2 e e' t Hs Hwf Hap _.
      eapply (capprox_csubstitutive andel); eauto; done.

      intros G1 s G2 p M M' T H Hwf Hap _.
      eapply (capprox_csubstitutive ander); eauto; done.

    split.
      intros G e e' t Hwf Hcc _.
      apply capproxt_comp_weak_ifwf; auto; [].
      apply (tcc_mono_wf _#3 (fun G e e' t => wf G -> capproxt G e e' t) _ (fun G p M M' T => wf G -> capproxm G p M M' T)); auto using same_sregular; done.

      intros G p M M' T Hwf Hcc _.
      apply capproxm_comp_weak_ifwf; auto; [].
      apply (mcc_mono_wf _#3 (fun G e e' t => wf G -> capproxt G e e' t) _ (fun G p M M' T => wf G -> capproxm G p M M' T)); auto using same_sregular; done.
assert (consistent
         (fun G p M M' T => wf G -> capproxm G p M M' T)) as Hcons.
  intros M M' T Hap Hhalt.
  eapply capprox_consistent; eauto using wf_nil; done.
split.
  intros G e e' t Hap.
  exact (capproxt_univ _#6 Hsc Hcons Hap).

  intros G p M M' T Hap.
  exact (capproxm_univ _#7 Hsc Hcons Hap).
Qed.
      

Lemma capproxt_with_wf :
  forall G e e' t,
    (wf G -> capproxt G e e' t)
    -> capproxt G e e' t.
Proof.
exact (capprox_with_wf andel).
Qed.


Lemma capproxm_with_wf :
  forall G p M M' T,
    (wf G -> capproxm G p M M' T)
    -> capproxm G p M M' T.
Proof.
exact (capprox_with_wf ander).
Qed.


Lemma weaken1_capproxt :
  forall G A e e' t,
    capproxt G e e' t
    -> capproxt (G; A) (shift1 e) (shift1 e') (shift1 t).
Proof.
intros G A e e' t Hap.
apply capproxt_with_wf; [].
intro Hwf.
eapply subst_capproxt; eauto using weaken1, wf_tail; done.
Qed.


Lemma weaken1_capproxm :
  forall G A p M M' T,
    capproxm G p M M' T
    -> capproxm (G; A) p (shift1 M) (shift1 M') (shift1 T).
Proof.
intros G A p M M' T Hap.
apply capproxm_with_wf; [].
intro Hwf.
eapply subst_capproxm; eauto using weaken1, wf_tail; done.
Qed.


Lemma deq_capprox_wclo :
  wclorel (deq capproxt capproxm).
Proof.
intros G m m' A A' H.
#4 invert H; try (intros; subst; simp_sub; eauto using deq; done).

(* cn *)
intros k H0 H1 <-.
simp_sub.
eapply deq_cn; eauto using weaken1_ofc, weaken1_eqc; done.

(* tm *)
intros t Hm <-.
simp_sub.
apply deq_tm; [].
apply weaken1_capproxt; auto; done.

(* md *)
intros T c Hm Hfst <-.
simp_sub.
#2 apply deq_md.
  apply weaken1_capproxm; auto; done.

  eapply subst_fstm; eauto; [].
  apply subof_sh; eauto with sequence; done.

(* umd *)
intros c Hm <-.
simp_sub.
apply deq_umd; auto; [].
eapply subst_fstm; eauto; [].
apply subof_sh; eauto with sequence; done.
Qed.


Lemma capproxt_comp_weak :
  forall G e e' t,
    tcc samek samec sames capproxt capproxm G e e' t 
    -> capproxt G e e' t.
Proof.
intros.
apply capproxt_with_wf; [].
intro.
apply capproxt_comp_weak_ifwf; auto; done.
Qed.


Lemma capproxm_comp_weak :
  forall G p M M' T,
    mcc samek samec sames capproxt capproxm G p M M' T
    -> capproxm G p M M' T.
Proof.
intros.
apply capproxm_with_wf; [].
intro.
apply capproxm_comp_weak_ifwf; auto; done.
Qed.


Lemma capproxm_forget_pure :
  forall G p M M' T,
    capproxm G pure M M' T
    -> capproxm G p M M' T.
Proof.
intros G p M M' T H.
destruct p; auto; [].
apply capproxm_comp_weak; apply mcc_forget; auto.
Qed.


Lemma capproxm_forget_impure :
  forall G p M M' T,
    capproxm G p M M' T
    -> capproxm G impure M M' T.
Proof.
intros G p M M' T H.
destruct p; auto; [].
apply capproxm_comp_weak; apply mcc_forget; auto.
Qed.


Lemma capproxt_comp :
  forall G e e' t,
    tcc eqk eqc eqs capproxt capproxm G e e' t
    -> capproxt G e e' t.
Proof.
intros G e e' t Htcc.
apply capproxt_with_wf; [].
#5 induct Htcc;
try (intros; apply capproxt_comp_weak; eauto using tcc; done).

(* lam *)
intros G t1 t1' t2 e e' Heqc Hap Hwf.
so (eqc_reg _#4 Hwf Heqc) as (Hoft1 & Hoft1' & _).
so (capproxt_reg _#4 (wf_tm Hwf Hoft1) Hap) as (_ & Hofe').
so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 (wf_tm Hwf Hoft1) Hofe')) as Hoft2.
simp_sub in Hoft2.
#2 apply (capproxt_trans _ _ (tm_lam t1 e')).
  apply capproxt_comp_weak; apply tcc_lam; auto with same; done.

  #3 apply treduce_capproxt.
    apply oft_lam; auto; done.
  
    #2 apply (@oft_equiv _ (cn_arrow t1' t2)).
      apply oft_lam; auto; [].
      eapply oft_conv_0t; eauto; done.
  
      apply eqc_arrow; auto with static; done.
  
    apply treduce_lam; [].
    apply treduce_refl; done.

(* plam *)
intros G k k' t e e' Heqk Hap Hwf.
so (eqk_reg _#3 Hwf Heqk) as (Hofk & Hofk').
so (capproxt_reg _#4 (wf_cn Hwf Hofk) Hap) as (_ & Hofe').
so (oft_reg _#3 (wf_cn Hwf Hofk) Hofe') as Hoft.
#2 apply (capproxt_trans _ _ (tm_plam k e')).
  apply capproxt_comp_weak; apply tcc_plam; auto with same; done.

  #3 apply treduce_capproxt.
    apply oft_plam; auto; done.

    #2 apply (@oft_equiv _ (cn_all k' t)).
      apply oft_plam; auto; [].
      eapply oft_conv_0c; eauto; done.

      so (subst_ofc _#5 (subof_conv_0 _#3 (eqk_symm Heqk)) Hoft) as Hoft'; simp_sub in Hoft'.
      apply eqc_all; auto with static; done.

  apply treduce_plam; [].
  apply treduce_refl; done.

(* papp *)
intros G k t e e' c c' Hap Heqc Hwf.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & _).
so (capproxt_reg _#4 Hwf Hap) as (_ & Hofe').
so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hofe')) as (kk & H).
invert H; intros _ Hoft _; clear kk H.
#2 apply (capproxt_trans _ _ (tm_papp e' c)).
  apply capproxt_comp_weak; eapply tcc_papp; eauto with same; done.

  #3 apply treduce_capproxt.
    eapply oft_papp; eauto; done.

    #2 apply (@oft_equiv _ (subst1 c' t)).
      eapply oft_papp; eauto; done.

      apply eqc_symm; [].
      eapply funct_cn_type; eauto; done.

    apply treduce_papp1; [].
    apply treduce_refl; done.

(* pack *)
intros G k k' t t' c c' e e' Heqc Heqk Hap Heqt Hwf.
so (eqk_reg _#3 Hwf Heqk) as (Hofk & _).
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & _).
so (eqc_reg _#4 (wf_cn Hwf Hofk) Heqt) as (Hoft & Hoft' & _).
so (capproxt_reg _#4 Hwf Hap) as (_ & Hofe').
#2 apply (capproxt_trans _ _ (tm_pack c e' k t)).
  apply capproxt_comp_weak; eapply tcc_pack; eauto with same; done.

  #3 apply treduce_capproxt.
    apply oft_pack; auto; done.

    #2 apply (@oft_equiv _ (cn_exists k' t')).
      #3 apply oft_pack.
        eapply ofc_subsume; eauto with static; done.

        eapply oft_equiv; eauto; [].
        eapply funct_eqc_type; eauto; done.

        so (subst_ofc _#5 (subof_conv_0 _#3 (eqk_symm Heqk)) Hoft') as H; simp_sub in H.
        assumption.

      apply eqc_symm; [].
      apply eqc_exists; auto; done.

    apply treduce_pack; [].
    apply treduce_refl; done.

(* fix *)
intros G t t' e e' Heqt Hap Hwf.
so (eqc_reg _#4 Hwf Heqt) as (Hoft & Hoft' & _).
so (capproxt_reg _#4 Hwf Hap) as (_ & Hofe').
#2 apply (capproxt_trans _ _ (tm_fix t e')).
  apply capproxt_comp_weak; eapply tcc_fix; eauto with same; done.

  #3 apply treduce_capproxt.
    apply oft_fix; auto; done.

    #2 apply (@oft_equiv _ t').
      apply oft_fix; [].
      eapply oft_equiv; eauto with static; done.

      apply eqc_symm; assumption.

    apply treduce_fix1; [].
    apply treduce_refl; done.
Qed.


Lemma capproxm_comp :
  forall G p M M' T,
    mcc eqk eqc eqs capproxt capproxm G p M M' T
    -> capproxm G p M M' T.
Proof.
intros G p M M' T Hmcc.
apply capproxm_with_wf; [].
#6 induct Hmcc;
try (intros; apply capproxm_comp_weak; eauto using mcc; done).

(* stat *)
intros G c c' k Heqc Hwf.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & _).
eapply mreduce_capproxm_pure; eauto with static; [].
apply mreduce_stat; done.

(* lamgn *)
intros G T1 T1' T2 k1 M M' Heqs Hfst Hap Hwf.
so (eqs_reg _#3 Hwf Heqs) as (Hofs & Hofs').
so (capproxm_reg _#5 (wf_md Hwf Hofs Hfst) Hap) as (_ & HofM').
so (can_fsts _ _ Hofs') as (k1' & Hfst').
so (eqs_fst _#5 Hwf Heqs Hfst Hfst') as Heqk.
so (subst_ofs _#4 (unshift_md _#3) (ofm_reg _#4 (wf_md Hwf Hofs Hfst) HofM')) as HofT2; simp_sub in HofT2.
#2 apply (capproxm_trans _ _ _ (md_lamgn T1 M')).
  apply capproxm_comp_weak; eapply mcc_lamgn; eauto with same; done.

  #2 eapply mreduce_capproxm_pure; eauto with static.
    #2 apply (@ofm_subsume _ _ (sg_pigen T1' T2)).
      eapply ofm_lamgn; eauto; [].
      eapply ofm_conv_01m; eauto; [].
      apply weaken1_eqs; auto; done.

      apply subs_refl; apply eqs_symm; [].
      eapply eqs_pigen; eauto using eqs_refl; done.

    apply mreduce_lamgn; [].
    apply mreduce_refl; done.

(* lamap *)
intros G T1 T1' T2 k1 M M' Heqs Hfst Hap Hwf.
so (eqs_reg _#3 Hwf Heqs) as (Hofs & Hofs').
so (capproxm_reg _#5 (wf_md Hwf Hofs Hfst) Hap) as (_ & HofM').
so (can_fsts _ _ Hofs') as (k1' & Hfst').
so (eqs_fst _#5 Hwf Heqs Hfst Hfst') as Heqk.
so (wf_md Hwf Hofs Hfst) as Hwf'.
so (subst_ofs _#4 (unshift_md _#3) (ofm_reg _#4 Hwf' HofM')) as HofT2; simp_sub in HofT2.
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
so (can_fstm _#3 Hwf' HofM') as (cw & Hfstm).
so (subordf_md_cn' _#4 (fstm_fcn _#3 (wf_fcontext _ Hwf') (ofm_fmd _#4 (wf_fcontext _ Hwf') HofM') Hfstm)).
forget (subst1 nonsense cw) as c; subst cw.
so (subst_ofc _#5 (unshift_md _#3) (ofm_fst _#6 Hwf' HofM' Hfstm (subst_fsts _ _ sh1 Hfst2))) as Hofc; simp_sub in Hofc.
#2 apply (capproxm_trans _ _ _ (md_lamap T1 M')).
  apply capproxm_comp_weak; eapply mcc_lamap; eauto with same; done.

  #3 apply (mreduce_capproxm_pure _#4 (cn_lam k1 c) (cn_lam k1' c) (kd_pi k1 k2)); eauto with static.
    #2 apply (@ofm_subsume _ _ (sg_piapp T1' T2)).
      eapply ofm_lamap; eauto; [].
      eapply ofm_conv_01m; eauto; [].
      apply weaken1_eqs; auto; done.

      apply subs_refl; apply eqs_symm; [].
      eapply eqs_piapp; eauto using eqs_refl; done.

    so (subst_fstm _#5 (subof_aof_tof _#3 (subof_conv_01m _#5 (eqk_symm Heqk) (weaken1_eqs _#4 (eqs_symm Heqs)))) Hfstm) as Hfstm'; simp_sub in Hfstm'.
    apply fstm_lamap; auto; done.

    apply mreduce_lamap; [].
    apply mreduce_refl; done.

(* unpack *)
intros G k t T T' e e' M M' Hap1 Hap2 Heqs Hwf.
so (capproxt_reg _#4 Hwf Hap1) as (_ & Hofe').
so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hofe')) as (kk & H).
invert H; intros Hofk Hoft _; clear kk H.
so (capproxm_reg _#5 (wf_tm (wf_cn Hwf Hofk) Hoft) Hap2) as (_ & HofM').
so (eqs_reg _#3 Hwf Heqs) as (HofT & HofT').
#2 apply (capproxm_trans _ _ _ (md_unpack e' M' T)).
  apply capproxm_comp_weak; eapply mcc_unpack; eauto with same; done.

  #3 apply mreduce_capproxm_impure.
    eapply ofm_unpack; eauto; done.

    #2 apply (@ofm_subsume _ _ T').
      eapply ofm_unpack; eauto; [].
      eapply ofm_subsume; eauto; [].
      apply subs_refl; [].
      apply (subst_eqs _#5 (weaken2 _#4)); auto; done.

      apply subs_refl; apply eqs_symm; assumption.

    #2 apply mreduce_unpack1.
      apply treduce_refl; done.

      apply mreduce_refl; done.

(* letm *)
intros G T1 T2 T2' k1 M1 M1' M2 M2' Hap1 Hfst Hap2 Heqs Hwf.
so (capproxm_reg _#5 Hwf Hap1) as (_ & HofM1').
so (ofm_reg _#4 Hwf HofM1') as HofT1.
so (capproxm_reg _#5 (wf_md Hwf HofT1 Hfst) Hap2) as (_ & HofM2').
so (eqs_reg _#3 Hwf Heqs) as (HofT2 & HofT2').
#2 apply (capproxm_trans _ _ _ (md_letm M1' M2' T2)).
  apply capproxm_comp_weak; eapply mcc_letm; eauto with same; done.

  #3 apply mreduce_capproxm_impure.
    eapply ofm_letm; eauto; done.

    #2 apply (@ofm_subsume _ _ T2').
      eapply ofm_letm; eauto; [].
      eapply ofm_subsume; eauto; [].
      apply subs_refl; [].
      apply (subst_eqs _#5 (weaken2 _#4)); auto; done.

      apply subs_refl; apply eqs_symm; assumption.

    apply mreduce_letm1; apply mreduce_refl; done.

(* seal *)
intros G T T' M M' Hap Heqs Hwf.
so (capproxm_reg _#5 Hwf Hap) as (_ & HofM').
so (eqs_reg _#3 Hwf Heqs) as (HofT & HofT').
#2 apply (capproxm_trans _ _ _ (md_seal M' T)).
  apply capproxm_comp_weak; eapply mcc_seal; eauto with same; done.

  #3 apply mreduce_capproxm_impure.
    eapply ofm_seal; eauto; done.

    #2 apply (@ofm_subsume _ _ T').
      eapply ofm_seal; eauto; [].
      eapply ofm_subsume; eauto; [].
      apply subs_refl; assumption.

      apply subs_refl; apply eqs_symm; assumption.

    apply mreduce_seal1; apply mreduce_refl; done.
Qed.


Lemma capprox_compatible_strong :
  compatible eqk eqc eqs capproxt capproxm.
Proof.
split.
  intros; apply capproxt_comp; auto; done.

  intros; apply capproxm_comp; auto; done.
Qed.


(* Contextual equivalence *)

Lemma ceqt_univ :
  forall Rt Rm G e e' t,
    congruence Rt Rm
    -> consistent Rm
    -> Rt G e e' t
    -> ceqt G e e' t.
Proof.
intros Rt Rm G e e' t Hcong Hcons HR.
destruct Hcong as (Hsc & (Hsymm & _)).
split.
  eapply capproxt_univ; eauto; done.

  eapply capproxt_univ; eauto; [].
  apply Hsymm; assumption.
Qed.


Lemma ceqm_univ :
  forall Rt Rm G p M M' T,
    congruence Rt Rm
    -> consistent Rm
    -> Rm G p M M' T
    -> ceqm G p M M' T.
Proof.
intros Rt Rm G p M M' T Hcong Hcons HR.
destruct Hcong as (Hsc & (_ & Hsymm)).
split.
  eapply capproxm_univ; eauto; done.

  eapply capproxm_univ; eauto; [].
  apply Hsymm; assumption.
Qed.


Lemma ceqt_univ_quasi :
  forall Rt Rm G e e' t,
    quasicongruence Rt Rm
    -> consistent Rm
    -> consistent (fun G p M M' T => Rm G p M' M T)
    -> Rt G e e' t
    -> ceqt G e e' t.
Proof.
intros Rt Rm G e e' t Hquasi Hcon Hcon' HR.
#3 apply (ceqt_univ (tstc Rt) (mstc Rm)).
  apply stc_congruence; auto; done.

  apply mstc_consistent; auto; done.

  eapply tstc_intro; eauto; done.
Qed.


Lemma ceqm_univ_quasi :
  forall Rt Rm G p M M' T,
    quasicongruence Rt Rm
    -> consistent Rm
    -> consistent (fun G p M M' T => Rm G p M' M T)
    -> Rm G p M M' T
    -> ceqm G p M M' T.
Proof.
intros Rt Rm G p M M' T Hquasi Hcon Hcon' HR.
#3 apply (ceqm_univ (tstc Rt) (mstc Rm)).
  apply stc_congruence; auto; done.

  apply mstc_consistent; auto; done.

  eapply mstc_intro; eauto; done.
Qed.


Lemma ceqt_with_wf :
  forall G e e' t,
    (wf G -> ceqt G e e' t)
    -> ceqt G e e' t.
Proof.
intros G e e' t Heq.
split.
  apply capproxt_with_wf; [].
  intro Hwf.
  exact (Heq Hwf andel).

  apply capproxt_with_wf; [].
  intro Hwf.
  exact (Heq Hwf ander).
Qed.


Lemma ceqm_with_wf :
  forall G p M M' T,
    (wf G -> ceqm G p M M' T)
    -> ceqm G p M M' T.
Proof.
intros G p M M' T Heq.
split.
  apply capproxm_with_wf; [].
  intro Hwf.
  exact (Heq Hwf andel).

  apply capproxm_with_wf; [].
  intro Hwf.
  exact (Heq Hwf ander).
Qed.


Lemma ceq_semicongruence :
  semicongruence ceqt ceqm.
Proof.
#2 apply conjunction_semicongruence.    
  exact capprox_semicongruence.

  apply transpose_semicongruence; exact capprox_semicongruence.
Qed.


Theorem ceq_congruence :
  congruence ceqt ceqm.
Proof.
split.
  exact ceq_semicongruence.

  split.
    intros G t e e' (? & ?).
    split; assumption.

    intros G p M M' T (? & ?).
    split; assumption.
Qed.


Theorem ceq_consistent :
  consistent ceqm.
Proof.
intros M M' T Hceq Hhalt.
destruct Hceq.
eapply capprox_consistent; eauto; done.
Qed.


Lemma ceqt_reg :
  forall G e e' t,
    wf G
    -> ceqt G e e' t
    -> oft G e t /\ oft G e' t.
Proof.
exact (ceq_semicongruence andel andel).
Qed.


Lemma ceqm_reg :
  forall G p M M' T,
    wf G
    -> ceqm G p M M' T
    -> ofm G p M T /\ ofm G p M' T.
Proof.
exact (ceq_semicongruence andel ander andel).
Qed.


Lemma ceqm_fst :
  forall G M M' T,
    wf G
    -> ceqm G pure M M' T
    -> exists k c c',
         fsts T k
         /\ fstm G M c
         /\ fstm G M' c'
         /\ eqc G c c' k.
Proof.
exact (ceq_semicongruence andel ander ander).
Qed.


Lemma ceqm_fst' :
  forall G M M' T k c c',
    wf G
    -> ceqm G pure M M' T
    -> fsts T k
    -> fstm G M c
    -> fstm G M' c'
    -> eqc G c c' k.
Proof.
intros G M M' T k c c' Hwf Hceq Hfst Hfstm Hfstm'.
so (ceqm_fst _#4 Hwf Hceq) as (ka & ca & ca' & Hfsta & Hfstma & Hfstma' & Heqc).
so (fsts_fun _#3 Hfst Hfsta); subst ka.
so (fstm_fun _#4 Hfstm Hfstma); subst ca.
so (fstm_fun _#4 Hfstm' Hfstma'); subst ca'.
assumption.
Qed.


Lemma ceqt_refl :
  forall G e t, oft G e t -> ceqt G e e t.
Proof.
exact (ceq_semicongruence ander andel andel).
Qed.


Lemma ceqm_refl :
  forall G p M T, ofm G p M T -> ceqm G p M M T.
Proof.
exact (ceq_semicongruence ander andel ander).
Qed.


Lemma ceqt_symm :
  forall G e e' t, ceqt G e e' t -> ceqt G e' e t.
Proof.
intros G e e' t (? & ?).
split; assumption.
Qed.


Lemma ceqm_symm :
  forall G p M M' T, ceqm G p M M' T -> ceqm G p M' M T.
Proof.
intros G p M M' T (? & ?).
split; assumption.
Qed.


Lemma ceqt_trans :
  forall G e1 e2 e3 t,
    ceqt G e1 e2 t
    -> ceqt G e2 e3 t
    -> ceqt G e1 e3 t.
Proof.
intros.
eapply (ceq_semicongruence ander ander andel andel); eauto.
Qed.


Lemma ceqm_trans :
  forall G p M1 M2 M3 T,
    ceqm G p M1 M2 T
    -> ceqm G p M2 M3 T
    -> ceqm G p M1 M3 T.
Proof.
intros.
eapply (ceq_semicongruence ander ander andel ander); eauto.
Qed.


Lemma subst_ceqt :
  forall G1 s G2 e e' t,
    subof aof G1 s G2 
    -> wf G2
    -> ceqt G2 e e' t
    -> ceqt G1 (subst s e) (subst s e') (subst s t).
Proof.
exact (ceq_semicongruence ander ander ander andel andel).
Qed.


Lemma subst_ceqm :
  forall G1 s G2 p M M' T,
    subof aof G1 s G2 
    -> wf G2
    -> ceqm G2 p M M' T
    -> ceqm G1 p (subst s M) (subst s M') (subst s T).
Proof.
exact (ceq_semicongruence ander ander ander andel ander).
Qed.


Lemma ceq_compatible_strong :
  compatible eqk eqc eqs ceqt ceqm.
Proof.
#2 apply conjunction_compatible.
  exact capprox_compatible_strong.

  #7 apply transpose_compatible.
    do2 2 split; auto with static; done.

    exact capprox_cregular.

    do2 2 split; eauto using subst_eqk, subst_eqc, subst_eqs; done.

    exact capprox_csubstitutive.

    eauto with static; done.

    do2 2 split; eauto with static; done.

    split.
      intros G e e' t _ Htcc.
      apply capproxt_comp; [].
      refine (tcc_mono _#19 Htcc); eauto with static; done.

      intros G p M M' T _ Hmcc.
      apply capproxm_comp; [].
      refine (mcc_mono _#20 Hmcc); eauto with static; done.
Qed.


Lemma ceqt_comp :
  forall G e e' t,
    tcc eqk eqc eqs ceqt ceqm G e e' t
    -> ceqt G e e' t.
Proof.
intros.
apply ceqt_with_wf; [].
intro.
apply ceq_compatible_strong; auto; done.
Qed.


Lemma ceqm_comp :
  forall G p M M' T,
    mcc eqk eqc eqs ceqt ceqm G p M M' T
    -> ceqm G p M M' T.
Proof.
intros.
apply ceqm_with_wf; [].
intro.
apply ceq_compatible_strong; auto; done.
Qed.


Lemma ceqt_halt :
  forall e e' t,
    ceqt nil e e' t
    -> thalt nil e
    -> thalt nil e'.
Proof.
intros e e' t Hceqt Hhalt.
destruct Hceqt.
eapply capproxt_halt; eauto.
Qed.


Lemma ceqm_halt :
  forall M M' T,
    ceqm nil impure M M' T
    -> mhalt nil M
    -> mhalt nil M'.
Proof.
exact ceq_consistent.
Qed.


Lemma ceqm_halt_iff :
  forall M M' T,
    ceqm nil impure M M' T
    -> (mhalt nil M <-> mhalt nil M').
Proof.
intros M M' T Heqm.
split; intro; eapply ceqm_halt; eauto using ceqm_symm.
Qed.
  

Lemma ceqm_forget_pure :
  forall G p M M' T,
    ceqm G pure M M' T
    -> ceqm G p M M' T.
Proof.
intros G p M M' T H.
destruct p; auto; [].
apply ceqm_comp; apply mcc_forget; auto.
Qed.


Lemma ceqm_forget_impure :
  forall G p M M' T,
    ceqm G p M M' T
    -> ceqm G impure M M' T.
Proof.
intros G p M M' T H.
destruct p; auto; [].
apply ceqm_comp; apply mcc_forget; auto.
Qed.


(* Evaluation implies contextual equivalence *)

Lemma step_reduce :
  (forall G e e',
     tstep G e e'
     -> treduce G e e')
  /\
  (forall G M M',
     mstep G M M'
     -> mreduce G M M').
Proof.
exploit (step_ind treduce mreduce);
intros; eauto with reduce.
Qed.


Lemma tstep_treduce :
  forall G e e',
    tstep G e e'
    -> treduce G e e'.
Proof.
exact (step_reduce andel).
Qed.


Lemma mstep_mreduce :
  forall G M M',
    mstep G M M'
    -> mreduce G M M'.
Proof.
exact (step_reduce ander).
Qed.


Lemma tstep_capproxt :
  forall G e e' t,
    oft G e t
    -> tstep G e e'
    -> capproxt G e e' t.
Proof.
intros G e e' t Hof Hstep.
apply capproxt_with_wf; [].
intro Hwf.
apply treduce_capproxt; eauto 6 using subst_oft, subst_tstep, subof_aof_dof, subof_aof_oof, preservation_tm, tstep_treduce.
Qed.


Lemma anti_tstep_capproxt :
  forall G e e' t,
    oft G e t
    -> tstep G e e'
    -> capproxt G e' e t.
Proof.
intros G e e' t Hof Hstep.
apply capproxt_with_wf; [].
intro Hwf.
apply anti_treduce_capproxt; eauto 6 using subst_oft, subst_tstep, subof_aof_dof, subof_aof_oof, preservation_tm, tstep_treduce.
Qed.


Lemma mstep_capproxm :
  forall G p M M' T,
    ofm G p M T
    -> mstep G M M'
    -> capproxm G p M M' T.
Proof.
intros G p M M' T Hof Hstep.
apply capproxm_with_wf; [].
intros Hwf.
destruct p.
  so (can_fsts _ _ (ofm_reg _#4 Hwf Hof)) as (k & Hfst).
  so (can_fstm _#3 Hwf Hof) as (c & Hfstm).
  so (ofm_fst _#6 Hwf Hof Hfstm Hfst) as Hofc.
  so (preservation_fst _#7 Hwf Hof Hstep Hfst Hfstm) as (c' & Hfstm' & Heqc).
  eapply mreduce_capproxm_pure; eauto 6 using subst_ofm, subof_aof_dof, subof_aof_oof, preservation_md, mstep_mreduce; done.

  apply mreduce_capproxm_impure; eauto 6 using subst_ofm, subst_mstep, subof_aof_dof, subof_aof_oof, preservation_md, mstep_mreduce; done.
Qed.


Lemma msteps_capproxm :
  forall G p M M' T,
    ofm G p M T
    -> star (mstep G) M M'
    -> capproxm G p M M' T.
Proof.
intros G p M M' T Hof Hsteps.
apply capproxm_with_wf; [].
intro Hwf.
revert Hof.
induct Hsteps.

(* refl *)
intros M Hof.
apply capproxm_refl; auto; done.

(* step *)
intros M1 M2 M3 H12 _ IH Hof1.
eapply capproxm_trans; eauto using mstep_capproxm, preservation_md; done.
Qed.


Lemma anti_mstep_capproxm :
  forall G p M M' T,
    ofm G p M T
    -> mstep G M M'
    -> capproxm G p M' M T.
Proof.
intros G p M M' T Hof Hstep.
apply capproxm_with_wf; [].
intro Hwf.
destruct p.
  so (can_fsts _ _ (ofm_reg _#4 Hwf Hof)) as (k & Hfst).
  so (can_fstm _#3 Hwf Hof) as (c & Hfstm).
  so (ofm_fst _#6 Hwf Hof Hfstm Hfst) as Hofc.
  so (preservation_fst _#7 Hwf Hof Hstep Hfst Hfstm) as (c' & Hfstm' & Heqc).
  eapply anti_mreduce_capproxm_pure; eauto 6 using subst_ofm, subof_aof_dof, subof_aof_oof, preservation_md, mstep_mreduce; done.

  apply anti_mreduce_capproxm_impure; eauto 6 using subst_ofm, subst_mstep, subof_aof_dof, subof_aof_oof, preservation_md, mstep_mreduce; done.
Qed.


Lemma anti_tsteps_capproxt :
  forall G e e' t,
    oft G e t
    -> star (tstep G) e e'
    -> capproxt G e' e t.
Proof.
intros G e e' t Hof Hsteps.
apply capproxt_with_wf; [].
intros Hwf.
revert Hof.
induct Hsteps.

(* refl *)
intros e Hof.
apply capproxt_refl; auto; done.

(* step *)
intros e1 e2 e3 H12 _ IH Hof1.
eapply capproxt_trans; eauto using anti_tstep_capproxt, preservation_tm; done.
Qed.


Lemma anti_msteps_capproxm :
  forall G p M M' T,
    ofm G p M T
    -> star (mstep G) M M'
    -> capproxm G p M' M T.
Proof.
intros G p M M' T Hof Hsteps.
apply capproxm_with_wf; [].
intros Hwf.
revert Hof.
induct Hsteps.

(* refl *)
intros M Hof.
apply capproxm_refl; auto; done.

(* step *)
intros M1 M2 M3 H12 _ IH Hof1.
eapply capproxm_trans; eauto using anti_mstep_capproxm, preservation_md; done.
Qed.


Lemma anti_tsteps_capproxt_trans :
  forall G e1 e2 e3 t,
    star (tstep G) e2 e1
    -> capproxt G e2 e3 t
    -> capproxt G e1 e3 t.
Proof.
intros G e1 e2 e3 t Hsteps Hap.
apply capproxt_with_wf; [].
intro Hwf.
eapply capproxt_trans; eauto; [].
so (capproxt_reg _#4 Hwf Hap andel) as Hof.
eapply anti_tsteps_capproxt; eauto.
Qed.


Lemma anti_msteps_capproxm_trans :
  forall G p M1 M2 M3 t,
    star (mstep G) M2 M1
    -> capproxm G p M2 M3 t
    -> capproxm G p M1 M3 t.
Proof.
intros G p M1 M2 M3 t Hsteps Hap.
apply capproxm_with_wf; [].
intro Hwf.
eapply capproxm_trans; eauto; [].
so (capproxm_reg _#5 Hwf Hap andel) as Hof.
eapply anti_msteps_capproxm; eauto.
Qed.


Lemma tstep_ceqt :
  forall G e e' t,
    oft G e t
    -> tstep G e e'
    -> ceqt G e e' t.
Proof.
intros G e e' t Hof Hstep.
split.
  apply tstep_capproxt; auto; done.

  apply anti_tstep_capproxt; auto; done.
Qed.


Lemma mstep_ceqm :
  forall G p M M' T,
    ofm G p M T
    -> mstep G M M'
    -> ceqm G p M M' T.
Proof.
intros G p M M' T Hof Hstep.
split.
  apply mstep_capproxm; auto; done.

  apply anti_mstep_capproxm; auto; done.
Qed.


Lemma tsteps_ceqt :
  forall G e e' t,
    oft G e t
    -> star (tstep G) e e'
    -> ceqt G e e' t.
Proof.
intros G e e' t Hof Hsteps.
apply ceqt_with_wf; [].
intro Hwf.
revert Hof.
induct Hsteps.

(* refl *)
auto using ceqt_refl; done.

(* step *)
intros e1 e2 e3 Hstep _ IH Hof.
so (IH (preservation_tm _#4 Hwf Hof Hstep)) as H23.
eapply ceqt_trans; eauto; [].
apply tstep_ceqt; auto; done.
Qed.


Lemma msteps_ceqm :
  forall G p M M' t,
    ofm G p M t
    -> star (mstep G) M M'
    -> ceqm G p M M' t.
Proof.
intros G p M M' t Hof Hsteps.
apply ceqm_with_wf; [].
intro Hwf.
revert Hof.
induct Hsteps.

(* refl *)
auto using ceqm_refl; done.

(* step *)
intros M1 M2 M3 Hstep _ IH Hof.
so (IH (preservation_md _#5 Hwf Hof Hstep)) as H23.
eapply ceqm_trans; eauto; [].
apply mstep_ceqm; auto; done.
Qed.


Lemma ceqt_tsteps_trans :
  forall G e1 e2 e3 t,
    ceqt G e1 e2 t
    -> star (tstep G) e2 e3
    -> ceqt G e1 e3 t.
Proof.
intros G e1 e2 e3 t H12 Hsteps.
apply ceqt_with_wf; [].
intro Hwf.
eapply ceqt_trans; eauto; [].
apply tsteps_ceqt; auto; [].
exact (ceqt_reg _#4 Hwf H12 ander).
Qed.
    

Lemma anti_tsteps_ceqt_trans :
  forall G e1 e2 e3 t,
    star (tstep G) e2 e1
    -> ceqt G e2 e3 t
    -> ceqt G e1 e3 t.
Proof.
intros G e1 e2 e3 t H12 H23.
apply ceqt_symm; [].
eapply ceqt_tsteps_trans; eauto using ceqt_symm.
Qed.


Lemma ceqm_msteps_trans :
  forall G p M1 M2 M3 T,
    ceqm G p M1 M2 T
    -> star (mstep G) M2 M3
    -> ceqm G p M1 M3 T.
Proof.
intros G p M1 M2 M3 t H12 Hsteps.
apply ceqm_with_wf; [].
intro Hwf.
eapply ceqm_trans; eauto; [].
apply msteps_ceqm; auto; [].
exact (ceqm_reg _#5 Hwf H12 ander).
Qed.


Lemma anti_msteps_ceqm_trans :
  forall G p M1 M2 M3 T,
    star (mstep G) M2 M1
    -> ceqm G p M2 M3 T
    -> ceqm G p M1 M3 T.
Proof.
intros G p M1 M2 M3 T H12 H23.
apply ceqm_symm; [].
eapply ceqm_msteps_trans; eauto using ceqm_symm.
Qed.


(* Functionality *)

Lemma self_respecting_with_ceqm :
  forall Rt (R : mreln),
    cregular Rt R
    -> csubstitutive Rt R
    -> weakly_compatible samek samec sames Rt R
    -> (forall G p M1 M2 M3 T,
          ceqm G p M1 M2 T
          -> R G p M2 M3 T
          -> R G p M1 M3 T)
    -> (forall G p M1 M2 M3 T,
          R G p M1 M2 T
          -> ceqm G p M2 M3 T
          -> R G p M1 M3 T)
    -> self_respecting R.
Proof.
intros Rt R (_ & Hreg & Hfstreg) (_ & Hsubst) ( _ & Hcomp) Htransl Htransr G M M' T c Hwf HofT Hap HfstM.
so (Hreg _#5 Hwf Hap) as (HofM & HofM').
so (Hfstreg _#4 Hwf Hap) as (k & d & c' & Hfst & Hfstd & HfstM' & Heqc).
so (fstm_fun _#4 HfstM Hfstd); subst d; clear Hfstd.
so (ofs_hsg _ _ HofT) as Hhsg.
so (ofs_fst _#3 HofT Hfst) as Hofk.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & _).
set (magic := (md_lamap (sg_piapp sg_one (shift1 T)) (md_appap (var 0) md_star))).
assert (ofm G pure magic
          (sg_piapp (sg_piapp sg_one (shift1 T)) (singles (cn_app (var 0) cn_star) (shift1 T)))) as Hofmagic.
  eapply ofm_lamap; eauto using weaken1_ofs, subst_fsts with static; [].
  unfold shift1; rewrite -> subst_singles; auto using subst_hsg; simp_sub; [].
  #3 apply singles_ofm.
    repl (sg_piapp sg_one (subst (sh 2) T)) with (shift1 (sg_piapp sg_one (shift1 T))) by (simp_sub; reflexivity).
    apply wf_md; eauto using weaken1_ofs, subst_fsts with static; done.

    so (@ofm_var (G; cl_cn (kd_pi kd_one (shift1 k)); cl_md (sg_piapp sg_one (subst (sh 2) T)) (var 0)) _#3 index_0) as H1.
    simp_sub in H1.
    so (ofm_appap H1 ofm_star fstm_star) as H2.
    simp_sub in H2.
    assumption.

    so (@fstm_var (G; cl_cn (kd_pi kd_one (shift1 k)); cl_md (sg_piapp sg_one (subst (sh 2) T)) (var 0)) _#3 index_0) as H1.
    simp_sub in H1.
    so (fstm_appap H1 fstm_star) as H2.
    assumption.
#2 apply (Htransl _#3 (md_appap magic (md_lamap sg_one (subst (sh 2) M)))).
  apply ceqm_symm; [].
  #2 apply msteps_ceqm.
    so (subst_ofm _#6 (weaken2 _ _ (cl_cn kd_one) (cl_md sg_one (var 0))) HofM) as HofMx.
    so (subst_fstm _#5 (weaken2 _ _ (cl_cn kd_one) (cl_md sg_one (var 0))) HfstM) as HfstMx.
    replsub (subst (sh 2) T) with (shift1 (shift1 T)) in HofMx.
    replsub (subst (sh 2) c) with (shift1 (shift1 c)) in HfstMx.
    so (ofm_appap Hofmagic (ofm_lamap ofs_one fsts_one HofMx) (fstm_lamap fsts_one HfstMx)) as H1.
    unfold subst1 in H1; rewrite -> subst_singles in H1; auto using shift1_hsg; [].
    simp_sub in H1.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply singles_funct; eauto; [].
    exploit (@eqc_beta G kd_one (shift1 k) cn_star (shift1 c)) as H2; eauto using weaken1_ofc with static; [].
    simp_sub in H2.
    assumption.

    #2 eapply star_step.
      eapply mstep_appap3; eauto with dynamic; [].
      so (subst_fstm _#5 (weaken2 _ _ (cl_cn kd_one) (cl_md (shift1 sg_one) (var 0))) HfstM) as HfstMx.
      replsub (subst (sh 2) c) with (shift1 (shift1 c)) in HfstMx.
      eapply fstm_lamap; eauto with static; done.

      simp_sub.
      #2 eapply star_step.
        eapply mstep_appap3; eauto with static dynamic; done.

        simp_sub.
        apply star_refl; done.
  #2 apply (Htransr _#3 (md_appap magic (md_lamap sg_one (subst (sh 2) M')))).
    apply Hcomp; auto; [].
    #2 apply (mccw_subsume _#7 (subst1 (cn_lam kd_one (shift1 c)) (singles (cn_app (var 0) cn_star) (shift1 T)))).
      apply Hcomp; auto; [].
      #2 eapply mccw_appap2; eauto.
        apply Hcomp; auto; [].
        eapply mccw_lamap; eauto using samesrefl with static; [].
        simp_sub.
        eapply Hsubst; eauto using weaken2; done.

        apply fstm_lamap; eauto with static; [].
        simp_sub.
        eapply subst_fstm; eauto using weaken2; done.

      unfold subst1; rewrite -> subst_singles; auto using shift1_hsg; [].
      simp_sub.
      apply subs_refl; [].
      eapply singles_funct; eauto; [].
      exploit (@eqc_beta G kd_one (shift1 k) cn_star (shift1 c)) as H2; eauto using weaken1_ofc with static; [].
      simp_sub in H2.
      assumption.
  
    #2 apply msteps_ceqm.
      apply (@ofm_subsume _ _ (singles c' T)).
        so (subst_ofm _#6 (weaken2 _ _ (cl_cn kd_one) (cl_md sg_one (var 0))) HofM') as HofMx'.
        so (subst_fstm _#5 (weaken2 _ _ (cl_cn kd_one) (cl_md sg_one (var 0))) HfstM') as HfstMx'.
        replsub (subst (sh 2) T) with (shift1 (shift1 T)) in HofMx'.
        replsub (subst (sh 2) c') with (shift1 (shift1 c')) in HfstMx'.
        so (ofm_appap Hofmagic (ofm_lamap ofs_one fsts_one HofMx') (fstm_lamap fsts_one HfstMx')) as H1.
        unfold subst1 in H1; rewrite -> subst_singles in H1; auto using shift1_hsg; [].
        simp_sub in H1.
        eapply ofm_subsume; eauto; [].
        apply subs_refl; [].
        eapply singles_funct; eauto; [].
        exploit (@eqc_beta G kd_one (shift1 k) cn_star (shift1 c')) as H2; eauto using weaken1_ofc with static; [].
        simp_sub in H2.
        assumption.

        apply subs_refl; [].
        eapply singles_funct; eauto with static; done.
        
      #2 eapply star_step.
        eapply mstep_appap3; eauto with dynamic; [].
        so (subst_fstm _#5 (weaken2 _ _ (cl_cn kd_one) (cl_md (shift1 sg_one) (var 0))) HfstM') as HfstMx'.
        replsub (subst (sh 2) c') with (shift1 (shift1 c')) in HfstMx'.
        eapply fstm_lamap; eauto with static; done.
  
        simp_sub.
        #2 eapply star_step.
          eapply mstep_appap3; eauto with static dynamic; done.
  
          simp_sub.
          apply star_refl; done.
Qed.


Lemma capproxm_self_respecting :
  self_respecting capproxm.
Proof.
#2 eapply self_respecting_with_ceqm; eauto using capprox_cregular, capprox_csubstitutive, capprox_weakly_compatible.
  intros G p M1 M2 M3 T (H12 & _) H23.
  eapply capproxm_trans; eauto; done.

  intros G p M1 M2 M3 T H12 (H23 & _).
  eapply capproxm_trans; eauto; done.
Qed.


Lemma ceqm_self_respecting :
  self_respecting ceqm.
Proof.
so ceq_semicongruence as (Hreg & Hrefl & _ & Hsubst & Hcomp).
eapply self_respecting_with_ceqm; eauto using ceqm_trans, weaken_compatible; done.
Qed.


Lemma funct_capprox_tm :
  forall G1 G2 s s' e t,
    wf G1
    -> wf G2
    -> subrel (deq capproxt capproxm) G1 s s' G2
    -> oft G2 e t
    -> capproxt G1 (subst s e) (subst s' e) (subst s t).
Proof.
#4 eapply subst_deq_tm; eauto.
  exact capprox_creflexive.

  exact capprox_csubstitutive.

  exact capprox_compatible_strong.

  exact capproxm_self_respecting.
Qed.


Lemma funct_capprox_md :
  forall G1 G2 p s s' M T,
    wf G1
    -> wf G2
    -> subrel (deq capproxt capproxm) G1 s s' G2
    -> ofm G2 p M T
    -> capproxm G1 p (subst s M) (subst s' M) (subst s T).
Proof.
#4 eapply subst_deq_md; eauto.
  exact capprox_creflexive.

  exact capprox_csubstitutive.

  exact capprox_compatible_strong.

  exact capproxm_self_respecting.
Qed.


Lemma funct_ceq_tm :
  forall G1 G2 s s' e t,
    wf G1
    -> wf G2
    -> subrel (deq ceqt ceqm) G1 s s' G2
    -> oft G2 e t
    -> ceqt G1 (subst s e) (subst s' e) (subst s t).
Proof.
#4 eapply subst_deq_tm; eauto.
  exact (ceq_semicongruence ander andel).

  exact (ceq_semicongruence ander ander ander andel).

  exact ceq_compatible_strong.

  exact ceqm_self_respecting.
Qed.


Lemma funct_ceq_md :
  forall G1 G2 p s s' M T,
    wf G1
    -> wf G2
    -> subrel (deq ceqt ceqm) G1 s s' G2
    -> ofm G2 p M T
    -> ceqm G1 p (subst s M) (subst s' M) (subst s T).
Proof.
#4 eapply subst_deq_md; eauto.
  exact (ceq_semicongruence ander andel).

  exact (ceq_semicongruence ander ander ander andel).

  exact ceq_compatible_strong.

  exact ceqm_self_respecting.
Qed.


Lemma funct_ceq_tm_tm :
  forall G t1 t2 e1 e1' e2,
    wf G
    -> ceqt G e1 e1' t1
    -> oft (G; cl_tm t1) e2 (shift1 t2)
    -> ceqt G (subst1 e1 e2) (subst1 e1' e2) t2.
Proof.
intros G t1 t2 e1 e1' e2 Hwf Hceq Hofe2.
replsub t2 with (subst1 e1 (shift1 t2)).
#2 apply (funct_ceq_tm _ (G; cl_tm t1)); auto.
  apply wf_tm; auto; [].
  exact (oft_reg _#3 Hwf (ceqt_reg _#4 Hwf Hceq andel)).

  apply subrel_dot; auto using subrel_id; [].
  simp_sub; apply deq_tm; auto; done.
Qed.


Lemma funct_ceq_md_md :
  forall G p T1 T2 k1 M1 M1' M2 c c',
    wf G
    -> ceqm G pure M1 M1' T1
    -> fsts T1 k1
    -> fstm G M1 c
    -> fstm G M1' c'
    -> ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) p M2 (shift1 T2)
    -> ceqm G p (subst (dot M1 (dot c id)) M2) (subst (dot M1' (dot c' id)) M2) (subst1 c T2).
Proof.
intros G p T1 T2 k1 M1 M1' M2 c c' Hwf Hceq Hfst HfstM1 HfstM1' Hofe2.
replsub (subst1 c T2) with (subst (dot M1 (dot c id)) (shift1 T2)).
so (ceqm_fst' _#7 Hwf Hceq Hfst HfstM1 HfstM1') as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & _).
#2 apply (funct_ceq_md _ (G; cl_cn k1; cl_md (shift1 T1) (var 0))); auto.
  apply wf_md; auto; [].
  exact (ofm_reg _#4 Hwf (ceqm_reg _#5 Hwf Hceq andel)).

  #2 apply subrel_dot.
    apply subrel_dot; auto using subrel_id; [].
    simp_sub; apply deq_cn; auto; done.

    simp_sub; apply deq_md; auto; done.
Qed.
