
Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Formation.
Require Import Declarative.
Require Import Determinism.
Require Import Congruence.
Require Import Contextual.
Require Import Progress.


Definition diverge (G : context) (e : (* tm *) term) :=
  forall e',
    star (tstep G) e e'
    -> exists e'',
         tstep G e' e''.


Lemma diverge_step :
  forall G e e',
    tstep G e e'
    -> diverge G e
    -> diverge G e'.
Proof.
intros G e1 e2 Hstep Hdiverge.
intros e3 Hsteps.
exact (Hdiverge _ (star_step _ Hstep Hsteps)).
Qed.


Lemma diverge_value_contra :
  forall G e,
    diverge G e
    -> tvalue G e
    -> False.
Proof.
intros G e Hdiverge Hval.
so (Hdiverge _ (star_refl _)) as (e' & Hstep).
eapply tstep_value_contra; eauto; done.
Qed.    


Lemma diverge_halt_contra :
  forall G e,
    diverge G e
    -> thalt G e
    -> False.
Proof.
intros G e Hdiverge Hhalt.
destruct Hhalt as (v & Hsteps & Hval).
so (Hdiverge _ Hsteps) as (e' & Hstep).
eapply tstep_value_contra; eauto; done.
Qed.    


Lemma subst_diverge :
  forall G1 s G2 e,
    subof oof G1 s G2
    -> diverge G2 e
    -> diverge G1 (subst s e).
Proof.
intros G1 s G2 e Hs Hdiv.
intros f Hsteps.
assert (exists e', f = subst s e' /\ star (tstep G2) e e') as (e' & -> & Hsteps').
  remember (subst s e) as m eqn:Heqm.
  revert e Heqm Hdiv.
  #2 induct Hsteps.
    (* refl *)
    intros f e -> Hdiv.
    eauto using star_refl; done.

    (* step *)
    intros f1 f2 f3 Hstep _ IH e1 -> Hdiv.
    so (Hdiv _ (star_refl _)) as (e2 & Hstep').
    so (subst_tstep _#5 Hs Hstep') as Hstep''.
    so (tstep_fun _#4 Hstep Hstep''); subst f2.
    so (IH e2 (eq_refl _) (diverge_step _#3 Hstep' Hdiv)) as (e3 & -> & Hsteps).
    eauto using star_step; done.
so (Hdiv _ Hsteps') as (e'' & Hstep).
so (subst_tstep _#5 Hs Hstep) as Hstep'.
eauto.
Qed.



Inductive tcrip : context -> (* tm *) term -> (* tm *) term -> Prop :=
| tcrip_bot {G e e'}
    : diverge G e
      -> tcrip G e e'

| tcrip_id {G e}
    : tcrip G e e

| tcrip_lam {G t e e'}
    : tcrip (G; cl_utm) e e'
      -> tcrip G (tm_lam t e) (tm_lam t e')

| tcrip_app {G e1 e1' e2 e2'}
    : tcrip G e1 e1'
      -> tcrip G e2 e2'
      -> tcrip G (tm_app e1 e2) (tm_app e1' e2')

| tcrip_pair {G e1 e1' e2 e2'}
    : tcrip G e1 e1'
      -> tcrip G e2 e2'
      -> tcrip G (tm_pair e1 e2) (tm_pair e1' e2')

| tcrip_pi1 {G e e'}
    : tcrip G e e'
      -> tcrip G (tm_pi1 e) (tm_pi1 e')

| tcrip_pi2 {G e e'}
    : tcrip G e e'
      -> tcrip G (tm_pi2 e) (tm_pi2 e')

| tcrip_plam {G k e e'}
    : tcrip (G; cl_ucn) e e'
      -> tcrip G (tm_plam k e) (tm_plam k e')

| tcrip_papp {G e e' c}
    : tcrip G e e'
      -> tcrip G (tm_papp e c) (tm_papp e' c)

| tcrip_pack {G c e e' k t}
    : tcrip G e e'
      -> tcrip G (tm_pack c e k t) (tm_pack c e' k t)

| tcrip_unpack {G e1 e1' e2 e2'}
    : tcrip G e1 e1'
      -> tcrip (G; cl_ucn; cl_utm) e2 e2'
      -> tcrip G (tm_unpack e1 e2) (tm_unpack e1' e2')

| tcrip_fix {G t e e'}
    : tcrip G e e'
      -> tcrip G (tm_fix t e) (tm_fix t e')

| tcrip_lett {G e1 e1' e2 e2'}
    : tcrip G e1 e1'
      -> tcrip (G; cl_utm) e2 e2'
      -> tcrip G (tm_lett e1 e2) (tm_lett e1' e2')

| tcrip_letm {G M M' e e'}
    : mcrip G M M'
      -> tcrip (G; cl_ucn; cl_umd (var 0)) e e'
      -> tcrip G (tm_letm M e) (tm_letm M' e')

| tcrip_ext {G M M'}
    : mcrip G M M'
      -> tcrip G (tm_ext M) (tm_ext M')


with mcrip : context -> (* md *) term -> (* md *) term -> Prop :=
| mcrip_id {G M}
    : mcrip G M M

| mcrip_dyn {G e e'}
    : tcrip G e e'
      -> mcrip G (md_dyn e) (md_dyn e')

| mcrip_lamgn {G T M M'}
    : mcrip (G; cl_ucn; cl_umd (var 0)) M M'
      -> mcrip G (md_lamgn T M) (md_lamgn T M')

| mcrip_appgn {G M1 M1' M2 M2'}
    : mcrip G M1 M1'
      -> mcrip G M2 M2'
      -> mcrip G (md_appgn M1 M2) (md_appgn M1' M2')

| mcrip_lamap {G T M M'}
    : mcrip (G; cl_ucn; cl_umd (var 0)) M M'
      -> mcrip G (md_lamap T M) (md_lamap T M')

| mcrip_appap {G M1 M1' M2 M2'}
    : mcrip G M1 M1'
      -> mcrip G M2 M2'
      -> mcrip G (md_appap M1 M2) (md_appap M1' M2')

| mcrip_pair {G M1 M1' M2 M2'}
    : mcrip G M1 M1'
      -> mcrip G M2 M2'
      -> mcrip G (md_pair M1 M2) (md_pair M1' M2')

| mcrip_pi1 {G M M'}
    : mcrip G M M'
      -> mcrip G (md_pi1 M) (md_pi1 M')

| mcrip_pi2 {G M M'}
    : mcrip G M M'
      -> mcrip G (md_pi2 M) (md_pi2 M')

| mcrip_unpack {G e e' M M' T}
    : tcrip G e e'
      -> mcrip (G; cl_ucn; cl_utm) M M'
      -> mcrip G (md_unpack e M T) (md_unpack e' M' T)

| mcrip_lett {G e1 e1' M2 M2'}
    : tcrip G e1 e1'
      -> mcrip (G; cl_utm) M2 M2'
      -> mcrip G (md_lett e1 M2) (md_lett e1' M2')

| mcrip_letm {G M1 M1' M2 M2' T}
    : mcrip G M1 M1'
      -> mcrip (G; cl_ucn; cl_umd (var 0)) M2 M2'
      -> mcrip G (md_letm M1 M2 T) (md_letm M1' M2' T)

| mcrip_seal {G M M' T}
    : mcrip G M M'
      -> mcrip G (md_seal M T) (md_seal M' T)
.


Scheme tcrip_mut_ind := Minimality for tcrip Sort Prop
with   mcrip_mut_ind := Minimality for mcrip Sort Prop.
Combined Scheme cripple_ind from tcrip_mut_ind, mcrip_mut_ind.


Lemma subst_tcrip_mcrip :
  (forall G1 s G2 e e',
     subof oof G1 s G2
     -> tcrip G2 e e'
     -> tcrip G1 (subst s e) (subst s e'))
  /\
  (forall G1 s G2 M M',
     subof oof G1 s G2
     -> mcrip G2 M M'
     -> mcrip G1 (subst s M) (subst s M')).
Proof.
#12 exploit
  (cripple_ind
     (fun G2 e e' =>
        forall G1 s,
          subof oof G1 s G2
          -> tcrip G1 (subst s e) (subst s e'))
     (fun G2 M M' =>
        forall G1 s,
          subof oof G1 s G2
          -> mcrip G1 (subst s M) (subst s M'))) as Hind;
try (intros; simp_sub; eauto using subof, oof, tcrip, mcrip; done).

(* bot *)
intros G e e' Hdiverge G1 s Hs.
eapply tcrip_bot; eauto using subst_diverge; done.

(* lam *)
intros G t e e' _ IH G1 s Hs.
simp_sub.
so (subof_wclo_bind _ oof_behaved oof_wclo _#4 iscl_utm Hs).
apply tcrip_lam; auto; done.

(* plam *)
intros G k e e' _ IH G1 s Hs.
simp_sub.
so (subof_wclo_bind _ oof_behaved oof_wclo _#4 iscl_ucn Hs).
apply tcrip_plam; auto; done.

(* unpack *)
intros G e1 e1' e2 e2' _ IH1 _ IH2 G1 s Hs.
simp_sub.
so (subof_wclo_bind2 _ oof_behaved oof_wclo _#5 iscl_ucn iscl_utm Hs) as Hs'.
simp_sub in Hs'.
apply tcrip_unpack; auto; done.

(* lett *)
intros G e1 e1' e2 e2' _ IH1 _ IH2 G1 s Hs.
simp_sub.
so (subof_wclo_bind _ oof_behaved oof_wclo _#4 iscl_utm Hs).
apply tcrip_lett; auto; done.

(* letm *)
intros G M M' e e' _ IH1 _ IH2 G1 s Hs.
simp_sub.
so (subof_wclo_bind2 _ oof_behaved oof_wclo _#4 (cl_umd (var 0)) iscl_ucn iscl_umd Hs) as Hs'.
simp_sub in Hs'.
apply tcrip_letm; auto; done.

(* lamgn *)
intros G T M M' _ IH G1 s Hs.
simp_sub.
so (subof_wclo_bind2 _ oof_behaved oof_wclo _#4 (cl_umd (var 0)) iscl_ucn iscl_umd Hs) as Hs'.
simp_sub in Hs'.
apply mcrip_lamgn; auto; done.

(* lamap *)
intros G T M M' _ IH G1 s Hs.
simp_sub.
so (subof_wclo_bind2 _ oof_behaved oof_wclo _#4 (cl_umd (var 0)) iscl_ucn iscl_umd Hs) as Hs'.
simp_sub in Hs'.
apply mcrip_lamap; auto; done.

(* unpack *)
intros G e e' M M' T _ IH1 _ IH2 G1 s Hs.
simp_sub.
so (subof_wclo_bind2 _ oof_behaved oof_wclo _#5 iscl_ucn iscl_utm Hs) as Hs'.
simp_sub in Hs'.
apply mcrip_unpack; auto; done.

(* lett *)
intros G e1 e1' M2 M2' _ IH1 _ IH2 G1 s Hs.
simp_sub.
so (subof_wclo_bind _ oof_behaved oof_wclo _#4 iscl_utm Hs).
apply mcrip_lett; auto; done.

(* letm *)
intros G M1 M1' M2 M2' T _ IH1 _ IH2 G1 s Hs.
simp_sub.
so (subof_wclo_bind2 _ oof_behaved oof_wclo _#4 (cl_umd (var 0)) iscl_ucn iscl_umd Hs) as Hs'.
simp_sub in Hs'.
apply mcrip_letm; auto; done.

(* epilogue *)
destruct Hind; split; intros; eauto; done.
Qed.


Lemma subst_tcrip :
  forall G1 s G2 e e',
    subof oof G1 s G2
    -> tcrip G2 e e'
    -> tcrip G1 (subst s e) (subst s e').
Proof.
exact (subst_tcrip_mcrip andel).
Qed.


Lemma subst_mcrip :
  forall G1 s G2 M M',
    subof oof G1 s G2
    -> mcrip G2 M M'
    -> mcrip G1 (subst s M) (subst s M').
Proof.
exact (subst_tcrip_mcrip ander).
Qed.


Inductive crq : context -> term -> term -> (* cl *) term -> Prop :=
| crq_cn {G c k}
    : crq G c c (cl_cn k)

| crq_tm {G e e' t}
    : tcrip G e e'
      -> tvalue G e
      -> crq G e e' (cl_tm t)

| crq_md {G M M' T c}
    : mcrip G M M'
      -> mvalue G M
      -> fstm G M c
      -> crq G M M' (cl_md T c)

| crq_ucn {G c}
    : crq G c c cl_ucn

| crq_utm {G e e'}
    : tcrip G e e' 
      -> tvalue G e
      -> crq G e e' cl_utm

| crq_umd {G M M' c}
    : mcrip G M M'
      -> mvalue G M
      -> fstm G M c
      -> crq G M M' (cl_umd c)
.


Lemma crq_behaved :
  behavedrel crq.
Proof.
intros G i C HC Hindex.
destruct HC; simp_sub; eauto using crq, tcrip, mcrip with static dynamic.
Qed.


Lemma crq_wclo :
  wclorel crq.
Proof.
intros G m m' C C' H.
#4 invert H; clear H;
try (intros; subst; simp_sub; eauto using crq; done).

(* tm *)
intros; subst; simp_sub; unfold shift1; eauto 8 using crq, subst_tcrip, subst_tvalue, subof_sh with sequence; done.

(* md *)
intros; subst; simp_sub.
apply crq_md; unfold shift1; eauto 8 using subst_mcrip, subst_mvalue, subst_fstm, subof_sh with sequence; done.

(* utm *)
intros; subst; simp_sub; unfold shift1; eauto 8 using crq, subst_tcrip, subst_tvalue, subof_sh with sequence; done.

(* umd *)
intros; subst; simp_sub.
apply crq_umd; unfold shift1; eauto 8 using subst_mcrip, subst_mvalue, subst_fstm, subof_sh with sequence; done.
Qed.


Lemma crq_oof :
  forall G m m' C,
    crq G m m' C
    -> oof G m C.
Proof.
intros G m m' C H.
invert H; intros; subst; eauto using oof.
Qed.


Lemma crq_sid :
  forall G m m' C,
    crq G m m' C
    -> sid G m m' C.
Proof.
intros G m m' C H.
invert H; intros; subst; eauto using sid.
Qed.


Lemma subrel_crq_sid :
  forall G1 s s' G2,
    subrel crq G1 s s' G2
    -> subrel sid G1 s s' G2.
Proof.
eauto using subrel_mono, crq_sid.
Qed.


Lemma crip_prefunct :
  (forall G1 s s' G2 e,
     ftm G2 e
     -> subrel crq G1 s s' G2
     -> tcrip G1 (subst s e) (subst s' e))
  /\
  (forall G1 s s' G2 M,
     fmd G2 M
     -> subrel crq G1 s s' G2
     -> mcrip G1 (subst s M) (subst s' M)).
Proof.
exploit
  (dynamic_form_ind
     (fun G2 e =>
        forall G1 s s',
          subrel crq G1 s s' G2
          -> tcrip G1 (subst s e) (subst s' e))
     (fun G2 M =>
        forall G1 s s',
          subrel crq G1 s s' G2
          -> mcrip G1 (subst s M) (subst s' M))) as Hind;
try (intros; simp_sub; eauto using tcrip, mcrip; done).

(* var *)
intros G2 i t Hindex G1 s s' Hss.
simp_sub.
so (subrel_index _ crq_behaved _#6 iscl_tm Hss Hindex) as H.
simp_sub in H.
invert H; auto; done.

(* varu *)
intros G2 i Hindex G1 s s' Hss.
simp_sub.
so (subrel_index _ crq_behaved _#6 iscl_utm Hss Hindex) as H.
simp_sub in H.
invert H; auto; done.

(* lam *)
intros G t e Ht _ IH G1 s s' Hss.
simp_sub.
so (sid_fcn _#5 (subrel_crq_sid _#4 Hss) Ht) as Heq.
rewrite <- Heq; [].
so (subrel_wclo_bind _ crq_behaved crq_wclo _#5 iscl_utm Hss) as Hss'.
apply tcrip_lam; auto; done.

(* plam *)
intros G k e Hk _ IH G1 s s' Hss.
simp_sub.
so (sid_fkd _#5 (subrel_crq_sid _#4 Hss) Hk) as Heq.
rewrite <- Heq; [].
so (subrel_wclo_bind _ crq_behaved crq_wclo _#5 iscl_ucn Hss) as Hss'.
apply tcrip_plam; auto; done.

(* papp *)
intros G e c _ IH Hc G1 s s' Hss.
simp_sub.
so (sid_fcn _#5 (subrel_crq_sid _#4 Hss) Hc) as Heq.
rewrite <- Heq; [].
apply tcrip_papp; auto; done.

(* pack *)
intros G k t c e Hc _ IH Hk Ht G1 s s' Hss.
simp_sub.
so (sid_fcn _#5 (subrel_crq_sid _#4 Hss) Hc) as Heqc.
so (sid_fkd _#5 (subrel_crq_sid _#4 Hss) Hk) as Heqk.
so (subrel_wclo_bind _ crq_behaved crq_wclo _#5 iscl_ucn Hss) as Hss'.
simp_sub in Hss'.
so (sid_fcn _#5 (subrel_crq_sid _#4 Hss') Ht) as Heqt.
rewrite <- Heqc, <- Heqk, <- Heqt; [].
apply tcrip_pack; auto; done.

(* unpack *)
intros G e1 e2 _ IH1 _ IH2 G1 s s' Hss.
simp_sub.
so (subrel_wclo_bind2 _ crq_behaved crq_wclo _#6 iscl_ucn iscl_utm Hss).
apply tcrip_unpack; eauto; done.

(* fix *)
intros G t e Ht _ IH G1 s s' Hss.
simp_sub.
so (sid_fcn _#5 (subrel_crq_sid _#4 Hss) Ht) as Heq.
rewrite <- Heq; [].
apply tcrip_fix; auto; done.

(* lett *)
intros G e1 e2 _ IH1 _ IH2 G1 s s' Hss.
simp_sub.
so (subrel_wclo_bind _ crq_behaved crq_wclo _#5 iscl_utm Hss).
apply tcrip_lett; auto; done.

(* letm *)
intros G M e _ IH1 _ IH2 G1 s s' Hss.
simp_sub.
so (subrel_wclo_bind2 _ crq_behaved crq_wclo _#5 (cl_umd (var 0)) iscl_ucn iscl_umd Hss) as Hss'.
simp_sub in Hss'.
apply tcrip_letm; eauto; done.

(* var *)
intros G2 i t c Hindex G1 s s' Hss.
simp_sub.
so (subrel_index _ crq_behaved _#6 iscl_md Hss Hindex) as H.
simp_sub in H.
invert H; auto; done.

(* varu *)
intros G2 i c Hindex G1 s s' Hss.
simp_sub.
so (subrel_index _ crq_behaved _#6 iscl_umd Hss Hindex) as H.
simp_sub in H.
invert H; auto; done.

(* stat *)
intros G c Hc G1 s s' Hss.
simp_sub.
so (sid_fcn _#5 (subrel_crq_sid _#4 Hss) Hc) as Heq.
rewrite <- Heq; [].
apply mcrip_id; done.

(* lamgn *)
intros G T M HT _ IH G1 s s' Hss.
simp_sub.
so (sid_fsg _#5 (subrel_crq_sid _#4 Hss) HT) as Heq.
rewrite <- Heq; [].
so (subrel_wclo_bind2 _ crq_behaved crq_wclo _#5 (cl_umd (var 0)) iscl_ucn iscl_umd Hss) as Hss'.
simp_sub in Hss'.
apply mcrip_lamgn; auto; done.

(* lamap *)
intros G T M HT _ IH G1 s s' Hss.
simp_sub.
so (sid_fsg _#5 (subrel_crq_sid _#4 Hss) HT) as Heq.
rewrite <- Heq; [].
so (subrel_wclo_bind2 _ crq_behaved crq_wclo _#5 (cl_umd (var 0)) iscl_ucn iscl_umd Hss) as Hss'.
simp_sub in Hss'.
apply mcrip_lamap; auto; done.

(* unpack *)
intros G e M T _ IH1 _ IH2 HT G1 s s' Hss.
simp_sub.
so (sid_fsg _#5 (subrel_crq_sid _#4 Hss) HT) as Heq.
rewrite <- Heq; [].
so (subrel_wclo_bind2 _ crq_behaved crq_wclo _#6 iscl_ucn iscl_utm Hss).
apply mcrip_unpack; eauto; done.

(* lett *)
intros G e M _ IH1 _ IH2 G1 s s' Hss.
simp_sub.
so (subrel_wclo_bind _ crq_behaved crq_wclo _#5 iscl_utm Hss).
apply mcrip_lett; auto; done.

(* letm *)
intros G M1 M2 T _ IH1 _ IH2 HT G1 s s' Hss.
simp_sub.
so (sid_fsg _#5 (subrel_crq_sid _#4 Hss) HT) as Heq.
rewrite <- Heq; [].
so (subrel_wclo_bind2 _ crq_behaved crq_wclo _#5 (cl_umd (var 0)) iscl_ucn iscl_umd Hss) as Hss'.
simp_sub in Hss'.
apply mcrip_letm; eauto; done.

(* seal *)
intros G T M _ IH HT G1 s s' Hss.
simp_sub.
so (sid_fsg _#5 (subrel_crq_sid _#4 Hss) HT) as Heq.
rewrite <- Heq; [].
apply mcrip_seal; auto; done.

(* epilogue *)
destruct Hind; split; intros; eauto; done.
Qed.


Lemma crip_funct :
  (forall G1 s s' G2 e e',
     ftm G2 e
     -> subrel crq G1 s s' G2
     -> tcrip G2 e e'
     -> tcrip G1 (subst s e) (subst s' e'))
  /\
  (forall G1 s s' G2 M M',
     fmd G2 M
     -> subrel crq G1 s s' G2
     -> mcrip G2 M M'
     -> mcrip G1 (subst s M) (subst s' M'))
.     
Proof.
#18 exploit
  (cripple_ind
     (fun G2 e e' =>
        forall G1 s s',
          ftm G2 e
          -> subrel crq G1 s s' G2
          -> tcrip G1 (subst s e) (subst s' e'))
     (fun G2 M M' =>
        forall G1 s s',
          fmd G2 M
          -> subrel crq G1 s s' G2
          -> mcrip G1 (subst s M) (subst s' M'))) as Hind;
try (intros;
     match goal with
     | H : fmd _ _ |- _ =>
       invert H; clear H; intros

     | H : ftm _ _ |- _ =>
       invert H; clear H; intros
     end;
     simp_sub;
     eauto using subrel, crq, tcrip, mcrip; done).

(* bot *)
intros G2 e e' Hdiverge G1 s s' _ Hss.
so (subrel_subof_mono _ _ crq_oof _#4 Hss) as Hs.
so (subst_diverge _#4 Hs Hdiverge) as Hdiverge'.
eapply tcrip_bot; eauto; done.

(* id *)
intros G e G1 s s' He Hss.
exact (crip_prefunct andel _#5 He Hss).

(* lam *)
intros G t e e' _ IH G1 s s' Htm Hss.
invert Htm; [].
intros Ht He.
simp_sub.
so (sid_fcn _#5 (subrel_crq_sid _#4 Hss) Ht) as Heq.
rewrite <- Heq; [].
apply tcrip_lam; [].
apply IH; auto; [].
replsub (G1; cl_utm) with (G1; subst s cl_utm).
apply subrel_wclo_bind; auto using crq_behaved, crq_wclo, iscl; done.

(* plam *)
intros G k e e' _ IH G1 s s' Htm Hss.
invert Htm; [].
intros Hk He.
simp_sub.
so (sid_fkd _#5 (subrel_crq_sid _#4 Hss) Hk) as Heq.
rewrite <- Heq; [].
apply tcrip_plam; [].
apply IH; auto; [].
replsub (G1; cl_ucn) with (G1; subst s cl_ucn).
apply subrel_wclo_bind; auto using crq_behaved, crq_wclo, iscl; done.

(* papp *)
intros G e e' c _ IH G1 s s' Htm Hss.
invert Htm; [].
intros He Hc.
simp_sub.
so (sid_fcn _#5 (subrel_crq_sid _#4 Hss) Hc) as Heq.
rewrite <- Heq; [].
apply tcrip_papp; [].
apply IH; auto; done.

(* pack *)
intros G c e e' k t _ IH G1 s s' Htm Hss.
invert Htm; [].
intros Hc He Hk Ht.
simp_sub.
so (sid_fcn _#5 (subrel_crq_sid _#4 Hss) Hc) as Heqc.
so (sid_fkd _#5 (subrel_crq_sid _#4 Hss) Hk) as Heqk.
so (subrel_wclo_bind _ crq_behaved crq_wclo _#5 iscl_ucn Hss) as Hss'.
simp_sub in Hss'.
so (sid_fcn _#5 (subrel_crq_sid _#4 Hss') Ht) as Heqt.
rewrite <- Heqc, <- Heqk, <- Heqt; [].
apply tcrip_pack; [].
apply IH; auto; done.

(* unpack *)
intros G e1 e1' e2 e2' _ IH1 _ IH2 G1 s s' Htm Hss.
invert Htm; [].
intros He1 He2.
simp_sub.
so (subrel_wclo_bind2 _ crq_behaved crq_wclo _#6 iscl_ucn iscl_utm Hss).
apply tcrip_unpack; eauto; done.

(* fix *)
intros G t e e' _ IH G1 s s' Htm Hss.
invert Htm; [].
intros Ht He.
simp_sub.
so (sid_fcn _#5 (subrel_crq_sid _#4 Hss) Ht) as Heq.
rewrite <- Heq; [].
apply tcrip_fix; auto; done.

(* lett *)
intros G e1 e1' e2 e2' _ IH1 _ IH2 G1 s s' Htm Hss.
invert Htm; [].
intros He1 He2.
simp_sub.
so (subrel_wclo_bind _ crq_behaved crq_wclo _#5 iscl_utm Hss).
apply tcrip_lett; auto; done.

(* letm *)
intros G M M' e e' _ IH1 _ IH2 G1 s s' Htm Hss.
invert Htm; [].
intros HM He.
simp_sub.
so (subrel_wclo_bind2 _ crq_behaved crq_wclo _#5 (cl_umd (var 0)) iscl_ucn iscl_umd Hss) as Hss'.
simp_sub in Hss'.
apply tcrip_letm; eauto; done.

(* id *)
intros G M G1 s s' HM Hss.
exact (crip_prefunct ander _#5 HM Hss).

(* lamgn *)
intros G T M M' _ IH G1 s s' Hmd Hss.
invert Hmd; [].
intros HT HM.
simp_sub.
so (sid_fsg _#5 (subrel_crq_sid _#4 Hss) HT) as Heq.
rewrite <- Heq; [].
so (subrel_wclo_bind2 _ crq_behaved crq_wclo _#5 (cl_umd (var 0)) iscl_ucn iscl_umd Hss) as Hss'.
simp_sub in Hss'.
apply mcrip_lamgn; auto; done.

(* lamap *)
intros G T M M' _ IH G1 s s' Hmd Hss.
invert Hmd; [].
intros HT HM.
simp_sub.
so (sid_fsg _#5 (subrel_crq_sid _#4 Hss) HT) as Heq.
rewrite <- Heq; [].
so (subrel_wclo_bind2 _ crq_behaved crq_wclo _#5 (cl_umd (var 0)) iscl_ucn iscl_umd Hss) as Hss'.
simp_sub in Hss'.
apply mcrip_lamap; auto; done.

(* unpack *)
intros G e e' M M' T _ IH1 _ IH2 G1 s s' Hmd Hss.
invert Hmd; [].
intros He HM HT.
simp_sub.
so (sid_fsg _#5 (subrel_crq_sid _#4 Hss) HT) as Heq.
rewrite <- Heq; [].
so (subrel_wclo_bind2 _ crq_behaved crq_wclo _#6 iscl_ucn iscl_utm Hss).
apply mcrip_unpack; eauto; done.

(* lett *)
intros G e1 e1' M2 M2' _ IH1 _ IH2 G1 s s' Hmd Hss.
invert Hmd; [].
intros He1 HM2.
simp_sub.
so (subrel_wclo_bind _ crq_behaved crq_wclo _#5 iscl_utm Hss).
apply mcrip_lett; auto; done.

(* letm *)
intros G M1 M1' M2 M2' T _ IH1 _ IH2 G1 s s' Hmd Hss.
invert Hmd; [].
intros HM1 HM2 HT.
simp_sub.
so (sid_fsg _#5 (subrel_crq_sid _#4 Hss) HT) as Heq.
rewrite <- Heq; [].
so (subrel_wclo_bind2 _ crq_behaved crq_wclo _#5 (cl_umd (var 0)) iscl_ucn iscl_umd Hss) as Hss'.
simp_sub in Hss'.
apply mcrip_letm; eauto; done.

(* seal *)
intros G M M' T _ IH G1 s s' Hmd Hss.
invert Hmd; [].
intros HM HT.
simp_sub.
so (sid_fsg _#5 (subrel_crq_sid _#4 Hss) HT) as Heq.
rewrite <- Heq; [].
apply mcrip_seal; auto; done.

(* epilogue *)
destruct Hind; split; intros; eauto; done.
Qed.


Lemma tcrip_funct :
  forall G1 s s' G2 e e',
    ftm G2 e
    -> subrel crq G1 s s' G2
    -> tcrip G2 e e'
    -> tcrip G1 (subst s e) (subst s' e').
Proof.
exact (crip_funct andel).
Qed.


Lemma mcrip_funct :
  forall G1 s s' G2 M M',
    fmd G2 M
    -> subrel crq G1 s s' G2
    -> mcrip G2 M M'
    -> mcrip G1 (subst s M) (subst s' M').
Proof.
exact (crip_funct ander).
Qed.


Lemma tcrip_val :
  forall G e e',
    tcrip G e e'
    -> tvalue G e
    -> tvalue G e'.
Proof.
intros G e e' Hcrip Hval.
revert Hval.
induct Hcrip;
try (intros;
     trivial;
     match goal with
     | H : tvalue _ _ |- _ =>
         invert H; clear H; intros
     end;
     eauto with dynamic; done); [].
(* bot *)
intros G e1 e2 Hdiverge Hval1.
exfalso.
so (Hdiverge e1 (star_refl _)) as (e2' & Hstep).
eapply tstep_value_contra; eauto; done.
Qed.


Lemma mcrip_val :
  forall G M M',
    mcrip G M M'
    -> mvalue G M
    -> mvalue G M'.
Proof.
intros G M M' Hcrip Hval.
revert Hval.
induct Hcrip;
try (intros;
     trivial;
     match goal with
     | H : mvalue _ _ |- _ =>
         invert H; clear H; intros
     end;
     eauto using tcrip_val with dynamic; done).
Qed.


Lemma mcrip_fstm :
  forall G M N c,
    mcrip G M N
    -> fstm G M c
    -> fstm G N c.
Proof.
intros G M N c Hcrip Hfst.
revert N Hcrip.
induct Hfst;
try (intros; invert Hcrip; intros; subst; eauto with static; done); [].

(* lam *)
intros G T M k c HfstT _ IH N Hcrip.
assert (exists M', mcrip (G; cl_ucn; cl_umd (var 0)) M M' /\ md_lamap T M' = N) as (M' & Mcrip & <-).
  #2 invert Hcrip.
    intros <-.
    exists M.
    split; auto using mcrip_id; done.

    eauto; done.
apply fstm_lamap; auto; [].
apply IH; [].
rewrite <- (subst_dot01_sh M); [].
rewrite <- (subst_dot01_sh M'); [].
eapply subst_mcrip; eauto; [].
#2 apply subof_dot.
  apply subof_dot; auto using subof_sh with sequence; [].
  simp_sub.
  apply oof_ucn; done.

  simp_sub.
  #2 apply oof_umd.
    eapply mvalue_var; eauto with sequence; done.

    replsub (var 1) with (shift1 (var 0)).
    eapply fstm_var; eauto with sequence; done.
Qed.


Lemma cripple_step :
  (forall G e e' f,
     ftm G e
     -> tcrip G e f
     -> tstep G e e'
     -> exists f',
          tcrip G e' f'
          /\ star (tstep G) f f')
  /\
  (forall G M M' N,
     fmd G M
     -> mcrip G M N
     -> mstep G M M'
     -> exists N',
          mcrip G M' N'
          /\ star (mstep G) N N').
Proof.
exploit 
  (cripple_ind
     (fun G e f =>
        forall e',
          ftm G e
          -> tstep G e e'
          -> exists f', tcrip G e' f' /\ star (tstep G) f f')
     (fun G M N =>
        forall M',
          fmd G M
          -> mstep G M M'
          -> exists N', mcrip G M' N' /\ star (mstep G) N N')) as Hind.

(* bot *)
intros G e f Hdiverge e' Htm Hstep1.
exists f.
split.
  eapply tcrip_bot; eauto; [].
  eapply diverge_step; eauto; done.

  apply star_refl; done.

(* id *)
intros G e e' _ Hstep.
exists e'.
split.
  apply tcrip_id; done.

  apply star_one; auto; done.

(* lam *)
intros G t e f H H0 e' _ Hstep.
invert Hstep; done.

(* app *)
intros G e1 f1 e2 f2 Hcrip1 IH1 Hcrip2 IH2 e Htm Hstep.
invert Htm; [].
intros He1 He2.
#3 invert Hstep.
  intros e1' Hstep1 <-.
  so (IH1 _ He1 Hstep1) as (f1' & Hcrip1' & Hstep1').
  exists (tm_app f1' f2).
  split.
    apply tcrip_app; auto; done.

    apply (star_map' _ _ (fun f => tm_app f _)); auto with dynamic; done.

  intros e2' Hval Hstep2 <-.
  so (IH2 _ He2 Hstep2) as (f2' & Hcrip2' & Hstep2').
  exists (tm_app f1 f2').
  split.
    apply tcrip_app; auto; done.

    apply (star_map' _ _ (fun f => tm_app _ f)); eauto using tcrip_val with dynamic; done.

  intros t e3 Hval <- <-.
  invert He1; [].
  intros _ He3.
  assert (exists u f3, f1 = tm_lam u f3 /\ tcrip (G; cl_utm) e3 f3) as (u & f3 & -> & Hcrip3).
    invert Hcrip1; eauto using tcrip_id; [].
    intros Hdiverge.
    exfalso.
    eapply diverge_value_contra; eauto with dynamic; done.
  exists (subst1 f2 f3).
  split.
    eapply tcrip_funct; eauto; [].
    apply subrel_dot; auto using subrel_id; [].
    simp_sub.
    apply crq_utm; auto; done.

    apply star_one; [].
    so (tcrip_val _#3 Hcrip2 Hval) as Hval'.
    apply tstep_app3; auto; done.

(* pair *)
intros G e1 f1 e2 f2 Hcrip1 IH1 Hcrip2 IH2 e' Htm Hstep.
invert Htm; [].
intros He1 He2.
#2 invert Hstep.
  intros e1' Hstep1 <-.
  so (IH1 _ He1 Hstep1) as (f1' & Hcrip1' & Hsteps).
  exists (tm_pair f1' f2).
  split.
    apply tcrip_pair; auto; done.

    apply (star_map' _ _ (fun f => tm_pair f _)); eauto with dynamic; done.

  intros e2' Hval1 Hstep2 <-.
  so (IH2 _ He2 Hstep2) as (f2' & Hcrip2' & Hsteps).
  exists (tm_pair f1 f2').
  split.
    apply tcrip_pair; auto; done.

    so (tcrip_val _#3 Hcrip1 Hval1) as Hval1'.
    apply (star_map' _ _ (fun f => tm_pair _ f)); eauto with dynamic; done.

(* pi1 *)
intros G e1 f1 Hcrip IH e' Htm Hstep.
invert Htm; [].
intros He1.
#2 invert Hstep.
  intros e1' Hstep1 <-.
  so (IH _ He1 Hstep1) as (f1' & Hcrip1' & Hsteps).
  exists (tm_pi1 f1').
  split.
    apply tcrip_pi1; auto; done.

    apply (star_map' _ _ tm_pi1); eauto with dynamic; done.

  intros v2 Hval1 Hval2 <-.
  rename e' into v1.
  assert (exists w1 w2, f1 = tm_pair w1 w2 /\ tcrip G v1 w1 /\ tcrip G v2 w2) as (w1 & w2 & -> & Hcrip1 & Hcrip2).
    invert Hcrip; eauto using tcrip_id; [].
    intros Hdiverge.
    exfalso.
    eapply diverge_value_contra; eauto with dynamic; done.
  exists w1.
  split; trivial; [].
  so (tcrip_val _#3 Hcrip1 Hval1) as Hval1'.
  so (tcrip_val _#3 Hcrip2 Hval2) as Hval2'.
  apply star_one; auto with dynamic; done.

(* pi2 *)
intros G e1 f1 Hcrip IH e' Htm Hstep.
invert Htm; [].
intros He1.
#2 invert Hstep.
  intros e1' Hstep1 <-.
  so (IH _ He1 Hstep1) as (f1' & Hcrip1' & Hsteps).
  exists (tm_pi2 f1').
  split.
    apply tcrip_pi2; auto; done.

    apply (star_map' _ _ tm_pi2); eauto with dynamic; done.

  intros v1 Hval1 Hval2 <-.
  rename e' into v2.
  assert (exists w1 w2, f1 = tm_pair w1 w2 /\ tcrip G v1 w1 /\ tcrip G v2 w2) as (w1 & w2 & -> & Hcrip1 & Hcrip2).
    invert Hcrip; eauto using tcrip_id; [].
    intros Hdiverge.
    exfalso.
    eapply diverge_value_contra; eauto with dynamic; done.
  exists w2.
  split; trivial; [].
  so (tcrip_val _#3 Hcrip1 Hval1) as Hval1'.
  so (tcrip_val _#3 Hcrip2 Hval2) as Hval2'.
  apply star_one; auto with dynamic; done.

(* plam *)
intros G k e1 f1 _ _  e' _ Hstep.
invert Hstep; done.

(* papp *)
intros G e f c Hcrip IH g Htm Hstep.
invert Htm; [].
intros He _.
#2 invert Hstep.
  intros e' Hstep' <-.
  so (IH _ He Hstep') as (f' & Hcrip' & Hsteps).
  exists (tm_papp f' c).
  split.
    apply tcrip_papp; auto; done.

    apply (star_map' _ _ (fun f => tm_papp f _)); auto with dynamic; done.

  intros k e1 <- <-.
  invert He; [].
  intros _ He1.
  assert (exists l f1, f = tm_plam l f1 /\ tcrip (G; cl_ucn) e1 f1) as (l & f1 & -> & Hcrip1).
    invert Hcrip; eauto using tcrip_id; [].
    intros Hdiverge.
    exfalso.
    eapply diverge_value_contra; eauto with dynamic; done.
  exists (subst1 c f1).
  split.
    eapply tcrip_funct; eauto; [].
    apply subrel_dot; auto using subrel_id; [].
    simp_sub.
    apply crq_ucn; done.

    apply star_one; auto with dynamic; done.

(* pack *)
intros G c e f k t _ IH g Htm Hstep.
invert Htm; [].
intros _ He _ _.
invert Hstep; [].
intros e' Hstep' <-.
so (IH _ He Hstep') as (f' & Hcrip & Hsteps).
exists (tm_pack c f' k t).
split.
  apply tcrip_pack; auto; done.

  apply (star_map' _ _ (fun f => tm_pack _ f _ _)); auto with dynamic; done.

(* unpack *)
intros G e1 f1 e2 f2 Hcrip1 IH Hcrip2 _ g Htm Hstep.
invert Htm; [].
intros He1 He2.
#2 invert Hstep.
  intros e1' Hstep1 <-.
  so (IH _ He1 Hstep1) as (f1' & Hcrip1' & Hsteps).
  exists (tm_unpack f1' f2).
  split.
    apply tcrip_unpack; auto; done.

    apply (star_map' _ _ (fun f => tm_unpack f _)); auto with dynamic; done.

  intros c v k t Hval <- <-.
  assert (exists w, f1 = tm_pack c w k t /\ tcrip G v w) as (w & -> & Hcrip).
    invert Hcrip1; eauto using tcrip_id; [].
    intros Hdiv.
    exfalso.
    eapply diverge_value_contra; eauto with dynamic; done.
  exists (subst (dot w (dot c id)) f2).
  split.
    eapply tcrip_funct; eauto; [].
    #2 apply subrel_dot.
      apply subrel_dot; auto using subrel_id; [].
      simp_sub.
      apply crq_ucn; done.

      simp_sub.
      apply crq_utm; auto; done.

    so (tcrip_val _#3 Hcrip Hval).
    apply star_one; auto with dynamic; done.

(* fix *)
intros G t e f Hcrip IH g _ Hstep.
invert Hstep; [].
intros <-.
exists (tm_app f (tm_lam cn_unit (shift1 (tm_fix t f)))).
split.
  apply tcrip_app; auto; [].
  apply tcrip_lam; auto; [].
  eapply subst_tcrip; eauto using tcrip; [].
  apply subof_sh; auto with sequence; done.

  apply star_one; auto with dynamic; done.

(* lett *)
intros G e1 f1 e2 f2 Hcrip1 IH Hcrip2 _ g Htm Hstep.
invert Htm; [].
intros He1 He2.
#2 invert Hstep.
  intros e1' Hstep1 <-.
  so (IH _ He1 Hstep1) as (f1' & Hcrip1' & Hsteps).
  exists (tm_lett f1' f2).
  split.
    eauto using tcrip; done.

    apply (star_map' _ _ (fun f => tm_lett f _)); auto with dynamic; done.

  intros Hval <-.
  exists (subst1 f1 f2).
  split.
    eapply tcrip_funct; eauto; [].
    apply subrel_dot; auto using subrel_id; [].
    simp_sub.
    apply crq_utm; auto; done.

    so (tcrip_val _#3 Hcrip1 Hval).
    apply star_one; auto with dynamic; done.

(* letm *)
intros G M N e f Hcrip1 IH Hcrip2 _ g Htm Hstep.
invert Htm; [].
intros HM He.
#2 invert Hstep.
  intros M' Hstep' <-.
  so (IH _ HM Hstep') as (N' & Hcrip1' & Hsteps).
  exists (tm_letm N' f).
  split.
    eapply tcrip_letm; eauto; done.

    eapply (star_mono_map _ _ _ (fun N => tm_letm N _)); eauto with dynamic; done.

  intros c Hval Hfst <-.
  so (mcrip_fstm _#4 Hcrip1 Hfst) as Hfst'.
  exists (subst (dot N (dot c id)) f).
  split.
    eapply tcrip_funct; eauto; [].
    #2 apply subrel_dot.
      apply subrel_dot; auto using subrel_id; [].
      simp_sub.
      apply crq_ucn; done.

      simp_sub.
      apply crq_umd; auto; done.

    so (mcrip_val _#3 Hcrip1 Hval).
    apply star_one; auto with dynamic; done.

(* ext *)
intros G M N Hcrip IH e' Htm Hstep.
invert Htm; [].
intros HM.
#2 invert Hstep.
  intros M' Hstep' <-.
  so (IH _ HM Hstep') as (N' & Hcrip' & Hsteps).
  exists (tm_ext N').
  split.
    apply tcrip_ext; auto; done.

    eapply (star_mono_map _ _ _ tm_ext); eauto with dynamic; done.

  intros Hval <-.
  assert (exists f', tcrip G e' f' /\ N = md_dyn f') as (f' & Hcrip' & ->).
    invert Hcrip; eauto using tcrip_id; done.
  exists f'.
  split; auto; [].
  apply star_one; [].
  apply tstep_ext2; [].
  eapply tcrip_val; eauto; done.

(* id *)
intros G M M' _ Hstep.
exists M'.
split.
  apply mcrip_id; done.

  apply star_one; auto; done.

(* dyn *)
intros G e f _ IH M' Hmd Hstep.
invert Hmd; [].
intros He.
invert Hstep; [].
intros e' Hstepe <-.
so (IH _ He Hstepe) as (f' & Hcrip & Hsteps).
exists (md_dyn f').
split.
  apply mcrip_dyn; auto; done.

  eapply (star_mono_map _#3 md_dyn); eauto with dynamic; done.

(* lamgn *)
intros G T M N _ _ M' _ Hstep.
invert Hstep; done.

(* appgn *)
intros G M1 N1 M2 N2 Hcrip1 IH1 Hcrip2 IH2 M' Hmd Hstep.
invert Hmd; [].
intros HM1 HM2.
#3 invert Hstep.
  intros M1' Hstep1 <-.
  so (IH1 _ HM1 Hstep1) as (N1' & Hcrip1' & Hsteps).
  exists (md_appgn N1' N2).
  split.
    apply mcrip_appgn; auto; done.

    eapply (star_mono_map _#3 (fun M => md_appgn M _)); eauto with dynamic; done.

  intros M2' Hval1 Hstep2 <-.
  so (IH2 _ HM2 Hstep2) as (N2' & Hcrip2' & Hsteps).
  exists (md_appgn N1 N2').
  split.
    apply mcrip_appgn; auto; done.

    so (mcrip_val _#3 Hcrip1 Hval1) as Hval1'.
    eapply (star_mono_map _#3 (fun M => md_appgn _ M)); eauto with dynamic; done.

  intros T M3 c Hval2 HfstM2 <- <-.
  invert HM1; [].
  intros _ HM3.
  assert (exists N3, mcrip (G; cl_ucn; cl_umd (var 0)) M3 N3 /\ N1 = md_lamgn T N3) as (N3 & Hcrip3 & ->).
    invert Hcrip1; eauto using mcrip_id; done.
  exists (subst (dot N2 (dot c id)) N3).
  split.
    eapply mcrip_funct; eauto; [].
    #2 apply subrel_dot.
      apply subrel_dot; auto using subrel_id; [].
      simp_sub.
      apply crq_ucn; auto; done.

      simp_sub.
      apply crq_umd; auto; done.

    apply star_one; [].
    so (mcrip_val _#3 Hcrip2 Hval2) as Hval2'.
    so (mcrip_fstm _#4 Hcrip2 HfstM2) as HfstN2.
    apply mstep_appgn3; auto; done.

(* lamap *)
intros G T M N _ _ M' _ Hstep.
invert Hstep; done.

(* appap *)
intros G M1 N1 M2 N2 Hcrip1 IH1 Hcrip2 IH2 M' Hmd Hstep.
invert Hmd; [].
intros HM1 HM2.
#3 invert Hstep.
  intros M1' Hstep1 <-.
  so (IH1 _ HM1 Hstep1) as (N1' & Hcrip1' & Hsteps).
  exists (md_appap N1' N2).
  split.
    apply mcrip_appap; auto; done.

    eapply (star_mono_map _#3 (fun M => md_appap M _)); eauto with dynamic; done.

  intros M2' Hval1 Hstep2 <-.
  so (IH2 _ HM2 Hstep2) as (N2' & Hcrip2' & Hsteps).
  exists (md_appap N1 N2').
  split.
    apply mcrip_appap; auto; done.

    so (mcrip_val _#3 Hcrip1 Hval1) as Hval1'.
    eapply (star_mono_map _#3 (fun M => md_appap _ M)); eauto with dynamic; done.

  intros T M3 c Hval2 HfstM2 <- <-.
  invert HM1; [].
  intros _ HM3.
  assert (exists N3, mcrip (G; cl_ucn; cl_umd (var 0)) M3 N3 /\ N1 = md_lamap T N3) as (N3 & Hcrip3 & ->).
    invert Hcrip1; eauto using mcrip_id; done.
  exists (subst (dot N2 (dot c id)) N3).
  split.
    eapply mcrip_funct; eauto; [].
    #2 apply subrel_dot.
      apply subrel_dot; auto using subrel_id; [].
      simp_sub.
      apply crq_ucn; auto; done.

      simp_sub.
      apply crq_umd; auto; done.

    apply star_one; [].
    so (mcrip_val _#3 Hcrip2 Hval2) as Hval2'.
    so (mcrip_fstm _#4 Hcrip2 HfstM2) as HfstN2.
    apply mstep_appap3; auto; done.

(* pair *)
intros G M1 N1 M2 N2 Hcrip1 IH1 Hcrip2 IH2 M' Hmd Hstep.
invert Hmd; [].
intros HM1 HM2.
#2 invert Hstep.
  intros M1' Hstep1 <-.
  so (IH1 _ HM1 Hstep1) as (N1' & Hcrip1' & Hsteps).
  exists (md_pair N1' N2).
  split.
    apply mcrip_pair; auto; done.

    eapply (star_mono_map _#3 (fun M => md_pair M _)); eauto with dynamic; done.

  intros M1' Hval1 Hstep2 <-.
  so (IH2 _ HM2 Hstep2) as (N2' & Hcrip2' & Hsteps).
  exists (md_pair N1 N2').
  split.
    apply mcrip_pair; auto; done.

    so (mcrip_val _#3 Hcrip1 Hval1) as Hval1'.
    eapply (star_mono_map _#3 (fun M => md_pair _ M)); eauto with dynamic; done.

(* pi1 *)
intros G M1 N1 Hcrip IH M' Hmd Hstep.
invert Hmd; [].
intros HM1.
#2 invert Hstep.
  intros M1' Hstep1 <-.
  so (IH _ HM1 Hstep1) as (N1' & Hcrip1' & Hsteps).
  exists (md_pi1 N1').
  split.
    apply mcrip_pi1; auto; done.

    apply (star_map' _ _ md_pi1); eauto with dynamic; done.

  intros V2 Hval1 Hval2 <-.
  rename M' into V1.
  assert (exists W1 W2, mcrip G V1 W1 /\ mcrip G V2 W2 /\ N1 = md_pair W1 W2) as (W1 & W2 & Hcrip1 & Hcrip2 & ->).
    invert Hcrip; eauto using mcrip_id; done.
  exists W1.
  split; trivial; [].
  so (mcrip_val _#3 Hcrip1 Hval1) as Hval1'.
  so (mcrip_val _#3 Hcrip2 Hval2) as Hval2'.
  apply star_one; auto with dynamic; done.

(* pi2 *)
intros G M1 N1 Hcrip IH M' Hmd Hstep.
invert Hmd; [].
intros HM1.
#2 invert Hstep.
  intros M1' Hstep1 <-.
  so (IH _ HM1 Hstep1) as (N1' & Hcrip1' & Hsteps).
  exists (md_pi2 N1').
  split.
    apply mcrip_pi2; auto; done.

    apply (star_map' _ _ md_pi2); eauto with dynamic; done.

  intros V1 Hval1 Hval2 <-.
  rename M' into V2.
  assert (exists W1 W2, mcrip G V1 W1 /\ mcrip G V2 W2 /\ N1 = md_pair W1 W2) as (W1 & W2 & Hcrip1 & Hcrip2 & ->).
    invert Hcrip; eauto using mcrip_id; done.
  exists W2.
  split; trivial; [].
  so (mcrip_val _#3 Hcrip1 Hval1) as Hval1'.
  so (mcrip_val _#3 Hcrip2 Hval2) as Hval2'.
  apply star_one; auto with dynamic; done.

(* unpack *)
intros G e1 f1 M2 N2 T Hcrip1 IH Hcrip2 _ M' Hmd Hstep.
invert Hmd; [].
intros He1 HM2.
#2 invert Hstep.
  intros e1' Hstep1 <-.
  so (IH _ He1 Hstep1) as (f1' & Hcrip1' & Hsteps).
  exists (md_unpack f1' N2 T).
  split.
    apply mcrip_unpack; auto; done.

    eapply (star_mono_map _ _ _ (fun f => md_unpack f _ _)); eauto with dynamic; done.

  intros c v k t Hval <- <-.
  assert (exists w, f1 = tm_pack c w k t /\ tcrip G v w) as (w & -> & Hcrip).
    invert Hcrip1; eauto using tcrip_id; [].
    intros Hdiv.
    exfalso.
    eapply diverge_value_contra; eauto with dynamic; done.
  exists (subst (dot w (dot c id)) N2).
  split.
    eapply mcrip_funct; eauto; [].
    #2 apply subrel_dot.
      apply subrel_dot; auto using subrel_id; [].
      simp_sub.
      apply crq_ucn; done.

      simp_sub.
      apply crq_utm; auto; done.

    so (tcrip_val _#3 Hcrip Hval).
    apply star_one; auto with dynamic; done.

(* lett *)
intros G e1 f1 M2 N2 Hcrip1 IH Hcrip2 _ M' Hmd Hstep.
invert Hmd; [].
intros He1 HM2.
#2 invert Hstep.
  intros e1' Hstep1 <-.
  so (IH _ He1 Hstep1) as (f1' & Hcrip1' & Hsteps).
  exists (md_lett f1' N2).
  split.
    eauto using mcrip; done.

    eapply (star_mono_map _ _ _ (fun f => md_lett f _)); eauto with dynamic; done.

  intros Hval <-.
  exists (subst1 f1 N2).
  split.
    eapply mcrip_funct; eauto; [].
    apply subrel_dot; auto using subrel_id; [].
    simp_sub.
    apply crq_utm; auto; done.

    so (tcrip_val _#3 Hcrip1 Hval).
    apply star_one; auto with dynamic; done.

(* letm *)
intros G M1 N1 M2 N2 T Hcrip1 IH Hcrip2 _ M' Hmd Hstep.
invert Hmd; [].
intros HM1 HM2.
#2 invert Hstep.
  intros M1' Hstep' <-.
  so (IH _ HM1 Hstep') as (N1' & Hcrip1' & Hsteps).
  exists (md_letm N1' N2 T).
  split.
    eapply mcrip_letm; eauto; done.

    eapply (star_mono_map _ _ _ (fun N => md_letm N _ _)); eauto with dynamic; done.

  intros c Hval Hfst <-.
  so (mcrip_fstm _#4 Hcrip1 Hfst) as Hfst'.
  exists (subst (dot N1 (dot c id)) N2).
  split.
    eapply mcrip_funct; eauto; [].
    #2 apply subrel_dot.
      apply subrel_dot; auto using subrel_id; [].
      simp_sub.
      apply crq_ucn; done.

      simp_sub.
      apply crq_umd; auto; done.

    so (mcrip_val _#3 Hcrip1 Hval).
    apply star_one; auto with dynamic; done.

(* seal *)
intros G M1 N1 T Hcrip _ M' _ Hstep.
invert Hstep; [].
intros <-.
exists N1.
split.
  assumption.

  apply star_one; auto with dynamic; done.

(* epilogue *)
destruct Hind; split; intros; eauto.
Qed.


(* This really only needs syntactic formation (not typing), but I didn't want 
   to go to the trouble to prove preservation for syntactic formation.
*)
Lemma mcrip_halt :
  forall G M M' T,
    wf G
    -> ofm G impure M T
    -> mcrip G M M'
    -> mhalt G M
    -> mhalt G M'.
Proof.
intros G M M' T Hwf Hof Hcrip Hhalt.
destruct Hhalt as (V & Hsteps & Hval).
assert (exists V',
          mcrip G V V'
          /\ star (mstep G) M' V') as (V' & Hcrip' & Hsteps').
  clear Hval.
  revert M' T Hwf Hof Hcrip.
  #2 induct Hsteps.
    (* refl *)
    intros M M' _ _ _ Hcrip.
    exists M'.
    split.
      assumption.

      apply star_refl; done.

    (* step *)
    intros M1 M2 M3 Hstep _ IH M1' T Hwf Hof1 Hcrip1.
    so (ofm_fmd _#4 (wf_fcontext _ Hwf) Hof1) as HM1.
    so (cripple_step ander _#4 HM1 Hcrip1 Hstep) as (M2' & Hcrip2 & Hsteps).    
    so (Preservation.preservation_md _#5 Hwf Hof1 Hstep) as Hof2.
    so (IH _ _ Hwf Hof2 Hcrip2) as (M3' & Hcrip3 & Hsteps').
    exists M3'.
    split.
      assumption.

      eapply star_trans; eauto; done.
exists V'.
split.
  assumption.

  eapply mcrip_val; eauto; done.
Qed.


Local Ltac invsame :=
  repeat
    (match goal with
     | H : samek _ _ _ |- _ =>
       invert H; clear H; intros H <-; []

     | H : samec _ _ _ _ |- _ =>
       invert H; clear H; intros H <-; []

     | H : sames _ _ _ |- _ =>
       invert H; clear H; intros H <-; []
     end).


Lemma crip_compat :
  (forall G e e' t, tcc samek samec sames (fun G e e' t => tcrip G e e') (fun G p M M' T => mcrip G M M') G e e' t -> tcrip G e e')
  /\
  (forall G p M M' T, mcc samek samec sames (fun G e e' t => tcrip G e e') (fun G p M M' T => mcrip G M M') G p M M' T -> mcrip G M M').
Proof.
assert (forall G t, subof oof (G; cl_utm) (under 1 id) (G; cl_tm t)) as Hsuboft.
  intros G t.
  rewrite -> under_succ; [].
  #2 apply subof_dot.
    simp_sub.
    apply subof_sh; auto with sequence; done.

    simp_sub.
    apply oof_tm; eauto with dynamic sequence; done.
assert (forall G k t, subof oof (G; cl_ucn; cl_utm) (under 2 id) (G; cl_cn k; cl_tm t)) as Hsubofct.
  intros G k t.  
  rewrite -> under_succ2; [].
  simp_sub.
  #2 apply subof_dot.
    apply subof_dot; auto using subof_sh with sequence; [].
    simp_sub.
    apply oof_cn; done.

    simp_sub.
    apply oof_tm; eauto with dynamic sequence; done.
assert (forall G k T c, subof oof (G; cl_ucn; cl_umd c) (under 2 id) (G; cl_cn k; cl_md T c)) as Hsubofm.
  intros G k T C.
  rewrite -> under_succ2; [].
  simp_sub.
  #2 apply subof_dot.
    apply subof_dot; auto using subof_sh with sequence; [].
    simp_sub.
    apply oof_cn; done.

    simp_sub.
    apply oof_md; eauto with dynamic sequence; [].
    rewrite -> subst_dot1_sh; [].
    eapply fstm_varu; eauto with sequence; done.
split.


intros G e e' t H.
#5 induct H;
try (intros; invsame; eauto using tcrip; done).

(* lam *)
intros G t1 t1' t2 e e' Hofc Hcrip.
invsame.
apply tcrip_lam; [].
so (subst_tcrip _#5 (Hsuboft _ _) Hcrip) as H.
simp_sub in H.
assumption.

(* plam *)
intros G k k' t e e' Hofk Hcrip.
invsame.
apply tcrip_plam; [].
#2 sor (refine (subst_tcrip (G; cl_ucn) (under 1 id) _#3 _ Hcrip)) as H.
  rewrite -> under_succ; [].
  #2 apply subof_dot.
    simp_sub.
    apply subof_sh; auto with sequence; done.

    simp_sub.
    apply oof_cn; done.
simp_sub in H.
assumption.

(* unpack *)
intros G k t t' e1 e1' e2 e2' Hcrip1 Hcrip2 _.
apply tcrip_unpack; auto; [].
so (subst_tcrip _#5 (Hsubofct _#3) Hcrip2) as H.
simp_sub in H.
assumption.

(* lett *)
intros G t1 t2 e1 e1' e2 e2' Hcrip1 Hcrip2.
apply tcrip_lett; auto; [].
so (subst_tcrip _#5 (Hsuboft _ _) Hcrip2) as H.
simp_sub in H.
assumption.

(* letm *)
intros G T k t M M' e e' Hcrip1 _ Hcrip2 _.
apply tcrip_letm; auto; [].
so (subst_tcrip _#5 (Hsubofm _#4) Hcrip2) as H.
simp_sub in H.
assumption.



intros G p M M' T H.
#5 induct H; try (intros; invsame; eauto using mcrip; done).

(* lamgn *)
intros G T1 T1' T2 k1 M M' Hofs _ Hcrip.
invsame.
apply mcrip_lamgn; [].
so (subst_mcrip _#5 (Hsubofm _#4) Hcrip) as H.
simp_sub in H.
assumption.

(* lamap *)
intros G T1 R1' T2 k1 M M' Hofs _ Hcrip.
invsame.
apply mcrip_lamap; [].
so (subst_mcrip _#5 (Hsubofm _#4) Hcrip) as H.
simp_sub in H.
assumption.

(* unpack *)
intros G k t T T'' e e' M M' Hcrip1 Hcrip2 Hofs.
invsame.
apply mcrip_unpack; auto; [].
so (subst_mcrip _#5 (Hsubofct _#3) Hcrip2) as H.
simp_sub in H.
assumption.

(* lett *)
intros G p t T e e' M M' Hcrip1 Hcrip2.
apply mcrip_lett; auto; [].
so (subst_mcrip _#5 (Hsuboft _ _) Hcrip2) as H.
simp_sub in H.
assumption.

(* letm *)
intros G T1 T2 T2' k1 M1 M1' M2 M2' Hcrip1 _ Hcrip2 HofT2.
invsame.
apply mcrip_letm; auto; [].
so (subst_mcrip _#5 (Hsubofm _#4) Hcrip2) as H.
simp_sub in H.
assumption.
Qed.


Lemma crip_quasicongruence :
  quasicongruence (treg (fun G e e' t => tcrip G e e')) (mreg (fun G p M M' T => mcrip G M M')).
Proof.
#3 apply reg_quasicongruence.
  split.
    intros; apply tcrip_id; done.

    intros; apply mcrip_id; done.

  split.
    intros; eapply subst_tcrip; eauto using subof_aof_oof; done.

    intros; eapply subst_mcrip; eauto using subof_aof_oof; done.

  split; intros; [eapply (crip_compat andel) | eapply (crip_compat ander)]; eauto; done.
Qed.


Lemma crip_consistent :
  consistent (mreg (fun G p M M' T => mcrip G M M')).
Proof.
apply reg_consistent; [].
intros M M' T Hof Hof' Hcrip Hhalt.
exact (mcrip_halt _#4 wf_nil Hof Hcrip Hhalt).
Qed.


Lemma diverge_capproxt :
  forall G e e' t,
    oft G e t
    -> oft G e' t
    -> diverge G e
    -> capproxt G e e' t.
Proof.
intros G e e' t Hof Hof' Hdiv.
refine (capproxt_univ_quasi _#6 crip_quasicongruence crip_consistent _); [].
do2 2 split; eauto; [].
apply tcrip_bot; auto; done.
Qed.
