
Require Import Tactics.
Require Import Sequence.
Require Import Subst.
Require Import Static.
Require Import Relation.
Export Syntax.


Inductive tvalue : context -> (* tm *) term -> Prop :=
| tvalue_var {G i t}
    : index i G (cl_tm t)
      -> tvalue G (var i)

| tvalue_varu {G i}
    : index i G cl_utm
      -> tvalue G (var i)

| tvalue_star {G}
    : tvalue G tm_star

| tvalue_lam {G t e}
    : tvalue G (tm_lam t e)

| tvalue_pair {G v1 v2}
    : tvalue G v1
      -> tvalue G v2
      -> tvalue G (tm_pair v1 v2)

| tvalue_plam {G k e}
    : tvalue G (tm_plam k e)

| tvalue_pack {G c v k t}
    : tvalue G v
      -> tvalue G (tm_pack c v k t)
.


Inductive mvalue : context -> (* md *) term -> Prop :=
| mvalue_var {G i T c}
    : index i G (cl_md T c)
      -> mvalue G (var i)

| mvalue_varu {G i c}
    : index i G (cl_umd c)
      -> mvalue G (var i)

| mvalue_star {G}
    : mvalue G md_star

| mvalue_stat {G c}
    : mvalue G (md_stat c)

| mvalue_dyn {G v}
    : tvalue G v
      -> mvalue G (md_dyn v)

| mvalue_lamgn {G T M}
    : mvalue G (md_lamgn T M)

| mvalue_lamap {G T M}
    : mvalue G (md_lamap T M)

| mvalue_pair {G V1 V2}
    : mvalue G V1
      -> mvalue G V2
      -> mvalue G (md_pair V1 V2)
.


Inductive tstep : context -> (* tm *) term -> (* tm *) term -> Prop :=
| tstep_app1 {G e1 e1' e2}
    : tstep G e1 e1'
      -> tstep G (tm_app e1 e2) (tm_app e1' e2)

| tstep_app2 {G v1 e2 e2'}
    : tvalue G v1
      -> tstep G e2 e2'
      -> tstep G (tm_app v1 e2) (tm_app v1 e2')

| tstep_app3 {G t e1 v2}
    : tvalue G v2
      -> tstep G (tm_app (tm_lam t e1) v2) (subst1 v2 e1)

| tstep_pair1 {G e1 e1' e2}
    : tstep G e1 e1'
      -> tstep G (tm_pair e1 e2) (tm_pair e1' e2)

| tstep_pair2 {G v1 e2 e2'}
    : tvalue G v1
      -> tstep G e2 e2'
      -> tstep G (tm_pair v1 e2) (tm_pair v1 e2')

| tstep_pi11 {G e e'}
    : tstep G e e'
      -> tstep G (tm_pi1 e) (tm_pi1 e')

| tstep_pi12 {G v1 v2}
    : tvalue G v1
      -> tvalue G v2
      -> tstep G (tm_pi1 (tm_pair v1 v2)) v1

| tstep_pi21 {G e e'}
    : tstep G e e'
      -> tstep G (tm_pi2 e) (tm_pi2 e')

| tstep_pi22 {G v1 v2}
    : tvalue G v1
      -> tvalue G v2
      -> tstep G (tm_pi2 (tm_pair v1 v2)) v2

| tstep_papp1 {G e e' c}
    : tstep G e e'
      -> tstep G (tm_papp e c) (tm_papp e' c)

| tstep_papp2 {G k e c}
    : tstep G (tm_papp (tm_plam k e) c) (subst1 c e)

| tstep_pack {G c e e' k t}
    : tstep G e e'
      -> tstep G (tm_pack c e k t) (tm_pack c e' k t)

| tstep_unpack1 {G e1 e1' e2}
    : tstep G e1 e1'
      -> tstep G (tm_unpack e1 e2) (tm_unpack e1' e2)

| tstep_unpack2 {G c v k t e}
    : tvalue G v
      -> tstep G (tm_unpack (tm_pack c v k t) e) (subst (dot v (dot c id)) e)

| tstep_fix {G t e}
    : tstep G (tm_fix t e) (tm_app e (tm_lam cn_unit (shift1 (tm_fix t e))))

| tstep_lett1 {G e1 e1' e2}
    : tstep G e1 e1'
      -> tstep G (tm_lett e1 e2) (tm_lett e1' e2)

| tstep_lett2 {G v1 e2}
    : tvalue G v1
      -> tstep G (tm_lett v1 e2) (subst1 v1 e2)

| tstep_letm1 {G M M' e}
    : mstep G M M'
      -> tstep G (tm_letm M e) (tm_letm M' e)

| tstep_letm2 {G V e c}
    : mvalue G V
      -> fstm G V c
      -> tstep G (tm_letm V e) (subst (dot V (dot c id)) e)

| tstep_ext1 {G M M'}
    : mstep G M M'
      -> tstep G (tm_ext M) (tm_ext M')

| tstep_ext2 {G v}
    : tvalue G v
      -> tstep G (tm_ext (md_dyn v)) v


with mstep : context -> (* md *) term -> (* md *) term -> Prop :=
| mstep_dyn {G e e'}
    : tstep G e e'
      -> mstep G (md_dyn e) (md_dyn e')

| mstep_appgn1 {G M1 M1' M2}
    : mstep G M1 M1'
      -> mstep G (md_appgn M1 M2) (md_appgn M1' M2)

| mstep_appgn2 {G V1 M2 M2'}
    : mvalue G V1
      -> mstep G M2 M2'
      -> mstep G (md_appgn V1 M2) (md_appgn V1 M2')

| mstep_appgn3 {G T M1 V2 c2}
    : mvalue G V2
      -> fstm G V2 c2
      -> mstep G (md_appgn (md_lamgn T M1) V2) (subst (dot V2 (dot c2 id)) M1)

| mstep_appap1 {G M1 M1' M2}
    : mstep G M1 M1'
      -> mstep G (md_appap M1 M2) (md_appap M1' M2)

| mstep_appap2 {G V1 M2 M2'}
    : mvalue G V1
      -> mstep G M2 M2'
      -> mstep G (md_appap V1 M2) (md_appap V1 M2')

| mstep_appap3 {G T M1 V2 c2}
    : mvalue G V2
      -> fstm G V2 c2
      -> mstep G (md_appap (md_lamap T M1) V2) (subst (dot V2 (dot c2 id)) M1)

| mstep_pair1 {G M1 M1' M2}
    : mstep G M1 M1'
      -> mstep G (md_pair M1 M2) (md_pair M1' M2)

| mstep_pair2 {G V1 M2 M2'}
    : mvalue G V1
      -> mstep G M2 M2'
      -> mstep G (md_pair V1 M2) (md_pair V1 M2')

| mstep_pi11 {G M M'}
    : mstep G M M'
      -> mstep G (md_pi1 M) (md_pi1 M')

| mstep_pi12 {G V1 V2}
    : mvalue G V1
      -> mvalue G V2
      -> mstep G (md_pi1 (md_pair V1 V2)) V1

| mstep_pi21 {G M M'}
    : mstep G M M'
      -> mstep G (md_pi2 M) (md_pi2 M')

| mstep_pi22 {G V1 V2}
    : mvalue G V1
      -> mvalue G V2
      -> mstep G (md_pi2 (md_pair V1 V2)) V2

| mstep_unpack1 {G e e' M T}
    : tstep G e e'
      -> mstep G (md_unpack e M T) (md_unpack e' M T)

| mstep_unpack2 {G c v k t M T}
    : tvalue G v
      -> mstep G (md_unpack (tm_pack c v k t) M T) (subst (dot v (dot c id)) M)

| mstep_lett1 {G e1 e1' M2}
    : tstep G e1 e1'
      -> mstep G (md_lett e1 M2) (md_lett e1' M2)

| mstep_lett2 {G v1 M2}
    : tvalue G v1
      -> mstep G (md_lett v1 M2) (subst1 v1 M2)

| mstep_letm1 {G M1 M1' M2 T}
    : mstep G M1 M1'
      -> mstep G (md_letm M1 M2 T) (md_letm M1' M2 T)

| mstep_letm2 {G V M T c}
    : mvalue G V
      -> fstm G V c
      -> mstep G (md_letm V M T) (subst (dot V (dot c id)) M)

| mstep_seal {G M T}
    : mstep G (md_seal M T) M
.


Scheme tstep_mut_ind := Minimality for tstep Sort Prop
with   mstep_mut_ind := Minimality for mstep Sort Prop.
Combined Scheme step_ind from tstep_mut_ind, mstep_mut_ind.


Definition thalt G e :=
  exists v, star (tstep G) e v /\ tvalue G v.


Definition mhalt G M :=
  exists V, star (mstep G) M V /\ mvalue G V.


Hint Constructors tvalue mvalue tstep mstep : dynamic.

