
Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import TSubst.


(* Sorts *)

Inductive fkd : context -> term -> Prop :=
| fkd_type {G}
    : fkd G kd_type

| fkd_sing {G c}
    : fcn G c
      -> fkd G (kd_sing c)

| fkd_pi {G k1 k2}
    : fkd G k1
      -> fkd (G; cl_ucn) k2
      -> fkd G (kd_pi k1 k2)

| fkd_sigma {G k1 k2}
    : fkd G k1
      -> fkd (G; cl_ucn) k2
      -> fkd G (kd_sigma k1 k2)

| fkd_one {G}
    : fkd G kd_one

with fcn : context -> term -> Prop :=
| fcn_var {G i k}
    : index i G (cl_cn k)
      -> fcn G (var i)

| fcn_varu {G i}
    : index i G cl_ucn
      -> fcn G (var i)

| fcn_lam {G k c}
    : fkd G k
      -> fcn (G; cl_ucn) c
      -> fcn G (cn_lam k c)

| fcn_app {G c1 c2}
    : fcn G c1
      -> fcn G c2
      -> fcn G (cn_app c1 c2)

| fcn_pair {G c1 c2}
    : fcn G c1
      -> fcn G c2
      -> fcn G (cn_pair c1 c2)

| fcn_pi1 {G c}
    : fcn G c
      -> fcn G (cn_pi1 c)

| fcn_pi2 {G c}
    : fcn G c
      -> fcn G (cn_pi2 c)

| fcn_star {G}
    : fcn G cn_star

| fcn_unit {G}
    : fcn G cn_unit

| fcn_arrow {G t1 t2}
    : fcn G t1
      -> fcn G t2
      -> fcn G (cn_arrow t1 t2)

| fcn_prod {G t1 t2}
    : fcn G t1
      -> fcn G t2
      -> fcn G (cn_prod t1 t2)

| fcn_all {G k t}
    : fkd G k
      -> fcn (G; cl_ucn) t
      -> fcn G (cn_all k t)

| fcn_exists {G k t}
    : fkd G k
      -> fcn (G; cl_ucn) t
      -> fcn G (cn_exists k t)
.


Scheme fkd_mut_ind := Minimality for fkd Sort Prop
with   fcn_mut_ind := Minimality for fcn Sort Prop.
Combined Scheme static_form_ind from fkd_mut_ind, fcn_mut_ind.


Inductive fsg : context -> term -> Prop :=
| fsg_one {G}
    : fsg G sg_one

| fsg_stat {G k}
    : fkd G k
      -> fsg G (sg_stat k)

| fsg_dyn {G t}
    : fcn G t
      -> fsg G (sg_dyn t)

| fsg_pigen {G T1 T2}
    : fsg G T1
      -> fsg (G; cl_ucn) T2
      -> fsg G (sg_pigen T1 T2)

| fsg_piapp {G T1 T2}
    : fsg G T1
      -> fsg (G; cl_ucn) T2
      -> fsg G (sg_piapp T1 T2)

| fsg_sigma {G T1 T2}
    : fsg G T1
      -> fsg (G; cl_ucn) T2
      -> fsg G (sg_sigma T1 T2)
.




Inductive ftm : context -> term -> Prop :=
| ftm_var {G i t}
    : index i G (cl_tm t)
      -> ftm G (var i)

| ftm_varu {G i}
    : index i G cl_utm
      -> ftm G (var i)

| ftm_star {G}
    : ftm G tm_star

| ftm_lam {G t e}
    : fcn G t
      -> ftm (G; cl_utm) e
      -> ftm G (tm_lam t e)

| ftm_app {G e1 e2}
    : ftm G e1
      -> ftm G e2
      -> ftm G (tm_app e1 e2)

| ftm_pair {G e1 e2}
    : ftm G e1
      -> ftm G e2
      -> ftm G (tm_pair e1 e2)

| ftm_pi1 {G e}
    : ftm G e
      -> ftm G (tm_pi1 e)

| ftm_pi2 {G e}
    : ftm G e
      -> ftm G (tm_pi2 e)

| ftm_plam {G k e}
    : fkd G k
      -> ftm (G; cl_ucn) e
      -> ftm G (tm_plam k e)

| ftm_papp {G e c}
    : ftm G e
      -> fcn G c
      -> ftm G (tm_papp e c)

| ftm_pack {G k t c e}
    : fcn G c
      -> ftm G e
      -> fkd G k
      -> fcn (G; cl_ucn) t
      -> ftm G (tm_pack c e k t)

| ftm_unpack {G e1 e2}
    : ftm G e1
      -> ftm (G; cl_ucn; cl_utm) e2
      -> ftm G (tm_unpack e1 e2)

| ftm_fix {G t e}
    : fcn G t
      -> ftm G e
      -> ftm G (tm_fix t e)

| ftm_lett {G e1 e2}
    : ftm G e1
      -> ftm (G; cl_utm) e2
      -> ftm G (tm_lett e1 e2)

| ftm_letm {G M e}
    : fmd G M
      -> ftm (G; cl_ucn; cl_umd (var 0)) e
      -> ftm G (tm_letm M e)

| ftm_ext {G M}
    : fmd G M
      -> ftm G (tm_ext M)


with fmd : context -> term -> Prop :=
| fmd_var {G i T c}
    : index i G (cl_md T c)
      -> fmd G (var i)

| fmd_varu {G i c}
    : index i G (cl_umd c)
      -> fmd G (var i)

| fmd_star {G}
    : fmd G md_star

| fmd_stat {G c}
    : fcn G c
      -> fmd G (md_stat c)

| fmd_dyn {G e}
    : ftm G e
      -> fmd G (md_dyn e)

| fmd_lamgn {G T M}
    : fsg G T
      -> fmd (G; cl_ucn; cl_umd (var 0)) M
      -> fmd G (md_lamgn T M)

| fmd_appgn {G M1 M2}
    : fmd G M1
      -> fmd G M2
      -> fmd G (md_appgn M1 M2)

| fmd_lamap {G T M}
    : fsg G T
      -> fmd (G; cl_ucn; cl_umd (var 0)) M
      -> fmd G (md_lamap T M)

| fmd_appap {G M1 M2}
    : fmd G M1
      -> fmd G M2
      -> fmd G (md_appap M1 M2)

| fmd_pair {G M1 M2}
    : fmd G M1
      -> fmd G M2
      -> fmd G (md_pair M1 M2)

| fmd_pi1 {G M}
    : fmd G M
      -> fmd G (md_pi1 M)

| fmd_pi2 {G M}
    : fmd G M
      -> fmd G (md_pi2 M)

| fmd_unpack {G e M T}
    : ftm G e
      -> fmd (G; cl_ucn; cl_utm) M
      -> fsg G T
      -> fmd G (md_unpack e M T)

| fmd_lett {G e M}
    : ftm G e
      -> fmd (G; cl_utm) M
      -> fmd G (md_lett e M)

| fmd_letm {G M1 M2 T}
    : fmd G M1
      -> fmd (G; cl_ucn; cl_umd (var 0)) M2
      -> fsg G T
      -> fmd G (md_letm M1 M2 T)

| fmd_seal {G T M}
    : fmd G M
      -> fsg G T
      -> fmd G (md_seal M T)
.


Scheme ftm_mut_ind := Minimality for ftm Sort Prop
with   fmd_mut_ind := Minimality for fmd Sort Prop.
Combined Scheme dynamic_form_ind from ftm_mut_ind, fmd_mut_ind.


Inductive fcl : context -> term -> Prop :=
| fcl_cn {G k}
    : fkd G k
      -> fcl G (cl_cn k)

| fcl_tm {G t}
    : fcn G t
      -> fcl G (cl_tm t)

| fcl_md {G T c}
    : fsg G T
      -> fcn G c
      -> fcl G (cl_md T c)

| fcl_ucn {G}
    : fcl G cl_ucn

| fcl_utm {G}
    : fcl G cl_utm

| fcl_umd {G c}
    : fcn G c
      -> fcl G (cl_umd c)
.


Hint Constructors fkd fcn fsg ftm fmd fcl : formation.


(* Substitution *)

(* substitutends for formation of static terms *)
Inductive sfof : context -> term -> term -> Prop :=
| sfof_cn {G c k}
    : fcn G c
      -> sfof G c (cl_cn k)

| sfof_tm {G e t}
    : sfof G e (cl_tm t)

| sfof_md {G M T c}
    : sfof G M (cl_md T c)

| sfof_ucn {G c}
    : fcn G c
      -> sfof G c cl_ucn

| sfof_utm {G t}
    : sfof G t cl_utm

| sfof_umd {G t c}
    : sfof G t (cl_umd c)
.


(* substitutends for formation of terms *)
Inductive fof : context -> term -> term -> Prop :=
| fof_cn {G c k}
    : fcn G c
      -> fof G c (cl_cn k)

| fof_tm {G e t}
    : ftm G e
      -> fof G e (cl_tm t)

| fof_md {G M T c}
    : fmd G M
      -> fstm G M c
      -> fof G M (cl_md T c)

| fof_ucn {G c}
    : fcn G c
      -> fof G c cl_ucn

| fof_utm {G t}
    : ftm G t
      -> fof G t cl_utm

| fof_umd {G M c}
    : fmd G M
      -> fstm G M c
      -> fof G M (cl_umd c)
.


Lemma sfof_behaved : 
  behaved sfof.
Proof.
intros G i c Hc Hindex.
destruct Hc; simp_sub; eauto using sfof with formation.
Qed.


Lemma fof_behaved : 
  behaved fof.
Proof.
intros G i c Hc Hindex.
destruct Hc; simp_sub; eauto using fof with static formation.
Qed.


Lemma fof_sfof :
  forall G1 s G2, fof G1 s G2 -> sfof G1 s G2.
Proof.
intros G1 s G2 Hof.
invert Hof; intros; subst; eauto using sfof.
Qed.


Lemma fof_tof :
  forall G1 s G2, fof G1 s G2 -> tof G1 s G2.
Proof.
intros G1 s G2 Hof.
invert Hof; intros; subst; eauto using tof.
Qed.


Lemma subof_fof_sfof :
  forall G1 s G2,
    subof fof G1 s G2
    -> subof sfof G1 s G2.
Proof.
eauto using subof_mono, fof_sfof.
Qed.


Lemma subof_wc_bind_sfof_ucn :
  forall G G' s,
    subof (wc sfof) G s G'
    -> subof (wc sfof) (G; cl_ucn) (dot (var 0) (compose s sh1)) (G'; cl_ucn).
Proof.
intros G G' s Hsubof.
replsub (G; cl_ucn) with (G; subst s cl_ucn).
apply subof_wc_bind; auto with static; [].
exact sfof_behaved.
Qed.


Lemma subst_fkd_fcn_wc :
  (forall G1 s G2 k,
     subof (wc sfof) G1 s G2
     -> fkd G2 k
     -> fkd G1 (subst s k))
  /\
  (forall G1 s G2 c,
     subof (wc sfof) G1 s G2
     -> fcn G2 c
     -> fcn G1 (subst s c)).
Proof.
#3 exploit
  (static_form_ind
     (fun G2 k =>
        forall G1 s,
          subof (wc sfof) G1 s G2
          -> fkd G1 (subst s k))
     (fun G2 c =>
        forall G1 s,
          subof (wc sfof) G1 s G2
          -> fcn G1 (subst s c))) as Hind;
try (intros;
     simp_sub;
     eauto using subof_wc_bind_sfof_ucn with formation;
     done).

(* var *)
intros G2 i k Hindex G1 s Hs.
simp_sub.
assert (subof sfof G1 s G2) as Hs'.
  apply (subof_mono (wc sfof)); auto; [].
  intros; apply wc_elim; auto; done.
so (subof_index _ sfof_behaved _#5 iscl_cn Hs' Hindex) as Hsfof.
simp_sub in Hsfof.
invert Hsfof; auto; done.

(* varu *)
intros G2 i Hindex G1 s Hs.
simp_sub.
assert (subof sfof G1 s G2) as Hs'.
  apply (subof_mono (wc sfof)); auto; [].
  intros; apply wc_elim; auto; done.
so (subof_index _ sfof_behaved _#5 iscl_ucn Hs' Hindex) as Hsfof.
simp_sub in Hsfof.
invert Hsfof; auto; done.

(* epilogue *)
destruct Hind.
split; intros; eauto.
Qed.


Lemma subst_fkd_wc :
  forall G1 s G2 k,
    subof (wc sfof) G1 s G2
    -> fkd G2 k
    -> fkd G1 (subst s k).
Proof.
exact (subst_fkd_fcn_wc andel).
Qed.


Lemma subst_fcn_wc :
  forall G1 s G2 c,
    subof (wc sfof) G1 s G2
    -> fcn G2 c
    -> fcn G1 (subst s c).
Proof.
exact (subst_fkd_fcn_wc ander).
Qed.


Lemma subst_sfof_wc :
  forall G1 G2 s m C,
    subof (wc sfof) G1 s G2
    -> sfof G2 m C
    -> sfof G1 (subst s m) (subst s C).
Proof.
intros G1 G2 s t c Hs H.
destruct H; simp_sub; eauto using subst_fcn_wc, sfof.
Qed.


Lemma sfof_wc :
  forall G m C,
    sfof G m C -> wc sfof G m C.
Proof.
exact (wc_intro _ subst_sfof_wc).
Qed.


Lemma subst_fkd :
  forall G1 G2 s k,
    subof sfof G1 s G2
    -> fkd G2 k
    -> fkd G1 (subst s k).
Proof.
intros G1 G2 s k Hs Hof.
so (subst_fkd_fcn_wc andel).
eauto using subof_mono, sfof_wc.
Qed.


Lemma subst_fcn :
  forall G1 G2 s c,
    subof sfof G1 s G2
    -> fcn G2 c
    -> fcn G1 (subst s c).
Proof.
intros G1 G2 s k Hs Hof.
so (subst_fkd_fcn_wc ander).
eauto using subof_mono, sfof_wc.
Qed.


Lemma sfof_wclo :
  wclo sfof.
Proof.
intros G t C C' H.
invert H; intros; subst; simp_sub; eauto using sfof, subst_fcn, weaken1.

(* cn *)
apply sfof_cn; [].
eapply subst_fcn; eauto using weaken1; done.

(* ucn *)
apply sfof_ucn; [].
eapply subst_fcn; eauto using weaken1; done.
Qed.


Lemma subof_bind_sfof_ucn :
  forall G G' s,
    subof sfof G s G'
    -> subof sfof (G; cl_ucn) (dot (var 0) (compose s sh1)) (G'; cl_ucn).
Proof.
intros G G' s Hsubof.
replsub (G; cl_ucn) with (G; subst s cl_ucn).
apply subof_wclo_bind; eauto using sfof_behaved, sfof_wclo with static.
Qed.


Lemma subst_fsg :
  forall G1 G2 s T,
    subof sfof G1 s G2
    -> fsg G2 T
    -> fsg G1 (subst s T).
Proof.
intros G1 G2 s T Hs H.
revert G1 s Hs.
induct H;
intros;
simp_sub;
eauto using subst_fkd, subst_fcn, subof_bind_sfof_ucn with formation.
Qed.


Lemma subst_fcl :
  forall G1 G2 s A,
    subof sfof G1 s G2
    -> fcl G2 A
    -> fcl G1 (subst s A).
Proof.
intros G1 G2 s A Hs H.
invert H; intros; subst; eauto using fcl, subst_fkd, subst_fcn, subst_fsg.
Qed.


Lemma subof_wc_bind_fof_ucn :
  forall G G' s,
    subof (wc fof) G s G'
    -> subof (wc fof) (G; cl_ucn) (dot (var 0) (compose s sh1)) (G'; cl_ucn).
Proof.
intros G G' s Hsubof.
replsub (G; cl_ucn) with (G; subst s cl_ucn).
apply subof_wc_bind; auto with static; [].
exact fof_behaved.
Qed.


Lemma subof_wc_bind_fof_utm :
  forall G G' s,
    subof (wc fof) G s G'
    -> subof (wc fof) (G; cl_utm) (dot (var 0) (compose s sh1)) (G'; cl_utm).
Proof.
intros G G' s Hsubof.
replsub (G; cl_utm) with (G; subst s cl_utm).
apply subof_wc_bind; auto with static; [].
exact fof_behaved.
Qed.


Lemma subof_wc_bind_fof_umd :
  forall G G' s c,
    subof (wc fof) G s G'
    -> subof (wc fof) (G; cl_umd (subst s c)) (dot (var 0) (compose s sh1)) (G'; cl_umd c).
Proof.
intros G G' s c Hsubof.
repl (cl_umd (subst s c)) with (subst s (cl_umd c)) by (simp_sub; reflexivity).
apply subof_wc_bind; auto with static; [].
exact fof_behaved.
Qed.


Lemma subst_ftm_fmd_wc :
  (forall G1 G2 s e,
     subof (wc fof) G1 s G2
     -> ftm G2 e
     -> ftm G1 (subst s e))
  /\
  (forall G1 G2 s M,
     subof (wc fof) G1 s G2
     -> fmd G2 M
     -> fmd G1 (subst s M)).
Proof.
assert (forall G1 G2 s, subof (wc fof) G1 s G2 -> subof sfof G1 s G2) as Hsubfof.
  intros G1 G2 s H.
  refine (subof_mono _#6 H); [].
  intros.
  apply fof_sfof; [].
  apply wc_elim; auto; done.
#12 exploit
  (dynamic_form_ind
     (fun G2 e =>
        forall G1 s,
          subof (wc fof) G1 s G2
          -> ftm G1 (subst s e))
     (fun G2 M =>
        forall G1 s,
          subof (wc fof) G1 s G2
          -> fmd G1 (subst s M))) as Hind;
try (intros;
     simp_sub;
     eauto 6 using subst_fkd, subst_fcn, subst_fsg, subof_wc_bind_fof_ucn, subof_wc_bind_fof_utm with formation;
     done).

(* var *)
intros G2 i t Hindex G1 s Hs.
simp_sub.
assert (subof fof G1 s G2) as Hs'.
  apply (subof_mono (wc fof)); auto; [].
  intros; apply wc_elim; auto; done.
so (subof_index _ fof_behaved _#5 iscl_tm Hs' Hindex) as H.
simp_sub in H.
invert H; auto; done.

(* varu *)
intros G2 i Hindex G1 s Hs.
simp_sub.
assert (subof fof G1 s G2) as Hs'.
  apply (subof_mono (wc fof)); auto; [].
  intros; apply wc_elim; auto; done.
so (subof_index _ fof_behaved _#5 iscl_utm Hs' Hindex) as H.
simp_sub in H.
invert H; auto; done.

(* pack *)
intros; simp_sub.
eapply ftm_pack; eauto using subst_fkd, subst_fcn, subof_bind_sfof_ucn with formation; done.

(* unpack *)
intros G e1 e2 H H0 H1 H2 G1 s Hs.
simp_sub.
so (subof_wc_bind_fof_ucn _#3 Hs) as Hs'.
so (subof_wc_bind_fof_utm _#3 Hs') as Hs''.
simp_sub in Hs''.
eapply ftm_unpack; eauto; done.

(* letm *)
intros G M e H H0 H1 H2 G1 s Hs.
simp_sub.
so (subof_wc_bind_fof_ucn _#3 Hs) as Hs'.
so (subof_wc_bind_fof_umd _#3 (var 0) Hs') as Hs''.
simp_sub in Hs''.
eapply ftm_letm; eauto; done.

(* var *)
intros G2 i T c Hindex G1 s Hs.
simp_sub.
assert (subof fof G1 s G2) as Hs'.
  apply (subof_mono (wc fof)); auto; [].
  intros; apply wc_elim; auto; done.
so (subof_index _ fof_behaved _#5 iscl_md Hs' Hindex) as H.
simp_sub in H.
invert H; auto; done.

(* varu *)
intros G2 i c Hindex G1 s Hs.
simp_sub.
assert (subof fof G1 s G2) as Hs'.
  apply (subof_mono (wc fof)); auto; [].
  intros; apply wc_elim; auto; done.
so (subof_index _ fof_behaved _#5 iscl_umd Hs' Hindex) as H.
simp_sub in H.
invert H; auto; done.

(* lamgn *)
intros G T M H H0 H1 G1 s Hs.
simp_sub.
so (subof_wc_bind_fof_ucn _#3 Hs) as Hs'.
so (subof_wc_bind_fof_umd _#3 (var 0) Hs') as Hs''.
simp_sub in Hs''.
eapply fmd_lamgn; eauto using subst_fsg; done.

(* lamap *)
intros G T M H H0 H1 G1 s Hs.
simp_sub.
so (subof_wc_bind_fof_ucn _#3 Hs) as Hs'.
so (subof_wc_bind_fof_umd _#3 (var 0) Hs') as Hs''.
simp_sub in Hs''.
eapply fmd_lamap; eauto using subst_fsg; done.

(* unpack *)
intros G e M T H H0 H1 H2 H3 G1 s Hs.
simp_sub.
so (subof_wc_bind_fof_ucn _#3 Hs) as Hs'.
so (subof_wc_bind_fof_utm _#3 Hs') as Hs''.
simp_sub in Hs''.
eapply fmd_unpack; eauto using subst_fsg; done.

(* letm *)
intros G M1 M2 T H H0 H1 H2 H3 G1 s Hs.
simp_sub.
so (subof_wc_bind_fof_ucn _#3 Hs) as Hs'.
so (subof_wc_bind_fof_umd _#3 (var 0) Hs') as Hs''.
simp_sub in Hs''.
eapply fmd_letm; eauto using subst_fsg; done.

(* epilogue *)
destruct Hind; split; eauto.
Qed.


Lemma subst_ftm_wc :
  forall G1 G2 s e,
    subof (wc fof) G1 s G2
    -> ftm G2 e
    -> ftm G1 (subst s e).
Proof.
exact (subst_ftm_fmd_wc andel).
Qed.


Lemma subst_fmd_wc :
  forall G1 G2 s M,
    subof (wc fof) G1 s G2
    -> fmd G2 M
    -> fmd G1 (subst s M).
Proof.
exact (subst_ftm_fmd_wc ander).
Qed.


Lemma subst_fof_wc :
  forall G1 G2 s m A,
    subof (wc fof) G1 s G2
    -> fof G2 m A
    -> fof G1 (subst s m) (subst s A).
Proof.
intros G1 G2 s t A Hs Hfof.
assert (subof sfof G1 s G2).
  apply (subof_mono (wc fof)); auto; [].
  intros.
  apply fof_sfof; [].
  eapply wc_elim; eauto; done.
assert (subof (wc tof) G1 s G2).
  eapply subof_mono; [| eexact Hs].
  intros; eauto using fof_tof, wc_mono; done.
destruct Hfof; simp_sub; eauto using subst_fcn, subst_ftm_wc, subst_fmd_wc, subst_fstm_wc, fof.
Qed.


Lemma fof_wc :
  forall G m C,
    fof G m C -> wc fof G m C.
Proof.
exact (wc_intro _ subst_fof_wc).
Qed.


Lemma subst_ftm :
  forall G1 G2 s e,
    subof fof G1 s G2
    -> ftm G2 e
    -> ftm G1 (subst s e).
Proof.
intros G1 G2 s e Hs H.
so (subst_ftm_fmd_wc andel).
eauto using subof_mono, fof_wc.
Qed.


Lemma subst_fmd :
  forall G1 G2 s M,
    subof fof G1 s G2
    -> fmd G2 M
    -> fmd G1 (subst s M).
Proof.
intros G1 G2 s e Hs H.
so (subst_ftm_fmd_wc ander).
eauto using subof_mono, fof_wc.
Qed.


Lemma fof_wclo :
  wclo fof.
Proof.
intros G t C C' H.
#6 invert H; intros; subst; simp_sub; eauto 10 using subst_fcn, weaken1.

(* cn *)
intros; subst; simp_sub.
apply fof_cn; [].
eapply subst_fcn; eauto using weaken1; done.

(* tm *)
intros; subst; simp_sub.
apply fof_tm; [].
eapply subst_ftm; eauto using weaken1; done.

(* md *)
intros; subst; simp_sub.
#2 apply fof_md.
  eapply subst_fmd; eauto using weaken1; done.

  eapply subst_fstm; eauto using weaken1; done.

(* ucn *)
intros; subst; simp_sub.
apply fof_ucn; [].
eapply subst_fcn; eauto using weaken1; done.

(* utm *)
intros; subst; simp_sub.
apply fof_utm; [].
eapply subst_ftm; eauto using weaken1; done.

(* umd *)
intros; subst; simp_sub.
#2 apply fof_umd.
  eapply subst_fmd; eauto using weaken1; done.

  eapply subst_fstm; eauto using weaken1; done.
Qed.


Lemma subof_sfof_ucn_cn :
  forall G G' s k,
    subof sfof G s G'
    -> subof sfof (G; cl_ucn) (dot (var 0) (compose s sh1)) (G'; cl_cn k).
Proof.
intros G G' s k H.
replsub (G; cl_ucn) with (G; subst s cl_ucn).
#2 apply subof_dot.
  apply subof_wclo_weaken1; auto using sfof_wclo; done.

  simp_sub.
  apply sfof_cn; [].
  eapply fcn_varu; eauto with sequence; done.
Qed.


Lemma subof_sfof_cn_ucn :
  forall G G' s k,
    subof sfof G s G'
    -> subof sfof (G; cl_cn (subst s k)) (dot (var 0) (compose s sh1)) (G'; cl_ucn).
Proof.
intros G G' s k H.
replsub (G; cl_cn (subst s k)) with (G; subst s (cl_cn k)).
#2 apply subof_dot.
  apply subof_wclo_weaken1; auto using sfof_wclo; done.

  simp_sub.
  apply sfof_ucn; [].
  eapply fcn_var; eauto with sequence; done.
Qed.


Lemma subof_fof_ucn_cn :
  forall G G' s k,
    subof fof G s G'
    -> subof fof (G; cl_ucn) (dot (var 0) (compose s sh1)) (G'; cl_cn k).
Proof.
intros G G' s k H.
replsub (G; cl_ucn) with (G; subst s cl_ucn).
#2 apply subof_dot.
  apply subof_wclo_weaken1; auto using fof_wclo; done.

  simp_sub.
  apply fof_cn; [].
  eapply fcn_varu; eauto with sequence; done.
Qed.


Lemma subof_fof_cn_ucn :
  forall G G' s k,
    subof fof G s G'
    -> subof fof (G; cl_cn (subst s k)) (dot (var 0) (compose s sh1)) (G'; cl_ucn).
Proof.
intros G G' s k H.
replsub (G; cl_cn (subst s k)) with (G; subst s (cl_cn k)).
#2 apply subof_dot.
  apply subof_wclo_weaken1; auto using fof_wclo; done.

  simp_sub.
  apply fof_ucn; [].
  eapply fcn_var; eauto with sequence; done.
Qed.


Lemma subof_fof_utm_tm :
  forall G G' s t,
    subof fof G s G'
    -> subof fof (G; cl_utm) (dot (var 0) (compose s sh1)) (G'; cl_tm t).
Proof.
intros G G' s t H.
replsub (G; cl_utm) with (G; subst s cl_utm).
#2 apply subof_dot.
  apply subof_wclo_weaken1; auto using fof_wclo; done.

  simp_sub.
  apply fof_tm; [].
  eapply ftm_varu; eauto with sequence; done.
Qed.


Lemma subof_fof_umd_md :
  forall G G' s T c,
    subof fof G s G'
    -> subof fof (G; cl_umd (subst s c)) (dot (var 0) (compose s sh1)) (G'; cl_md T c).
Proof.
intros G G' s T c H.
replsub (G; cl_umd (subst s c)) with (G; subst s (cl_umd c)).
#2 apply subof_dot.
  apply subof_wclo_weaken1; auto using fof_wclo; done.

  simp_sub.
  #2 apply fof_md.
    eapply fmd_varu; eauto with sequence; done.

    rewrite -> subst_compose; [].
    apply fstm_varu; eauto with sequence; done.
Qed.


Lemma subof_fof_md_umd :
  forall G G' s T c,
    subof fof G s G'
    -> subof fof (G; cl_md (subst s T) (subst s c)) (dot (var 0) (compose s sh1)) (G'; cl_umd c).
Proof.
intros G G' s T c H.
replsub (G; cl_md (subst s T) (subst s c)) with (G; subst s (cl_md T c)).
#2 apply subof_dot.
  apply subof_wclo_weaken1; auto using fof_wclo; done.

  simp_sub.
  #2 apply fof_umd.
    eapply fmd_var; eauto with sequence; done.

    rewrite -> subst_compose; [].
    eapply fstm_var; eauto with sequence; done.
Qed.


Lemma subst1_fkd :
  forall G c k,
    fcn G c
    -> fkd (G; cl_ucn) k
    -> fkd G (subst1 c k).
Proof.
intros G c k H1 H2.
eapply subst_fkd; eauto; [].
apply subof_dot; eauto using subof_id; [].
simp_sub.
apply sfof_ucn; auto; done.
Qed.


Lemma subst1_fcn :
  forall G c c',
    fcn G c
    -> fcn (G; cl_ucn) c'
    -> fcn G (subst1 c c').
Proof.
intros G c c' H1 H2.
eapply subst_fcn; eauto; [].
apply subof_dot; eauto using subof_id; [].
simp_sub.
apply sfof_ucn; auto; done.
Qed.


Lemma subst1_fsg :
  forall G c T,
    fcn G c
    -> fsg (G; cl_ucn) T
    -> fsg G (subst1 c T).
Proof.
intros G c c' H1 H2.
eapply subst_fsg; eauto; [].
apply subof_dot; eauto using subof_id; [].
simp_sub.
apply sfof_ucn; auto; done.
Qed.


Lemma unshift_tm_sfof :
  forall G t,
    subof sfof G (dot nonsense id) (G; cl_tm t).
Proof.
intros G t.
apply subof_dot; auto with static; [].
simp_sub.
apply sfof_tm.
Qed.


Lemma unshift_utm_sfof :
  forall G,
    subof sfof G (dot nonsense id) (G; cl_utm).
Proof.
intros G.
apply subof_dot; auto with static; [].
simp_sub.
apply sfof_utm.
Qed.


Lemma unshift_md_sfof :
  forall G T c,
    subof sfof G (dot nonsense id) (G; cl_md T c).
Proof.
intros G T c.
apply subof_dot; auto with static; [].
simp_sub.
apply sfof_md.
Qed.


(* Typing implies formation *)

Lemma static_formation :
  (forall G k,
     ofk G k
     -> fkd G k)
  /\
  (forall G c k,
     ofc G c k
     -> fcn G c).
Proof.
assert (forall G k k', fkd (G; cl_cn k) k' -> fkd (G; cl_ucn) k') as Hkimpl.
  intros G k k' H.
  so (subst_fkd _#4 (subof_sfof_ucn_cn _#4 (subof_id _ _)) H) as H'.
  simp_sub in H'.
  assumption.
assert (forall G k c, fcn (G; cl_cn k) c -> fcn (G; cl_ucn) c) as Hcimpl.
  intros G k c H.
  so (subst_fcn _#4 (subof_sfof_ucn_cn _#4 (subof_id _ _)) H) as H'.
  simp_sub in H'.
  assumption.
#2 exploit
  (static_ind
     (fun G k => fkd G k)
     (fun _ _ _ => True)
     (fun _ _ _ => True)
     (fun G c k => fcn G c)
     (fun _ _ _ _ => True)) as Hind;
try (intros;
     eauto with formation;
     done).

(* extsigma *)
intros G k1 k2 c _ IH _ _ _ _.
invert IH; auto; done.

(* epilogue *)
destruct_all Hind; eauto; done.
Qed.


Lemma ofk_fkd :
  forall G k, ofk G k -> fkd G k.
Proof.
exact (static_formation andel).
Qed.


Lemma ofc_fcn :
  forall G c k, ofc G c k -> fcn G c.
Proof.
exact (static_formation ander).
Qed.


Lemma ofs_fsg :
  forall G T, ofs G T -> fsg G T.
Proof.
assert (forall G k T, fsg (G; cl_cn k) T -> fsg (G; cl_ucn) T) as Himpl.
  intros G k T H.
  so (subst_fsg _#4 (subof_sfof_ucn_cn _#4 (subof_id _ _)) H) as H'.
  simp_sub in H'.
  assumption.
intros G T H.
induct H; eauto using ofk_fkd, ofc_fcn with formation.
Qed.


Inductive fcontext : context -> Prop :=
| fcontext_nil
    : fcontext nil

| fcontext_cn {G k}
    : fcontext G
      -> fkd G k
      -> fcontext (G; cl_cn k)

| fcontext_tm {G t}
    : fcontext G
      -> fcn G t
      -> fcontext (G; cl_tm t)

| fcontext_md {G T c}
    : fcontext G
      -> fsg G T
      -> fcn G c
      -> fcontext (G; cl_md T c)

| fcontext_ucn {G}
    : fcontext G
      -> fcontext (G; cl_ucn)

| fcontext_utm {G}
    : fcontext G
      -> fcontext (G; cl_utm)

| fcontext_umd {G c}
    : fcontext G
      -> fcn G c
      -> fcontext (G; cl_umd c)
.


Hint Constructors fcontext : formation.


Lemma fcontext_tail :
  forall G C, fcontext (G; C) -> fcontext G.
Proof.
intros G C H.
invert H; auto.
Qed.


Lemma static_formation2 :
  (forall G k k', fcontext G -> eqk G k k' -> fkd G k /\ fkd G k')
  /\
  (forall G k k', fcontext G -> subk G k k' -> fkd G k /\ fkd G k')
  /\
  (forall G c k, fcontext G -> ofc G c k -> fkd G k)
  /\
  (forall G c c' k, fcontext G -> eqc G c c' k -> fcn G c /\ fcn G c' /\ fkd G k).
Proof.
assert (forall G k m, fkd (G; cl_cn k) m -> fkd (G; cl_ucn) m) as Himplk.
  intros G m T H.
  so (subst_fkd _#4 (subof_sfof_ucn_cn _#4 (subof_id _ _)) H) as H'.
  simp_sub in H'.
  assumption.
assert (forall G k m, fcn (G; cl_cn k) m -> fcn (G; cl_ucn) m) as Himplc.
  intros G m T H.
  so (subst_fcn _#4 (subof_sfof_ucn_cn _#4 (subof_id _ _)) H) as H'.
  simp_sub in H'.
  assumption.
#3 exploit
  (static_ind
     (fun _ _ => True)
     (fun G k k' => fcontext G -> fkd G k /\ fkd G k')
     (fun G k k' => fcontext G -> fkd G k /\ fkd G k')
     (fun G _ k => fcontext G -> fkd G k)
     (fun G c c' k => fcontext G -> fcn G c /\ fcn G c' /\ fkd G k)) as Hind;
try (trivial;
     intros;
     repeat
       (match goal with
        | H : _ /\ _ |- _ =>
          destruct H
     
        | H : fcontext _ -> _ |- _ =>
          lapply H; [clear H; intro | eauto with formation; done]
     
        | H : fkd _ (kd_pi _ _) |- _ =>
          invert H; []; clear H; intros
     
        | H : fkd _ (kd_sigma _ _) |- _ =>
          invert H; []; clear H; intros
     
        | H : fkd _ (kd_sing _) |- _ =>
          invert H; []; clear H; intros
        end);
     repeat2 split;
     eauto 8 using ofk_fkd, ofc_fcn, subst1_fkd, subst1_fcn with formation;
     done).

(* var *)
intros G i k Hindex HG.
remember (cl_cn k) as C eqn:HeqC.
revert HG HeqC.
#2 induct Hindex.
  intros C G HG ->.
  invert HG; [].
  intros _ Hk.
  so (subst_fkd _#4 (weaken1 _ _ (cl_cn k)) Hk).
  assumption.

  intros i D G C _ IH HG HeqC.
  so (IH (fcontext_tail _ _ HG) HeqC) as H.
  so (subst_fkd _#4 (weaken1 _ _ D) H) as H'.
  simp_sub in H'.
  assumption.

(* extsigma *)
intros G k1 k2 c c' _ IH1 _ _ Hofk _ HG.
destruct (IH1 HG) as (H1 & H2 & Hk1).
invert H1; []; intros.
invert H2; []; intros.
eauto 8 using ofk_fkd with formation; done.

(* epilogue *)
destruct_all Hind; do2 2 split; eauto; done.
Qed.


Lemma eqk_freg :
  forall G k k', fcontext G -> eqk G k k' -> fkd G k /\ fkd G k'.
Proof.
exact (static_formation2 andel).
Qed.


Lemma subk_freg :
  forall G k k', fcontext G -> subk G k k' -> fkd G k /\ fkd G k'.
Proof.
exact (static_formation2 ander andel).
Qed.


Lemma ofc_freg :
  forall G c k, fcontext G -> ofc G c k -> fcn G c /\ fkd G k.
Proof.
intros G c k HG H.
split.
  eapply ofc_fcn; eauto; done.

  eapply (static_formation2 ander ander andel); eauto; done.
Qed.


Lemma eqc_freg :
  forall G c c' k, fcontext G -> eqc G c c' k -> fcn G c /\ fcn G c' /\ fkd G k.
Proof.
exact (static_formation2 ander ander ander).
Qed.


Lemma fsts_fkd :
  forall G T k, fsg G T -> fsts T k -> fkd G k.
Proof.
intros G T k HT Hfst.
revert G HT.
induct Hfst;
intros;
try (let H := findhyp fsg in invert H);
eauto with formation.
Qed.


Lemma fcontext_twin :
  forall G T k,
    fcontext G
    -> fsg G T
    -> fsts T k
    -> fcontext (G; cl_cn k; cl_md (shift1 T) (var 0)).
Proof.
intros G T k HG HT Hfst.
so (fsts_fkd _#3 HT Hfst) as Hk.
#2 apply fcontext_md; eauto with formation.
  eapply subst_fsg; eauto using weaken1; done.

  eapply fcn_var; eauto using index_0; done.
Qed.


Lemma wf_fcontext :
  forall G, wf G -> fcontext G.
Proof.
intros G Hwf.
induct Hwf; eauto using fcontext, ofk_fkd, ofc_fcn, ofs_fsg, fcontext_twin.
Qed.


Lemma fcontext_index :
  forall G i A,
    fcontext G
    -> index i G A
    -> fcl G (subst (sh (S i)) A).
Proof.
intros G i A HG Hindex.
revert HG.
induct Hindex.

(* 0 *)
intros A G HG.
assert (fcl G A) as H.
  invert HG; intros; subst; simp_sub; auto using fcl; done.
eapply subst_fcl; eauto using subof_sh with sequence; done.

(* S *)
intros i B G A _ IH HG.
replsub (subst (sh (S (S i))) A) with (subst sh1 (subst (sh (S i)) A)).
so (IH (fcontext_tail _ _ HG)) as H.
eapply subst_fcl; eauto using subof_sh with sequence; done.
Qed.


Lemma eqs_freg :
  forall G T T', fcontext G -> eqs G T T' -> fsg G T /\ fsg G T'.
Proof.
assert (forall G k m, fsg (G; cl_cn k) m -> fsg (G; cl_ucn) m) as Himpl.
  intros G m T H.
  so (subst_fsg _#4 (subof_sfof_ucn_cn _#4 (subof_id _ _)) H) as H'.
  simp_sub in H'.
  assumption.
intros G T T' HG H.
revert HG.
#2 induct H;
try
  (intros;
   repeat
     (match goal with
      | H : _ /\ _ |- _ =>
        destruct H

      | H : fcontext _ -> _ |- _ =>
        lapply H; [clear H; intro | eauto using fsts_fkd with formation; done]
      end);
   split;
   eauto using ofs_fsg with formation;
   done).

(* stat *)
intros G k k' H HG.
so (eqk_freg _#3 HG H) as (? & ?).
eauto with formation; done.

(* dyn *)
intros G k k' H HG.
so (eqc_freg _#4 HG H) as (? & ? & _).
eauto with formation; done.
Qed.


Lemma subs_freg :
  forall G T T', fcontext G -> subs G T T' -> fsg G T /\ fsg G T'.
Proof.
assert (forall G k m, fsg (G; cl_cn k) m -> fsg (G; cl_ucn) m) as Himpl.
  intros G m T H.
  so (subst_fsg _#4 (subof_sfof_ucn_cn _#4 (subof_id _ _)) H) as H'.
  simp_sub in H'.
  assumption.
intros G T T' HG H.
revert HG.
#2 induct H;
try
  (intros;
   repeat
     (match goal with
      | H : _ /\ _ |- _ =>
        destruct H

      | H : fcontext _ -> _ |- _ =>
        lapply H; [clear H; intro | eauto using fsts_fkd with formation; done]
      end);
   split;
   eauto using ofs_fsg with formation;
   done).

(* refl *)
intros G T T' H HG.
apply eqs_freg; auto; done.

(* stat *)
intros G k k' H HG.
so (subk_freg _#3 HG H) as (? & ?).
eauto with formation; done.
Qed.


Lemma fstm_fcn :
  forall G M c, fcontext G -> fmd G M -> fstm G M c -> fcn G c.
Proof.
intros G M c HG HM Hfst.
revert HG HM.
induct Hfst;
try
  (intros;
   repeat
     (match goal with
      | H : _ /\ _ |- _ =>
        destruct H

      | H : fcontext _ -> _ |- _ =>
        lapply H; [clear H; intro | eauto using fsts_fkd with formation; done]

      | H : fmd _ ?M |- _ =>
        isapp M; invert H; clear H; intros
      end);
    eauto with formation;
    done).

(* var *)
intros G i T c Hindex HG HM.
clear HM.
remember (cl_md T c) as C eqn:HeqC.
revert HG HeqC.
#2 induct Hindex.
  intros C G HG ->.
  invert HG; [].
  intros _ _ H.
  so (subst_fcn _#4 (weaken1 _ _ (cl_md T c)) H).
  assumption.

  intros i D G C _ IH HG HeqC.
  so (IH (fcontext_tail _ _ HG) HeqC) as H.
  so (subst_fcn _#4 (weaken1 _ _ D) H) as H'.
  simp_sub in H'.
  assumption.

(* varu *)
intros G i c Hindex HG HM.
clear HM.
remember (cl_umd c) as C eqn:HeqC.
revert HG HeqC.
#2 induct Hindex.
  intros C G HG ->.
  invert HG; [].
  intros _ H.
  so (subst_fcn _#4 (weaken1 _ _ (cl_umd c)) H).
  assumption.

  intros i D G C _ IH HG HeqC.
  so (IH (fcontext_tail _ _ HG) HeqC) as H.
  so (subst_fcn _#4 (weaken1 _ _ D) H) as H'.
  simp_sub in H'.
  assumption.

(* lam *)
intros G T M k c Hfst _ IH HG Hfmd.
invert Hfmd; [].
intros HT HM.
so (fsts_fkd _#3 HT Hfst) as Hk.
so (fcontext_md (fcontext_cn HG Hk) (subst_fsg _#4 (weaken1 _#3) HT) (fcn_var index_0)) as HG'.
so (subst_fmd _#4 (subof_fof_md_umd _#3 (shift1 T) (var 0) (subof_fof_cn_ucn _#3 k (subof_id _ _))) HM) as HM'.
simp_sub in HM'.
rewrite -> subst_dot01_sh in HM'; [].
so (IH HG' HM') as Hc.
so (subst_fcn _#4 (unshift_md_sfof _#3) Hc) as Hc'.
so (subst_fcn _#4 (subof_sfof_ucn_cn _#4 (subof_id _ _)) Hc') as Hc''.
simp_sub in Hc''.
eauto with formation; done.

(* lett *)
intros G e M c _ IH HG Hfmd.
invert Hfmd; [].
intros He HM.
so (IH (fcontext_utm HG) HM) as Hc.
so (subst_fcn _#4 (unshift_utm_sfof _) Hc) as Hc'.
simp_sub in Hc'.
assumption.
Qed.


Lemma dynamic_formation :
  (forall G e t, fcontext G -> oft G e t -> ftm G e /\ fcn G t)
  /\
  (forall G p M T, fcontext G -> ofm G p M T -> fmd G M /\ fsg G T).
Proof.
exploit
  (dynamic_ind
     (fun G e t => fcontext G -> ftm G e /\ fcn G t)
     (fun G p M T => fcontext G -> fmd G M /\ fsg G T)) as Hind;
try (intros;
     repeat
       (match goal with
        | H : _ /\ _ |- _ =>
          destruct H

        | H : fcontext _ -> _ |- _ =>
          lapply H; [clear H; intro | eauto with formation; done]

        | H : fcn _ ?t |- _ =>
          isapp t; invert H; clear H; intros

        | H : fsg _ ?T |- _ =>
          isapp T; invert H; clear H; intros
        end);
     split;
     eauto using ofc_fcn, subst1_fcn with formation;
     done).

(* var *)
intros G i t Hindex HG.
split; eauto with formation; [].
remember (cl_tm t) as C eqn:HeqC.
revert HG HeqC.
#2 induct Hindex.
  intros C G HG ->.
  invert HG; [].
  intros _ Ht.
  so (subst_fcn _#4 (weaken1 _ _ (cl_tm t)) Ht).
  assumption.

  intros i D G C _ IH HG HeqC.
  so (IH (fcontext_tail _ _ HG) HeqC) as H.
  so (subst_fcn _#4 (weaken1 _ _ D) H) as H'.
  simp_sub in H'.
  assumption.

(* lam *)
intros G t1 t2 e Hof1 _ IH HG.
so (ofc_fcn _#3 Hof1) as Ht1.
so (IH (fcontext_tm HG Ht1)) as (He & Ht2).
so (subst_ftm _#4 (subof_fof_utm_tm _#4 (subof_id _ _)) He) as H.
simp_sub in H.
so (subst_fcn _#4 (unshift_tm_sfof _ _) Ht2) as Ht2'.
simp_sub in Ht2'.
eauto with formation; done.

(* plam *)
intros G k t e Hof1 _ IH HG.
so (ofk_fkd _ _ Hof1) as Hk.
so (IH (fcontext_cn HG (ofk_fkd _ _ Hof1))) as (He & Ht).
so (subst_ftm _#4 (subof_fof_ucn_cn _#4 (subof_id _ _)) He) as He'.
simp_sub in He'.
so (subst_fcn _#4 (subof_sfof_ucn_cn _#4 (subof_id _ _)) Ht) as Ht'.
simp_sub in Ht'.
eauto with formation; done.

(* pack *)
intros G k t c e Hofc _ IH Hoft HG.
so (IH HG) as (Hofe & _).
so (subst_fcn _#4 (subof_sfof_ucn_cn _#4 (subof_id _ _)) (ofc_fcn _#3 Hoft)) as Ht.
simp_sub in Ht.
so (ofc_freg _#3 HG Hofc) as (Hc & Hk).
eauto with formation; done.

(* unpack *)
intros G k t t' e1 e2 _ IH1 _ IH2 Hoft' HG.
so (IH1 HG) as (He1 & H).
invert H; []; intros Hk Ht.
so (subst_fcn _#4 (subof_sfof_cn_ucn _#3 k (subof_id _ _)) Ht) as Ht'.
simp_sub in Ht'.
so (IH2 (fcontext_tm (fcontext_cn HG Hk) Ht')) as (He2 & _).
so (subst_ftm _#4 (subof_fof_utm_tm _#4 (subof_fof_ucn_cn _#4 (subof_id _ _))) He2) as He2'.
simp_sub in He2'.
rewrite -> subst_dot01_sh in He2'; [].
eauto using ofc_fcn with formation; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 HG.
so (IH1 HG) as (He1 & Ht1).
so (IH2 (fcontext_tm HG Ht1)) as (He2 & Ht2).
so (subst_ftm _#4 (subof_fof_utm_tm _#4 (subof_id _ _)) He2) as He2'.
simp_sub in He2'.
so (subst_fcn _#4 (unshift_tm_sfof _ _) Ht2) as Ht2'.
simp_sub in Ht2'.
eauto with formation; done.

(* equiv *)
intros G t t' e _ IH Heqc HG.
so (IH HG) as (? & _).
so (eqc_freg _#4 HG Heqc) as (_ & ? & _).
auto; done.

(* letm *)
intros G T k t M e _ IH1 Hfst _ IH2 Hoft HG.
so (IH1 HG) as (HM & HT).
so (fsts_fkd _#3 HT Hfst) as Hk.
so (fcontext_md (fcontext_cn HG Hk) (subst_fsg _#4 (weaken1 _#3) HT) (fcn_var index_0)) as HG'.
so (IH2 HG') as (He & Ht).
so (subst_ftm _#4 (subof_fof_umd_md _#5 (subof_fof_ucn_cn _#4 (subof_id _ _))) He) as He'.
simp_sub in He'.
rewrite -> subst_dot01_sh in He'; [].
assert (subof sfof G (dot nonsense (dot cn_star id)) (G; cl_cn k; cl_md (shift1 T) (var 0))) as Hsub.
  #2 apply subof_dot.
    apply subof_dot; eauto using subof_id; [].
    simp_sub.
    apply sfof_cn; eauto with formation; done.

    simp_sub.
    apply sfof_md; done.
so (subst_fcn _#4 Hsub Ht) as Ht'.
simp_sub in Ht'.
eauto with formation; done.

(* var *)
intros G i T c Hindex HG.
split; eauto with formation; [].
remember (cl_md T c) as C eqn:HeqC.
revert HG HeqC.
#2 induct Hindex.
  intros C G HG ->.
  invert HG; [].
  intros _ H _.
  so (subst_fsg _#4 (weaken1 _ _ (cl_md T c)) H).
  assumption.

  intros i D G C _ IH HG HeqC.
  so (IH (fcontext_tail _ _ HG) HeqC) as H.
  so (subst_fsg _#4 (weaken1 _ _ D) H) as H'.
  simp_sub in H'.
  assumption.

(* stat *)
intros G c k Hof HG.
so (ofc_freg _#3 HG Hof) as (Hc & Hk).
eauto with formation; done.

(* lamgn *)
intros G T1 T2 k1 M Hof1 Hfst _ IH HG.
so (ofs_fsg _ _ Hof1) as HT1.
so (fsts_fkd _#3 HT1 Hfst) as Hk1.
so (fcontext_md (fcontext_cn HG Hk1) (subst_fsg _#4 (weaken1 _#3) HT1) (fcn_var index_0)) as HG'.
so (IH HG') as (HM & HT2).
so (subst_fmd _#4 (subof_fof_umd_md _#5 (subof_fof_ucn_cn _#4 (subof_id _ _))) HM) as HM'.
simp_sub in HM'.
rewrite -> subst_dot01_sh in HM'; [].
so (subst_fsg _#4 (subof_sfof_ucn_cn _#4 (subof_id _ _)) (subst_fsg _#4 (unshift_md_sfof _#3) HT2)) as HT2'.
simp_sub in HT2'.
eauto with formation; done.

(* appgn *)
intros G T1 T2 M1 M2 c2 _ IH1 _ IH2 Hfst HG.
so (IH1 HG) as (HM1 & H).
invert H; []; intros HT1 HT2.
so (IH2 HG) as (HM2 & _).
so (fstm_fcn _#3 HG HM2 Hfst) as Hc2.
eauto using subst1_fsg with formation; done.

(* lamap *)
intros G T1 T2 k1 M Hof1 Hfst _ IH HG.
so (ofs_fsg _ _ Hof1) as HT1.
so (fsts_fkd _#3 HT1 Hfst) as Hk1.
so (fcontext_md (fcontext_cn HG Hk1) (subst_fsg _#4 (weaken1 _#3) HT1) (fcn_var index_0)) as HG'.
so (IH HG') as (HM & HT2).
so (subst_fmd _#4 (subof_fof_umd_md _#5 (subof_fof_ucn_cn _#4 (subof_id _ _))) HM) as HM'.
simp_sub in HM'.
rewrite -> subst_dot01_sh in HM'; [].
so (subst_fsg _#4 (subof_sfof_ucn_cn _#4 (subof_id _ _)) (subst_fsg _#4 (unshift_md_sfof _#3) HT2)) as HT2'.
simp_sub in HT2'.
eauto with formation; done.

(* appgn *)
intros G p T1 T2 M1 M2 c2 _ IH1 _ IH2 Hfst HG.
so (IH1 HG) as (HM1 & H).
invert H; []; intros HT1 HT2.
so (IH2 HG) as (HM2 & _).
so (fstm_fcn _#3 HG HM2 Hfst) as Hc2.
eauto using subst1_fsg with formation; done.

(* pair *)
intros G p T1 T2 M1 M2 _ IH1 _ IH2 HG.
so (IH1 HG) as (HM1 & HT1).
so (IH2 HG) as (HM2 & HT2).
so (subst_fsg _#4 (weaken1 _ _ cl_ucn) HT2) as HT2'.
eauto with formation; done.

(* pi2 *)
intros G T1 T2 M c _ IH Hfstm HG.
so (IH HG) as (HM & H).
invert H; []; intros HT1 HT2.
so (fstm_fcn _#3 HG HM Hfstm).
eauto using subst1_fsg with formation; done.

(* unpack *)
intros G k t T e M _ IH1 _ IH2 HofT HG.
so (IH1 HG) as (He & H).
invert H; []; intros Hk Ht.
so (subst_fcn _#4 (subof_sfof_cn_ucn _#3 k (subof_id _ _)) Ht) as Ht'.
simp_sub in Ht'.
so (IH2 (fcontext_tm (fcontext_cn HG Hk) Ht')) as (HM & _).
so (subst_fmd _#4 (subof_fof_utm_tm _#4 (subof_fof_ucn_cn _#4 (subof_id _ _))) HM) as HM'.
simp_sub in HM'.
rewrite -> subst_dot01_sh in HM'; [].
eauto using ofs_fsg with formation; done.

(* lett *)
intros G p t T e M _ IH1 _ IH2 HG.
so (IH1 HG) as (He & Ht).
so (IH2 (fcontext_tm HG Ht)) as (HM & HT).
so (subst_fmd _#4 (subof_fof_utm_tm _#4 (subof_id _ _)) HM) as HM'.
simp_sub in HM'.
so (subst_fsg _#4 (unshift_tm_sfof _ _) HT) as HT'.
simp_sub in HT'.
eauto with formation; done.

(* letm *)
intros G T1 T2 k1 M1 M2 _ IH1 Hfst _ IH2 HofT2 HG.
so (IH1 HG) as (HM1 & HT1).
so (fsts_fkd _#3 HT1 Hfst) as Hk1.
so (fcontext_md (fcontext_cn HG Hk1) (subst_fsg _#4 (weaken1 _#3) HT1) (fcn_var index_0)) as HG'.
so (IH2 HG') as (HM2 & _).
so (subst_fmd _#4 (subof_fof_umd_md _#5 (subof_fof_ucn_cn _#4 (subof_id _ _))) HM2) as HM2'.
simp_sub in HM2'.
rewrite -> subst_dot01_sh in HM2'; [].
eauto using ofs_fsg with formation; done.

(* extstat *)
intros G k k' M c _ IH _ Hofc HG.
so (IH HG) as (HM & _).
so (ofc_freg _#3 HG Hofc) as (_ & Hk).
eauto with formation; done.

(* extpi *)
intros G T1 T2 T3 k1 M _ IH1 Hfst _ IH2 HG.
so (IH1 HG) as (HM & H).
invert H; []; intros HT1 _.
so (fsts_fkd _#3 HT1 Hfst) as Hk1.
so (fcontext_md (fcontext_cn HG Hk1) (subst_fsg _#4 (weaken1 _#3) HT1) (fcn_var index_0)) as HG'.
so (IH2 HG') as (_ & HT2).
so (subst_fsg _#4 (subof_sfof_ucn_cn _#4 (subof_id _ _)) (subst_fsg _#4 (unshift_md_sfof _#3) HT2)) as HT2'.
simp_sub in HT2'.
eauto with formation; done.

(* extsigma *)
intros G T1 T2 M _ IH1 _ IH2 HG.
so (IH1 HG) as (H & HT1).
invert H; []; intro HM.
so (IH2 HG) as (_ & HT2).
so (subst_fsg _#4 (weaken1 _ _ cl_ucn) HT2) as HT2'.
eauto with formation; done.

(* subsume *)
intros G p T T' M _ IH Hsubs HG.
so (IH HG) as (HM & _).
so (subs_freg _#3 HG Hsubs) as (_ & HT').
auto; done.
Qed.


Lemma oft_freg :
  forall G e t, fcontext G -> oft G e t -> ftm G e /\ fcn G t.
Proof.
exact (dynamic_formation andel).
Qed.


Lemma ofm_freg :
  forall G p M T, fcontext G -> ofm G p M T -> fmd G M /\ fsg G T.
Proof.
exact (dynamic_formation ander).
Qed.


Lemma oft_ftm :
  forall G e t, fcontext G -> oft G e t -> ftm G e.
Proof.
intros G e t HG Hof.
exact (oft_freg _ _ _ HG Hof andel).
Qed.


Lemma ofm_fmd :
  forall G p M T, fcontext G -> ofm G p M T -> fmd G M.
Proof.
intros G p M T HG Hof.
exact (ofm_freg _ _ _ _ HG Hof andel).
Qed.


(* Closed terms *)

Definition closed (n : nat) (m : term) := subst (under n sh1) m = m.


Lemma subst_closed :
  forall i m s,
    closed i m
    -> subst (under i s) m = m.
Proof.
intros i m s Hclos.
induct s.

(* dot *)
intros m' s IH.
rewrite <- Hclos at 1; [].
rewrite <- subst_compose; [].
rewrite -> dist_compose_under; [].
simp_sub.
assumption.

(* sh *)
intros n.
induct n.
  (* 0 *)
  simp_sub.
  reflexivity.
  
  (* S *)
  intros n IH.
  replace (S n) with (1 + n) by reflexivity.
  rewrite <- (compose_sh_sh 1 n); [].
  rewrite <- dist_compose_under; [].
  rewrite -> subst_compose; [].
  rewrite -> Hclos; [].
  assumption.
Qed.


Lemma subst_closed_0 :
  forall m s,
    closed 0 m
    -> subst s m = m.
Proof.
intros m s Hclos.
so (subst_closed _ _ s Hclos) as H.
simp_sub in H.
assumption.
Qed.


Lemma fkd_fcn_closed :
  (forall G k,
     fkd G k
     -> closed (length G) k)
  /\
  (forall G c,
     fcn G c
     -> closed (length G) c).
Proof.
#2 exploit
  (static_form_ind
     (fun G m => closed (length G) m)
     (fun G m => closed (length G) m)) as Hind; unfold closed;
try (intros; simp_sub; f_equal; rewrite <- ?under_succ; auto; done).

(* var *)
intros G i k Hindex.
apply subst_under_var; [].
eapply index_length; eauto; done.

(* varu *)
intros G i Hindex.
apply subst_under_var; [].
eapply index_length; eauto; done.
Qed.


Lemma fkd_closed :
  forall G k,
    fkd G k
    -> closed (length G) k.
Proof.
exact (fkd_fcn_closed andel).
Qed.


Lemma fcn_closed :
  forall G c,
    fcn G c
    -> closed (length G) c.
Proof.
exact (fkd_fcn_closed ander).
Qed.


Lemma fsg_closed :
  forall G T,
    fsg G T
    -> closed (length G) T.
Proof.
intros G T H.
#2 induct H; unfold closed;
try (intros; simp_sub; f_equal; rewrite <- ?under_succ; eauto; done).

(* stat *)
intros; simp_sub; f_equal; apply fkd_closed; eauto; done.

(* dyn *)
intros; simp_sub; f_equal; apply fcn_closed; eauto; done.
Qed.


Lemma ftm_fmd_closed :
  (forall G e,
     ftm G e
     -> closed (length G) e)
  /\
  (forall G M,
     fmd G M
     -> closed (length G) M).
Proof.
so fkd_closed as Hkd.
so fcn_closed as Hcn.
so fsg_closed as Hsg.
assert (forall G A c, fcn (G; A) c -> closed (S (length G)) c) as Hcn'.
  intros G A c H.
  exact (fcn_closed (G; A) _ H).
unfold closed in Hkd, Hcn, Hcn', Hsg.
assert (forall n s, under (S (S n)) s = dot (var 0) (dot (var 1) (compose (under n s) (sh 2)))) as Hunder2.
  intros n s.
  rewrite -> under_succ; [].
  rewrite -> under_succ; [].
  simp_sub.
  reflexivity.
#4 exploit
  (dynamic_form_ind
     (fun G m => closed (length G) m)
     (fun G m => closed (length G) m)) as Hind; unfold closed;
try (intros; simp_sub; f_equal; rewrite <- ?under_succ; rewrite <- ?Hunder2; eauto; done).

(* var *)
intros G i t Hindex.
apply subst_under_var; [].
eapply index_length; eauto; done.

(* varu *)
intros G i Hindex.
apply subst_under_var; [].
eapply index_length; eauto; done.

(* var *)
intros G i T c Hindex.
apply subst_under_var; [].
eapply index_length; eauto; done.

(* varu *)
intros G i c Hindex.
apply subst_under_var; [].
eapply index_length; eauto; done.
Qed.


Lemma ftm_closed :
  forall G e,
    ftm G e
    -> closed (length G) e.
Proof.
exact (ftm_fmd_closed andel).
Qed.


Lemma fmd_closed :
  forall G M,
    fmd G M
    -> closed (length G) M.
Proof.
exact (ftm_fmd_closed ander).
Qed.


Lemma fcontext_closed :
  forall G, fcontext G -> subst_ctx sh1 G = G.
Proof.
intros G H.
induct H.

(* nil *)
reflexivity.

(* cn *)
intros G k _ IH Hk.
simpl.
f_equal; auto; [].
simp_sub.
f_equal; [].
apply fkd_closed; auto; done.

(* tm *)
intros G t _ IH Hc.
simpl.
f_equal; auto; [].
simp_sub.
f_equal; [].
apply fcn_closed; auto; done.

(* md *)
intros G T c _ IH HT Hc.
simpl.
f_equal; auto; [].
simp_sub.
#2 f_equal.
  apply fsg_closed; auto; done.

  apply fcn_closed; auto; done.

(* ucn *)
intros G _ IH.
simpl.
f_equal; auto; done.

(* utm *)
intros G _ IH.
simpl.
f_equal; auto; done.

(* umd *)
intros G c _ IH Hc.
simpl.
f_equal; auto; [].
simp_sub.
f_equal; [].
apply fcn_closed; auto; done.
Qed.


Lemma fcontext_subst_ctx :
  forall G s, fcontext G -> subst_ctx s G = G.
Proof.
intros G s HG.
so (fcontext_closed _ HG) as Hclos.
clear HG.
induct s.

(* dot *)
intros m' s IH.
rewrite <- Hclos at 1; [].
rewrite <- subst_ctx_compose; [].
simp_sub.
assumption.

(* sh *)
intros n.
induct n.
  (* 0 *)
  simp_sub.
  apply subst_ctx_id; done.
  
  (* S *)
  intros n IH.
  replace (S n) with (1 + n) by reflexivity.
  rewrite <- (compose_sh_sh 1 n); [].
  rewrite -> subst_ctx_compose; [].
  fold sh1.
  rewrite -> Hclos; [].
  assumption.
Qed.


(* Subordination *)

(* statically identical *)
Inductive sid : context -> term -> term -> (* cl *) term -> Prop :=
| sid_cn {G c k}
    : sid G c c (cl_cn k)

| sid_tm {G e e' t}
    : sid G e e' (cl_tm t)

| sid_md {G M M' T c}
    : sid G M M' (cl_md T c)

| sid_ucn {G c}
    : sid G c c cl_ucn

| sid_utm {G t t'}
    : sid G t t' cl_utm

| sid_umd {G M M' c}
    : sid G M M' (cl_umd c)
.

  
Lemma subrel_sid_tm :
  forall G t m m',
    subrel sid G (dot m id) (dot m' id) (G; cl_tm t).
Proof.
intros G t m m'.
apply subrel_dot; eauto using subrel_sh with sequence; [].
simp_sub.
apply sid_tm; done.
Qed.


Lemma subrel_sid_md :
  forall G T c m m',
    subrel sid G (dot m id) (dot m' id) (G; cl_md T c).
Proof.
intros G T c m m'.
apply subrel_dot; eauto using subrel_sh with sequence; [].
simp_sub.
apply sid_md; done.
Qed.


Lemma subrel_sid_bind_tm :
  forall G t m m',
    subrel sid (G; cl_tm t) (dot m sh1) (dot m' sh1) (G; cl_tm t).
Proof.
intros G t m m'.
apply subrel_dot; eauto using subrel_sh with sequence; [].
simp_sub.
apply sid_tm; done.
Qed.


Lemma subrel_sid_bind_md :
  forall G T c m m',
    subrel sid (G; cl_md T c) (dot m sh1) (dot m' sh1) (G; cl_md T c).
Proof.
intros G T c m m'.
apply subrel_dot; eauto using subrel_sh with sequence; [].
simp_sub.
apply sid_md; done.
Qed.


Lemma sid_behaved :
  behavedrel sid.
Proof.
intros G i C HC Hindex.
destruct HC; simp_sub; eauto using sid with static.
Qed.


Lemma sid_wclo :
  wclorel sid.
Proof.
intros G m m' C C' H.
invert H; try (intros; subst; simp_sub; eauto using sid; done).
Qed.


Lemma sid_fkd_fcn :
  (forall G1 s s' G2 k,
     subrel sid G1 s s' G2
     -> fkd G2 k
     -> subst s k = subst s' k)
  /\
  (forall G1 s s' G2 c,
     subrel sid G1 s s' G2
     -> fcn G2 c
     -> subst s c = subst s' c).
Proof.
so (subrel_wclo_bind _ sid_behaved sid_wclo).
exploit
  (static_form_ind
     (fun G2 k =>
        forall G1 s s',
          subrel sid G1 s s' G2
          -> subst s k = subst s' k)
     (fun G2 c =>
        forall G1 s s',
          subrel sid G1 s s' G2
          -> subst s c = subst s' c)) as Hind;
try (intros; eauto; simp_sub; f_equal; eauto using subrel, sid, iscl; done).

(* var *)
intros G i k Hindex G1 s s' Hs.
simp_sub.
so (subrel_index _ sid_behaved _#6 iscl_cn Hs Hindex) as Hsid.
simp_sub in Hsid.
invert Hsid; auto; done.

(* varu *)
intros G i Hindex G1 s s' Hs.
simp_sub.
so (subrel_index _ sid_behaved _#6 iscl_ucn Hs Hindex) as Hsid.
simp_sub in Hsid.
invert Hsid; auto; done.

(* epilogue *)
destruct_all Hind.
split; intros; eauto; done.
Qed.


Lemma sid_fkd :
  forall G1 s s' G2 k,
    subrel sid G1 s s' G2
    -> fkd G2 k
    -> subst s k = subst s' k.
Proof.
exact (sid_fkd_fcn andel).
Qed.


Lemma sid_fcn :
  forall G1 s s' G2 c,
    subrel sid G1 s s' G2
    -> fcn G2 c
    -> subst s c = subst s' c.
Proof.
exact (sid_fkd_fcn ander).
Qed.


Lemma sid_fsg :
  forall G1 s s' G2 T,
    subrel sid G1 s s' G2
    -> fsg G2 T
    -> subst s T = subst s' T.
Proof.
so (subrel_wclo_bind _ sid_behaved sid_wclo).
intros G1 s s' G2 T Hs Hof.
revert G1 s s' Hs.
induct Hof;
try (intros; eauto; simp_sub; f_equal; eauto using subrel, sid, iscl, sid_fkd, sid_fcn; done).
Qed.  


Lemma sid_ofk :
  forall G1 s s' G2 k,
    subrel sid G1 s s' G2
    -> ofk G2 k
    -> subst s k = subst s' k.
Proof.
intros G1 s s' G2 k Hs Hof.
eauto using sid_fkd, ofk_fkd.
Qed.


Lemma sid_ofc :
  forall G1 s s' G2 c k,
    subrel sid G1 s s' G2
    -> ofc G2 c k
    -> subst s c = subst s' c.
Proof.
intros G1 s s' G2 c k Hs Hof.
eauto using sid_fcn, ofc_fcn.
Qed.


Lemma sid_ofs :
  forall G1 s s' G2 T,
    subrel sid G1 s s' G2
    -> ofs G2 T
    -> subst s T = subst s' T.
Proof.
intros G1 s s' G2 T Hs Hof.
eauto using sid_fsg, ofs_fsg.
Qed.  


Lemma subordf_utm_kd :
  forall G k,
    fkd (G; cl_utm) k
    -> k = subst (dot nonsense sh1) k.
Proof.
intros G k H.
rewrite <- subst_dot0_sh at 1.
eapply (sid_fkd (G; cl_utm) _ _ (G; cl_utm)); eauto; [].
apply subrel_dot; eauto using subrel with sequence; [].
simp_sub.
apply sid_utm; done.
Qed.


Lemma subordf_utm_kd' :
  forall G k,
    fkd (G; cl_utm) k
    -> k = shift1 (subst1 nonsense k).
Proof.
intros G k Hk.
simp_sub.
eapply subordf_utm_kd; eauto.
Qed.


Lemma subordf_utm_cn :
  forall G c,
    fcn (G; cl_utm) c
    -> c = subst (dot nonsense sh1) c.
Proof.
intros G c H.
rewrite <- subst_dot0_sh at 1.
eapply (sid_fcn (G; cl_utm) _ _ (G; cl_utm)); eauto; [].
apply subrel_dot; eauto using subrel with sequence; [].
simp_sub.
apply sid_utm; done.
Qed.


Lemma subordf_utm_cn' :
  forall G c,
    fcn (G; cl_utm) c
    -> c = shift1 (subst1 nonsense c).
Proof.
intros G c Hc.
simp_sub.
eapply subordf_utm_cn; eauto.
Qed.


Lemma subordf_utm_sg :
  forall G T,
    fsg (G; cl_utm) T
    -> T = subst (dot nonsense sh1) T.
Proof.
intros G T H.
rewrite <- subst_dot0_sh at 1.
eapply (sid_fsg (G; cl_utm) _ _ (G; cl_utm)); eauto; [].
apply subrel_dot; eauto using subrel with sequence; [].
simp_sub.
apply sid_utm; done.
Qed.


Lemma subordf_utm_sg' :
  forall G T,
    fsg (G; cl_utm) T
    -> T = shift1 (subst1 nonsense T).
Proof.
intros.
simp_sub.
eapply subordf_utm_sg; eauto.
Qed.


Lemma subordf_umd_kd :
  forall G d k,
    fkd (G; cl_umd d) k
    -> k = subst (dot nonsense sh1) k.
Proof.
intros G d k H.
rewrite <- subst_dot0_sh at 1.
eapply (sid_fkd (G; cl_umd d) _ _ (G; cl_umd d)); eauto; [].
apply subrel_dot; eauto using subrel with sequence; [].
simp_sub.
apply sid_umd; done.
Qed.


Lemma subordf_umd_cn :
  forall G d c,
    fcn (G; cl_umd d) c
    -> c = subst (dot nonsense sh1) c.
Proof.
intros G d c H.
rewrite <- subst_dot0_sh at 1.
eapply (sid_fcn (G; cl_umd d) _ _ (G; cl_umd d)); eauto; [].
apply subrel_dot; eauto using subrel with sequence; [].
simp_sub.
apply sid_umd; done.
Qed.


Lemma subordf_umd_sg :
  forall G d T,
    fsg (G; cl_umd d) T
    -> T = subst (dot nonsense sh1) T.
Proof.
intros G d T H.
rewrite <- subst_dot0_sh at 1.
eapply (sid_fsg (G; cl_umd d) _ _ (G; cl_umd d)); eauto; [].
apply subrel_dot; eauto using subrel with sequence; [].
simp_sub.
apply sid_umd; done.
Qed.


Lemma subordf_tm_kd :
  forall G t k,
    fkd (G; cl_tm t) k
    -> k = subst (dot nonsense sh1) k.
Proof.
intros G t k H.
so (subst_fkd _#4 (subof_fof_sfof _#3 (subof_fof_utm_tm _#4 (subof_id _ _))) H) as H'.
simp_sub in H'.
eapply subordf_utm_kd; eauto.
Qed.


Lemma subordf_tm_kd' :
  forall G t k,
    fkd (G; cl_tm t) k
    -> k = shift1 (subst1 nonsense k).
Proof.
intros G k Hk.
simp_sub.
eapply subordf_tm_kd; eauto.
Qed.


Lemma subordf_tm_cn :
  forall G t c,
    fcn (G; cl_tm t) c
    -> c = subst (dot nonsense sh1) c.
Proof.
intros G t c H.
so (subst_fcn _#4 (subof_fof_sfof _#3 (subof_fof_utm_tm _#4 (subof_id _ _))) H) as H'.
simp_sub in H'.
eapply subordf_utm_cn; eauto.
Qed.


Lemma subordf_tm_cn' :
  forall G t c,
    fcn (G; cl_tm t) c
    -> c = shift1 (subst1 nonsense c).
Proof.
intros G c Hc.
simp_sub.
eapply subordf_tm_cn; eauto.
Qed.


Lemma subordf_tm_sg :
  forall G t T,
    fsg (G; cl_tm t) T
    -> T = subst (dot nonsense sh1) T.
Proof.
intros G t c H.
so (subst_fsg _#4 (subof_fof_sfof _#3 (subof_fof_utm_tm _#4 (subof_id _ _))) H) as H'.
simp_sub in H'.
eapply subordf_utm_sg; eauto.
Qed.


Lemma subordf_tm_sg' :
  forall G t T,
    fsg (G; cl_tm t) T
    -> T = shift1 (subst1 nonsense T).
Proof.
intros G t T H.
simp_sub.
eapply subordf_tm_sg; eauto.
Qed.


Lemma subordf_md_kd :
  forall G T d k,
    fkd (G; cl_md T d) k
    -> k = subst (dot nonsense sh1) k.
Proof.
intros G T d c H.
so (subst_fkd _#4 (subof_fof_sfof _#3 (subof_fof_umd_md _#5 (subof_id _ _))) H) as H'.
simp_sub in H'.
eapply subordf_umd_kd; eauto.
Qed.


Lemma subordf_md_kd' :
  forall G T d k,
    fkd (G; cl_md T d) k
    -> k = shift1 (subst1 nonsense k).
Proof.
intros G T d k H.
simp_sub.
eapply subordf_md_kd; eauto.
Qed.


Lemma subordf_md_cn :
  forall G T d c,
    fcn (G; cl_md T d) c
    -> c = subst (dot nonsense sh1) c.
Proof.
intros G T d c H.
so (subst_fcn _#4 (subof_fof_sfof _#3 (subof_fof_umd_md _#5 (subof_id _ _))) H) as H'.
simp_sub in H'.
eapply subordf_umd_cn; eauto.
Qed.


Lemma subordf_md_cn' :
  forall G T d c,
    fcn (G; cl_md T d) c
    -> c = shift1 (subst1 nonsense c).
Proof.
intros G T d c H.
simp_sub.
eapply subordf_md_cn; eauto.
Qed.


Lemma subordf_md_sg :
  forall G T d U,
    fsg (G; cl_md T d) U
    -> U = subst (dot nonsense sh1) U.
Proof.
intros G T d U H.
so (subst_fsg _#4 (subof_fof_sfof _#3 (subof_fof_umd_md _#5 (subof_id _ _))) H) as H'.
simp_sub in H'.
eapply subordf_umd_sg; eauto.
Qed.


Lemma subord_tm_cn :
  forall G t c k,
    ofc (G; cl_tm t) c k
    -> c = subst (dot nonsense sh1) c.
Proof.
intros G t c k Hof.
eapply subordf_tm_cn; eauto using ofc_fcn.
Qed.


Lemma subord_tm_sg :
  forall G t T,
    ofs (G; cl_tm t) T
    -> T = subst (dot nonsense sh1) T.
Proof.
intros G t T Hof.
eapply subordf_tm_sg; eauto using ofs_fsg.
Qed.


Lemma subord_md_cn :
  forall G T d c k,
    ofc (G; cl_md T d) c k
    -> c = subst (dot nonsense sh1) c.
Proof.
intros G T d c k Hof.
eapply subordf_md_cn; eauto using ofc_fcn.
Qed.


Lemma subord_md_sg :
  forall G T d U,
    ofs (G; cl_md T d) U
    -> U = subst (dot nonsense sh1) U.
Proof.
intros G T d U Hof.
eapply subordf_md_sg; eauto using ofs_fsg.
Qed.


(* Weaker forms of formation that don't require a context. *)

Inductive hkd : term -> Prop :=
| hkd_type
    : hkd kd_type

| hkd_sing {c}
    : hkd (kd_sing c)

| hkd_pi {k1 k2}
    : hkd k1
      -> hkd k2
      -> hkd (kd_pi k1 k2)

| hkd_sigma {k1 k2}
    : hkd k1
      -> hkd k2
      -> hkd (kd_sigma k1 k2)

| hkd_one
    : hkd kd_one
.


Inductive hsg : term -> Prop :=
| hsg_one
    : hsg sg_one

| hsg_stat {k}
    : hkd k
      -> hsg (sg_stat k)

| hsg_dyn {t}
    : hsg (sg_dyn t)

| hsg_pigen {T1 T2}
    : hsg T1
      -> hsg T2
      -> hsg (sg_pigen T1 T2)

| hsd_piapp {T1 T2}
    : hsg T1
      -> hsg T2
      -> hsg (sg_piapp T1 T2)

| hsd_sigma {T1 T2}
    : hsg T1
      -> hsg T2
      -> hsg (sg_sigma T1 T2)
.


Lemma fkd_hkd :
  forall G k, fkd G k -> hkd k.
Proof.
intros G k H.
induct H; eauto using hkd.
Qed.


Lemma fsg_hsg :
  forall G T, fsg G T -> hsg T.
Proof.
intros G T H.
induct H; eauto using hsg, fkd_hkd.
Qed.


Lemma ofk_hkd :
  forall G k, ofk G k -> hkd k.
Proof.
intros G k Hof.
eauto using ofk_fkd, fkd_hkd.
Qed.


Lemma eqk_hkd :
  forall G k k',
    eqk G k k'
    -> hkd k /\ hkd k'.
Proof.
intros G k k' Heqk.
induct Heqk.

(* refl *)
intros G k H.
split; eapply ofk_hkd; eauto; done.

(* symm *)
intros G k k' _ (? & ?).
split; auto; done.

(* trans *)
intros G k1 k2 k3 _ (? & _) _ (_ & ?).
split; auto; done.

(* sing *)
intros G c c' H.
auto using hkd; done.

(* pi *)
intros G k1 k1' k2 k2' _ (? & ?) _ (? & ?).
auto using hkd; done.

(* sigma *)
intros G k1 k1' k2 k2' _ (? & ?) _ (? & ?).
auto using hkd; done.
Qed.


Lemma subk_hkd :
  forall G k k',
    subk G k k'
    -> hkd k /\ hkd k'.
Proof.
intros G k k' Hsubk.
induct Hsubk.

(* refl *)
intros G k k' Heqk.
eapply eqk_hkd; eauto; done.

(* trans *)
intros G k1 k2 k3 _ (? & _) _ (_ & ?).
split; auto; done.

(* sing_t *)
intros G c H.
auto using hkd; done.

(* pi *)
intros G k1 k1' k2 k2' _ (? & ?) _ (? & ?) _.
auto using hkd; done.

(* sigma *)
intros G k1 k1' k2 k2' _ (? & ?) _ (? & ?) _.
auto using hkd; done.
Qed.


Lemma subst_hkd :
  forall s k, hkd k -> hkd (subst s k).
Proof.
intros s k H.
revert s.
induct H; intros; simp_sub; auto using hkd.
Qed.


Lemma subst1_hkd :
  forall c k, hkd k -> hkd (subst1 c k).
Proof.
unfold subst1.
auto using subst_hkd.
Qed.


Lemma subst_hsg :
  forall s T, hsg T -> hsg (subst s T).
Proof.
intros s T H.
revert s.
induct H; intros; simp_sub; auto using hsg, subst_hkd.
Qed.


Lemma shift1_hsg :
  forall T, hsg T -> hsg (shift1 T).
Proof.
intros T H.
unfold shift1.
apply subst_hsg; auto.
Qed.


Lemma ofs_hsg :
  forall G T, ofs G T -> hsg T.
Proof.
intros G k Hof.
eauto using ofs_fsg, fsg_hsg.
Qed.


Lemma eqs_hsg :
  forall G T T',
    eqs G T T'
    -> hsg T /\ hsg T'.
Proof.
intros G T T' H.
induct H;
intros;
repeat (match goal with
        | H : _ /\ _ |- _ => destruct H
        | H : eqk _ _ _ |- _ => so (eqk_hkd _#3 H); clear H
        end);
split; eauto using hsg, ofs_hsg.
Qed.


Lemma subs_hsg :
  forall G T T',
    subs G T T'
    -> hsg T /\ hsg T'.
Proof.
intros G T T' H.
induct H;
intros;
repeat (match goal with
        | H : _ /\ _ |- _ => destruct H
        | H : subk _ _ _ |- _ => so (subk_hkd _#3 H); clear H
        | H : eqs _ _ _ |- _ => so (eqs_hsg _#3 H); clear H
        end);
split; eauto using hsg.
Qed.




(* TODO: Clean up self_respecting definition.  Clean up fundamental_theorem. *)
