
Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import Formation.
Require Import TSubst.
Require Import Determinism.
Require Import Regularity.
Require Import Inversion.
Require Import Preservation.
Require Import Congruence.
Require Import Contextual.
Require Import Divergence.
Require Import Compactness.
Require Import Closure.
Require Import Candidate.
Require Import Logical.
Require Import ModelRegular.
Require Import ModelRegularSg.
Require Import Biorthogonal.
Require Import Monad.
Require Inversion.
Require Progress.
Require Extensionality.


Inductive fintctx : context -> env -> Prop :=
| fintctx_nil
    : fintctx nil nil

| fintctx_cn {G k E c c' Q}
    : fintctx G E
      -> intks k E c c' Q
      -> fintctx (G; cl_cn k) (E; Sta c c' Q)

| fintctx_tm {G t E v v'}
    : fintctx G E
      -> intt t E v v'
      -> fintctx (G; cl_tm t) (E; Dyn v v')

| fintctx_md {G k T E c c' Q V V'}
    : fintctx G E
      -> ints T E V V' Q
      -> fsts T k
      -> fstm nil V c
      -> fstm nil V' c'
      -> fintctx (G; cl_cn k; cl_md (shift1 T) (var 0)) (E; Sta c c' Q; Dyn V V')
.


Definition leqt (G : context) (e e' : (* tm *) term) (t : (* tp *) term) :=
  wf G
  -> (oft G e t /\ oft G e' t
      /\
      (forall E,
         fintctx G E
         -> tev (subst (left E) t) (subst (right E) t) (intt t E) (subst (left E) e) (subst (right E) e'))).


Definition leqm (G : context) (p : purity) (M M' : (* md *) term) (T : (* sg *) term) :=
  wf G
  -> (ofm G p M T /\ ofm G p M' T
      /\
      (match p with
       | pure =>
           (forall k c c',
              fsts T k
              -> fstm G M c
              -> fstm G M' c'
              -> eqc G c c' k)
           /\
           (forall E c,
              fintctx G E
              -> fstm G M c
              -> mpuev (subst (left E) T) (subst (right E) T) (ints T E) (subst (left E) M) (subst (right E) M') (intc c E))
    
       | impure =>
           (forall E,
              fintctx G E
              -> miev (subst (left E) T) (subst (right E) T) (ints T E) (subst (left E) M) (subst (right E) M'))
       end)).


Lemma fintctx_stat :
  forall G E,
    wf G
    -> fintctx G E
    -> intctx G E E.
Proof.
intros G E Hwf H.
revert Hwf.
induct H.

(* nil *)
intros _.
apply intctx_nil; done.

(* cn *)
intros G k E c c' Q _ IH Hc Hwf.
simpl.
invert Hwf; [].
intros Hwf' Hofk.
so (intk_reg _#6 Hc) as (Hof & Hof').
apply intctx_cn; auto with static; done.

(* tm *)
intros G t E v v' _ IH Hv Hwf.
invert Hwf; [].
intros Hwf' _.
simpl.
apply intctx_tm; auto; done.

(* md *)
intros G k T E c c' Q V V' _ IH HintV Hfst HfstV HfstV' Hwf.
invert Hwf; [].
intros Ta Hwf' HofT _ Heq.
f_apply (fun x => subst1 nonsense x) in Heq as Heq'.
simp_sub in Heq'.
subst Ta; clear Heq.
simpl.
apply intctx_md; [].
so (IH Hwf') as HintE.
so (ints_fst _#9 Hwf' HofT HintE Hfst HfstV HfstV' HintV) as Hintc.
so (intk_reg _#6 Hintc) as (Hof & Hof').
apply intctx_cn; auto with static; done.
Qed.


Lemma fintctx_fenv :
  forall G E,
    fintctx G E -> fenv G E.
Proof.
intros G E H.
induct H; auto using fenv.
Qed.


Lemma left_subof_aof :
  forall G E,
    wf G
    -> fintctx G E
    -> subof aof nil (left E) G.
Proof.
intros G E Hwf H.
revert Hwf.
induct H.

(* nil *)
intros _.
simpl.
apply subof_id; done.

(* cn *)
intros G k E c c' Q HintE IH Hint Hwf.
invert Hwf; [].
intros Hwf' Hofk.
simpl.
apply subof_dot; auto; [].
simp_sub.
apply aof_cn; [].
so (intk_reg _#6 Hint) as (Hof & Hof').
assumption.

(* tm *)
intros G t E v v' HintE IH Hintv Hwf.
invert Hwf; [].
intros Hwf' Hoft.
simpl.
so (intt_tvrel _#3 Hwf' Hoft (fintctx_stat _ _ Hwf' HintE) _ _ Hintv) as (Hofv & _ & Hval & _).
apply subof_dot; auto; [].
simp_sub.
apply aof_tm; auto; done.

(* md *)
intros G k T E c c' Q V V' HintE IH HintV Hfst HfstV HfstV' Hwf.
invert Hwf; [].
intros T' Hwf' HofT _ Heq.
f_apply (fun T => subst1 nonsense T) in Heq as Heq'.
simp_sub in Heq'.
subst T'; clear Heq.
simpl.
so (ofs_fst _#3 HofT Hfst) as Hofk.
so (ints_reg _#5 HintV) as (HofV & _).
so (ofm_fst _#6 wf_nil HofV HfstV (subst_fsts _ _ (left E) Hfst)) as Hofc.
#2 apply subof_dot.
  apply subof_dot; auto; [].
  simp_sub.
  apply aof_cn; [].
  assumption.

  simp_sub.
  apply aof_md; auto; [].
  exact (ints_val _#6 Hwf' HofT (fintctx_stat _ _ Hwf' HintE) HintV andel).
Qed.


Lemma right_subof_aof :
  forall G E,
    wf G
    -> fintctx G E
    -> subof aof nil (right E) G.
Proof.
intros G E Hwf H.
revert Hwf.
induct H.

(* nil *)
intros _.
simpl.
apply subof_id; done.

(* cn *)
intros G k E c c' Q HintE IH Hint Hwf.
invert Hwf; [].
intros Hwf' Hofk.
simpl.
apply subof_dot; auto; [].
simp_sub.
apply aof_cn; [].
so (intk_reg _#6 Hint) as (Hof & Hof').
assumption.

(* tm *)
intros G t E v v' HintE IH Hintv Hwf.
invert Hwf; [].
intros Hwf' Hoft.
simpl.
so (intt_tvrel _#3 Hwf' Hoft (fintctx_stat _ _ Hwf' HintE) _ _ Hintv) as (_ & Hofv & _ & Hval).
apply subof_dot; auto; [].
simp_sub.
apply aof_tm; auto; done.

(* md *)
intros G k T E c c' Q V V' HintE IH HintV Hfst HfstV HfstV' Hwf.
invert Hwf; [].
intros T' Hwf' HofT _ Heq.
f_apply (fun T => subst1 nonsense T) in Heq as Heq'.
simp_sub in Heq'.
subst T'; clear Heq.
simpl.
so (ofs_fst _#3 HofT Hfst) as Hofk.
so (ints_reg _#5 HintV) as (_ & HofV').
so (ofm_fst _#6 wf_nil HofV' HfstV' (subst_fsts _ _ (right E) Hfst)) as Hofc.
#2 apply subof_dot.
  apply subof_dot; auto; [].
  simp_sub.
  apply aof_cn; [].
  assumption.

  simp_sub.
  apply aof_md; auto; [].
  exact (ints_val _#6 Hwf' HofT (fintctx_stat _ _ Hwf' HintE) HintV ander).
Qed.


Lemma left_subof_dof :
  forall G E,
    wf G
    -> fintctx G E
    -> subof dof nil (left E) G.
Proof.
intros; apply subof_aof_dof; auto using left_subof_aof.
Qed.


Lemma right_subof_dof :
  forall G E,
    wf G
    -> fintctx G E
    -> subof dof nil (right E) G.
Proof.
intros; apply subof_aof_dof; auto using right_subof_aof.
Qed.


Lemma left_subof_oof :
  forall G E,
    wf G
    -> fintctx G E
    -> subof oof nil (left E) G.
Proof.
intros; apply subof_aof_oof; auto using left_subof_aof.
Qed.


Lemma right_subof_oof :
  forall G E,
    wf G
    -> fintctx G E
    -> subof oof nil (right E) G.
Proof.
intros; apply subof_aof_oof; auto using right_subof_aof.
Qed.


Lemma fintctx_tail :
  forall G A E ent,
    wf (G; A)
    -> fintctx (G; A) (E; ent)
    -> fintctx G E.
Proof.
intros G A E ent Hwf H.
invert H; clear H; try (intros; subst; auto; done); [].
intros G' k T E' c c' Q V V' HintE' HintV Hfst HfstV HfstV' <- <- <- <-.
invert Hwf; [].
intros T' Hwf' Hofs _ Heq.
f_apply (fun z => subst1 nonsense z) in Heq as Heq'.
simp_sub in Heq'; subst T'; clear Heq.
apply fintctx_cn; auto; [].
eapply (ints_fst G'); eauto using fintctx_stat; done.
Qed.


Lemma leqt_reg :
  forall G e e' t,
    wf G
    -> leqt G e e' t
    -> oft G e t /\ oft G e' t.
Proof.
intros G e e' t Hwf H.
destruct (H Hwf) as (H1 & H2 & _).
auto; done.
Qed.


Lemma leqm_reg :
  forall G p M M' T,
    wf G
    -> leqm G p M M' T
    -> ofm G p M T /\ ofm G p M' T.
Proof.
intros G p M M' T Hwf H.
destruct (H Hwf) as (H1 & H2 & _).
auto; done.
Qed.


Lemma leqm_fst :
  forall G M M' T,
    wf G
    -> leqm G pure M M' T
    -> exists k c c',
         fsts T k
         /\ fstm G M c
         /\ fstm G M' c'
         /\ eqc G c c' k.
Proof.
intros G M M' T Hwf H.
so (H Hwf) as (Hof & Hof' & Hrel).
so (can_fsts _ _ (ofm_reg _#4 Hwf Hof)) as (k & Hfst).
so (can_fstm _#3 Hwf Hof) as (c & Hfstm).
so (can_fstm _#3 Hwf Hof') as (c' & Hfstm').
exists k, c, c'.
so (Hrel andel _ _ _ Hfst Hfstm Hfstm') as Heqc.
auto; done.
Qed.


Lemma leqm_fst' :
  forall G M M' T k c c',
    wf G
    -> leqm G pure M M' T
    -> fsts T k
    -> fstm G M c
    -> fstm G M' c'
    -> eqc G c c' k.
Proof.
intros G M M' T k c c' Hwf H Hfst Hfstm Hfstm'.
exact (H Hwf anderrl _#3 Hfst Hfstm Hfstm').
Qed.


Lemma leq_cregular :
  cregular leqt leqm.
Proof.
do2 2 split; auto using leqt_reg, leqm_reg, leqm_fst.
Qed.


Lemma tcc_formation' :
  forall Rt Rm G e e' t,
    cregular Rt Rm
    -> wf G
    -> tcc samek samec sames Rt Rm G e e' t
    -> oft G e t /\ oft G e' t.
Proof.
intros Rt Rm G e e' t Hreg Hwf Hcc.
eapply tcc_formation; eauto; [].
#3 refine (tcc_mono _#10 _#5 _#4 Hcc); auto.
  intros G' k k' H.
  invert H; intros; subst; auto with static; done.

  intros G' c c' k H.
  invert H; intros; subst; auto with static; done.

  intros G' T T' H.
  invert H; intros; subst; auto with static; done.
Qed.


Lemma mcc_formation' :
  forall Rt Rm G p M M' T,
    cregular Rt Rm
    -> wf G
    -> mcc samek samec sames Rt Rm G p M M' T
    -> ofm G p M T /\ ofm G p M' T
       /\ (p = pure ->
             exists k c c',
               fsts T k
               /\ fstm G M c
               /\ fstm G M' c'
               /\ eqc G c c' k).
Proof.
intros Rt Rm G p M M' T Hreg Hwf Hcc.
eapply mcc_formation; eauto; [].
#3 refine (mcc_mono _#10 _#5 _#5 Hcc); auto.
  intros G' k k' H.
  invert H; intros; subst; auto with static; done.

  intros G' c c' k H.
  invert H; intros; subst; auto with static; done.

  intros G' ? ? H.
  invert H; intros; subst; auto with static; done.
Qed.


Lemma leqt_comp :
  forall G e e' t,
    tcc samek samec sames leqt leqm G e e' t
    -> leqt G e e' t.
Proof.
intros G e e' t Hcc Hwf.
so (tcc_formation' _#6 leq_cregular Hwf Hcc) as (Hof & Hof').
do2 2 split; auto; [].
clear Hof Hof'.
revert Hwf.
induct Hcc.

(* lam *)
intros G t1 t1' t2 e e' Hsame IH Hwf E HintE.
invert Hsame; clear Hsame; [].
intros Hoft1 <-.
so (leqt_reg _#4 (wf_tm Hwf Hoft1) IH) as (Hofe & Hofe').
so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 (wf_tm Hwf Hoft1) Hofe)) as Hoft2.
simp_sub in Hoft2.
assert (oft nil (subst (left E) (tm_lam t1 e)) (subst (left E) (cn_arrow t1 t2))) as Hoflaml.
  eapply subst_oft; eauto using left_subof_dof with static; done.
assert (oft nil (subst (right E) (tm_lam t1 e')) (subst (right E) (cn_arrow t1 t2))) as Hoflamr.
  eapply subst_oft; eauto using right_subof_dof with static; done.
apply tev_intro; auto; try (simp_sub; auto with dynamic; done); [].
rewrite -> intt_arrow_unroll; [].
apply arrow_interp_; auto; try (simp_sub; auto with dynamic; done); [].
intros w w' Hintw.
so (intt_tvrel _#3 Hwf Hoft1 (fintctx_stat _ _ Hwf HintE) _ _ Hintw) as (Hofw & Hofw' & Hvalw & Hvalw').
#5 apply (anti_tsteps_tev _#5 (subst (dot w (left E)) e) (subst (dot w' (right E)) e')).
  simp_sub; simp_sub in Hoflaml.
  eapply oft_app; eauto; done.

  simp_sub; simp_sub in Hoflamr.
  eapply oft_app; eauto; done.

  simp_sub.
  #2 eapply star_step.
    apply tstep_app3; auto; done.

    simp_sub.
    apply star_refl; done.

  simp_sub.
  #2 eapply star_step.
    apply tstep_app3; auto; done.

    simp_sub.
    apply star_refl; done.

  #3 exploit (fun H => IH H anderr (E; Dyn w w')) as Hintapp.
    apply wf_tm; auto; done.

    apply fintctx_tm; auto; done.

    simpl in Hintapp.
    simp_sub in Hintapp.
    unfold shift1 in Hintapp.
    force_exact Hintapp.
    f_equal; [].
    eapply subst_intt; eauto using ofc_fcn, fintctx_fenv, subenv with sequence; done.

(* app *)
intros G t1 t2 e1 e1' e2 e2' IH1 IH2 Hwf E HintE.
so (leqt_reg _#4 Hwf IH1) as (Hofe1 & Hofe1').
so (leqt_reg _#4 Hwf IH2) as (Hofe2 & Hofe2').
so (IH1 Hwf anderr _ HintE) as Hint1.
so (IH2 Hwf anderr _ HintE) as Hint2.
clear IH1 IH2.
simp_sub.
simp_sub in Hint1.
eapply (monad_tev_tev _#6 (fun z => tm_app z (subst (left E) e2)) (fun z => tm_app z (subst (right E) e2'))); eauto using subst_oft, left_subof_dof, right_subof_dof with static dynamic; [].
intros v1 v1' Hintv1.
rewrite -> intt_arrow_unroll in Hintv1; [].
invert Hintv1; [].
intros Hofv1 Hofv1' Hvalv1 Hvalv1' Hintapp.
simp_sub in Hofv1.
simp_sub in Hofv1'.
eapply (monad_tev_tev _#6 (fun z => tm_app v1 z) (fun z => tm_app v1' z)); eauto using subst_oft, left_subof_dof, right_subof_dof with static dynamic; done.

(* pair *)
intros G t1 t2 e1 e1' e2 e2' IH1 IH2 Hwf E HintE.
so (leqt_reg _#4 Hwf IH1) as (Hofe1 & Hofe1').
so (leqt_reg _#4 Hwf IH2) as (Hofe2 & Hofe2').
so (IH1 Hwf anderr _ HintE) as Hint1.
so (IH2 Hwf anderr _ HintE) as Hint2.
clear IH1 IH2.
simp_sub.
eapply (monad_tev_tev _#6 (fun z => tm_pair z (subst (left E) e2)) (fun z => tm_pair z (subst (right E) e2'))); eauto using subst_oft, left_subof_dof, right_subof_dof with static dynamic; [].
intros v1 v1' Hintv1.
so (intt_tvrel _#3 Hwf (oft_reg _#3 Hwf Hofe1) (fintctx_stat _ _ Hwf HintE) _ _ Hintv1) as (Hofv1 & Hofv1' & Hvalv1 & Hvalv1').
eapply (monad_tev_tev _#6 (fun z => tm_pair v1 z) (fun z => tm_pair v1' z)); eauto using subst_oft, left_subof_dof, right_subof_dof with static dynamic; [].
intros v2 v2' Hintv2.
so (intt_tvrel _#3 Hwf (oft_reg _#3 Hwf Hofe2) (fintctx_stat _ _ Hwf HintE) _ _ Hintv2) as (Hofv2 & Hofv2' & Hvalv2 & Hvalv2').
apply tev_intro; auto with static dynamic; [].
rewrite -> intt_prod_unroll; [].
apply prod_interp_; intros; simp_sub; auto with static dynamic; done.

(* pi1 *)
intros G t1 t2 e e' IH Hwf E HintE.
so (leqt_reg _#4 Hwf IH) as (Hofe & Hofe').
so (IH Hwf anderr _ HintE) as Hint.
clear IH.
simp_sub in Hint.
simp_sub.
eapply (monad_tev_tev _#6 tm_pi1 tm_pi1); eauto using subst_oft, left_subof_dof, right_subof_dof with static dynamic; [].
intros v v' Hintv.
rewrite -> intt_prod_unroll in Hintv; [].
invert Hintv; [].
intros v1 v1' v2 v2' Hofv Hofv' Hvalv Hvalv' Hintv1 Hintv2 <- <-.
invert Hvalv; [].
intros Hvalv1 Hvalv2.
invert Hvalv'; [].
intros Hvalv1' Hvalv2'.
simp_sub in Hofv; simp_sub in Hofv'.
so (pair_oft_inversion_prod _#5 wf_nil Hofv) as (Hofv1 & Hofv2).
so (pair_oft_inversion_prod _#5 wf_nil Hofv') as (Hofv1' & Hofv2').
eapply tev_intro_steps; eauto using star_one with static dynamic; done.

(* pi2 *)
intros G t1 t2 e e' IH Hwf E HintE.
so (leqt_reg _#4 Hwf IH) as (Hofe & Hofe').
so (IH Hwf anderr _ HintE) as Hint.
clear IH.
simp_sub in Hint.
simp_sub.
eapply (monad_tev_tev _#6 tm_pi2 tm_pi2); eauto using subst_oft, left_subof_dof, right_subof_dof with static dynamic; [].
intros v v' Hintv.
rewrite -> intt_prod_unroll in Hintv; [].
invert Hintv; [].
intros v1 v1' v2 v2' Hofv Hofv' Hvalv Hvalv' Hintv1 Hintv2 <- <-.
invert Hvalv; [].
intros Hvalv1 Hvalv2.
invert Hvalv'; [].
intros Hvalv1' Hvalv2'.
simp_sub in Hofv; simp_sub in Hofv'.
so (pair_oft_inversion_prod _#5 wf_nil Hofv) as (Hofv1 & Hofv2).
so (pair_oft_inversion_prod _#5 wf_nil Hofv') as (Hofv1' & Hofv2').
eapply tev_intro_steps; eauto using star_one with static dynamic; done.

(* plam *)
intros G k k' t e e' Hsame IH Hwf E HintE.
invert Hsame; clear Hsame; [].
intros Hofk <-.
so (leqt_reg _#4 (wf_cn Hwf Hofk) IH) as (Hofe & Hofe').
simp_sub.
so (wf_cn Hwf Hofk) as Hwf'.
so (oft_reg _#3 Hwf' Hofe) as Hoft.
assert (oft nil (subst (left E) (tm_plam k e)) (subst (left E) (cn_all k t))) as Hoflaml.
  eapply subst_oft; eauto using left_subof_dof with static; done.
assert (oft nil (subst (right E) (tm_plam k e')) (subst (right E) (cn_all k t))) as Hoflamr.
  eapply subst_oft; eauto using right_subof_dof with static; done.
simp_sub in Hoflaml; simp_sub in Hoflamr.
apply tev_intro; auto with dynamic; [].
rewrite -> intt_all_unroll; [].
apply all_interp_; auto with dynamic; try (simp_sub; auto with static; done); [].
intros c c' Q Hintc.
so (intk_reg _#6 Hintc) as (Hofc & Hofc').
assert (fintctx (G; cl_cn k) (E; Sta c c' Q)) as HintE'.
  apply fintctx_cn; auto; done.
#5 apply (tpclosed_tev_ceqt _ _ _ (subst (dot c (left E)) e) (subst (dot c' (right E)) e')).
  exact (intt_tvrel _#3 Hwf' Hoft (fintctx_stat _ _ Hwf' HintE')).

  eapply intt_tpclosed; eauto using fintctx_stat; done.

  apply ceqt_symm; [].
  #2 apply tstep_ceqt.
    replsub (subst (dot c (left E)) t) with (subst1 c (subst (dot (var 0) (compose (left E) sh1)) t)).
    eapply oft_papp; eauto; done.

    #2 eapply rel_quest.
      apply tstep_papp2; auto; done.
      
      simp_sub.
      reflexivity.

  apply ceqt_symm; [].
  #2 apply tstep_ceqt.
    replsub (subst (dot c' (right E)) t) with (subst1 c' (subst (dot (var 0) (compose (right E) sh1)) t)).
    eapply oft_papp; eauto; done.

    #2 eapply rel_quest.
      apply tstep_papp2; auto; done.
      
      simp_sub.
      reflexivity.

  exploit (fun H => IH H anderr (E; Sta c c' Q)); auto; done.

(* papp *)
intros G k t e e' c c' IH Hsame Hwf E HintE.
invert Hsame; clear Hsame; [].
intros Hofc <-.
so (leqt_reg _#4 Hwf IH) as (Hofe & Hofe').
so (IH Hwf anderr _ HintE) as Hint.
clear IH.
simp_sub.
simp_sub in Hint.
#3 eapply (monad_tev_tev _#6 (fun z => tm_papp z (subst (left E) c)) (fun z => tm_papp z (subst (right E) c))); eauto with dynamic.
  intros f Hof.
  replsub (subst (dot (subst (left E) c) (left E)) t) with (subst1 (subst (left E) c) (subst (dot (var 0) (compose (left E) sh1)) t)).
  eapply oft_papp; eauto using subst_ofc, left_subof, fintctx_stat; done.

  intros f Hof.
  replsub (subst (dot (subst (right E) c) (right E)) t) with (subst1 (subst (right E) c) (subst (dot (var 0) (compose (right E) sh1)) t)).
  eapply oft_papp; eauto using subst_ofc, right_subof, fintctx_stat; done.

  intros v v' Hintv.
  rewrite -> intt_all_unroll in Hintv; [].
  invert Hintv; [].
  intros Hofv Hofv' Hvalv Hvalv' Hintapp.
  unfold subst1.
  assert (subenv E (dot c id) (E; Sta (subst (left E) c) (subst (right E) c) (intc c E))) as Hsubenv.
    eapply subenv_dot; eauto using ofc_fcn, fintctx_fenv, subenv_sh with sequence; done.
  so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hofe)) as (l & Hofcr).
  invert Hofcr; [].
  intros _ Hoft _; clear l Hofcr.
  rewrite -> (subst_intt _#5 (ofc_fcn _#3 Hoft) (fenv_cn (fintctx_fenv _ _ HintE)) Hsubenv); [].
  apply Hintapp; [].
  eapply intc_fun; eauto using fintctx_stat; done.

(* pack *)
intros G k k' t t' c c' e e' Hsamec Hsamek IH Hsamet Hwf E HintE.
invert Hsamec; invert Hsamek; invert Hsamet; clear Hsamec Hsamek Hsamet; [].
intros Hoft <- Hofk <- Hofc <-.
so (leqt_reg _#4 Hwf IH) as (Hofe & Hofe').
so (IH Hwf anderr _ HintE) as Hint.
clear IH.
simp_sub.
#3 eapply (monad_tev_tev _#6 (fun z => tm_pack (subst (left E) c) z (subst (left E) k) (subst (dot (var 0) (compose (left E) sh1)) t)) (fun z => tm_pack (subst (right E) c) z (subst (right E) k) (subst (dot (var 0) (compose (right E) sh1)) t))); eauto with dynamic.
  intros f Hof.
  apply oft_pack; eauto using subst_ofc, subst_ofc_type, subof_bind_sof_cn, left_subof, fintctx_stat; [].
  simp_sub.
  simp_sub in Hof.
  assumption.

  intros f Hof.
  apply oft_pack; eauto using subst_ofc, subst_ofc_type, subof_bind_sof_cn, right_subof, fintctx_stat; [].
  simp_sub.
  simp_sub in Hof.
  assumption.

  intros v v' Hintv.
  so (intt_tvrel _#3 Hwf (oft_reg _#3 Hwf Hofe) (fintctx_stat _ _ Hwf HintE) _ _ Hintv) as (Hofv & Hofv' & Hvalv & Hvalv').
  replsub (subst (left E) (subst1 c t)) with (subst1 (subst (left E) c) (subst (dot (var 0) (compose (left E) sh1)) t)) in Hofv.
  replsub (subst (right E) (subst1 c t)) with (subst1 (subst (right E) c) (subst (dot (var 0) (compose (right E) sh1)) t)) in Hofv'.
  #3 apply tev_intro; eauto with dynamic.
    apply oft_pack; eauto using subst_ofc, subst_ofc_type, subof_bind_sof_cn, left_subof, fintctx_stat with static dynamic; done.

    apply oft_pack; eauto using subst_ofc, subst_ofc_type, subof_bind_sof_cn, right_subof, fintctx_stat with static dynamic; done.

    rewrite -> intt_exists_unroll; [].
    unfold exists_interp_st.
    simp_sub.
    apply tpst_expand; auto using exists_interp_tvrel; [].
    #4 eapply exists_interp_; eauto with dynamic.
      simp_sub.
      apply oft_pack; eauto using subst_ofc, subst_ofc_type, subof_bind_sof_cn, left_subof, fintctx_stat; done.

      simp_sub.
      apply oft_pack; eauto using subst_ofc, subst_ofc_type, subof_bind_sof_cn, right_subof, fintctx_stat; done.

      eapply intc_fun; eauto using fintctx_stat; done.

      assert (subenv E (dot c id) (E; Sta (subst (left E) c) (subst (right E) c) (intc c E))) as Hsubenv.
        eapply subenv_dot; eauto using ofc_fcn, fintctx_fenv, subenv_sh with sequence; done.
      unfold subst1 in Hintv.
      rewrite -> (subst_intt _#5 (ofc_fcn _#3 Hoft) (fenv_cn (fintctx_fenv _ _ HintE)) Hsubenv) in Hintv; [].
      assumption.

(* unpack *)
intros G k t t' e1 e1' e2 e2' IH1 IH2 Hoft' Hwf E HintE.
so (leqt_reg _#4 Hwf IH1) as (Hofe1 & Hofe1').
so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hofe1)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _; clear l Hofcr.
so (leqt_reg _#4 (wf_tm (wf_cn Hwf Hofk) Hoft) IH2) as (Hofe2 & Hofe2').
so (IH1 Hwf anderr _ HintE) as Hint1.
clear IH1.
simp_sub.
simp_sub in Hint1.
rewrite -> intt_exists_unroll in Hint1; [].
unfold exists_interp_st in Hint1.
simp_sub in Hint1.
assert (closed 2 (subst (under 2 (left E)) e2)) as Hcle2.
  so (ftm_closed _ _ (oft_ftm _#3 fcontext_nil (subst_oft _#5 (left_subof_dof _ _ Hwf HintE) (oft_unpack Hofe1 Hofe2 Hoft')))) as H.
  simpl in H.
  unfold closed in H.
  simp_sub in H.
  unfold closed.
  simp_sub.
  injection H; auto; done.
assert (closed 2 (subst (under 2 (right E)) e2')) as Hcle2'.
  so (ftm_closed _ _ (oft_ftm _#3 fcontext_nil (subst_oft _#5 (right_subof_dof _ _ Hwf HintE) (oft_unpack Hofe1' Hofe2' Hoft')))) as H.
  simpl in H.
  unfold closed in H.
  simp_sub in H.
  unfold closed.
  simp_sub.
  injection H; auto; done.
#8 eapply (monad_tpstev_tev_tpclosed _#6 (fun z => tm_unpack z _) (fun z => tm_unpack z _)); eauto with dynamic.
  apply exists_interp_tvrel; done.

  eapply intt_tvrel; eauto using fintctx_stat; done.

  eapply intt_tpclosed; eauto using fintctx_stat; done.

  intros s e.
  simp_sub.
  f_equal; [].
  so (subst_closed _ _ s Hcle2) as H.
  simp_sub in H.
  exact H.

  intros s e.
  simp_sub.
  f_equal; [].
  so (subst_closed _ _ s Hcle2') as H.
  simp_sub in H.
  exact H.

  intros G' i e Heqi Hof.
  assert (subof dof G' (compose (left E) (sh i)) G) as Hsubof.
    eapply subof_dof_trans; eauto using left_subof_dof; [].
    subst i; auto using subof_sh, truncate_all; done.
  simp_sub.
  simp_sub in Hof.
  #3 eapply oft_unpack.
    exact Hof.

    simp_sub.
    so (subst_closed _ _ (sh i) Hcle2) as Heq.
    simp_sub in Heq.
    rewrite <- Heq; [].
    so (subst_oft _#5 (subof_bind_dof_cn_tm _#5 Hsubof) Hofe2) as H.
    simp_sub in H.
    exact H.

    eapply subst_ofc_type; eauto using subof_dof_sof; done.

  intros G' i e Heqi Hof.
  assert (subof dof G' (compose (right E) (sh i)) G) as Hsubof.
    eapply subof_dof_trans; eauto using right_subof_dof; [].
    subst i; auto using subof_sh, truncate_all; done.
  simp_sub.
  simp_sub in Hof.
  #3 eapply oft_unpack.
    exact Hof.

    simp_sub.
    so (subst_closed _ _ (sh i) Hcle2') as Heq.
    simp_sub in Heq.
    rewrite <- Heq; [].
    so (subst_oft _#5 (subof_bind_dof_cn_tm _#5 Hsubof) Hofe2') as H.
    simp_sub in H.
    exact H.

    eapply subst_ofc_type; eauto using subof_dof_sof; done.

  intros w w' Hintw.
  simp_sub.
  invert Hintw; [].
  intros u u' c c' l l' s s' Q Hofw Hofw' Hvalw Hvalw' Hintc Hintv <- <-.
  clear Hintw.
  so (intt_tvrel _#3 (wf_cn Hwf Hofk) Hoft (fintctx_stat _ _ (wf_cn Hwf Hofk) (fintctx_cn HintE Hintc)) _ _ Hintv) as (_ & _ & Hvalu & Hvalu').
  exploit (fun H => IH2 H anderr (E; Sta c c' Q; Dyn u u')) as Hinte; auto using wf, fintctx; [].
  simpl in Hinte; simp_sub in Hinte.
  rewrite -> (subst_intt _ _ E _ _ (ofc_fcn _#3 Hoft') (fintctx_fenv _ _ HintE)) in Hinte; auto using subenv_sh with sequence; [].
  #4 eapply anti_tsteps_tev; eauto.
    simp_sub in Hofw.
    #2 eapply oft_unpack; eauto.
      simp_sub.
      replsub (subst (compose (left E) (sh 2)) t') with (subst (dot (var 0) (dot (var 1) (compose (left E) (sh 2)))) (subst (sh 2) t')).
      eapply subst_oft; eauto; [].
      apply subof_bind_dof_cn_tm; auto using left_subof_dof; done.

      eapply subst_ofc_type; eauto using left_subof, fintctx_stat; done.

    simp_sub in Hofw.
    #2 eapply oft_unpack; eauto.
      simp_sub.
      replsub (subst (compose (right E) (sh 2)) t') with (subst (dot (var 0) (dot (var 1) (compose (right E) (sh 2)))) (subst (sh 2) t')).
      eapply subst_oft; eauto; [].
      apply subof_bind_dof_cn_tm; auto using right_subof_dof; done.

      eapply subst_ofc_type; eauto using right_subof, fintctx_stat; done.

    #2 eapply star_step.
      apply tstep_unpack2; auto; done.

      simp_sub.
      apply star_refl; done.

    #2 eapply star_step.
      apply tstep_unpack2; auto; done.

      simp_sub.
      apply star_refl; done.

(* fix *)
intros G t t' e e' Hsame IH Hwf E HintE.
invert Hsame; clear Hsame; [].
intros Hoft <-.
so (leqt_reg _#4 Hwf IH) as (Hofe & Hofe').
so (IH Hwf anderr _ HintE) as Hinte; clear IH.
cut (intt (cn_arrow cn_unit t) E (subst (left E) (fixv t e)) (subst (right E) (fixv t e'))).
  clear Hinte.
  intro Hintf.
  unfold fixv in Hintf.
  simp_sub in Hintf.
  simp_sub.
  rewrite -> intt_arrow_unroll in Hintf; [].
  invert Hintf; clear Hintf; [].
  intros _ _ _ _ Hintapp.
  #2 exploit (Hintapp tm_star tm_star) as Hint; clear Hintapp.
    rewrite -> intt_unit_unroll; [].
    apply unit_interp_; done.

    #2 eapply tsteps_tev; eauto.
      #2 eapply star_step.
        eapply tstep_app3; eauto with dynamic; done.

        simp_sub.
        apply star_refl; done.

      #2 eapply star_step.
        eapply tstep_app3; eauto with dynamic; done.

        simp_sub.
        apply star_refl; done.
repl (subst (left E) (fixv t e)) with (subst1 (fixv (subst (left E) t) (subst (left E) e)) (var 0)) by (unfold fixv; simp_sub; reflexivity).
repl (subst (right E) (fixv t e')) with (subst1 (fixv (subst (right E) t) (subst (right E) e')) (var 0)) by (unfold fixv; simp_sub; reflexivity).
#8 apply (tpclosed_admissible (cn_arrow cn_unit (subst (left E) t)) (cn_arrow cn_unit (subst (right E) t))).
  eapply intt_tpclosed; eauto using fintctx_stat with static; done.

  so (subst_oft _#5 (left_subof_dof _ _ Hwf HintE) Hofe) as H.
  simp_sub in H.
  assumption.

  so (subst_oft _#5 (right_subof_dof _ _ Hwf HintE) Hofe') as H.
  simp_sub in H.
  assumption.

  apply oft_var; auto with sequence; done.

  apply oft_var; auto with sequence; done.

  eapply tvalue_var; eauto with sequence; done.

  eapply tvalue_var; eauto with sequence; done.

  assert (forall j, oft nil (fixnv j (subst (left E) t) (subst (left E) e)) (subst (left E) (cn_arrow cn_unit t))) as Hofleft.
    intro j.
    #2 apply oft_fixnv.
      eapply subst_ofc_type; eauto using left_subof, fintctx_stat; done.

      so (subst_oft _#5 (left_subof_dof _ _ Hwf HintE) Hofe) as H.
      simp_sub in H.
      assumption.
  assert (forall j, oft nil (fixnv j (subst (right E) t) (subst (right E) e')) (subst (right E) (cn_arrow cn_unit t))) as Hofright.
    intro j.
    #2 apply oft_fixnv.
      eapply subst_ofc_type; eauto using right_subof, fintctx_stat; done.

      so (subst_oft _#5 (right_subof_dof _ _ Hwf HintE) Hofe') as H.
      simp_sub in H.
      assumption.
  intro j.
  simp_sub.
  induct j.
    (* 0 *)
    rewrite -> intt_arrow_unroll; [].
    #3 apply arrow_interp_; auto.
      unfold fixnv; apply tvalue_lam; done.

      unfold fixnv; apply tvalue_lam; done.

      intros w w' Hintw.
      invert Hintw; [].
      intros <- <-.
      #5 apply (anti_tsteps_tev _#5 (bottom (subst (left E) t)) (bottom (subst (right E) t))).
        so (Hofleft 0) as H.
        simp_sub in H.
        eapply oft_app; eauto with static; done.

        so (Hofleft 0) as H.
        simp_sub in H.
        eapply oft_app; eauto with static; done.

        unfold fixnv, fixn.
        #2 eapply star_step.
          apply tstep_app3; auto with dynamic; done.

          simp_sub.
          apply star_refl; done.

        unfold fixnv, fixn.
        #2 eapply star_step.
          apply tstep_app3; auto with dynamic; done.

          simp_sub.
          apply star_refl; done.

        #2 do2 3 split; eauto using oft_bottom, subst_ofc_type, left_subof, right_subof, fintctx_stat.
          split.
            intro Hhalt.
            exfalso.
            eapply diverge_halt_contra; eauto using bottom_diverge; done.
          
            intro Hhalt.
            exfalso.
            eapply diverge_halt_contra; eauto using bottom_diverge; done.
          
          intros v v' Hsteps _ Hval _.
          exfalso.
          assert (thalt nil (bottom (subst (left E) t))) as H by (eexists; eauto).
          eapply diverge_halt_contra; eauto using bottom_diverge; done.

    (* S *)
    intros j IH.
    rewrite -> intt_arrow_unroll; [].
    #3 apply arrow_interp_; auto.
      unfold fixnv; apply tvalue_lam; done.

      unfold fixnv; apply tvalue_lam; done.

      intros w w' Hintw.
      invert Hintw; [].
      intros <- <-.
      #5 apply (anti_tsteps_tev _#5 (tm_app (subst (left E) e) (fixnv j (subst (left E) t) (subst (left E) e))) (tm_app (subst (right E) e') (fixnv j (subst (right E) t) (subst (right E) e')))).
        so (Hofleft (S j)) as H.
        simp_sub in H.
        eapply oft_app; eauto with static; done.

        so (Hofleft (S j)) as H.
        simp_sub in H.
        eapply oft_app; eauto with static; done.

        unfold fixnv.
        #2 eapply star_step.
          apply tstep_app3; auto with dynamic; done.

          simp_sub.
          simpl fixn.
          apply star_refl; done.

        unfold fixnv.
        #2 eapply star_step.
          apply tstep_app3; auto with dynamic; done.

          simp_sub.
          simpl fixn.
          apply star_refl; done.

        #3 eapply (monad_tev_tev _#6 (fun z => tm_app z (fixnv j (subst (left E) t) (subst (left E) e))) (fun z => tm_app z (fixnv j (subst (right E) t) (subst (right E) e')))); eauto with dynamic.
          intros f Hof.
          simp_sub in Hof.
          eapply oft_app; eauto; done.

          intros f Hof.
          simp_sub in Hof.
          eapply oft_app; eauto; done.

          intros f f' Hintf.
          rewrite -> intt_arrow_unroll in Hintf; [].
          invert Hintf; [].
          intros _ _ _ _ Hintapp.
          eapply Hintapp; eauto; done.

(* lett *)
intros G t1 t2 e1 e1' e2 e2' IH1 IH2 Hwf E HintE.
so (leqt_reg _#4 Hwf IH1) as (Hofe1 & Hofe1').
so (oft_reg _#3 Hwf Hofe1) as Hoft1.
so (leqt_reg _#4 (wf_tm Hwf Hoft1) IH2) as (Hofe2 & Hofe2').
so (oft_reg _#3 Hwf (oft_lett Hofe1 Hofe2)) as Hoft2.
so (IH1 Hwf anderr _ HintE) as Hint1; clear IH1.
simp_sub.
#3 eapply (monad_tev_tev _#6 (fun z => tm_lett z (subst (dot (var 0) (compose (left E) sh1)) e2)) (fun z => tm_lett z (subst (dot (var 0) (compose (right E) sh1)) e2'))); eauto with dynamic.
  intros e Hofe.
  eapply oft_lett; eauto; [].
  so (subst_oft _#5 (subof_bind_dof_tm _#4 (left_subof_dof _ _ Hwf HintE)) Hofe2) as H.
  simp_sub in H.
  simp_sub.
  assumption.

  intros e Hofe.
  eapply oft_lett; eauto; [].
  so (subst_oft _#5 (subof_bind_dof_tm _#4 (right_subof_dof _ _ Hwf HintE)) Hofe2') as H.
  simp_sub in H.
  simp_sub.
  assumption.

  intros v1 v1' Hintv1.
  clear e1 Hofe1 Hint1.
  so (intt_tvrel _#3 Hwf Hoft1 (fintctx_stat _ _ Hwf HintE) _ _ Hintv1) as (Hofv1 & Hofv1' & Hvalv1 & Hvalv1').
  #5 eapply (anti_tsteps_tev _#5 (subst (dot v1 (left E)) e2) (subst (dot v1' (right E)) e2')).
    eapply oft_lett; eauto; [].
    so (subst_oft _#5 (subof_bind_dof_tm _#4 (left_subof_dof _ _ Hwf HintE)) Hofe2) as H.
    simp_sub in H.
    simp_sub.
    assumption.

    eapply oft_lett; eauto; [].
    so (subst_oft _#5 (subof_bind_dof_tm _#4 (right_subof_dof _ _ Hwf HintE)) Hofe2') as H.
    simp_sub in H.
    simp_sub.
    assumption.

    #2 eapply star_step.
      apply tstep_lett2; auto; done.

      simp_sub.
      apply star_refl; done.

    #2 eapply star_step.
      apply tstep_lett2; auto; done.

      simp_sub.
      apply star_refl; done.

    #3 exploit (fun H => IH2 H anderr (E; Dyn v1 v1')) as Hint2.
      apply wf_tm; auto; done.

      apply fintctx_tm; auto; done.

      simpl in Hint2.
      simp_sub in Hint2.
      force_exact Hint2.
      f_equal; [].
      eapply subst_intt; eauto using ofc_fcn, fintctx_fenv, subenv with sequence; done.

(* equiv *)
intros G t1 t2 e e' IH Heqc Hwf E HintE.
so (leqt_reg _#4 Hwf IH) as (Hofe & Hofe').
so (IH Hwf anderr _ HintE) as Hint.
rewrite <- (eqc_intt _#5 Hwf Heqc (fintctx_stat _ _ Hwf HintE)); [].
rewrite <- (tev_eqc _#4 (intt t1 E) (subst_eqc_type _#5 (left_subof _#3 (fintctx_stat _ _ Hwf HintE)) Heqc) (subst_eqc_type _#5 (right_subof _#3 (fintctx_stat _ _ Hwf HintE)) Heqc)); [].
assumption.

(* letm *)
intros G T k t M M' e e' IH1 Hfst IH2 Hoft Hwf E HintE.
so (leqm_reg _#5 Hwf IH1) as (HofM & HofM').
so (ofm_reg _#4 Hwf HofM) as HofT.
so (leqt_reg _#4 (wf_md Hwf HofT Hfst) IH2) as (Hofe & Hofe').
so (IH1 Hwf anderr _ HintE) as HintM; clear IH1.
simp_sub.
unfold miev, mi in HintM.
assert (closed 2 (subst (under 2 (left E)) e)) as Hcle.
  so (ftm_closed _ _ (oft_ftm _#3 fcontext_nil (subst_oft _#5 (left_subof_dof _ _ Hwf HintE) (oft_letm HofM Hfst Hofe Hoft)))) as H.
  simpl in H.
  unfold closed in H.
  simp_sub in H.
  unfold closed.
  simp_sub.
  injection H; auto; done.
assert (closed 2 (subst (under 2 (right E)) e')) as Hcle'.
  so (ftm_closed _ _ (oft_ftm _#3 fcontext_nil (subst_oft _#5 (right_subof_dof _ _ Hwf HintE) (oft_letm HofM' Hfst Hofe' Hoft)))) as H.
  simpl in H.
  unfold closed in H.
  simp_sub in H.
  unfold closed.
  simp_sub.
  injection H; auto; done.
#8 eapply (monad_mpstev_tev_tpclosed _#6 (fun z => tm_letm z _) (fun z => tm_letm z _)); eauto with dynamic; clear M HofM HintM.
  intros W W' HW.
  destruct HW as (Q & HintW).
  so (ints_reg _#5 HintW) as (HofW & HofW').
  so (ints_val _#6 Hwf HofT (fintctx_stat _ _ Hwf HintE) HintW) as (HvalW & HvalW').
  auto with static; done.

  eapply intt_tvrel; eauto using fintctx_stat; done.

  eapply intt_tpclosed; eauto using fintctx_stat; done.

  intros s M.
  simp_sub.
  f_equal; [].
  so (subst_closed _ _ s Hcle) as H.
  simp_sub in H.
  exact H.

  intros s M.
  simp_sub.
  f_equal; [].
  so (subst_closed _ _ s Hcle') as H.
  simp_sub in H.
  exact H.

  intros G' i M Heqi Hof.
  assert (subof dof G' (compose (left E) (sh i)) G) as Hsubof.
    eapply subof_dof_trans; eauto using left_subof_dof; [].
    subst i; auto using subof_sh, truncate_all; done.
  simp_sub.
  simp_sub in Hof.
  #4 eapply oft_letm.
    exact Hof.

    apply subst_fsts; eauto; done.

    simp_sub.
    so (subst_closed _ _ (sh i) Hcle) as Heq.
    simp_sub in Heq.
    rewrite <- Heq; [].
    so (subst_oft _#5 (subof_bind_dof_cn_md _#5 Hsubof) Hofe) as H.
    simp_sub in H.
    exact H.

    eapply subst_ofc_type; eauto using subof_dof_sof; done.

  intros G' i M Heqi Hof.
  assert (subof dof G' (compose (right E) (sh i)) G) as Hsubof.
    eapply subof_dof_trans; eauto using right_subof_dof; [].
    subst i; auto using subof_sh, truncate_all; done.
  simp_sub.
  simp_sub in Hof.
  #4 eapply oft_letm.
    exact Hof.

    apply subst_fsts; eauto; done.

    simp_sub.
    so (subst_closed _ _ (sh i) Hcle') as Heq.
    simp_sub in Heq.
    rewrite <- Heq; [].
    so (subst_oft _#5 (subof_bind_dof_cn_md _#5 Hsubof) Hofe') as H.
    simp_sub in H.
    exact H.

    eapply subst_ofc_type; eauto using subof_dof_sof; done.

  intros V V' HV.
  destruct HV as (Q & HintV).
  so (ints_reg _#5 HintV) as (HofV & HofV').
  so (ints_val _#6 Hwf HofT (fintctx_stat _ _ Hwf HintE) HintV) as (HvalV & HvalV').
  so (can_fstm _#3 wf_nil HofV) as (c & HfstV).
  so (can_fstm _#3 wf_nil HofV') as (c' & HfstV').
  #5 apply (anti_tsteps_tev _#5 (subst (dot V (dot c (left E))) e) (subst (dot V' (dot c' (right E))) e')).
    eapply oft_letm; eauto using ofm_forget, subst_fsts, subst_ofc_type, left_subof, fintctx_stat; [].
    so (subst_oft _#5 (subof_bind_dof_cn_md _#5 (left_subof_dof _ _ Hwf HintE)) Hofe) as H.
    simp_sub in H.
    simp_sub.
    assumption.

    eapply oft_letm; eauto using ofm_forget, subst_fsts, subst_ofc_type, right_subof, fintctx_stat; [].
    so (subst_oft _#5 (subof_bind_dof_cn_md _#5 (right_subof_dof _ _ Hwf HintE)) Hofe') as H.
    simp_sub in H.
    simp_sub.
    assumption.

    #2 eapply star_step.
      eapply tstep_letm2; eauto; done.

      simp_sub.
      apply star_refl; done.

    #2 eapply star_step.
      eapply tstep_letm2; eauto; done.

      simp_sub.
      apply star_refl; done.

    exploit (fun H => IH2 H anderr (E; Sta c c' Q; Dyn V V')) as Hinte; auto using wf, fintctx; [].
    simpl in Hinte; simp_sub in Hinte.
    rewrite -> (subst_intt _ _ E _ _ (ofc_fcn _#3 Hoft) (fintctx_fenv _ _ HintE)) in Hinte; auto using subenv_sh with sequence; done.

(* ext *)
intros G t M M' IH Hwf E HintE.
so (leqm_reg _#5 Hwf IH) as (HofM & HofM').
invert (ofm_reg _#4 Hwf HofM); [].
intros Hoft.
so (IH Hwf anderr _ HintE) as HintM; clear IH.
simp_sub in HintM.
unfold miev, mi in HintM.
#6 eapply (monad_mpstev_tev_tpclosed _#6 tm_ext tm_ext); eauto with dynamic; clear M HofM HintM.
  intros V V' (Q & HintV).
  so (ints_val _#6 Hwf (ofs_dyn Hoft) (fintctx_stat _ _ Hwf HintE) HintV) as (Hval & Hval').
  simpl in HintV.
  invert HintV; intros; subst; auto; done.

  eapply intt_tvrel; eauto using fintctx_stat; done.

  eapply intt_tpclosed; eauto using fintctx_stat; done.

  intros G' i M Heqi Hof.
  simp_sub in Hof; simp_sub.
  apply oft_ext; auto; done.

  intros G' i M Heqi Hof.
  simp_sub in Hof; simp_sub.
  apply oft_ext; auto; done.

  intros V V' HV.
  destruct HV as (Q & HV).
  invert HV; [].
  intros v v' _ _ Hv <- <- <-.
  so (intt_tvrel _#3 Hwf Hoft (fintctx_stat _ _ Hwf HintE) v v' Hv) as (Hofv & Hofv' & Hval & Hval').
  apply (anti_tsteps_tev _#5 v v'); auto using star_one with static dynamic; [].
  apply tev_intro; auto; done.
Qed.


Lemma leqm_comp :
  forall G p M M' T,
    mcc samek samec sames leqt leqm G p M M' T
    -> leqm G p M M' T.
Proof.
intros G p M M' T Hcc Hwf.
so (mcc_formation' _#7 leq_cregular Hwf Hcc) as (Hof & Hof' & Hfstof).
do2 2 split; auto; [].
cut (match p with
     | pure =>
         forall E c,
           wf G
           -> fintctx G E
           -> fstm G M c
           -> mpuev (subst (left E) T) (subst (right E) T) (ints T E) (subst (left E) M) (subst (right E) M') (intc c E)
     | impure =>
         forall E,
           wf G
           -> fintctx G E
           -> miev (subst (left E) T) (subst (right E) T) (ints T E) (subst (left E) M) (subst (right E) M')
     end).
  intro Hprop.
  destruct p; auto; [].
  (* pure *)
  so (Hfstof (eq_refl _)) as (k & c & c' & Hfst & Hfstm & Hfstm' & Heqc).
  split; auto; [].
  intros ka ca ca' Hfsta Hfstma Hfstma'.
  so (fsts_fun _#3 Hfst Hfsta); subst ka.
  so (fstm_fun _#4 Hfstm Hfstma); subst ca.
  so (fstm_fun _#4 Hfstm' Hfstma'); subst ca'.
  assumption.
clear Hof Hof' Hfstof Hwf.
induct Hcc.

(* stat *)
intros G c c' k Hsame E d Hwf HintE Hfstm.
invert Hsame; clear Hsame; [].
intros Hofc <-.
invert Hfstm; clear Hfstm; [].
intros <-.
simp_sub.
simpl.
unfold mpuev.
apply mevp_intro; eauto using subst_ofc, left_subof, right_subof, fintctx_stat with static dynamic; [].
apply sg_stat_interp_; [].
unfold intks.
eapply intc_fun; eauto using fintctx_stat; done.

(* dyn *)
intros G e e' t IH E c Hwf HintE Hfstm.
so (leqt_reg _#4 Hwf IH) as (Hofe & Hofe').
so (oft_reg _#3 Hwf Hofe) as Hoft.
invert Hfstm; clear Hfstm; [].
intros <-.
simp_sub.
simpl.
so (IH Hwf anderr _ HintE) as Hinte; clear IH.
unfold mpuev.
eapply monad_tev_mevp; eauto with static dynamic; [].
clear e Hofe Hinte.  
intros v v' Hv.
so (intt_tvrel _#3 Hwf Hoft (fintctx_stat _ _ Hwf HintE) _ _ Hv) as (Hofv & Hofv' & Hval & Hval').
apply mevp_intro; auto with static dynamic; [].
apply sg_dyn_interp_; auto with static; done.

(* lamgn *)
intros G T1 T1' T2 k1 M M' Hsame Hfst1 IH E c Hwf HintE Hfstm.
invert Hsame; clear Hsame; [].
intros HofT1 <-.
so (leqm_reg _#5 (wf_md Hwf HofT1 Hfst1) IH) as (HofM & HofM').
so (subst_ofs _#4 (unshift_md _#3) (ofm_reg _#4 (wf_md Hwf HofT1 Hfst1) HofM)) as HofT2.
simp_sub in HofT2.
invert Hfstm; clear Hfstm; [].
intros <-.
unfold mpuev.
#3 apply mevp_intro; eauto using subst_ofm, left_subof_dof, right_subof_dof with static.
  simp_sub.
  apply mvalue_lamgn; done.

  simp_sub.
  apply mvalue_lamgn; done.
simpl.
#3 apply sg_pigen_interp_; eauto using subst_ofm, left_subof_dof, right_subof_dof with static.
  simp_sub.
  apply mvalue_lamgn; done.

  simp_sub.
  apply mvalue_lamgn; done.
intros W W' Q HW.
so (ints_reg _#5 HW) as (HofW & HofW').
so (ints_val _#6 Hwf HofT1 (fintctx_stat _ _ Hwf HintE) HW) as (HvalW & HvalW').
so (can_fstm_value _#4 wf_nil HofW HvalW) as (c & HfstW).
so (can_fstm_value _#4 wf_nil HofW' HvalW') as (c' & HfstW').
so (ints_fst _#9 Hwf HofT1 (fintctx_stat _ _ Hwf HintE) Hfst1 HfstW HfstW' HW) as Hc.
exists c, c'.
do2 2 split; auto; [].
#2 exploit (fun H => IH H anderr (E; Sta c c' Q; Dyn W W')) as Hint; auto using wf.
  apply fintctx_md; auto; done.
simpl in Hint.
simp_sub in Hint.
repl (ints (shift1 T2) (E; Sta c c' Q; Dyn W W')) with (ints T2 (E; Sta c c' Q)) in Hint.
  unfold shift1.
  symmetry.
  #2 eapply subst_ints; eauto using ofs_fsg.
    eapply fintctx_fenv; eauto using fintctx; done.

    apply subenv_sh; auto with sequence; done.
unfold miev; unfold miev in Hint.
#4 eapply anti_msteps_mev; eauto.
  #2 eapply rel_quest.
    eapply ofm_appgn; eauto; [].
    apply ofm_forget; [].
    so (subst_ofm _#6 (left_subof_dof _ _ Hwf HintE) (ofm_lamgn HofT1 Hfst1 HofM)) as H.
    simp_sub; simp_sub in H; exact H.

    simp_sub.
    reflexivity.

  #2 eapply rel_quest.
    eapply ofm_appgn; eauto; [].
    apply ofm_forget; [].
    so (subst_ofm _#6 (right_subof_dof _ _ Hwf HintE) (ofm_lamgn HofT1 Hfst1 HofM')) as H.
    simp_sub; simp_sub in H; exact H.

    simp_sub.
    reflexivity.

  simp_sub.
  #2 eapply star_step.
    eapply mstep_appgn3; eauto; done.

    simp_sub.
    apply star_refl; done.

  simp_sub.
  #2 eapply star_step.
    eapply mstep_appgn3; eauto; done.

    simp_sub.
    apply star_refl; done.

(* appgn *)
intros G T1 T2 M1 M1' M2 M2' c2 IH1 IH2 HfstM2 E Hwf HintE.
so (leqm_reg _#5 Hwf IH1) as (HofM1 & HofM1').
so (leqm_reg _#5 Hwf IH2) as (HofM2 & HofM2').
so (ofm_reg _#4 Hwf HofM1) as Hofpi.
invert Hofpi; [].
intros k1 HofT1 Hfst1 HofT2.
so (can_fstm _#3 Hwf HofM2') as (c2' & HfstM2').
so (leqm_fst' _#7 Hwf IH2 Hfst1 HfstM2 HfstM2') as Heqc.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
so (ofm_fst _#6 Hwf HofM2 HfstM2 Hfst1) as Hofc2.
so (IH1 Hwf anderr _ HintE) as HintM1; clear IH1.
so (IH2 Hwf anderrr E c2 HintE HfstM2) as HintM2; clear IH2.
rewrite -> (miev_eqs _#5 (eqs_refl (subst_ofs _#4 (left_subof _#3 (fintctx_stat _ _ Hwf HintE)) (subst1_ofs _#4 Hofc2 HofT2))) (subst_eqs _#5 (right_subof _#3 (fintctx_stat _ _ Hwf HintE)) (funct_sg _#5 Hwf Heqc HofT2))); [].
unfold miev in HintM1; unfold mpuev in HintM2; unfold miev.
unfold mi in HintM1; unfold mi.
simp_sub.
assert (closed 0 (subst (left E) M2)) as HclM2.
  apply (fmd_closed nil); [].
  eapply ofm_fmd; eauto using subst_ofm, fcontext, left_subof_dof; done.
assert (closed 0 (subst (right E) M2')) as HclM2'.
  apply (fmd_closed nil); [].
  eapply ofm_fmd; eauto using subst_ofm, fcontext, right_subof_dof; done.
#8 eapply (monad_mpstev_mev_mpclosed _#6 (fun z => md_appgn z _) (fun z => md_appgn z _)); eauto with dynamic.
  intros V V' (Q & HV).
  so (ints_reg _#5 HV) as (Hof & Hof').
  so (ints_val _#6 Hwf Hofpi (fintctx_stat _ _ Hwf HintE) HV) as (Hval & Hval').
  auto using ofm_forget; done.

  apply mpst_mvrel; done.

  apply mpst_idem; done.

  intros s M.
  simp_sub.
  f_equal; [].
  so (subst_closed _ _ s HclM2) as H.
  simp_sub in H.
  exact H.

  intros s M.
  simp_sub.
  f_equal; [].
  so (subst_closed _ _ s HclM2') as H.
  simp_sub in H.
  exact H.

  intros G' i m Heqi Hof.
  simp_sub in Hof.
  rewrite <- (subst_closed _ _ (sh i) HclM2); [].
  simp_sub.
  so (subof_dof_trans _#5 (subof_close _ G') (left_subof_dof _ _ Hwf HintE)) as Hsubof.
  rewrite -> Heqi in Hsubof.
  #2 eapply rel_quest.
    #2 eapply ofm_appgn; eauto.
      eapply subst_ofm; eauto; done.

      eapply subst_fstm; eauto using subof_dof_tof; done.

    simp_sub.
    reflexivity.

  intros G' i m Heqi Hof.
  simp_sub in Hof.
  rewrite <- (subst_closed _ _ (sh i) HclM2'); [].
  simp_sub.
  so (subof_dof_trans _#5 (subof_close _ G') (right_subof_dof _ _ Hwf HintE)) as Hsubof.
  rewrite -> Heqi in Hsubof.
  #2 eapply rel_quest.
    #2 eapply ofm_appgn; eauto.
      eapply subst_ofm; eauto; done.

      eapply subst_fstm; eauto using subof_dof_tof; done.

    simp_sub.
    reflexivity.

  clear M1 HofM1 HintM1.
  intros V1 V1' HV1.
  destruct HV1 as (Q & HV1).
  so (ints_reg _#5 HV1) as (HofV1 & HofV1').
  so (ints_val _#6 Hwf Hofpi (fintctx_stat _ _ Hwf HintE) HV1) as (HvalV1 & HvalV1').
  #3 eapply (monad_mevp_mev _#4 (subst (left E) c2) (subst (right E) c2') (subst (left E) k1) (subst (right E) k1) _ _ (fun z => md_appgn _ z) (fun z => md_appgn _ z)); eauto using subst_fsts, subst_fstm, left_subof_dof, right_subof_dof, subof_dof_tof with dynamic.
    intros m Hofm Hfstm.
    rewrite -> split_dot, -> subst_compose.
    simp_sub in HofV1.
    eapply ofm_appgn; eauto using ofm_forget; done.

    intros m Hofm Hfstm.
    rewrite -> split_dot, -> subst_compose.
    simp_sub in HofV1'.
    #2 eapply rel_quest.
      eapply ofm_appgn; eauto using ofm_forget; done.

      simp_sub.
      reflexivity.

    intros V2 V2' d d' HfstV2 HfstV2' Heqd Heqd' HV2.
    simpl in HV2.
    so (funct_sg _#5 wf_nil Heqd (subst_ofs _#4 (subof_bind_sof_cn _#4 (left_subof _#3 (fintctx_stat _ _ Hwf HintE))) HofT2)) as Heqs.
    so (funct_sg _#5 wf_nil Heqd' (subst_ofs _#4 (subof_bind_sof_cn _#4 (right_subof _#3 (fintctx_stat _ _ Hwf HintE))) HofT2)) as Heqs'.
    simp_sub in Heqs; simp_sub in Heqs'.
    rewrite -> (mev_eqs _#5 Heqs Heqs'); [].
    rewrite -> (mpst_eqs _#5 Heqs Heqs'); [].
    simpl in HV1.
    invert HV1; [].
    intros _ _ _ _ Hincl <-.
    so (Hincl _ _ _ HV2) as (da & da' & HfstV2a & HfstV2a' & Hintapp).
    so (fstm_fun _#4 HfstV2 HfstV2a); subst da; clear HfstV2a.
    so (fstm_fun _#4 HfstV2' HfstV2a'); subst da'; clear HfstV2a'.
    #2 rewrite <- (ints_fun (G; cl_cn k1) T2 (E; Sta (subst (left E) c2) (subst (right E) c2) (intc c2 E)) (E; Sta d d' (intc c2 E))) in Hintapp; eauto using wf.
      rewrite <- (subst_ints _#5 (ofs_fsg _ _ HofT2) (fenv_cn (fintctx_fenv _ _ HintE)) (subenv_dot _#5 (ofc_fcn _#3 Hofc2) (fintctx_fenv _ _ HintE) (subenv_sh truncate_0))) in Hintapp; [].
      exact Hintapp.

      #2 apply intctx_cn; auto using fintctx_stat.
        eapply eqc_trans; eauto; [].
        eapply subst_eqc; eauto using fintctx_stat, right_subof; done.

        eapply intc_fun; eauto using fintctx_stat, ofm_fst; done.
      
(* lamap *)
intros G T1 T1' T2 k1 M M' Hsame Hfst1 IH E da Hwf HintE Hfstlam.
invert Hsame; clear Hsame; [].
intros HofT1 <-.
so (leqm_reg _#5 (wf_md Hwf HofT1 Hfst1) IH) as (HofM & HofM').
so (subst_ofs _#4 (unshift_md _#3) (ofm_reg _#4 (wf_md Hwf HofT1 Hfst1) HofM)) as HofT2.
simp_sub in HofT2.
invert Hfstlam; clear Hfstlam; [].
intros k1a d Hfst1a HfstM <-.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (can_fstm _#3 Hwf (ofm_lamap HofT1 Hfst1 HofM')) as (da' & Hfstlam').
invert Hfstlam'; clear Hfstlam'; [].
intros k1a d' Hfst1a HfstM' _.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
unfold mpuev.
#3 apply mevp_intro; eauto using subst_ofm, left_subof_dof, right_subof_dof with static.
  simp_sub.
  apply mvalue_lamap; done.

  simp_sub.
  apply mvalue_lamap; done.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
so (subst_eqc _#6 (unshift_md _#3) (leqm_fst' _#7 (wf_md Hwf HofT1 Hfst1) IH (subst_fsts _#3 Hfst2) HfstM HfstM')) as Heqd.
simp_sub in Heqd.
so (wf_cn Hwf Hofk1) as Hwf'.
so (eqc_reg _#4 Hwf' Heqd) as (Hofd & Hofd' & _).
so (ofc_reg _#3 Hwf' Hofd) as Hofk2.
rewrite -> (intc_lam_unroll _#5 (wf_cn Hwf Hofk1) Hofd (fintctx_stat _ _ Hwf HintE)); [].
fold precand.
simpl ints.
#4 eapply sg_piapp_interp_; eauto using subst_ofm, left_subof_dof, right_subof_dof, subst_fstm, subof_dof_tof with static.
  simp_sub.
  apply mvalue_lamap; done.

  simp_sub.
  apply mvalue_lamap; done.

  unfold intks.
  simpl.
  simp_sub.
  #3 apply pi_interp_.
    simp_sub.
    apply ofc_lam; eauto using subst_ofk, subst_ofc, left_subof, fintctx_stat, subof_bind_sof_cn; done.

    simp_sub.
    apply ofc_lam; eauto using subst_ofk, subst_ofc, right_subof, fintctx_stat, subof_bind_sof_cn; done.

    intros b b' Q Q' b'' b''' Hb Heqc Heqc'.
    so (eqc_reg _#4 wf_nil Heqc) as (Hofb & _).
    so (eqc_reg _#4 wf_nil Heqc') as (Hofb' & _).
    assert (intctx (G; cl_cn k1) (E; Sta b b' (Cand (simp k1) Q)) (E; Sta b'' b''' (Cand (simp k1) Q'))) as HintEx by (apply intctx_cn; auto using fintctx_stat).
    rewrite -> (cand_eta _ _ (ofc_intc_csk _#4 Hwf' Hofd (intctx_refl _#3 Hwf' HintEx))); [].
    rewrite -> (cand_eta _ _ (ofc_intc_csk _#4 Hwf' Hofd (intctx_refl _#3 Hwf' (intctx_symm _#3 Hwf' HintEx)))); [].
    so (intc_fun _#5 Hwf' Hofd HintEx) as Hintd.
    simpl in Hintd.
    #2 eapply intk_defeq; eauto.
      simpl.
      apply eqc_symm; [].
      rewrite -> !(split_dot_compose b); [].
      apply eqc_beta; eauto using subst_ofc, subof_bind_sof_cn, left_subof, fintctx_stat; done.

      simpl.
      #2 apply (@eqc_trans _ _ (subst (dot b' (right E)) d')).
        eapply subst_eqc; eauto; [].
        apply subof_dot; eauto using right_subof, fintctx_stat; [].
        simp_sub.
        apply sof_cn; auto; done.

        apply eqc_symm; [].
        rewrite -> !(split_dot_compose b'); [].
        apply eqc_beta; eauto using subst_ofc, subof_bind_sof_cn, right_subof, fintctx_stat; done.
intros W W' Q HW.
so (ints_reg _#5 HW) as (HofW & HofW').
so (ints_val _#6 Hwf HofT1 (fintctx_stat _ _ Hwf HintE) HW) as (HvalW & HvalW').
so (can_fstm _#3 wf_nil HofW) as (c & HfstW).
so (can_fstm _#3 wf_nil HofW') as (c' & HfstW').
so (ints_fst _#9 Hwf HofT1 (fintctx_stat _ _ Hwf HintE) Hfst1 HfstW HfstW' HW) as Hc.
exists c, c'.
do2 2 split; auto; [].
#2 exploit (fun H => IH H anderrr (E; Sta c c' Q; Dyn W W') (shift1 d)) as Hint; auto using wf.
  apply fintctx_md; auto; done.
simpl in Hint.
simp_sub in Hint.
repl (ints (shift1 T2) (E; Sta c c' Q; Dyn W W')) with (ints T2 (E; Sta c c' Q)) in Hint.
  unfold shift1.
  symmetry.
  #2 eapply subst_ints; eauto using ofs_fsg.
    eapply fintctx_fenv; eauto using fintctx; done.

    apply subenv_sh; auto with sequence; done.
rewrite -> (cand_eta _ _ (intk_csk _#6 Hc andel)); [].
rewrite -> (cand_eta _ _ (ofc_intc_csk _#4 Hwf' Hofd (fintctx_stat _ _ Hwf' (fintctx_cn HintE Hc)))); [].
unfold shift1 in Hint; rewrite -> (subst_intc (G; cl_cn k1) _ (E; Sta c c' Q)) in Hint; eauto using ofc_fcn, fintctx_fenv, fenv, subenv_sh with sequence; [].
unfold mpuev; unfold mpuev in Hint.
#4 eapply anti_msteps_mevp; eauto.
  #2 eapply rel_quest.
    eapply ofm_appap; eauto; [].
    so (subst_ofm _#6 (left_subof_dof _ _ Hwf HintE) (ofm_lamap HofT1 Hfst1 HofM)) as H.
    simp_sub; simp_sub in H; exact H.

    simp_sub.
    reflexivity.

  #2 eapply rel_quest.
    eapply ofm_appap; eauto; [].
    so (subst_ofm _#6 (right_subof_dof _ _ Hwf HintE) (ofm_lamap HofT1 Hfst1 HofM')) as H.
    simp_sub; simp_sub in H; exact H.

    simp_sub.
    reflexivity.

  simp_sub.
  #2 eapply star_step.
    eapply mstep_appap3; eauto; done.

    simp_sub.
    apply star_refl; done.

  simp_sub.
  #2 eapply star_step.
    eapply mstep_appap3; eauto; done.

    simp_sub.
    apply star_refl; done.

(* appap *)
intros G p T1 T2 M1 M1' M2 M2' c2 IH1 IH2 HfstM2.
#2 destruct p.
  (* pure *)
  intros E b Hwf HintE Hfstb.
  so (leqm_reg _#5 Hwf IH1) as (HofM1 & HofM1').
  so (leqm_reg _#5 Hwf IH2) as (HofM2 & HofM2').
  invert Hfstb; clear Hfstb; [].
  intros c1 c2a HfstM1 HfstM2a <-.
  so (fstm_fun _#4 HfstM2 HfstM2a); subst c2a; clear HfstM2a.
  so (can_fstm _#3 Hwf HofM1') as (c1' & HfstM1').
  so (can_fstm _#3 Hwf HofM2') as (c2' & HfstM2').
  so (ofm_reg _#4 Hwf HofM1) as Hofpi.
  invert Hofpi; clear Hofpi; [].
  intros k1 HofT1 Hfst1 HofT2.
  so (can_fsts _ _ HofT2) as (k2 & Hfst2).
  so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
  so (ofs_fst _#3 HofT2 Hfst2) as Hofk2.
  so (wf_cn Hwf Hofk1) as Hwf'.
  so (leqm_fst' _#7 Hwf IH2 Hfst1 HfstM2 HfstM2') as Heqc2.
  so (eqc_reg _#4 Hwf Heqc2) as (Hofc2 & _).
  so (IH1 Hwf anderrr _ _ HintE HfstM1) as Hint1; clear IH1.
  so (IH2 Hwf anderrr _ _ HintE HfstM2) as Hint2; clear IH2.
  rewrite -> (mpuev_eqs _#5 (eqs_refl (subst_ofs _#4 (left_subof _#3 (fintctx_stat _ _ Hwf HintE)) (subst1_ofs _#4 Hofc2 HofT2))) (subst_eqs _#5 (right_subof _#3 (fintctx_stat _ _ Hwf HintE)) (funct_sg _#5 Hwf Heqc2 HofT2))); [].
  simp_sub.
  unfold mpuev in Hint1, Hint2; unfold mpuev.
  #3 eapply (monad_mevp_mevp (subst (left E) (sg_piapp T1 T2)) (subst (right E) (sg_piapp T1 T2)) _#8 (fun z => md_appap z _) (fun z => md_appap z _)); eauto using fsts_piapp, subst_fsts, subst_fstm, left_subof_dof, right_subof_dof, subof_dof_tof with dynamic.
    intros m Hof Hfstm.
    simp_sub in Hof.
    #2 eapply rel_quest.
      eapply ofm_appap; eauto using subst_ofm, subst_fstm, left_subof_dof, subof_dof_tof; done.

      simp_sub.
      reflexivity.

    intros m Hof Hfstm.
    simp_sub in Hof.
    #2 eapply rel_quest.
      eapply ofm_appap; eauto using subst_ofm, subst_fstm, right_subof_dof, subof_dof_tof; done.

      simp_sub.
      reflexivity.

    intros V1 V1' d1 d1' HfstV1 HfstV1' Heqd1 Heqd1' HV1.
    simpl in HV1.
    invert HV1; clear HV1; [].
    intros d1a d1a' k1a k2a F HofV1 HofV1' HvalV1 HvalV1' HfstV1a HfstV1a' Hfst1a Hfst2a HF Hincl Heqcand.
    so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
    so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a; clear Hfst2a.
    so (fstm_fun _#4 HfstV1 HfstV1a); subst d1a; clear HfstV1a.
    so (fstm_fun _#4 HfstV1' HfstV1a'); subst d1a'; clear HfstV1a'.
    so (mvalue_pure _#4 HvalV1 HofV1) as H; renameover H into HofV1.
    so (mvalue_pure _#4 HvalV1' HofV1') as H; renameover H into HofV1'.
    #3 eapply (monad_mevp_mevp (subst (left E) T1) (subst (right E) T1) _#8 (fun z => md_appap _ z) (fun z => md_appap _ z)); eauto using subst_fsts, subst_fstm, left_subof_dof, right_subof_dof, subof_dof_tof with dynamic.
      intros m Hof Hfstm.
      simp_sub in HofV1.
      #2 eapply rel_quest.
        eapply ofm_appap; eauto using subst_ofm, subst_fstm, left_subof_dof, subof_dof_tof; done.

        simp_sub.
        reflexivity.

      intros m Hof Hfstm.
      simp_sub in HofV1'.
      #2 eapply rel_quest.
        eapply ofm_appap; eauto using subst_ofm, subst_fstm, right_subof_dof, subof_dof_tof; done.

        simp_sub.
        reflexivity.

      intros V2 V2' d2 d2' HfstV2 HfstV2' Heqd2 Heqd2' HV2.
      simpl in HV2.
      so (ints_reg _#5 HV2) as (HofV2 & HofV2').
      so (Hincl _ _ _ HV2) as (d2a & d2a' & HfstV2a & HfstV2a' & Hintapp); clear Hincl.
      so (fstm_fun _#4 HfstV2 HfstV2a); subst d2a; clear HfstV2a.
      so (fstm_fun _#4 HfstV2' HfstV2a'); subst d2a'; clear HfstV2a'.
      assert (eqc nil (subst (right E) c2) d2' (subst (right E) k1)) as Heqd2''.
        eapply eqc_trans; eauto; [].
        eapply subst_eqc; eauto using right_subof, fintctx_stat; done.
      unfold mpuev in Hintapp.
      simpl intc.
      rewrite <- Heqcand; clear Heqcand; [].
      (* reconcile substitution into T2 *)
      unfold subst1; rewrite -> (subst_ints _#5 (ofs_fsg _ _ HofT2) (fenv_cn (fintctx_fenv _ _ HintE)) (subenv_dot _#5 (ofc_fcn _#3 Hofc2) (fintctx_fenv _ _ HintE) (subenv_id _))); [].
      (* reconcile environment *)
      assert (intctx (G; cl_cn k1)
                (E; Sta (subst (left E) c2) (subst (right E) c2) (intc c2 E))
                (E; Sta d2 d2' (intc c2 E))) as Hintctx.
        so (ints_fst _#9 Hwf HofT1 (fintctx_stat _ _ Hwf HintE) Hfst1 HfstV2 HfstV2' HV2) as Hd2.
        apply intctx_symm; auto using wf; [].
        apply intctx_cn; auto using fintctx_stat, eqc_symm; done.
      rewrite -> (ints_fun (G; cl_cn k1) T2 (E; Sta (subst (left E) c2) (subst (right E) c2) (intc c2 E)) (E; Sta d2 d2' (intc c2 E))); auto using wf; [].
      (* reconcile the candidate (a bit messy) *)
      #2 apply (mevp_funct _ _ (fun V V' => ints T2 (E; Sta d2 d2' (intc c2 E)) V V' (Cand (simp k2) (F d2 d2' (out (simp k1) (intc c2 E)))))).
        intros V V' Hsteps Hsteps' _ _ Hint.
        (* show that the candidates are equivalent (alas, not equal) *)
        unfold intks in HF; invert HF; clear HF; [].
        intros Fa Faa _ _ HinclF HeqFa HeqFaa.
        so (cand_inject _#3 HeqFa); subst Fa.
        so (cand_inject _#3 HeqFaa); subst Faa.
        clear HeqFa HeqFaa.
        so (ints_fst _#9 Hwf HofT1 (fintctx_stat _ _ Hwf HintE) Hfst1 HfstV2 HfstV2' HV2) as Hd2.
        rewrite <- (cand_eta _ _ (ofc_intc_csk _#4 Hwf Hofc2 (fintctx_stat _ _ Hwf HintE))) in Hd2; [].
        so (HinclF _#6 Hd2 (eqc_symm Heqd2) (eqc_symm Heqd2'')) as HequivFa; clear HinclF.
        rewrite -> (cand_eta _ _ (ofc_intc_csk _#4 Hwf Hofc2 (fintctx_stat _ _ Hwf HintE))) in HequivFa; [].
        (* adjust the equivalence for the constructors we need *)
        simp_sub in HofV1; simp_sub in HofV1'.
        so (preservations_fst _#7 wf_nil (ofm_appap HofV1 HofV2 HfstV2) Hsteps (subst_fsts _#3 (subst_fsts _#3 Hfst2)) (fstm_appap HfstV1 HfstV2)) as (e & Hfste & Heqe).
        so (preservations_fst _#7 wf_nil (ofm_appap HofV1' HofV2' HfstV2') Hsteps' (subst_fsts _#3 (subst_fsts _#3 Hfst2)) (fstm_appap HfstV1' HfstV2')) as (e' & Hfste' & Heqe').
        simp_sub in Heqe; simp_sub in Heqe'.
        so (intk_defeq _#10 Hwf' Hofk2 (intctx_symm _#3 Hwf' Hintctx) HequivFa Heqe Heqe') as HequivFa'; clear HequivFa.
        eapply ints_equiv_cand; eauto using intctx_refl, intctx_symm; done.

        (* finish up *)
        simp_sub.
        #2 rewrite -> (mevp_eqs _ (subst (dot d2 (left E)) T2) _ (subst (dot d2' (right E)) T2)); auto.
          rewrite -> !(split_dot_compose _ (left E)); [].
          eapply funct_sg; eauto using wf, subst_ofs, subof_bind_sof_cn, left_subof, fintctx_stat; done.

          rewrite -> !(split_dot_compose _ (right E)); [].
          eapply funct_sg; eauto using wf, subst_ofs, subof_bind_sof_cn, right_subof, fintctx_stat; done.

  (* impure *)
  intros E Hwf HintE.
  so (leqm_reg _#5 Hwf IH1) as (HofM1 & HofM1').
  so (leqm_reg _#5 Hwf IH2) as (HofM2 & HofM2').
  so (ofm_reg _#4 Hwf HofM1) as Hofpi.
  invert Hofpi; [].
  intros k1 HofT1 Hfst1 HofT2.
  so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
  so (can_fstm _#3 Hwf HofM2') as (c2' & HfstM2').
  so (leqm_fst' _#7 Hwf IH2 Hfst1 HfstM2 HfstM2') as Heqc2.
  so (eqc_reg _#4 Hwf Heqc2) as (Hofc2 & Hofc2' & _).
  so (IH1 Hwf anderr E HintE) as HintM1; clear IH1.
  so (IH2 Hwf anderrr E c2 HintE HfstM2) as HintM2; clear IH2.
  rewrite -> (miev_eqs _#5 (eqs_refl (subst_ofs _#4 (left_subof _#3 (fintctx_stat _ _ Hwf HintE)) (subst1_ofs _#4 Hofc2 HofT2))) (subst_eqs _#5 (right_subof _#3 (fintctx_stat _ _ Hwf HintE)) (funct_sg _#5 Hwf Heqc2 HofT2))); [].
  unfold miev in HintM1; unfold mpuev in HintM2; unfold miev.
  unfold mi in HintM1; unfold mi.
  simp_sub.
  assert (closed 0 (subst (left E) M2)) as HclM2.
    apply (fmd_closed nil); [].
    eapply ofm_fmd; eauto using subst_ofm, fcontext, left_subof_dof; done.
  assert (closed 0 (subst (right E) M2')) as HclM2'.
    apply (fmd_closed nil); [].
    eapply ofm_fmd; eauto using subst_ofm, fcontext, right_subof_dof; done.
  #8 eapply (monad_mpstev_mev_mpclosed _#6 (fun z => md_appap z _) (fun z => md_appap z _)); eauto with dynamic.
    intros V V' (Q & HV).
    so (ints_reg _#5 HV) as (Hof & Hof').
    so (ints_val _#6 Hwf Hofpi (fintctx_stat _ _ Hwf HintE) HV) as (Hval & Hval').
    auto using ofm_forget; done.
  
    apply mpst_mvrel; done.
  
    apply mpst_idem; done.
  
    intros s M.
    simp_sub.
    f_equal; [].
    so (subst_closed _ _ s HclM2) as H.
    simp_sub in H.
    exact H.
  
    intros s M.
    simp_sub.
    f_equal; [].
    so (subst_closed _ _ s HclM2') as H.
    simp_sub in H.
    exact H.
  
    intros G' i m Heqi Hof.
    simp_sub in Hof.
    rewrite <- (subst_closed _ _ (sh i) HclM2); [].
    simp_sub.
    so (subof_dof_trans _#5 (subof_close _ G') (left_subof_dof _ _ Hwf HintE)) as Hsubof.
    rewrite -> Heqi in Hsubof.
    #2 eapply rel_quest.
      #2 eapply ofm_appap; eauto.
        eapply subst_ofm; eauto; done.
  
        eapply subst_fstm; eauto using subof_dof_tof; done.
  
      simp_sub.
      reflexivity.
  
    intros G' i m Heqi Hof.
    simp_sub in Hof.
    rewrite <- (subst_closed _ _ (sh i) HclM2'); [].
    simp_sub.
    so (subof_dof_trans _#5 (subof_close _ G') (right_subof_dof _ _ Hwf HintE)) as Hsubof.
    rewrite -> Heqi in Hsubof.
    #2 eapply rel_quest.
      #2 eapply ofm_appap; eauto.
        eapply subst_ofm; eauto; done.
  
        eapply subst_fstm; eauto using subof_dof_tof; done.
  
      simp_sub.
      reflexivity.
  
    clear M1 HofM1 HintM1.
    intros V1 V1' HV1.
    destruct HV1 as (Q & HV1).
    so (ints_reg _#5 HV1) as (HofV1 & HofV1').
    so (ints_val _#6 Hwf Hofpi (fintctx_stat _ _ Hwf HintE) HV1) as (HvalV1 & HvalV1').
    #3 eapply (monad_mevp_mev _#4 (subst (left E) c2) (subst (right E) c2') (subst (left E) k1) (subst (right E) k1) _ _ (fun z => md_appap _ z) (fun z => md_appap _ z)); eauto using subst_fsts, subst_fstm, left_subof_dof, right_subof_dof, subof_dof_tof with dynamic.
      intros m Hofm Hfstm.
      rewrite -> split_dot, -> subst_compose.
      simp_sub in HofV1.
      eapply ofm_appap; eauto using ofm_forget; done.
  
      intros m Hofm Hfstm.
      rewrite -> split_dot, -> subst_compose.
      simp_sub in HofV1'.
      eapply ofm_appap; eauto using ofm_forget; done.
  
      intros V2 V2' d d' HfstV2 HfstV2' Heqc Heqc' HV2.
      simpl in HV2.
      so (funct_sg _#5 wf_nil Heqc (subst_ofs _#4 (subof_bind_sof_cn _#4 (left_subof _#3 (fintctx_stat _ _ Hwf HintE))) HofT2)) as Heqs.
      so (funct_sg _#5 wf_nil Heqc' (subst_ofs _#4 (subof_bind_sof_cn _#4 (right_subof _#3 (fintctx_stat _ _ Hwf HintE))) HofT2)) as Heqs'.
      simp_sub in Heqs; simp_sub in Heqs'.
      rewrite -> (mev_eqs _#5 Heqs Heqs'); [].
      rewrite -> (mpst_eqs _#5 Heqs Heqs'); [].
      simpl in HV1.
      invert HV1; [].
      intros c c' k1a k2 F _ _ _ _ _ _ Hfst1a Hfst2 _ Hincl <-.
      so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
      so (Hincl _ _ _ HV2) as (da & da' & HfstV2a & HfstV2a' & Hintapp).
      so (fstm_fun _#4 HfstV2 HfstV2a); subst da; clear HfstV2a.
      so (fstm_fun _#4 HfstV2' HfstV2a'); subst da'; clear HfstV2a'.
      #2 rewrite <- (ints_fun (G; cl_cn k1) T2 (E; Sta (subst (left E) c2) (subst (right E) c2) (intc c2 E)) (E; Sta d d' (intc c2 E))) in Hintapp; eauto using wf.
        rewrite <- (subst_ints _#5 (ofs_fsg _ _ HofT2) (fenv_cn (fintctx_fenv _ _ HintE)) (subenv_dot _#5 (ofc_fcn _#3 Hofc2) (fintctx_fenv _ _ HintE) (subenv_sh truncate_0))) in Hintapp; [].
        eapply mpuev_miev; eauto; [].
        intros V V' (Q' & HV).
        so (ints_reg _#5 HV) as (Hof & Hof').
        so (ints_val _#6 Hwf (subst1_ofs _#4 Hofc2 HofT2) (fintctx_stat _ _ Hwf HintE) HV) as (Hval & Hval').
        #2 do2 3 split; auto.
          simp_sub in Hof.
          apply ofm_forget; [].
          eapply ofm_subsume; eauto; [].
          apply subs_refl; [].
          rewrite -> !(split_dot_compose _ (left E)); [].
          eapply funct_sg; eauto using wf, subst_ofs, subof_bind_sof_cn, left_subof, fintctx_stat; done.

          simp_sub in Hof'.
          apply ofm_forget; [].
          eapply ofm_subsume; eauto; [].
          apply subs_refl; [].
          rewrite -> !(split_dot_compose _ (right E)); [].
          eapply funct_sg; eauto using wf, subst_ofs, subof_bind_sof_cn, right_subof, fintctx_stat; [].
          eapply eqc_trans; eauto; [].
          eapply subst_eqc; eauto using right_subof, fintctx_stat; done.

        #2 apply intctx_cn; auto using fintctx_stat.
          eapply eqc_trans; eauto; [].
          eapply subst_eqc; eauto using right_subof, fintctx_stat; done.

          eapply intc_fun; eauto using fintctx_stat, ofm_fst; done.

(* pair *)
intros G p T1 T2 M1 M1' M2 M2' IH1 IH2.
#2 destruct p.
  (* pure *)
  intros E b Hwf HintE Hfstpair.
  so (leqm_reg _#5 Hwf IH1) as (HofM1 & HofM1').
  so (leqm_reg _#5 Hwf IH2) as (HofM2 & HofM2').
  invert Hfstpair; clear Hfstpair; [].
  intros c1 c2 HfstM1 HfstM2 <-.
  so (can_fstm _#3 Hwf HofM1') as (c1' & HfstM1').
  so (can_fstm _#3 Hwf HofM2') as (c2' & HfstM2').
  so (ofm_reg _#4 Hwf HofM1) as HofT1.
  so (ofm_reg _#4 Hwf HofM2) as HofT2.
  so (can_fsts _ _ HofT1) as (k1 & Hfst1).
  so (can_fsts _ _ HofT2) as (k2 & Hfst2).
  so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
  so (ofs_fst _#3 HofT2 Hfst2) as Hofk2.
  so (ofm_fst _#6 Hwf HofM1 HfstM1 Hfst1) as Hofc1.
  so (ofm_fst _#6 Hwf HofM2 HfstM2 Hfst2) as Hofc2.
  so (IH1 Hwf anderrr _ _ HintE HfstM1) as Hint1.
  so (IH2 Hwf anderrr _ _ HintE HfstM2) as Hint2.
  clear IH1 IH2.
  unfold mpuev in Hint1, Hint2; unfold mpuev.
  simp_sub.
  #3 eapply (monad_mevp_mevp (subst (left E) T1) (subst (right E) T1) _#8 (fun z => md_pair z _) (fun z => md_pair z _)); eauto using subst_fsts, subst_fstm, subof_dof_tof, left_subof_dof, right_subof_dof with dynamic.
    intros m Hofm _.
    #2 eapply rel_quest.
      eapply ofm_pair; eauto using subst_ofm, left_subof_dof; done.

      simp_sub.
      reflexivity.

    intros m Hofm _.
    #2 eapply rel_quest.
      eapply ofm_pair; eauto using subst_ofm, right_subof_dof; done.

      simp_sub.
      reflexivity.

    intros V1 V1' d1 d1' HfstV1 HfstV1' Heqd1 Heqd1' HV1.
    simpl in HV1.
    so (ints_reg _#5 HV1) as (HofV1 & HofV1').
    so (ints_val _#6 Hwf HofT1 (fintctx_stat _ _ Hwf HintE) HV1) as (HvalV1 & HvalV1').
    #3 eapply (monad_mevp_mevp (subst (left E) T2) (subst (right E) T2) _#8 (fun z => md_pair _ z) (fun z => md_pair _ z)); eauto using subst_fsts, subst_fstm, subof_dof_tof, left_subof_dof, right_subof_dof with dynamic.
      intros m Hofm _.
      #2 eapply rel_quest.
        eapply ofm_pair; eauto using subst_ofm, left_subof_dof; done.
  
        simp_sub.
        reflexivity.
  
      intros m Hofm _.
      #2 eapply rel_quest.
        eapply ofm_pair; eauto using subst_ofm, right_subof_dof; done.
  
        simp_sub.
        reflexivity.

      intros V2 V2' d2 d2' HfstV2 HfstV2' Heqd2 Heqd2' HV2.
      simpl in HV2.
      so (ints_reg _#5 HV2) as (HofV2 & HofV2').
      so (ints_val _#6 Hwf HofT2 (fintctx_stat _ _ Hwf HintE) HV2) as (HvalV2 & HvalV2').
      #3 apply mevp_intro; auto with dynamic.
        #2 eapply rel_quest.
          eapply ofm_pair; eauto; done.
    
          simp_sub.
          reflexivity.

        #2 eapply rel_quest.
          eapply ofm_pair; eauto; done.
    
          simp_sub.
          reflexivity.

        simpl.
        rewrite <- (cand_eta _ _ (ofc_intc_csk _#4 Hwf Hofc1 (fintctx_stat _ _ Hwf HintE))); [].
        rewrite <- (cand_eta _ _ (ofc_intc_csk _#4 Hwf Hofc2 (fintctx_stat _ _ Hwf HintE))); [].
        repl (simp k2) with (simp (shift1 k2)).
          apply subst_simp; eauto using ofk_hkd; done.
        #4 eapply sg_sigma_interp_; eauto using subst_fsts.
          #2 eapply rel_quest.
            eapply ofm_pair; eauto using ofm_forget; done.

            simp_sub.
            reflexivity.

          #2 eapply rel_quest.
            eapply ofm_pair; eauto using ofm_forget; done.

            simp_sub.
            reflexivity.

          simpl.
          rewrite -> (cand_eta _ _ (ofc_intc_csk _#4 Hwf Hofc1 (fintctx_stat _ _ Hwf HintE))); [].
          exact HV1.

          simpl.
          unfold shift1.
          rewrite -> (subst_ints _#5 (ofs_fsg _ _ HofT2) (fintctx_fenv _ _ HintE) (subenv_sh (truncate_S truncate_0))); [].
          rewrite -> (subst_simp _ _ (ofk_hkd _ _ Hofk2)); [].
          rewrite -> (cand_eta _ _ (ofc_intc_csk _#4 Hwf Hofc2 (fintctx_stat _ _ Hwf HintE))); [].
          exact HV2.
          
  (* impure *)
  intros E Hwf HintE.
  so (leqm_reg _#5 Hwf IH1) as (HofM1 & HofM1').
  so (leqm_reg _#5 Hwf IH2) as (HofM2 & HofM2').
  so (ofm_reg _#4 Hwf HofM1) as HofT1.
  so (ofm_reg _#4 Hwf HofM2) as HofT2.
  so (can_fsts _ _ HofT1) as (k1 & Hfst1).
  so (can_fsts _ _ HofT2) as (k2 & Hfst2).
  so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
  so (ofs_fst _#3 HofT2 Hfst2) as Hofk2.
  so (IH1 Hwf anderr _ HintE) as Hint1.
  so (IH2 Hwf anderr _ HintE) as Hint2.
  clear IH1 IH2.
  unfold miev, mi in Hint1, Hint2; unfold miev, mi.
  simp_sub.
  assert (closed 0 (subst (left E) M2)) as HclM2.
    apply (fmd_closed nil); [].
    eapply ofm_fmd; eauto using subst_ofm, fcontext, left_subof_dof; done.
  assert (closed 0 (subst (right E) M2')) as HclM2'.
    apply (fmd_closed nil); [].
    eapply ofm_fmd; eauto using subst_ofm, fcontext, right_subof_dof; done.
  #8 eapply (monad_mpstev_mev_mpclosed _#6 (fun z => md_pair z _) (fun z => md_pair z _)); eauto with dynamic.
    intros V V' (Q & HV).
    so (ints_reg _#5 HV) as (Hof & Hof').
    so (ints_val _#6 Hwf HofT1 (fintctx_stat _ _ Hwf HintE) HV) as (Hval & Hval').
    auto using ofm_forget; done.

    apply mpst_mvrel; done.

    apply mpst_idem; done.

    intros s M.
    simp_sub.
    f_equal; [].
    so (subst_closed _ _ s HclM2) as H.
    simp_sub in H.
    exact H.
  
    intros s M.
    simp_sub.
    f_equal; [].
    so (subst_closed _ _ s HclM2') as H.
    simp_sub in H.
    exact H.

    intros G' i m Heqi Hof.
    simp_sub in Hof.
    rewrite <- (subst_closed _ _ (sh i) HclM2); [].
    simp_sub.
    so (subof_dof_trans _#5 (subof_close _ G') (left_subof_dof _ _ Hwf HintE)) as Hsubof.
    rewrite -> Heqi in Hsubof.
    #2 eapply rel_quest.
      eapply ofm_pair; eauto; [].
      eapply subst_ofm; eauto; done.
  
      simp_sub.
      reflexivity.

    intros G' i m Heqi Hof.
    simp_sub in Hof.
    rewrite <- (subst_closed _ _ (sh i) HclM2'); [].
    simp_sub.
    so (subof_dof_trans _#5 (subof_close _ G') (right_subof_dof _ _ Hwf HintE)) as Hsubof.
    rewrite -> Heqi in Hsubof.
    #2 eapply rel_quest.
      eapply ofm_pair; eauto; [].
      eapply subst_ofm; eauto; done.
  
      simp_sub.
      reflexivity.

    intros V1 V1' (Q1 & HV1).
    clear M1 HofM1 Hint1.
    so (ints_reg _#5 HV1) as (HofV1 & HofV1').
    so (ints_val _#6 Hwf HofT1 (fintctx_stat _ _ Hwf HintE) HV1) as (HvalV1 & HvalV1').
    assert (closed 0 V1) as HclV1.
      apply (fmd_closed nil); [].
      eapply ofm_fmd; eauto using fcontext_nil; done.
    assert (closed 0 V1') as HclV1'.
      apply (fmd_closed nil); [].
      eapply ofm_fmd; eauto using fcontext_nil; done.
    #8 eapply (monad_mpstev_mev_mpclosed _#6 (fun z => md_pair _ z) (fun z => md_pair _ z)); eauto with dynamic.
      intros V V' (Q & HV).
      so (ints_reg _#5 HV) as (Hof & Hof').
      so (ints_val _#6 Hwf HofT2 (fintctx_stat _ _ Hwf HintE) HV) as (Hval & Hval').
      auto using ofm_forget; done.
  
      apply mpst_mvrel; done.
  
      apply mpst_idem; done.
  
      intros s M.
      simp_sub.
      f_equal; [].
      exact (subst_closed _ _ s HclV1).
    
      intros s M.
      simp_sub.
      f_equal; [].
      exact (subst_closed _ _ s HclV1').
  
      intros G' i m Heqi Hof.
      simp_sub in Hof.
      rewrite <- (subst_closed _ _ (sh i) HclV1); [].
      simp_sub.
      so (subof_dof_trans _#5 (subof_close _ G') (left_subof_dof _ _ Hwf HintE)) as Hsubof.
      rewrite -> Heqi in Hsubof.
      #2 eapply rel_quest.
        eapply ofm_pair; eauto; [].
        eapply subst_ofm; eauto using ofm_forget; [].
        subst i; apply subof_close; done.
    
        simp_sub.
        reflexivity.
    
      intros G' i m Heqi Hof.
      simp_sub in Hof.
      rewrite <- (subst_closed _ _ (sh i) HclV1'); [].
      simp_sub.
      so (subof_dof_trans _#5 (subof_close _ G') (left_subof_dof _ _ Hwf HintE)) as Hsubof.
      rewrite -> Heqi in Hsubof.
      #2 eapply rel_quest.
        eapply ofm_pair; eauto; [].
        eapply subst_ofm; eauto using ofm_forget; [].
        subst i; apply subof_close; done.
    
        simp_sub.
        reflexivity.
    
      intros V2 V2' (Q2 & HV2).
      simpl in HV2.
      so (ints_reg _#5 HV2) as (HofV2 & HofV2').
      so (ints_val _#6 Hwf HofT2 (fintctx_stat _ _ Hwf HintE) HV2) as (HvalV2 & HvalV2').
      #3 apply mev_intro; auto with dynamic.
        #2 eapply rel_quest.
          eapply ofm_pair; eauto using ofm_forget; done.
    
          simp_sub.
          reflexivity.

        #2 eapply rel_quest.
          eapply ofm_pair; eauto using ofm_forget; done.
    
          simp_sub.
          reflexivity.

        #2 apply mpst_expand.
          intros V V' (Q & HV).
          so (ints_reg _#5 HV) as (HofV & HofV').
          so (ints_val _#6 Hwf (ofs_sigma HofT1 Hfst1 (weaken1_ofs _#3 HofT2)) (fintctx_stat _ _ Hwf HintE) HV) as (Hval & Hval').
          simp_sub in HofV; simp_sub in HofV'.
          do2 3 split; auto using ofm_forget; done.

          exists (Cand (sk_sigma (simp k1) (simp (shift1 k2))) (out (simp k1) Q1, out (simp (shift1 k2)) Q2)).
          simpl.
          so (can_fstm _#3 wf_nil HofV1) as (d1 & HfstV1).
          so (can_fstm _#3 wf_nil HofV1') as (d1' & HfstV1').
          #4 eapply sg_sigma_interp_; eauto using subst_fsts.
            #2 eapply rel_quest.
              eapply ofm_pair; eauto using ofm_forget; done.

              simp_sub.
              reflexivity.

            #2 eapply rel_quest.
              eapply ofm_pair; eauto using ofm_forget; done.

              simp_sub.
              reflexivity.

            simpl.
            rewrite -> cand_eta; auto; [].
            eapply (ints_csk _ T1); eauto using fintctx_stat; done.

            simpl.
            rewrite -> (subst_ints _#5 (ofs_fsg _ _ HofT2) (fintctx_fenv _ _ HintE) (subenv_sh (truncate_S truncate_0))); [].
            unfold shift1; rewrite -> (subst_simp _ _ (ofk_hkd _ _ Hofk2)); [].
            rewrite -> cand_eta; auto; [].
            eapply (ints_csk _ T2); eauto using fintctx_stat; done.
            
(* pi1 *)
intros G T1 T2 M M' IH E b Hwf HintE Hfstm.
so (leqm_reg _#5 Hwf IH) as (HofM & HofM').
invert Hfstm; clear Hfstm; [].
intros c HfstM <-.
so (ofm_reg _#4 Hwf HofM) as Hofpi.
so (can_fstm _#3 Hwf HofM') as (c' & HfstM').
invert Hofpi; clear Hofpi; [].
intros k1 HofT1 Hfst1 HofT2.
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
so (IH Hwf anderrr _ _ HintE HfstM) as Hint.
clear IH.
unfold mpuev in Hint; unfold mpuev.
simp_sub.
eapply (monad_mevp_mevp (subst (left E) (sg_sigma T1 T2)) (subst (right E) (sg_sigma T1 T2)) _#8 md_pi1 md_pi1); eauto using subst_fsts, subst_fstm, subof_dof_tof, left_subof_dof, right_subof_dof with static dynamic; [].
intros V V' d d' HfstV HfstV' Heqd Heqd' HV.
simpl in HV.
invert HV; [].
intros V1 V1' V2 V2' d1 d1' k1a k2a P HofV HofV' HvalV1 HvalV2 HvalV1' HvalV2' HfstV1 HfstV1' Hfst1a Hfst2a HV1 HV2 <- <- Heqcand.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a; clear Hfst2a.
invert HfstV; clear HfstV; [].
intros d1a d2 HfstV1a HfstV2 <-.
invert HfstV'; clear HfstV'; [].
intros d1a' d2' HfstV1a' HfstV2' <-.
so (fstm_fun _#4 HfstV1 HfstV1a); subst d1a; clear HfstV1a.
so (fstm_fun _#4 HfstV1' HfstV1a'); subst d1a'; clear HfstV1a'.
so (mvalue_pure _#4 (mvalue_pair HvalV1 HvalV2) HofV) as H.
renameover H into HofV.
so (mvalue_pure _#4 (mvalue_pair HvalV1' HvalV2') HofV') as H.
renameover H into HofV'.
simp_sub in HofV; simp_sub in HofV'.
apply (mevp_intro_steps _#4 V1 V1' _); eauto using star_one with static dynamic; [].
simpl.
rewrite <- Heqcand; [].
exact HV1.

(* pi2 *)
intros G T1 T2 M M' c IH HfstM E b Hwf HintE Hfstm.
so (leqm_reg _#5 Hwf IH) as (HofM & HofM').
invert Hfstm; clear Hfstm; [].
intros ca HfstMa <-.
so (fstm_fun _#4 HfstM HfstMa); subst ca; clear HfstMa.
so (can_fstm _#3 Hwf HofM') as (c' & HfstM').
so (ofm_reg _#4 Hwf HofM) as Hofpi.
invert Hofpi; clear Hofpi; [].
intros k1 HofT1 Hfst1 HofT2.
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
so (ofs_fst _#3 HofT2 Hfst2) as Hofk2.
so (leqm_fst' _#7 Hwf IH (fsts_sigma Hfst1 Hfst2) HfstM HfstM') as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & _).
so (IH Hwf anderrr _ _ HintE HfstM) as Hint.
clear IH.
unfold mpuev in Hint; unfold mpuev.
simp_sub.
#3 eapply (monad_mevp_mevp (subst (left E) (sg_sigma T1 T2)) (subst (right E) (sg_sigma T1 T2)) _#8 md_pi2 md_pi2); eauto using subst_fsts, subst_fstm, subof_dof_tof, left_subof_dof, right_subof_dof with static dynamic.
  intros m Hof Hfstm.
  simp_sub in Hof.
  #2 eapply rel_quest.
    eapply ofm_pi2; eauto; done.
    
    simp_sub.
    reflexivity.

  intros m Hof Hfstm.
  simp_sub in Hof.
  #2 eapply ofm_subsume.
    eapply ofm_pi2; eauto; done.
    
    simp_sub.
    so (subst_eqs _#5 (right_subof _#3 (fintctx_stat _ _ Hwf HintE)) (funct_sg _#5 Hwf (eqc_pi1 Heqc) HofT2)) as H.
    simp_sub in H.
    auto with static; done.
intros V V' d d' HfstV HfstV' Heqd Heqd' HV.
simpl in HV.
invert HV; [].
intros V1 V1' V2 V2' d1 d1' k1a k2a P HofV HofV' HvalV1 HvalV2 HvalV1' HvalV2' HfstV1 HfstV1' Hfst1a Hfst2a HV1 HV2 <- <- Heqcand.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a; clear Hfst2a.
invert HfstV; clear HfstV; [].
intros d1a d2 HfstV1a HfstV2 <-.
invert HfstV'; clear HfstV'; [].
intros d1a' d2' HfstV1a' HfstV2' <-.
so (fstm_fun _#4 HfstV1 HfstV1a); subst d1a; clear HfstV1a.
so (fstm_fun _#4 HfstV1' HfstV1a'); subst d1a'; clear HfstV1a'.
so (mvalue_pure _#4 (mvalue_pair HvalV1 HvalV2) HofV) as H.
renameover H into HofV.
so (mvalue_pure _#4 (mvalue_pair HvalV1' HvalV2') HofV') as H.
renameover H into HofV'.
simp_sub in HofV; simp_sub in HofV'.
so (Inversion.pair_ofm_inversion_sigma_pure' _#6 wf_nil HofV HfstV1) as (HofV1 & HofV2).
so (Inversion.pair_ofm_inversion_sigma_pure' _#6 wf_nil HofV' HfstV1') as (HofV1' & HofV2').
#3 apply (mevp_intro_steps _#4 V2 V2' _); eauto using star_one with dynamic.
  #2 eapply ofm_subsume.
    eapply ofm_pi2; eauto with static; done.

    simp_sub.
    apply subs_refl; [].
    rewrite -> !(split_dot_compose _ (left E)); [].
    eapply funct_sg; eauto using wf_nil, subst_ofs, subof_bind_sof_cn, left_subof, fintctx_stat; [].
    eapply eqc_pi1; eauto; [].
    apply eqc_symm; [].
    simp_sub in Heqd.
    exact Heqd.

  #2 eapply ofm_subsume.
    eapply ofm_pi2; eauto with static; done.

    simp_sub.
    apply subs_refl; [].
    rewrite -> !(split_dot_compose _ (right E)); [].
    eapply funct_sg; eauto using wf_nil, subst_ofs, subof_bind_sof_cn, right_subof, fintctx_stat; [].
    eapply eqc_pi1; eauto; [].
    apply eqc_symm; [].
    simp_sub in Heqd'.
    eapply eqc_trans; eauto; [].
    so (subst_eqc _#6 (right_subof _#3 (fintctx_stat _ _ Hwf HintE)) Heqc) as H.
    simp_sub in H.
    exact H.
unfold subst1; rewrite -> (subst_ints _#5 (ofs_fsg _ _ HofT2) (fenv_cn (fintctx_fenv _ _ HintE)) (subenv_dot _#5 (fcn_pi1 (ofc_fcn _#3 Hofc)) (fintctx_fenv _ _ HintE) (subenv_id _))); [].
simpl intc.
rewrite <- Heqcand; [].
rewrite -> (ints_fun (G; cl_cn k1) _ _ (E; Sta d1 d1' (Cand (simp k1) (fst P)))); eauto using wf; [].
#3 apply intctx_cn; eauto using fintctx_stat.
  simp_sub.
  #2 apply (@eqc_trans _ _ (cn_pi1 (cn_pair d1 d2))).
    eapply eqc_pi1; eauto; done.

    #2 eapply eqc_beta1.
      eapply (ofm_fst _ _ V1); eauto using wf, subst_fsts; done.

      eapply (ofm_fst _ _ V2); eauto using wf, subst_fsts; done.

  simp_sub.
  #2 apply (@eqc_trans _ _ (cn_pi1 (subst (right E) c'))).
    exact (eqc_pi1 (subst_eqc _#6 (right_subof _#3 (fintctx_stat _ _ Hwf HintE)) Heqc)).

    #2 apply (@eqc_trans _ _ (cn_pi1 (cn_pair d1' d2'))).
      eapply eqc_pi1; eauto; done.
  
      #2 eapply eqc_beta1.
        eapply (ofm_fst _ _ V1'); eauto using wf, subst_fsts; done.
  
        eapply (ofm_fst _ _ V2'); eauto using wf, subst_fsts; done.

  so (intc_fun _#5 Hwf Hofc (fintctx_stat _ _ Hwf HintE)) as Hintc.
  invert Hintc; [].
  intros P1 P2 _ _ Hintc1 _ Heqcand1 Heqcand2.
  repl (Cand (simp k1) (fst P)) with (intc (cn_pi1 c) E).
    simpl intc.
    rewrite <- Heqcand; [].
    reflexivity.
  repl (Cand (simp k1) (fst P1)) with (intc (cn_pi1 c) E) in Hintc1.
    simpl intc.
    rewrite <- Heqcand1; [].
    reflexivity.
  repl (Cand (simp k1) (fst P2)) with (intc (cn_pi1 c) E) in Hintc1.
    simpl intc.
    rewrite <- Heqcand2; [].
    reflexivity.
  simp_sub.
  exact Hintc1.

(* unpack *)
intros G k t T T' e e' M M' IH1 IH2 Hsame E Hwf HintE.
invert Hsame; clear Hsame; [].
intros HofT <-.
so (leqt_reg _#4 Hwf IH1) as (Hofe & Hofe').
so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hofe)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _; clear l Hofcr.
so (leqm_reg _#5 (wf_tm (wf_cn Hwf Hofk) Hoft) IH2) as (HofM & HofM').
so (IH1 Hwf anderr _ HintE) as Hinte; clear IH1.
rewrite -> intt_exists_unroll in Hinte; [].
unfold exists_interp_st in Hinte.
unfold miev, mi.
simp_sub in Hinte.
simp_sub.
assert (closed 2 (subst (under 2 (left E)) M)) as HclM.
  so (fmd_closed _ _ (ofm_fmd _#4 fcontext_nil (subst_ofm _#6 (left_subof_dof _ _ Hwf HintE) (ofm_unpack Hofe HofM HofT)))) as H.
  simpl in H.
  unfold closed in H.
  simp_sub in H.
  unfold closed.
  simp_sub.
  injection H; auto; done.
assert (closed 2 (subst (under 2 (right E)) M')) as HclM'.
  so (fmd_closed _ _ (ofm_fmd _#4 fcontext_nil (subst_ofm _#6 (right_subof_dof _ _ Hwf HintE) (ofm_unpack Hofe' HofM' HofT)))) as H.
  simpl in H.
  unfold closed in H.
  simp_sub in H.
  unfold closed.
  simp_sub.
  injection H; auto; done.
assert (forall s, subof dof nil s G -> closed 0 (subst s T)) as HclT.
  intros s Hs.
  exact (fsg_closed _ _ (ofs_fsg _ _ (subst_ofs _#4 (subof_dof_sof _#3 Hs) HofT))).
#8 eapply (monad_tpstev_mev_mpclosed _#6 (fun z => md_unpack z _ _) (fun z => md_unpack z _ _)); eauto with dynamic.
  apply exists_interp_tvrel; done.

  eapply mpst_mvrel; eauto using fintctx_stat; done.

  apply mpst_idem; done.

  intros s m.
  simp_sub.
  #2 f_equal.
    so (subst_closed _ _ s HclM) as H.
    simp_sub in H.
    exact H.

    so (subst_closed _ _ s (HclT (left E) (left_subof_dof _ _ Hwf HintE))) as H.
    simp_sub in H.
    exact H.

  intros s m.
  simp_sub.
  #2 f_equal.
    so (subst_closed _ _ s HclM') as H.
    simp_sub in H.
    exact H.

    so (subst_closed _ _ s (HclT (right E) (right_subof_dof _ _ Hwf HintE))) as H.
    simp_sub in H.
    exact H.

  intros G' i m Heqi Hof.
  assert (subof dof G' (compose (left E) (sh i)) G) as Hsubof.
    eapply subof_dof_trans; eauto using left_subof_dof; [].
    subst i; auto using subof_sh, truncate_all; done.
  simp_sub.
  simp_sub in Hof.
  so (subst_closed _ _ (sh i) (HclT (left E) (left_subof_dof _ _ Hwf HintE))) as HeqT.
  simp_sub in HeqT.
  rewrite <- HeqT; [].
  #3 eapply ofm_unpack.
    exact Hof.

    simp_sub.
    so (subst_closed _ _ (sh i) HclM) as Heq.
    simp_sub in Heq.
    rewrite <- Heq; [].
    so (subst_ofm _#6 (subof_bind_dof_cn_tm _#5 Hsubof) HofM) as H.
    simp_sub in H.
    exact H.

    eapply subst_ofs; eauto using subof_dof_sof; done.

  intros G' i m Heqi Hof.
  assert (subof dof G' (compose (right E) (sh i)) G) as Hsubof.
    eapply subof_dof_trans; eauto using right_subof_dof; [].
    subst i; auto using subof_sh, truncate_all; done.
  simp_sub.
  simp_sub in Hof.
  so (subst_closed _ _ (sh i) (HclT (right E) (right_subof_dof _ _ Hwf HintE))) as HeqT.
  simp_sub in HeqT.
  rewrite <- HeqT; [].
  #3 eapply ofm_unpack.
    exact Hof.

    simp_sub.
    so (subst_closed _ _ (sh i) HclM') as Heq.
    simp_sub in Heq.
    rewrite <- Heq; [].
    so (subst_ofm _#6 (subof_bind_dof_cn_tm _#5 Hsubof) HofM') as H.
    simp_sub in H.
    exact H.

    eapply subst_ofs; eauto using subof_dof_sof; done.

  intros w w' Hintw.
  simp_sub.
  invert Hintw; [].
  intros u u' c c' l l' s s' Q Hofw Hofw' Hvalw Hvalw' Hintc Hintv <- <-.
  clear Hintw.
  so (intt_tvrel _#3 (wf_cn Hwf Hofk) Hoft (fintctx_stat _ _ (wf_cn Hwf Hofk) (fintctx_cn HintE Hintc)) _ _ Hintv) as (_ & _ & Hvalu & Hvalu').
  exploit (fun H => IH2 H anderr (E; Sta c c' Q; Dyn u u')) as HintM; auto using wf, fintctx; [].
  simpl in HintM; simp_sub in HintM.
  rewrite -> (subst_ints _ _ E _ _ (ofs_fsg _ _ HofT) (fintctx_fenv _ _ HintE)) in HintM; auto using subenv_sh with sequence; [].
  #4 eapply anti_msteps_mev; eauto.
    simp_sub in Hofw.
    #2 eapply ofm_unpack; eauto.
      simp_sub.
      replsub (subst (compose (left E) (sh 2)) T) with (subst (dot (var 0) (dot (var 1) (compose (left E) (sh 2)))) (subst (sh 2) T)).
      eapply subst_ofm; eauto; [].
      apply subof_bind_dof_cn_tm; auto using left_subof_dof; done.

      eapply subst_ofs; eauto using left_subof, fintctx_stat; done.

    simp_sub in Hofw.
    #2 eapply ofm_unpack; eauto.
      simp_sub.
      replsub (subst (compose (right E) (sh 2)) T) with (subst (dot (var 0) (dot (var 1) (compose (right E) (sh 2)))) (subst (sh 2) T)).
      eapply subst_ofm; eauto; [].
      apply subof_bind_dof_cn_tm; auto using right_subof_dof; done.

      eapply subst_ofs; eauto using right_subof, fintctx_stat; done.

    #2 eapply star_step.
      apply mstep_unpack2; auto; done.

      simp_sub.
      apply star_refl; done.

    #2 eapply star_step.
      apply mstep_unpack2; auto; done.

      simp_sub.
      apply star_refl; done.

(* lett *)
intros G p t T e e' M M' IH1 IH2.
destruct p.
  (* pure *)
  intros E c Hwf HintE Hfstm.
  so (leqt_reg _#4 Hwf IH1) as (Hofe & Hofe').
  so (oft_reg _#3 Hwf Hofe) as Hoft.
  so (leqm_reg _#5 (wf_tm Hwf Hoft) IH2) as (HofM & HofM').
  invert Hfstm; clear Hfstm; [].
  intros HfstM.
  so (subst_fstm _#5 (subof_tm_utm_tof _ t) HfstM) as H.
  simp_sub in H.
  renameover H into HfstM.
  so (ofm_reg _#4 Hwf (ofm_lett Hofe HofM)) as HofT.
  so (IH1 Hwf anderr _ HintE) as Hinte; clear IH1.
  unfold mpuev.
  simp_sub.
  #3 eapply (monad_tev_mevp _#6 (fun z => md_lett z _) (fun z => md_lett z _)); eauto with dynamic.
    intros m Hof.
    eapply ofm_lett; eauto; [].
    so (subst_ofm _#6 (subof_bind_dof_tm _#4 (left_subof_dof _ _ Hwf HintE)) HofM) as H.
    simp_sub in H.
    simp_sub.
    assumption.

    intros m Hof.
    eapply ofm_lett; eauto; [].
    so (subst_ofm _#6 (subof_bind_dof_tm _#4 (right_subof_dof _ _ Hwf HintE)) HofM') as H.
    simp_sub in H.
    simp_sub.
    assumption.

    intros v v' Hintv1.
    clear e Hofe Hinte.
    so (intt_tvrel _#3 Hwf Hoft (fintctx_stat _ _ Hwf HintE) _ _ Hintv1) as (Hofv1 & Hofv1' & Hvalv1 & Hvalv1').
    #5 eapply (anti_msteps_mevp _#5 (subst (dot v (left E)) M) (subst (dot v' (right E)) M')).
      eapply ofm_lett; eauto; [].
      so (subst_ofm _#6 (subof_bind_dof_tm _#4 (left_subof_dof _ _ Hwf HintE)) HofM) as H.
      simp_sub in H.
      simp_sub.
      assumption.
  
      eapply ofm_lett; eauto; [].
      so (subst_ofm _#6 (subof_bind_dof_tm _#4 (right_subof_dof _ _ Hwf HintE)) HofM') as H.
      simp_sub in H.
      simp_sub.
      assumption.
  
      #2 eapply star_step.
        apply mstep_lett2; auto; done.
  
        simp_sub.
        apply star_refl; done.
  
      #2 eapply star_step.
        apply mstep_lett2; auto; done.
  
        simp_sub.
        apply star_refl; done.
  
      #4 exploit (fun H => IH2 H anderrr (E; Dyn v v') (shift1 c)) as HintM.
        apply wf_tm; auto; done.
  
        apply fintctx_tm; auto; done.

        exact HfstM.

        simpl in HintM.
        simp_sub in HintM.
        unfold mpuev in HintM.
        unfold shift1 in HintM; rewrite -> (subst_ints _#5 (ofs_fsg _ _ HofT) (fintctx_fenv _ _ HintE) (subenv_sh (truncate_S truncate_0))) in HintM; [].
        so (can_fsts _ _ HofT) as (k & Hfst).
        so (subst_ofc _#5 (unshift_tm _ _) (ofm_fst _#6 (wf_tm Hwf Hoft) HofM HfstM (subst_fsts _#3 Hfst))) as Hofc.
        simp_sub in Hofc.
        rewrite -> (subst_intc _#5 (ofc_fcn _#3 Hofc) (fintctx_fenv _ _ HintE) (subenv_sh (truncate_S truncate_0))) in HintM; [].
        exact HintM.

  (* impure *)
  intros E Hwf HintE.
  so (leqt_reg _#4 Hwf IH1) as (Hofe & Hofe').
  so (oft_reg _#3 Hwf Hofe) as Hoft.
  so (leqm_reg _#5 (wf_tm Hwf Hoft) IH2) as (HofM & HofM').
  so (ofm_reg _#4 Hwf (ofm_lett Hofe HofM)) as HofT.
  so (IH1 Hwf anderr _ HintE) as Hinte; clear IH1.
  unfold miev.
  simp_sub.
  #3 eapply (monad_tev_mev _#6 (fun z => md_lett z _) (fun z => md_lett z _)); eauto with dynamic.
    intros m Hof.
    eapply ofm_lett; eauto; [].
    so (subst_ofm _#6 (subof_bind_dof_tm _#4 (left_subof_dof _ _ Hwf HintE)) HofM) as H.
    simp_sub in H.
    simp_sub.
    assumption.

    intros m Hof.
    eapply ofm_lett; eauto; [].
    so (subst_ofm _#6 (subof_bind_dof_tm _#4 (right_subof_dof _ _ Hwf HintE)) HofM') as H.
    simp_sub in H.
    simp_sub.
    assumption.

    intros v v' Hintv1.
    clear e Hofe Hinte.
    so (intt_tvrel _#3 Hwf Hoft (fintctx_stat _ _ Hwf HintE) _ _ Hintv1) as (Hofv1 & Hofv1' & Hvalv1 & Hvalv1').
    #5 eapply (anti_msteps_mev _#5 (subst (dot v (left E)) M) (subst (dot v' (right E)) M')).
      eapply ofm_lett; eauto; [].
      so (subst_ofm _#6 (subof_bind_dof_tm _#4 (left_subof_dof _ _ Hwf HintE)) HofM) as H.
      simp_sub in H.
      simp_sub.
      assumption.
  
      eapply ofm_lett; eauto; [].
      so (subst_ofm _#6 (subof_bind_dof_tm _#4 (right_subof_dof _ _ Hwf HintE)) HofM') as H.
      simp_sub in H.
      simp_sub.
      assumption.
  
      #2 eapply star_step.
        apply mstep_lett2; auto; done.
  
        simp_sub.
        apply star_refl; done.
  
      #2 eapply star_step.
        apply mstep_lett2; auto; done.
  
        simp_sub.
        apply star_refl; done.
  
      #3 exploit (fun H => IH2 H anderr (E; Dyn v v')) as HintM.
        apply wf_tm; auto; done.
  
        apply fintctx_tm; auto; done.

        simpl in HintM.
        simp_sub in HintM.
        unfold miev in HintM.
        unfold shift1 in HintM; rewrite -> (subst_ints _#5 (ofs_fsg _ _ HofT) (fintctx_fenv _ _ HintE) (subenv_sh (truncate_S truncate_0))) in HintM; [].
        exact HintM.

(* letm *)
intros G T1 T2 T2' k1 M1 M1' M2 M2' IH1 Hfst1 IH2 Hsame E Hwf HintE.
invert Hsame; clear Hsame; [].
intros HofT2 <-.
so (leqm_reg _#5 Hwf IH1) as (HofM1 & HofM1').
so (ofm_reg _#4 Hwf HofM1) as HofT1.
so (leqm_reg _#5 (wf_md Hwf HofT1 Hfst1) IH2) as (HofM2 & HofM2').
so (IH1 Hwf anderr _ HintE) as HintM1; clear IH1.
simp_sub.
unfold miev, mi in HintM1.
assert (closed 2 (subst (under 2 (left E)) M2)) as HclM2.
  so (fmd_closed _ _ (ofm_fmd _#4 fcontext_nil (subst_ofm _#6 (left_subof_dof _ _ Hwf HintE) (ofm_letm HofM1 Hfst1 HofM2 HofT2)))) as H.
  simpl in H.
  unfold closed in H.
  simp_sub in H.
  unfold closed.
  simp_sub.
  injection H; auto; done.
assert (closed 2 (subst (under 2 (right E)) M2')) as HclM2'.
  so (fmd_closed _ _ (ofm_fmd _#4 fcontext_nil (subst_ofm _#6 (right_subof_dof _ _ Hwf HintE) (ofm_letm HofM1' Hfst1 HofM2' HofT2)))) as H.
  simpl in H.
  unfold closed in H.
  simp_sub in H.
  unfold closed.
  simp_sub.
  injection H; auto; done.
assert (forall s, subof dof nil s G -> closed 0 (subst s T2)) as HclT2.
  intros s Hs.
  exact (fsg_closed _ _ (ofs_fsg _ _ (subst_ofs _#4 (subof_dof_sof _#3 Hs) HofT2))).
unfold miev, mi.
#8 eapply (monad_mpstev_mev_mpclosed _#6 (fun z => md_letm z _ _) (fun z => md_letm z _ _)); eauto with dynamic; clear M1 HofM1 HintM1.
  intros W W' HW.
  destruct HW as (Q & HintW).
  so (ints_reg _#5 HintW) as (HofW & HofW').
  so (ints_val _#6 Hwf HofT1 (fintctx_stat _ _ Hwf HintE) HintW) as (HvalW & HvalW').
  auto with static; done.

  apply mpst_mvrel; done.

  apply mpst_idem; done.

  intros s M.
  simp_sub.
  #2 f_equal.
    so (subst_closed _ _ s HclM2) as H.
    simp_sub in H.
    exact H.

    so (subst_closed _ _ s (HclT2 _ (left_subof_dof _ _ Hwf HintE))) as H.
    simp_sub in H.
    exact H.

  intros s M.
  simp_sub.
  #2 f_equal.
    so (subst_closed _ _ s HclM2') as H.
    simp_sub in H.
    exact H.

    so (subst_closed _ _ s (HclT2 _ (right_subof_dof _ _ Hwf HintE))) as H.
    simp_sub in H.
    exact H.

  intros G' i M Heqi Hof.
  assert (subof dof G' (compose (left E) (sh i)) G) as Hsubof.
    eapply subof_dof_trans; eauto using left_subof_dof; [].
    subst i; auto using subof_sh, truncate_all; done.
  simp_sub.
  simp_sub in Hof.
  so (subst_closed _ _ (sh i) (HclT2 (left E) (left_subof_dof _ _ Hwf HintE))) as HeqT2.
  simp_sub in HeqT2.
  rewrite <- HeqT2; [].
  #4 eapply ofm_letm.
    exact Hof.

    apply subst_fsts; eauto; done.

    simp_sub.
    so (subst_closed _ _ (sh i) HclM2) as Heq.
    simp_sub in Heq.
    rewrite <- Heq; [].
    so (subst_ofm _#6 (subof_bind_dof_cn_md _#5 Hsubof) HofM2) as H.
    simp_sub in H.
    exact H.

    eapply subst_ofs; eauto using subof_dof_sof; done.

  intros G' i M Heqi Hof.
  assert (subof dof G' (compose (right E) (sh i)) G) as Hsubof.
    eapply subof_dof_trans; eauto using right_subof_dof; [].
    subst i; auto using subof_sh, truncate_all; done.
  simp_sub.
  simp_sub in Hof.
  so (subst_closed _ _ (sh i) (HclT2 (right E) (right_subof_dof _ _ Hwf HintE))) as HeqT2.
  simp_sub in HeqT2.
  rewrite <- HeqT2; [].
  #4 eapply ofm_letm.
    exact Hof.

    apply subst_fsts; eauto; done.

    simp_sub.
    so (subst_closed _ _ (sh i) HclM2') as Heq.
    simp_sub in Heq.
    rewrite <- Heq; [].
    so (subst_ofm _#6 (subof_bind_dof_cn_md _#5 Hsubof) HofM2') as H.
    simp_sub in H.
    exact H.

    eapply subst_ofs; eauto using subof_dof_sof; done.

  intros V V' HV.
  destruct HV as (Q & HintV).
  so (ints_reg _#5 HintV) as (HofV & HofV').
  so (ints_val _#6 Hwf HofT1 (fintctx_stat _ _ Hwf HintE) HintV) as (HvalV & HvalV').
  so (can_fstm _#3 wf_nil HofV) as (c & HfstV).
  so (can_fstm _#3 wf_nil HofV') as (c' & HfstV').
  #5 apply (anti_msteps_mev _#5 (subst (dot V (dot c (left E))) M2) (subst (dot V' (dot c' (right E))) M2')).
    eapply ofm_letm; eauto using ofm_forget, subst_fsts, subst_ofs, left_subof, fintctx_stat; [].
    so (subst_ofm _#6 (subof_bind_dof_cn_md _#5 (left_subof_dof _ _ Hwf HintE)) HofM2) as H.
    simp_sub in H.
    simp_sub.
    assumption.

    eapply ofm_letm; eauto using ofm_forget, subst_fsts, subst_ofs, right_subof, fintctx_stat; [].
    so (subst_ofm _#6 (subof_bind_dof_cn_md _#5 (right_subof_dof _ _ Hwf HintE)) HofM2') as H.
    simp_sub in H.
    simp_sub.
    assumption.

    #2 eapply star_step.
      eapply mstep_letm2; eauto; done.

      simp_sub.
      apply star_refl; done.

    #2 eapply star_step.
      eapply mstep_letm2; eauto; done.

      simp_sub.
      apply star_refl; done.

    exploit (fun H => IH2 H anderr (E; Sta c c' Q; Dyn V V')) as Hinte; auto using wf, fintctx; [].
    simpl in Hinte; simp_sub in Hinte.
    rewrite -> (subst_ints _ _ E _ _ (ofs_fsg _#2 HofT2) (fintctx_fenv _ _ HintE)) in Hinte; auto using subenv_sh with sequence; done.

(* seal *)
intros G T T' M M' IH Hsame E Hwf HintE.
invert Hsame; clear Hsame; [].
intros HofT <-.
so (leqm_reg _#5 Hwf IH) as (HofM & HofM').
so (IH Hwf anderr _ HintE) as Hint; clear IH.
simp_sub.
unfold miev in Hint; unfold miev.
eapply anti_msteps_mev; eauto using star_one, subst_ofm, left_subof_dof, right_subof_dof with static dynamic; done.

(* forget *)
intros G T M M' IH E Hwf HintE.
so (leqm_reg _#5 Hwf IH) as (HofM & HofM').
so (can_fstm _#3 Hwf HofM) as (c & HfstM).
so (IH Hwf anderrr _ _ HintE HfstM) as Hint; clear IH.
unfold mpuev in Hint; unfold miev.
destruct Hint as (Hhaltiff & Hrel & Hof & Hof').
do2 3 split; auto using ofm_forget; [].
intros V V' Hsteps Hsteps' Hval Hval'.
so (Hrel _ _ Hsteps Hsteps' Hval Hval') as HintV.
unfold mi.
eapply mpst_expand; eauto; [].
intros W W' (Q & HW).
so (ints_reg _#5 HW) as (HofW & HofW').
so (ints_val _#6 Hwf (ofm_reg _#4 Hwf HofM) (fintctx_stat _ _ Hwf HintE) HW) as (HvalW & HvalW').
auto using ofm_forget; done.

(* subsume *)
intros G p T T' M M' IH Hsubs.
destruct p.
  (* pure *)
  intros E c Hwf HintE HfstM.
  so (leqm_reg _#5 Hwf IH) as (HofM & HofM').
  so (IH Hwf anderrr _ _ HintE HfstM) as Hint.
  unfold mpuev in Hint; unfold mpuev.
  apply (mevp_mono_sg (subst (left E) T) (subst (right E) T)); eauto using subst_subs, left_subof, right_subof, fintctx_stat; [].
  eapply mevp_mono; eauto; [].
  intros N N' HintN.
  eapply subs_ints; eauto using fintctx_stat; done.

  (* impure *)
  intros E Hwf HintE.
  so (leqm_reg _#5 Hwf IH) as (HofM & HofM').
  so (IH Hwf anderr _ HintE) as Hint.
  unfold miev in Hint; unfold miev.
  apply (mev_mono_sg (subst (left E) T) (subst (right E) T)); eauto using subst_subs, left_subof, right_subof, fintctx_stat; [].
  apply (mev_mono _ _ (mi (subst (left E) T) (subst (right E) T) (ints T E))); auto; [].
  intros N N' HintN.
  unfold mi in HintN; unfold mi.
  eapply mpst_mono_sg; eauto using subst_subs, left_subof, right_subof, fintctx_stat; [].
  eapply mpst_mono; eauto; [].
  intros W W' (Q & HintW).
  exists Q.
  eapply subs_ints; eauto using fintctx_stat; done.
Qed.


Lemma leq_compat :
  compatible samek samec sames leqt leqm.
Proof.
split.
  intros; apply leqt_comp; auto; done.

  intros; apply leqm_comp; auto; done.
Qed.


Lemma leqt_with_wf :
  forall G e e' t,
    (wf G -> leqt G e e' t)
    -> leqt G e e' t.
Proof.
intros G e e' t H Hwf.
exact (H Hwf Hwf).
Qed.


Lemma leqm_with_wf :
  forall G p M M' T,
    (wf G -> leqm G p M M' T)
    -> leqm G p M M' T.
Proof.
intros G p M M' T H Hwf.
exact (H Hwf Hwf).
Qed.


Theorem fundamental :
  (forall G e t,
     oft G e t
     -> leqt G e e t)
  /\
  (forall G p M T,
     ofm G p M T
     -> leqm G p M M T).
Proof.
assert (forall G p M T,
          ofm G p M T
          -> (match p with
              | pure =>
                  wf G
                  -> forall E c,
                       fintctx G E
                       -> fstm G M c
                       -> mpuev (subst (left E) T) (subst (right E) T) (ints T E) (subst (left E) M) (subst (right E) M) (intc c E)
   
              | impure =>
                  wf G
                  -> forall E,
                       fintctx G E
                       -> miev (subst (left E) T) (subst (right E) T) (ints T E) (subst (left E) M) (subst (right E) M)
           end)
          -> leqm G p M M T) as Hdom.
  intros G p M T Hof H.
  destruct p.
    intro Hwf.
    do2 2 split; auto with static; [].
    split.
      intros k c c' Hfst HfstM HfstM'.
      so (fstm_fun _#4 HfstM HfstM'); subst c'.
      so (ofm_fst _#6 Hwf Hof HfstM Hfst).
      apply eqc_refl; auto; done.

      apply H; auto; done.

    intro Hwf.
    do2 2 split; auto with static; done.
#9 exploit
  (dynamic_ind
     (fun G e t => leqt G e e t)
     (fun G p M T => leqm G p M M T)) as Hind;
  try (intros;
       first [apply leqt_with_wf | apply leqm_with_wf];
       intro Hwf;
       repeat
         (match goal with
          | H : ofc _ _ _ |- _ => so (ofc_reg _#3 Hwf H); revert H

          | H : ofm _ _ _ _ |- _ => so (ofm_reg _#4 Hwf H); revert H
          end);
       intros;
       first [apply leqt_comp | apply leqm_comp];
       eauto using tcc, mcc with same;
       done).

(* var *)
intros G i t Hindex Hwf.
do2 2 split; auto with static; [].
intros E HintE.
remember (cl_tm t) as A eqn:HeqA.
revert t HeqA E HintE Hwf.
#2 induct Hindex.
  (* 0 *)
  intros A G t -> E HintE Hwf.
  invert HintE; [].
  intros E' v v' HintE' Hintv <-.
  invert Hwf; [].
  intros Hwf' Hoft.
  simpl.
  simp_sub.
  repl (intt (shift1 t) (E'; Dyn v v')) with (intt t E').
    symmetry.
    eapply subst_intt; eauto using ofc_fcn, fintctx_fenv, subenv with sequence; done.
  so (intt_tvrel _#3 Hwf' Hoft (fintctx_stat _ _ Hwf' HintE') _ _ Hintv) as (Hofv & Hofv' & Hval & Hval').
  apply tev_intro; auto; done.

  (* S *)
  intros i B G A Hindex IH t -> E HintE Hwf.
  destruct E as [| ent E'].
    invert HintE; done.
  so (fintctx_tail _#4 Hwf HintE) as HintE'.
  so (wf_tail _ _ Hwf) as Hwf'.
  so (IH _ (eq_refl _) E' HintE' Hwf') as Htev.
  simp_sub in Htev.
  assert (intt (subst (sh (S (S i))) t) (E'; ent) = intt (subst (sh (S i)) t) E') as Heq.
    simp_sub.
    replsub (subst (sh (2 + i)) t) with (subst sh1 (subst (sh (1 + i)) t)).
    eapply subst_intt; eauto using fintctx_fenv, subenv with sequence; [].
    eapply ofc_fcn; eauto using wf_index_tm; done.
  rewrite -> Heq; [].
  destruct ent as [m m' | c c' Q]; simpl; simp_sub; assumption.

(* unit *)
intros G.
do2 2 split; auto with static; [].
intros E HintE.
apply tev_intro; auto with static dynamic; [].
rewrite -> intt_unit_unroll; [].
apply unit_interp_; done.

(* fix *)
intros G t e Hof IH.
apply leqt_with_wf; intro Hwf; [].
so (oft_reg _#3 Hwf Hof) as Hofarr.
so (ofc_ofcr _#3 Hofarr) as (l & Hofcr).
invert Hofcr; [].
intros _ Hoft _.
apply leqt_comp; eauto using tcc with same; done.

(* var *)
intros G i T c Hindex.
apply Hdom; eauto with static; [].
intros Hwf E d HintE Hfstm.
so (fstm_fun _#4 Hfstm (fstm_var Hindex)).
subst d; clear Hfstm.
remember (cl_md T c) as A eqn:HeqA.
revert HeqA E HintE Hwf.
#2 induct Hindex.
  (* 0 *)
  intros A G -> E HintE Hwf.
  invert HintE; [].
  intros G' k U E' d d' Q V V' HintE' HintV HfstU HfstV HfstV' <- <- <- <-.
  invert Hwf; [].
  intros U' Hwf' HofU _ Heq.
  f_apply (fun z => subst1 nonsense z) in Heq as Heq'.
  simp_sub in Heq'.
  subst U'; clear Heq.
  simpl.
  simp_sub.
  rewrite -> (subst_ints _#5 (ofs_fsg _ _ HofU) (fintctx_fenv _ _ HintE') (subenv_sh (truncate_S (truncate_S truncate_0)))); [].
  unfold mpuev.
  so (ints_reg _#5 HintV) as (HofV & HofV').
  so (ints_val _#6 Hwf' HofU (fintctx_stat _ _ Hwf' HintE') HintV) as (Hval & Hval').
  apply mevp_intro; auto; done.

  (* S *)
  intros i B G A Hindex IH -> E HintE Hwf.
  destruct E as [| ent E'].
    invert HintE; done.
  so (fintctx_tail _#4 Hwf HintE) as HintE'.
  so (wf_tail _ _ Hwf) as Hwf'.
  so (IH (eq_refl _) E' HintE' Hwf') as H.
  simp_sub in H.
  so (wf_index_md _#4 Hwf' Hindex) as (HofT & Heqc).
  assert (ints (subst (sh (S (S i))) T) (E'; ent) = ints (subst (sh (S i)) T) E') as Heq.
    simp_sub.
    replsub (subst (sh (2 + i)) T) with (subst sh1 (subst (sh (1 + i)) T)).
    eapply subst_ints; eauto using fintctx_fenv, subenv with sequence; [].
    eapply ofs_fsg; eauto; done.
  rewrite -> Heq; [].
  subst c.
  destruct ent as [m m' | d d' Q]; simp_sub in H; simpl intc in H; simpl; simp_sub; exact H.

(* star *)
intros G.
apply Hdom; eauto with static; [].
intros Hwf E c HintE Hfstm.
invert Hfstm; [].
intros <-.
simpl.
simp_sub.
unfold mpuev.
apply mevp_intro; auto with static dynamic; [].
apply sg_one_interp_; auto; done.

(* extstat *)
intros G k k' M c HofM IH HfstM Hofc.
apply Hdom; eauto with static; [].
intros Hwf E ca HintE HfstMa.
so (fstm_fun _#4 HfstM HfstMa); subst ca; clear HfstMa.
so (ofc_reg _#3 Hwf Hofc) as Hofk.
so (IH Hwf anderrr _ _ HintE HfstM) as HintM; clear IH.
unfold mpuev in HintM; unfold mpuev.
invert HintM; clear HintM; [].
intros Hhaltiff (Hrel & _).
split; auto; [].
do2 2 split.
  intros V V' Hsteps Hsteps' HvalV HvalV'.
  so (subst_ofm _#6 (left_subof_dof _ _ Hwf HintE) (ofm_extstat HofM HfstM Hofc)) as HofMl.
  so (subst_ofm _#6 (right_subof_dof _ _ Hwf HintE) (ofm_extstat HofM HfstM Hofc)) as HofMr.
  simp_sub in HofMl; simp_sub in HofMr.
  so (preservations_md _#5 wf_nil HofMl Hsteps) as HofV.
  so (preservations_md _#5 wf_nil HofMr Hsteps') as HofV'.
  invert (Progress.canon_md _#3 HofV HvalV); [].
  intros d <-.
  invert (Progress.canon_md _#3 HofV' HvalV'); [].
  intros d' <-.
  apply sg_stat_interp_; [].
  so (preservations_fst _#7 wf_nil HofMl Hsteps fsts_stat (subst_fstm _#5 (subof_dof_tof _#3 (left_subof_dof _ _ Hwf HintE)) HfstM)) as (b & Hfstb & Heqb).
  so (preservations_fst _#7 wf_nil HofMr Hsteps' fsts_stat (subst_fstm _#5 (subof_dof_tof _#3 (right_subof_dof _ _ Hwf HintE)) HfstM)) as (b' & Hfstb' & Heqb').
  invert Hfstb; clear Hfstb; [].
  intros <-.
  invert Hfstb'; clear Hfstb'; [].
  intros <-.
  unfold intks.
  so (intc_fun _#5 Hwf Hofc (fintctx_stat _ _ Hwf HintE)) as Hint.
  eapply intk_defeq; eauto using fintctx_stat; done.

  exact (subst_ofm _#6 (left_subof_dof _ _ Hwf HintE) (ofm_extstat HofM HfstM Hofc)).

  exact (subst_ofm _#6 (right_subof_dof _ _ Hwf HintE) (ofm_extstat HofM HfstM Hofc)).

(* extpi *)
intros G T1 T2 T3 k1 M HofM IH1 Hfst1 Hofapp IH2.
apply Hdom; eauto with static; [].
intros Hwf E c HintE HfstM.
so (ofm_reg _#4 Hwf HofM) as Hofpi.
invert Hofpi; clear Hofpi; [].
intros k1a HofT1 Hfst1a HofT3.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (can_fsts _ _ HofT3) as (k3 & Hfst3).
so (ofm_fst _#6 Hwf HofM HfstM (fsts_piapp Hfst1 Hfst3)) as Hofcalt.
so (subst_ofs _#4 (unshift_md _#3) (ofm_reg _#4 (wf_md Hwf HofT1 Hfst1) Hofapp)) as HofT2.
simp_sub in HofT2.
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
so (ofs_fst _#3 HofT2 Hfst2) as Hofk2.
so (subst_ofc _#5 (unshift_md _#3) (ofm_fst _#6 (wf_md Hwf HofT1 Hfst1) Hofapp (fstm_appap (subst_fstm _#5 (weaken2 _#4) HfstM) (fstm_var index_0)) (subst_fsts _#3 Hfst2))) as Hofcapp.
simp_sub in Hofcapp.
so (ofc_extpi Hofcalt Hofcapp) as Hofc.
so (intc_fun _#5 Hwf Hofc (fintctx_stat _ _ Hwf HintE)) as Hintc.
so (IH1 Hwf anderrr _ _ HintE HfstM) as HintM; clear IH1.
unfold mpuev in HintM.
destruct HintM as (Hhaltiff & _ & _).
do2 3 split; eauto using subst_ofm, left_subof_dof, right_subof_dof with static; [].
intros V V' Hsteps Hsteps' HvalV HvalV'.
so (subst_ofm _#6 (left_subof_dof _ _ Hwf HintE) (ofm_extpi HofM Hfst1 Hofapp)) as HofMl.
so (subst_ofm _#6 (right_subof_dof _ _ Hwf HintE) (ofm_extpi HofM Hfst1 Hofapp)) as HofMr.
so (preservations_fst _#7 wf_nil HofMl Hsteps (subst_fsts _#3 (fsts_piapp Hfst1 Hfst2)) (subst_fstm _#5 (subof_dof_tof _#3 (left_subof_dof _ _ Hwf HintE)) HfstM)) as (d & HfstV & Heqd).
so (preservations_fst _#7 wf_nil HofMr Hsteps' (subst_fsts _#3 (fsts_piapp Hfst1 Hfst2)) (subst_fstm _#5 (subof_dof_tof _#3 (right_subof_dof _ _ Hwf HintE)) HfstM)) as (d' & HfstV' & Heqd').
simpl ints.
invert Hintc; [].
intros F _ _ _ _ Heqcand _.
rewrite <- Heqcand; [].
#4 eapply sg_piapp_interp_; eauto.
  apply ofm_forget; [].
  eapply preservations_md; eauto using wf_nil; done.

  apply ofm_forget; [].
  eapply preservations_md; eauto using wf_nil; done.

  unfold intks.
  rewrite -> Heqcand; [].
  eapply intk_defeq; eauto using fintctx_stat with static; done.

  intros W W' Q HintW.
  so (ints_reg _#5 HintW) as (HofW & HofW').
  so (ints_val _#6 Hwf HofT1 (fintctx_stat _ _ Hwf HintE) HintW) as (HvalW & HvalW').
  so (can_fstm _#3 wf_nil HofW) as (b & HfstW).
  so (can_fstm _#3 wf_nil HofW') as (b' & HfstW').
  exists b, b'.
  do2 2 split; auto; [].
  #3 exploit (fun H => IH2 H anderrr (E; Sta b b' Q; Dyn W W') (cn_app (subst (sh 2) c) (var 1))) as Hintapp; eauto using wf.
    apply fintctx_md; auto; done.

    apply fstm_appap; eauto using subst_fstm, weaken2; [].
    replsub (var 1) with (shift1 (var 0)).
    eapply fstm_var; eauto with sequence; done.

    simpl left in Hintapp; simpl right in Hintapp.
    simp_sub in Hintapp.
    rewrite -> (subst_ints _#5 (ofs_fsg _ _ HofT2) (fenv_cn (fintctx_fenv _ _ HintE)) (subenv_sh (truncate_S truncate_0))) in Hintapp; [].
    replsub (cn_app (subst (sh 2) c) (var 1)) with (subst (sh 1) (cn_app (shift1 c) (var 0))) in Hintapp.
    rewrite -> (subst_intc _#5 (ofc_fcn _#3 Hofcapp) (fenv_cn (fintctx_fenv _ _ HintE)) (subenv_sh (truncate_S truncate_0))) in Hintapp; [].
    simpl intc in Hintapp.
    rewrite -> (subst_intc _#5 (ofc_fcn _#3 Hofc) (fintctx_fenv _ _ HintE) (subenv_sh (truncate_S truncate_0))) in Hintapp; [].
    rewrite <- Heqcand in Hintapp; [].
    simp_sub in Hintapp.
    unfold mpuev in Hintapp; unfold mpuev.
    #2 eapply msteps_mevp; eauto.
      eapply (star_map' _ _ (fun z => md_appap z _)); eauto with dynamic; done.

      eapply (star_map' _ _ (fun z => md_appap z _)); eauto with dynamic; done.

Lemma mpi1_halt_iff :
  forall p M T1 T2,
    ofm nil p M (sg_sigma T1 T2)
    -> (mhalt nil M <-> mhalt nil (md_pi1 M)).
Proof.
intros p M T1 T2 Hof.
split.
  intros (V & Hsteps & Hval).
  invert (Progress.canon_md _#3 (preservations_md _#5 wf_nil Hof Hsteps) Hval); [].
  intros V1 V2 <-.
  invert Hval; [].
  intros Hval1 Hval2.
  exists V1.
  split; auto; [].
  #2 apply (star_trans _#3 (md_pi1 (md_pair V1 V2))).
    eapply star_map'; eauto with dynamic; done.

    apply star_one.
    apply mstep_pi12; auto; done.

  intro H.
  eapply mpi1_halt_inversion'; eauto; done.
Qed.

(* extsigma *)
intros G T1 T2 M Hof1 IH1 Hof2 IH2.
apply Hdom; eauto with static; [].
intros Hwf E c HintE HfstM.
so (ofm_reg _#4 Hwf Hof1) as HofT1.
so (ofm_reg _#4 Hwf Hof2) as HofT2.
so (can_fsts _ _ HofT1) as (k1 & Hfst1).
so (can_fsts _ _ HofT2) as (k2 & Hfst2).
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
so (ofs_fst _#3 HofT2 Hfst2) as Hofk2.
so (ofm_fst _#6 Hwf (ofm_extsigma Hof1 Hof2) HfstM (fsts_sigma Hfst1 (subst_fsts _#3 Hfst2))) as Hofc.
so (IH1 Hwf anderrr _ _ HintE (fstm_pi1 HfstM)) as Hint1.
so (IH2 Hwf anderrr _ _ HintE (fstm_pi2 HfstM)) as Hint2.
clear IH1 IH2.
#2 do2 3 split; eauto using subst_ofm, left_subof_dof, right_subof_dof with static.
  rewrite -> (mpi1_halt_iff _#4 (subst_ofm _#6 (left_subof_dof _ _ Hwf HintE) (ofm_extsigma Hof1 Hof2))); [].
  rewrite -> (mpi1_halt_iff _#4 (subst_ofm _#6 (right_subof_dof _ _ Hwf HintE) (ofm_extsigma Hof1 Hof2))); [].
  exact (Hint1 andel).

  intros V V' Hsteps Hsteps' HvalV HvalV'.
  invert (Progress.canon_md _#3 (preservations_md _#5 wf_nil (subst_ofm _#6 (left_subof_dof _ _ Hwf HintE) (ofm_extsigma Hof1 Hof2)) Hsteps) HvalV); [].
  intros V1 V2 <-.
  invert (Progress.canon_md _#3 (preservations_md _#5 wf_nil (subst_ofm _#6 (right_subof_dof _ _ Hwf HintE) (ofm_extsigma Hof1 Hof2)) Hsteps') HvalV'); [].
  intros V1' V2' <-.
  simp_sub in Hint1; simp_sub in Hint2.
  invert HvalV; [].
  intros HvalV1 HvalV2.
  invert HvalV'; [].
  intros HvalV1' HvalV2'.
  so (preservations_md _#5 wf_nil (subst_ofm _#6 (left_subof_dof _ _ Hwf HintE) (ofm_extsigma Hof1 Hof2)) Hsteps) as HofV.
  so (preservations_md _#5 wf_nil (subst_ofm _#6 (right_subof_dof _ _ Hwf HintE) (ofm_extsigma Hof1 Hof2)) Hsteps') as HofV'.
  simp_sub in HofV; simp_sub in HofV'.
  so (pair_ofm_inversion_sigma_pure _#5 wf_nil HofV) as (c1 & _ & HfstV1 & _).
  so (pair_ofm_inversion_sigma_pure _#5 wf_nil HofV') as (c1' & _ & HfstV1' & _).
  so (intc_fun _#5 Hwf Hofc (fintctx_stat _ _ Hwf HintE)) as Hintc.
  simpl intk in Hintc.
  invert Hintc; [].
  intros P _ _ _ _ _ Heqcand _.
  rewrite <- Heqcand; [].
  simpl ints.
  #4 eapply sg_sigma_interp_; eauto using subst_fsts.
    simp_sub.
    apply ofm_forget; auto; done.

    simp_sub.
    apply ofm_forget; auto; done.

    #3 exploit (Hint1 anderl V1 V1') as HintV1; auto.
      #2 apply (star_trans _#3 (md_pi1 (md_pair V1 V2))).
        eapply star_map'; eauto with dynamic; done.
  
        apply star_one; auto with dynamic; done.
  
      #2 apply (star_trans _#3 (md_pi1 (md_pair V1' V2'))).
        eapply star_map'; eauto with dynamic; done.
  
        apply star_one; auto with dynamic; done.

      simpl in HintV1.
      rewrite <- Heqcand in HintV1; [].
      exact HintV1.

    #3 exploit (Hint2 anderl V2 V2') as HintV2; auto.
      #2 apply (star_trans _#3 (md_pi2 (md_pair V1 V2))).
        eapply star_map'; eauto with dynamic; done.
  
        apply star_one; auto with dynamic; done.
  
      #2 apply (star_trans _#3 (md_pi2 (md_pair V1' V2'))).
        eapply star_map'; eauto with dynamic; done.
  
        apply star_one; auto with dynamic; done.
    
      simpl in HintV2.
      rewrite <- Heqcand in HintV2; [].
      rewrite -> (subst_ints _#5 (ofs_fsg _ _ HofT2) (fintctx_fenv _ _ HintE) (subenv_sh (truncate_S truncate_0))); [].
      exact HintV2.

(* epilogue *)
split; intros; apply Hind; auto; done.
Qed.


Lemma leqt_refl :
  forall G e t,
    oft G e t
    -> leqt G e e t.
Proof.
exact (fundamental andel).
Qed.


Lemma leqm_refl :
  forall G p M T,
    ofm G p M T
    -> leqm G p M M T.
Proof.
exact (fundamental ander).
Qed.


Lemma leq_creflexive :
  creflexive leqt leqm.
Proof.
split; auto using leqt_refl, leqm_refl.
Qed.


Lemma can_subenv :
  forall G1 G2 s E1,
    subof aof G1 s G2
    -> wf G1
    -> wf G2
    -> fintctx G1 E1
    -> exists s' E2,
         fintctx G2 E2
         /\ subenv E1 s' E2
         /\ eqsub s s'.
Proof.
intros G1 G2 s E1 Hs Hwf1 Hwf HintE1.
revert s Hs.
induct Hwf.

(* nil *)
intros s Hs.
invert Hs; [].
intros n Htrunc <-.
exists (sh n), nil.
do2 2 split.
  apply fintctx_nil; done.

  apply subenv_sh; [].
  so (truncate_length _#4 Htrunc) as Heq.
  simpl in Heq.
  replace (n + 0) with n in Heq by omega.
  rewrite -> (intctx_length _#3 (fintctx_stat _ _ Hwf1 HintE1)) in Heq; [].
  rewrite <- Heq; [].
  apply truncate_all; done.

  apply eqsub_refl; done.

(* cn *)
intros G2 k Hwf IH Hofk s Hs.
so (subof_invert_as_dot _#5 aof_behaved iscl_cn Hs) as (c & s' & Hs' & Hofm & Heq).
simp_sub in Hofm.
invert Hofm; [].
intro Hofc.
so (IH s' Hs') as (s'' & E2 & HintE2 & Hsubenv & Heq').
exists (dot c s''), (E2; Sta (subst (left E1) c) (subst (right E1) c) (intc c E1)).
do2 2 split.
  apply fintctx_cn; auto; [].
  unfold intks.
  rewrite <- (subst_intk G2 E1 E2 s'' k); auto using ofk_fkd, fintctx_fenv; [].
  rewrite -> (subst_eqsub _#3 Heq') in Hofc; [].
  apply (intc_fun G1); auto using fintctx_stat; done.

  eapply subenv_dot; eauto using fintctx_fenv, ofc_fcn; done.

  eapply eqsub_trans; eauto using eqsub_dot, eqsub_refl; done.

(* tm *)
intros G2 t Hwf IH Hoft s Hs.
so (subof_invert_as_dot _#5 aof_behaved iscl_tm Hs) as (v & s' & Hs' & Hofm & Heq).
simp_sub in Hofm.
invert Hofm; [].
intros Hofv Hval.
so (IH s' Hs') as (s'' & E2 & HintE2 & Hsubenv & Heq').
exists (dot v s''), (E2; Dyn (subst (left E1) v) (subst (right E1) v)).
do2 2 split.
  apply fintctx_tm; auto; [].
  rewrite <- (subst_intt G2 E1 E2 s'' t); eauto using ofc_fcn, fintctx_fenv; [].
  rewrite -> (subst_eqsub _#3 Heq') in Hofv; [].
  so (leqt_refl _#3 Hofv Hwf1 anderr _ HintE1) as Hint.
  destruct Hint as (_ & Hrel & _).
  exploit (Hrel (subst (left E1) v) (subst (right E1) v)); eauto using star_refl, subst_tvalue, left_subof_oof, right_subof_oof; done.

  apply subenv_empty; auto; done.

  eapply eqsub_trans; eauto using eqsub_dot, eqsub_refl; done.

(* md *)
intros G2 k T Hwf IH HofT Hfst s Hs.
so (subof_invert_as_dot _#5 aof_behaved iscl_md Hs) as (V & s1 & Hs1 & Hofm & Heq1).
so (subof_invert_as_dot _#5 aof_behaved iscl_cn Hs1) as (c & s2 & Hs2 & Hofm' & Heq2).
rewrite -> (subst_eqsub _#3 Heq2) in Hofm; [].
simp_sub in Hofm.
invert Hofm; [].
intros HofV Hval HfstV.
so (IH s2 Hs2) as (s2' & E2 & HintE2 & Hsubenv & Heq2').
exists (dot V (dot c s2')), (E2; Sta (subst (left E1) c) (subst (right E1) c) (intc c E1); Dyn (subst (left E1) V) (subst (right E1) V)).
do2 2 split.
  apply fintctx_md; eauto using subst_fstm, subof_dof_tof, left_subof_dof, right_subof_dof; [].
  rewrite <- (subst_ints G2 E1 E2 s2' T); eauto using ofs_fsg, fintctx_fenv; [].
  rewrite -> (subst_eqsub _#3 Heq2') in HofV; [].
  so (leqm_refl _#4 HofV Hwf1 anderrr _ _ HintE1 HfstV) as Hint.
  destruct Hint as (_ & Hrel & _).
  exploit (Hrel (subst (left E1) V) (subst (right E1) V)); eauto using star_refl, subst_mvalue, left_subof_oof, right_subof_oof; done.

  apply subenv_empty; [].
  eapply subenv_dot; eauto using fintctx_fenv, wf_fcontext, ofm_fmd, fstm_fcn; done.

  eauto using eqsub_trans, eqsub_dot, eqsub_refl; done.
Qed.


Lemma subst_leqt :
  forall G1 s G2 e e' t,
    subof aof G1 s G2 
    -> wf G2
    -> leqt G2 e e' t
    -> leqt G1 (subst s e) (subst s e') (subst s t).
Proof.
intros G1 s G2 e e' t Hs Hwf Heq.
so (Heq Hwf) as (Hof & Hof' & Hrel).
clear Heq.
intro Hwf1.
do2 2 split; eauto using subst_oft, subof_aof_dof; [].
intros E HintE.
so (can_subenv _#4 Hs Hwf1 Hwf HintE) as (s' & E' & HintE' & Hsubenv & Heqsub).
rewrite -> !(subst_eqsub _#3 Heqsub); [].
rewrite -> (subst_intt _#5 (ofc_fcn _#3 (oft_reg _#3 Hwf Hof)) (fintctx_fenv _ _ HintE') Hsubenv); [].
so (Hrel E' HintE') as Hint.
simp_sub.
rewrite -> (subenv_left _#3 Hsubenv); [].
rewrite -> (subenv_right _#3 Hsubenv); [].
exact Hint.
Qed.


Lemma subst_leqm :
  forall G1 s G2 p M M' T,
    subof aof G1 s G2 
    -> wf G2
    -> leqm G2 p M M' T
    -> leqm G1 p (subst s M) (subst s M') (subst s T).
Proof.
intros G1 s G2 p M M' T Hs Hwf Heq.
so (Heq Hwf) as (Hof & Hof' & Hrel).
clear Heq.
intro Hwf1.
do2 2 split; eauto using subst_ofm, subof_aof_dof; [].
destruct p.
  (* pure *)
  so (can_fsts _ _ (ofm_reg _#4 Hwf Hof)) as (l & Hfstl).
  so (can_fstm _#3 Hwf Hof) as (d & Hfstd).
  so (can_fstm _#3 Hwf Hof') as (d' & Hfstd').
  so (ofm_fst _#6 Hwf Hof Hfstd Hfstl) as Hofd.
  so (Hrel andel l d d' Hfstl Hfstd Hfstd') as Heqc.
  split.
    intros k c c' Hfst Hfstm Hfstm'.
    so (fsts_fun _#3 Hfst (subst_fsts _#3 Hfstl)); subst k.
    so (fstm_fun _#4 Hfstm (subst_fstm _#5 (subof_aof_tof _#3 Hs) Hfstd)); subst c.
    so (fstm_fun _#4 Hfstm' (subst_fstm _#5 (subof_aof_tof _#3 Hs) Hfstd')); subst c'.
    eapply subst_eqc; eauto using subof_aof_sof; done.

    intros E c HintE Hfstm.
    so (fstm_fun _#4 Hfstm (subst_fstm _#5 (subof_aof_tof _#3 Hs) Hfstd)); subst c.
    so (can_subenv _#4 Hs Hwf1 Hwf HintE) as (s' & E' & HintE' & Hsubenv & Heqsub).
    rewrite -> !(subst_eqsub _#3 Heqsub); [].
    rewrite -> (subst_ints _#5 (ofs_fsg _ _ (ofm_reg _#4 Hwf Hof)) (fintctx_fenv _ _ HintE') Hsubenv); [].
    rewrite -> (subst_intc _#5 (ofc_fcn _#3 Hofd) (fintctx_fenv _ _ HintE') Hsubenv); [].
    so (Hrel ander E' d HintE' Hfstd) as Hint.
    simp_sub.
    rewrite -> (subenv_left _#3 Hsubenv); [].
    rewrite -> (subenv_right _#3 Hsubenv); [].
    exact Hint.

  (* impure *)
  intros E HintE.
  so (can_subenv _#4 Hs Hwf1 Hwf HintE) as (s' & E' & HintE' & Hsubenv & Heqsub).
  rewrite -> !(subst_eqsub _#3 Heqsub); [].
  rewrite -> (subst_ints _#5 (ofs_fsg _ _ (ofm_reg _#4 Hwf Hof)) (fintctx_fenv _ _ HintE') Hsubenv); [].
  so (Hrel E' HintE') as Hint.
  simp_sub.
  rewrite -> (subenv_left _#3 Hsubenv); [].
  rewrite -> (subenv_right _#3 Hsubenv); [].
  exact Hint.
Qed.


Lemma leq_csubstitutive :
  csubstitutive leqt leqm.
Proof.
split; eauto using subst_leqt, subst_leqm.
Qed.


Lemma leq_quasicongruence :
  quasicongruence leqt leqm.
Proof.
do2 3 split.
  exact leq_cregular.

  exact leq_creflexive.

  exact leq_csubstitutive.

  exact leq_compat.
Qed.


Lemma leq_consistent :
  consistent leqm.
Proof.
intros M M' T Hleqm Hhalt.
so (Hleqm wf_nil anderr nil fintctx_nil andel) as H.
simp_sub in H.
apply H; auto; done.
Qed.


Lemma leq_consistent_transpose :
  consistent (fun G p M M' T => leqm G p M' M T).
Proof.
intros M M' T Hleqm Hhalt.
so (Hleqm wf_nil anderr nil fintctx_nil andel) as H.
simp_sub in H.
apply H; auto; done.
Qed.


Lemma leqt_implies_ceqt :
  forall G e e' t,
    leqt G e e' t
    -> ceqt G e e' t.
Proof.
intros G e e' t H.
eapply ceqt_univ_quasi; eauto using leq_quasicongruence, leq_consistent, leq_consistent_transpose.
Qed.


Lemma leqm_implies_ceqm :
  forall G p M M' T,
    leqm G p M M' T
    -> ceqm G p M M' T.
Proof.
intros G p M M' T H.
eapply ceqm_univ_quasi; eauto using leq_quasicongruence, leq_consistent, leq_consistent_transpose.
Qed.


Lemma ceqt_implies_leqt :
  forall G e e' t,
    ceqt G e e' t
    -> leqt G e e' t.
Proof.
intros G e e' t Hceqt.
intro Hwf.
so (ceqt_reg _#4 Hwf Hceqt) as (Hofe & Hofe').
so (oft_reg _#3 Hwf Hofe) as Hoft.
so (leqt_refl _#3 Hofe) as Hleqt.
do2 2 split; auto; [].
intros E HintE.
#5 apply (tpclosed_tev_ceqt _ _ _ (subst (left E) e) (subst (right E) e)).
  eapply intt_tvrel; eauto using fintctx_stat; done.

  eapply intt_tpclosed; eauto using fintctx_stat; done.

  eapply subst_ceqt; eauto using left_subof_aof, ceqt_refl; done.

  eapply subst_ceqt; eauto using right_subof_aof; done.

  exact (Hleqt Hwf anderr _ HintE).
Qed.


Lemma ceqm_implies_leqm :
  forall G p M M' T,
    ceqm G p M M' T
    -> leqm G p M M' T.
Proof.
intros G p M M' T Hceqm.
intro Hwf.
so (ceqm_reg _#5 Hwf Hceqm) as (HofM & HofM').
so (ofm_reg _#4 Hwf HofM) as HofT.
so (leqm_refl _#4 HofM) as Hleqm.
do2 2 split; auto; [].
destruct p.
  split.
    intros k c c' Hfst HfstM HfstM'.
    exact (ceqm_fst' _#7 Hwf Hceqm Hfst HfstM HfstM').

    intros E c HintE HfstM.
    so (Hleqm Hwf anderrr _ _ HintE HfstM) as HM.
    unfold mpuev in HM; unfold mpuev.
    #4 eapply mevp_ceqm; eauto.
      intros V V' Hint.
      so (ints_reg _#5 Hint) as (HofV & HofV').
      so (ints_val _#6 Hwf HofT (fintctx_stat _ _ Hwf HintE) Hint) as (Hval & Hval').
      auto with static; done.

      intros V1 V2 V1' V2' Hceq Hceq' HvalV1' HvalV2' Hint.
      so (can_fsts _ _ HofT) as (k & Hfst).
      so (ints_reg _#5 Hint) as (HofV & HofV').
      so (can_fstm _#3 wf_nil HofV) as (d & HfstV).
      so (can_fstm _#3 wf_nil HofV') as (d' & HfstV').
      eapply (ints_ceqm _#3 V1 V2); eauto using fintctx_stat; [].
      eapply ints_fst; eauto using fintctx_stat; done.

      apply ceqm_refl; [].
      eapply subst_ofm; eauto using left_subof_dof with static; done.

      eapply subst_ceqm; eauto using right_subof_aof; done.

  intros E HintE.
  #5 apply (mpclosed_mev_ceqm _ _ _ (subst (left E) M) (subst (right E) M)).
    unfold mi.
    apply mpst_mvrel; done.

    unfold mi.
    apply mpst_idem; done.
  
    eapply subst_ceqm; eauto using left_subof_aof, ceqm_refl; done.
  
    eapply subst_ceqm; eauto using right_subof_aof; done.
  
    exact (Hleqm Hwf anderr _ HintE).
Qed.


Theorem ceqt_is_leqt : ceqt = leqt.
Proof.
Extensionality.fextensionality 4.
intros.
Extensionality.pextensionality.
  intro.
  apply ceqt_implies_leqt; auto; done.

  intro.
  apply leqt_implies_ceqt; auto; done.
Qed.


Theorem ceqm_is_leqm : ceqm = leqm.
Proof.
Extensionality.fextensionality 5.
intros.
Extensionality.pextensionality.
  intro.
  apply ceqm_implies_leqm; auto; done.

  intro.
  apply leqm_implies_ceqm; auto; done.
Qed.
