
Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Regularity.
Require Import Declarative.
Require Import Determinism.
Require Import Principal.
Require Import Consistent.
Require Import Preservation.
Require Formation.
Require Singleton.
Require Sound.


Lemma lam_oft_inversion :
  forall G t1 e t,
    wf G
    -> oft G (tm_lam t1 e) t
    -> exists t2,
         eqc G t (cn_arrow t1 t2) kd_type
         /\ oft (G; cl_tm t1) e (shift1 t2).
Proof.
intros G t1 e t Hwf Hof.
so (eqc_refl (oft_reg _#3 Hwf Hof)) as Heqt.
remember (tm_lam t1 e) as b eqn:Heqb.
revert Hof Hwf Heqt Heqb.
generalize t at 1 3.
intros t' Hof.
induct Hof; try (intros; discriminate Heqb); eauto with static; [].
intros G t1' t2 e' Hoft1 Hofe _ Hwf Heqt Heqb.
injection Heqb.
intros; subst e' t1'.
exists t2.
split; auto; done.
Qed.


Lemma lam_oft_inversion_arrow :
  forall G t1 t2 t1' e,
    wf G
    -> oft G (tm_lam t1' e) (cn_arrow t1 t2)
    -> eqc G t1' t1 kd_type
       /\ oft (G; cl_tm t1) e (shift1 t2).
Proof.
intros G t1 t2 t1' e Hwf Hof.
so (lam_oft_inversion _#4 Hwf Hof) as (t2' & Heqc & Hofe).
so (arrow_injective _#5 Hwf Heqc) as (Heq1 & Heq2).
split.
  apply eqc_symm; assumption.

  eapply oft_conv_0t; eauto using eqc_symm; [].
  eapply oft_equiv; eauto; [].
  eapply weaken1_eqc_type; eauto using eqc_symm; done.
Qed.


Lemma app_oft_inversion :
  forall G e1 e2 t,
    wf G
    -> oft G (tm_app e1 e2) t
    -> exists t',
         oft G e1 (cn_arrow t' t)
         /\ oft G e2 t'.
Proof.
intros G e1 e2 t Hwf Hof.
so (eqc_refl (oft_reg _#3 Hwf Hof)) as Heqt.
remember (tm_app e1 e2) as b eqn:Heqb.
revert Hof Hwf Heqt Heqb.
generalize t at 1 3.
intros t' Hof.
induct Hof; try (intros; discriminate Heqb); eauto with static; [].
intros G t1 t2 e1' e2' Hof1 _ Hof2 _ Hwf Heqt Heqb.
injection Heqb; intros; subst e1' e2'.
exists t1.
split; auto; [].
so (oft_reg _#3 Hwf Hof2).
eapply oft_equiv; eauto with static; done.
Qed.


Lemma app_halt_inversion :
  forall i e1 e2 v,
    expon (tstep nil) i (tm_app e1 e2) v
    -> tvalue nil v
    -> exists i1 i2 i3 t e3 v2,
         i1 < i
         /\ i2 < i
         /\ i3 < i
         /\ tvalue nil v2
         /\ expon (tstep nil) i1 e1 (tm_lam t e3)
         /\ expon (tstep nil) i2 e2 v2
         /\ expon (tstep nil) i3 (subst1 v2 e3) v.
Proof.
intros i e1 e2 v Hsteps Hval.
remember (tm_app e1 e2) as b eqn:Heqb.
revert e1 e2 Heqb Hval.
induct Hsteps.

(* 0 *)
intros b e1 e2 -> Hval.
invert Hval; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH e1 e2 -> Hval.
#3 invert Hstep.
  intros e1' Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & i3 & t & e3 & v2 & Hval2 & Hlt1 & Hlt2 & Hlt3 & Hsteps1 & Hsteps2 & Hsteps3).
  exists (S i1), i2, i3.
  #2 do 3 eexists; do2 6 split; eauto.
    omega.

    eapply expon_step; eauto; done.

  intros e2' Hval1 Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & i3 & t & e3 & v2 & Hval2 & Hlt1 & Hlt2 & Hlt3 & Hsteps1 & Hsteps2 & Hsteps3).
  exists i1, (S i2), i3.
  #2 do 3 eexists; do2 6 split; eauto.
    omega.

    eapply expon_step; eauto; done.

  intros t e3 Hval2 <- <-.
  exists 0, 0, i.
  #4 do 3 eexists; do2 6 split; eauto.
    omega.

    omega.

    apply expon_refl; done.

    apply expon_refl; done.
Qed.


Lemma app_halt_inversion' :
  forall e1 e2,
    thalt nil (tm_app e1 e2)
    -> thalt nil e1 /\ thalt nil e2.
Proof.
intros e1 e2 (v & Hsteps & Hval).
so (star_expon _#4 Hsteps) as (i & Hsteps').
so (app_halt_inversion _#4 Hsteps' Hval) as (i1 & i2 & _ & t & e3 & v2 & _ & _ & _ & Hval2 & Hsteps1 & Hsteps2 & _).
split; eexists; eauto using expon_star, tvalue_lam.
Qed.


Lemma pair_oft_inversion :
  forall G e1 e2 t,
    wf G
    -> oft G (tm_pair e1 e2) t
    -> exists t1 t2,
         eqc G t (cn_prod t1 t2) kd_type
         /\ oft G e1 t1
         /\ oft G e2 t2.
Proof.
intros G e1 e2 t Hwf Hof.
so (eqc_refl (oft_reg _#3 Hwf Hof)) as Heqt.
remember (tm_pair e1 e2) as b eqn:Heqb.
revert Hof Hwf Heqt Heqb.
generalize t at 1 3.
intros t' Hof.
induct Hof; try (intros; discriminate Heqb); eauto with static; [].
intros G t1 t2 e1' e2' Hof1 _ Hof2 _ Hwf Heqt Heqb.
injection Heqb.
intros; subst e1' e2'.
eauto.
Qed.


Lemma pair_oft_inversion_prod :
  forall G e1 e2 t1 t2,
    wf G
    -> oft G (tm_pair e1 e2) (cn_prod t1 t2)
    -> oft G e1 t1
       /\ oft G e2 t2.
Proof.
intros G e1 e2 t1 t2 Hwf Hof.
so (pair_oft_inversion _#4 Hwf Hof) as (t1' & t2' & Heqc & Hof1 & Hof2).
so (prod_injective _#5 Hwf Heqc) as (Heqc1 & Heqc2).
split; eauto with static.
Qed.


Lemma pair_halt_inversion :
  forall i e1 e2 v,
    expon (tstep nil) i (tm_pair e1 e2) v
    -> tvalue nil v
    -> exists i1 i2 v1 v2,
         i1 <= i
         /\ i2 <= i
         /\ tvalue nil v1
         /\ tvalue nil v2
         /\ expon (tstep nil) i1 e1 v1
         /\ expon (tstep nil) i2 e2 v2
         /\ v = tm_pair v1 v2.
Proof.
intros i e1 e2 v Hsteps Hval.
remember (tm_pair e1 e2) as b eqn:Heqb.
revert e1 e2 Heqb Hval.
induct Hsteps.

(* 0 *)
intros x e1 e2 -> Hval.
invert Hval; [].
intros Hval1 Hval2.
exists 0, 0, e1, e2.
repeat2 split; auto using expon_refl; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH e1 e2 -> Hval.
#2 invert Hstep.
  intros e1' Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & v1 & v2 & Hleq1 & Hleq2 & Hval1 & Hval2 & Hsteps1 & Hsteps2 & ->).
  exists (S i1), i2, v1, v2.
  do2 6 split; eauto using expon_step; omega.

  intros e1' Hval1 Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & v1 & v2 & Hleq1 & Hleq2 & Hval1' & Hval2 & Hsteps1 & Hsteps2 & ->).
  exists i1, (S i2), v1, v2.
  do2 6 split; eauto using expon_step; omega.
Qed.


Lemma pi1_oft_inversion :
  forall G e t,
    wf G
    -> oft G (tm_pi1 e) t
    -> exists t', oft G e (cn_prod t t').
Proof.
intros G e t Hwf Hof.
so (eqc_refl (oft_reg _#3 Hwf Hof)) as Heqt.
remember (tm_pi1 e) as b eqn:Heqb.
revert Hof Hwf Heqt Heqb.
generalize t at 1 2.
intros t' Hof.
induct Hof; try (intros; discriminate Heqb); eauto with static; [].
intros G t1 t2 e' Hof' _ Hwf Heqt Heqb.
injection Heqb; intros; subst e'.
exists t2.
eapply oft_equiv; eauto; [].
apply eqc_prod; auto; [].
apply eqc_refl; [].
so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hof')) as (kk & Hofcr).
invert Hofcr; auto; done.
Qed.


Lemma pi2_oft_inversion :
  forall G e t,
    wf G
    -> oft G (tm_pi2 e) t
    -> exists t', oft G e (cn_prod t' t).
Proof.
intros G e t Hwf Hof.
so (eqc_refl (oft_reg _#3 Hwf Hof)) as Heqt.
remember (tm_pi2 e) as b eqn:Heqb.
revert Hof Hwf Heqt Heqb.
generalize t at 1 2.
intros t' Hof.
induct Hof; try (intros; discriminate Heqb); eauto with static; [].
intros G t1 t2 e' Hof' _ Hwf Heqt Heqb.
injection Heqb; intros; subst e'.
exists t1.
eapply oft_equiv; eauto; [].
apply eqc_prod; auto; [].
apply eqc_refl; [].
so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hof')) as (kk & Hofcr).
invert Hofcr; auto; done.
Qed.


Lemma pi1_halt_inversion :
  forall i e v,
    expon (tstep nil) i (tm_pi1 e) v
    -> tvalue nil v
    -> exists i' v',
         i' < i  (* in fact i = S i' *)
         /\ tvalue nil v'
         /\ expon (tstep nil) i' e (tm_pair v v').
Proof.
intros i e v Hsteps Hval.
remember (tm_pi1 e) as b eqn:Heqb.
revert e Heqb Hval.
induct Hsteps.

(* 0 *)
intros x e -> Hval.
invert Hval; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH e -> Hval.
#2 invert Hstep.
  intros e' Hstep' <-.
  so (IH _ (eq_refl _) Hval) as (i' & v2 & Hlt & Hval2 & Hsteps').
  exists (S i'), v2.
  do2 2 split; eauto using expon_step; omega.

  intros v2 Hval1 Hval2 <-.
  exists i, v2.
  do2 2 split; auto; [].
  apply (expon_map' _ _ (fun b => tm_pair b v2)); auto; [].
  auto with dynamic.
Qed.
         

Lemma pi2_halt_inversion :
  forall i e v,
    expon (tstep nil) i (tm_pi2 e) v
    -> tvalue nil v
    -> exists i' v',
         i' < i  (* in fact i = S i' *)
         /\ tvalue nil v'
         /\ expon (tstep nil) i' e (tm_pair v' v).
Proof.
intros i e v Hsteps Hval.
remember (tm_pi2 e) as b eqn:Heqb.
revert e Heqb Hval.
induct Hsteps.

(* 0 *)
intros x e -> Hval.
invert Hval; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH e -> Hval.
#2 invert Hstep.
  intros e' Hstep' <-.
  so (IH _ (eq_refl _) Hval) as (i' & v2 & Hlt & Hval2 & Hsteps').
  exists (S i'), v2.
  do2 2 split; eauto using expon_step; omega.

  intros v1 Hval1 Hval2 <-.
  exists i, v1.
  do2 2 split; auto; [].
  apply (expon_map' _ _ (fun b => tm_pair v1 b)); auto; [].
  auto with dynamic.
Qed.


Lemma plam_oft_inversion :
  forall G k e u,
    wf G
    -> oft G (tm_plam k e) u
    -> exists t,
         eqc G u (cn_all k t) kd_type
         /\ oft (G; cl_cn k) e t.
Proof.
intros G k e u Hwf Hof.
so (eqc_refl (oft_reg _#3 Hwf Hof)) as Hequ.
remember (tm_plam k e) as b eqn:Heqb.
revert Hof Hwf Hequ Heqb.
generalize u at 1 3.
intros t Hof.
induct Hof; try (intros; discriminate Heqb); eauto with static; [].
intros G k' t e' Hofk Hofe _ Hwf Hequ Heqb.
injection Heqb; intros; subst k' e'.
eauto.
Qed.


Lemma plam_oft_inversion_all :
  forall G l t k e,
    wf G
    -> oft G (tm_plam k e) (cn_all l t)
    -> eqk G k l
       /\ oft (G; cl_cn l) e t.
Proof.
intros G l t k e Hwf Hof.
so (plam_oft_inversion _#4 Hwf Hof) as (u & Heqc & Hofe).
so (all_injective _#5 Hwf Heqc) as (Heqk & Heqt).
split.
  apply eqk_symm; auto; done.

  apply (@oft_equiv _ u); auto using eqc_symm; [].
  eapply oft_conv_0c; eauto using eqk_symm; done.
Qed.


Lemma papp_oft_inversion :
  forall G e c u,
    wf G
    -> oft G (tm_papp e c) u
    -> exists k t,
         oft G e (cn_all k t)
         /\ ofc G c k
         /\ eqc G u (subst1 c t) kd_type.
Proof.
intros G e c u Hwf Hof.
so (eqc_refl (oft_reg _#3 Hwf Hof)) as Hequ.
remember (tm_papp e c) as b eqn:Heqb.
revert Hof Hwf Hequ Heqb.
generalize u at 1 3.
intros t Hof.
induct Hof; try (intros; discriminate Heqb); eauto with static; [].
intros G k t e' c' Hofe _ Hofc Hwf Hequ Heqb.
injection Heqb; intros; subst e' c'.
eauto.
Qed.


Lemma papp_halt_inversion :
  forall i e c v,
    expon (tstep nil) i (tm_papp e c) v
    -> tvalue nil v
    -> exists i1 i2 k e',
         i1 < i
         /\ i2 < i
         /\ expon (tstep nil) i1 e (tm_plam k e')
         /\ expon (tstep nil) i2 (subst1 c e') v.
Proof.
intros i e c v Hsteps Hval.
remember (tm_papp e c) as b eqn:Heqb.
revert e c Heqb Hval.
induct Hsteps.

(* 0 *)
intros x e c -> Hval.
invert Hval; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH e c -> Hval.
#2 invert Hstep.
  intros e' Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & k & e'' & Hlt1 & Hlt2 & Hsteps1 & Hsteps2).
  exists (S i1), i2, k, e''.
  do2 3 split; eauto using expon_step; omega.

  intros k e' <- <-.
  exists 0, i, k, e'.
  do2 3 split; eauto using expon_refl; omega.
Qed.


Lemma pack_oft_inversion :
  forall G c e k t u,
    wf G
    -> oft G (tm_pack c e k t) u
    -> eqc G u (cn_exists k t) kd_type
       /\ ofc G c k
       /\ oft G e (subst1 c t).
Proof.
intros G c e k s u Hwf Hof.
so (eqc_refl (oft_reg _#3 Hwf Hof)) as Hequ.
remember (tm_pack c e k s) as b eqn:Heqb.
revert Hof Hwf Hequ Heqb.
generalize u at 1 3.
intros t Hof.
induct Hof; try (intros; discriminate Heqb); eauto with static; [].
intros G k' s' c' e' Hofc Hofe _ Hofs Hwf Hequ Heqb.
injection Heqb; intros; subst c' e' k' s'.
auto.
Qed.


Lemma pack_oft_inversion_exists :
  forall G c e k t l u,
    wf G
    -> oft G (tm_pack c e k t) (cn_exists l u)
    -> eqk G k l
       /\ eqc (G; cl_cn l) t u kd_type
       /\ ofc G c l
       /\ oft G e (subst1 c u).
Proof.
intros G c e k t l u Hwf Hof.
so (pack_oft_inversion _#6 Hwf Hof) as (Heqc & Hofc & Hofe).
so (exists_injective _#5 Hwf Heqc) as (Heqk & Heqt).
do2 3 split.
  apply eqk_symm; assumption.

  apply eqc_symm; assumption.

  eapply ofc_subsume; eauto with static; done.

  eapply oft_equiv; eauto; [].
  apply eqc_symm; [].
  eapply subst1_eqc_type; eauto; [].
  so (subst_eqc _#6 (subof_conv_0 _#3 (eqk_symm Heqk)) Heqt) as H.
  simp_sub in H.
  assumption.
Qed.


Lemma pack_halt_inversion :
  forall i c e k t v,
    expon (tstep nil) i (tm_pack c e k t) v
    -> tvalue nil v
    -> exists v',
         tvalue nil v'
         /\ expon (tstep nil) i e v'
         /\ v = tm_pack c v' k t.
Proof.
intros i c e k t v Hsteps Hval.
remember (tm_pack c e k t) as b eqn:Heqb.
revert c e k t Heqb Hval.
induct Hsteps.

(* 0 *)
intros x c e k t -> Hval.
invert Hval; [].
intros Hval'.
exists e; auto using expon_refl; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH c e k t -> Hval.
invert Hstep; [].
intros e' Hstep' <-.
so (IH _#4 (eq_refl _) Hval) as (v' & Hval' & Hsteps' & ->).
eauto using expon_step.
Qed.


Lemma unpack_oft_inversion :
  forall G e1 e2 u,
    wf G
    -> oft G (tm_unpack e1 e2) u
    -> exists k t,
         oft G e1 (cn_exists k t)
         /\ oft (G; cl_cn k; cl_tm t) e2 (subst (sh 2) u).
Proof.
intros G e1 e2 u Hwf Hof.
so (eqc_refl (oft_reg _#3 Hwf Hof)) as Hequ.
remember (tm_unpack e1 e2) as b eqn:Heqb.
revert Hof Hwf Hequ Heqb.
generalize u at 1 2.
intros t Hof.
induct Hof; try (intros; discriminate Heqb); eauto with static; [].
intros G k t t' e1' e2' Hof1 _ Hof2 _ Hoft' Hwf Hequ Heqb.
injection Heqb; intros; subst e1' e2'.
exists k, t.
split; trivial; [].
eapply oft_equiv; eauto; [].
eapply subst_eqc_type; eauto; [].
apply weaken2; auto; done.
Qed.


Lemma unpack_halt_inversion :
  forall i e1 e2 v,
    expon (tstep nil) i (tm_unpack e1 e2) v
    -> tvalue nil v
    -> exists i1 i2 c v' k t,
         i1 < i
         /\ i2 < i
         /\ tvalue nil v'
         /\ expon (tstep nil) i1 e1 (tm_pack c v' k t)
         /\ expon (tstep nil) i2 (subst (dot v' (dot c id)) e2) v.
Proof.
intros i e1 e2 v Hsteps Hval.
remember (tm_unpack e1 e2) as b eqn:Heqb.
revert e1 e2 Heqb Hval.
induct Hsteps.

(* 0 *)
intros x e1 e2 -> Hval.
invert Hval; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH e1 e2 -> Hval.
#2 invert Hstep.
  intros e1' Hstep' <-.
  so (IH e1' e2 (eq_refl _) Hval) as (i1 & i2 & c & v' & k & t & Hlt1 & Hlt2 & Hval' & Hsteps1 & Hsteps2).
  exists (S i1), i2, c, v', k, t.
  do2 4 split; eauto using expon_step; omega.

  intros c v k t Hval' <- <-.
  exists 0, i, c, v, k, t.
  do2 4 split; eauto using expon_refl; omega.
Qed.


Lemma fix_oft_inversion :
  forall G t e u,
    wf G
    -> oft G (tm_fix t e) u
    -> eqc G u t kd_type
       /\ oft G e (cn_arrow (cn_arrow cn_unit t) t).
Proof.
intros G t e u Hwf Hof.
so (eqc_refl (oft_reg _#3 Hwf Hof)) as Hequ.
remember (tm_fix t e) as b eqn:Heqb.
revert Hof Hwf Hequ Heqb.
generalize u at 1 3.
intros t' Hof.
induct Hof; try (intros; discriminate Heqb); eauto with static; [].
intros G t' e' Hof _ Hwf Hequ Heqb.
injection Heqb; intros; subst t' e'.
auto.
Qed.


Lemma lett_oft_inversion :
  forall G e1 e2 u,
    wf G
    -> oft G (tm_lett e1 e2) u
    -> exists t,
         oft G e1 t
         /\ oft (G; cl_tm t) e2 (shift1 u).
Proof.
intros G e1 e2 u Hwf Hof.
so (eqc_refl (oft_reg _#3 Hwf Hof)) as Hequ.
remember (tm_lett e1 e2) as b eqn:Heqb.
revert Hof Hwf Hequ Heqb.
generalize u at 1 2.
intros t Hof.
induct Hof; try (intros; discriminate Heqb); eauto with static; [].
intros G t1 t2 e1' e2' Hof1 _ Hof2 _ Hwf Hequ Heqb.
injection Heqb; intros; subst e1' e2'.
exists t1.
split; trivial; [].
eapply oft_equiv; eauto; [].
apply weaken1_eqc_type; assumption.
Qed.


Lemma lett_halt_inversion :
  forall i e1 e2 v,
    expon (tstep nil) i (tm_lett e1 e2) v
    -> tvalue nil v
    -> exists i1 i2 v',
         i1 < i
         /\ i2 < i
         /\ tvalue nil v'
         /\ expon (tstep nil) i1 e1 v'
         /\ expon (tstep nil) i2 (subst1 v' e2) v.
Proof.
intros i e1 e2 v Hsteps Hval.
remember (tm_lett e1 e2) as b eqn:Heqb.
revert e1 e2 Heqb Hval.
induct Hsteps.

(* 0 *)
intros x e1 e2 -> Hval.
invert Hval; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH e1 e2 -> Hval.
#2 invert Hstep.
  intros e1' Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & v' & Hlt1 & Hlt2 & Hval' & Hsteps1 & Hsteps2).
  exists (S i1), i2, v'.
  do2 4 split; eauto using expon_step; omega.

  intros Hval1 <-.
  exists 0, i, e1.
  do2 4 split; eauto using expon_refl; omega.
Qed.


Lemma lett_halt_inversion' :
  forall e1 e2,
    thalt nil (tm_lett e1 e2)
    -> thalt nil e1.
Proof.
intros e1 e2 (v & Hsteps & Hval).
so (star_expon _#4 Hsteps) as (i & Hsteps').
so (lett_halt_inversion _#4 Hsteps' Hval) as (i1 & _ & v' & _ & _ & Hval1 & Hsteps1 & _).
eexists; eauto using expon_star.
Qed.


Lemma letm_oft_inversion :
  forall G M e u,
    wf G
    -> oft G (tm_letm M e) u
    -> exists T k,
         ofm G impure M T
         /\ fsts T k
         /\ oft (G; cl_cn k; cl_md (shift1 T) (var 0)) e (subst (sh 2) u).
Proof.
intros G M e u Hwf Hof.
so (eqc_refl (oft_reg _#3 Hwf Hof)) as Hequ.
remember (tm_letm M e) as b eqn:Heqb.
revert Hof Hwf Hequ Heqb.
generalize u at 1 2.
intros t Hof.
induct Hof; try (intros; discriminate Heqb); eauto with static; [].
intros G T k t M' e' HofM Hfst Hofe _ Hoft Hwf Hequ Heqb.
injection Heqb; intros; subst M' e'.
exists T, k.
do2 2 split; auto; [].
eapply oft_equiv; eauto; [].
eapply subst_eqc_type; eauto; [].
apply weaken2; done.
Qed.


Lemma letm_halt_inversion :
  forall i M e v,
    expon (tstep nil) i (tm_letm M e) v
    -> tvalue nil v
    -> exists i1 i2 V c,
         i1 < i
         /\ i2 < i
         /\ mvalue nil V
         /\ fstm nil V c
         /\ expon (mstep nil) i1 M V
         /\ expon (tstep nil) i2 (subst (dot V (dot c id)) e) v.
Proof.
intros i M e v Hsteps Hval.
remember (tm_letm M e) as b eqn:Heqb.
revert M e Heqb Hval.
induct Hsteps.

(* 0 *)
intros x M e -> Hval.
invert Hval.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH M e -> Hval.
#2 invert Hstep.
  intros M' Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & V & c & Hlt1 & Hlt2 & Hval' & Hfst & Hsteps1 & Hsteps2).
  exists (S i1), i2, V, c.
  do2 5 split; eauto using expon_step; omega.

  intros c Hval' Hfst <-.
  exists 0, i, M, c.
  do2 5 split; eauto using expon_refl; omega.
Qed.


Lemma letm_halt_inversion' :
  forall M e,
    thalt nil (tm_letm M e)
    -> mhalt nil M.
Proof.
intros M e (v & Hsteps & Hval).
so (star_expon _#4 Hsteps) as (i & Hsteps').
so (letm_halt_inversion _#4 Hsteps' Hval) as (i1 & _ & V & _ & _ & _ & Hval1 & _ & Hsteps1 & _).
eexists; eauto using expon_star.
Qed.


Lemma ext_oft_inversion :
  forall G M u,
    wf G
    -> oft G (tm_ext M) u
    -> ofm G impure M (sg_dyn u).
Proof.
intros G M u Hwf Hof.
so (eqc_refl (oft_reg _#3 Hwf Hof)) as Hequ.
remember (tm_ext M) as b eqn:Heqb.
revert Hof Hwf Hequ Heqb.
generalize u at 1 2.
intros t Hof.
induct Hof; try (intros; discriminate Heqb); eauto with static; [].
intros G t M' H Hwf Hequ Heqb.
injection Heqb; intros; subst M'.
eapply ofm_subsume; eauto; [].
apply subs_refl; apply eqs_dyn; assumption.
Qed.


Lemma ext_halt_inversion :
  forall i M v,
    expon (tstep nil) i (tm_ext M) v
    -> tvalue nil v
    -> exists i',
         i' < i  (* in fact i = S i' *)
         /\ expon (mstep nil) i' M (md_dyn v).
Proof.
intros i M v Hsteps Hval.
remember (tm_ext M) as b eqn:Heqb.
revert M Heqb Hval.
induct Hsteps.

(* 0 *)
intros x M -> Hval.
invert Hval; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH M -> Hval.
#2 invert Hstep.
  intros M' Hstep' <-.
  so (IH _ (eq_refl _) Hval) as (i' & Hlt & Hsteps').
  exists (S i'); split; eauto using expon_step; omega.

  intros Hval' <-.
  exists i.
  split.
    omega.

    apply (expon_map _ _ (tstep nil) _ md_dyn); eauto with dynamic; done.
Qed.


Lemma dyn_ofm_inversion :
  forall G p e U,
    wf G
    -> ofm G p (md_dyn e) U
    -> exists t,
         sg_dyn t = U
         /\ oft G e t.
Proof.
intros G p e U Hwf Hof.
so (prins_complete _#4 Hwf Hof) as (p' & T & Hprin & _ & Hsubs).
invert Hprin; [].
intros t Hprint <- <-.
so (subs_dyn_invert _#3 Hsubs) as (t' & -> & Heqt).
exists t'.
split; trivial; [].
so (print_sound _#3 Hwf Hprint) as Hof'.
eauto with static; done.
Qed.


Lemma dyn_ofm_inversion_dyn :
  forall G p e t,
    wf G
    -> ofm G p (md_dyn e) (sg_dyn t)
    -> oft G e t.
Proof.
intros G p e t Hwf Hof.
so (dyn_ofm_inversion _#4 Hwf Hof) as (t' & Heq & Hof').
injection Heq; intros; subst t'.
assumption.
Qed.


Lemma dyn_halt_inversion :
  forall i e V,
    expon (mstep nil) i (md_dyn e) V
    -> mvalue nil V
    -> exists v,
         expon (tstep nil) i e v
         /\ V = md_dyn v.
Proof.
intros i e V Hsteps Hval.
remember (md_dyn e) as b eqn:Heqb.
revert e Heqb Hval.
induct Hsteps.

(* 0 *)
intros x e -> Hval.
invert Hval; [].
intros Hval'.
eauto using expon_refl; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH e -> Hval.
invert Hstep; [].
intros e' Hstep' <-.
so (IH _ (eq_refl _) Hval) as (v' & Hsteps' & ->).
eauto using expon_step; done.
Qed.


Lemma dyn_halt_inversion' :
  forall e,
    mhalt nil (md_dyn e)
    -> thalt nil e.
Proof.
intros e (V & Hsteps & Hval).
so (star_expon _#4 Hsteps) as (i & Hsteps').
so (dyn_halt_inversion i e V Hsteps' Hval) as (v & Hsteps'' & ->).
invert Hval; [].
intros Hval'.
eexists; split; eauto using expon_star; done.
Qed.


Lemma lamgn_ofm_inversion :
  forall G p T1 M U,
    wf G
    -> ofm G p (md_lamgn T1 M) U
    -> exists T2 k1,
         subs G (sg_pigen T1 T2) U
         /\ fsts T1 k1
         /\ ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M (shift1 T2).
Proof.
intros G p T1 M U Hwf Hof.
so (prins_complete _#4 Hwf Hof) as (p' & T & Hprin & _ & Hsubs).
invert Hprin; [].
intros T2 k1 p'' HofT1 Hfst Hprins <- <-.
so (prins_sound _#4 (wf_md Hwf HofT1 Hfst) Hprins) as Hof'.
exists T2, k1.
do2 2 split; auto; [].
eapply ofm_forget_impure; eauto; done.
Qed.


Lemma lamgn_ofm_inversion_pigen :
  forall G p T1 T2 T1' M,
    wf G
    -> ofm G p (md_lamgn T1' M) (sg_pigen T1 T2)
    -> exists k1,
         subs G T1 T1'
         /\ fsts T1 k1
         /\ ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M (shift1 T2).
Proof.
intros G p T1 T2 T1' M Hwf Hof.
so (lamgn_ofm_inversion _#5 Hwf Hof) as (T2' & k1' & Hsubs & HfstT1' & Hof').
so (subs_pigen_invert'' _#5 Hwf Hsubs) as (k1 & Hsubs1 & HfstT1 & Hsubs2).
exists k1.
do2 2 split; auto; [].
refine (ofm_subsume (ofm_subsume_01m _#8 (subs_fst _#5 Hwf Hsubs1 HfstT1 HfstT1') (weaken1_subs _#4 Hsubs1) Hof') _); [].
apply weaken1_subs; assumption.
Qed.


Lemma appgn_ofm_inversion :
  forall G p M1 M2 T,
    wf G
    -> ofm G p (md_appgn M1 M2) T
    -> exists T1 T2 c,
         ofm G impure M1 (sg_pigen T1 T2)
         /\ ofm G pure M2 T1
         /\ fstm G M2 c
         /\ subs G (subst1 c T2) T
         /\ p = impure.
Proof.
intros G p M1 M2 U Hwf Hof.
so (prins_complete _#4 Hwf Hof) as (p' & T & Hprins & Hpleq & Hsubs).
invert Hprins; [].
intros T1 T2 T1' c2 p'' Hprin1 Hprin2 Hsubs1 HfstM2 <- <-.
so (prins_sound _#4 Hwf Hprin1) as Hof1.
so (prins_sound _#4 Hwf Hprin2) as Hof2.
destruct p; [invert Hpleq; done |].
exists T1, T2, c2.
#2 do2 4 split; auto.
  eapply ofm_forget_impure; eauto; done.

  eapply ofm_subsume; eauto; done.
Qed.


Lemma appgn_halt_inversion :
  forall i M1 M2 V,
    expon (mstep nil) i (md_appgn M1 M2) V
    -> mvalue nil V
    -> exists i1 i2 i3 T M3 V2 c,
         i1 < i
         /\ i2 < i
         /\ i3 < i
         /\ mvalue nil V2
         /\ expon (mstep nil) i1 M1 (md_lamgn T M3)
         /\ expon (mstep nil) i2 M2 V2
         /\ fstm nil V2 c
         /\ expon (mstep nil) i3 (subst (dot V2 (dot c id)) M3) V.
Proof.
intros i M1 M2 V Hsteps Hval.
remember (md_appgn M1 M2) as b eqn:Heqb.
revert M1 M2 Heqb Hval.
induct Hsteps.

(* 0 *)
intros x M1 M2 -> Hval.
invert Hval; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH M1 M2 -> Hval.
#3 invert Hstep.
  intros M1' Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & i3 & T & M3 & V2 & c2 & Hlt1 & Hlt2 & Hlt3 & Hval2 & Hsteps1 & Hsteps2 & HfstV2 & Hsteps3).
  exists (S i1), i2, i3, T, M3, V2, c2.
  do2 7 split; eauto using expon_step; omega.

  intros M2' Hval1 Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & i3 & T & M3 & V2 & c2 & Hlt1 & Hlt2 & Hlt3 & Hval2 & Hsteps1 & Hsteps2 & HfstV2 & Hsteps3).
  exists i1, (S i2), i3, T, M3, V2, c2.
  do2 7 split; eauto using expon_step; omega.

  intros T M3 c2 Hval2 HfstM2 <- <-.
  exists 0, 0, i, T, M3, M2, c2.
  do2 7 split; eauto using expon_refl; omega.
Qed.


Lemma appgn_halt_inversion' :
  forall M1 M2,
    mhalt nil (md_appgn M1 M2)
    -> mhalt nil M1 /\ mhalt nil M2.
Proof.
intros M1 M2 (V & Hsteps & Hval).
so (star_expon _#4 Hsteps) as (i & Hsteps').
so (appgn_halt_inversion _#4 Hsteps' Hval) as (i1 & i2 & _ & t & e3 & v2 & _ & _ & _ & _ & Hval2 & Hsteps1 & Hsteps2 & _).
split; eexists; eauto using expon_star, mvalue_lamgn.
Qed.


Lemma lamap_ofm_inversion :
  forall G p T1 M U,
    wf G
    -> ofm G p (md_lamap T1 M) U
    -> exists T2 k1,
         subs G (sg_piapp T1 T2) U
         /\ fsts T1 k1
         /\ ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) pure M (shift1 T2).
Proof.
intros G p T1 M U Hwf Hof.
so (prins_complete _#4 Hwf Hof) as (p' & T & Hprin & _ & Hsubs).
invert Hprin; [].
intros T2 k1 HofT1 Hfst Hprins <- <-.
so (prins_sound _#4 (wf_md Hwf HofT1 Hfst) Hprins) as Hof'.
exists T2, k1.
do2 2 split; auto; done.
Qed.


Lemma lamap_ofm_inversion_piapp :
  forall G p T1 T2 T1' M,
    wf G
    -> ofm G p (md_lamap T1' M) (sg_piapp T1 T2)
    -> exists k1,
         subs G T1 T1'
         /\ fsts T1 k1
         /\ ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) pure M (shift1 T2).
Proof.
intros G p T1 T2 T1' M Hwf Hof.
so (lamap_ofm_inversion _#5 Hwf Hof) as (T2' & k1' & Hsubs & HfstT1' & Hof').
so (subs_piapp_invert'' _#5 Hwf Hsubs) as (k1 & Hsubs1 & HfstT1 & Hsubs2).
exists k1.
do2 2 split; auto; [].
refine (ofm_subsume (ofm_subsume_01m _#8 (subs_fst _#5 Hwf Hsubs1 HfstT1 HfstT1') (weaken1_subs _#4 Hsubs1) Hof') _); [].
apply weaken1_subs; assumption.
Qed.


Lemma appap_ofm_inversion :
  forall G p M1 M2 T,
    wf G
    -> ofm G p (md_appap M1 M2) T
    -> exists T1 T2 c,
         ofm G p M1 (sg_piapp T1 T2)
         /\ ofm G pure M2 T1
         /\ fstm G M2 c
         /\ subs G (subst1 c T2) T.
Proof.
intros G p M1 M2 U Hwf Hof.
so (prins_complete _#4 Hwf Hof) as (p' & T & Hprins & Hpleq & Hsubs).
invert Hprins; [].
intros T1 T2 T1' c2 Hprin1 Hprin2 Hsubs1 HfstM2 <-.
so (prins_sound _#4 Hwf Hprin1) as Hof1.
so (prins_sound _#4 Hwf Hprin2) as Hof2.
exists T1, T2, c2.
#2 do2 3 split; auto.
  eapply ofm_forget_leq; eauto; done.

  eapply ofm_subsume; eauto; done.
Qed.


Lemma appap_halt_inversion :
  forall i M1 M2 V,
    expon (mstep nil) i (md_appap M1 M2) V
    -> mvalue nil V
    -> exists i1 i2 i3 T M3 V2 c,
         i1 < i
         /\ i2 < i
         /\ i3 < i
         /\ mvalue nil V2
         /\ expon (mstep nil) i1 M1 (md_lamap T M3)
         /\ expon (mstep nil) i2 M2 V2
         /\ fstm nil V2 c
         /\ expon (mstep nil) i3 (subst (dot V2 (dot c id)) M3) V.
Proof.
intros i M1 M2 V Hsteps Hval.
remember (md_appap M1 M2) as b eqn:Heqb.
revert M1 M2 Heqb Hval.
induct Hsteps.

(* 0 *)
intros x M1 M2 -> Hval.
invert Hval; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH M1 M2 -> Hval.
#3 invert Hstep.
  intros M1' Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & i3 & T & M3 & V2 & c2 & Hlt1 & Hlt2 & Hlt3 & Hval2 & Hsteps1 & Hsteps2 & HfstV2 & Hsteps3).
  exists (S i1), i2, i3, T, M3, V2, c2.
  do2 7 split; eauto using expon_step; omega.

  intros M2' Hval1 Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & i3 & T & M3 & V2 & c2 & Hlt1 & Hlt2 & Hlt3 & Hval2 & Hsteps1 & Hsteps2 & HfstV2 & Hsteps3).
  exists i1, (S i2), i3, T, M3, V2, c2.
  do2 7 split; eauto using expon_step; omega.

  intros T M3 c2 Hval2 HfstM2 <- <-.
  exists 0, 0, i, T, M3, M2, c2.
  do2 7 split; eauto using expon_refl; omega.
Qed.


Lemma pair_ofm_inversion :
  forall G p M1 M2 U,
    wf G
    -> ofm G p (md_pair M1 M2) U
    -> exists T1 T2,
         subs G (sg_sigma T1 (shift1 T2)) U
         /\ ofm G p M1 T1
         /\ ofm G p M2 T2.
Proof.
intros G p M1 M2 U Hwf Hof.
so (prins_complete _#4 Hwf Hof) as (p' & T & Hprin & Hpleq & Hsubs).
invert Hprin; [].
intros p1 p2 T1 T2 Hprin1 Hprin2 <- <-.
so (prins_sound _#4 Hwf Hprin1) as Hof1.
so (prins_sound _#4 Hwf Hprin2) as Hof2.
exists T1, T2.
do2 2 split; eauto using ofm_forget_leq, purity_join_leq1, purity_join_leq2.
Qed.


Lemma pair_ofm_inversion_1 :
  forall G p M1 M2 T1 T2,
    wf G
    -> ofm G p (md_pair M1 M2) (sg_sigma T1 T2)
    -> ofm G p M1 T1.
Proof.
intros G p M1 M2 T1 T2 Hwf Hof.
so (pair_ofm_inversion _#5 Hwf Hof) as (T1' & T2' & Hsubs & Hof1 & _).
so (subs_sigma_invert' _#4 Hwf Hsubs) as (T1'' & T2'' & _ & Heq & _ & Hsubs1 & _).
injection Heq; intros; subst T1'' T2''.
eapply ofm_subsume; eauto; done.
Qed.


Lemma pair_ofm_inversion_sigma_pure :
  forall G M1 M2 T1 T2,
    wf G
    -> ofm G pure (md_pair M1 M2) (sg_sigma T1 T2)
    -> exists c1,
         ofm G pure M1 T1
         /\ fstm G M1 c1
         /\ ofm G pure M2 (subst1 c1 T2).
Proof.
intros G M1 M2 T1 T2 Hwf Hof.
so (pair_ofm_inversion _#5 Hwf Hof) as (T1' & T2' & Hsubs & Hof1 & Hof2).
so (subs_sigma_invert' _#4 Hwf Hsubs) as (T1'' & T2'' & k1' & Heq & HfstT1' & Hsubs1 & Hsubs2).
injection Heq; clear Heq; intros <- <-.
so (can_fstm _#3 Hwf Hof1) as (c & HfstM1).
so (ofm_fst _#6 Hwf Hof1 HfstM1 HfstT1') as Hofc.
so (subst1_subs _#5 Hofc Hsubs2) as Hsubs2x; simp_sub in Hsubs2x.
exists c.
do2 2 split; eauto using ofm_subsume.
Qed.


Lemma pair_ofm_inversion_sigma_pure' :
  forall G M1 M2 T1 T2 c1,
    wf G
    -> ofm G pure (md_pair M1 M2) (sg_sigma T1 T2)
    -> fstm G M1 c1
    -> ofm G pure M1 T1
       /\ ofm G pure M2 (subst1 c1 T2).
Proof.
intros G M1 M2 T1 T2 c1 Hwf Hof Hfst.
so (pair_ofm_inversion_sigma_pure _#5 Hwf Hof) as (c1a & Hof1 & Hfsta & Hof2).
so (fstm_fun _#4 Hfst Hfsta); subst c1a.
auto.
Qed.


Lemma mpair_halt_inversion :
  forall i M1 M2 V,
    expon (mstep nil) i (md_pair M1 M2) V
    -> mvalue nil V
    -> exists i1 i2 V1 V2,
         i1 <= i
         /\ i2 <= i
         /\ mvalue nil V1
         /\ mvalue nil V2
         /\ expon (mstep nil) i1 M1 V1
         /\ expon (mstep nil) i2 M2 V2
         /\ V = md_pair V1 V2.
Proof.
intros i M1 M2 V Hsteps Hval.
remember (md_pair M1 M2) as b eqn:Heqb.
revert M1 M2 Heqb Hval.
induct Hsteps.

(* 0 *)
intros x M1 M2 -> Hval.
invert Hval; [].
intros Hval1 Hval2.
exists 0, 0, M1, M2.
repeat2 split; auto using expon_refl; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH M1 M2 -> Hval.
#2 invert Hstep.
  intros M1' Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & V1 & V2 & Hleq1 & Hleq2 & Hval1 & Hval2 & Hsteps1 & Hsteps2 & ->).
  exists (S i1), i2, V1, V2.
  do2 6 split; eauto using expon_step; omega.

  intros M1' Hval1 Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & V1 & V2 & Hleq1 & Hleq2 & Hval1' & Hval2 & Hsteps1 & Hsteps2 & ->).
  exists i1, (S i2), V1, V2.
  do2 6 split; eauto using expon_step; omega.
Qed.


(* It's messy to get the nice conclusion without a subsignature.  Not sure it was worth it. *)
Lemma pi1_ofm_inversion :
  forall G p M T,
    wf G
    -> ofm G p (md_pi1 M) T
    -> exists T',
         ofm G pure M (sg_sigma T T').
Proof.
intros G p M U Hwf Hof.
so (prins_complete _#4 Hwf Hof) as (p' & T1 & Hprin & _ & Hsubs).
invert Hprin; [].
intros T2 Hprin' <-.
so (prins_sound _#4 Hwf Hprin') as HofM.
so (can_fstm _#3 Hwf HofM) as (c & HfstM).
so (Singleton.singles_ofm _#4 Hwf HofM HfstM) as HofMs.
simpl in HofMs.
so (Formation.ofs_hsg _ _ (ofm_reg _#4 Hwf HofM)) as Hsig.
invert Hsig; [].
intros _ HT2; clear Hsig.
rewrite -> (Singleton.subst_singles _#3 HT2) in HofMs; [].
simp_sub in HofMs.
repl (Singleton.singles (cn_pi2 (shift1 c)) (subst (dot (cn_pi1 (shift1 c)) sh1) T2)) with (shift1 (Singleton.singles (cn_pi2 c) (subst1 (cn_pi1 c) T2))) in HofMs.
  unfold shift1, subst1.
  rewrite -> (Singleton.subst_singles _#3 (Formation.subst_hsg _ _ HT2)); [].
  simp_sub.
  reflexivity.
so (ofm_reg _#4 Hwf HofM) as Hofsigma.
invert Hofsigma; [].
intros k HofT1 HfstT1 HofT2.
so (can_fsts _ _ (ofm_reg _#4 Hwf Hof)) as (l & HfstU).
exists (shift1 (Singleton.singles (cn_pi2 c) (subst1 (cn_pi1 c) T2))).
eapply ofm_subsume; eauto; [].
so (Singleton.singles_fst (cn_pi1 c) _ _ HfstT1) as HfstT1s.
so (ofm_reg _#4 Hwf HofMs) as Hofsigmas.
invert Hofsigmas; [].
intros ks _ HfstT1s' HofT2sw.
so (fsts_fun _#3 HfstT1s HfstT1s'); subst ks.
so (subst1_ofs _#4 (inh_ofc _ _ (Singleton.single_ofk _#3 Hwf (ofm_fst _#6 Hwf (ofm_pi1 HofM) (fstm_pi1 HfstM) HfstT1))) HofT2sw) as HofT2s.
simp_sub in HofT2s.
#3 eapply subs_sigma; eauto.
  eapply subs_trans; eauto; [].
  eapply Singleton.singles_subs; eauto; [].
  refine (ofm_fst _ pure _#4 Hwf _ (fstm_pi1 HfstM) HfstT1); [].
  eapply ofm_pi1; eauto; done.

  apply subs_refl; apply eqs_refl; [].
  apply weaken1_ofs; auto; done.

  apply weaken1_ofs; auto; done.
Qed.


Lemma pi2_ofm_inversion :
  forall G p M U,
    wf G
    -> ofm G p (md_pi2 M) U
    -> exists T1 T2 c,
         ofm G pure M (sg_sigma T1 T2)
         /\ fstm G M c
         /\ subs G (subst1 (cn_pi1 c) T2) U.
Proof.
intros G p M U Hwf Hof.
so (prins_complete _#4 Hwf Hof) as (p' & T & Hprin & _ & Hsubs).
invert Hprin; [].
intros T1 T2 c Hprin' HfstM <- <-.
so (prins_sound _#4 Hwf Hprin') as Hof'.
do 3 eexists; eauto.
Qed.


Lemma mpi1_halt_inversion :
  forall i M V,
    expon (mstep nil) i (md_pi1 M) V
    -> mvalue nil V
    -> exists i' V',
         i' < i  (* in fact i = S i' *)
         /\ mvalue nil V'
         /\ expon (mstep nil) i' M (md_pair V V').
Proof.
intros i M V Hsteps Hval.
remember (md_pi1 M) as b eqn:Heqb.
revert M Heqb Hval.
induct Hsteps.

(* 0 *)
intros x M -> Hval.
invert Hval; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH M -> Hval.
#2 invert Hstep.
  intros M' Hstep' <-.
  so (IH _ (eq_refl _) Hval) as (i' & V2 & Hlt & Hval2 & Hsteps').
  exists (S i'), V2.
  do2 2 split; eauto using expon_step; omega.

  intros V2 Hval1 Hval2 <-.
  exists i, V2.
  do2 2 split; auto; [].
  apply (expon_map' _ _ (fun b => md_pair b V2)); auto; [].
  auto with dynamic.
Qed.


Lemma mpi2_halt_inversion :
  forall i M V,
    expon (mstep nil) i (md_pi2 M) V
    -> mvalue nil V
    -> exists i' V',
         i' < i  (* in fact i = S i' *)
         /\ mvalue nil V'
         /\ expon (mstep nil) i' M (md_pair V' V).
Proof.
intros i M V Hsteps Hval.
remember (md_pi2 M) as b eqn:Heqb.
revert M Heqb Hval.
induct Hsteps.

(* 0 *)
intros x M -> Hval.
invert Hval; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH M -> Hval.
#2 invert Hstep.
  intros M' Hstep' <-.
  so (IH _ (eq_refl _) Hval) as (i' & V1 & Hlt & Hval2 & Hsteps').
  exists (S i'), V1.
  do2 2 split; eauto using expon_step; omega.

  intros V1 Hval1 Hval2 <-.
  exists i, V1.
  do2 2 split; auto; [].
  apply (expon_map' _ _ (fun b => md_pair V1 b)); auto; [].
  auto with dynamic.
Qed.


Lemma mpi1_halt_inversion' :
  forall M,
    mhalt nil (md_pi1 M)
    -> mhalt nil M.
Proof.
intros M (V & Hsteps & Hval).
so (star_expon _#4 Hsteps) as (i & Hsteps').
so (mpi1_halt_inversion _#3 Hsteps' Hval) as (i' & V' & _ & Hval' & Hsteps'').
eexists; eauto using expon_star with dynamic.
Qed.


Lemma unpack_ofm_inversion :
  forall G p e M T U,
    wf G
    -> ofm G p (md_unpack e M T) U
    -> exists k t,
         oft G e (cn_exists k t)
         /\ ofm (G; cl_cn k; cl_tm t) impure M (subst (sh 2) T)
         /\ subs G T U
         /\ p = impure.
Proof.
intros G p e M T U Hwf Hof.
so (prins_complete _#4 Hwf Hof) as (p' & T' & Hprins & Hpleq & Hsubs).
invert Hprins; [].
intros p'' u k t T'' Hprin1 Hnorm Hprin2 Hsubs' Hofs <- <-.
so (print_sound _#3 Hwf Hprin1) as Hof1.
so (Sound.headnorm_sound _#4 Hwf (oft_reg _#3 Hwf Hof1) Hnorm) as Heqc.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _; clear Hofcr l.
so (prins_sound _#4 (wf_tm (wf_cn Hwf Hofk) Hoft) Hprin2) as Hof2.
destruct Hpleq as [Heq | ->]; [discriminate Heq |].
exists k, t.
#2 do2 3 split; auto.
  eapply oft_equiv; eauto; done.

  eapply ofm_forget_impure; eapply ofm_subsume; eauto; done.
Qed.


Lemma munpack_halt_inversion :
  forall i e M T V,
    expon (mstep nil) i (md_unpack e M T) V
    -> mvalue nil V
    -> exists i1 i2 c v k t,
         i1 < i
         /\ i2 < i
         /\ tvalue nil v
         /\ expon (tstep nil) i1 e (tm_pack c v k t)
         /\ expon (mstep nil) i2 (subst (dot v (dot c id)) M) V.
Proof.
intros i e M T V Hsteps Hval.
remember (md_unpack e M T) as b eqn:Heqb.
revert e M T Heqb Hval.
induct Hsteps.

(* 0 *)
intros x e M T -> Hval.
invert Hval; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH e M T -> Hval.
#2 invert Hstep.
  intros e' Hstep' <-.
  so (IH e' M T (eq_refl _) Hval) as (i1 & i2 & c & v' & k & t & Hlt1 & Hlt2 & Hval' & Hsteps1 & Hsteps2).
  exists (S i1), i2, c, v', k, t.
  do2 4 split; eauto using expon_step; omega.

  intros c v k t Hval' <- <-.
  exists 0, i, c, v, k, t.
  do2 4 split; eauto using expon_refl; omega.
Qed.


Lemma lett_ofm_inversion :
  forall G p e M U,
    wf G
    -> ofm G p (md_lett e M) U
    -> exists t,
         oft G e t
         /\ ofm (G; cl_tm t) p M (shift1 U).
Proof.
intros G p e M U Hwf Hof.
so (prins_complete _#4 Hwf Hof) as (p' & T & Hprin & Hpleq & Hsubs).
invert Hprin; [].
intros t T' Hprin1 Hprin2 <-.
so (print_sound _#3 Hwf Hprin1) as Hof1.
so (prins_sound _#4 (wf_tm Hwf (oft_reg _#3 Hwf Hof1)) Hprin2) as Hof2.
exists t.
split; auto; [].
eapply ofm_forget_leq; eauto; [].
eapply ofm_subsume; eauto; [].
so (subst_subs _#5 (weaken1 _ _ (cl_tm t)) Hsubs) as Hsubsw.
simp_sub in Hsubsw.
rewrite <- (Formation.subord_tm_sg _#3 (ofm_reg _#4 (wf_tm Hwf (oft_reg _#3 Hwf Hof1)) Hof2)) in Hsubsw; [].
assumption.
Qed.


Lemma mlett_halt_inversion :
  forall i e M V,
    expon (mstep nil) i (md_lett e M) V
    -> mvalue nil V
    -> exists i1 i2 v,
         i1 < i
         /\ i2 < i
         /\ tvalue nil v
         /\ expon (tstep nil) i1 e v
         /\ expon (mstep nil) i2 (subst1 v M) V.
Proof.
intros i e M V Hsteps Hval.
remember (md_lett e M) as b eqn:Heqb.
revert e M Heqb Hval.
induct Hsteps.

(* 0 *)
intros x e M -> Hval.
invert Hval; done.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH e M -> Hval.
#2 invert Hstep.
  intros e' Hstep' <-.
  so (IH _ _ (eq_refl _) Hval) as (i1 & i2 & V' & Hlt1 & Hlt2 & Hval' & Hsteps1 & Hsteps2).
  exists (S i1), i2, V'.
  do2 4 split; eauto using expon_step; omega.

  intros Hval1 <-.
  exists 0, i, e.
  do2 4 split; eauto using expon_refl; omega.
Qed.


Lemma mlett_halt_inversion' :
  forall e M,
    mhalt nil (md_lett e M)
    -> thalt nil e.
Proof.
intros M e (v & Hsteps & Hval).
so (star_expon _#4 Hsteps) as (i & Hsteps').
so (mlett_halt_inversion _#4 Hsteps' Hval) as (i1 & _ & v1 & _ & _ & Hval1 & Hsteps1 & _).
eexists; eauto using expon_star.
Qed.


Lemma letm_ofm_inversion :
  forall G p M1 M2 T U,
    wf G
    -> ofm G p (md_letm M1 M2 T) U
    -> exists T' k,
         ofm G impure M1 T'
         /\ fsts T' k
         /\ ofm (G; cl_cn k; cl_md (shift1 T') (var 0)) impure M2 (subst (sh 2) T)
         /\ subs G T U
         /\ p = impure.
Proof.
intros G p M1 M2 T U Hwf Hof.
so (prins_complete _#4 Hwf Hof) as (p' & T' & Hprin & Hpleq & Hsubs).
invert Hprin; [].
intros T1 T2 k1 p1 p2 Hprin1 HfstT1 Hprin2 Hsubs' HofT <- <-.
so (prins_sound _#4 Hwf Hprin1) as Hof1.
so (ofm_reg _#4 Hwf Hof1) as HofT1.
so (prins_sound _#4 (wf_md Hwf HofT1 HfstT1) Hprin2) as Hof2.
destruct Hpleq as [Heq | ->]; [discriminate Heq |].
exists T1, k1.
#2 do2 4 split; auto.
  eapply ofm_forget_impure; eassumption.

  eapply ofm_subsume; eauto; [].
  eapply ofm_forget_impure; eassumption.
Qed.


Lemma mletm_halt_inversion :
  forall i M1 M2 T V,
    expon (mstep nil) i (md_letm M1 M2 T) V
    -> mvalue nil V
    -> exists i1 i2 V' c,
         i1 < i
         /\ i2 < i
         /\ mvalue nil V'
         /\ fstm nil V' c
         /\ expon (mstep nil) i1 M1 V'
         /\ expon (mstep nil) i2 (subst (dot V' (dot c id)) M2) V.
Proof.
intros i M1 M2 T V Hsteps Hval.
remember (md_letm M1 M2 T) as b eqn:Heqb.
revert M1 M2 T Heqb Hval.
induct Hsteps.

(* 0 *)
intros x M1 M2 T -> Hval.
invert Hval.

(* S *)
intros i b1 b2 b3 Hstep Hsteps IH M1 M2 T -> Hval.
#2 invert Hstep.
  intros M' Hstep' <-.
  so (IH _ _ _ (eq_refl _) Hval) as (i1 & i2 & V1 & c1 & Hlt1 & Hlt2 & Hval' & Hfst & Hsteps1 & Hsteps2).
  exists (S i1), i2, V1, c1.
  do2 5 split; eauto using expon_step; omega.

  intros c Hval' Hfst <-.
  exists 0, i, M1, c.
  do2 5 split; eauto using expon_refl; omega.
Qed.


Lemma mletm_halt_inversion' :
  forall M1 M2 T,
    mhalt nil (md_letm M1 M2 T)
    -> mhalt nil M1.
Proof.
intros M1 M2 T (V & Hsteps & Hval).
so (star_expon _#4 Hsteps) as (i & Hsteps').
so (mletm_halt_inversion _#5 Hsteps' Hval) as (i1 & _ & v' & _ & _ & _ & Hval1 & _ & Hsteps1 & _).
eexists; eauto using expon_star.
Qed.


Lemma seal_ofm_inversion :
  forall G p M T U,
    wf G
    -> ofm G p (md_seal M T) U
    -> ofm G impure M T
       /\ subs G T U
       /\ p = impure.
Proof.
intros G p M T U Hwf Hof.
so (prins_complete _#4 Hwf Hof) as (p' & T' & Hprin & Hpleq & Hsubs).
invert Hprin; [].
intros T'' p'' Hprin' Hsubs' <- <-.
so (prins_sound _#4 Hwf Hprin') as Hof'.
destruct Hpleq as [Heq | ->]; [discriminate Heq |].
do2 2 split; eauto with static; [].
eapply ofm_subsume; eauto; [].
eapply ofm_forget_impure; eassumption.
Qed.
