
Require Import Tactics.
Require Import Extensionality.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import Formation.
Require Import TSubst.
Require Import Regularity.
Require Import Closure.
Require Import Candidate.
Require Import Skof.
Require Contextual.


Inductive type_interp : env -> per :=
| type_interp_ {E t t' R}
    : ofc nil t kd_type
      -> ofc nil t' kd_type
      -> tvrel t t' R
      -> tpclosed t t' R
      -> type_interp E t t' (Cand sk_type R) (Cand sk_type R)
.


Inductive sing_interp
  (c : (* cn *) term) (Ic : env -> trelation) : env -> per :=
| sing_interp_ {E t t'}
    : eqc nil t (subst (left E) c) kd_type
      -> eqc nil t' (subst (right E) c) kd_type
      -> tvrel t t' (Ic E)
      -> tpclosed t t' (Ic E)
      -> sing_interp c Ic E t t' (Cand sk_type (Ic E)) (Cand sk_type (Ic E))
.


Inductive pi_interp
  (k1 k2 : (* kd *) term) (Ik1 Ik2 : env -> per) : env -> per :=
| pi_interp_ {E c c' F F'}
    : ofc nil c (subst (left E) (kd_pi k1 k2))
      -> ofc nil c' (subst (right E) (kd_pi k1 k2))
      -> (forall d d' Q Q' d'' d''',
            Ik1 E d d' (Cand (simp k1) Q) (Cand (simp k1) Q')
            -> eqc nil d d'' (subst (left E) k1)
            -> eqc nil d' d''' (subst (right E) k1)
            -> Ik2 (E; Sta d d' (Cand (simp k1) Q)) (cn_app c d) (cn_app c' d') (Cand (simp k2) (F d d' Q)) (Cand (simp k2) (F' d'' d''' Q')))
      -> pi_interp k1 k2 Ik1 Ik2 E c c' (Cand (sk_pi (simp k1) (simp k2)) F) (Cand (sk_pi (simp k1) (simp k2)) F')
.


Inductive sigma_interp
  (k1 k2 : (* kd *) term) (Ik1 Ik2 : env -> per) : env -> per :=
| sigma_interp_ {E c c' P P'}
    : ofc nil c (subst (left E) (kd_sigma k1 k2))
      -> ofc nil c' (subst (right E) (kd_sigma k1 k2))
      -> Ik1 E (cn_pi1 c) (cn_pi1 c') (Cand (simp k1) (fst P)) (Cand (simp k1) (fst P'))
      -> Ik2 (E; Sta (cn_pi1 c) (cn_pi1 c') (Cand (simp k1) (fst P))) (cn_pi2 c) (cn_pi2 c') (Cand (simp k2) (snd P)) (Cand (simp k2) (snd P'))
      -> sigma_interp k1 k2 Ik1 Ik2 E c c' (Cand (sk_sigma (simp k1) (simp k2)) P) (Cand (sk_sigma (simp k1) (simp k2)) P')
.


Inductive one_interp : env -> per :=
| one_interp_ {E c c'}
    : ofc nil c kd_one
      -> ofc nil c' kd_one
      -> one_interp E c c' (Cand sk_one tt) (Cand sk_one tt)
.


Inductive arrow_interp
  (t1 t2 : (* cn *) term) (It1 It2 : env -> trelation) : env -> trelation :=
| arrow_interp_ {E v v'}
    : oft nil v (subst (left E) (cn_arrow t1 t2))
      -> oft nil v' (subst (right E) (cn_arrow t1 t2))
      -> tvalue nil v
      -> tvalue nil v'
      -> (forall w w',
            It1 E w w'
            -> tev (subst (left E) t2) (subst (right E) t2) (It2 E) (tm_app v w) (tm_app v' w'))
      -> arrow_interp t1 t2 It1 It2 E v v'
.


Inductive prod_interp
  (t1 t2 : (* cn *) term) (It1 It2 : env -> trelation) : env -> trelation :=
| prod_interp_ {E v1 v1' v2 v2'}
    : oft nil (tm_pair v1 v2) (subst (left E) (cn_prod t1 t2))
      -> oft nil (tm_pair v1' v2') (subst (right E) (cn_prod t1 t2))
      -> tvalue nil (tm_pair v1 v2)
      -> tvalue nil (tm_pair v1' v2')
      -> It1 E v1 v1'
      -> It2 E v2 v2'
      -> prod_interp t1 t2 It1 It2 E (tm_pair v1 v2) (tm_pair v1' v2')
.


Inductive all_interp
  (k : (* kd *) term) (t : (* cn *) term) (Ik : env -> per) (It : env -> trelation) : env -> trelation :=
| all_interp_ {E v v'}
    : oft nil v (subst (left E) (cn_all k t))
      -> oft nil v' (subst (right E) (cn_all k t))
      -> tvalue nil v
      -> tvalue nil v'
      -> (forall c c' Q,
            Ik E c c' Q Q
            -> tev (subst (dot c (left E)) t) (subst (dot c' (right E)) t)
               (It (E; Sta c c' Q)) (tm_papp v c) (tm_papp v' c'))
      -> all_interp k t Ik It E v v'
.


Inductive exists_interp
  (k : (* kd *) term) (t : (* cn *) term) (Ik : env -> per) (It : env -> trelation) : env -> trelation :=
| exists_interp_ {E v v' c c' l l' u u' Q}
    : oft nil (tm_pack c v l u) (subst (left E) (cn_exists k t))
      -> oft nil (tm_pack c' v' l' u') (subst (right E) (cn_exists k t))
      -> tvalue nil (tm_pack c v l u)
      -> tvalue nil (tm_pack c' v' l' u')
      -> Ik E c c' Q Q
      -> It (E; Sta c c' Q) v v'
      -> exists_interp k t Ik It E (tm_pack c v l u) (tm_pack c' v' l' u')
.


Definition exists_interp_st
  (k : (* kd *) term) (t : (* cn *) term) (Ik : env -> per) (It : env -> trelation) (E : env) : trelation :=
  tpst (subst (left E) (cn_exists k t)) (subst (right E) (cn_exists k t))
  (exists_interp k t Ik It E).
  

Inductive unit_interp : env -> trelation :=
| unit_interp_ {E}
    : unit_interp E tm_star tm_star
.


Definition flatten_env (E : env) : list (option skind) :=
  map (fun x => (match x with Dyn _ _ => None | Sta _ _ (Cand sk _) => Some sk end)) E.


Fixpoint intk (k : (* kd *) term) (E : env) {struct k} : per :=
  (match k with
   | kd_type =>
       type_interp E

   | kd_sing c =>
       sing_interp c (fun E => out sk_type (intc c E)) E

   | kd_pi k1 k2 =>
       pi_interp k1 k2 (intk k1) (intk k2) E

   | kd_sigma k1 k2 =>
       sigma_interp k1 k2 (intk k1) (intk k2) E

   | kd_one =>
       one_interp E

   | _ => (fun _ _ _ _ => False)
   end)

with intc (c : (* cn *) term) (E : env) {struct c} : cand :=
  (match c with
   | var i =>
       (match nth i E (Dyn nonsense nonsense) with
        | Dyn _ _ => Cand sk_one tt
        | Sta _ _ Q => Q
        end)

   | cn_lam k1 c' =>
       (* The problem here is we have to know the codomain (up to simple types),
          before we define the function.  The easy way to get the codomain is
          to interpret the body, but we can't interpret the body until we know
          the arguments, which we can't get until we define the function.
       *)
       let sk := skof (flatten_env E; Some (simp k1)) c'
       in
         Cand (sk_pi (simp k1) sk)
         (fun d d' Q => out sk (intc c' (E; Sta d d' (Cand (simp k1) Q))))

   | cn_app c1 c2 =>
       (match intc c1 E with
        | Cand (sk_pi sk sk') Q1 =>
            Cand sk' (Q1 (subst (left E) c2) (subst (right E) c2) (out sk (intc c2 E)))

        | _ => Cand sk_one tt
        end)

   | cn_pair c1 c2 =>
       (match intc c1 E with
        | Cand sk1 Q1 =>
            (match intc c2 E with
             | Cand sk2 Q2 =>
                 Cand (sk_sigma sk1 sk2) (Q1, Q2)
             end)
        end)

   | cn_pi1 c' =>
       (match intc c' E with
        | Cand (sk_sigma sk1 sk2) Q =>
            Cand sk1 (fst Q)

        | _ => Cand sk_one tt
        end)

   | cn_pi2 c' =>
       (match intc c' E with
        | Cand (sk_sigma sk1 sk2) Q =>
            Cand sk2 (snd Q)

        | _ => Cand sk_one tt
        end)

   | cn_star =>
       Cand sk_one tt

   | cn_unit =>
       Cand sk_type (unit_interp E)

   | cn_arrow t1 t2 =>
       Cand sk_type
       (arrow_interp t1 t2 (fun E => out sk_type (intc t1 E)) (fun E => out sk_type (intc t2 E)) E)

   | cn_prod t1 t2 =>
       Cand sk_type
       (prod_interp t1 t2 (fun E => out sk_type (intc t1 E)) (fun E => out sk_type (intc t2 E)) E)

   | cn_all k t =>
       Cand sk_type
       (all_interp k t (intk k) (fun E => out sk_type (intc t E)) E)

   | cn_exists k t =>
       Cand sk_type
       (exists_interp_st k t (intk k) (fun E => out sk_type (intc t E)) E)

   | _ => Cand sk_one tt
   end)
.


Definition intks (k : (* kd *) term) (E : env) (c c' : (* cn *) term) (Q : cand) :=
  intk k E c c' Q Q.


Definition intt (t : (* cn *) term) (E : env) : trelation :=
  out sk_type (intc t E).


Definition mi (T T' : (* sg *) term) (R : mrelation) :=
  mpst T T' (fun V V' => exists Q, R V V' Q).


Definition miev (T T' : (* sg *) term) (R : mrelation) :=
  mev T T' (mi T T' R).


Definition mpuev (T T' : (* sg *) term) (R : mrelation) (M M' : (* md *) term) (Q : cand) :=
  mevp T T' (fun V V' => R V V' Q) M M'.


Lemma mpuev_miev :
  forall T T' (R : mrelation) M M' Q,
    mvrel T T' (fun V V' => exists Q, R V V' Q)
    -> mpuev T T' R M M' Q
    -> miev T T' R M M'.
Proof.
intros T T' R M M' Q HofR Hevp.
unfold mpuev in Hevp.
unfold miev.
so (mevp_forget _#5 Hevp) as Hev; clear Hevp.
refine (mev_mono _#7 Hev); [].
clear M M' Hev.
intros N N' HN.
unfold mi.
apply mpst_expand; eauto.
Qed.


Inductive sg_one_interp : env -> mrelation :=
| sg_one_interp_ {E}
    : sg_one_interp E md_star md_star (Cand sk_one tt)
.


Inductive sg_stat_interp (k : (* kd *) term) : env -> mrelation :=
| sg_stat_interp_ {E c c' Q}
    : intks k E c c' Q
      -> sg_stat_interp k E (md_stat c) (md_stat c') Q
.


Inductive sg_dyn_interp (t : (* cn *) term) : env -> mrelation :=
| sg_dyn_interp_ {E v v'}
    : ofm nil impure (md_dyn v) (sg_dyn (subst (left E) t))
      -> ofm nil impure (md_dyn v') (sg_dyn (subst (right E) t))
      -> intt t E v v'
      -> sg_dyn_interp t E (md_dyn v) (md_dyn v') (Cand sk_one tt)
.


Inductive sg_pigen_interp
  (T1 T2 : (* sg *) term) (IT1 IT2 : env -> mrelation) : env -> mrelation :=
| sg_pigen_interp_ {E V V'}
    : ofm nil impure V (subst (left E) (sg_pigen T1 T2))
      -> ofm nil impure V' (subst (right E) (sg_pigen T1 T2))
      -> mvalue nil V
      -> mvalue nil V'
      -> (forall W W' Q,
            IT1 E W W' Q
            -> exists c c',
                 fstm nil W c
                 /\ fstm nil W' c'
                 /\ miev (subst (dot c (left E)) T2) (subst (dot c' (right E)) T2)
                    (IT2 (E; Sta c c' Q)) (md_appgn V W) (md_appgn V' W'))
      -> sg_pigen_interp T1 T2 IT1 IT2 E V V' (Cand sk_one tt)
.


Inductive sg_piapp_interp
  (T1 T2 : (* sg *) term) (IT1 IT2 : env -> mrelation) : env -> mrelation :=
| sg_piapp_interp_ {E V V' c c' k1 k2} {F : precand (sk_pi (simp k1) (simp k2))}
    : ofm nil impure V (subst (left E) (sg_piapp T1 T2))
      -> ofm nil impure V' (subst (right E) (sg_piapp T1 T2))
      -> mvalue nil V
      -> mvalue nil V'
      -> fstm nil V c
      -> fstm nil V' c'
      -> fsts T1 k1
      -> fsts T2 k2
      -> intks (kd_pi k1 k2) E c c' (Cand (sk_pi (simp k1) (simp k2)) F) 
      -> (forall W W' Q,
            IT1 E W W' Q
            -> exists d d',
                 fstm nil W d
                 /\ fstm nil W' d'
                 /\ mpuev (subst (dot d (left E)) T2) (subst (dot d' (right E)) T2)
                    (IT2 (E; Sta d d' Q)) (md_appap V W) (md_appap V' W') (Cand (simp k2) (F d d' (out (simp k1) Q))))
      -> sg_piapp_interp T1 T2 IT1 IT2 E V V' (Cand (sk_pi (simp k1) (simp k2)) F)
.


Inductive sg_sigma_interp
  (T1 T2 : (* sg *) term) (IT1 IT2 : env -> mrelation) : env -> mrelation :=
| sg_sigma_interp_ {E V1 V1' V2 V2' c1 c1' k1 k2} {P : precand (sk_sigma (simp k1) (simp k2))}
    : ofm nil impure (md_pair V1 V2) (subst (left E) (sg_sigma T1 T2))
      -> ofm nil impure (md_pair V1' V2') (subst (right E) (sg_sigma T1 T2))
      -> mvalue nil V1
      -> mvalue nil V2
      -> mvalue nil V1'
      -> mvalue nil V2'
      -> fstm nil V1 c1
      -> fstm nil V1' c1'
      -> fsts T1 k1
      -> fsts T2 k2
      -> IT1 E V1 V1' (Cand (simp k1) (fst P))
      -> IT2 (E; Sta c1 c1' (Cand (simp k1) (fst P))) V2 V2' (Cand (simp k2) (snd P))
      -> sg_sigma_interp T1 T2 IT1 IT2 E (md_pair V1 V2) (md_pair V1' V2') (Cand (sk_sigma (simp k1) (simp k2)) P)
.


Fixpoint ints (T : (* sg *) term) (E : env) {struct T} : mrelation :=
  (match T with
   | sg_one =>
       sg_one_interp E

   | sg_stat k =>
       sg_stat_interp k E

   | sg_dyn t =>
       sg_dyn_interp t E

   | sg_pigen T1 T2 =>
       sg_pigen_interp T1 T2 (ints T1) (ints T2) E

   | sg_piapp T1 T2 =>
       sg_piapp_interp T1 T2 (ints T1) (ints T2) E

   | sg_sigma T1 T2 =>
       sg_sigma_interp T1 T2 (ints T1) (ints T2) E

   | _ => (fun _ _ _ => False)
   end).


Inductive intctx : context -> env -> env -> Prop :=
| intctx_nil
    : intctx nil nil nil

| intctx_cn {G k E E' c1 c1' c2 c2' Q Q'}
    : intctx G E E'
      -> eqc nil c1 c1' (subst (left E) k)
      -> eqc nil c2 c2' (subst (right E) k)
      -> intk k E c1 c2 Q Q'
      -> intctx (G; cl_cn k) (E; Sta c1 c2 Q) (E'; Sta c1' c2' Q')

| intctx_tm {G t E E' m1 m2 m1' m2'}
    : intctx G E E'
      -> intctx (G; cl_tm t) (E; Dyn m1 m2) (E'; Dyn m1' m2')

| intctx_md {G T c E E' m1 m2 m1' m2'}
    : intctx G E E'
      -> intctx (G; cl_md T c) (E; Dyn m1 m2) (E'; Dyn m1' m2')
.


Lemma intctx_length :
  forall G E1 E2,
    intctx G E1 E2
    -> length G = length E1.
Proof.
intros G E1 E2 H.
induct H; intros; simpl; omega.
Qed.


Lemma type_interp_eq :
  forall E c c' Q Q',
    intk kd_type E c c' Q Q'
    -> Q = Q'.
Proof.
intros E c c' Q Q' H.
invert H; intros; subst; auto.
Qed.


Lemma type_interp_tvrel :
  forall E c c' Q Q',
    intk kd_type E c c' Q Q'
    -> tvrel c c' (out sk_type Q).
Proof.
intros E c c' Q Q' Hint.
invert Hint; [].
intros R _ _ Htvrel _ <- <-.
simpl.
rewrite -> cast_id; [].
assumption.
Qed.


Lemma type_interp_tpclosed :
  forall E c c' Q Q',
    intk kd_type E c c' Q Q'
    -> tpclosed c c' (out sk_type Q).
Proof.
intros E c c' Q Q' Hint.
invert Hint; [].
intros R _ _ _ Hcl <- <-.
simpl.
rewrite -> cast_id; [].
assumption.
Qed.


Lemma sing_interp_eq :
  forall E b c c' Q Q',
    intk (kd_sing b) E c c' Q Q'
    -> Q = Q'.
Proof.
intros E b c c' Q Q' H.
invert H; intros; subst; auto.
Qed.


Lemma sing_interp_tvrel :
  forall E b c c' Q Q',
    intk (kd_sing b) E c c' Q Q'
    -> tvrel c c' (out sk_type Q).
Proof.
intros E b c c' Q Q' Hint.
invert Hint; [].
intros _ _ Htvrel _ <- <-.
simpl.
rewrite -> cast_id; [].
assumption.
Qed.


Lemma sing_interp_tpclosed :
  forall E b c c' Q Q',
    intk (kd_sing b) E c c' Q Q'
    -> tpclosed c c' (out sk_type Q).
Proof.
intros E b c c' Q Q' Hint.
invert Hint; [].
intros _ _ _ Hcl <- <-.
simpl.
rewrite -> cast_id; [].
assumption.
Qed.


Lemma intt_unit_unroll :
  forall E,
    intt cn_unit E = unit_interp E.
Proof.
intros E.
unfold intt.
simpl.
rewrite -> cast_id; [].
reflexivity.
Qed.


Lemma intt_arrow_unroll :
  forall t1 t2 E,
    intt (cn_arrow t1 t2) E = arrow_interp t1 t2 (intt t1) (intt t2) E.
Proof.
intros t1 t2 E.
unfold intt.
simpl.
rewrite -> cast_id; [].
reflexivity.
Qed.


Lemma intt_prod_unroll :
  forall t1 t2 E,
    intt (cn_prod t1 t2) E = prod_interp t1 t2 (intt t1) (intt t2) E.
Proof.
intros t1 t2 E.
unfold intt.
simpl.
rewrite -> cast_id; [].
reflexivity.
Qed.


Lemma intt_all_unroll :
  forall k t E,
    intt (cn_all k t) E = all_interp k t (intk k) (intt t) E.
Proof.
intros t1 t2 E.
unfold intt.
simpl.
rewrite -> cast_id; [].
reflexivity.
Qed.


Lemma intt_exists_unroll :
  forall k t E,
    intt (cn_exists k t) E = exists_interp_st k t (intk k) (intt t) E.
Proof.
intros t1 t2 E.
unfold intt.
simpl.
rewrite -> cast_id; [].
reflexivity.
Qed.


Hint Rewrite -> intt_unit_unroll intt_arrow_unroll intt_prod_unroll intt_all_unroll intt_exists_unroll : unfold_int.

Ltac unfold_int := autorewrite with unfold_int.


Lemma intc_skind :
  forall c E sk Q,
    intc c E = Cand sk Q
    -> sk = skof (flatten_env E) c.
Proof.
intros c.
#5 induct c;
try
  (intros;
   (match goal with
    | H : intc _ _ = _ |- _ => injection H
    end);
   intros; subst; reflexivity).

(* var *)
intros i E sk Q Heq.
simpl in Heq.
simpl.
revert E sk Q Heq.
#2 induct i.
  (* 0 *)
  intros E sk Q Heq.
  #2 destruct E as [| x E'].
    simpl in Heq.
    injection Heq.
    intros <-.
    simpl.
    reflexivity.

    #2 destruct x as [| sk' Q'].
      simpl in Heq.
      injection Heq.
      intros <-.
      simpl.
      reflexivity.

      simpl in Heq.
      subst c.
      simpl.
      reflexivity.

  (* S *)
  intros i IH E sk Q Heq.
  #2 destruct E as [| x E'].
    simpl in Heq.
    injection Heq.
    intros <-.
    simpl.
    reflexivity.

    simpl in Heq.
    simpl.
    eauto; done.

(* app *)
intros c1 IH1 c2 _ E sk Q Heq.
simpl in Heq.
simpl.
remember (intc c1 E) as C1 eqn:HeqC1.
destruct C1 as [sk1 Q1].
so (IH1 _#3 (eqsymm HeqC1)) as Heqsk1.
rewrite <- Heqsk1; [].
destruct sk1; try (injection Heq; intros; subst; auto); done.

(* pair *)
intros c1 IH1 c2 IH2 E sk Q Heq.
simpl in Heq.
simpl.
remember (intc c1 E) as C1 eqn:HeqC1.
destruct C1 as [sk1 Q1].
remember (intc c2 E) as C2 eqn:HeqC2.
destruct C2 as [sk2 Q2].
injection Heq; intros <-.
f_equal; eauto; done.

(* pi1 *)
intros c IH E sk Q Heq.
simpl in Heq.
simpl.
remember (intc c E) as C1 eqn:HeqC1.
destruct C1 as [sk1 Q1].
so (IH _#3 (eqsymm HeqC1)) as Heqsk1.
rewrite <- Heqsk1; [].
destruct sk1; try (injection Heq; intros; subst; auto); done.

(* pi2 *)
intros c IH E sk Q Heq.
simpl in Heq.
simpl.
remember (intc c E) as C1 eqn:HeqC1.
destruct C1 as [sk1 Q1].
so (IH _#3 (eqsymm HeqC1)) as Heqsk1.
rewrite <- Heqsk1; [].
destruct sk1; try (injection Heq; intros; subst; auto); done.
Qed.


Lemma intk_regular :
  forall E k c c' sk sk' Q Q',
    intk k E c c' (Cand sk Q) (Cand sk' Q')
    -> ofc nil c (subst (left E) k)
       /\ ofc nil c' (subst (right E) k)
       /\ sk = simp k
       /\ sk' = simp k.
Proof.
intros E k c c' sk sk' Q Q' Hintk.
revert c c' sk sk' Q Q' Hintk.
#5 case k; clear k; try (intros; destruct Hintk; done).

(* type *)
intros c c' sk sk' Q Q' Hintk.
invert Hintk; [].
intros _ ? ? _ _ <- <-.
do2 3 split; auto; done.

(* sing *)
intros b c c' sk sk' Q Q' Hintk.
simpl in Hintk.
remember (Cand sk Q) as P eqn:HeqP.
remember (Cand sk' Q') as P' eqn:HeqP'.
invert Hintk; [].
intros Heqc Heqc' _ _ <- <-.
so (cand_ext_1 _#4 HeqP); subst sk.
so (cand_ext_1 _#4 HeqP'); subst sk'.
#2 do2 3 split; auto.
  simp_sub.
  #2 refine (ofc_subsume (ofc_sing _) _).
    so (eqc_reg _#4 wf_nil Heqc) as (Hofc & _).
    assumption.

    apply subk_refl; [].
    apply eqk_sing; [].
    assumption.

  simp_sub.
  #2 refine (ofc_subsume (ofc_sing _) _).
    so (eqc_reg _#4 wf_nil Heqc') as (Hofc & _).
    assumption.

    apply subk_refl; [].
    apply eqk_sing; [].
    assumption.

(* pi *)
intros k1 k2 c c' sk sk' Q Q' Hintk.
invert Hintk; [].
intros _ _ Hofc Hofc' _ <- <-.
do2 3 split; auto; done.

(* sigma *)
intros k1 k2 c c' sk sk' Q Q' Hintk.
invert Hintk; [].
intros _ _ Hofc Hofc' _ _ <- <-.
do2 3 split; auto; done.

(* one *)
intros c c' sk sk' Q Q' Hintk.
invert Hintk; [].
intros Hofc Hofc' <- <-.
do2 3 split; auto; done.
Qed.


Lemma intk_reg :
  forall E k c c' Q Q',
    intk k E c c' Q Q'
    -> ofc nil c (subst (left E) k)
       /\ ofc nil c' (subst (right E) k).
Proof.
intros E k c c' Q Q' Hint.
destruct Q as (sk, Z).
destruct Q' as (sk', Z').
so (intk_regular _#8 Hint) as (H & H' & _).
auto.
Qed.


Lemma intk_csk :
  forall k E c c' Q Q',
    intk k E c c' Q Q'
    -> csk Q = simp k
       /\ csk Q' = simp k.
Proof.
intros k E c c' Q Q' Hint.
destruct Q as [sk Z].
destruct Q' as [sk' Z'].
simpl.
so (intk_regular _#8 Hint) as (_ & _ & Heq & Heq').
auto; done.
Qed.


Lemma intk_csk' :
  forall k E c c' Q Q',
    intk k E c c' Q Q'
    -> csk Q = csk Q'.
Proof.
intros k E c c' Q Q' H.
so (intk_csk _#6 H) as (Heq & Heq').
etransitivity; eauto; symmetry; auto.
Qed.


Lemma intctx_flatten :
  forall G E E',
    intctx G E E'
    -> flatten_ctx G = flatten_env E
       /\ flatten_ctx G = flatten_env E'.
Proof.
intros G E E' H.
induct H.

(* nil *)
auto; done.

(* cn *)
intros G k E E' c1 c1' c2 c2' Q Q' _ (IH1 & IH2) _ _ Hint.
destruct Q as [sk Z].
destruct Q' as [sk' Z'].
simpl.
rewrite <- IH1; [].
rewrite <- IH2; [].
so (intk_regular _#8 Hint) as (_ & _ & Heq & Heq').
rewrite -> Heq, -> Heq'; [].
split; reflexivity.

(* tm *)
intros G t E E' ? ? ? ? _ (IH1 & IH2).
split; simpl; f_equal; auto; done.

(* md *)
intros G T c E E' ? ? ? ? _ (IH1 & IH2).
split; simpl; f_equal; auto; done.
Qed.


Lemma intc_csk :
  forall c E,
    csk (intc c E) = skof (flatten_env E) c.
Proof.
intros c E.
remember (intc c E) as Q eqn:HeqQ.
destruct Q as [sk Z].
simpl.
eapply intc_skind; symmetry; eassumption.
Qed.


Lemma ofc_intc_csk :
  forall G c k E,
    wf G
    -> ofc G c k
    -> intctx G E E
    -> csk (intc c E) = simp k.
Proof.
intros G c k E Hwf Hofc HintE.
so (ofc_skof _#3 Hwf Hofc) as Heq.
rewrite -> (intctx_flatten _ _ _ HintE andel) in Heq; [].
rewrite <- Heq; [].
apply intc_csk; done.
Qed.


Lemma ofm_intc_csk :
  forall G M T c k E,
    wf G
    -> ofm G pure M T
    -> fstm G M c
    -> fsts T k
    -> intctx G E E
    -> csk (intc c E) = simp k.
Proof.
intros G M T c k E Hwf Hofm Hfstm Hfst Hint.
eapply ofc_intc_csk; eauto using ofm_fst.
Qed.


Lemma intc_lam_unroll :
  forall G k1 k2 c E,
    wf (G; cl_cn k1)
    -> ofc (G; cl_cn k1) c k2
    -> intctx G E E
    -> intc (cn_lam k1 c) E
       =
       Cand (sk_pi (simp k1) (simp k2)) (fun d d' Q => out (simp k2) (intc c (E; Sta d d' (Cand (simp k1) Q)))).
Proof.
intros G k1 k2 c E Hwf Hofc HintE.
simpl.
fold precand.
replace (Skof.skof (flatten_env E; Some (simp k1)) c) with (simp k2).
  reflexivity.
symmetry.
rewrite <- (intctx_flatten _#3 HintE andel); [].
so (ofc_skof _#3 Hwf Hofc) as H.
simpl in H.
exact H.
Qed.


(* weakening closure built in *)
Inductive subenv : env -> sub -> env -> Prop :=
| subenv_dot_wc {E E' c s Q}
    : subenv E s E'
      -> (forall E'' n, truncate n E'' E -> Q = intc (subst (sh n) c) E'')
      -> subenv E (dot c s) (E'; Sta (subst (left E) c) (subst (right E) c) Q)

| subenv_empty {E E' s m}
    : subenv E s E'
      -> subenv E (dot m s) (E'; Dyn (subst (left E) m) (subst (right E) m))

| subenv_sh {E E' n}
    : truncate n E E'
      -> subenv E (sh n) E'
.


Lemma subenv_id :
  forall E, subenv E id E.
Proof.
intros E.
apply subenv_sh; auto with sequence.
Qed.


Lemma subenv_left :
  forall E E' s,
    subenv E s E'
    -> compose s (left E) = left E'.
Proof.
intros E E' s Hs.
#3 induct Hs.

(* dot *)
intros E E' c s Q Hs IH _.
simpl.
f_equal; auto; done.

(* empty *)
intros E E' s m _ IH.
simpl.
simp_sub.
f_equal; auto; done.

(* sh *)
intros E E' n Htrunc.
#2 induct Htrunc.

(* 0 *)
intros E.
simp_sub.
reflexivity.

(* S *)
intros n ent E E' Htrunc IH.
repl (S n) with (n + 1) by omega.
rewrite <- compose_sh_sh; [].
rewrite -> compose_assoc; [].
rewrite <- IH; [].
f_equal; [].
destruct ent as [|]; reflexivity.
Qed.


Lemma subenv_right :
  forall E E' s,
    subenv E s E'
    -> compose s (right E) = right E'.
Proof.
intros E E' s Hs.
#3 induct Hs.

(* dot *)
intros E E' c s Q Hs IH _.
simpl.
f_equal; auto; done.

(* empty *)
intros E E' s m _ IH.
simpl.
simp_sub.
f_equal; auto; done.

(* sh *)
intros E E' n Htrunc.
#2 induct Htrunc.

(* 0 *)
intros E.
simp_sub.
reflexivity.

(* S *)
intros n ent E E' Htrunc IH.
repl (S n) with (n + 1) by omega.
rewrite <- compose_sh_sh; [].
rewrite -> compose_assoc; [].
rewrite <- IH; [].
f_equal; [].
destruct ent as [|]; reflexivity.
Qed.


Lemma subenv_weaken :
  forall E E' s ent,
    subenv E s E'
    -> subenv (E; ent) (compose s sh1) E'.
Proof.
intros E E' s ent Hs.
induct Hs.

(* dot *)
intros E E' c s Q Hs IH HQ.
simp_sub.
repl (subst (left E) c) with (subst (left (E; ent)) (shift1 c)).
  destruct ent; simpl; simp_sub; reflexivity.
repl (subst (right E) c) with (subst (right (E; ent)) (shift1 c)).
  destruct ent; simpl; simp_sub; reflexivity.
apply subenv_dot_wc; auto; [].
intros E'' n Htrunc.
simp_sub.
apply HQ; [].
replace (1 + n) with (n + 1) by omega.
eapply truncate_sum; eauto with sequence; done.

(* empty *)
intros E E' s m _ IH.
simp_sub.
repl (subst (left E) m) with (subst (left (E; ent)) (shift1 m)).
  destruct ent; simpl; simp_sub; reflexivity.
repl (subst (right E) m) with (subst (right (E; ent)) (shift1 m)).
  destruct ent; simpl; simp_sub; reflexivity.
apply subenv_empty; auto; done.

(* sh *)
intros E E' n Htrunc.
simp_sub.
apply subenv_sh; [].
auto with sequence; done.
Qed.


Lemma subenv_under :
  forall E E' s c c' Q,
    subenv E s E'
    -> subenv (E; Sta c c' Q) (dot (var 0) (compose s sh1)) (E'; Sta c c' Q).
Proof.
intros E E' s c c' Q Hs.
replace (E'; Sta c c' Q) with (E'; Sta (subst (left (E; Sta c c' Q)) (var 0)) (subst (right (E; Sta c c' Q)) (var 0)) (intc (var 0) (E; Sta c c' Q))) by auto.
#2 apply subenv_dot_wc.
  apply subenv_weaken; auto; done.

  intros E'' n Htrunc.
  simpl.
  rewrite -> (index_nth _#5 (truncate_index _#5 Htrunc)); [].
  reflexivity.
Qed.


Inductive fenv : context -> env -> Prop :=
| fenv_nil
    : fenv nil nil

| fenv_cn {G k E c c' Q}
    : fenv G E
      -> fenv (G; cl_cn k) (E; Sta c c' Q)

| fenv_ucn {G E c c' Q}
    : fenv G E
      -> fenv (G; cl_ucn) (E; Sta c c' Q)

| fenv_tm {G t E m m'}
    : fenv G E
      -> fenv (G; cl_tm t) (E; Dyn m m')

| fenv_md {G T c E m m'}
    : fenv G E
      -> fenv (G; cl_md T c) (E; Dyn m m').


Lemma intctx_fenv :
  forall G E E',
    intctx G E E'
    -> fenv G E.
Proof.
intros G E E' H.
induct H; auto using fenv.
Qed.


Lemma intctx_fenv' :
  forall G E E',
    intctx G E E'
    -> fenv G E'.
Proof.
intros G E E' H.
induct H; auto using fenv.
Qed.


Lemma subst_int :
  (forall G E E' s k,
     fkd G k
     -> fenv G E'
     -> subenv E s E'
     -> intk (subst s k) E = intk k E')
  /\
  (forall G E E' s c,
     fcn G c
     -> fenv G E'
     -> subenv E s E'
     -> intc (subst s c) E = intc c E')
.
Proof.
exploit
  (static_form_ind
     (fun G k =>
        forall E E' s,
          fenv G E'
          -> subenv E s E'
          -> intk (subst s k) E = intk k E')
     (fun G c =>
        forall E E' s,
          fenv G E'
          -> subenv E s E'
          -> intc (subst s c) E = intc c E'))
  as Hind.

(* type *)
intros G E E' s Henv Hs.
fextensionality 4.
intros b b' Q Q'.
pextensionality.
  intro Hint.
  simp_sub in Hint.
  invert Hint; [].
  intros R Hofb Hofb' Hrel Hclosed <- <-.
  eapply type_interp_; eauto; done.

  intro Hint.
  invert Hint; [].
  intros R Hofb Hofb' Hrel Hclosed <- <-.
  eapply type_interp_; eauto; done.

(* sing *)
intros G c _ IH E E' s Henv Hs.
fextensionality 4.
intros b b' Q Q'.
simp_sub.
simpl.
pextensionality.
  intro Hint.
  invert Hint; clear Hint; [].
  intros Heqb Heqb' Hrel Hcl <- <-.
  simp_sub in Heqb.
  simp_sub in Heqb'.
  so (IH _ _ _ Henv Hs) as Heq.
  rewrite -> Heq; [].
  rewrite -> Heq in Hrel, Hcl; [].
  #2 apply sing_interp_; auto.
    rewrite -> (subenv_left _#3 Hs) in Heqb; [].
    assumption.

    rewrite -> (subenv_right _#3 Hs) in Heqb'; [].
    assumption.

  intro Hint.
  invert Hint; clear Hint; [].
  intros Heqb Heqb' Hrel Hcl <- <-.
  so (IH _ _ _ Henv Hs) as Heq.
  rewrite <- Heq; [].
  rewrite <- Heq in Hrel, Hcl; [].
  #2 apply sing_interp_; auto.
    rewrite <- subst_compose; [].
    rewrite -> (subenv_left _#3 Hs); [].
    assumption.

    rewrite <- subst_compose; [].
    rewrite -> (subenv_right _#3 Hs); [].
    assumption.

(* pi *)
intros G k1 k2 Hk1 IH1 Hk2 IH2 E E' s Henv Hs.
so (fkd_hkd _ _ Hk1) as H; renameover H into Hk1.
so (fkd_hkd _ _ Hk2) as H; renameover H into Hk2.
fextensionality 4.
intros c c' Z Z'.
simp_sub.
pextensionality.
  intro H.
  simpl in H.
  invert H; clear H; [].
  intros F F' Hofc Hofc' Hintapp <- <-.
  simpl.
  revert F F' Hintapp.
  repl (simp (subst s k1)) with (simp k1).
    symmetry; apply subst_simp; auto; done.
  repl (simp (subst (dot (var 0) (compose s sh1)) k2)) with (simp k2).
    symmetry; apply subst_simp; auto; done.
  intros F F' Hintapp.
  #3 apply pi_interp_.
    rewrite <- (subenv_left _#3 Hs); [].
    simp_sub.
    simp_sub in Hofc.
    assumption.

    rewrite <- (subenv_right _#3 Hs); [].
    simp_sub.
    simp_sub in Hofc'.
    assumption.

    intros d d' Q Q' d'' d''' Hintk1 Heqd Heqd'.
    so (IH2 (E; Sta d d' (Cand (simp k1) Q)) (E'; Sta d d' (Cand (simp k1) Q)) (dot (var 0) (compose s sh1)) (fenv_ucn Henv) (subenv_under _#6 Hs)) as H2; clear IH2.
    rewrite <- H2; clear H2; [].
    #3 apply Hintapp.
      so (IH1 E E' s Henv Hs) as H1; clear IH1.
      rewrite -> H1; [].
      exact Hintk1.

      rewrite <- (subenv_left _#3 Hs) in Heqd; [].
      simp_sub.
      assumption.

      rewrite <- (subenv_right _#3 Hs) in Heqd'; [].
      simp_sub.
      assumption.

  intro H.
  simpl in H.
  invert H; clear H; [].
  intros F F' Hofc Hofc' Hintapp <- <-.
  simpl.
  revert F F' Hintapp.
  repl (simp k1) with (simp (subst s k1)).
    apply subst_simp; auto; done.
  repl (simp k2) with (simp (subst (dot (var 0) (compose s sh1)) k2)).
    apply subst_simp; auto; done.
  intros F F' Hintapp.
  #3 apply pi_interp_.
    rewrite <- (subenv_left _#3 Hs) in Hofc; [].
    simp_sub in Hofc.
    simp_sub.
    assumption.

    rewrite <- (subenv_right _#3 Hs) in Hofc'; [].
    simp_sub in Hofc'.
    simp_sub.
    assumption.

    intros d d' Q Q' d'' d''' Hintk1 Heqd Heqd'.
    so (IH2 (E; Sta d d' (Cand _ Q)) (E'; Sta d d' (Cand _ Q)) (dot (var 0) (compose s sh1)) (fenv_ucn Henv) (subenv_under _#6 Hs)) as H2; clear IH2.
    rewrite -> H2; clear H2; [].
    #3 apply Hintapp; clear Hintapp.
      so (IH1 E E' s Henv Hs) as H1; clear IH1.
      rewrite <- H1; [].
      assumption.

      rewrite <- (subenv_left _#3 Hs); [].
      simp_sub in Heqd.
      assumption.

      rewrite <- (subenv_right _#3 Hs); [].
      simp_sub in Heqd'.
      assumption.

(* sigma *)
intros G k1 k2 Hk1 IH1 Hk2 IH2 E E' s Henv Hs.
so (fkd_hkd _ _ Hk1) as H; renameover H into Hk1.
so (fkd_hkd _ _ Hk2) as H; renameover H into Hk2.
fextensionality 4.
intros c c' Z Z'.
simp_sub.
pextensionality.
  intro H.
  simpl in H.
  invert H; clear H; [].
  intros P P' Hofc Hofc' Heqpi1 Heqpi2 <- <-.
  simpl.
  revert P P' Heqpi1 Heqpi2.
  repl (simp (subst s k1)) with (simp k1).
    symmetry; apply subst_simp; auto; done.
  repl (simp (subst (dot (var 0) (compose s sh1)) k2)) with (simp k2).
    symmetry; apply subst_simp; auto; done.
  intros P P' Heqpi1 Heqpi2.
  #4 apply sigma_interp_.
    simp_sub in Hofc.
    rewrite <- (subenv_left _#3 Hs); [].
    simp_sub.
    assumption.

    simp_sub in Hofc'.
    rewrite <- (subenv_right _#3 Hs); [].
    simp_sub.
    assumption.

    so (IH1 _#3 Henv Hs) as H1.
    rewrite <- H1; clear H1; [].
    assumption.

    so (IH2 _#3 (fenv_ucn Henv) (subenv_under _ _ _ (cn_pi1 c) (cn_pi1 c') (Cand (simp k1) (fst P)) Hs)) as H2; clear IH2.
    rewrite <- H2; clear H2; [].
    assumption.

  intro H.
  simpl in H.
  invert H; clear H; [].
  intros P P' Hofc Hofc' Heqpi1 Heqpi2 <- <-.
  simpl.
  revert P P' Heqpi1 Heqpi2.
  repl (simp k1) with (simp (subst s k1)).
    apply subst_simp; auto; done.
  repl (simp k2) with (simp (subst (dot (var 0) (compose s sh1)) k2)).
    apply subst_simp; auto; done.
  intros P P' Heqpi1 Heqpi2.
  #4 apply sigma_interp_.
    simp_sub in Hofc.
    rewrite <- (subenv_left _#3 Hs) in Hofc; [].
    simp_sub in Hofc.
    simp_sub.
    assumption.

    simp_sub in Hofc'.
    rewrite <- (subenv_right _#3 Hs) in Hofc'; [].
    simp_sub in Hofc'.
    simp_sub.
    assumption.

    so (IH1 _#3 Henv Hs) as H1.
    rewrite -> H1; clear H1; [].
    assumption.

    so (IH2 _#3 (fenv_ucn Henv) (subenv_under _ _ _ (cn_pi1 c) (cn_pi1 c') (Cand (simp (subst s k1)) (fst P)) Hs)) as H2; clear IH2.
    rewrite -> H2; clear H2; [].
    assumption.

(* one *)
intros _ E E' s _ Hs.
fextensionality 4.
intros b b' Q Q'.
pextensionality.
  intro Hint.
  simp_sub in Hint.
  invert Hint; clear Hint; [].
  intros Hofb Hofb' <- <-.
  apply one_interp_; auto; done.

  intro Hint.
  invert Hint; clear Hint; [].
  intros Hofb Hofb' <- <-.
  simp_sub.
  apply one_interp_; auto; done.

(* var *)
intros G i k Hindex E1 E2 s Henv Hs.
simp_sub.
simpl.
revert G i Hindex Henv.
#3 induct Hs.
  (* dot *)
  intros E E' c s Q Hs IH HQ G i Hindex Henv.
  destruct i as [| i].
    (* 0 *)
    simp_sub.
    simpl.
    so (HQ E 0 truncate_0) as H.
    simp_sub in H.
    symmetry; assumption.
    
    (* S *)
    rewrite -> project_dot; [].
    simpl.
    invert Hindex; [].
    intros A G' Hindex' <-.
    assert (fenv G' E') as Henv' by (invert Henv; auto).
    eapply IH; eauto; done.

  (* empty *)
  intros E E' s m _ IH G i Hindex Henv.
  destruct i as [| i].
    (* 0 *)
    exfalso.
    invert Hindex; [].
    intros G' <-.
    invert Henv; done.

    rewrite -> project_dot; [].
    simpl.
    invert Hindex; [].
    intros A G' Hindex' <-.
    assert (fenv G' E') as Henv' by (invert Henv; auto).
    eapply IH; eauto; done.

  (* sh *)
  intros E E' n Htrunc G i Hindex Henv.
  simp_sub.
  simpl.
  rewrite -> (truncate_nth _#6 Htrunc); [].
  reflexivity.

(* varu *)
intros G i Hindex E1 E2 s Henv Hs.
simp_sub.
simpl.
revert G i Hindex Henv.
#3 induct Hs.
  (* dot *)
  intros E E' c s Q Hs IH HQ G i Hindex Henv.
  destruct i as [| i].
    (* 0 *)
    simp_sub.
    simpl.
    so (HQ E 0 truncate_0) as H.
    simp_sub in H.
    symmetry; assumption.
    
    (* S *)
    rewrite -> project_dot; [].
    simpl.
    invert Hindex; [].
    intros A G' Hindex' <-.
    assert (fenv G' E') as Henv' by (invert Henv; auto).
    eapply IH; eauto; done.

  (* empty *)
  intros E E' s m _ IH G i Hindex Henv.
  destruct i as [| i].
    (* 0 *)
    exfalso.
    invert Hindex; [].
    intros G' <-.
    invert Henv; done.

    rewrite -> project_dot; [].
    simpl.
    invert Hindex; [].
    intros A G' Hindex' <-.
    assert (fenv G' E') as Henv' by (invert Henv; auto).
    eapply IH; eauto; done.

  (* sh *)
  intros E E' n Htrunc G i Hindex Henv.
  simp_sub.
  simpl.
  rewrite -> (truncate_nth _#6 Htrunc); [].
  reflexivity.

(* lam *)
intros G k c Hk IHk _ IHc E E' s Henv Hs.
so (fkd_hkd _ _ Hk) as H; renameover H into Hk.
simp_sub.
simpl.
rewrite -> subst_simp; auto; [].
fold (precand (simp k)).
repl (skof (flatten_env E; Some (simp k)) (subst (dot (var 0) (compose s sh1)) c)) with (skof (flatten_env E'; Some (simp k)) c).
  (* This is super-ugly, but it gets us the result for free. *)
  so (nonsense) as b.
  so (bogus_precand (simp k)) as Z.
  so (IHc _ _ _ (fenv_ucn Henv) (subenv_under _#3 b b (Cand (simp k) Z) Hs)) as Heq1.
  remember (intc c (E'; Sta b b (Cand (simp k) Z))) as Q eqn:Heq2.
  destruct Q as [sk Z'].
  so (intc_skind _#4 Heq1) as Heqsk1.
  so (intc_skind _#4 (eqsymm Heq2)) as Heqsk2.
  simpl in Heqsk1.
  simpl in Heqsk2.
  rewrite <- Heqsk1; [].
  rewrite <- Heqsk2; [].
  reflexivity.

  f_equal; [].
  fextensionality 3.
  intros b b' Q.
  f_equal; [].
  apply IHc; auto using fenv_ucn, subenv_under; done.

(* app *)
intros G c1 c2 _ IH1 _ IH2 E E' s Henv Hs.
simp_sub.
so (IH1 _#3 Henv Hs) as H1; clear IH1.
simpl.
rewrite -> H1; clear H1; [].
forget (intc c1 E') as Q.
destruct Q as [sk Z].
revert Z.
case sk; auto; [].
intros sk1 sk2 Z.
f_equal; [].
#3 f_equal.
  simp_sub.
  f_equal; apply subenv_left; auto; done.

  simp_sub.
  f_equal; apply subenv_right; auto; done.

  f_equal; [].
  apply IH2; auto; done.

(* pair *)
intros G c1 c2 _ IH1 _ IH2 E E' s Henv Hs.
simp_sub.
simpl.
rewrite (IH1 _#3 Henv Hs); [].
forget (intc c1 E') as Q.
destruct Q as [sk1 Z1].
rewrite (IH2 _#3 Henv Hs); [].
reflexivity.

(* pi1 *)
intros G c _ IH E E' s Henv Hs.
simp_sub.
simpl.
rewrite -> (IH _#3 Henv Hs); [].
forget (intc c E') as Q.
destruct Q as [sk Z].
revert Z.
case sk; auto; done.

(* pi2 *)
intros G c _ IH E E' s Henv Hs.
simp_sub.
simpl.
rewrite -> (IH _#3 Henv Hs); [].
forget (intc c E') as Q.
destruct Q as [sk Z].
revert Z.
case sk; auto; done.

(* star *)
intros G E E' s _ Hs.
simp_sub.
simpl.
reflexivity.

(* unit *)
intros _ E E' s _ Hs.
simp_sub.
simpl.
f_equal; [].
fextensionality 2.
intros e e'.
pextensionality.
  intro H.
  invert H; clear H; [].
  intros <- <-.
  apply unit_interp_; done.

  intro H.
  invert H; clear H; [].
  intros <- <-.
  apply unit_interp_; done.

(* arrow *)
intros G t1 t2 _ IH1 _ IH2 E E' s Henv Hs.
simp_sub.
simpl.
f_equal; [].
fextensionality 2.
intros v v'.
pextensionality.
  intro H; invert H; clear H; [].
  intros Hofv Hofv' Hval Hval' Hrelv.
  #3 apply arrow_interp_; auto.
    simp_sub in Hofv.
    simp_sub.
    rewrite -> (subenv_left _#3 Hs) in Hofv; [].
    assumption.
    
    simp_sub in Hofv'.
    simp_sub.
    rewrite -> (subenv_right _#3 Hs) in Hofv'; [].
    assumption.

    intros w w' Hrelw.
    simp_sub in Hrelv.
    rewrite -> (subenv_left _#3 Hs) in Hrelv; [].
    rewrite -> (subenv_right _#3 Hs) in Hrelv; [].
    rewrite -> (IH2 _#3 Henv Hs) in Hrelv; [].
    apply Hrelv; [].
    rewrite -> (IH1 _#3 Henv Hs); [].
    assumption.

  intro H; invert H; clear H; [].
  intros Hofv Hofv' Hval Hval' Hrelv.
  #3 apply arrow_interp_; auto.
    simp_sub in Hofv.
    simp_sub.
    rewrite -> (subenv_left _#3 Hs); [].
    assumption.
    
    simp_sub in Hofv'.
    simp_sub.
    rewrite -> (subenv_right _#3 Hs); [].
    assumption.

    intros w w' Hrelw.
    simp_sub.
    rewrite -> (subenv_left _#3 Hs); [].
    rewrite -> (subenv_right _#3 Hs); [].
    rewrite -> (IH2 _#3 Henv Hs); [].
    apply Hrelv; [].
    rewrite -> (IH1 _#3 Henv Hs) in Hrelw; [].
    assumption.

(* prod *)
intros G t1 t2 _ IH1 _ IH2 E E' s Henv Hs.
simp_sub.
simpl.
f_equal; [].
fextensionality 2.
intros v v'.
pextensionality.
  intro H; invert H; clear H; [].
  intros v1 v1' v2 v2' Hofv Hofv' Hval Hval' Hrel1 Hrel2 <- <-.
  simp_sub in Hofv.
  simp_sub in Hofv'.
  rewrite -> (subenv_left _#3 Hs) in Hofv; [].
  rewrite -> (subenv_right _#3 Hs) in Hofv'; [].
  rewrite -> (IH1 _#3 Henv Hs) in Hrel1; [].
  rewrite -> (IH2 _#3 Henv Hs) in Hrel2; [].
  eapply prod_interp_; eauto; done.

  intro H; invert H; clear H; [].
  intros v1 v1' v2 v2' Hofv Hofv' Hval Hval' Hrel1 Hrel2 <- <-.
  simp_sub in Hofv.
  simp_sub in Hofv'.
  rewrite <- (subenv_left _#3 Hs) in Hofv; [].
  rewrite <- (subenv_right _#3 Hs) in Hofv'; [].
  rewrite <- (IH1 _#3 Henv Hs) in Hrel1; [].
  rewrite <- (IH2 _#3 Henv Hs) in Hrel2; [].
  eapply prod_interp_; simp_sub; eauto; done.

(* all *)
intros G k t _ IHk _ IHc E E' s Henv Hs.
simp_sub.
simpl.
f_equal; [].
fextensionality 2.
intros v v'.
pextensionality.
  intro H; invert H; clear H; [].
  intros Hofv Hofv' Hval Hval' Hrelv.
  simp_sub in Hofv.
  simp_sub in Hofv'.
  rewrite <- compose_assoc in Hofv; [].
  rewrite <- compose_assoc in Hofv'; [].
  rewrite -> (subenv_left _#3 Hs) in Hofv; [].
  rewrite -> (subenv_right _#3 Hs) in Hofv'; [].
  #3 apply all_interp_; auto.
    simp_sub; assumption.

    simp_sub; assumption.

    intros c c' Q HQ.
    rewrite <- (IHk _#3 Henv Hs) in HQ; [].
    so (Hrelv c c' Q HQ) as H.
    simp_sub in H.
    rewrite -> (subenv_left _#3 Hs) in H; [].
    rewrite -> (subenv_right _#3 Hs) in H; [].
    rewrite -> (IHc _#3 (fenv_ucn Henv) (subenv_under _#6 Hs)) in H; [].
    exact H.

  intro H; invert H; clear H; [].
  intros Hofv Hofv' Hval Hval' Hrelv.
  simp_sub in Hofv.
  simp_sub in Hofv'.
  rewrite <- (subenv_left _#3 Hs) in Hofv; [].
  rewrite <- (subenv_right _#3 Hs) in Hofv'; [].
  rewrite -> compose_assoc in Hofv; [].
  rewrite -> compose_assoc in Hofv'; [].
  #3 apply all_interp_; auto.
    simp_sub; assumption.

    simp_sub; assumption.

    intros c c' Q HQ.
    rewrite -> (IHk _#3 Henv Hs) in HQ; [].
    so (Hrelv c c' Q HQ) as H.
    simp_sub.
    rewrite -> (subenv_left _#3 Hs); [].
    rewrite -> (subenv_right _#3 Hs); [].
    rewrite -> (IHc _#3 (fenv_ucn Henv) (subenv_under _#6 Hs)); [].
    exact H.

(* exists *)
intros G k t _ IHk _ IHc E E' s Henv Hs.
simp_sub.
simpl.
f_equal; [].
assert (exists_interp k t (intk k) (fun E'' => out sk_type (intc t E'')) E'
        =
        exists_interp (subst s k) (subst (dot (var 0) (compose s sh1)) t) (intk (subst s k))
        (fun E'' => out sk_type (intc (subst (dot (var 0) (compose s sh1)) t) E'')) E) as Heq.
  fextensionality 2.
  intros v v'.
  pextensionality.
    intro H; invert H; clear H; [].
    intros w w' c c' l l' u u' Q Hofv Hofv' Hval Hval' HrelQ Hrelw <- <-.
    simp_sub in Hofv.
    simp_sub in Hofv'.
    #4 apply (exists_interp_ _#13 Q); auto.
      simp_sub.
      rewrite <- compose_assoc; [].
      rewrite -> (subenv_left _#3 Hs); [].
      assumption.
  
      simp_sub.
      rewrite <- compose_assoc; [].
      rewrite -> (subenv_right _#3 Hs); [].
      assumption.

      rewrite -> (IHk _#3 Henv Hs); [].
      assumption.
  
      rewrite -> (IHc _#3 (fenv_ucn Henv) (subenv_under _#6 Hs)); [].
      assumption.

    intro H; invert H; clear H; [].
    intros w w' c c' l l' u u' Q Hofv Hofv' Hval Hval' HrelQ Hrelw <- <-.
    simp_sub in Hofv.
    simp_sub in Hofv'.
    #4 apply (exists_interp_ _#13 Q); auto.
      simp_sub.
      rewrite <- (subenv_left _#3 Hs); [].
      rewrite -> compose_assoc; [].
      assumption.
  
      simp_sub.
      rewrite <- (subenv_right _#3 Hs); [].
      rewrite -> compose_assoc; [].
      assumption.

      rewrite <- (IHk _#3 Henv Hs); [].
      assumption.
  
      rewrite <- (IHc _#3 (fenv_ucn Henv) (subenv_under _#6 Hs)); [].
      assumption.
fextensionality 2.
intros v v'.
unfold exists_interp_st.
pextensionality.
  intro H.
  destruct H as (Hrelv & Hofv & Hofv' & Hvalv & Hvalv').
  simp_sub in Hofv.
  simp_sub in Hofv'.
  rewrite <- compose_assoc in Hofv; [].
  rewrite <- compose_assoc in Hofv'; [].
  rewrite -> (subenv_left _#3 Hs) in Hofv; [].
  rewrite -> (subenv_right _#3 Hs) in Hofv'; [].
  #3 do2 4 split; auto.
    intros f f' Hrelf.
    apply Hrelv; [].
    simp_sub.
    rewrite <- !compose_assoc; [].
    rewrite -> (subenv_left _#3 Hs); [].
    rewrite -> (subenv_right _#3 Hs); [].
    simp_sub in Hrelf.
    rewrite -> Heq in Hrelf; [].
    exact Hrelf.

    simp_sub; assumption.

    simp_sub; assumption.

  intro H.
  destruct H as (Hrelv & Hofv & Hofv' & Hvalv & Hvalv').
  simp_sub in Hofv.
  simp_sub in Hofv'.
  #3 do2 4 split; auto.
    intros f f' Hrelf.
    apply Hrelv; [].
    simp_sub.
    rewrite <- (subenv_left _#3 Hs); [].
    rewrite <- (subenv_right _#3 Hs); [].
    rewrite -> !compose_assoc; [].
    simp_sub in Hrelf.
    rewrite <- Heq in Hrelf; [].
    exact Hrelf.

    simp_sub.
    rewrite <- compose_assoc; [].
    rewrite -> (subenv_left _#3 Hs); [].
    assumption.

    simp_sub.
    rewrite <- compose_assoc; [].
    rewrite -> (subenv_right _#3 Hs); [].
    assumption.

(* peroration *)
destruct Hind; split; intros; eauto; done.
Qed.


Lemma subst_intk :
  forall G E E' s k,
    fkd G k
    -> fenv G E'
    -> subenv E s E'
    -> intk (subst s k) E = intk k E'.
Proof.
exact (subst_int andel).
Qed.  


Lemma subst_intc :
  forall G E E' s c,
    fcn G c
    -> fenv G E'
    -> subenv E s E'
    -> intc (subst s c) E = intc c E'.
Proof.
exact (subst_int ander).
Qed.  


Lemma subst_intks :
  forall G E E' s k,
    fkd G k
    -> fenv G E'
    -> subenv E s E'
    -> intks (subst s k) E = intks k E'.
Proof.
intros G E E' s k Hk Henv Hs.
unfold intks.
rewrite -> (subst_intk _#5 Hk Henv Hs).
reflexivity.
Qed.


Lemma subst_intt :
  forall G E E' s t,
    fcn G t
    -> fenv G E'
    -> subenv E s E'
    -> intt (subst s t) E = intt t E'.
Proof.
intros G E E' s t Ht Henv Hs.
unfold intt.
rewrite -> (subst_intc _#5 Ht Henv Hs).
reflexivity.
Qed.


Lemma subst_ints :
  forall G E E' s T,
    fsg G T
    -> fenv G E'
    -> subenv E s E'
    -> ints (subst s T) E = ints T E'.
Proof.
intros G E E' s T HT Henv Hs.
revert E E' s Henv Hs.
#6 induct HT.

(* one *)
intros _ E E' s _ Hs.
simp_sub.
fextensionality 3.
intros V V' Q.
pextensionality.
  intro H.
  invert H; clear H; [].
  intros <- <- <-.
  apply sg_one_interp_; done.

  intro H.
  invert H; clear H; [].
  intros <- <- <-.
  apply sg_one_interp_; done.

(* stat *)
intros G k Hk E E' s Henv Hs.
simp_sub.
fextensionality 3.
intros V V' Q.
pextensionality.
  intro H; invert H; clear H; [].
  intros c c' Hintc <- <-.
  apply sg_stat_interp_; [].
  rewrite -> (subst_intks _#5 Hk Henv Hs) in Hintc; [].
  assumption.

  intro H; invert H; clear H; [].
  intros c c' Hintc <- <-.
  apply sg_stat_interp_; [].
  rewrite <- (subst_intks _#5 Hk Henv Hs) in Hintc; [].
  assumption.

(* dyn *)
intros G t Ht E E' s Henv Hs.
simp_sub.
fextensionality 3.
intros V V' Q.
pextensionality.
  intro H; invert H; clear H; [].
  intros v v' Hof Hof' Hintt <- <- <-.
  rewrite -> (subst_intt _#5 Ht Henv Hs) in Hintt; [].
  simp_sub in Hof.
  simp_sub in Hof'.
  rewrite -> (subenv_left _#3 Hs) in Hof; [].
  rewrite -> (subenv_right _#3 Hs) in Hof'; [].
  apply sg_dyn_interp_; auto; done.

  intro H; invert H; clear H; [].
  intros v v' Hof Hof' Hintt <- <- <-.
  rewrite <- (subst_intt _#5 Ht Henv Hs) in Hintt; [].
  simp_sub in Hof.
  simp_sub in Hof'.
  rewrite <- (subenv_left _#3 Hs) in Hof; [].
  rewrite <- (subenv_right _#3 Hs) in Hof'; [].
  apply sg_dyn_interp_; simp_sub; auto; done.

(* pigen *)
intros G T1 T2 _ IH1 _ IH2 E E' s Henv Hs.
simp_sub.
fextensionality 3.
intros V V' Q.
pextensionality.
  intro H; invert H; clear H; [].
  intros HofV HofV' HvalV HvalV' Hintapp <-.
  simp_sub in HofV.
  simp_sub in HofV'.
  #3 apply sg_pigen_interp_; auto.
    rewrite <- ! (subenv_left _#3 Hs); [].
    simp_sub.
    assumption.

    rewrite <- ! (subenv_right _#3 Hs); [].
    simp_sub.
    assumption.

    intros W W' Q Hint.
    fold ints in Hint.
    fold ints.
    rewrite <- (IH1 _#3 Henv Hs) in Hint; [].
    so (Hintapp _#3 Hint) as (c & c' & Hfst & Hfst' & H); clear Hintapp Hint.
    exists c, c'.
    do2 2 split; auto; [].
    simp_sub in H.
    rewrite -> (subenv_left _#3 Hs) in H; [].
    rewrite -> (subenv_right _#3 Hs) in H; [].
    force_exact H; clear H.
    f_equal; [].
    apply IH2; auto using subenv_under, fenv_ucn; done.

  intro H; invert H; clear H; [].
  intros HofV HofV' HvalV HvalV' Hintapp <-.
  #3 apply sg_pigen_interp_; auto.
    simp_sub.
    rewrite <- ! (subenv_left _#3 Hs) in HofV; [].
    simp_sub in HofV.
    assumption.

    simp_sub.
    rewrite <- ! (subenv_right _#3 Hs) in HofV'; [].
    simp_sub in HofV'.
    assumption.

    intros W W' Q Hint.
    fold ints in Hint.
    fold ints.
    rewrite -> (IH1 _#3 Henv Hs) in Hint; [].
    so (Hintapp _#3 Hint) as (c & c' & Hfst & Hfst' & H); clear Hintapp Hint.
    exists c, c'.
    do2 2 split; auto; [].
    simp_sub.
    rewrite -> (subenv_left _#3 Hs); [].
    rewrite -> (subenv_right _#3 Hs); [].
    force_exact H; clear H.
    f_equal; [].
    symmetry.
    apply IH2; auto using subenv_under, fenv_ucn; done.

(* piapp *)
intros G T1 T2 HT1 IH1 HT2 IH2 E E' s Henv Hs.
simp_sub.
fextensionality 3.
intros V V' Q.
pextensionality.
  intro H; invert H; clear H; [].
  intros c c' k1s k2s F HofV HofV' Hval Hval' Hfstm Hfstm' Hfst1s Hfst2s HintF Hintapp <-.
  so (hsg_fsts _ (fsg_hsg _ _ HT1)) as (k1 & Hfst1).
  so (fsts_fun _#3 Hfst1s (subst_fsts _#3 Hfst1)); subst k1s; clear Hfst1s.
  so (hsg_fsts _ (fsg_hsg _ _ HT2)) as (k2 & Hfst2).
  so (fsts_fun _#3 Hfst2s (subst_fsts _#3 Hfst2)); subst k2s; clear Hfst2s.
  so (fsts_fkd _#3 HT1 Hfst1) as Hk1.
  so (fsts_fkd _#3 HT2 Hfst2) as Hk2.
  revert F HintF Hintapp.
  rewrite -> (subst_simp s _ (fkd_hkd _ _ Hk1)); [].
  rewrite -> (subst_simp (dot (var 0) (compose s sh1)) _ (fkd_hkd _ _ Hk2)); [].
  intros F HintF Hintapp.
  #4 apply (sg_piapp_interp_ _#7 c c'); auto.
    simp_sub in HofV.
    rewrite <- (subenv_left _#3 Hs).
    simp_sub.
    assumption.

    simp_sub in HofV'.
    rewrite <- (subenv_right _#3 Hs).
    simp_sub.
    assumption.

    rewrite <- (subst_intks _#5 (fkd_pi Hk1 Hk2) Henv Hs); [].
    simp_sub.
    assumption.

    fold ints.
    intros W W' Q Hint1.
    rewrite <- (IH1 _#3 Henv Hs) in Hint1; [].
    so (Hintapp _#3 Hint1) as (d & d' & HfstW & HfstW' & Hint2); clear Hintapp Hint1.
    exists d, d'.
    do2 2 split; auto; [].
    simp_sub in Hint2.
    rewrite <- (subenv_left _#3 Hs); [].
    rewrite <- (subenv_right _#3 Hs); [].
    force_exact Hint2.
    f_equal; [].
    apply IH2; auto using subenv_under, fenv_ucn; done.

  intro H; invert H; clear H; [].
  intros c c' k1 k2 F HofV HofV' Hval Hval' Hfstm Hfstm' Hfst1 Hfst2 HintF Hintapp <-.
  so (fsts_fkd _#3 HT1 Hfst1) as Hk1.
  so (fsts_fkd _#3 HT2 Hfst2) as Hk2.
  revert F HintF Hintapp.
  rewrite <- (subst_simp s _ (fkd_hkd _ _ Hk1)); [].
  rewrite <- (subst_simp (dot (var 0) (compose s sh1)) _ (fkd_hkd _ _ Hk2)); [].
  intros F HintF Hintapp.
  #4 apply (sg_piapp_interp_ _#7 c c'); auto using subst_fsts.
    rewrite <- (subenv_left _#3 Hs) in HofV; [].
    simp_sub in HofV.
    simp_sub.
    assumption.
    
    rewrite <- (subenv_right _#3 Hs) in HofV'; [].
    simp_sub in HofV'.
    simp_sub.
    assumption.

    rewrite <- (subst_intks _#5 (fkd_pi Hk1 Hk2) Henv Hs) in HintF; [].
    simp_sub in HintF.
    assumption.

    fold ints.
    intros W W' Q Hint1.
    rewrite -> (IH1 _#3 Henv Hs) in Hint1; [].
    so (Hintapp _#3 Hint1) as (d & d' & HfstW & HfstW' & Hint2); clear Hintapp Hint1.
    exists d, d'.
    do2 2 split; auto; [].
    simp_sub.
    rewrite <- (subenv_left _#3 Hs) in Hint2; [].
    rewrite <- (subenv_right _#3 Hs) in Hint2; [].
    force_exact Hint2.
    f_equal; [].
    symmetry.
    apply IH2; auto using subenv_under, fenv_ucn; done.

(* sigma *)
intros G T1 T2 HT1 IH1 HT2 IH2 E E' s Henv Hs.
simp_sub.
fextensionality 3.
intros V V' Q.
pextensionality.
  intro H; invert H; clear H; [].
  intros V1 V1' V2 V2' c1 c1' k1s k2s P Hof Hof' Hval1 Hval2 Hval1' Hval2' HfstV1 HfstV1' Hfst1s Hfst2s Hint1 Hint2 <- <- <-.
  so (hsg_fsts _ (fsg_hsg _ _ HT1)) as (k1 & Hfst1).
  so (fsts_fun _#3 Hfst1s (subst_fsts _#3 Hfst1)); subst k1s; clear Hfst1s.
  so (hsg_fsts _ (fsg_hsg _ _ HT2)) as (k2 & Hfst2).
  so (fsts_fun _#3 Hfst2s (subst_fsts _#3 Hfst2)); subst k2s; clear Hfst2s.
  so (fsts_fkd _#3 HT1 Hfst1) as Hk1.
  so (fsts_fkd _#3 HT2 Hfst2) as Hk2.
  revert P Hint1 Hint2.
  rewrite -> (subst_simp s _ (fkd_hkd _ _ Hk1)); [].
  rewrite -> (subst_simp (dot (var 0) (compose s sh1)) _ (fkd_hkd _ _ Hk2)); [].
  intros P Hint1 Hint2.
  #4 apply (sg_sigma_interp_ _#9 c1 c1'); auto.
    rewrite <- (subenv_left _#3 Hs); [].
    simp_sub.
    simp_sub in Hof.
    assumption.

    rewrite <- (subenv_right _#3 Hs); [].
    simp_sub.
    simp_sub in Hof'.
    assumption.

    fold ints.
    rewrite <- (IH1 _#3 Henv Hs); [].
    assumption.

    fold ints.
    rewrite <- (IH2 _#3 (fenv_ucn Henv) (subenv_under _#3 c1 c1' (Cand (simp k1) (fst P)) Hs)); [].
    assumption.

  intro H; invert H; clear H; [].
  intros V1 V1' V2 V2' c1 c1' k1 k2 P Hof Hof' Hval1 Hval2 Hval1' Hval2' HfstV1 HfstV1' Hfst1 Hfst2 Hint1 Hint2 <- <- <-.
  so (fsts_fkd _#3 HT1 Hfst1) as Hk1.
  so (fsts_fkd _#3 HT2 Hfst2) as Hk2.
  revert P Hint1 Hint2.
  rewrite <- (subst_simp s _ (fkd_hkd _ _ Hk1)); [].
  rewrite <- (subst_simp (dot (var 0) (compose s sh1)) _ (fkd_hkd _ _ Hk2)); [].
  intros P Hint1 Hint2.
  #4 apply (sg_sigma_interp_ _#9 c1 c1'); auto using subst_fsts.
    rewrite <- (subenv_left _#3 Hs) in Hof; [].
    simp_sub.
    simp_sub in Hof.
    assumption.

    rewrite <- (subenv_right _#3 Hs) in Hof'; [].
    simp_sub.
    simp_sub in Hof'.
    assumption.

    fold ints.
    rewrite <- (IH1 _#3 Henv Hs) in Hint1; [].
    assumption.

    fold ints.
    rewrite <- (IH2 _#3 (fenv_ucn Henv) (subenv_under _#3 c1 c1' (Cand (simp (subst s k1)) (fst P)) Hs)) in Hint2; [].
    assumption.
Qed.


Lemma subenv_dot :
  forall G E E' c s,
    fcn G c
    -> fenv G E
    -> subenv E s E'
    -> subenv E (dot c s) (E'; Sta (subst (left E) c) (subst (right E) c) (intc c E)).
Proof.
intros G E E' c s Hc Henv Hs.
apply subenv_dot_wc; auto; [].
intros E'' n Htrunc.
symmetry.
apply (subst_intc G E''); auto; [].
apply subenv_sh; auto; done.
Qed.


Lemma subst1_intc :
  forall G E b c,
    fcn G b
    -> fcn (G; cl_ucn) c
    -> fenv G E
    -> intc (subst1 b c) E = intc c (E; Sta (subst (left E) b) (subst (right E) b) (intc b E)).
Proof.
intros G E b c Hb Hc Henv.
apply (subst_intc (G; cl_ucn)); auto using fenv_ucn; [].
apply (subenv_dot G); auto; [].
apply subenv_sh; auto with sequence; done.
Qed.


Lemma weaken1_intc :
  forall G E ent c,
    fcn G c
    -> fenv G E
    -> intc (shift1 c) (E; ent) = intc c E.
Proof.
intros G E ent c Henv Hc.
eapply subst_intc; eauto; [].
apply subenv_sh; auto with sequence.
Qed.


Lemma intctx_left :
  forall G E E',
    intctx G E E'
    -> subrel seq nil (left E) (left E') G.
Proof.
intros G E E' H.
induct H.

(* nil *)
simpl.
apply subrel_id; done.

(* cn *)
intros G k E E' c1 c1' c2 c2' Q Q' _ IH Heq1 Heq2 _.
simpl.
apply subrel_dot; auto; [].
simp_sub.
apply seq_cn; auto; done.

(* tm *)
intros G t E E' m1 m2 m1' m2' _ IH.
simpl.
apply subrel_dot; auto; [].
simp_sub.
apply seq_tm; auto; done.

(* md *)
intros G T c E E' m1 m2 m1' m2' _ IH.
simpl.
apply subrel_dot; auto; [].
simp_sub.
apply seq_md; auto; done.
Qed.


Lemma intctx_right :
  forall G E E',
    intctx G E E'
    -> subrel seq nil (right E) (right E') G.
Proof.
intros G E E' H.
induct H.

(* nil *)
simpl.
apply subrel_id; done.

(* cn *)
intros G k E E' c1 c1' c2 c2' Q Q' _ IH Heq1 Heq2 _.
simpl.
apply subrel_dot; auto; [].
simp_sub.
apply seq_cn; auto; done.

(* tm *)
intros G t E E' m1 m2 m1' m2' _ IH.
simpl.
apply subrel_dot; auto; [].
simp_sub.
apply seq_tm; auto; done.

(* md *)
intros G T c E E' m1 m2 m1' m2' _ IH.
simpl.
apply subrel_dot; auto; [].
simp_sub.
apply seq_md; auto; done.
Qed.


Lemma left_subof :
  forall G E E',
    intctx G E E'
    -> subof sof nil (left E) G.
Proof.
intros G E E' H.
induct H.

(* nil *)
simpl.
apply subof_id; done.

(* cn *)
intros G k E E' c1 c1' c2 c2' Q Q' _ IH Heqc _ _.
simpl.
apply subof_dot; auto; [].
simp_sub.
apply sof_cn; [].
exact (eqc_reg _#4 wf_nil Heqc andel).

(* tm *)
intros G t E E' m1 m2 m1' m2' _ IH.
simpl.
apply subof_dot; auto; [].
simp_sub.
apply sof_tm; done.

(* md *)
intros G T c E E' m1 m2 m1' m2' _ IH.
simpl.
apply subof_dot; auto; [].
simp_sub.
apply sof_md; done.
Qed.


Lemma right_subof :
  forall G E E',
    intctx G E E'
    -> subof sof nil (right E) G.
Proof.
intros G E E' H.
induct H.

(* nil *)
simpl.
apply subof_id; done.

(* cn *)
intros G k E E' c1 c1' c2 c2' Q Q' _ IH _ Heqc _.
simpl.
apply subof_dot; auto; [].
simp_sub.
apply sof_cn; [].
exact (eqc_reg _#4 wf_nil Heqc andel).

(* tm *)
intros G t E E' m1 m2 m1' m2' _ IH.
simpl.
apply subof_dot; auto; [].
simp_sub.
apply sof_tm; done.

(* md *)
intros G T c E E' m1 m2 m1' m2' _ IH.
simpl.
apply subof_dot; auto; [].
simp_sub.
apply sof_md; done.
Qed.


Lemma mi_eqs :
  forall T1 T1' T2 T2' R,
    eqs nil T1 T1'
    -> eqs nil T2 T2'
    -> mi T1 T2 R = mi T1' T2' R.
Proof.
intros T1 T1' T2 T2' R Heq1 Heq2.
unfold mi.
apply mpst_eqs; auto.
Qed.


Lemma mi_mvrel :
  forall T T' R,
    mvrel T T' (mi T T' R).
Proof.
intros T T' R.
unfold mi.
apply mpst_mvrel.
Qed.


Lemma mi_mpclosed :
  forall T T' R,
    mpclosed T T' (mi T T' R).
Proof.
intros T T' R.
unfold mi.
apply mpst_idem.
Qed.


Lemma mi_mono :
  forall T T' (R1 R2 : mrelation),
    (forall M M' Q, R1 M M' Q -> R2 M M' Q)
    -> forall M M', mi T T' R1 M M' -> mi T T' R2 M M'.
Proof.
unfold mi.
intros T T' R1 R2 Hincl M M' HM.
eapply mpst_mono; eauto; [].
intros V V' HV.
destruct HV as (Q & HV).
exists Q.
eapply Hincl; eauto; done.
Qed.


Lemma mi_mono_sg :
  forall T T' U U' R,
    subs nil T U
    -> subs nil T' U'
    -> forall V V', mi T T' R V V' -> mi U U' R V V'.
Proof.
intros T T' U U' R Hsubs Hsubs' V V' HV.
unfold mi in HV.
unfold mi.
eapply mpst_mono_sg; eauto.
Qed.


Lemma miev_eqs :
  forall T1 T1' T2 T2' R,
    eqs nil T1 T1'
    -> eqs nil T2 T2'
    -> miev T1 T2 R = miev T1' T2' R.
Proof.
intros T1 T1' T2 T2' R Heq1 Heq2.
unfold miev.
rewrite -> (mi_eqs _#5 Heq1 Heq2); [].
apply mev_eqs; auto; done.
Qed.


Lemma miev_ceqm :
  forall T1 T2 R M1 M2 M1' M2',
    Contextual.ceqm nil impure M1 M1' T1
    -> Contextual.ceqm nil impure M2 M2' T2
    -> miev T1 T2 R M1 M2
    -> miev T1 T2 R M1' M2'.
Proof.
intros T1 T2 R M1 M2 M1' M2' Heqm1 Heqm2 HM.
unfold miev; unfold miev in HM.
eapply mpclosed_mev_ceqm; eauto using mi_mvrel, mi_mpclosed.
Qed.


Lemma mpuev_mono :
  forall T T' (R1 R2 : mrelation),
    (forall M M' Q, R1 M M' Q -> R2 M M' Q)
    -> forall M M' Q, mpuev T T' R1 M M' Q -> mpuev T T' R2 M M' Q.
Proof.
unfold mpuev.
intros T T' R1 R2 Hincl M M' Q HM.
destruct HM as (Hhalt & Hrel & Hof & Hof').
do2 3 split; auto.
Qed.


Lemma mpuev_mono_sg :
  forall T T' U U' R,
    subs nil T U
    -> subs nil T' U'
    -> forall M M' Q, mpuev T T' R M M' Q -> mpuev U U' R M M' Q.
Proof.
intros T T' U U' R Hsubs Hsubst' M M' Q HM.
destruct HM as (Hhalt & Hrel & Hof & Hof').
do2 3 split; eauto with static.
Qed.


Lemma mpuev_eqs :
  forall T1 T1' T2 T2' R,
    eqs nil T1 T1'
    -> eqs nil T2 T2'
    -> mpuev T1 T2 R = mpuev T1' T2' R.
Proof.
intros T1 T1' T2 T2' R Heq1 Heq2.
fextensionality 3.
intros V V' Q.
pextensionality.
  intros (H1 & H2 & Hof & Hof').
  do2 3 split; eauto with static; done.

  intros (H1 & H2 & Hof & Hof').
  do2 3 split; eauto with static; done.
Qed.
