
Require Import Tactics.
Require Import Extensionality.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import Formation.
Require Import TSubst.
Require Import Regularity.
Require Import Closure.
Require Import Candidate.
Require Import Skof.
Require Import Logical.
Require Import Biorthogonal.
Require Contextual.


(* model regular at k under G *)
Definition mrkat (G : context) (k : term) : Prop :=
  forall E E',
    intctx G E E'
    -> intk k E = intk k E'
       /\ (forall c c' Q1 Q2, intk k E c c' Q1 Q2 -> intk k E c c' Q2 Q1)
       /\ (forall c c' Q1 Q2 Q3, intk k E c c' Q1 Q2 -> intk k E c c' Q2 Q3 -> intk k E c c' Q1 Q3).


(* model regular up to k under G *)
Inductive mrk : context -> (* kd *) term -> Prop :=
| mrk_type {G}
    : mrkat G kd_type
      -> mrk G kd_type

| mrk_sing {G c}
    : ofc G c kd_type
      -> mrkat G (kd_sing c)
      -> mrk G (kd_sing c)

| mrk_pi {G k1 k2}
    : mrkat G (kd_pi k1 k2)
      -> mrk G k1
      -> mrk (G; cl_cn k1) k2
      -> mrk G (kd_pi k1 k2)

| mrk_sigma {G k1 k2}
    : mrkat G (kd_sigma k1 k2)
      -> mrk G k1
      -> mrk (G; cl_cn k1) k2
      -> mrk G (kd_sigma k1 k2)

| mrk_one {G}
    : mrkat G kd_one
      -> mrk G kd_one
.


(* model regular up to G *)
(* for convenience, make mr include wf *)
Inductive mr : context -> Prop :=
| mr_nil
    : mr nil

| mr_cn {G k}
    : mr G
      -> mrk G k
      -> mr (G; cl_cn k)

| mr_tm {G t}
    : mr G
      -> ofc G t kd_type
      -> mr (G; cl_tm t)

| mr_md {G G' k T}
    : mr G
      -> ofs G' T
      -> fsts T k
      -> G = G'; cl_cn k
      -> mr (G; cl_md (shift1 T) (var 0))
.


Lemma mrk_ofk :
  forall G k, mrk G k -> ofk G k.
Proof.
intros G k H.
induct H; auto with static.
Qed.


Lemma mrk_mrkat :
  forall G k, mrk G k -> mrkat G k.
Proof.
intros G k H.
induct H; auto.
Qed.


Lemma mr_wf :
  forall G, mr G -> wf G.
Proof.
intros G H.
induct H; eauto using wf, mrk_ofk; [].
intros G G' k T _ Hwf Hofs Hfst ->.
apply wf_md; eauto using wf_tail; done.
Qed.


Lemma mr_tail :
  forall G A,
    mr (G; A) -> mr G.
Proof.
intros G A H.
invert H; intros; subst; eauto using mr_cn.
Qed.


Lemma mr_fcontext :
  forall G, mr G -> fcontext G.
Proof.
auto using mr_wf, wf_fcontext.
Qed.


Lemma intk_fun_mr :
  forall G k E E',
    mrk G k
    -> intctx G E E'
    -> intk k E = intk k E'.
Proof.
intros G k E E' Hmrk Hint.
exact (mrk_mrkat _ _ Hmrk _ _ Hint andel).
Qed.


Lemma intk_symm_mr :
  forall G k E E' c c' Q Q',
    mrk G k
    -> intctx G E E'
    -> intk k E c c' Q Q'
    -> intk k E c c' Q' Q.
Proof.
intros G k E E' c c' Q Q' Hmrk Hint Hintk.
exact (mrk_mrkat _ _ Hmrk _ _ Hint ander andel _#4 Hintk).
Qed.


Lemma intk_trans_mr :
  forall G k E E' c c' Q1 Q2 Q3,
    mrk G k
    -> intctx G E E'
    -> intk k E c c' Q1 Q2
    -> intk k E c c' Q2 Q3
    -> intk k E c c' Q1 Q3.
Proof.
intros G k E E' c c' Q1 Q2 Q3 Hmrk Hint Hint12 Hint23.
exact (mrk_mrkat _ _ Hmrk _ _ Hint ander ander _#5 Hint12 Hint23).
Qed.


Lemma intk_trans_mr_r :
  forall G k E c c' Q1 Q2 Q3,
    mrk G k
    -> intctx G E E
    -> intk k E c c' Q1 Q2
    -> intk k E c c' Q2 Q3
    -> intk k E c c' Q1 Q3.
Proof.
intros; eapply intk_trans_mr; eauto.
Qed.


Lemma intk_refl_mr :
  forall G k E E' c c' Q Q',
    mrk G k
    -> intctx G E E'
    -> intk k E c c' Q Q'
    -> intk k E c c' Q Q.
Proof.
intros G k E E' c c' Q Q' Hmrk HintE Hint.
eapply intk_trans_mr; eauto; [].
eapply intk_symm_mr; eauto; done.
Qed.


Lemma intctx_refl_mr :
  forall G E E',
     mr G
     -> intctx G E E' 
     -> intctx G E E.
Proof.
intros G E E' Hmr Hint.
revert E E' Hint.
induct Hmr.

(* nil *)
intros E E' H.
invert H.
intros <- <-.
apply intctx_nil; done.

(* cn *)
intros G k _ IH Hmrk E1 E2 Hint.
invert Hint; clear Hint; [].
intros E1' E2' c1 c1' c2 c2' Q Q' HintG Heqc Heqc' Hintk <- <-.
#4 apply intctx_cn; auto.
  eapply IH; eauto; done.

  apply (@eqc_trans _ _ c1'); auto with static; done.

  apply (@eqc_trans _ _ c2'); auto with static; done.

  so (mrk_mrkat _ _ Hmrk _ _ HintG) as (_ & Hsymm & Htrans).
  apply (Htrans _#3 Q'); auto; done.

(* tm *)
intros G t _ IH _ E1 E2 Hint.
invert Hint; clear Hint; [].
intros E1' E2' m1 m2 m1' m2' Hint <- <-.
apply intctx_tm; eauto; done.

(* md *)
intros G G' k T _ IH Hofs Hfsts -> E E' Hint.
invert Hint; clear Hint; [].
intros E1' E2' m1 m2 m1' m2' Hint <- <-.
apply intctx_md; eauto; done.
Qed.


Lemma intk_defeq_mr :
  forall G k E E' c1 c2 c1' c2' Q Q',
    mr G
    -> mrk G k
    -> intctx G E E'
    -> intk k E c1 c2 Q Q'
    -> eqc nil c1 c1' (subst (left E) k)
    -> eqc nil c2 c2' (subst (right E) k)
    -> intk k E c1' c2' Q Q'.
Proof.
intros G k E E' c1 c2 c1' c2' Q Q' Hmr Hmrk HintE Hintk Heqc1 Heqc2.
so (intctx_refl_mr _#3 Hmr HintE) as HintE'.
renameover HintE' into HintE.
revert E c1 c2 c1' c2' Q Q' Hmr HintE Hintk Heqc1 Heqc2.
induct Hmrk.

(* type *)
intros G _ E c1 c2 c1' c2' Q Q' _ _ Hintk Heqc1 Heqc2.
simp_sub in Heqc1; simp_sub in Heqc2.
invert Hintk; clear Hintk; [].
intros R Hofc1 Hofc2 Hrel Hclose <- <-.
#4 apply type_interp_.
  exact (eqc_reg _#4 wf_nil Heqc1 ander andel).

  exact (eqc_reg _#4 wf_nil Heqc2 ander andel).

  eapply tvrel_eqc; eauto; done.

  eapply tpclosed_eqc; eauto; done.

(* sing *)
intros G c Hofc _ E c1 c2 c1' c2' Q Q' _ HintE Hintk Heqc1 Heqc2.
simp_sub in Heqc1; simp_sub in Heqc2.
invert Hintk; clear Hintk; [].
intros Heqc1' Heqc2' Hrel Hcl <- <-.
simpl.
#4 apply sing_interp_; auto.
  apply (@eqc_trans _ _ c1); auto; [].
  apply eqc_symm; [].
  eapply eqc_subsume; eauto; [].
  apply subk_sing_t; [].
  exact (eqc_reg _#4 wf_nil Heqc1' ander andel).

  apply (@eqc_trans _ _ c2); auto; [].
  apply eqc_symm; [].
  eapply eqc_subsume; eauto; [].
  apply subk_sing_t; [].
  exact (eqc_reg _#4 wf_nil Heqc2' ander andel).

  #2 eapply tvrel_eqc; eauto.
    eapply eqc_subsume; eauto using subst_ofc_type, left_subof with static; done.

    eapply eqc_subsume; eauto using subst_ofc_type, right_subof with static; done.

  #2 eapply tpclosed_eqc; eauto.
    eapply eqc_subsume; eauto using subst_ofc_type, left_subof with static; done.

    eapply eqc_subsume; eauto using subst_ofc_type, right_subof with static; done.

(* pi *)
intros G k1 k2 _ Hmrk1 IH1 _ IH2 E c1 c2 c1' c2' Q Q' Hmr HintE Hintk Heqc1 Heqc2.
invert Hintk; clear Hintk; [].
intros F F' Hofc1 Hofc2 Hintapp <- <-.
#3 apply pi_interp_.
  exact (eqc_reg _#4 wf_nil Heqc1 ander andel).

  exact (eqc_reg _#4 wf_nil Heqc2 ander andel).

  fold intk.
  intros d d' Q Q' d'' d''' Hintk1 Heqd Heqd'.
  so (Hintapp _#6 Hintk1 Heqd Heqd') as Hint.
  #4 refine (IH2 _#9 Hint _ _).
    apply mr_cn; auto; done.

    #3 apply intctx_cn; auto.
      apply (@eqc_trans _ _ d''); auto with static; done.

      apply (@eqc_trans _ _ d'''); auto with static; done.

      eapply intk_refl_mr; eauto; done.

    simpl.
    simp_sub in Heqc1.
    replace (subst (dot d (left E)) k2) with (subst1 d (subst (dot (var 0) (compose (left E) sh1)) k2)) by (simp_sub; reflexivity).
    eapply eqc_app; eauto; [].
    apply (@eqc_trans _ _ d''); auto with static; done.

    simpl.
    simp_sub in Heqc2.
    replace (subst (dot d' (right E)) k2) with (subst1 d' (subst (dot (var 0) (compose (right E) sh1)) k2)) by (simp_sub; reflexivity).
    eapply eqc_app; eauto; [].
    apply (@eqc_trans _ _ d'''); auto with static; done.

(* sigma *)
intros G k1 k2 _ Hmrk1 IH1 Hmrk2 IH2 E c1 c2 c1' c2' Q Q' Hmr HintE Hintk Heqc1 Heqc2.
invert Hintk; clear Hintk; [].
intros P P' Hofc1 Hofc2 Hint1 Hint2 <- <-.
#4 apply sigma_interp_; fold intk.
  exact (eqc_reg _#4 wf_nil Heqc1 ander andel).

  exact (eqc_reg _#4 wf_nil Heqc2 ander andel).

  eapply IH1; eauto with static; done.

  #5 apply (IH2 (E; Sta (cn_pi1 c1') (cn_pi1 c2') (Cand (simp k1) (fst P))) (cn_pi2 c1) (cn_pi2 c2) (cn_pi2 c1') (cn_pi2 c2') (Cand (simp k2) (snd P)) (Cand (simp k2) (snd P'))).
    apply mr_cn; auto; done.

    #3 apply intctx_cn; auto.
      apply eqc_refl; [].
      exact (ofc_pi1 (eqc_reg _#4 wf_nil Heqc1 ander andel)).

      apply eqc_refl; [].
      exact (ofc_pi1 (eqc_reg _#4 wf_nil Heqc2 ander andel)).

      #2 eapply IH1; eauto using intk_refl_mr.
        eapply eqc_pi1; eauto; done.

        eapply eqc_pi1; eauto; done.

    assert (intctx (G; cl_cn k1) (E; Sta (cn_pi1 c1) (cn_pi1 c2) (Cand (simp k1) (fst P))) (E; Sta (cn_pi1 c1') (cn_pi1 c2') (Cand (simp k1) (fst P)))) as HintGk1.
      #4 apply intctx_cn.
        eapply intctx_refl_mr; eauto; done.

        eapply eqc_pi1; eauto; done.

        eapply eqc_pi1; eauto; done.

        so (mrk_mrkat _ _ Hmrk1 _ _ HintE) as (_ & Hsymm & Htrans).
        apply (Htrans _ _ _ (Cand (simp k1) (fst P'))); auto; done.
    so (intk_fun_mr _#4 Hmrk2 HintGk1) as Heqint2.
    rewrite -> Heqint2 in Hint2; [].
    assumption.

    simpl.
    simp_sub in Heqc1.
    apply eqc_symm; [].
    replace (subst (dot (cn_pi1 c1') (left E)) k2) with (subst1 (cn_pi1 c1') (subst (dot (var 0) (compose (left E) sh1)) k2)) by (simp_sub; reflexivity).
    eapply eqc_pi2; eauto with static; done.
  
    simpl.
    simp_sub in Heqc2.
    apply eqc_symm; [].
    replace (subst (dot (cn_pi1 c2') (right E)) k2) with (subst1 (cn_pi1 c2') (subst (dot (var 0) (compose (right E) sh1)) k2)) by (simp_sub; reflexivity).
    eapply eqc_pi2; eauto with static; done.

(* one *)
intros G H E c1 c2 c1' c2' Q Q' Hmr HintE Hintk Heqc1 Heqc2.
simp_sub in Heqc1; simp_sub in Heqc2.
invert Hintk; clear Hintk; [].
intros _ _ <- <-.
simpl.
#2 apply one_interp_.
  exact (eqc_reg _#4 wf_nil Heqc1 ander andel).

  exact (eqc_reg _#4 wf_nil Heqc2 ander andel).
Qed.


Lemma intk_defeq_mr_r :
  forall G k E c1 c2 c1' c2' Q Q',
    mr G
    -> mrk G k
    -> intctx G E E
    -> intk k E c1 c2 Q Q'
    -> eqc nil c1 c1' (subst (left E) k)
    -> eqc nil c2 c2' (subst (right E) k)
    -> intk k E c1' c2' Q Q'.
Proof.
intros.
eapply intk_defeq_mr; eauto.
Qed.


Lemma intctx_per :
  forall G,
    mr G
    -> (forall E E', intctx G E E' -> intctx G E' E)
       /\ (forall E1 E2 E3, intctx G E1 E2 -> intctx G E2 E3 -> intctx G E1 E3).
Proof.
intros G H.
induct H.

(* nil *)
split.
  intros E E' H.
  invert H; [].
  intros <- <-.
  apply intctx_nil; done.

  intros E1 E2 E3 H12 H23.
  invert H12; []; intros <- <-.
  invert H23; []; intros <-.
  apply intctx_nil; done.

(* cn *)
intros G k Hmr (IHs & IHt) Hmrk.
so (mrk_ofk _ _ Hmrk) as Hofk.
split.
  intros E1 E2 H.
  invert H; clear H; [].
  intros E1' E2' c1 c1' c2 c2' Q Q' HintG Heqc1 Heqc2 Hintk <- <-.
  #4 apply intctx_cn.
    apply IHs; auto; done.

    apply eqc_symm; [].
    eapply eqc_subsume; eauto; [].
    apply subk_refl; [].
    eapply seq_ofk; eauto using wf_nil; [].
    apply intctx_left; auto; done.

    apply eqc_symm; [].
    eapply eqc_subsume; eauto; [].
    apply subk_refl; [].
    eapply seq_ofk; eauto using wf_nil; [].
    apply intctx_right; auto; done.

    so (intk_fun_mr _#4 Hmrk HintG) as Heq.
    rewrite <- Heq; [].
    apply (intk_symm_mr _#8 Hmrk HintG); [].
    eapply intk_defeq_mr; eauto; done.

  intros E1 E2 E3 H12 H23.
  invert H12; clear H12; [].
  intros E1' E2' c1 c2 c1' c2' Q1 Q2 H12' Heqc12 Heqc12' HintQ12 <- <-.
  invert H23; clear H23; [].
  intros E3' c3 c3' Q3 H23' Heqc23 Heqc23' HintQ23 <-.
  #4 apply intctx_cn.
    eapply IHs; eauto; done.

    eapply eqc_trans; eauto; [].
    eapply eqc_subsume; eauto; [].
    apply subk_refl; [].
    apply eqk_symm; [].
    eapply seq_ofk; eauto using wf_nil; [].
    apply intctx_left; auto; done.

    eapply eqc_trans; eauto; [].
    eapply eqc_subsume; eauto; [].
    apply subk_refl; [].
    apply eqk_symm; [].
    eapply seq_ofk; eauto using wf_nil; [].
    apply intctx_right; auto; done.

    apply (intk_trans_mr _#7 Q2 _ Hmrk H12'); auto; [].
    apply (intk_defeq_mr G _ _ E2' c2 c2'); auto with static; [].
    rewrite -> (intk_fun_mr _#4 Hmrk H12'); [].
    assumption.

(* tm *)
intros G t _ (IHs & IHt) _.
split.
  intros E1 E2 H.
  invert H; clear H; [].
  intros E1' E2' m1 m2 m1' m2' H <- <-.
  apply intctx_tm; auto; done.

  intros E1 E2 E3 H12 H23.
  invert H12; clear H12; [].
  intros E1' E2' m1 m2 m1' m2' H12' <- <-.
  invert H23; clear H23; [].
  intros E3' m3 m3' H23' <-.
  apply intctx_tm; eauto; done.

(* md *)
intros G G' k T _ (IHs & IHt) Hofs Hfst ->.
split.
  intros E1 E2 H.
  invert H; clear H; [].
  intros E1' E2' m1 m2 m1' m2' H <- <-.
  apply intctx_md; auto; done.

  intros E1 E2 E3 H12 H23.
  invert H12; clear H12; [].
  intros E1' E2' m1 m2 m1' m2' H12' <- <-.
  invert H23; clear H23; [].
  intros E3' m3 m3' H23' <-.
  apply intctx_md; eauto; done.
Qed.


Lemma intctx_symm_mr :
  forall G E E',
    mr G
    -> intctx G E E'
    -> intctx G E' E.
Proof.
intros G E E' Hmr Hint.
exact (intctx_per _ Hmr andel _ _ Hint).
Qed.


Lemma intctx_trans_mr :
  forall G E1 E2 E3,
    mr G
    -> intctx G E1 E2
    -> intctx G E2 E3
    -> intctx G E1 E3.
Proof.
intros G E1 E2 E3 Hmr H12 H23.
exact (intctx_per _ Hmr ander _ _ _ H12 H23).
Qed.


Lemma mrk_type_intro :
  forall G, mrk G kd_type.
Proof.
intros G.
apply mrk_type; [].
intros E E' _.
do2 2 split.
  simpl.
  fextensionality 4.
  intros c c' Q Q'.
  pextensionality.
    intro H; invert H; clear H; [].
    intros; subst; apply type_interp_; auto; done.

    intro H; invert H; clear H; [].
    intros; subst; apply type_interp_; auto; done.

  simpl.
  intros c c' Q1 Q2 Hint.
  invert Hint; clear Hint; [].
  intros; subst; apply type_interp_; auto; done.

  simpl.
  intros c c' Q1 Q2 Q3 H12 H23.
  invert H12; clear H12; [].
  intros R Hofc Hofc' HofR Hcl <- <-.
  invert H23; clear H23; [].
  intros R' _ _ HofR' Hcl' Heq <-.
  so (cand_inject _#3 Heq); subst R'.
  apply type_interp_; auto; done.
Qed.


Lemma mrk_one_intro :
  forall G, mrk G kd_one.
Proof.
intros G.
apply mrk_one; [].
intros E E' _.
do2 2 split.
  simpl.
  fextensionality 4.
  intros c c' Q Q'.
  pextensionality.
    intro H; invert H; clear H; [].
    intros; subst; apply one_interp_; auto; done.

    intro H; invert H; clear H; [].
    intros; subst; apply one_interp_; auto; done.

  simpl.
  intros c c' Q1 Q2 H.
  invert H; clear H; [].
  intros; subst; apply one_interp_; auto; done.

  simpl.
  intros c c' Q1 Q2 Q3 H12 H23.
  invert H12; invert H23; intros; subst; apply one_interp_; auto; done.
Qed.


Lemma mrk_sing_intro' :
  forall G c,
    ofc G c kd_type
    -> (forall E E', intctx G E E' -> intk (kd_sing c) E = intk (kd_sing c) E')
    -> mrk G (kd_sing c).
Proof.
intros G c Hof Hfun.
apply mrk_sing; auto; [].
intros E E' HintE.
do2 2 split.
  apply Hfun; auto; done.

  simpl.
  intros b b' Q1 Q2.
  intro H; invert H; clear H; [].
  intros Hofb Hofb' Hrel Hcl <- <-.
  apply sing_interp_; auto; done.

  simpl.
  intros b b' Q1 Q2 Q3 H12 H23.
  invert H12; clear H12; [].
  intros Hofb Hofb' Hrel Hcl <- _.
  invert H23; clear H23; [].
  intros _ _ _ _ <- <-.
  apply sing_interp_; auto; done.
Qed.


Lemma mrk_sing_intro :
  forall G c,
    ofc G c kd_type
    -> (forall E E', intctx G E E' -> intc c E = intc c E')
    -> mrk G (kd_sing c).
Proof.
intros G c Hof Hfun.
apply mrk_sing_intro'; auto; [].
intros E E' HintE.
simpl.
fextensionality 4.
intros b b' Q Q'.
pextensionality.
  intro H; invert H; clear H; [].
  intros Heqb Heqb' Hrel Hcl <- <-.
  rewrite -> (Hfun _ _ HintE); [].
  #4 apply sing_interp_.
    eapply eqc_trans; eauto; [].
    replsub (kd_type) with (subst (left E) kd_type).
    eapply seq_ofc; eauto using wf_nil, intctx_left; done.

    eapply eqc_trans; eauto; [].
    replsub (kd_type) with (subst (right E) kd_type).
    eapply seq_ofc; eauto using wf_nil, intctx_right; done.

    rewrite <- (Hfun _ _ HintE); [].
    assumption.

    rewrite <- (Hfun _ _ HintE); [].
    assumption.

  intro H; invert H; clear H; [].
  intros Heqb Heqb' Hrel Hcl <- <-.
  rewrite <- (Hfun _ _ HintE); [].
  #4 apply sing_interp_.
    eapply eqc_trans; eauto; [].
    apply eqc_symm; [].
    replsub (kd_type) with (subst (left E) kd_type).
    eapply seq_ofc; eauto using wf_nil, intctx_left; done.

    eapply eqc_trans; eauto; [].
    apply eqc_symm; [].
    replsub (kd_type) with (subst (right E) kd_type).
    eapply seq_ofc; eauto using wf_nil, intctx_right; done.

    rewrite -> (Hfun _ _ HintE); [].
    assumption.

    rewrite -> (Hfun _ _ HintE); [].
    assumption.
Qed.


Lemma mrk_pi_intro :
  forall G k1 k2,
    mr G
    -> mrk G k1
    -> mrk (G; cl_cn k1) k2
    -> mrk G (kd_pi k1 k2).
Proof.
intros G k1 k2 Hmr Hmrk1 Hmrk2.
apply mrk_pi; auto; [].
intros E E' HintE.
do2 2 split.
  fextensionality 4.
  intros c c' Q Q'.
  pextensionality.
    intro H; invert H; clear H; [].
    intros F F' Hofc Hofc' Hintapp <- <-.
    simpl.
    #3 apply pi_interp_.
      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      eapply (seq_ofk _#3 G); auto using wf_nil, intctx_left; [].
      apply ofk_pi; auto using mrk_ofk; done.

      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      eapply (seq_ofk _#3 G); auto using wf_nil, intctx_right; [].
      apply ofk_pi; auto using mrk_ofk; done.

      intros d d' Q Q' d'' d''' Hint1 Heqd Heqd'.      
      so (intk_fun_mr _#4 Hmrk1 HintE) as Heqint1.
      #2 exploit (intk_fun_mr (G; cl_cn k1) k2 (E; Sta d d' (Cand (simp k1) Q)) (E'; Sta d d' (Cand (simp k1) Q))) as Heqint2; auto.
        #3 apply intctx_cn; auto.
          apply eqc_refl; [].
          #2 apply (@ofc_subsume _ (subst (left E') k1)).
            exact (eqc_reg _#4 wf_nil Heqd andel).

            apply subk_refl; [].
            apply eqk_symm; [].
            apply (seq_ofk _#3 G); auto using wf_nil, mrk_ofk; [].
            apply intctx_left; auto; done.

          apply eqc_refl; [].
          #2 apply (@ofc_subsume _ (subst (right E') k1)).
            exact (eqc_reg _#4 wf_nil Heqd' andel).

            apply subk_refl; [].
            apply eqk_symm; [].
            apply (seq_ofk _#3 G); auto using wf_nil, mrk_ofk; [].
            apply intctx_right; auto; done.

          rewrite -> Heqint1; [].
          eapply intk_refl_mr; eauto; [].
          apply intctx_symm_mr; eauto; done.
      rewrite <- Heqint2; [].
      #3 apply Hintapp.
        rewrite <- Heqint1 in Hint1; [].
        assumption.

        eapply eqc_subsume; eauto; [].
        apply subk_refl; [].
        apply eqk_symm; [].
        apply (seq_ofk _#3 G); auto using wf_nil, mrk_ofk; [].
        apply intctx_left; auto; done.

        eapply eqc_subsume; eauto; [].
        apply subk_refl; [].
        apply eqk_symm; [].
        apply (seq_ofk _#3 G); auto using wf_nil, mrk_ofk; [].
        apply intctx_right; auto; done.

    intro H; invert H; clear H; [].
    intros F F' Hofc Hofc' Hintapp <- <-.
    simpl.
    #3 apply pi_interp_.
      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      #2 eapply (seq_ofk _#3 G); auto using wf_nil.
        apply intctx_left; [].
        apply intctx_symm_mr; auto; done.

        apply ofk_pi; auto using mrk_ofk; done.

      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      #2 eapply (seq_ofk _#3 G); auto using wf_nil.
        apply intctx_right; [].
        apply intctx_symm_mr; auto; done.

        apply ofk_pi; auto using mrk_ofk; done.

      intros d d' Q Q' d'' d''' Hint1 Heqd Heqd'.      
      so (intk_fun_mr _#4 Hmrk1 HintE) as Heqint1.
      #2 exploit (intk_fun_mr (G; cl_cn k1) k2 (E; Sta d d' (Cand (simp k1) Q)) (E'; Sta d d' (Cand (simp k1) Q))) as Heqint2; auto.
        #3 apply intctx_cn; auto.
          apply eqc_refl; [].
          exact (eqc_reg _#4 wf_nil Heqd andel).

          apply eqc_refl; [].
          exact (eqc_reg _#4 wf_nil Heqd' andel).

          eapply intk_refl_mr; eauto; done.
      rewrite -> Heqint2; [].
      #3 apply Hintapp.
        rewrite <- Heqint1; [].
        assumption.

        eapply eqc_subsume; eauto; [].
        apply subk_refl; [].
        apply (seq_ofk _#3 G); auto using wf_nil, mrk_ofk; [].
        apply intctx_left; auto; done.

        eapply eqc_subsume; eauto; [].
        apply subk_refl; [].
        apply (seq_ofk _#3 G); auto using wf_nil, mrk_ofk; [].
        apply intctx_right; auto; done.

  intros c c' Q1 Q2 Hint.
  invert Hint; clear Hint; [].
  intros F F' Hofc Hofc' Hintapp <- <-.
  simpl.
  apply pi_interp_; auto; [].
  intros d d' Q Q' d'' d''' Hint1 Heqd Heqd'.
  assert (mr (G; cl_cn k1)) as Hmr' by (apply mr_cn; auto).
  assert (intctx (G; cl_cn k1) (E; Sta d d' (Cand (simp k1) Q)) (E; Sta d'' d''' (Cand (simp k1) Q'))) as HintE'.
    apply intctx_cn; eauto using intctx_refl_mr; done.
  apply (intk_symm_mr (G; cl_cn k1) _ _ (E; Sta d d' (Cand (simp k1) Q))); eauto using intctx_refl_mr; [].
  rewrite -> (intk_fun_mr _#4 Hmrk2 HintE'); [].
  #4 apply (intk_defeq_mr_r (G; cl_cn k1) _ _ (cn_app c d'') (cn_app c' d''')); auto.
    eapply intctx_refl_mr; auto; [].
    apply intctx_symm_mr; eauto; done.

    #3 apply Hintapp.
      eapply intk_symm_mr; eauto; [].
      apply (intk_defeq_mr_r G _ _ d d'); eauto using intctx_refl_mr; done.

      apply eqc_symm; auto; done.

      apply eqc_symm; auto; done.

    simpl.
    replsub (subst (dot d'' (left E)) k2) with (subst1 d'' (subst (dot (var 0) (compose (left E) sh1)) k2)).
    simp_sub in Hofc.
    eapply eqc_app; eauto using eqc_refl; [].
    apply eqc_symm; auto; done.

    simpl.
    replsub (subst (dot d''' (right E)) k2) with (subst1 d''' (subst (dot (var 0) (compose (right E) sh1)) k2)).
    simp_sub in Hofc'.
    eapply eqc_app; eauto using eqc_refl; [].
    apply eqc_symm; auto; done.

  simpl.
  intros c c' Q1 Q2 Q3 Hint12 Hint23.
  invert Hint12; clear Hint12; [].
  intros F1 F2 Hofc Hofc' Hintapp <- <-.
  invert Hint23; clear Hint23; [].
  intros F2' F3 _ _ Hintapp' Heq <-.
  so (cand_inject _#3 Heq); subst F2'; clear Heq.
  apply pi_interp_; auto; [].
  intros d d' Q Q' d'' d''' Hint1 Heqd Heqd'.
  assert (intctx (G; cl_cn k1) (E; Sta d d' (Cand (simp k1) Q)) (E; Sta d d' (Cand (simp k1) Q))) as HintE'.
    #2 apply intctx_cn; eauto using intctx_refl_mr, intk_refl_mr.
      eapply eqc_trans; eauto using eqc_symm; done.

      eapply eqc_trans; eauto using eqc_symm; done.
  apply (intk_trans_mr_r (G; cl_cn k1) _ _ _ _ _ (Cand (simp k2) (F2 d d' Q))); auto; [].
  #2 apply Hintapp; eauto using intk_refl_mr.
    eapply eqc_trans; eauto using eqc_symm; done.

    eapply eqc_trans; eauto using eqc_symm; done.
Qed.


Lemma mrk_sigma_intro :
  forall G k1 k2,
    mr G
    -> mrk G k1
    -> mrk (G; cl_cn k1) k2
    -> mrk G (kd_sigma k1 k2).
Proof.
intros G k1 k2 Hmr Hmrk1 Hmrk2.
apply mrk_sigma; auto; [].
intros E E' HintE.
do2 2 split.
  fextensionality 4.
  intros c c' Q Q'.
  pextensionality.
    intro H; invert H; clear H; [].
    intros P P' Hofc Hofc' Hint1 Hint2 <- <-.
    simpl.
    #4 apply sigma_interp_.
      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      apply (seq_ofk _#3 G); auto using wf_nil, intctx_left; [].
      apply ofk_sigma; auto using mrk_ofk; done.

      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      apply (seq_ofk _#3 G); auto using wf_nil, intctx_right; [].
      apply ofk_sigma; auto using mrk_ofk; done.

      rewrite <- (intk_fun_mr _#4 Hmrk1 HintE); [].
      assumption.

      assert (intctx (G; cl_cn k1) (E; Sta (cn_pi1 c) (cn_pi1 c') (Cand (simp k1) (fst P))) (E'; Sta (cn_pi1 c) (cn_pi1 c') (Cand (simp k1) (fst P)))) as HintE'.
        #3 apply intctx_cn; auto.
          simp_sub in Hofc.
          eapply eqc_pi1; eauto using eqc_refl; done.

          simp_sub in Hofc'.
          eapply eqc_pi1; eauto using eqc_refl; done.

          eapply intk_refl_mr; eauto; done.
      rewrite <- (intk_fun_mr _#4 Hmrk2 HintE'); [].
      assumption.

    intro H; invert H; clear H; [].
    intros P P' Hofc Hofc' Hint1 Hint2 <- <-.
    simpl.
    #4 apply sigma_interp_.
      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      apply eqk_symm; [].
      apply (seq_ofk _#3 G); auto using wf_nil, intctx_left; [].
      apply ofk_sigma; auto using mrk_ofk; done.

      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      apply eqk_symm; [].
      apply (seq_ofk _#3 G); auto using wf_nil, intctx_right; [].
      apply ofk_sigma; auto using mrk_ofk; done.

      rewrite -> (intk_fun_mr _#4 Hmrk1 HintE); [].
      assumption.

      assert (intctx (G; cl_cn k1) (E'; Sta (cn_pi1 c) (cn_pi1 c') (Cand (simp k1) (fst P))) (E; Sta (cn_pi1 c) (cn_pi1 c') (Cand (simp k1) (fst P)))) as HintE'.
        #4 apply intctx_cn.
          apply intctx_symm_mr; auto; done.

          simp_sub in Hofc.
          eapply eqc_pi1; eauto using eqc_refl; done.

          simp_sub in Hofc'.
          eapply eqc_pi1; eauto using eqc_refl; done.

          eapply intk_refl_mr; eauto using intctx_symm_mr; done.
      rewrite <- (intk_fun_mr _#4 Hmrk2 HintE'); [].
      assumption.

  intros c c' Q1 Q2 H.
  invert H; clear H; [].
  intros P P' Hofc Hofc' Hint1 Hint2 <- <-.
  simpl.
  #2 apply sigma_interp_; auto.
    eapply intk_symm_mr; eauto; done.

    assert (intctx (G; cl_cn k1) (E; Sta (cn_pi1 c) (cn_pi1 c') (Cand (simp k1) (fst P'))) (E; Sta (cn_pi1 c) (cn_pi1 c') (Cand (simp k1) (fst P)))) as HintE'.
      #4 apply intctx_cn.
        eapply intctx_refl_mr; eauto; done.

        apply eqc_refl; [].
        eapply ofc_pi1; eauto; done.

        apply eqc_refl; [].
        eapply ofc_pi1; eauto; done.

        eapply intk_symm_mr; eauto; done.
    rewrite -> (intk_fun_mr _#4 Hmrk2 HintE'); [].
    eapply intk_symm_mr; eauto; [].
    apply intctx_symm_mr; eauto using mr_cn; done.

  intros c c' Q1 Q2 Q3 H12 H23.
  invert H12; clear H12; [].
  intros P1 P2 Hofc Hofc' Hint11 Hint21 <- <-.
  invert H23; clear H23; [].
  intros P2' P3 _ _ Hint12 Hint22 Heq <-.
  so (cand_inject _#3 Heq); subst P2'; clear Heq.
  simpl.
  #2 apply sigma_interp_; auto.
    eapply intk_trans_mr; eauto; done.

    assert (intctx (G; cl_cn k1) (E; Sta (cn_pi1 c) (cn_pi1 c') (Cand (simp k1) (fst P1))) (E; Sta (cn_pi1 c) (cn_pi1 c') (Cand (simp k1) (fst P2)))) as HintE'.
      #4 apply intctx_cn.
        eapply intctx_refl_mr; eauto; done.

        apply eqc_refl; [].
        eapply ofc_pi1; eauto; done.

        apply eqc_refl; [].
        eapply ofc_pi1; eauto; done.

        assumption.
    eapply intk_trans_mr; eauto; [].
    rewrite -> (intk_fun_mr _#4 Hmrk2 HintE'); [].
    assumption.
Qed.


Inductive mrcl : context -> (* cn *) term -> (* cl *) term -> Prop :=
| mrcl_cn {G c k}
    : ofc G c k
      -> (forall E E', intctx G E E' -> intk k E (subst (left E) c) (subst (right E) c) (intc c E) (intc c E'))
      -> mrcl G c (cl_cn k)

| mrcl_tm {G e t}
    : mrcl G e (cl_tm t)

| mrcl_md {G M T c}
    : mrcl G M (cl_md T c)

| mrcl_ucn {G c}
    : fcn G c
      -> mrcl G c cl_ucn

| mrcl_utm {G t}
    : mrcl G t cl_utm

| mrcl_umd {G t c}
    : mrcl G t (cl_umd c)
.


Lemma mrcl_sfof :
  forall G m A,
    mrcl G m A -> sfof G m A.
Proof.
intros G m A H.
induct H; eauto using sfof, ofc_fcn; done.
Qed.


Lemma mrcl_sof :
  forall G m A,
    mrcl G m A -> sof G m A.
Proof.
intros G m A H.
induct H; eauto using sof; done.
Qed.


Lemma subof_mrcl_sfof :
  forall G1 s G2,
    subof mrcl G1 s G2
    -> subof sfof G1 s G2.
Proof.
eauto using subof_mono, mrcl_sfof.
Qed.


Lemma subof_mrcl_sof :
  forall G1 s G2,
    subof mrcl G1 s G2
    -> subof sof G1 s G2.
Proof.
eauto using subof_mono, mrcl_sof.
Qed.


Lemma mrcl_behaved :
  forall G i A,
    fcontext G
    -> iscl A
    -> index i G A
    -> mrcl G (var i) (subst (sh (S i)) A).
Proof.
intros G i A HG Hiscl Hindex.
#2 invert Hiscl; clear Hiscl; try (intros; subst; eauto using mrcl; done).
  intros k <-.
  simp_sub.
  #2 apply mrcl_cn.
    apply ofc_var; auto; done.
  intros E E' HintE.
  remember (cl_cn k) as A eqn:HeqA.
  revert HG HeqA E E' HintE.
  induct Hindex.
    (* 0 *)
    intros A G HG -> E1 E2 HintE.
    invert HintE; clear HintE; [].
    intros E1' E2' c1 c1' c2 c2' Q Q' HintE' Heqc1 Heqc2 Hintk <- <-.
    simpl.
    simp_sub.
    assert (subenv (E1'; Sta c1 c2 Q) sh1 E1') as Hsh1.
      apply subenv_sh; auto with sequence; done.
    assert (fkd G k) as Hk.
      invert HG; intros; assumption.
    unfold shift1.
    rewrite -> (subst_intk _#5 Hk (intctx_fenv _#3 HintE') Hsh1); assumption.
    
    (* S *)
    intros i B G A Hindex IH HG HeqA E1 E2 HintE.
    so (IH (fcontext_tail _ _ HG) HeqA) as IH'; clear IH.
    subst A.
    assert (exists E1' E2' ent1 ent2, intctx G E1' E2' /\ E1 = E1'; ent1 /\ E2 = E2'; ent2) as (E1' & E2' & ent1 & ent2 & HintE' & -> & ->).
      invert HintE; intros; do 4 eexists; eauto; done.
    so (IH' _ _ HintE') as H.
    simp_sub.
    simp_sub in H.
    replsub (subst (sh (2 + i)) k) with (subst sh1 (subst (sh (1 + i)) k)).
    assert (subenv (E1'; ent1) sh1 E1') as Hsh1.
      apply subenv_sh; auto with sequence; done.
    repl (1 + i) with (S i) by reflexivity.
    so (fcontext_index _#3 (fcontext_tail _ _ HG) Hindex) as Hcl.
    simp_sub in Hcl; invert Hcl; clear Hcl; intro Hk.
    rewrite -> (subst_intk _#5 Hk (intctx_fenv _#3 HintE') Hsh1); [].
    force_exact H.
    f_equal.
      destruct ent1; simpl; auto; done.
    
      destruct ent1; simpl; auto; done.

  intros <-.
  simp_sub.
  apply mrcl_ucn; eauto using fcn; done.
Qed.


Lemma mrcl_wclo :
  forall G m B A,
    fcontext G
    -> mrcl G m B
    -> mrcl (G; A) (shift1 m) (shift1 B).
Proof.
intros G m B A HG Hmrcl.
#2 invert Hmrcl; try (intros; subst; simp_sub; auto using mrcl; done).
  rename m into c.
  intros k Hofc Hint <-.
  simp_sub.
  #2 apply mrcl_cn.
    eapply weaken1_ofc; assumption.
  
    intros E1 E2 HintE.
    assert (exists E1' E2' ent1 ent2, intctx G E1' E2' /\ E1 = E1'; ent1 /\ E2 = E2'; ent2) as (E1' & E2' & ent1 & ent2 & HintE' & -> & ->).
      invert HintE; intros; do 4 eexists; eauto; done.
    assert (subenv (E1'; ent1) sh1 E1') as Hsh1.
      apply subenv_sh; auto with sequence; done.
    so (ofc_freg _#3 HG Hofc) as (Hc & Hk).
    unfold shift1.
    rewrite -> (subst_intk _#5 Hk (intctx_fenv _#3 HintE') Hsh1); [].
    force_exact (Hint _ _ HintE').
    f_equal.
      destruct ent1; simpl; simp_sub; auto; done.
    
      destruct ent1; simpl; simp_sub; auto; done.
  
      rewrite -> (subst_intc _#5 Hc (intctx_fenv _#3 HintE') Hsh1); reflexivity.
  
      assert (subenv (E2'; ent2) sh1 E2') as Hsh1'.
        apply subenv_sh; auto with sequence; done.
      rewrite -> (subst_intc _#5 Hc (intctx_fenv' _#3 HintE') Hsh1'); reflexivity.

  intros Hm <-.
  simp_sub.
  apply mrcl_ucn; [].
  eapply subst_fcn; eauto using subof_sh with sequence; done.
Qed.


Lemma subof_mrcl_weaken1 :
  forall G1 G2 s A,
    fcontext G1
    -> subof mrcl G1 s G2
    -> subof mrcl (G1; A) (compose s sh1) G2.
Proof.
intros G1 G2 s A HG1 Hs.
revert HG1.
induct Hs.

(* dot *)
intros G1 t s B G' _ IH Hmrcl HG.
simp_sub.
apply subof_dot; eauto; [].
rewrite -> subst_compose; [].
apply mrcl_wclo; auto; done.

(* sh *)
intros G i G' Htrunc _.
simp_sub.
apply subof_sh; eauto with sequence; done.
Qed.


Lemma subof_bind_mrcl :
  forall G G' s k,
    fcontext G
    -> fkd G' k
    -> subof mrcl G s G'
    -> subof mrcl (G; cl_cn (subst s k)) (dot (var 0) (compose s sh1)) (G'; cl_cn k).
Proof.
intros G G' s k HG Hk Hs.
#2 apply subof_dot.
  apply subof_mrcl_weaken1; auto; done.

  replsub (subst (compose s sh1) (cl_cn k)) with (subst sh1 (subst s (cl_cn k))).
  #3 apply mrcl_behaved.
    apply fcontext_cn; auto; [].
    eapply subst_fkd; eauto using subof_mrcl_sfof; done.

    simp_sub; auto using iscl; done.

    simp_sub.
    apply index_0; done.
Qed.


Lemma intctx_truncate :
  forall G1 G2 E1 E1' n,
    intctx G1 E1 E1'
    -> truncate n G1 G2
    -> exists E2 E2',
         intctx G2 E2 E2'
         /\ truncate n E1 E2
         /\ truncate n E1' E2'.
Proof.
intros G1 G2 E1 E1' n Hint Htrunc.
revert E1 E1' Hint.
induct Htrunc.

(* 0 *)
intros G E E' Hint.
exists E, E'.
auto with sequence; done.

(* S *)
intros n A G1 G2 _ IH E1 E1' Hint.
assert (exists E2 E2' ent ent', intctx G1 E2 E2' /\ E1 = E2; ent /\ E1' = E2'; ent') as (E2 & E2' & ent & ent' & Hint2 & -> & ->).
  invert Hint; intros; subst; do 4 eexists; eauto; done.
so (IH _ _ Hint2) as (E3 & E3' & Hint3 & Htrunc & Htrunc').
exists E3, E3'.
do2 2 split; auto with sequence; done.
Qed.


(* The first "technical lemma". *)
Lemma intctx_sub :
  forall G1 G2 s E1 E1',
    wf G2
    -> subof mrcl G1 s G2
    -> intctx G1 E1 E1'
    -> exists s' E2 E2',
         subrel sid G1 s s' G2
         /\ intctx G2 E2 E2'
         /\ subenv E1 s' E2
         /\ subenv E1' s' E2'.
Proof.
intros G1 G2 s E1 E1' HG2 Hs Hint.
revert E1 E1' HG2 Hint.
induct Hs.

(* dot *)
intros G1 m s A G2 Hs IH Hmrcl E1 E1' HG2a Hint1.
so (IH _ _ (wf_tail _ _ HG2a) Hint1) as (s' & E2 & E2' & Hsid & Hint2 & Hsubenv & Hsubenv'); clear IH.
#3 invert HG2a; clear HG2a.
  (* cn *)
  intros k HG2 Hofk <-.
  rename m into c.
  so (ofk_fkd _ _ Hofk) as Hk; clear Hofk.
  simp_sub in Hmrcl.
  invert Hmrcl; clear Hmrcl; [].
  intros Hofc Hfun.
  so (sid_fkd _#5 Hsid Hk) as Heq.
  exists (dot c s').
  exists (E2; Sta (subst (left E1) c) (subst (right E1) c) (intc c E1)).
  exists (E2'; Sta (subst (left E1') c) (subst (right E1') c) (intc c E1')).
  do2 3 split.
    apply subrel_dot; auto; [].
    simp_sub.
    apply sid_cn; done.

    #3 apply intctx_cn; auto.
      rewrite <- (subenv_left _#3 Hsubenv); [].
      rewrite -> subst_compose; [].
      rewrite <- Heq; [].
      eapply seq_ofc; eauto using wf_nil; [].
      apply intctx_left; eauto; done.

      rewrite <- (subenv_right _#3 Hsubenv); [].
      rewrite -> subst_compose; [].
      rewrite <- Heq; [].
      eapply seq_ofc; eauto using wf_nil; [].
      apply intctx_right; eauto; done.

      rewrite <- (subst_intk _#5 Hk (intctx_fenv _#3 Hint2) Hsubenv); [].
      rewrite <- Heq; [].
      apply Hfun; auto; done.

    eapply subenv_dot; eauto using ofc_fcn, intctx_fenv; done.

    eapply subenv_dot; eauto using ofc_fcn, intctx_fenv'; done.

  (* tm *)
  intros t _ _ <-.
  exists (dot nonsense s'), (E2; Dyn (subst (left E1) nonsense) (subst (right E1) nonsense)), (E2'; Dyn (subst (left E1') nonsense) (subst (right E1') nonsense)).
  do2 3 split.
    apply subrel_dot; auto; [].
    simp_sub.
    apply sid_tm; done.

    apply intctx_tm; auto; done.

    apply subenv_empty; auto; done.

    apply subenv_empty; auto; done.

  (* md *)
  intros G k T _ _ _ <- <-.
  exists (dot nonsense s'), (E2; Dyn (subst (left E1) nonsense) (subst (right E1) nonsense)), (E2'; Dyn (subst (left E1') nonsense) (subst (right E1') nonsense)).
  do2 3 split.
    apply subrel_dot; auto; [].
    simp_sub.
    apply sid_md; done.

    apply intctx_md; auto; done.

    apply subenv_empty; auto; done.

    apply subenv_empty; auto; done.

(* sh *)
intros G1 n G2 Htrunc E1 E1' _ Hint.
so (intctx_truncate _#5 Hint Htrunc) as (E2 & E2' & Hint2 & HtruncE & HtruncE').
exists (sh n), E2, E2'.
do2 3 split; auto using subenv_sh; [].
apply subrel_sh; auto; done.
Qed.


(* The second "technical lemma". *)
Lemma subst_mrk :
  forall G1 G2 s k,
    mr G1
    -> wf G2
    -> subof mrcl G1 s G2
    -> mrk G2 k
    -> mrk G1 (subst s k).
Proof.
intros G1 G2 s k Hmr HG2 Hs H.
revert G1 s Hmr HG2 Hs.
induct H.

(* type *)
intros G2 _ G1 s _ _ Hs.
simp_sub.
apply mrk_type_intro; done.

(* sing *)
intros G2 c Hofc Hmrkat G1 s _ HG2 Hs.
simp_sub.
#2 apply mrk_sing_intro'.
  eapply subst_ofc_type; eauto using subof_mrcl_sof; done.

  intros E E' HintE.
  so (intctx_sub _#5 HG2 Hs HintE) as (s' & E2 & E2' & Hsid & HintE2 & Hsubenv & Hsubenv').
  so (sid_ofc _#6 Hsid Hofc) as Heq.
  rewrite -> Heq; [].
  so (subst_intk _#5 (fkd_sing (ofc_fcn _#3 Hofc)) (intctx_fenv _#3 HintE2) Hsubenv) as H.
  simp_sub in H.
  rewrite -> H; clear H.
  so (subst_intk _#5 (fkd_sing (ofc_fcn _#3 Hofc)) (intctx_fenv' _#3 HintE2) Hsubenv') as H.
  simp_sub in H.
  rewrite -> H; clear H.
  exact (Hmrkat _ _ HintE2 andel).

(* pi *)
intros G2 k1 k2 _ Hmrk1 IH1 _ IH2 G1 s Hmr HG2 Hs.
simp_sub.
apply mrk_pi_intro; auto; [].
#3 apply IH2.
  apply mr_cn; auto; done.

  apply wf_cn; auto using mrk_ofk; done.

  apply subof_bind_mrcl; auto using ofk_fkd, mrk_ofk, mr_fcontext; done.

(* sigma *)
intros G2 k1 k2 _ Hmrk1 IH1 _ IH2 G1 s Hmr HG2 Hs.
simp_sub.
apply mrk_sigma_intro; auto; [].
#3 apply IH2.
  apply mr_cn; auto; done.

  apply wf_cn; auto using mrk_ofk; done.

  apply subof_bind_mrcl; auto using ofk_fkd, mrk_ofk, mr_fcontext; done.

(* one *)
intros G2 _ G1 s _ _ Hs.
simp_sub.
apply mrk_one_intro; done.
Qed.


Lemma mrk_conv :
  forall G k1 k1' k2,
    wf G
    -> mr G
    -> eqk G k1 k1'
    -> mrk G k1'
    -> (forall E, intctx G E E -> intk k1 E = intk k1' E)
    -> mrk (G; cl_cn k1) k2
    -> mrk (G; cl_cn k1') k2.
Proof.
intros G k1 k1' k2 Hwf Hmr Heqk Hmrk1' Hfun Hmrk2.
replsub k2 with (subst (dot (var 0) sh1) k2).
#2 apply (subst_mrk _ (G; cl_cn k1)); auto using mr_cn.
  apply wf_cn; auto; [].
  exact (eqk_reg _#3 Hwf Heqk andel).
apply subof_dot; eauto using subof_sh with sequence; [].
simp_sub.
#2 apply mrcl_cn.
  eapply ofc_subsume; eauto using ofc_var with sequence; [].
  apply weaken1_subk; eauto with static; done.

  intros Ex Ex' HintGx.
  invert HintGx; [].
  intros E E' c1 c1' c2 c2' Q Q' HintE Heqc1 Heqc2 Hintk1' <- <-.
  simpl.
  simp_sub.
  assert (subenv (E; Sta c1 c2 Q) sh1 E) as Hsh1 by (apply subenv_sh; auto with sequence).
  so (subst_intk _#5 (eqk_freg _#3 (wf_fcontext _ Hwf) Heqk andel) (intctx_fenv _#3 HintE) Hsh1) as Heq.
  unfold shift1.
  rewrite -> Heq; [].
  rewrite -> Hfun; eauto using intctx_refl_mr.
Qed.


Lemma intk_fst_fun :
  forall G k c c',
    mr G
    -> mrk G k
    -> eqc G c c' k
    -> (forall E E',
          intctx G E E'
          -> intk k E (subst (left E) c) (subst (right E) c) (intc c E) (intc c' E'))
    -> (forall E E',
          intctx G E E'
          -> intk k E (subst (left E) c) (subst (right E) c) (intc c E) (intc c E')).
Proof.
intros G k c c' Hmr Hmrk Heqc Hfun E E' HintE.
eapply (intk_trans_mr _#7 (intc c' E')); eauto; [].
eapply intk_symm_mr; eauto; [].
rewrite -> (intk_fun_mr _#4 Hmrk HintE); [].
so (Hfun _ _ (intctx_refl_mr _ _ _ Hmr (intctx_symm_mr _ _ _ Hmr HintE))) as H.
so (eqc_reg _#4 (mr_wf _ Hmr) Heqc) as (Hofc & _).
#2 eapply intk_defeq_mr; eauto using intctx_symm_mr, intctx_refl_mr; clear H.
  eapply seq_ofc; eauto using wf_nil; [].
  apply intctx_left; auto using intctx_symm_mr; done.

  eapply seq_ofc; eauto using wf_nil; [].
  apply intctx_right; auto using intctx_symm_mr; done.
Qed.
    

Lemma intk_snd_fun :
  forall G k c c',
    mr G
    -> mrk G k
    -> eqc G c c' k
    -> (forall E E',
          intctx G E E'
          -> intk k E (subst (left E) c) (subst (right E) c) (intc c E) (intc c' E'))
    -> (forall E E',
          intctx G E E'
          -> intk k E (subst (left E) c') (subst (right E) c') (intc c' E) (intc c' E')).
Proof.
intros G k c c' Hmr Hmrk Heqc Hfun E E' HintE.
eapply intk_fst_fun; eauto using eqc_symm; [].
clear E E' HintE.
intros E E' HintE.
eapply intk_symm_mr; eauto; [].
so (Hfun _ _ (intctx_symm_mr _#3 Hmr HintE)) as H.
rewrite -> (intk_fun_mr _#4 Hmrk HintE); [].
#2 eapply intk_defeq_mr; eauto using intctx_symm_mr.
  eapply (seq_eqc nil _ _ G); eauto using mr_wf, wf_nil; [].
  apply intctx_left; auto using intctx_symm_mr; done.

  eapply (seq_eqc nil _ _ G); eauto using mr_wf, wf_nil; [].
  apply intctx_right; auto using intctx_symm_mr; done.
Qed.


Lemma unit_interp_fun :
  forall E E',
    unit_interp E = unit_interp E'.
Proof.
cut (forall E E' v v', unit_interp E v v' -> unit_interp E' v v').
  intros H E E'.
  fextensionality 2.
  intros v v'.
  pextensionality; eauto; done.
intros E E' v v' H.  
invert H; clear H; [].
intros; subst; apply unit_interp_.
Qed.


Lemma unit_interp_tvrel :
  forall E,
    tvrel cn_unit cn_unit (unit_interp E).
Proof.
intros E.
unfold tvrel.
intros v v' H.
invert H; [].
intros <- <-.
do2 3 split; auto with static dynamic.
Qed.


Lemma unit_versions :
  forall E,
    unit_interp E = unit_rel.
Proof.
intro E.
fextensionality 2; intros v v'; pextensionality; intro H; invert H; intros; subst; eauto using unit_interp_, unit_rel_.
Qed.


Lemma arrow_interp_fun :
  forall G t1 t1' t2 t2' It1 It2 E E',
    mr G
    -> eqc G t1 t1' kd_type
    -> eqc G t2 t2' kd_type
    -> intctx G E E'
    -> (forall E E', intctx G E E' -> It1 E = It1 E')
    -> (forall E E', intctx G E E' -> It2 E = It2 E')
    -> arrow_interp t1 t2 It1 It2 E = arrow_interp t1' t2' It1 It2 E'.
Proof.
intros G t1 t1' t2 t2' R1 R2 E E' Hmr Heqt1 Heqt2 HintE Hfun1 Hfun2.
revert t1 t1' t2 t2' E E' Heqt1 Heqt2 HintE.
cut (forall t1 t1' t2 t2' E E' v v',
       eqc G t1 t1' kd_type
       -> eqc G t2 t2' kd_type
       -> intctx G E E'
       -> arrow_interp t1 t2 R1 R2 E v v'
       -> arrow_interp t1' t2' R1 R2 E' v v').
  intros H t1 t1' t2 t2' E E' Heqt1 Heqt2 HintE.
  fextensionality 2.
  intros v v'.
  pextensionality; eauto using intctx_symm_mr, eqc_symm; done.
intros t1 t1' t2 t2' E E' v v' Heqt1 Heqt2 HintE H.
invert H; clear H; [].
intros Hof Hof' Hval Hval' Hrel.
#3 apply arrow_interp_; auto.
  eapply oft_equiv; eauto; [].
  eapply (seq_eqc_type nil _ _ G); eauto using wf_nil, mr_wf, intctx_left with static; done.

  eapply oft_equiv; eauto; [].
  eapply (seq_eqc_type nil _ _ G); eauto using wf_nil, mr_wf, intctx_right with static; done.

  intros w w' Hrelw.
  rewrite <- (Hfun1 _ _ HintE) in Hrelw.
  rewrite <- (Hfun2 _ _ HintE).
  so (Hrel _ _ Hrelw) as H.
  #2 rewrite <- (tev_eqc (subst (left E) t2) _ (subst (right E) t2)); auto.
    eapply (seq_eqc_type nil _ _ G); eauto using wf_nil, mr_wf, intctx_left; done.

    eapply (seq_eqc_type nil _ _ G); eauto using wf_nil, mr_wf, intctx_right; done.
Qed.


Lemma arrow_versions :
  forall t1 t2 Ic1 Ic2 E,
    arrow_interp t1 t2 Ic1 Ic2 E = arrow_rel (subst (left E) t1) (subst (right E) t1) (subst (left E) t2) (subst (right E) t2) (Ic1 E) (Ic2 E).
Proof.
intros t1 t2 Ic1 Ic2 E.
fextensionality 2; intros v v'; pextensionality; intro H; invert H; auto using arrow_interp_, arrow_rel_; done.
Qed.


Lemma arrow_interp_tvrel :
  forall t1 t2 E Ic1 Ic2,
    tvrel (cn_arrow (subst (left E) t1) (subst (left E) t2)) (cn_arrow (subst (right E) t1) (subst (right E) t2)) (arrow_interp t1 t2 Ic1 Ic2 E).
Proof.
intros t1 t2 E Ic1 Ic2.
rewrite -> !arrow_versions; [].
apply arrow_rel_tvrel; auto.
Qed.


Lemma prod_interp_fun :
  forall G t1 t1' t2 t2' It1 It2 E E',
    mr G
    -> eqc G t1 t1' kd_type
    -> eqc G t2 t2' kd_type
    -> intctx G E E'
    -> (forall E E', intctx G E E' -> It1 E = It1 E')
    -> (forall E E', intctx G E E' -> It2 E = It2 E')
    -> prod_interp t1 t2 It1 It2 E = prod_interp t1' t2' It1 It2 E'.
Proof.
intros G t1 t1' t2 t2' R1 R2 E E' Hmr Heqt1 Heqt2 HintE Hfun1 Hfun2.
revert t1 t1' t2 t2' E E' Heqt1 Heqt2 HintE.
cut (forall t1 t1' t2 t2' E E' v v',
       eqc G t1 t1' kd_type
       -> eqc G t2 t2' kd_type
       -> intctx G E E'
       -> prod_interp t1 t2 R1 R2 E v v'
       -> prod_interp t1' t2' R1 R2 E' v v').
  intros H t1 t1' t2 t2' E E' Heqt1 Heqt2 HintE.
  fextensionality 2.
  intros v v'.
  pextensionality; eauto using intctx_symm_mr, eqc_symm; done.
intros t1 t1' t2 t2' E E' v v' Heqt1 Heqt2 HintE H.
invert H; clear H; [].
intros v1 v1' v2 v2' Hofv Hofv' Hvalv Hvalv' Hv1 Hv2 <- <-.
#4 apply prod_interp_; auto.
  eapply oft_equiv; eauto; [].
  eapply (seq_eqc_type nil _ _ G); eauto using wf_nil, mr_wf, intctx_left with static; done.

  eapply oft_equiv; eauto; [].
  eapply (seq_eqc_type nil _ _ G); eauto using wf_nil, mr_wf, intctx_right with static; done.

  erewrite <- Hfun1; eauto; done.

  erewrite <- Hfun2; eauto; done.
Qed.


Lemma prod_versions :
  forall t1 t2 Ic1 Ic2 E,
    prod_interp t1 t2 Ic1 Ic2 E = prod_rel (subst (left E) t1) (subst (right E) t1) (subst (left E) t2) (subst (right E) t2) (Ic1 E) (Ic2 E).
Proof.
intros t1 t2 Ic1 Ic2 E.
fextensionality 2; intros v v'.
pextensionality; intros H; invert H; intros; subst; auto using prod_rel_, prod_interp_.
Qed.
  

Lemma prod_interp_tvrel :
  forall t1 t2 E Ic1 Ic2,
    tvrel (cn_prod (subst (left E) t1) (subst (left E) t2)) (cn_prod (subst (right E) t1) (subst (right E) t2)) (prod_interp t1 t2 Ic1 Ic2 E).
Proof.
intros t1 t2 E Ic1 Ic2.
rewrite -> !prod_versions; [].
apply prod_rel_tvrel; auto.
Qed.


Lemma all_interp_fun :
  forall G k k' t t' It It' E E',
    mr G
    -> mrk G k
    -> eqk G k k'
    -> eqc (G; cl_cn k) t t' kd_type
    -> (forall E E', intctx G E E' -> intk k E = intk k' E')
    -> (forall E E', intctx (G; cl_cn k) E E' -> It E = It' E')
    -> intctx G E E'
    -> all_interp k t (intk k) It E = all_interp k' t' (intk k') It' E'.
Proof.
cut (forall G k k' t t' It It' E E' v v',
       mr G
       -> eqk G k k'
       -> eqc (G; cl_cn k) t t' kd_type
       -> (forall E E', intctx G E E' -> intk k E = intk k' E')
       -> (forall E E', intctx (G; cl_cn k) E E' -> It E = It' E')
       -> intctx G E E'
       -> all_interp k t (intk k) It E v v'
       -> all_interp k' t' (intk k') It' E' v v').
  intros Hprop G k k' t t' It It' E E' Hmr Hmrk Heqk Heqt Hfunk Hfunt HintE.
  fextensionality 2.
  intros v v'.
  pextensionality.
    intro H.
    eapply Hprop; eauto; done.

    intro H.
    #3 eapply Hprop; eauto using eqk_symm, intctx_symm_mr; clear H.
      apply eqc_symm; [].
      so (subst_eqc _#6 (subof_conv_0 _#3 (eqk_symm Heqk)) Heqt) as H.
      simp_sub in H.
      assumption.

      clear E E' HintE.
      intros E E' HintE.
      symmetry.
      apply Hfunk; auto using intctx_symm_mr; done.

      clear E E' HintE.
      intros E E' HintE.
      symmetry.
      apply Hfunt; [].
      apply intctx_symm_mr; auto using mr_cn; [].
      invert HintE; [].
      intros D D' c1 c1' c2 c2' Q Q' HintD Heqc1 Heqc2 Hintk <- <-.
      #3 apply intctx_cn; auto.
        eapply eqc_subsume; eauto; [].
        apply subk_refl; [].
        apply eqk_symm; [].
        eapply subst_eqk; eauto using left_subof; done.
      
        eapply eqc_subsume; eauto; [].
        apply subk_refl; [].
        apply eqk_symm; [].
        eapply subst_eqk; eauto using right_subof; done.
      
        rewrite -> (Hfunk _ _ (intctx_refl_mr _#3 Hmr HintD)); [].
        assumption.
intros G k k' t t' It It' E E' v v' Hmr Heqk Heqt Hfunk Hfunt HintE H.
invert H; clear H; [].
intros Hofv Hofv' Hvalv Hvalv' Hrel.
#3 apply all_interp_; auto.
  eapply oft_equiv; eauto; [].
  apply (seq_eqc_type _ _ _ G); auto using mr_wf, wf_nil, intctx_left with static; done.

  eapply oft_equiv; eauto; [].
  apply (seq_eqc_type _ _ _ G); auto using mr_wf, wf_nil, intctx_right with static; done.

  intros c c' Q HQ.
  rewrite <- (Hfunk _ _ HintE) in HQ; [].
  so (intk_reg _#6 HQ) as (Hofc & Hofc').
  #2 rewrite <- (Hfunt (E; Sta c c' Q)).
    #3 rewrite <- (tev_eqc (subst (dot c (left E)) t) (subst (dot c (left E')) t') (subst (dot c' (right E)) t) (subst (dot c' (right E')) t')).
      apply Hrel; auto; done.

      #2 eapply (seq_eqc_type _#3 (G; cl_cn k)); eauto using wf_nil.
        apply wf_cn; auto using mr_wf; [].
        exact (eqk_reg _#3 (mr_wf _ Hmr) Heqk andel).

        #2 apply subrel_dot.
          apply intctx_left; auto; done.

          simp_sub.
          apply seq_cn; [].
          apply eqc_refl; auto; done.

      #2 eapply (seq_eqc_type _#3 (G; cl_cn k)); eauto using wf_nil.
        apply wf_cn; auto using mr_wf; [].
        exact (eqk_reg _#3 (mr_wf _ Hmr) Heqk andel).

        #2 apply subrel_dot.
          apply intctx_right; auto; done.

          simp_sub.
          apply seq_cn; [].
          apply eqc_refl; auto; done.

    apply intctx_cn; auto with static; done.
Qed.


Lemma all_versions :
  forall k t Ik It E,
    all_interp k t Ik It E =
    all_rel cand (subst (left E) k) (subst (right E) k) (subst (dot (var 0) (compose (left E) sh1)) t) (subst (dot (var 0) (compose (right E) sh1)) t) (fun c c' Q => Ik E c c' Q Q) (fun c c' Q v v' => It (E; Sta c c' Q) v v').
Proof.
intros k t Ik It E.
fextensionality 2; intros v v'.
pextensionality.
  intros H; invert H; clear H.
  simp_sub.
  intros.
  apply all_rel_; auto; [].
  intros; simp_sub; auto; done.

  intros H; invert H; clear H.
  simp_sub.
  intros ? ? ? ? H.
  apply all_interp_; simp_sub; auto; [].
  intros c c' Q Hc.
  so (H _ _ _ Hc) as H'.
  simp_sub in H'.
  assumption.
Qed.


Lemma all_interp_tvrel :
  forall k t E Ik It,
    tvrel
    (cn_all (subst (left E) k) (subst (dot (var 0) (compose (left E) sh1)) t))
    (cn_all (subst (right E) k) (subst (dot (var 0) (compose (right E) sh1)) t))
    (all_interp k t Ik It E).
Proof.
intros k t E Ik It.
rewrite -> !all_versions; [].
apply all_rel_tvrel; auto.
Qed.


Lemma exists_interp_fun :
  forall G k k' t t' It It' E E',
    mr G
    -> mrk G k
    -> eqk G k k'
    -> eqc (G; cl_cn k) t t' kd_type
    -> (forall E E', intctx G E E' -> intk k E = intk k' E')
    -> (forall E E', intctx (G; cl_cn k) E E' -> It E = It' E')
    -> intctx G E E'
    -> exists_interp k t (intk k) It E = exists_interp k' t' (intk k') It' E'.
Proof.
cut (forall G k k' t t' It It' E E' v v',
       mr G
       -> eqk G k k'
       -> eqc (G; cl_cn k) t t' kd_type
       -> (forall E E', intctx G E E' -> intk k E = intk k' E')
       -> (forall E E', intctx (G; cl_cn k) E E' -> It E = It' E')
       -> intctx G E E'
       -> exists_interp k t (intk k) It E v v'
       -> exists_interp k' t' (intk k') It' E' v v').
  intros Hprop G k k' t t' It It' E E' Hmr Hmrk Heqk Heqt Hfunk Hfunt HintE.
  fextensionality 2.
  intros v v'.
  pextensionality.
    intro H.
    eapply Hprop; eauto; done.

    intro H.
    #3 eapply Hprop; eauto using eqk_symm, intctx_symm_mr; clear H.
      apply eqc_symm; [].
      so (subst_eqc _#6 (subof_conv_0 _#3 (eqk_symm Heqk)) Heqt) as H.
      simp_sub in H.
      assumption.

      clear E E' HintE.
      intros E E' HintE.
      symmetry.
      apply Hfunk; auto using intctx_symm_mr; done.

      clear E E' HintE.
      intros E E' HintE.
      symmetry.
      apply Hfunt; [].
      apply intctx_symm_mr; auto using mr_cn; [].
      invert HintE; [].
      intros D D' c1 c1' c2 c2' Q Q' HintD Heqc1 Heqc2 Hintk <- <-.
      #3 apply intctx_cn; auto.
        eapply eqc_subsume; eauto; [].
        apply subk_refl; [].
        apply eqk_symm; [].
        eapply subst_eqk; eauto using left_subof; done.
      
        eapply eqc_subsume; eauto; [].
        apply subk_refl; [].
        apply eqk_symm; [].
        eapply subst_eqk; eauto using right_subof; done.
      
        rewrite -> (Hfunk _ _ (intctx_refl_mr _#3 Hmr HintD)); [].
        assumption.
intros G k k' t t' It It' E E' v v' Hmr Heqk Heqt Hfunk Hfunt HintE H.
invert H; clear H; [].
intros w w' c c' l l' u u' Q Hofv Hofv' Hvalv Hvalv' HQ Hv <- <-.
so (intk_reg _#6 HQ) as (Hofc & Hofc').
#4 apply (exists_interp_ _#13 Q); auto.
  eapply oft_equiv; eauto; [].
  apply (seq_eqc_type _ _ _ G); auto using mr_wf, wf_nil, intctx_left with static; done.

  eapply oft_equiv; eauto; [].
  apply (seq_eqc_type _ _ _ G); auto using mr_wf, wf_nil, intctx_right with static; done.

  rewrite <- (Hfunk _ _ HintE); [].
  assumption.

  rewrite <- (Hfunt (E; Sta c c' Q)); auto; [].
  apply intctx_cn; auto with static; done.
Qed.


Lemma exists_interp_st_fun :
  forall G k k' t t' It It' E E',
    mr G
    -> mrk G k
    -> eqk G k k'
    -> eqc (G; cl_cn k) t t' kd_type
    -> (forall E E', intctx G E E' -> intk k E = intk k' E')
    -> (forall E E', intctx (G; cl_cn k) E E' -> It E = It' E')
    -> intctx G E E'
    -> exists_interp_st k t (intk k) It E = exists_interp_st k' t' (intk k') It' E'.
Proof.
intros G k k' t t' It It' E E' Hmr Hmrk Heqk Heqt Hfunk Hfunt HintE.
unfold exists_interp_st.
rewrite <- (exists_interp_fun G k k' t t' It It' E E'); auto; [].
#2 apply tpst_eqc.
  eapply (seq_eqc_type _#3 G); eauto using mr_wf, wf_nil, intctx_left, eqc_exists; done.

  eapply (seq_eqc_type _#3 G); eauto using mr_wf, wf_nil, intctx_right, eqc_exists; done.
Qed.


Lemma exists_interp_tvrel :
  forall k t E Ik It,
    tvrel
    (cn_exists (subst (left E) k) (subst (dot (var 0) (compose (left E) sh1)) t))
    (cn_exists (subst (right E) k) (subst (dot (var 0) (compose (right E) sh1)) t))
    (exists_interp k t Ik It E).
Proof.
intros k t E Ik It.
intros v v' Hv.
invert Hv; [].
intros w w' c c' l l' u u' Q Hof Hof' Hval Hval' _ _ <- <-.
simp_sub in Hof; simp_sub in Hof'.
do2 3 split; auto.
Qed.


Theorem model_regular :
  (forall G k,
     mr G
     -> ofk G k
     -> mrk G k)
  /\
  (forall G k1 k2,
     mr G
     -> eqk G k1 k2
     -> mrk G k1
        /\ mrk G k2
        /\ (forall E, intctx G E E -> intk k1 E = intk k2 E))
  /\
  (forall G k1 k2,
     mr G
     -> subk G k1 k2
     -> mrk G k1
        /\ mrk G k2
        /\ (forall E c1 c2 Q1 Q2,
              intctx G E E
              -> intk k1 E c1 c2 Q1 Q2
              -> intk k2 E c1 c2 Q1 Q2))
  /\
  (forall G c k,
     mr G
     -> ofc G c k
     -> mrk G k
        /\ (forall E E',
              intctx G E E'
              -> intk k E (subst (left E) c) (subst (right E) c) (intc c E) (intc c E')))
  /\
  (forall G c1 c2 k,
     mr G
     -> eqc G c1 c2 k
     -> mrk G k
        /\ (forall E E',
              intctx G E E'
              -> intk k E (subst (left E) c1) (subst (right E) c1) (intc c1 E) (intc c2 E')))
.
Proof.
exploit
  (static_ind
     (fun G k =>
        mr G
        -> mrk G k)
     (fun G k1 k2 =>
        mr G
        -> mrk G k1
           /\ mrk G k2
           /\ (forall E, intctx G E E -> intk k1 E = intk k2 E))
     (fun G k1 k2 =>
        mr G
        -> mrk G k1
           /\ mrk G k2
           /\ (forall E c1 c2 Q1 Q2,
                 intctx G E E
                 -> intk k1 E c1 c2 Q1 Q2
                 -> intk k2 E c1 c2 Q1 Q2))
     (fun G c k =>
        mr G
        -> mrk G k
           /\ (forall E E',
                 intctx G E E'
                 -> intk k E (subst (left E) c) (subst (right E) c) (intc c E) (intc c E')))
     (fun G c1 c2 k =>
        mr G
        -> mrk G k
           /\ (forall E E',
                 intctx G E E'
                 -> intk k E (subst (left E) c1) (subst (right E) c1) (intc c1 E) (intc c2 E'))))
  as Hind.

(* type *)
intros G _.
apply mrk_type_intro; done.

(* sing *)
intros G c Hofc IH Hmr.
apply mrk_sing_intro; auto; [].
intros E E' HintG.
so (IH Hmr ander _ _ HintG) as Hint.
simpl in Hint.
invert Hint; clear Hint; [].
intros R _ _ _ _ <- <-.
reflexivity.

(* pi *)
intros G k1 k2 _ IH1 _ IH2 Hmr.
apply mrk_pi_intro; auto using mr_cn, wf_cn, mrk_ofk; done.

(* sigma *)
intros G k1 k2 _ IH1 _ IH2 Hmr.
apply mrk_sigma_intro; auto using mr_cn, wf_cn, mrk_ofk; done.

(* one *)
intros G _.
apply mrk_one_intro; done.

(* refl *)
intros G k Hofk IH Hmr.
auto; done.

(* symm *)
intros G k1 k2 Heqk IH Hmr.
so (IH Hmr) as (Hmrk1 & Hmrk2 & Hfun).
do2 2 split; auto; [].
intros; symmetry; auto; done.

(* trans *)
intros G k1 k2 k3 _ IH1 _ IH2 Hmr.
so (IH1 Hmr) as (Hmrk1 & Hmrk2 & Hfun).
so (IH2 Hmr) as (_ & Hmrk3 & Hfun').
do2 2 split; auto; [].
intros.
transitivity (intk k2 E); auto; done.

(* sing *)
intros G c c' Heqc IH Hmr.
so (IH Hmr) as (_ & H); renameover H into IH.
do2 2 split.
  #2 apply mrk_sing_intro; auto.
    exact (eqc_reg _#4 (mr_wf _ Hmr) Heqc andel).

    intros E E' HintG.
    transitivity (intc c' E').
      eapply type_interp_eq; eauto; done.

      symmetry.
      eapply type_interp_eq; eauto using intctx_refl_mr, intctx_symm_mr; done.

  #2 apply mrk_sing_intro; auto.
    exact (eqc_reg _#4 (mr_wf _ Hmr) Heqc ander andel).

    intros E E' HintG.
    transitivity (intc c E').
      symmetry.
      eapply type_interp_eq; eauto using intctx_symm_mr; done.

      eapply type_interp_eq; eauto using intctx_refl_mr, intctx_symm_mr; done.

  intros E HintG.
  so (IH _ _ HintG) as Hint; clear IH.
  simpl in Hint.
  simpl.
  fextensionality 4.
  intros b b' Q Q'.
  pextensionality.
    intro H; invert H; clear H; [].
    intros Heqb Heqb' Hrel Hcl <- <-.
    rewrite -> (type_interp_eq _#5 Hint); [].
    #4 apply sing_interp_.
      apply (@eqc_trans _ _ (subst (left E) c)); auto; [].
      eapply subst_eqc_type; eauto using left_subof; done.

      apply (@eqc_trans _ _ (subst (right E) c)); auto; [].
      eapply subst_eqc_type; eauto using right_subof; done.

      rewrite <- (type_interp_eq _#5 Hint); [].
      assumption.

      rewrite <- (type_interp_eq _#5 Hint); [].
      assumption.

    intro H; invert H; clear H; [].
    intros Heqb Heqb' Hrel Hcl <- <-.
    rewrite <- (type_interp_eq _#5 Hint); [].
    #4 apply sing_interp_.
      apply (@eqc_trans _ _ (subst (left E) c')); auto; [].
      apply eqc_symm; [].
      eapply subst_eqc_type; eauto using left_subof; done.

      apply (@eqc_trans _ _ (subst (right E) c')); auto; [].
      apply eqc_symm; [].
      eapply subst_eqc_type; eauto using right_subof; done.

      rewrite -> (type_interp_eq _#5 Hint); [].
      assumption.

      rewrite -> (type_interp_eq _#5 Hint); [].
      assumption.

(* pi *)
intros G k1 k1' k2 k2' Heqk1 IH1 Heqk2 IH2 Hmr.
so (IH1 Hmr) as (Hmrk1 & Hmrk1' & Hfun1); clear IH1.
so (mrk_ofk _ _ Hmrk1) as Hofk1.
so (IH2 (mr_cn Hmr Hmrk1)) as (Hmrk2 & Hmrk2' & Hfun2); clear IH2.
do2 2 split.
  apply mrk_pi_intro; auto; done.

  apply mrk_pi_intro; auto; [].
  eapply mrk_conv; eauto using mr_wf; done.
  
  intros E HintG.
  simpl.
  fextensionality 4.
  intros c c' Q Q'.
  pextensionality.
    intro H; invert H; clear H; [].
    intros F F' Hofc Hofc' Hintapp <- <-.
    revert F F' Hintapp.
    rewrite -> (eqk_simp _#3 Heqk1); [].
    rewrite -> (eqk_simp _#3 Heqk2); [].
    intros F F' Hintapp.
    #3 apply pi_interp_.
      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      eapply subst_eqk; eauto using left_subof with static; done.

      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      eapply subst_eqk; eauto using right_subof with static; done.

      intros d d' Q Q' d'' d''' Hintk1 Heqd Heqd'.
      assert (eqc nil d d'' (subst (left E) k1)) as Heqda.
        eapply eqc_subsume; eauto; [].
        apply subk_refl; [].
        apply eqk_symm; [].
        eapply subst_eqk; eauto using left_subof with static; done.
      assert (eqc nil d' d''' (subst (right E) k1)) as Heqd'a.
        eapply eqc_subsume; eauto; [].
        apply subk_refl; [].
        apply eqk_symm; [].
        eapply subst_eqk; eauto using right_subof with static; done.
      revert F F' Q Q' Hintapp Hintk1.
      rewrite <- (eqk_simp _#3 Heqk1); [].
      rewrite <- (eqk_simp _#3 Heqk2); [].
      intros F F' Q Q' Hintapp Hintk1.
      assert (intctx (G; cl_cn k1) (E; Sta d d' (Cand (simp k1) Q)) (E; Sta d d' (Cand (simp k1) Q))) as HintGx.
        apply intctx_cn; eauto with static; [].
        rewrite -> (Hfun1 _ HintG); [].
        eapply intk_refl_mr; eauto; done.
      rewrite <- (Hfun2 _ HintGx); [].
      apply Hintapp; auto; [].
      rewrite -> (Hfun1 _ HintG); [].
      assumption.

    intro H; invert H; clear H; [].
    intros F F' Hofc Hofc' Hintapp <- <-.
    revert F F' Hintapp.
    rewrite <- (eqk_simp _#3 Heqk1); [].
    rewrite <- (eqk_simp _#3 Heqk2); [].
    intros F F' Hintapp.
    #3 apply pi_interp_.
      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      apply eqk_symm; [].
      eapply subst_eqk; eauto using left_subof with static; done.

      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      apply eqk_symm; [].
      eapply subst_eqk; eauto using right_subof with static; done.

      intros d d' Q Q' d'' d''' Hintk1 Heqd Heqd'.
      assert (eqc nil d d'' (subst (left E) k1')) as Heqda.
        eapply eqc_subsume; eauto; [].
        apply subk_refl; [].
        eapply subst_eqk; eauto using left_subof with static; done.
      assert (eqc nil d' d''' (subst (right E) k1')) as Heqd'a.
        eapply eqc_subsume; eauto; [].
        apply subk_refl; [].
        eapply subst_eqk; eauto using right_subof with static; done.
      revert F F' Q Q' Hintapp Hintk1.
      rewrite -> (eqk_simp _#3 Heqk1); [].
      rewrite -> (eqk_simp _#3 Heqk2); [].
      intros F F' Q Q' Hintapp Hintk1.
      assert (intctx (G; cl_cn k1) (E; Sta d d' (Cand (simp k1') Q)) (E; Sta d d' (Cand (simp k1') Q))) as HintGx.
        apply intctx_cn; eauto with static; [].
        eapply intk_refl_mr; eauto; done.
      rewrite -> (Hfun2 _ HintGx); [].
      apply Hintapp; auto; [].
      rewrite <- (Hfun1 _ HintG); [].
      assumption.

(* sigma *)
intros G k1 k1' k2 k2' Heqk1 IH1 Heqk2 IH2 Hmr.
so (IH1 Hmr) as (Hmrk1 & Hmrk1' & Hfun1); clear IH1.
so (mrk_ofk _ _ Hmrk1) as Hofk1.
so (IH2 (mr_cn Hmr Hmrk1)) as (Hmrk2 & Hmrk2' & Hfun2); clear IH2.
do2 2 split.
  apply mrk_sigma_intro; auto; done.

  apply mrk_sigma_intro; auto; [].
  eapply mrk_conv; eauto using mr_wf; done.
  
  intros E HintG.
  simpl.
  fextensionality 4.
  intros c c' Q Q'.
  pextensionality.
    intro H; invert H; clear H; [].
    intros P P' Hofc Hofc' Hint1 Hint2 <- <-.
    revert P P' Hint1 Hint2.
    rewrite -> (eqk_simp _#3 Heqk1); [].
    rewrite -> (eqk_simp _#3 Heqk2); [].
    intros P P' Hint1 Hint2.
    #4 apply sigma_interp_.
      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      eapply subst_eqk; eauto with static; [].
      eapply left_subof; eauto; done.
      
      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      eapply subst_eqk; eauto with static; [].
      eapply right_subof; eauto; done.

      rewrite <- (Hfun1 _ HintG); [].
      assumption.

      assert (intctx (G; cl_cn k1) (E; Sta (cn_pi1 c) (cn_pi1 c') (Cand (simp k1') (fst P))) (E; Sta (cn_pi1 c) (cn_pi1 c') (Cand (simp k1') (fst P)))) as HintGx.
        #3 apply intctx_cn; auto.
          apply eqc_refl; [].
          simp_sub in Hofc.
          eapply ofc_pi1; eauto; done.
  
          apply eqc_refl; [].
          simp_sub in Hofc.
          eapply ofc_pi1; eauto; done.

          eapply intk_refl_mr; eauto; done.
      rewrite <- (Hfun2 _ HintGx); [].
      assumption.

    intro H; invert H; clear H; [].
    intros P P' Hofc Hofc' Hint1 Hint2 <- <-.
    revert P P' Hint1 Hint2.
    rewrite <- (eqk_simp _#3 Heqk1); [].
    rewrite <- (eqk_simp _#3 Heqk2); [].
    intros P P' Hint1 Hint2.
    #4 apply sigma_interp_.
      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      apply eqk_symm; [].
      eapply subst_eqk; eauto with static; [].
      eapply left_subof; eauto; done.
      
      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      apply eqk_symm; [].
      eapply subst_eqk; eauto with static; [].
      eapply right_subof; eauto; done.

      rewrite -> (Hfun1 _ HintG); [].
      assumption.

      assert (intctx (G; cl_cn k1) (E; Sta (cn_pi1 c) (cn_pi1 c') (Cand (simp k1) (fst P))) (E; Sta (cn_pi1 c) (cn_pi1 c') (Cand (simp k1) (fst P)))) as HintGx.
        #3 apply intctx_cn; auto.
          apply eqc_refl; [].
          simp_sub in Hofc.
          eapply ofc_subsume; eauto using ofc_pi1; [].
          apply subk_refl; apply eqk_symm; [].
          eapply subst_eqk; eauto with static; [].
          eapply left_subof; eauto; done.

          apply eqc_refl; [].
          simp_sub in Hofc.
          eapply ofc_subsume; eauto using ofc_pi1; [].
          apply subk_refl; apply eqk_symm; [].
          eapply subst_eqk; eauto with static; [].
          eapply right_subof; eauto; done.
  
          rewrite -> (Hfun1 _ HintG); [].
          eapply intk_refl_mr; eauto; done.
      rewrite -> (Hfun2 _ HintGx); [].
      assumption.

(* refl *)
intros G k k' Heqk IH Hmr.          
so (IH Hmr) as (Hmrk & Hmrk' & Hfun); clear IH.
do2 2 split; auto; [].
intros E c1 c2 Q1 Q2 HintE Hintk.
rewrite <- (Hfun _ HintE); [].
assumption.

(* trans *)
intros G k1 k2 k3 Hsubk IH1 _ IH2 Hmr.
so (IH1 Hmr) as (Hmrk1 & Hmrk2 & Hfun1); clear IH1.
so (IH2 Hmr) as (_ & Hmrk3 & Hfun2); clear IH2.
do2 2 split; auto; done.

(* sing_t *)
intros G c Hofc IH Hmr.
so (IH Hmr) as (_ & Hfun); clear IH.
do2 2 split.
  apply mrk_sing_intro; auto; [].
  intros E E' HintE.
  exact (type_interp_eq _#5 (Hfun _ _ HintE)).

  exact (mrk_type_intro _).

  intros E c1 c2 Q1 Q2 HintE Hint.
  simpl in Hint.
  invert Hint; clear Hint; [].
  intros Heqc1 Heqc2 _ _ <- <-.
  so (Hfun _ _ HintE) as Hintt.
  invert Hintt; [].
  intros R _ _ _ _ Heq _.
  rewrite <- Heq; [].
  unfold out; rewrite -> cast_id; [].
  rewrite -> Heq; [].
  #2 eapply intk_defeq_mr; eauto using mrk_type_intro.
    simp_sub; auto with static; done.

    simp_sub; auto with static; done.

(* pi *)
intros G k1 k1' k2 k2' Hsubk1 IH1 Hsubk2 IH2 Hofk2 IH3 Hmr.    
so (IH1 Hmr) as (Hmrk1' & Hmrk1 & Hfun1); clear IH1.
exploit IH2 as (_ & Hmrk2' & Hfun2); clear IH2; eauto using wf, mr, mrk_ofk; [].
exploit IH3 as Hmrk2; clear IH3; eauto using wf, mr, mrk_ofk; [].
do2 2 split.
  apply mrk_pi_intro; auto; done.

  apply mrk_pi_intro; auto; done.

  intros E c1 c2 Q1 Q2 HintE H.
  invert H; clear H; [].
  intros F F' Hofc1 Hofc2 Hintapp <- <-.
  revert F F' Hintapp.
  rewrite <- (subk_simp _#3 Hsubk1); [].
  rewrite -> (subk_simp _#3 Hsubk2); [].
  intros F F' Hintapp.
  simpl.
  #3 apply pi_interp_.
    eapply ofc_subsume; eauto; [].
    eapply subst_subk; eauto using left_subof with static; done.

    eapply ofc_subsume; eauto; [].
    eapply subst_subk; eauto using right_subof with static; done.

    intros d d' Q Q' d'' d''' Hintk1 Heqd Heqd'.
    assert (eqc nil d d'' (subst (left E) k1)) as Heqda.
      eapply eqc_subsume; eauto; [].
      eapply subst_subk; eauto using left_subof with static; done.
    assert (eqc nil d' d''' (subst (right E) k1)) as Heqd'a.
      eapply eqc_subsume; eauto; [].
      eapply subst_subk; eauto using right_subof with static; done.
    revert F F' Q Q' Hintapp Hintk1.
    rewrite -> (subk_simp _#3 Hsubk1); [].
    rewrite <- (subk_simp _#3 Hsubk2); [].
    intros F F' Q Q' Hintapp Hintk1.
    #2 apply Hfun2.
      apply intctx_cn; eauto with static; [].
      eapply intk_refl_mr; eauto; done.

      apply Hintapp; auto; done.

(* sigma *)
intros G k1 k1' k2 k2' Hsubk1 IH1 Hsubk2 IH2 Hofk2' IH3 Hmr.
so (IH1 Hmr) as (Hmrk1' & Hmrk1 & Hfun1); clear IH1.
exploit IH2 as (Hmrk2 & _ & Hfun2); clear IH2; eauto using wf, mr, mrk_ofk; [].
exploit IH3 as Hmrk2'; clear IH3; eauto using wf, mr, mrk_ofk; [].
do2 2 split.
  apply mrk_sigma_intro; auto; done.

  apply mrk_sigma_intro; auto; done.

  intros E c1 c2 Q1 Q2 HintE H.
  invert H; clear H; [].
  intros P P' Hofc Hofc' Hint1 Hint2 <- <-.
  revert P P' Hint1 Hint2.
  rewrite -> (subk_simp _#3 Hsubk1); [].
  rewrite -> (subk_simp _#3 Hsubk2); [].
  intros P P' Hint1 Hint2.
  simpl.
  #4 apply sigma_interp_.
    eapply ofc_subsume; eauto; [].
    eapply subst_subk; eauto with static; [].
    eapply left_subof; eauto; done.
    
    eapply ofc_subsume; eauto; [].
    eapply subst_subk; eauto with static; [].
    eapply right_subof; eauto; done.

    apply Hfun1; auto; done.

    #2 apply Hfun2.
      #3 apply intctx_cn; auto.
        apply eqc_refl; [].
        simp_sub in Hofc.
        eapply ofc_pi1; eauto; done.

        apply eqc_refl; [].
        simp_sub in Hofc.
        eapply ofc_pi1; eauto; done.

        eapply intk_refl_mr; eauto; done.

      assumption.

(* var *)
intros G i k Hindex Hmr.
remember (cl_cn k) as A eqn:HeqA.
revert Hmr HeqA.
induct Hindex.
  (* 0 *)
  intros A G Hmr ->.
  invert Hmr; [].
  intros Hmr' Hmrk.
  split.
    apply (subst_mrk _ G); eauto using mr_wf; [].
    apply subof_sh; auto with sequence; done.

    intros E E' HintE.
    invert HintE; clear HintE; [].
    intros E1 E1' c1 c1' c2 c2' Q Q' HintE1 Heqc1 Heqc2 Hintk <- <-.
    simpl.
    simp_sub.
    assert (subenv (E1; Sta c1 c2 Q) sh1 E1) as Hsubenv.
      apply subenv_sh; auto with sequence; done.
    unfold shift1.
    rewrite -> (subst_intk _#5 (ofk_fkd _ _ (mrk_ofk _ _ Hmrk)) (intctx_fenv _#3 HintE1) Hsubenv); [].
    assumption.

  (* S *)
  intros i B G A _ IH Hmr HeqA.
  so (IH (mr_tail _ _ Hmr) HeqA) as (Hmrk & Hfun); clear IH.
  subst A.
  split.
    replsub (subst (sh (S (S i))) k) with (subst sh1 (subst (sh (S i)) k)).
    apply (subst_mrk _ G); eauto using mr_tail, mr_wf; [].
    apply subof_sh; auto with sequence; done.

    intros E E' HintE.
    assert (exists E1 E1' ent ent', intctx G E1 E1' /\ E = E1; ent /\ E' = E1'; ent') as (E1 & E1' & ent & ent' & HintE' & -> & ->).
      invert HintE; intros; subst; do 4 eexists; eauto; done.
    replsub (subst (sh (S (S i))) k) with (subst sh1 (subst (sh (S i)) k)).
    assert (subenv (E1; ent) sh1 E1) as Hsubenv.
      apply subenv_sh; auto with sequence; done.
    rewrite -> (subst_intk _#5 (ofk_fkd _ _ (mrk_ofk _ _ Hmrk)) (intctx_fenv _#3 HintE') Hsubenv); [].
    force_exact (Hfun _ _ HintE').
    #2 f_equal.
      destruct ent; simpl; simp_sub; reflexivity.

      destruct ent; simpl; simp_sub; reflexivity.

(* lam *)
intros G k1 k2 c Hofk1 IH1 Hofc IH2 Hmr.
so (IH1 Hmr) as Hmrk1; clear IH1.
so (IH2 (mr_cn Hmr Hmrk1)) as (Hmrk2 & Hfun); clear IH2.
split.
  apply mrk_pi_intro; auto; done.

  intros E E' HintE.
  simpl.
  fold precand.
  rewrite <- (intctx_flatten _#3 HintE andel); [].
  rewrite <- (intctx_flatten _#3 HintE ander); [].
  replace (flatten_ctx G; Some (simp k1)) with (flatten_ctx (G; cl_cn k1)) by reflexivity.
  rewrite -> (ofc_skof _#3 (wf_cn (mr_wf _ Hmr) Hofk1) Hofc); [].
  assert (ofc G (cn_lam k1 c) (kd_pi k1 k2)) as Hofclam by (auto with static).
  #3 apply pi_interp_.
    eapply subst_ofc; eauto using left_subof; done.

    eapply subst_ofc; eauto using right_subof; done.

    intros d d' Q Q' d'' d''' Hint1 Heqd Heqd'.
    assert (intctx (G; cl_cn k1) (E; Sta d d' (Cand (simp k1) Q)) (E'; Sta d'' d''' (Cand (simp k1) Q'))) as HintEx by (apply intctx_cn; auto).
    so (Hfun _ _ HintEx) as Hint2.
    simpl in Hint2.
    rewrite -> (cand_eta _ _ (intk_csk _#6 Hint2 andel)); [].
    rewrite -> (cand_eta _ _ (intk_csk _#6 Hint2 ander)); [].
    #2 eapply (intk_defeq_mr (G; cl_cn k1)); eauto using mr.
      simpl.
      simp_sub.
      apply eqc_symm; [].
      replsub (subst (dot d (left E)) c) with (subst1 d (subst (dot (var 0) (compose (left E) sh1)) c)).
      replsub (subst (dot d (left E)) k2) with (subst1 d (subst (dot (var 0) (compose (left E) sh1)) k2)).
      #2 apply eqc_beta.
        eapply subst_ofc; eauto; [].
        apply subof_bind_sof_cn; eauto using left_subof; done.

        exact (eqc_reg _#4 wf_nil Heqd andel).

      simpl.
      simp_sub.
      apply eqc_symm; [].
      replsub (subst (dot d' (right E)) c) with (subst1 d' (subst (dot (var 0) (compose (right E) sh1)) c)).
      replsub (subst (dot d' (right E)) k2) with (subst1 d' (subst (dot (var 0) (compose (right E) sh1)) k2)).
      #2 apply eqc_beta.
        eapply subst_ofc; eauto; [].
        apply subof_bind_sof_cn; eauto using right_subof; done.

        exact (eqc_reg _#4 wf_nil Heqd' andel).
    
(* app *)
intros G k1 k2 c1 c2 Hofc1 IH1 Hofc2 IH2 Hmr.
so (IH1 Hmr) as (Hmrkpi & Hfun1); clear IH1.
so (IH2 Hmr) as (Hmrk1 & Hfun2); clear IH2.
invert Hmrkpi; [].
intros _ _ Hmrk2.
split.  
  #2 eapply subst_mrk; eauto.
    auto using wf_cn, mr_wf, mrk_ofk; done.

    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply mrcl_cn; auto; done.

  intros E E' HintE.
  so (Hfun1 _ _ HintE) as Hint1.
  so (Hfun2 _ _ HintE) as Hint2.
  simpl in Hint1.
  invert Hint1; clear Hint1; [].
  intros F F' _ _ Hintapp Heq Heq'.
  simpl.
  rewrite <- Heq, <- Heq'; [].
  clear Heq Heq'.
  simp_sub.
  #4 exploit (Hintapp (subst (left E) c2) (subst (right E) c2) (out (simp k1) (intc c2 E)) (out (simp k1) (intc c2 E')) (subst (left E') c2) (subst (right E') c2)) as Hint.
    rewrite -> (cand_eta _ _ (intk_csk _#6 Hint2 andel)); [].
    rewrite -> (cand_eta _ _ (intk_csk _#6 Hint2 ander)); [].
    assumption.

    eapply seq_ofc; eauto using wf_nil; [].
    apply intctx_left; auto; done.

    eapply seq_ofc; eauto using wf_nil; [].
    apply intctx_right; auto; done.

    clear Hintapp.
    assert (subenv E (dot c2 id) (E; Sta (subst (left E) c2) (subst (right E) c2) (Cand (simp k1) (out (simp k1) (intc c2 E))))) as Hsubenv.
      rewrite -> (cand_eta _ _ (intk_csk _#6 Hint2 andel)); [].
      apply (subenv_dot G); eauto using ofc_fcn, intctx_fenv; [].
      apply subenv_sh; auto with sequence; done.
    unfold subst1.
    rewrite -> (subst_intk _#5 (ofk_fkd _ _ (mrk_ofk _ _ Hmrk2)) (fenv_cn (intctx_fenv _#3 HintE)) Hsubenv).
    assumption.

(* pair *)
intros G k1 k2 c1 c2 Hofc1 IH1 Hofc2 IH2 Hofk2 IH3 Hmr.
so (IH1 Hmr) as (Hmrk1 & Hfun1); clear IH1.
so (IH2 Hmr) as (_ & Hfun2); clear IH2.
so (IH3 (mr_cn Hmr Hmrk1)) as Hmrk2; clear IH3.
split.
  apply mrk_sigma_intro; auto; done.

  intros E E' HintE.
  so (Hfun1 _ _ HintE) as Hint1; clear Hfun1.
  so (Hfun2 _ _ HintE) as Hint2.
  simpl.
  remember (intc c1 E) as Q1 eqn:HeqQ1.
  remember (intc c1 E') as Q1' eqn:HeqQ1'.
  remember (intc c2 E) as Q2 eqn:HeqQ2.
  remember (intc c2 E') as Q2' eqn:HeqQ2'.
  destruct Q1 as [sk1 Z1].
  destruct Q1' as [sk1' Z1'].
  destruct Q2 as [sk2 Z2].
  destruct Q2' as [sk2' Z2'].
  so (intk_regular _#8 Hint1) as (_ & _ & -> & ->).
  so (intk_regular _#8 Hint2) as (_ & _ & Heq & Heq').
  unfold subst1 in Heq; rewrite -> (subst_simp _ _ (fkd_hkd _ _ (ofk_fkd _ _ Hofk2))) in Heq.
  unfold subst1 in Heq'; rewrite -> (subst_simp _ _ (fkd_hkd _ _ (ofk_fkd _ _ Hofk2))) in Heq'.
  subst sk2 sk2'.
  assert (ofc G (cn_pair c1 c2) (kd_sigma k1 k2)) as Hofpair by (eauto with static).
  clear HeqQ1' HeqQ2 HeqQ2'.
  assert (eqc nil (cn_pi1 (cn_pair (subst (left E) c1) (subst (left E) c2))) (subst (left E) c1) (subst (left E) k1)) as Heqbeta1.
    #2 apply (@eqc_beta1 _ _ (subst (left E) (subst1 c1 k2))).
      eapply subst_ofc; eauto using left_subof; done.

      eapply subst_ofc; eauto using left_subof; done.
  assert (eqc nil (cn_pi1 (cn_pair (subst (right E) c1) (subst (right E) c2))) (subst (right E) c1) (subst (right E) k1)) as Heqbeta1'.
    #2 apply (@eqc_beta1 _ _ (subst (right E) (subst1 c1 k2))).
      eapply subst_ofc; eauto using right_subof; done.

      eapply subst_ofc; eauto using right_subof; done.
  #4 apply sigma_interp_.
    eapply subst_ofc; eauto using left_subof; done.

    eapply subst_ofc; eauto using right_subof; done.

    simpl.
    #2 eapply intk_defeq_mr; eauto.
      simp_sub.
      apply eqc_symm; auto; done.

      simp_sub.
      apply eqc_symm; auto; done.

    simpl.
    simp_sub.
    assert (intctx (G; cl_cn k1) (E; Sta (subst (left E) c1) (subst (right E) c1) (Cand (simp k1) Z1)) (E; Sta (subst (left E) (cn_pi1 (cn_pair c1 c2))) (subst (right E) (cn_pi1 (cn_pair c1 c2))) (Cand (simp k1) Z1))) as HintEx.
      #4 apply intctx_cn.
        eapply intctx_refl_mr; eauto; done.

        simp_sub; auto with static; done.

        simp_sub; auto with static; done.

        eapply intk_refl_mr; eauto; done.
    so (mrk_mrkat _ _ Hmrk2 _ _ HintEx andel) as H.
    simp_sub in H.
    rewrite <- H; [].
    clear H.
    assert (subenv E (dot c1 id) (E; Sta (subst (left E) c1) (subst (right E) c1) (intc c1 E))) as Hsubenv.
      apply (subenv_dot G); eauto using ofc_fcn, intctx_fenv; [].
      apply subenv_sh; auto with sequence; done.
    unfold subst1 in Hint2; rewrite -> (subst_intk _#5 (ofk_fkd _ _ Hofk2) (fenv_cn (intctx_fenv _#3 HintE)) Hsubenv) in Hint2; [].
    rewrite <- HeqQ1 in Hint2; [].
    #2 eapply (intk_defeq_mr (G; cl_cn k1)); eauto using mr_cn.
      simpl.
      apply eqc_symm; [].
      #2 apply (@eqc_beta2 _ (subst (left E) k1)).
        eapply subst_ofc; eauto using left_subof; done.

        so (subst_ofc _#5 (left_subof _#3 HintE) Hofc2) as H; simp_sub in H.
        assumption.

      simpl.
      apply eqc_symm; [].
      #2 apply (@eqc_beta2 _ (subst (right E) k1)).
        eapply subst_ofc; eauto using right_subof; done.

        so (subst_ofc _#5 (right_subof _#3 HintE) Hofc2) as H; simp_sub in H.
        assumption.

(* pi1 *)
intros G k1 k2 c Hofc IH Hmr.
so (IH Hmr) as (Hmrksig & Hfun); clear IH.
invert Hmrksig; [].
intros _ Hmrk1 Hmrk2.
split; auto; [].
intros E E' HintE.
so (Hfun _ _ HintE) as Hint.
simpl in Hint.
invert Hint; clear Hint; [].
intros P P' _ _ Hint1 Hint2 Heq Heq'.
simpl.
rewrite <- Heq, <- Heq'; [].
assumption.

(* pi2 *)
intros G k1 k2 c Hofc IH Hmr.
so (IH Hmr) as (Hmrksig & Hfun); clear IH.
invert Hmrksig; [].
intros _ Hmrk1 Hmrk2.
split.
  eapply (subst_mrk _ (G; cl_cn k1)); eauto using mr_wf, wf_cn, mrk_ofk; [].
  apply subof_dot; eauto using subof_id with sequence; [].
  simp_sub.
  apply mrcl_cn; eauto with static; [].
  intros E E' HintE.
  so (Hfun _ _ HintE) as Hint.
  simpl in Hint.
  invert Hint; [].
  intros P P' _ _ Hint1 Hint2 Heq Heq'.
  simpl.
  rewrite <- Heq, <- Heq'; [].
  assumption.

  intros E E' HintE.
  simp_sub.
  so (Hfun _ _ HintE) as Hint.
  simpl in Hint.
  invert Hint; clear Hint; [].
  intros P P' _ _ Hint1 Hint2 Heq Heq'.
  simpl.
  rewrite <- Heq, <- Heq'; [].
  assert (subenv E (dot (cn_pi1 c) id) (E; Sta (subst (left E) (cn_pi1 c)) (subst (right E) (cn_pi1 c)) (intc (cn_pi1 c) E))) as Hsubenv.
    apply (subenv_dot G); eauto using subenv_sh, intctx_fenv with sequence; [].
    apply fcn_pi1; eauto using ofc_fcn; done.
  simpl in Hsubenv.
  rewrite <- Heq in Hsubenv; [].
  simp_sub in Hsubenv.
  unfold subst1.
  rewrite -> (subst_intk _#5 (ofk_fkd _ _ (mrk_ofk _ _ Hmrk2)) (fenv_cn (intctx_fenv _#3 HintE)) Hsubenv); [].
  assumption.

(* star *)
intros G H.    
split.
  apply mrk_one_intro; done.

  intros E E' HintE.
  simp_sub.
  simpl.
  apply one_interp_; auto with static; done.

(* unit *)
intros G Hmr.
split.
  apply mrk_type_intro; done.

  intros E E' HintE.
  simpl.
  rewrite <- (unit_interp_fun E E'); [].
  #2 apply type_interp_; try (simp_sub; auto with static; done).
    simp_sub.
    eapply unit_interp_tvrel; done.

    simp_sub.
    rewrite -> unit_versions; [].
    exact unit_rel_tpclosed.

(* arrow *)
intros G t1 t2 Hoft1 IH1 Hoft2 IH2 Hmr.
so (IH1 Hmr ander) as Hfun1; clear IH1.
so (IH2 Hmr ander) as Hfun2; clear IH2.
split; [apply mrk_type_intro; done |].
intros E E' HintE.
simpl.
simp_sub.
#3 rewrite <- (arrow_interp_fun _#9 Hmr (eqc_refl Hoft1) (eqc_refl Hoft2) HintE).
  #4 apply type_interp_.
    eauto 6 using subst_ofc_type, left_subof with static; done.

    eauto 6 using subst_ofc_type, right_subof with static; done.

    eapply arrow_interp_tvrel; eauto using intctx_refl_mr; done.

    rewrite -> arrow_versions; [].
    eapply arrow_rel_tpclosed; eauto using subst_ofc_type, left_subof, right_subof, type_interp_tvrel, type_interp_tpclosed; done.

  clear E E' HintE.
  intros E E' HintE.
  f_equal; eauto using type_interp_eq; done.

  clear E E' HintE.
  intros E E' HintE.
  f_equal; eauto using type_interp_eq; done.

(* prod *)
intros G t1 t2 Hoft1 IH1 Hoft2 IH2 Hmr.
so (IH1 Hmr ander) as Hfun1; clear IH1.
so (IH2 Hmr ander) as Hfun2; clear IH2.
split; [apply mrk_type_intro; done |].
intros E E' HintE.
simpl.
simp_sub.
#3 rewrite <- (prod_interp_fun _#9 Hmr (eqc_refl Hoft1) (eqc_refl Hoft2) HintE).
  #4 apply type_interp_.
    eauto 6 using subst_ofc_type, left_subof with static; done.

    eauto 6 using subst_ofc_type, right_subof with static; done.

    eapply prod_interp_tvrel; eauto using intctx_refl_mr; done.

    rewrite -> prod_versions; [].
    eapply prod_rel_tpclosed; eauto using subst_ofc_type, left_subof, right_subof, type_interp_tvrel, type_interp_tpclosed; done.

  clear E E' HintE.
  intros E E' HintE.
  f_equal; eauto using type_interp_eq; done.

  clear E E' HintE.
  intros E E' HintE.
  f_equal; eauto using type_interp_eq; done.

(* all *)
intros G k t Hofk IH1 Hoft IH2 Hmr.
so (IH1 Hmr) as Hmrk; clear IH1.
so (IH2 (mr_cn Hmr Hmrk)) as (_ & Hfun); clear IH2.
split; auto using mrk_type_intro; [].
intros E E' HintE.
simpl.
simp_sub.
#3 rewrite <- (all_interp_fun G k k t t (fun E => out sk_type (intc t E)) _ E E'); auto with static.
  #4 apply type_interp_.
    so (subst_ofc _#5 (left_subof _#3 HintE) (ofc_all Hofk Hoft)) as H.
    simp_sub in H.
    assumption.

    so (subst_ofc _#5 (right_subof _#3 HintE) (ofc_all Hofk Hoft)) as H.
    simp_sub in H.
    assumption.

    eapply all_interp_tvrel; eauto; done.

    rewrite -> all_versions; [].
    #3 eapply all_rel_tpclosed; eauto using subst_ofk, left_subof, right_subof.
      eapply subst_ofc_type; eauto; [].
      apply subof_bind_sof_cn; auto; [].
      eapply left_subof; eauto; done.

      eapply subst_ofc_type; eauto; [].
      apply subof_bind_sof_cn; auto; [].
      eapply right_subof; eauto; done.

      intros c c' Q Hc.
      so (intk_reg _#6 Hc) as (Hofc & Hofc').
      #2 exploit (Hfun (E; Sta c c' Q) (E; Sta c c' Q)) as H.
        apply intctx_cn; eauto using intctx_refl_mr, eqc_refl; done.
      simpl in H.
      invert H; clear H; [].
      intros R _ _ HofR Hcl Heq _.
      #2 do2 3 split; auto.
        simp_sub.
        rewrite <- Heq; [].
        simpl.
        rewrite -> cast_id; [].
        assumption.

        simp_sub.
        rewrite <- Heq; [].
        simpl.
        rewrite -> cast_id; [].
        assumption.

  clear E E' HintE.
  intros E E' HintE.
  eapply intk_fun_mr; eauto; done.

  clear E E' HintE.
  intros E E' HintE.
  f_equal; eauto using type_interp_eq; done.

(* exists *)
intros G k t Hofk IH1 Hoft IH2 Hmr.
so (IH1 Hmr) as Hmrk; clear IH1.
so (IH2 (mr_cn Hmr Hmrk)) as (_ & Hfunt).
split; auto using mrk_type_intro; [].
intros E E' HintE.
simpl.
simp_sub.
#3 rewrite <- (exists_interp_st_fun G k k t t (fun E => out sk_type (intc t E)) _ E E'); auto with static.
  #4 apply type_interp_.
    so (subst_ofc _#5 (left_subof _#3 HintE) (ofc_exists Hofk Hoft)) as H.
    simp_sub in H.
    assumption.

    so (subst_ofc _#5 (right_subof _#3 HintE) (ofc_exists Hofk Hoft)) as H.
    simp_sub in H.
    assumption.

    intros v v' Hv.
    so Hv as (_ & H).
    simp_sub in H.
    exact H.

    unfold exists_interp_st.
    simp_sub.
    apply tpst_idem; done.

  clear E E' HintE.
  intros E E' HintE.
  eapply intk_fun_mr; eauto; done.

  clear E E' HintE.
  intros E E' HintE.
  f_equal; eauto using type_interp_eq; done.

(* sing *)
intros G c Hofc IH Hmr.
so (IH Hmr) as (_ & Hfun); clear IH.
split.
  apply mrk_sing_intro; auto; [].
  intros E E' HintE.
  exact (type_interp_eq _#5 (Hfun _ _ HintE)).

  intros E E' HintE.
  simpl.
  so (Hfun _ _ HintE) as H.
  simpl in H.
  rewrite <- (type_interp_eq _#5 H); [].
  invert H; clear H; [].
  intros R _ _ _ _ Heq _.
  rewrite <- (cand_eta' _#3 Heq); [].
  invert (Hfun _ _ HintE); [].
  intros R' _ _ HofR Hcl HeqR' _.
  rewrite <- Heq in HeqR'; [].
  so (cand_inject _#3 HeqR'); subst R'; clear HeqR'.
  rewrite <- (cand_beta sk_type R) in HofR, Hcl; [].
  rewrite -> Heq in HofR, Hcl; [].
  #2 apply sing_interp_; auto.
    apply eqc_refl; [].
    eapply subst_ofc_type; eauto using left_subof; done.

    apply eqc_refl; [].
    eapply subst_ofc_type; eauto using right_subof; done.

(* extpi *)
intros G k1 k2 k3 c Hofcalt IH1 Hofapp IH2 Hmr.
so (ofc_extpi Hofcalt Hofapp) as Hofc.
so (IH1 Hmr) as (Hmrkpi & Hfun); clear IH1.
invert Hmrkpi; intros _ Hmrk1 _.
so (IH2 (mr_cn Hmr Hmrk1)) as (Hmrk2 & Hfunapp); clear IH2.
split.
  apply mrk_pi_intro; auto; done.

  intros E E' HintE.
  simpl.
  so (Hfun _ _ HintE) as Hintpi.
  simpl in Hintpi.
  invert Hintpi; clear Hintpi; [].
  repl (simp k3) with (simp k2).
    so (ofc_skof _#3 (mr_wf _ Hmr) Hofcalt) as Heq.
    so (ofc_skof _#3 (mr_wf _ Hmr) Hofc) as Heq'.
    rewrite -> Heq' in Heq; clear Heq'.
    simpl in Heq.
    injection Heq; auto; done.
  intros F F' _ _ _ Heq Heq'.
  rewrite <- Heq, <- Heq'; [].
  #3 apply pi_interp_.
    eapply subst_ofc; eauto using left_subof; done.

    eapply subst_ofc; eauto using right_subof; done.

    intros d d' Q Q' d'' d''' Hint1 Heqd Heqd'.
    #2 exploit (Hfunapp (E; Sta d d' (Cand (simp k1) Q)) (E'; Sta d'' d''' (Cand (simp k1) Q'))) as Hint2; clear Hfunapp.
      apply intctx_cn; eauto; done.

      simpl left in Hint2.
      simpl right in Hint2.
      simp_sub in Hint2.
      force_exact Hint2; clear Hint2.
      #2 f_equal.
        simpl.
        rewrite -> (weaken1_intc G _ _ c); eauto using ofc_fcn, intctx_fenv; [].
        rewrite <- Heq; [].
        simp_sub.
        rewrite -> cast_id; [].
        reflexivity.

        simpl.
        rewrite -> (weaken1_intc G _ _ c); eauto using ofc_fcn, intctx_fenv'; [].
        rewrite <- Heq'; [].
        simp_sub.
        rewrite -> cast_id; [].
        reflexivity.

(* extsigma *)
intros G k1 k2 c Hofc1 IH1 Hofc2 IH2 Hofk2 IH3 Hmr.       
so (IH1 Hmr) as (Hmrk1 & Hfun1); clear IH1.
so (IH2 Hmr) as (_ & Hfun2); clear IH2.
so (IH3 (mr_cn Hmr Hmrk1)) as Hmrk2; clear IH3.
so (ofc_extsigma Hofc1 Hofc2 Hofk2) as Hofc.
split.
  apply mrk_sigma_intro; auto; done.

  intros E E' HintE.
  simpl.
  so (ofc_skof _#3 (mr_wf _ Hmr) Hofc) as Hskofeq.
  simpl in Hskofeq.
  remember (intc c E) as Q eqn:Heq.
  destruct Q as [sk P].
  so (intc_skind _#4 (eqsymm Heq)) as Heqsk.
  rewrite <- (intctx_flatten _#3 HintE andel) in Heqsk; [].
  rewrite -> Hskofeq in Heqsk; [].
  subst sk.
  remember (intc c E') as Q eqn:Heq'.
  destruct Q as [sk P'].
  so (intc_skind _#4 (eqsymm Heq')) as Heqsk.
  rewrite <- (intctx_flatten _#3 HintE ander) in Heqsk; [].
  rewrite -> Hskofeq in Heqsk; [].
  subst sk.
  clear Hskofeq.
  #4 apply sigma_interp_.
    eapply subst_ofc; eauto using left_subof; done.

    eapply subst_ofc; eauto using right_subof; done.

    so (Hfun1 _ _ HintE) as Hint.
    simp_sub in Hint.
    simpl in Hint.
    rewrite <- Heq, <- Heq' in Hint; [].
    assumption.

    so (Hfun2 _ _ HintE) as Hint.
    simp_sub in Hint.
    simpl in Hint.
    rewrite <- Heq, <- Heq' in Hint; [].
    assert (subenv E (dot (cn_pi1 c) id) (E; Sta (subst (left E) (cn_pi1 c)) (subst (right E) (cn_pi1 c)) (intc (cn_pi1 c) E))) as Hsubenv.
      eapply subenv_dot; eauto using subenv_sh, ofc_fcn, intctx_fenv with sequence; done.
    unfold subst1 in Hint; rewrite -> (subst_intk _#5 (ofk_fkd _ _ Hofk2) (fenv_cn (intctx_fenv _#3 HintE)) Hsubenv) in Hint; [].
    simp_sub in Hint.
    simpl in Hint.
    rewrite <- Heq in Hint; [].
    assumption.

(* subsume *)
intros G k k' c Hofc IH1 Hsubk IH2 Hmr.
so (IH1 Hmr) as (Hmrk & Hfun); clear IH1.
so (IH2 Hmr) as (_ & Hmrk' & Hincl); clear IH2.
split; auto; [].
intros E E' HintE.
apply Hincl; eauto using intctx_refl_mr; done.

(* refl *)
intros G c k Hofc IH Hmr.
so (IH Hmr) as (Hmrk & Hfun); clear IH.
split; auto; done.

(* symm *)
intros G c1 c2 k Heqc IH Hmr.
so (IH Hmr) as (Hmrk & Hfun); clear IH.
split; auto; [].
intros E E' HintE.
eapply intk_symm_mr; eauto; [].
so (Hfun _ _ (intctx_symm_mr _#3 Hmr HintE)) as Hint.
rewrite -> (intk_fun_mr _#4 Hmrk HintE); [].
#3 eapply intk_defeq_mr_r; eauto.
  eauto using intctx_refl_mr, intctx_symm_mr; done.

  #3 eapply seq_eqc; eauto.
    exact wf_nil.

    apply mr_wf; auto; done.

    apply intctx_left; auto using intctx_symm_mr; done.

  #3 eapply seq_eqc; eauto.
    exact wf_nil.

    apply mr_wf; auto; done.

    apply intctx_right; auto using intctx_symm_mr; done.

(* trans *)
intros G c1 c2 c3 k Heqc12 IH1 Heqc23 IH2 Hmr.
so (IH1 Hmr) as (Hmrk & Hfun12); clear IH1.
so (IH2 Hmr) as (_ & Hfun23); clear IH2.
split; auto; [].
intros E E' HintE.
so (Hfun12 _ _ (intctx_refl_mr _#3 Hmr HintE)) as Hint12.
so (Hfun23 _ _ HintE) as Hint23.
eapply intk_trans_mr; eauto; [].
#2 eapply intk_defeq_mr; eauto.
  apply eqc_symm; [].
  eapply subst_eqc; eauto using left_subof; done.

  apply eqc_symm; [].
  eapply subst_eqc; eauto using right_subof; done.

(* lam *)
intros G k1 k1' k2 c c' Heqk1 IH1 Heqc IH2 Hmr.
so (IH1 Hmr) as (Hmrk1 & Hmrk1' & Hfun1); clear IH1.
so (IH2 (mr_cn Hmr Hmrk1)) as (Hmrk2 & Hfun2); clear IH2.
so (mrk_ofk _ _ Hmrk1) as Hofk1.
so (mrk_ofk _ _ Hmrk1') as Hofk1'.
split.
  apply mrk_pi_intro; auto; done.

  intros E E' HintE.
  simpl.
  fold precand.
  rewrite <- (intctx_flatten _#3 HintE andel); [].
  rewrite <- (intctx_flatten _#3 HintE ander); [].
  repl (simp k1') with (simp k1).
    eapply eqk_simp; eauto; done.
  replace (flatten_ctx G; Some (simp k1)) with (flatten_ctx (G; cl_cn k1)) by reflexivity.
  so (eqc_reg _#4 (wf_cn (mr_wf _ Hmr) Hofk1) Heqc) as (Hofc & Hofc' & _).
  rewrite -> (ofc_skof _#3 (wf_cn (mr_wf _ Hmr) Hofk1) Hofc); [].
  rewrite -> (ofc_skof _#3 (wf_cn (mr_wf _ Hmr) Hofk1) Hofc'); [].
  assert (ofc G (cn_lam k1 c) (kd_pi k1 k2)) as Hofclam by (auto with static).
  #3 apply pi_interp_.
    eapply subst_ofc; eauto using left_subof; done.

    eapply subst_ofc; eauto using right_subof; done.

    intros d d' Q Q' d'' d''' Hint1 Heqd Heqd'.
    assert (intctx (G; cl_cn k1) (E; Sta d d' (Cand (simp k1) Q)) (E'; Sta d'' d''' (Cand (simp k1) Q'))) as HintEx by (apply intctx_cn; auto).
    so (Hfun2 _ _ HintEx) as Hint2.
    simpl in Hint2.
    rewrite -> (cand_eta _ _ (intk_csk _#6 Hint2 andel)); [].
    rewrite -> (cand_eta _ _ (intk_csk _#6 Hint2 ander)); [].
    #2 eapply (intk_defeq_mr (G; cl_cn k1)); eauto using mr.
      simpl.
      simp_sub.
      apply eqc_symm; [].
      replsub (subst (dot d (left E)) c) with (subst1 d (subst (dot (var 0) (compose (left E) sh1)) c)).
      replsub (subst (dot d (left E)) k2) with (subst1 d (subst (dot (var 0) (compose (left E) sh1)) k2)).
      #2 apply eqc_beta.
        eapply subst_ofc; eauto; [].
        apply subof_bind_sof_cn; eauto using left_subof; done.

        exact (eqc_reg _#4 wf_nil Heqd andel).

      simpl.
      simp_sub.
      apply eqc_symm; [].
      replsub (subst (dot d' (right E)) c) with (subst1 d' (subst (dot (var 0) (compose (right E) sh1)) c)).
      replsub (subst (dot d' (right E)) k2) with (subst1 d' (subst (dot (var 0) (compose (right E) sh1)) k2)).
      #2 apply eqc_beta.
        eapply subst_ofc; eauto; [].
        apply subof_bind_sof_cn; eauto using right_subof; done.

        exact (eqc_reg _#4 wf_nil Heqd' andel).

(* app *)
intros G k1 k2 c1 c2 c1' c2' Heqc1 IH1 Heqc2 IH2 Hmr.
so (IH1 Hmr) as (Hmrkpi & Hfun1); clear IH1.
so (IH2 Hmr) as (Hmrk1 & Hfun2); clear IH2.
invert Hmrkpi; [].
intros _ _ Hmrk2.
so (eqc_reg _#4 (mr_wf _ Hmr) Heqc2) as (Hofc2 & Hofc2' & _).
split.  
  #2 eapply subst_mrk; eauto.
    auto using wf_cn, mr_wf, mrk_ofk; done.

    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply mrcl_cn; auto; [].
    intros E E' HintE.
    eapply intk_fst_fun; eauto; done.

  intros E E' HintE.
  so (Hfun1 _ _ HintE) as Hint1.
  so (Hfun2 _ _ HintE) as Hint2.
  simpl in Hint1.
  invert Hint1; clear Hint1; [].
  intros F F' _ _ Hintapp Heq Heq'.
  simpl.
  rewrite <- Heq, <- Heq'; [].
  clear Heq Heq'.
  simp_sub.
  #4 exploit (Hintapp (subst (left E) c2) (subst (right E) c2) (out (simp k1) (intc c2 E)) (out (simp k1) (intc c2' E')) (subst (left E') c2') (subst (right E') c2')) as Hint.
    rewrite -> (cand_eta _ _ (intk_csk _#6 Hint2 andel)); [].
    rewrite -> (cand_eta _ _ (intk_csk _#6 Hint2 ander)); [].
    assumption.

    eapply (seq_eqc nil _ _ G); eauto using wf_nil, mr_wf; [].
    apply intctx_left; auto; done.

    eapply (seq_eqc nil _ _ G); eauto using wf_nil, mr_wf; [].
    apply intctx_right; auto; done.

    clear Hintapp.
    assert (subenv E (dot c2 id) (E; Sta (subst (left E) c2) (subst (right E) c2) (Cand (simp k1) (out (simp k1) (intc c2 E))))) as Hsubenv.
      rewrite -> (cand_eta _ _ (intk_csk _#6 Hint2 andel)); [].
      apply (subenv_dot G); eauto using ofc_fcn, intctx_fenv; [].
      apply subenv_sh; auto with sequence; done.
    unfold subst1.
    rewrite -> (subst_intk _#5 (ofk_fkd _ _ (mrk_ofk _ _ Hmrk2)) (fenv_cn (intctx_fenv _#3 HintE)) Hsubenv); [].
    assumption.

(* pair *)
intros G k1 k2 c1 c1' c2 c2' Heqc1 IH1 Heqc2 IH2 Hofk2 IH3 Hmr.
so (IH1 Hmr) as (Hmrk1 & Hfun1); clear IH1.
so (IH2 Hmr) as (_ & Hfun2); clear IH2.
so (IH3 (mr_cn Hmr Hmrk1)) as Hmrk2; clear IH3.
split.
  apply mrk_sigma_intro; auto; done.

  intros E E' HintE.
  so (Hfun1 _ _ HintE) as Hint1; clear Hfun1.
  so (Hfun2 _ _ HintE) as Hint2.
  simpl.
  remember (intc c1 E) as Q1 eqn:HeqQ1.
  remember (intc c1' E') as Q1' eqn:HeqQ1'.
  remember (intc c2 E) as Q2 eqn:HeqQ2.
  remember (intc c2' E') as Q2' eqn:HeqQ2'.
  destruct Q1 as [sk1 Z1].
  destruct Q1' as [sk1' Z1'].
  destruct Q2 as [sk2 Z2].
  destruct Q2' as [sk2' Z2'].
  so (intk_regular _#8 Hint1) as (_ & _ & -> & ->).
  so (intk_regular _#8 Hint2) as (_ & _ & Heq & Heq').
  unfold subst1 in Heq; rewrite -> (subst_simp _ _ (fkd_hkd _ _ (ofk_fkd _ _ Hofk2))) in Heq.
  unfold subst1 in Heq'; rewrite -> (subst_simp _ _ (fkd_hkd _ _ (ofk_fkd _ _ Hofk2))) in Heq'.
  subst sk2 sk2'.
  so (eqc_reg _#4 (mr_wf _ Hmr) Heqc1) as (Hofc1 & _).
  so (eqc_reg _#4 (mr_wf _ Hmr) Heqc2) as (Hofc2 & _).
  assert (ofc G (cn_pair c1 c2) (kd_sigma k1 k2)) as Hofpair by (eauto with static).
  clear HeqQ1' HeqQ2 HeqQ2'.
  assert (eqc nil (cn_pi1 (cn_pair (subst (left E) c1) (subst (left E) c2))) (subst (left E) c1) (subst (left E) k1)) as Heqbeta1.
    #2 apply (@eqc_beta1 _ _ (subst (left E) (subst1 c1 k2))).
      eapply subst_ofc; eauto using left_subof; done.

      eapply subst_ofc; eauto using left_subof; done.
  assert (eqc nil (cn_pi1 (cn_pair (subst (right E) c1) (subst (right E) c2))) (subst (right E) c1) (subst (right E) k1)) as Heqbeta1'.
    #2 apply (@eqc_beta1 _ _ (subst (right E) (subst1 c1 k2))).
      eapply subst_ofc; eauto using right_subof; done.

      eapply subst_ofc; eauto using right_subof; done.
  #4 apply sigma_interp_.
    eapply subst_ofc; eauto using left_subof; done.

    eapply subst_ofc; eauto using right_subof; done.

    simpl.
    #2 eapply intk_defeq_mr; eauto.
      simp_sub.
      apply eqc_symm; auto; done.

      simp_sub.
      apply eqc_symm; auto; done.

    simpl.
    simp_sub.
    assert (intctx (G; cl_cn k1) (E; Sta (subst (left E) c1) (subst (right E) c1) (Cand (simp k1) Z1)) (E; Sta (subst (left E) (cn_pi1 (cn_pair c1 c2))) (subst (right E) (cn_pi1 (cn_pair c1 c2))) (Cand (simp k1) Z1))) as HintEx.
      #4 apply intctx_cn.
        eapply intctx_refl_mr; eauto; done.

        simp_sub; auto with static; done.

        simp_sub; auto with static; done.

        eapply intk_refl_mr; eauto; done.
    so (mrk_mrkat _ _ Hmrk2 _ _ HintEx andel) as H.
    simp_sub in H.
    rewrite <- H; [].
    clear H.
    assert (subenv E (dot c1 id) (E; Sta (subst (left E) c1) (subst (right E) c1) (intc c1 E))) as Hsubenv.
      apply (subenv_dot G); eauto using ofc_fcn, intctx_fenv; [].
      apply subenv_sh; auto with sequence; done.
    unfold subst1 in Hint2; rewrite -> (subst_intk _#5 (ofk_fkd _ _ Hofk2) (fenv_cn (intctx_fenv _#3 HintE)) Hsubenv) in Hint2; [].
    rewrite <- HeqQ1 in Hint2; [].
    #2 eapply (intk_defeq_mr (G; cl_cn k1)); eauto using mr_cn.
      simpl.
      apply eqc_symm; [].
      #2 apply (@eqc_beta2 _ (subst (left E) k1)).
        eapply subst_ofc; eauto using left_subof; done.

        so (subst_ofc _#5 (left_subof _#3 HintE) Hofc2) as H; simp_sub in H.
        assumption.

      simpl.
      apply eqc_symm; [].
      #2 apply (@eqc_beta2 _ (subst (right E) k1)).
        eapply subst_ofc; eauto using right_subof; done.

        so (subst_ofc _#5 (right_subof _#3 HintE) Hofc2) as H; simp_sub in H.
        assumption.

(* pi1 *)
intros G k1 k2 c c' Heqc IH Hmr.
so (IH Hmr) as (Hmrksigma & Hfun); clear IH.
invert Hmrksigma; [].
intros _ Hmrk1 Hmrk2.
split; auto; [].
intros E E' HintE.
so (Hfun _ _ HintE) as Hint.
simpl in Hint.
invert Hint; clear Hint; [].
intros P P' _ _ Hint1 Hint2 Heq Heq'.
simpl.
rewrite <- Heq, <- Heq'; [].
assumption.

(* pi2 *)
intros G k1 k2 c c' Heqc IH Hmr.
so (IH Hmr) as (Hmrksigma & Hfun); clear IH.
invert Hmrksigma; [].
intros _ Hmrk1 Hmrk2.
so (eqc_reg _#4 (mr_wf _ Hmr) Heqc) as (Hofc & _).
split.
  eapply (subst_mrk _ (G; cl_cn k1)); eauto using mr_wf, wf_cn, mrk_ofk; [].
  apply subof_dot; eauto using subof_id with sequence; [].
  simp_sub.
  apply mrcl_cn; eauto with static; [].
  intros E E' HintE.
  so (intk_fst_fun _#4 Hmr Hmrksigma Heqc Hfun _ _ HintE) as Hint.
  simpl in Hint.
  invert Hint; [].
  intros P P' _ _ Hint1 Hint2 Heq Heq'.
  simpl.
  rewrite <- Heq, <- Heq'; [].
  assumption.

  intros E E' HintE.
  simp_sub.
  so (Hfun _ _ HintE) as Hint.
  simpl in Hint.
  invert Hint; clear Hint; [].
  intros P P' _ _ Hint1 Hint2 Heq Heq'.
  simpl.
  rewrite <- Heq, <- Heq'; [].
  assert (subenv E (dot (cn_pi1 c) id) (E; Sta (subst (left E) (cn_pi1 c)) (subst (right E) (cn_pi1 c)) (intc (cn_pi1 c) E))) as Hsubenv.
    apply (subenv_dot G); eauto using subenv_sh, intctx_fenv with sequence; [].
    apply fcn_pi1; eauto using ofc_fcn; done.
  simpl in Hsubenv.
  rewrite <- Heq in Hsubenv; [].
  simp_sub in Hsubenv.
  unfold subst1.
  rewrite -> (subst_intk _#5 (ofk_fkd _ _ (mrk_ofk _ _ Hmrk2)) (fenv_cn (intctx_fenv _#3 HintE)) Hsubenv); [].
  assumption.

(* arrow *)
intros G t1 t1' t2 t2' Heqt1 IH1 Heqt2 IH2 Hmr.
so (IH1 Hmr) as (_ & Hfun1); clear IH1.
so (IH2 Hmr) as (_ & Hfun2); clear IH2.
split; [apply mrk_type_intro; done |].
intros E E' HintE.
simpl.
simp_sub.
so (eqc_reg _#4 (mr_wf _ Hmr) Heqt1) as (Hoft1 & Hoft1' & _).
so (eqc_reg _#4 (mr_wf _ Hmr) Heqt2) as (Hoft2 & Hoft2' & _).
#3 rewrite <- (arrow_interp_fun _#9 Hmr Heqt1 Heqt2 HintE).
  rewrite -> !arrow_versions; [].
  rewrite <- (type_interp_eq _#5 (Hfun1 _ _ (intctx_refl_mr _#3 Hmr HintE))); [].
  rewrite <- (type_interp_eq _#5 (Hfun2 _ _ (intctx_refl_mr _#3 Hmr HintE))); [].
  rewrite <- !(arrow_versions _ _ (fun E => out sk_type (intc t1 E)) (fun E => out sk_type (intc t2 E))); [].
  #4 apply type_interp_.
    eauto 6 using subst_ofc_type, left_subof with static; done.

    eauto 6 using subst_ofc_type, right_subof with static; done.

    eapply arrow_interp_tvrel; eauto using intctx_refl_mr; done.

    rewrite -> arrow_versions; [].
    eapply arrow_rel_tpclosed; eauto using subst_ofc_type, left_subof, right_subof, type_interp_tvrel, type_interp_tpclosed; done.

  clear E E' HintE.
  intros E E' HintE.
  so (intk_snd_fun _#4 Hmr (mrk_type_intro _) Heqt1 Hfun1 _ _ HintE).
  f_equal; eauto using type_interp_eq; done.

  clear E E' HintE.
  intros E E' HintE.
  so (intk_snd_fun _#4 Hmr (mrk_type_intro _) Heqt2 Hfun2 _ _ HintE).
  f_equal; eauto using type_interp_eq; done.

(* prod *)
intros G t1 t1' t2 t2' Heqt1 IH1 Heqt2 IH2 Hmr.
so (IH1 Hmr) as (_ & Hfun1); clear IH1.
so (IH2 Hmr) as (_ & Hfun2); clear IH2.
split; [apply mrk_type_intro; done |].
intros E E' HintE.
simpl.
simp_sub.
so (eqc_reg _#4 (mr_wf _ Hmr) Heqt1) as (Hoft1 & Hoft1' & _).
so (eqc_reg _#4 (mr_wf _ Hmr) Heqt2) as (Hoft2 & Hoft2' & _).
#3 rewrite <- (prod_interp_fun _#9 Hmr Heqt1 Heqt2 HintE).
  rewrite -> !prod_versions; [].
  rewrite <- (type_interp_eq _#5 (Hfun1 _ _ (intctx_refl_mr _#3 Hmr HintE))); [].
  rewrite <- (type_interp_eq _#5 (Hfun2 _ _ (intctx_refl_mr _#3 Hmr HintE))); [].
  rewrite <- !(prod_versions _ _ (fun E => out sk_type (intc t1 E)) (fun E => out sk_type (intc t2 E))); [].
  #4 apply type_interp_.
    eauto 6 using subst_ofc_type, left_subof with static; done.

    eauto 6 using subst_ofc_type, right_subof with static; done.

    eapply prod_interp_tvrel; eauto using intctx_refl_mr; done.

    rewrite -> prod_versions; [].
    eapply prod_rel_tpclosed; eauto using subst_ofc_type, left_subof, right_subof, type_interp_tvrel, type_interp_tpclosed; done.

  clear E E' HintE.
  intros E E' HintE.
  so (intk_snd_fun _#4 Hmr (mrk_type_intro _) Heqt1 Hfun1 _ _ HintE).
  f_equal; eauto using type_interp_eq; done.

  clear E E' HintE.
  intros E E' HintE.
  so (intk_snd_fun _#4 Hmr (mrk_type_intro _) Heqt2 Hfun2 _ _ HintE).
  f_equal; eauto using type_interp_eq; done.

(* all *)
intros G k k' t t' Heqk IH1 Heqt IH2 Hmr.
so (IH1 Hmr) as (Hmrk & Hmrk' & Hfunk); clear IH1.
so (IH2 (mr_cn Hmr Hmrk)) as (_ & Hfunt); clear IH2.
so (eqk_reg _#3 (mr_wf _ Hmr) Heqk) as (Hofk & Hofk').
so (eqc_reg _#4 (wf_cn (mr_wf _ Hmr) Hofk) Heqt) as (Hoft & Hoft' & _).
split; auto using mrk_type_intro; [].
intros E E' HintE.
simpl.
simp_sub.
#3 rewrite <- (all_interp_fun G k k' t t' (fun E => out sk_type (intc t E)) _ E E'); auto with static.
  #4 apply type_interp_.
    so (subst_ofc _#5 (left_subof _#3 HintE) (ofc_all Hofk Hoft)) as H.
    simp_sub in H.
    assumption.

    so (subst_ofc _#5 (right_subof _#3 HintE) (ofc_all Hofk Hoft)) as H.
    simp_sub in H.
    assumption.

    eapply all_interp_tvrel; eauto; done.

    rewrite -> all_versions; [].
    #3 eapply all_rel_tpclosed; eauto using subst_ofk, left_subof, right_subof.
      eapply subst_ofc_type; eauto; [].
      apply subof_bind_sof_cn; auto; [].
      eapply left_subof; eauto; done.

      eapply subst_ofc_type; eauto; [].
      apply subof_bind_sof_cn; auto; [].
      eapply right_subof; eauto; done.

      intros c c' Q Hc.
      so (intk_reg _#6 Hc) as (Hofc & Hofc').
      #2 exploit (Hfunt (E; Sta c c' Q) (E; Sta c c' Q)) as H.
        apply intctx_cn; eauto using intctx_refl_mr, eqc_refl; done.
      simpl in H.
      invert H; clear H; [].
      intros R _ _ HofR Hcl Heq _.
      #2 do2 3 split; auto.
        simp_sub.
        rewrite <- Heq; [].
        simpl.
        rewrite -> cast_id; [].
        assumption.

        simp_sub.
        rewrite <- Heq; [].
        simpl.
        rewrite -> cast_id; [].
        assumption.

  clear E E' HintE.
  intros E E' HintE.
  transitivity (intk k E').
    eapply intk_fun_mr; eauto; done.

    apply Hfunk; eauto using intctx_refl_mr, intctx_symm_mr; done.

  clear E E' HintE.
  intros E E' HintE.
  f_equal; [].
  so (Hfunt _ _ HintE) as H.
  simpl in H.
  exact (type_interp_eq _#5 H).

(* exists *)
intros G k k' t t' Heqk IH1 Heqt IH2 Hmr.
so (IH1 Hmr) as (Hmrk & Hmrk' & Hfunk); clear IH1.
so (IH2 (mr_cn Hmr Hmrk)) as (_ & Hfunt); clear IH2.
so (eqk_reg _#3 (mr_wf _ Hmr) Heqk) as (Hofk & Hofk').
so (eqc_reg _#4 (wf_cn (mr_wf _ Hmr) Hofk) Heqt) as (Hoft & Hoft' & _).
split; auto using mrk_type_intro; [].
intros E E' HintE.
simpl.
simp_sub.
#3 rewrite <- (exists_interp_st_fun G k k' t t' (fun E => out sk_type (intc t E)) _ E E'); auto with static.
  #4 apply type_interp_.
    so (subst_ofc _#5 (left_subof _#3 HintE) (ofc_exists Hofk Hoft)) as H.
    simp_sub in H.
    assumption.

    so (subst_ofc _#5 (right_subof _#3 HintE) (ofc_exists Hofk Hoft)) as H.
    simp_sub in H.
    assumption.

    intros v v' Hv.
    so Hv as (_ & H).
    simp_sub in H.
    exact H.

    unfold exists_interp_st.
    simp_sub.
    apply tpst_idem; done.

  clear E E' HintE.
  intros E E' HintE.
  transitivity (intk k E').
    eapply intk_fun_mr; eauto; done.

    apply Hfunk; eauto using intctx_refl_mr, intctx_symm_mr; done.

  clear E E' HintE.
  intros E E' HintE.
  f_equal; [].
  so (Hfunt _ _ HintE) as H.
  simpl in H.
  exact (type_interp_eq _#5 H).

(* sing *)
intros G c c' Heqc IH Hmr.
so (IH Hmr) as (_ & Hfun); clear IH.
so (eqc_reg _#4 (mr_wf _ Hmr) Heqc) as (Hofc & _).
split.
  apply mrk_sing_intro; auto; [].
  intros E E' HintE.
  so (intk_fst_fun _#4 Hmr (mrk_type_intro _) Heqc Hfun _ _ HintE).
  eapply type_interp_eq; eauto; done.

  intros E E' HintE.
  so (Hfun _ _ HintE) as Hint.
  simpl in Hint.
  invert Hint; clear Hint; [].
  intros R _ _ HofR Hcl Heq <-.
  rewrite <- Heq; [].
  simpl.
  rewrite <- (cand_beta sk_type R); [].
  rewrite -> Heq; [].
  rewrite <- (cand_beta sk_type R) in HofR, Hcl; [].
  rewrite -> Heq in HofR, Hcl; [].
  apply sing_interp_; eauto using eqc_refl, subst_ofc_type, left_subof, right_subof; done.

(* singelim *)
intros G c c' Hofc IH Hmr.
so (IH Hmr) as (Hmrk & Hfun); clear IH.
split.
  apply mrk_type_intro; done.

  intros E E' HintE.
  so (Hfun _ _ HintE) as Hint.
  rewrite -> (intk_fun_mr _#4 Hmrk HintE) in Hint; [].
  invert Hint; [].
  intros _ _ _ _ Heq' Heq.
  rewrite <- Heq in Hint; [].
  invert (ofc_reg _#3 (mr_wf _ Hmr) Hofc); [].
  intros Hofc'.
  so (ofc_intc_csk _#4 (mr_wf _ Hmr) Hofc' (intctx_refl_mr _#3 Hmr (intctx_symm_mr _#3 Hmr HintE))) as Hcsk.
  simpl in Hcsk.
  rewrite -> (cand_eta _ _ Hcsk) in Hint; [].
  clear Hcsk.
  rewrite <- (sing_interp_eq _#6 Hint); [].
  simpl.
  so (ofc_intc_csk _#4 (mr_wf _ Hmr) Hofc (intctx_refl_mr _#3 Hmr HintE)) as Hcsk.
  simpl in Hcsk.
  rewrite <- (cand_eta _ _ Hcsk); [].
  clear Hcsk.
  assert (ofc G c kd_type) as Hofct.
    eapply ofc_subsume; eauto with static; done.
  #4 apply type_interp_.
    eapply subst_ofc_type; eauto using left_subof; done.

    eapply subst_ofc_type; eauto using right_subof; done.
 
    eapply sing_interp_tvrel; eauto; done.

    eapply sing_interp_tpclosed; eauto; done.
    
(* extpi *)
intros G k1 k2 k3 k3' c c' Hofcalt IH1 Hofcalt' IH2 Heqcapp IH3 Hmr.
so (eqc_extpi Hofcalt Hofcalt' Heqcapp) as Heqc.
so (eqc_reg _#4 (mr_wf _ Hmr) Heqc) as (Hofc & Hofc' & _).
so (IH1 Hmr) as (Hmrkalt & Hfunalt); clear IH1.
so (IH2 Hmr) as (Hmrkalt' & Hfunalt'); clear IH2.
invert Hmrkalt; [].
intros _ Hmrk1 _.
so (IH3 (mr_cn Hmr Hmrk1)) as (Hmrk2 & Hfun); clear IH3.
split; auto using mrk_pi_intro; [].
intros E E' HintE.
so (Hfunalt _ _ HintE) as Hintalt; clear Hfunalt.
simpl in Hintalt.
invert Hintalt; clear Hintalt; [].
repl (simp k3) with (simp k2).
  so (ofc_skof _#3 (mr_wf _ Hmr) Hofcalt) as Heq.
  so (ofc_skof _#3 (mr_wf _ Hmr) Hofc) as Heq'.
  rewrite -> Heq' in Heq; clear Heq'.
  simpl in Heq.
  injection Heq; auto; done.
intros F _ _ _ _ HeqF _.
so (Hfunalt' _ _ HintE) as Hintalt'; clear Hfunalt'.
simpl in Hintalt'.
invert Hintalt'; clear Hintalt'; [].
repl (simp k3') with (simp k2).
  so (ofc_skof _#3 (mr_wf _ Hmr) Hofcalt') as Heq.
  so (ofc_skof _#3 (mr_wf _ Hmr) Hofc') as Heq'.
  rewrite -> Heq' in Heq; clear Heq'.
  simpl in Heq.
  injection Heq; auto; done.
intros _ F' _ _ _ _ HeqF'.
rewrite <- HeqF, <- HeqF'; [].
simpl.
#3 apply pi_interp_.
  eapply subst_ofc; eauto using left_subof; done.

  eapply subst_ofc; eauto using right_subof; done.

  intros d d' Q Q' d'' d''' Hint1 Heqd Heqd'.
  #2 exploit (Hfun (E; Sta d d' (Cand (simp k1) Q)) (E'; Sta d'' d''' (Cand (simp k1) Q'))) as Hint2.
    apply intctx_cn; auto; done.
  clear Hfun.
  simpl left in Hint2.
  simpl right in Hint2.
  simp_sub in Hint2.
  simpl in Hint2.
  simp_sub in Hint2.
  rewrite -> (subst_intc _#5 (ofc_fcn _#3 Hofc) (intctx_fenv _#3 HintE) (subenv_sh (truncate_S truncate_0))) in Hint2; [].
  rewrite <- HeqF in Hint2; [].
  rewrite -> (subst_intc _#5 (ofc_fcn _#3 Hofc') (intctx_fenv' _#3 HintE) (subenv_sh (truncate_S truncate_0))) in Hint2; [].
  rewrite <- HeqF' in Hint2; [].
  rewrite -> !cast_id in Hint2; [].
  assumption.

(* extpiw *)
intros G k1 k2 k3 c c' Heqcalt IH1 Heqcapp IH2 Hmr.
so (eqc_extpiw Heqcalt Heqcapp) as Heqc.
so (eqc_reg _#4 (mr_wf _ Hmr) Heqc) as (Hofc & Hofc' & _).
so (IH1 Hmr) as (Hmrkalt & Hfunalt); clear IH1.
invert Hmrkalt; [].
intros _ Hmrk1 _.
so (IH2 (mr_cn Hmr Hmrk1)) as (Hmrk2 & Hfun); clear IH2.
split; auto using mrk_pi_intro; [].
intros E E' HintE.
so (Hfunalt _ _ HintE) as Hintalt; clear Hfunalt.
simpl in Hintalt.
invert Hintalt; clear Hintalt; [].
repl (simp k3) with (simp k2).
  so (ofc_skof _#3 (mr_wf _ Hmr) (eqc_reg _#4 (mr_wf _ Hmr) Heqcalt andel)) as Heq.
  so (ofc_skof _#3 (mr_wf _ Hmr) Hofc) as Heq'.
  rewrite -> Heq' in Heq; clear Heq'.
  simpl in Heq.
  injection Heq; auto; done.
intros F F' _ _ _ HeqF HeqF'.
rewrite <- HeqF, <- HeqF'; [].
simpl.
#3 apply pi_interp_.
  eapply subst_ofc; eauto using left_subof; done.

  eapply subst_ofc; eauto using right_subof; done.

  intros d d' Q Q' d'' d''' Hint1 Heqd Heqd'.
  #2 exploit (Hfun (E; Sta d d' (Cand (simp k1) Q)) (E'; Sta d'' d''' (Cand (simp k1) Q'))) as Hint2.
    apply intctx_cn; auto; done.
  clear Hfun.
  simpl left in Hint2.
  simpl right in Hint2.
  simp_sub in Hint2.
  simpl in Hint2.
  simp_sub in Hint2.
  rewrite -> (subst_intc _#5 (ofc_fcn _#3 Hofc) (intctx_fenv _#3 HintE) (subenv_sh (truncate_S truncate_0))) in Hint2; [].
  rewrite <- HeqF in Hint2; [].
  rewrite -> (subst_intc _#5 (ofc_fcn _#3 Hofc') (intctx_fenv' _#3 HintE) (subenv_sh (truncate_S truncate_0))) in Hint2; [].
  rewrite <- HeqF' in Hint2; [].
  rewrite -> !cast_id in Hint2; [].
  assumption.

(* extsigma *)
intros G k1 k2 c c' Heqc1 IH1 Heqc2 IH2 Hofk2 IH3 Hmr.
so (eqc_extsigma Heqc1 Heqc2 Hofk2) as Heqc.
so (eqc_reg _#4 (mr_wf _ Hmr) Heqc) as (Hofc & Hofc' & _).
so (IH1 Hmr) as (Hmrk1 & Hfun1); clear IH1.
so (IH2 Hmr) as (_ & Hfun2); clear IH2.
so (IH3 (mr_cn Hmr Hmrk1)) as Hmrk2; clear IH3.
split; auto using mrk_sigma_intro; [].
intros E E' HintE.
so (Hfun1 _ _ HintE) as Hint1; clear Hfun1.
so (Hfun2 _ _ HintE) as Hint2; clear Hfun2.
so (ofc_skof _#3 (mr_wf _ Hmr) Hofc) as Hskofeq.
simpl in Hskofeq.
remember (intc c E) as Q eqn:Heq.
destruct Q as [sk P].
so (intc_skind _#4 (eqsymm Heq)) as Heqsk.
rewrite <- (intctx_flatten _#3 HintE andel) in Heqsk; [].
rewrite -> Hskofeq in Heqsk; [].
subst sk.
clear Hskofeq.
so (ofc_skof _#3 (mr_wf _ Hmr) Hofc') as Hskofeq.
simpl in Hskofeq.
remember (intc c' E') as Q eqn:Heq'.
destruct Q as [sk P'].
so (intc_skind _#4 (eqsymm Heq')) as Heqsk.
rewrite <- (intctx_flatten _#3 (intctx_symm_mr _#3 Hmr HintE) andel) in Heqsk; [].
rewrite -> Hskofeq in Heqsk; [].
subst sk.
clear Hskofeq.
simpl.
#4 apply sigma_interp_.
  eapply subst_ofc; eauto using left_subof; done.

  eapply subst_ofc; eauto using right_subof; done.

  simpl in Hint1.
  rewrite <- Heq in Hint1; [].
  rewrite <- Heq' in Hint1; [].
  assumption.

  simpl in Hint2.
  rewrite <- Heq in Hint2; [].
  rewrite <- Heq' in Hint2; [].
  simp_sub in Hint2.
  assert (subenv E (dot (cn_pi1 c) id) (E; Sta (subst (left E) (cn_pi1 c)) (subst (right E) (cn_pi1 c)) (intc (cn_pi1 c) E))) as Hsubenv.
    eapply subenv_dot; eauto using subenv_sh, ofc_fcn, intctx_fenv with sequence formation; done.
  simpl in Hsubenv.
  rewrite <- Heq in Hsubenv; [].
  simp_sub in Hsubenv.
  unfold subst1 in Hint2.
  rewrite -> (subst_intk _#5 (ofk_fkd _ _ Hofk2) (fenv_cn (intctx_fenv _#3 HintE)) Hsubenv) in Hint2; [].
  assumption.

(* extone *)
intros G c c' Hofc IH1 Hofc' IH2 Hmr.
so (IH1 Hmr) as (_ & Hfun1); clear IH1.
so (IH2 Hmr) as (_ & Hfun2); clear IH2.
split; auto using mrk_one_intro; [].
intros E E' HintE.
so (Hfun1 _ _ HintE) as Hint1.
simpl in Hint1.
invert Hint1; [].
intros _ _ Heq _.
so (Hfun2 _ _ HintE) as Hint2.
simpl in Hint2.
invert Hint2; [].
intros _ _ _ Heq'.
rewrite <- Heq, <- Heq'.
#2 apply one_interp_.
  replsub kd_one with (subst (left E) kd_one).
  eapply subst_ofc; eauto using left_subof; done.

  replsub kd_one with (subst (right E) kd_one).
  eapply subst_ofc; eauto using right_subof; done.

(* subsume *)
intros G c c' k k' Heqc IH1 Hsubk IH2 Hmr.
so (IH1 Hmr) as (Hmrk & Hfun); clear IH1.
so (IH2 Hmr) as (_ & Hmrk' & Hincl); clear IH2.
split; auto; [].
intros E E' HintE.
apply Hincl; eauto using intctx_refl_mr; done.

(* beta *)
intros G k1 k2 c1 c2 Hofc2 IH1 Hofc1 IH2 Hmr.
so (IH2 Hmr) as (Hmrk1 & Hfun1); clear IH2.
so (IH1 (mr_cn Hmr Hmrk1)) as (Hmrk2 & Hfun2); clear IH1.
assert (mrk G (subst1 c1 k2)) as Hmrk2x.
  #2 eapply subst_mrk; eauto.
    auto using wf_cn, mr_wf, mrk_ofk; done.

    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply mrcl_cn; auto; done.
split; auto; [].
intros E E' HintE.
so (Hfun1 _ _ HintE) as Hint1; clear Hfun1.
#2 exploit (Hfun2 (E; Sta (subst (left E) c1) (subst (right E) c1) (intc c1 E)) (E'; Sta (subst (left E') c1) (subst (right E') c1) (intc c1 E'))) as Hint2.
  #2 apply intctx_cn; auto.
    eapply seq_ofc; eauto using intctx_left, wf_nil; done.

    eapply seq_ofc; eauto using intctx_right, wf_nil; done.
simpl in Hint2.
assert (subenv E (dot c1 id) (E; Sta (subst (left E) c1) (subst (right E) c1) (intc c1 E))) as Hsubenv.
  eapply subenv_dot; eauto using subenv_sh, ofc_fcn, intctx_fenv with sequence; done.
assert (subenv E' (dot c1 id) (E'; Sta (subst (left E') c1) (subst (right E') c1) (intc c1 E'))) as Hsubenv'.
  eapply subenv_dot; eauto using subenv_sh, ofc_fcn, intctx_fenv' with sequence; done.
rewrite <- (subst_intk _#5 (ofk_fkd _ _ (mrk_ofk _ _ Hmrk2)) (fenv_cn (intctx_fenv _#3 HintE)) Hsubenv) in Hint2; [].
rewrite <- (subst_intc _#5 (ofc_fcn _#3 Hofc2) (fenv_cn (intctx_fenv _#3 HintE)) Hsubenv) in Hint2; [].
rewrite <- (subst_intc _#5 (ofc_fcn _#3 Hofc2) (fenv_cn (intctx_fenv' _#3 HintE)) Hsubenv') in Hint2; [].
repl (intc (cn_app (cn_lam k1 c2) c1) E) with (intc (subst1 c1 c2) E).
  simpl.
  #2 rewrite -> (cand_eta _ (skof (flatten_env E; Some (simp k1)) c2)).
    #2 rewrite -> (cand_eta _ (simp k1)).
      rewrite <- (subst_intc _#5 (ofc_fcn _#3 Hofc2) (fenv_cn (intctx_fenv _#3 HintE)) Hsubenv); [].
      reflexivity.

      eapply (ofc_intc_csk G); eauto using mr_wf, intctx_refl_mr; done.

    apply intc_csk; done.
so (eqc_beta Hofc2 Hofc1) as Hbeta.
#2 eapply intk_defeq_mr; eauto.
  simp_sub.
  apply eqc_symm; [].
  so (subst_eqc _#6 (left_subof _#3 HintE) Hbeta) as H.
  simp_sub in H.
  assumption.

  simp_sub.
  apply eqc_symm; [].
  so (subst_eqc _#6 (right_subof _#3 HintE) Hbeta) as H.
  simp_sub in H.
  assumption.

(* beta1 *)
intros G k1 k2 c1 c2 Hofc1 IH1 Hofc2 IH2 Hmr.
so (IH1 Hmr) as (Hmrk1 & Hfun1); clear IH1.
so (IH2 Hmr) as (Hmrk2 & Hfun2); clear IH2.
split; auto; [].
intros E E' HintE.
simpl.
remember (intc c1 E) as Q1 eqn:HeqQ1.
remember (intc c2 E) as Q2 eqn:HeqQ2.
destruct Q1 as [sk1 Z1].
destruct Q2 as [sk2 Z2].
simpl.
rewrite -> HeqQ1; [].
so (Hfun1 _ _ HintE) as Hint.
so (eqc_beta1 Hofc1 Hofc2) as Hbeta.
#2 eapply intk_defeq_mr; eauto.
  simp_sub.
  apply eqc_symm; [].
  so (subst_eqc _#6 (left_subof _#3 HintE) Hbeta) as H.
  simp_sub in H.
  assumption.

  simp_sub.
  apply eqc_symm; [].
  so (subst_eqc _#6 (right_subof _#3 HintE) Hbeta) as H.
  simp_sub in H.
  assumption.

(* beta2 *)
intros G k1 k2 c1 c2 Hofc1 IH1 Hofc2 IH2 Hmr.
so (IH1 Hmr) as (Hmrk1 & Hfun1); clear IH1.
so (IH2 Hmr) as (Hmrk2 & Hfun2); clear IH2.
split; auto; [].
intros E E' HintE.
simpl.
remember (intc c1 E) as Q1 eqn:HeqQ1.
remember (intc c2 E) as Q2 eqn:HeqQ2.
destruct Q1 as [sk1 Z1].
destruct Q2 as [sk2 Z2].
simpl.
rewrite -> HeqQ2; [].
so (Hfun2 _ _ HintE) as Hint.
so (eqc_beta2 Hofc1 Hofc2) as Hbeta.
#2 eapply intk_defeq_mr; eauto.
  simp_sub.
  apply eqc_symm; [].
  so (subst_eqc _#6 (left_subof _#3 HintE) Hbeta) as H.
  simp_sub in H.
  assumption.

  simp_sub.
  apply eqc_symm; [].
  so (subst_eqc _#6 (right_subof _#3 HintE) Hbeta) as H.
  simp_sub in H.
  assumption.

(* epilogue *)
destruct_all Hind; do2 4 split; auto; done.
Qed.


Lemma wf_mr :
  forall G, wf G -> mr G.
Proof.
intros G Hwf.
induct Hwf.

(* nil *)
exact mr_nil.

(* cn *)
intros G k _ Hmr Hofk.
apply mr_cn; auto; [].
exact (model_regular andel _ _ Hmr Hofk).

(* tm *)
intros G t _ Hmr Hoft.
apply mr_tm; auto; done.

(* md *)
intros G k T _ Hmr Hofs Hfst.
eapply mr_md; eauto; [].
apply mr_cn; auto; [].
apply (model_regular andel); auto; [].
eapply ofs_fst; eauto; done.
Qed.


Lemma ofk_mrk :
  forall G k, wf G -> ofk G k -> mrk G k.
Proof.
intros G k Hwf Hofk.
apply (model_regular andel); auto using wf_mr.
Qed.


Lemma intctx_refl :
  forall G E E',
     wf G
     -> intctx G E E' 
     -> intctx G E E.
Proof.
intros; eapply intctx_refl_mr; eauto using wf_mr.
Qed.


Lemma intctx_symm :
  forall G E E',
    wf G
    -> intctx G E E'
    -> intctx G E' E.
Proof.
intros; eapply intctx_symm_mr; eauto using wf_mr.
Qed.


Lemma intctx_trans :
  forall G E1 E2 E3,
    wf G
    -> intctx G E1 E2
    -> intctx G E2 E3
    -> intctx G E1 E3.
Proof.
intros; eapply intctx_trans_mr; eauto using wf_mr.
Qed.


Lemma intk_refl :
  forall G k E E' c c' Q Q',
    wf G
    -> ofk G k
    -> intctx G E E'
    -> intk k E c c' Q Q'
    -> intk k E c c' Q Q.
Proof.
intros; eapply intk_refl_mr; eauto using wf_mr, ofk_mrk.
Qed.


Lemma intk_symm :
  forall G k E E' c c' Q Q',
    wf G
    -> ofk G k
    -> intctx G E E'
    -> intk k E c c' Q Q'
    -> intk k E c c' Q' Q.
Proof.
intros; eapply intk_symm_mr; eauto using wf_mr, ofk_mrk.
Qed.


Lemma intk_trans :
  forall G k E E' c c' Q1 Q2 Q3,
    wf G
    -> ofk G k
    -> intctx G E E'
    -> intk k E c c' Q1 Q2
    -> intk k E c c' Q2 Q3
    -> intk k E c c' Q1 Q3.
Proof.
intros; eapply intk_trans_mr; eauto using wf_mr, ofk_mrk.
Qed.


Lemma intk_fun :
  forall G k E E',
    wf G
    -> ofk G k
    -> intctx G E E'
    -> intk k E = intk k E'.
Proof.
intros; eapply intk_fun_mr; eauto using wf_mr, ofk_mrk.
Qed.


Lemma intk_defeq :
  forall G k E E' c1 c2 c1' c2' Q Q',
    wf G
    -> ofk G k
    -> intctx G E E'
    -> intk k E c1 c2 Q Q'
    -> eqc nil c1 c1' (subst (left E) k)
    -> eqc nil c2 c2' (subst (right E) k)
    -> intk k E c1' c2' Q Q'.
Proof.
intros; eapply intk_defeq_mr; eauto using wf_mr, ofk_mrk.
Qed.


Lemma eqk_intk :
  forall G k k' E E',
    wf G
    -> eqk G k k'
    -> intctx G E E'
    -> intk k E = intk k' E'.
Proof.
intros G k k' E E' Hwf Heqk HintE.
transitivity (intk k E').
  eapply intk_fun; eauto; [].
  exact (eqk_reg _#3 Hwf Heqk andel).

  refine (model_regular ander andel _#5 ander ander _ _); eauto using wf_mr, intctx_symm, intctx_refl; done.
Qed.


Lemma subk_intk :
  forall G k k' E E',
    wf G
    -> subk G k k'
    -> intctx G E E'
    -> forall V V' Q Q', intk k E V V' Q Q' -> intk k' E' V V' Q Q'.
Proof.
intros G k k' E E' Hwf Hsubk HintE V V' Q Q' HV.
rewrite <- (intk_fun _#4 Hwf (subk_reg _#3 Hwf Hsubk ander) HintE); [].
exact (model_regular anderrl _#3 (wf_mr _ Hwf) Hsubk anderr _#5 (intctx_refl _#3 Hwf HintE) HV).
Qed.


Lemma intc_fun :
  forall G c k E E',
    wf G
    -> ofc G c k
    -> intctx G E E'
    -> intk k E (subst (left E) c) (subst (right E) c) (intc c E) (intc c E').
Proof.
intros G c k E E' Hwf Hofc HintE.
refine (model_regular ander ander ander andel _#5 ander _ _ _); eauto using wf_mr.
Qed.


Lemma eqc_intc :
  forall G c c' k E E',
    wf G
    -> eqc G c c' k
    -> intctx G E E'
    -> intk k E (subst (left E) c) (subst (right E) c) (intc c E) (intc c' E').
Proof.
intros G c c' k E E' Hwf Heqc HintE.
refine (model_regular ander ander ander ander _#6 ander _#3); eauto using wf_mr.
Qed.


Lemma intt_fun :
  forall G t E E',
    wf G
    -> ofc G t kd_type
    -> intctx G E E'
    -> intt t E = intt t E'.
Proof.
intros G t E E' Hwf Hof HintE.
so (intc_fun _#5 Hwf Hof HintE) as Hint.
simpl in Hint.
so (type_interp_eq _#5 Hint) as Heq.
unfold intt.
rewrite <- Heq; auto.
Qed.


Lemma eqc_intt :
  forall G t t' E E',
    wf G
    -> eqc G t t' kd_type
    -> intctx G E E'
    -> intt t E = intt t' E'.
Proof.
intros G t t' E E' Hwf Heqt HintE.
so (eqc_intc _#6 Hwf Heqt HintE) as Hint.
simpl in Hint.
so (type_interp_eq _#5 Hint) as Heq.
unfold intt.
rewrite <- Heq; auto.
Qed.


Lemma intt_tvrel :
  forall G t E,
    wf G
    -> ofc G t kd_type
    -> intctx G E E
    -> tvrel (subst (left E) t) (subst (right E) t) (intt t E).
Proof.
intros G t E Hwf Hof HintE.
so (intc_fun _#5 Hwf Hof HintE) as Hint.
simpl in Hint.
invert Hint; [].
intros R _ _ HofR _ Heq _.
unfold intt.
rewrite <- Heq; [].
simpl.
rewrite -> cast_id; [].
assumption.
Qed.


Lemma intt_reg :
  forall G t E v v',
    wf G
    -> ofc G t kd_type
    -> intctx G E E
    -> intt t E v v'
    -> oft nil v (subst (left E) t)
       /\ oft nil v' (subst (right E) t).
Proof.
intros G t E v v' Hwf Hoft HintE Hv.
so (intt_tvrel _#3 Hwf Hoft HintE _ _ Hv) as (Hofv & Hofv' & _).
auto.
Qed.


Lemma intt_tpclosed :
  forall G t E,
    wf G
    -> ofc G t kd_type
    -> intctx G E E
    -> tpclosed (subst (left E) t) (subst (right E) t) (intt t E).
Proof.
intros G t E Hwf Hof HintE.
so (intc_fun _#5 Hwf Hof HintE) as Hint.
simpl in Hint.
invert Hint; [].
intros R _ _ _ Hcl Heq _.
unfold intt.
rewrite <- Heq; [].
simpl.
rewrite -> cast_id; [].
assumption.
Qed.


Lemma intt_ceqt :
  forall G t E v1 v1' v2 v2',
    wf G
    -> ofc G t kd_type
    -> intctx G E E
    -> Contextual.ceqt nil v1 v1' (subst (left E) t)
    -> Contextual.ceqt nil v2 v2' (subst (right E) t)
    -> tvalue nil v1'
    -> tvalue nil v2'
    -> intt t E v1 v2
    -> intt t E v1' v2'.
Proof.
intros G t E v1 v1' v2 v2' Hwf Hofct HintE Heqt1 Heqt2 Hval1' Hval2' Hint.
eapply tpst_ceqt; eauto using intt_tvrel, intt_tpclosed.
Qed.


Lemma intks_fun :
  forall G k E E',
    wf G
    -> ofk G k
    -> intctx G E E'
    -> intks k E = intks k E'.
Proof.
intros G k E E' Hwf Hof HintE.
unfold intks.
rewrite <- (intk_fun G _ E E'); auto.
Qed.
