
Require Import Tactics.
Require Import Extensionality.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Regularity.
Require Import Congruence.
Require Import Contextual.
Require Import Closure.
Require Import Candidate.
Require Import Logical.
Require Import ModelRegular.
Require Inversion.
Require Progress.
Require Preservation.


Lemma ints_reg :
  forall E T V V' Q,
    ints T E V V' Q
    -> ofm nil pure V (subst (left E) T)
       /\ ofm nil pure V' (subst (right E) T).
Proof.
intros E T V V' Q Hint.
revert E V V' Q Hint.
#6 induct T; try (intros; destruct Hint; done).

(* one *)
intros E V V' Q Hint.
invert Hint; [].
intros <- <- <-.
simp_sub.
split; auto with static; done.

(* stat *)
intros k _ E V V' Q Hint.
simpl in Hint.
invert Hint; clear Hint; [].
intros c c' Hintk <- <-.
unfold intks in Hintk.
so (intk_reg _#6 Hintk) as (Hofc & Hofc').
simp_sub.
split; auto with static; done.

(* dyn *)
intros t _ E V V' Q Hint.
invert Hint; clear Hint; [].
intros v v' Hof Hof' Hintt <- <- <-.
so (Inversion.dyn_ofm_inversion_dyn _#4 wf_nil Hof) as Hofv.
so (Inversion.dyn_ofm_inversion_dyn _#4 wf_nil Hof') as Hofv'.
simp_sub.
split; auto with static; done.

(* pigen *)
intros T1 _ T2 _ E V V' Q Hint.
invert Hint; clear Hint; [].
intros HofV HofV' HvalV HvalV' Hrel <-.
split; eauto using Preservation.mvalue_pure; done.

(* piapp *)
intros T1 _ T2 _ E V V' Q Hint.
invert Hint; clear Hint; [].
intros c c' k1 k2 F HofV HofV' HvalV HvalV' HfstV HfstV' Hfst1 Hfst2 Hintk Hrel <-.
split; eauto using Preservation.mvalue_pure; done.

(* sigma *)
intros T1 IH1 T2 IH2 E V V' Q Hint.
invert Hint; clear Hint; [].
intros V1 V1' V2 V2' c1 c1' k1 k2 P Hof Hof' HvalV1 HvalV2 HvalV1' HvalV2' HfstV1 HfstV1' Hfst1 Hfst2 Hint1 Hint2 <- <- <-.
so (IH1 _#4 Hint1) as (HofV1 & HofV1').
so (IH2 _#4 Hint2) as (HofV2 & HofV2').
split; eauto using Preservation.mvalue_pure with dynamic; done.
Qed.


Lemma ints_val :
  forall G E T V V' Q,
    wf G
    -> ofs G T
    -> intctx G E E
    -> ints T E V V' Q
    -> mvalue nil V
       /\ mvalue nil V'.
Proof.
intros G E T V V' Q Hwf Hof HintE Hint.
revert E V V' Q Hint HintE Hwf.
induct Hof.

(* one *)
intros G E V V' Q Hint _ _.
simpl in Hint.
invert Hint; [].
intros <- <- <-.
auto with dynamic; done.

(* stat *)
intros G t _ E V V' Q Hint _ _.
simpl in Hint.
invert Hint; [].
intros c c' _ <- <-.
auto with dynamic; done.

(* dyn *)
intros G t Hof E V V' Q Hint HintE Hwf.
simpl in Hint.
invert Hint; [].
intros v v' _ _ Hintv <- <- <-.
so (intt_tvrel _#3 Hwf Hof HintE v v' Hintv) as (_ & _ & Hofv & Hofv').
auto with dynamic; done.

(* pigen *)
intros G T1 T2 k1 _ _ _ _ _ E V V' Q Hint HintE Hwf.
simpl in Hint.
invert Hint; [].
intros; auto; done.

(* piapp *)
intros G T1 T2 k1 _ _ _ _ _ E V V' Q Hint HintE Hwf.
simpl in Hint.
invert Hint; [].
intros; auto; done.

(* sigma *)
intros G T1 T2 k1 _ _ _ _ _ E V V' Q Hint HintE Hwf.
simpl in Hint.
invert Hint; [].
intros; subst; auto with dynamic; done.
Qed.


Lemma ints_fst :
  forall G T E V V' Q c c' k,
    wf G
    -> ofs G T
    -> intctx G E E
    -> fsts T k
    -> fstm nil V c
    -> fstm nil V' c'
    -> ints T E V V' Q
    -> intk k E c c' Q Q.
Proof.
intros G T E V V' Q c c' k Hwf Hof HintE Hfst HfstV HfstV' HintV.
revert E V V' Q c c' k Hwf HintE Hfst HfstV HfstV' HintV.
induct Hof.

(* one *)
intros G E V V' Q c c' k Hwf HintE Hfsts HfstV HfstV' Hint.
invert Hfsts; clear Hfsts; [].
intros <-.
simpl in Hint.
invert Hint; [].
intros <- <- <-.
invert HfstV; invert HfstV'; clear HfstV HfstV'; [].
intros <- <-.
simpl.
apply one_interp_; auto with static; done.

(* stat *)
intros G k Hofk E V V' Q c c' ka Hwf HintE Hfst HfstV HfstV' Hint.
invert Hfst; clear Hfst; [].
intros <-.
invert Hint; [].
intros ca ca' Hintc <- <-.
invert HfstV; invert HfstV'; [].
intros -> ->.
assumption.

(* dyn *)
intros G t Hoft E V V' Q c c' k Hwf HintE Hfst HfstV HfstV' Hints.
invert Hfst; clear Hfst; [].
intros <-.
invert Hints; [].
intros v v' Hofv Hofv' Hint <- <- <-.
invert HfstV; invert HfstV'; [].
intros <- <-.
simpl.
apply one_interp_; auto with static; done.

(* pigen *)
intros G T1 T2 k1 HofT1 _ Hfst1 HofT2 _ E V V' Q c c' k Hwf HintE Hfst HfstV HfstV' HQ.
so (fsts_fun _#3 Hfst fsts_pigen); subst k; clear Hfst.
invert HQ; [].
intros HofV HofV' _ _ _ <-.
simpl.
simp_sub in HofV; simp_sub in HofV'.
so (ofm_fst _#6 wf_nil HofV HfstV fsts_pigen) as Hofc.
so (ofm_fst _#6 wf_nil HofV' HfstV' fsts_pigen) as Hofc'.
apply one_interp_; auto; done.

(* piapp *)
intros G T1 T2 k1 HofT1 _ Hfst1 HofT2 _ E V V' Q c c' k Hwf HintE Hfstpi HfstV HfstV' HV.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
invert Hfstpi; clear Hfstpi; [].
intros k1a k2 Hfst1a Hfst2 <-.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
simpl in HV.
invert HV; clear HV; [].
intros ca ca' k1a k2a F _ _ _ _ HfstVa HfstV'a Hfst1a Hfst2a Hint _ <-.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a; clear Hfst2a.
so (fstm_fun _#4 HfstV HfstVa); subst ca; clear HfstVa.
so (fstm_fun _#4 HfstV' HfstV'a); subst ca'; clear HfstV'a.
assumption.

(* sigma *)
intros G T1 T2 k1 HofT1 IH1 Hfst1 HofT2 IH2 E V V' Q c c' k Hwf HintE Hofsigma HfstV HfstV' HV.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
invert Hofsigma; clear Hofsigma; [].
intros k1a k2 Hfst1a Hfst2 <-.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
simpl in HV.
invert HV; clear HV; [].
intros V1 V1' V2 V2' c1 c1' k1a k2a P Hof Hof' HvalV1 HvalV2 HvalV1' HvalV2' HfstV1 HfstV1' Hfst1a Hfst2a HintV1 HintV2 <- <- <-.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a; clear Hfst2a.
invert HfstV; [].
intros c1a c2 HfstV1a HfstV2 <-.
so (fstm_fun _#4 HfstV1 HfstV1a); subst c1a; clear HfstV1a.
invert HfstV'; [].
intros c1a' c2' HfstV1a' HfstV2' <-.
so (fstm_fun _#4 HfstV1' HfstV1a'); subst c1a'; clear HfstV1a'.
so (Preservation.mvalue_pure _#4 (mvalue_pair HvalV1 HvalV2) Hof) as H.
simp_sub in H.
so (Inversion.pair_ofm_inversion_sigma_pure _#5 wf_nil H) as (c1a & HofV1 & HfstV1a & HofV2).
clear H.
so (fstm_fun _#4 HfstV1 HfstV1a); subst c1a; clear HfstV1a.
so (Preservation.mvalue_pure _#4 (mvalue_pair HvalV1' HvalV2') Hof') as H.
simp_sub in H.
so (Inversion.pair_ofm_inversion_sigma_pure _#5 wf_nil H) as (c1a' & HofV1' & HfstV1a' & HofV2').
clear H.
so (fstm_fun _#4 HfstV1' HfstV1a'); subst c1a'; clear HfstV1a'.
simp_sub in HofV2.
simp_sub in HofV2'.
so (ofm_fst _#6 wf_nil HofV1 HfstV1 (subst_fsts _#3 Hfst1)) as Hofc1.
so (ofm_fst _#6 wf_nil HofV1' HfstV1' (subst_fsts _#3 Hfst1)) as Hofc1'.
so (ofm_fst _#6 wf_nil HofV2 HfstV2 (subst_fsts _#3 Hfst2)) as Hofc2.
so (ofm_fst _#6 wf_nil HofV2' HfstV2' (subst_fsts _#3 Hfst2)) as Hofc2'.
simpl.
#4 apply sigma_interp_; auto.
  exact (ofm_fst _#6 wf_nil Hof (fstm_pair HfstV1 HfstV2) (subst_fsts _#3 (fsts_sigma Hfst1 Hfst2))).

  exact (ofm_fst _#6 wf_nil Hof' (fstm_pair HfstV1' HfstV2') (subst_fsts _#3 (fsts_sigma Hfst1 Hfst2))).

  so (IH1 _#7 Hwf HintE Hfst1 HfstV1 HfstV1' HintV1) as Hintc1.
  #2 eapply intk_defeq; eauto.
    apply eqc_symm; [].
    eapply eqc_beta1; eauto; done.

    apply eqc_symm; [].
    eapply eqc_beta1; eauto; done.

  assert (intctx (G; cl_cn k1) (E; Sta c1 c1' (Cand (simp k1) (fst P))) (E; Sta (cn_pi1 (cn_pair c1 c2)) (cn_pi1 (cn_pair c1' c2')) (Cand (simp k1) (fst P)))) as HintEx.
    #2 apply intctx_cn; eauto using intctx_refl.
      apply eqc_symm; [].
      eapply eqc_beta1; eauto; done.

      apply eqc_symm; [].
      eapply eqc_beta1; eauto; done.
  so (IH2 _#7 (wf_cn Hwf Hofk1) (intctx_refl _#3 (wf_cn Hwf Hofk1) HintEx) Hfst2 HfstV2 HfstV2' HintV2) as Hintc2.
  so (ofs_fst _#3 HofT2 Hfst2) as Hofk2.
  cut (intk k2 (E; Sta c1 c1' (Cand (simp k1) (fst P))) (cn_pi2 (cn_pair c1 c2)) (cn_pi2 (cn_pair c1' c2')) (Cand (simp k2) (snd P)) (Cand (simp k2) (snd P))).
    intro H.
    rewrite -> (intk_fun _#4 (wf_cn Hwf Hofk1) Hofk2 HintEx) in H.
    exact H.
  #2 eapply (intk_defeq (G; cl_cn k1)); eauto using wf_cn.
    simpl left.
    apply eqc_symm; [].
    eapply eqc_beta2; eauto; done.

    simpl right.
    apply eqc_symm; [].
    eapply eqc_beta2; eauto; done.
Qed.


Lemma ints_csk :
  forall G T E V V' Q k,
    wf G
    -> ofs G T
    -> ints T E V V' Q
    -> intctx G E E
    -> fsts T k
    -> csk Q = simp k.
Proof.
intros G T E V V' Q k Hwf Hofs Hints HintE Hfst.
so (ints_reg _#5 Hints) as (Hof & Hof').
so (Preservation.can_fstm _#3 wf_nil Hof) as (c & HfstV).
so (Preservation.can_fstm _#3 wf_nil Hof') as (c' & HfstV').
so (ints_fst _#9 Hwf Hofs HintE Hfst HfstV HfstV' Hints) as Hintk.
exact (intk_csk _#6 Hintk andel).
Qed.


Theorem ints_fun :
  forall G T E E',
    wf G
    -> ofs G T
    -> intctx G E E'
    -> ints T E = ints T E'.
Proof.
intros G T E E' Hwf Hof HintE.
revert E E' Hwf HintE.
induct Hof.

(* one *)
intros G E E' Hwf HintE.
simpl.
fextensionality 3.
intros V V' Q.
pextensionality.
  intro H; invert H; clear H; [].
  intros <- <- <-.
  apply sg_one_interp_; done.

  intro H; invert H; clear H; [].
  intros <- <- <-.
  apply sg_one_interp_; done.

(* stat *)
intros G k Hofk E E' Hwf HintE.
simpl.
fextensionality 3.
intros V V' Q.
pextensionality.
  intro H; invert H; clear H; [].
  intros c c' Hint <- <-.
  apply sg_stat_interp_; auto; [].
  unfold intks.
  unfold intks in Hint.
  rewrite <- (intk_fun _#4 Hwf Hofk HintE); [].
  assumption.

  intro H; invert H; clear H; [].
  intros c c' Hint <- <-.
  apply sg_stat_interp_; auto; [].
  unfold intks.
  unfold intks in Hint.
  rewrite -> (intk_fun _#4 Hwf Hofk HintE); [].
  assumption.

(* dyn *)
intros G t Hoft E E' Hwf HintE.
simpl.
fextensionality 3.
intros V V' Q.
pextensionality.
  intro H; invert H; clear H; [].
  intros v v' Hofv Hofv' Hint <- <- <-.
  #3 apply sg_dyn_interp_; auto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_dyn; [].
    eapply seq_ofc_type; eauto using intctx_left with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_dyn; [].
    eapply seq_ofc_type; eauto using intctx_right with static; done.

    erewrite <- intt_fun; eauto; done.

  intro H; invert H; clear H; [].
  intros v v' Hofv Hofv' Hint <- <- <-.
  #3 apply sg_dyn_interp_; auto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    apply eqs_dyn; [].
    eapply seq_ofc_type; eauto using intctx_left with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    apply eqs_dyn; [].
    eapply seq_ofc_type; eauto using intctx_right with static; done.

    erewrite -> intt_fun; eauto; done.

(* pigen *)
intros G T1 T2 k1 HofT1 IH1 Hfst1 HofT2 IH2 E E' Hwf HintE.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
fextensionality 3.
intros V V' Q.
simpl.
pextensionality.
  intro H; invert H; clear H; [].
  intros HofV HofV' HvalV HvalV' Hrel <-.
  #3 apply sg_pigen_interp_; auto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply seq_ofs; eauto using intctx_left with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply seq_ofs; eauto using intctx_right with static; done.

    intros W W' Q HW.
    rewrite <- (IH1 _ _ Hwf HintE) in HW; [].
    so (ints_reg _#5 HW) as (HofW & HofW').
    so (Hrel _ _ _ HW) as (c & c' & HfstW & HfstW' & HVW).
    so (ints_fst _#9 Hwf HofT1 (intctx_refl _#3 Hwf HintE) Hfst1 HfstW HfstW' HW) as Hint.
    clear IH1.
    exists c, c'.
    do2 2 split; auto; [].
    so (ofm_fst _#6 wf_nil HofW HfstW (subst_fsts _#3 Hfst1)) as Hofc.
    so (ofm_fst _#6 wf_nil HofW' HfstW' (subst_fsts _#3 Hfst1)) as Hofc'.
    rewrite <- (fun H1 H2 => IH2 (E; Sta c c' Q) (E'; Sta c c' Q) H1 H2); auto using intctx_cn with static; [].
    #2 erewrite <- (miev_eqs _#5); [eexact HVW | ..].
      eapply seq_ofs; eauto using wf_nil; [].
      apply subrel_dot; auto using intctx_left; [].
      simp_sub.
      apply seq_cn; auto with static; done.

      eapply seq_ofs; eauto using wf_nil; [].
      apply subrel_dot; auto using intctx_right; [].
      simp_sub.
      apply seq_cn; auto with static; done.

  intro H; invert H; clear H; [].
  intros HofV HofV' HvalV HvalV' Hrel <-.
  #3 apply sg_pigen_interp_; auto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    eapply seq_ofs; eauto using intctx_left with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    eapply seq_ofs; eauto using intctx_right with static; done.

    intros W W' Q HW.
    so (ints_reg _#5 HW) as (HofW & HofW').
    rewrite -> (IH1 _ _ Hwf HintE) in HW; [].
    so (Hrel _ _ _ HW) as (c & c' & HfstW & HfstW' & HVW).
    so (ints_fst _#9 Hwf HofT1 (intctx_refl _#3 Hwf (intctx_symm _#3 Hwf HintE)) Hfst1 HfstW HfstW' HW) as Hint.
    clear IH1.
    exists c, c'.
    do2 2 split; auto; [].
    so (ofm_fst _#6 wf_nil HofW HfstW (subst_fsts _#3 Hfst1)) as Hofc.
    so (ofm_fst _#6 wf_nil HofW' HfstW' (subst_fsts _#3 Hfst1)) as Hofc'.
    #2 rewrite <- (fun H1 H2 => IH2 (E; Sta c c' Q) (E'; Sta c c' Q) H1 H2) in HVW; auto with static.
      #2 erewrite <- (miev_eqs _#5); [eexact HVW | ..].
        apply eqs_symm; [].
        eapply seq_ofs; eauto using wf_nil; [].
        apply subrel_dot; auto using intctx_left; [].
        simp_sub.
        apply seq_cn; auto with static; done.

        apply eqs_symm; [].
        eapply seq_ofs; eauto using wf_nil; [].
        apply subrel_dot; auto using intctx_right; [].
        simp_sub.
        apply seq_cn; auto with static; done.

      apply intctx_cn; auto with static; [].
      rewrite -> (intk_fun _#4 Hwf Hofk1 HintE); auto; done.

(* piapp *)
intros G T1 T2 k1 HofT1 IH1 Hfst1 HofT2 IH2 E E' Hwf HintE.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
fextensionality 3.
intros V V' Q.
simpl.
pextensionality.
  intro H; invert H; clear H; [].
  intros c c' k1a k2 F HofV HofV' HvalV HvalV' HfstV HfstV' Hfst1a Hfst2 Hintk Hrel <-.
  so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
  so (ofs_fst _#3 HofT2 Hfst2) as Hofk2.
  #4 eapply sg_piapp_interp_; eauto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply seq_ofs; eauto using intctx_left with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply seq_ofs; eauto using intctx_right with static; done.

    rewrite <- (intks_fun _#4 Hwf (ofk_pi Hofk1 Hofk2) HintE); [].
    assumption.

    intros W W' Q HW.
    rewrite <- (IH1 _ _ Hwf HintE) in HW; [].
    so (Hrel _#3 HW) as (d & d' & HfstW & HfstW' & HVW).
    so (ints_fst _#9 Hwf HofT1 (intctx_refl _#3 Hwf HintE) Hfst1 HfstW HfstW' HW) as Hd.
    so (intk_reg _#6 Hd) as (Hofd & Hofd').
    exists d, d'.
    do2 2 split; auto; [].
    assert (intctx (G; cl_cn k1) (E; Sta d d' Q) (E'; Sta d d' Q)) as HintEx by (apply intctx_cn; auto with static).
    rewrite <- (IH2 _ _ (wf_cn Hwf Hofk1) HintEx); clear IH2; [].
    #2 erewrite <- mpuev_eqs; eauto.
      eapply seq_ofs; eauto with static; [].
      apply subrel_dot; auto using intctx_left; [].
      simp_sub.
      apply seq_cn; auto with static; done.

      eapply seq_ofs; eauto with static; [].
      apply subrel_dot; auto using intctx_right; [].
      simp_sub.
      apply seq_cn; auto with static; done.

  intro H; invert H; clear H; [].
  intros c c' k1a k2 F HofV HofV' HvalV HvalV' HfstV HfstV' Hfst1a Hfst2 Hintk Hrel <-.
  so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
  so (ofs_fst _#3 HofT2 Hfst2) as Hofk2.
  #4 eapply sg_piapp_interp_; eauto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    eapply seq_ofs; eauto using intctx_left with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    eapply seq_ofs; eauto using intctx_right with static; done.

    rewrite -> (intks_fun _#4 Hwf (ofk_pi Hofk1 Hofk2) HintE); [].
    assumption.

    intros W W' Q HW.
    so HW as HW'.
    rewrite -> (IH1 _ _ Hwf HintE) in HW'; [].
    so (Hrel _#3 HW') as (d & d' & HfstW & HfstW' & HVW).
    clear HW'.
    so (ints_fst _#9 Hwf HofT1 (intctx_refl _#3 Hwf HintE) Hfst1 HfstW HfstW' HW) as Hd.
    so (intk_reg _#6 Hd) as (Hofd & Hofd').
    exists d, d'.
    do2 2 split; auto; [].
    assert (intctx (G; cl_cn k1) (E; Sta d d' Q) (E'; Sta d d' Q)) as HintEx by (apply intctx_cn; auto with static).
    rewrite -> (IH2 _ _ (wf_cn Hwf Hofk1) HintEx); clear IH2; [].
    #2 erewrite <- mpuev_eqs; eauto.
      apply eqs_symm; [].
      eapply seq_ofs; eauto with static; [].
      apply subrel_dot; auto using intctx_left; [].
      simp_sub.
      apply seq_cn; auto with static; done.

      apply eqs_symm; [].
      eapply seq_ofs; eauto with static; [].
      apply subrel_dot; auto using intctx_right; [].
      simp_sub.
      apply seq_cn; auto with static; done.

(* sigma *)
intros G T1 T2 k1 HofT1 IH1 Hfst1 HofT2 IH2 E E' Hwf HintE.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
fextensionality 3.
intros V V' Q.
pextensionality.
  intro H; invert H; clear H; [].
  intros V1 V1' V2 V2' c1 c1' k1a k2 P Hof Hof' HvalV1 HvalV2 HvalV1' HvalV2' HfstV1 HfstV1' Hfst1a Hfst2 HintV1 HintV2 <- <- <-.
  so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
  simpl.
  #4 eapply sg_sigma_interp_; eauto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply seq_ofs; eauto using intctx_left with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply seq_ofs; eauto using intctx_right with static; done.

    rewrite <- (IH1 _ _ Hwf HintE); [].
    assumption.

    assert (intctx (G; cl_cn k1) (E; Sta c1 c1' (Cand (simp k1) (fst P))) (E'; Sta c1 c1' (Cand (simp k1) (fst P)))) as HintEx.
      #3 apply intctx_cn; auto with static.
        simp_sub in Hof.
        exact (eqc_refl (ofm_fst _#6 wf_nil (Inversion.pair_ofm_inversion_1 _#6 wf_nil Hof) HfstV1 (subst_fsts _#3 Hfst1))).

        simp_sub in Hof'.
        exact (eqc_refl (ofm_fst _#6 wf_nil (Inversion.pair_ofm_inversion_1 _#6 wf_nil Hof') HfstV1' (subst_fsts _#3 Hfst1))).

        exact (ints_fst _#9 Hwf HofT1 (intctx_refl _#3 Hwf HintE) Hfst1 HfstV1 HfstV1' HintV1).
    rewrite <- (IH2 _ _ (wf_cn Hwf Hofk1) HintEx); [].
    assumption.

  intro H; invert H; clear H; [].
  intros V1 V1' V2 V2' c1 c1' k1a k2 P Hof Hof' HvalV1 HvalV2 HvalV1' HvalV2' HfstV1 HfstV1' Hfst1a Hfst2 HintV1 HintV2 <- <- <-.
  so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
  simpl.
  #4 eapply sg_sigma_interp_; eauto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    eapply seq_ofs; eauto using intctx_left with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    eapply seq_ofs; eauto using intctx_right with static; done.

    rewrite -> (IH1 _ _ Hwf HintE); [].
    assumption.

    assert (intctx (G; cl_cn k1) (E'; Sta c1 c1' (Cand (simp k1) (fst P))) (E; Sta c1 c1' (Cand (simp k1) (fst P)))) as HintEx.
      #3 apply intctx_cn; auto using intctx_symm with static.
        simp_sub in Hof.
        exact (eqc_refl (ofm_fst _#6 wf_nil (Inversion.pair_ofm_inversion_1 _#6 wf_nil Hof) HfstV1 (subst_fsts _#3 Hfst1))).

        simp_sub in Hof'.
        exact (eqc_refl (ofm_fst _#6 wf_nil (Inversion.pair_ofm_inversion_1 _#6 wf_nil Hof') HfstV1' (subst_fsts _#3 Hfst1))).

        exact (ints_fst _#9 Hwf HofT1 (intctx_refl _#3 Hwf (intctx_symm _#3 Hwf HintE)) Hfst1 HfstV1 HfstV1' HintV1).
    rewrite <- (IH2 _ _ (wf_cn Hwf Hofk1) HintEx); [].
    assumption.
Qed.


Lemma ints_ceqm :
  forall G T E V V' V'' V''' Q Q' c c' k,
    wf G
    -> ofs G T
    -> intctx G E E
    -> fsts T k
    -> fstm nil V c
    -> fstm nil V' c'
    -> ceqm nil pure V V'' (subst (left E) T)
    -> ceqm nil pure V' V''' (subst (right E) T)
    -> mvalue nil V''
    -> mvalue nil V'''
    -> ints T E V V' Q
    -> intk k E c c' Q Q'
    -> ints T E V'' V''' Q'.
Proof.
intros G T E V V' V'' V''' Q Q' c c' k Hwf HofT HintE HfstT HfstV HfstV' HeqmV HeqmV' HvalV'' HvalV''' HintV Hintc.
revert E Hwf HintE V V' V'' V''' Q Q' c c' k HfstT HfstV HfstV' HeqmV HeqmV' HvalV'' HvalV''' HintV Hintc.
induct HofT.

(* one *)
intros G E Hwf HintE V V' V'' V''' Q Q' c c' k Hfsts HfstV HfstV' HeqV HeqV' HvalV'' HvalV''' Hints Hintk.
invert Hints; [].
intros <- <- <-.
simpl.
so (ceqm_reg _#5 wf_nil HeqV) as (_ & HofV'').
so (ceqm_reg _#5 wf_nil HeqV') as (_ & HofV''').
simp_sub in HofV''.
simp_sub in HofV'''.
invert Hfsts; [].
invert (Progress.canon_md _#3 HofV'' HvalV''); [].
invert (Progress.canon_md _#3 HofV''' HvalV'''); [].
intros <- <- <-.
invert Hintk; [].
intros _ _ <-.
apply sg_one_interp_; done.

(* stat *)
intros G k Hofk E Hwf HintE V V' V'' V''' Q Q' c c' ka Hfst HfstV HfstV' HeqmV HeqmV' HvalV'' HvalV''' Hints Hintk.
simp_sub in HeqmV.
simp_sub in HeqmV'.
invert Hfst; clear Hfst; [].
intros <-.
simpl.
so (ceqm_reg _#5 wf_nil HeqmV) as (_ & HofV'').
so (ceqm_reg _#5 wf_nil HeqmV') as (_ & HofV''').
invert (Progress.canon_md _#3 HofV'' HvalV''); [].
invert (Progress.canon_md _#3 HofV''' HvalV'''); [].
intros c''' <- c'' <-.
so (ceqm_fst' _#7 wf_nil HeqmV fsts_stat HfstV fstm_stat) as Heqc.
so (ceqm_fst' _#7 wf_nil HeqmV' fsts_stat HfstV' fstm_stat) as Heqc'.
apply sg_stat_interp_; [].
unfold intks.
eapply intk_defeq; eauto; [].
eapply intk_refl; eauto; [].
eapply intk_symm; eauto; done.

(* dyn *)
intros G t Hoft E Hwf HintE V V' V'' V''' Q Q' c c' k Hfst HfstV HfstV' HeqmV HeqmV' HvalV'' HvalV''' Hints Hintk.
invert Hfst; clear Hfst; [].
intros <-.
simpl in Hintk.
invert Hintk; clear Hintk; [].
intros _ _ <- <-.
simpl in Hints.
invert Hints; clear Hints; [].
intros v v' _ _ Hint <- <-.
so (intt_tvrel _#3 Hwf Hoft (intctx_refl _#3 Hwf HintE) _ _ Hint) as (Hofv & Hofv' & Hvalv & Hvalv').
simp_sub in HeqmV; simp_sub in HeqmV'.
so (ceqm_reg _#5 wf_nil HeqmV) as (_ & HofV'').
so (ceqm_reg _#5 wf_nil HeqmV') as (_ & HofV''').
invert (Progress.canon_md _#3 HofV'' HvalV''); [].
invert (Progress.canon_md _#3 HofV''' HvalV'''); [].
intros e''' <- e'' <-.
invert HvalV''; clear HvalV''; [].
invert HvalV'''; clear HvalV'''; [].
intros Hvale''' Hvale''.
simpl.
apply sg_dyn_interp_; auto with static; [].
#2 eapply intt_ceqt; eauto using intctx_refl.
  #2 apply (ceqt_trans _ _ (tm_ext (md_dyn v))).
    apply ceqt_symm; [].
    apply tstep_ceqt; auto with static dynamic; done.

    #2 apply (ceqt_trans _ _ (tm_ext (md_dyn e''))).
      apply ceqt_comp; [].
      apply tcc_ext; [].
      apply ceqm_comp; [].
      apply mcc_forget; auto; done.

      apply tstep_ceqt; auto with static dynamic; done.

  #2 apply (ceqt_trans _ _ (tm_ext (md_dyn v'))).
    apply ceqt_symm; [].
    apply tstep_ceqt; auto with static dynamic; done.

    #2 apply (ceqt_trans _ _ (tm_ext (md_dyn e'''))).
      apply ceqt_comp; [].
      apply tcc_ext; [].
      apply ceqm_comp; [].
      apply mcc_forget; auto; done.

      apply tstep_ceqt; auto with static dynamic; done.

(* pigen *)
intros G T1 T2 k1 HofT1 _ Hfst1 HofT2 _ E Hwf HintE V V' V'' V''' Q Q' c c' k Hfst HfstV HfstV' HeqmV HeqmV' HvalV'' HvalV''' Hints Hintk.
so (fsts_fun _#3 Hfst fsts_pigen); subst k; clear Hfst.
simpl in Hintk.
invert Hintk; [].
intros _ _ <- <-.
simpl in Hints.
invert Hints; [].
intros HofV HofV' HvalV HvalV' Hrel.
simpl.
#3 apply sg_pigen_interp_; auto.
  exact (ofm_forget (ceqm_reg _#5 wf_nil HeqmV ander)).

  exact (ofm_forget (ceqm_reg _#5 wf_nil HeqmV' ander)).
intros W W' Q HQ.
so (Hrel _ _ _ HQ) as (d & d' & HfstW & HfstW' & HVW).
exists d, d'.
do2 2 split; auto; [].
#2 eapply miev_ceqm; eauto.
  apply ceqm_comp; [].
  rewrite -> split_dot_compose; [].
  #2 apply (@mcc_appgn _#6 (subst (left E) T1)); auto.
    simp_sub in HeqmV.
    auto using ceqm_comp, mcc_forget; done.

    apply ceqm_refl; [].
    exact (ints_reg _#5 HQ andel).

  apply ceqm_comp; [].
  rewrite -> split_dot_compose; [].
  #2 apply (@mcc_appgn _#6 (subst (right E) T1)); auto.
    simp_sub in HeqmV'.
    auto using ceqm_comp, mcc_forget; done.

    apply ceqm_refl; [].
    exact (ints_reg _#5 HQ ander).

(* piapp *)
intros G T1 T2 k1 HofT1 _ Hfst1 HofT2 IH E Hwf HintE V V' V'' V''' Q Q' c c' k Hfstpi HfstV HfstV' HeqmV HeqmV' HvalV'' HvalV''' HintV Hintc.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
invert Hfstpi; clear Hfstpi; [].
intros k1a k2 Hfst1a Hfst2 <-.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (ofs_fst _#3 HofT2 Hfst2) as Hofk2.
simpl in HintV.
invert HintV; clear HintV; [].
intros ca ca' k1a k2a F HofV HofV' HvalV HvalV' HfstVa HfstV'a Hfst1a Hfst2a _ Hrel <-.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a; clear Hfst2a.
so (fstm_fun _#4 HfstV HfstVa); subst ca; clear HfstVa.
so (fstm_fun _#4 HfstV' HfstV'a); subst ca'; clear HfstV'a.
invert Hintc; [].
intros Fa F' _ _ Hrelc Heq <-.
so (cand_inject _#3 Heq); subst Fa; clear Heq.
so (ceqm_fst _#4 wf_nil HeqmV) as (kpia & ca & c'' & Hfstpia & HfstVa & HfstV'' & Heqc).
so (fsts_fun _#3 (subst_fsts _#3 (fsts_piapp Hfst1 Hfst2)) Hfstpia); subst kpia; clear Hfstpia.
so (fstm_fun _#4 HfstV HfstVa); subst ca; clear HfstVa.
so (ceqm_fst _#4 wf_nil HeqmV') as (kpia & c'a & c''' & Hfstpia & HfstV'a & HfstV''' & Heqc').
so (fsts_fun _#3 (subst_fsts _#3 (fsts_piapp Hfst1 Hfst2)) Hfstpia); subst kpia; clear Hfstpia.
so (fstm_fun _#4 HfstV' HfstV'a); subst c'a; clear HfstV'a.
simpl.
#4 eapply sg_piapp_interp_; eauto.
  exact (ofm_forget (ceqm_reg _#5 wf_nil HeqmV ander)).

  exact (ofm_forget (ceqm_reg _#5 wf_nil HeqmV' ander)).

  unfold intks.
  apply (intk_defeq G _ E E c c'); auto with static; [].
  eapply intk_refl; eauto using intk_symm with static; done.

  intros W W' Q HW.
  so (Hrel _#3 HW) as (d & d' & HfstW & HfstW' & HVW); clear Hrel.
  so (ints_fst _#9 Hwf HofT1 HintE Hfst1 HfstW HfstW' HW) as Hd.
  so (intk_reg _#6 Hd) as (Hofd & Hofd').
  exists d, d'.
  do2 2 split; auto; [].
  assert (intctx (G; cl_cn k1) (E; Sta d d' Q) (E; Sta d d' Q)) as HintEx by (apply intctx_cn; auto with static).
  so (ints_reg _#5 HW andel) as HofW.
  so (ints_reg _#5 HW ander) as HofW'.
  assert (ceqm nil pure (md_appap V W) (md_appap V'' W) (subst (dot d (left E)) T2)) as HeqmVW.
    rewrite -> split_dot_compose; [].
    apply ceqm_comp; [].
    eapply mcc_appap; eauto using ceqm_refl; [].
    simp_sub in HeqmV.
    assumption.
  assert (ceqm nil pure (md_appap V' W') (md_appap V''' W') (subst (dot d' (right E)) T2)) as HeqmVW'.
    rewrite -> split_dot_compose; [].
    apply ceqm_comp; [].
    eapply mcc_appap; eauto using ceqm_refl; [].
    simp_sub in HeqmV'.
    assumption.
  unfold mpuev.
  do2 3 split.
    rewrite -> (ceqm_halt_iff (md_appap V'' W) (md_appap V W) (subst (dot d (left E)) T2)); auto using ceqm_symm, ceqm_forget_pure; [].
    rewrite <- (ceqm_halt_iff (md_appap V' W') (md_appap V''' W') (subst (dot d' (right E)) T2)); auto using ceqm_forget_pure; [].
    exact (HVW andel).

    intros U'' U''' HstepsU'' HstepsU''' HvalU'' HvalU'''.
    lapply (ceqm_halt _#3 (ceqm_forget_impure _#5 (ceqm_symm _#5 HeqmVW))); [| unfold mhalt; eauto].
    intros (U & HstepsU & HvalU).
    lapply (ceqm_halt _#3 (ceqm_forget_impure _#5 (ceqm_symm _#5 HeqmVW'))); [| unfold mhalt; eauto].
    intros (U' & HstepsU' & HvalU').
    so (HVW anderl U U' HstepsU HstepsU' HvalU HvalU') as HU.
    simp_sub in HofV.
    so (ofm_appap HofV HofW HfstW) as H.
    simp_sub in H.
    so (Preservation.preservations_fst _#7 wf_nil H HstepsU (subst_fsts _#3 Hfst2) (fstm_appap HfstV HfstW)) as (b & HfstU & Heqb).
    clear H.
    so (ofm_appap HofV' HofW' HfstW') as H.
    simp_sub in H.
    so (Preservation.preservations_fst _#7 wf_nil H HstepsU' (subst_fsts _#3 Hfst2) (fstm_appap HfstV' HfstW')) as (b' & HfstU' & Heqb').
    clear H.
    #3 apply (IH _ (wf_cn Hwf Hofk1) HintEx U U' _ _ (Cand (simp k2) (F d d' (out (simp k1) Q))) _ b b' k2); auto.
      eapply anti_msteps_ceqm_trans; eauto; [].
      eapply ceqm_msteps_trans; eauto; done.

      eapply anti_msteps_ceqm_trans; eauto; [].
      eapply ceqm_msteps_trans; eauto; done.

      clear IH.
      so (cand_eta _ _ (intk_csk _#6 Hd andel)) as HeqQ.
      rewrite <- HeqQ in Hd.
      so (Hrelc _#6 Hd (eqc_refl Hofd) (eqc_refl Hofd')) as Hintcd.
      rewrite -> HeqQ in Hintcd at 1; [].
      eapply (intk_defeq (G; cl_cn k1)); eauto with static; done.

    exact (ceqm_reg _#5 wf_nil HeqmVW ander).

    exact (ceqm_reg _#5 wf_nil HeqmVW' ander).

(* sigma *)
intros G T1 T2 k1 HofT1 IH1 Hfst1 HofT2 IH2 E Hwf HintE V V' V'' V''' Q Q' c c' k Hfstsigma HfstV HfstV' HeqmV HeqmV' HvalV'' HvalV''' HintV Hintc.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
invert Hfstsigma; clear Hfstsigma; [].
intros k1a k2 Hfst1a Hfst2 <-.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
simpl in Hintc.
invert HintV; clear HintV; [].
intros V1 V1' V2 V2' c1 c1' k1a k2a P Hof Hof' HvalV1 HvalV2 HvalV1' HvalV2' HfstV1 HfstV1' Hfst1a Hfst2a HintV1 HintV2 <- <- <-.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a; clear Hfst2a.
invert Hintc; clear Hintc; [].
intros Pa P' _ _ Hintc1 Hintc2 Heq <-.
so (cand_inject _#3 Heq); subst Pa; clear Heq.
invert HfstV; [].
intros c1a c2 HfstV1a HfstV2 <-.
so (fstm_fun _#4 HfstV1 HfstV1a); subst c1a; clear HfstV1a.
invert HfstV'; [].
intros c1a' c2' HfstV1a' HfstV2' <-.
so (fstm_fun _#4 HfstV1' HfstV1a'); subst c1a'; clear HfstV1a'.
so (Progress.canon_md _#3 (ceqm_reg _#5 wf_nil HeqmV ander) HvalV'') as H.
simp_sub in H.
invert H; clear H; [].
intros V1'' V2'' <-.
so (Progress.canon_md _#3 (ceqm_reg _#5 wf_nil HeqmV' ander) HvalV''') as H.
simp_sub in H.
invert H; clear H; [].
intros V1''' V2''' <-.
so (ceqm_fst _#4 wf_nil HeqmV) as (_ & _ & c'' & _ & _ & H & _).
invert H; clear H; [].
intros c1'' c2'' HfstV1'' HfstV2'' <-.
so (ceqm_fst _#4 wf_nil HeqmV') as (_ & _ & c''' & _ & _ & H & _).
invert H; clear H; [].
intros c1''' c2''' HfstV1''' HfstV2''' <-.
invert HvalV''; clear HvalV''; [].
intros HvalV1'' HvalV2''.
invert HvalV'''; clear HvalV'''; [].
intros HvalV1''' HvalV2'''.
simpl.
simp_sub in Hof.
simp_sub in Hof'.
so (Inversion.pair_ofm_inversion_sigma_pure' _#6 wf_nil (Preservation.mvalue_pure _#4 (mvalue_pair HvalV1 HvalV2) Hof) HfstV1) as (HofV1 & HofV2).
so (Inversion.pair_ofm_inversion_sigma_pure' _#6 wf_nil (Preservation.mvalue_pure _#4 (mvalue_pair HvalV1' HvalV2') Hof') HfstV1') as (HofV1' & HofV2').
simp_sub in HofV2.
simp_sub in HofV2'.
so (ofm_fst _#6 wf_nil HofV1 HfstV1 (subst_fsts _#3 Hfst1)) as Hofc1.
so (ofm_fst _#6 wf_nil HofV1' HfstV1' (subst_fsts _#3 Hfst1)) as Hofc1'.
so (ofm_fst _#6 wf_nil HofV2 HfstV2 (subst_fsts _#3 Hfst2)) as Hofc2.
so (ofm_fst _#6 wf_nil HofV2' HfstV2' (subst_fsts _#3 Hfst2)) as Hofc2'.
assert (ceqm nil pure V1 V1'' (subst (left E) T1)) as HeqmV1.
  #2 apply (anti_msteps_ceqm_trans _#3 (md_pi1 (md_pair V1 V2))).
    apply star_one; [].
    apply mstep_pi12; auto; done.
    
    #2 apply (ceqm_msteps_trans _#3 (md_pi1 (md_pair V1'' V2''))).
      apply ceqm_comp; [].
      simp_sub in HeqmV.
      eapply mcc_pi1; eauto; done.

      apply star_one; [].
      apply mstep_pi12; auto; done.
assert (ceqm nil pure V1' V1''' (subst (right E) T1)) as HeqmV1'.
  #2 apply (anti_msteps_ceqm_trans _#3 (md_pi1 (md_pair V1' V2'))).
    apply star_one; [].
    apply mstep_pi12; auto; done.
    
    #2 apply (ceqm_msteps_trans _#3 (md_pi1 (md_pair V1''' V2'''))).
      apply ceqm_comp; [].
      simp_sub in HeqmV'.
      eapply mcc_pi1; eauto; done.

      apply star_one; [].
      apply mstep_pi12; auto; done.
#4 eapply sg_sigma_interp_; eauto.
  exact (ofm_forget (ceqm_reg _#5 wf_nil HeqmV ander)).

  exact (ofm_forget (ceqm_reg _#5 wf_nil HeqmV' ander)).

  eapply (IH1 _ Hwf HintE V1 V1' _ _ (Cand (simp k1) (fst P))); eauto; [].
  eapply intk_defeq; eauto with static; done.

  assert (intctx (G; cl_cn k1) (E; Sta c1 c1' (Cand (simp k1) (fst P))) (E; Sta c1'' c1''' (Cand (simp k1) (fst P')))) as HintEx.
    #3 apply intctx_cn; auto.
      exact (ceqm_fst' _#7 wf_nil HeqmV1 (subst_fsts _#3 Hfst1) HfstV1 HfstV1'').

      exact (ceqm_fst' _#7 wf_nil HeqmV1' (subst_fsts _#3 Hfst1) HfstV1' HfstV1''').

      eapply intk_defeq; eauto with static; done.
  rewrite <- (ints_fun (G; cl_cn k1) T2 (E; Sta c1 c1' (Cand (simp k1) (fst P)))); auto with static; [].
  #3 eapply (IH2 _ (wf_cn Hwf Hofk1) (intctx_refl _#3 (wf_cn Hwf Hofk1) HintEx) V2 V2' _ _ (Cand (simp k2) (snd P))); eauto.
    #2 apply (anti_msteps_ceqm_trans _#3 (md_pi2 (md_pair V1 V2))).
      apply star_one; [].
      apply mstep_pi22; auto; done.
      
      #2 apply (ceqm_msteps_trans _#3 (md_pi2 (md_pair V1'' V2''))).
        apply ceqm_comp; [].
        simp_sub in HeqmV.
        simpl left.
        #2 apply (mcc_subsume _#7 (subst (dot (cn_pi1 (cn_pair c1 c2)) (left E)) T2)).
          rewrite -> split_dot_compose; [].
          apply ceqm_comp; [].
          eapply mcc_pi2; eauto; done.

          apply subs_refl; [].
          apply eqs_symm; [].
          rewrite -> !(split_dot_compose _ (left E)); [].
          #2 apply (seq_ofs _#3 (nil; cl_cn (subst (left E) k1))); auto using wf_nil.
            apply subrel_dot; auto using subrel_id; [].
            simp_sub.
            apply seq_cn; eauto with static; done.
            
            eapply subst_ofs; eauto; [].
            apply subof_bind_sof_cn; eauto using left_subof; done.

        apply star_one; [].
        apply mstep_pi22; auto; done.

    #2 apply (anti_msteps_ceqm_trans _#3 (md_pi2 (md_pair V1' V2'))).
      apply star_one; [].
      apply mstep_pi22; auto; done.
      
      #2 apply (ceqm_msteps_trans _#3 (md_pi2 (md_pair V1''' V2'''))).
        apply ceqm_comp; [].
        simp_sub in HeqmV'.
        simpl right.
        #2 apply (mcc_subsume _#7 (subst (dot (cn_pi1 (cn_pair c1' c2')) (right E)) T2)).
          rewrite -> split_dot_compose; [].
          apply ceqm_comp; [].
          eapply mcc_pi2; eauto; done.

          apply subs_refl; [].
          apply eqs_symm; [].
          rewrite -> !(split_dot_compose _ (right E)); [].
          #2 apply (seq_ofs _#3 (nil; cl_cn (subst (right E) k1))); auto using wf_nil.
            apply subrel_dot; auto using subrel_id; [].
            simp_sub.
            apply seq_cn; eauto with static; done.
            
            eapply subst_ofs; eauto; [].
            apply subof_bind_sof_cn; eauto using right_subof; done.

        apply star_one; [].
        apply mstep_pi22; auto; done.

    so (ofs_fst _#3 HofT2 Hfst2) as Hofk2.
    #2 rewrite <- (intk_fun (G; cl_cn k1) k2 (E; Sta c1 c1' (Cand (simp k1) (fst P)))) in Hintc2; auto with static.
      eapply (intk_defeq (G; cl_cn k1)); eauto with static; done.

      apply intctx_cn; eauto with static; [].
      eapply (intk_refl _#7 (Cand (simp k1) (fst P'))); eauto; [].
      eapply intk_defeq; eauto with static; done.
Qed.


Lemma ints_equiv_cand :
  forall G T E V V' Q Q' c c' k,
    wf G
    -> ofs G T
    -> intctx G E E
    -> fsts T k
    -> fstm nil V c
    -> fstm nil V' c'
    -> ints T E V V' Q
    -> intk k E c c' Q Q'
    -> ints T E V V' Q'.
Proof.
intros G T E V V' Q Q' c c' k Hwf HofT HintE Hfst HfstV HfstV' HV Heqcand.
so (ints_reg _#5 HV) as (HofV & HofV').
so (ints_val _#6 Hwf HofT HintE HV) as (Hval & Hval').
eapply (ints_ceqm _ _ _ V V' V V'); eauto using ceqm_refl; done.
Qed.


Lemma eqs_ints :
  forall G E T T',
    wf G
    -> intctx G E E
    -> eqs G T T'
    -> ints T E = ints T' E.
Proof.
intros G E T T' Hwf HintE Heqs.
revert E Hwf HintE.
induct Heqs.

(* refl *)
intros G T Hof E Hwf HintE.
reflexivity.

(* symm *)
intros G T T' _ IH E Hwf HintE.
symmetry; auto; done.

(* trans *)
intros G T1 T2 T3 _ IH1 _ IH2 E Hwf HintE.
transitivity (ints T2 E); auto; done.

(* stat *)
intros G k k' Heqk E Hwf HintE.
simpl.
fextensionality 3.
intros V V' Q.
pextensionality.
  intro H; invert H; clear H; [].
  intros c c' Hc <- <-.
  apply sg_stat_interp_; auto; [].
  unfold intks.
  unfold intks in Hc.
  erewrite -> eqk_intk in Hc; eauto; done.

  intro H; invert H; clear H; [].
  intros c c' Hc <- <-.
  apply sg_stat_interp_; auto; [].
  unfold intks.
  unfold intks in Hc.
  erewrite <- eqk_intk in Hc; eauto; done.

(* dyn *)
intros G t t' Heqt E Hwf HintE.
simpl.
fextensionality 3.
intros V V' Q.
pextensionality.
  intro H; invert H; clear H; [].
  intros v v' Hof Hof' Hv <- <- <-.
  #3 apply sg_dyn_interp_; auto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_dyn; [].
    eapply subst_eqc_type; eauto using left_subof; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_dyn; [].
    eapply subst_eqc_type; eauto using right_subof; done.

    erewrite <- eqc_intt; eauto; done.

  intro H; invert H; clear H; [].
  intros v v' Hof Hof' Hv <- <- <-.
  #3 apply sg_dyn_interp_; auto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    apply eqs_dyn; [].
    eapply subst_eqc_type; eauto using left_subof; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    apply eqs_dyn; [].
    eapply subst_eqc_type; eauto using right_subof; done.

    erewrite -> eqc_intt; eauto; done.

(* pigen *)
intros G T1 T1' k1 T2 T2' Heqs1 IH1 Hfst1 Heqs2 IH2 E Hwf HintE.
so (ofs_fst _#3 (eqs_reg _#3 Hwf Heqs1 andel) Hfst1) as Hofk1.
fextensionality 3.
intros V V' Q.
simpl.
pextensionality.
  intro H; invert H; clear H; [].
  intros HofV HofV' HvalV HvalV' Hrel <-.
  #3 apply sg_pigen_interp_; auto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply subst_eqs; eauto using left_subof with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply subst_eqs; eauto using right_subof with static; done.

    intros W W' Q HW.
    rewrite <- (IH1 _ Hwf HintE) in HW; [].
    so (ints_reg _#5 HW) as (HofW & HofW').
    so (Hrel _ _ _ HW) as (c & c' & HfstW & HfstW' & HVW).
    so (ints_fst _#9 Hwf (eqs_reg _#3 Hwf Heqs1 andel) (intctx_refl _#3 Hwf HintE) Hfst1 HfstW HfstW' HW) as Hint.
    exists c, c'.
    do2 2 split; auto; [].
    so (ofm_fst _#6 wf_nil HofW HfstW (subst_fsts _#3 Hfst1)) as Hofc.
    so (ofm_fst _#6 wf_nil HofW' HfstW' (subst_fsts _#3 Hfst1)) as Hofc'.
    rewrite <- (IH2 (E; Sta c c' Q)); auto using intctx_cn with static; [].
    #2 erewrite <- (miev_eqs _#5); [eexact HVW | ..].
      eapply subst_eqs; eauto; [].
      apply subof_dot; eauto using left_subof; [].
      simp_sub.
      apply sof_cn; auto; done.

      eapply subst_eqs; eauto; [].
      apply subof_dot; eauto using right_subof; [].
      simp_sub.
      apply sof_cn; auto; done.

  intro H; invert H; clear H; [].
  intros HofV HofV' HvalV HvalV' Hrel <-.
  #3 apply sg_pigen_interp_; auto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply subst_eqs; eauto using left_subof with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply subst_eqs; eauto using right_subof with static; done.

    intros W W' Q HW.
    so (ints_reg _#5 HW) as (HofW & HofW').
    so HW as HW'.
    rewrite -> (IH1 _ Hwf HintE) in HW'; [].
    so (Hrel _ _ _ HW') as (c & c' & HfstW & HfstW' & HVW).
    clear HW'.
    so (ints_fst _#9 Hwf (eqs_reg _#3 Hwf Heqs1 andel) (intctx_refl _#3 Hwf HintE) Hfst1 HfstW HfstW' HW) as Hint.
    exists c, c'.
    do2 2 split; auto; [].
    so (ofm_fst _#6 wf_nil HofW HfstW (subst_fsts _#3 Hfst1)) as Hofc.
    so (ofm_fst _#6 wf_nil HofW' HfstW' (subst_fsts _#3 Hfst1)) as Hofc'.
    rewrite -> (IH2 (E; Sta c c' Q)); auto using intctx_cn with static; [].
    #2 erewrite <- (miev_eqs _#5); [eexact HVW | ..].
      apply eqs_symm; [].
      eapply subst_eqs; eauto; [].
      apply subof_dot; eauto using left_subof; [].
      simp_sub.
      apply sof_cn; auto; done.

      apply eqs_symm; [].
      eapply subst_eqs; eauto; [].
      apply subof_dot; eauto using right_subof; [].
      simp_sub.
      apply sof_cn; auto; done.

(* piapp *)
intros G T1 T1' k1 T2 T2' Heqs1 IH1 Hfst1 Heqs2 IH2 E Hwf HintE.
so (eqs_fst_o _#3 Hwf Heqs1) as (k1a & k1' & Hfst1a & Hfst1' & Heqk1).
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (eqk_reg _#3 Hwf Heqk1) as (Hofk1 & _).
so (eqs_fst_o _#3 (wf_cn Hwf Hofk1) Heqs2) as (k2 & k2' & Hfst2 & Hfst2' & Heqk2).
so (eqk_reg _#3 (wf_cn Hwf Hofk1) Heqk2 andel) as Hofk2.
fextensionality 3.
intros V V' Q.
simpl.
pextensionality.
  intro H; invert H; clear H; [].
  intros c c' k1a k2a F HofV HofV' HvalV HvalV' HfstV HfstV' Hfst1a Hfst2a Hintk Hrel <-.
  so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
  so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a; clear Hfst2a.
  revert F Hintk Hrel.
  rewrite -> (eqk_simp _#3 Heqk1); [].
  rewrite -> (eqk_simp _#3 Heqk2); [].
  intros F Hintk Hrel.
  #4 eapply sg_piapp_interp_; eauto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply subst_eqs; eauto using left_subof with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply subst_eqs; eauto using right_subof with static; done.
    unfold intks in Hintk; unfold intks.
    rewrite <- (eqk_intk _#5 Hwf (eqk_pi Heqk1 Heqk2) HintE); [].
    assumption.

    intros W W' Q HW.
    rewrite <- (IH1 _ Hwf HintE) in HW; [].
    so (Hrel _#3 HW) as (d & d' & HfstW & HfstW' & HVW).
    so (ints_fst _#9 Hwf (eqs_reg _#3 Hwf Heqs1 andel) (intctx_refl _#3 Hwf HintE) Hfst1 HfstW HfstW' HW) as Hd.
    so (intk_reg _#6 Hd) as (Hofd & Hofd').
    exists d, d'.
    do2 2 split; auto; [].
    assert (intctx (G; cl_cn k1) (E; Sta d d' Q) (E; Sta d d' Q)) as HintEx by (apply intctx_cn; auto with static).
    rewrite <- (IH2 _ (wf_cn Hwf Hofk1) HintEx); [].
    #2 erewrite <- mpuev_eqs; eauto.
      eapply subst_eqs; eauto; [].
      apply subof_dot; eauto using left_subof; [].
      simp_sub.
      apply sof_cn; auto; done.

      eapply subst_eqs; eauto; [].
      apply subof_dot; eauto using right_subof; [].
      simp_sub.
      apply sof_cn; auto; done.

  intro H; invert H; clear H; [].
  intros c c' k1a' k2a' F HofV HofV' HvalV HvalV' HfstV HfstV' Hfst1a' Hfst2a' Hintk Hrel <-.
  so (fsts_fun _#3 Hfst1' Hfst1a'); subst k1a'; clear Hfst1a'.
  so (fsts_fun _#3 Hfst2' Hfst2a'); subst k2a'; clear Hfst2a'.
  revert F Hintk Hrel.
  rewrite <- (eqk_simp _#3 Heqk1); [].
  rewrite <- (eqk_simp _#3 Heqk2); [].
  intros F Hintk Hrel.
  #4 eapply sg_piapp_interp_; eauto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    eapply subst_eqs; eauto using left_subof with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    eapply subst_eqs; eauto using right_subof with static; done.
    unfold intks in Hintk; unfold intks.
    rewrite -> (eqk_intk _#5 Hwf (eqk_pi Heqk1 Heqk2) HintE); [].
    assumption.

    intros W W' Q HW.
    so HW as HW'.
    rewrite -> (IH1 _ Hwf HintE) in HW'; [].
    so (Hrel _#3 HW') as (d & d' & HfstW & HfstW' & HVW).
    clear HW'.
    so (ints_fst _#9 Hwf (eqs_reg _#3 Hwf Heqs1 andel) (intctx_refl _#3 Hwf HintE) Hfst1 HfstW HfstW' HW) as Hd.
    so (intk_reg _#6 Hd) as (Hofd & Hofd').
    exists d, d'.
    do2 2 split; auto; [].
    assert (intctx (G; cl_cn k1) (E; Sta d d' Q) (E; Sta d d' Q)) as HintEx by (apply intctx_cn; auto with static).
    rewrite -> (IH2 _ (wf_cn Hwf Hofk1) HintEx); [].
    #2 erewrite <- mpuev_eqs; eauto.
      apply eqs_symm; [].
      eapply subst_eqs; eauto; [].
      apply subof_dot; eauto using left_subof; [].
      simp_sub.
      apply sof_cn; auto; done.

      apply eqs_symm; [].
      eapply subst_eqs; eauto; [].
      apply subof_dot; eauto using right_subof; [].
      simp_sub.
      apply sof_cn; auto; done.

(* sigma *)
intros G T1 T1' k1 T2 T2' Heqs1 IH1 Hfst1 Heqs2 IH2 E Hwf HintE.
so (eqs_fst_o _#3 Hwf Heqs1) as (k1a & k1' & Hfst1a & Hfst1' & Heqk1).
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (eqk_reg _#3 Hwf Heqk1) as (Hofk1 & _).
so (eqs_fst_o _#3 (wf_cn Hwf Hofk1) Heqs2) as (k2 & k2' & Hfst2 & Hfst2' & Heqk2).
so (eqk_reg _#3 (wf_cn Hwf Hofk1) Heqk2 andel) as Hofk2.
fextensionality 3.
intros V V' Q.
pextensionality.
  intro H; invert H; clear H; [].
  intros V1 V1' V2 V2' c1 c1' k1a k2a P Hof Hof' HvalV1 HvalV2 HvalV1' HvalV2' HfstV1 HfstV1' Hfst1a Hfst2a HintV1 HintV2 <- <- <-.
  so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
  so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a; clear Hfst2a.
  simpl.
  revert P HintV1 HintV2.
  rewrite -> (eqk_simp _#3 Heqk1); [].
  rewrite -> (eqk_simp _#3 Heqk2); [].
  intros P HintV1 HintV2.
  #4 eapply sg_sigma_interp_; eauto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply subst_eqs; eauto using left_subof with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    eapply subst_eqs; eauto using right_subof with static; done.

    rewrite <- (IH1 _ Hwf HintE); [].
    assumption.

    assert (intctx (G; cl_cn k1) (E; Sta c1 c1' (Cand (simp k1') (fst P))) (E; Sta c1 c1' (Cand (simp k1') (fst P)))) as HintEx.
      #3 apply intctx_cn; auto with static.
        simp_sub in Hof.
        exact (eqc_refl (ofm_fst _#6 wf_nil (Inversion.pair_ofm_inversion_1 _#6 wf_nil Hof) HfstV1 (subst_fsts _#3 Hfst1))).

        simp_sub in Hof'.
        exact (eqc_refl (ofm_fst _#6 wf_nil (Inversion.pair_ofm_inversion_1 _#6 wf_nil Hof') HfstV1' (subst_fsts _#3 Hfst1))).

        exact (ints_fst _#9 Hwf (eqs_reg _#3 Hwf Heqs1 andel) (intctx_refl _#3 Hwf HintE) Hfst1 HfstV1 HfstV1' HintV1).
    rewrite <- (IH2 _ (wf_cn Hwf Hofk1) HintEx); [].
    assumption.

  intro H; invert H; clear H; [].
  intros V1 V1' V2 V2' c1 c1' k1a' k2a' P Hof Hof' HvalV1 HvalV2 HvalV1' HvalV2' HfstV1 HfstV1' Hfst1a' Hfst2a' HintV1 HintV2 <- <- <-.
  so (fsts_fun _#3 Hfst1' Hfst1a'); subst k1a'; clear Hfst1a'.
  so (fsts_fun _#3 Hfst2' Hfst2a'); subst k2a'; clear Hfst2a'.
  simpl.
  revert P HintV1 HintV2.
  rewrite <- (eqk_simp _#3 Heqk1); [].
  rewrite <- (eqk_simp _#3 Heqk2); [].
  intros P HintV1 HintV2.
  #4 eapply sg_sigma_interp_; eauto.
    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    eapply subst_eqs; eauto using left_subof with static; done.

    eapply ofm_subsume; eauto; [].
    apply subs_refl; [].
    apply eqs_symm; [].
    eapply subst_eqs; eauto using right_subof with static; done.

    rewrite -> (IH1 _ Hwf HintE); [].
    assumption.

    assert (intctx (G; cl_cn k1) (E; Sta c1 c1' (Cand (simp k1) (fst P))) (E; Sta c1 c1' (Cand (simp k1) (fst P)))) as HintEx.
      #3 apply intctx_cn; auto with static.
        simp_sub in Hof.
        apply eqc_refl; [].
        so (ofm_fst _#6 wf_nil (Inversion.pair_ofm_inversion_1 _#6 wf_nil Hof) HfstV1 (subst_fsts _#3 Hfst1')) as Hofc.
        eapply ofc_subsume; eauto; [].
        apply subk_refl; apply eqk_symm; [].
        eapply subst_eqk; eauto using left_subof; done.

        simp_sub in Hof'.
        apply eqc_refl; [].
        so (ofm_fst _#6 wf_nil (Inversion.pair_ofm_inversion_1 _#6 wf_nil Hof') HfstV1' (subst_fsts _#3 Hfst1')) as Hofc.
        eapply ofc_subsume; eauto; [].
        apply subk_refl; apply eqk_symm; [].
        eapply subst_eqk; eauto using right_subof; done.

        rewrite <- (IH1 _ Hwf HintE) in HintV1; [].
        exact (ints_fst _#9 Hwf (eqs_reg _#3 Hwf Heqs1 andel) (intctx_refl _#3 Hwf HintE) Hfst1 HfstV1 HfstV1' HintV1).
    rewrite -> (IH2 _ (wf_cn Hwf Hofk1) HintEx); [].
    assumption.
Qed.


Lemma subs_ints :
  forall G E T T',
    wf G
    -> intctx G E E
    -> subs G T T'
    -> forall V V' Q, ints T E V V' Q -> ints T' E V V' Q.
Proof.
intros G E T T' Hwf HintE Hsubs.
revert E Hwf HintE.
induct Hsubs.

(* refl *)
intros G T T' Heqs E Hwf HintE V V' Q HV.
rewrite <- (eqs_ints _#4 Hwf HintE Heqs); [].
assumption.

(* trans *)
auto; done.

(* stat *)
intros G k k' Hsubk E Hwf HintE V V' Q HV.
simpl in HV.
invert HV; clear HV; [].
intros c c' Hintc <- <-.
simpl.
apply sg_stat_interp_; auto; [].
unfold intks in Hintc; unfold intks.
eapply subk_intk; eauto; done.

(* pigen *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 IH1 Hfst1' Hsubs2 IH2 Hfst1 HofsT2 E Hwf HintE V V' Q HV.
so (subs_reg _#3 Hwf Hsubs1) as (HofT1' & HofT1).
so (ofs_fst _#3 HofT1' Hfst1') as Hofk1'.
simpl in HV.
invert HV; clear HV; [].
intros HofV HofV' HvalV HvalV' Hrel <-.
simpl.
#3 apply sg_pigen_interp_; auto.
  eapply ofm_subsume; eauto; [].
  eapply subst_subs; eauto using left_subof with static; done.

  eapply ofm_subsume; eauto; [].
  eapply subst_subs; eauto using right_subof with static; done.

  intros W W' Q HW.
  so (IH1 _ Hwf HintE _#3 HW) as HW'.
  so (ints_reg _#5 HW) as (HofW & HofW').
  so (Hrel _ _ _ HW') as (c & c' & HfstW & HfstW' & HVW).
  so (ints_fst _#9 Hwf HofT1' (intctx_refl _#3 Hwf HintE) Hfst1' HfstW HfstW' HW) as Hint.
  exists c, c'.
  do2 2 split; auto; [].
  so (ofm_fst _#6 wf_nil HofW HfstW (subst_fsts _#3 Hfst1')) as Hofc.
  so (ofm_fst _#6 wf_nil HofW' HfstW' (subst_fsts _#3 Hfst1')) as Hofc'.
  unfold miev.
  unfold miev in HVW.
  assert (subs nil (subst (dot c (left E)) T2) (subst (dot c (left E)) T2')) as Hsubs2x.
    eapply subst_subs; eauto; [].
    apply subof_dot; eauto using left_subof; [].
    simp_sub.
    apply sof_cn; auto; done.
  assert (subs nil (subst (dot c' (right E)) T2) (subst (dot c' (right E)) T2')) as Hsubs2x'.
    eapply subst_subs; eauto; [].
    apply subof_dot; eauto using right_subof; [].
    simp_sub.
    apply sof_cn; auto; done.
  eapply mev_mono_sg; eauto; [].
  refine (mev_mono _#7 HVW); [].
  clear W W' HW HW' HofW HofW' HfstW HfstW' HVW.
  intros M M' HM.
  eapply mi_mono_sg; eauto; [].
  refine (mi_mono _#7 HM); [].
  intros N N' S HN.
  apply IH2; auto with static; [].
  apply intctx_cn; auto with static; done.

(* piapp *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 IH1 Hfst1' Hsubs2 IH2 Hfst1 HofT2 E Hwf HintE V V' Q HV.
so (subs_reg _#3 Hwf Hsubs1) as (HofT1' & HofT1).
so (ofs_fst _#3 HofT1' Hfst1') as Hofk1'.
simpl in HV.
invert HV; clear HV; [].
intros c c' k1a k2 F HofV HofV' HvalV HvalV' HfstV HfstV' Hfst1a Hfst2 Hintc Hrel <-.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (subs_fst _#5 Hwf Hsubs1 Hfst1' Hfst1) as Hsubk1.
so (subs_fst_o _#3 (wf_cn Hwf Hofk1') Hsubs2) as (k2a & k2' & Hfst2a & Hfst2' & Hsubk2).
so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a; clear Hfst2a.
revert F Hintc Hrel.
rewrite <- (subk_simp _#3 Hsubk1); [].
rewrite -> (subk_simp _#3 Hsubk2); [].
intros F Hintc Hrel.
simpl.
#4 apply (sg_piapp_interp_ _#7 c c'); auto.
  eapply ofm_subsume; eauto using subst_subs, left_subof with static; done.

  eapply ofm_subsume; eauto using subst_subs, right_subof with static; done.

  eapply subk_intk; eauto; [].
  apply subk_pi; eauto; [].
  eapply ofs_fst; eauto; done.

  intros W W' Q HW.
  so (IH1 _ Hwf HintE _ _ _ HW) as HW'.
  so (Hrel _#3 HW') as (d & d' & HfstW & HfstW' & HVW).
  exists d, d'.
  do2 2 split; auto; [].
  so (ints_reg _#5 HW) as (HofW & HofW').
  so (ofm_fst _#6 wf_nil HofW HfstW (subst_fsts _#3 Hfst1')) as Hofd.
  so (ofm_fst _#6 wf_nil HofW' HfstW' (subst_fsts _#3 Hfst1')) as Hofd'.
  assert (subs nil (subst (dot d (left E)) T2) (subst (dot d (left E)) T2')) as Hsubs2x.
    eapply subst_subs; eauto; [].
    apply subof_dot; eauto using left_subof; [].
    simp_sub.
    apply sof_cn; auto; done.
  assert (subs nil (subst (dot d' (right E)) T2) (subst (dot d' (right E)) T2')) as Hsubs2x'.
    eapply subst_subs; eauto; [].
    apply subof_dot; eauto using right_subof; [].
    simp_sub.
    apply sof_cn; auto; done.
  eapply mpuev_mono_sg; eauto; [].
  refine (mpuev_mono _#8 HVW); [].
  intros N N' S HN.
  apply IH2; auto with static; [].
  apply intctx_cn; auto with static; [].
  apply (ints_fst G T1' E W W'); auto; done.

(* sigma *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 IH1 Hfst1 Hsubs2 IH2 Hfst1' HofT2' E Hwf HintE V V' Q HV.
so (subs_reg _#3 Hwf Hsubs1) as (HofT1 & HofT1').
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
simpl in HV.
invert HV; clear HV; [].
intros V1 V1' V2 V2' c1 c1' k1a k2 P Hof Hof' HvalV1 HvalV2 HvalV1' HvalV2' HfstV1 HfstV1' Hfst1a Hfst2 Hrel1 Hrel2 <- <- <-.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a; clear Hfst1a.
so (subs_fst _#5 Hwf Hsubs1 Hfst1 Hfst1') as Hsubk1.
so (subs_fst_o _#3 (wf_cn Hwf Hofk1) Hsubs2) as (k2a & k2' & Hfst2a & Hfst2' & Hsubk2).
so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a; clear Hfst2a.
revert P Hrel1 Hrel2.
rewrite -> (subk_simp _#3 Hsubk1); [].
rewrite -> (subk_simp _#3 Hsubk2); [].
intros P Hrel1 Hrel2.
simpl.
#3 eapply sg_sigma_interp_; eauto.
  eapply ofm_subsume; eauto; [].
  eapply subst_subs; eauto using left_subof with static; done.

  eapply ofm_subsume; eauto; [].
  eapply subst_subs; eauto using right_subof with static; done.
  apply IH2; auto with static; [].
  so (ints_fst _#9 Hwf HofT1 HintE Hfst1 HfstV1 HfstV1' Hrel1) as Hintk.
  so (intk_reg _#6 Hintk) as (Hofc1 & Hofc1').
  apply intctx_cn; auto with static; done.
Qed.
