
Require Import Tactics.
Require Import Extensionality.
Require Import Sequence.
Require Import Relation.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Formation.
Require Import Regularity.
Require Import Declarative.
Require Import Preservation.
Require Import Progress.
Require Import Inversion.
Require Import Determinism.
Require Import Closure.
Require Import Biorthogonal.
Require Singleton.
Require Compactness.


(* Ev monad *)

Lemma ev_intro_steps :
  forall value step e e' v v' (R : term -> term -> Prop),
    deterministic value step
    -> star step e v
    -> star step e' v'
    -> value v
    -> value v'
    -> R v v'
    -> ev value step R e e'.
Proof.
intros value step e e' v v' R Hdeterm Hstepsv Hstepsv' Hvalv Hvalv' HRv.
split.
  split; intro; eexists; eauto; done.

  intros w w' Hstepsw Hstepsw' Hvalw Hvalw'.
  so (value_unique _ _ Hdeterm _ _ _ Hstepsv Hstepsw Hvalv Hvalw); subst w.
  so (value_unique _ _ Hdeterm _ _ _ Hstepsv' Hstepsw' Hvalv' Hvalw'); subst w'.
  assumption.
Qed.


Definition congruous (ok value : term -> Prop) (step : term -> term -> Prop) :=
  (forall m m', ok m -> step m m' -> ok m')
  /\ (forall m, ok m -> value m \/ exists m', step m m')
  /\ deterministic value step.


Lemma halt_evalctx :
  forall ok value1 value2 step1 step2 F,
    congruous ok value1 step1
    -> deterministic value2 step2
    -> (forall m m', step1 m m' -> step2 (F m) (F m'))
    -> forall m, ok m -> halt value2 step2 (F m) -> halt value1 step1 m.
Proof.
intros ok value1 value2 step1 step2 F (Hpres1 & Hprog1 & _) (Hdetv2 & Hdet2) HstepF m Hok Hhalt.
destruct Hhalt as (v & Hsteps & Hval).
remember (F m) as d eqn:Heqd.
revert m Hok Hval Heqd.
induct Hsteps.

(* refl *)
intros d m Hok Hval2 ->.
so (Hprog1 _ Hok) as [Hval1 | (n & Hstep)].
  exists m.
  auto using star_refl; done.

  exfalso.
  so (HstepF _ _ Hstep) as Hstep'.
  eapply Hdetv2; eauto; done.

(* step *)
intros d e f Hstep2 _ IH m Hok Hval ->.
so (Hprog1 _ Hok) as [Hval1 | (n & Hstep1)].
  exists m; auto using star_refl; done.

  so (HstepF _ _ Hstep1) as Hstep2'.
  so (Hdet2 _ _ _ Hstep2 Hstep2'); subst e.
  so (Hpres1 _ _ Hok Hstep1) as Hok'.
  so (IH _ Hok' Hval (eq_refl _)) as (o & Hsteps & Hvalo).
  exists o; eauto using star_step; done.
Qed.


Lemma monad_ev_ev :
  forall ok value1 value2 step1 step2 (R1 R2 : term -> term -> Prop) F F',
    congruous ok value1 step1
    -> deterministic value2 step2
    -> (forall m m', step1 m m' -> step2 (F m) (F m'))
    -> (forall m m', step1 m m' -> step2 (F' m) (F' m'))
    -> (forall v v', R1 v v' -> ev value2 step2 R2 (F v) (F' v'))
    -> (forall m m', ok m -> ok m' -> ev value1 step1 R1 m m' -> ev value2 step2 R2 (F m) (F' m')).
Proof.
intros ok value1 value2 step1 step2 R1 R2 F F' Hcong1 Hdeterm2.
revert R1 R2 F F'.
assert (forall (R1 R2 : term -> term -> Prop) F F',
          (forall m m', step1 m m' -> step2 (F m) (F m'))
          -> (forall m m', step1 m m' -> step2 (F' m) (F' m'))
          -> (forall v v', R1 v v' -> ev value2 step2 R2 (F v) (F' v'))
          -> forall m m', ok m -> ok m' -> ev value1 step1 R1 m m' -> halt value2 step2 (F m) -> halt value2 step2 (F' m')) as Hhaltcond.
  intros R1 R2 F F' HstepF HstepF' Hincl m m' Hokm Hofm' Hm HhaltFm.
  destruct Hm as (Hhaltiff & HRval).
  so (halt_evalctx _#6 Hcong1 Hdeterm2 HstepF m Hokm HhaltFm) as Hhaltm.
  so (Hhaltiff andel Hhaltm) as Hhaltm'.
  destruct Hhaltm as (v & Hsteps & Hval).
  destruct Hhaltm' as (v' & Hsteps' & Hval').
  so (HRval _ _ Hsteps Hsteps' Hval Hval') as HRv.
  so (Hincl _ _ HRv) as HRFv.
  destruct HRFv as (HhaltiffFv & _).
  lapply (HhaltiffFv andel).
    intro HhaltFv'.
    eapply anti_steps_halt; eauto; [].
    eapply star_map; eauto; done.

    eapply steps_halt; eauto; [].
    eapply star_map; eauto; done.
intros R1 R2 F F' HstepF HstepF' Hincl m m' Hok Hok' Hm.
assert (halt value2 step2 (F m) <-> halt value2 step2 (F' m')) as HhaltiffFm.
  split.
    intro HhaltFm.
    refine (Hhaltcond R1 R2 F F' _ _ _ m m' _ _ _ _); eauto; done.

    intro HhaltFm'.
    #2 refine (Hhaltcond (transpose R1) (transpose R2) F' F _ _ _ m' m _ _ _ _); eauto.
      intros v v' HR1v.
      unfold transpose in HR1v.
      rewrite -> ev_transpose; [].
      unfold transpose.
      auto; done.

      rewrite -> ev_transpose; [].
      auto.
destruct Hm as (Hhaltiff & HRval).
split; auto; [].
intros w w' HstepsFm HstepsFm' Hvalw Hvalw'.
so (halt_evalctx _#6 Hcong1 Hdeterm2 HstepF m Hok (eintro w (conj HstepsFm Hvalw))) as Hhaltm.
so (Hhaltiff andel Hhaltm) as Hhaltm'.
destruct Hhaltm as (v & Hstepsm & Hvalv).
destruct Hhaltm' as (v' & Hstepsm' & Hvalv').
so (HRval _ _ Hstepsm Hstepsm' Hvalv Hvalv') as HR1v.
so (Hincl _ _ HR1v) as HR2Fv.
destruct HR2Fv as (_ & HR2val).
#2 exploit (HR2val w w'); auto.
  eapply steps_confluent_val; eauto; [].
  eapply star_map; eauto; done.

  eapply steps_confluent_val; eauto; [].
  eapply star_map; eauto; done.
Qed.



Definition okt (e : term) := exists t, oft nil e t.
Definition okm (p : purity) (M : term) := exists T, ofm nil p M T.


Lemma congruous_tm : congruous okt (tvalue nil) (tstep nil).
Proof.
do2 2 split.
  intros e e' (t & Hof) Hstep.
  exists t.
  eapply preservation_tm; eauto using wf_nil; done.

  intros e (t & Hof).
  eapply progress_tm; eauto; done.

  exact (deterministic_tm nil).
Qed.


Lemma congruous_md : forall p, congruous (okm p) (mvalue nil) (mstep nil).
Proof.
do2 2 split.
  intros M M' (T & Hof) Hstep.
  exists T.
  eapply preservation_md; eauto using wf_nil; done.

  intros M (T & Hof).
  eapply progress_md; eauto; done.

  exact (deterministic_md nil).
Qed.


Lemma tev_intro_steps :
  forall t t' e e' v v' (R : term -> term -> Prop),
    oft nil e t
    -> oft nil e' t'
    -> star (tstep nil) e v
    -> star (tstep nil) e' v'
    -> tvalue nil v
    -> tvalue nil v'
    -> R v v'
    -> tev t t' R e e'.
Proof.
intros t t' e e' v v' R Hofe Hofe' Hstepsv Hstepsv' Hvalv Hvalv' HRv.
destruct (ev_intro_steps _#7 (deterministic_tm nil) Hstepsv Hstepsv' Hvalv Hvalv' HRv).
do2 3 split; auto; done.
Qed.


Lemma mev_intro_steps :
  forall t t' e e' v v' (R : term -> term -> Prop),
    ofm nil impure e t
    -> ofm nil impure e' t'
    -> star (mstep nil) e v
    -> star (mstep nil) e' v'
    -> mvalue nil v
    -> mvalue nil v'
    -> R v v'
    -> mev t t' R e e'.
Proof.
intros t t' e e' v v' R Hofe Hofe' Hstepsv Hstepsv' Hvalv Hvalv' HRv.
destruct (ev_intro_steps _#7 (deterministic_md nil) Hstepsv Hstepsv' Hvalv Hvalv' HRv).
do2 3 split; auto; done.
Qed.


Lemma mevp_intro_steps :
  forall t t' e e' v v' (R : term -> term -> Prop),
    ofm nil pure e t
    -> ofm nil pure e' t'
    -> star (mstep nil) e v
    -> star (mstep nil) e' v'
    -> mvalue nil v
    -> mvalue nil v'
    -> R v v'
    -> mevp t t' R e e'.
Proof.
intros t t' e e' v v' R Hofe Hofe' Hstepsv Hstepsv' Hvalv Hvalv' HRv.
destruct (ev_intro_steps _#7 (deterministic_md nil) Hstepsv Hstepsv' Hvalv Hvalv' HRv).
do2 3 split; auto; done.
Qed.


Lemma tev_intro :
  forall t t' v v' (R : term -> term -> Prop),
    oft nil v t
    -> oft nil v' t'
    -> tvalue nil v
    -> tvalue nil v'
    -> R v v'
    -> tev t t' R v v'.
Proof.
intros; eapply tev_intro_steps; eauto using star_refl.
Qed.


Lemma mev_intro :
  forall t t' v v' (R : term -> term -> Prop),
    ofm nil impure v t
    -> ofm nil impure v' t'
    -> mvalue nil v
    -> mvalue nil v'
    -> R v v'
    -> mev t t' R v v'.
Proof.
intros; eapply mev_intro_steps; eauto using star_refl.
Qed.


Lemma mevp_intro :
  forall t t' v v' (R : term -> term -> Prop),
    ofm nil pure v t
    -> ofm nil pure v' t'
    -> mvalue nil v
    -> mvalue nil v'
    -> R v v'
    -> mevp t t' R v v'.
Proof.
intros; eapply mevp_intro_steps; eauto using star_refl.
Qed.

Lemma monad_tev_tev :
  forall t1 t1' t2 t2' (R1 R2 : term -> term -> Prop) F F',
    (forall m, oft nil m t1 -> oft nil (F m) t2)
    -> (forall m, oft nil m t1' -> oft nil (F' m) t2')
    -> (forall m m', tstep nil m m' -> tstep nil (F m) (F m'))
    -> (forall m m', tstep nil m m' -> tstep nil (F' m) (F' m'))
    -> (forall v v', R1 v v' -> tev t2 t2' R2 (F v) (F' v'))
    -> (forall m m', tev t1 t1' R1 m m' -> tev t2 t2' R2 (F m) (F' m')).
Proof.
intros t1 t1' t2 t2' R1 R2 F F' HofF HofF' HstepF HstepF' Hincl m m' Hm.
destruct Hm as (Hiff & HRval & Hofm & Hofm').
assert (ev (tvalue nil) (tstep nil) R2 (F m) (F' m')) as Hev.
  #6 apply (monad_ev_ev okt (tvalue nil) (tvalue nil) (tstep nil) (tstep nil) R1 R2); auto.
    exact congruous_tm.

    exact (deterministic_tm nil).

    intros v v' Hv.
    so (Hincl v v' Hv) as (Hiff' & HRval' & _).
    split; auto; done.

    exists t1; eauto; done.

    exists t1'; eauto; done.

    split; auto; done.
destruct Hev as (Hiff' & HRval').
do2 3 split; auto; done.
Qed.


Lemma monad_mev_mev :
  forall t1 t1' t2 t2' (R1 R2 : term -> term -> Prop) F F',
    (forall m, ofm nil impure m t1 -> ofm nil impure (F m) t2)
    -> (forall m, ofm nil impure m t1' -> ofm nil impure (F' m) t2')
    -> (forall m m', mstep nil m m' -> mstep nil (F m) (F m'))
    -> (forall m m', mstep nil m m' -> mstep nil (F' m) (F' m'))
    -> (forall v v', R1 v v' -> mev t2 t2' R2 (F v) (F' v'))
    -> (forall m m', mev t1 t1' R1 m m' -> mev t2 t2' R2 (F m) (F' m')).
Proof.
intros t1 t1' t2 t2' R1 R2 F F' HofF HofF' HstepF HstepF' Hincl m m' Hm.
destruct Hm as (Hiff & HRval & Hofm & Hofm').
assert (ev (mvalue nil) (mstep nil) R2 (F m) (F' m')) as Hev.
  #6 apply (monad_ev_ev (okm impure) (mvalue nil) (mvalue nil) (mstep nil) (mstep nil) R1 R2); auto.
    exact (congruous_md impure).

    exact (deterministic_md nil).

    intros v v' Hv.
    so (Hincl v v' Hv) as (Hiff' & HRval' & _).
    split; auto; done.

    exists t1; eauto; done.

    exists t1'; eauto; done.

    split; auto; done.
destruct Hev as (Hiff' & HRval').
do2 3 split; auto; done.
Qed.


Lemma monad_mevp_mev :
  forall t1 t1' t2 t2' c c' k1 k1' (R1 R2 : term -> term -> Prop) F F',
    fsts t1 k1
    -> fsts t1' k1'
    -> (forall m, ofm nil pure m t1 -> fstm nil m c -> ofm nil impure (F m) t2)
    -> (forall m, ofm nil pure m t1' -> fstm nil m c' -> ofm nil impure (F' m) t2')
    -> (forall m m', mstep nil m m' -> mstep nil (F m) (F m'))
    -> (forall m m', mstep nil m m' -> mstep nil (F' m) (F' m'))
    -> (forall v v' d d', fstm nil v d -> fstm nil v' d' -> eqc nil c d k1 -> eqc nil c' d' k1' -> R1 v v' -> mev t2 t2' R2 (F v) (F' v'))
    -> (forall m m', fstm nil m c -> fstm nil m' c' -> mevp t1 t1' R1 m m' -> mev t2 t2' R2 (F m) (F' m')).
Proof.
intros t1 t1' t2 t2' c c' k1 k1' R1 R2 F F' Hfst Hfst' HofF HofF' HstepF HstepF' Hincl m m' Hfstm Hfstm' Hm.
destruct Hm as (Hiff & HRval & Hofm & Hofm').
set (R1' := fun w w' => R1 w w' /\ exists d d', fstm nil w d /\ fstm nil w' d' /\ eqc nil c d k1 /\ eqc nil c' d' k1').
assert (ev (mvalue nil) (mstep nil) R2 (F m) (F' m')) as Hev.
  #6 apply (monad_ev_ev (okm pure) (mvalue nil) (mvalue nil) (mstep nil) (mstep nil) R1' R2); auto.
    exact (congruous_md pure).

    exact (deterministic_md nil).

    intros v v' Hv.
    destruct Hv as (Hv & d & d' & Hfstv & Hfstv' & Heqc & Heqc').
    so (Hincl v v' d d' Hfstv Hfstv' Heqc Heqc' Hv) as (Hiff' & HRval' & _).
    split; auto; done.

    exists t1; eauto; done.

    exists t1'; eauto; done.

    split; auto; [].
    intros v v' Hsteps Hsteps' Hval Hval'.
    split; auto; [].
    so (preservations_fst _#7 wf_nil Hofm Hsteps Hfst Hfstm) as (d & Hfstv & Heqc).
    so (preservations_fst _#7 wf_nil Hofm' Hsteps' Hfst' Hfstm') as (d' & Hfstv' & Heqc').
    exists d, d'.
    auto; done.
destruct Hev as (Hiff' & HRval').
do2 3 split; auto; done.
Qed.


Lemma monad_mevp_mevp :
  forall t1 t1' t2 t2' c c' k1 k1' (R1 R2 : term -> term -> Prop) F F',
    fsts t1 k1
    -> fsts t1' k1'
    -> (forall m, ofm nil pure m t1 -> fstm nil m c -> ofm nil pure (F m) t2)
    -> (forall m, ofm nil pure m t1' -> fstm nil m c' -> ofm nil pure (F' m) t2')
    -> (forall m m', mstep nil m m' -> mstep nil (F m) (F m'))
    -> (forall m m', mstep nil m m' -> mstep nil (F' m) (F' m'))
    -> (forall v v' d d', fstm nil v d -> fstm nil v' d' -> eqc nil c d k1 -> eqc nil c' d' k1' -> R1 v v' -> mevp t2 t2' R2 (F v) (F' v'))
    -> (forall m m', fstm nil m c -> fstm nil m' c' -> mevp t1 t1' R1 m m' -> mevp t2 t2' R2 (F m) (F' m')).
Proof.
intros t1 t1' t2 t2' c c' k1 k1' R1 R2 F F' Hfst Hfst' HofF HofF' HstepF HstepF' Hincl m m' Hfstm Hfstm' Hm.
destruct Hm as (Hiff & HRval & Hofm & Hofm').
set (R1' := fun w w' => R1 w w' /\ exists d d', fstm nil w d /\ fstm nil w' d' /\ eqc nil c d k1 /\ eqc nil c' d' k1').
assert (ev (mvalue nil) (mstep nil) R2 (F m) (F' m')) as Hev.
  #6 apply (monad_ev_ev (okm pure) (mvalue nil) (mvalue nil) (mstep nil) (mstep nil) R1' R2); auto.
    exact (congruous_md pure).

    exact (deterministic_md nil).

    intros v v' Hv.
    destruct Hv as (Hv & d & d' & Hfstv & Hfstv' & Heqc & Heqc').
    so (Hincl v v' d d' Hfstv Hfstv' Heqc Heqc' Hv) as (Hiff' & HRval' & _).
    split; auto; done.

    exists t1; eauto; done.

    exists t1'; eauto; done.

    split; auto; [].
    intros v v' Hsteps Hsteps' Hval Hval'.
    split; auto; [].
    so (preservations_fst _#7 wf_nil Hofm Hsteps Hfst Hfstm) as (d & Hfstv & Heqc).
    so (preservations_fst _#7 wf_nil Hofm' Hsteps' Hfst' Hfstm') as (d' & Hfstv' & Heqc').
    exists d, d'.
    auto; done.
destruct Hev as (Hiff' & HRval').
do2 3 split; auto; done.
Qed.


Lemma monad_mev_tev :
  forall t1 t1' t2 t2' (R1 R2 : term -> term -> Prop) F F',
    (forall m, ofm nil impure m t1 -> oft nil (F m) t2)
    -> (forall m, ofm nil impure m t1' -> oft nil (F' m) t2')
    -> (forall m m', mstep nil m m' -> tstep nil (F m) (F m'))
    -> (forall m m', mstep nil m m' -> tstep nil (F' m) (F' m'))
    -> (forall v v', R1 v v' -> tev t2 t2' R2 (F v) (F' v'))
    -> (forall m m', mev t1 t1' R1 m m' -> tev t2 t2' R2 (F m) (F' m')).
Proof.
intros t1 t1' t2 t2' R1 R2 F F' HofF HofF' HstepF HstepF' Hincl m m' Hm.
destruct Hm as (Hiff & HRval & Hofm & Hofm').
assert (ev (tvalue nil) (tstep nil) R2 (F m) (F' m')) as Hev.
  #6 apply (monad_ev_ev (okm impure) (mvalue nil) (tvalue nil) (mstep nil) (tstep nil) R1 R2); auto.
    exact (congruous_md impure).

    exact (deterministic_tm nil).

    intros v v' Hv.
    so (Hincl v v' Hv) as (Hiff' & HRval' & _).
    split; auto; done.

    exists t1; eauto; done.

    exists t1'; eauto; done.

    split; auto; done.
destruct Hev as (Hiff' & HRval').
do2 3 split; auto; done.
Qed.


Lemma monad_tev_mev :
  forall t1 t1' t2 t2' (R1 R2 : term -> term -> Prop) F F',
    (forall m, oft nil m t1 -> ofm nil impure (F m) t2)
    -> (forall m, oft nil m t1' -> ofm nil impure (F' m) t2')
    -> (forall m m', tstep nil m m' -> mstep nil (F m) (F m'))
    -> (forall m m', tstep nil m m' -> mstep nil (F' m) (F' m'))
    -> (forall v v', R1 v v' -> mev t2 t2' R2 (F v) (F' v'))
    -> (forall m m', tev t1 t1' R1 m m' -> mev t2 t2' R2 (F m) (F' m')).
Proof.
intros t1 t1' t2 t2' R1 R2 F F' HofF HofF' HstepF HstepF' Hincl m m' Hm.
destruct Hm as (Hiff & HRval & Hofm & Hofm').
assert (ev (mvalue nil) (mstep nil) R2 (F m) (F' m')) as Hev.
  #6 apply (monad_ev_ev okt (tvalue nil) (mvalue nil) (tstep nil) (mstep nil) R1 R2); auto.
    exact congruous_tm.

    exact (deterministic_md nil).

    intros v v' Hv.
    so (Hincl v v' Hv) as (Hiff' & HRval' & _).
    split; auto; done.

    exists t1; eauto; done.

    exists t1'; eauto; done.

    split; auto; done.
destruct Hev as (Hiff' & HRval').
do2 3 split; auto; done.
Qed.


Lemma monad_tev_mevp :
  forall t1 t1' t2 t2' (R1 R2 : term -> term -> Prop) F F',
    (forall m, oft nil m t1 -> ofm nil pure (F m) t2)
    -> (forall m, oft nil m t1' -> ofm nil pure (F' m) t2')
    -> (forall m m', tstep nil m m' -> mstep nil (F m) (F m'))
    -> (forall m m', tstep nil m m' -> mstep nil (F' m) (F' m'))
    -> (forall v v', R1 v v' -> mevp t2 t2' R2 (F v) (F' v'))
    -> (forall m m', tev t1 t1' R1 m m' -> mevp t2 t2' R2 (F m) (F' m')).
Proof.
intros t1 t1' t2 t2' R1 R2 F F' HofF HofF' HstepF HstepF' Hincl m m' Hm.
destruct Hm as (Hiff & HRval & Hofm & Hofm').
assert (ev (mvalue nil) (mstep nil) R2 (F m) (F' m')) as Hev.
  #6 apply (monad_ev_ev okt (tvalue nil) (mvalue nil) (tstep nil) (mstep nil) R1 R2); auto.
    exact congruous_tm.

    exact (deterministic_md nil).

    intros v v' Hv.
    so (Hincl v v' Hv) as (Hiff' & HRval' & _).
    split; auto; done.

    exists t1; eauto; done.

    exists t1'; eauto; done.

    split; auto; done.
destruct Hev as (Hiff' & HRval').
do2 3 split; auto; done.
Qed.


(* Biorthogonality monad *)

Lemma monad_tpst_tev_tpclosed :
  forall t1 t1' t2 t2' (R1 R2 : term -> term -> Prop) F F',
    tvrel t1 t1' R1
    -> tvrel t2 t2' R2
    -> tpclosed t2 t2' R2
    -> (forall s e, subst s (F e) = F (subst s e))
    -> (forall s e, subst s (F' e) = F' (subst s e))
    -> (forall G i e, length G = i -> oft G e (subst (sh i) t1) -> oft G (F e) (subst (sh i) t2))
    -> (forall G i e, length G = i -> oft G e (subst (sh i) t1') -> oft G (F' e) (subst (sh i) t2'))
    -> (forall e e', tstep nil e e' -> tstep nil (F e) (F e'))
    -> (forall e e', tstep nil e e' -> tstep nil (F' e) (F' e'))
    -> (forall v v', R1 v v' -> tev t2 t2' R2 (F v) (F' v'))
    -> (forall v v', tpst t1 t1' R1 v v' -> tev t2 t2' R2 (F v) (F' v')).
Proof.
intros t1 t1' t2 t2' R1 R2 F F' HofR1 HofR2 HclR2 HsubF HsubF' HofF HofF' HstepF HstepF' Hincl v v' Hv.
so (tpst_tvrel _#5 Hv) as (Hofv & Hofv' & Hvalv & Hvalv').
so (oft_reg _#3 wf_nil Hofv) as Hoft1.
so (oft_reg _#3 wf_nil Hofv') as Hoft1'.
rewrite <- (tpclosed_eq_tpst _#3 HofR2 HclR2); [].
#3 apply tev_tpst_intro; auto.
  replsub t2 with (subst (sh 0) t2).
  eapply HofF; eauto; [].
  simp_sub.
  assumption.

  replsub t2' with (subst (sh 0) t2').
  eapply HofF'; eauto; [].
  simp_sub.
  assumption.
assert (oft nil (tm_lam t1 (F (var 0))) (cn_arrow t1 t2)) as Hoflam.
  apply oft_lam; auto; [].
  apply HofF; auto; [].
  apply oft_var; auto with sequence; done.
assert (oft nil (tm_lam t1' (F' (var 0))) (cn_arrow t1' t2')) as Hoflam'.
  apply oft_lam; auto; [].
  apply HofF'; auto; [].
  apply oft_var; auto with sequence; done.
assert (forall w, tvalue nil w -> star (tstep nil) (tm_app (tm_lam t1 (F (var 0))) w) (F w)) as Heval.
  intros w Hval.
  #2 eapply star_step.
    apply tstep_app3; auto; done.

    unfold subst1; rewrite -> HsubF; [].
    simp_sub.
    apply star_refl; done.
assert (forall w, tvalue nil w -> star (tstep nil) (tm_app (tm_lam t1' (F' (var 0))) w) (F' w)) as Heval'.
  intros w Hval.
  #2 eapply star_step.
    apply tstep_app3; auto; done.

    unfold subst1; rewrite -> HsubF'; [].
    simp_sub.
    apply star_refl; done.
intros f f' Hf.
apply (biorthogonality_lemma t2 t2' t1 t1' R2 R1 f f' v v' (tm_lam t1 (F (var 0))) (tm_lam t1' (F' (var 0))) (F v) (F' v')); auto; [].
apply arrow_rel_; auto with dynamic; [].
intros w w' Hw.
so (HofR1 _ _ Hw) as (Hofw & Hofw' & Hvalw & Hvalw').
eapply anti_tsteps_tev; eauto with static; done.
Qed.


Lemma monad_mpst_mev_mpclosed :
  forall T1 T1' T2 T2' (R1 R2 : term -> term -> Prop) F F',
    mvrel T1 T1' R1
    -> mvrel T2 T2' R2
    -> mpclosed T2 T2' R2
    -> (forall s M, subst s (F M) = F (subst s M))
    -> (forall s M, subst s (F' M) = F' (subst s M))
    -> (forall G i M, length G = i -> ofm G pure M (subst (sh i) T1) -> ofm G impure (F M) (subst (sh i) T2))
    -> (forall G i M, length G = i -> ofm G pure M (subst (sh i) T1') -> ofm G impure (F' M) (subst (sh i) T2'))
    -> (forall M M', mstep nil M M' -> mstep nil (F M) (F M'))
    -> (forall M M', mstep nil M M' -> mstep nil (F' M) (F' M'))
    -> (forall V V', R1 V V' -> mev T2 T2' R2 (F V) (F' V'))
    -> (forall V V', mpst T1 T1' R1 V V' -> mev T2 T2' R2 (F V) (F' V')).
Proof.
intros T1 T1' T2 T2' R1 R2 F F' HofR1 HofR2 HclR2 HsubF HsubF' HofF HofF' HstepF HstepF' Hincl V V' HV.
so (mpst_mvrel _#5 HV) as (HofV & HofV' & HvalV & HvalV').
so (can_fstm_value _#4 wf_nil HofV HvalV) as (c & HfstV).
so (can_fstm_value _#4 wf_nil HofV' HvalV') as (c' & HfstV').
so (ofm_reg _#4 wf_nil HofV) as HofT1.
so (ofm_reg _#4 wf_nil HofV') as HofT1'.
so (can_fsts _ _ HofT1) as (k1 & Hfst1).
so (can_fsts _ _ HofT1') as (k1' & Hfst1').
rewrite <- (mpclosed_eq_mpst _#3 HofR2 HclR2); [].
#3 apply mev_mpst_intro; auto.
  replsub T2 with (subst (sh 0) T2).
  eapply HofF; eauto; [].
  simp_sub.
  eapply mvalue_pure; eauto; done.

  replsub T2' with (subst (sh 0) T2').
  eapply HofF'; eauto; [].
  simp_sub.
  eapply mvalue_pure; eauto; done.
assert (ofm nil impure (md_lamgn T1 (F (var 0))) (sg_pigen T1 (shift1 T2))) as Hoflam.
  apply ofm_forget; [].
  eapply ofm_lamgn; eauto; [].
  simp_sub.
  apply HofF; auto; [].
  replsub (subst (sh 2) T1) with (subst (sh 1) (shift1 T1)).
  eapply ofm_var; eauto with sequence; done.
assert (ofm nil impure (md_lamgn T1' (F' (var 0))) (sg_pigen T1' (shift1 T2'))) as Hoflam'.
  apply ofm_forget; [].
  eapply ofm_lamgn; eauto; [].
  simp_sub.
  apply HofF'; auto; [].
  replsub (subst (sh 2) T1') with (subst (sh 1) (shift1 T1')).
  eapply ofm_var; eauto with sequence; done.
assert (forall W d, mvalue nil W -> fstm nil W d -> star (mstep nil) (md_appgn (md_lamgn T1 (F (var 0))) W) (F W)) as Heval.
  intros W d Hval Hfstm.
  #2 eapply star_step.
    eapply mstep_appgn3; eauto; done.

    rewrite -> HsubF; [].
    simp_sub.
    apply star_refl; done.
assert (forall W d, mvalue nil W -> fstm nil W d -> star (mstep nil) (md_appgn (md_lamgn T1' (F' (var 0))) W) (F' W)) as Heval'.
  intros W d Hval Hfstm.
  #2 eapply star_step.
    eapply mstep_appgn3; eauto; done.

    rewrite -> HsubF'; [].
    simp_sub.
    apply star_refl; done.
intros G G' HG.
apply (biorthogonality_lemma_md T2 T2' T1 T1' R2 R1 G G' V V' (md_lamgn T1 (F (var 0))) (md_lamgn T1' (F' (var 0))) (F V) (F' V')); eauto; [].
apply pigen_rel_; auto with dynamic; [].
intros W W' HW.
so (HofR1 _ _ HW) as (HofW & HofW' & HvalW & HvalW').
so (mvalue_pure _#4 HvalW HofW) as HofWp.
so (mvalue_pure _#4 HvalW' HofW') as HofWp'.
renameover HofWp into HofW.
renameover HofWp' into HofW'.
so (can_fstm _#3 wf_nil HofW) as (d & HfstW).
so (can_fstm _#3 wf_nil HofW') as (d' & HfstW').
#2 eapply anti_msteps_mev; eauto.
  #2 eapply rel_quest.
    eapply ofm_appgn; eauto; done.

    simp_sub.
    reflexivity.
  
  #2 eapply rel_quest.
    eapply ofm_appgn; eauto; done.

    simp_sub.
    reflexivity.
Qed.


Lemma thalt_under_dyn :
  forall e,
    thalt nil e <-> mhalt nil (md_dyn e).
Proof.
intro e.
split.
  intro Hhalt.
  destruct Hhalt as (v & Hsteps & Hval).
  exists (md_dyn v).
  split; auto with dynamic; [].
  eapply star_map; eauto with dynamic; done.

  intro Hhalt.
  apply dyn_halt_inversion'; auto; done.
Qed.


Lemma dyn_eval_form :
  forall e M,
    star (mstep nil) (md_dyn e) M
    -> exists e',
         M = md_dyn e'
         /\ star (tstep nil) e e'.
Proof.
intros e M Hsteps.
remember (md_dyn e) as N eqn:HeqN.
revert e HeqN.
induct Hsteps.

(* refl *)
intros N e ->.
exists e.
auto using star_refl; done.

(* step *)
intros M1 M2 M3 Hstep _ IH e1 ->.
invert Hstep; [].
intros e2 Hstep' <-.
so (IH e2 (eq_refl _)) as (e3 & Heq & Hsteps'').
eauto using star_step; done.
Qed.


Lemma tev_under_dyn :
  forall t t' R e e',
    tev t t' R e e'
    <->
    mev (sg_dyn t) (sg_dyn t') (dyn_rel t t' R) (md_dyn e) (md_dyn e').
Proof.
intros t t' R e e'.
split.
  intro He.
  destruct He as (Hhaltiff & Hrel & Hofe & Hofe').
  #2 do2 3 split; auto with static.
    rewrite <- !thalt_under_dyn; [].
    assumption.

    intros V V' HstepsV HstepsV' HvalV HvalV'.
    so (dyn_eval_form _ _ HstepsV) as (v & -> & Hsteps).
    so (dyn_eval_form _ _ HstepsV') as (v' & -> & Hsteps').
    invert HvalV; [].
    intro Hvalv.
    invert HvalV'; [].
    intro Hvalv'.
    clear HstepsV HstepsV' HvalV HvalV'.
    so (preservations_tm _#4 wf_nil Hofe Hsteps) as Hofv.
    so (preservations_tm _#4 wf_nil Hofe' Hsteps') as Hofv'.
    apply dyn_rel_; auto; done.

  intro H.
  destruct H as (Hhaltiff & Hrel & Hof & Hof').
  so (dyn_ofm_inversion_dyn _#4 wf_nil Hof) as Hofe.
  so (dyn_ofm_inversion_dyn _#4 wf_nil Hof') as Hofe'.
  clear Hof Hof'.
  #2 do2 3 split; auto.
    rewrite -> !thalt_under_dyn; [].
    assumption.

    intros v v' Hsteps Hsteps' Hvalv Hvalv'.
    #3 exploit (Hrel (md_dyn v) (md_dyn v')) as Hdyn; auto with dynamic.
      eapply star_map; eauto with dynamic; done.

      eapply star_map; eauto with dynamic; done.

      invert Hdyn; auto; done.
Qed.


Lemma inhabitation_md :
  forall G T,
    wf G
    -> ofs G T
    -> exists M, ofm G pure M T.
Proof.
intros G T Hwf Hof.
revert Hwf.
induct Hof.

(* one *)
intros G _.
exists md_star; auto with static; done.

(* stat *)
intros G k Hofk _.
so (inhabitation _ _ Hofk) as (c & Hofc).
exists (md_stat c); auto with static; done.

(* dyn *)
intros G t Hoft _.
exists (md_dyn (Compactness.bottom t)).
auto using Compactness.oft_bottom with static; done.

(* pigen *)
intros G T1 T2 k1 HofT1 _ Hfst1 _ IH Hwf.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
so (IH (wf_cn Hwf Hofk1)) as (M & HofM).
exists (md_lamgn T1 (shift1 M)).
eapply ofm_lamgn; eauto; [].
apply weaken1_ofm; apply ofm_forget; auto; done.

(* piapp *)
intros G T1 T2 k1 HofT1 _ Hfst1 _ IH Hwf.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
so (IH (wf_cn Hwf Hofk1)) as (M & HofM).
exists (md_lamap T1 (shift1 M)).
eapply ofm_lamap; eauto; [].
apply weaken1_ofm; auto; done.

(* sigma *)
intros G T1 T2 k1 HofT1 IH1 Hfst1 HofT2 IH2 Hwf.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
so (IH1 Hwf) as (M1 & HofM1).
so (IH2 (wf_cn Hwf Hofk1)) as (M2 & HofM2).
so (can_fstm _#3 Hwf HofM1) as (c1 & HfstM1).
so (ofm_fst _#6 Hwf HofM1 HfstM1 Hfst1) as Hofc1.
so (subst1_ofm_cn _#6 Hofc1 HofM2) as HofM2x.
exists (md_pair M1 (subst1 c1 M2)).
eapply Singleton.ofm_pair_dep; eauto; done.
Qed.


Lemma monad_mpst_tev_tpclosed :
  forall T1 T1' t2 t2' (R1 R2 : term -> term -> Prop) F F',
    mvrel T1 T1' R1
    -> tvrel t2 t2' R2
    -> tpclosed t2 t2' R2
    -> (forall s M, subst s (F M) = F (subst s M))
    -> (forall s M, subst s (F' M) = F' (subst s M))
    -> (forall G i M, length G = i -> ofm G pure M (subst (sh i) T1) -> oft G (F M) (subst (sh i) t2))
    -> (forall G i M, length G = i -> ofm G pure M (subst (sh i) T1') -> oft G (F' M) (subst (sh i) t2'))
    -> (forall M M', mstep nil M M' -> tstep nil (F M) (F M'))
    -> (forall M M', mstep nil M M' -> tstep nil (F' M) (F' M'))
    -> (forall V V', R1 V V' -> tev t2 t2' R2 (F V) (F' V'))
    -> (forall V V', mpst T1 T1' R1 V V' -> tev t2 t2' R2 (F V) (F' V')).
Proof.
intros T1 T1' t2 t2' R1 R2 F F' HofR1 HofR2 HclR2 HsubF HsubF' HofF HofF' HstepF HstepF' Hincl V V' HV.
assert (ofc nil t2 kd_type) as Hoft2.
  so (inhabitation_md _ _ wf_nil (ofm_reg _#4 wf_nil (HV anderl))) as (M & HofM).
  #2 exploit (HofF nil 0 M (eq_refl _)) as H.
    simp_sub; assumption.

    simp_sub in H.
    exact (oft_reg _#3 wf_nil H).
assert (ofc nil t2' kd_type) as Hoft2'.
  so (inhabitation_md _ _ wf_nil (ofm_reg _#4 wf_nil (HV anderrl))) as (M & HofM).
  #2 exploit (HofF' nil 0 M (eq_refl _)) as H.
    simp_sub; assumption.

    simp_sub in H.
    exact (oft_reg _#3 wf_nil H).
rewrite -> tev_under_dyn; [].
revert V V' HV.
#10 apply (monad_mpst_mev_mpclosed T1 T1' (sg_dyn t2) (sg_dyn t2') R1 (dyn_rel t2 t2' R2) (fun z => md_dyn (F z)) (fun z => md_dyn (F' z))).
  assumption.

  apply dyn_rel_mvrel; done.

  apply dyn_rel_mpclosed; auto; done.

  intros s M.
  simp_sub.
  rewrite -> HsubF; [].
  reflexivity.

  intros s M.
  simp_sub.
  rewrite -> HsubF'; [].
  reflexivity.

  intros G i M Heq Hof.
  simp_sub.
  auto with static; done.

  intros G i M Heq Hof.
  simp_sub.
  auto with static; done.

  intros M M' Hstep.
  auto with dynamic; done.

  intros M M' Hstep.
  auto with dynamic; done.

  intros V V' HV.
  rewrite <- tev_under_dyn; [].
  apply Hincl; auto; done.
Qed.


Lemma mhalt_into_term :
  forall T M,
    ofm nil impure M T
    -> (mhalt nil M <-> thalt nil (tm_letm M tm_star)).
Proof.
intros T M Hof.
split.
  intro Hhalt.
  destruct Hhalt as (V & Hsteps & Hval).
  so (can_fstm_value _#4 wf_nil (preservations_md _#5 wf_nil Hof Hsteps) Hval) as (c & Hfst).
  exists tm_star.
  split; auto with dynamic; [].
  #2 apply (star_trans _#3 (tm_letm V tm_star)).
    eapply (star_map _#4 (fun z => tm_letm z _)); eauto with dynamic; done.

    #2 eapply star_step.
      eapply tstep_letm2; eauto; done.

      simp_sub.
      apply star_refl; done.

  intro Hhalt.
  eapply letm_halt_inversion'; eauto; done.
Qed.


Lemma tpst_implies_dyn_mpst :
  forall t t' R v v',
    tvrel t t' R
    -> tpst t t' R v v'
    -> mpst (sg_dyn t) (sg_dyn t') (dyn_rel t t' R) (md_dyn v) (md_dyn v').
Proof.
intros t t' R v v' HofR Hv.
destruct Hv as (Hincl & Hofv & Hofv' & Hvalv & Hvalv').
do2 4 split; auto with static dynamic; [].
intros F F' HF.
so HF as (_ & HofF & HofF' & _).
assert (ofc nil t kd_type) as Hoft.
  so (ofm_reg _#4 wf_nil HofF) as H.
  invert H; [].
  intros _ H' _ _.
  invert H'; auto; done.
assert (ofc nil t' kd_type) as Hoft'.
  so (ofm_reg _#4 wf_nil HofF') as H.
  invert H; [].
  intros _ H' _ _.
  invert H'; auto; done.
#2 rewrite -> (mhalt_into_term sg_one (md_appgn F (md_dyn v))).
  #2 rewrite -> (mhalt_into_term sg_one (md_appgn F' (md_dyn v'))).
    #2 rewrite <- (tsteps_thalt_iff _ (tm_app (tm_lam t (tm_letm (md_appgn (shift1 F) (md_dyn (var 0))) tm_star)) v) (tm_letm (md_appgn F (md_dyn v)) tm_star)).
      #2 rewrite <- (tsteps_thalt_iff _ (tm_app (tm_lam t' (tm_letm (md_appgn (shift1 F') (md_dyn (var 0))) tm_star)) v') (tm_letm (md_appgn F' (md_dyn v')) tm_star)).
        apply Hincl; [].
        clear v v' Hincl Hofv Hofv' Hvalv Hvalv'.
        #3 do2 4 split; auto with dynamic.
          intros v v' Hv.
          so (HofR _ _ Hv) as (Hofv & Hofv' & Hvalv & Hvalv').
          #2 rewrite -> (tsteps_thalt_iff _ (tm_app (tm_lam t (tm_letm (md_appgn (shift1 F) (md_dyn (var 0))) tm_star)) v) (tm_letm (md_appgn F (md_dyn v)) tm_star)).
            #2 rewrite -> (tsteps_thalt_iff _ (tm_app (tm_lam t' (tm_letm (md_appgn (shift1 F') (md_dyn (var 0))) tm_star)) v') (tm_letm (md_appgn F' (md_dyn v')) tm_star)).
              #2 rewrite <- (mhalt_into_term sg_one (md_appgn F (md_dyn v))).
                #2 rewrite <- (mhalt_into_term sg_one (md_appgn F' (md_dyn v'))).
                  apply (HF andel); [].
                  apply dyn_rel_; auto; done.

                  #2 eapply rel_quest.
                    eapply ofm_appgn; eauto with static; done.

                    simp_sub.
                    reflexivity.

                #2 eapply rel_quest.
                  eapply ofm_appgn; eauto with static; done.

                  simp_sub.
                  reflexivity.
                  
              #2 eapply star_step.
                apply tstep_app3; auto with dynamic; done.

                simp_sub.
                apply star_refl; done.

            #2 eapply star_step.
              apply tstep_app3; auto with dynamic; done.

              simp_sub.
              apply star_refl; done.

          apply oft_lam; auto; [].
          simp_sub.
          apply (@oft_letm _ sg_one kd_one); auto with static; [].
          #2 eapply rel_quest.
            #2 eapply ofm_appgn; eauto with static.
              #2 eapply rel_quest.
                eapply weaken1_ofm; eauto; done.

                simp_sub.
                reflexivity.

              apply ofm_dyn; [].
              apply oft_var; auto with sequence; done.

            simp_sub.
            reflexivity.

          apply oft_lam; auto; [].
          simp_sub.
          apply (@oft_letm _ sg_one kd_one); auto with static; [].
          #2 eapply rel_quest.
            #2 eapply ofm_appgn; eauto with static.
              #2 eapply rel_quest.
                eapply weaken1_ofm; eauto; done.
  
                simp_sub.
                reflexivity.
  
              apply ofm_dyn; [].
              apply oft_var; auto with sequence; done.
  
            simp_sub.
            reflexivity.
  
        #2 eapply star_step.
          apply tstep_app3; auto; done.
  
          simp_sub.
          apply star_refl; done.
  
      #2 eapply star_step.
        apply tstep_app3; auto; done.
  
        simp_sub.
        apply star_refl; done.
  
    #2 eapply rel_quest.
      eapply ofm_appgn; eauto with static; done.
  
      simp_sub.
      reflexivity.
  
  #2 eapply rel_quest.
    eapply ofm_appgn; eauto with static; done.
  
    simp_sub.
    reflexivity.
Qed.


Lemma monad_tpst_mev_mpclosed :
  forall t1 t1' T2 T2' (R1 R2 : term -> term -> Prop) F F',
    tvrel t1 t1' R1
    -> mvrel T2 T2' R2
    -> mpclosed T2 T2' R2
    -> (forall s M, subst s (F M) = F (subst s M))
    -> (forall s M, subst s (F' M) = F' (subst s M))
    -> (forall G i e, length G = i -> oft G e (subst (sh i) t1) -> ofm G impure (F e) (subst (sh i) T2))
    -> (forall G i e, length G = i -> oft G e (subst (sh i) t1') -> ofm G impure (F' e) (subst (sh i) T2'))
    -> (forall e e', tstep nil e e' -> mstep nil (F e) (F e'))
    -> (forall e e', tstep nil e e' -> mstep nil (F' e) (F' e'))
    -> (forall v v', R1 v v' -> mev T2 T2' R2 (F v) (F' v'))
    -> (forall v v', tpst t1 t1' R1 v v' -> mev T2 T2' R2 (F v) (F' v')).
Proof.
intros t1 t1' T2 T2' R1 R2 F F' HofR1 HofR2 HclR2 HsubF HsubF' HofF HofF' HstepF HstepF' Hincl.
cut (forall V V',
       mpst (sg_dyn t1) (sg_dyn t1') (dyn_rel t1 t1' R1) V V'
       -> mev T2 T2' R2 (F (tm_ext V)) (F' (tm_ext V'))).
  clear Hincl.
  intros Hincl v v' Hv.
  so Hv as (_ & Hofv & Hofv' & Hvalv & Hvalv').
  #3 apply (msteps_mev _#3 (F (tm_ext (md_dyn v))) (F' (tm_ext (md_dyn v')))).
    eapply star_map; eauto using star_one with dynamic; done.

    eapply star_map; eauto using star_one with dynamic; done.

    apply Hincl; [].

    apply tpst_implies_dyn_mpst; auto; done.
#10 apply (monad_mpst_mev_mpclosed (sg_dyn t1) (sg_dyn t1') T2 T2' (dyn_rel t1 t1' R1) R2 (fun z => F (tm_ext z)) (fun z => F' (tm_ext z))).
  apply dyn_rel_mvrel; done.

  assumption.
  
  assumption.

  intros s M.
  rewrite -> HsubF; [].
  simp_sub.
  reflexivity.

  intros s M.
  rewrite -> HsubF'; [].
  simp_sub.
  reflexivity.

  intros G i M Heq Hof.
  simp_sub in Hof.
  apply HofF; auto with static; done.

  intros G i M Heq Hof.
  simp_sub in Hof.
  apply HofF'; auto with static; done.

  intros M M' Hstep.
  apply HstepF; auto with dynamic; done.

  intros M M' Hstep.
  apply HstepF'; auto with dynamic; done.

  intros V V' HV.
  invert HV; [].
  intros v v' Hofv Hofv' Hvalv Hvalv' Hv <- <-.
  #2 apply (anti_msteps_mev _#5 (F v) (F' v')); eauto using star_one with dynamic.
    replsub T2 with (subst (sh 0) T2).
    apply HofF; auto; [].
    simp_sub.
    auto with static; done.

    replsub T2' with (subst (sh 0) T2').
    apply HofF'; auto; [].
    simp_sub.
    auto with static; done.
Qed.


(* Consolidations *)

Lemma monad_tpstev_tev_tpclosed :
  forall t1 t1' t2 t2' (R1 R2 : term -> term -> Prop) F F',
    tvrel t1 t1' R1
    -> tvrel t2 t2' R2
    -> tpclosed t2 t2' R2
    -> (forall s e, subst s (F e) = F (subst s e))
    -> (forall s e, subst s (F' e) = F' (subst s e))
    -> (forall G i e, length G = i -> oft G e (subst (sh i) t1) -> oft G (F e) (subst (sh i) t2))
    -> (forall G i e, length G = i -> oft G e (subst (sh i) t1') -> oft G (F' e) (subst (sh i) t2'))
    -> (forall e e', tstep nil e e' -> tstep nil (F e) (F e'))
    -> (forall e e', tstep nil e e' -> tstep nil (F' e) (F' e'))
    -> (forall v v', R1 v v' -> tev t2 t2' R2 (F v) (F' v'))
    -> (forall e e', tev t1 t1' (tpst t1 t1' R1) e e' -> tev t2 t2' R2 (F e) (F' e')).
Proof.
intros t1 t1' t2 t2' R1 R2 F F' HofR1 HofR2 HclosedR2 HsubF HsubF' HofF HofF' HstepF HstepF' Hincl.
#3 apply monad_tev_tev; auto.
  intros m Hof.
  so (HofF nil 0 m (eq_refl _)) as H.
  simp_sub in H.
  auto; done.

  intros m Hof.
  so (HofF' nil 0 m (eq_refl _)) as H.
  simp_sub in H.
  auto; done.

  apply monad_tpst_tev_tpclosed; auto; done.
Qed.


Lemma monad_mpstev_tev_tpclosed :
  forall T1 T1' t2 t2' (R1 R2 : term -> term -> Prop) F F',
    mvrel T1 T1' R1
    -> tvrel t2 t2' R2
    -> tpclosed t2 t2' R2
    -> (forall s M, subst s (F M) = F (subst s M))
    -> (forall s M, subst s (F' M) = F' (subst s M))
    -> (forall G i M, length G = i -> ofm G impure M (subst (sh i) T1) -> oft G (F M) (subst (sh i) t2))
    -> (forall G i M, length G = i -> ofm G impure M (subst (sh i) T1') -> oft G (F' M) (subst (sh i) t2'))
    -> (forall M M', mstep nil M M' -> tstep nil (F M) (F M'))
    -> (forall M M', mstep nil M M' -> tstep nil (F' M) (F' M'))
    -> (forall V V', R1 V V' -> tev t2 t2' R2 (F V) (F' V'))
    -> (forall M M', mev T1 T1' (mpst T1 T1' R1) M M' -> tev t2 t2' R2 (F M) (F' M')).
Proof.
intros T1 T1' t2 t2' R1 R2 F F' HofR1 HofR2 HclosedR2 HsubF HsubF' HofF HofF' HstepF HstepF' Hincl.
#3 apply monad_mev_tev; auto.
  intros m Hof.
  so (HofF nil 0 m (eq_refl _)) as H.
  simp_sub in H.
  auto; done.

  intros m Hof.
  so (HofF' nil 0 m (eq_refl _)) as H.
  simp_sub in H.
  auto; done.

  apply monad_mpst_tev_tpclosed; auto using ofm_forget_pure; done.
Qed.


Lemma monad_mpstev_mev_mpclosed :
  forall T1 T1' T2 T2' (R1 R2 : term -> term -> Prop) F F',
    mvrel T1 T1' R1
    -> mvrel T2 T2' R2
    -> mpclosed T2 T2' R2
    -> (forall s M, subst s (F M) = F (subst s M))
    -> (forall s M, subst s (F' M) = F' (subst s M))
    -> (forall G i M, length G = i -> ofm G impure M (subst (sh i) T1) -> ofm G impure (F M) (subst (sh i) T2))
    -> (forall G i M, length G = i -> ofm G impure M (subst (sh i) T1') -> ofm G impure (F' M) (subst (sh i) T2'))
    -> (forall M M', mstep nil M M' -> mstep nil (F M) (F M'))
    -> (forall M M', mstep nil M M' -> mstep nil (F' M) (F' M'))
    -> (forall V V', R1 V V' -> mev T2 T2' R2 (F V) (F' V'))
    -> (forall M M', mev T1 T1' (mpst T1 T1' R1) M M' -> mev T2 T2' R2 (F M) (F' M')).
Proof.
intros T1 T1' t2 t2' R1 R2 F F' HofR1 HofR2 HclosedR2 HsubF HsubF' HofF HofF' HstepF HstepF' Hincl.
#3 apply monad_mev_mev; auto.
  intros m Hof.
  so (HofF nil 0 m (eq_refl _)) as H.
  simp_sub in H.
  auto; done.

  intros m Hof.
  so (HofF' nil 0 m (eq_refl _)) as H.
  simp_sub in H.
  auto; done.

  apply monad_mpst_mev_mpclosed; auto using ofm_forget_pure; done.
Qed.


Lemma monad_tpstev_mev_mpclosed :
  forall t1 t1' T2 T2' (R1 R2 : term -> term -> Prop) F F',
    tvrel t1 t1' R1
    -> mvrel T2 T2' R2
    -> mpclosed T2 T2' R2
    -> (forall s M, subst s (F M) = F (subst s M))
    -> (forall s M, subst s (F' M) = F' (subst s M))
    -> (forall G i e, length G = i -> oft G e (subst (sh i) t1) -> ofm G impure (F e) (subst (sh i) T2))
    -> (forall G i e, length G = i -> oft G e (subst (sh i) t1') -> ofm G impure (F' e) (subst (sh i) T2'))
    -> (forall e e', tstep nil e e' -> mstep nil (F e) (F e'))
    -> (forall e e', tstep nil e e' -> mstep nil (F' e) (F' e'))
    -> (forall v v', R1 v v' -> mev T2 T2' R2 (F v) (F' v'))
    -> (forall e e', tev t1 t1' (tpst t1 t1' R1) e e' -> mev T2 T2' R2 (F e) (F' e')).
Proof.
intros T1 T1' t2 t2' R1 R2 F F' HofR1 HofR2 HclosedR2 HsubF HsubF' HofF HofF' HstepF HstepF' Hincl.
#3 apply monad_tev_mev; auto.
  intros m Hof.
  so (HofF nil 0 m (eq_refl _)) as H.
  simp_sub in H.
  auto; done.

  intros m Hof.
  so (HofF' nil 0 m (eq_refl _)) as H.
  simp_sub in H.
  auto; done.

  apply monad_tpst_mev_mpclosed; auto; done.
Qed.
