
Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Formation.
Require Import Regularity.
Require Import Declarative.
Require Import SubjectReduction.
Require Import Consistent.
Require Import Relation.


Inductive oftr : context -> (* tm *) term -> (* cn *) term -> Prop :=
| oftr_var {G i t}
    : index i G (cl_tm t)
      -> oftr G (var i) (subst (sh (S i)) t)

| oftr_star {G}
    : oftr G tm_star cn_unit

| oftr_lam {G t1 t2 e}
    : ofc G t1 kd_type
      -> oft (G; cl_tm t1) e (shift1 t2)
      -> oftr G (tm_lam t1 e) (cn_arrow t1 t2)

| oftr_app {G t1 t2 e1 e2}
    : oft G e1 (cn_arrow t1 t2)
      -> oft G e2 t1
      -> oftr G (tm_app e1 e2) t2

| oftr_pair {G t1 t2 e1 e2}
    : oft G e1 t1
      -> oft G e2 t2
      -> oftr G (tm_pair e1 e2) (cn_prod t1 t2)

| oftr_pi1 {G t1 t2 e}
    : oft G e (cn_prod t1 t2)
      -> oftr G (tm_pi1 e) t1

| oftr_pi2 {G t1 t2 e}
    : oft G e (cn_prod t1 t2)
      -> oftr G (tm_pi2 e) t2

| oftr_plam {G k t e}
    : ofk G k
      -> oft (G; cl_cn k) e t
      -> oftr G (tm_plam k e) (cn_all k t)

| oftr_papp {G k t e c}
    : oft G e (cn_all k t)
      -> ofc G c k
      -> oftr G (tm_papp e c) (subst1 c t)

| oftr_pack {G k t c e}
    : ofc G c k
      -> oft G e (subst1 c t)
      -> ofc (G; cl_cn k) t kd_type
      -> oftr G (tm_pack c e k t) (cn_exists k t)

| oftr_unpack {G k t t' e1 e2}
    : oft G e1 (cn_exists k t)
      -> oft (G; cl_cn k; cl_tm t) e2 (subst (sh 2) t')
      -> ofc G t' kd_type
      -> oftr G (tm_unpack e1 e2) t'

| oftr_fix {G t e}
    : oft G e (cn_arrow (cn_arrow cn_unit t) t)
      -> oftr G (tm_fix t e) t

| oftr_lett {G t1 t2 e1 e2}
    : oft G e1 t1
      -> oft (G; cl_tm t1) e2 (shift1 t2)
      -> oftr G (tm_lett e1 e2) t2

| oftr_letm {G T k t M e}
    : ofm G impure M T
      -> fsts T k
      -> oft (G; cl_cn k; cl_md (shift1 T) (var 0)) e (subst (sh 2) t)
      -> ofc G t kd_type
      -> oftr G (tm_letm M e) t

| oftr_ext {G t M}
    : ofm G impure M (sg_dyn t)
      -> oftr G (tm_ext M) t
.


Lemma oft_invert :
  forall G e t,
    wf G
    -> oft G e t
    -> exists t',
         eqc G t' t kd_type
         /\ oftr G e t'.
Proof.
intros G e t Hwf Hof.
revert Hwf.
induct Hof; try (intros; eexists; split; eauto using oftr, oft_reg with static; done); [].
intros G t t' e _ IH Heqc Hwf.
so (IH Hwf) as (t'' & Heqc' & Hoftr).
eauto with static.
Qed.


Lemma ofm_forget_pure {G p T M} :
  ofm G pure M T
  -> ofm G p M T.
Proof.
destruct p; auto with static.
Qed.


Lemma ofm_forget_impure {G p T M} :
  ofm G p M T
  -> ofm G impure M T.
Proof.
destruct p; auto with static.
Qed.


Lemma can_fstm :
  forall G M T,
    wf G
    -> ofm G pure M T
    -> exists c, fstm G M c.
Proof.
intros G M T Hwf Hof.
remember pure as p eqn:Heqp.
revert Hwf Heqp.
#6 induct Hof;
try (intros; try discriminate; eauto with static; done).

(* lamap *)
intros G T1 T2 k1 M HofT1 Hfst1 Hof IH Hwf _.
thus (wf (G; cl_cn k1; cl_md (shift1 T1) (var 0))) as Hwf' by @wf_md.
so (IH Hwf' (eq_refl _)) as (c & Hfstm).
exists (cn_lam k1 (subst1 nonsense c)).
apply fstm_lamap; auto; [].
simp_sub.
so (can_fsts _ _ (ofm_reg _#4 Hwf' Hof)) as (k2w & Hfst2w).
so (ofm_fst _#6 Hwf' Hof Hfstm Hfst2w) as Hofc.
so (sid_ofc _#6 (subrel_sid_bind_md _ _ _ nonsense (var 0)) Hofc) as Heq.
simp_sub in Heq.
rewrite -> Heq.
assumption.

(* appap *)
intros G p T1 T2 M1 M2 c2 Hof1 IH1 Hof2 IH2 HfstM2 Hwf ->.
so (IH1 Hwf (eq_refl _)) as (c1 & Hfst1).
exists (cn_app c1 c2).
apply fstm_appap; auto; done.

(* pair *)
intros G p T1 T2 M1 M2 _ IH1 _ IH2 Hwf ->.
so (IH1 Hwf (eq_refl _)) as (c1 & Hfst1).
so (IH2 Hwf (eq_refl _)) as (c2 & Hfst2).
exists (cn_pair c1 c2).
apply fstm_pair; auto; done.

(* pi1 *)
intros G T1 T2 M _ IH Hwf _.
so (IH Hwf (eq_refl _)) as (c & Hfst).
eauto with static; done.

(* lett *)
intros G p t T e M Hofe HofM IH Hwf ->.
so (wf_tm Hwf (oft_reg _#3 Hwf Hofe)) as Hwf'.
so (IH Hwf' (eq_refl _)) as (c & Hfst).
#2 sor (refine (subst_fstm (G; cl_utm) _ (under 1 id) _ _ _ Hfst)) as Hfst'.
  simp_sub.
  apply subof_dot; eauto using subof_sh with sequence; [].
  simp_sub.
  apply tof_tm; done.
simp_sub in Hfst'.
exists (subst1 nonsense c).
apply fstm_lett; [].
simp_sub.
so (can_fsts _ _ (ofm_reg _#4 Hwf' HofM)) as (kw & HfstTw).
so (ofm_fst _#6 Hwf' HofM Hfst HfstTw) as Hofc.
so (sid_ofc _#6 (subrel_sid_bind_tm _ _ nonsense (var 0)) Hofc) as Heq.
simp_sub in Heq.
rewrite -> Heq.
assumption.

(* extsigma *)
intros G T1 T2 M _ IH _ _ Hwf _.
so (IH Hwf (eq_refl _)) as (c & Hfst).
invert Hfst; [].
intros c' Hfst' _.
exists c'.
assumption.
Qed.


Lemma mvalue_pure :
  forall G p V T,
    mvalue G V
    -> ofm G p V T
    -> ofm G pure V T.
Proof.
intros G p V T Hval Hof.
revert Hval.
induct Hof;
intros; invert Hval; eauto with static.
Qed.


Lemma can_fstm_value :
  forall G p M T,
    wf G
    -> ofm G p M T
    -> mvalue G M
    -> exists c, fstm G M c.
Proof.
intros G p M T Hwf Hof Hval.
so (mvalue_pure _#4 Hval Hof) as Hof'.
exact (can_fstm _#3 Hwf Hof').
Qed.


Lemma mstep_fst :
  forall G M M' c,
    mstep G M M'
    -> fstm G M c
    -> fstm G M' c
       \/
       exists c',
         fstm G M' c'
         /\ reduce G c c'.
Proof.
intros G M M' c Hstep Hfst.
revert c Hfst.
#12 induct Hstep;
try (intros;
     match goal with
     | H : fstm _ _ _ |- _ => invert H; done
     end;
     done).

(* dyn *)
intros G e e' H c Hfst.
invert Hfst; [].
intros <-.
left.
apply fstm_dyn; done.

(* appap1 *)
intros G M1 M1' M2 _ IH c Hfst.
invert Hfst; [].
intros c1 c2 Hfst1 Hfst2 <-.
so (IH _ Hfst1) as [Hfst1' | (c1' & Hfst1' & Hred)].
  left; auto with static; done.

  right; eauto using reduce with static; done.

(* appap2 *)
intros G M1 M1' M2 _ _ IH c Hfst.
invert Hfst; [].
intros c1 c2 Hfst1 Hfst2 <-.
so (IH _ Hfst2) as [Hfst2' | (c2' & Hfst2' & Hred)].
  left; auto with static; done.

  right; eauto using reduce with static; done.

(* appap3 *)
intros G T M1 V2 c2 _ HfstV2 c Hfst.
right.
invert Hfst; [].
intros c' c2' Hfst' HfstV2' <-.
so (fstm_fun _ _ _ _ HfstV2 HfstV2'); subst c2'.
invert Hfst'; [].
intros k d HfstT HfstM1 <-.
exists (subst1 c2 d).
split.
  #2 sor (refine (subst_fstm G _ (dot V2 (dot c2 id)) _ _ _ HfstM1)) as HfstM1x.
    #2 apply subof_dot.
      apply subof_dot; auto with static; [].
      simp_sub.
      apply tof_cn; done.

      simp_sub.
      apply tof_md; auto; done.
  
    simp_sub in HfstM1x.
    assumption.

  apply reduce_beta; done.

(* pair1 *)
intros G M1 M1' M2 _ IH c Hfst.
invert Hfst; [].
intros c1 c2 Hfst1 Hfst2 <-.
so (IH _ Hfst1) as [| (c1' & Hfst1' & Hred)].
  left; auto with static; done.

  right; eauto using reduce with static; done.

(* pair2 *)
intros G M1 M1' M2 _ _ IH c Hfst.
invert Hfst; [].
intros c1 c2 Hfst1 Hfst2 <-.
so (IH _ Hfst2) as [| (c2' & Hfst2' & Hred)].
  left; auto with static; done.

  right; eauto using reduce with static; done.

(* pi11 *)
intros G M M' _ IH c Hfst.
invert Hfst; [].
intros c' Hfst' <-.
so (IH _ Hfst') as [| (c'' & Hfst'' & Hred)].
  left; auto with static; done.

  right; eauto using reduce with static; done.

(* pi12 *)
intros G V1 V2 _ _ c Hfst.
right.
invert Hfst; [].
intros c' Hfst' <-.
invert Hfst'; [].
intros c1 c2 Hfst1 Hfst2 <-.
eauto using reduce; done.

(* pi21 *)
intros G M M' _ IH c Hfst.
invert Hfst; [].
intros c' Hfst' <-.
so (IH _ Hfst') as [| (c'' & Hfst'' & Hred)].
  left; auto with static; done.

  right; eauto using reduce with static; done.

(* pi22 *)
intros G V1 V2 _ _ c Hfst.
right.
invert Hfst; [].
intros c' Hfst' <-.
invert Hfst'; [].
intros c1 c2 Hfst1 Hfst2 <-.
eauto using reduce; done.

(* lett1 *)
intros G e1 e1' M2 _ c Hfst.
left.
invert Hfst; [].
intro Hfst'.
apply fstm_lett; auto; done.

(* lett2 *)
intros G v1 M2 H c Hfst.
left.
invert Hfst; [].
intros Hfst'.
sor (refine (subst_fstm G _ (dot v1 id) _ _ _ Hfst')) as Hfst''; eauto using tof with static; [].
simp_sub in Hfst''.
assumption.
Qed.


Lemma preservation_fst :
  forall G p M M' T c k,
    wf G
    -> ofm G p M T
    -> mstep G M M'
    -> fsts T k
    -> fstm G M c
    -> exists c',
         fstm G M' c'
         /\ eqc G c c' k.
Proof.
intros G p M M' T c k Hwf HofM Hstep Hfst HfstM.
so (ofm_fst _#6 Hwf HofM HfstM Hfst) as Hofc.
so (mstep_fst _#4 Hstep HfstM) as [HfstM' | (c' & HfstM' & Hred)].
  eauto with static; done.

  exists c'.
  eauto using subject_reduction; done.
Qed.


Theorem preservation :
  (forall G e e' t,
     wf G
     -> oft G e t
     -> tstep G e e'
     -> oft G e' t)
  /\
  (forall G p M M' T,
     wf G
     -> ofm G p M T
     -> mstep G M M'
     -> ofm G p M' T).
Proof.
exploit
  (dynamic_ind
     (fun G e t =>
        forall e',
          wf G
          -> tstep G e e'
          -> oft G e' t)
     (fun G p M T =>
        wf G
        -> (forall M',
              mstep G M M'
              -> ofm G p M' T)
           /\
           (forall e t,
              M = md_dyn e
              -> T = sg_dyn t
              -> oft G e t)
           /\
           (forall U N T1 T2,
              M = md_lamgn U N
              -> T = sg_pigen T1 T2
              -> exists k1,
                   fsts T1 k1
                   /\ ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure N (shift1 T2))
           /\
           (forall U N T1 T2,
              M = md_lamap U N
              -> T = sg_piapp T1 T2
              -> exists k1,
                   fsts T1 k1
                   /\ ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) pure N (shift1 T2))
           /\
           (forall N1 N2 T1 T2,
              M = md_pair N1 N2
              -> T = sg_sigma T1 T2
              -> mvalue G N1
              -> mvalue G N2
              -> exists c1,
                   ofm G p N1 T1
                   /\ fstm G N1 c1
                   /\ ofm G p N2 (subst1 c1 T2)))) as Hind.

(* var *)
intros G i t H e' _ Hstep.
invert Hstep; done.

(* star *)
intros G e' _ Hstep.
invert Hstep; done.

(* lam *)
intros G t1 t2 e Hoft Hofe IH d Hwf Hstep.
invert Hstep; done.

(* app *)
intros G t1 t2 e1 e2 Hof1 IH1 Hof2 IH2 e' Hwf Hstep.
invert Hstep; try (intros; subst; eauto with static; done); [].
intros t1' d _ <- <-.
so (oft_invert _#3 Hwf Hof1) as (? & Heqc & Hoftr).
invert Hoftr; [].
intros t2' Hoft Hofd <-.
so (arrow_injective _#5 Hwf Heqc) as (Heqt1 & Heqt2).
#2 sor (refine (subst_oft G _ (dot e2 id) _ _ _ Hofd)) as Hofdx.
  apply subof_dot; auto with static; [].
  simp_sub.
  apply dof_tm; eauto with static; done.

  simp_sub in Hofdx.
  eauto with static; done.

(* pair *)
intros G t1 t2 e1 e2 Hof1 IH1 Hof2 IH2 e' Hwf Hstep.
invert Hstep; intros; subst; eauto with static; done.

(* pi1 *)
intros G t1 t2 e Hof IH d Hwf Hstep.
invert Hstep; try (intros; subst; eauto with static; done); [].
intros v2 Hval1 Hval2 <-.
rename d into v1.
so (oft_invert _#3 Hwf Hof) as (t & Heqc & Hoftr).
invert Hoftr; [].
intros t1' t2' Hof1 Hof2 <-.
so (prod_injective _#5 Hwf Heqc) as (Heqt1 & Heqt2).
eauto with static; done.

(* pi2 *)
intros G t1 t2 e Hof IH d Hwf Hstep.
invert Hstep; try (intros; subst; eauto with static; done); [].
intros v1 Hval1 Hval2 <-.
rename d into v2.
so (oft_invert _#3 Hwf Hof) as (t & Heqc & Hoftr).
invert Hoftr; [].
intros t1' t2' Hof1 Hof2 <-.
so (prod_injective _#5 Hwf Heqc) as (Heqt1 & Heqt2).
eauto with static; done.

(* plam *)
intros G k t e Hofk Hofe IH e' Hwf Hstep.
invert Hstep; done.

(* papp *)
intros G k t e c Hofe IH Hofc e' Hwf Hstep.
invert Hstep; try (intros; subst; eauto with static; done); [].
intros k' d <- <-.
so (oft_invert _#3 Hwf Hofe) as (? & Heqc & Hoftr).
invert Hoftr; [].
intros t' Hofk' Hoft' <-.
so (all_injective _#5 Hwf Heqc) as (Heqk & Heqt).
#2 apply (@oft_equiv _ (subst1 c t')).
  eapply subst_oft; eauto; [].
  apply subof_dot; auto with static; [].
  simp_sub.
  apply dof_cn; eauto with static; done.

  repl kd_type with (subst1 c kd_type) by (simp_sub; reflexivity).
  eapply subst_eqc; eauto; [].
  apply subof_dot; auto with static; [].
  simp_sub.
  apply sof_cn; eauto with static; done.

(* pack *)
intros G k t c e Hofc Hofe IH Hoft e' Hwf Hstep.
invert Hstep; intros; subst; eauto with static; done.

(* unpack *)
intros G k t u e1 e2 Hofe1 IH1 Hofe2 IH2 Hofu e' Hwf Hstep.
invert Hstep; try (intros; subst; eauto with static; done); [].
intros c v k' t' _ <- <-.
so (oft_invert _#3 Hwf Hofe1) as (? & Heqc & Hoftr).
invert Hoftr; [].
intros Hofc Hofv Hoft' <-.
so (exists_injective _#5 Hwf Heqc) as (Heqk & Heqt).
#2 sor (refine (subst_oft G _ (dot v (dot c id)) _ _ _ Hofe2)) as Hofe2x.
  #2 apply subof_dot.
    apply subof_dot; auto with static; [].
    simp_sub.
    apply dof_cn; eauto with static; done.

    simp_sub.
    apply dof_tm; [].
    eapply oft_equiv; eauto; [].
    repl kd_type with (subst1 c kd_type) by (simp_sub; reflexivity).
    eapply subst1_eqc; eauto; done.

  simp_sub in Hofe2x.
  assumption.

(* fix *)
intros G t e Hof _ e' Hwf Hstep.
invert Hstep; [].
intros <-.
eapply oft_app; eauto; [].
apply oft_lam; auto with static; [].
apply (subst_oft _ G sh1); eauto using subof_sh with static sequence; done.

(* lett *)
intros G t1 t2 e1 e2 Hof1 IH1 Hof2 _ e' Hwf Hstep.
invert Hstep; try (intros; subst; eauto with static; done); [].
intros _ <-.
#2 sor (refine (subst_oft G _ (dot e1 id) _ _ _ Hof2)) as Hof2x.
  apply subof_dot; auto with static; [].
  simp_sub.
  apply dof_tm; assumption.

  simp_sub in Hof2x.
  assumption.

(* equiv *)
intros G t t' e _ IH Heqc e' Hwf Hstep.
eauto with static; done.

(* letm *)
intros G T k t M e HofM IH1 Hfst Hofe IH2 Hoft e' Hwf Hstep.
#2 invert Hstep.
  intros M' Hstep' <-.
  so (IH1 Hwf andel _ Hstep') as HofM'.
  eapply oft_letm; eauto; done.

  intros c Hval HfstM <-.
  #2 sor (refine (subst_oft G _ (dot M (dot c id)) _ _ _ Hofe)) as Hofex.
    #2 apply subof_dot.
      apply subof_dot; auto with static; [].
      simp_sub.
      apply dof_cn; eauto using ofm_fst; done.

      simp_sub.
      apply dof_md; eauto using mvalue_pure; done.

    simp_sub in Hofex.
    assumption.

(* ext *)
intros G t M Hof IH e' Hwf Hstep.              
#2 invert Hstep.
  intros M' Hstep' <-.
  apply oft_ext; [].
  exact (IH Hwf andel _ Hstep').

  intros _ <-.
  exact (IH Hwf ander andel _ _ (eq_refl _) (eq_refl _)).

(* var *)
intros G i T c _ Hwf.
do2 4 split; try (intros; discriminate); [].
intros M' Hstep.
invert Hstep; done.

(* star *)
intros G Hwf.
do2 4 split; try (intros; discriminate); [].
intros ? Hstep.
invert Hstep; done.

(* stat *)
intros G c k Hof Hwf.
do2 4 split; try (intros; discriminate); [].
intros ? Hstep.
invert Hstep; done.

(* dyn *)
intros G e t Hof IH Hwf.
#2 do2 4 split; try (intros; discriminate).
  intros M' Hstep.
  invert Hstep; [].
  intros e' Hstep' <-.
  apply ofm_dyn; [].
  eapply IH; eauto; done.

  intros e' t' Heq Heq'.
  injection Heq; injection Heq'; intros; subst e' t'.
  assumption.

(* lamgn *)
intros G T1 T2 k1 M HofT1 Hfst1 Hof IH Hwf.
#2 do2 4 split; try (intros; discriminate).
  intros M' Hstep.
  invert Hstep; done.

  intros U N T1' T2' Heq Heq'.
  injection Heq; injection Heq'; clear Heq Heq'.
  intros; subst U N T1' T2'.
  exists k1; auto; done.

(* appgn *)
intros G T1 T2 M1 M2 c2 Hof1 IH1 Hof2 IH2 HfstM2 Hwf.
do2 4 split; try (intros; discriminate); [].
intros M' Hstep.
so (IH1 Hwf andel) as IH1'.
so (IH2 Hwf andel) as IH2'.
#3 invert Hstep.
  intros; subst; eauto with static; done.

  intros M2' _ Hstep' <-.
  so (IH2 Hwf andel _ Hstep') as Hof2'.
  clear IH1.
  so (ofm_reg _#4 Hwf Hof1) as Hofs.
  invert Hofs; clear Hofs; [].
  intros k1 HofT1 Hfst HofT2.
  so (preservation_fst _#7 Hwf Hof2 Hstep' Hfst HfstM2) as (c2' & HfstM2' & Heqc).
  #2 apply (@ofm_subsume _ _ (subst1 c2' T2)).
    eapply ofm_appgn; eauto; done.

    apply subs_refl; [].
    apply eqs_symm; [].
    eapply funct_sg; eauto; done.

  clear IH2.
  intros T N c2' _ HfstM2' <- <-.
  so (fstm_fun _#4 HfstM2 HfstM2'); subst c2'.
  so (IH1 Hwf ander ander andel _#4 (eq_refl _) (eq_refl _)) as (k1 & Hfst & HofN).
  clear IH1.
  #2 sor (refine (subst_ofm G _ (dot M2 (dot c2 id)) _ _ _ _ HofN)) as H.
    #2 apply subof_dot.
      apply subof_dot; auto with static; [].
      simp_sub.
      apply dof_cn; [].
      exact (ofm_fst _#6 Hwf Hof2 HfstM2 Hfst).

      simp_sub.
      apply dof_md; auto; done.

    simp_sub in H.
    assumption.

(* lamap *)
intros G T1 T2 k1 M HofT1 Hfst1 Hof IH Hwf.
#2 do2 4 split; try (intros; discriminate).
  intros M' Hstep.
  invert Hstep; done.

  intros U N T1' T2' Heq Heq'.
  injection Heq; injection Heq'; clear Heq Heq'.
  intros; subst U N T1' T2'.
  exists k1; auto; done.

(* appap *)
intros G p T1 T2 M1 M2 c2 Hof1 IH1 Hof2 IH2 HfstM2 Hwf.
do2 4 split; try (intros; discriminate); [].
intros M' Hstep.
so (IH1 Hwf andel) as IH1'.
so (IH2 Hwf andel) as IH2'.
#3 invert Hstep.
  intros; subst; eauto with static; done.

  intros M2' _ Hstep' <-.
  so (IH2 Hwf andel _ Hstep') as Hof2'.
  clear IH1.
  so (ofm_reg _#4 Hwf Hof1) as Hofs.
  invert Hofs; clear Hofs; [].
  intros k1 HofT1 Hfst HofT2.
  so (preservation_fst _#7 Hwf Hof2 Hstep' Hfst HfstM2) as (c2' & HfstM2' & Heqc).
  #2 apply (@ofm_subsume _ _ (subst1 c2' T2)).
    eapply ofm_appap; eauto; done.

    apply subs_refl; [].
    apply eqs_symm; [].
    eapply funct_sg; eauto; done.

  clear IH2.
  intros T N c2' _ HfstM2' <- <-.
  so (fstm_fun _#4 HfstM2 HfstM2'); subst c2'.
  so (IH1 Hwf ander ander ander andel _#4 (eq_refl _) (eq_refl _)) as (k1 & Hfst & HofN).
  clear IH1.
  #2 sor (refine (subst_ofm G _ (dot M2 (dot c2 id)) _ _ _ _ HofN)) as H.
    #2 apply subof_dot.
      apply subof_dot; auto with static; [].
      simp_sub.
      apply dof_cn; [].
      exact (ofm_fst _#6 Hwf Hof2 HfstM2 Hfst).

      simp_sub.
      apply dof_md; auto; done.

    simp_sub in H.
    exact (ofm_forget_pure H).

(* pair *)
intros G p T1 T2 M1 M2 Hof1 IH1 Hof2 IH2 Hwf.
#2 do2 4 split; try (intros; discriminate).
  intros M' Hstep.
  #2 invert Hstep.
    intros M1' Hstep' <-.
    so (IH1 Hwf andel _ Hstep') as Hof1'.
    eapply ofm_pair; eauto; done.

    intros M2' _ Hstep' <-.
    so (IH2 Hwf andel _ Hstep') as Hof2'.
    eapply ofm_pair; eauto; done.

  intros N1 N2 T1' T2' Heq Heq' Hval1 _.
  injection Heq; injection Heq'; clear Heq Heq'; intros; subst N1 N2 T1' T2'.
  so (mvalue_pure _#4 Hval1 Hof1) as Hof1'.
  so (can_fstm _#3 Hwf Hof1') as (c1 & HfstM1).
  exists c1.
  do2 2 split; auto; [].
  simp_sub.
  assumption.

(* pi1 *)
intros G T1 T2 M Hof IH Hwf.
do2 4 split; try (intros; discriminate); [].
intros N Hstep.
#2 invert Hstep.
  intros M' Hstep' <-.
  apply (@ofm_pi1 _ _ T2); [].
  apply (IH Hwf andel); auto; done.

  rename N into V1.
  intros V2 Hval1 Hval2 <-.
  so (IH Hwf ander ander ander ander _#4 (eq_refl _) (eq_refl _) Hval1 Hval2) as (k & Hfsts & Hof1 & Hof2).
  assumption.

(* pi2 *)
intros G T1 T2 M c Hof IH HfstM Hwf.
do2 4 split; try (intros; discriminate); [].
intros N Hstep.
#2 invert Hstep.
  intros M' Hstep' <-.
  so (ofm_reg _#4 Hwf Hof) as Hofs.
  invert Hofs; clear Hofs.
  intros k1 HofT1 Hfst1 HofT2.
  so (can_fsts _ _ HofT2) as (k2 & Hfst2).
  so (IH Hwf andel _ Hstep') as Hof'.
  so (preservation_fst _#7 Hwf Hof Hstep' (fsts_sigma Hfst1 Hfst2) HfstM) as (c' & HfstM' & Heqc).
  clear IH.
  so (ofm_pi2 Hof' HfstM') as Hof''.
  eapply ofm_subsume; eauto; [].
  apply subs_refl; [].
  apply eqs_symm; [].
  eapply funct_sg; eauto; [].
  eapply eqc_pi1; eauto; done.

  rename N into V2.
  intros V1 Hval1 Hval2 <-.
  invert HfstM.
  intros c1 c2 HfstV1 HfstV2 <-.
  so (IH Hwf ander ander ander ander _#4 (eq_refl _) (eq_refl _) Hval1 Hval2) as (c1' & Hof1 & HfstV1' & Hof2).
  clear IH.
  so (fstm_fun _#4 HfstV1 HfstV1'); clear HfstV1'; subst c1'.
  eapply ofm_subsume; eauto; [].
  apply subs_refl; [].
  apply eqs_symm; [].
  so (ofm_reg _#4 Hwf Hof) as Hofs.
  invert Hofs; [].
  intros k1 HofT1 Hfst1 HofT2.
  so (ofm_fst _#6 Hwf Hof1 HfstV1 Hfst1) as Hofc1.
  eapply funct_sg; eauto; [].
  so (can_fsts _ _ HofT2) as (k2 & Hfst2).
  apply (@eqc_beta1 _ _ (subst1 c1 k2)); eauto; [].
  so (subst_fsts _ _ (dot c1 id) Hfst2) as Hfst2x.
  exact (ofm_fst _#6 Hwf Hof2 HfstV2 Hfst2x).

(* unpack *)
intros G k t T e M Hofe IH1 HofM IH2 HofT Hwf.
do2 4 split; try (intros; discriminate); [].
intros M' Hstep.
#2 invert Hstep.
  intros e' Hstep' <-.
  so (IH1 _ Hwf Hstep') as Hofe'.
  eapply ofm_unpack; eauto; done.

  intros c v k' t' _ <- <-.
  so (oft_invert _#3 Hwf Hofe) as (? & Heqc & Hoftr).
  invert Hoftr; [].
  intros Hofc Hofv Hoft' <-.
  so (exists_injective _#5 Hwf Heqc) as (Heqk & Heqt).
  #2 sor (refine (subst_ofm G _ (dot v (dot c id)) _ _ _ _ HofM)) as HofMx.
    #2 apply subof_dot.
      apply subof_dot; auto with static; [].
      simp_sub.
      apply dof_cn; eauto with static; done.

      simp_sub.
      apply dof_tm; [].
      eapply oft_equiv; eauto; [].
      repl kd_type with (subst1 c kd_type) by (simp_sub; reflexivity).
      eapply subst1_eqc; eauto; done.

    simp_sub in HofMx.
    assumption.

(* lett *)
intros G p t T e M Hofe IH1 HofM IH2 Hwf.
do2 4 split; try (intros; discriminate); [].
intros M' Hstep.
#2 invert Hstep.
  intros e1' Hstep' <-.
  so (IH1 _ Hwf Hstep') as Hofe'.
  eapply ofm_lett; eauto; done.

  intros _ <-.
  #2 sor (refine (subst_ofm G _ (dot e id) _ _ _ _ HofM)) as HofM'.
    apply subof_dot; auto with static; [].
    simp_sub.
    auto using dof; done.

    simp_sub in HofM'.
    assumption.

(* letm *)
intros G T1 T2 k1 M1 M2 Hof1 IH1 Hfst1 Hof2 IH2 HofT2 Hwf.
do2 4 split; try (intros; discriminate); [].
intros M' Hstep.
#2 invert Hstep.
  intros M1' Hstep' <-.
  clear IH2.
  so (IH1 Hwf andel _ Hstep') as Hof1'.
  eapply ofm_letm; eauto; done.

  clear IH1 IH2.
  intros c Hval HfstM1 <-.
  #2 sor (refine (subst_ofm G _ (dot M1 (dot c id)) _ _ _ _ Hof2)) as Hof2'.
    #2 apply subof_dot.
      apply subof_dot; auto with static; [].
      simp_sub.
      apply dof_cn; [].
      eapply ofm_fst; eauto; done.

      simp_sub.
      apply dof_md; eauto using mvalue_pure; done.

    simp_sub in Hof2'.
    assumption.

(* seal *)
intros G T M Hof IH Hwf.
do2 4 split; try (intros; discriminate); [].
intros M' Hstep.
invert Hstep; [].
intros <-.
assumption.

(* extstat *)
intros G k k' M c Hof IH Hfst Hofc Hwf.
do2 4 split; try (intros; discriminate); [].
intros M' Hstep.
so (IH Hwf andel _ Hstep) as Hof'.
so (mstep_fst _#4 Hstep Hfst) as [Hfst' | (c' & Hfst' & Hred)].
  eapply ofm_extstat; eauto; done.

  so (subject_reduction _#4 Hwf Hofc Hred) as Heqc.
  so (eqc_reg _#4 Hwf Heqc) as (_ & Hofc' & _).
  eapply ofm_extstat; eauto; done.

(* extpi *)
intros G T1 T2 T3 k1 M Hof1 IH1 Hfst Hof2 IH2 Hwf.
assert (wf (G; cl_cn k1; cl_md (shift1 T1) (var 0))) as Hwf'.
  apply wf_md; auto; [].
  so (ofm_reg _#4 Hwf Hof1) as Hofs.
  invert Hofs; auto; done.
#2 do2 4 split; try (intros; discriminate).
  intros M' Hstep.
  so (IH1 Hwf andel _ Hstep) as Hof1'.
  assert (mstep (G; cl_cn k1; cl_md (shift1 T1) (var 0)) (md_appap (subst (sh 2) M) (var 0)) (md_appap (subst (sh 2) M') (var 0))) as Hstep'.
    clear IH1 IH2.
    apply mstep_appap1; [].
    eapply subst_mstep; eauto; [].
    apply subof_sh; auto with sequence; done.
  so (IH2 Hwf' andel _ Hstep') as Hof2'.
  eapply ofm_extpi; eauto; done.

  clear IH1.
  intros U N T1' T2' -> Heq.
  injection Heq; intros; subst T1' T2'.
  #3 exploit (@mstep_appap3 (G; cl_cn k1; cl_md (shift1 T1) (var 0)) (subst (sh 2) U) (subst (dot (var 0) (dot (var 1) (sh 4))) N) (var 0) (var 1)) as Hstep.
    eapply mvalue_var; eauto with sequence; done.

    so (@fstm_var (G; cl_cn k1; cl_md (shift1 T1) (var 0)) 0 _ _ index_0) as H.
    simp_sub in H.
    assumption.
  simp_sub in Hstep.
  rewrite -> subst_dot01_sh in Hstep.
  simp_sub in IH2.
  so (IH2 Hwf' andel _ Hstep) as H.
  exists k1; auto; done.

(* extsigma *)
intros G T1 T2 M Hof1 IH1 Hof2 IH2 Hwf.
#2 do2 4 split; try (intros; discriminate).
  intros M' Hstep.
  so (IH1 Hwf andel _ (mstep_pi11 Hstep)) as Hof1'.
  so (IH2 Hwf andel _ (mstep_pi21 Hstep)) as Hof2'.
  apply ofm_extsigma; auto; done.

  intros M1 M2 T1' T2' -> Heq Hval1 Hval2.
  injection Heq; clear Heq; intros; subst T1' T2'.
  so (IH1 Hwf andel _ (mstep_pi12 Hval1 Hval2)) as Hof1'.
  so (IH2 Hwf andel _ (mstep_pi22 Hval1 Hval2)) as Hof2'.
  clear IH1 IH2.
  so (can_fstm _#3 Hwf Hof1') as (c1 & HfstM1).
  exists c1.
  do2 2 split; auto; [].
  simp_sub.
  assumption.

(* forget *)
intros G T M _ IH Hwf.
do2 4 split.
  intros M' Hstep.
  apply ofm_forget; [].
  apply (IH Hwf andel); auto; done.

  exact (IH Hwf ander andel).

  exact (IH Hwf ander ander andel).

  exact (IH Hwf ander ander ander andel).

  intros N1 N2 T1 T2 Heq Heq' Hval1 Hval2.
  so (IH Hwf ander ander ander ander N1 N2 T1 T2 Heq Heq' Hval1 Hval2) as (c1 & Hof1 & Hfstm & Hof2).
  exists c1; auto using ofm_forget; done.

(* subsume *)
intros G p T T' M _ IH Hsubs Hwf.
do2 4 split.
  intros M' Hstep.
  eapply ofm_subsume; eauto; [].
  apply (IH Hwf andel); auto; done.

  (* dyn *)
  intros e t -> ->.
  so (subs_dyn_invert' _#3 Hsubs) as (t' & -> & Heqc).
  so (IH Hwf ander andel _ _ (eq_refl _) (eq_refl _)).
  eapply oft_equiv; eauto with static; done.

  (* pigen *)
  intros U N T1 T2 -> ->.
  so (subs_pigen_invert' _#4 Hwf Hsubs) as (T1' & T2' & k1 & -> & HfstT1 & Hsubs1 & Hsubs2).
  exists k1; split; auto; [].
  so (IH Hwf ander ander andel _#4 (eq_refl _) (eq_refl _)) as (k1' & HfstT1' & Hof).
  clear IH.
  #2 sor (refine (subst_ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) _ (under 2 id) _ _ _ _ Hof)) as Hof'.
    rewrite -> under_succ; [].
    #2 apply subof_dot.
      simp_sub.
      apply subof_dot; auto with static sequence; [].
      simp_sub.
      apply dof_cn; [].
      apply (@ofc_subsume _ (subst (sh 2) k1)); auto with static sequence; [].
      refine (subst_subk _#5 (weaken2 _#4) _); [].
      eapply subs_fst; eauto; done.

      simp_sub.
      #2 apply dof_md.
        #2 apply (@ofm_subsume _ _ (shift1 (shift1 T1))).
          eapply ofm_var; eauto with sequence; done.

          simp_sub.
          exact (subst_subs _#5 (weaken2 _#4) Hsubs1).

        repl (var 1) with (subst (sh 1) (var 0)) by (simp_sub; reflexivity).
        eapply fstm_var; eauto with sequence; done.

    simp_sub in Hof'.
    apply (@ofm_subsume _ _ (shift1 T2')); auto; [].
    apply (subst_subs _#5 (weaken1 _#3)); auto; done.

  (* piapp *)
  intros U N T1 T2 -> ->.
  so (subs_piapp_invert' _#4 Hwf Hsubs) as (T1' & T2' & k1 & -> & HfstT1 & Hsubs1 & Hsubs2).
  exists k1; split; auto; [].
  so (IH Hwf ander ander ander andel _#4 (eq_refl _) (eq_refl _)) as (k1' & HfstT1' & Hof).
  clear IH.
  #2 sor (refine (subst_ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) _ (under 2 id) _ _ _ _ Hof)) as Hof'.
    rewrite -> under_succ; [].
    #2 apply subof_dot.
      simp_sub.
      apply subof_dot; auto with static sequence; [].
      simp_sub.
      apply dof_cn; [].
      apply (@ofc_subsume _ (subst (sh 2) k1)); auto with static sequence; [].
      refine (subst_subk _#5 (weaken2 _#4) _); [].
      eapply subs_fst; eauto; done.

      simp_sub.
      #2 apply dof_md.
        #2 apply (@ofm_subsume _ _ (shift1 (shift1 T1))).
          eapply ofm_var; eauto with sequence; done.

          simp_sub.
          exact (subst_subs _#5 (weaken2 _#4) Hsubs1).

        repl (var 1) with (subst (sh 1) (var 0)) by (simp_sub; reflexivity).
        eapply fstm_var; eauto with sequence; done.

    simp_sub in Hof'.
    apply (@ofm_subsume _ _ (shift1 T2')); auto; [].
    apply (subst_subs _#5 (weaken1 _#3)); auto; done.
    
  (* sigma *)
  intros N1 N2 T1 T2 -> -> Hval1 Hval2.
  so (subs_sigma_invert' _#4 Hwf Hsubs) as (T1' & T2' & k1' & -> & Hfst1' & Hsubs1 & Hsubs2).
  so (can_fsts _ _ (subs_reg _#3 Hwf Hsubs1 ander)) as (k1 & Hfst1).
  so (IH Hwf ander ander ander ander _#4 (eq_refl _) (eq_refl _) Hval1 Hval2) as (c1 & Hof1 &  Hfstm & Hof2).
  #3 exists c1; do2 2 split.
    eapply ofm_subsume; eauto; done.

    assumption.

    eapply ofm_subsume; eauto; [].
    eapply subst1_subs; eauto; [].
    exact (ofm_fst _#6 Hwf Hof1 Hfstm Hfst1').

(* epilogue *)
destruct Hind as (IH1 & IH2).
split.
  eauto; done.

  intros G p M M' T Hwf Hofm Hstep.
  exact (IH2 _#4 Hofm Hwf andel _ Hstep).
Qed.


Lemma preservation_tm :
  forall G e e' t,
    wf G
    -> oft G e t
    -> tstep G e e'
    -> oft G e' t.
Proof.
exact (preservation andel).
Qed.


Lemma preservation_md :
  forall G p M M' T,
    wf G
    -> ofm G p M T
    -> mstep G M M'
    -> ofm G p M' T.
Proof.
exact (preservation ander).
Qed.


Lemma preservations_tm :
  forall G e e' t,
    wf G
    -> oft G e t
    -> star (tstep G) e e'
    -> oft G e' t.
Proof.
intros G e e' t Hwf Hof Hsteps.
revert Hof.
induct Hsteps; eauto using preservation_tm.
Qed.


Lemma preservations_md :
  forall G p M M' T,
    wf G
    -> ofm G p M T
    -> star (mstep G) M M'
    -> ofm G p M' T.
Proof.
intros G p M M' t Hwf Hof Hsteps.
revert Hof.
induct Hsteps; eauto using preservation_md.
Qed.


Lemma preservations_fst :
  forall G p M M' T c k,
    wf G
    -> ofm G p M T
    -> star (mstep G) M M'
    -> fsts T k
    -> fstm G M c
    -> exists c',
         fstm G M' c'
         /\ eqc G c c' k.
Proof.
intros G p M M' T c k Hwf HofM Hsteps HfstT HfstM.
revert c HofM HfstM.
induct Hsteps.

(* refl *)
intros M c HofM HfstM.
so (ofm_fst _#6 Hwf HofM HfstM HfstT) as Hofc.
eauto using eqc_refl; done.

(* step *)
intros M1 M2 M3 Hstep _ IH c1 HofM1 HfstM.
so (preservation_fst _#7 Hwf HofM1 Hstep HfstT HfstM) as (c2 & HfstM2 & Heqc).
so (preservation_md _#5 Hwf HofM1 Hstep) as HofM2.
so (IH _ HofM2 HfstM2) as (c3 & HfstM3 & Heqc').
eauto using eqc_trans; done.
Qed.


Lemma preservations_fst' :
  forall G p M M' T c c' k,
    wf G
    -> ofm G p M T
    -> star (mstep G) M M'
    -> fsts T k
    -> fstm G M c
    -> fstm G M' c'
    -> eqc G c c' k.
Proof.
intros G p M M' T c c' k Hwf HofM Hsteps HfstT HfstM HfstM'.
so (preservations_fst _#7 Hwf HofM Hsteps HfstT HfstM) as (c'' & HfstM'' & Heqc).
so (fstm_fun _#4 HfstM' HfstM''); subst c''.
assumption.
Qed.
