
Require Import Tactics.
Require Import Sequence.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import Regularity.
Require Import Declarative.
Require Import Consistent.
Require Import Preservation.


Inductive tcf : (* tm *) term -> (* cn *) term -> Prop :=
| tcf_unit
    : tcf tm_star cn_unit

| tcf_arrow {t e t1 t2}
    : tcf (tm_lam t e) (cn_arrow t1 t2)

| tcf_prod {e1 e2 t1 t2}
    : tcf (tm_pair e1 e2) (cn_prod t1 t2)

| tcf_all {k e k' t}
    : tcf (tm_plam k e) (cn_all k' t)

| tcf_exists {c e k t k' t'}
    : tcf (tm_pack c e k t) (cn_exists k' t')
  
.


Inductive mcf : (* md *) term -> (* cn *) term -> Prop :=
| mcf_one
    : mcf md_star sg_one

| mcf_stat {c k}
    : mcf (md_stat c) (sg_stat k)

| msf_dyn {e t}
    : mcf (md_dyn e) (sg_dyn t)

| mcf_pigen {T M T1 T2}
    : mcf (md_lamgn T M) (sg_pigen T1 T2)

| mcf_piapp {T M T1 T2}
    : mcf (md_lamap T M) (sg_piapp T1 T2)

| mcf_sigma {M1 M2 T1 T2}
    : mcf (md_pair M1 M2) (sg_sigma T1 T2)
.


Lemma canon_tm :
  forall e t,
    oft nil e t
    -> similar t t
    -> tvalue nil e
    -> tcf e t.
Proof.
intros e t Hof Hsimt Hval.
so (oft_invert _#3 wf_nil Hof) as (t' & Heqc & Hofr).
assert (similar t' t') as Hsimt'.
  invert Hofr; try (intros; subst; invert Hval; done); try (intros; subst; auto using similar; done); [].
  (* var *)
  intros ? ? Hindex _ _.
  invert Hindex; done.
so (consistent _#3 wf_nil Heqc Hsimt' Hsimt) as Hsim.
invert Hofr;
try (intros; subst; invert Hval; done);
try (intros; subst; invert Hsim; intros; subst; auto using tcf; done).
(* var *)
intros i t'' Hindex _ _.
invert Hindex; done.
Qed.


Lemma mcf_subs :
  forall M T T',
    subs nil T T'
    -> mcf M T
    -> mcf M T'.
Proof.
intros M T T' Hsubs Hcf.
(* We use this contortion, rather than just inverting Hcf, so that
   we don't have to establish the converses for subs_X_invert'.
*)
#6 invert (subs_reg _#3 wf_nil Hsubs ander).

(* one *)
intros <-.
so (subs_one_invert' _ _ Hsubs); subst T.
invert Hcf; intros; subst; auto using mcf; done.

(* stat *)
intros k _ <-.
so (subs_stat_invert' _#3 Hsubs) as (k' & -> & _).
invert Hcf; intros; subst; auto using mcf; done.

(* dyn *)
intros t _ <-.
so (subs_dyn_invert' _#3 Hsubs) as (t' & -> & _).
invert Hcf; intros; subst; auto using mcf; done.

(* pigen *)
intros T1 T2 _ _ _ _ <-.
so (subs_pigen_invert' _#4 wf_nil Hsubs) as (T1' & T2' & _ & -> & _).
invert Hcf; intros; subst; auto using mcf; done.

(* piapp *)
intros T1 T2 _ _ _ _ <-.
so (subs_piapp_invert' _#4 wf_nil Hsubs) as (T1' & T2' & _ & -> & _).
invert Hcf; intros; subst; auto using mcf; done.

(* sigma *)
intros T1 T2 _ _ _ _ <-.
so (subs_sigma_invert' _#4 wf_nil Hsubs) as (T1' & T2' & _ & -> & _).
invert Hcf; intros; subst; auto using mcf; done.
Qed.


Inductive elimcon : ((* md *) term -> (* md *) term) -> Prop :=
| elimcon_null 
    : elimcon (fun M => M)

| elimcon_pi1 {E}
    : elimcon E
      -> elimcon (fun M => md_pi1 (E M))
.


(* Induction principle that adds an extra premise to the extsigma case. *)
Lemma dynamic_ind' :
  forall (Pt : context -> term -> term -> Prop)
         (Pm : context -> purity -> term -> term -> Prop),
    (forall (G : list term) (i : nat) (t : term),
     index i G (cl_tm t) -> Pt G (var i) (subst (sh (S i)) t)) ->
    (forall G : context, Pt G tm_star cn_unit) ->
    (forall G (t1 t2 e : term),
     ofc G t1 kd_type ->
     oft (G; cl_tm t1) e (shift1 t2) ->
     Pt (G; cl_tm t1) e (shift1 t2) -> Pt G (tm_lam t1 e) (cn_arrow t1 t2)) ->
    (forall G (t1 t2 e1 e2 : term),
     oft G e1 (cn_arrow t1 t2) ->
     Pt G e1 (cn_arrow t1 t2) ->
     oft G e2 t1 -> Pt G e2 t1 -> Pt G (tm_app e1 e2) t2) ->
    (forall G (t1 t2 e1 e2 : term),
     oft G e1 t1 ->
     Pt G e1 t1 ->
     oft G e2 t2 -> Pt G e2 t2 -> Pt G (tm_pair e1 e2) (cn_prod t1 t2)) ->
    (forall G (t1 t2 e : term),
     oft G e (cn_prod t1 t2) -> Pt G e (cn_prod t1 t2) -> Pt G (tm_pi1 e) t1) ->
    (forall G (t1 t2 e : term),
     oft G e (cn_prod t1 t2) -> Pt G e (cn_prod t1 t2) -> Pt G (tm_pi2 e) t2) ->
    (forall G (k t e : term),
     ofk G k ->
     oft (G; cl_cn k) e t ->
     Pt (G; cl_cn k) e t -> Pt G (tm_plam k e) (cn_all k t)) ->
    (forall G (k t e c : term),
     oft G e (cn_all k t) ->
     Pt G e (cn_all k t) -> ofc G c k -> Pt G (tm_papp e c) (subst1 c t)) ->
    (forall G (k t c e : term),
     ofc G c k ->
     oft G e (subst1 c t) ->
     Pt G e (subst1 c t) ->
     ofc (G; cl_cn k) t kd_type -> Pt G (tm_pack c e k t) (cn_exists k t)) ->
    (forall G (k t t' e1 e2 : term),
     oft G e1 (cn_exists k t) ->
     Pt G e1 (cn_exists k t) ->
     oft (G; cl_cn k; cl_tm t) e2 (subst (sh 2) t') ->
     Pt (G; cl_cn k; cl_tm t) e2 (subst (sh 2) t') ->
     ofc G t' kd_type -> Pt G (tm_unpack e1 e2) t') ->
    (forall G (t e : term),
     oft G e (cn_arrow (cn_arrow cn_unit t) t) ->
     Pt G e (cn_arrow (cn_arrow cn_unit t) t) -> Pt G (tm_fix t e) t) ->
    (forall G (t1 t2 e1 e2 : term),
     oft G e1 t1 ->
     Pt G e1 t1 ->
     oft (G; cl_tm t1) e2 (shift1 t2) ->
     Pt (G; cl_tm t1) e2 (shift1 t2) -> Pt G (tm_lett e1 e2) t2) ->
    (forall G (t t' e : term),
     oft G e t -> Pt G e t -> eqc G t t' kd_type -> Pt G e t') ->
    (forall G (T k t M e : term),
     ofm G impure M T ->
     Pm G impure M T ->
     fsts T k ->
     oft (G; cl_cn k; cl_md (shift1 T) (var 0)) e (subst (sh 2) t) ->
     Pt (G; cl_cn k; cl_md (shift1 T) (var 0)) e (subst (sh 2) t) ->
     ofc G t kd_type -> Pt G (tm_letm M e) t) ->
    (forall G (t M : term),
     ofm G impure M (sg_dyn t) ->
     Pm G impure M (sg_dyn t) -> Pt G (tm_ext M) t) ->
    (forall (G : list term) (i : nat) (T c : term),
     index i G (cl_md T c) -> Pm G pure (var i) (subst (sh (S i)) T)) ->
    (forall G : context, Pm G pure md_star sg_one) ->
    (forall G (c k : term),
     ofc G c k -> Pm G pure (md_stat c) (sg_stat k)) ->
    (forall G (e t : term),
     oft G e t -> Pt G e t -> Pm G pure (md_dyn e) (sg_dyn t)) ->
    (forall G (T1 T2 k1 M : term),
     ofs G T1 ->
     fsts T1 k1 ->
     ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M (shift1 T2) ->
     Pm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M (shift1 T2) ->
     Pm G pure (md_lamgn T1 M) (sg_pigen T1 T2)) ->
    (forall G (T1 T2 M1 M2 c2 : term),
     ofm G impure M1 (sg_pigen T1 T2) ->
     Pm G impure M1 (sg_pigen T1 T2) ->
     ofm G pure M2 T1 ->
     Pm G pure M2 T1 ->
     fstm G M2 c2 -> Pm G impure (md_appgn M1 M2) (subst1 c2 T2)) ->
    (forall G (T1 T2 k1 M : term),
     ofs G T1 ->
     fsts T1 k1 ->
     ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) pure M (shift1 T2) ->
     Pm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) pure M (shift1 T2) ->
     Pm G pure (md_lamap T1 M) (sg_piapp T1 T2)) ->
    (forall G p (T1 T2 M1 M2 c2 : term),
     ofm G p M1 (sg_piapp T1 T2) ->
     Pm G p M1 (sg_piapp T1 T2) ->
     ofm G pure M2 T1 ->
     Pm G pure M2 T1 ->
     fstm G M2 c2 -> Pm G p (md_appap M1 M2) (subst1 c2 T2)) ->
    (forall G p (T1 T2 M1 M2 : term),
     ofm G p M1 T1 ->
     Pm G p M1 T1 ->
     ofm G p M2 T2 ->
     Pm G p M2 T2 -> Pm G p (md_pair M1 M2) (sg_sigma T1 (shift1 T2))) ->
    (forall G (T1 T2 M : term),
     ofm G pure M (sg_sigma T1 T2) ->
     Pm G pure M (sg_sigma T1 T2) -> Pm G pure (md_pi1 M) T1) ->
    (forall G (T1 T2 M c : term),
     ofm G pure M (sg_sigma T1 T2) ->
     Pm G pure M (sg_sigma T1 T2) ->
     fstm G M c -> Pm G pure (md_pi2 M) (subst1 (cn_pi1 c) T2)) ->
    (forall G (k t T e M : term),
     oft G e (cn_exists k t) ->
     Pt G e (cn_exists k t) ->
     ofm (G; cl_cn k; cl_tm t) impure M (subst (sh 2) T) ->
     Pm (G; cl_cn k; cl_tm t) impure M (subst (sh 2) T) ->
     ofs G T -> Pm G impure (md_unpack e M T) T) ->
    (forall G p (t T e M : term),
     oft G e t ->
     Pt G e t ->
     ofm (G; cl_tm t) p M (shift1 T) ->
     Pm (G; cl_tm t) p M (shift1 T) -> Pm G p (md_lett e M) T) ->
    (forall G (T1 T2 k1 M1 M2 : term),
     ofm G impure M1 T1 ->
     Pm G impure M1 T1 ->
     fsts T1 k1 ->
     ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M2
       (subst (sh 2) T2) ->
     Pm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M2
       (subst (sh 2) T2) -> ofs G T2 -> Pm G impure (md_letm M1 M2 T2) T2) ->
    (forall G (T M : term),
     ofm G impure M T -> Pm G impure M T -> Pm G impure (md_seal M T) T) ->
    (forall G (k k' M c : term),
     ofm G pure M (sg_stat k') ->
     Pm G pure M (sg_stat k') ->
     fstm G M c -> ofc G c k -> Pm G pure M (sg_stat k)) ->
    (forall G (T1 T2 T3 k1 M : term),
     ofm G pure M (sg_piapp T1 T3) ->
     Pm G pure M (sg_piapp T1 T3) ->
     fsts T1 k1 ->
     ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) pure
       (md_appap (subst (sh 2) M) (var 0)) (shift1 T2) ->
     Pm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) pure
       (md_appap (subst (sh 2) M) (var 0)) (shift1 T2) ->
     Pm G pure M (sg_piapp T1 T2)) ->
    (forall G (T1 T2 T3 T4 M : term),
     ofm G pure M (sg_sigma T3 T4) ->
     Pm G pure M (sg_sigma T3 T4) ->
     ofm G pure (md_pi1 M) T1 ->
     Pm G pure (md_pi1 M) T1 ->
     ofm G pure (md_pi2 M) T2 ->
     Pm G pure (md_pi2 M) T2 -> Pm G pure M (sg_sigma T1 (shift1 T2))) ->
    (forall G (T M : term),
     ofm G pure M T -> Pm G pure M T -> Pm G impure M T) ->
    (forall G p (T T' M : term),
     ofm G p M T -> Pm G p M T -> subs G T T' -> Pm G p M T') ->
    (forall G (t t0 : term), oft G t t0 -> Pt G t t0) /\
    (forall G p (t t0 : term), ofm G p t t0 -> Pm G p t t0).
Proof.
intros Pt Pm.
intros.
#3 exploit
  (dynamic_ind
     Pt
     (fun G p M T =>
        True
        /\
        Pm G p M T
        /\
        (forall E M',
           elimcon E
           -> p = pure
           -> M = E (md_pi1 M')
           -> exists T3 T4,
                ofm G pure M' (sg_sigma T3 T4)
                /\ Pm G pure M' (sg_sigma T3 T4)))) as Hind;
  try (intros; 
       repeat (lazymatch goal with
               | H : True /\ _ |- _ => destruct H as (_ & ? & ?)
               end);
       try (split; [| split]);
       try (intros; let HE := findhyp elimcon in invert HE; intros; subst; discriminate);
       eauto; done).

(* pi1 *)
intros G T1 T2 M Hof (_ & IH & IH').
do2 2 split; eauto; [].
intros E M' HE _ Heq.
#2 invert HE.
  intros <-.
  injection Heq; intros; subst M'.
  eauto; done.

  intros E' HE' <-.
  injection Heq.
  intro.
  apply (IH' E'); auto; done.

(* extsigma *)
intros G T1 T2 M Hof1 (_ & IH1 & IH1') Hof2 (_ & IH2 & _).
so (IH1' _ _ elimcon_null (eq_refl _) (eq_refl _)) as (T3 & T4 & Hof & Hm).
do2 2 split; eauto; [].
intros E M' HE _ ->.
apply (IH1' _ _ (elimcon_pi1 HE)); eauto; done.

(* epilogue *)
destruct Hind as (? & Hindm).
split; eauto; [].
intros G p t M Hof.
so (Hindm _#4 Hof) as (_ & ? & _).
assumption.
Qed.


Lemma canon_md :
  forall p M T,
    ofm nil p M T
    -> mvalue nil M
    -> mcf M T.
Proof.
#6 exploit
  (dynamic_ind'
     (fun _ _ _ => True)
     (fun G p M T =>
        G = nil
        -> mvalue nil M
        -> mcf M T)) as Hind;
  trivial;
  try (intros; let Hval := findhyp mvalue in invert Hval; done);
  auto using mcf.

(* var *)
intros G i T c Hindex -> _.
invert Hindex; done.

(* stat *)
intros G k k' M c _ IH _ _ -> Hval.
so (IH (eq_refl _) Hval) as Hcf.
invert Hcf; intros; subst; auto using mcf; done.

(* extpi *)
intros G T1 T2 T3 k1 M _ IH _ _ _ -> Hval.
so (IH (eq_refl _) Hval) as Hcf.
invert Hcf; intros; subst; auto using mcf; done.

(* extsigma *)
intros G T1 T2 T3 T4 M _ IH _ _ _ _ -> Hval.
so (IH (eq_refl _) Hval) as Hcf.
invert Hcf; intros; subst; auto using mcf; done.

(* subsume *)
intros G p T T' M _ IH Hsubs -> Hval.
so (IH (eq_refl _) Hval) as Hcf.
eapply mcf_subs; eauto; done.

(* epilogue *)
destruct Hind.
intros; eauto; done.
Qed.


Theorem progress :
  (forall e t,
     oft nil e t
     -> tvalue nil e \/ exists e', tstep nil e e')
  /\
  (forall p M T,
     ofm nil p M T
     -> mvalue nil M \/ exists M', mstep nil M M').
Proof.
#23 exploit
  (dynamic_ind
     (fun G e t =>
        G = nil
        -> tvalue nil e \/ exists e', tstep nil e e')
     (fun G p M T =>
        G = nil
        -> mvalue nil M \/ exists M', mstep nil M M')) as Hind;
try (intros; left; auto with dynamic; done);
eauto with dynamic.

(* var *)
intros G i t Hindex ->.
invert Hindex; done.

(* app *)
intros G t1 t2 e1 e2 Hof1 IH1 _ IH2 ->.
right.
so (IH1 eqrefl) as [Hval1 | ()]; eauto with dynamic; [].
so (IH2 eqrefl) as [Hval2 | ()]; eauto with dynamic; [].
so (canon_tm _ _ Hof1 similar_arrow Hval1) as Hcf.
invert Hcf; [].
intros t e <-.
eauto with dynamic; done.

(* pair *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 ->.
so (IH1 eqrefl) as [Hval1 | ()]; eauto with dynamic; [].
so (IH2 eqrefl) as [Hval2 | ()]; eauto with dynamic; done.

(* pi1 *)
intros G t1 t2 e Hof IH ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_tm _ _ Hof similar_prod Hval) as Hcf.
invert Hcf; [].
intros e1 e2 <-.
invert Hval; [].
eauto with dynamic; done.

(* pi2 *)
intros G t1 t2 e Hof IH ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_tm _ _ Hof similar_prod Hval) as Hcf.
invert Hcf; [].
intros e1 e2 <-.
invert Hval; [].
eauto with dynamic; done.

(* papp *)
intros G k t e c Hof IH _ ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_tm _ _ Hof similar_all Hval) as Hcf.
invert Hcf; [].
intros; subst; eauto with dynamic; done.

(* pack *)
intros G k t c e _ _ IH _ ->.
so (IH eqrefl) as [| ()]; eauto with dynamic; done.

(* unpack *)
intros G k t t' e1 e2 Hof IH _ _ _ ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_tm _ _ Hof similar_exists Hval) as Hcf.
invert Hcf; [].
intros c e k0 t0 <-.
invert Hval; [].
eauto with dynamic; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH _ _ ->.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; done.

(* letm *)
intros G T k t M e Hof IH _ _ _ _ ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (mvalue_pure _#4 Hval Hof) as Hof'.
so (can_fstm _#3 wf_nil Hof') as (c & Hfst).
eauto with dynamic; done.

(* ext *)
intros G t M Hof IH ->.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_md _#3 Hof Hval) as Hcf.
invert Hcf; [].
intros e <-.
invert Hval; [].
eauto with dynamic; done.

(* var *)
intros G i T c Hindex ->.
invert Hindex; done.

(* dyn *)
intros G e t _ IH ->.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; done.

(* appgn *)
intros G T1 T2 M1 M2 c2 Hof IH1 _ IH2 Hfst ->.
right.
so (IH1 eqrefl) as [Hval1 | ()]; eauto with dynamic; [].
so (IH2 eqrefl) as [Hval2 | ()]; eauto with dynamic; [].
so (canon_md _ _ _ Hof Hval1) as Hcf.
invert Hcf; [].
intros T M <-.
eauto with dynamic; done.

(* appap *)
intros G p T1 T2 M1 M2 c2 Hof IH1 _ IH2 Hfst ->.
right.
so (IH1 eqrefl) as [Hval1 | ()]; eauto with dynamic; [].
so (IH2 eqrefl) as [Hval2 | ()]; eauto with dynamic; [].
so (canon_md _ _ _ Hof Hval1) as Hcf.
invert Hcf; [].
intros T M <-.
eauto with dynamic; done.

(* pair *)
intros G p T1 T2 M1 M2 _ IH1 _ IH2 ->.
so (IH1 eqrefl) as [Hval1 | ()]; eauto with dynamic; [].
so (IH2 eqrefl) as [Hval2 | (M2' & Hstep2)]; eauto with dynamic; done.

(* pi1 *)
intros G T1 T2 M Hof IH ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_md _ _ _ Hof Hval) as Hcf.
invert Hcf; [].
intros M1 M2 <-.
invert Hval; [].
eauto with dynamic; done.

(* pi2 *)
intros G T1 T2 M c Hof IH _ ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_md _ _ _ Hof Hval) as Hcf.
invert Hcf; [].
intros M1 M2 <-.
invert Hval; [].
eauto with dynamic; done.

(* unpack *)
intros G k t T e M Hof IH _ _ _ ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_tm _ _ Hof similar_exists Hval) as Hcf.
invert Hcf; [].
intros c e0 k0 t0 <-.
invert Hval; [].
eauto with dynamic; done.

(* lett *)
intros G p t T e M _ IH _ _ ->.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; done.

(* letm *)
intros G T1 T2 k1 M1 M2 Hof IH _ _ _ _ ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (mvalue_pure _#4 Hval Hof) as Hof'.
so (can_fstm _#3 wf_nil Hof') as (c & Hfst).
eauto with dynamic; done.

(* extsigma *)
intros G T1 T2 M _ IH _ _ ->.
so (IH eqrefl) as [Hval | (N & Hstep)].
  invert Hval; auto; done.
  
  #2 invert Hstep.
    eauto; done.

    intros; subst.
    eauto with dynamic; done.

(* epilogue *)
destruct_all Hind.
split; eauto; done.
Qed.


Lemma progress_tm :    
  forall e t,
    oft nil e t
    -> tvalue nil e \/ exists e', tstep nil e e'.
Proof.
exact (progress andel).
Qed.


Lemma progress_md :
  forall p M T,
    ofm nil p M T
    -> mvalue nil M \/ exists M', mstep nil M M'.
Proof.
exact (progress ander).
Qed.
