
Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import TSubst.
Require Import Formation.
Require Import Regularity.
Require Import Declarative.


Fixpoint single (c : (* cn *) term) (k : (* kd *) term) {struct k} :=
  (match k with
   | kd_pi k1 k2 =>
       kd_pi k1 (single (cn_app (shift1 c) (var 0)) k2)

   | kd_sigma k1 k2 =>
       kd_sigma
       (single (cn_pi1 c) k1) 
       (subst (dot (cn_pi1 (shift1 c)) sh1) (single (cn_pi2 (shift1 c)) k2))

   | kd_one => kd_one

   | _ => kd_sing c
   end).


(* Doesn't need any premises. *)
Lemma shift_under_single :
  forall i j c k,
    subst (under i (sh j)) (single c k) = single (subst (under i (sh j)) c) (subst (under i (sh j)) k).
Proof.
intros i j c k.
revert i c.
#3 induct k; try (intros; simp_sub; simpl; auto; done).

(* var *)
intros n i c.
simp_sub.
so (Compare_dec.dec_lt n i) as [Hlt | Hgeq].
  rewrite -> project_under_lt; auto; done.

  rewrite -> project_under_ge; [| omega].
  simp_sub.
  simpl; reflexivity.

(* pi *)
intros k1 _ k2 IH i c.
simp_sub.
simpl.
simp_sub.
f_equal; [].
rewrite <- under_succ; [].
rewrite -> IH; [].
simp_sub.
reflexivity.

(* sigma *)
intros k1 IH1 k2 IH2 i c.
simp_sub.
simpl.
simp_sub.
#2 f_equal.
  rewrite -> IH1; [].
  simp_sub.
  reflexivity.

  rewrite -> split_dot.
  rewrite -> subst_compose.
  repl (dot (cn_pi1 (subst (compose (under i (sh j)) sh1) c)) sh1) with (compose (dot (var 0) (sh 2)) (dot (cn_pi1 (subst (compose (under i (sh j)) sh1) c)) id)) by (simp_sub; reflexivity).
  rewrite -> (subst_compose _ _ (dot (var 0) (sh 2))).
  f_equal; [].
  simp_sub.
  repl (dot (var 0) (compose (under i (sh j)) (sh 2))) with (compose (dot (var 0) (compose (under i (sh j)) sh1)) (dot (var 0) (sh 2))) by (simp_sub; reflexivity).
  rewrite <- under_succ.
  rewrite -> subst_compose.
  rewrite -> IH2; [].
  f_equal; [].
  simp_sub.
  reflexivity.
Qed.


Lemma subst_single :
  forall s c k,
    hkd k
    -> subst s (single c k) = single (subst s c) (subst s k).
Proof.
intros s c k Hhkd.
revert s c.
induct Hhkd.

(* type *)
intros.
simp_sub.
simpl.
simp_sub.
reflexivity.

(* sing *)
intros b s c.
simp_sub.
simpl.
simp_sub.
reflexivity.

(* pi *)
intros k1 k2 _ _ _ IH s c.
simp_sub.
simpl.
simp_sub.
f_equal; [].
rewrite -> IH; [].
simp_sub.
reflexivity.

(* sigma *)
intros k1 k2 _ IH1 _ IH2 s c.
simp_sub.
simpl.
simp_sub.
#2 f_equal.
  rewrite -> IH1; [].
  simp_sub.
  reflexivity.

  repl (cn_pi2 (subst (compose s sh1) c)) with (subst (dot (var 0) (compose s sh1)) (shift1 (cn_pi2 c))) by (simp_sub; reflexivity).
  rewrite <- IH2; [].
  simp_sub.
  reflexivity.

(* one *)
intros s c.
simp_sub.
simpl.
simp_sub.
reflexivity.
Qed.


Lemma single_omnibus :
  forall G c k,
    wf G
    -> ofc G c k
    -> subk G (single c k) k
       /\ eqc (G; cl_cn (single c k)) (var 0) (shift1 c) (shift1 (single c k)).
Proof.
intros G c k Hwf Hofc.
so (ofk_hkd _ _ (ofc_reg _#3 Hwf Hofc)) as Hhkd.
remember (skel k) as sk eqn:Hskel.
revert G c k Hskel Hhkd Hwf Hofc.
induct sk.

(* type *)
intros G c k Hskel Hhkd Hwf Hofc.
invert Hhkd; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros <-.
split.
  simpl.
  apply subk_sing_t; auto; done.
  
  simpl.
  simp_sub.
  apply eqc_symm; [].
  apply eqc_sing; [].
  apply eqc_symm; [].
  apply eqc_singelim; [].
  repl (kd_sing (shift1 c)) with (shift1 (kd_sing c)) by (simp_sub; reflexivity).
  apply ofc_var; auto using index_0; done.

(* sing *)
intros G c k Hskel Hhkd Hwf Hofc.
invert Hhkd; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros b <-.
split.
  simpl.
  apply subk_refl; [].
  apply eqk_sing; [].
  apply eqc_singelim; auto; done.

  simpl.
  simp_sub.
  apply eqc_symm; [].
  apply eqc_sing; [].
  apply eqc_symm; [].
  apply eqc_singelim; [].
  repl (kd_sing (shift1 c)) with (shift1 (kd_sing c)) by (simp_sub; reflexivity).
  apply ofc_var; auto using index_0; done.

(* pi *)
intros sk1 _ sk2 IH2 G c k Hskel Hhkd Hwf Hofc.
invert Hhkd; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros k1 k2 _ Hhkd2 <-.
simpl in Hskel.
injection Hskel.
intros Hskel2 _.
so (ofc_reg _#3 Hwf Hofc) as Hofk.
invert Hofk.
intros Hofk1 Hofk2.
split.
  simpl.
  assert (ofc (G; cl_cn k1) (cn_app (shift1 c) (var 0)) k2) as Hofapp.
    exploit (@ofc_app (G; cl_cn k1) (shift1 k1) (subst (dot (var 0) (sh 2)) k2) (shift1 c) (var 0)) as H.
      so (weaken1_ofc _ (cl_cn k1) _ _ Hofc) as Hofcw.
      simp_sub in Hofcw.
      assumption.
  
      apply ofc_var; auto using index_0; done.
  
      simp_sub in H.
      assumption.
  #2 apply subk_pi; auto with static.
    apply IH2; auto with static; done.

    so (IH2 _ _ _ Hskel2 Hhkd2 (wf_cn Hwf Hofk1) Hofapp andel) as Hsubk.
    so (subk_reg _#3 (wf_cn Hwf Hofk1) Hsubk) as (Hofks & _).
    assumption.

  simpl.
  simp_sub.
  #3 eapply eqc_extpi.
    so (@ofc_var (G; cl_cn (kd_pi k1 (single (cn_app (shift1 c) (var 0)) k2))) _ _ index_0) as H.
    simp_sub in H.
    eassumption.

    so (weaken1_ofc _ (cl_cn (kd_pi k1 (single (cn_app (shift1 c) (var 0)) k2))) _ _ Hofc) as H.
    simp_sub in H.
    eassumption.

    simp_sub.
    exploit (IH2 (G; cl_cn k1) (cn_app (shift1 c) (var 0)) k2) as (_ & Heqc); auto with static.
      so (weaken1_ofc _ (cl_cn k1) _ _ Hofc) as Hofcw.
      simp_sub in Hofcw.
      so (ofc_app Hofcw (ofc_var index_0)) as H.
      simp_sub in H.
      assumption.

      cut (subof sof
            (G; cl_cn (kd_pi k1 (single (cn_app (shift1 c) (var 0)) k2)); cl_cn (shift1 k1))
            (dot (cn_app (var 1) (var 0)) (dot (var 0) (sh 2)))
            (G; cl_cn k1; cl_cn (single (cn_app (shift1 c) (var 0)) k2))).
        intro Hsubof.
        so (subst_eqc _#6 Hsubof Heqc) as Heqc'.
        simp_sub in Heqc'.
        assumption.

        #2 apply subof_dot.
          #2 apply subof_dot.
            apply subof_sh; auto with sequence; done.

            simp_sub.
            apply sof_cn; [].
            repl (subst (sh 2) k1) with (shift1 (shift1 k1)) by (simp_sub; reflexivity).
            apply ofc_var; auto with sequence; done.

          simp_sub.
          apply sof_cn; [].
          so (@ofc_var (G; cl_cn (kd_pi k1 (single (cn_app (shift1 c) (var 0)) k2)); cl_cn (shift1 k1)) _ _ (index_S index_0)) as Hofvar1.
          so (@ofc_var (G; cl_cn (kd_pi k1 (single (cn_app (shift1 c) (var 0)) k2)); cl_cn (shift1 k1)) _ _ index_0) as Hofvar0.
          simp_sub in Hofvar1.
          simp_sub in Hofvar0.
          so (ofc_app Hofvar1 Hofvar0) as Hofapp.
          simp_sub in Hofapp.
          assumption.
      
(* sigma *)
intros sk1 IH1 sk2 IH2 G c k Hskel Hhkd Hwf Hofc.
invert Hhkd; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros k1 k2 Hhkd1 Hhkd2 <-.
simpl in Hskel.
injection Hskel.
intros Hskel2 Hskel1.
split.
  simpl.
  so (ofc_reg _#3 Hwf Hofc) as Hofk.
  invert Hofk.
  intros Hofk1 Hofk2.
  so (IH1 _ _ _ Hskel1 Hhkd1 Hwf (ofc_pi1 Hofc)) as (Hsubk1 & Heqc1).
  so (subk_reg _#3 Hwf Hsubk1) as (Hofk1s & _).
  apply subk_sigma; eauto with static; [].
  #2 apply (@subk_trans _ _ (subst (dot (cn_pi1 (shift1 c)) sh1) k2)).
    rewrite -> subst_single; auto; [].
    simp_sub.
    #3 apply IH2; eauto with static.
      rewrite -> subst_skel; auto; done.

      apply subst_hkd; auto; done.
      
      so (ofc_pi2 Hofc) as Hofpi2.
      so (weaken1_ofc _ (cl_cn (single (cn_pi1 c) k1)) _ _ Hofpi2) as Hofpi2w.
      simp_sub in Hofpi2w.
      assumption.

    assert (subof sof (G; cl_cn (single (cn_pi1 c) k1); cl_cn (shift1 k1)) (dot (var 0) (sh 2)) (G; cl_cn k1)) as Hsubof.
      #2 apply subof_dot.
        apply subof_sh; auto with sequence; done.
  
        simp_sub.
        apply sof_cn; [].
        replace (subst (sh 2) k1) with (subst (sh 1) (shift1 k1)) by (simp_sub; reflexivity).
        apply ofc_var; auto using index_0; done.
    so (subst_ofk _#4 Hsubof Hofk2) as Hofk2w.
    so (eqc_subsume Heqc1 (weaken1_subk _#4 Hsubk1)) as Heqc1'.
    so (funct_kd _#5 (wf_cn Hwf Hofk1s) Heqc1' Hofk2w) as Heqk.
    simp_sub in Heqk.
    apply subk_refl; [].
    apply eqk_symm; assumption.

  simpl.
  simp_sub.
  #3 apply eqc_extsigma.
    #2 exploit (IH1 G (cn_pi1 c) k1) as (_ & Heqc1); auto.
      eapply ofc_pi1; eauto; done.
    cut (subof sof
         (G; cl_cn (kd_sigma (single (cn_pi1 c) k1) (subst (dot (cn_pi1 (shift1 c)) sh1) (single (cn_pi2 (shift1 c)) k2))))
         (dot (cn_pi1 (var 0)) sh1)
         (G; cl_cn (single (cn_pi1 c) k1))).
      intro Hsubof.
      so (subst_eqc _#6 Hsubof Heqc1) as H.
      simp_sub in H.
      assumption.

      #2 apply subof_dot.
        apply subof_sh; auto with sequence; done.

        simp_sub.
        apply sof_cn; [].
        so (@ofc_var (G; cl_cn (kd_sigma (single (cn_pi1 c) k1) (subst (dot (cn_pi1 (shift1 c)) sh1) (single (cn_pi2 (shift1 c)) k2)))) _ _ index_0) as Hofvar.
        simp_sub in Hofvar.
        so (ofc_pi1 Hofvar) as H.
        simp_sub in H.
        assumption.

    rewrite -> !subst_single; auto; [].
    simp_sub.
    #4 exploit (IH2 G (cn_pi2 c) (subst1 (cn_pi1 c) k2)) as (_ & Heqc2); auto.
      unfold subst1.
      rewrite -> subst_skel; auto; done.

      apply subst_hkd; auto; done.

      eapply ofc_pi2; eauto; done.
    cut (subof sof
         (G; cl_cn (kd_sigma (single (cn_pi1 c) k1) (single (cn_pi2 (shift1 c)) (subst (dot (cn_pi1 (shift1 c)) sh1) k2)))) 
         (dot (cn_pi2 (var 0)) sh1)
         (G; cl_cn (single (cn_pi2 c) (subst1 (cn_pi1 c) k2)))).
      intro Hsubof.
      clear IH1 IH2.
      so (subst_eqc _#6 Hsubof Heqc2) as H.
      unfold shift1 in H.
      rewrite -> !subst_single in H; auto using subst_hkd, subst1_hkd.
      simp_sub in H.
      unfold subst1.
      rewrite -> subst_single; auto using subst_hkd; [].
      simp_sub.
      assumption.

      #2 apply subof_dot.
        apply subof_sh; auto with sequence; done.

        simp_sub.
        apply sof_cn; [].
        so (@ofc_var (G; cl_cn (kd_sigma (single (cn_pi1 c) k1) (single (cn_pi2 (shift1 c)) (subst (dot (cn_pi1 (shift1 c)) sh1) k2)))) _ _ index_0) as Hofvar.
        simp_sub in Hofvar.
        so (ofc_pi2 Hofvar) as H.
        clear IH1 IH2.
        simp_sub in H.
        rewrite -> subst_single in H; auto using subst_hkd; [].
        simp_sub in H.
        unfold shift1.
        rewrite -> subst_single; auto using subst_hkd, subst1_hkd; [].
        simp_sub.
        assumption.

    unfold shift1.
    rewrite -> !subst_single; auto; [].
    simp_sub.
    #5 exploit (IH2 G (cn_pi2 c) (subst1 (cn_pi1 c) k2)) as (Hsubk & _).
      unfold subst1.
      rewrite -> subst_skel; auto; done.

      apply subst1_hkd; auto; done.

      assumption.

      eapply ofc_pi2; eauto; done.
    so (subk_reg _#3 Hwf Hsubk) as (Hofks & _); clear Hsubk.
    so (weaken_ofk _ (nil; cl_cn (kd_sigma (single (cn_pi1 c) k1) (single (cn_pi2 (shift1 c)) (subst (dot (cn_pi1 (shift1 c)) sh1) k2))); cl_cn (single (cn_pi1 (shift1 c)) (shift1 k1))) _ Hofks) as Hofksw.
    autorewrite with canonlist in Hofksw.
    simpl length in Hofksw.
    rewrite -> subst_single in Hofksw; auto using subst1_hkd; [].
    simp_sub in Hofksw.
    assumption.

(* one *)
intros G c k Hskel Hhkd Hwf Hofc.
invert Hhkd; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros <-.
split.
  simpl.
  apply subk_refl; [].
  apply eqk_refl; [].
  apply ofk_one; done.

  simpl.
  simp_sub.
  #2 apply eqc_extone.
    so (@ofc_var (G; cl_cn kd_one) _ _ index_0) as Hofvar.
    simp_sub in Hofvar.
    assumption.

    repl kd_one with (shift1 kd_one) by (simp_sub; reflexivity).
    apply weaken1_ofc; auto; done.
Qed.


Lemma single_eqc :
  forall G c k,
    wf G
    -> ofc G c k
    -> eqc (G; cl_cn (single c k)) (var 0) (shift1 c) (shift1 (single c k)).
Proof.
intros.
eapply single_omnibus; eauto.
Qed.


Lemma single_subk :
  forall G c k,
    wf G
    -> ofc G c k
    -> subk G (single c k) k.
Proof.
intros.
eapply single_omnibus; eauto.
Qed.


Lemma single_eqc_weak :
  forall G c k,
    wf G
    -> ofc G c k
    -> eqc (G; cl_cn (single c k)) (var 0) (shift1 c) (shift1 k).
Proof.
intros G c k Hwf Hofc.
so (single_eqc _#3 Hwf Hofc) as Heqc.
so (single_subk _#3 Hwf Hofc) as Hsubk.
eapply eqc_subsume; eauto; [].
apply weaken1_subk; auto; done.
Qed.


Lemma single_ofk :
  forall G c k,
    wf G
    -> ofc G c k
    -> ofk G (single c k).
Proof.
intros G c k Hwf Hof.
so (single_subk _#3 Hwf Hof) as Hsubk.
so (subk_reg _#3 Hwf Hsubk) as (? & _).
assumption.
Qed.


Lemma single_ofc :
  forall G c k,
    wf G
    -> ofc G c k
    -> ofc G c (single c k).
Proof.
intros G c k Hwf Hofc.
so (single_ofk _#3 Hwf Hofc) as Hofk.
so (single_eqc _#3 Hwf Hofc) as Heqc.
sor (refine (eqc_reg _#4 _ Heqc)) as (_ & Hofc' & _); auto with static; [].
so (inhabitation _ _ Hofk) as (d & Hofd).
so (subst1_ofc _#5 Hofd Hofc') as Hofc''.
simp_sub in Hofc''.
assumption.
Qed.


Lemma single_funct :
  forall G c c' k,
    wf G
    -> eqc G c c' k
    -> eqk G (single c k) (single c' k).
Proof.
intros G c c' k Hwf Heqc.
so (eqc_reg _#4 Hwf Heqc) as (_ & _ & Hofk).
so (single_ofk _#3 (wf_cn Hwf Hofk) (ofc_var index_0)) as Hofks.
so (funct_kd _#5 Hwf Heqc Hofks) as Heqk.
unfold subst1 in Heqk.
rewrite -> !subst_single in Heqk; eauto using subst_hkd, ofk_hkd; [].
simp_sub in Heqk.
assumption.
Qed.


Lemma single_ofc_intro :
  forall G c c' k,
    wf G
    -> eqc G c c' k
    -> ofc G c' (single c k).
Proof.
intros G c c' k Hwf Heqc.
so (single_ofc _#3 Hwf (eqc_reg _#4 Hwf Heqc ander andel)) as Hofc.
eapply ofc_subsume; eauto; [].
apply subk_refl; [].
apply eqk_symm; [].
apply single_funct; auto; done.
Qed.


Lemma single_eqc_intro :
  forall G c c' k,
    wf G
    -> eqc G c c' k
    -> eqc G c c'(single c k).
Proof.
intros G c c' k Hwf Heqc.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & _).
so (single_eqc _#3 Hwf Hofc) as Heqc'.
so (subst1_eqc _#6 (single_ofc_intro _#4 Hwf Heqc) Heqc') as Heqc''.
simp_sub in Heqc''.
apply eqc_symm; assumption.
Qed.


Lemma single_resp_eqk :
  forall G c k k',
    wf G
    -> ofc G c k
    -> eqk G k k'
    -> eqk G (single c k) (single c k').
Proof.
cut (forall G1 G2 s c k k',
       subof sof G1 s G2
       -> wf G1
       -> ofc G1 c (subst s k)
       -> eqk G2 k k'
       -> eqk G1 (single c (subst s k)) (single c (subst s k'))).
  intro Hprop.
  intros G c k k' Hwf Hofc Heqk.
  #2 exploit (Hprop G G id c k k') as H; simp_sub; auto.
    apply subof_id; done.
    
    simp_sub in H.
    assumption.
intros G1 G2 s c k k' Hsubof Hwf Hofc Heqk.
revert G1 s c Hsubof Hwf Hofc.
induct Heqk.

(* refl *)
intros G2 k H G1 s c Hsubof Hwf Hofc.
apply eqk_refl; [].
apply single_ofk; auto; done.

(* symm *)
intros G2 k k' Heqk IH G1 s c Hsubof Hwf Hofc.
apply eqk_symm; [].
apply IH; eauto; [].
eapply ofc_subsume; eauto; [].
apply subk_refl; [].
apply eqk_symm; [].
eapply subst_eqk; eauto; done.

(* trans *)
intros G k1 k2 k3 Heqk1 IH1 Heqk2 IH2 G1 s c Hsubof Hwf Hofc.
so (IH1 _#3 Hsubof Hwf Hofc) as Heqk1s.
#2 sor (refine (IH2 _ _ c Hsubof Hwf _)) as Heqk2s.
  eapply ofc_subsume; eauto; [].
  apply subk_refl; [].
  eapply subst_eqk; eauto; done.
eapply eqk_trans; eauto; done.  

(* sing *)
intros G2 c c' Heqc G1 s b Hsubof Hwf Hofb.
simp_sub in Hofb.
simp_sub.
simpl.
apply eqk_refl; [].
apply ofk_sing; [].
eapply ofc_subsume; eauto; [].
apply subk_sing_t; [].
so (subst_eqc _#6 Hsubof Heqc) as Heqc'.
simp_sub in Heqc'.
exact (eqc_reg _#4 Hwf Heqc' andel).

(* pi *)
intros G2 k1 k1' k2 k2' Heqk1 IH1 Heqk2 IH2 G1 s c Hsubof Hwf Hofc.
simp_sub.
simpl.
#2 apply eqk_pi; auto.
  eapply subst_eqk; eauto; done.

  simp_sub in Hofc.
  #3 apply IH2.
    apply subof_bind_sof_cn; auto; done.

    apply wf_cn; auto; [].
    so (ofc_reg _#3 Hwf Hofc) as Hofk.
    invert Hofk.
    intros Hofk1 _.
    assumption.

    so (weaken1_ofc _ (cl_cn (subst s k1)) _ _ Hofc) as Hofcw.
    simp_sub in Hofcw.
    so (@ofc_var (G1; cl_cn (subst s k1)) 0 _ index_0) as Hofvar.
    simp_sub in Hofvar.
    so (ofc_app Hofcw Hofvar) as H.
    simp_sub in H.
    assumption.

(* sigma *)
intros G2 k1 k1' k2 k2' Heqk1 IH1 Heqk2 IH2 G1 s c Hsubof Hwf Hofc.
simp_sub.
simpl.
#2 apply eqk_sigma.
  apply IH1; eauto with static; done.

  so (eqk_hkd _#3 Heqk2) as (Hhkd & Hhkd').
  rewrite -> !subst_single; eauto using subst_hkd; [].
  simp_sub.
  simp_sub in Hofc.
  #3 apply IH2.
    #2 apply subof_dot.
      so (subof_weaken_sof _ (nil; cl_cn (single (cn_pi1 c) (subst s k1))) _ _ Hsubof) as Hsubof'.
      simpl in Hsubof'.
      assumption.

      simp_sub.
      apply sof_cn; [].
      so (weaken1_ofc _ (cl_cn (single (cn_pi1 c) (subst s k1))) _ _ (ofc_pi1 Hofc)) as H.
      simp_sub in H.
      assumption.

    apply wf_cn; auto; [].
    apply single_ofk; eauto with static; done.

    so (weaken1_ofc _ (cl_cn (single (cn_pi1 c) (subst s k1))) _ _ (ofc_pi2 Hofc)) as H.
    simp_sub in H.
    assumption.
Qed.


Lemma single_resp :
  forall G c c' k k',
    wf G
    -> eqc G c c' k
    -> eqk G k k'
    -> eqk G (single c k) (single c' k').
Proof.
intros G c c' k k' Hwf Heqc Heqk.
#2 apply (@eqk_trans _ _ (single c k')).
  apply single_resp_eqk; auto; [].
  exact (eqc_reg _#4 Hwf Heqc andel).

  apply single_funct; eauto with static; done.
Qed.


Lemma single_mono :
  forall G c k k',
    wf G
    -> ofc G c k
    -> subk G k k'
    -> subk G (single c k) (single c k').
Proof.
cut (forall G1 G2 s c k k',
       subof sof G1 s G2
       -> wf G1
       -> ofc G1 c (subst s k)
       -> subk G2 k k'
       -> subk G1 (single c (subst s k)) (single c (subst s k'))).
  intro Hprop.
  intros G c k k' Hwf Hofc Hsubk.
  #2 exploit (Hprop G G id c k k') as H; simp_sub; auto.
    apply subof_id; done.
    
    simp_sub in H.
    assumption.
intros G1 G2 s c k k' Hsubof Hwf Hofc Hsubk.
revert G1 s c Hsubof Hwf Hofc.
induct Hsubk.

(* refl *)
intros G2 k k' Heqk G1 s c Hsubof Hwf Hofc.
apply subk_refl; [].
apply single_resp_eqk; auto; [].
eapply subst_eqk; eauto; done.

(* trans *)
intros G k1 k2 k3 Hsubk1 IH1 Hsubk2 IH2 G1 s c Hsubof Hwf Hofc.
so (IH1 _#3 Hsubof Hwf Hofc) as Hsubk1s.
#2 sor (refine (IH2 _ _ c Hsubof Hwf _)) as Hsubk2s.
  eapply ofc_subsume; eauto; [].
  eapply subst_subk; eauto; done.
eapply subk_trans; eauto; done.  

(* sing_t *)
intros G c Hofc G1 s b Hsubof Hwf Hofb.
simp_sub.
simpl.
apply subk_refl; [].
apply eqk_refl; [].
apply ofk_sing; [].
simp_sub in Hofb.
eapply ofc_subsume; eauto; [].
apply subk_sing_t; [].
so (subst_ofc _#5 Hsubof Hofc) as Hofc'.
simp_sub in Hofc'; assumption.

(* pi *)
intros G2 k1 k1' k2 k2' Hsubk1 IH1 Hsubk2 IH2 Hofk2 G1 s c Hsubof Hwf Hofc.
simp_sub.
simpl.
#3 apply subk_pi; auto.
  eapply subst_subk; eauto; done.

  simp_sub in Hofc.
  #3 apply IH2.
    apply subof_bind_sof_cn; auto; done.

    apply wf_cn; auto; [].
    so (subk_reg _#3 Hwf (subst_subk _#5 Hsubof Hsubk1)) as (H & _).
    assumption.

    so (weaken1_ofc _ (cl_cn (subst s k1')) _ _ Hofc) as Hofcw.
    simp_sub in Hofcw.
    so (@ofc_var (G1; cl_cn (subst s k1')) 0 _ index_0) as Hofvar.
    simp_sub in Hofvar.
    sor (refine (ofc_app Hofcw (ofc_subsume Hofvar _))) as H.
      rewrite -> !subst_compose; [].
      apply weaken1_subk; [].
      eapply subst_subk; eauto; done.
    simp_sub in H.
    assumption.

  #2 apply single_ofk.
    apply wf_cn; auto; [].
    exact (subk_reg _#3 Hwf (subst_subk _#5 Hsubof Hsubk1) ander).

    so (weaken1_ofc _ (cl_cn (subst s k1)) _ _ Hofc) as Hofcw.
    simp_sub in Hofcw.
    so (@ofc_var (G1; cl_cn (subst s k1)) 0 _ index_0) as Hofvar.
    simp_sub in Hofvar.
    so (ofc_app Hofcw Hofvar) as H.
    simp_sub in H.
    assumption.

(* sigma *)
intros G2 k1 k1' k2 k2' Hsubk1 IH1 Hsubk2 IH2 Hofk2' G1 s c Hsubof Hwf Hofc.
simp_sub.
simpl.
so (subk_hkd _#3 Hsubk2) as (Hhkd & Hhkd').
#3 apply subk_sigma.
  apply IH1; eauto with static; done.

  rewrite -> !subst_single; eauto using subst_hkd; [].
  simp_sub.
  simp_sub in Hofc.
  #3 apply IH2.
    #2 apply subof_dot.
      so (subof_weaken_sof _ (nil; cl_cn (single (cn_pi1 c) (subst s k1))) _ _ Hsubof) as Hsubof'.
      simpl in Hsubof'.
      assumption.

      simp_sub.
      apply sof_cn; [].
      so (weaken1_ofc _ (cl_cn (single (cn_pi1 c) (subst s k1))) _ _ (ofc_pi1 Hofc)) as H.
      simp_sub in H.
      assumption.

    apply wf_cn; auto; [].
    apply single_ofk; eauto with static; done.

    so (weaken1_ofc _ (cl_cn (single (cn_pi1 c) (subst s k1))) _ _ (ofc_pi2 Hofc)) as H.
    simp_sub in H.
    assumption.

  rewrite -> subst_single; eauto using subst_hkd; [].
  simp_sub.
  simp_sub in Hofc.
  #2 apply single_ofk.
    apply wf_cn; auto; [].
    apply single_ofk; auto; [].
    refine (@ofc_subsume _ _ (subst s k1') _ (ofc_pi1 Hofc) _); [].
    eapply subst_subk; eauto; done.

    so (ofc_pi2 Hofc) as Hofpi2.
    simp_sub in Hofpi2.
    #2 sor (refine (@ofc_subsume _ _ (subst (dot (cn_pi1 c) s) k2') _ Hofpi2 _)) as Hofpi2'.
      eapply subst_subk; eauto; [].
      apply subof_dot; auto; [].
      simp_sub.
      apply sof_cn; [].
      eapply ofc_pi1; eauto; done.
    so (weaken1_ofc _ (cl_cn (single (cn_pi1 c) (subst s k1'))) _ _ Hofpi2') as H.
    simp_sub in H.
    assumption.
Qed.


Fixpoint singles (c : (* cn *) term) (T : (* sg *) term) {struct T} :=
  (match T with
   | sg_stat k =>
       sg_stat (single c k)

   | sg_piapp T1 T2 =>
       sg_piapp T1 (singles (cn_app (shift1 c) (var 0)) T2)

   | sg_sigma T1 T2 =>
       sg_sigma
       (singles (cn_pi1 c) T1)
       (subst (dot (cn_pi1 (shift1 c)) sh1) (singles (cn_pi2 (shift1 c)) T2))

   | sg_dyn _ => T
       
   | sg_pigen _ _ => T

   | _ => sg_one
   end).


Inductive skeletons : Set :=
| sks_one   : skeletons
| sks_stat  : skeletons
| sks_dyn   : skeletons
| sks_pigen : skeletons -> skeletons -> skeletons
| sks_piapp : skeletons -> skeletons -> skeletons
| sks_sigma : skeletons -> skeletons -> skeletons
.


Fixpoint skels (T : (* sg *) term) {struct T} : skeletons :=
  (match T with
   | sg_stat _ => sks_stat
   | sg_dyn _ => sks_dyn
   | sg_pigen T1 T2 => sks_pigen (skels T1) (skels T2)
   | sg_piapp T1 T2 => sks_piapp (skels T1) (skels T2)
   | sg_sigma T1 T2 => sks_sigma (skels T1) (skels T2)
   | _ => sks_one
   end).


Lemma subst_skels :
  forall s T, hsg T -> skels (subst s T) = skels T.
Proof.
intros s T Hsg.
revert s.
induct Hsg; intros; simp_sub; simpl; try f_equal; auto.
Qed.


(* Doesn't need any premises. *)
Lemma shift_under_singles :
  forall i j c T,
    subst (under i (sh j)) (singles c T) = singles (subst (under i (sh j)) c) (subst (under i (sh j)) T).
Proof.
intros i j c T.
revert i c.
#5 induct T; try (intros; simp_sub; simpl; auto; done).

(* var *)
intros n i c.
simp_sub.
so (Compare_dec.dec_lt n i) as [Hlt | Hgeq].
  rewrite -> project_under_lt; auto; done.

  rewrite -> project_under_ge; [| omega].
  simp_sub.
  simpl; reflexivity.

(* stat *)
intros k _ i c.
simp_sub.
simpl.
simp_sub.
rewrite -> shift_under_single; [].
reflexivity.

(* pigen *)
intros T1 _ T2 IH i c.
simp_sub.
simpl.
simp_sub.
reflexivity.

(* piapp *)
intros T1 _ T2 IH i c.
simp_sub.
simpl.
simp_sub.
f_equal; [].
rewrite <- under_succ; [].
rewrite -> IH; [].
simp_sub.
reflexivity.

(* sigma *)
intros T1 IH1 T2 IH2 i c.
simp_sub.
simpl.
simp_sub.
#2 f_equal.
  rewrite -> IH1; [].
  simp_sub.
  reflexivity.

  rewrite -> split_dot.
  rewrite -> subst_compose.
  repl (dot (cn_pi1 (subst (compose (under i (sh j)) sh1) c)) sh1) with (compose (dot (var 0) (sh 2)) (dot (cn_pi1 (subst (compose (under i (sh j)) sh1) c)) id)) by (simp_sub; reflexivity).
  rewrite -> (subst_compose _ _ (dot (var 0) (sh 2))).
  f_equal; [].
  simp_sub.
  repl (dot (var 0) (compose (under i (sh j)) (sh 2))) with (compose (dot (var 0) (compose (under i (sh j)) sh1)) (dot (var 0) (sh 2))) by (simp_sub; reflexivity).
  rewrite <- under_succ.
  rewrite -> subst_compose.
  rewrite -> IH2; [].
  f_equal; [].
  simp_sub.
  reflexivity.
Qed.


Lemma subst_singles :
  forall s c T,
    hsg T
    -> subst s (singles c T) = singles (subst s c) (subst s T).
Proof.
intros s c T Hhsg.
revert c s.
#3 induct Hhsg;
try (intros; simpl; simp_sub; auto; done).

(* stat *)
intros; simpl; simp_sub.
rewrite -> subst_single; auto; done.

(* piapp *)
intros T1 T2 _ _ _ IH c s.
simpl; simp_sub.
f_equal; [].
rewrite -> IH; [].
simp_sub; reflexivity.

(* sigma *)
intros T1 T2 _ IH1 _ IH2 c s.
simpl; simp_sub.
#2 f_equal.
  rewrite -> IH1; [].
  simp_sub; reflexivity.

  replsub (cn_pi2 (subst (compose s sh1) c)) with (subst (dot (var 0) (compose s sh1)) (shift1 (cn_pi2 c))).
  rewrite <- IH2; [].
  simp_sub.
  reflexivity.
Qed.


Lemma singles_fst :
  forall c T k,
    fsts T k
    -> fsts (singles c T) (single c k).
Proof.
intros c T k Hfst.
revert c.
induct Hfst; try (intros; simpl; eauto with static; done); [].
(* sigma *)
intros T1 T2 k1 k2 Hfst1 IH1 Hfst2 IH2 c.
simpl.
apply fsts_sigma; auto; [].
apply subst_fsts; auto; done.
Qed.


Lemma singles_subs :
  forall G T k c,
    wf G
    -> ofs G T
    -> fsts T k
    -> ofc G c k
    -> subs G (singles c T) T.
Proof.
intros G T k c Hwf Hofs Hfsts Hofc.
remember (skels T) as sk eqn:Hskel.
revert G T k c Hwf Hofs Hfsts Hofc Hskel.
induct sk;
try (intros; subst; simpl;
     invert Hofs; intros; subst; simpl in Hskel; try (discriminate Hskel);
     eauto with static; done).

(* stat *)
intros G T k c Hwf Hofs Hfsts Hofc Hskel.
invert Hofs; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros k' Hofk <-.
invert Hfsts; [].
intros; subst k'.
simpl.
apply subs_stat; [].
apply single_subk; auto; done.

(* piapp *)
intros sk1 _ sk2 IH2 G T k c Hwf Hofs Hfsts Hofc Hskel.
invert Hofs; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros T1 T2 k1 HofT1 Hfst1 HofT2 <-.
injection Hskel.
intros Hskel2 Hskel1.
invert Hfsts.
intros k1' k2 Hfst1' Hfst2 <-.
so (fsts_fun _#3 Hfst1 Hfst1'); subst k1'; clear Hfst1'.
simpl.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
assert (subs (G; cl_cn k1) (singles (cn_app (shift1 c) (var 0)) T2) T2) as Hsubs.
  eapply IH2; eauto with static; [].
  so (subst_ofc _#5 (weaken1 _ _ (cl_cn k1)) Hofc) as H1; simp_sub in H1.
  so (ofc_app H1 (ofc_var index_0)) as H2; simp_sub in H2.
  assumption.
eapply subs_piapp; eauto with static; [].
so (subs_reg _#3 (wf_cn Hwf Hofk1) Hsubs) as (? & _).
assumption.
             
(* sigma *)
intros sk1 IH1 sk2 IH2 G T k c Hwf Hofs Hfsts Hofc Hskel.
invert Hofs; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros T1 T2 k1 HofT1 Hfst1 HofT2 <-.
injection Hskel.
intros Hskel2 Hskel1.
invert Hfsts.
intros k1' k2 Hfst1' Hfst2 <-.
so (fsts_fun _#3 Hfst1 Hfst1'); subst k1'; clear Hfst1'.
simpl.
rewrite -> subst_singles; eauto using ofs_hsg; [].
simp_sub.
assert (subs G (singles (cn_pi1 c) T1) T1) as HsubT1 by (eauto with static).
so (subs_reg _#3 Hwf HsubT1) as (HofT1s & _).
so (singles_fst (cn_pi1 c) _ _ Hfst1) as Hfst1s.
eapply subs_sigma; eauto; [].
apply (@subs_trans _ _ (subst (dot (cn_pi1 (shift1 c)) sh1) T2)).
  cut (subs G (singles (cn_pi2 c) (subst1 (cn_pi1 c) T2)) (subst1 (cn_pi1 c) T2)).
    intro H1.
    so (subst_subs _#5 (weaken1 _ _ (cl_cn (single (cn_pi1 c) k1))) H1) as H2.
    rewrite -> subst_singles in H2; eauto using subst_hsg, ofs_hsg; [].
    simp_sub in H2.
    assumption.
  #3 eapply IH2; eauto with static.
    eapply subst1_ofs; eauto with static; done.
  
    apply subst_fsts; auto; done.
  
    unfold subst1; rewrite -> subst_skels; eauto using ofs_hsg; done.

  apply subs_refl; [].
  assert (eqc (G; cl_cn (single (cn_pi1 c) k1)) (var 0) (shift1 (cn_pi1 c)) (shift1 k1)) as Heqc.
    apply single_eqc_weak; eauto with static; done.
  #3 exploit (subof_under_sof (G; cl_cn (single (cn_pi1 c) k1)) G (nil; cl_cn k1) sh1) as Hs.
    apply Forall_cons; auto using iscl_cn; done.

    apply subof_sh; auto with sequence; done.
  simpl in Hs.
  so (subst_ofs _#4 Hs HofT2) as HofT2'.
  simp_sub in HofT2'.
  so (ofs_fst _#3 HofT1s Hfst1s) as Hofk1s.
  so (funct_sg _#5 (wf_cn Hwf Hofk1s) Heqc HofT2') as Heqs.
  simp_sub in Heqs.
  apply eqs_symm; [].
  assumption.
Qed.


Lemma singles_ofs :
  forall G T k c,
    wf G
    -> ofs G T
    -> fsts T k
    -> ofc G c k
    -> ofs G (singles c T).
Proof.
intros G T k c Hwf Hofs Hfsts Hofc.
so (singles_subs _#4 Hwf Hofs Hfsts Hofc) as Hsubs.
so (subs_reg _#3 Hwf Hsubs andel).
assumption.
Qed.


Lemma singles_ofm :
  forall G M T c,
    wf G
    -> ofm G pure M T
    -> fstm G M c
    -> ofm G pure M (singles c T).
Proof.
intros G M T c Hwf Hof Hfstm.
remember (skels T) as sk eqn:Hskel.
so (ofs_hsg _ _ (ofm_reg _#4 Hwf Hof)) as Hhsg.
revert G M T c Hwf Hhsg Hof Hfstm Hskel.
induct sk.

(* one *)
intros G M T c Hwf Hhsg Hof Hfstm Hskel.
invert Hhsg; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros <-.
simpl.
assumption.

(* stat *)
intros G M T c Hwf Hhsg Hof Hfstm Hskel.
invert Hhsg; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros k _ <-.
simpl.
eapply ofm_extstat; eauto; [].
apply single_ofc; auto; [].
eapply ofm_fst; eauto with static; done.

(* dyn *)
intros G M T c Hwf Hhsg Hof Hfstm Hskel.
invert Hhsg; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros t <-.
simpl.
assumption.

(* pigen *)
intros sk1 _ sk2 IH2 G M T c Hwf Hhsg Hof Hfstm Hskel.
invert Hhsg; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros T1 T2 Hhsg1 Hhsg2 <-.
simpl.
assumption.

(* piapp *)
intros sk1 _ sk2 IH2 G M T c Hwf Hhsg Hof Hfstm Hskel.
invert Hhsg; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros T1 T2 Hhsg1 Hhsg2 <-.
injection Hskel.
intros Hskel2 Hskel1.
simpl.
so (ofm_reg _#4 Hwf Hof) as Hofpi.
invert Hofpi; [].
intros k1 HofT1 Hfst1 _.
eapply ofm_extpi; eauto; [].
unfold shift1; rewrite -> subst_singles; auto; [].
simp_sub.
#4 apply IH2; auto with static.
  unfold shift1; apply subst_hsg; auto; done.

  repl (shift1 T2) with (subst1 (var 1) (subst (dot (var 0) (sh 3)) T2)).
    simp_sub.
    replsub (dot (var 1) (sh 2)) with (compose (dot (var 0) sh1) sh1).
    rewrite -> subst_compose.
    unfold shift1; f_equal; [].
    simp_sub.
    reflexivity.
  #2 eapply ofm_appap; eauto with static sequence.
    simp_sub.
    so (subst_ofm _#6 (weaken2 _ _ (cl_cn k1) (cl_md (shift1 T1) (var 0))) Hof) as H; simp_sub in H.
    assumption.

    replsub (var 1) with (shift1 (var 0)).
    eapply fstm_var; eauto using index_0; done.

  #2 apply fstm_appap.
    eapply subst_fstm; eauto using weaken2; done.

    replsub (var 1) with (shift1 (var 0)).
    eapply fstm_var; eauto using index_0; done.

  unfold shift1; rewrite -> subst_skels; eauto; done.

(* sigma *)
intros sk1 IH1 sk2 IH2 G M T c Hwf Hhsg Hof Hfstm Hskel.
invert Hhsg; try (intros; subst; simpl in Hskel; discriminate Hskel); [].
intros T1 T2 Hhsg1 Hhsg2 <-.
injection Hskel.
intros Hskel2 Hskel1.
simpl.
replsub (subst (dot (cn_pi1 (shift1 c)) sh1) (singles (cn_pi2 (shift1 c)) T2)) with (shift1 (subst1 (cn_pi1 c) (singles (cn_pi2 (shift1 c)) T2))).
#2 apply ofm_extsigma.
  apply IH1; eauto with static; done.

  unfold subst1; rewrite -> subst_singles; auto; [].
  simp_sub.
  apply IH2; eauto using subst_hsg with static; [].
  unfold subst1; rewrite -> subst_skels; auto; done.
Qed.


Lemma singles_funct :
  forall G c c' T k,
    wf G
    -> ofs G T
    -> fsts T k
    -> eqc G c c' k
    -> eqs G (singles c T) (singles c' T).
Proof.
intros G c c' T k Hwf HofT Hfst Heqc.
so (ofs_fst _#3 HofT Hfst) as Hofk.
so (weaken1_ofs _ (cl_cn k) _ HofT) as HofTw.
so (singles_ofs _#4 (wf_cn Hwf Hofk) HofTw (subst_fsts _ _ sh1 Hfst) (ofc_var index_0)) as Hofs.
so (funct_sg _#5 Hwf Heqc Hofs) as Heqs.
unfold subst1 in Heqs; rewrite -> !subst_singles in Heqs; eauto using ofs_hsg, subst_hsg; [].
simp_sub in Heqs.
assumption.
Qed.


Lemma singles_resp_eqs :
  forall G c T T' k,
    wf G
    -> eqs G T T'
    -> fsts T k
    -> ofc G c k
    -> eqs G (singles c T) (singles c T').
Proof.
cut (forall G1 G2 s c T T' k,
       subof sof G1 s G2
       -> wf G1
       -> fsts T k
       -> ofc G1 c (subst s k)
       -> eqs G2 T T'
       -> eqs G1 (singles c (subst s T)) (singles c (subst s T'))).
  intros H G c T T' k Hwf Heqs Hfst Hofc.
  replsub k with (subst id k) in Hofc.
  so (H _#7 (subof_id _ _) Hwf Hfst Hofc Heqs) as H'.
  simp_sub in H'.
  assumption.
intros G1 G2 s c T T' k Hs Hwf Hfst Hofc Heqs.
revert G1 s c k Hs Hwf Hfst Hofc.
induct Heqs.

(* refl *)
intros G T HofT G1 s c k Hs Hwf Hfst Hofc.
apply eqs_refl; eauto using singles_ofs, subst_ofs, subst_fsts; done.

(* symm *)
intros G T T' Heqs IH G1 s c k' Hs Hwf Hfst' Hofc.
so (can_fsts_hsg _ (eqs_hsg _#3 Heqs andel)) as (k & Hfst).
apply eqs_symm; [].
eapply IH; eauto; [].
eapply ofc_subsume; eauto; [].
apply subk_refl; apply eqk_symm; [].
so (subst_eqs _#5 Hs Heqs) as Heqs'.
eapply eqs_fst; eauto using subst_fsts; done.

(* trans *)
intros G T1 T2 T3 Heqs12 IH1 Heqs23 IH2 G1 s c k1 Hs Hwf Hfst1 Hofc.
#2 apply (@eqs_trans _ _ (singles c (subst s T2))).
  eapply IH1; eauto; done.

  so (can_fsts_hsg _ (eqs_hsg _#3 Heqs12 ander)) as (k2 & Hfst2).
  eapply IH2; eauto; [].
  eapply ofc_subsume; eauto; [].
  apply subk_refl; [].
  so (subst_eqs _#5 Hs Heqs12).
  eapply eqs_fst; eauto using subst_fsts; done.

(* stat *)
intros G k1 k2 Heqk G1 s c k1' Hs Hwf Hfst Hofc.
invert Hfst; [].
intro; subst k1'.
simp_sub.
simpl.
apply eqs_stat; eauto using single_resp_eqk, subst_eqk; done.

(* dyn *)
intros G t t' Heqt G1 s c k Hs Hwf Hfst Hofc.
simp_sub.
simpl.
apply eqs_dyn; [].
eapply subst_eqc_type; eauto; done.

(* pigen *)
intros G T1 T1' k1 T2 T2' Heqs1 IH1 Hfst1 Heqs2 IH2 G1 s c k Hs Hwf Hfstpi Hofc.
simp_sub.
simpl.
eapply eqs_pigen; eauto using subst_eqs, subst_fsts; [].
eapply subst_eqs; eauto; [].
apply subof_bind_sof_cn; auto; done.

(* piapp *)
intros G T1 T1' k1 T2 T2' Heqs1 IH1 Hfst1 Heqs2 IH2 G1 s c k Hs Hwf Hfstpi Hofc.
simp_sub.
simpl.
eapply eqs_piapp; eauto using subst_eqs, subst_fsts; [].
invert Hfstpi; [].
intros k1' k2 Hfst1' Hfst2 <-.
so (fsts_fun _#3 Hfst1 Hfst1'); subst k1'.
so (eqs_reg _#3 Hwf (subst_eqs _#5 Hs Heqs1)) as (HofT1x & _).
so (ofs_fst _#3 HofT1x (subst_fsts _ _ _ Hfst1)) as Hofk1x.
simp_sub in Hofc.
eapply IH2; eauto using subof_bind_sof_cn, wf_cn; [].
so (subst_ofc _#5 (weaken1 _ _ (cl_cn (subst s k1))) Hofc) as Hofcw; simp_sub in Hofcw.
rewrite -> (subst_compose k1) in Hofcw.
so (ofc_app Hofcw (ofc_var index_0)) as H.
simp_sub in H.
assumption.

(* sigma *)
intros G T1 T1' k1 T2 T2' Heqs1 IH1 Hfst1 Heqs2 IH2 G1 s c k Hs Hwf Hfstsig Hofc.
simp_sub.
simpl.
invert Hfstsig; [].
intros k1' k2 Hfst1' Hfst2 <-.
so (fsts_fun _#3 Hfst1 Hfst1'); subst k1'.
so (eqs_hsg _#3 Heqs2) as (Hhsg2 & Hhsg2').
rewrite -> !subst_singles; eauto using subst_hsg; [].
simp_sub.
so (singles_fst (cn_pi1 c) _ _ (subst_fsts _ _ s Hfst1)) as Hfstss.
simp_sub in Hofc.
#2 eapply eqs_sigma; eauto.
  eapply IH1; eauto with static; done.

  #3 eapply IH2; eauto.
    #2 apply subof_dot.
      so (subof_weaken_sof _ (nil; cl_cn (single (cn_pi1 c) (subst s k1))) _ _ Hs) as Hss.
      assumption.

      simp_sub.
      apply sof_cn; [].
      sor (refine (subst_ofc _#5 (weaken1 _ _ _) (ofc_pi1 Hofc))) as H; [].
      simp_sub in H.
      eassumption.

    apply wf_cn; auto; [].
    apply single_ofk; eauto using ofc_pi1; done.

    so (ofc_pi2 Hofc) as H1; simp_sub in H1.
    sor (refine (subst_ofc _#5 (weaken1 _ _ _) H1)) as H2; simp_sub in H2.
    eassumption.
Qed.


Lemma singles_mono :
  forall G c T T' k,
    wf G
    -> subs G T T'
    -> fsts T k
    -> ofc G c k
    -> subs G (singles c T) (singles c T').
Proof.
cut (forall G1 G2 s c T T' k,
       subof sof G1 s G2
       -> wf G1
       -> fsts T k
       -> ofc G1 c (subst s k)
       -> subs G2 T T'
       -> subs G1 (singles c (subst s T)) (singles c (subst s T'))).
  intros H G c T T' k Hwf Heqs Hfst Hofc.
  replsub k with (subst id k) in Hofc.
  so (H _#7 (subof_id _ _) Hwf Hfst Hofc Heqs) as H'.
  simp_sub in H'.
  assumption.
intros G1 G2 s c T T' k Hs Hwf Hfst Hofc Hsubs.
revert G1 s c k Hs Hwf Hfst Hofc.
induct Hsubs.

(* refl *)
intros G T T' HofT G1 s c k Hs Hwf Hfst Hofc.
apply subs_refl; [].
eapply singles_resp_eqs; eauto using singles_ofs, subst_eqs, subst_fsts; done.

(* trans *)
intros G T1 T2 T3 Hsubs12 IH1 Hsubs23 IH2 G1 s c k1 Hs Hwf Hfst1 Hofc.
#2 apply (@subs_trans _ _ (singles c (subst s T2))).
  eapply IH1; eauto; done.

  so (can_fsts_hsg _ (subs_hsg _#3 Hsubs12 ander)) as (k2 & Hfst2).
  eapply IH2; eauto; [].
  eapply ofc_subsume; eauto; [].
  so (subst_subs _#5 Hs Hsubs12).
  eapply subs_fst; eauto using subst_fsts; done.

(* stat *)
intros G k1 k2 Heqk G1 s c k1' Hs Hwf Hfst Hofc.
invert Hfst; [].
intro; subst k1'.
simp_sub.
simpl.
apply subs_stat; eauto using single_mono, subst_subk; done.

(* pigen *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 IH1 Hfst1' Hsubs2 IH2 Hfst1 HofsT2 G1 s c k Hs Hwf Hfstpi Hofc.
simp_sub.
simpl.
#2 eapply subs_pigen; eauto using subst_subs, subst_fsts.
  eapply subst_subs; eauto; [].
  apply subof_bind_sof_cn; auto; done.

  eapply subst_ofs; eauto; [].
  apply subof_bind_sof_cn; auto; done.

(* piapp *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 IH1 Hfst1' Hsubs2 IH2 Hfst1 HofsT2 G1 s c k Hs Hwf Hfstpi Hofc.
simp_sub.
simpl.
invert Hfstpi; [].
intros l k2 Hfstl Hfst2 <-.
so (fsts_fun _#3 Hfst1 Hfstl); subst l.
#2 eapply subs_piapp; eauto using subst_subs, subst_fsts.
  so (subs_reg _#3 Hwf (subst_subs _#5 Hs Hsubs1)) as (HofT1x' & _).
  so (ofs_fst _#3 HofT1x' (subst_fsts _ _ _ Hfst1')) as Hofk1x'.
  simp_sub in Hofc.
  eapply IH2; eauto using subof_bind_sof_cn, wf_cn; [].
  so (subst_ofc _#5 (weaken1 _ _ (cl_cn (subst s k1'))) Hofc) as Hofcw; simp_sub in Hofcw.
  rewrite -> (subst_compose k1) in Hofcw.
  so (subs_fst _#5 Hwf (subst_subs _#5 Hs Hsubs1) (subst_fsts _#3 Hfst1') (subst_fsts _#3 Hfst1)) as H1.
  sor (refine (subst_subk _#5 (weaken1 _#3) H1)) as H2.
  so (ofc_subsume (ofc_var index_0) H2) as H3.
  so (ofc_app Hofcw H3) as H4; simp_sub in H4.
  assumption.

  so (ofs_fst _#3 (subs_reg _#3 Hwf (subst_subs _#5 Hs Hsubs1) ander) (subst_fsts _#3 Hfst1)) as Hofk1.
  so (ofs_fst _#3 (subs_reg _#3 Hwf (subst_subs _#5 Hs Hsubs1) andel) (subst_fsts _#3 Hfst1')) as Hofk1'.
  #2 eapply singles_ofs; eauto using subst_fsts, wf_cn.
    eapply subst_ofs; eauto; [].
    apply subof_bind_sof_cn; auto; done.

    so (subst_ofc _#5 (weaken1 _ _ (cl_cn (subst s k1))) Hofc) as Hofcw; simp_sub in Hofcw.
    rewrite -> (subst_compose k1) in Hofcw.
    so (ofc_app Hofcw (ofc_var index_0)) as H; simp_sub in H.
    assumption.

(* sigma *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 IH1 Hfst1 Hsubs2 IH2 Hfst1' HofsT2' G1 s c k Hs Hwf Hfstsig Hofc.
simp_sub.
simpl.
invert Hfstsig; [].
intros l k2 Hfstl Hfst2 <-.
so (fsts_fun _#3 Hfst1 Hfstl); subst l; clear Hfstl.
so (subs_hsg _#3 Hsubs2) as (Hhsg2 & Hhsg2').
rewrite -> !subst_singles; eauto using subst_hsg; [].
simp_sub.
so (singles_fst (cn_pi1 c) _ _ (subst_fsts _ _ s Hfst1)) as Hfstss.
so (singles_fst (cn_pi1 c) _ _ (subst_fsts _ _ s Hfst1')) as Hfstss'.
simp_sub in Hofc.
#3 eapply subs_sigma; eauto.
  eapply IH1; eauto with static; done.

  #3 eapply IH2; eauto.
    #2 apply subof_dot.
      so (subof_weaken_sof _ (nil; cl_cn (single (cn_pi1 c) (subst s k1))) _ _ Hs) as Hss.
      assumption.

      simp_sub.
      apply sof_cn; [].
      sor (refine (subst_ofc _#5 (weaken1 _ _ _) (ofc_pi1 Hofc))) as H; [].
      simp_sub in H.
      eassumption.

    apply wf_cn; auto; [].
    apply single_ofk; eauto using ofc_pi1; done.

    so (ofc_pi2 Hofc) as H1; simp_sub in H1.
    sor (refine (subst_ofc _#5 (weaken1 _ _ _) H1)) as H2; simp_sub in H2.
    eassumption.

  so (can_fsts_hsg _ Hhsg2') as (k2' & Hfst2').
  #3 eapply singles_ofs; eauto using subst_fsts.
    apply wf_cn; auto; [].
    apply single_ofk; auto; [].
    eapply ofc_subsume; eauto with static; [].
    eapply subs_fst; eauto using subst_fsts, subst_subs; done.

    cut (ofs G1 (subst (dot (cn_pi1 c) s) T2')).
      intros H.
      sor (refine (subst_ofs _#4 (weaken1 _#3) H)) as H'; simp_sub in H'.
      eassumption.
    eapply subst_ofs; eauto; [].
    apply subof_dot; auto; [].
    simp_sub.
    apply sof_cn; [].
    eapply ofc_subsume; eauto with static; [].
    eapply subs_fst; eauto using subst_fsts, subst_subs; done.

    cut (ofc G1 (cn_pi2 c) (subst (dot (cn_pi1 c) s) k2')).
      intros H.
      sor (refine (subst_ofc _#5 (weaken1 _#3) H)) as H'; simp_sub in H'.
      eassumption.
    eapply ofc_subsume; eauto with static; [].
    simp_sub.
    eapply subs_fst; eauto using subst_fsts; [].
    eapply subst_subs; eauto; [].
    apply subof_dot; auto; [].
    simp_sub.
    apply sof_cn; eauto with static; done.
Qed.


Lemma ofm_pair_dep :
  forall G p T1 T2 M1 M2 k1 c1,
    wf G
    -> ofm G pure M1 T1
    -> fsts T1 k1
    -> fstm G M1 c1
    -> ofm G p M2 (subst1 c1 T2)
    -> ofs (G; cl_cn k1) T2
    -> ofm G p (md_pair M1 M2) (sg_sigma T1 T2).
Proof.
intros G p T1 T2 M1 M2 k1 c1 Hwf HofM1 Hfst1 HfstM1 HofM2 HofT2.
so (ofm_reg _#4 Hwf HofM1) as HofT1.
so (ofm_fst _#6 Hwf HofM1 HfstM1 Hfst1) as Hofc1.
#2 apply (@ofm_subsume _ _ (sg_sigma (singles c1 T1) (shift1 (subst1 c1 T2)))).
  apply ofm_pair; auto; [].
  apply ofm_forget_pure; [].
  apply singles_ofm; auto; done.

  eapply subs_sigma; eauto using singles_fst, singles_subs; [].
  apply subs_refl; [].
  simp_sub.
  cut (eqs (G; cl_cn (single c1 k1)) (subst1 (shift1 c1) (subst (under 1 sh1) T2)) (subst1 (var 0) (subst (under 1 sh1) T2))).
    intro H.
    simp_sub in H.
    exact H.
  apply (funct_sg _ (shift1 k1)).
    apply wf_cn; auto using single_ofk; done.

    apply eqc_symm; [].
    apply single_eqc_weak; auto; done.

    eapply subst_ofs; eauto; [].
    simp_sub.
    apply subof_bind_sof_cn; eauto; [].
    apply subof_sh; auto with sequence; done.
Qed.
