
Require Import Tactics.
Require Import Sequence.
Require Import Subst.
Require Import Static.
Require Import Formation.
Require Import Declarative.
Require Import Singleton.
Require Import SubjectReduction.
Require Import Candidate.


Definition flatten_ctx (G : context) : list (option skind) :=
  map (fun x => (match x with cl_cn k => Some (simp k) | _ => None end)) G.


Fixpoint skof (L : list (option skind)) (c : (* cn *) term) {struct c} : skind :=
  (match c with
   | var i =>
       (match nth i L None with
        | Some sk => sk
        | None => sk_one
        end)

   | cn_lam k1 c' =>
       sk_pi (simp k1) (skof (L; Some (simp k1)) c')

   | cn_app c1 c2 =>
       (match skof L c1 with
        | sk_pi sk sk' => sk'
        | _ => sk_one
        end)

   | cn_pair c1 c2 =>
       sk_sigma (skof L c1) (skof L c2)

   | cn_pi1 c' =>
       (match skof L c' with
        | sk_sigma sk1 sk2 => sk1
        | _ => sk_one
        end)

   | cn_pi2 c' =>
       (match skof L c' with
        | sk_sigma sk1 sk2 => sk2
        | _ => sk_one
        end)

   | cn_star => sk_one

   | cn_unit => sk_type
   | cn_arrow t1 t2 => sk_type
   | cn_prod t1 t2 => sk_type
   | cn_all k t => sk_type
   | cn_exists k t => sk_type

   | _ => sk_one
   end).


Lemma simp_single :
  forall c k,
    hkd k -> simp (single c k) = simp k.
Proof.
intros c k Hk.
remember (skel k) as skl eqn:Heqskl.
revert c k Hk Heqskl.
induct skl.

(* type *)
intros c k Hk Heqskl.
invert Hk; try (intros; subst; discriminate Heqskl); [].
intros <-.
simpl.
reflexivity.

(* sing *)
intros c k Hk Heqskl.
invert Hk; try (intros; subst; discriminate Heqskl); [].
intros ? <-.
simpl.
reflexivity.

(* pi *)
intros skl1 _ skl2 IH c k Hk Heqskl.
invert Hk; try (intros; subst; discriminate Heqskl); [].
intros k1 k2 _ Hk2 <-.
simpl.
injection Heqskl.
intros Heq2 Heq1.
f_equal; auto; done.

(* sigma *)
intros skl1 IH1 skl2 IH2 c k Hk Heqskl.
invert Hk; try (intros; subst; discriminate Heqskl); [].
intros k1 k2 Hk1 Hk2 <-.
simpl.
injection Heqskl.
intros Heq2 Heq1.
f_equal; auto; [].
rewrite -> (subst_single _#3 Hk2); [].
simp_sub.
transitivity (simp (subst (dot (cn_pi1 (shift1 c)) sh1) k2)).
  apply IH2; auto using subst_hkd; [].
  rewrite -> subst_skel; auto; done.

  rewrite -> subst_simp; auto; done.

(* one *)
intros c k Hk Heqskl.
invert Hk; try (intros; subst; discriminate Heqskl); [].
intros <-.
simpl.
reflexivity.
Qed.


Lemma prink_skof :
  forall G c k,
    wf G
    -> prink G c k
    -> skof (flatten_ctx G) c = simp k.
Proof.
intros G c k Hwf H.
revert Hwf.
#6 induct H; try reflexivity.

(* var *)
intros G i k Hindex Hwf.
simpl.
assert (hkd k) as Hk.
  so (wf_fcontext _ Hwf) as HG.
  clear Hwf; revert HG.
  remember (cl_cn k) as A eqn:HeqA.
  revert HeqA.
  #2 induct Hindex.
    (* 0 *)
    intros A G -> H.
    invert H; eauto using fkd_hkd; done.
  
    (* S *)
    intros i B G A _ IH HeqA HG.
    eauto using fcontext_tail; done.
rewrite -> (simp_single _ _ (subst_hkd _ _ Hk)); [].
rewrite -> (subst_simp _ _ Hk); [].
repl (nth i (flatten_ctx G) None) with (Some (simp k)).
  rewrite -> (map_nth (fun x => (match x with cl_cn k => Some (simp k) | _ => None end)) G nonsense); [].
  rewrite -> (index_nth _#5 Hindex); [].
  reflexivity.
reflexivity.

(* lam *)
intros G k1 k2 c Hofk _ IH Hwf.
simpl.
f_equal; [].
apply IH; auto using wf; done.

(* app *)
intros G k1 k2 c1 c2 Hprin1 IH Hofc2 Hwf.
simpl.
rewrite -> (IH Hwf); [].
simpl.
unfold subst1.
symmetry.
apply subst_simp; [].
so (ofc_freg _#3 (wf_fcontext _ Hwf) (prink_sound _#3 Hwf Hprin1) ander) as Hk.
invert Hk; eauto using fkd_hkd; done.

(* pair *)
intros G k1 k2 c1 c2 _ IH1 Hprin2 IH2 Hwf.
simpl.
f_equal; auto; [].
unfold shift1.
so (fkd_hkd _ _ (ofc_freg _#3 (wf_fcontext _ Hwf) (prink_sound _#3 Hwf Hprin2) ander)) as Hk2.
rewrite -> subst_simp; auto; done.

(* pi1 *)
intros G k1 k2 c _ IH Hwf.
simpl.
rewrite -> (IH Hwf); [].
simpl.
reflexivity.

(* pi2 *)
intros G k1 k2 c Hprin IH Hwf.
simpl.
rewrite -> (IH Hwf); [].
simpl.
so (fkd_hkd _ _ (ofc_freg _#3 (wf_fcontext _ Hwf) (prink_sound _#3 Hwf Hprin) ander)) as Hk.
invert Hk; []; intros Hk1 Hk2.
unfold subst1.
rewrite -> subst_simp; auto; done.
Qed.


Lemma ofc_skof :
  forall G c k,
    wf G
    -> ofc G c k
    -> skof (flatten_ctx G) c = simp k.
Proof.
intros G c k Hwf Hofc.
so (prink_complete _#3 Hwf Hofc) as (k' & Hprin & Hsubk).
rewrite -> (prink_skof _#3 Hwf Hprin); [].
eapply subk_simp; eauto; done.
Qed.
