
Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import TSubst.
Require Import Regularity.
Require Import Algorithm.
Require Import Declarative.
Require Import SubjectReduction.


(* We could generalize this, but this is all we need. *)
Inductive ctxeq : context -> context -> Prop :=
| ctxeq_nil
    : ctxeq nil nil

| ctxeq_cn {G G' k k'}
    : ctxeq G G'
      -> eqk G k k'
      -> ctxeq (G; cl_cn k) (G'; cl_cn k')

| ctxeq_tm {G G' t}
    : ctxeq G G'
      -> ofc G t kd_type
      -> ctxeq (G; cl_tm t) (G'; cl_tm t)

| ctxeq_md {G G' T k}
    : ctxeq G G'
      -> ofs G T
      -> fsts T k
      -> ctxeq (G; cl_cn k; cl_md (shift1 T) (var 0)) (G'; cl_cn k; cl_md (shift1 T) (var 0))
.


Hint Constructors ctxeq : static.


Lemma ctxeq_refl :
  forall G, wf G -> ctxeq G G.
Proof.
intros G Hwf.
induct Hwf; eauto with static.
Qed.


Lemma ctxeq_sub :
  forall G G',
    ctxeq G G'
    -> subof sof G' (under (length G) id) G.
Proof.
intros G G' Hctxeq.
induct Hctxeq.

(* nil *)
simpl; auto with static; done.

(* cn *)
intros G G' k k' _ IH Heqk.
simpl length.
rewrite -> under_succ.
#2 apply subof_dot.
  so (subof_weaken_sof _ (nil; cl_cn k') _ _ IH) as Hsubof.
  simpl in Hsubof.
  assumption.

  simp_sub.
  apply sof_cn; [].
  rewrite -> subst_compose.
  rewrite -> subst_under_id.
  so (subst_eqk _#5 IH Heqk) as Heqk'.
  rewrite -> !subst_under_id in Heqk'.
  so (weaken_eqk _ (nil; cl_cn k') _ _ Heqk') as Heqk''.
  simpl length in Heqk''.
  autorewrite with canonlist in Heqk''.
  eauto using index_0 with static; done.

(* tm *)
intros G G' t _ IH _.
simpl length.
rewrite -> under_succ.
#2 apply subof_dot.
  so (subof_weaken_sof _ (nil; cl_tm t) _ _ IH) as Hsubof.
  simpl in Hsubof.
  assumption.

  simp_sub.
  apply sof_tm; done.

(* md *)
intros G G' T k _ IH _ _.
simpl length.
rewrite -> under_succ.
#2 apply subof_dot.
  rewrite -> under_succ.
  simp_sub.
  #2 apply subof_dot.
    so (subof_weaken_sof _ (nil; cl_cn k; cl_md (shift1 T) (var 0)) _ _ IH) as Hsubof.
    simpl in Hsubof.
    assumption.

    simp_sub.
    apply sof_cn; [].
    rewrite -> subst_compose.
    rewrite -> subst_under_id.
    apply ofc_var; [].
    apply index_S; [].
    apply index_0; done.

  simp_sub.
  apply sof_md; done.
Qed.


Lemma ctxeq_ofk :
  forall G G' k,
    ctxeq G G'
    -> ofk G k
    -> ofk G' k.
Proof.
intros G G' k Hctxeq Heqk.
so (ctxeq_sub _ _ Hctxeq) as Hsubof.
so (subst_ofk _#4 Hsubof Heqk) as Heqk'.
rewrite -> !subst_under_id in Heqk'.
assumption.
Qed.


Lemma ctxeq_eqk :
  forall G G' k k',
    ctxeq G G'
    -> eqk G k k'
    -> eqk G' k k'.
Proof.
intros G G' k k' Hctxeq Heqk.
so (ctxeq_sub _ _ Hctxeq) as Hsubof.
so (subst_eqk _#5 Hsubof Heqk) as Heqk'.
rewrite -> !subst_under_id in Heqk'.
assumption.
Qed.


Lemma ctxeq_ofc :
  forall G G' c k,
    ctxeq G G'
    -> ofc G c k
    -> ofc G' c k.
Proof.
intros G G' c k Hctxeq Hofc.
so (ctxeq_sub _ _ Hctxeq) as Hsubof.
so (subst_ofc _#5 Hsubof Hofc) as Hofc'.
rewrite -> !subst_under_id in Hofc'.
assumption.
Qed.


Lemma ctxeq_eqc :
  forall G G' c c' k,
    ctxeq G G'
    -> eqc G c c' k
    -> eqc G' c c' k.
Proof.
intros G G' c c' k Hctxeq Heqc.
so (ctxeq_sub _ _ Hctxeq) as Hsubof.
so (subst_eqc _#6 Hsubof Heqc) as Heqc'.
rewrite -> !subst_under_id in Heqc'.
assumption.
Qed.


Lemma ctxeq_symm :
  forall G G', ctxeq G G' -> ctxeq G' G.
Proof.
intros G G' Hctxeq.
induct Hctxeq.

(* nil *)
apply ctxeq_nil; done.

(* cn *)
intros G G' k k' Hctxeq IH Heqk.
apply ctxeq_cn; auto; [].
eapply ctxeq_eqk; eauto; [].
apply eqk_symm; auto; done.

(* tm *)
intros G G' t Hctxeq IH Hofc.
apply ctxeq_tm; auto; [].
eapply ctxeq_ofc; eauto; done.

(* md *)
intros G G' T k Hctxeq IH Hofs Hfst.
apply ctxeq_md; auto; [].
so (subst_ofs _#4 (ctxeq_sub _ _ Hctxeq) Hofs) as Hofs'.
rewrite -> !subst_under_id in Hofs'.
assumption.
Qed.


Lemma ctxeq_reg :
  forall G G', ctxeq G G' -> wf G /\ wf G'.
Proof.
intros G G' Hctxeq.
induct Hctxeq.

(* nil *)
auto with static; done.

(* cn *)
intros G G' k k' Hctxeq (Hwf & Hwf') Heqk.
so (eqk_reg _#3 Hwf Heqk) as (Hofk & Hofk').
split; eauto using ctxeq_ofk with static; done.

(* tm *)
intros G G' t Hctxeq (Hwf & Hwf') Heqc.
split; eauto using ctxeq_ofc with static; done.

(* md *)
intros G G' T k Hctxeq (Hwf & Hwf') Hofs Hfsts.
so (subst_ofs _#4 (ctxeq_sub _ _ Hctxeq) Hofs) as Hofs'.
rewrite -> subst_under_id in Hofs'.
split; eauto with static; done.
Qed.


Lemma ctxeq_tail :
  forall G G' A A',
    ctxeq (G; A) (G'; A')
    -> ctxeq G G'.
Proof.
intros G G' A A' Hctxeq.
invert Hctxeq; try (intros; auto; done); [].
intros G1 G1' T k Hctxeq' Hofs Hfsts <- <- <- <-.
apply ctxeq_cn; auto; [].
apply eqk_refl; [].
eapply ofs_fst; eauto; done.
Qed.



Lemma headred_reduce :
  forall G c c',
    headred G c c'
    -> reduce G c c'.
Proof.
intros G c c' Hred.
induct Hred; eauto using reduce.
Qed.


Lemma headred_sound :
  forall G c c' k,
    wf G
    -> ofc G c k
    -> headred G c c'
    -> eqc G c c' k.
Proof.
intros G c c' k Hwf Hofc Hred.
apply subject_reduction; auto using headred_reduce; done.
Qed.


Lemma headnorm_sound :
  forall G c c' k,
    wf G
    -> ofc G c k
    -> headnorm G c c'
    -> eqc G c c' k.
Proof.
intros G c c' k Hwf Hof Hnorm.
revert Hwf Hof.
induct Hnorm.

(* more *)
intros G c c' c'' Hred _ IH Hwf Hof.
so (headred_sound _#4 Hwf Hof Hred) as Heqc.
eapply eqc_trans; eauto; [].
apply IH; auto; [].
exact (eqc_reg _#4 Hwf Heqc ander andel).

(* done *)
intros G c H Hwf Hof.
apply eqc_refl; auto; done.
Qed.


Lemma eqc_pieta :
  forall G k1 k2 c,
    ofk G k1
    -> ofc G c (kd_pi k1 k2)
    -> eqc G c (cn_lam k1 (cn_app (shift1 c) (var 0))) (kd_pi k1 k2).
Proof.
intros G k1 k2 c Hofk1 Hofc.
#3 apply (@eqc_extpi _#3 k2 k2).
  assumption.

  apply ofc_lam; auto; [].
  #3 exploit (@ofc_app (G; cl_cn k1) (shift1 k1) (subst (dot (var 0) (sh 2)) k2) (shift1 c) (var 0)) as H.
    so (weaken1_ofc _ (cl_cn k1) _ _ Hofc) as H.
    simp_sub in H.
    assumption.

    apply ofc_var; auto using index_0; done.

    simp_sub in H.
    assumption.

  simp_sub.
  apply eqc_symm; [].
  #3 exploit (@eqc_beta (G; cl_cn k1) (shift1 k1) (subst (dot (var 0) (sh 2)) k2) (var 0) (cn_app (subst (sh 2) c) (var 0))) as H.
    rewrite -> split_dot.
    rewrite -> subst_compose.
    #2 apply (@ofc_app _ (subst (sh 2) k1)).
      so (weaken_ofc _ (nil; cl_cn k1; cl_cn (shift1 k1)) _ _ Hofc) as Hofc'.
      autorewrite with canonlist in Hofc'.
      simpl length in Hofc'.
      simp_sub in Hofc'.
      assumption.

      repl (subst (sh 2) k1) with (subst (sh 1) (shift1 k1)) by (simp_sub; reflexivity).
      apply ofc_var; auto using index_0; done.

    apply ofc_var; auto using index_0; done.

    simp_sub in H.
    assumption.
Qed.


Lemma eqc_sigmaeta :
  forall G k1 k2 c,
    ofk (G; cl_cn k1) k2
    -> ofc G c (kd_sigma k1 k2)
    -> eqc G c (cn_pair (cn_pi1 c) (cn_pi2 c)) (kd_sigma k1 k2).
Proof.
intros G k1 k2 c Hofk2 Hofc.
#2 apply eqc_extsigma; auto.
  apply eqc_symm; [].
  eapply eqc_beta1; eauto with static; done.

  apply eqc_symm; [].
  eapply eqc_beta2; eauto with static; done.
Qed.


Theorem alg_sound :
  (forall G G' c c' k k',
     ctxeq G G'
     -> eqk G k k'
     -> ofc G c k
     -> ofc G' c' k'
     -> aeq G G' c c' k k'
     -> eqc G c c' k)
  /\
  (forall G G' p p' l l' k k',
     ctxeq G G'
     -> ofc G p l
     -> ofc G' p' l'
     -> peq G G' p p' k k'
     -> eqk G k k' /\ eqc G p p' k)
  /\
  (forall G G' k k',
     ctxeq G G'
     -> ofk G k
     -> ofk G' k'
     -> aeqk G G' k k'
     -> eqk G k k').
Proof.
exploit
  (alg_ind
     (fun G G' c c' k k' =>
        ctxeq G G'
        -> eqk G k k'
        -> ofc G c k
        -> ofc G' c' k'
        -> eqc G c c' k)
     (fun G G' p p' k k' =>
        forall l l',
          ctxeq G G'
          -> ofc G p l
          -> ofc G' p' l'
          -> eqk G k k' /\ eqc G p p' k)
     (fun G G' k k' =>
        ctxeq G G'
        -> ofk G k
        -> ofk G' k'
        -> eqk G k k')) as Hind.

(* type *)
intros G G' c c' p p' Hnorm Hnorm' _ IH Hctxeq Heqk Hofc Hofc'.
so (ctxeq_reg _ _ Hctxeq) as (Hwf & Hwf').
so (headnorm_sound _#4 Hwf Hofc Hnorm) as Heq_cp.
so (headnorm_sound _#4 Hwf' Hofc' Hnorm') as Heq_cp'.
so (ctxeq_eqc _#5 (ctxeq_symm _ _ Hctxeq) Heq_cp') as Heq_cp''.
#3 exploit (IH kd_type kd_type) as (_ & Heqp); auto.
  eapply eqc_reg; eauto; done.

  eapply eqc_reg; eauto; done.

  eauto using eqc_trans, eqc_symm; done.

(* sing *)
intros G G' c c' b b' Hctxeq Heqk Hofc Hofc'.
so (sing_injective _#3 Heqk) as Heqc.
#2 apply (@eqc_trans _ _ b).
  apply eqc_symm; [].
  apply eqc_sing; [].
  apply eqc_symm; [].
  apply eqc_singelim; auto; done.

  apply eqc_sing; [].
  eapply eqc_trans; eauto; [].
  apply eqc_symm; [].
  apply eqc_singelim; [].
  eapply ctxeq_ofc; eauto; [].
  apply ctxeq_symm; auto; done.

(* pi *)
intros G G' c c' k1 k1' k2 k2' _ IH Hctxeq Heqk Hofc1 Hofc2.
so (ctxeq_reg _ _ Hctxeq) as (Hwf & _).
so (pi_injective _#5 Heqk) as (Heqk1 & Heqk2).
so (eqk_reg _#3 Hwf Heqk1) as (Hofk1 & _).
#2 apply (@eqc_trans _ _ (cn_lam k1 (cn_app (shift1 c) (var 0)))).
  apply eqc_pieta; auto; done.

  apply eqc_symm; [].
  #2 apply (@eqc_trans _ _ (cn_lam k1 (cn_app (shift1 c') (var 0)))).
    apply eqc_pieta; auto; [].
    #2 apply (@ofc_subsume _ (kd_pi k1' k2')).
      eapply ctxeq_ofc; eauto using ctxeq_symm; done.

      apply subk_refl; [].
      apply eqk_symm; auto; done.

    apply eqc_lam; auto with static; [].
    apply eqc_symm; [].
    #3 apply IH; auto.
      apply ctxeq_cn; auto; done.

      rewrite <- (subst_dot0_sh k2).
      rewrite -> split_dot.
      rewrite -> subst_compose.
      #2 apply (@ofc_app _ (shift1 k1)).
        so (weaken1_ofc _ (cl_cn k1) _ _ Hofc1) as Hofc1'.
        simp_sub in Hofc1'.
        simp_sub.
        assumption.

        apply ofc_var; auto using index_0; done.

      rewrite <- (subst_dot0_sh k2').
      rewrite -> split_dot.
      rewrite -> subst_compose.
      #2 apply (@ofc_app _ (shift1 k1')).
        so (weaken1_ofc _ (cl_cn k1') _ _ Hofc2) as Hofc2'.
        simp_sub in Hofc2'.
        simp_sub.
        assumption.

        apply ofc_var; auto using index_0; done.

(* sigma *)
intros G G' c c' k1 k1' k2 k2' _ IH1 _ IH2 Hctxeq Heqk Hofc Hofc'.
so (ctxeq_reg _ _ Hctxeq) as (Hwf & _).
so (sigma_injective _#5 Heqk) as (Heqk1 & Heqk2).
so (eqk_reg _#3 Hwf Heqk1) as (Hofk1 & _).
so (eqk_reg _#3 (wf_cn Hwf Hofk1) Heqk2) as (Hofk2 & _).
#2 apply (@eqc_trans _ _ (cn_pair (cn_pi1 c) (cn_pi2 c))).
  apply eqc_sigmaeta; auto; done.

  apply eqc_symm; [].
  #2 apply (@eqc_trans _ _ (cn_pair (cn_pi1 c') (cn_pi2 c'))).
    apply eqc_sigmaeta; auto; [].
    #2 apply (@ofc_subsume _ (kd_sigma k1' k2')).
      eapply ctxeq_ofc; eauto using ctxeq_symm; done.

      apply subk_refl; [].
      apply eqk_symm; auto; done.

    apply eqc_symm; [].
    exploit IH1 as Heqpi1; eauto with static; [].
    apply eqc_pair; auto; [].
    apply IH2; eauto with static; [].
    #2 apply (@eqk_trans _ _ (subst1 (cn_pi1 c') k2)).
      eapply funct_kd; eauto; done.

      eapply subst1_eqk; eauto; [].
      eapply eqc_reg; eauto; done.

(* one *)
intros G G' c c' Hctxeq _ Hofc Hofc'.
apply eqc_extone; auto; [].
eapply ctxeq_ofc; eauto using ctxeq_symm; done.

(* var *)
intros G G' i k k' Hindex Hindex' _ _ Hctxeq _ _.
split.
  revert G G' Hctxeq Hindex Hindex'.
  #2 induct i.
    intros G G' Hctxeq Hindex Hindex'.
    invert Hindex; [].
    intros G1 <-.
    invert Hindex'; [].
    intros G1' <-.
    invert Hctxeq.
    intros _ Heqk.
    apply weaken1_eqk; auto; done.

    intros i IH G G' Hctxeq Hindex Hindex'.
    invert Hindex; [].
    intros A G1 Hindex1 <-.
    invert Hindex'; [].
    intros A' G1' Hindex1' <-.
    so (IH _ _ (ctxeq_tail _#4 Hctxeq) Hindex1 Hindex1') as Heqk.
    so (weaken1_eqk _ A _ _ Heqk) as Heqk'.
    simp_sub in Heqk'.
    assumption.

  apply eqc_refl; [].
  apply ofc_var; auto; done.

(* unit *)
intros G G' l l' _ _ _.
auto with static; done.

(* arrow *)
intros G G' t1 t1' t2 t2' _ IH1 _ IH2 l l' Hctxeq Hof Hof'.
split; auto with static; [].
so (ofc_ofcr _#3 Hof) as (e & Hofr).
so (ofc_ofcr _#3 Hof') as (e' & Hofr').
invert Hofr.
intros Hoft1 Hoft2 <-.
invert Hofr'.
intros Hoft1' Hoft2' <-.
auto with static; done.

(* prod *)
intros G G' t1 t1' t2 t2' _ IH1 _ IH2 l l' Hctxeq Hof Hof'.
split; auto with static; [].
so (ofc_ofcr _#3 Hof) as (e & Hofr).
so (ofc_ofcr _#3 Hof') as (e' & Hofr').
invert Hofr.
intros Hoft1 Hoft2 <-.
invert Hofr'.
intros Hoft1' Hoft2' <-.
auto with static; done.

(* all *)
intros G G' k t k' t' _ IH1 _ IH2 l l' Hctxeq Hof Hof'.
split; auto with static; [].
so (ofc_ofcr _#3 Hof) as (e & Hofr).
so (ofc_ofcr _#3 Hof') as (e' & Hofr').
invert Hofr.
intros Hofk Hoft <-.
invert Hofr'.
intros Hofk' Hoft' <-.
auto with static; done.

(* exists *)
intros G G' k t k' t' _ IH1 _ IH2 l l' Hctxeq Hof Hof'.
split; auto with static; [].
so (ofc_ofcr _#3 Hof) as (e & Hofr).
so (ofc_ofcr _#3 Hof') as (e' & Hofr').
invert Hofr.
intros Hofk Hoft <-.
invert Hofr'.
intros Hofk' Hoft' <-.
auto with static; done.

(* app *)
intros G G' p p' c c' k1 k1' k2 k2' Hpeq IH1 Haeq IH2 l l' Hctxeq Hof Hof'.
so (ofc_ofcr _#3 Hof) as (e & Hofr).
so (ofc_ofcr _#3 Hof') as (e' & Hofr').
invert Hofr.
intros l1 l2 Hofp Hofc <-.
invert Hofr'.
intros l1' l2' Hofp' Hofc' <-.
clear Hof Hof' Hofr Hofr'.
so (peq_natk _#6 Hpeq) as (Hnatk & Hnatk').
so (natk_complete _#3 (natk_path _#3 Hnatk) (ctxeq_reg _ _ Hctxeq andel) Hofp) as (e & Hnatke & Hofpk & Hsubk).
so (natk_complete _#3 (natk_path _#3 Hnatk') (ctxeq_reg _ _ Hctxeq ander) Hofp') as (e' & Hnatke' & Hofpk' & Hsubk').
clear Hofp Hofp'.
so (natk_fun _#4 Hnatk Hnatke); subst e.
so (natk_fun _#4 Hnatk' Hnatke'); subst e'.
clear Hnatk Hnatk' Hnatke Hnatke'.
simpl in Hsubk.
simpl in Hsubk'.
so (pi_injective_subk _#5 Hsubk) as (Hsubk1 & _).
so (pi_injective_subk _#5 Hsubk') as (Hsubk1' & _).
clear Hsubk Hsubk'.
so (IH1 _ _ Hctxeq Hofpk Hofpk') as (Heqk & Heqp).
so (pi_injective _#5 Heqk) as (Heqk1 & Heqk2); clear Heqk.
so (IH2 Hctxeq Heqk1 (ofc_subsume Hofc Hsubk1) (ofc_subsume Hofc' Hsubk1')) as Heqc.
split; eauto with static; [].
eapply funct_eqk; eauto; [].
exact (ctxeq_reg _ _ Hctxeq andel).

(* pi1 *)
intros G G' p p' k1 k1' k2 k2' Hpeq IH l l' Hctxeq Hof Hof'.
so (ofc_ofcr _#3 Hof) as (l1 & Hofr).
so (ofc_ofcr _#3 Hof') as (l1' & Hofr').
invert Hofr.
intros l2 Hofp.
invert Hofr'.
intros l2' Hofp'.
clear Hof Hof' Hofr Hofr'.
so (IH _ _ Hctxeq Hofp Hofp') as (Heqk & Heqc).
so (sigma_injective _#5 Heqk) as (Heqk1 & Heqk2).
eauto with static; done.

(* pi2 *)
intros G G' p p' k1 k1' k2 k2' Hpeq IH l l' Hctxeq Hof Hof'.
so (ofc_ofcr _#3 Hof) as (e & Hofr).
so (ofc_ofcr _#3 Hof') as (e' & Hofr').
invert Hofr.
intros l1 l2 Hofp <-.
invert Hofr'.
intros l1' l2' Hofp' <-.
clear Hof Hof' Hofr Hofr'.
so (IH _ _ Hctxeq Hofp Hofp') as (Heqk & Heqc).
so (sigma_injective _#5 Heqk) as (Heqk1 & Heqk2).
split; eauto with static; [].
eapply funct_eqk; eauto with static; [].
exact (ctxeq_reg _ _ Hctxeq andel).

(* star *)
auto with static; done.

(* type *)
auto with static; done.

(* sing *)
intros G G' c c' _ IH Hctxeq Hof Hof'.
invert Hof.
intros Hofc.
invert Hof'.
intros Hofc'.
so (IH Hctxeq (eqk_refl ofk_type) Hofc Hofc') as Heqc.
apply eqk_sing; auto; done.

(* pi *)
intros G G' k1 k1' k2 k2' _ IH1 _ IH2 Hctxeq Hof Hof'.
invert Hof.
intros Hofk1 Hofk2.
invert Hof'.
intros Hofk1' Hofk2'.
so (IH1 Hctxeq Hofk1 Hofk1') as Heqk1.
sor (refine (IH2 _ Hofk2 Hofk2')) as Heqk2; auto using ctxeq_cn; [].
apply eqk_pi; auto; done.

(* sigma *)
intros G G' k1 k1' k2 k2' _ IH1 _ IH2 Hctxeq Hof Hof'.
invert Hof.
intros Hofk1 Hofk2.
invert Hof'.
intros Hofk1' Hofk2'.
so (IH1 Hctxeq Hofk1 Hofk1') as Heqk1.
sor (refine (IH2 _ Hofk2 Hofk2')) as Heqk2; auto using ctxeq_cn; [].
apply eqk_sigma; auto; done.

(* one *)
auto with static; done.

(* epilogue *)
destruct_all Hind.
do2 2 split; intros; eauto; done.
Qed.


Lemma aeq_sound :
  forall G G' c c' k k',
    ctxeq G G'
    -> eqk G k k'
    -> ofc G c k
    -> ofc G' c' k'
    -> aeq G G' c c' k k'
    -> eqc G c c' k.
Proof.
exact (alg_sound andel).
Qed.


Lemma aeqk_sound :
  forall G G' k k',
    ctxeq G G'
    -> ofk G k
    -> ofk G' k'
    -> aeqk G G' k k'
    -> eqk G k k'.
Proof.
exact (alg_sound ander ander).
Qed.
