
Require Import Tactics.
Require Import Subst.
Export Syntax.
Require Import Sequence.


Inductive iscl : term -> Prop :=
| iscl_cn {t}
    : iscl (cl_cn t)

| iscl_tm {t}
    : iscl (cl_tm t)

| iscl_md {t t'}
    : iscl (cl_md t t')

| iscl_ucn
    : iscl cl_ucn

| iscl_utm
    : iscl cl_utm

| iscl_umd {c}
    : iscl (cl_umd c)
.

Hint Constructors iscl : static.


(* depends on static hints *)
Lemma subst_cl :
  forall s c, iscl c -> iscl (subst s c).
Proof.
intros s c Hc.
destruct Hc; simp_sub; auto with static.
Qed.



Inductive ofk : context -> (* kd *) term -> Prop :=
| ofk_type {G}
    : ofk G kd_type

| ofk_sing {G c}
    : ofc G c kd_type
      -> ofk G (kd_sing c)

| ofk_pi {G k1 k2}
    : ofk G k1
      -> ofk (G; cl_cn k1) k2
      -> ofk G (kd_pi k1 k2)

| ofk_sigma {G k1 k2}
    : ofk G k1
      -> ofk (G; cl_cn k1) k2
      -> ofk G (kd_sigma k1 k2)

| ofk_one {G}
    : ofk G kd_one


with eqk : context -> (* kd *) term -> (* kd *) term -> Prop :=
| eqk_refl {G k}
    : ofk G k
      -> eqk G k k

| eqk_symm {G k k'}
    : eqk G k k'
      -> eqk G k' k

| eqk_trans {G k1 k2 k3}
    : eqk G k1 k2
      -> eqk G k2 k3
      -> eqk G k1 k3

| eqk_sing {G c c'}
    : eqc G c c' kd_type
      -> eqk G (kd_sing c) (kd_sing c')

| eqk_pi {G k1 k1' k2 k2'}
    : eqk G k1 k1'
      -> eqk (G; cl_cn k1) k2 k2'
      -> eqk G (kd_pi k1 k2) (kd_pi k1' k2')

| eqk_sigma {G k1 k1' k2 k2'}
    : eqk G k1 k1'
      -> eqk (G; cl_cn k1) k2 k2'
      -> eqk G (kd_sigma k1 k2) (kd_sigma k1' k2')


with subk : context -> (* kd *) term -> (* kd *) term -> Prop :=
| subk_refl {G k k'}
    : eqk G k k'
      -> subk G k k'

| subk_trans {G k1 k2 k3}
    : subk G k1 k2
      -> subk G k2 k3
      -> subk G k1 k3

| subk_sing_t {G c}
    : ofc G c kd_type
      -> subk G (kd_sing c) kd_type

| subk_pi {G k1 k1' k2 k2'}
    : subk G k1' k1
      -> subk (G; cl_cn k1') k2 k2'
      -> ofk (G; cl_cn k1) k2
      -> subk G (kd_pi k1 k2) (kd_pi k1' k2')

| subk_sigma {G k1 k1' k2 k2'}
    : subk G k1 k1'
      -> subk (G; cl_cn k1) k2 k2'
      -> ofk (G; cl_cn k1') k2'
      -> subk G (kd_sigma k1 k2) (kd_sigma k1' k2')


with ofc : context -> (* cn *) term -> (* kd *) term -> Prop :=
| ofc_var {G i k}
    : index i G (cl_cn k)
      -> ofc G (var i) (subst (sh (S i)) k)

| ofc_lam {G k1 k2 c}
    : ofk G k1
      -> ofc (G; cl_cn k1) c k2
      -> ofc G (cn_lam k1 c) (kd_pi k1 k2)

| ofc_app {G k1 k2 c1 c2}
    : ofc G c1 (kd_pi k1 k2)
      -> ofc G c2 k1
      -> ofc G (cn_app c1 c2) (subst1 c2 k2)

| ofc_pair {G k1 k2 c1 c2}
    : ofc G c1 k1
      -> ofc G c2 (subst1 c1 k2)
      -> ofk (G; cl_cn k1) k2
      -> ofc G (cn_pair c1 c2) (kd_sigma k1 k2)

| ofc_pi1 {G k1 k2 c}
    : ofc G c (kd_sigma k1 k2)
      -> ofc G (cn_pi1 c) k1

| ofc_pi2 {G k1 k2 c}
    : ofc G c (kd_sigma k1 k2)
      -> ofc G (cn_pi2 c) (subst1 (cn_pi1 c) k2)

| ofc_star {G}
    : ofc G cn_star kd_one

| ofc_unit {G}
    : ofc G cn_unit kd_type

| ofc_arrow {G t1 t2}
    : ofc G t1 kd_type
      -> ofc G t2 kd_type
      -> ofc G (cn_arrow t1 t2) kd_type

| ofc_prod {G t1 t2}
    : ofc G t1 kd_type
      -> ofc G t2 kd_type
      -> ofc G (cn_prod t1 t2) kd_type

| ofc_all {G k t}
    : ofk G k
      -> ofc (G; cl_cn k) t kd_type
      -> ofc G (cn_all k t) kd_type

| ofc_exists {G k t}
    : ofk G k
      -> ofc (G; cl_cn k) t kd_type
      -> ofc G (cn_exists k t) kd_type

| ofc_sing {G c}
    : ofc G c kd_type
      -> ofc G c (kd_sing c)

| ofc_extpi {G k1 k2 k3 c}
    : ofc G c (kd_pi k1 k3)
      -> ofc (G; cl_cn k1) (cn_app (shift1 c) (var 0)) k2
      -> ofc G c (kd_pi k1 k2)

| ofc_extsigma {G k1 k2 c}
    : ofc G (cn_pi1 c) k1
      -> ofc G (cn_pi2 c) (subst1 (cn_pi1 c) k2)
      -> ofk (G; cl_cn k1) k2
      -> ofc G c (kd_sigma k1 k2)

| ofc_subsume {G k k' c}
    : ofc G c k
      -> subk G k k'
      -> ofc G c k'


with eqc : context -> (* cn *) term -> (* cn *) term -> (* kd *) term -> Prop :=
| eqc_refl {G c k}
    : ofc G c k
      -> eqc G c c k

| eqc_symm {G c c' k}
    : eqc G c c' k 
      -> eqc G c' c k

| eqc_trans {G c1 c2 c3 k}
    : eqc G c1 c2 k
      -> eqc G c2 c3 k
      -> eqc G c1 c3 k

| eqc_lam {G k1 k1' k2 c c'}
    : eqk G k1 k1'
      -> eqc (G; cl_cn k1) c c' k2
      -> eqc G (cn_lam k1 c) (cn_lam k1' c') (kd_pi k1 k2)

| eqc_app {G k1 k2 c1 c2 c1' c2'}
    : eqc G c1 c1' (kd_pi k1 k2)
      -> eqc G c2 c2' k1
      -> eqc G (cn_app c1 c2) (cn_app c1' c2') (subst1 c2 k2)

| eqc_pair {G k1 k2 c1 c1' c2 c2'}
    : eqc G c1 c1' k1
      -> eqc G c2 c2' (subst1 c1 k2)
      -> ofk (G; cl_cn k1) k2
      -> eqc G (cn_pair c1 c2) (cn_pair c1' c2') (kd_sigma k1 k2)

| eqc_pi1 {G k1 k2 c c'}
    : eqc G c c' (kd_sigma k1 k2)
      -> eqc G (cn_pi1 c) (cn_pi1 c') k1

| eqc_pi2 {G k1 k2 c c'}
    : eqc G c c' (kd_sigma k1 k2)
      -> eqc G (cn_pi2 c) (cn_pi2 c') (subst1 (cn_pi1 c) k2)

| eqc_arrow {G t1 t1' t2 t2'}
    : eqc G t1 t1' kd_type
      -> eqc G t2 t2' kd_type
      -> eqc G (cn_arrow t1 t2) (cn_arrow t1' t2') kd_type

| eqc_prod {G t1 t1' t2 t2'}
    : eqc G t1 t1' kd_type
      -> eqc G t2 t2' kd_type
      -> eqc G (cn_prod t1 t2) (cn_prod t1' t2') kd_type

| eqc_all {G k k' t t'}
    : eqk G k k'
      -> eqc (G; cl_cn k) t t' kd_type
      -> eqc G (cn_all k t) (cn_all k' t') kd_type

| eqc_exists {G k k' t t'}
    : eqk G k k'
      -> eqc (G; cl_cn k) t t' kd_type
      -> eqc G (cn_exists k t) (cn_exists k' t') kd_type

| eqc_sing {G c c'}
    : eqc G c c' kd_type
      -> eqc G c c' (kd_sing c)

| eqc_singelim {G c c'}
    : ofc G c (kd_sing c')
      -> eqc G c c' kd_type

| eqc_extpi {G k1 k2 k3 k3' c c'}
    : ofc G c (kd_pi k1 k3)
      -> ofc G c' (kd_pi k1 k3')
      -> eqc (G; cl_cn k1) (cn_app (shift1 c) (var 0)) (cn_app (shift1 c') (var 0)) k2
      -> eqc G c c' (kd_pi k1 k2)

| eqc_extpiw {G k1 k2 k3 c c'}
    : eqc G c c' (kd_pi k1 k3)
      -> eqc (G; cl_cn k1) (cn_app (shift1 c) (var 0)) (cn_app (shift1 c') (var 0)) k2
      -> eqc G c c' (kd_pi k1 k2)

| eqc_extsigma {G k1 k2 c c'}
    : eqc G (cn_pi1 c) (cn_pi1 c') k1
      -> eqc G (cn_pi2 c) (cn_pi2 c') (subst1 (cn_pi1 c) k2)
      -> ofk (G; cl_cn k1) k2
      -> eqc G c c' (kd_sigma k1 k2)

| eqc_extone {G c c'}
    : ofc G c kd_one
      -> ofc G c' kd_one
      -> eqc G c c' kd_one

| eqc_subsume {G c c' k k'}
    : eqc G c c' k
      -> subk G k k'
      -> eqc G c c' k'

| eqc_beta {G k1 k2 c1 c2}
    : ofc (G; cl_cn k1) c2 k2
      -> ofc G c1 k1
      -> eqc G (cn_app (cn_lam k1 c2) c1) (subst1 c1 c2) (subst1 c1 k2)

| eqc_beta1 {G k1 k2 c1 c2}
    : ofc G c1 k1
      -> ofc G c2 k2
      -> eqc G (cn_pi1 (cn_pair c1 c2)) c1 k1
    
| eqc_beta2 {G k1 k2 c1 c2}
    : ofc G c1 k1
      -> ofc G c2 k2
      -> eqc G (cn_pi2 (cn_pair c1 c2)) c2 k2
.


Scheme ofk_mut_ind := Minimality for ofk Sort Prop
with   eqk_mut_ind := Minimality for eqk Sort Prop
with   subk_mut_ind := Minimality for subk Sort Prop
with   ofc_mut_ind := Minimality for ofc Sort Prop
with   eqc_mut_ind := Minimality for eqc Sort Prop.
Combined Scheme static_ind from ofk_mut_ind, eqk_mut_ind, subk_mut_ind, ofc_mut_ind, eqc_mut_ind.


Inductive fsts : (* sg *) term -> (* kd *) term -> Prop :=
| fsts_one
    : fsts sg_one kd_one

| fsts_stat {k}
    : fsts (sg_stat k) k

| fsts_dyn {t}
    : fsts (sg_dyn t) kd_one

| fsts_pigen {T1 T2}
    : fsts (sg_pigen T1 T2) kd_one

| fsts_piapp {T1 T2 k1 k2}
    : fsts T1 k1
      -> fsts T2 k2
      -> fsts (sg_piapp T1 T2) (kd_pi k1 k2)

| fsts_sigma {T1 T2 k1 k2}
    : fsts T1 k1
      -> fsts T2 k2
      -> fsts (sg_sigma T1 T2) (kd_sigma k1 k2)
.


Inductive ofs : context -> (* sg *) term -> Prop :=
| ofs_one {G}
    : ofs G sg_one

| ofs_stat {G k}
    : ofk G k
      -> ofs G (sg_stat k)

| ofs_dyn {G t}
    : ofc G t kd_type
      -> ofs G (sg_dyn t)

| ofs_pigen {G T1 T2 k1}
    : ofs G T1
      -> fsts T1 k1
      -> ofs (G; cl_cn k1) T2
      -> ofs G (sg_pigen T1 T2)

| ofs_piapp {G T1 T2 k1}
    : ofs G T1
      -> fsts T1 k1
      -> ofs (G; cl_cn k1) T2
      -> ofs G (sg_piapp T1 T2)

| ofs_sigma {G T1 T2 k1}
    : ofs G T1
      -> fsts T1 k1
      -> ofs (G; cl_cn k1) T2
      -> ofs G (sg_sigma T1 T2)
.


Inductive eqs : context -> (* sg *) term -> (* sg *) term -> Prop :=
| eqs_refl {G T}
    : ofs G T
      -> eqs G T T

| eqs_symm {G T T'}
    : eqs G T T'
      -> eqs G T' T

| eqs_trans {G T1 T2 T3}
    : eqs G T1 T2
      -> eqs G T2 T3
      -> eqs G T1 T3

| eqs_stat {G k k'}
    : eqk G k k'
      -> eqs G (sg_stat k) (sg_stat k')

| eqs_dyn {G t t'}
    : eqc G t t' kd_type
      -> eqs G (sg_dyn t) (sg_dyn t')

| eqs_pigen {G T1 T1' k1 T2 T2'}
    : eqs G T1 T1'
      -> fsts T1 k1
      -> eqs (G; cl_cn k1) T2 T2'
      -> eqs G (sg_pigen T1 T2) (sg_pigen T1' T2')

| eqs_piapp {G T1 T1' k1 T2 T2'}
    : eqs G T1 T1'
      -> fsts T1 k1
      -> eqs (G; cl_cn k1) T2 T2'
      -> eqs G (sg_piapp T1 T2) (sg_piapp T1' T2')

| eqs_sigma {G T1 T1' k1 T2 T2'}
    : eqs G T1 T1'
      -> fsts T1 k1
      -> eqs (G; cl_cn k1) T2 T2'
      -> eqs G (sg_sigma T1 T2) (sg_sigma T1' T2')
.


Inductive subs : context -> (* sg *) term -> (* sg *) term -> Prop :=
| subs_refl {G T T'}
    : eqs G T T'
      -> subs G T T'

| subs_trans {G T1 T2 T3}
    : subs G T1 T2
      -> subs G T2 T3
      -> subs G T1 T3

| subs_stat {G k k'}
    : subk G k k'
      -> subs G (sg_stat k) (sg_stat k')

| subs_pigen {G T1 T1' k1 k1' T2 T2'}
    : subs G T1' T1
      -> fsts T1' k1'
      -> subs (G; cl_cn k1') T2 T2'
      -> fsts T1 k1
      -> ofs (G; cl_cn k1) T2
      -> subs G (sg_pigen T1 T2) (sg_pigen T1' T2')

| subs_piapp {G T1 T1' k1 k1' T2 T2'}
    : subs G T1' T1
      -> fsts T1' k1'
      -> subs (G; cl_cn k1') T2 T2'
      -> fsts T1 k1
      -> ofs (G; cl_cn k1) T2
      -> subs G (sg_piapp T1 T2) (sg_piapp T1' T2')

| subs_sigma {G T1 T1' k1 k1' T2 T2'}
    : subs G T1 T1'
      -> fsts T1 k1
      -> subs (G; cl_cn k1) T2 T2'
      -> fsts T1' k1'
      -> ofs (G; cl_cn k1') T2'
      -> subs G (sg_sigma T1 T2) (sg_sigma T1' T2')
.


Inductive purity : Set :=
| pure : purity
| impure : purity
.


Inductive fstm : context -> (* md *) term -> (* cn *) term -> Prop :=
| fstm_var {G i T c}
    : index i G (cl_md T c)
      -> fstm G (var i) (subst (sh (S i)) c)

| fstm_varu {G i c}
    : index i G (cl_umd c)
      -> fstm G (var i) (subst (sh (S i)) c)

| fstm_star {G}
    : fstm G md_star cn_star

| fstm_stat {G c}
    : fstm G (md_stat c) c

| fstm_dyn {G e}
    : fstm G (md_dyn e) cn_star

| fstm_lamgn {G T M}
    : fstm G (md_lamgn T M) cn_star

| fstm_lamap {G T M k c}
    : fsts T k
      -> fstm (G; cl_cn k; cl_md (shift1 T) (var 0)) M (shift1 c)
      -> fstm G (md_lamap T M) (cn_lam k c)

| fstm_appap {G M1 M2 c1 c2}
    : fstm G M1 c1
      -> fstm G M2 c2
      -> fstm G (md_appap M1 M2) (cn_app c1 c2)

| fstm_pair {G M1 M2 c1 c2}
    : fstm G M1 c1
      -> fstm G M2 c2
      -> fstm G (md_pair M1 M2) (cn_pair c1 c2)

| fstm_pi1 {G M c}
    : fstm G M c
      -> fstm G (md_pi1 M) (cn_pi1 c)

| fstm_pi2 {G M c}
    : fstm G M c
      -> fstm G (md_pi2 M) (cn_pi2 c)

| fstm_lett {G e M c}
    : fstm (G; cl_utm) M (shift1 c)
      -> fstm G (md_lett e M) c
.


Inductive oft : context -> (* tm *) term -> (* cn *) term -> Prop :=
| oft_var {G i t}
    : index i G (cl_tm t)
      -> oft G (var i) (subst (sh (S i)) t)

| oft_star {G}
    : oft G tm_star cn_unit

| oft_lam {G t1 t2 e}
    : ofc G t1 kd_type
      -> oft (G; cl_tm t1) e (shift1 t2)
      -> oft G (tm_lam t1 e) (cn_arrow t1 t2)

| oft_app {G t1 t2 e1 e2}
    : oft G e1 (cn_arrow t1 t2)
      -> oft G e2 t1
      -> oft G (tm_app e1 e2) t2

| oft_pair {G t1 t2 e1 e2}
    : oft G e1 t1
      -> oft G e2 t2
      -> oft G (tm_pair e1 e2) (cn_prod t1 t2)

| oft_pi1 {G t1 t2 e}
    : oft G e (cn_prod t1 t2)
      -> oft G (tm_pi1 e) t1

| oft_pi2 {G t1 t2 e}
    : oft G e (cn_prod t1 t2)
      -> oft G (tm_pi2 e) t2

| oft_plam {G k t e}
    : ofk G k
      -> oft (G; cl_cn k) e t
      -> oft G (tm_plam k e) (cn_all k t)

| oft_papp {G k t e c}
    : oft G e (cn_all k t)
      -> ofc G c k
      -> oft G (tm_papp e c) (subst1 c t)

| oft_pack {G k t c e}
    : ofc G c k
      -> oft G e (subst1 c t)
      -> ofc (G; cl_cn k) t kd_type
      -> oft G (tm_pack c e k t) (cn_exists k t)

| oft_unpack {G k t t' e1 e2}
    : oft G e1 (cn_exists k t)
      -> oft (G; cl_cn k; cl_tm t) e2 (subst (sh 2) t')
      -> ofc G t' kd_type
      -> oft G (tm_unpack e1 e2) t'

| oft_fix {G t e}
    : oft G e (cn_arrow (cn_arrow cn_unit t) t)
      -> oft G (tm_fix t e) t

| oft_lett {G t1 t2 e1 e2}
    : oft G e1 t1
      -> oft (G; cl_tm t1) e2 (shift1 t2)
      -> oft G (tm_lett e1 e2) t2

| oft_equiv {G t t' e}
    : oft G e t
      -> eqc G t t' kd_type
      -> oft G e t'

| oft_letm {G T k t M e}
    : ofm G impure M T
      -> fsts T k
      -> oft (G; cl_cn k; cl_md (shift1 T) (var 0)) e (subst (sh 2) t)
      -> ofc G t kd_type
      -> oft G (tm_letm M e) t

| oft_ext {G t M}
    : ofm G impure M (sg_dyn t)
      -> oft G (tm_ext M) t


with ofm : context -> purity -> (* cn *) term -> (* kd *) term -> Prop :=
| ofm_var {G i T c}
    : index i G (cl_md T c)
      -> ofm G pure (var i) (subst (sh (S i)) T)

| ofm_star {G}
    : ofm G pure md_star sg_one

| ofm_stat {G c k}
    : ofc G c k
      -> ofm G pure (md_stat c) (sg_stat k)

| ofm_dyn {G e t}
    : oft G e t
      -> ofm G pure (md_dyn e) (sg_dyn t)

| ofm_lamgn {G T1 T2 k1 M}
    : ofs G T1
      -> fsts T1 k1
      -> ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M (shift1 T2)
      -> ofm G pure (md_lamgn T1 M) (sg_pigen T1 T2)

| ofm_appgn {G T1 T2 M1 M2 c2}
    : ofm G impure M1 (sg_pigen T1 T2)
      -> ofm G pure M2 T1
      -> fstm G M2 c2
      -> ofm G impure (md_appgn M1 M2) (subst1 c2 T2)

| ofm_lamap {G T1 T2 k1 M}
    : ofs G T1
      -> fsts T1 k1
      -> ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) pure M (shift1 T2)
      -> ofm G pure (md_lamap T1 M) (sg_piapp T1 T2)

| ofm_appap {G p T1 T2 M1 M2 c2}
    : ofm G p M1 (sg_piapp T1 T2)
      -> ofm G pure M2 T1
      -> fstm G M2 c2
      -> ofm G p (md_appap M1 M2) (subst1 c2 T2)

| ofm_pair {G p T1 T2 M1 M2}
    : ofm G p M1 T1
      -> ofm G p M2 T2
      -> ofm G p (md_pair M1 M2) (sg_sigma T1 (shift1 T2))

| ofm_pi1 {G T1 T2 M}
    : ofm G pure M (sg_sigma T1 T2)
      -> ofm G pure (md_pi1 M) T1

| ofm_pi2 {G T1 T2 M c}
    : ofm G pure M (sg_sigma T1 T2)
      -> fstm G M c
      -> ofm G pure (md_pi2 M) (subst1 (cn_pi1 c) T2)

| ofm_unpack {G k t T e M}
    : oft G e (cn_exists k t)
      -> ofm (G; cl_cn k; cl_tm t) impure M (subst (sh 2) T)
      -> ofs G T
      -> ofm G impure (md_unpack e M T) T

| ofm_lett {G p t T e M}
    : oft G e t
      -> ofm (G; cl_tm t) p M (shift1 T)
      -> ofm G p (md_lett e M) T

| ofm_letm {G T1 T2 k1 M1 M2}
    : ofm G impure M1 T1
      -> fsts T1 k1
      -> ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M2 (subst (sh 2) T2)
      -> ofs G T2
      -> ofm G impure (md_letm M1 M2 T2) T2

| ofm_seal {G T M}
    : ofm G impure M T
      -> ofm G impure (md_seal M T) T

| ofm_extstat {G k k' M c}
    : ofm G pure M (sg_stat k')
      -> fstm G M c
      -> ofc G c k
      -> ofm G pure M (sg_stat k)

| ofm_extpi {G T1 T2 T3 k1 M}
    : ofm G pure M (sg_piapp T1 T3)
      -> fsts T1 k1
      -> ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) pure (md_appap (subst (sh 2) M) (var 0)) (shift1 T2)
      -> ofm G pure M (sg_piapp T1 T2)

| ofm_extsigma {G T1 T2 M}
    : ofm G pure (md_pi1 M) T1
      -> ofm G pure (md_pi2 M) T2
      -> ofm G pure M (sg_sigma T1 (shift1 T2))

| ofm_forget {G T M}
    : ofm G pure M T
      -> ofm G impure M T

| ofm_subsume {G p T T' M}
    : ofm G p M T
      -> subs G T T'
      -> ofm G p M T'
.


Scheme oft_mut_ind := Minimality for oft Sort Prop
with   ofm_mut_ind := Minimality for ofm Sort Prop.
Combined Scheme dynamic_ind from oft_mut_ind, ofm_mut_ind.


Inductive wf : context -> Prop :=
| wf_nil
    : wf nil

| wf_cn {G k}
    : wf G
      -> ofk G k
      -> wf (G; cl_cn k)

| wf_tm {G t}
    : wf G
      -> ofc G t kd_type
      -> wf (G; cl_tm t)

| wf_md {G k T}
    : wf G
      -> ofs G T
      -> fsts T k
      -> wf (G; cl_cn k; cl_md (shift1 T) (var 0))
.


Hint Constructors fsts ofk eqk subk ofc eqc fsts ofs eqs subs fstm oft ofm wf : static.


Definition isclx := iscl.
