
Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import TSubst.
Require Import Formation.
Require Import Regularity.
Require Import Declarative.
Require Import Singleton.
Require Import Algorithm.


Lemma natk_complete :
  forall G p k,
    path p
    -> wf G
    -> ofc G p k
    -> exists l,
         natk G p l
         /\ ofc G p l
         /\ subk G (single p l) k.
Proof.
intros G p k Hpath Hwf Hofp.
cut ((exists l,
        natk G p l
        /\ subk G (single p l) k)
     /\
     (forall E p',
        elimcon E
        -> p = E p'
        -> exists l,
             natk G p' l
             /\ ofc G p' l)).
  intros ((l & Hnatk & Hsubk) & Hofall).
  exists l.
  do2 2 split; auto; [].
  so (Hofall _ _ elimcon_null (eq_refl _)) as (l' & Hnatk' & Hofc).
  so (natk_fun _#4 Hnatk Hnatk').
  subst l'.
  assumption.
revert Hpath Hwf.
induct Hofp.

(* var *)
intros G i k Hindex Hpath Hwf.
split.
  exists (subst (sh (S i)) k).
  split; auto with alg static; [].
  apply single_subk; auto with static; done.

  intros E q Helimcon Heq.
  invert Helimcon; try (intros; subst; discriminate Heq); [].
  intros; subst E; subst q.
  exists (subst (sh (S i)) k).
  split; auto with alg static; done.

(* lam *)
intros G k1 k2 c H H0 H1 Hpath.
invert Hpath; done.

(* app *)
intros G k1 k2 p c _ IH Hofc _ Hpath Hwf.
invert Hpath; clear Hpath.
intros Hpath'.
so (IH Hpath' Hwf) as ((l & Hnatk & Hsubk) & Hofall).
so (subk_pi_invert' _#4 Hsubk) as (e1 & e2 & Heq & Hsubk1 & Hsubk2).
assert (exists l1 l2, l = kd_pi l1 l2) as (l1 & l2 & ->).
  destruct l; try (discriminate Heq); [].
  eauto; done.
simpl in Heq.
injection Heq; clear Heq; intros; subst e1 e2.
so (Hofall _ p elimcon_null (eq_refl _)) as (e & Hnatk' & Hofp).
so (natk_fun _#4 Hnatk Hnatk') as Heq; subst e; clear Hnatk'.
split.
  exists (subst1 c l2).
  split; eauto with alg; [].
  so (subst1_subk _#5 Hofc Hsubk2) as Hsubk2'.
  eapply subk_trans; eauto; [].
  apply subk_refl; [].
  so (ofk_hkd _ _ (ofc_reg _#3 Hwf Hofp)) as Hhkd.
  invert Hhkd; intros _ Hhkd2.
  unfold subst1.
  rewrite -> subst_single; auto; [].
  simp_sub.
  apply eqk_refl; [].
  apply single_ofk; eauto with static; done.

  intros E p' Helimcon HeqEp'.
  invert Helimcon; clear Helimcon; intros; subst E; try (discriminate HeqEp'); [].
  subst p'.
  exists (subst1 c l2).
  split; eauto with alg static; done.

(* pair *)
intros G k1 k2 c1 c2 H H0 H1 H2 H3 Hpath Hwf.
invert Hpath; done.

(* pi1 *)
intros G k1 k2 p _ IH Hpath Hwf.
invert Hpath; clear Hpath.
intro Hpath'.
so (IH Hpath' Hwf) as ((l & Hnatk & Hsubk) & Hofall).
so (subk_sigma_invert' _#4 Hsubk) as (e1 & e2 & Heq & Hsubk1 & Hsubk2).
assert (exists l1 l2, l = kd_sigma l1 l2) as (l1 & l2 & ->).
  destruct l; try (discriminate Heq); [].
  eauto; done.
simpl in Heq.
injection Heq; clear Heq; intros; subst e1 e2.
so (Hofall _ p elimcon_null (eq_refl _)) as (e & Hnatk' & Hofp).
so (natk_fun _#4 Hnatk Hnatk') as Heq; subst e; clear Hnatk'.
split.
  exists l1.
  split; eauto with alg; done.

  intros E p' Helimcon HeqEp'.
  #2 invert Helimcon.
    intros <-.
    subst p'.
    exists l1.
    split; eauto with alg static; done.

    intros E' Helimcon' <-.
    injection HeqEp'; intro HeqE'p'.
    exact (Hofall E' p' Helimcon' HeqE'p').

(* pi2 *)
intros G k1 k2 p _ IH Hpath Hwf.
invert Hpath; clear Hpath.
intro Hpath'.
so (IH Hpath' Hwf) as ((l & Hnatk & Hsubk) & Hofall).
so (subk_sigma_invert' _#4 Hsubk) as (e1 & e2 & Heq & Hsubk1 & Hsubk2).
assert (exists l1 l2, l = kd_sigma l1 l2) as (l1 & l2 & ->).
  destruct l; try (discriminate Heq); [].
  eauto; done.
simpl in Heq.
injection Heq; clear Heq; intros; subst e1 e2.
so (Hofall _ p elimcon_null (eq_refl _)) as (e & Hnatk' & Hofp).
so (natk_fun _#4 Hnatk Hnatk') as Heq; subst e; clear Hnatk'.
split.
  exists (subst1 (cn_pi1 p) l2).
  split; eauto with alg; [].
  #2 sor (refine (subst1_subk _ (cn_pi1 p) _#4 Hsubk2)) as Hsubk2'.
    apply single_ofc; auto; [].
    eapply ofc_pi1; eauto; done.
  
    so (ofk_hkd _ _ (ofc_reg _#3 Hwf Hofp)) as Hhkdl.
    invert Hhkdl; intros _ Hhkdl2.
    simp_sub in Hsubk2'.
    unfold subst1 in Hsubk2'.
    rewrite -> subst_single in Hsubk2'; auto; [].
    simp_sub in Hsubk2'.
    assumption.

  intros E p' Helimcon HeqEp'.
  invert Helimcon; clear Helimcon; intros; subst E; try (discriminate HeqEp'); [].
  subst p'.
  exists (subst1 (cn_pi1 p) l2).
  split; eauto with alg static; done.
  
(* star *)
intros G Hpath Hwf.
split.
  exists kd_one.
  auto with static alg; done.

  intros E p' Helimcon HeqEp'.
  invert Helimcon; clear Helimcon; intros; subst E; try (discriminate HeqEp'); [].
  subst p'.
  exists kd_one.
  split; eauto with alg static; done.

(* unit *)
intros G Hpath Hwf.
split.
  exists kd_type.
  split; auto with static alg; [].
  simpl.
  auto with static; done.

  intros E p' Helimcon HeqEp'.
  invert Helimcon; clear Helimcon; intros; subst E; try (discriminate HeqEp'); [].
  subst p'.
  exists kd_type.
  split; eauto with alg static; done.

(* arrow *)
intros G Hpath Hwf.
split.
  exists kd_type.
  split; auto with static alg; [].
  simpl.
  auto with static; done.

  intros E p' Helimcon HeqEp'.
  invert Helimcon; clear Helimcon; intros; subst E; try (discriminate HeqEp'); [].
  subst p'.
  exists kd_type.
  split; eauto with alg static; done.

(* prod *)
intros G Hpath Hwf.
split.
  exists kd_type.
  split; auto with static alg; [].
  simpl.
  auto with static; done.

  intros E p' Helimcon HeqEp'.
  invert Helimcon; clear Helimcon; intros; subst E; try (discriminate HeqEp'); [].
  subst p'.
  exists kd_type.
  split; eauto with alg static; done.

(* forall *)
intros G Hpath Hwf.
split.
  exists kd_type.
  split; auto with static alg; [].
  simpl.
  auto with static; done.

  intros E p' Helimcon HeqEp'.
  invert Helimcon; clear Helimcon; intros; subst E; try (discriminate HeqEp'); [].
  subst p'.
  exists kd_type.
  split; eauto with alg static; done.

(* exists *)
intros G Hpath Hwf.
split.
  exists kd_type.
  split; auto with static alg; [].
  simpl.
  auto with static; done.

  intros E p' Helimcon HeqEp'.
  invert Helimcon; clear Helimcon; intros; subst E; try (discriminate HeqEp'); [].
  subst p'.
  exists kd_type.
  split; eauto with alg static; done.

(* sing *)
intros G p _ IH Hpath Hwf.
so (IH Hpath Hwf) as ((l & Hnatk & Hsubk) & Hofall).
so (Hofall _ p elimcon_null (eq_refl _)) as (e & Hnatk' & Hofp).
so (natk_fun _#4 Hnatk Hnatk') as Heq; subst e; clear Hnatk'.
assert (single p l = kd_sing p) as Heq.
  so (subk_type_invert' _ _ Hwf Hsubk) as [Heq | (b & Heq & _)].
    destruct l; discriminate Heq.

    so (ofk_hkd _ _ (ofc_reg _#3 Hwf Hofp)) as Hhkd.
    #2 invert Hhkd; try (intros; subst; discriminate Heq).
      intros <-.
      simpl; reflexivity.

      intros b' <-.
      simpl; reflexivity.
split.
  exists l.
  split; auto; [].
  rewrite -> Heq.
  apply subk_refl; [].
  apply eqk_refl; [].
  rewrite <- Heq.
  exact (subk_reg _#3 Hwf Hsubk andel).

  exact Hofall.

(* extpi *)
intros G k1 k2 k3 p Hofpalt IH1 _ IH2 Hpath Hwf.
so (ofc_reg _#3 Hwf Hofpalt) as Hofkalt.
invert Hofkalt; intros Hofk1 _.
so (IH1 Hpath Hwf) as ((l & Hnatk & Hsubkalt) & Hofall).
so (subk_reg _#3 Hwf Hsubkalt) as (Hofl & _).
so (subk_pi_invert' _#4 Hsubkalt) as (e1 & e2 & Heq & Hsubk1 & Hsubk3).
clear Hsubkalt.
assert (exists l1 l2, l = kd_pi l1 l2) as (l1 & l2 & ->).
  destruct l; try (discriminate Heq); [].
  eauto; done.
simpl in Heq.
injection Heq; clear Heq; intros; subst e1 e2.
so (IH2 (path_app (shift_path _ _ Hpath)) (wf_cn Hwf Hofk1)) as ((l' & Hnatk' & Hsubk) & _).
split.
  exists (kd_pi l1 l2).
  split; auto; [].
  simpl.
  #2 apply subk_pi; auto.
    so (weaken_natk _#5 (extend_1 _ (cl_cn k1)) Hnatk) as Hnatkw.
    simp_sub in Hnatkw.
    so (natk_fun _#4 (natk_app Hnatkw) Hnatk') as Heq.
    subst l'.
    simp_sub in Hsubk.
    assumption.
  
    simpl in Hofl.
    invert Hofl; auto; done.

  exact Hofall.

(* extsigma *)
intros G k1 k2 p _ IH1 _ IH2 Hofk2 Hpath Hwf.
so (IH1 (path_pi1 Hpath) Hwf) as ((l1 & Hnatk1 & Hsubk1) & Hofall).
so (IH2 (path_pi2 Hpath) Hwf) as ((e & Hnatk2 & Hsubk2) & _).
invert Hnatk1.
intros l2' Hnatk.
invert Hnatk2.
intros l1' l2 Hnatk' <-.
so (natk_fun _#4 Hnatk Hnatk') as Heq.
injection Heq; clear Heq; intros; subst l1' l2'.
clear Hnatk'.
so (Hofall _ _ (elimcon_pi1 elimcon_null) (eq_refl _)) as (l & Hnatk' & Hofp).
so (natk_fun _#4 Hnatk Hnatk') as Heq.
subst l.
clear Hnatk'.
split.
  exists (kd_sigma l1 l2).
  split; auto; [].
  simpl.
  apply subk_sigma; auto; [].
  so (ofk_hkd _ _ (ofc_reg _#3 Hwf Hofp)) as Hhkd.
  invert Hhkd; intros Hhkd1  Hhkd2.
  so (ofk_hkd _ _ Hofk2) as Hhkd2'.
  rewrite -> subst_single; auto; [].
  simp_sub.
  #2 apply (@subk_trans _ _ (subst (dot (cn_pi1 (shift1 p)) sh1) k2)).
    so (weaken1_subk _ (cl_cn (single (cn_pi1 p) l1)) _ _ Hsubk2) as Hsubk2w.
    unfold shift1 in Hsubk2w.
    rewrite -> !subst_single in Hsubk2w; eauto using subst_hkd; [].
    simp_sub in Hsubk2w.
    assumption.

    apply subk_refl; [].
    apply eqk_symm; [].
    #3 exploit (subof_under_sof (G; cl_cn (single (cn_pi1 p) l1)) G (nil; cl_cn k1) sh1) as Hsubof; auto.
      apply Forall_cons; auto with static; done.

      apply subof_sh; auto with sequence; done.
    simpl in Hsubof.
    so (subst_ofk _#4 Hsubof Hofk2) as Hofk2w.
    simp_sub in Hofk2w.
    so (single_eqc _#3 Hwf (ofc_pi1 Hofp)) as Heqc.
    unfold shift1 in Heqc; rewrite -> subst_single in Heqc; auto; [].
    simp_sub in Heqc.
    so (weaken1_subk _ (cl_cn (single (cn_pi1 p) l1)) _ _ Hsubk1) as Hsubk1w.
    unfold shift1 in Hsubk1w; rewrite -> subst_single in Hsubk1w; auto; [].
    simp_sub in Hsubk1w.
    #2 sor (refine (funct_kd _#6 (eqc_subsume Heqc Hsubk1w) Hofk2w)) as H.
      apply wf_cn; auto; [].
      apply single_ofk; eauto with static; done.
    simp_sub in H.
    assumption.

  intros E q Helimcon Heqp.
  refine (Hofall _ _ (elimcon_pi1 Helimcon) _); [].
  apply f_equal; auto; done.

(* subsume *)
intros G k k' p _ IH Hsubk Hpath Hwf.
so (IH Hpath Hwf) as ((l & Hnatk & Hsubk') & Hofall).
split.
  exists l.
  eauto with static; done.

  exact Hofall.
Qed.


Lemma natk_sound :
  forall G p k c,
    wf G
    -> ofc G p k
    -> natk G p (kd_sing c)
    -> eqc G p c k.
Proof.
intros G p k c Hwf Hofc Hnatk.
so (natk_complete _#3 (natk_path _#3 Hnatk) Hwf Hofc) as (l & Hnatk' & Hofp' & Hsubk).
so (natk_fun _#4 Hnatk Hnatk'); subst l.
simpl in Hsubk.
eapply eqc_subsume; eauto; [].
apply eqc_sing; [].
apply eqc_singelim; auto; done.
Qed.


Inductive prink : context -> (* cn *) term -> (* kd *) term -> Prop :=
| prink_var {G i k}
    : index i G (cl_cn k)
      -> prink G (var i) (single (var i) (subst (sh (S i)) k))

| prink_lam {G k1 k2 c}
    : ofk G k1
      -> prink (G; cl_cn k1) c k2
      -> prink G (cn_lam k1 c) (kd_pi k1 k2)

| prink_app {G k1 k2 c1 c2}
    : prink G c1 (kd_pi k1 k2)
      -> ofc G c2 k1
      -> prink G (cn_app c1 c2) (subst1 c2 k2)

| prink_pair {G k1 k2 c1 c2}
    : prink G c1 k1
      -> prink G c2 k2
      -> prink G (cn_pair c1 c2) (kd_sigma k1 (shift1 k2))

| prink_pi1 {G k1 k2 c}
    : prink G c (kd_sigma k1 k2)
      -> prink G (cn_pi1 c) k1

| prink_pi2 {G k1 k2 c}
    : prink G c (kd_sigma k1 k2)
      -> prink G (cn_pi2 c) (subst1 (cn_pi1 c) k2)

| prink_star {G}
    : prink G cn_star kd_one

| prink_unit {G}
    : prink G cn_unit (kd_sing cn_unit)

| prink_arrow {G c1 c2}
    : ofc G c1 kd_type
      -> ofc G c2 kd_type
      -> prink G (cn_arrow c1 c2) (kd_sing (cn_arrow c1 c2))

| prink_prod {G c1 c2}
    : ofc G c1 kd_type
      -> ofc G c2 kd_type
      -> prink G (cn_prod c1 c2) (kd_sing (cn_prod c1 c2))

| prink_all {G k c}
    : ofk G k
      -> ofc (G; cl_cn k) c kd_type
      -> prink G (cn_all k c) (kd_sing (cn_all k c))

| prink_exists {G k c}
    : ofk G k
      -> ofc (G; cl_cn k) c kd_type
      -> prink G (cn_exists k c) (kd_sing (cn_exists k c))
.


(* no dependent sigmas *)
Inductive nds : (* kd *) term -> Prop :=
| nds_type
    : nds kd_type

| nds_sing {c}
    : nds (kd_sing c)

| nds_pi {k1 k2}
    : nds k2
      -> nds (kd_pi k1 k2)

| nds_sigma {k1 k2}
    : nds k1
      -> nds k2
      -> nds (kd_sigma k1 (shift1 k2))

| nds_one
    : nds kd_one.


Lemma subst_nds :
  forall s k,
    nds k
    -> nds (subst s k).
Proof.
intros s k Hnds.
revert s.
induct Hnds; try (intros; simp_sub; eauto using nds; done); [].

(* sigma *)
intros k1 k2 _ IH1 _ IH2 s.
simp_sub.
rewrite -> subst_compose; [].
apply nds_sigma; auto; done.
Qed.


Lemma single_nds :
  forall c k, nds (single c k).
Proof.
intros c k.
revert c.
induct k; try (intros; simpl; eauto using nds; done); [].
(* sigma *)
intros k1 IH1 k2 IH2 c.
simpl.
repl (dot (cn_pi1 (shift1 c)) sh1) with (compose (dot (cn_pi1 c) id) sh1) by (simp_sub; reflexivity).
rewrite -> subst_compose; [].
apply nds_sigma; auto using subst_nds; done.
Qed.


Lemma prink_nds :
  forall G c k,
    prink G c k
    -> nds k.
Proof.
intros G c k Hprink.
#4 induct Hprink; try (intros; eauto using nds; done).

(* var *)
intros.
apply single_nds; done.

(* app *)
intros G k1 k2 c1 c2 _ IH _.
apply subst_nds; [].
invert IH; auto; done.

(* pi1 *)
intros G k1 k2 c _ IH.
invert IH; auto; done.

(* pi2 *)
intros G k1 k2 c _ IH.
apply subst_nds; [].
invert IH.
intros.
subst k2.
apply subst_nds; auto; done.
Qed.


Lemma prink_fun :
  forall G c k k',
    prink G c k
    -> prink G c k'
    -> k = k'.
Proof.
intros G c k k' Hprink Hprink'.
revert k' Hprink'.
induct Hprink.

(* var *)
intros G i k Hindex k' Hprink'.
invert Hprink'; [].
intros k'' Hindex' <-.
so (index_fun _#5 Hindex Hindex') as Heq.
injection Heq.
intros; subst k''.
reflexivity.

(* lam *)
intros G k1 k2 c _ _ IH k' Hprink'.
invert Hprink'; [].
intros k2' _ Hprink'' <-.
f_equal; auto; done.

(* app *)
intros G k1 k2 c1 c2 _ IH _ k' Hprink'.
invert Hprink'; [].
intros k1' k2' Hprink'' _ <-.
f_equal; auto; [].
so (IH _ Hprink'') as Heq.
injection Heq; auto; done.

(* pair *)
intros G k1 k2 c1 c2 _ IH1 _ IH2 k' Hprink'.
invert Hprink'; [].
intros k1' k2' Hprink1 Hprink2 <-.
f_equal; auto; [].
f_equal; auto; done.

(* pi1 *)
intros G k1 k2 c _ IH k' Hprink'.
invert Hprink'; [].
intros k2' Hprink''.
so (IH _ Hprink'') as Heq.
injection Heq; auto; done.

(* pi2 *)
intros G k1 k2 c _ IH k' Hprink'.
invert Hprink'; [].
intros k1' k2' Hprink'' <-.
f_equal; [].
so (IH _ Hprink'') as Heq.
injection Heq; auto; done.

(* star *)
intros G k' Hprink'.
invert Hprink'; auto; done.

(* unit *)
intros G k' Hprink'.
invert Hprink'; auto; done.

(* arrow *)
intros G c1 c2 H H0 k' Hprink'.
invert Hprink'; auto; done.

(* prod *)
intros G c1 c2 H H0 k' Hprink'.
invert Hprink'; auto; done.

(* all *)
intros G c1 c2 H H0 k' Hprink'.
invert Hprink'; auto; done.

(* exists *)
intros G c1 c2 H H0 k' Hprink'.
invert Hprink'; auto; done.
Qed.


Lemma weaken_prink_gen :
  forall i j G1 G2 G3 c k,
    length G2 = i
    -> length G3 = j
    -> Forall iscl G3
    -> prink (G1;; G3) c k
    -> prink (G1;; G2;; subst_ctx (sh i) G3) (subst (under j (sh i)) c) (subst (under j (sh i)) k).
Proof.
intros i j G1 G2 G3 c k Hlen2 Hlen3 Hallcl Hprink.
remember (G1;; G3) as G eqn:HeqG.
revert j G3 Hallcl Hlen3 HeqG.
induct Hprink.

(* var *)
intros G n k Hindex j G3 Hallcl Hlen3 ->.
simp_sub.
so (Compare_dec.lt_dec n j) as [Hlt | Hnlt].
  rewrite -> project_under_lt; auto; [].
  so (index_app_lt _#5 (eqconv (eqsymm Hlen3) Hlt) Hindex) as Hindex'.
  so (index_subst_ctx _#3 (sh i) Hindex') as Hindex''.
  rewrite -> Hlen3 in Hindex''.
  so (index_app_left _ _ (G1;; G2) _ _ Hindex'') as Hindex'''.
  simp_sub in Hindex'''.
  so (prink_var Hindex''') as Hprink.
  simp_sub in Hprink.
  rewrite -> shift_under_single; [].
  simp_sub.
  rewrite -> compose_sh_under_leq; [| omega].
  repl (j - (1 + n)) with (j - n - 1) by omega.
  rewrite -> project_under_lt; [| omega].
  assumption.

  assert (n >= j) as Hgeq by omega; clear Hnlt.
  rewrite -> project_under_ge; auto; [].
  simp_sub.
  repl (j + i + (n - j)) with (n + i) by omega.
  so (index_app_ge _#5 (eqconv (eqsymm Hlen3) Hgeq) Hindex) as Hindex'.
  so (index_app_right _ (G2;; subst_ctx (sh i) G3) _#3 Hindex') as Hindex''.
  rewrite -> app_length in Hindex''; [].
  rewrite -> subst_ctx_length in Hindex''; [].
  rewrite -> Hlen2 in Hindex''; [].
  rewrite -> Hlen3 in Hindex''; [].
  rewrite <- app_assoc in Hindex''; [].
  repl (n - j + (j + i)) with (n + i) in Hindex'' by omega.
  so (prink_var Hindex'') as Hprink.
  rewrite -> shift_under_single; [].
  simp_sub.
  rewrite -> compose_sh_under_geq; auto; [].
  simp_sub.
  repl (1 + n - j + i + j) with (S (n + i)) by omega.
  rewrite -> project_under_ge; [| omega].
  simp_sub.
  repl (j + i + (n - j)) with (n + i) by omega.
  assumption.

(* lam *)
intros G k1 k2 c Hofk1 _ IH j G3 Hallcl Hlen3 ->.
simp_sub.
#2 apply prink_lam.
  eapply subst_ofk; eauto; [].
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

  rewrite <- under_succ; [].
  rewrite -> app_comm_cons; [].
  repl (subst_ctx (sh i) G3; cl_cn (subst (under j (sh i)) k1)) with (subst_ctx (sh i) (G3; cl_cn k1)) by (simpl; simp_sub; subst; reflexivity).
  #2 apply IH; auto.
    apply Forall_cons; auto with static; done.

    simpl; auto; done.

(* app *)
intros G k1 k2 c1 c2 _ IH Hofc2 j G3 Hallcl Hlen3 ->.
simp_sub.
so (IH _ _ Hallcl Hlen3 (eq_refl _)) as Hprink.
simp_sub in Hprink.
#2 sor (refine (@prink_app _#4 (subst (under j (sh i)) c2) Hprink _)) as H.
  eapply subst_ofc; eauto; [].
  subst j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  subst i.
  apply app_truncate; done.

  simp_sub in H.
  assumption.

(* pair *)
intros G k1 k2 c1 c2 _ IH1 _ IH2 j G3 Hallcl Hlen3 ->.
simp_sub.
so (IH1 _ _ Hallcl Hlen3 (eq_refl _)) as Hprink1.
so (IH2 _ _ Hallcl Hlen3 (eq_refl _)) as Hprink2.
so (prink_pair Hprink1 Hprink2).
simp_sub in H.
assumption.

(* pi1 *)
intros G k1 k2 c _ IH j G3 Hallcl Hlen3 ->.
so (IH _ _ Hallcl Hlen3 (eq_refl _)) as Hprink.
simp_sub in Hprink.
exact (prink_pi1 Hprink).

(* pi2 *)
intros G k1 k2 c _ IH j G3 Hallcl Hlen3 ->.
so (IH _ _ Hallcl Hlen3 (eq_refl _)) as Hprink.
simp_sub in Hprink.
so (prink_pi2 Hprink) as Hprink'.
simp_sub in Hprink'.
simp_sub.
assumption.

(* star *)
intros G j G3 Hallcl Hlen3 ->.
simp_sub.
exact prink_star.

(* unit *)
intros G j G3 Hallcl Hlen3 ->.
simp_sub.
exact prink_unit.

(* arrow *)
intros G c1 c2 Hof1 Hof2 j G3 Hallcl Hlen3 ->.
simp_sub.
assert (subof sof (G1;; G2;; subst_ctx (sh i) G3) (under j (sh i)) (G1;; G3)) as Hsubof.
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.
apply prink_arrow; replace kd_type with (subst (under j (sh i)) kd_type) by (simp_sub; reflexivity); eapply subst_ofc; eauto; done.

(* prod *)
intros G c1 c2 Hof1 Hof2 j G3 Hallcl Hlen3 ->.
simp_sub.
assert (subof sof (G1;; G2;; subst_ctx (sh i) G3) (under j (sh i)) (G1;; G3)) as Hsubof.
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.
apply prink_prod; replace kd_type with (subst (under j (sh i)) kd_type) by (simp_sub; reflexivity); eapply subst_ofc; eauto; done.

(* all *)
intros G k c Hofk Hofc j G3 Hallcl Hlen3 ->.
simp_sub.
assert (subof sof (G1;; G2;; subst_ctx (sh i) G3) (under j (sh i)) (G1;; G3)) as Hsubof.
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.
#2 apply prink_all.
  eapply subst_ofk; eauto; done.

  replace kd_type with (subst (dot (var 0) (compose (under j (sh i)) sh1)) kd_type) by (simp_sub; reflexivity); [].
  eapply subst_ofc; eauto; [].
  apply subof_bind_sof_cn; auto; done.
  
(* exists *)
intros G k c Hofk Hofc j G3 Hallcl Hlen3 ->.
simp_sub.
assert (subof sof (G1;; G2;; subst_ctx (sh i) G3) (under j (sh i)) (G1;; G3)) as Hsubof.
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.
#2 apply prink_exists.
  eapply subst_ofk; eauto; done.

  replace kd_type with (subst (dot (var 0) (compose (under j (sh i)) sh1)) kd_type) by (simp_sub; reflexivity); [].
  eapply subst_ofc; eauto; [].
  apply subof_bind_sof_cn; auto; done.
Qed.


Lemma weaken1_prink :
  forall G c k A,
    prink G c k
    -> prink (G; A) (shift1 c) (shift1 k).
Proof.
intros G c k A Hprink.
exploit (weaken_prink_gen 1 0 G (nil; A) nil c k); auto; done.
Qed.


Lemma prink_sound :
  forall G c k,
    wf G
    -> prink G c k
    -> ofc G c k.
Proof.
intros G c k Hwf Hprink.
revert Hwf.
#2 induct Hprink; eauto with static.

(* var *)
intros G i k Hindex Hwf.
apply single_ofc; auto with static; done.

(* pair *)
intros G k1 k2 c1 c2 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as Hofc1.
so (IH2 Hwf) as Hofc2.
#2 apply ofc_pair; auto.
  simp_sub.
  assumption.

  apply weaken1_ofk; [].
  exact (ofc_reg _#3 Hwf Hofc2).
Qed.


Lemma prink_complete_strong :
  forall G c k,
    wf G
    -> ofc G c k
    -> exists k',
         prink G c k'
         /\ subk G k' (single c k).
Proof.
intros G c k Hwf Hofc.
revert Hwf.
induct Hofc.

(* var *)
intros G i k Hindex Hwf.
exists (single (var i) (subst (sh (S i)) k)).
split.
  apply prink_var; auto; done.

  apply subk_refl; [].
  apply eqk_refl; [].
  apply single_ofk; auto with static; done.

(* lam *)
intros G k1 k2 c Hofk Hofc IH Hwf.
so (IH (wf_cn Hwf Hofk)) as (l2 & Hprink & Hsubk); clear IH.
exists (kd_pi k1 l2).
split.
  apply prink_lam; auto; done.

  simpl.
  #2 apply subk_pi; auto with static.
    eapply subk_trans; eauto; [].
    apply subk_refl; [].
    apply single_funct; auto with static; [].
    apply eqc_symm; [].
    simp_sub.
    #3 exploit (@eqc_beta (G; cl_cn k1) (shift1 k1) (subst (dot (var 0) (sh 2)) k2) (var 0) (subst (dot (var 0) (sh 2)) c)) as H.
      #2 apply (subst_ofc _ (G; cl_cn k1)).
        apply subof_dot; eauto with static sequence; [].
        simp_sub.
        apply sof_cn; [].
        repl (subst (sh 2) k1) with (subst (sh 1) (subst (sh 1) k1)) by (simp_sub; reflexivity).
        apply ofc_var; [].
        apply index_0; done.
        
        #2 apply (@ofc_subsume _ l2).
          apply prink_sound; auto with static; done.

          eapply subk_trans; eauto; [].
          apply single_subk; auto with static; done.

      apply ofc_var; auto using index_0; done.

      simp_sub in H.
      assumption.

    exact (subk_reg _#3 (wf_cn Hwf Hofk) Hsubk andel).

(* app *)
intros G k1 k2 c1 c2 Hofc1 IH Hofc2 _ Hwf.
so (IH Hwf) as (l & Hprink & Hsubk); clear IH.
simpl in Hsubk.
so (subk_pi_invert' _#4 Hsubk) as (l1 & l2 & -> & Hsubk1 & Hsubk2).
exists (subst1 c2 l2).
split.
  eapply prink_app; eauto with static; done.

  so (subst1_subk _#5 Hofc2 Hsubk2) as Hsubk2x.
  so (ofk_hkd _ _ (ofc_reg _#3 Hwf Hofc1)) as Hhkd.
  invert Hhkd.
  intros _ Hhkd2.
  unfold subst1 in Hsubk2x; rewrite -> subst_single in Hsubk2x; auto; [].
  simp_sub in Hsubk2x.
  assumption.

(* pair *)
intros G k1 k2 c1 c2 Hofc1 IH1 Hofc2 IH2 Hofk2 Hwf.
so (IH1 Hwf) as (l1 & Hprink1 & Hsubk1); clear IH1.
so (IH2 Hwf) as (l2 & Hprink2 & Hsubk2); clear IH2.
exists (kd_sigma l1 (shift1 l2)).
split.
  apply prink_pair; auto; done.

  simpl.
  exploit (@ofc_pair G k1 (shift1 (subst1 c1 k2)) c1 c2) as Hofc.
    assumption.

    simp_sub.
    assumption.

    apply weaken1_ofk; [].
    exact (ofc_reg _#3 Hwf Hofc2).
  #3 apply subk_sigma.
    eapply subk_trans; eauto; [].
    apply subk_refl; [].
    apply single_funct; auto; [].
    apply eqc_symm; [].
    eapply eqc_beta1; eauto; done.

    so (weaken1_subk _ (cl_cn l1) _ _ Hsubk2) as Hsubk2w.
    eapply subk_trans; eauto; [].
    unfold shift1; rewrite -> !subst_single; eauto using subst_hkd, ofk_hkd; [].
    simp_sub.
    apply subk_refl; [].
    so (subk_reg _#3 Hwf Hsubk1) as (Hofl1 & _).
    #2 apply single_resp; auto with static.
      apply eqc_symm; [].
      #2 apply (@eqc_beta2 _ (shift1 k1)).
        apply weaken1_ofc; auto; done.

        repl (subst (dot (shift1 c1) sh1) k2) with (shift1 (subst1 c1 k2)) by (simp_sub; reflexivity).
        apply weaken1_ofc; auto; done.
      
      cut (eqk G (subst1 c1 k2) (subst1 (cn_pi1 (cn_pair c1 c2)) k2)).
        intro Heqk.
        so (weaken1_eqk _ (cl_cn l1) _ _ Heqk) as Heqkw.
        simp_sub in Heqkw.
        assumption.
      eapply funct_kd; eauto; [].
      apply eqc_symm; [].
      eapply eqc_beta1; eauto; done.

    simp_sub.
    rewrite -> subst_single; eauto using subst_hkd, ofk_hkd; [].
    simp_sub.
    #2 exploit (single_ofk G (cn_pi2 (cn_pair c1 c2)) (subst1 (cn_pi1 (cn_pair c1 c2)) k2)) as Hofk; auto.
      so (ofc_pi2 Hofc) as Hofpi2.
      simp_sub in Hofpi2.
      eapply ofc_subsume; eauto; [].
      apply subk_refl; [].
      apply eqk_symm; [].
      eapply funct_kd; eauto; [].
      eapply eqc_beta1; eauto; done.

      so (weaken1_ofk _ (cl_cn (single (cn_pi1 (cn_pair c1 c2)) k1)) _ Hofk) as Hofkw.
      simp_sub in Hofkw.
      unfold shift1 in Hofkw; rewrite -> subst_single in Hofkw; eauto using subst_hkd, ofk_hkd; [].
      simp_sub in Hofkw.
      assumption.

(* pi1 *)
intros G k1 k2 c Hofc IH Hwf.
so (IH Hwf) as (l & Hprink & Hsubk); clear IH.
simpl in Hsubk.
so (subk_sigma_invert' _#4 Hsubk) as (l1 & l2 & -> & Hsubk1 & Hsubk2).
exists l1.
split.
  eapply prink_pi1; eauto; done.

  assumption.

(* pi2 *)
intros G k1 k2 c Hofc IH Hwf.
so (IH Hwf) as (l & Hprink & Hsubk); clear IH.
simpl in Hsubk.
so (subk_sigma_invert' _#4 Hsubk) as (l1 & l2 & -> & Hsubk1 & Hsubk2).
exists (subst1 (cn_pi1 c) l2).
split.
  eapply prink_pi2; eauto; done.

  so (ofk_hkd _ _ (ofc_reg _#3 Hwf Hofc)) as Hhkd.
  invert Hhkd; clear Hhkd.
  intros Hhkd1 Hhkd2.
  rewrite -> subst_single in Hsubk2; auto; [].
  simp_sub in Hsubk2.
  so (prink_sound _#3 Hwf Hprink) as Hofc'.
  so (subst1_subk _#5 (ofc_pi1 Hofc') Hsubk2) as Hsubk2x.
  unfold subst1 in Hsubk2x; rewrite -> subst_single in Hsubk2x; auto using subst_hkd; [].
  simp_sub in Hsubk2x.
  assumption.

(* star *)
intros G Hwf.
exists kd_one.
split.
  apply prink_star; done.

  simpl.
  auto with static; done.

(* unit *)
intros G Hwf.
eexists.
split.
  apply prink_unit; done.

  eauto with static; done.

(* arrow *)
intros G t1 t2 Hof1 _ Hof2 _ Hwf.
eexists.
split.
  apply prink_arrow; auto; done.

  simpl.
  eauto with static; done.

(* prod *)
intros G t1 t2 Hof1 _ Hof2 _ Hwf.
eexists.
split.
  apply prink_prod; auto; done.

  simpl.
  eauto with static; done.

(* all *)
intros G k t Hofk Hofc _ Hwf.  
eexists.
split.
  apply prink_all; auto; done.

  simpl.
  eauto with static; done.

(* exists *)
intros G k t Hofk Hofc _ Hwf.  
eexists.
split.
  apply prink_exists; auto; done.

  simpl.
  eauto with static; done.

(* sing *)
intros G c Hofc IH Hwf.
so (IH Hwf) as (l & Hprink & Hsubk).
exists l.
split; auto; done.

(* extpi *)
intros G k1 k2 k3 c Hof1 IH1 Hof2 IH2 Hwf.
so (ofc_reg _#3 Hwf Hof1) as Hofk.
invert Hofk; clear Hofk.
intros Hofk1 _.
so (IH1 Hwf) as (l & Hprink & Hsubkl).
simpl in Hsubkl.
so (subk_pi_invert' _#4 Hsubkl) as (l1 & l2 & -> & Hsubk1 & _).
exists (kd_pi l1 l2).
split; auto; [].
simpl.
#2 apply subk_pi; auto.
  so (IH2 (wf_cn Hwf Hofk1)) as (e & Hprinkapp & Hsubk2); clear IH2.
  invert Hprinkapp.
  intros e1 e2 Hprinkw Hofcvar <-.
  so (weaken1_prink _#3 (cl_cn k1) Hprink) as Hprinkw'.
  so (prink_fun _#4 Hprinkw Hprinkw') as Heq.
  simp_sub in Heq.
  invert Heq; clear Heq; intros; subst e1 e2.
  clear Hprinkw Hprinkw' Hprinkapp.
  simp_sub in Hsubk2.
  assumption.

  so (ofc_reg _#3 Hwf (prink_sound _#3 Hwf Hprink)) as Hofl.
  invert Hofl.
  intros Hofl1 Hofl2.
  assumption.

(* extsigma *)
intros G k1 k2 c Hof1 IH1 Hof2 IH2 Hofk2 Hwf.
so (IH1 Hwf) as (l1 & Hprink1 & Hsubk1).
invert Hprink1; clear Hprink1.
intros e Hprink.
invert (prink_nds _#3 Hprink).
intros l2 _ _ <-.
so (IH2 Hwf) as (e & Hprink2 & Hsubk2).
invert Hprink2; clear Hprink2.
intros l1' l2' Hprink' <-.
so (prink_fun _#4 Hprink Hprink') as Heq.
injection Heq; clear Heq.
intros; subst l1' l2'.
clear Hprink'.
exists (kd_sigma l1 (shift1 l2)).
split; auto; [].
simpl.
#2 apply subk_sigma; auto.
  rewrite -> subst_single; eauto using subst_hkd, ofk_hkd; [].
  simp_sub.
  simp_sub in Hsubk2.
  so (weaken1_subk _ (cl_cn l1) _ _ Hsubk2) as Hsubk2w.
  unfold shift1 in Hsubk2w; rewrite -> !subst_single in Hsubk2w; eauto using subst_hkd, ofk_hkd; [].
  simp_sub in Hsubk2w.
  assumption.

  rewrite -> subst_single; eauto using subst_hkd, ofk_hkd; [].
  simp_sub.
  apply single_ofk; auto using single_ofk with static; [].
  so (weaken1_ofc _ (cl_cn (single (cn_pi1 c) k1)) _ _ Hof2) as H.
  simp_sub in H.
  assumption.

(* subsume *)
intros G k k' c Hof IH Hsubk Hwf.
so (IH Hwf) as (l & Hprink & Hsubk').
eexists; split; eauto; [].
eapply subk_trans; eauto; [].
apply single_mono; auto; done.
Qed.


Lemma prink_complete :
  forall G c k,
    wf G
    -> ofc G c k
    -> exists k',
         prink G c k'
         /\ subk G k' k.
Proof.
intros G c k Hwf Hofc.
so (prink_complete_strong _#3 Hwf Hofc) as (k' & Hprink & Hsubk).
exists k'.
split; auto; [].
eapply subk_trans; eauto; [].
apply single_subk; auto; done.
Qed.


Inductive reduce : context -> (* cn *) term -> (* cn *) term -> Prop :=
| reduce_app1 {G c1 c1' c2}
    : reduce G c1 c1'
      -> reduce G (cn_app c1 c2) (cn_app c1' c2)

| reduce_app2 {G c1 c2 c2'}
    : reduce G c2 c2'
      -> reduce G (cn_app c1 c2) (cn_app c1 c2')

| reduce_beta {G k1 c1 c2}
    : reduce G (cn_app (cn_lam k1 c1) c2) (subst1 c2 c1)

| reduce_pair1 {G c1 c1' c2}
    : reduce G c1 c1'
      -> reduce G (cn_pair c1 c2) (cn_pair c1' c2)

| reduce_pair2 {G c1 c2 c2'}
    : reduce G c2 c2'
      -> reduce G (cn_pair c1 c2) (cn_pair c1 c2')

| reduce_pi1 {G c c'}
    : reduce G c c'
      -> reduce G (cn_pi1 c) (cn_pi1 c')

| reduce_beta1 {G c1 c2}
    : reduce G (cn_pi1 (cn_pair c1 c2)) c1

| reduce_pi2 {G c c'}
    : reduce G c c'
      -> reduce G (cn_pi2 c) (cn_pi2 c')

| reduce_beta2 {G c1 c2}
    : reduce G (cn_pi2 (cn_pair c1 c2)) c2

| reduce_sing {G c c'}
    : natk G c (kd_sing c')
      -> reduce G c c'
.


Lemma weaken_reduce_gen :
  forall i j G1 G2 G3 c c',
    length G2 = i
    -> length G3 = j
    -> reduce (G1;; G3) c c'
    -> reduce (G1;; G2;; subst_ctx (sh i) G3) (subst (under j (sh i)) c) (subst (under j (sh i)) c').
Proof.
intros i j G1 G2 G3 c c' Hlen2 Hlen3 Hred.
remember (G1;; G3) as G eqn:HeqG.
revert HeqG.
#2 induct Hred; try (simp_sub; intros; subst; eauto using reduce; done).

(* beta *)
intros G k1 c1 c2 ->.
simp_sub.
so (@reduce_beta
      (G1;; G2;; subst_ctx (sh i) G3)
      (subst (under j (sh i)) k1)
      (subst (dot (var 0) (compose (under j (sh i)) sh1)) c1)
      (subst (under j (sh i)) c2)) as Hred.
simp_sub in Hred.
assumption.

(* sing *)
intros G c c' Hnatk ->.
apply reduce_sing; [].
so (weaken_natk_gen _#7 Hlen2 Hlen3 Hnatk) as Hnatk'.
simp_sub in Hnatk'.
assumption.
Qed.


Lemma weaken_reduce :
  forall i G1 G2 c c',
    extend i G1 G2
    -> reduce G1 c c'
    -> reduce G2 (subst (sh i) c) (subst (sh i) c').
Proof.
intros i G1 G2 c c' Hext Hred.
destruct Hext as (G & -> & Hlen).
exploit (weaken_reduce_gen i 0 G1 G nil c c'); auto; done.
Qed.


Lemma subject_reduction :
  forall G c c' k,
    wf G
    -> ofc G c k
    -> reduce G c c'
    -> eqc G c c' k.
Proof.
intros G c c' k Hwf Hofc Hred.
revert Hwf c' Hred.
induct Hofc.

(* var *)
intros G i k Hindex Hwf c' Hred.
invert Hred; [].
intro Hnatk.
exact (natk_sound _#4 Hwf (ofc_var Hindex) Hnatk).

(* lam *)
intros G k1 k2 c H H0 H1 Hwf c' Hred.
invert Hred; [].
intro Hnatk.
invert Hnatk; done.

(* app *)
intros G k1 k2 c1 c2 Hofc1 IH1 Hofc2 IH2 Hwf c' Hred.
#4 invert Hred.
  intros c1' Hred' <-.
  eapply eqc_app; eauto with static; done.

  intros c2' Hred' <-.
  eapply eqc_app; eauto with static; done.

  intros k d <- <-.
  clear IH1 IH2.
  so (prink_complete _#3 Hwf (ofc_app Hofc1 Hofc2)) as (l & Hprink & Hsubk).
  invert Hprink; clear Hprink.
  intros l1 l2 Hprink' Hofc2' <-.
  invert Hprink'; clear Hprink'.
  intros Hofk Hprink'' <-.
  so (prink_sound _#3 (wf_cn Hwf Hofk) Hprink'') as Hofd; clear Hprink''.
  eapply eqc_subsume; eauto; [].
  apply eqc_beta; auto; done.
  
  intro Hnatk.
  exact (natk_sound _#4 Hwf (ofc_app Hofc1 Hofc2) Hnatk).

(* pair *)
intros G k1 k2 c1 c2 H H0 H1 H2 H3 Hwf c' Hred.
#3 invert Hred.
  intros c1' Hred' <-.
  eapply eqc_pair; eauto with static; done.

  intros c2' Hred' <-.
  eapply eqc_pair; eauto with static; done.

  intro Hnatk.
  invert Hnatk; done.

(* pi1 *)
intros G k1 k2 b Hofc IH Hwf c' Hred.
#3 invert Hred.
  intros b' Hred' <-.
  eapply eqc_pi1; eauto with static; done.

  intros c2 <-.
  rename c' into c1.
  clear IH.
  so (prink_complete _#3 Hwf (ofc_pi1 Hofc)) as (l1 & Hprink & Hsubk).
  invert Hprink; clear Hprink.
  intros e Hprink'.
  invert Hprink'; clear Hprink'.
  intros l2 Hprink1 Hprink2 <-.
  eapply eqc_beta1; eauto using prink_sound with static; done.

  intro Hnatk.
  exact (natk_sound _#4 Hwf (ofc_pi1 Hofc) Hnatk).

(* pi2 *)
intros G k1 k2 b Hofc IH Hwf c' Hred.
#3 invert Hred.
  intros b' Hred' <-.
  eapply eqc_pi2; eauto with static; done.

  intros c1 <-.
  rename c' into c2.
  clear IH.
  so (prink_complete _#3 Hwf (ofc_pi2 Hofc)) as (l & Hprink & Hsubk).
  invert Hprink; clear Hprink.
  intros l1 e Hprink' <-.
  invert Hprink'; clear Hprink'.
  intros l2 Hprink1 Hprink2 <-.
  simp_sub in Hsubk.
  eapply eqc_beta2; eauto using prink_sound with static; done.

  intro Hnatk.
  exact (natk_sound _#4 Hwf (ofc_pi2 Hofc) Hnatk).

(* star *)
intros G Hwf c' Hred.
invert Hred; [].
intro Hnatk.
invert Hnatk; done.

(* unit *)
intros G Hwf c' Hred.
invert Hred; [].
intro Hnatk.
invert Hnatk; done.

(* arrow *)
intros G t1 t2 _ _ _ _ Hwf c' Hred.
invert Hred; [].
intro Hnatk.
invert Hnatk; done.

(* prod *)
intros G t1 t2 _ _ _ _ Hwf c' Hred.
invert Hred; [].
intro Hnatk.
invert Hnatk; done.

(* all *)
intros G k t _ _ _ Hwf c' Hred.
invert Hred; [].
intro Hnatk.
invert Hnatk; done.

(* exists *)
intros G k t _ _ _ Hwf c' Hred.
invert Hred; [].
intro Hnatk.
invert Hnatk; done.

(* sing *)
intros G c _ IH Hwf c' Hred.
apply eqc_sing; [].
apply IH; auto; done.

(* extpi *)
intros G k1 k2 k3 c Hof1 IH1 Hof2 IH2 Hwf c' Hred.
so (IH1 Hwf _ Hred) as Heqc1.
so (ofc_reg _#3 Hwf Hof1) as Hofk.
invert Hofk.
intros Hofk1 _.
#3 sor (refine (IH2 _ (cn_app (shift1 c') (var 0)) _)) as Heqc2.
  auto with static; done.

  apply reduce_app1; [].
  eapply weaken_reduce; eauto using extend_1; done.
eapply eqc_extpiw; eauto; done.

(* extsigma *)
intros G k1 k2 c Hof1 IH1 Hof2 IH2 Hofk2 Hwf c' Hred.
so (IH1 Hwf _ (reduce_pi1 Hred)) as Heqc1.
so (IH2 Hwf _ (reduce_pi2 Hred)) as Heqc2.
apply eqc_extsigma; auto; done.

(* subsume *)
intros G k k' c _ IH Hsubk Hwf c' Hred.
eapply eqc_subsume; auto; done.
Qed.


Lemma pair_ofc_inversion_sigma :
  forall G c1 c2 k1 k2,
    wf G
    -> ofc G (cn_pair c1 c2) (kd_sigma k1 k2)
    -> ofc G c1 k1
       /\ ofc G c2 (subst1 c1 k2).
Proof.
intros G c1 c2 k1 k2 Hwf Hof.
so (prink_complete _#3 Hwf Hof) as (k' & Hprink & Hsubk).
invert Hprink; [].
intros k1' k2' Hprin1 Hprin2 <-.
so (subk_sigma_invert' _#4 Hsubk) as (k1'' & k2'' & Heq & Hsubk1 & Hsubk2).
injection Heq; intros; subst k1'' k2''; clear Heq.
split.
  eapply ofc_subsume; eauto; [].
  apply prink_sound; auto; done.

  so (subst1_subk _#5 (prink_sound _#3 Hwf Hprin1) Hsubk2) as Hsubk2'.
  simp_sub in Hsubk2'.
  eapply ofc_subsume; eauto; [].
  apply prink_sound; auto; done.
Qed.
