
Require Import TraverseTactics.

Inductive term : Set :=
| var : nat -> term

(* kinds *)

| kd_type : term
| kd_sing : (* cn *) term -> term
| kd_pi : term -> (* cn => *) term -> term
| kd_sigma : term -> (* cn => *) term -> term
| kd_one : term

(* constructors *)

| cn_lam : (* kd *) term -> (* cn => *) term -> term
| cn_app : term -> term -> term
| cn_pair : term -> term -> term
| cn_pi1 : term -> term
| cn_pi2 : term -> term
| cn_star : term

| cn_unit : term
| cn_arrow : term -> term -> term
| cn_prod : term -> term -> term
| cn_all : (* kd *) term -> (* cn => *) term -> term
| cn_exists : (* kd *) term -> (* cn => *) term -> term

(* terms *)

| tm_star : term
| tm_lam : (* cn *) term -> (* tm => *) term -> term
| tm_app : term -> term -> term
| tm_pair : term -> term -> term
| tm_pi1 : term -> term
| tm_pi2 : term -> term
| tm_plam : (* kd *) term -> (* cn => *) term -> term
| tm_papp : term -> (* cn *) term -> term
| tm_pack : (* cn *) term -> term -> (* kd *) term -> (* cn => cn *) term -> term
| tm_unpack : term -> (* cn => tm => *) term -> term
| tm_fix : (* cn *) term -> term -> term
| tm_lett : term -> (* tm => *) term -> term
| tm_letm : (* md *) term -> (* cn => md => *) term -> term
| tm_ext : (* md *) term -> term

(* signatures *)
| sg_one : term
| sg_stat : (* kd *) term -> term
| sg_dyn : (* cn *) term -> term
| sg_pigen : term -> (* cn => *) term -> term
| sg_piapp : term -> (* cn => *) term -> term
| sg_sigma : term -> (* cn => *) term -> term

(* modules *)
| md_star : term
| md_stat : (* cn *) term -> term
| md_dyn : (* tm *) term -> term
| md_lamgn : (* sg *) term -> (* cn => md => *) term -> term
| md_appgn : term -> term -> term
| md_lamap : (* sg *) term -> (* cn => md => *) term -> term
| md_appap : term -> term -> term
| md_pair : term -> term -> term
| md_pi1 : term -> term
| md_pi2 : term -> term
| md_unpack : term -> (* cn => tm => *) term -> (* sg *) term -> term
| md_lett : (* tm *) term -> (* tm => *) term -> term
| md_letm : term -> (* cn => md => *) term -> (* sg *) term -> term
| md_seal : term -> (* sg *) term -> term

(* classifiers *)

| cl_cn : (* kd *) term -> term
| cl_tm : (* cn *) term -> term
| cl_md : (* sg *) term -> (* cn *) term -> term
| cl_ucn : term
| cl_utm : term
| cl_umd : (* cn *) term -> term

(* nonsense *)

| nonsense : term

(* untyped lambda terms *)

| lm_lam : (* lm => *) term -> term
| lm_app : term -> term -> term
.


Definition context := list term.


Fixpoint traverse (S:Set) (enter : S -> S) (resolve : S -> nat -> term) (s : S) (t : term) {struct t} : term :=
  (match t with
   | var i => resolve s i

   | kd_type => kd_type
   | kd_sing t1 => kd_sing (traverse S enter resolve s t1)
   | kd_pi t1 t2 => kd_pi (traverse S enter resolve s t1) (traverse S enter resolve (enter s) t2)
   | kd_sigma t1 t2 => kd_sigma (traverse S enter resolve s t1) (traverse S enter resolve (enter s) t2)
   | kd_one => kd_one

   | cn_lam t1 t2 => cn_lam (traverse S enter resolve s t1) (traverse S enter resolve (enter s) t2)
   | cn_app t1 t2 => cn_app (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | cn_pair t1 t2 => cn_pair (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | cn_pi1 t1 => cn_pi1 (traverse S enter resolve s t1)
   | cn_pi2 t1 => cn_pi2 (traverse S enter resolve s t1)
   | cn_star => cn_star
   | cn_unit => cn_unit
   | cn_arrow t1 t2 => cn_arrow (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | cn_prod t1 t2 => cn_prod (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | cn_all t1 t2 => cn_all (traverse S enter resolve s t1) (traverse S enter resolve (enter s) t2)
   | cn_exists t1 t2 => cn_exists (traverse S enter resolve s t1) (traverse S enter resolve (enter s) t2)

   | tm_star => tm_star
   | tm_lam t1 t2 => tm_lam (traverse S enter resolve s t1) (traverse S enter resolve (enter s) t2)
   | tm_app t1 t2 => tm_app (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | tm_pair t1 t2 => tm_pair (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | tm_pi1 t1 => tm_pi1 (traverse S enter resolve s t1)
   | tm_pi2 t1 => tm_pi2 (traverse S enter resolve s t1)
   | tm_plam t1 t2 => tm_plam (traverse S enter resolve s t1) (traverse S enter resolve (enter s) t2)
   | tm_papp t1 t2 => tm_papp (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | tm_pack t1 t2 t3 t4 => tm_pack (traverse S enter resolve s t1) (traverse S enter resolve s t2) (traverse S enter resolve s t3) (traverse S enter resolve (enter s) t4)
   | tm_unpack t1 t2 => tm_unpack (traverse S enter resolve s t1) (traverse S enter resolve (enter (enter s)) t2)
   | tm_fix t1 t2 => tm_fix (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | tm_lett t1 t2 => tm_lett (traverse S enter resolve s t1) (traverse S enter resolve (enter s) t2)
   | tm_letm t1 t2 => tm_letm (traverse S enter resolve s t1) (traverse S enter resolve (enter (enter s)) t2)
   | tm_ext t1 => tm_ext (traverse S enter resolve s t1)

   | sg_one => sg_one
   | sg_stat t1 => sg_stat (traverse S enter resolve s t1)
   | sg_dyn t1 => sg_dyn (traverse S enter resolve s t1)
   | sg_pigen t1 t2 => sg_pigen (traverse S enter resolve s t1) (traverse S enter resolve (enter s) t2)
   | sg_piapp t1 t2 => sg_piapp (traverse S enter resolve s t1) (traverse S enter resolve (enter s) t2)
   | sg_sigma t1 t2 => sg_sigma (traverse S enter resolve s t1) (traverse S enter resolve (enter s) t2)

   | md_star => md_star
   | md_stat t1 => md_stat (traverse S enter resolve s t1)
   | md_dyn t1 => md_dyn (traverse S enter resolve s t1)
   | md_lamgn t1 t2 => md_lamgn (traverse S enter resolve s t1) (traverse S enter resolve (enter (enter s)) t2)
   | md_appgn t1 t2 => md_appgn (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | md_lamap t1 t2 => md_lamap (traverse S enter resolve s t1) (traverse S enter resolve (enter (enter s)) t2)
   | md_appap t1 t2 => md_appap (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | md_pair t1 t2 => md_pair (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | md_pi1 t1 => md_pi1 (traverse S enter resolve s t1)
   | md_pi2 t1 => md_pi2 (traverse S enter resolve s t1)
   | md_unpack t1 t2 t3 => md_unpack (traverse S enter resolve s t1) (traverse S enter resolve (enter (enter s)) t2) (traverse S enter resolve s t3)
   | md_lett t1 t2 => md_lett (traverse S enter resolve s t1) (traverse S enter resolve (enter s) t2)
   | md_letm t1 t2 t3 => md_letm (traverse S enter resolve s t1) (traverse S enter resolve (enter (enter s)) t2) (traverse S enter resolve s t3)
   | md_seal t1 t2 => md_seal (traverse S enter resolve s t1) (traverse S enter resolve s t2)

   | cl_cn t1 => cl_cn (traverse S enter resolve s t1)
   | cl_tm t1 => cl_tm (traverse S enter resolve s t1)
   | cl_md t1 t2 => cl_md (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   | cl_ucn => cl_ucn
   | cl_utm => cl_utm
   | cl_umd t1 => cl_umd (traverse S enter resolve s t1)

   | nonsense => nonsense

   | lm_lam t1 => lm_lam (traverse S enter resolve (enter s) t1)
   | lm_app t1 t2 => lm_app (traverse S enter resolve s t1) (traverse S enter resolve s t2)
   end).


Lemma traverse_var :
  forall S enter resolve s i,
    traverse S enter resolve s (var i) = resolve s i.
Proof.
auto.
Qed.


Lemma traverse_parametric :
  forall (S:Set) (S':Set) (R : S -> S' -> Prop) enter enter' resolve resolve' s s' t,
    (forall s s', R s s' -> R (enter s) (enter' s'))
    -> (forall s s' i, R s s' -> resolve s i = resolve' s' i)
    -> R s s'
    -> traverse S enter resolve s t = traverse S' enter' resolve' s' t.
Proof.
prove_traverse_parametric term_ind.
Qed.


Lemma traverse_id :
  forall (S:Set) (R : S -> Prop) enter resolve s t,
    (forall s, R s -> R (enter s))
    -> (forall s i, R s -> resolve s i = var i)
    -> R s
    -> traverse S enter resolve s t = t.
Proof.
prove_traverse_id term_ind.
Qed.


Lemma traverse_compose :
  forall (S:Set) (S':Set) enter enter' resolve resolve' s s' t,
    traverse S enter resolve s (traverse S' enter' resolve' s' t)
    = traverse (S * S')
      (fun p => let (s, s') := p in (enter s, enter' s'))
      (fun p i => let (s, s') := p in traverse S enter resolve s (resolve' s' i))
      (s, s') t.
Proof.
prove_traverse_compose term_ind.
Qed.


Definition termx := term.
Definition varx := var.
