
Require Import Tactics.
Require Import Sequence.
Require Export TSubstGeneral.
Require Import Static.
Require Import Dynamic.


(* substitutends for static terms *)
Inductive sof : context -> term -> term -> Prop :=
| sof_cn {G c k}
    : ofc G c k
      -> sof G c (cl_cn k)

| sof_tm {G e t}
    : sof G e (cl_tm t)

| sof_md {G M T c}
    : sof G M (cl_md T c)

| sof_ucn {G t}
    : sof G t cl_ucn

| sof_utm {G t}
    : sof G t cl_utm

| sof_umd {G t c}
    : sof G t (cl_umd c)
.


(* substitutends for terms *)
Inductive dof : context -> term -> term -> Prop :=
| dof_cn {G c k}
    : ofc G c k
      -> dof G c (cl_cn k)

| dof_tm {G e t}
    : oft G e t
      -> dof G e (cl_tm t)

| dof_md {G M T c}
    : ofm G pure M T
      -> fstm G M c
      -> dof G M (cl_md T c)

| dof_ucn {G t}
    : dof G t cl_ucn

| dof_utm {G t}
    : dof G t cl_utm

| dof_umd {G M c}
    : fstm G M c
      -> dof G M (cl_umd c)
.


(* substitutends for twinning *)
Inductive tof : context -> term -> term -> Prop :=
| tof_cn {G c k}
    : tof G c (cl_cn k)

| tof_tm {G e t}
    : tof G e (cl_tm t)

| tof_md {G M T c}
    : fstm G M c
      -> tof G M (cl_md T c)

| tof_ucn {G t}
    : tof G t cl_ucn

| tof_utm {G t}
    : tof G t cl_utm

| tof_umd {G M c}
    : fstm G M c
      -> tof G M (cl_umd c)
.



(* substitutends for operational semantics *)
Inductive oof : context -> term -> term -> Prop :=
| oof_cn {G c k}
    : oof G c (cl_cn k)

| oof_tm {G e t}
    : tvalue G e
      -> oof G e (cl_tm t)

| oof_md {G M T c}
    : mvalue G M
      -> fstm G M c
      -> oof G M (cl_md T c)

| oof_ucn {G t}
    : oof G t cl_ucn

| oof_utm {G e}
    : tvalue G e
      -> oof G e cl_utm

| oof_umd {G M c}
    : mvalue G M
      -> fstm G M c
      -> oof G M (cl_umd c)
.


(* substitutends for all *)
Inductive aof : context -> term -> term -> Prop :=
| aof_cn {G c k}
    : ofc G c k
      -> aof G c (cl_cn k)

| aof_tm {G e t}
    : oft G e t
      -> tvalue G e
      -> aof G e (cl_tm t)

| aof_md {G M T c}
    : ofm G pure M T
      -> mvalue G M
      -> fstm G M c
      -> aof G M (cl_md T c)

| aof_ucn {G c}
    : aof G c cl_ucn

| aof_utm {G e}
    : tvalue G e
      -> aof G e cl_utm

| aof_umd {G M c}
    : mvalue G M
      -> fstm G M c
      -> aof G M (cl_umd c)
.



Lemma sof_behaved : 
  behaved sof.
Proof.
intros G i c Hc Hindex.
destruct Hc; simp_sub; auto using sof with static.
Qed.


Lemma dof_behaved : 
  behaved dof.
Proof.
intros G i t Ht Hindex.
destruct Ht; simp_sub; eauto using dof with static.
Qed.


Lemma tof_behaved : 
  behaved tof.
Proof.
intros G i t Ht Hindex.
destruct Ht; simp_sub; eauto using tof with static.
Qed.


Lemma oof_behaved : 
  behaved oof.
Proof.
intros G i t Ht Hindex.
destruct Ht; simp_sub; eauto using oof with static dynamic.
Qed.


Lemma aof_behaved : 
  behaved aof.
Proof.
intros G i t Ht Hindex.
destruct Ht; simp_sub; eauto using aof with static dynamic.
Qed.


Lemma dof_sof :
  forall G1 s G2, dof G1 s G2 -> sof G1 s G2.
Proof.
intros G1 s G2 Hof.
invert Hof; intros; subst; auto using sof.
Qed.


Lemma dof_tof :
  forall G1 s G2, dof G1 s G2 -> tof G1 s G2.
Proof.
intros G1 s G2 Hof.
invert Hof; intros; subst; auto using tof.
Qed.


Lemma oof_tof :
  forall G1 s G2, oof G1 s G2 -> tof G1 s G2.
Proof.
intros G1 s G2 Hof.
invert Hof; intros; subst; auto using tof.
Qed.


Lemma aof_sof :
  forall G m C, aof G m C -> sof G m C.
Proof.
intros G m C Hof.
invert Hof; intros; subst; auto using sof.
Qed.


Lemma aof_dof :
  forall G m C, aof G m C -> dof G m C.
Proof.
intros G m C Hof.
invert Hof; intros; subst; auto using dof.
Qed.


Lemma aof_tof :
  forall G m C, aof G m C -> tof G m C.
Proof.
intros G m C Hof.
invert Hof; intros; subst; auto using tof.
Qed.


Lemma aof_oof :
  forall G1 s G2, aof G1 s G2 -> oof G1 s G2.
Proof.
intros G1 s G2 Hof.
invert Hof; intros; subst; auto using oof.
Qed.


Lemma subof_dof_sof :
  forall G1 s G2,
    subof dof G1 s G2
    -> subof sof G1 s G2.
Proof.
eauto using subof_mono, dof_sof.
Qed.


Lemma subof_dof_tof :
  forall G1 s G2,
    subof dof G1 s G2
    -> subof tof G1 s G2.
Proof.
eauto using subof_mono, dof_tof.
Qed.


Lemma subof_oof_tof :
  forall G1 s G2,
    subof oof G1 s G2
    -> subof tof G1 s G2.
Proof.
eauto using subof_mono, oof_tof.
Qed.


Lemma subof_aof_sof :
  forall G1 s G2,
    subof aof G1 s G2
    -> subof sof G1 s G2.
Proof.
eauto using subof_mono, aof_sof.
Qed.


Lemma subof_aof_dof :
  forall G1 s G2,
    subof aof G1 s G2
    -> subof dof G1 s G2.
Proof.
eauto using subof_mono, aof_dof.
Qed.


Lemma subof_aof_oof :
  forall G1 s G2,
    subof aof G1 s G2
    -> subof oof G1 s G2.
Proof.
eauto using subof_mono, aof_oof.
Qed.


Lemma subof_aof_tof :
  forall G1 s G2,
    subof aof G1 s G2
    -> subof tof G1 s G2.
Proof.
eauto using subof_mono, aof_tof.
Qed.


Lemma subof_wc_bind_sof_cn :
  forall G G' s k,
    subof (wc sof) G s G'
    -> subof (wc sof) (G; cl_cn (subst s k)) (dot (var 0) (compose s sh1)) (G'; cl_cn k).
Proof.
intros G G' s k Hsubof.
repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
apply subof_wc_bind; auto with static; [].
exact sof_behaved.
Qed.


Lemma subof_wc_bind_dof_cn :
  forall G G' s k,
    subof (wc dof) G s G'
    -> subof (wc dof) (G; cl_cn (subst s k)) (dot (var 0) (compose s sh1)) (G'; cl_cn k).
Proof.
intros G G' s k Hsubof.
repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
apply subof_wc_bind; auto with static; [].
exact dof_behaved.
Qed.


Lemma subof_wc_bind_dof_tm :
  forall G G' s t,
    subof (wc dof) G s G'
    -> subof (wc dof) (G; cl_tm (subst s t)) (dot (var 0) (compose s sh1)) (G'; cl_tm t).
Proof.
intros G G' s t Hsubof.
repl (cl_tm (subst s t)) with (subst s (cl_tm t)) by (simp_sub; reflexivity).
apply subof_wc_bind; auto with static; [].
exact dof_behaved.
Qed.


Lemma subof_wc_bind_dof_md :
  forall G G' s T c,
    subof (wc dof) G s G'
    -> subof (wc dof) (G; cl_md (subst s T) (subst s c)) (dot (var 0) (compose s sh1)) (G'; cl_md T c).
Proof.
intros G G' s T c Hsubof.
repl (cl_md (subst s T) (subst s c)) with (subst s (cl_md T c)) by (simp_sub; reflexivity).
apply subof_wc_bind; auto with static; [].
exact dof_behaved.
Qed.


Lemma subof_wc_bind_dof_utm :
  forall G G' s,
    subof (wc dof) G s G'
    -> subof (wc dof) (G; cl_utm) (dot (var 0) (compose s sh1)) (G'; cl_utm).
Proof.
intros G G' s Hsubof.
repl (cl_utm) with (subst s cl_utm) by (simp_sub; reflexivity).
apply subof_wc_bind; auto with static; [].
exact dof_behaved.
Qed.


Lemma subof_wc_bind_tof_utm :
  forall G G' s,
    subof (wc tof) G s G'
    -> subof (wc tof) (G; cl_utm) (dot (var 0) (compose s sh1)) (G'; cl_utm).
Proof.
intros G G' s Hsubof.
repl (cl_utm) with (subst s cl_utm) by (simp_sub; reflexivity).
apply subof_wc_bind; auto with static; [].
exact tof_behaved.
Qed.


Lemma subof_wc_bind_dof_cn_md :
  forall G G' s k T,
    subof (wc dof) G s G'
    -> subof (wc dof) (G; cl_cn (subst s k); cl_md (subst (dot (var 0) (compose s sh1)) T) (var 0)) (dot (var 0) (dot (var 1) (compose s (sh 2)))) (G'; cl_cn k; cl_md T (var 0)).
Proof.
intros G G' s k T Hsubof.
repl (cl_md (subst (dot (var 0) (compose s sh1)) T) (var 0)) with (subst (dot (var 0) (compose s sh1)) (cl_md T (var 0))) by (simp_sub; reflexivity).
repl (dot (var 0) (dot (var 1) (compose s (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s sh1)) sh1)) by (simp_sub; reflexivity).
apply subof_wc_bind; auto using dof_behaved with static; [].
repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
apply subof_wc_bind; auto using dof_behaved with static; done.
Qed.


Lemma subof_wc_bind_tof_cn_md :
  forall G G' s k T,
    subof (wc tof) G s G'
    -> subof (wc tof) (G; cl_cn (subst s k); cl_md (subst (dot (var 0) (compose s sh1)) T) (var 0)) (dot (var 0) (dot (var 1) (compose s (sh 2)))) (G'; cl_cn k; cl_md T (var 0)).
Proof.
intros G G' s k T Hsubof.
repl (cl_md (subst (dot (var 0) (compose s sh1)) T) (var 0)) with (subst (dot (var 0) (compose s sh1)) (cl_md T (var 0))) by (simp_sub; reflexivity).
repl (dot (var 0) (dot (var 1) (compose s (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s sh1)) sh1)) by (simp_sub; reflexivity).
apply subof_wc_bind; auto using tof_behaved with static; [].
repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
apply subof_wc_bind; auto using tof_behaved with static; done.
Qed.


Lemma subof_tm_utm_oof :
  forall G t,
    subof oof (G; cl_tm t) (under 1 id) (G; cl_utm).
Proof.
intros G t.
#2 apply subof_dot.
  apply subof_sh; auto with sequence; done.

  simp_sub.
  apply oof_utm; eauto with dynamic sequence.
Qed.


Lemma subof_utm_tm_oof :
  forall G t,
    subof oof (G; cl_utm) (under 1 id) (G; cl_tm t).
Proof.
intros G t.
#2 apply subof_dot.
  apply subof_sh; auto with sequence; done.

  simp_sub.
  apply oof_tm; auto with dynamic sequence.
Qed.


Lemma subof_tm_utm_tof :
  forall G t,
    subof tof (G; cl_tm t) (under 1 id) (G; cl_utm).
Proof.
intros; apply subof_oof_tof; apply subof_tm_utm_oof.
Qed.


Lemma subof_utm_tm_tof :
  forall G t,
    subof tof (G; cl_utm) (under 1 id) (G; cl_tm t).
Proof.
intros; apply subof_oof_tof; apply subof_utm_tm_oof.
Qed.


Lemma subst_ofk_eqk_subk_ofc_eqc_wc :
  (forall G1 G2 s k,
     subof (wc sof) G1 s G2
     -> ofk G2 k
     -> ofk G1 (subst s k))
  /\
  (forall G1 G2 s k k',
     subof (wc sof) G1 s G2
     -> eqk G2 k k'
     -> eqk G1 (subst s k) (subst s k'))
  /\
  (forall G1 G2 s k k',
     subof (wc sof) G1 s G2
     -> subk G2 k k'
     -> subk G1 (subst s k) (subst s k'))
  /\
  (forall G1 G2 s c k,
     subof (wc sof) G1 s G2
     -> ofc G2 c k
     -> ofc G1 (subst s c) (subst s k))
  /\
  (forall G1 G2 s c c' k,
     subof (wc sof) G1 s G2
     -> eqc G2 c c' k
     -> eqc G1 (subst s c) (subst s c') (subst s k)).
Proof.
#15 exploit
  (static_ind
     (fun G2 k =>
        forall G1 s,
          subof (wc sof) G1 s G2
          -> ofk G1 (subst s k))
     (fun G2 k k' =>
        forall G1 s,
          subof (wc sof) G1 s G2
          -> eqk G1 (subst s k) (subst s k'))
     (fun G2 k k' =>
        forall G1 s,
          subof (wc sof) G1 s G2
          -> subk G1 (subst s k) (subst s k'))
     (fun G2 c k =>
        forall G1 s,
          subof (wc sof) G1 s G2
          -> ofc G1 (subst s c) (subst s k))
     (fun G2 c c' k =>
        forall G1 s,
          subof (wc sof) G1 s G2
          -> eqc G1 (subst s c) (subst s c') (subst s k))) as Hind;
  [.. | destruct_all Hind; repeat2 split; eauto; done];
try (intros;
     simp_sub;
     eauto using subof_wc_bind_sof_cn with static;
     done).

(* subk_pi *)
intros; simp_sub; eauto 7 using subof_wc_bind_sof_cn with static; done.

(* subk_sigma *)
intros; simp_sub; eauto 7 using subof_wc_bind_sof_cn with static; done.

(* var *)
intros G2 i k Hindex G1 s Hs.
simp_sub.
assert (subof sof G1 s G2) as Hs'.
  apply (subof_mono (wc sof)); auto; [].
  intros; apply wc_elim; auto; done.
so (subof_index _ sof_behaved _#5 iscl_cn Hs' Hindex) as Hsof.
simp_sub in Hsof.
invert Hsof; auto; done.

(* app *)
intros G2 k1 k2 c1 c2 _ IH1 _ IH2 G1 s Hs.
simp_sub.
repl (subst (dot (subst s c2) s) k2) with (subst1 (subst s c2) (subst (dot (var 0) (compose s sh1)) k2)).
  simp_sub; reflexivity.
eapply ofc_app; eauto; [].
so (IH1 _ _ Hs) as H1.
simp_sub in H1.
assumption.

(* pair *)
intros G2 k1 k2 c1 c2 _ IH1 _ IH2 _ IH3 G1 s Hs.
intros; simp_sub.
apply ofc_pair; eauto using subof_wc_bind_sof_cn with static; [].
simp_sub.
so (IH2 _ _ Hs) as H2.
simp_sub in H2.
assumption.

(* pi2 *)
intros G2 k1 k2 c _ IH G1 s Hs.
simp_sub.
repl (subst (dot (cn_pi1 (subst s c)) s) k2) with (subst1 (cn_pi1 (subst s c)) (subst (dot (var 0) (compose s sh1)) k2)).
  simp_sub; reflexivity.
eapply ofc_pi2; eauto; [].
so (IH _ _ Hs) as H.
simp_sub in H.
eassumption.

(* extpi *)
intros G2 k1 k2 k3 c _ IH1 _ IH2 G1 s Hs.
simp_sub.
eapply ofc_extpi; eauto; [].
simp_sub.
#2 eexploit IH2 as H2.
  apply subof_wc_bind_sof_cn; eauto; done.
simp_sub in H2.
assumption.

(* extsigma *)
intros G2 k1 k2 c _ IH1 _ IH2 _ IH3 G1 s Hs.
simp_sub.
#2 eapply ofc_extsigma; eauto.
  simp_sub.
  so (IH2 _ _ Hs) as H2.
  simp_sub in H2.
  assumption.

  #2 eexploit IH3 as H3.
    apply subof_wc_bind_sof_cn; eauto; done.
  simp_sub in H3.
  assumption.

(* app *)
intros G2 k1 k2 c1 c2 c1' c2' _ IH1 _ IH2 G1 s Hs.
simp_sub.
repl (subst (dot (subst s c2) s) k2) with (subst1 (subst s c2) (subst (dot (var 0) (compose s sh1)) k2)).
  simp_sub; reflexivity.
eapply eqc_app; eauto; [].
so (IH1 _ _ Hs) as H1.
so (IH2 _ _ Hs) as H2.
simp_sub in H1; simp_sub in H2.
assumption.

(* pair *)
intros G2 k1 k2 c1 c1' c2 c2' _ IH1 _ IH2 _ IH3 G1 s Hs.
simp_sub.
apply eqc_pair; eauto using subof_wc_bind_sof_cn with static; [].
simp_sub.
so (IH2 _ _ Hs) as H2.
simp_sub in H2.
assumption.

(* pi2 *)
intros G2 k1 k2 c c' _ IH G1 s Hs.
simp_sub.
repl (subst (dot (cn_pi1 (subst s c)) s) k2) with (subst1 (cn_pi1 (subst s c)) (subst (dot (var 0) (compose s sh1)) k2)).
  simp_sub; reflexivity.
eapply eqc_pi2; eauto; [].
so (IH _ _ Hs) as H.
simp_sub in H.
eassumption.

(* extpi *)
intros G2 k1 k2 k3 k3' c c' _ IH1 _ IH2 _ IH3 G1 s Hs.
simp_sub.
eapply eqc_extpi; eauto; [].
simp_sub.
#2 eexploit IH3 as H3.
  apply subof_wc_bind_sof_cn; eauto; done.
simp_sub in H3.
assumption.

(* extpiw *)
intros G2 k1 k2 k3 c c' _ IH1 _ IH2 G1 s Hs.
simp_sub.
eapply eqc_extpiw; eauto; [].
simp_sub.
#2 eexploit IH2 as H2.
  apply subof_wc_bind_sof_cn; eauto; done.
simp_sub in H2.
assumption.

(* extsigma *)
intros G2 k1 k2 c c' _ IH1 _ IH2 _ IH3 G1 s Hs.
simp_sub.
#2 eapply eqc_extsigma; eauto.
  simp_sub.
  so (IH2 _ _ Hs) as H2.
  simp_sub in H2.
  assumption.

  #2 eexploit IH3 as H3.
    apply subof_wc_bind_sof_cn; eauto; done.
  simp_sub in H3.
  assumption.

(* beta *)
intros G2 k1 k2 c1 c2 _ IH1 _ IH2 G1 s Hs.
simp_sub.
repl (subst (dot (subst s c1) s) c2) with (subst1 (subst s c1) (subst (dot (var 0) (compose s sh1)) c2)).
  simp_sub; reflexivity.
repl (subst (dot (subst s c1) s) k2) with (subst1 (subst s c1) (subst (dot (var 0) (compose s sh1)) k2)).
  simp_sub; reflexivity.
apply eqc_beta; auto using subof_wc_bind_sof_cn; done.
Qed.


Lemma subst_ofc_wc :
  forall G1 G2 s c k,
    subof (wc sof) G1 s G2
    -> ofc G2 c k
    -> ofc G1 (subst s c) (subst s k).
Proof.
so subst_ofk_eqk_subk_ofc_eqc_wc as H.
destruct_all H.
assumption.
Qed.


Lemma subst_sof_wc :
  forall G1 G2 s t c,
    subof (wc sof) G1 s G2
    -> sof G2 t c
    -> sof G1 (subst s t) (subst s c).
Proof.
intros G1 G2 s t c Hs H.
destruct H; simp_sub; eauto using subst_ofc_wc, sof.
Qed.


Lemma sof_wc :
  forall G t c,
    sof G t c -> wc sof G t c.
Proof.
exact (wc_intro _ subst_sof_wc).
Qed.


Lemma subst_ofk :
  forall G1 G2 s k,
    subof sof G1 s G2
    -> ofk G2 k
    -> ofk G1 (subst s k).
Proof.
intros G1 G2 s k Hs Hof.
so subst_ofk_eqk_subk_ofc_eqc_wc as H.
destruct_all H.
eauto using subof_mono, sof_wc.
Qed.


Lemma subst_eqk :
  forall G1 G2 s k k',
    subof sof G1 s G2
    -> eqk G2 k k'
    -> eqk G1 (subst s k) (subst s k').
Proof.
intros G1 G2 s k k' Hs H.
so subst_ofk_eqk_subk_ofc_eqc_wc as Hsubst.
destruct_all Hsubst.
eauto using subof_mono, sof_wc.
Qed.


Lemma subst_subk :
  forall G1 G2 s k k',
    subof sof G1 s G2
    -> subk G2 k k'
    -> subk G1 (subst s k) (subst s k').
Proof.
intros G1 G2 s k k' Hs H.
so subst_ofk_eqk_subk_ofc_eqc_wc as Hsubst.
destruct_all Hsubst.
eauto using subof_mono, sof_wc.
Qed.


Lemma subst_ofc :
  forall G1 G2 s c k,
    subof sof G1 s G2
    -> ofc G2 c k
    -> ofc G1 (subst s c) (subst s k).
Proof.
intros G1 G2 s c k Hs H.
so subst_ofk_eqk_subk_ofc_eqc_wc as Hsubst.
destruct_all Hsubst.
eauto using subof_mono, sof_wc.
Qed.


Lemma subst_eqc :
  forall G1 G2 s c c' k,
    subof sof G1 s G2
    -> eqc G2 c c' k
    -> eqc G1 (subst s c) (subst s c') (subst s k).
Proof.
intros G1 G2 s c k Hs H.
so subst_ofk_eqk_subk_ofc_eqc_wc as Hsubst.
destruct_all Hsubst.
eauto using subof_mono, sof_wc.
Qed.


Lemma subst_ofc_type :
  forall G1 G2 s c,
    subof sof G1 s G2
    -> ofc G2 c kd_type
    -> ofc G1 (subst s c) kd_type.
Proof.
intros G1 G2 s c Hs Hof.
so (subst_ofc _#5 Hs Hof) as Hof'.
simp_sub in Hof'.
assumption.
Qed.


Lemma subst_eqc_type :
  forall G1 G2 s c c',
    subof sof G1 s G2
    -> eqc G2 c c' kd_type
    -> eqc G1 (subst s c) (subst s c') kd_type.
Proof.
intros G1 G2 s c c' Hs Heq.
so (subst_eqc _#6 Hs Heq) as Heq'.
simp_sub in Heq'.
assumption.
Qed.


Lemma unshift_tm :
  forall G t,
    subof sof G (dot nonsense id) (G; cl_tm t).
Proof.
intros G t.
apply subof_dot; auto with static; [].
simp_sub.
auto using sof; done.
Qed.


Lemma unshift_md :
  forall G T c,
    subof sof G (dot nonsense id) (G; cl_md T c).
Proof.
intros G T c.
apply subof_dot; auto with static; [].
simp_sub.
auto using sof; done.
Qed.


Lemma subst1_ofk :
  forall G c k k',
    ofc G c k
    -> ofk (G; cl_cn k) k'
    -> ofk G (subst1 c k').
Proof.
intros.
eapply subst_ofk; eauto; [].
apply subof_dot; auto using subof_id; [].
simp_sub.
auto using sof.
Qed.


Lemma subst1_eqk :
  forall G c k k1 k2,
    ofc G c k
    -> eqk (G; cl_cn k) k1 k2
    -> eqk G (subst1 c k1) (subst1 c k2).
Proof.
intros.
eapply subst_eqk; eauto; [].
apply subof_dot; auto using subof_id; [].
simp_sub.
auto using sof.
Qed.


Lemma subst1_subk :
  forall G c k k1 k2,
    ofc G c k
    -> subk (G; cl_cn k) k1 k2
    -> subk G (subst1 c k1) (subst1 c k2).
Proof.
intros.
eapply subst_subk; eauto; [].
apply subof_dot; auto using subof_id; [].
simp_sub.
auto using sof.
Qed.


Lemma subst1_ofc :
  forall G c k c' k',
    ofc G c k
    -> ofc (G; cl_cn k) c' k'
    -> ofc G (subst1 c c') (subst1 c k').
Proof.
intros.
eapply subst_ofc; eauto; [].
apply subof_dot; auto using subof_id; [].
simp_sub.
auto using sof.
Qed.


Lemma subst1_eqc :
  forall G c k c1 c2 k',
    ofc G c k
    -> eqc (G; cl_cn k) c1 c2 k'
    -> eqc G (subst1 c c1) (subst1 c c2) (subst1 c k').
Proof.
intros.
eapply subst_eqc; eauto; [].
apply subof_dot; auto using subof_id; [].
simp_sub.
auto using sof.
Qed.


Lemma subst1_eqc_type :
  forall G c k c1 c2,
    ofc G c k
    -> eqc (G; cl_cn k) c1 c2 kd_type
    -> eqc G (subst1 c c1) (subst1 c c2) kd_type.
Proof.
intros G c k c1 c2 Hof Heqc.
replsub kd_type with (subst1 c kd_type).
eapply subst1_eqc; eauto; done.
Qed.


Lemma weaken_ofk :
  forall G1 G2 k,
    ofk G1 k
    -> ofk (G1;; G2) (subst (sh (length G2)) k).
Proof.
intros G1 G2 k H.
eapply subst_ofk; eauto; [].
apply subof_sh; [].
apply app_truncate; auto; done.
Qed.


Lemma weaken_eqk :
  forall G1 G2 k k',
    eqk G1 k k'
    -> eqk (G1;; G2) (subst (sh (length G2)) k) (subst (sh (length G2)) k').
Proof.
intros G1 G2 k k' H.
eapply subst_eqk; eauto; [].
apply subof_sh; [].
apply app_truncate; auto; done.
Qed.


Lemma weaken_subk :
  forall G1 G2 k k',
    subk G1 k k'
    -> subk (G1;; G2) (subst (sh (length G2)) k) (subst (sh (length G2)) k').
Proof.
intros G1 G2 k k' H.
eapply subst_subk; eauto; [].
apply subof_sh; [].
apply app_truncate; auto; done.
Qed.


Lemma weaken_ofc :
  forall G1 G2 c k,
    ofc G1 c k
    -> ofc (G1;; G2) (subst (sh (length G2)) c) (subst (sh (length G2)) k).
Proof.
intros G1 G2 c k H.
eapply subst_ofc; eauto; [].
apply subof_sh; [].
apply app_truncate; auto; done.
Qed.


Lemma weaken_eqc :
  forall G1 G2 c c' k,
    eqc G1 c c' k
    -> eqc (G1;; G2) (subst (sh (length G2)) c) (subst (sh (length G2)) c') (subst (sh (length G2)) k).
Proof.
intros G1 G2 c c' k H.
eapply subst_eqc; eauto; [].
apply subof_sh; [].
apply app_truncate; auto; done.
Qed.


Lemma weaken1_subk :
  forall G A k k',
    subk G k k'
    -> subk (G; A) (shift1 k) (shift1 k').
Proof.
intros G A k k' H.
so (weaken_subk _ (nil; A) _ _ H) as H'.
simpl in H'.
assumption.
Qed.


Lemma weaken1_ofk :
  forall G A k,
    ofk G k
    -> ofk (G; A) (shift1 k).
Proof.
intros G A k H.
so (weaken_ofk _ (nil; A) _ H) as H'.
simpl in H'.
assumption.
Qed.


Lemma weaken1_eqk :
  forall G A k k',
    eqk G k k'
    -> eqk (G; A) (shift1 k) (shift1 k').
Proof.
intros G A k k' H.
so (weaken_eqk _ (nil; A) _ _ H) as H'.
simpl in H'.
assumption.
Qed.


Lemma weaken1_ofc :
  forall G A c k,
    ofc G c k
    -> ofc (G; A) (shift1 c) (shift1 k).
Proof.
intros G A c k H.
so (weaken_ofc _ (nil; A) _ _ H) as H'.
simpl in H'.
assumption.
Qed.


Lemma weaken1_ofc_type :
  forall G A t,
    ofc G t kd_type
    -> ofc (G; A) (shift1 t) kd_type.
Proof.
intros G A t H.
replsub kd_type with (shift1 kd_type).
apply weaken1_ofc; auto; done.
Qed.


Lemma weaken1_eqc :
  forall G A c c' k,
    eqc G c c' k
    -> eqc (G; A) (shift1 c) (shift1 c') (shift1 k).
Proof.
intros G A c c' k H.
so (weaken_eqc _ (nil; A) _ _ _ H) as H'.
simpl in H'.
assumption.
Qed.


Lemma weaken1_eqc_type :
  forall G A c c',
    eqc G c c' kd_type
    -> eqc (G; A) (shift1 c) (shift1 c') kd_type.
Proof.
intros G A c c' H.
so (weaken1_eqc _ A _ _ _ H) as H'.
simp_sub in H'.
assumption.
Qed.


Lemma sof_wclo :
  wclo sof.
Proof.
intros G t C C' H.
invert H; intros; subst; eauto using sof, weaken1_ofc.
Qed.


Lemma subof_bind_sof_cn :
  forall G G' s k,
    subof sof G s G'
    -> subof sof (G; cl_cn (subst s k)) (dot (var 0) (compose s sh1)) (G'; cl_cn k).
Proof.
intros G G' s k Hsubof.
repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
apply subof_wclo_bind; eauto using sof_behaved, sof_wclo with static.
Qed.


Lemma subof_subsume_0_aof :
  forall G k k',
    subk G k k'
    -> subof aof (G; cl_cn k) (under 1 id) (G; cl_cn k').
Proof.
intros G k k' Heqk.
#2 apply subof_dot.
  apply subof_sh; auto with sequence; done.

  simp_sub.
  apply aof_cn; [].
  #2 apply (@ofc_subsume _ (shift1 k)).
    apply ofc_var; auto using index_0; done.

    apply weaken1_subk; auto; done.
Qed.


Lemma subof_subsume_0 :
  forall G k k',
    subk G k k'
    -> subof sof (G; cl_cn k) (under 1 id) (G; cl_cn k').
Proof.
auto using subof_subsume_0_aof, subof_aof_sof.
Qed.


Lemma subof_conv_0_aof :
  forall G k k',
    eqk G k k'
    -> subof aof (G; cl_cn k) (under 1 id) (G; cl_cn k').
Proof.
auto using subof_subsume_0_aof, subk_refl.
Qed.


Lemma subof_conv_0 :
  forall G k k',
    eqk G k k'
    -> subof sof (G; cl_cn k) (under 1 id) (G; cl_cn k').
Proof.
auto using subof_conv_0_aof, subof_aof_sof.
Qed.


Lemma subof_under_sof :
  forall G1 G1' G2 s,
    Forall iscl G2
    -> subof sof G1 s G1'
    -> subof sof (G1;; subst_ctx s G2) (under (length G2) s) (G1';; G2).
Proof.
intros G1 G1' G2 s Hiscl Hsubof.
apply subof_under; auto using sof_behaved; [].
intros; apply sof_wc; auto.
Qed.


Lemma subof_weaken_sof :
  forall G1 G2 s G,
    subof sof G1 s G
    -> subof sof (G1;; G2) (compose s (sh (length G2))) G.
Proof.
intros G1 G2 s G Hsubof.
apply subof_wclo_weaken; auto using sof_wclo.
Qed.


Lemma subst_fsts :
  forall T k s,
    fsts T k
    -> fsts (subst s T) (subst s k).
Proof.
intros T k s Hfst.
revert s.
induct Hfst; try (intros; simp_sub; auto with static; done).
Qed.


Lemma subst_ofs :
  forall G1 G2 s T,
    subof sof G1 s G2
    -> ofs G2 T
    -> ofs G1 (subst s T).
Proof.
intros G1 G2 s T Hsubof Hofs.
revert G1 s Hsubof.
induct Hofs.

(* one *)
intros G2 G1 s Hsubof.
simp_sub.
auto with static; done.

(* stat *)
intros G2 k Hofk G1 s Hsubof.
simp_sub.
eauto using subst_ofk with static; done.

(* dyn *)
intros G2 c Hofk G1 s Hsubof.
simp_sub.
apply ofs_dyn; [].
repl kd_type with (subst s kd_type) by (simp_sub; reflexivity).
eapply subst_ofc; eauto; done.

(* pigen *)
intros G2 T1 T2 k1 _ IH1 Hfst _ IH2 G1 s Hsubof.
simp_sub.
eapply ofs_pigen; eauto using subst_fsts; [].
apply IH2; [].
apply subof_bind_sof_cn; auto; done.

(* piapp *)
intros G2 T1 T2 k1 _ IH1 Hfst _ IH2 G1 s Hsubof.
simp_sub.
eapply ofs_piapp; eauto using subst_fsts; [].
apply IH2; [].
apply subof_bind_sof_cn; auto; done.

(* sigma *)
intros G2 T1 T2 k1 _ IH1 Hfst _ IH2 G1 s Hsubof.
simp_sub.
eapply ofs_sigma; eauto using subst_fsts; [].
apply IH2; [].
apply subof_bind_sof_cn; auto; done.
Qed.


Lemma subst_eqs :
  forall G1 G2 s T T',
    subof sof G1 s G2
    -> eqs G2 T T'
    -> eqs G1 (subst s T) (subst s T').
Proof.
intros G1 G2 s T T' Hsubof Heqs.
revert G1 s Hsubof.
induct Heqs.

(* refl *)
intros G T H G1 s Hsubof.
apply eqs_refl; [].
eapply subst_ofs; eauto; done.

(* symm *)
intros G T T' _ IH G1 s Hsubof.
apply eqs_symm; [].
apply IH; auto; done.

(* trans *)
intros G T1 T2 T3 _ IH1 _ IH2 G1 s Hsubof.
eapply eqs_trans; eauto; done.

(* stat *)
intros G k k' H G1 s Hsubof.
simp_sub.
apply eqs_stat; eauto using subst_eqk; done.

(* dyn *)
intros G t t' H G1 s Hsubof.
simp_sub.
apply eqs_dyn; [].
repl kd_type with (subst s kd_type) by (simp_sub; reflexivity).
eauto using subst_eqc; done.

(* pigen *)
intros G T1 T1' k1 T2 T2' _ IH1 HfstT1 _ IH2 G1 s Hsubof.
simp_sub.
eapply eqs_pigen; eauto using subst_fsts; [].
apply IH2; [].
apply subof_bind_sof_cn; auto; done.

(* piapp *)
intros G T1 T1' k1 T2 T2' _ IH1 HfstT1 _ IH2 G1 s Hsubof.
simp_sub.
eapply eqs_piapp; eauto using subst_fsts; [].
apply IH2; [].
apply subof_bind_sof_cn; auto; done.

(* sigma *)
intros G T1 T1' k1 T2 T2' _ IH1 HfstT1 _ IH2 G1 s Hsubof.
simp_sub.
eapply eqs_sigma; eauto using subst_fsts; [].
apply IH2; [].
apply subof_bind_sof_cn; auto; done.
Qed.


Lemma subst_subs :
  forall G1 G2 s T T',
    subof sof G1 s G2
    -> subs G2 T T'
    -> subs G1 (subst s T) (subst s T').
Proof.
intros G1 G2 s T T' Hsubof Hsubs.
revert G1 s Hsubof.
induct Hsubs.

(* refl *)
intros G T T' Heqs G1 s Hsubof.
apply subs_refl; eauto using subst_eqs; done.

(* trans *)
intros G T1 T2 T3 _ IH1 _ IH2 G1 s Hsubof.
eapply subs_trans; eauto; done.

(* stat *)
intros G k k' H G1 s Hsubof.
simp_sub.
apply subs_stat; eauto using subst_subk; done.

(* pigen *)
intros G T1 T1' k1 k1' T2 T2' _ IH1 Hfst' _ IH2 Hfst Hofs G1 s Hsubof.
simp_sub.
eapply subs_pigen; eauto using subst_fsts, subof_bind_sof_cn, subst_ofs; done.

(* piapp *)
intros G T1 T1' k1 k1' T2 T2' _ IH1 Hfst' _ IH2 Hfst Hofs G1 s Hsubof.
simp_sub.
eapply subs_piapp; eauto using subst_fsts, subof_bind_sof_cn, subst_ofs; done.

(* piapp *)
intros G T1 T1' k1 k1' T2 T2' H H0 H1 H2 H3 H4 H5 G1 s Hsubof.
simp_sub.
eapply subs_sigma; eauto using subst_fsts, subof_bind_sof_cn, subst_ofs; done.
Qed.


Lemma subst1_ofs :
  forall G c k T,
    ofc G c k
    -> ofs (G; cl_cn k) T
    -> ofs G (subst1 c T).
Proof.
intros.
eapply subst_ofs; eauto; [].
apply subof_dot; auto using subof_id; [].
simp_sub.
auto using sof.
Qed.


Lemma weaken_ofs :
  forall G1 G2 T,
    ofs G1 T
    -> ofs (G1;; G2) (subst (sh (length G2)) T).
Proof.
intros G1 G2 T H.
eapply subst_ofs; eauto; [].
apply subof_sh; [].
apply app_truncate; auto; done.
Qed.


Lemma weaken1_ofs :
  forall G A T,
    ofs G T
    -> ofs (G; A) (shift1 T).
Proof.
intros G A T H.
so (weaken_ofs _ (nil; A) _ H) as H'.
simpl in H'.
assumption.
Qed.


Lemma weaken2_ofs :
  forall G A B T,
    ofs G T
    -> ofs (G; A; B) (subst (sh 2) T).
Proof.
intros.
eapply subst_ofs; eauto using subof_sh with sequence.
Qed.


Lemma subst1_eqs :
  forall G c k T1 T2,
    ofc G c k
    -> eqs (G; cl_cn k) T1 T2
    -> eqs G (subst1 c T1) (subst1 c T2).
Proof.
intros.
eapply subst_eqs; eauto; [].
apply subof_dot; auto using subof_id; [].
simp_sub.
auto using sof; done.
Qed.

Lemma subst1_subs :
  forall G c k T1 T2,
    ofc G c k
    -> subs (G; cl_cn k) T1 T2
    -> subs G (subst1 c T1) (subst1 c T2).
Proof.
intros.
eapply subst_subs; eauto; [].
apply subof_dot; auto using subof_id; [].
simp_sub.
auto using sof; done.
Qed.


Lemma weaken1_eqs :
  forall G A T T',
    eqs G T T'
    -> eqs (G; A) (shift1 T) (shift1 T').
Proof.
intros G A k k' H.
eapply subst_eqs; eauto using weaken1.
Qed.


Lemma weaken1_subs :
  forall G A T T',
    subs G T T'
    -> subs (G; A) (shift1 T) (shift1 T').
Proof.
intros G A k k' H.
eapply subst_subs; eauto using weaken1.
Qed.


Lemma subof_conv_0t :
  forall G t t',
    eqc G t t' kd_type
    -> subof aof (G; cl_tm t) (under 1 id) (G; cl_tm t').
Proof.
intros G t t' Heqc.
#2 apply subof_dot.
  apply subof_sh; auto with sequence; done.

  simp_sub.
  #2 apply aof_tm.
    #2 apply (@oft_equiv _ (shift1 t)).
      apply oft_var; auto using index_0; done.
  
      repl kd_type with (shift1 kd_type) by (simp_sub; reflexivity).
      apply weaken1_eqc; auto; done.

    eapply tvalue_var; auto using index_0; done.
Qed.


Lemma subof_subsume_0m :
  forall G T T' c,
    subs G T T'
    -> subof aof (G; cl_md T c) (under 1 id) (G; cl_md T' c).
Proof.
intros G T T' c Hsubs.
#2 apply subof_dot.
  apply subof_sh; auto with sequence; done.

  simp_sub.
  #3 apply aof_md.
    #2 apply (@ofm_subsume _ _ (shift1 T)).
      eapply ofm_var; auto using index_0; done.
  
      apply weaken1_subs; auto; done.

    eapply mvalue_var; auto using index_0; done.

    eapply fstm_var; eauto using index_0; done.
Qed.    


Lemma subof_subsume_01m :
  forall G k k' T T',
    subk G k k'
    -> subs (G; cl_cn k) T T'
    -> subof aof (G; cl_cn k; cl_md T (var 0)) (under 2 id) (G; cl_cn k'; cl_md T' (var 0)).
Proof.
intros G k k' T T' Hsubk Hsubs.
simp_sub.
#2 apply subof_dot.
  apply subof_dot; eauto using subof_sh with sequence; [].
  simp_sub.
  apply aof_cn; [].
  #2 apply (@ofc_subsume _ (subst (sh 2) k)).
    apply ofc_var; auto with sequence; done.

    eapply subst_subk; eauto; [].
    apply subof_sh; eauto with sequence; done.

  simp_sub.
  #3 apply aof_md.
    repl (dot (var 1) (sh 2)) with (compose (dot (var 0) sh1) sh1) by (simp_sub; reflexivity).
    rewrite -> subst_compose; [].
    simp_sub.
    #2 apply (@ofm_subsume _ _ (shift1 T)).
      eapply ofm_var; eauto with static sequence; done.

      eapply subst_subs; eauto; [].
      apply subof_sh; eauto with sequence; done.

    eapply mvalue_var; eauto with sequence; done.

    repl (var 1) with (shift1 (var 0)) by (simp_sub; reflexivity).
    eapply fstm_var; eauto with sequence; done.
Qed.


Lemma subof_conv_01m :
  forall G k k' T T',
    eqk G k k'
    -> eqs (G; cl_cn k) T T'
    -> subof aof (G; cl_cn k; cl_md T (var 0)) (under 2 id) (G; cl_cn k'; cl_md T' (var 0)).
Proof.
auto using subof_subsume_01m, subk_refl, subs_refl.
Qed.


Lemma subst_fstm_wc :
  forall G1 G2 s M c,
    subof (wc tof) G1 s G2
    -> fstm G2 M c
    -> fstm G1 (subst s M) (subst s c).
Proof.
intros G1 G2 s M c Hsubof Hfst.
revert G1 s Hsubof.
induct Hfst; try (intros; simp_sub; auto with static; done).

(* var *)
intros G2 i T c Hindex G1 s Hsubof.
simp_sub.
assert (subof tof G1 s G2) as Hsubof'.
  apply (subof_mono (wc tof)); auto; [].
  intros; apply wc_elim; auto; done.
so (subof_index _ tof_behaved _#5 iscl_md Hsubof' Hindex) as Hsof.
simp_sub in Hsof.
invert Hsof; auto; done.

(* var *)
intros G2 i c Hindex G1 s Hsubof.
simp_sub.
assert (subof tof G1 s G2) as Hsubof'.
  apply (subof_mono (wc tof)); auto; [].
  intros; apply wc_elim; auto; done.
so (subof_index _ tof_behaved _#5 iscl_umd Hsubof' Hindex) as Hsof.
simp_sub in Hsof.
invert Hsof; auto; done.

(* lam *)
intros G2 T M k c Hfsts _ IH G1 s Hsubof.
simp_sub.
apply fstm_lamap; auto using subst_fsts; [].
simp_sub.
so (subof_wc_bind_tof_cn_md _#3 k (shift1 T) Hsubof) as Hsubof'.
simp_sub in Hsubof'.
so (IH _ _ Hsubof') as H.
simp_sub in H.
assumption.

(* lett *)
intros G2 e M c _ IH G1 s Hsubof.
simp_sub.
apply fstm_lett; [].
simp_sub.
so (subof_wc_bind_tof_utm _#3 Hsubof) as Hsubof'.
so (IH _ _ Hsubof') as H.
simp_sub in H.
assumption.
Qed.


Lemma subst_oft_ofm_wc :
  (forall G1 G2 s e t,
     subof (wc dof) G1 s G2
     -> oft G2 e t
     -> oft G1 (subst s e) (subst s t))
  /\
  (forall G1 G2 s p M T,
     subof (wc dof) G1 s G2
     -> ofm G2 p M T
     -> ofm G1 p (subst s M) (subst s T)).
Proof.
assert (forall G1 s G2, subof (wc dof) G1 s G2 -> subof sof G1 s G2) as Hsd.
  intros.
  apply (subof_mono (wc dof)); auto; [].
  intros.
  apply dof_sof; [].
  eapply wc_elim; eauto; done.
assert (forall G1 G2 s c, subof (wc dof) G1 s G2 -> ofc G2 c kd_type -> ofc G1 (subst s c) kd_type) as Hsubtp.
  intros G1 G2 s c Hsubof Hofc.
  repl kd_type with (subst s kd_type) by (simp_sub; reflexivity).
  eapply subst_ofc; eauto; done.
#26 exploit
  (dynamic_ind
     (fun G2 e t =>
        forall G1 s,
          subof (wc dof) G1 s G2
          -> oft G1 (subst s e) (subst s t))
     (fun G2 p M T =>
        forall G1 s,
          subof (wc dof) G1 s G2
          -> ofm G1 p (subst s M) (subst s T))) as Hind;
try (intros;
     simp_sub;
     eauto using subst_ofs with static;
     done).

(* var *)
intros G2 i t Hindex G1 s Hs.
simp_sub.
assert (subof dof G1 s G2) as Hs'.
  apply (subof_mono (wc dof)); auto; [].
  intros; apply wc_elim; auto; done.
so (subof_index _ dof_behaved _#5 iscl_tm Hs' Hindex) as Hsof.
simp_sub in Hsof.
invert Hsof; auto; done.

(* lam *)
intros G2 t1 t2 e Hoft1 _ IH G1 s Hsubof.
simp_sub.
#2 apply oft_lam.
  eapply Hsubtp; eauto; done.

  so (subof_wc_bind_dof_tm _#3 t1 Hsubof) as Hsubof'.
  so (IH _ _ Hsubof') as H.
  simp_sub in H.
  simp_sub.
  assumption.

(* plam *)
intros G2 k t e Hofk _ IH G1 s Hs.
simp_sub.
apply oft_plam; eauto using subst_ofk; [].
apply IH; [].
apply subof_wc_bind_dof_cn; auto; done.

(* papp *)
intros G k t e c _ IH Hofc G1 s Hs.
simp_sub.
so (IH _ _ Hs) as H.
simp_sub in H.
rewrite -> split_dot.
rewrite -> subst_compose.
eapply oft_papp; eauto; [].
eapply subst_ofc; eauto; done.

(* pack *)
intros G k t c e Hofc _ IH Hoft G1 s Hs.
simp_sub.
#2 apply oft_pack; eauto using subst_ofc.
  simp_sub.
  so (IH _ _ Hs) as H.
  simp_sub in H.
  assumption.

  repl kd_type with (subst (dot (var 0) (compose s sh1)) kd_type) by (simp_sub; reflexivity).
  eapply subst_ofc; eauto; [].
  apply subof_bind_sof_cn; [].
  apply Hsd; auto; done.

(* unpack *)
intros G k t t' e1 e2 _ IH1 _ IH2 Hofc G1 s Hs.
simp_sub.
so (IH1 _ _ Hs) as Hof1.
simp_sub in Hof1.
refine (oft_unpack Hof1 _ _); eauto; [].
so (subof_wc_bind_dof_cn _ _ _ k Hs) as Hs'.
so (subof_wc_bind_dof_tm _ _ _ t Hs') as Hs''.
so (IH2 _ _ Hs'') as Hof2.
simp_sub in Hof2.
simp_sub.
assumption.

(* fix *)
intros G t e _ IH G1 s Hs.
simp_sub.
apply oft_fix; eauto; [].
apply IH; auto; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 G1 s Hs.
simp_sub.
eapply oft_lett; eauto; [].
simp_sub.
repl (compose s sh1) with (compose sh1 (dot (var 0) (compose s sh1))) by (simp_sub; reflexivity).
rewrite -> subst_compose.
apply IH2; [].
apply subof_wc_bind_dof_tm; auto; done.

(* equiv *)
intros G t t' e _ IH Hsubs G1 s Hs.
so (subst_eqc _#6 (Hsd _ _ _ Hs) Hsubs).
eapply oft_equiv; eauto; done.

(* letm *)
intros G T k t M e _ IH1 Hfsts _ IH2 Hofc G1 s Hs.
simp_sub.
eapply oft_letm; eauto using subst_fsts; [].
so (IH2 _ _ (subof_wc_bind_dof_cn_md _#5 Hs)) as H.
simp_sub in H.
simp_sub.
assumption.

(* var *)
intros G2 i T c Hindex G1 s Hs.
simp_sub.
assert (subof dof G1 s G2) as Hs'.
  apply (subof_mono (wc dof)); auto; [].
  intros; apply wc_elim; auto; done.
so (subof_index _ dof_behaved _#5 iscl_md Hs' Hindex) as Hsof.
simp_sub in Hsof.
invert Hsof; auto; done.

(* stat *)
intros G c k H G1 s Hs.
simp_sub.
apply ofm_stat; [].
eapply subst_ofc; eauto; done.

(* lamgn *)
intros G T1 T2 k1 M HofT1 Hfsts _ IH G1 s Hs.
simp_sub.
eapply ofm_lamgn; eauto using subst_ofs, subst_fsts; [].
so (IH _ _ (subof_wc_bind_dof_cn_md _#5 Hs)) as H.
simp_sub in H.
simp_sub.
assumption.

(* appgn *)
intros G T1 T2 M1 M2 c2 _ IH1 _ IH2 Hfstm G1 s Hs.
simp_sub.
rewrite -> split_dot, subst_compose.
so (IH1 _ _ Hs) as H.
simp_sub in H.
eapply ofm_appgn; eauto; [].
eapply subst_fstm_wc; eauto; [].
eapply subof_mono; [| eassumption].
intros; eauto using wc_mono, dof_tof; done.

(* lamap *)
intros G T1 T2 k1 M HofT1 Hfsts _ IH G1 s Hs.
simp_sub.
eapply ofm_lamap; eauto using subst_ofs, subst_fsts; [].
so (IH _ _ (subof_wc_bind_dof_cn_md _#5 Hs)) as H.
simp_sub in H.
simp_sub.
assumption.

(* appap *)
intros G p T1 T2 M1 M2 c2 _ IH1 _ IH2 Hfstm G1 s Hs.
simp_sub.
rewrite -> split_dot, subst_compose.
so (IH1 _ _ Hs) as H.
simp_sub in H.
eapply ofm_appap; eauto using subst_fstm_wc; [].
eapply subst_fstm_wc; eauto; [].
eapply subof_mono; [| eassumption].
intros; eauto using wc_mono, dof_tof; done.

(* pair *)
intros G p T1 T2 M1 M2 _ IH1 _ IH2 G1 s Hs.
simp_sub.
rewrite -> subst_compose.
apply ofm_pair; auto; done.

(* pi2 *)
intros G T1 T2 M c _ IH Hfstm G1 s Hs.
simp_sub.
so (IH _ _ Hs) as Hof.
simp_sub in Hof.
#2 sor (refine (ofm_pi2 Hof (subst_fstm_wc _#5 _ Hfstm))) as H.
  eapply subof_mono; [| eassumption].
  intros; eauto using wc_mono, dof_tof; done.
simp_sub in H.
assumption.

(* unpack *)
intros G k t T e M _ IH1 _ IH2 Hofc G1 s Hs.
simp_sub.
so (IH1 _ _ Hs) as Hof1.
simp_sub in Hof1.
refine (ofm_unpack Hof1 _ _); eauto using subst_ofs; [].
so (subof_wc_bind_dof_cn _ _ _ k Hs) as Hs'.
so (subof_wc_bind_dof_tm _ _ _ t Hs') as Hs''.
so (IH2 _ _ Hs'') as Hof2.
simp_sub in Hof2.
simp_sub.
assumption.

(* lett *)
intros G p t T e M _ IH1 _ IH2 G1 s Hs.
simp_sub.
eapply ofm_lett; eauto; [].
simp_sub.
repl (compose s sh1) with (compose sh1 (dot (var 0) (compose s sh1))) by (simp_sub; reflexivity).
rewrite -> subst_compose.
apply IH2; [].
apply subof_wc_bind_dof_tm; auto; done.

(* letm *)
intros G T1 T2 k1 M1 M2 _ IH1 Hfsts _ IH2 Hofs G1 s Hs.
simp_sub.
eapply ofm_letm; eauto using subst_fsts, subst_ofs; [].
so (IH2 _ _ (subof_wc_bind_dof_cn_md _#5 Hs)) as H.
simp_sub in H.
simp_sub.
assumption.

(* extstat *)
intros G k k' M c _ IH Hfstm Hofc G1 s Hs.
simp_sub.
eapply ofm_extstat; eauto using subst_ofc with static; [].
eapply subst_fstm_wc; eauto; [].
eapply subof_mono; [| eassumption].
intros; eauto using wc_mono, dof_tof; done.

(* extpi *)
intros G T1 T2 T2' k1 M _ IH1 Hfsts _ IH2 G1 s Hs.
simp_sub.
eapply ofm_extpi; eauto using subst_fsts; [].
simp_sub.
so (IH2 _ _ (subof_wc_bind_dof_cn_md _#5 Hs)) as H.
simp_sub in H.
assumption.

(* extsigma *)
intros G T1 T2 M _ IH1 _ IH2 G1 s Hs.
simp_sub.
rewrite -> subst_compose.
#2 apply ofm_extsigma.
  so (IH1 _ _ Hs) as H.
  simp_sub in H.
  assumption.

  so (IH2 _ _ Hs) as H.
  simp_sub in H.
  assumption.

(* subsume *)
intros G p T T' M _ IH Hsubs G1 s Hs.
so (subst_subs _#5 (Hsd _ _ _ Hs) Hsubs).
eapply ofm_subsume; eauto; done.

(* epilogue *)
destruct_all Hind.
split; eauto; done.
Qed.


Lemma subst_oft_wc :
  forall G1 G2 s e t,
    subof (wc dof) G1 s G2
    -> oft G2 e t
    -> oft G1 (subst s e) (subst s t).
Proof.
so subst_oft_ofm_wc as H.
destruct_all H.
assumption.
Qed.


Lemma subst_ofm_wc :
  forall G1 G2 s p M T,
    subof (wc dof) G1 s G2
    -> ofm G2 p M T
    -> ofm G1 p (subst s M) (subst s T).
Proof.
so subst_oft_ofm_wc as H.
destruct_all H.
assumption.
Qed.


Lemma subst_dof_wc :
  forall G1 G2 s t A,
    subof (wc dof) G1 s G2
    -> dof G2 t A
    -> dof G1 (subst s t) (subst s A).
Proof.
intros G1 G2 s t A Hs Hdof.
assert (subof sof G1 s G2).
  apply (subof_mono (wc dof)); auto; [].
  intros.
  apply dof_sof; [].
  eapply wc_elim; eauto; done.
assert (subof (wc tof) G1 s G2).
  eapply subof_mono; [| eexact Hs].
  intros; eauto using dof_tof, wc_mono; done.
destruct Hdof; simp_sub; eauto using subst_ofc, subst_oft_wc, subst_ofm_wc, subst_fstm_wc, dof.
Qed.


Lemma subst_tof_wc :
  forall G1 G2 s t A,
    subof (wc tof) G1 s G2
    -> tof G2 t A
    -> tof G1 (subst s t) (subst s A).
Proof.
intros G1 G2 s t A Hs Htof.
destruct Htof; simp_sub; eauto using subst_fstm_wc, tof.
Qed.


Lemma dof_wc :
  forall G t A,
    dof G t A -> wc dof G t A.
Proof.
exact (wc_intro _ subst_dof_wc).
Qed.


Lemma tof_wc :
  forall G t A,
    tof G t A -> wc tof G t A.
Proof.
exact (wc_intro _ subst_tof_wc).
Qed.


Lemma subof_under_tof :
  forall G1 G1' G2 s,
    Forall iscl G2
    -> subof tof G1 s G1'
    -> subof tof (G1;; subst_ctx s G2) (under (length G2) s) (G1';; G2).
Proof.
intros G1 G1' G2 s Hiscl Hsubof.
apply subof_under; auto using tof_behaved; [].
intros; apply tof_wc; auto.
Qed.


Lemma subst_fstm :
  forall G1 G2 s M c,
    subof tof G1 s G2
    -> fstm G2 M c
    -> fstm G1 (subst s M) (subst s c).
Proof.
intros G1 G2 s M c Hsubof Hfst.
eauto using subst_fstm_wc, subof_mono, tof_wc.
Qed.


Lemma subst_oft :
  forall G1 G2 s e t,
    subof dof G1 s G2
    -> oft G2 e t
    -> oft G1 (subst s e) (subst s t).
Proof.
so subst_oft_ofm_wc as H.
destruct_all H.
eauto using subof_mono, dof_wc.
Qed.


Lemma subst_ofm :
  forall G1 G2 s p M T,
    subof dof G1 s G2
    -> ofm G2 p M T
    -> ofm G1 p (subst s M) (subst s T).
Proof.
so subst_oft_ofm_wc as H.
destruct_all H.
eauto using subof_mono, dof_wc.
Qed.


Lemma subst1_oft_cn :
  forall G c k e t,
    ofc G c k
    -> oft (G; cl_cn k) e t
    -> oft G (subst1 c e) (subst1 c t).
Proof.
intros.
eapply subst_oft; eauto; [].
apply subof_dot; auto using subof_id; [].
simp_sub.
apply dof_cn; auto; done.
Qed.


Lemma subst1_oft_tm :
  forall G e t e' t',
    oft G e t
    -> oft (G; cl_tm t) e' t'
    -> oft G (subst1 e e') (subst1 e t').
Proof.
intros.
eapply subst_oft; eauto; [].
apply subof_dot; auto using subof_id; [].
simp_sub.
apply dof_tm; auto; done.
Qed.


Lemma subst1_ofm_cn :
  forall G c k p M T,
    ofc G c k
    -> ofm (G; cl_cn k) p M T
    -> ofm G p (subst1 c M) (subst1 c T).
Proof.
intros.
eapply subst_ofm; eauto; [].
apply subof_dot; auto using subof_id; [].
simp_sub.
apply dof_cn; auto; done.
Qed.


Lemma subst1_ofm_tm :
  forall G e t p M T,
    oft G e t
    -> ofm (G; cl_tm t) p M T
    -> ofm G p (subst1 e M) (subst1 e T).
Proof.
intros.
eapply subst_ofm; eauto; [].
apply subof_dot; auto using subof_id; [].
simp_sub.
apply dof_tm; auto; done.
Qed.


Lemma subst1_fstm_tm :
  forall G e t M c,
    fstm (G; cl_tm t) M c
    -> fstm G (subst1 e M) (subst1 e c).
Proof.
intros G e t M c Hfst.
eapply subst_fstm; eauto; [].
apply subof_dot; auto using subof_id; [].
simp_sub.
apply tof_tm; done.
Qed.


Lemma weaken1_oft :
  forall G A e t,
    oft G e t
    -> oft (G; A) (shift1 e) (shift1 t).
Proof.
intros G A e t Hof.
eapply subst_oft; eauto using subof_sh with sequence.
Qed.


Lemma weaken2_oft :
  forall G A B e t,
    oft G e t
    -> oft (G; A; B) (subst (sh 2) e) (subst (sh 2) t).
Proof.
intros.
eapply subst_oft; eauto using subof_sh with sequence.
Qed.


Lemma weaken1_ofm :
  forall G A p M T,
    ofm G p M T
    -> ofm (G; A) p (shift1 M) (shift1 T).
Proof.
intros G A p M T Hof.
eapply subst_ofm; eauto using subof_sh with sequence.
Qed.


Lemma weaken2_ofm :
  forall G A B p M T,
    ofm G p M T
    -> ofm (G; A; B) p (subst (sh 2) M) (subst (sh 2) T).
Proof.
intros.
eapply subst_ofm; eauto using subof_sh with sequence.
Qed.


Lemma weaken3_ofm :
  forall G A B C p M T,
    ofm G p M T
    -> ofm (G; A; B; C) p (subst (sh 3) M) (subst (sh 3) T).
Proof.
intros.
eapply subst_ofm; eauto using subof_sh with sequence.
Qed.


Lemma oft_conv_0c :
  forall G k k' e t,
    eqk G k k'
    -> oft (G; cl_cn k) e t
    -> oft (G; cl_cn k') e t.
Proof.
intros G k k' e t Heq Hof.
so (subst_oft _#5 (subof_aof_dof _#3 (subof_conv_0_aof _#3 (eqk_symm Heq))) Hof) as H.
simp_sub in H.
assumption.
Qed.


Lemma oft_conv_0t :
  forall G t1 t1' e t2,
    eqc G t1 t1' kd_type
    -> oft (G; cl_tm t1) e t2
    -> oft (G; cl_tm t1') e t2.
Proof.
intros G t1 t1' e t2 Heq Hof.
so (subst_oft _#5 (subof_aof_dof _#3 (subof_conv_0t _#3 (eqc_symm Heq))) Hof) as H.
simp_sub in H.
assumption.
Qed.


Lemma ofm_subsume_01m :
  forall G k k' T T' p M U,
    subk G k' k
    -> subs (G; cl_cn k') T' T
    -> ofm (G; cl_cn k; cl_md T (var 0)) p M U
    -> ofm (G; cl_cn k'; cl_md T' (var 0)) p M U.
Proof.
intros G k k' T T' p M U Hsubk Hsubs Hof.
so (subst_ofm _#6 (subof_aof_dof _#3 (subof_subsume_01m _#5 Hsubk Hsubs)) Hof) as H.
simp_sub in H.
assumption.
Qed.


Lemma ofm_conv_01m :
  forall G k k' T T' p M U,
    eqk G k k'
    -> eqs (G; cl_cn k) T T'
    -> ofm (G; cl_cn k; cl_md T (var 0)) p M U
    -> ofm (G; cl_cn k'; cl_md T' (var 0)) p M U.
Proof.
intros G k k' T T' p M U Heqk Heqs Hof.
eapply ofm_subsume_01m; eauto with static; [].
apply subs_refl; [].
apply eqs_symm; [].
so (subst_eqs _#5 (subof_conv_0 _#3 (eqk_symm Heqk)) Heqs) as H.
simp_sub in H.
assumption.
Qed.


Lemma dof_wclo :
  wclo dof.
Proof.
intros G t C C' H.
invert H; intros; subst; simp_sub; unfold shift1; eauto using dof, subst_ofc, subst_oft, subst_ofm, subst_fstm, weaken1.
Qed.


Lemma subof_weaken_dof :
  forall G1 G2 s G,
    subof dof G1 s G
    -> subof dof (G1;; G2) (compose s (sh (length G2))) G.
Proof.
intros G1 G2 s G Hsubof.
apply subof_wclo_weaken; auto using dof_wclo.
Qed.


Lemma subof_bind_dof_cn :
  forall G G' s k,
    subof dof G s G'
    -> subof dof (G; cl_cn (subst s k)) (dot (var 0) (compose s sh1)) (G'; cl_cn k).
Proof.
intros G G' s k H.
replsub (cl_cn (subst s k)) with (subst s (cl_cn k)).
apply subof_wclo_bind; eauto using dof_behaved, dof_wclo with static.
Qed.


Lemma subof_bind_dof_tm :
  forall G G' s t,
    subof dof G s G'
    -> subof dof (G; cl_tm (subst s t)) (dot (var 0) (compose s sh1)) (G'; cl_tm t).
Proof.
intros G G' s t H.
replsub (cl_tm (subst s t)) with (subst s (cl_tm t)).
apply subof_wclo_bind; eauto using dof_behaved, dof_wclo with static.
Qed.


Lemma subof_bind_dof_cn_tm :
  forall G G' s k t,
    subof dof G s G'
    -> subof dof (G; cl_cn (subst s k); cl_tm (subst (dot (var 0) (compose s sh1)) t)) (dot (var 0) (dot (var 1) (compose s (sh 2)))) (G'; cl_cn k; cl_tm t).
Proof.
intros G G' s k t Hsubof.
repl (cl_tm (subst (dot (var 0) (compose s sh1)) t)) with (subst (dot (var 0) (compose s sh1)) (cl_tm t)) by (simp_sub; reflexivity).
repl (dot (var 0) (dot (var 1) (compose s (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s sh1)) sh1)) by (simp_sub; reflexivity).
apply subof_wclo_bind; auto using dof_behaved, dof_wclo with static; [].
repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
apply subof_wclo_bind; auto using dof_behaved, dof_wclo with static; done.
Qed.


Lemma subof_bind_dof_cn_md :
  forall G G' s k T,
    subof dof G s G'
    -> subof dof (G; cl_cn (subst s k); cl_md (subst (dot (var 0) (compose s sh1)) T) (var 0)) (dot (var 0) (dot (var 1) (compose s (sh 2)))) (G'; cl_cn k; cl_md T (var 0)).
Proof.
intros G G' s k T Hsubof.
repl (cl_md (subst (dot (var 0) (compose s sh1)) T) (var 0)) with (subst (dot (var 0) (compose s sh1)) (cl_md T (var 0))) by (simp_sub; reflexivity).
repl (dot (var 0) (dot (var 1) (compose s (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s sh1)) sh1)) by (simp_sub; reflexivity).
apply subof_wclo_bind; auto using dof_behaved, dof_wclo with static; [].
repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
apply subof_wclo_bind; auto using dof_behaved, dof_wclo with static; done.
Qed.


Lemma subst_tvalue :
  forall G1 G2 s v,
    subof oof G1 s G2
    -> tvalue G2 v
    -> tvalue G1 (subst s v).
Proof.
intros G1 G2 s v Hs Hval.
revert G1 s Hs.
#2 induct Hval; try (intros; eauto using tvalue; done).
  (* var *)
  intros G2 i t Hindex G1 s Hs.
  simp_sub.
  so (subof_index _ oof_behaved _#5 iscl_tm Hs Hindex) as Haof.
  simp_sub in Haof.
  invert Haof; auto; done.

  (* var *)
  intros G2 i Hindex G1 s Hs.
  simp_sub.
  so (subof_index _ oof_behaved _#5 iscl_utm Hs Hindex) as Haof.
  simp_sub in Haof.
  invert Haof; auto; done.
Qed.


Lemma subst_mvalue :
  forall G1 G2 s V,
    subof oof G1 s G2
    -> mvalue G2 V
    -> mvalue G1 (subst s V).
Proof.
intros G1 G2 s V Hs Hval.
revert G1 s Hs.
#2 induct Hval; try (intros; eauto using subst_tvalue, mvalue; done).
  (* var *)
  intros G2 i T c Hindex G1 s Hs.
  simp_sub.
  so (subof_index _ oof_behaved _#5 iscl_md Hs Hindex) as Hoof.
  simp_sub in Hoof.
  invert Hoof; auto; done.

  (* var *)
  intros G2 i c Hindex G1 s Hs.
  simp_sub.
  so (subof_index _ oof_behaved _#5 iscl_umd Hs Hindex) as Hoof.
  simp_sub in Hoof.
  invert Hoof; auto; done.
Qed.


Lemma aof_wclo :
  wclo aof.
Proof.
intros G t C C' H.
invert H; try (intros; subst; simp_sub; eauto 9 using aof, weaken1_ofc; done).

(* tm *)
intros; subst; simp_sub.
#2 apply aof_tm.
  eapply subst_oft; eauto using subof with sequence; done.

  eapply subst_tvalue; eauto using subof with sequence; done.

(* md *)
intros; subst; simp_sub.
#3 apply aof_md.
  eapply subst_ofm; eauto using subof with sequence; done.

  eapply subst_mvalue; eauto using subof with sequence; done.

  eapply subst_fstm; eauto; [].
  #2 eapply subof_mono.
    eauto using oof_tof; done.

    apply subof_sh; auto with sequence; done.

(* utm *)
intros; subst; simp_sub.
apply aof_utm; [].
eapply subst_tvalue; eauto using subof with sequence; done.

(* umd *)
intros; subst; simp_sub.
#2 apply aof_umd.
  eapply subst_mvalue; eauto using subof with sequence; done.

  eapply subst_fstm; eauto; [].
  #2 eapply subof_mono.
    eauto using aof_tof; done.

    apply subof_sh; auto with sequence; done.
Qed.


Lemma subof_bind_aof_tm :
  forall G G' s t,
    subof aof G s G'
    -> subof aof (G; cl_tm (subst s t)) (dot (var 0) (compose s sh1)) (G'; cl_tm t).
Proof.
intros G G' s t Hsubof.
repl (cl_tm (subst s t)) with (subst s (cl_tm t)) by (simp_sub; reflexivity).
apply subof_wclo_bind; eauto using aof_behaved, aof_wclo with static.
Qed.


Lemma subof_bind_aof_cn :
  forall G G' s k,
    subof aof G s G'
    -> subof aof (G; cl_cn (subst s k)) (dot (var 0) (compose s sh1)) (G'; cl_cn k).
Proof.
intros G G' s k Hsubof.
repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
apply subof_wclo_bind; eauto using aof_behaved, aof_wclo with static.
Qed.


Lemma subof_bind_aof_md :
  forall G G' s T c,
    subof aof G s G'
    -> subof aof (G; cl_md (subst s T) (subst s c)) (dot (var 0) (compose s sh1)) (G'; cl_md T c).
Proof.
intros G G' s T c Hsubof.
repl (cl_md (subst s T) (subst s c)) with (subst s (cl_md T c)) by (simp_sub; reflexivity).
apply subof_wclo_bind; eauto using aof_behaved, aof_wclo, iscl_md.
Qed.


Lemma subof_bind_aof_cn_tm :
  forall G G' s k t,
    subof aof G s G'
    -> subof aof (G; cl_cn (subst s k); cl_tm (subst (dot (var 0) (compose s sh1)) t)) (dot (var 0) (dot (var 1) (compose s (sh 2)))) (G'; cl_cn k; cl_tm t).
Proof.
intros G G' s k t Hsubof.
repl (cl_tm (subst (dot (var 0) (compose s sh1)) t)) with (subst (dot (var 0) (compose s sh1)) (cl_tm t)) by (simp_sub; reflexivity).
repl (dot (var 0) (dot (var 1) (compose s (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s sh1)) sh1)) by (simp_sub; reflexivity).
apply subof_wclo_bind; auto using aof_behaved, aof_wclo with static; [].
repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
apply subof_wclo_bind; auto using aof_behaved, aof_wclo with static; done.
Qed.


Lemma subof_bind_aof_cn_md :
  forall G G' s k T,
    subof aof G s G'
    -> subof aof (G; cl_cn (subst s k); cl_md (subst (dot (var 0) (compose s sh1)) T) (var 0)) (dot (var 0) (dot (var 1) (compose s (sh 2)))) (G'; cl_cn k; cl_md T (var 0)).
Proof.
intros G G' s k T Hsubof.
repl (cl_md (subst (dot (var 0) (compose s sh1)) T) (var 0)) with (subst (dot (var 0) (compose s sh1)) (cl_md T (var 0))) by (simp_sub; reflexivity).
repl (dot (var 0) (dot (var 1) (compose s (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s sh1)) sh1)) by (simp_sub; reflexivity).
apply subof_wclo_bind; auto using aof_behaved, aof_wclo with static; [].
repl (cl_cn (subst s k)) with (subst s (cl_cn k)) by (simp_sub; reflexivity).
apply subof_wclo_bind; auto using aof_behaved, aof_wclo with static; done.
Qed.


Lemma subof_bind_aof_cn_md' :
  forall G G' s k T,
    subof aof G s G'
    -> subof aof (G; cl_cn (subst s k); cl_md (shift1 (subst s T)) (var 0)) (dot (var 0) (dot (var 1) (compose s (sh 2)))) (G'; cl_cn k; cl_md (shift1 T) (var 0)).
Proof.
intros G G' s k T Hsubof.
repl (shift1 (subst s T)) with (subst (dot (var 0) (compose s sh1)) (shift1 T)) by (simp_sub; reflexivity).
apply subof_bind_aof_cn_md; assumption.
Qed.


Lemma subst_tstep_mstep :
  (forall G1 s G2 e e',
     subof oof G1 s G2
     -> tstep G2 e e'
     -> tstep G1 (subst s e) (subst s e'))
  /\
  (forall G1 s G2 M M',
     subof oof G1 s G2
     -> mstep G2 M M'
     -> mstep G1 (subst s M) (subst s M')).
Proof.
#12 exploit
  (step_ind
     (fun G2 e e' =>
        forall G1 s,
          subof oof G1 s G2
          -> tstep G1 (subst s e) (subst s e'))
     (fun G2 M M' =>
        forall G1 s,
          subof oof G1 s G2
          -> mstep G1 (subst s M) (subst s M'))) as Hind;
try (intros;
     simp_sub;
     eauto using subst_tvalue, subst_mvalue with dynamic;
     done).

(* app3 *)
intros G t e1 v2 Hval G1 s Hs.
simp_sub.
rewrite -> (split_dot _ s).
rewrite -> subst_compose.
apply tstep_app3; eauto using subst_tvalue; done.

(* papp2 *)
intros G k e c G1 s Hs.
simp_sub.
rewrite -> (split_dot _ s).
rewrite -> subst_compose.
apply tstep_papp2; eauto using subst_tvalue; done.

(* unpack2 *)
intros G c v k t e Hval G1 s Hs.
simp_sub.
so (@tstep_unpack2 G1 (subst s c) (subst s v) (subst s k) (subst (under 1 s) t) (subst (under 2 s) e) (subst_tvalue _#4 Hs Hval)) as H.
simp_sub in H.
assumption.

(* fix *)
intros G t e G1 s Hs.
simp_sub.
so (@tstep_fix G1 (subst s t) (subst s e)) as H.
simp_sub in H.
assumption.

(* lett *)
intros G v1 e2 Hval G1 s Hs.
simp_sub.
rewrite -> (split_dot _ s), -> subst_compose.
apply tstep_lett2; eauto using subst_tvalue; done.

(* letm *)
intros G V e c Hval Hfst G1 s Hs.
simp_sub.
so (@tstep_letm2 G1 (subst s V) (subst (under 2 s) e) (subst s c) (subst_mvalue _#4 Hs Hval) (subst_fstm _#5 (subof_oof_tof _#3 Hs) Hfst)) as H.
simp_sub in H.
assumption.

(* appgn *)
intros G T M1 V2 c2 Hval Hfst G1 s Hs.
simp_sub.
so (@mstep_appgn3 G1 (subst s T) (subst (under 2 s) M1) (subst s V2) (subst s c2) (subst_mvalue _#4 Hs Hval) (subst_fstm _#5 (subof_oof_tof _#3 Hs) Hfst)) as H.
simp_sub in H.
assumption.

(* appap *)
intros G T M1 V2 c2 Hval Hfst G1 s Hs.
simp_sub.
so (@mstep_appap3 G1 (subst s T) (subst (under 2 s) M1) (subst s V2) (subst s c2) (subst_mvalue _#4 Hs Hval) (subst_fstm _#5 (subof_oof_tof _#3 Hs) Hfst)) as H.
simp_sub in H.
assumption.

(* unpack2 *)
intros G c v k t M T Hval G1 s Hs.
simp_sub.
so (@mstep_unpack2 G1 (subst s c) (subst s v) (subst s k) (subst (under 1 s) t) (subst (under 2 s) M) (subst s T) (subst_tvalue _#4 Hs Hval)) as H.
simp_sub in H.
assumption.

(* lett2 *)
intros G v1 M2 Hval G1 s Hs.
simp_sub.
rewrite -> (split_dot _ s), -> subst_compose.
apply mstep_lett2; eauto using subst_tvalue; done.

(* letm *)
intros G V M T c Hval Hfst G1 s Hs.
simp_sub.
so (@mstep_letm2 G1 (subst s V) (subst (under 2 s) M) (subst s T) (subst s c) (subst_mvalue _#4 Hs Hval) (subst_fstm _#5 (subof_oof_tof _#3 Hs) Hfst)) as H.
simp_sub in H.
assumption.

(* epilogue *)
destruct_all Hind.
split; eauto; done.
Qed.


Lemma subst_tstep :
  forall G1 s G2 e e',
    subof oof G1 s G2
    -> tstep G2 e e'
    -> tstep G1 (subst s e) (subst s e').
Proof.
exact (subst_tstep_mstep andel).
Qed.


Lemma subst_mstep :
  forall G1 s G2 M M',
    subof oof G1 s G2
    -> mstep G2 M M'
    -> mstep G1 (subst s M) (subst s M').
Proof.
exact (subst_tstep_mstep ander).
Qed.


Lemma oof_wclo :
  wclo oof.
Proof.
intros G t C C' H.
invert H; try (intros; subst; simp_sub; eauto 9 using oof; done).

(* tm *)
intros; subst; simp_sub.
apply oof_tm; [].
eapply subst_tvalue; eauto using subof with sequence; done.

(* md *)
intros; subst; simp_sub.
#2 apply oof_md.
  eapply subst_mvalue; eauto using subof with sequence; done.

  eapply subst_fstm; eauto; [].
  #2 eapply subof_mono.
    eauto using oof_tof; done.

    apply subof_sh; auto with sequence; done.

(* utm *)
intros; subst; simp_sub.
apply oof_utm; [].
eapply subst_tvalue; eauto using subof with sequence; done.

(* umd *)
intros; subst; simp_sub.
#2 apply oof_umd.
  eapply subst_mvalue; eauto using subof with sequence; done.

  eapply subst_fstm; eauto; [].
  #2 eapply subof_mono.
    eauto using oof_tof; done.

    apply subof_sh; auto with sequence; done.
Qed.


Lemma subst_sof :
  forall G1 G2 s m C,
    subof sof G1 s G2
    -> sof G2 m C
    -> sof G1 (subst s m) (subst s C).
Proof.
intros G1 G2 s m C Hs H.
destruct H; simp_sub; eauto 8 using sof, subst_ofc, subst_ofk.
Qed.


Lemma subof_sof_trans :
  forall G1 G2 G3 s s',
    subof sof G1 s G2
    -> subof sof G2 s' G3
    -> subof sof G1 (compose s' s) G3.
Proof.
intros G1 G2 G3 s s' H12 H23.
revert G1 s H12.
#2 induct H23.

(* dot *)
intros G2 t s' C G3 _ IH Hsof G1 s H12.
simp_sub.
apply subof_dot; auto; [].
rewrite -> subst_compose; [].
eapply subst_sof; eauto; done.

(* sh *)
intros G2 n G3 Htrunc.
#2 induct Htrunc.
  (* 0 *)
  intros G2 G1 s H12.
  simp_sub.
  assumption.

  (* S *)
  intros i C G2 G3 Htrunc IH G1 s H12.
  #2 invert H12.
    intros m s' Hs Hsof <-.
    simp_sub.
    auto; done.

    intros n Htrunc' <-.
    simp_sub.
    apply subof_sh; [].
    repl (1 + (i + n)) with (n + S i) by omega.
    eapply truncate_sum; eauto; [].
    apply truncate_S; auto; done.
Qed.


Lemma subst_dof :
  forall G1 G2 s m C,
    subof dof G1 s G2
    -> dof G2 m C
    -> dof G1 (subst s m) (subst s C).
Proof.
intros G1 G2 s m C Hs H.
destruct H; simp_sub; eauto 8 using dof, subst_ofc, subst_ofk, subst_oft, subst_ofm, subst_fstm, subst_tvalue, subst_mvalue, subof_dof_sof, subof_dof_tof.
Qed.


Lemma subof_dof_trans :
  forall G1 G2 G3 s s',
    subof dof G1 s G2
    -> subof dof G2 s' G3
    -> subof dof G1 (compose s' s) G3.
Proof.
intros G1 G2 G3 s s' H12 H23.
revert G1 s H12.
#2 induct H23.

(* dot *)
intros G2 t s' C G3 _ IH Hdof G1 s H12.
simp_sub.
apply subof_dot; auto; [].
rewrite -> subst_compose; [].
eapply subst_dof; eauto; done.

(* sh *)
intros G2 n G3 Htrunc.
#2 induct Htrunc.
  (* 0 *)
  intros G2 G1 s H12.
  simp_sub.
  assumption.

  (* S *)
  intros i C G2 G3 Htrunc IH G1 s H12.
  #2 invert H12.
    intros m s' Hs Hdof <-.
    simp_sub.
    auto; done.

    intros n Htrunc' <-.
    simp_sub.
    apply subof_sh; [].
    repl (1 + (i + n)) with (n + S i) by omega.
    eapply truncate_sum; eauto; [].
    apply truncate_S; auto; done.
Qed.


Lemma subst_aof :
  forall G1 G2 s m C,
    subof aof G1 s G2
    -> aof G2 m C
    -> aof G1 (subst s m) (subst s C).
Proof.
intros G1 G2 s m C Hs H.
destruct H; simp_sub; eauto 8 using aof, subst_ofc, subst_ofk, subst_oft, subst_ofm, subst_fstm, subst_tvalue, subst_mvalue, subof_aof_sof, subof_aof_dof, subof_aof_oof, subof_aof_tof.
Qed.


Lemma subof_aof_trans :
  forall G1 G2 G3 s s',
    subof aof G1 s G2
    -> subof aof G2 s' G3
    -> subof aof G1 (compose s' s) G3.
Proof.
intros G1 G2 G3 s s' H12 H23.
revert G1 s H12.
#2 induct H23.

(* dot *)
intros G2 t s' C G3 _ IH Haof G1 s H12.
simp_sub.
apply subof_dot; auto; [].
rewrite -> subst_compose; [].
eapply subst_aof; eauto; done.

(* sh *)
intros G2 n G3 Htrunc.
#2 induct Htrunc.
  (* 0 *)
  intros G2 G1 s H12.
  simp_sub.
  assumption.

  (* S *)
  intros i C G2 G3 Htrunc IH G1 s H12.
  #2 invert H12.
    intros m s' Hs Haof <-.
    simp_sub.
    auto; done.

    intros n Htrunc' <-.
    simp_sub.
    apply subof_sh; [].
    repl (1 + (i + n)) with (n + S i) by omega.
    eapply truncate_sum; eauto; [].
    apply truncate_S; auto; done.
Qed.
