
Require Import Tactics.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import Regularity.
Require Import Relation.


Lemma tstep_value_contra :
  forall G e e',
    tstep G e e'
    -> tvalue G e
    -> False.
Proof.
intros G e e' Hstep Hval.
revert e' Hstep.
induct Hval; try (intros; invert Hstep; eauto; done).
Qed.


Lemma mstep_value_contra :
  forall G M M',
    mstep G M M'
    -> mvalue G M
    -> False.
Proof.
intros G M M' Hstep Hval.
revert M' Hstep.
induct Hval; try (intros; invert Hstep; eauto using tstep_value_contra; done).
Qed.


Lemma step_fun :
  (forall G e e' e'',
     tstep G e e'
     -> tstep G e e''
     -> e' = e'')
  /\
  (forall G M M' M'',
     mstep G M M'
     -> mstep G M M''
     -> M' = M'').
Proof.
#5 exploit
  (step_ind
     (fun G e e' => forall e'', tstep G e e'' -> e' = e'')
     (fun G M M' => forall M'', mstep G M M'' -> M' = M'')) as Hind;
try
  (intros;
   repeat
     (match goal with
      | H : tstep _ ?e _ |- _ =>
        isapp e; invert H; clear H; intros; subst

      | H : mstep _ ?M _ |- _ =>
        isapp M; invert H; clear H; intros; subst
      end);
   try
     (match goal with
      | H1 : tstep _ ?e _, H2 : tvalue _ ?e |- _ =>
        exfalso; exact (tstep_value_contra _#3 H1 H2)

      | H1 : mstep _ ?M _, H2 : mvalue _ ?M |- _ =>
        exfalso; exact (mstep_value_contra _#3 H1 H2)
      end);
   f_equal; eauto; done).

(* letm2 *)
intros G V e c Hval Hfst e'' Hstep.
#2 invert Hstep.
  intros M1'' Hstep1' <-.
  exfalso; eapply mstep_value_contra; eauto; done.

  intros c' _ Hfst' <-.
  so (fstm_fun _#4 Hfst Hfst'); subst c'.
  reflexivity.

(* appgn3 *)
intros G T M1 V2 c2 Hval Hfst M'' Hstep.
#3 invert Hstep; clear Hstep.
  intros M1' Hstep1' <-.
  invert Hstep1'; done.

  intros; exfalso; eapply mstep_value_contra; eauto; done.

  intros c2' _ Hfst' <-.
  so (fstm_fun _#4 Hfst Hfst'); subst c2'.
  reflexivity.

(* appap3 *)
intros G T M1 V2 c2 Hval Hfst M'' Hstep.
#3 invert Hstep; clear Hstep.
  intros M1' Hstep1' <-.
  invert Hstep1'; done.

  intros; exfalso; eapply mstep_value_contra; eauto; done.

  intros c2' _ Hfst' <-.
  so (fstm_fun _#4 Hfst Hfst'); subst c2'.
  reflexivity.

(* letm2 *)
intros G V M T c Hval Hfst M'' Hstep.
#2 invert Hstep; clear Hstep.
  intros M1'' Hstep1' <-.
  exfalso; eapply mstep_value_contra; eauto; done.

  intros c' _ Hfst' <-.
  so (fstm_fun _#4 Hfst Hfst'); subst c'.
  reflexivity.

(* epilogue *)
destruct Hind; split; intros; eauto; done.
Qed.


Lemma tstep_fun :
  forall G e e' e'',
    tstep G e e'
    -> tstep G e e''
    -> e' = e''.
Proof.
exact (step_fun andel).
Qed.


Lemma mstep_fun :
  forall G M M' M'',
    mstep G M M'
    -> mstep G M M''
    -> M' = M''.
Proof.
exact (step_fun ander).
Qed.


Definition deterministic (value : term -> Prop) (step : term -> term -> Prop) :=
  (forall m m', step m m' -> value m -> False)
  /\ (forall m m1 m2, step m m1 -> step m m2 -> m1 = m2).



Lemma deterministic_tm : forall G, deterministic (tvalue G) (tstep G).
Proof.
intro G.
split; eauto using tstep_value_contra, tstep_fun.
Qed.



Lemma deterministic_md : forall G, deterministic (mvalue G) (mstep G).
Proof.
intro G.
split; eauto using mstep_value_contra, mstep_fun.
Qed.



Definition halt (value : term -> Prop) (step : term -> term -> Prop) (m : term) :=
  exists v, star step m v /\ value v.


Lemma steps_halt :
  forall value step ,
    deterministic value step
    -> forall m m',
         star step m m'
         -> halt value step m
         -> halt value step m'.
Proof.
intros value step (Hdetv & Hdet) e e' Hsteps Hhalt.
destruct Hhalt as (v & Hsteps' & Hval).
revert Hsteps'.
induct Hsteps.

(* refl *)
intros e Hsteps'.
eexists; eauto; done.

(* step *)
intros e1 e2 e3 Hstep _ IH Hsteps'.
#2 invert Hsteps'.
  intros ->.
  exfalso.
  eapply Hdetv; eauto; done.

  intros e1' Hstep'.
  so (Hdet _#3 Hstep Hstep'); subst e1'.
  assumption.
Qed.    


Lemma steps_confluent :
  forall value step,
    deterministic value step
    -> forall e e1 e2,
         star step e e1
         -> star step e e2
         -> star step e1 e2 \/ star step e2 e1.
Proof.
intros value step (_ & Hdet) e e1 e2 Hsteps1 Hsteps2.
revert e2 Hsteps2.
induct Hsteps1.

(* refl *)
intros e e2 Hsteps2.
left; assumption.

(* step *)
intros e e1' e1 Hstep1 Hsteps1 IH e2 Hsteps2.
so (star_plus _#4 Hsteps2) as [Heq | Hsteps2'].
  subst e2.
  right.
  eapply star_step; eauto; done.

  destruct Hsteps2' as (e2' & Hstep2 & Hsteps2').
  so (Hdet _#3 Hstep1 Hstep2); subst e2'.
  exact (IH _ Hsteps2').
Qed.


Lemma steps_confluent_val :
  forall value step,
    deterministic value step
    -> forall e e' v,
         star step e e'
         -> star step e v
         -> value v
         -> star step e' v.
Proof.
intros value step Hdet e d v Hstepsed Hstepsev Hval.
so (steps_confluent _ _ Hdet _#3 Hstepsed Hstepsev) as [H | Hstepvd]; auto; [].
so (star_plus _#4 Hstepvd) as [Heq | Hstepvd'].
  subst d.
  apply star_refl; done.

  exfalso.
  destruct Hstepvd' as (x & Hstep & _).
  eapply (Hdet andel); eauto; done.
Qed.


Lemma value_unique :
  forall value step,
    deterministic value step
    -> forall e v1 v2,
         star step e v1
         -> star step e v2
         -> value v1
         -> value v2
         -> v1 = v2.
Proof.
intros value step Hdet e v1 v2 Hsteps1 Hsteps2 Hval1 Hval2.
so (steps_confluent_val _ _ Hdet _#3 Hsteps1 Hsteps2 Hval2) as Hsteps.
so (star_plus _#4 Hsteps) as [Heq | Hsteps'].
  assumption.

  exfalso.
  destruct Hsteps' as (x & Hstep & _).
  eapply (Hdet andel); eauto; done.
Qed.


(* Not actually a determinism property, but naturally grouped with these. *)
Lemma anti_steps_halt :
  forall (value : term -> Prop) (step : term -> term -> Prop) e e',
    star step e' e
    -> halt value step e
    -> halt value step e'.
Proof.
intros value step e e' Hsteps Hhalt.
destruct Hhalt as (v & Hsteps' & Hval).
eexists; eauto using star_trans.
Qed.


Lemma steps_halt_iff :
  forall value step,
    deterministic value step
    -> forall e e',
         star step e e'
         -> (halt value step e <-> halt value step e').
Proof.
intros value step Hdet e e' Hsteps.
split.
  intro H.
  eapply steps_halt; eauto; done.

  intro H.
  eapply anti_steps_halt; eauto; done.
Qed.


(* Not actually a determinism property, but naturally grouped with these. *)
Lemma value_halt :
  forall (value : term -> Prop) (step : term -> term -> Prop)  v,
    value v
    -> halt value step v.
Proof.
intros value step v Hval.
exists v.
auto using star_refl.
Qed.



(* Instances *)

Lemma tsteps_thalt :
  forall G e e',
    star (tstep G) e e'
    -> thalt G e
    -> thalt G e'.
Proof.
intro G.
exact (steps_halt _ _ (deterministic_tm G)).
Qed.


Lemma msteps_mhalt :
  forall G e e',
    star (mstep G) e e'
    -> mhalt G e
    -> mhalt G e'.
Proof.
intro G.
exact (steps_halt _ _ (deterministic_md G)).
Qed.


Lemma tsteps_confluent :
  forall G e e1 e2,
    star (tstep G) e e1
    -> star (tstep G) e e2
    -> star (tstep G) e1 e2 \/ star (tstep G) e2 e1.
Proof.
intro G.
exact (steps_confluent _ _ (deterministic_tm G)).
Qed.


Lemma tsteps_confluent_val :
  forall G e e' v,
    star (tstep G) e e'
    -> star (tstep G) e v
    -> tvalue G v
    -> star (tstep G) e' v.
Proof.
intro G.
exact (steps_confluent_val _ _ (deterministic_tm G)).
Qed.


Lemma tvalue_unique :
  forall G e v1 v2,
    star (tstep G) e v1
    -> star (tstep G) e v2
    -> tvalue G v1
    -> tvalue G v2
    -> v1 = v2.
Proof.
intro G.
exact (value_unique _ _ (deterministic_tm G)).
Qed.


Lemma anti_tsteps_thalt :
  forall G e e',
    star (tstep G) e' e
    -> thalt G e
    -> thalt G e'.
Proof.
intro G.
exact (anti_steps_halt _ _).
Qed.


Lemma anti_msteps_mhalt :
  forall G e e',
    star (mstep G) e' e
    -> mhalt G e
    -> mhalt G e'.
Proof.
intro G.
exact (anti_steps_halt _ _).
Qed.


Lemma tsteps_thalt_iff :
  forall G e e',
    star (tstep G) e e'
    -> (thalt G e <-> thalt G e').
Proof.
intro G.
exact (steps_halt_iff _ _ (deterministic_tm G)).
Qed.


Lemma msteps_mhalt_iff :
  forall G e e',
    star (mstep G) e e'
    -> (mhalt G e <-> mhalt G e').
Proof.
intro G.
exact (steps_halt_iff _ _ (deterministic_md G)).
Qed.


Lemma tvalue_thalt :
  forall G v,
    tvalue G v
    -> thalt G v.
Proof.
intro G.
exact (value_halt _ _).
Qed.
