
Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import TSubst.
Require Compare_dec.


(* statically equivalent (strong) *)
Inductive seqs : context -> term -> term -> (* cl *) term -> Prop :=
| seqs_cn {G c c' k}
    : ofc G c k
      -> eqc G c c' k
      -> seqs G c c' (cl_cn k)

| seqs_tm {G e e' t}
    : seqs G e e' (cl_tm t)

| seqs_md {G M M' T c}
    : seqs G M M' (cl_md T c)

| seqs_ucn {G c c'}
    : seqs G c c' cl_ucn

| seqs_utm {G t t'}
    : seqs G t t' cl_utm

| seqs_umd {G M M' c}
    : seqs G M M' (cl_umd c)
.


Lemma seqs_sof :
  forall G m m' C,
    seqs G m m' C
    -> sof G m C.
Proof.
intros G m m' C H.
invert H; intros; subst; simp_sub; eauto using sof; done.
Qed.


Lemma seqs_behaved :
  behavedrel seqs.
Proof.
intros G i C HC Hindex.
destruct HC; simp_sub; eauto using seqs with static.
Qed.


Lemma seqs_wclo :
  wclorel seqs.
Proof.
intros G m m' C C' H.
invert H; try (intros; subst; simp_sub; eauto using seqs; done).

(* cn *)
intros k Hofc Heqc <-.
simp_sub.
#2 apply seqs_cn.
  apply weaken1_ofc; assumption.

  apply weaken1_eqc; assumption.
Qed.


Lemma seqs_ofk_ofc :
  (forall G1 s s' G2 k,
     subrel seqs G1 s s' G2
     -> ofk G2 k
     -> eqk G1 (subst s k) (subst s' k))
  /\
  (forall G1 s s' G2 c k,
     subrel seqs G1 s s' G2
     -> ofc G2 c k
     -> eqc G1 (subst s c) (subst s' c) (subst s k)).
Proof.
assert (forall G1 G2 s s' k,
          subrel seqs G1 s s' G2
          -> subrel seqs (G1; cl_cn (subst s k)) (dot (var 0) (compose s sh1)) (dot (var 0) (compose s' sh1)) (G2; cl_cn k)) as Hbind.
  intros G1 G2 s s' k Hs.
  so (subrel_wclo_bind _ seqs_behaved seqs_wclo _#4 (cl_cn k) iscl_cn Hs) as H.
  simp_sub in H.
  assumption.
#8 exploit
  (static_ind
     (fun G2 k =>
        forall G1 s s',
          subrel seqs G1 s s' G2
          -> eqk G1 (subst s k) (subst s' k))
     (fun _ _ _ => True)
     (fun _ _ _ => True)
     (fun G2 c k =>
        forall G1 s s',
          subrel seqs G1 s s' G2
          -> eqc G1 (subst s c) (subst s' c) (subst s k))
     (fun _ _ _ _ => True)) as Hind;
eauto;
try (intros; simp_sub; eauto using subrel, seqs, iscl with static; done).

(* var *)
intros G i k Hindex G1 s s' Hs.
simp_sub.
so (subrel_index _ seqs_behaved _#6 iscl_cn Hs Hindex) as Hseqs.
simp_sub in Hseqs.
invert Hseqs; auto; done.

(* app *)
intros G k1 k2 c1 c2 _ IH1 _ IH2 G1 s s' Hs.
simp_sub.
rewrite -> split_dot; [].
rewrite -> subst_compose; [].
eapply eqc_app; eauto; [].
so (IH1 _ _ _ Hs) as H.
simp_sub in H.
assumption.

(* pair *)
intros G k1 k2 c1 c2 _ IH1 _ IH2 Hofk _ G1 s s' Hs.
simp_sub.
#2 apply eqc_pair; eauto.
  simp_sub.
  so (IH2 _ _ _ Hs) as H.
  simp_sub in H.
  assumption.

  eapply subst_ofk; eauto; [].
  refine (subrel_subof_mono _#7 (Hbind _#5 Hs)); [].
  eauto using seqs_sof; done.

(* pi2 *)
intros G k1 k2 c _ IH G1 s s' Hs.
simp_sub.
rewrite -> split_dot; rewrite -> subst_compose; [].
eapply eqc_pi2; eauto; [].
so (IH _ _ _ Hs) as H.
simp_sub in H.
eassumption.

(* extpi *)
intros G k1 k2 k3 c _ IH1 _ IH2 G1 s s' Hs.
simp_sub.
eapply eqc_extpiw; eauto; [].
simp_sub.
so (IH2 _ _ _ (Hbind _#5 Hs)) as H.
simp_sub in H.
assumption.

(* extsigma *)
intros G k1 k2 c _ IH1 _ IH2 Hofk _ G1 s s' Hs.
simp_sub.
#2 eapply eqc_extsigma; eauto.
  simp_sub.
  so (IH2 _ _ _ Hs) as H.
  simp_sub in H.
  assumption.

  eapply subst_ofk; eauto; [].
  refine (subrel_subof_mono _#7 (Hbind _#5 Hs)); [].
  eauto using seqs_sof; done.

(* subsume *)
intros G k k' c _ IH Hsubk _ G1 s s' Hs.
so (IH _ _ _ Hs) as H.
eapply eqc_subsume; eauto; [].
eapply subst_subk; eauto; [].
refine (subrel_subof_mono _#7 Hs); [].
eauto using seqs_sof; done.

(* epilogue *)
destruct_all Hind.
eauto; done.
Qed.


Lemma seqs_ofk :
  forall G1 s s' G2 k,
    subrel seqs G1 s s' G2
    -> ofk G2 k
    -> eqk G1 (subst s k) (subst s' k).
Proof.
exact (seqs_ofk_ofc andel).
Qed.


Lemma seqs_ofc :
  forall G1 s s' G2 c k,
    subrel seqs G1 s s' G2
    -> ofc G2 c k
    -> eqc G1 (subst s c) (subst s' c) (subst s k).
Proof.
exact (seqs_ofk_ofc ander).
Qed.


Lemma seqs_ofs :
  forall G1 s s' G2 T,
    subrel seqs G1 s s' G2
    -> ofs G2 T
    -> eqs G1 (subst s T) (subst s' T).
Proof.
assert (forall G1 G2 s s' k,
          subrel seqs G1 s s' G2
          -> subrel seqs (G1; cl_cn (subst s k)) (dot (var 0) (compose s sh1)) (dot (var 0) (compose s' sh1)) (G2; cl_cn k)) as Hbind.
  intros G1 G2 s s' k Hs.
  so (subrel_wclo_bind _ seqs_behaved seqs_wclo _#4 (cl_cn k) iscl_cn Hs) as H.
  simp_sub in H.
  assumption.
intros G1 s s' G2 T Hs HofT.
revert G1 s s' Hs.
induct HofT;
eauto;
try (intros; simp_sub; eauto 6 using subrel, seqs, iscl, seqs_ofk, subst_fsts with static; done); [].

(* dyn *)
intros G t Hof G1 s s' Hs.
simp_sub.
apply eqs_dyn; [].
repl kd_type with (subst s kd_type) by (simp_sub; reflexivity).
eapply seqs_ofc; eauto; done.
Qed.


Lemma funct_kd_reg :
  forall G k c1 c2 k',
    ofc G c1 k
    -> eqc G c1 c2 k
    -> ofk (G; cl_cn k) k'
    -> eqk G (subst1 c1 k') (subst1 c2 k').
Proof.
intros G k c1 c2 k' Hofc Heqc H.
eapply seqs_ofk; eauto; [].
apply subrel_dot; eauto using subrel_sh with sequence; [].
simp_sub.
apply seqs_cn; auto; done.
Qed.


Lemma funct_cn_reg :
  forall G k c1 c2 c' k',
    ofc G c1 k
    -> eqc G c1 c2 k
    -> ofc (G; cl_cn k) c' k'
    -> eqc G (subst1 c1 c') (subst1 c2 c') (subst1 c1 k').
Proof.
intros G k c1 c2 c' k' Hofs Heqc H.
eapply seqs_ofc; eauto; [].
apply subrel_dot; eauto using subrel_sh with sequence; [].
simp_sub.
apply seqs_cn; auto; done.
Qed.


Lemma funct_sg_reg :
  forall G k c1 c2 T,
    ofc G c1 k
    -> eqc G c1 c2 k
    -> ofs (G; cl_cn k) T
    -> eqs G (subst1 c1 T) (subst1 c2 T).
Proof.
intros G k c1 c2 T Hofc Heqc H.
eapply seqs_ofs; eauto; [].
apply subrel_dot; eauto using subrel_sh with sequence; [].
simp_sub.
apply seqs_cn; auto; done.
Qed.
